#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#include "htimpl.h"

/* No more than 20 hash tables can be handled simutanuously 
 * If there is already 20 hash tables are handled
 * the ht_create() function will return NULL
 */ 
#define MAXTABLENUM 20
#define MAXOVERFLOW 10000

#ifndef BITSPERBYTE
# define BITSPERBYTE 8
#endif

#ifndef LONGBITS
# define LONGBITS (sizeof (long) * BITSPERBYTE)
#endif

/* local variables */
static struct ht_header *htable[MAXTABLENUM];
static int ht_id = 0;
/* used to allocate space for the hash table */
static int bucket_size;
static void *space_pool[MAXTABLENUM];
/* used to allocate space for overflow bucket */
static void *overflow_pool;
static unsigned long overflow_pt;

/* Prototype for local functions */
static unsigned long next_prime __ARGS ((unsigned long int seed));
static int is_prime __ARGS ((unsigned long int candidate));
static unsigned long compute_hashval __ARGS ((struct ht_header *htab,
                                              const dict_key key, 
                                              size_t keylen));


/*
 * Return value:
 * Not NULL: the handle of the hash table
 * NULL    : fail to create the hash table
 */
dict_h ht_create (
			dict_function oo_compare,
			dict_function ko_compare,
			int flags,   /* Always unique keys */
			int *errnop, /* We don't need to consider this */
			struct ht_args *args
		) 
{
  long init_size;
  int i;

  if (ht_id >= MAXTABLENUM) 
    return NULL;

  htable[ht_id] = (struct ht_header *)malloc(sizeof(struct ht_header));

  /* Set the htable[ht_id]->args */
  /* We need the size to be a prime.  */
  init_size = next_prime (args->ht_table_entries);
  htable[ht_id]->args.ht_table_entries = init_size;
  htable[ht_id]->args.ht_bucket_entries = args->ht_bucket_entries;
  htable[ht_id]->args.ht_objvalue = args->ht_objvalue;
  htable[ht_id]->args.ht_keyvalue = args->ht_keyvalue;

  /* Set the htable[ht_id]->dh */
  htable[ht_id]->dh.oo_comp = oo_compare;
  htable[ht_id]->dh.ko_comp = ko_compare;
  htable[ht_id]->dh.flags = flags;
  htable[ht_id]->dh.errnop = errnop;  

  /* Allocate space for hash table entries */
  htable[ht_id]->table = \
           (tabent_s *)malloc(htable[ht_id]->args.ht_table_entries \
                *sizeof(tabent_s));

  /* 
   * Allocate space for all the buckets 
   * The structure for this space is like the following:
   *
   *   |------------------|
   *   | Bucket 1         | 
   *   |------------------|
   *   | n (void pointer) |
   *   | pointers to key  |
   *   | pointers to data |
   *   |------------------|
   *   | Bucket 2         |
   *   |------------------|
   *   | n (void pointer) |
   *   | pointers to key  |
   *   | pointers to data |
   *   |------------------|
   *   | ........
   */
  bucket_size = 2 * htable[ht_id]->args.ht_bucket_entries*(sizeof(void *));
  space_pool[ht_id] = malloc( \
                       (htable[ht_id]->args.ht_table_entries * \
                         ((sizeof(bucket_s)) + bucket_size)));
  for (i=0; i<htable[ht_id]->args.ht_table_entries; i++) {
    htable[ht_id]->table[i].head_bucket = (bucket_s *)((int *)space_pool[ht_id] + \
                     (i * ((sizeof(bucket_s)) + bucket_size)/sizeof(int *))); 
	htable[ht_id]->table[i].head_bucket->next = NULL;
    htable[ht_id]->table[i].n_free = htable[ht_id]->args.ht_bucket_entries;
  /* 
   * The data stored in bucket is just the pointer to the real data 
   * So the space allocated to the data is just 
   * (ht_bucket_entries * sizeof(void *), the space between this 
   * bucket and the next bucket is reserved for data pointer.
   */
  }

  overflow_pool = malloc(MAXOVERFLOW * ((sizeof(bucket_s)) + bucket_size));
  overflow_pt = 0;
  ht_id++;
  return ((dict_h)htable[ht_id-1]);
}

/*
 * Retrun value:
 * 0: successfully inserted
 * -1 : overflow space overflow, insert failed
 */
int ht_insert	(
			dict_h hh,
			dict_obj obj
		)
{
  unsigned long hval;
  size_t objlen;
  bucket_s *t_buck;
  bucket_s *new_buck;
  int *n_free;
  int bk_size;

/*
 * Get the bucket index the object should be inserted into
 */
  objlen = strlen((char *)obj);
  hval = ((header_s *)hh)->args.ht_objvalue(obj) % \
         ((header_s *)hh)->args.ht_table_entries;

/*
 * Store the pointer of the object in the bucket
 * First: find the last bucket in the bucket link
 *        (means the next pointer of the bucket is NULL)
 * Second: based on the n_free of the table entry we know where
 *        to insert the pointer in. If n_free is equal to 0
 *        we should create a new bucket
 * Third: Insert the pointer in and change the n_free number
 */
  t_buck = ((header_s *)hh)->table[hval].head_bucket;
  while (t_buck->next != NULL)
    t_buck = t_buck->next;

  bk_size = ((header_s *)hh)->args.ht_bucket_entries;
  n_free = &(((header_s *)hh)->table[hval].n_free);
  if (*n_free == 0) {
    if (overflow_pt >= MAXOVERFLOW)
      return -1;
    new_buck = (bucket_s *)((int *)overflow_pool + \
                            (overflow_pt * (sizeof(bucket_s) + bucket_size)/sizeof(int *))); 
    new_buck->next = NULL;
    overflow_pt++;
    t_buck->next = new_buck;
    t_buck = t_buck->next;
    *n_free = bk_size;
  }

  *(BUCKET_OBJECTS(t_buck) + 2*(bk_size - *n_free)*sizeof(void *)) = obj;
  *(BUCKET_OBJECTS(t_buck) + (2*(bk_size - *n_free)+1)*sizeof(void *)) = obj;
  (*n_free)--;
  return 0;
}

dict_obj ht_search(
			dict_h hh,
			dict_key key
		)
{
  unsigned long hval;
  size_t keylen;
  bucket_s *t_buck;
  int finished = 0;
  unsigned n_free, n_search, i;
  dict_obj target;
  int bk_size;

/*
 * Get the bucket index the object should be inserted into
 */
  keylen = strlen((char *)key);
  hval = ((header_s *)hh)->args.ht_keyvalue(key) % \
         ((header_s *)hh)->args.ht_table_entries;

/*
 * Search all the entries in the bucket link
 * Until the key is found or no more entries
 * If found, return the object
 * else, return NULL
 */
  t_buck = ((header_s *)hh)->table[hval].head_bucket;
  n_free = ((header_s *)hh)->table[hval].n_free;
  bk_size = ((header_s *)hh)->args.ht_bucket_entries;

  while (!finished) {
    if (t_buck->next == NULL) {
      n_search = bk_size - n_free;
      finished = 1;
    }
    else {
      n_search = bk_size;
    }
    for (i=0; i<n_search; i++) {
      target = \
        *(BUCKET_OBJECTS(t_buck) + 2*i*sizeof(void *));
      if (((header_s *)hh)->dh.ko_comp(key, target) == 0)
        return *(BUCKET_OBJECTS(t_buck) + (2*i+1)*sizeof(void *));
    }

    t_buck = t_buck->next;
  }
  return NULL;
}

/*
 * Free the space used by the hashtable
 */
void ht_destroy	(
			dict_h hh
		)
{
/*
 * Find the id of the hashtable
 * Free the space by the id, but the overflow space are not freed
 * Not good for this version.
 * This version still need to be improved
 */
  int i;

  for (i=0; i<ht_id; i++) 
    if (htable[i] == hh)
      break;

  if (i == ht_id) {
  /* Invalid hash table handler */
    return;
  }
  else {
    free(htable[i]);
    free(space_pool[i]);
    if (ht_id > 1) {
      htable[i] = htable[ht_id-1];
      space_pool[i] = space_pool[ht_id-1];
      ht_id--;
    }
  }
}

/*
 * Note, this won't be used.
 * Just for mentioing the method
 */
static unsigned long
compute_hashval (htab, key, keylen) 
     struct ht_header *htab;
     const dict_key key;
     size_t keylen;
{
  size_t cnt;
  unsigned long int hval, g;

  /* Compute the hash value for the given string.  The algorithm
     is taken from [Aho,Sethi,Ullman].  */
  cnt = 0;
  hval = keylen;
  while (cnt < keylen)
    {
      hval <<= 4;
      hval += (unsigned long int) *(((char *) key) + cnt++);
      g = hval & ((unsigned long) 0xf << (LONGBITS - 4));
      if (g != 0)
        {
          hval ^= g >> (LONGBITS - 8);
          hval ^= g;
        }
    }
  hval = hval != 0 ? hval : ~((unsigned long) 0);
  return (hval % htab->args.ht_table_entries);
}

static unsigned long
next_prime (seed)
     unsigned long int seed;
{
  /* Make it definitely odd.  */
  seed |= 1;

  while (!is_prime (seed))
    seed += 2;

  return seed;
}


static int
is_prime (candidate)
     unsigned long int candidate;
{
  /* No even number and none less than 10 will be passed here.  */
  unsigned long int divn = 3;
  unsigned long int sq = divn * divn;

  while (sq < candidate && candidate % divn != 0)
    {
      ++divn;
      sq += 4 * divn;
      ++divn;
    }

  return candidate % divn != 0;
}

