/* cpl-exception.h - exceptions thrown while parsing or running CPL scripts. */
/* Copyright 2001 Columbia University.  All rights reserved. */

#include <string>
#include <strstream>

#include <sax/Locator.hpp>
#include <sax/SAXParseException.hpp>
#include <util/XMLException.hpp>

#include "cpl-utils.h"
#include "cpl-parse.h"
#include "cpl-exception.h"

using namespace std;

/* Lots and lots of constructors */

/*
 * CPLException::CPLException(CPLException::error_type)
 * Construct a basic CPL exception from just a type.
 */
CPLException::CPLException(CPLException::error_type t) :
  type(t)
{
  message = error_type_string(t);
} /* CPLException::CPLException(CPLException::error_type) */


/*
 * CPLException::CPLException(CPLException::error_type,
 *                            const Locator* const)
 * Construct a CPL exception given a type and an XML locator.
 */
CPLException::CPLException(CPLException::error_type t,
                           const Locator* const l) :
  type(t)
{
  message = error_type_string(t) + " at " + locator_string(l);
} /* CPLException::CPLException(CPLException::error_type, 
   *                            const Locator* const) */


/*
 * CPLException::CPLException(CPLException::error_type,
 *                            const char* const)
 * Construct a CPL exception given a type and a caller-supplied message.
 */
CPLException::CPLException(CPLException::error_type t,
                           const char* const m) :
  type(t)
{
  message = error_type_string(t) + ": " + m;
} /* CPLException::CPLException(CPLException::error_type,
   *                            const char* const) */


/*
 * CPLException::CPLException(CPLException::error_type, const Locator* const, 
 *                            const char* const)
 * Construct a CPL exception given a type, an XML locator, and a
 * caller-supplied message.
 */
CPLException::CPLException(CPLException::error_type t, const Locator* const l, 
                           const char* const m) :
  type(t)
{
  message = error_type_string(t) + ": " + m + " at " + locator_string(l);
} /* CPLException::CPLException(CPLException::error_type,
   *                            const Locator* const, const char* const) */


/*
 * CPLException::CPLException(CPLException::error_type, const Locator* const,
 *                            const qualified_name&)
 * Construct a CPL exception given a type, an XML locator, and a qualified
 * XML name giving the invalid type.
 */
CPLException::CPLException(CPLException::error_type t, const Locator* const l,
                           const qualified_name& n) :
  type(t)
{
  message = error_type_string(t) + " " + qualified_name_string(n) +
    " at " + locator_string(l);
} /* CPLException::CPLException(CPLException::error_type, const Locator* const,
   *                            const qualified_name&) */


/*
 * CPLException::CPLException(CPLException::error_type, const Locator* const,
 *                            const qualified_name&, const qualified_name&)
 * Construct a CPL exception, given a type, an XML locator, and two
 * qualified XML names given the invalid type and its context.
 */
CPLException::CPLException(CPLException::error_type t, const Locator* const l,
                           const qualified_name& n, const qualified_name& n2) :
  type(t)
{
  message = error_type_string(t) + " " + qualified_name_string(n) +
    " in " + qualified_name_string(n2) + " at " + locator_string(l);
} /* CPLException::CPLException(CPLException::error_type, const Locator* const,
   *                            const qualified_name&, const qualified_name&) */


/*
 * CPLException::CPLException(CPLException::error_type, const Locator* const,
 *                            const string&, const qualified_name&, 
 *                            const qualified_name&)
 * Construct a CPL exception, given a type, an XML locator,
 * an attribute value, and two qualified XML names giving the attribute
 * and its enclosing node.
 */
CPLException::CPLException(CPLException::error_type t, const Locator* const l,
                           const string& v, const qualified_name& n,
                           const qualified_name& n2) :
  type(t)
{
  message = error_type_string(t) + " \'" + v + "\' for " +
    qualified_name_string(n) + " in " + qualified_name_string(n2) +
    " at " + locator_string(l);
} /* CPLException::CPLException(CPLException::error_type, const Locator* const,
   *                            const string&, const qualified_name&, 
   *                            const qualified_name&) */



/*
 * CPLException::CPLException(CPLException::error_type, const Locator* const,
 *                            const char *const, const qualified_name&)
 * Construct a CPL exception given a type, an XML locator, a message,
 * and a qualified XML name giving the error's context.
 */
CPLException::CPLException(CPLException::error_type t, const Locator* const l,
                           const char* const m, const qualified_name& n) :
  type(t)
{
  message = error_type_string(t) + ": " + m + " in " +
    qualified_name_string(n) + " at " + locator_string(l);
} /* CPLException::CPLException(CPLException::error_type, const Locator* const,
   *                            const char* const, const qualified_name&) */


/*
 * CPLException(CPLException::error_type, const string&,
 *              const cpl_parser*)
 * Construct a CPL exception given a type, an unqualified XML name giving
 * the error's context, and a parser with current node and locator
 * information.
 */
CPLException::CPLException(CPLException::error_type t, const string& n,
                           const cpl_parser* p) :
  type(t)
{
  qualified_name qual = qualified_name(string(), n);
  const element_handler* current_handler = 
    p->getStackCurrent()->getHandler();
  qualified_name context = qualified_name(current_handler->getSpace(),
                                          current_handler->getName());

  message = error_type_string(t) + " " + qualified_name_string(qual) + 
    " in " + qualified_name_string(context) + " at " +
    locator_string(p->getLocator());
} /* CPLException(CPLException::error_type, const string&,
   *              const cpl_parser*) */


/*
 * CPLException::CPLException(CPLException::error_type, const string&,
 *                            const string&, const cpl_parser*)
 * Construct a CPL exception given an attribute value, an unqualified XML name
 * giving the attribute, and a CPL parser with current node and locator
 * information.
 */
CPLException::CPLException(CPLException::error_type t, const string& v,
                           const string& n, const cpl_parser* p)
{
  qualified_name qual = qualified_name(string(), n);
  const element_handler* current_handler = 
    p->getStackCurrent()->getHandler();
  qualified_name context = qualified_name(current_handler->getSpace(),
                                          current_handler->getName());

  message = error_type_string(t) + " \'" + v + "\' for " +
    qualified_name_string(qual) + " in " + qualified_name_string(context) +
    " at " + locator_string(p->getLocator());
} /* CPLException::CPLException(CPLException::error_type, const string&,
   *                            const string&, const cpl_parser*) */

  
/*
 * CPLException::CPLException(CPLException::error_type, const string&,
 *                            const string&, const char * const,
 *                            const cpl_parser*)
 * Construct a CPL exception given an attribute value, an unqualified XML name
 * giving the attribute, a caller-supplied message, and a CPL parser with
 * current node and locator information.
 */
CPLException::CPLException(CPLException::error_type t, const string& v,
                           const string& n, const char* const m,
                           const cpl_parser* p)
{
  qualified_name qual = qualified_name(string(), n);
  const element_handler* current_handler = 
    p->getStackCurrent()->getHandler();
  qualified_name context = qualified_name(current_handler->getSpace(),
                                          current_handler->getName());

  message = error_type_string(t) + " \'" + v + "\' for " +
    qualified_name_string(qual) + " in " + qualified_name_string(context) +
    ": " + m + " at " + locator_string(p->getLocator());
} /* CPLException::CPLException(CPLException::error_type, const string&,
   *                            const string&, const char * const,
   *                            const cpl_parser*) */
  
/*
 * CPLException(CPLException::error_type, const string&,
 *              const Locator* const, const element_handler*)
 * Construct a CPL exception given a type, an unqualified XML name giving
 * the error's context, an XML locator, and an element_handler with the
 * relevant node's namespace and type.
 */
CPLException::CPLException(CPLException::error_type t, const string& n,
                           const Locator* const l, const element_handler* e) :
  type(t)
{
  qualified_name qual = qualified_name(string(), n);
  qualified_name context = qualified_name(e->getSpace(),
                                          e->getName());

  message = error_type_string(t) + " " + qualified_name_string(qual) + 
    " in " + qualified_name_string(context) + " at " +
    locator_string(l);
} /* CPLException(CPLException::error_type, const string&,
   *              const Locator* const, const element_handler*) */


/*
 * CPLException::CPLException(CPLException::error_type, const string&,
 *                            const string&, const Locator* const,
 *                            const element_handler*)
 * Construct a CPL exception given an attribute value, an unqualified XML name
 * giving the attribute, an XML locator, and an element_handler with the
 * relevant node's namepace and type.
 */
CPLException::CPLException(CPLException::error_type t, const string& v,
                           const string& n, const Locator* const l,
                           const element_handler *e)
{
  qualified_name qual = qualified_name(string(), n);
  qualified_name context = qualified_name(e->getSpace(),
                                          e->getName());

  message = error_type_string(t) + " \'" + v + "\' for " +
    qualified_name_string(qual) + " in " + qualified_name_string(context) +
    " at " + locator_string(l);
} /* CPLException::CPLException(CPLException::error_type, const string&,
   *                            const string&, const Locator* const,
   *                            const element_handler*) */

  
/*
 * CPLException::CPLException(const SAXParseException)
 * Construct a CPL exception from a SAXParseException.  This gives
 * us a message and a locator.
 */
CPLException::CPLException(const SAXParseException& e) :
  type(CPLException::ParserException)
{
  ostrstream out;

  out << error_type_string(CPLException::ParserException) << ": ";
  out << cpl_utils::transcode(e.getMessage());
  out << " at ";
    
  out << cpl_utils::transcode(e.getSystemId()) << ":";
  out << e.getLineNumber() << ":" << e.getColumnNumber() << ends;
    
  message = string(out.str());
  out.rdbuf()->freeze(0);
} /* CPLException::CPLException(const SAXParseException) */


/*
 * CPLException::CPLException(const SAXException)
 * Construct a CPL exception from a generic SAXException.
 */
CPLException::CPLException(const SAXException& e) :
  type(CPLException::ParserException)
{
  message = error_type_string(CPLException::ParserException) + ": " +
    cpl_utils::transcode(e.getMessage());
} /* CPLException::CPLException(const SAXException) */


/*
 * CPLException::CPLException(const XMLException)
 * Construct a CPL exception from a generic XMLException.
 */
CPLException::CPLException(const XMLException& e) :
  type(CPLException::ParserException)
{
  ostrstream out;

  out << error_type_string(CPLException::ParserException) << ": ";
  out << cpl_utils::transcode(e.getMessage());
  if (e.getSrcFile()[0] != '\0' || e.getSrcLine() != 0) {
    out << " at " << e.getSrcFile() << ":" << e.getSrcLine();
  }
  out << ends;
    
  message = string(out.str());
  out.rdbuf()->freeze(0);
} /* CPLException::CPLException(const XMLException) */



/*
 * CPLException::error_type_string(CPLException::error_type)
 * Return a string corresponding to the CPL error class.
 * (Static function.)
 */
string CPLException::error_type_string(CPLException::error_type error)
{
  switch(error) {
  case InternalError:
    return "Internal error";
  case DuplicateAttr:
    return "Duplicate attribute";
  case DuplicateChild:
    return "Illegally duplicated subtag"; 
  case DuplicateSubaction:
    return "Duplication subaction ID";
  case MissingAttr:
    return "Missing attribute";
  case MissingChild:
    return "Missing subtag";
  case UnknownAttrVal:
    return "Unsupported attribute value";
  case MalformedAttr:
    return "Mal-formed attribute value";
  case InvalidTopTag:
    return "Invalid top-level tag";
  case InvalidChild:
    return "Invalid subtag";
  case InvalidAttr:
    return "Invalid attribute";
  case TagMismatch:
    return "Open/close tag mismatch";
  case TooDeep:
    return "Too many levels of nesting";
  case TooManySubactions:
    return "Too many subactions defined";
  case TooManyTimezones:
    return "Too many timezones defined";
  case TooManyTimeRanges:
    return "Too many time ranges in argument";
  case UnknownSub:
    return "Unknown subaction ID";
  case UnknownZone:
    return "Unknown timezone ID";
  case IllegalComparitor:
    return "Illegal comparitor in context";
  case ParserException:
    return "XML parsing exception";
  case UnknownError:
  default:
    return "Unknown error";
  }
} /* CPLException::error_type_string(CPLException::error_type) */


/*
 * CPLException::locator_string(Locator)
 * Given a SAX locator, return a string describing the location it
 * indicates.
 * (Static function.)
 */
string CPLException::locator_string(const Locator* const l)
{
  ostrstream out;

  out << cpl_utils::transcode(l->getSystemId()) << ":";
  out << l->getLineNumber() << ":" << l->getColumnNumber() << ends;

  string ret(out.str());
  out.rdbuf()->freeze(0);
  return ret;
} /* CPLException::locator_string(Locator) */


/*
 * CPLException::qualified_name_string(const qualified_name&)
 * Given a qualified name, return a string describing the qualified name --
 * XML name and optional namespace -- it describes.
 */
string CPLException::qualified_name_string(const qualified_name& n)
{
  string ret = "\'";
  if (n.first.length() != 0) {
    ret += "[" + n.first + "]:";
  }
  ret += n.second;
  ret += "\'";

  return ret;
} /* CPLException::qualified_name_string(const qualified_name&) */
