/* cpl-exception.h - exceptions thrown while parsing or running CPL scripts. */
/* Copyright 2001 Columbia University.  All rights reserved. */

#ifndef CPL_EXCEPTION_H_
#define CPL_EXCEPTION_H_

#include <exception>
#include <stdexcept>
#include <string>
#include <utility> /* pair<> */

/* "Anonymous" classes that can be used to create a CPLException. */
class Locator;
class XMLException;
class SAXException;
class SAXParseException;
class cpl_parser;
class element_handler;

/* N.B.: This must be the same as the definition in cpl-parse.h */
typedef std::pair<std::string, std::string>     qualified_name;

/* Errors thrown by CPL */
class CPLException : public std::exception {
public:
  typedef enum {
    UnknownError,
    InternalError,
    DuplicateAttr,
    DuplicateChild,
    DuplicateSubaction,
    MissingAttr,
    MissingChild,
    UnknownAttrVal,
    MalformedAttr,
    InvalidTopTag,
    InvalidChild,
    InvalidAttr,
    TagMismatch,
    TooDeep,
    TooManySubactions,
    TooManyTimezones,
    TooManyTimeRanges,
    UnknownSub,
    UnknownZone,
    IllegalComparitor,
    ParserException
  } error_type;

private:
  const error_type type;
  std::string message;

protected:
  static std::string error_type_string(CPLException::error_type);
  static std::string locator_string(const Locator* const);
  static std::string qualified_name_string(const qualified_name&);

public:
  /* Lots and lots of constructors */
  CPLException(CPLException::error_type);
  CPLException(CPLException::error_type, const Locator* const);
  CPLException(CPLException::error_type, const char* const);
  CPLException(CPLException::error_type, const Locator* const, 
               const char* const);
  CPLException(CPLException::error_type, const Locator* const,
               const qualified_name&);
  CPLException(CPLException::error_type, const Locator* const,
               const qualified_name&, const qualified_name&);
  CPLException(CPLException::error_type, const Locator* const,
               const std::string&,
               const qualified_name&, const qualified_name&);
  CPLException(CPLException::error_type, const Locator* const,
               const char* const, const qualified_name&);

  CPLException(CPLException::error_type, const std::string&,
               const cpl_parser*);
  CPLException(CPLException::error_type, const std::string&,
               const std::string&, const cpl_parser*);
  CPLException(CPLException::error_type, const std::string&,
               const std::string&, const char* const,
               const cpl_parser*);
  
  CPLException(CPLException::error_type, const std::string&,
               const Locator* const, const element_handler*);
  CPLException(CPLException::error_type, const std::string&,
               const std::string&, const Locator* const,
               const element_handler*);
  
  CPLException(const SAXParseException&);
  CPLException(const SAXException&);
  CPLException(const XMLException&);

  virtual CPLException::error_type get_type() const { return type; }
  virtual const char* what() const throw() { return message.c_str(); }

  virtual ~CPLException() throw() { }
};

#endif /* CPL_EXCEPTION_H_ */
