/* cpl-nodes-action.cpp - handlers and definitions for CPL action tags. */
/* Copyright 2001 Columbia University.  All rights reserved. */

#include "cpl-exception.h"

#include "cpl-templatefns.h"

#include "cpl-nodes.h"
#include "cpl-nodes-internal.h"
#include "cpl-nodes-action.h"

using namespace std;

/* Code for <redirect>. */

/*
 * redirect_node::redirect_node()
 * Initialize a new <redirect> node.  Set permanent to false.
 */
redirect_node::redirect_node() :
  cpl_node()
{
  permanent = false;
} /* redirect_node::redirect_node() */


/*
 * redirect_handler::allocate_node() const
 * Allocate and return a new redirect_node.
 */
cpl_node_ptr redirect_handler::allocate_node() const
{
  return cpl_node_ptr(new redirect_node());
} /* redirect_handler::allocate_node() const */


/*
 * redirect_handler::redirect_handler()
 * Fill in the subnode and attribute maps for a redirect node.
 * A redirect has no subnodes; leave these blank.
 */
redirect_handler::redirect_handler() :
  cpl_node_handler(NULL, "redirect")
{
  insert_att_map("permanent", &redirect_handler::set_permanent);
} /* redirect_handler::redirect_handler() */


/*
 * redirect_handler::set_permanent(cpl_parser*, const string&)
 * Handler function for the permanent attribute of a <redirect>.
 * (Static function)
 */
void redirect_handler::set_permanent(cpl_parser* parser, const string& value)
{
  redirect_node* current_node =
    dynamic_cast_cpl_node_ptr<redirect_node*>(parser->getCurrentNode(),
                                              "redirect_handler::set_permanent");

  current_node->permanent = 
    get_map_value<bool>(yes_no_map, value, parser, "permanent");
} /* redirect_handler::set_permanent(cpl_parser*, const string&) */


/* Code for <proxy>. */

/*
 * proxy_handler::allocate_node()
 * Allocate and return a new proxy_node.
 */
cpl_node_ptr proxy_handler::allocate_node() const
{
  return cpl_node_ptr(new proxy_node());
} /* proxy_handler::allocate_node() */


/*
 * proxy_handler::proxy_handler()
 * Fill in the subnode and attribute maps for a proxy node.
 */
proxy_handler::proxy_handler() :
  cpl_node_handler(NULL, "proxy")
{
  INSERT_SIMPLE_OUTPUT_SUBNODE(proxy, busy);
  INSERT_SIMPLE_OUTPUT_SUBNODE(proxy, noanswer);
  INSERT_SIMPLE_OUTPUT_SUBNODE(proxy, redirection);
  INSERT_SIMPLE_OUTPUT_SUBNODE(proxy, failure);
  INSERT_SIMPLE_OUTPUT_SUBNODE(proxy, default);

  insert_att_map("timeout", &proxy_handler::set_timeout);
  insert_att_map("recurse", &proxy_handler::set_recurse);
  insert_att_map("ordering", &proxy_handler::set_ordering);
} /* proxy_handler::proxy_handler() */


/*
 * proxy_handler::set_timeout(cpl_parser*, const string&)
 * Set the timeout attribute of a <proxy> node.
 * (Static function)
 */
void proxy_handler::set_timeout(cpl_parser* parser, const string& value)
{
  proxy_node* current_node =
    dynamic_cast_cpl_node_ptr<proxy_node*>(parser->getCurrentNode(),
                                              "proxy_handler::set_timeout");

  int timeout = read_string<int>(value, parser, "timeout");
  if (timeout < 0) {
    throw CPLException(CPLException::MalformedAttr, value, "timeout",
                       parser);
  }

  current_node->timeout = timeout;
} /* proxy_handler::set_timeout(cpl_parser*, const string&) */


/*
 * proxy_handler::set_recurse(cpl_parser*, const string&)
 * Handler function for the recurse attribute of a <proxy>.
 * (Static function)
 */
void proxy_handler::set_recurse(cpl_parser* parser, const string& value)
{
  proxy_node* current_node =
    dynamic_cast_cpl_node_ptr<proxy_node*>(parser->getCurrentNode(),
                                              "proxy_handler::set_recurse");

  current_node->recurse = 
    get_map_value<bool>(yes_no_map, value, parser, "recurse");
} /* proxy_handler::set_recurse(cpl_parser*, const string&) */


/*
 * proxy_handler::set_ordering(cpl_parser*, const string&)
 * Handler function for the ordering attribute of a <proxy>.
 * (Static function)
 */
void proxy_handler::set_ordering(cpl_parser* parser, const string& value)
{
  proxy_node* current_node =
    dynamic_cast_cpl_node_ptr<proxy_node*>(parser->getCurrentNode(),
                                              "proxy_handler::set_ordering");

  current_node->ordering = 
    get_map_value<proxy_node::order>(proxy_order_map, value, parser,
                                     "ordering");
} /* proxy_handler::set_ordering(cpl_parser*, const string&) */


/* The map for proxy orderings. */
std::map<std::string, proxy_node::order> proxy_order_map;

/*
 * init_proxy_order_map()
 * Initialize a map for proxy_node::order values.
 */
void init_proxy_order_map()
{
  proxy_order_map["parallel"] = proxy_node::parallel;
  proxy_order_map["sequential"] = proxy_node::sequential;
  proxy_order_map["first-only"] = proxy_node::first_only;
} /* init_proxy_order_map() */


/*
 * proxy_handler::end(cpl_parser *) const
 * End a proxy node.  If we have a noanswer or default output, and timeout
 * is still -1, set timeout to the default noanswer timeout (20).
 */
void proxy_handler::end(cpl_parser *parser) const
{
  proxy_node* current_node =
    dynamic_cast_cpl_node_ptr<proxy_node*>(parser->getCurrentNode(),
                                            "proxy_handler::end");

  if (current_node->timeout == -1
      && (current_node->noanswer.get() != NULL
          || current_node->default_output.get() != NULL)) {
    current_node->timeout = 20;
  }
} /* proxy_handler::end(cpl_parser *) */


/*
 * proxy_node::proxy_node()
 * Construct a new proxy node.  Set its defaults appropriately.
 * The timeout of "-1" means unlimited, unless the noanswer output is set.
 * (This case gets fixed up in proxy_handler::end().)
 */
proxy_node::proxy_node()
{
  timeout = -1;
  recurse = true;
  ordering = parallel;
} /* proxy_node::proxy_node() */



DEFINE_SIMPLE_OUTPUT_HANDLER_END(proxy, busy)
DEFINE_SIMPLE_OUTPUT_HANDLER_END(proxy, noanswer)
DEFINE_SIMPLE_OUTPUT_HANDLER_END(proxy, redirection)
DEFINE_SIMPLE_OUTPUT_HANDLER_END(proxy, failure)


/*
 * proxy_default_handler::end(cpl_parser*) const
 * End a <default> output of a <proxy> node.
 * We can't use DEFINE_SIMPLE_OUTPUT_HANDLER_END for this, because
 * the proxy_node data member is named "default_output".  ("default" is
 * a C++ reserved word.)
 */
void proxy_default_handler::end(cpl_parser *parser) const
{
  proxy_node* parent_node =
    dynamic_cast_cpl_node_ptr<proxy_node*>(parser->getParentNode(),
                                           "proxy_default_handler::end");
  if (parent_node->default_output.get() != NULL) {
    throw CPLException(CPLException::DuplicateChild, "default",
                       parser->getLocator(),
                       parser->getStackParent()->getHandler());
  }
  parent_node->default_output = parser->getCurrentOutput();
} /* proxy_default_handler::end(cpl_parser*) const */


/* Code for <reject>. */

/*
 * reject_node::reject_node()
 * Initialize a new <reject> node.  The return status is unknown at this
 * point.
 */
reject_node::reject_node() : 
  cpl_node()
{
  status_type = unknown;
} /* reject_node::reject_node() */


/*
 * reject_handler::allocate_node() const
 * Allocate and return a new reject_node.
 */
cpl_node_ptr reject_handler::allocate_node() const
{
  return cpl_node_ptr(new reject_node());
} /* reject_handler::allocate_node() const */


/*
 * reject_handler::reject_handler()
 * Fill in attribute maps for a <reject> node.
 * <reject> has no subnodes; leave these empty.
 */
reject_handler::reject_handler() : 
  cpl_node_handler(NULL, "reject")
{
  insert_att_map("status", &reject_handler::set_status);
  insert_att_map("reason", &reject_handler::set_reason);
} /* reject_handler::reject_handler() */


/*
 * reject_handler::set_status(cpl_parser* parser, const string& value)
 * Handler function for the status attribute of a <reject>.
 * (Static function).
 */
void reject_handler::set_status(cpl_parser* parser, const string& value)
{
  reject_node* current_node =
    dynamic_cast_cpl_node_ptr<reject_node*>(parser->getCurrentNode(),
                                            "reject_handler::set_status");

  std::map<std::string, reject_node::symbolic_reject_status>::const_iterator
    it;

  it = reject_status_map.find(value);

  if (it != reject_status_map.end()) {
    /* It was a symbolic value. */
    current_node->status_type = reject_node::symbolic;
    current_node->symbolic_status = it->second;
    return;
  }

  /* Try to treat it as a numeric value. */
  int status = read_string<int>(value, parser, "status");
  if (status < 400 || status >= 700) {
    throw CPLException(CPLException::MalformedAttr, value, "status", parser);
  }
  
  current_node->status_type = reject_node::numeric;
  current_node->numeric_status = status;
} /* reject_handler::set_status(cpl_parser* parser, const string& value) */


/* The map for reject statuses. */
std::map<std::string, reject_node::symbolic_reject_status> reject_status_map;

/*
 * init_reject_status_map()
 * Initialize a map for reject_node::symbolic_reject_status values.
 */
void init_reject_status_map()
{
  reject_status_map["busy"] = reject_node::busy;
  reject_status_map["notfound"] = reject_node::notfound;
  reject_status_map["reject"] = reject_node::reject;
  reject_status_map["error"] = reject_node::error;
} /* init_reject_status_map() */


/*
 * reject_handler::set_reason(cpl_parser*, const string&)
 * Handler function for the reason attribute of a <reject>.
 * (Static function).
 */
void reject_handler::set_reason(cpl_parser* parser, const string& value)
{
  reject_node* current_node =
    dynamic_cast_cpl_node_ptr<reject_node*>(parser->getCurrentNode(),
                                            "reject_handler::set_reason");
  current_node->reason = value;
} /* reject_handler::set_reason(cpl_parser *, const string&) */


/*
 * reject_handler::end(cpl_parser*)
 * End a reject node.  Make sure that a status has been set.
 */
void reject_handler::end(cpl_parser* parser) const
{
  reject_node* current_node =
    dynamic_cast_cpl_node_ptr<reject_node*>(parser->getCurrentNode(),
                                            "reject_handler::end");
  
  if (current_node->status_type == reject_node::unknown) {
    throw CPLException(CPLException::MissingAttr, "status", parser);
  }
} /* reject_handler::end(cpl_parser*) */
 

/*
 * Singleton handlers.
 */

redirect_handler the_redirect_handler;

proxy_handler    the_proxy_handler;
DEFINE_SINGLETON_SIMPLE_OUTPUT_HANDLER(proxy, busy);
DEFINE_SINGLETON_SIMPLE_OUTPUT_HANDLER(proxy, noanswer);
DEFINE_SINGLETON_SIMPLE_OUTPUT_HANDLER(proxy, redirection);
DEFINE_SINGLETON_SIMPLE_OUTPUT_HANDLER(proxy, failure);
DEFINE_SINGLETON_SIMPLE_OUTPUT_HANDLER(proxy, default);


reject_handler    the_reject_handler;
