/* cpl-nodes-internal.h - internal properties of all CPL nodes. */
/* Copyright 2001 Columbia University.  All rights reserved. */

#ifndef CPL_NODES_INTERNAL_H_
#define CPL_NODES_INTERNAL_H_

#include "cpl-nodes.h"

/*
 * Base class for CPL node and output handlers.
 * This exists only so we can have a generic function which populates
 * the child map with all CPL nodes.
 */
class cpl_base_handler : public element_handler {
protected:
  void output_all_cpl_nodes();

public:
  cpl_base_handler(const char *const space, const char *const name) :
    element_handler(space, name) { }

  virtual ~cpl_base_handler() { }
};


/*
 * Base class for generic CPL node handlers.
 */
class cpl_node_handler : public cpl_base_handler {
protected:
  virtual cpl_node_ptr allocate_node() const = 0;

public:
  cpl_node_handler(const char *const space, const char *const name) :
    cpl_base_handler(space, name) { }

  virtual ~cpl_node_handler() { };

  virtual void start(cpl_parser*) const;
};


/*
 * Base class for generic CPL output handlers.
 * Next node is any CPL node.
 */
class cpl_output_handler : public cpl_base_handler {
public:
  cpl_output_handler(const char *const space, const char *const name);

  virtual ~cpl_output_handler() { };
};

/* Generic value maps. */
extern std::map<std::string, bool> yes_no_map;
extern void init_yes_no_map();


/*
 * Support macro for a simple CPL output handler (one which has no
 * attributes).
 * Declare the output handler class, and its singleton handler.
 * This is the output "OUTPUT" of the node "NODE".
 * You must write 'void <NODE>_<OUTPUT>_handler::end(cpl_parser*) const',
 * which will normally insert the current_output into the appropriate private
 * member of the parent.  (You can use DEFINE_SIMPLE_OUTPUT_HANDLER_END for
 * this.)
 */
#define DECLARE_SIMPLE_OUTPUT_HANDLER(NODE, OUTPUT) \
class NODE ## _ ## OUTPUT ## _handler : public cpl_output_handler { \
public: \
  NODE ## _ ## OUTPUT ## _handler() : cpl_output_handler(NULL, #OUTPUT) \
    { } \
  virtual void end(cpl_parser*) const; \
};

/*
 * Support macro for a simple CPL output handler (one which has no attributes).
 * Define the end function for the output "OUTPUT" of the node "NODE".
 * This assumes that the parent class has a cpl_node_ptr data member named
 * "OUTPUT", and that this class has access to it.  (You can use the
 * FRIEND_SIMPLE_OUTPUT_HANDLER macro below to accomplish this.)
 * This also assumes that the handler does not need to do any special
 * error checking.
 */
#define DEFINE_SIMPLE_OUTPUT_HANDLER_END(NODE, OUTPUT) \
void NODE ## _ ## OUTPUT ## _handler::end(cpl_parser *parser) const \
{ \
  NODE ## _node* parent_node = \
    dynamic_cast_cpl_node_ptr<NODE ## _node*>(parser->getParentNode(), \
                              #NODE "_" #OUTPUT "_handler::end"); \
  if (parent_node->OUTPUT.get() != NULL) { \
    throw CPLException(CPLException::DuplicateChild, #OUTPUT, \
                       parser->getLocator(), \
                       parser->getStackParent()->getHandler()); \
  } \
  parent_node->OUTPUT = parser->getCurrentOutput(); \
}


/*
 * Declare a simple output handler class to be a friend of a generic class.
 */
#define FRIEND_SIMPLE_OUTPUT_HANDLER(NODE, OUTPUT) \
  friend class NODE ## _ ## OUTPUT ## _handler

/*
 * Declare an extern singleton simple output handler.
 */
#define DECLARE_EXTERN_SINGLETON_SIMPLE_OUTPUT_HANDLER(NODE, OUTPUT) \
extern NODE ## _ ## OUTPUT ## _handler the_ ## NODE ## _ ## OUTPUT ## _handler

/*
 * Define a singleton simple output handler.
 */
#define DEFINE_SINGLETON_SIMPLE_OUTPUT_HANDLER(NODE, OUTPUT) \
NODE ## _ ## OUTPUT ## _handler the_ ## NODE ## _ ## OUTPUT ## _handler


/*
 * Insert a pointer to a singleton simple output handler into an node's
 * subnode map.
 */
#define INSERT_SIMPLE_OUTPUT_SUBNODE(NODE, OUTPUT) \
  insert_subnode_map(#OUTPUT, &the_ ## NODE ## _ ## OUTPUT ##_handler)



#endif /* CPL_NODES_INTERNAL_H_ */
