/* cpl-nodes-location.cpp - handlers and definitions for CPL location tags. */
/* Copyright 2001 Columbia University.  All rights reserved. */

#include <strstream>

#include "cpl-exception.h"

#include "cpl-nodes.h"
#include "cpl-nodes-location.h"
#include "cpl-templatefns.h"

using namespace std;

/* <location> node */

/*
 * location_handler::allocate_node()
 * Allocate and return a new location_node.
 */
cpl_node_ptr location_handler::allocate_node() const
{
  return cpl_node_ptr(new location_node());
} /* location_handler::allocate_node() */


/*
 * location_handler::location_handler()
 * Fill in the subnode and attribute maps for a location node.
 * Outputs all CPL nodes.  Takes a 'url' attribute.
 */
location_handler::location_handler() :
  cpl_node_handler(NULL, "location")
{
  output_all_cpl_nodes();

  insert_att_map("url", &location_handler::set_url);
  insert_att_map("priority", &location_handler::set_priority);
  insert_att_map("clear", &location_handler::set_clear);
} /* location_handler::location_handler() */


/*
 * location_handler::set_url(cpl_parser*, const string&)
 * Handler function for the url attribute of a <location>.
 * (Static function)
 * TODO: represent URLs as url_t's, not strings.
 */
void location_handler::set_url(cpl_parser* parser, const string& value)
{
  location_node* current_node =
    dynamic_cast_cpl_node_ptr<location_node*>(parser->getCurrentNode(),
                                              "location_handler::set_url");

  current_node->url = value;
} /* location_handler::set_url(cpl_parser*, const std::string&) */


/*
 * location_handler::set_priority(cpl_parser*, const string&)
 * Handler function for the priority attribute of a <location>.
 * (Static function)
 */
void location_handler::set_priority(cpl_parser* parser, const string& value)
{
  location_node* current_node =
    dynamic_cast_cpl_node_ptr<location_node*>(parser->getCurrentNode(),
                                              "location_handler::set_priority");
  float priority = read_string<float>(value, parser, "priority");

  if (priority < 0.0 || priority > 1.0) {
    throw CPLException(CPLException::MalformedAttr, value, "priority",
                       parser);
  }

  current_node->priority = priority;
} /* location_handler::set_priority(cpl_parser*, const string&) */


/*
 * location_handler::set_clear(cpl_parser*, const string&)
 * Handler function for the clear attribute of a <location>.
 * (Static function)
 */
void location_handler::set_clear(cpl_parser* parser, const string& value)
{
  location_node* current_node =
    dynamic_cast_cpl_node_ptr<location_node*>(parser->getCurrentNode(),
                                              "location_handler::set_clear");

  current_node->clear = 
    get_map_value<bool>(yes_no_map, value, parser, "clear");
} /* location_handler::set_clear(cpl_parser*, const string&) */


/*
 * location_handler::end(cpl_parser *) const
 * End a location node.  Check to make sure url is set, then insert the
 * current output into the node's output. 
 */
void location_handler::end(cpl_parser *parser) const
{
  location_node* current_node =
    dynamic_cast_cpl_node_ptr<location_node*>(parser->getCurrentNode(),
                                              "location_handler::end");

  if (current_node->url.length() == 0) {
    throw CPLException(CPLException::MissingAttr, "url", parser);
  }

  current_node->output = parser->getCurrentOutput();
} /* location_handler::end(cpl_parser *) */


/*
 * location_node::location_node()
 * Initialize a location node.  Set the priority to 1.0.
 */
location_node::location_node() : 
  cpl_node()
{
  priority = 1.0;
  clear = false;
} /* location_node::location_node() */


/*
 * location_node::run()
 * Run a location node.  Insert the specified location in the location list,
 * and set the next node result to the single output.
 * XXX: do the location list.
 */
cpl_run_result location_node::run()
{
  cpl_run_result res;

  res.result_type = cpl_run_result::result_next_node;
  res.next_node = output;
  
  return res;
} /* location_node::run() */


/* <lookup> node */

/*
 * lookup_handler::allocate_node()
 * Allocate and return a new lookup_node.
 */
cpl_node_ptr lookup_handler::allocate_node() const
{
  return cpl_node_ptr(new lookup_node());
} /* lookup_handler::allocate_node() */


/*
 * lookup_handler::lookup_handler()
 * Fill in the subnode and attribute maps for a lookup node.
 */
lookup_handler::lookup_handler() :
  cpl_node_handler(NULL, "lookup")
{
  INSERT_SIMPLE_OUTPUT_SUBNODE(lookup, success);
  INSERT_SIMPLE_OUTPUT_SUBNODE(lookup, notfound);
  INSERT_SIMPLE_OUTPUT_SUBNODE(lookup, failure);

  insert_att_map("source", &lookup_handler::set_source);
  insert_att_map("timeout", &lookup_handler::set_timeout);
  insert_att_map("use", &lookup_handler::set_use);
  insert_att_map("ignore", &lookup_handler::set_ignore);
  insert_att_map("clear", &lookup_handler::set_clear);
} /* lookup_handler::lookup_handler() */


/*
 * lookup_handler::set_source(cpl_parser*, const string&)
 * Set the source attribute of a <lookup> node.
 * (Static function)
 * XXX TODO: verify that the source is either a known URL source format,
 * or 'registration'.  Convert to a url_t or enum.
 */
void lookup_handler::set_source(cpl_parser* parser, const string& value)
{
  lookup_node* current_node =
    dynamic_cast_cpl_node_ptr<lookup_node*>(parser->getCurrentNode(),
                                              "lookup_handler::set_source");

  current_node->source = value;
} /* lookup_handler::set_source(cpl_parser*, const string&) */


/*
 * lookup_handler::set_timeout(cpl_parser*, const string&)
 * Set the timeout attribute of a <lookup> node.
 * (Static function)
 */
void lookup_handler::set_timeout(cpl_parser* parser, const string& value)
{
  lookup_node* current_node =
    dynamic_cast_cpl_node_ptr<lookup_node*>(parser->getCurrentNode(),
                                              "lookup_handler::set_timeout");

  int timeout = read_string<int>(value, parser, "timeout");
  if (timeout < 0) {
    throw CPLException(CPLException::MalformedAttr, value, "timeout",
                       parser);
  }

  current_node->timeout = timeout;
} /* lookup_handler::set_timeout(cpl_parser*, const string&) */


/*
 * lookup_handler::set_use(cpl_parser*, const string&)
 * Set the use attribute of a <lookup> node.
 * (Static function)
 * XXX: change to do a proper set
 */
void lookup_handler::set_use(cpl_parser* parser, const string& value)
{
  lookup_node* current_node =
    dynamic_cast_cpl_node_ptr<lookup_node*>(parser->getCurrentNode(),
                                              "lookup_handler::set_use");

  current_node->use = value;
} /* lookup_handler::set_use(cpl_parser*, const string&) */


/*
 * lookup_handler::set_ignore(cpl_parser*, const string&)
 * Set the ignore attribute of a <lookup> node.
 * (Static function)
 * XXX: change to do a proper set
 */
void lookup_handler::set_ignore(cpl_parser* parser, const string& value)
{
  lookup_node* current_node =
    dynamic_cast_cpl_node_ptr<lookup_node*>(parser->getCurrentNode(),
                                              "lookup_handler::set_ignore");

  current_node->ignore = value;
} /* lookup_handler::set_ignore(cpl_parser*, const string&) */


/*
 * lookup_handler::set_clear(cpl_parser*, const string&)
 * Handler function for the clear attribute of a <lookup>.
 * (Static function)
 */
void lookup_handler::set_clear(cpl_parser* parser, const string& value)
{
  lookup_node* current_node =
    dynamic_cast_cpl_node_ptr<lookup_node*>(parser->getCurrentNode(),
                                              "lookup_handler::set_clear");

  current_node->clear = 
    get_map_value<bool>(yes_no_map, value, parser, "clear");
} /* lookup_handler::set_clear(cpl_parser*, const string&) */


/*
 * lookup_handler::end(cpl_parser *) const
 * End a lookup node.  Check to make sure source is set, and that
 * use and ignore are not both set.
 */
void lookup_handler::end(cpl_parser *parser) const
{
  lookup_node* current_node =
    dynamic_cast_cpl_node_ptr<lookup_node*>(parser->getCurrentNode(),
                                            "lookup_handler::end");

  if (current_node->source.length() == 0) {
    throw CPLException(CPLException::MissingAttr, "source", parser);
  }

  if (current_node->use.length() > 0 && current_node->ignore.length() > 0) {
    throw CPLException(CPLException::DuplicateAttr,
                       parser->getLocator(),
                       "\'use\' and \'ignore\'",
                       qualified_name(getSpace(), getName()));
  }
} /* lookup_handler::end(cpl_parser *) */


/*
 * lookup_node::lookup_node()
 * Construct a <lookup> node.  Initialize the timeout to 30 seconds.
 */
lookup_node::lookup_node() :
  cpl_node()
{
  timeout = 30;
} /* lookup_node::lookup_node() */


DEFINE_SIMPLE_OUTPUT_HANDLER_END(lookup, success)
DEFINE_SIMPLE_OUTPUT_HANDLER_END(lookup, notfound)
DEFINE_SIMPLE_OUTPUT_HANDLER_END(lookup, failure)


/* <remove-location> node */

/*
 * remove_location_handler::allocate_node()
 * Allocate and return a new remove_location_node.
 */
cpl_node_ptr remove_location_handler::allocate_node() const
{
  return cpl_node_ptr(new remove_location_node());
} /* remove_location_handler::allocate_node() */


/*
 * remove_location_handler::remove_location_handler()
 * Fill in the subnode and attribute maps for a remove-location node.
 * Outputs all CPL nodes.  Takes a 'url' attribute.
 */
remove_location_handler::remove_location_handler() :
  cpl_node_handler(NULL, "remove-location")
{
  output_all_cpl_nodes();

  insert_att_map("location", &remove_location_handler::set_location);
  insert_att_map("param", &remove_location_handler::set_param);
  insert_att_map("value", &remove_location_handler::set_value);
} /* remove_location_handler::remove_location_handler() */


/*
 * remove_location_handler::set_location(cpl_parser*, const string&)
 * Handler function for the location attribute of a <remove-location>.
 * (Static function)
 * TODO: represent URLs as url_t's, not strings.
 */
void remove_location_handler::set_location(cpl_parser* parser,
                                           const string& value)
{
  remove_location_node* current_node =
    dynamic_cast_cpl_node_ptr<remove_location_node*>
    (parser->getCurrentNode(), "remove__handler::set_location");

  current_node->location = value;
} /* remove_location_handler::set_location(cpl_parser*, const std::string&) */


/*
 * remove_location_handler::set_param(cpl_parser*, const string&)
 * Handler function for the param attribute of a <remove-location>.
 * (Static function)
 * TODO: represent param and value as pairs.
 */
void remove_location_handler::set_param(cpl_parser* parser,
                                           const string& value)
{
  remove_location_node* current_node =
    dynamic_cast_cpl_node_ptr<remove_location_node*>
    (parser->getCurrentNode(), "remove_location_handler::set_param");

  current_node->param = value;
} /* remove_location_handler::set_param(cpl_parser*, const std::string&) */


/*
 * remove_location_handler::set_value(cpl_parser*, const string&)
 * Handler function for the value attribute of a <remove-location>.
 * (Static function)
 * TODO: represent param and value as pairs.
 */
void remove_location_handler::set_value(cpl_parser* parser,
                                           const string& value)
{
  remove_location_node* current_node =
    dynamic_cast_cpl_node_ptr<remove_location_node*>
    (parser->getCurrentNode(), "remove_location_handler::set_value");

  current_node->value = value;
} /* remove_location_handler::set_value(cpl_parser*, const std::string&) */


/*
 * remove_location_handler::end(cpl_parser*)
 * 
 * End a remove-location node.
 * Validate the parameters, then insert the current output into the node's
 * output.
 * TODO: validate that the numbers of parameters and values match.
 */
void remove_location_handler::end(cpl_parser *parser) const
{
  remove_location_node* current_node =
    dynamic_cast_cpl_node_ptr<remove_location_node*>
    (parser->getCurrentNode(), "remove_location_handler::end");

  // XXX validation goes here

  current_node->output = parser->getCurrentOutput();
} /* remove_location_handler::end(cpl_parser *) */

/*
 * Singleton handlers.
 */

location_handler the_location_handler;

lookup_handler the_lookup_handler;
DEFINE_SINGLETON_SIMPLE_OUTPUT_HANDLER(lookup, success);
DEFINE_SINGLETON_SIMPLE_OUTPUT_HANDLER(lookup, notfound);
DEFINE_SINGLETON_SIMPLE_OUTPUT_HANDLER(lookup, failure);

remove_location_handler the_remove_location_handler;
