/* cpl-nodes-nca.cpp - handlers and definitions for CPL non-signalling
 * operation tags.   ("NCA" stands for "non-call actions", the old name for
 * this class of tags.)
 */ 
/* Copyright 2001 Columbia University.  All rights reserved. */

#include <strstream>

#include "cpl-exception.h"

#include "cpl-nodes.h"
#include "cpl-nodes-nca.h"
#include "cpl-templatefns.h"

using namespace std;

/* <mail> node */

/*
 * mail_handler::allocate_node()
 * Allocate and return a new mail_node.
 */
cpl_node_ptr mail_handler::allocate_node() const
{
  return cpl_node_ptr(new mail_node());
} /* mail_handler::allocate_node() */


/*
 * mail_handler::mail_handler()
 * Fill in the subnode and attribute maps for a mail node.
 * Outputs all CPL nodes.
 */
mail_handler::mail_handler() :
  cpl_node_handler(NULL, "mail")
{
  output_all_cpl_nodes();

  insert_att_map("url", &mail_handler::set_url);
} /* mail_handler::mail_handler() */


/*
 * mail_handler::set_url(cpl_parser*, const string&)
 * Handler function for the url attribute of a <mail>.
 * (Static function)
 * TODO: represent URLs as url_t's, not strings.
 * TODO: make sure that the URL is a mailto: URL.
 */
void mail_handler::set_url(cpl_parser* parser, const string& value)
{
  mail_node* current_node =
    dynamic_cast_cpl_node_ptr<mail_node*>(parser->getCurrentNode(),
                                              "mail_handler::set_url");

  current_node->url = value;
} /* mail_handler::set_url(cpl_parser*, const std::string&) */


/*
 * mail_handler::end(cpl_parser *) const
 * End a mail node.  Check to make sure url is set, then insert the
 * current output into the node's output. 
 */
void mail_handler::end(cpl_parser *parser) const
{
  mail_node* current_node =
    dynamic_cast_cpl_node_ptr<mail_node*>(parser->getCurrentNode(),
                                          "mail_handler::end");

  if (current_node->url.length() == 0) {
    throw CPLException(CPLException::MissingAttr, "url", parser);
  }

  current_node->output = parser->getCurrentOutput();
} /* mail_handler::end(cpl_parser *) */


/* <log> node */

/*
 * log_handler::allocate_node()
 * Allocate and return a new log_node.
 */
cpl_node_ptr log_handler::allocate_node() const
{
  return cpl_node_ptr(new log_node());
} /* log_handler::allocate_node() */


/*
 * log_handler::log_handler()
 * Fill in the subnode and attribute maps for a log node.
 * Outputs all CPL nodes.
 */
log_handler::log_handler() :
  cpl_node_handler(NULL, "log")
{
  output_all_cpl_nodes();

  insert_att_map("name", &log_handler::set_name);
  insert_att_map("comment", &log_handler::set_comment);
} /* log_handler::log_handler() */


/*
 * log_handler::set_name(cpl_parser*, const string&)
 * Handler function for the name attribute of a <log>.
 * (Static function)
 */
void log_handler::set_name(cpl_parser* parser, const string& value)
{
  log_node* current_node =
    dynamic_cast_cpl_node_ptr<log_node*>(parser->getCurrentNode(),
                                         "log_handler::set_name");

  current_node->name = value;
} /* log_handler::set_name(cpl_parser*, const std::string&) */


/*
 * log_handler::set_comment(cpl_parser*, const string&)
 * Handler function for the comment attribute of a <log>.
 * (Static function)
 */
void log_handler::set_comment(cpl_parser* parser, const string& value)
{
  log_node* current_node =
    dynamic_cast_cpl_node_ptr<log_node*>(parser->getCurrentNode(),
                                         "log_handler::set_comment");

  current_node->comment = value;
} /* log_handler::set_comment(cpl_parser*, const std::string&) */


/*
 * log_handler::end(cpl_parser *) const
 * End a log node.  Insert the current output into the node's output. 
 */
void log_handler::end(cpl_parser *parser) const
{
  log_node* current_node =
    dynamic_cast_cpl_node_ptr<log_node*>(parser->getCurrentNode(),
                                          "log_handler::end");

  current_node->output = parser->getCurrentOutput();
} /* log_handler::end(cpl_parser *) */


/*
 * Singleton handlers.
 */

mail_handler the_mail_handler;
log_handler the_log_handler;
