/* cpl-nodes-switch.cpp - handlers and definitions for CPL switch tags. */
/* Copyright 2001 Columbia University.  All rights reserved. */

#include "cpl-exception.h"

#include "cpl-templatefns.h"

#include "cpl-nodes.h"
#include "cpl-nodes-internal.h"
#include "cpl-nodes-switch.h"

using namespace std;

/*
 * switch_node::append_output(const switch_output_ptr&)
 * Append a switch_output to the switch_node's output vector.
 */
void switch_node::append_output(const switch_output_ptr& out)
{
  outputs.push_back(out);
} /* switch_node::append_output(const switch_output_ptr&) */


/*
 * switch_handler::switch_handler(const char * const, const char * const)
 * Initialize a generic switch_handler -- insert the <notfound> and
 * <otherwise> outputs.
 */
switch_handler::switch_handler(const char * const space,
                               const char * const name) :
  cpl_node_handler(space, name)
{
  insert_subnode_map("otherwise", &the_otherwise_handler);
  insert_subnode_map("notfound",  &the_notfound_handler);
} /* switch_handler::switch_handler(const char * const, const char * const) */
  

/*
 * switch_output_handler::new_stack_frame() const
 * Allocate a new parse stack frame for this switch output handler.  switch
 * outputs need additional data beyond that of generic CPL handlers, so this
 * i subclassed to provide a switch_stackelement.
 */
cpl_parser_stackelement_ptr switch_output_handler::new_stack_frame() const
{
  return cpl_parser_stackelement_ptr(new switch_stackelement(this));
} /* switch_handler::new_stack_frame() const */


/*
 * switch_output_handler::start(cpl_parser *) const
 * Start handling a switch output.  Allocate an appropriate switch_output, 
 * and put it into the stack frame.
 */
void switch_output_handler::start(cpl_parser * parser) const
{
  switch_node* parent_node =
    dynamic_cast_cpl_node_ptr<switch_node*>(parser->getParentNode(),
                                            "switch_output_handler::start");

  switch_output_ptr swout = allocate_output(parent_node);

  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "switch_output_handler::start");

  current_frame->current_switch_output = swout;
} /* switch_output_handler::start(cpl_parser *) const */


/*
 * switch_output_handler::end(cpl_parser *) const
 * Finish handling a switch output.  Insert the output in the parent
 * node's output set. 
 */
void switch_output_handler::end(cpl_parser * parser) const
{
  switch_node* parent_node =
    dynamic_cast_cpl_node_ptr<switch_node*>(parser->getParentNode(),
                                            "switch_output_handler::end");

  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "switch_output_handler::end");

  current_frame->current_switch_output->output = parser->getCurrentOutput();

  parent_node->append_output(current_frame->current_switch_output);
} /* switch_output_handler::end(cpl_parser *) const */


/*
 * Generic outputs: <notfound> and <otherwise>
 */


/*
 * notfound_handler::notfound_handler()
 * Initialize a notfound_handler.
 */
notfound_handler::notfound_handler() : 
  switch_output_handler(NULL, "notfound")
{
  // Nothing else to do
} /* notfound_handler::notfound_handler() */


/*
 * notfound_handler::allocate_output(const switch_node * const) const
 * Allocate and return a new notfound_output.  Pass it its parent.
 */
switch_output_ptr
notfound_handler::allocate_output(const switch_node * const par) const
{
  return switch_output_ptr(new otherwise_output(par));
} /* notfound_handler::allocate_output(const switch_node * const) const */


/*
 * otherwise_handler::otherwise_handler()
 * Initialize a otherwise_handler.
 */
otherwise_handler::otherwise_handler() : 
  switch_output_handler(NULL, "otherwise")
{
  // Nothing else to do
} /* otherwise_handler::otherwise_handler() */


/*
 * otherwise_handler::allocate_output(const switch_node * const) const
 * Allocate and return a new otherwise_output.  Pass it its parent.
 */
switch_output_ptr
otherwise_handler::allocate_output(const switch_node * const par) const
{
  return switch_output_ptr(new otherwise_output(par));
} /* otherwise_handler::allocate_output(const switch_node * const) const */


/*
 * <string-switch> and <string>
 */

/*
 * string_switch_handler::allocate_node() const
 * Allocate a string_switch node.
 */
cpl_node_ptr string_switch_handler::allocate_node() const
{
  return cpl_node_ptr(new string_switch_node());
} /* string_switch_handler::allocate_node() const */


/*
 * string_switch_handler::string_switch_handler()
 * Initialize a string_switch_handler.  In addition to the generic handler
 * fields, we also support the <string> output and the "field" attribute.
 */
string_switch_handler::string_switch_handler() :
  switch_handler(NULL, "string-switch")
{
  insert_subnode_map("string",  &the_string_handler);

  insert_att_map("field", &string_switch_handler::set_field);
} /* string_switch_handler::string_switch_handler() */


/*
 * string_switch_handler::set_field(cpl_parser*, const string&)
 * Handler function for the field attribute of a <string-switch>
 */
void string_switch_handler::set_field(cpl_parser *parser, const string& value)
{
  string_switch_node* current_node =
    dynamic_cast_cpl_node_ptr<string_switch_node*>
    (parser->getCurrentNode(), "string_switch_handler::set_field");

  current_node->field = 
    get_map_value<string_switch_node::string_field>(string_field_map,
                                                    value, parser, "field");
} /* string_switch_handler::set_field(cpl_parser*, const string&) */


/* The map for string fields. */
std::map<std::string, string_switch_node::string_field> string_field_map;

/*
 * init_string_field_map()
 * Initalize a map for string_switch_node::string_field values.
 */
void init_string_field_map()
{
  string_field_map["subject"] = string_switch_node::subject;
  string_field_map["organization"] = string_switch_node::organization;
  string_field_map["user-agent"] = string_switch_node::user_agent;
  string_field_map["language"] = string_switch_node::language;
  string_field_map["display"] = string_switch_node::display;
} /* init_string_field_map() */


/*
 * string_switch_handler::end(cpl_parser *) const
 * End a <string-switch> node.  Make sure that its field has been set.
 */
void string_switch_handler::end(cpl_parser * parser) const
{
  string_switch_node* current_node =
    dynamic_cast_cpl_node_ptr<string_switch_node*>
    (parser->getCurrentNode(),"string_switch_handler::end");

  if (current_node->field == string_switch_node::no_field) {
    throw CPLException(CPLException::MissingAttr, "field", parser);
  }
} /* string_switch_handler::end(cpl_parser *) const */


/*
 * string_switch_node::string_switch_node()
 * Construct a <string-switch> node.  Initialize the field to no_field.
 */
string_switch_node::string_switch_node() :
  switch_node()
{
  field = no_field;
} /* string_switch_node::string_switch_node() */


/*
 * string_handler::string_handler()
 * Construct a <string> handler.  It's an output handler, plus it has the
 * 'is' and 'contains' outputs.
 */
string_handler::string_handler() :
  switch_output_handler(NULL, "string")
{
  insert_att_map("is", &string_handler::set_is);
  insert_att_map("contains", &string_handler::set_contains);
} /* string_handler::string_handler() */


/*
 * string_handler::allocate_output(const switch_node * const)
 * Allocate a string_output object.
 * Note that this will (correctly) throw an InternalError if the parent isn't
 * a <string-switch> node.
 */
switch_output_ptr
string_handler::allocate_output(const switch_node * const raw_par) const
{
  const string_switch_node* par =
    dynamic_cast_cpl_node_ptr<const string_switch_node*>
    (raw_par, "string_handler::allocate_output");
  
  return switch_output_ptr(new string_output(par));
} /* string_handler::allocate_output(const switch_node * const) */


/*
 * string_handler::set_is(cpl_parser*, const string&)
 * Handler function for the 'is' attribute of a <string>
 */
void string_handler::set_is(cpl_parser *parser, const string& value)
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "string_handler::set_is");

  string_output* current_output =
    dynamic_cast_switch_output_ptr<string_output*>
    (current_frame->getCurrentOutput(), "string_handler::set_is");

  if (current_output->match != string_output::unknown) {
    throw CPLException(CPLException::DuplicateAttr,
                       "\'is\' and \'contains\'",
                       parser);
  }

  current_output->match = string_output::is;
  current_output->pattern = value;
} /* string_handler::set_is(cpl_parser*, const string&) */


/*
 * string_handler::set_contains(cpl_parser*, const string&)
 * Handler function for the 'contains' attribute of a <string>
 */
void string_handler::set_contains(cpl_parser *parser, const string& value)
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "string_handler::set_contains");

  string_output* current_output =
    dynamic_cast_switch_output_ptr<string_output*>
    (current_frame->getCurrentOutput(), "string_handler::set_contains");

  if (current_output->match != string_output::unknown) {
    throw CPLException(CPLException::DuplicateAttr,
                       "\'is\' and \'contains\'",
                       parser);
  }

  current_output->match = string_output::contains;
  current_output->pattern = value;
} /* string_handler::set_contains(cpl_parser*, const string&) */


/*
 * string_handler::end(cpl_parser*) const
 * End a <string> output.  Check to make sure we've gotten either an 'is' or
 * 'contains' attribute, then call the switch_output_handler::end().
 */
void string_handler::end(cpl_parser * parser) const
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "string_handler::end");

  string_output* current_output =
    dynamic_cast_switch_output_ptr<string_output*>
    (current_frame->getCurrentOutput(), "string_handler::end");

  if (current_output->match == string_output::unknown) {
    throw CPLException(CPLException::MissingAttr,
                       "\'is\' or \'contains\'",
                       parser);
  }

  switch_output_handler::end(parser);
} /* string_handler::end(cpl_parser*) const */


/*
 * string_output::string_output(const string_switch_node *)
 * Construct a new string_output object.  Initialize the match_type
 * to 'unknown', and fill in the parent object.
 */
string_output::string_output(const string_switch_node * par) :
  parent(par)
{
  match = unknown;
} /* string_output::string_output(const string_switch_node *) */


/*
 * <address-switch> and <address>
 */

/*
 * address_switch_handler::allocate_node() const
 * Allocate a address_switch node.
 */
cpl_node_ptr address_switch_handler::allocate_node() const
{
  return cpl_node_ptr(new address_switch_node());
} /* address_switch_handler::allocate_node() const */


/*
 * address_switch_handler::address_switch_handler()
 * Initialize a address_switch_handler.  In addition to the generic handler
 * fields, we also support the <address> output and the "field" and
 * "subfield" attributes.
 */
address_switch_handler::address_switch_handler() :
  switch_handler(NULL, "address-switch")
{
  insert_subnode_map("address",  &the_address_handler);

  insert_att_map("field", &address_switch_handler::set_field);
  insert_att_map("subfield", &address_switch_handler::set_subfield);
} /* address_switch_handler::address_switch_handler() */


/*
 * address_switch_handler::set_field(cpl_parser*, const string&)
 * Handler function for the field attribute of a <address-switch>
 */
void address_switch_handler::set_field(cpl_parser *parser, const string& value)
{
  address_switch_node* current_node =
    dynamic_cast_cpl_node_ptr<address_switch_node*>
    (parser->getCurrentNode(), "address_switch_handler::set_field");

  current_node->field = 
    get_map_value<address_switch_node::address_field>(address_field_map,
                                                      value, parser, "field");
} /* address_switch_handler::set_field(cpl_parser*, const string&) */


/* The map for address fields. */
std::map<std::string, address_switch_node::address_field> address_field_map;

/*
 * init_address_field_map()
 * Initalize a map for address_switch_node::address_field values.
 */
void init_address_field_map()
{
  address_field_map["origin"] = address_switch_node::origin;
  address_field_map["destination"] = address_switch_node::destination;
  address_field_map["original-destination"] =
    address_switch_node::original_destination;
} /* init_address_field_map() */


/*
 * address_switch_handler::set_subfield(cpl_parser*, const string&)
 * Handler function for the subfield attribute of a <address-switch>
 */
void address_switch_handler::set_subfield(cpl_parser *parser,
                                          const string& value)
{
  address_switch_node* current_node =
    dynamic_cast_cpl_node_ptr<address_switch_node*>
    (parser->getCurrentNode(), "address_switch_handler::set_subfield");

  current_node->subfield = 
    get_map_value<address_switch_node::address_subfield>(address_subfield_map,
                                                         value, parser,
                                                         "subfield");
} /* address_switch_handler::set_subfield(cpl_parser*, const string&) */


/* The map for address subfields. */
std::map<std::string, address_switch_node::address_subfield> address_subfield_map;

/*
 * init_address_subfield_map()
 * Initalize a map for address_switch_node::address_subfield values.
 */
void init_address_subfield_map()
{
  address_subfield_map["address-type"] = address_switch_node::address_type;
  address_subfield_map["user"] = address_switch_node::user;
  address_subfield_map["host"] = address_switch_node::host;
  address_subfield_map["port"] = address_switch_node::port;
  address_subfield_map["tel"] = address_switch_node::tel;
  address_subfield_map["display"] = address_switch_node::display;
  address_subfield_map["password"] = address_switch_node::password;
  address_subfield_map["alias-type"] = address_switch_node::alias_type;
} /* init_address_subfield_map() */


/*
 * address_switch_handler::end(cpl_parser *) const
 * End a <address-switch> node.  Make sure that its field has been set.
 */
void address_switch_handler::end(cpl_parser * parser) const
{
  address_switch_node* current_node =
    dynamic_cast_cpl_node_ptr<address_switch_node*>
    (parser->getCurrentNode(),"address_switch_handler::end");

  if (current_node->field == address_switch_node::no_field) {
    throw CPLException(CPLException::MissingAttr, "field", parser);
  }
  /* It's okay if subfield is not set. */
} /* address_switch_handler::end(cpl_parser *) const */


/*
 * address_switch_node::address_switch_node()
 * Construct a <address-switch> node.  Initialize the field and subfield to
 * no_field and no_subfield respectively.
 */
address_switch_node::address_switch_node() :
  switch_node()
{
  field = no_field;
  subfield = no_subfield;
} /* address_switch_node::address_switch_node() */


/*
 * address_handler::address_handler()
 * Construct an <address> handler.  It's an output handler, plus it has the
 * 'is', 'contains', and 'subdomain-of' outputs.
 */
address_handler::address_handler() :
  switch_output_handler(NULL, "address")
{
  insert_att_map("is", &address_handler::set_is);
  insert_att_map("contains", &address_handler::set_contains);
  insert_att_map("subdomain-of", &address_handler::set_subdomain_of);
} /* address_handler::address_handler() */


/*
 * address_handler::allocate_output(const switch_node * const)
 * Allocate a address_output object.
 * Note that this will (correctly) throw an InternalError if the parent isn't
 * a <address-switch> node.
 */
switch_output_ptr
address_handler::allocate_output(const switch_node * const raw_par) const
{
  const address_switch_node* par =
    dynamic_cast_cpl_node_ptr<const address_switch_node*>
    (raw_par, "address_handler::allocate_output");
  
  return switch_output_ptr(new address_output(par));
} /* address_handler::allocate_output(const switch_node * const) */


/*
 * address_handler::set_is(cpl_parser*, const string&)
 * Handler function for the 'is' attribute of a <address>
 */
void address_handler::set_is(cpl_parser *parser, const string& value)
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "address_handler::set_is");

  address_output* current_output =
    dynamic_cast_switch_output_ptr<address_output*>
    (current_frame->getCurrentOutput(), "address_handler::set_is");

  if (current_output->match != address_output::unknown) {
    throw CPLException(CPLException::DuplicateAttr,
                       "\'is\', \'contains\', or \'subdomain-of\'",
                       parser);
  }

  current_output->match = address_output::is;
  current_output->pattern = value;
} /* address_handler::set_is(cpl_parser*, const string&) */


/*
 * address_handler::set_contains(cpl_parser*, const string&)
 * Handler function for the 'contains' attribute of a <address>
 */
void address_handler::set_contains(cpl_parser *parser, const string& value)
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "address_handler::set_contains");

  address_output* current_output =
    dynamic_cast_switch_output_ptr<address_output*>
    (current_frame->getCurrentOutput(), "address_handler::set_contains");

  if (current_output->match != address_output::unknown) {
    throw CPLException(CPLException::DuplicateAttr,
                       "\'is\', \'contains\', or \'subdomain-of\'",
                       parser);
  }

  /* Check to make sure that the parent node is matching the subfield
   * 'display'. */
  address_switch_node* parent_node =
    dynamic_cast_cpl_node_ptr<address_switch_node*>
    (parser->getParentNode(), "address_handler::set_contains");

  if (parent_node->subfield != address_switch_node::display) {
    throw CPLException(CPLException::IllegalComparitor,
                       "contains",
                       parser);
  }

  current_output->match = address_output::contains;
  current_output->pattern = value;
} /* address_handler::set_contains(cpl_parser*, const string&) */


/*
 * address_handler::set_subdomain_of(cpl_parser*, const string&)
 * Handler function for the 'subdomain-of' attribute of a <address>
 */
void address_handler::set_subdomain_of(cpl_parser *parser, const string& value)
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "address_handler::set_subdomain_of");

  address_output* current_output =
    dynamic_cast_switch_output_ptr<address_output*>
    (current_frame->getCurrentOutput(), "address_handler::set_subdomain_of");

  if (current_output->match != address_output::unknown) {
    throw CPLException(CPLException::DuplicateAttr,
                       "\'is\', \'contains\', or \'subdomain-of\'",
                       parser);
  }

  /* Check to make sure that the parent node is matching the subfields
   * 'host' or 'tel'. */
  address_switch_node* parent_node =
    dynamic_cast_cpl_node_ptr<address_switch_node*>
    (parser->getParentNode(), "address_handler::set_subdomain_of");

  if (parent_node->subfield != address_switch_node::host &&
      parent_node->subfield != address_switch_node::tel) {
    throw CPLException(CPLException::IllegalComparitor,
                       "subdomain-of",
                       parser);
  }

  current_output->match = address_output::subdomain_of;
  current_output->pattern = value;
} /* address_handler::set_subdomain_of(cpl_parser*, const string&) */


/*
 * address_handler::end(cpl_parser*) const
 * End a <address> output.  Check to make sure we've gotten one of an 'is',
 * 'contains' or 'subdomain-of' attribute, then call
 * switch_output_handler::end().
 */
void address_handler::end(cpl_parser * parser) const
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "address_handler::end");

  address_output* current_output =
    dynamic_cast_switch_output_ptr<address_output*>
    (current_frame->getCurrentOutput(), "address_handler::end");

  if (current_output->match == address_output::unknown) {
    throw CPLException(CPLException::MissingAttr,
                       "\'is\', \'contains\', or 'subdomain-of'",
                       parser);
  }

  switch_output_handler::end(parser);
} /* address_handler::end(cpl_parser*) const */


/*
 * address_output::address_output(const address_switch_node *)
 * Construct a new address_output object.  Initialize the match_type
 * to 'unknown', and fill in the parent object.
 */
address_output::address_output(const address_switch_node * par) :
  parent(par)
{
  match = unknown;
} /* address_output::address_output(const address_switch_node *) */


/*
 * <time-switch> and <time>
 */

/*
 * time_switch_handler::allocate_node() const
 * Allocate a time_switch node.
 */
cpl_node_ptr time_switch_handler::allocate_node() const
{
  return cpl_node_ptr(new time_switch_node());
} /* time_switch_handler::allocate_node() const */


/*
 * time_switch_handler::time_switch_handler()
 * Initialize a time_switch_handler.  In addition to the generic handler
 * fields, we also support the <time> output and the "tzid" and
 * "tzurl" attributes.
 */
time_switch_handler::time_switch_handler() :
  switch_handler(NULL, "time-switch")
{
  insert_subnode_map("time",  &the_time_handler);

  insert_att_map("tzid", &time_switch_handler::set_tzid);
  insert_att_map("tzurl", &time_switch_handler::set_tzurl);
} /* time_switch_handler::time_switch_handler() */


/*
 * time_switch_handler::set_tzid(cpl_parser*, const string&)
 * Handler function for the tzid attribute of a <time-switch>
 */
void time_switch_handler::set_tzid(cpl_parser *parser, const string& value)
{
  time_switch_node* current_node =
    dynamic_cast_cpl_node_ptr<time_switch_node*>
    (parser->getCurrentNode(), "time_switch_handler::set_tzid");

  /* Check if this is a tzid that our calendar class knows about */
  try {
    Calendar c;
    c.setTimeZone(value.c_str());
  }
  catch (exception& e) {
    throw CPLException(CPLException::UnknownZone, value, "tzid", parser);
  }

  current_node->tzid = value;
} /* time_switch_handler::set_tzid(cpl_parser*, const string&) */


/*
 * time_switch_handler::set_tzurl(cpl_parser*, const string&)
 * Handler function for the tzurl attribute of a <time-switch>
 */
void time_switch_handler::set_tzurl(cpl_parser *parser, const string& value)
{
  time_switch_node* current_node =
    dynamic_cast_cpl_node_ptr<time_switch_node*>
    (parser->getCurrentNode(), "time_switch_handler::set_tzurl");

  current_node->tzurl = value;
} /* time_switch_handler::set_tzurl(cpl_parser*, const string&) */


/*
 * time_switch_node::get_tz_string()
 * Get the TZ name (TZID) for this node.  Either the local tzid, or
 * the current TZ env var, or NULL.
 * XXX: should support TZURL better.
 */
const char * time_switch_node::get_tz_string()
{
  if (tzid.length() != 0) {
    return tzid.c_str();
  }
  else {
    return getenv("TZ");
  }
} /* time_switch_node::get_tz_string() */

/*
 * time_handler::time_handler()
 * Construct a <time> handler.  It's an output handler, plus it has a
 * whole bunch of outputs.
 */
time_handler::time_handler() :
  switch_output_handler(NULL, "time")
{
  insert_att_map("dtstart", &time_handler::set_dtstart);
  insert_att_map("dtend", &time_handler::set_dtend);
  insert_att_map("duration", &time_handler::set_duration);

  insert_att_map("freq", &time_handler::set_freq);
  insert_att_map("interval", &time_handler::set_interval);
  insert_att_map("until", &time_handler::set_until);
  insert_att_map("count", &time_handler::set_count);

  insert_att_map("bysecond", &time_handler::set_bysecond);
  insert_att_map("byminute", &time_handler::set_byminute);
  insert_att_map("byhour", &time_handler::set_byhour);
  insert_att_map("byday", &time_handler::set_byday);
  insert_att_map("bymonthday", &time_handler::set_bymonthday);
  insert_att_map("byyearday", &time_handler::set_byyearday);
  insert_att_map("byweekno", &time_handler::set_byweekno);
  insert_att_map("bymonth", &time_handler::set_bymonth);

  insert_att_map("wkst", &time_handler::set_wkst);
} /* time_handler::time_handler() */


/*
 * time_handler::allocate_output(const switch_node * const)
 * Allocate a time_output object.
 * Note that this will (correctly) throw an InternalError if the parent isn't
 * a <time-switch> node.
 */
switch_output_ptr
time_handler::allocate_output(const switch_node * const raw_par) const
{
  const time_switch_node* par =
    dynamic_cast_cpl_node_ptr<const time_switch_node*>
    (raw_par, "time_handler::allocate_output");
  
  return switch_output_ptr(new time_output(par));
} /* time_handler::allocate_output(const switch_node * const) */


/*
 * time_handler::set_dtstart(cpl_parser*, const string&)
 * Handler function for the 'dtstart' attribute of <time>
 */
void time_handler::set_dtstart(cpl_parser *parser, const string& value)
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "time_handler::set_dtstart");

  time_output* current_output =
    dynamic_cast_switch_output_ptr<time_output*>
    (current_frame->getCurrentOutput(), "time_handler::set_dtstart");

  time_switch_node* parent_node =
    dynamic_cast_cpl_node_ptr<time_switch_node*>
    (parser->getParentNode(), "time_handler::set_dtstart");

  try {
    Calendar dtStart = Recurrence::parseDateTime(value);
    if (value[value.length()] != 'Z') {
      /* If the last character is not Z, this is a local-time value. */
      dtStart.setTimeZone(parent_node->get_tz_string());
    }
    current_output->recurrence.setDtStart(dtStart);
  }
  catch (exception& e) {
    throw CPLException(CPLException::MalformedAttr, value, "dtstart",
                       e.what(), parser);
  }
} /* time_handler::set_dtstart(cpl_parser*, const string&) */


/*
 * time_handler::set_dtend(cpl_parser*, const string&)
 * Handler function for the 'dtend' attribute of <time>
 */
void time_handler::set_dtend(cpl_parser *parser, const string& value)
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "time_handler::set_dtend");

  time_output* current_output =
    dynamic_cast_switch_output_ptr<time_output*>
    (current_frame->getCurrentOutput(), "time_handler::set_dtend");

  time_switch_node* parent_node =
    dynamic_cast_cpl_node_ptr<time_switch_node*>
    (parser->getParentNode(), "time_handler::set_dtstart");

  if (current_output->recurrence.getDuration().getInterval() != 0) {
    throw CPLException(CPLException::DuplicateAttr,
                       "\'dtend\' and \'duration\'",
                       parser);
  }

  try {
    Calendar dtEnd = Recurrence::parseDateTime(value);
    if (value[value.length()] != 'Z') {
      /* If the last character is not Z, this is a local-time value. */
      dtEnd.setTimeZone(parent_node->get_tz_string());
    }
    current_output->recurrence.setDtEnd(dtEnd);
  }
  catch (exception& e) {
    throw CPLException(CPLException::MalformedAttr, value, "dtend",
                       e.what(), parser);
  }
} /* time_handler::set_dtend(cpl_parser*, const string&) */


/*
 * time_handler::set_duration(cpl_parser*, const string&)
 * Handler function for the 'duration' attribute of <time>
 */
void time_handler::set_duration(cpl_parser *parser, const string& value)
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "time_handler::set_duration");

  time_output* current_output =
    dynamic_cast_switch_output_ptr<time_output*>
    (current_frame->getCurrentOutput(), "time_handler::set_duration");

  if (current_output->recurrence.getDuration().getInterval() != 0) {
    throw CPLException(CPLException::DuplicateAttr,
                       "\'dtend\' and \'duration\'",
                       parser);
  }

  try {
    current_output->recurrence.setDuration(value);
  }
  catch (exception& e) {
    throw CPLException(CPLException::MalformedAttr, value, "duration",
                       e.what(), parser);
  }
} /* time_handler::set_duration(cpl_parser*, const string&) */


/*
 * time_handler::set_freq(cpl_parser*, const string&)
 * Handler function for the 'freq' attribute of <time>
 */
void time_handler::set_freq(cpl_parser *parser, const string& value)
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "time_handler::set_freq");

  time_output* current_output =
    dynamic_cast_switch_output_ptr<time_output*>
    (current_frame->getCurrentOutput(), "time_handler::set_freq");

  try {
    current_output->recurrence.setFrequency(value);
  }
  catch (exception& e) {
    throw CPLException(CPLException::MalformedAttr, value, "freq",
                       e.what(), parser);
  }
} /* time_handler::set_freq(cpl_parser*, const string&) */


/*
 * time_handler::set_interval(cpl_parser*, const string&)
 * Handler function for the 'interval' attribute of <time>
 */
void time_handler::set_interval(cpl_parser *parser, const string& value)
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "time_handler::set_interval");

  time_output* current_output =
    dynamic_cast_switch_output_ptr<time_output*>
    (current_frame->getCurrentOutput(), "time_handler::set_interval");

  try {
    current_output->recurrence.setInterval(value);
  }
  catch (exception& e) {
    throw CPLException(CPLException::MalformedAttr, value, "interval",
                       e.what(), parser);
  }
} /* time_handler::set_interval(cpl_parser*, const string&) */


/*
 * time_handler::set_until(cpl_parser*, const string&)
 * Handler function for the 'until' attribute of <time>
 */
void time_handler::set_until(cpl_parser *parser, const string& value)
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "time_handler::set_until");

  time_output* current_output =
    dynamic_cast_switch_output_ptr<time_output*>
    (current_frame->getCurrentOutput(), "time_handler::set_until");

  time_switch_node* parent_node =
    dynamic_cast_cpl_node_ptr<time_switch_node*>
    (parser->getParentNode(), "time_handler::set_dtstart");

  try {
    Calendar until = Recurrence::parseDateTime(value);
    if (value[value.length()] != 'Z') {
      /* If the last character is not Z, this is a local-time value. */
      until.setTimeZone(parent_node->get_tz_string());
    }
    current_output->recurrence.setUntil(&until);
  }
  catch (exception& e) {
    throw CPLException(CPLException::MalformedAttr, value, "until",
                       e.what(), parser);
  }
} /* time_handler::set_until(cpl_parser*, const string&) */


/*
 * time_handler::set_count(cpl_parser*, const string&)
 * Handler function for the 'count' attribute of <time>
 */
void time_handler::set_count(cpl_parser *parser, const string& value)
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "time_handler::set_count");

  time_output* current_output =
    dynamic_cast_switch_output_ptr<time_output*>
    (current_frame->getCurrentOutput(), "time_handler::set_count");

  try {
    current_output->recurrence.setCount(value);
  }
  catch (exception& e) {
    throw CPLException(CPLException::MalformedAttr, value, "count",
                       e.what(), parser);
  }
} /* time_handler::set_count(cpl_parser*, const string&) */


/*
 * time_handler::set_bysecond(cpl_parser*, const string&)
 * Handler function for the 'bysecond' attribute of <time>
 */
void time_handler::set_bysecond(cpl_parser *parser, const string& value)
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "time_handler::set_bysecond");

  time_output* current_output =
    dynamic_cast_switch_output_ptr<time_output*>
    (current_frame->getCurrentOutput(), "time_handler::set_bysecond");

  try {
    current_output->recurrence.setBySecond(value);
  }
  catch (exception& e) {
    throw CPLException(CPLException::MalformedAttr, value, "bysecond",
                       e.what(), parser);
  }
} /* time_handler::set_bysecond(cpl_parser*, const string&) */


/*
 * time_handler::set_byminute(cpl_parser*, const string&)
 * Handler function for the 'byminute' attribute of <time>
 */
void time_handler::set_byminute(cpl_parser *parser, const string& value)
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "time_handler::set_byminute");

  time_output* current_output =
    dynamic_cast_switch_output_ptr<time_output*>
    (current_frame->getCurrentOutput(), "time_handler::set_byminute");

  try {
    current_output->recurrence.setByMinute(value);
  }
  catch (exception& e) {
    throw CPLException(CPLException::MalformedAttr, value, "byminute",
                       e.what(), parser);
  }
} /* time_handler::set_byminute(cpl_parser*, const string&) */


/*
 * time_handler::set_byhour(cpl_parser*, const string&)
 * Handler function for the 'byhour' attribute of <time>
 */
void time_handler::set_byhour(cpl_parser *parser, const string& value)
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "time_handler::set_byhour");

  time_output* current_output =
    dynamic_cast_switch_output_ptr<time_output*>
    (current_frame->getCurrentOutput(), "time_handler::set_byhour");

  try {
    current_output->recurrence.setByHour(value);
  }
  catch (exception& e) {
    throw CPLException(CPLException::MalformedAttr, value, "byhour",
                       e.what(), parser);
  }
} /* time_handler::set_byhour(cpl_parser*, const string&) */


/*
 * time_handler::set_byday(cpl_parser*, const string&)
 * Handler function for the 'byday' attribute of <time>
 */
void time_handler::set_byday(cpl_parser *parser, const string& value)
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "time_handler::set_byday");

  time_output* current_output =
    dynamic_cast_switch_output_ptr<time_output*>
    (current_frame->getCurrentOutput(), "time_handler::set_byday");

  try {
    current_output->recurrence.setByDay(value);
  }
  catch (exception& e) {
    throw CPLException(CPLException::MalformedAttr, value, "byday",
                       e.what(), parser);
  }
} /* time_handler::set_byday(cpl_parser*, const string&) */


/*
 * time_handler::set_bymonthday(cpl_parser*, const string&)
 * Handler function for the 'bymonthday' attribute of <time>
 */
void time_handler::set_bymonthday(cpl_parser *parser, const string& value)
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "time_handler::set_bymonthday");

  time_output* current_output =
    dynamic_cast_switch_output_ptr<time_output*>
    (current_frame->getCurrentOutput(), "time_handler::set_bymonthday");

  try {
    current_output->recurrence.setByMonthDay(value);
  }
  catch (exception& e) {
    throw CPLException(CPLException::MalformedAttr, value, "bymonthday",
                       e.what(), parser);
  }
} /* time_handler::set_bymonthday(cpl_parser*, const string&) */


/*
 * time_handler::set_byyearday(cpl_parser*, const string&)
 * Handler function for the 'byyearday' attribute of <time>
 */
void time_handler::set_byyearday(cpl_parser *parser, const string& value)
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "time_handler::set_byyearday");

  time_output* current_output =
    dynamic_cast_switch_output_ptr<time_output*>
    (current_frame->getCurrentOutput(), "time_handler::set_byyearday");

  try {
    current_output->recurrence.setByYearDay(value);
  }
  catch (exception& e) {
    throw CPLException(CPLException::MalformedAttr, value, "byyearday",
                       e.what(), parser);
  }
} /* time_handler::set_byyearday(cpl_parser*, const string&) */


/*
 * time_handler::set_byweekno(cpl_parser*, const string&)
 * Handler function for the 'byweekno' attribute of <time>
 */
void time_handler::set_byweekno(cpl_parser *parser, const string& value)
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "time_handler::set_byweekno");

  time_output* current_output =
    dynamic_cast_switch_output_ptr<time_output*>
    (current_frame->getCurrentOutput(), "time_handler::set_byweekno");

  try {
    current_output->recurrence.setByWeekNo(value);
  }
  catch (exception& e) {
    throw CPLException(CPLException::MalformedAttr, value, "byweekno",
                       e.what(), parser);
  }
} /* time_handler::set_byweekno(cpl_parser*, const string&) */


/*
 * time_handler::set_bymonth(cpl_parser*, const string&)
 * Handler function for the 'bymonth' attribute of <time>
 */
void time_handler::set_bymonth(cpl_parser *parser, const string& value)
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "time_handler::set_bymonth");

  time_output* current_output =
    dynamic_cast_switch_output_ptr<time_output*>
    (current_frame->getCurrentOutput(), "time_handler::set_bymonth");

  try {
    current_output->recurrence.setByMonth(value);
  }
  catch (exception& e) {
    throw CPLException(CPLException::MalformedAttr, value, "bymonth",
                       e.what(), parser);
  }
} /* time_handler::set_bymonth(cpl_parser*, const string&) */


/*
 * time_handler::set_wkst(cpl_parser*, const string&)
 * Handler function for the 'wkst' attribute of <time>
 */
void time_handler::set_wkst(cpl_parser *parser, const string& value)
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "time_handler::set_wkst");

  time_output* current_output =
    dynamic_cast_switch_output_ptr<time_output*>
    (current_frame->getCurrentOutput(), "time_handler::set_wkst");

  try {
    current_output->recurrence.setWeekStart(value);
  }
  catch (exception& e) {
    throw CPLException(CPLException::MalformedAttr, value, "wkst",
                       e.what(), parser);
  }
} /* time_handler::set_wkst(cpl_parser*, const string&) */


/*
 * time_handler::end(cpl_parser*) const
 * End a <time> output.  XXX check that the recurrence is valid, then call
 * switch_output_handler::end().
 */
void time_handler::end(cpl_parser * parser) const
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "time_handler::end");

  time_output* current_output =
    dynamic_cast_switch_output_ptr<time_output*>
    (current_frame->getCurrentOutput(), "time_handler::end");

  if (current_output->recurrence.getDtStart().getTime() == 0) {
    throw CPLException(CPLException::MissingAttr, "dtstart", parser);
  }

  if (current_output->recurrence.getDuration().getInterval() == 0) {
    throw CPLException(CPLException::MissingAttr,
                       "\'dtend\' or \'duration\'",
                       parser);
  }

  // XXX TODO: Implement and call some consistency checks within class Recur.

  // XXX TODO: Check the parameter boundaries.

  switch_output_handler::end(parser);
} /* time_handler::end(cpl_parser*) const */


/*
 * time_output::time_output(const time_switch_node *)
 * Construct a new time_output object.  Fill in the parent object;
 * nothing else needs to explicitly be done here (I think).
 */
time_output::time_output(const time_switch_node * par) :
  parent(par)
{
  
} /* time_output::time_output(const time_switch_node *) */



/*
 * <priority-switch> and <priority>
 */

/*
 * priority_switch_handler::allocate_node() const
 * Allocate a priority_switch node.
 */
cpl_node_ptr priority_switch_handler::allocate_node() const
{
  return cpl_node_ptr(new priority_switch_node());
} /* priority_switch_handler::allocate_node() const */


/*
 * priority_switch_handler::priority_switch_handler()
 * Initialize a priority_switch_handler.  In addition to the generic handler
 * fields, we also support the <priority> output.  <priority-switch> has
 * no attributes.
 */
priority_switch_handler::priority_switch_handler() :
  switch_handler(NULL, "priority-switch")
{
  insert_subnode_map("priority",  &the_priority_handler);
} /* priority_switch_handler::priority_switch_handler() */


/*
 * priority_handler::priority_handler()
 * Construct an <priority> handler.  It's an output handler, plus it has the
 * 'less', 'greater', and 'equal' attributes.
 */
priority_handler::priority_handler() :
  switch_output_handler(NULL, "priority")
{
  insert_att_map("less", &priority_handler::set_less);
  insert_att_map("greater", &priority_handler::set_greater);
  insert_att_map("equal", &priority_handler::set_equal);
} /* priority_handler::priority_handler() */


/*
 * priority_handler::allocate_output(const switch_node * const)
 * Allocate a priority_output object.
 * Note that this will (correctly) throw an InternalError if the parent isn't
 * a <priority-switch> node.
 */
switch_output_ptr
priority_handler::allocate_output(const switch_node * const raw_par) const
{
  const priority_switch_node* par =
    dynamic_cast_cpl_node_ptr<const priority_switch_node*>
    (raw_par, "priority_handler::allocate_output");
  
  return switch_output_ptr(new priority_output(par));
} /* priority_handler::allocate_output(const switch_node * const) */


/*
 * priority_handler::set_less(cpl_parser*, const string&)
 * Handler function for the 'less' attribute of a <priority>
 */
void priority_handler::set_less(cpl_parser *parser, const string& value)
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "priority_handler::set_is");

  priority_output* current_output =
    dynamic_cast_switch_output_ptr<priority_output*>
    (current_frame->getCurrentOutput(), "priority_handler::set_is");

  if (current_output->match != priority_output::unknown) {
    throw CPLException(CPLException::DuplicateAttr,
                       "\'less\', \'greater\', or \'equal\'",
                       parser);
  }

  current_output->sym_value = 
    get_map_value<priority_output::match_value>(priority_value_map,
                                                value, parser, "less");
  current_output->match = priority_output::less;
} /* priority_handler::set_less(cpl_parser*, const string&) */


/*
 * priority_handler::set_greater(cpl_parser*, const string&)
 * Handler function for the 'greater' attribute of a <priority>
 */
void priority_handler::set_greater(cpl_parser *parser, const string& value)
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "priority_handler::set_is");

  priority_output* current_output =
    dynamic_cast_switch_output_ptr<priority_output*>
    (current_frame->getCurrentOutput(), "priority_handler::set_is");

  if (current_output->match != priority_output::unknown) {
    throw CPLException(CPLException::DuplicateAttr,
                       "\'less\', \'greater\', or \'equal\'",
                       parser);
  }

  current_output->sym_value = 
    get_map_value<priority_output::match_value>(priority_value_map,
                                                value, parser, "greater");
  current_output->match = priority_output::greater;
} /* priority_handler::set_greater(cpl_parser*, const string&) */


/*
 * priority_handler::set_equal(cpl_parser*, const string&)
 * Handler function for the 'equal' attribute of a <priority>
 */
void priority_handler::set_equal(cpl_parser *parser, const string& value)
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "priority_handler::set_is");

  priority_output* current_output =
    dynamic_cast_switch_output_ptr<priority_output*>
    (current_frame->getCurrentOutput(), "priority_handler::set_is");

  if (current_output->match != priority_output::unknown) {
    throw CPLException(CPLException::DuplicateAttr,
                       "\'less\', \'greater\', or \'equal\'",
                       parser);
  }

  std::map<std::string, priority_output::match_value>::const_iterator
    it;

  it = priority_value_map.find(value);

  if (it != priority_value_map.end()) {
    /* There was a symbolic value */
    current_output->sym_value = it->second;
  }
  else {
    current_output->sym_value = priority_output::unknown_val;
  }

  /* We always save the string value */
  current_output->str_value = value;

  current_output->match = priority_output::equal;
} /* priority_handler::set_equal(cpl_parser*, const string&) */


/* The map for priority fields. */

std::map<std::string, priority_output::match_value> priority_value_map;

/*
 * init_priority_value_map()
 * Initalize a map for priority_switch_node::priority_field values.
 */
void init_priority_value_map()
{
  priority_value_map["non-urgent"] = priority_output::non_urgent;
  priority_value_map["normal"] = priority_output::normal;
  priority_value_map["urgent"] = priority_output::urgent;
  priority_value_map["emergency"] = priority_output::emergency;
} /* init_priority_value_map() */


/*
 * priority_handler::end(cpl_parser*) const
 * End a <priority> output.  Check to make sure we've gotten one of a 'less',
 * 'greater' or 'equal' attribute, then call
 * switch_output_handler::end().
 */
void priority_handler::end(cpl_parser * parser) const
{
  switch_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<switch_stackelement*>
    (parser->getStackCurrent(), "priority_handler::end");

  priority_output* current_output =
    dynamic_cast_switch_output_ptr<priority_output*>
    (current_frame->getCurrentOutput(), "priority_handler::end");

  if (current_output->match == priority_output::unknown) {
    throw CPLException(CPLException::MissingAttr,
                       "\'less\', \'greater\', or 'equal'",
                       parser);
  }

  switch_output_handler::end(parser);
} /* priority_handler::end(cpl_parser*) const */


/*
 * priority_output::priority_output(const priority_switch_node *)
 * Construct a new priority_output object.  Initialize the match_type
 * to 'unknown', the match_value to 'unknown_val', and fill in the parent
 * object.
 */
priority_output::priority_output(const priority_switch_node * par) :
  parent(par)
{
  match = unknown;
  sym_value = unknown_val;
} /* priority_output::priority_output(const priority_switch_node *) */


/*
 * Singleton handlers.
 */
otherwise_handler the_otherwise_handler;
notfound_handler the_notfound_handler;

string_switch_handler the_string_switch_handler;
string_handler the_string_handler;

address_switch_handler the_address_switch_handler;
address_handler the_address_handler;

time_switch_handler the_time_switch_handler;
time_handler the_time_handler;

priority_switch_handler the_priority_switch_handler;
priority_handler the_priority_handler;

