/* cpl-nodes-switch.h - handlers and definitions for CPL switch tags. */
/* Copyright 2001 Columbia University.  All rights reserved. */

#ifndef CPL_NODES_SWITCH_H_
#define CPL_NODES_SWITCH_H_

#include "cpl-nodes-internal.h"

/* For Recurrence and friends, needed for time-switch */
#include "recur/recur.h"

/* Forward declarations. */
class switch_output;
typedef boost::shared_ptr<switch_output> switch_output_ptr;

/*
 * A generic switch node.
 */
class switch_node : public cpl_node {
protected:
  std::vector<switch_output_ptr> outputs;

public:
  void append_output(const switch_output_ptr&);
};


/*
 * A generic switch node handler.
 */
class switch_handler : public cpl_node_handler {
public:
  switch_handler(const char *const space, const char *const name);

  virtual ~switch_handler() { };
};


/*
 * A generic switch output.
 * Doesn't actually do anything currently; in the future it'll have a
 * run() method or some such.
 */
class switch_output {
private:
  cpl_node_ptr output;

public:
  virtual ~switch_output() { };

  friend class switch_output_handler;
};


/*
 * A generic switch output handler.
 */
class switch_output_handler : public cpl_output_handler {
public:
  switch_output_handler(const char * const name, const char * const space) :
    cpl_output_handler(name, space) { }

  virtual cpl_parser_stackelement_ptr new_stack_frame() const;

  virtual switch_output_ptr allocate_output(const switch_node * const) const
    = 0;

  virtual void start(cpl_parser*) const;
  virtual void end(cpl_parser*) const;
};


/*
 * A switch parse stack element.
 */
class switch_stackelement : public cpl_parser_stackelement {
private:
  switch_output_ptr current_switch_output;
public:
  switch_output_ptr getCurrentOutput() const { return current_switch_output; }

  switch_stackelement(const element_handler * h) :
    cpl_parser_stackelement(h) { }

  virtual ~switch_stackelement() { }

  friend class switch_output_handler;
};


/* Switch outputs that apply to all switch types */

/*
 * A handler for a <notfound> switch output.
 */
class notfound_handler : public switch_output_handler {
public:
  notfound_handler();

  virtual switch_output_ptr allocate_output(const switch_node * const) const;
};

/*
 * A <notfound> switch output.
 */
class notfound_output : public switch_output {
private:
  const switch_node * const parent;

public:
  notfound_output(const switch_node *const p) : parent(p) { }

  virtual ~notfound_output() { }
};

/*
 * A handler for an <otherwise> switch output.
 */
class otherwise_handler : public switch_output_handler {
public:
  otherwise_handler();

  virtual switch_output_ptr allocate_output(const switch_node *const) const;
};

/*
 * A <otherwise> switch output.
 */
class otherwise_output : public switch_output {
private:
  const switch_node * const parent;

public:
  otherwise_output(const switch_node *const p) : parent(p) { }

  virtual ~otherwise_output() { }
};


/* <string-switch> and <string> nodes. */

/*
 * A handler for a <string-switch> node.
 */
class string_switch_handler : public switch_handler {
private:
  virtual cpl_node_ptr allocate_node() const;

public:
  string_switch_handler();

  static void set_field(cpl_parser*, const std::string&);

  virtual void end(cpl_parser*) const;
};


/*
 * A <string-switch> node.
 */
class string_switch_node : public switch_node {
public:
  typedef enum {
    no_field, subject, organization, user_agent, language, display
  } string_field;

protected:
  string_field field;

public:
  string_switch_node();

  virtual ~string_switch_node() { }

  friend class string_switch_handler;
};

/*
 * A handler for a <string> output.
 */
class string_handler : public switch_output_handler {
public:
  string_handler();

  virtual switch_output_ptr allocate_output(const switch_node * const) const;

  static void set_is(cpl_parser*, const std::string&);
  static void set_contains(cpl_parser*, const std::string&);

  virtual void end(cpl_parser*) const;
};

/*
 * A <string> output.
 */
class string_output : public switch_output {
public:
  typedef enum { unknown, is, contains } match_type;

private:
  const string_switch_node * const parent;

  match_type match;
  std::string pattern;

public:
  string_output(const string_switch_node *const);

  virtual ~string_output() { }

  friend class string_handler;
};


/* Value maps for a <string-switch> node. */
extern std::map<std::string, string_switch_node::string_field>
string_field_map;
extern void init_string_field_map();



/* <address-switch> and <address> nodes. */

/*
 * A handler for a <address-switch> node.
 */
class address_switch_handler : public switch_handler {
private:
  virtual cpl_node_ptr allocate_node() const;

public:
  address_switch_handler();

  static void set_field(cpl_parser*, const std::string&);
  static void set_subfield(cpl_parser*, const std::string&);

  virtual void end(cpl_parser*) const;
};


/*
 * A <address-switch> node.
 */
class address_switch_node : public switch_node {
public:
  typedef enum {
    no_field, origin, destination, original_destination
  } address_field;

  typedef enum {
    no_subfield, address_type, user, host, port, tel, display,
    password, alias_type
  } address_subfield;

protected:
  address_field field;
  address_subfield subfield;

public:
  address_switch_node();

  virtual ~address_switch_node() { }

  friend class address_switch_handler;
  friend class address_handler;
};

/*
 * A handler for a <address> output.
 */
class address_handler : public switch_output_handler {
public:
  address_handler();

  virtual switch_output_ptr allocate_output(const switch_node * const) const;

  static void set_is(cpl_parser*, const std::string&);
  static void set_contains(cpl_parser*, const std::string&);
  static void set_subdomain_of(cpl_parser*, const std::string&);

  virtual void end(cpl_parser*) const;
};

/*
 * A <address> output.
 */
class address_output : public switch_output {
public:
  typedef enum { unknown, is, contains, subdomain_of } match_type;

private:
  const address_switch_node * const parent;

  match_type match;
  std::string pattern;

public:
  address_output(const address_switch_node *const);

  virtual ~address_output() { }

  friend class address_handler;
};


/* Value maps for a <address-switch> node. */
extern std::map<std::string, address_switch_node::address_field>
address_field_map;
extern std::map<std::string, address_switch_node::address_subfield>
address_subfield_map;
extern void init_address_field_map();
extern void init_address_subfield_map();


/* <time-switch> and <time> nodes. */

/*
 * A handler for a <time-switch> node.
 */
class time_switch_handler : public switch_handler {
private:
  virtual cpl_node_ptr allocate_node() const;

public:
  time_switch_handler();

  static void set_tzid(cpl_parser*, const std::string&);
  static void set_tzurl(cpl_parser*, const std::string&);
  
  /* Since all attributes of time-switch are optional, there's no need for
   * an end() method here. */
};


/*
 * A <time-switch> node.
 */
class time_switch_node : public switch_node {
protected:
  std::string tzid;
  std::string tzurl;

  const char * get_tz_string();

public:
  virtual ~time_switch_node() { }

  friend class time_switch_handler;
  friend class time_handler;
};

/*
 * A handler for a <time> output.
 */
class time_handler : public switch_output_handler {
public:
  time_handler();

  virtual switch_output_ptr allocate_output(const switch_node * const) const;

  static void set_dtstart(cpl_parser*, const std::string&);
  static void set_dtend(cpl_parser*, const std::string&);
  static void set_duration(cpl_parser*, const std::string&);

  static void set_freq(cpl_parser*, const std::string&);
  static void set_interval(cpl_parser*, const std::string&);
  static void set_until(cpl_parser*, const std::string&);
  static void set_count(cpl_parser*, const std::string&);

  static void set_bysecond(cpl_parser*, const std::string&);
  static void set_byminute(cpl_parser*, const std::string&);
  static void set_byhour(cpl_parser*, const std::string&);
  static void set_byday(cpl_parser*, const std::string&);
  static void set_bymonthday(cpl_parser*, const std::string&);
  static void set_byyearday(cpl_parser*, const std::string&);
  static void set_byweekno(cpl_parser*, const std::string&);
  static void set_bymonth(cpl_parser*, const std::string&);

  static void set_wkst(cpl_parser*, const std::string&);

  virtual void end(cpl_parser*) const;
};

/*
 * A <time> output.
 */
class time_output : public switch_output {
private:
  const time_switch_node * const parent;

  Recurrence recurrence;

public:
  time_output(const time_switch_node *const);

  virtual ~time_output() { }

  friend class time_handler;
};


/* <priority-switch> and <priority> nodes. */

/*
 * A handler for a <priority-switch> node.
 */
class priority_switch_handler : public switch_handler {
private:
  virtual cpl_node_ptr allocate_node() const;

public:
  priority_switch_handler();

  // No end() needed
};


/*
 * A <priority-switch> node.
 */
class priority_switch_node : public switch_node {
public:
  priority_switch_node() : switch_node() { }

  virtual ~priority_switch_node() { }

  friend class priority_switch_handler;
};


/*
 * A handler for a <priority> output.
 */
class priority_handler : public switch_output_handler {
public:
  priority_handler();

  virtual switch_output_ptr allocate_output(const switch_node * const) const;

  static void set_less(cpl_parser*, const std::string&);
  static void set_greater(cpl_parser*, const std::string&);
  static void set_equal(cpl_parser*, const std::string&);

  virtual void end(cpl_parser*) const;
};

/*
 * A <priority> output.
 */
class priority_output : public switch_output {
public:
  typedef enum { unknown, less, greater, equal } match_type;
  typedef enum { unknown_val, non_urgent, normal, urgent, emergency }
  match_value;

private:
  const priority_switch_node * const parent;

  match_type match;
  match_value sym_value;
  std::string str_value;

public:
  priority_output(const priority_switch_node *const);

  virtual ~priority_output() { }

  friend class priority_handler;
};


/* Value maps for a <priority> node. */
extern std::map<std::string, priority_output::match_value>
priority_value_map;
extern void init_priority_value_map();


/* switch-specific template functions */

/*
 * dynamic_cast_switch_output_ptr<type>(switch_output_ptr, string)
 * Dynamic-cast a switch_output_ptr to a plain pointer to a specific
 * subclass of switch_output.  If it's the wrong type, throw a
 * CPLException::InternalError.
 */
template <class Tgt_Type>
Tgt_Type dynamic_cast_switch_output_ptr(switch_output_ptr p,
                                        std::string context)
{
  Tgt_Type dst = dynamic_cast<Tgt_Type>(p.get());
  if (dst == NULL) {
    throw CPLException(CPLException::InternalError,
                       ("switch_output_ptr has incorrect subclass "
                        "in " + context).c_str());
  }
  return dst;
} /* dynamic_cast_cpl_node_ptr<type>(switch_output_ptr, string) */


/*
 * Singleton handlers.
 */

extern otherwise_handler the_otherwise_handler;
extern notfound_handler the_notfound_handler;

extern string_switch_handler the_string_switch_handler;
extern string_handler the_string_handler;

extern address_switch_handler the_address_switch_handler;
extern address_handler the_address_handler;

extern time_switch_handler the_time_switch_handler;
extern time_handler the_time_handler;

extern priority_switch_handler the_priority_switch_handler;
extern priority_handler the_priority_handler;

#endif /* CPL_NODES_SWITCH_H_ */
