/* cpl-nodes-toplevel.cpp - handlers and definitions for top-level CPL tags. */
/* Copyright 2001 Columbia University.  All rights reserved. */

#include "cpl-exception.h"

#include "cpl-nodes.h"
#include "cpl-nodes-toplevel.h"

#include "cpl-templatefns.h"

using namespace std;

/*
 * cpl_toplevel_handler::cpl_toplevel_handler()
 * Fill the subnode map for toplevel with CPL.
 * Leave the attribute map empty; there's no such thing as an attribute for
 * toplevel.
 */
cpl_toplevel_handler::cpl_toplevel_handler() : 
  element_handler(NULL, "[toplevel]")
{
  insert_subnode_map("cpl", &the_cpl_cpl_handler);
} /* cpl_toplevel_handler::cpl_toplevel_handler() */


/*
 * cpl_toplevel_handler::end(cpl_parser*) const
 * End the top level.  Make sure we've seen a <cpl> node.
 */
void cpl_toplevel_handler::end(cpl_parser* parser) const
{
  if (parser->getCPL() == NULL) {
    throw CPLException(CPLException::MissingChild, "cpl", parser);
  }
} /* cpl_toplevel_handler::end(cpl_parser*) const */


/*
 * cpl_cpl_handler::cpl_cpl_handler()
 * Fill in the subnode map for the <cpl> node.
 */
cpl_cpl_handler::cpl_cpl_handler() : 
  element_handler(NULL, "cpl")
{
  insert_subnode_map("ancillary", &the_cpl_ancillary_handler);
  insert_subnode_map("subaction", &the_cpl_subaction_handler);
  insert_subnode_map("outgoing", &the_cpl_outgoing_handler);
  insert_subnode_map("incoming", &the_cpl_incoming_handler);
} /* cpl_cpl_handler::cpl_cpl_handler() */


/*
 * cpl_cpl_handler::start(cpl_parser*) const
 * Allocate the cpl object. 
 * XXX: or elsewhere?
 */
void cpl_cpl_handler::start(cpl_parser* parser) const
{
  /* XXX */
} /* cpl_cpl_handler::start(cpl_parser*) const */


/*
 * cpl_cpl_handler::end(cpl_parser*) const
 * Make sure we've seen at least one <incoming> or <outgoing> node. 
 */
void cpl_cpl_handler::end(cpl_parser* parser) const
{
  if (parser->getCPL()->incoming.get() == NULL &&
      parser->getCPL()->outgoing.get() == NULL) {
    throw CPLException(CPLException::MissingChild, parser->getLocator(),
                       "either 'incoming' or 'outgoing' must be present",
                       qualified_name(getSpace(), getName()));
  }
} /* cpl_cpl_handler::end(cpl_parser*) const */


/*
 * cpl_ancillary_handler::cpl_ancillary_handler()
 * Initialize a CPL ancillary handler.  Nothing specific needs to be done here.
 */
cpl_ancillary_handler::cpl_ancillary_handler() :
  element_handler(NULL, "ancillary")
{
  // Nothing to do
} /* cpl_ancillary_handler::cpl_ancillary_handler() */


/*
 * cpl_subaction_handler::cpl_subaction_handler()
 * Initialize a CPL subaction node.  Support all the usual outputs,
 * and the "id" attribute.
 */
cpl_subaction_handler::cpl_subaction_handler() :
  cpl_output_handler(NULL, "subaction")
{
  insert_att_map("id", &cpl_subaction_handler::set_id);
} /* cpl_subaction_handler::cpl_subaction_handler() */


/*
 * cpl_subaction_handler::new_stack_frame() const
 * Allocate a stack frame for a subaction.
 */
cpl_parser_stackelement_ptr cpl_subaction_handler::new_stack_frame() const
{
  return cpl_parser_stackelement_ptr(new subaction_stackelement(this));
} /* cpl_subaction_handler::new_stack_frame() const */


/*
 * cpl_subaction_handler::set_id(cpl_parser*, const string&)
 * Set the ID of the current subaction to 'value'.  No need to explicitly
 * check for duplication: Xerces-C checks for duplicate attrs.
 */
void cpl_subaction_handler::set_id(cpl_parser* parser, const string& value)
{
  subaction_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<subaction_stackelement*>
    (parser->getStackCurrent(), "cpl_subaction_handler::set_id");

  current_frame->id = value;
} /* cpl_subaction_handler::set_id(cpl_parser*, const string&) */


/*
 * cpl_subaction_handler::end(cpl_parser*) const
 * Check to make sure that the action has had an ID defined, and that no
 * previous subaction used this ID.  If so, put the current output into the
 * CPL's subs map.
 */
void cpl_subaction_handler::end(cpl_parser* parser) const
{
  subaction_stackelement* current_frame =
    dynamic_cast_stackelement_ptr<subaction_stackelement*>
    (parser->getStackCurrent(), "cpl_subaction_handler::end");

  if (current_frame->id.length() == 0) {
    throw CPLException(CPLException::MissingAttr, "id", parser);
  }

  if (parser->getCPL()->subs.count(current_frame->id) != 0) {
    throw CPLException(CPLException::DuplicateSubaction, current_frame->id,
                       parser);
  }

  parser->getCPL()->subs[current_frame->id] =
    parser->getCurrentOutput();
} /* cpl_subaction_handler::set_id(cpl_parser*, const string&) */


/*
 * cpl_outgoing_handler::end(cpl_parser*) const
 * Insert the calculated outgoing node -- the parent stack frame's current
 * output -- into the CPL's outgoing slot.
 */
void cpl_outgoing_handler::end(cpl_parser* parser) const
{
  parser->getCPL()->outgoing = parser->getCurrentOutput();
} /* cpl_outgoing_handler::end(cpl_parser*) const */


/*
 * cpl_incoming_handler::end(cpl_parser*) const
 * Insert the calculated incoming node -- the parent stack frame's current
 * output -- into the CPL's incoming slot.
 */
void cpl_incoming_handler::end(cpl_parser* parser) const
{
  parser->getCPL()->incoming = parser->getCurrentOutput();
} /* cpl_incoming_handler::end(cpl_parser*) const */


/*
 * sub_handler::sub_handler()
 * Fill in the subnode and attribute maps for a sub node.
 * A sub has no (direct) subnodes; leave these blank.
 */
sub_handler::sub_handler() :
  cpl_node_handler(NULL, "sub")
{
  insert_att_map("ref", &sub_handler::set_ref);
} /* sub_handler::sub_handler() */


/*
 * sub_handler::allocate_node()
 * Allocate and return a new sub_node.
 */
cpl_node_ptr sub_handler::allocate_node() const
{
  return cpl_node_ptr(new sub_node());
} /* sub_handler::allocate_node() */


/*
 * sub_handler::set_ref(cpl_parser*, const string&)
 * Handler function for the ref attribute of a <sub>.
 * (Static function)
 */
void sub_handler::set_ref(cpl_parser* parser, const string& value)
{
  sub_node* current_node =
    dynamic_cast_cpl_node_ptr<sub_node*>(parser->getCurrentNode(),
                                         "sub_handler::set_ref");

  current_node->ref_name = value;
} /* sub_handler::set_ref(cpl_parser*, const std::string&) */


/*
 * sub_handler::end(cpl_parser *) const
 * End a sub node.  Check to make sure ref_name is set, then lookup ref
 * in the CPL's subs map.  If it's present, set ref to it; otherwise, throw
 * an exception.
 */
void sub_handler::end(cpl_parser *parser) const
{
  sub_node* current_node =
    dynamic_cast_cpl_node_ptr<sub_node*>(parser->getCurrentNode(),
                                              "sub_handler::end");

  if (current_node->ref_name.length() == 0) {
    throw CPLException(CPLException::MissingAttr, "ref", parser);
  }

  map<string, cpl_node_ptr>::const_iterator it;

  it = parser->getCPL()->subs.find(current_node->ref_name);
  if (it == parser->getCPL()->subs.end()) {
    throw CPLException(CPLException::UnknownSub, current_node->ref_name,
                       parser);
  }

  current_node->ref = it->second;
} /* sub_handler::end(cpl_parser *) */



/*
 * Singleton handlers.
 */
cpl_toplevel_handler the_cpl_toplevel_handler;

cpl_cpl_handler the_cpl_cpl_handler;

cpl_ancillary_handler the_cpl_ancillary_handler;
cpl_subaction_handler the_cpl_subaction_handler;
cpl_outgoing_handler the_cpl_outgoing_handler;
cpl_incoming_handler the_cpl_incoming_handler;

sub_handler the_sub_handler;
