/* cpl-parse.cpp - parse a CPL tree using Xerces-C. */
/* Copyright 2001 Columbia University.  All rights reserved. */

// System headers
#include <strstream>

// Xerces headers
#include <sax2/Attributes.hpp>
#include <sax2/SAX2XMLReader.hpp>
#include <sax2/XMLReaderFactory.hpp>
#include <sax/SAXParseException.hpp>
#include <framework/MemBufInputSource.hpp>

// Local headers
#include "cpl.h"
#include "cpl-exception.h"
#include "cpl-nodes.h"
#include "cpl-parse.h"
#include "cpl-utils.h"

using namespace boost;
using namespace std;

/*
 * element_handler::element_handler(const string, const string)
 * Construct an element_handler.
 * Subclasses of this class will want to fill in the maps.
 */
element_handler::element_handler(const string s, const string n) :
  space(s), name(n)
{

} /* element_handler::element_handler(const string, const string) */


/*
 * element_handler::element_handler(const char *const, const char *const)
 * Construct an element_handler with const char* const names.
 * Subclasses of this class will want to fill in the maps.
 */
element_handler::element_handler(const char *const s, const char *const n) :
  space(s != NULL ? s : ""), name(n != NULL ? n : "")
{

} /* element_handler::element_handler(const char *const, const char *const) */


/*
 * element_handler::effective_namespace(const string&)
 * Given a namespace, get the "effective" namespace -- i.e. for the 
 * null namespace, return the default CPL namespace.
 */
const string element_handler::effective_namespace(const string& space)
{
  if (space.length() == 0) {
    return string(cpl::default_cpl_xml_namespace);
  }
  return space;
} /* element_handler::effective_namespace(const string&) */


/*
 * element_handler::get_attr_fn(const qualified_name&)
 * Get the attribute_fn for the given qualified attribute name from
 * the attribute table.  If there is no such fn, return NULL.
 */
attribute_fn element_handler::get_attr_fn(const qualified_name& name) const
{
  qualified_name effective_name = name;
  effective_name.first = effective_namespace(name.first);

  attr_map::const_iterator the_fn = att_map.find(effective_name);
  if (the_fn != att_map.end()) {
    return ((*the_fn).second);
  }

  return NULL;
} /* element_handler::get_attr_fn(const qualified_name&) */


/*
 * element_handler::get_child_handler(const qualified_name&);
 * Get a subelement handler for the element.
 * Check the subelement table; if the
 * given subelement has a function registered, call it and return the resulting
 * object.
 * Otherwise, return NULL.
 */
const element_handler*
element_handler::get_child_handler(const qualified_name& name) const
{
  qualified_name effective_name = name;
  effective_name.first = effective_namespace(name.first);

  child_map::const_iterator the_handler = subnode_map.find(effective_name);
  if (the_handler != subnode_map.end()) {
    return ((*the_handler).second);
  }

  return NULL;
} /* element_handler::get_child_handler(const qualified_name&) */


/*
 * element_handler::new_stack_frame() const
 * Allocate a new stack frame.  Element handlers which need subclassed
 * stack frames to maintain local parse state should override this function.
 */
cpl_parser_stackelement_ptr element_handler::new_stack_frame() const
{
  return cpl_parser_stackelement_ptr(new cpl_parser_stackelement(this));
} /* element_handler::new_stack_frame() const */


/*
 * element_handler::start()
 * Start an element.
 * This does nothing; derived classes should override to do something
 * (like allocating appropriate data) if they need to.
 */
void element_handler::start(cpl_parser* p) const
{
  // Do nothing
} /* element_handler::start() */


/*
 * element_handler::end()
 * End an element.
 * This does nothing; derived classes should override to do something
 * (like inserting their data into a parent) if they need to.
 */
void element_handler::end(cpl_parser* p) const
{
  // Do nothing
} /* element_handler::end() */


/*
 * cpl_parser_stackelement::cpl_parser_stackelement(const element_handler*)
 * Construct a CPL parser stack element.
 * Not defined in-line because cpl_node is a forward declaration in
 * cpl_parse.h.
 */
cpl_parser_stackelement::cpl_parser_stackelement(const element_handler *h) :
  handler(h)
{
  // Nothing else to do
} /* cpl_parser_stackelement::cpl_parser_stackelement(const element_handler*) */


/*
 * cpl_parser_stackelement::~cpl_parser_stackelement()
 * Destruct a CPL parser stack element.
 * Not defined in-line because cpl_node is a forward declaration in
 * cpl_parse.h.
 */
cpl_parser_stackelement::~cpl_parser_stackelement()
{
  // Nothing to do
} /* cpl_parser_stackelement::~cpl_parser_stackelement() */


/*
 * cpl_parser::getStackCurrent() 
 * Get the current parser stack frame.
 */
cpl_parser_stackelement_ptr cpl_parser::getStackCurrent()
{
  if (parse_stack.empty()) {
    throw CPLException(CPLException::InternalError, getLocator(),
                       "Stack underflow");
  }
  return parse_stack.back();
} /* cpl_parser::getStackCurrent() */


/*
 * cpl_parser::getStackParent()
 * Get the parent of the current parser stack frame -- one below current.
 * Have to "break in" to the underlying implementation type for this.
 */
cpl_parser_stackelement_ptr cpl_parser::getStackParent()
{
  if (parse_stack.size() < 2) {
    throw CPLException(CPLException::InternalError, getLocator(),
                       "Stack underflow getting parent");
  }
  return *(parse_stack.end() - 2);
} /* cpl_parser::getStackParent() */


/*
 * cpl_parser::getStackCurrent() const
 * Get a const pointer to the current parser stack frame.
 */
const cpl_parser_stackelement_ptr cpl_parser::getStackCurrent() const
{
  if (parse_stack.empty()) {
    throw CPLException(CPLException::InternalError, getLocator(),
                       "Stack underflow");
  }
  return parse_stack.back();
} /* cpl_parser::getStackCurrent() const*/


/*
 * cpl_parser::getStackParent() const
 * Get a const pointer to the parent of the current parser stack frame.
 */
const cpl_parser_stackelement_ptr cpl_parser::getStackParent() const
{
  if (parse_stack.size() < 2) {
    throw CPLException(CPLException::InternalError, getLocator(),
                       "Stack underflow getting parent");
  }
  return *(parse_stack.end() - 2);
} /* cpl_parser::getStackParent() const */


/*
 * cpl_parser::setParentCurrentOutput(cpl_node_ptr)
 * Set the stack's current parent frame's current output.
 */
void cpl_parser::setParentCurrentOutput(cpl_node_ptr p) {
  if (getStackParent()->current_output.get() != NULL) {
    const element_handler* parent_h = getStackParent()->getHandler();
    const element_handler* current_h = getStackCurrent()->getHandler();
    throw CPLException(CPLException::DuplicateChild,
                       getLocator(),
                       qualified_name(parent_h->getSpace(),
                                      parent_h->getName()),
                       qualified_name(current_h->getSpace(),
                                      current_h->getName()));
  }

  getStackParent()->current_output = p;
} /* cpl_parser::setParentCurrentOutput(cpl_node_ptr) */


/*
 * cpl_parser::getCurrentOutput()
 * Get the stack's current frame's current output.
 */
cpl_node_ptr cpl_parser::getCurrentOutput() {
  return getStackCurrent()->current_output;
} /* cpl_parser::getCurrentOutput() */


/*
 * cpl_parser::setCurrentNode(cpl_node_ptr)
 * Set the stack's current frame's current node.
 */
void cpl_parser::setCurrentNode(cpl_node_ptr p) {
  getStackCurrent()->current_node = p;
} /* cpl_parser::setCurrentNode(cpl_node_ptr) */


/*
 * cpl_parser::getCurrentNode()
 * Get the stack's current frame's current node.
 */
cpl_node_ptr cpl_parser::getCurrentNode() {
  return getStackCurrent()->current_node;
} /* cpl_parser::getCurrentNode() */


/*
 * cpl_parser::getParentNode()
 * Get the stack's parent frame's current node.
 */
cpl_node_ptr cpl_parser::getParentNode() {
  return getStackParent()->current_node;
} /* cpl_parser::getParentNode() */


/*
 * cpl_parser::parse(const InputSource&)
 * Parse the CPL script on the input stream, and return it as a cpl_ptr.
 */
cpl_ptr cpl_parser::parse(const InputSource& stream)
{
  cpl* dest_cpl = new cpl();

  parse(stream, dest_cpl);

  return cpl_ptr(dest_cpl);
} /* cpl_parser::parse(const InputSource&) */


/*
 * cpl_parser::parse(const InputSource&, cpl*)
 * Parse the CPL script on the input stream, and insert it into the
 * given cpl.  The result is undefined if cpl already contains anything.
 */
void cpl_parser::parse(const InputSource& stream, cpl* dest_cpl)
{
  try {
    SAX2XMLReader* saxparser = XMLReaderFactory::createXMLReader();
    /* TODO: Set features?  Or are the defaults correct? */

    cpl_sax_handler handler(this);
    saxparser->setContentHandler(&handler);
    saxparser->setErrorHandler(&handler);
    saxparser->setEntityResolver(&handler);

    the_cpl = dest_cpl;

    saxparser->parse(stream);
  }
  catch (SAXException& e) {
    throw CPLException(e);
  }
} /* cpl_parser::parse(const InputSource&, cpl*) */


/*
 * cpl_sax_handler::startDocument()
 * A CPL has just started being parsed.  Push a new stack frame onto
 * the stack, pointing to the toplevel handler.
 */
void 
cpl_sax_handler::startDocument()
{
  cpl_parser_stackelement_ptr frame =
    the_cpl_toplevel_handler.new_stack_frame();
  the_parser->parse_stack.push_back(frame);
} /* cpl_sax_handler::startDocument() */


/*
 * cpl_sax_handler::endDocument()
 * A CPL has finished being parsed.  Cool!
 * Pop the final stack frame.
 * TODO: confirm that the final stack frame is a toplevel_handler?
 */
void 
cpl_sax_handler::endDocument()
{
  the_parser->parse_stack.pop_back();
} /* cpl_sax_handler::endDocument() */


/*
 * cpl_sax_handler::startElement(const XMLCh* const, const XMLCh* const,
 *                          const XMLCh* const, const Attributes&)
 * Xerces-C has discovered a new XML element.  Tell the current stack frame
 * about it, and based on what it says, allocate a new stack frame.
 * Then tell the new stack frame about its attributes.
 */
void
cpl_sax_handler::startElement(const XMLCh* const uri,
                         const XMLCh* const localname,
                         const XMLCh* const qname,
                         const Attributes& attrs)
{
  qualified_name element_name(cpl_utils::transcode(uri),
                      cpl_utils::transcode(localname));
  const element_handler* current_handler, *next_handler;

  current_handler = the_parser->getStackCurrent()->getHandler();

  next_handler = current_handler->get_child_handler(element_name);

  if (next_handler == NULL) {
    throw CPLException(CPLException::InvalidChild,
                       the_parser->getLocator(),
                       element_name,
                       qualified_name(current_handler->getSpace(),
                                      current_handler->getName()));
  }
                       

  the_parser->parse_stack.push_back(next_handler->new_stack_frame());

  next_handler->start(the_parser);

  for (unsigned int i = 0; i < attrs.getLength(); i++) {
    qualified_name attr_name(cpl_utils::transcode(attrs.getURI(i)),
                             cpl_utils::transcode(attrs.getLocalName(i)));

    attribute_fn fn = next_handler->get_attr_fn(attr_name);

    if (fn == NULL) {
      throw CPLException(CPLException::InvalidAttr,
                         the_parser->getLocator(),
                         attr_name,
                         element_name);
    }

    (*fn)(the_parser, cpl_utils::transcode(attrs.getValue(i)));
  }
} /* cpl_sax_handler::startElement(const XMLCh* const, const XMLCh* const,
   *                          const XMLCh* const, const Attributes&) */


/*
 * cpl_sax_handler::endElement(const XMLCh* const, const XMLCh* const,
 *                        const XMLCh* const)
 * Xerces-C has finished an XML element.  Tell the current stack frame
 * that it's done, then pop it.
 */
void
cpl_sax_handler::endElement(const XMLCh* const uri,
                       const XMLCh* const localname,
                       const XMLCh* const qname)
{
  cpl_parser_stackelement_ptr current = the_parser->getStackCurrent();

  current->getHandler()->end(the_parser);

  the_parser->parse_stack.pop_back();
} /* cpl_sax_handler::endElement(const XMLCh* const, const XMLCh* const,
   *                        const XMLCh* const) */


/*
 * cpl_sax_handler::setDocumentLocator(const Locator *const)
 * Xerces-C wants to let us know where the next event we're about to receive
 * is located in the document.
 */
void cpl_sax_handler::setDocumentLocator(const Locator *const loc) {
  the_parser->locator = loc;
} /* cpl_sax_handler::setDocumentLocator(const Locator *const) */


/*
 * cpl_sax_handler::error(const SAXParseException&)
 * Xerces-C wants to let us know about a recoverable error during parsing.
 * Be paranoid -- throw a corresponding CPLException.
 */
void cpl_sax_handler::error(const SAXParseException& e) {
  throw CPLException(e);
} /* cpl_sax_handler::error(const SAXParseException&) */


/*
 * cpl_sax_handler::fatalError(const SAXParseException&)
 * Xerces-C wants to let us know about a non-recoverable error during parsing.
 * Throw a corresponding CPLException.
 */
void cpl_sax_handler::fatalError(const SAXParseException& e) {
  throw CPLException(e);
} /* cpl_sax_handler::fatalError(const SAXParseException&) */


/*
 * cpl_sax_handler::warning(const SAXParseException&)
 * Xerces-C wants to let us know about a warning during parsing.
 * For now, throw a corresponding CPLException.  Probably want to turn
 * this off for actual production code?
 */
void cpl_sax_handler::warning(const SAXParseException& e) {
  throw CPLException(e);
} /* cpl_sax_handler::warning(const SAXParseException&) */


/*
 * cpl_sax_handler::resolveEntity(const XMLCh* const, const XMLCh* const)
 * Resolve external entities.  In particular, "resolve" the default CPL
 * DTD into a null entity -- it's purely advisory.
 */
InputSource* cpl_sax_handler::resolveEntity(const XMLCh* const publicId,
                                            const XMLCh* const systemId)
{
  if (cpl_utils::transcode(publicId) == 
      cpl::default_cpl_xml_public_id) {
    return new MemBufInputSource(NULL, 0, "[Internal CPL DTD]");
  }

  return NULL;
} /* cpl_sax_handler::resolveEntity(const XMLCh* const, const XMLCh* const) */

