/* cpl-parse.h - parse CPL scripts. */
/* Copyright 2001 Columbia University.  All rights reserved. */

#ifndef CPL_PARSE_H_
#define CPL_PARSE_H_

#include <string>
#include <vector>
#include <boost/smart_ptr.hpp>
#include <sax2/DefaultHandler.hpp>
#include <sax/Locator.hpp>
#include <sax/SAXParseException.hpp>
#include <map>

#include "cpl.h"

/* Forward declarations */
class element_handler;
class cpl_parser;
class cpl_parser_stackelement;
typedef boost::shared_ptr<cpl_parser_stackelement> cpl_parser_stackelement_ptr;


/* Utility types */
typedef std::pair<std::string, std::string>     qualified_name;
typedef void (*attribute_fn)(cpl_parser*, const std::string&);

typedef std::map<qualified_name, attribute_fn>             attr_map;
typedef std::map<qualified_name, const element_handler*>   child_map;

/*
 * Global information about how to parse a CPL element (node, output, or
 * toplevel).  Do-nothing base class. 
 * Instantiate a subclass for specific elements.
 */
class element_handler {
private:
  attr_map  att_map;
  child_map subnode_map;

protected:
  const std::string space;
  const std::string name;

  void insert_att_map(const std::string&, const std::string&, attribute_fn);
  void insert_att_map(const std::string&, attribute_fn);
  void insert_att_map(const qualified_name&, attribute_fn);

  void insert_subnode_map(const std::string&, const std::string&,
                          const element_handler*);
  void insert_subnode_map(const std::string&, const element_handler*);
  void insert_subnode_map(const qualified_name&, const element_handler*);

  static const std::string effective_namespace(const std::string& space);

public:
  element_handler(const std::string space, const std::string name);
  element_handler(const char *const space, const char *const name);
  virtual ~element_handler() { }

  virtual attribute_fn get_attr_fn(const qualified_name&) const;
  virtual const element_handler* get_child_handler(const qualified_name&)
    const;

  virtual cpl_parser_stackelement_ptr new_stack_frame() const;

  virtual void start(cpl_parser*) const;
  virtual void end(cpl_parser*) const;

  virtual const std::string getSpace() const { return space; }

  virtual const std::string getName() const { return name; }
};


/*
 * A stack frame for parsing the current CPL node.
 * Parsing for specific node types may need to subclass this.
 */
class cpl_parser_stackelement {
protected:
  const element_handler* handler;

public:
  cpl_node_ptr current_node;
  cpl_node_ptr current_output;

  cpl_parser_stackelement(const element_handler * h);

  const element_handler* getHandler() const { return handler; }

  virtual ~cpl_parser_stackelement();
};

typedef std::vector<cpl_parser_stackelement_ptr> cpl_parser_stack;

/*
 * Parse a CPL script.  This implements Xerces-C SAX2 function callbacks.
 */
class cpl_parser {
private:
  cpl_parser_stack parse_stack;

  cpl* the_cpl;

  const Locator* locator;

public:
  const Locator* const getLocator() const { return locator; }

  cpl* getCPL() const { return the_cpl; }

  void setParentCurrentOutput(cpl_node_ptr);
  cpl_node_ptr getCurrentOutput();

  void setCurrentNode(cpl_node_ptr);
  cpl_node_ptr getCurrentNode();
  cpl_node_ptr getParentNode();

  cpl_parser_stackelement_ptr getStackCurrent();
  cpl_parser_stackelement_ptr getStackParent();

  const cpl_parser_stackelement_ptr getStackCurrent() const;
  const cpl_parser_stackelement_ptr getStackParent() const;

  cpl_ptr parse(const InputSource& source);
  void parse(const InputSource& source, cpl* dest);

  friend class cpl_sax_handler;
};

/*
 * Xerces-C SAX2 handler for CPL parsing.
 */
class cpl_sax_handler : public DefaultHandler {
private:
  cpl_parser* the_parser;

public:
  cpl_sax_handler(cpl_parser *p) : the_parser(p) { }

  void startDocument();
  void endDocument();

  void startElement(const XMLCh* const uri,
                    const XMLCh* const localname,
                    const XMLCh* const qname,
                    const Attributes& attrs);

  void endElement(const XMLCh* const uri,
                  const XMLCh* const localname,
                  const XMLCh* const qname);

  void setDocumentLocator(const Locator *const loc);

  void error(const SAXParseException&);
  void fatalError(const SAXParseException&);
  void warning(const SAXParseException&);

  InputSource* resolveEntity(const XMLCh* const publicID,
                             const XMLCh* const systemID);
};


/* Inline functions */

/*
 * element_handler::insert_att_map(const std::string&, const std::string&, attribute_fn)
 * Insert the given attribute, in the given namespace, into the attribute map.
 * Typically called from initialize() in a derived class.
 */
inline
void element_handler::insert_att_map(const std::string& space,
                                     const std::string& name, attribute_fn f)
{
  insert_att_map(qualified_name(space, name), f);
} /* element_handler::insert_att_map(const std::string&, const std::string&,
   *                                 attribute_fn) */


/*
 * element_handler::insert_att_map(const std::string&, attribute_fn)
 * Insert the given attribute, in the default CPL namespace, into the
 * attribute map.
 * Typically called from initialize() in a derived class.
 */
inline
void element_handler::insert_att_map(const std::string& name, attribute_fn f)
{
  insert_att_map(cpl::default_cpl_xml_namespace, name, f);
} /* element_handler::insert_att_map(const std::string&, attribute_fn) */


/*
 * element_handler::insert_att_map(const qualified_name, attribute_fn)
 * Insert the given attribute namespace/name pair into the attribute map.
 * Typically called from initialize() in a derived class.
 */
inline
void element_handler::insert_att_map(const qualified_name& key,
                                     attribute_fn f) {
  att_map[key] = f;
} /* element_handler::insert_att_map(const qualified_name,
   *                                 attribute_fn) */


/*
 * element_handler::insert_subnode_map(const std::string&, const std::string&, 
 *                                     const element_handler*)
 * Insert the given subnode, in the given namespace, into the subnode map.
 * Typically called from initialize() in a derived class.
 */
inline
void element_handler::insert_subnode_map(const std::string& space,
                                         const std::string& name,
                                         const element_handler* h)
{
  insert_subnode_map(qualified_name(space, name), h);
} /* element_handler::insert_subnode_map(const std::string&, const std::string&,
   *                                     const element_handler*) */


/*
 * element_handler::insert_subnode_map(const std::string&, const element_handler*)
 * Insert the given subnode, in the default CPL namespace, into the
 * subnode map.
 * Typically called from initialize() in a derived class.
 */
inline
void element_handler::insert_subnode_map(const std::string& name,
                                         const element_handler* h)
{
  insert_subnode_map(cpl::default_cpl_xml_namespace, name, h);
} /* element_handler::insert_subnode_map(const std::string&, const element_handler*) */


/*
 * element_handler::insert_subnode_map(const qualified_name&,
 *                                     const element_handler*)
 * Insert the given subnode namespace/name pair into the subnode map.
 * Typically called from initialize() in a derived class.
 */
inline
void element_handler::insert_subnode_map(const qualified_name& key,
                                         const element_handler* f) {
  subnode_map[key] = f;
} /* element_handler::insert_subnode_map(const qualified_name&,
   *                                     const element_handler*) */





#endif /* CPL_PARSE_H_ */
