/* cpl-templatefns.h - template functions useful for CPL. */
/* Copyright 2001 Columbia University.  All rights reserved. */

#ifndef CPL_TEMPLATEFNS_H_
#define CPL_TEMPLATEFNS_H_

#include <strstream.h>
#include "cpl-parse.h"
#include "cpl-exception.h"

/*
 * dynamic_cast_cpl_node_ptr<type>(cpl_node_ptr, string)
 * Dynamic-cast a cpl_node_ptr to a plain pointer to a specific subclass of
 * cpl_node.  If it's the wrong type, throw a CPLException::InternalError.
 */
template <class Tgt_Type>
Tgt_Type dynamic_cast_cpl_node_ptr(cpl_node_ptr p, std::string context)
{
  Tgt_Type dst = dynamic_cast<Tgt_Type>(p.get());
  if (dst == NULL) {
    throw CPLException(CPLException::InternalError,
                       ("cpl_node_ptr has incorrect subclass "
                        "in " + context).c_str());
  }
  return dst;
} /* dynamic_cast_cpl_node_ptr<type>(cpl_node_ptr, string) */


/*
 * dynamic_cast_cpl_node_ptr<type>(const cpl_node *, string)
 * Dynamic-cast a const cpl_node* to a plain pointer to a specific subclass of
 * cpl_node.  If it's the wrong type, throw a CPLException::InternalError.
 */
template <class Tgt_Type>
Tgt_Type dynamic_cast_cpl_node_ptr(const cpl_node * p, std::string context)
{
  Tgt_Type dst = dynamic_cast<Tgt_Type>(p);
  if (dst == NULL) {
    throw CPLException(CPLException::InternalError,
                       ("cpl_node_ptr has incorrect subclass "
                        "in " + context).c_str());
  }
  return dst;
} /* dynamic_cast_cpl_node_ptr<type>(cpl_node_ptr, string) */


/*
 * dynamic_cast_stackelement_ptr<type>(cpl_parser_stackelement_ptr,
 *                                     const std::string&)
 * Dynamic-cast a cpl_parser_stackelement_ptr to a plain pointer to a
 * specific subclass of cpl_parser_stackelement.  If it's the wrong type,
 * throw a CPLException::InternalError.
 */
template <class Tgt_Type>
Tgt_Type dynamic_cast_stackelement_ptr(cpl_parser_stackelement_ptr p,
                                       const std::string& context)
{
  Tgt_Type dst = dynamic_cast<Tgt_Type>(p.get());
  if (dst == NULL) {
    throw CPLException(CPLException::InternalError,
                       ("cpl_parser_stackelement_ptr has incorrect subclass "
                        "in " + context).c_str());
  }
  return dst;
} /* dynamic_cast_stackelement_ptr<type>(cpl_parser_stackelement_ptr,
   *                                     const std::string&) */


/*
 * get_map_value<type>(const std::map<std::string, type>&, const std::string&,
 *                     cpl_parser *, const std::string&)
 * Given a value-as-string and a map, return the value-as-type.  If there is
 * none, throw a CPL MalformedAttr exception.
 */
template <class Tgt_Type>
Tgt_Type get_map_value(const std::map<std::string, Tgt_Type>& valmap,
                       const std::string& value,
                       cpl_parser *the_parser,
                       const std::string& att_name)
{
  std::map<std::string, Tgt_Type>::const_iterator it;

  it = valmap.find(value);
  if (it == valmap.end()) {
    throw CPLException(CPLException::MalformedAttr, value, att_name,
                       the_parser);
  }

  return it->second;
} /* get_map_value<type>(const std::map<std::string, type>&,
   *                     const std::string&,
   *                     cpl_parser *, const std::string&) */


/*
 * read_string<type>(const std::string&, cpl_parser *, const std::string&)
 * Read a 'type' from 'value'.  If operator>>(istream, type) does not
 * entirely consume 'value', throw a CPL exception.
 */
template <class Tgt_Type>
Tgt_Type read_string(const std::string& value, cpl_parser *parser,
                     const std::string& att_name)
{
  istrstream bufstr(value.c_str());
  Tgt_Type retval;
  char dummy; /* Try to read one extra character off the end.  If this
               * succeeds, value had garbage after retval. */

  if (!(bufstr >> retval) || (bufstr >> dummy)) {
    throw CPLException(CPLException::MalformedAttr, value, att_name,
                       parser);
  }
  return retval;
} /* read_string<type>(const std::string&, cpl_parser *, const std::string&) */

#endif /* CPL_TEMPLATEFNS_H_ */
