/* cpl.cpp - externally-accessible methods operating on a CPL. */
/* Copyright 2001 Columbia University.  All rights reserved. */

#include <sax/InputSource.hpp>
#include <framework/URLInputSource.hpp>
#include <framework/LocalFileInputSource.hpp>
#include <framework/MemBufInputSource.hpp>
#include <util/XMLString.hpp>

#include <util/PlatformUtils.hpp>

#include <cstring>

#include "cpl.h"
#include "cpl-parse.h"
#include "cpl-nodes.h"
#include "cpl-exception.h"

using namespace std;

/* Static data members of cpl */ 
const char cpl::default_cpl_xml_public_id[];
const char cpl::default_cpl_xml_namespace[];

/*
 * cpl::cpl()
 * Construct a new, empty CPL.
 * Nothing needs to be explicitly done here, but this needs to be defined
 * out-of-line to get the implicit constructors right, because cpl_node isn't
 * defined in cpl.h.
 */
cpl::cpl()
{
  /* Nothing to be done */
} /* cpl::cpl() */

/*
 * cpl::cpl(const char*)
 * Construct a CPL from a file or URI.
 */
cpl::cpl(const char *src)
{
  internal_parse(src);
} /* cpl::cpl(const char*) */


/*
 * cpl::cpl(const unsigned char*, size_t, bufname)
 * Construct a CPL from an in-memory buffer.
 */
cpl::cpl(const unsigned char *buf, size_t len, const char* bufname)
{
  internal_parse(buf, len, bufname);
} /* cpl::cpl(const unsigned char*, size_t, bufname) */


/*
 * cpl::~cpl()
 * Destruct a CPL.
 * I think nothing to be explicitly done here, but see cpl::cpl().
 */
cpl::~cpl()
{
  /* Nothing to be done */
} /* cpl::~cpl() */


/*
 * cpl::clear()
 * Reset a CPL back to its null state (as though it had been initialized
 * by the default constructor).
 */
void cpl::clear()
{
  subs.clear();
  incoming.reset();
  outgoing.reset();
} /* cpl::clear() */


/*
 * cpl::parse(char *)
 * Parse a file or URI.  Reset the CPL to the value parsed.
 */
void cpl::parse(const char *src)
{
  clear();

  internal_parse(src);
} /* cpl::parse(char *) */


/*
 * cpl::internal_parse(char *)
 * Parse a file or URI.
 */
void cpl::internal_parse(const char *src)
{
  InputSource* source = NULL;

  try {
    if (strchr(src, ':') != NULL) {
      source = new URLInputSource(src);
    }
    else {
      source = new LocalFileInputSource(XMLString::transcode(src));
    }

    internal_parse(*source);
  }
  catch (XMLException& e) {
    delete source;
    throw CPLException(e);
  }
  catch (...) {
    delete source;
    throw;
  }
} /* cpl::internal_parse(char *) */


/*
 * cpl::parse(const unsigned char*, size_t, const char*)
 * Parse an in-memory buffer.  Reset the CPL to the value parsed.
 */
void cpl::parse(const unsigned char* buf, size_t len, const char* bufname)
{
  clear();

  internal_parse(buf, len, bufname);
} /* cpl::parse(const unsigned char*, size_t, const char*) */


/*
 * cpl::internal_parse(const unsigned char*, size_t, const char *)
 * Parse an in-memory buffer.
 */
void cpl::internal_parse(const unsigned char* buf, size_t len,
                         const char* bufname)
{
  try {
    MemBufInputSource source(buf, len, bufname);

    internal_parse(source);
  }
  catch (XMLException& e) {
    throw CPLException(e);
  }
} /* cpl::internal_parse(const unsigned char*, size_t, const char*) */


/*
 * cpl::internal_parse(InputSource&)
 * Parse a CPL, from InputSource.  Internal (protected) function
 * called by all other parsers and constructors.
 */
void cpl::internal_parse(InputSource& source)
{
  cpl_parser parser;

  parser.parse(source, this);
} /* cpl::internal_parse(InputSource&) */


/*
 * class-static global variable
 */
bool cpl::initialized = false;

/*
 * cpl::initialize()
 * Statically initialize the CPL environment.
 */
void cpl::initialize()
{
  if (initialized) return;

  XMLPlatformUtils::Initialize();
  init_cpl_node_data();
  
  initialized = true;
} /* cpl::initialize() */


/*
 * cpl::cleanup()
 * Statically clean up the CPL environment.
 */
void cpl::cleanup()
{
  if (!initialized) return;

  XMLPlatformUtils::Terminate();
  free_cpl_node_data();

  initialized = false;
} /* cpl::cleanup() */
  
