/* calendar.cpp - Object-based wrapper around a 'struct tm', based on Java's
 * java.util.GregorianCalendar. */
/* Copyright 2001 Columbia University.  All rights reserved. */

#include "calendar.h"
#include "tztimeext.h"

#include <cstdlib>
#include <cerrno>

#include <stdexcept>
#include <strstream>

using namespace std;

/*
 * Calendar static const data members
 */
const int Calendar::DAYSPERWEEK;
const int Calendar::DAYSPERNYEAR;
const int Calendar::DAYSPERLYEAR;

/*
 * Calendar::Calendar()
 * Construct a new default calendar.  The current time is the epoch;
 * the timezone is the current getenv("TZ").
 */
Calendar::Calendar()
{
  init();

  the_time = 0;
  setTimeZone(getenv("TZ"));

  set_from_time_t();
} /* Calendar::Calendar() */


/*
 * Calendar::Calendar(time_t)
 * Construct a new calendar with the specified time.  The timezone is the
 * current getenv("TZ").
 */
Calendar::Calendar(time_t t)
{
  init();

  the_time = t;
  setTimeZone(getenv("TZ"));

  set_from_time_t();
} /* Calendar::Calendar(time_t) */


/*
 * Calendar::Calendar(string)
 * Construct a new calendar.  The time zone is the given string; the current
 * time is the epoch.
 */
Calendar::Calendar(const char * tzn)
{
  init();

  the_time = 0;
  setTimeZone(tzn);

  set_from_time_t();
} /* Calendar::Calendar(string) */


/*
 * Calendar::Calendar(time_t, const char *)
 * Construct a new calendar with the given time_t and time zone.
 */
Calendar::Calendar(time_t t, const char * tzn)
{
  init();

  the_time = t;
  setTimeZone(tzn);

  set_from_time_t();
} /* Calendar::Calendar(time_t, const char *) */


/*
 * Calendar::Calendar(int, Month, int, int, int, const char *)
 * Construct a new calendar with the given time_t and time zone.
 */
Calendar::Calendar(int y, Month mo, int d, int h, int m, int s)
{
  init();

  setTimeZone(getenv("TZ"));

  set(YEAR, y);
  set(MONTH, mo);
  set(DAY_OF_MONTH, d);
  set(HOUR_OF_DAY, h);
  set(MINUTE, m);
  set(SECOND, s);

  set_and_normalize();
} /* Calendar::Calendar(int, Month, int, int, int, const char *) */


/*
 * Calendar::init()
 * Set initial Calendar fields which aren't set by default constructors.
 */
void Calendar::init()
{
  firstDayOfWeek = MONDAY; // Defaults from ISO
  minimalDaysInFirstWeek = 4;

  memset(&tm, 0, sizeof(tm));
  tm.tm_isdst = -1;
  the_time = -1;

  normalized = false;
} /* Calendar::init() */
 

/*
 * Calendar::set_and_normalize()
 * Normalize 'tm', and set 'the_time'.  Throw an exception if
 * the time isn't representable as a time_t.
 */
void Calendar::set_and_normalize() const
{
  if (normalized)
    return;

  int err;

  err = time_make(const_cast<time_t*>(&the_time),
                  const_cast<struct tm*>(&tm),
                  *tz);
  if (err != 0) {
    if (err == EINVAL || err == ERANGE) {
      throw calendar_range_error(strerror(err));
    }
    else {
      throw calendar_internal_error(strerror(err));
    }
  }

  normalized = true;
} /* Calendar::set_and_normalize() */


/*
 * Calendar::set_from_time_t()
 * the_time is valid for this calendar object.  Set tm.
 */
void Calendar::set_from_time_t()
{
  int err;

  err = time_breakup(&tm, &the_time, *tz);
  if (err != 0) {
    ostrstream buf;
    
    buf << "Error calculating time: " << strerror(err) << ends;

    /* Shouldn't happen */
    calendar_internal_error e(buf.str());
    buf.rdbuf()->freeze(0);
    throw e;
  }

  normalized = true;
} /* Calendar::set_from_time_t() */


/*
 * Calendar::setTimeZone(const char *)
 * Fill in 'tz' based on 'tzname'.
 * Loses tm.tm_isdst state, naturally.
 */
void Calendar::setTimeZone(const char * tzn)
{
  TimeZone* tmp = new TimeZone(tzn);

  tz.reset(tmp);
  tm.tm_isdst = -1;

  normalized = false;
} /* Calendar::setTimeZone() */


/*
 * Calendar::getFirstDayOfWeek()
 * Return the calendar's current registered first day of the week.
 */
Calendar::Weekday Calendar::getFirstDayOfWeek() const
{
  return firstDayOfWeek;
} /* Calendar::getFirstDayOfWeek() */


/*
 * Calendar::setFirstDayOfWeek(Weekday)
 * Set the calendar's first day of the week.
 * This does not cause 'normalized' to become false, since week-based
 * things aren't stored in struct tm.
 */
void Calendar::setFirstDayOfWeek(Weekday w)
{
  if (w != SUNDAY && w != MONDAY && w != TUESDAY && w != WEDNESDAY &&
      w != THURSDAY && w != FRIDAY && w != SATURDAY) {
    throw calendar_range_error("Invalid first day of week");
  }
  firstDayOfWeek = w;
} /* Calendar::setFirstDayOfWeek(Weekday) */



/*
 * Calendar::getMinimalDaysInFirstWeek()
 * Get the minimal number of days in the first week of the year.  Used
 * in ISO terminology to define week 1.
 */
int Calendar::getMinimalDaysInFirstWeek() const
{
  return minimalDaysInFirstWeek;
} /* Calendar::getMinimalDaysInFirstWeek() */


/*
 * Calendar::setMinimalDaysInFirstWeek(int)
 * Set the minimal number of days in the first week of the year.  Used
 * in ISO terminology to define week 2.
 * This does not cause 'normalized' to become false, since week-based
 * things aren't stored in struct tm.
 */
void Calendar::setMinimalDaysInFirstWeek(int d)
{
  if (d < 1 || d > 7) {
    throw calendar_range_error("Invalid number of days for first week");
  }
  minimalDaysInFirstWeek = d;
} /* Calendar::setMinimalDaysInFirstWeek(int) */


/*
 * Calendar::getTime()
 * Return the current time.
 */
time_t Calendar::getTime() const
{
  set_and_normalize();

  return the_time;
} /* Calendar::getTime() */


/*
 * Calendar::setTime(time_t)
 * Set the calendar's current time.
 */
void Calendar::setTime(time_t t)
{
  the_time = t;
  set_from_time_t();
} /* Calendar::setTime(time_t) */


/*
 * Calendar::get(Field) const
 * Return field 'f' of the calendar.
 */
int Calendar::get(Field f) const
{
  set_and_normalize();

  switch (f) {
    /* Fields in struct tm */
  case SECOND:
    return tm.tm_sec;
  case MINUTE:
    return tm.tm_min;
  case HOUR_OF_DAY:
    return tm.tm_hour;
  case DAY_OF_WEEK:
    return tm.tm_wday;
  case DAY_OF_MONTH: /* eq. DATE */
    return tm.tm_mday;
  case DAY_OF_YEAR:
    return tm.tm_yday + 1;
  case MONTH:
    return tm.tm_mon;
  case YEAR:
    return tm.tm_year + 1900;

    /* Fields that must be calculated */
  case WEEK_OF_MONTH:
    return calc_week_of_month();
  case WEEK_OF_YEAR:
    return calc_week_of_year();
  default:
    throw calendar_range_error("Get: Unknown field");
  }
} /* Calendar::get(Field) const */


/*
 * Calendar::calc_week_of_period() const
 * Calculate the current week (ISO definition) of an enclosing period
 * (either a month or a year) that a given day falls on.  To calculate this,
 * all we need is the current day of the week, the number of the day within
 * that period (counting from 0), the first day of the week, and the number
 * of days in a first week.  The first two are passed as arguments; the
 * latter two are data members of Calendar.
 * Return the week of the period; 0 if during a fractional first week
 * of fewer than minimalDaysInFirstWeek days.  */
int Calendar::calc_week_of_period(int wday, int pday) const
{
  /* Determine the "relative" day of the week of the first day of the
   * period.  (Relative DOW is firstDayOfWeek=0 .. 6.*/
  int rel_1st = (wday - firstDayOfWeek - pday) % 7;
  /* And force it to be positive */
  if (rel_1st < 0) rel_1st += 7;

  /* Compute the week of the period, disregarding the possibly fractional
   * first week. */
  int week = (pday + rel_1st) / 7;

  /* Check how many days fall during the possibly fractional first week.  If
   * it is at least minimalDaysInFirstWeek, add 1 to all week numbers.
   * The number of days in the fractional first week is 7 - rel_1st. */
  if ((7 - rel_1st) >= minimalDaysInFirstWeek)
    ++week;

  return week;
} /* Calendar::calc_week_of_period() const */


/*
 * Calendar::calc_week_of_month() const
 * Calculate the current week of the month (ISO definition)
 */
int Calendar::calc_week_of_month() const
{
  if (!normalized) {
    throw calendar_internal_error("Not normalized in calc_week_of_year");
  }

  /* Week_of_month is defined to be 0 during the fractional first week,
   * so we map directly onto calc_week_of_period. */

  return calc_week_of_period(tm.tm_wday, tm.tm_mday - 1);
} /* Calendar::calc_week_of_month() const */


/*
 * Calendar::calc_week_of_year() const
 * Calculate the current week of the year (ISO definition).
 * Algorithm (though not code) from GNU Classpath.
 */
int Calendar::calc_week_of_year() const
{
  if (!normalized) {
    throw calendar_internal_error("Not normalized in calc_week_of_year");
  }

  /* Calculate the week within this year. */
  int week = calc_week_of_period(tm.tm_wday, tm.tm_yday);

  /* Two corrections may now be necessary for fractional weeks.  We can
   * be in the last week of the previous year, or the first week of the
   * next year. */
  if (week == 0) {
    /* We're in the last week of the previous year.  This may be either
     * week 52 or 53, so re-run through calc_week_of_period. */
    week =
      calc_week_of_period(tm.tm_wday,
                          tm.tm_yday + yearlengthBASE(tm.tm_year - 1));
  }
  else if (tm.tm_yday > 357) {
    /* We might be in the first week of the next year. */
    /* Check if we're in the last (possibly fractional) week of the year,
     * and if the last week of the year has fewer than minimalDaysInFirstWeek
     * days. */
    /* We're in the last week of the year if we're within (7 - rel_day) days
     * of the last day of the year. */
    /* The number of days in the last week of the year is rel_last + 1. */
    
    /* Calculate our relative day. */
    int rel_day = (tm.tm_wday - firstDayOfWeek + 7) % 7;

    /* Calculate the relative day of the last day of the year. */
    int rel_last = (rel_day + (yearlengthBASE(tm.tm_year) - 1 - tm.tm_yday))
        % 7;
    
    if ((tm.tm_yday + (7 - rel_day) > yearlengthBASE(tm.tm_year)) &&
        (rel_last + 1 < minimalDaysInFirstWeek)) {
      week = 1;
    }
  }

  return week;
} /* Calendar::calc_week_of_year() const */


/*
 * Calendar::set(Field, int)
 * Set field 'f' of the calendar to v.
 * Loses tm.tm_isdst state.
 */
void Calendar::set(Field f, int v)
{
  normalized = false;
  tm.tm_isdst = -1;

  switch (f) {
    /* Fields in struct tm */
  case SECOND:
    tm.tm_sec = v;
    break;
  case MINUTE:
    tm.tm_min = v;
    break;
  case HOUR_OF_DAY:
    tm.tm_hour = v;
    break;
  case DATE: /* eq. DAY_OF_MONTH */
    tm.tm_mday = v;
    break;
  case MONTH:
    tm.tm_mon = v;
    break;
  case YEAR:
    tm.tm_year = v - 1900;
    break;

    /* Fields that must be calculated */
  case DAY_OF_WEEK:
  case WEEK_OF_MONTH:
  case WEEK_OF_YEAR:
  case DAY_OF_YEAR:
    throw calendar_not_implemented("Calculated set not supported");
  default:
    throw calendar_range_error("Set: Unknown field");
  }
} /* Calendar::set(Field) */


/*
 * Calendar::clear(Field)
 * Clear field 'f' of the calendar.
 */
void Calendar::clear(Field f)
{
  switch (f) {
    /* Fields in struct tm */
  case SECOND:
  case MINUTE:
  case HOUR_OF_DAY:
  case DATE: /* eq. DAY_OF_MONTH */
  case MONTH:
  case YEAR:
    set(f, 0);
    break;

    /* Fields that must be calculated */
  case DAY_OF_WEEK:
  case WEEK_OF_MONTH:
  case WEEK_OF_YEAR:
  case DAY_OF_YEAR:
    /* Just re-set normalization */
    normalized = false;
    break;

  default:
    throw calendar_range_error("Clear: Unknown field");
  }
} /* Calendar::clear(Field) const */


/*
 * Calendar::add(Field, int)
 * Add 'inc' to field 'f' of the calendar.
 * Increments of units of a day or larger lose tm.tm_isdst state, so that
 * a day after 9:00 AM October 25 1997 EDT is 9:00 AM October 26 1997 EST.
 * Increments of less than a day keep this state, which is then re-normalized
 * by time_make, so that an hour after 1:30 AM October 26 1997 EDT is
 * "2:30 AM October 26 1997 EDT", or, normalized, 1:30 AM October 26 1997 EST.
 */
void Calendar::add(Field f, int increment)
{
  if (increment == 0) return;
  set_and_normalize();

  switch(f) {
    /* Fields less than a day */
  case YEAR:
    tm.tm_year += increment;
    tm.tm_isdst = -1;
    break;

  case MONTH:
    tm.tm_mon += increment;
    tm.tm_isdst = -1;
    break;

  case WEEK_OF_YEAR:
  case WEEK_OF_MONTH:
    // case DAY_OF_WEEK_IN_MONTH:
    tm.tm_mday += 7 * increment;
    tm.tm_isdst = -1;
    break;
    
  case DAY_OF_MONTH:
  case DAY_OF_YEAR:
  case DAY_OF_WEEK:
    tm.tm_mday += increment;
    tm.tm_isdst = -1;
    break;

  case HOUR_OF_DAY:
    // case HOUR:
    tm.tm_hour += increment;
    break;

  case MINUTE:
    tm.tm_min += increment;
    break;

  case SECOND:
    tm.tm_sec += increment;
    break;
    
  default:
    throw calendar_range_error("Add: Unknown field");
  }

  normalized = false;
  // Re-normalize
  set_and_normalize();
} /* Calendar::add(Field, int) */


/*
 * Calendar::getMaximum(Field) const
 * Get the maximum value that field f can have.
 */
int Calendar::getMaximum(Field f) const
{
  switch (f) {
  case SECOND:
    return 59; // Ignoring leap seconds
  case MINUTE:
    return 59;
  case HOUR_OF_DAY:
    return 23;
  case DAY_OF_WEEK:
    return SATURDAY;
  case DAY_OF_MONTH:
    return 31;
  case MONTH:
    return DECEMBER;
  case YEAR:
    return 2038;
  case WEEK_OF_MONTH:
    return 6;
  case WEEK_OF_YEAR:
    return 53;
  case DAY_OF_YEAR:
    return 366;
  default:
    throw calendar_range_error("GetMaximum: Unknown field");
  }
} /* Calendar::getMaximum(Field) const */


/*
 * Calendar::getLeastMaximum(Field) const
 * Get the least maximum value that field f can have.
 * (Ignoring the effects of the maximum representable time.)
 */
int Calendar::getLeastMaximum(Field f) const
{
  switch (f) {
  case DAY_OF_MONTH:
    return 28;
  case WEEK_OF_MONTH:
    return 4;
  case WEEK_OF_YEAR:
    return 52;
  case DAY_OF_YEAR:
    return 365;
  case SECOND:
  case MINUTE:
  case HOUR_OF_DAY:
  case DAY_OF_WEEK:
  case MONTH:
  case YEAR:
    return getMaximum(f);
  default:
    throw calendar_range_error("GetLeastMaximum: Unknown field");
  }
} /* Calendar::getMaximum(Field) const */


/*
 * Calendar::getActualMaximum(Field) const
 * Get the actual maxmum value for field f, if all other fields are held
 * constant.
 * XXX: Ignores the effects of the minimum representable time.
 */
int Calendar::getActualMaximum(Field f) const
{
  set_and_normalize();

  switch (f) {
  case DAY_OF_MONTH:
    if (tm.tm_mon == FEBRUARY)
      return isleap(tm.tm_year + 1900) ? 29 : 28;
    else if (tm.tm_mon < AUGUST)
      return 31 - (tm.tm_mon & 1);
    else 
      return 30 + (tm.tm_mon & 1);
  case DAY_OF_YEAR:
    return yearlengthBASE(tm.tm_year);

  case SECOND:
  case MINUTE:
  case HOUR_OF_DAY:
  case DAY_OF_WEEK:
  case MONTH:
  case YEAR:
    return getMaximum(f);

  case WEEK_OF_MONTH:
  case WEEK_OF_YEAR:
    throw calendar_not_implemented("Weird getActualMaximum not supported");

  default:
    throw calendar_range_error("GetActualMaximum: Unknown field");
  }
} /* Calendar::getMaximum(Field) const */


/*
 * Calendar::getMinimum(Field) const
 * Get the minimum value that field f can have.
 */
int Calendar::getMinimum(Field f) const
{
  switch (f) {
  case SECOND:
    return 0;
  case MINUTE:
    return 0;
  case HOUR_OF_DAY:
    return 0;
  case DAY_OF_WEEK:
    return SUNDAY;
  case DAY_OF_MONTH:
    return 0;
  case MONTH:
    return JANUARY;
  case YEAR:
    return 1901;
  case WEEK_OF_MONTH:
    return 0;
  case WEEK_OF_YEAR:
    return 1;
  case DAY_OF_YEAR:
    return 1;
  default:
    throw calendar_range_error("GetMinimum: Unknown field");
  }
} /* Calendar::getMinimum(Field) const */


/*
 * Calendar::getGreatestMinimum(Field) const
 * Get the greatest minimum value that field f can have.
 * (Ignoring the effects of the minimum representable time.)
 */
int Calendar::getGreatestMinimum(Field f) const
{
  return getMinimum(f);
} /* Calendar::getGreatestMinimum(Field) const */


/*
 * Calendar::getActualMinimum(Field) const
 * Get the actual minimum value for field f, if all other fields are held
 * constant.
 * XXX: Ignores the effects of the minimum representable time.
 */
int Calendar::getActualMinimum(Field f) const
{
  return getMinimum(f);
} /* Calendar::getActualMinimum(Field) const */


/**
 * Return a string representation of this calendar object. This method 
 * is intended to be used only for debugging purposes.
 * The returned string may be empty but may not be <code>NULL</code>.
 * 
 * @return  A string representation of this recurrence.
 */
string Calendar::toString() const
{
  ostrstream buffer;
  
  buffer << "Calendar";
  buffer << "[tm=tm[tm_sec=" << tm.tm_sec;
  buffer << ",tm_min=" << tm.tm_min;
  buffer << ",tm_min=" << tm.tm_min;
  buffer << ",tm_hour=" << tm.tm_hour;
  buffer << ",tm_mday=" << tm.tm_mday;
  buffer << ",tm_mon=" << tm.tm_mon;
  buffer << ",tm_year=" << tm.tm_year;
  buffer << ",tm_wday=" << tm.tm_wday;
  buffer << ",tm_yday=" << tm.tm_yday;
  buffer << ",tm_isdst=" << tm.tm_isdst;
  buffer << "],tz=" << *tz;
  buffer << ",the_time=" << the_time;
  buffer << ",normalized=" << normalized ? "true" : "false";
  buffer << ",firstDayOfWeek=" << firstDayOfWeek;
  buffer << ",minimalDaysInFirstWeek=" << minimalDaysInFirstWeek;
  buffer << "]";

  buffer << ends;
  
  string ret(buffer.str());
  buffer.rdbuf()->freeze(0);

  return ret;
}


/*
 * Calendar::format(const char *)
 * Return a string corresponding to the Calendar object, formatted (as
 * with strftime().)
 */
string Calendar::format(const char * fmt) const
{
  char *buf;
  size_t len = 0;

  set_and_normalize();

  /* Get the necessary length */
  buf = new char[len];
  len = strftime_z(buf, len, fmt, &tm, *tz);

  /* Actually fill in the buffer */
  delete[] buf;
  buf = new char[len];
  len = strftime_z(buf, len, fmt, &tm, *tz);

  string ret(buf);
  delete[] buf;

  return ret;
} /* Calendar::format(const char *) */
