/* calendar.h - Object-based wrapper around a 'struct tm'. */
/* Copyright 2001 Columbia University.  All rights reserved. */

#ifndef CALENDAR_H_
#define CALENDAR_H_

#include <time.h>
#include <string>
#include <exception>
#include <boost/smart_ptr.hpp>

#include "timezone.h"

class Calendar {
public:
  /* Agrees with tm.tm_wday */
  typedef enum {
    SUNDAY = 0, MONDAY, TUESDAY, WEDNESDAY, THURSDAY, FRIDAY,
    SATURDAY, NO_WEEKDAY = -1
  } Weekday;
  
  /* Agrees with tm.tm_mon */
  typedef enum {
    JANUARY = 0, FEBRUARY, MARCH, APRIL, MAY, JUNE,
    JULY, AUGUST, SEPTEMBER, OCTOBER, NOVEMBER, DECEMBER,
    NO_MONTH = -1
  } Month;

  typedef enum {
    SECOND, MINUTE, HOUR_OF_DAY, DAY_OF_WEEK, DATE, MONTH, YEAR,
    WEEK_OF_MONTH, WEEK_OF_YEAR, DAY_OF_YEAR,
    DAY_OF_MONTH = DATE
  } Field;

private:
  static const int DAYSPERWEEK = 7;
  static const int DAYSPERNYEAR = 365;
  static const int DAYSPERLYEAR = 366;

  static bool isleap(int y) {
    return (y % 4 == 0 && ((y % 100) != 0 || (y % 400) == 0));
  }
  static int yearlengthTRUE(int y) {
    return isleap(y) ? DAYSPERLYEAR : DAYSPERNYEAR;
  }
  static int yearlengthBASE(int y) {
    return yearlengthTRUE(y + 1900);
  }

private:
  mutable struct tm tm;
  mutable time_t the_time;
  mutable bool normalized;
  
  boost::shared_ptr<const TimeZone> tz;

  Weekday firstDayOfWeek;
  int minimalDaysInFirstWeek;
  
protected:
  void set_and_normalize() const;
  void set_from_time_t();

  void init();

  int calc_week_of_period(int, int) const;
  int calc_week_of_month() const;
  int calc_week_of_year() const;

public:
  Calendar();
  Calendar(time_t);
  Calendar(const char *);
  Calendar(time_t, const char *);
  Calendar(int y, Month mo, int d, int h = 0, int m = 0, int s = 0);

  void setTimeZone(const char *);

  Weekday getFirstDayOfWeek() const;
  void setFirstDayOfWeek(Weekday w);

  int getMinimalDaysInFirstWeek() const;
  void setMinimalDaysInFirstWeek(int d);

  time_t getTime() const;
  void setTime(time_t t);

  void add(Field f, int inc);

  int getMaximum(Field f) const;
  int getLeastMaximum(Field f) const;
  int getActualMaximum(Field f) const;

  int getMinimum(Field f) const;
  int getGreatestMinimum(Field f) const;
  int getActualMinimum(Field f) const;

  int get(Field f) const;
  void set(Field f, int val);
  void clear(Field f);

  std::string toString() const;

  std::string format(const char *) const;
};


class calendar_exception : public std::exception {
  std::string _what;
public:
  calendar_exception(const std::string& what_arg) :
    _what(what_arg) { }
  virtual const char* what () const throw() { return _what.c_str(); }
  ~calendar_exception() throw() { }
};

class timezone_error : public calendar_exception {
public:
  timezone_error(const std::string& what_arg) :
    calendar_exception(what_arg) { }
  ~timezone_error() throw() { }
};


class calendar_range_error : public calendar_exception {
public:
  calendar_range_error(const std::string& what_arg) :
    calendar_exception(what_arg) { }
  ~calendar_range_error() throw() { }
};
  

class calendar_internal_error : public calendar_exception {
public:
  calendar_internal_error(const std::string& what_arg) :
    calendar_exception(what_arg) { }
  ~calendar_internal_error() throw() { }
};


class calendar_not_implemented : public calendar_exception {
public:
  calendar_not_implemented(const std::string& what_arg) :
    calendar_exception(what_arg) { }
  ~calendar_not_implemented() throw() { }
};



#endif /* CALENDAR_H_ */
