/* dayandposition.cpp - Represent a day-of-the-week and its position within
 * a larger unit (month or year) */
/* Copyright 2001 Columbia University.  All rights reserved. */

/* This is a manual translation of the Java Cal-code package. */

#include "dayandposition.h"
#include <stdexcept>
#include <strstream>

using namespace std;

/**
 * <code>DayAndPosition</code> is a utility class for representing byday
 * parameters in iCal recurrences and CPL time-switches.<p>
 *
 * It simply holds two pieces of information: a day of the week, and a day
 * position. <p>
 * 
 * The day of the week is represented by the day-of-week parameters of the
 * class {@link Calendar}: one of {@link Calendar#SUNDAY}, {@link
 * Calendar#MONDAY}, {@link Calendar#TUESDAY}, {@link Calendar#WEDNESDAY},
 * {@link Calendar#THURSDAY}, {@link Calendar#FRIDAY}, or {@link
 * Calendar#SATURDAY}.  It may also have the value
 * {@link Calendar#NO_WEEKDAY} if it has not been initialized.<p>
 *
 * The day position is an integer with a value between -53 and 53,
 * representing the <em>n</em>th occurence of the day in the enclosing
 * period (a month or a year).  Positive values count from the beginning of
 * the period; negative values count from the end.  A zero value means that
 * no position is implied, and that every day of the week with the given day
 * value is intended.<p>
 *
 * If the position is intended to represent a position within a month, only
 * the positions -5 to 5 are meaningful.  However,
 * <code>DayAndPosition</code> has no knowledge of what period its position
 * represents.<p>
 *
 * Change history:<br>
 * <dl>
 * <dt><b>1.0</b></dt><dd>Initial version</dd>
 * </dl>
 * @see Recurrence
 * @see Recurrence#setByDay
 * @see Calendar
 * @version 1.0
 * @author Jonathan Lennox */

/**
 * Construct a DayAndPosition with no weekday and zero recurrence.
 */
DayAndPosition::DayAndPosition() {
  day = Calendar::NO_WEEKDAY;
  position = 0;
}


/**
 * Construct a DayAndPosition with the given weekday and recurrence.
 *
 * @param d The day
 * @param p The period
 * @throws invalid_argument  If either argument is out
 *         of range.
 */
DayAndPosition::DayAndPosition(Calendar::Weekday d, int p) {
  if (!isValidDayOfWeek(d))
    throw invalid_argument("Invalid day of week");
  if (!isValidDayPosition(p))
    throw invalid_argument("Invalid day position");
  day = d;
  position = p;
}


/**
 * Construct a DayAndPosition from the given string
 * 
 * @param s String
 * @throws invalid_argument If the given string does not describe a valid
 *         DayAndPosition
 */
DayAndPosition::DayAndPosition(const std::string& s)
{
  setString(s);
}


/**
 * Get the standardized string representation of the DayAndPosition
 *
 * @return The string representation, of the form [[-]num]DY
 */
std::string DayAndPosition::getString() const
{
  ostrstream buf;

  if (position != 0) {
    buf << position;
  }

  buf << generateDayOfWeek(day);

  string ret(buf.str());
  buf.rdbuf()->freeze(0);
  return ret;
}


/**
 * Set the DayAndPosition to the given string
 *
 * @param s String
 * @throws invalid_argument If the given string does not describe a valid
 *         DayAndPosition
 */
void DayAndPosition::setString(const std::string& s)
{
  int p = 0;
  Calendar::Weekday d;

  string::const_iterator it = s.begin(), end = s.end();

  if (it == end) {
    throw invalid_argument("Bad day-and-position string \""
                           + s + "\": no day-of-week");
  }

  bool negative = false;
  if (*it == '-') {
    negative = true;
    it++;
    if (it == end || !isdigit(*it)) {
      throw invalid_argument("Bad day-and-position string \""
                             + s + "\": minus sign but no digits");
    }
  }

  for(; it != end && isdigit(*it); it++) {
    p = p * 10 + (*it - '0');
  }
  if (negative) {
    p = -p;
  }

  if (it == end) {
    throw invalid_argument("Bad day-and-position string \""
                           + s + "\": no day-of-week");
  }

  const string daystr(it, s.end());

  d = parseDayOfWeek(daystr);

  if (d == Calendar::NO_WEEKDAY) {
    throw invalid_argument("Bad day-and-position string \""
                           + s + "\": invalid day-of-week");
  }

  position = p;
  day = d;

  return;
} 


/**
 * Write a DayAndPosition to an output stream
 * @param os  The output stream
 * @param dap The DayAndPosition
 * @returns The output stream, again
 */
ostream& operator<<(ostream& os, const DayAndPosition& dap)
{
  os << dap.getString();
  return os;
}


/**
 * Get the day of the week stored in this object.  The value will be
 * a day-of-week constant from {@link Calendar} or
 * {@link Calendar#NO_WEEKDAY}.
 *
 * @return The day of the week, or {@link Calendar#NO_WEEKDAY}.
 * @see Calendar#SUNDAY
 * @see Calendar#MONDAY
 * @see Calendar#TUESDAY
 * @see Calendar#WEDNESDAY
 * @see Calendar#THURSDAY
 * @see Calendar#FRIDAY
 * @see Calendar#SATURDAY 
 * @see Calendar#NO_WEEKDAY
 */
Calendar::Weekday DayAndPosition::getDayOfWeek() const {
  return day;
}


/**
 * Set the day of the week stored in this object.  The value must be
 * a day-of-week constant from {@link Calendar} or
 * {@link Calendar#NO_WEEKDAY}.
 *
 * @param d The day of the week, or {@link Calendar#NO_WEEKDAY}.
 * @see Calendar#SUNDAY
 * @see Calendar#MONDAY
 * @see Calendar#TUESDAY
 * @see Calendar#WEDNESDAY
 * @see Calendar#THURSDAY
 * @see Calendar#FRIDAY
 * @see Calendar#SATURDAY 
 * @see Calendar#NO_WEEKDAY
 * @throws invalid_argument  If an invalid weekday is
 *         given.
 */
void DayAndPosition::setDayOfWeek(Calendar::Weekday d) {
  if (!isValidDayOfWeek(d))
    throw invalid_argument("Invalid day of week");
  day = d;
}


/**
 * Get the day position stored in this object.  The value is an integer
 * between -53 and 53, representing the <em>n</em>th occurence of
 * a day within a larger period (a month or year).  A value of 0 means
 * every occurence of the day.<p>
 *
 * Thus, if <code>dayOfWeek</code> were {@link Calendar#THURSDAY}, a value
 * of 1 would mean means the first Thursday, 2 would mean the seocnd
 * Thursday, -1 would mean the last Thursday, and so forth, while 0 would
 * mean every Thursday.
 *
 * @return The day position.
 */
int DayAndPosition::getDayPosition() const {
  return position;
}


/**
 * Set the day position stored in this object.  The value must be an
 * integer between -53 and 53, representing the <em>n</em>th occurence of
 * a day within a larger period (a month or year).  A value of 0 means
 * every occurence of the day.
 *
 * @param p The day position.
 * @throws invalid_argument  If an invalid position is
 *         given.
 */
void DayAndPosition::setDayPosition(int p) {
  if (!isValidDayPosition(p)) throw invalid_argument("Invalid day position");
  position = p;
}


/**
 * Compare this day-and-position to the specified object.
 * The result is <code>true</code> if and only if the argument is
 * not <code>null</code> and is a <code>DayAndPosition</code> object that
 * represents the same day and position as this object.
 *
 * @param obj the object to compare with.
 * @return <code>true</code> if the objects are the same;
 *         <code>false</code> otherwise.
 */
bool DayAndPosition::operator==(const DayAndPosition& other) const {
  return getDayOfWeek() == other.getDayOfWeek() && 
    getDayPosition() == other.getDayPosition();
}


/**
 * Test if an integer is a valid day of the week.
 * The result is <code>true</code> if and only if the argument
 * is a valid weekday constant from {@link Calendar}, or
 * {@link Calendar#NO_WEEKDAY}.
 *
 * @param d The value to be tested.
 * @return  Whether the value is a valid weekday.
 * @see Calendar#SUNDAY
 * @see Calendar#MONDAY
 * @see Calendar#TUESDAY
 * @see Calendar#WEDNESDAY
 * @see Calendar#THURSDAY
 * @see Calendar#FRIDAY
 * @see Calendar#SATURDAY 
 * @see Calendar#NO_WEEKDAY
 */
bool DayAndPosition::isValidDayOfWeek(Calendar::Weekday d) {
  switch (d) {
  case Calendar::NO_WEEKDAY:
  case Calendar::SUNDAY:
  case Calendar::MONDAY:
  case Calendar::TUESDAY:
  case Calendar::WEDNESDAY:
  case Calendar::THURSDAY:
  case Calendar::FRIDAY:
  case Calendar::SATURDAY:
    return true;
  default:
    return false;
  }
}


/**
 * Test if an integer is a valid day position.
 * The result is <code>true</code> if and only if the argument
 * is in the range -53 to 53.
 *
 * @param p The value to be tested.
 * @return  Whether the value is a valid day position.
 */
bool DayAndPosition::isValidDayPosition(int p) {
  return (p >= -53 && p <= 53);
}


/**
 * Parse a two-char RFC 2445 day-of-week string.
 *
 * @param daystr The string to parse.
 * @return An appropriate Calendar::Weekday constant, or
 *         Calendar::NO_WEEKDAY if the value does not correspond to any
 *         weekday. 
 */
Calendar::Weekday DayAndPosition::parseDayOfWeek(const string& daystr)
{
  if (daystr.size() != 2) {
    return Calendar::NO_WEEKDAY;
  }

  /* XXX: worth changing to a map? */
  if (daystr == "SU") {
    return Calendar::SUNDAY;
  }
  else if (daystr == "MO") {
    return Calendar::MONDAY;
  }
  else if (daystr == "TU") {
    return Calendar::TUESDAY;
  }
  else if (daystr == "WE") {
    return Calendar::WEDNESDAY;
  }
  else if (daystr == "TH") {
    return Calendar::THURSDAY;
  }
  else if (daystr == "FR") {
    return Calendar::FRIDAY;
  }
  else if (daystr == "SA") {
    return Calendar::SATURDAY;
  }
  else {
    return Calendar::NO_WEEKDAY;
  }
}

/**
 * Return a string corresponding to the given day-of-the-week.
 * @param day  The day of the week
 * @returns A two-letter string, as in RFC 2445.
 */
string DayAndPosition::generateDayOfWeek(Calendar::Weekday day)
{
  switch (day) {
  case Calendar::SUNDAY:
    return "SU";

  case Calendar::MONDAY:
    return "MO";

  case Calendar::TUESDAY:
    return "TU";

  case Calendar::WEDNESDAY:
    return "WE";

  case Calendar::THURSDAY:
    return "TH";

  case Calendar::FRIDAY:
    return "FR";

  case Calendar::SATURDAY:
    return "SA";

  default:
    throw
      runtime_error("Internal error: Invalid day-of-week in "
                    "DayAndPosition::generateDayOfWeek()");
  }
}


/**
 * Return a string representation of this day-and-position. This method 
 * is intended to be used only for debugging purposes.
 * The returned string may be empty but may not be <code>null</code>.
 * 
 * @return  A string representation of this day-and-position.
 */
string DayAndPosition::toString() const {
  ostrstream buffer;

  buffer << "DayAndPosition";

  buffer << "[day=";
  buffer << day;
  buffer << ",position=";
  buffer << position;
  buffer << "]";

  buffer << ends;

  string ret(buffer.str());
  buffer.rdbuf()->freeze(0);

  return ret;
}

