/* duration.cpp - Represent an iCalendar duration. */
/* Copyright 2001 Columbia University.  All rights reserved. */

/* This is a manual translation of the Java Cal-code package. */

#include "duration.h"

#include <stdexcept>
#include <strstream>

using namespace std;

// Duration -- represent an iCal duration.
// TODO: Parsers

/**
 * <code>Duration</code> represents a duration field in an iCal recurrence
 * or a CPL time-switch.<p>
 *
 * It represents a duration of time, expressed as a number of days, hours,
 * minutes, and seconds, or a number of weeks.<p>
 *
 * <b>The iCal spec forbids a duration from simultaneously specifying weeks
 * and other units.</b>  The <code>Duration</code> class enforces this
 * restriction.  Trying to specify a week value when there are any non-zero
 * non-week values, or a non-week value when there is a non-zero week value,
 * will throw an <code>invalid_argument</code> exception.<p>
 *
 * Values may be specified which are larger than permitted for their
 * enclosing fields.  They will be normalized appropriately.<p>
 *
 * Change history:<br>
 * <dl>
 * <dt><b>1.0</b></dt><dd>Initial version</dd>
 * </dl>
 * @see Recurrence
 * @version 1.0
 * @author Jonathan Lennox
 */

/**
 * Construct a <code>Duration</code> object and initializes it to
 * a zero-length interval.
 */
Duration::Duration() {
  clear();
}


/**
 * Construct a <code>Duration</code> object with the specified duration.
 *
 * @param  d  The number of days.
 * @param  h  The number of hours.
 * @param  m  The number of minutes.
 * @param  s  The number of seconds.
 */
Duration::Duration(int d, int h, int m, int s) {
  clear();
  days = d;
  hours = h;
  minutes = m;
  seconds = s;
}


/**
 * Construct a <code>Duration</code> object with the specified duration.
 *
 * @param  h  The number of hours.
 * @param  m  The number of minutes.
 * @param  s  The number of seconds.
 */
Duration::Duration(int h, int m, int s)
{
  clear();
  hours = h;
  minutes = m;
  seconds = s;
}


/**
 * Construct a <code>Duration</code> object with the specified duration.
 *
 * @param  w  The number of weeks.
 */
Duration::Duration(int w) {
  clear();
  weeks = w;
}


/**
 * Construct a <code>Duration</code> object described by the given string.
 * A duration string is defined as follows, according to RFC 2445:
 * <pre>
 *     dur-value  = (["+"] / "-") "P" (dur-date / dur-time / dur-week)
 *
 *     dur-date   = dur-day [dur-time]
 *     dur-time   = "T" (dur-hour / dur-minute / dur-second)
 *     dur-week   = 1*DIGIT "W"
 *     dur-hour   = 1*DIGIT "H" [dur-minute]
 *     dur-minute = 1*DIGIT "M" [dur-second]
 *     dur-second = 1*DIGIT "S"
 *     dur-day    = 1*DIGIT "D"
 * </pre>
 *
 * @param  s  The string representation of the duration.
 * @throws invalid_argument If the given string does not describe a valid
 *         Duration
 */
Duration::Duration(const std::string& s)
{
  setString(s);
}


/**
 * Reset the duration to a zero-length interval.
 */
void Duration::clear() {
  weeks = 0;
  days = 0;
  hours = 0;
  minutes = 0;
  seconds = 0;
}


/**
 * Set the value of the <code>Duration</code> object to the given string.
 * A duration string is defined as follows, according to RFC 2445:
 * <pre>
 *     dur-value  = (["+"] / "-") "P" (dur-date / dur-time / dur-week)
 *
 *     dur-date   = dur-day [dur-time]
 *     dur-time   = "T" (dur-hour / dur-minute / dur-second)
 *     dur-week   = 1*DIGIT "W"
 *     dur-hour   = 1*DIGIT "H" [dur-minute]
 *     dur-minute = 1*DIGIT "M" [dur-second]
 *     dur-second = 1*DIGIT "S"
 *     dur-day    = 1*DIGIT "D"
 * </pre>
 *
 * @param  s  The string representation of the duration.
 * @throws invalid_argument If the given string does not describe a valid
 *         Duration
 */
void Duration::setString(const std::string& str)
{
  string::const_iterator s = str.begin();
  string::const_iterator end = str.end();
  
  int w = 0, d = 0, h = 0, m = 0, sec = 0;
  
  if (s == end) {
    throw invalid_argument("Bad duration string: missing \'P\'");
  }
  if (*s == '-') {
    throw invalid_argument("Bad duration string: negative durations not supported");
  }
  else if (*s == '+') {
    s++;
  }
  
  if (s == end || *s != 'P') {
    throw invalid_argument("Bad duration string: missing \'P\'");
  }
  s++;
  
  bool seen_T = false;
  bool seen_W = false;
  bool seen_non_W = false;
  bool value_has_digits = false;
  int value = 0;
  for (; s != end; s++) {
    if (isdigit(*s)) {
      value = value * 10 + (*s - '0');
      value_has_digits = true;
    }
    else {
      switch (*s) {
      case 'T':
        seen_T = true;
        break;
        
      case 'W':
        if (seen_T) {
          throw invalid_argument("Bad duration string: \'W\' after \'T\'");
        }
        else if (seen_non_W) {
          throw invalid_argument("Bad duration string: mixed \'W\' and non-\'W\'");
        }
        else if (!value_has_digits) {
          throw invalid_argument("Bad duration string: \'W\' not after value");
        }
        w = value;
        seen_W = true;
        break;
        
      case 'D':
        if (seen_T) {
          throw invalid_argument("Bad duration string: \'D\' after \'T\'");
        }
        else if (seen_W) {
          throw invalid_argument("Bad duration string: mixed \'W\' and non-\'W\'");
        }
        else if (!value_has_digits) {
          throw invalid_argument("Bad duration string: \'D\' not after value");
        }
        d = value;
        seen_non_W = true;
        break;
        
      case 'H':
      case 'M':
      case 'S':
        if (!seen_T) {
          throw invalid_argument("Bad duration string: \'"
                                 + string(s, s+1)
                                 + "\' not after \'T\'");
        }
        else if (seen_W) {
          throw invalid_argument("Bad duration string: mixed \'W\' and non-\'W\'");
        }
        else if (!value_has_digits) {
          throw invalid_argument("Bad duration string: \'"
                                 + string(s, s+1)
                                 + "\' not after value");
        }
        if (*s == 'H') {
          h = value;
        }
        else if (*s == 'M') {
          m = value;
        }
        else if (*s == 'S') {
          sec = value;
        }
        seen_non_W = true;
        break;
        
      default:
        throw invalid_argument("Bad duration string: unknown duration specifier \'"
                               + string(s, s + 1)
                               + "\'");
      }
      value = 0;
      value_has_digits = false;
    }
  }
  
  weeks = w;
  days = d;
  hours = h;
  minutes = m;
  seconds = sec;
  
  return;
}


/**
 * Get the standardized string representation of the Duration.
 *
 * @return The string representation, of the form
 * P(<num>W|[<num>D]T[<num>H][<num>M][<num>S]).  A zero-time duration is
 * represented as PT0S.
 */
std::string Duration::getString() const
{
  ostrstream buf;
  
  buf << "P";
  if (weeks != 0) {
    buf << weeks << "W";
  }
  if (days != 0) {
    buf << days << "D";
  }
  if (hours != 0 || minutes != 0 || seconds != 0) {
    buf << "T";
  }
  if (hours != 0) {
    buf << hours << "H";
  }
  if (minutes != 0) {
    buf << minutes << "M";
  }
  if (seconds != 0) {
    buf << seconds << "S";
  }
  if (weeks == 0 && days == 0 && hours == 0 && minutes == 0 && seconds == 0) {
    buf << "T0S";
  }
  buf << ends;

  string ret(buf.str());
  buf.rdbuf()->freeze(0);
  return ret;
}


/**
 * Get the weeks field of this duration.
 * @return The number of weeks.
 */
int Duration::getWeeks() const {
  return weeks;
}


/**
 * Set the weeks field of this duration.
 * @param w The number of weeks.
 * @throws out_of_range If <code>w</code> is negative.
 * @throws invalid_argument    If any non-week field is set.
 */
void Duration::setWeeks(int w) {
  if (w < 0) throw out_of_range("Week value out of range");
  checkWeeksOkay(w);
  weeks = w;
}


/**
 * Get the days field of this duration.
 * @return The number of days.
 */
int Duration::getDays() const {
  return days;
}


/**
 * Set the days field of this duration.
 * @param d The number of days.
 * @throws out_of_range If <code>d</code> is negative.
 * @throws invalid_argument If the week field is set.
 */
void Duration::setDays(int d) {
  if (d < 0) throw out_of_range("Day value out of range");
  checkNonWeeksOkay(d);
  days = d;
  normalize();
}


/**
 * Get the hours field of this duration.
 * @return The number of hours.
 */
int Duration::getHours() const {
  return hours;
}


/**
 * Set the hours field of this duration.
 * @param h The number of hours.
 * @throws out_of_range If <code>h</code> is negative.
 * @throws invalid_argument If the week field is set.
 */
void Duration::setHours(int h) {
  if (h < 0) throw out_of_range("Hour value out of range");
  checkNonWeeksOkay(h);
  hours = h;
  normalize();
}


/**
 * Get the minutes field of this duration.
 * @return The number of minutes.
 */
int Duration::getMinutes() const {
  return minutes;
}


/**
 * Set the minutes field of this duration.
 * @param m The number of minutes.
 * @throws out_of_range If <code>m</code> is negative.
 * @throws invalid_argument If the week field is set.
 */
void Duration::setMinutes(int m) {
  if (m < 0) throw out_of_range("Minute value out of range");
  checkNonWeeksOkay(m);
  minutes = m;
  normalize();
}


/**
 * Get the seconds field of this duration.
 * @return The number of seconds.
 */
int Duration::getSeconds() const {
  return seconds;
}


/**
 * Set the seconds field of this duration.
 * @param s The number of days.
 * @throws out_of_range If <code>s</code> is negative.
 * @throws invalid_argument If the week field is set.
 */
void Duration::setSeconds(int s) {
  if (s < 0) throw out_of_range("Second value out of range");
  checkNonWeeksOkay(s);
  seconds = s;
  normalize();
}


/**
 * Get the interval represented by this <code>Duration</code>, in
 * seconds.  Pure Gregorian time is used; leap-seconds and
 * daylight-savings time transitions are ignored.
 *
 * @return  the number of seconds equivalent to this duration.
 */
time_t Duration::getInterval() const {
  return
    seconds +
    minutes * SECONDS_PER_MINUTE +
    hours   * SECONDS_PER_HOUR +
    days    * SECONDS_PER_DAY +
    weeks   * SECONDS_PER_WEEK;
}


/**
 * Set the duration's interval to the specified number of seconds.
 * Pure-Gregorian time is used; leap-seconds and daylight-savings time
 * transitions are ignored.  
 * The duration is broken down into the <code>days</code>,
 * <code>hours</code>, <code>minutes</code>, and <code>seconds</code>
 * fields; the <code>weeks</code> field will be zero after this method is
 * called.
 *
 * @param seconds  The number of seconds for this duration.
 * @throws out_of_range If <code>millis</code> is negative.
 */
void Duration::setInterval(time_t interval) {
  if (interval < 0) {
    throw out_of_range("Negative-length interval");
  }
  clear();

  days = (int)(seconds / SECONDS_PER_DAY);
  seconds = (int)(seconds % SECONDS_PER_DAY);

  normalize();
}


/**
 * Normalize the duration, so that every field is within its legal range.
 */
void Duration::normalize() {
  minutes += seconds / SECONDS_PER_MINUTE;
  seconds %= SECONDS_PER_MINUTE;
  
  hours   += minutes / MINUTES_PER_HOUR;
  minutes %= MINUTES_PER_HOUR;

  /* N.B. do *not* normalize hours to days.  24 hours is not the same thing
   * as 1 day, if we're near a DST jump. */
}


/**
 * Check if it would be okay to set the weeks field of this duration.
 * This is allowed if none of the other fields are set, or if the
 * prospective value for the weeks field is zero.
 * @param f The prospective value for the weeks field
 * @throws invalid_argument If it wouldn't be okay.
 */
void Duration::checkWeeksOkay(int f) {
  if (f != 0 &&
      (days != 0 || hours != 0 || minutes != 0 || seconds != 0)) {
    throw
      invalid_argument("Weeks and non-weeks are incompatible");
  }
}


/**
 * Check if it would be okay to set a non-weeks field of this duration.
 * This is allowed if the weeks field is not set, or if the
 * prospective value for this non-weeks field is zero.
 * @param f The prospective value for the non-weeks field
 * @throws invalid_argument If it wouldn't be okay.
 */
void Duration::checkNonWeeksOkay(int f) {
  if (f != 0 && weeks != 0) {
    throw
      invalid_argument("Weeks and non-weeks are incompatible");
  }
}


/**
 * Return a string representation of this duration. This method 
 * is intended to be used only for debugging purposes.
 * The returned string may be empty but may not be <code>null</code>.
 * 
 * @return  a string representation of this duration.
 */
string Duration::toString() const {
  ostrstream buffer;
  
  buffer << "Duration";

  buffer << "[weeks=";
  buffer << weeks;
  buffer << ",days=";
  buffer << days;
  buffer << ",hours=";
  buffer << hours;
  buffer << ",minutes=";
  buffer << minutes;
  buffer << ",seconds=";
  buffer << seconds;
  buffer << "]";

  buffer << ends;

  string ret(buffer.str());
  buffer.rdbuf()->freeze(0);

  return ret;
}
