/* recur.cpp - Represent an iCalendar (subset) recurrence rule. */
/* Copyright 2001 Columbia University.  All rights reserved. */

/* This is a manual translation of the Java Cal-code package. */
// XXX TODO: Error-checking

#include <stdexcept>
#include <strstream>

#include "recur.h"

using namespace std;

/**
 * <code>Recurrence</code> represents a recurring interval of time.  It
 * corresponds to a CPL time-switch, and as such corresponds to a subset of
 * an iCal RRULE.<p>
 *
 * A recurring interval of time is represented by a start time, a duration
 * (equivalently, an ending time), a frequency, an interval, a termination
 * ("until") date, and a set of by* rules.  For more information on this,
 * see below, and the relevant specifications.<p>
 *
 * Change history:<br>
 * <dl>
 * <dt><b>1.0</b></dt><dd>Initial version</dd>
 * <dt><b>1.1</b></dt><dd>Added sub-day by* and intervals, and count</dd>
 * </dl>
 * @see <a href="http://www.rfc-editor.org/rfc/rfc2445.txt">RFC 2445</a>
 * @see <a href="http://www.ietf.org/internet-drafts/draft-ietf-iptel-cpl-03.txt">CPL</a> 
 * @version 1.1
 * @author Jonathan Lennox
 */


/* Constructors */

/**
 * Allocate a new Recurrence, with no recurrence frequency, starting
 * at time = 0, duration 0.
 */
Recurrence::Recurrence() 
{
  init();

  dtStart = new Calendar();
  duration = new Duration();
}


/**
 * Allocate a new Recurrence, with no recurrence frequency.
 *
 * @param start The start time, as a broken-down time.  (Only the calendar
 *              fields are used.  Use {@link #setWeekStart} rather
 *              than {@link Calendar#setFirstDayOfWeek} to set the
 *              first day of the week.
 * @param dur  The duration.
 * @see #setWeekStart
 */
Recurrence::Recurrence(const Calendar& start, const Duration& dur)
{
  init();

  setDtStart(start);

  duration = new Duration(dur);
}


/**
 * Allocate a new Recurrence, with the specified frequency.
 *
 * @param start The start time, as a broken-down time.  (Only the calendar
 *              fields are used.  Use {@link #setWeekStart} rather
 *              than {@link Calendar#setFirstDayOfWeek} to set the
 *              first day of the week.
 * @param dur   The duration.
 * @param freq  The recurrence frequency (one of
 *        {@link Recurrence#DAILY}, {@link Recurrence#WEEKLY},
 *        {@link Recurrence#MONTHLY}, {@link Recurrence#YEARLY},
 *        or {@link Recurrence#NO_RECURRENCE}). 
 * @see #setWeekStart
 */
Recurrence::Recurrence(const Calendar& start, const Duration& dur,
                       Frequency freq)
{
  init();

  setDtStart(start);

  duration = new Duration(dur);
  frequency = freq;
  
  interval = 1;
  // Everything else gets initialized to 0 or NULL, which is what we want.
}

/**
 * Copy constructor.
 */
Recurrence::Recurrence(const Recurrence& other)
{
  dtStart = (other.dtStart != NULL ? new Calendar(*other.dtStart) : NULL);
  duration = (other.duration != NULL ? new Duration(*other.duration) : NULL);
  frequency = other.frequency;
  interval = other.interval;
  until = (other.until != NULL ? new Calendar(*other.until) : NULL);
  count = other.count;

  bySecond = other.bySecond;
  byMinute = other.byMinute;
  byHour = other.byHour;
  byDay = other.byDay;
  byMonthDay = other.byMonthDay;
  byYearDay = other.byYearDay;
  byWeekNo = other.byWeekNo;
  byMonth = other.byMonth;
}

/**
 * Initialize fields of a recurrence.
 */
void Recurrence::init()
{
  dtStart = NULL;
  duration = NULL;
  frequency = NO_RECURRENCE;
  interval = 1;
  until = NULL;
  count = 0;
}


/**
 * De-allocate a Recurrence.
 */
Recurrence::~Recurrence() {
  delete dtStart;
  delete duration;
  delete until;
}


/**
 * Assignment operator.
 */
Recurrence& Recurrence::operator=(const Recurrence& other)
{
  delete dtStart;
  delete duration;
  delete until;

  dtStart = (other.dtStart != NULL ? new Calendar(*other.dtStart) : NULL);
  duration = (other.duration != NULL ? new Duration(*other.duration) : NULL);
  frequency = other.frequency;
  interval = other.interval;
  until = (other.until != NULL ? new Calendar(*other.until) : NULL);
  count = other.count;

  bySecond = other.bySecond;
  byMinute = other.byMinute;
  byHour = other.byHour;
  byDay = other.byDay;
  byMonthDay = other.byMonthDay;
  byYearDay = other.byYearDay;
  byWeekNo = other.byWeekNo;
  byMonth = other.byMonth;

  return *this;
}

/* Accessors */
/**
 * Get the start time of the recurrence.
 * @return The start time.
 */
Calendar Recurrence::getDtStart() const {
  return *dtStart;
}


/**
 * Get a string representation of the start time of the recurrence.
 * @return A string representing the start time.
 */
string Recurrence::getDtStartString() const {
  return generateDateTimeString(getDtStart());
}


/**
 * Set the start time of the recurrence.
 * @param start The start time, as a broken-down time.  (Only the calendar
 *              fields are used.  Use {@link #setWeekStart} rather
 *              than {@link Calendar#setFirstDayOfWeek} to set the
 *              first day of the week.
 * @see #setWeekStart
 */
void Recurrence::setDtStart(const Calendar& start) {
  Calendar::Weekday oldStart;
  if (dtStart != NULL) {
    oldStart = dtStart->getFirstDayOfWeek();
    delete dtStart;
  }
  else {
    oldStart = Calendar::MONDAY;
  }
  
  dtStart = new Calendar(start);
  
  setCalendarWeekStart(*dtStart, oldStart);
}


/**
 * Set the start time of the recurrence.
 * @param start The start time, as a string.
 * @throws invalid_argument If the given string does not describe a valid
 *                          Date-Time.
 */
void Recurrence::setDtStart(const string& start) {
  setDtStart(parseDateTime(start));
}


/**
 * Get the duration of the recurrence.
 * @return The duration.
 */
Duration Recurrence::getDuration() const {
  return *duration;
}


/**
 * Get the duration of the recurrence, represented as a string.
 * @return A string representing the duration.
 */
string Recurrence::getDurationString() const {
  return getDuration().getString();
}


/**
 * Set the duration of the recurrence.
 * @param d The duration.
 */
void Recurrence::setDuration(const Duration& d) {
  if (duration != NULL)
    delete duration;
  duration = new Duration(d);
}


/**
 * Set the duration of the recurrence.
 * @param str A string representing the duration.
 * @throws invalid_argument If the given string does not describe a valid
 *                          Duration.
 */
void Recurrence::setDuration(const string& str) {
  Duration d(str);
  setDuration(d);
}


/**
 * Get the end time of the recurrence.
 * The broken-down time of the returned dtEnd will be correct, though
 * its time in milliseconds may not be.
 * @return The end time.
 */
Calendar Recurrence::getDtEnd() const {
  /* Make dtEnd a cloned dtStart, so non-time fields of the Calendar
   * are accurate. */
  Calendar tempEnd(*dtStart);

  tempEnd.setTime(dtStart->getTime() +
                  duration->getInterval());
  return tempEnd;
}


/**
 * Get a string representing the end time of the recurrence.
 * @return A string representation of the end time.
 */
string Recurrence::getDtEndString() const {
  return generateDateTimeString(getDtEnd());
}


/**
 * Set the end time of the recurrence.
 * @param end The end time.
 */
void Recurrence::setDtEnd(const Calendar& end) {
  if (duration == NULL) {
    duration = new Duration();
  }
  duration->setInterval(end.getTime() -
                        dtStart->getTime());
}


/**
 * Set the end time of the recurrence.
 * @param end The end time, as a string.
 * @throws invalid_argument If the given string does not describe a valid
 *                          Date-Time.
 */
void Recurrence::setDtEnd(const string& end) {
  setDtEnd(parseDateTime(end));
}


/**
 * Get the frequency of the recurrence.
 * @return The recurrence frequency (one of
 *         {@link Recurrence#SECONDLY}, {@link Recurrence#MINUTELY},
 *         {@link Recurrence#HOURLY},
 *         {@link Recurrence#DAILY}, {@link Recurrence#WEEKLY},
 *         {@link Recurrence#MONTHLY}, {@link Recurrence#YEARLY},
 *         or {@link Recurrence#NO_RECURRENCE}).
 */
Recurrence::Frequency Recurrence::getFrequency() const {
  return frequency;
}


/**
 * Get the frequency of the recurrence, as a string.
 * @return A string representing the recurrence frequency (one of
 * "SECONDLY", "MINUTELY", "HOURLY", "DAILY", "WEEKLY", "MONTHLY",
 * "YEARLY", or "" representing no recurrence).
 */
string Recurrence::getFrequencyString() const {
  switch (frequency) {
  case SECONDLY:
    return "SECONDLY";
  case MINUTELY:
    return "MINUTELY";
  case HOURLY:
    return "HOURLY";
  case DAILY:
    return "DAILY";
  case WEEKLY:
    return "WEEKLY";
  case MONTHLY:
    return "MONTHLY";
  case YEARLY:
    return "YEARLY";
  case NO_RECURRENCE:
    return "";
  }
  throw runtime_error("Invalid frequency in getFrequencyString");
}


/**
 * Set the frequency of the recurrence.
 * @param freq The recurrence frequency (one of
 *         {@link Recurrence#SECONDLY}, {@link Recurrence#MINUTELY},
 *         {@link Recurrence#HOURLY},
 *         {@link Recurrence#DAILY}, {@link Recurrence#WEEKLY},
 *         {@link Recurrence#MONTHLY}, {@link Recurrence#YEARLY},
 *         or {@link Recurrence#NO_RECURRENCE}).
 * @throws invalid_argument if the frequency isn't one of
 *         the above values.
 */
void Recurrence::setFrequency(Frequency freq) {
  if (frequency != SECONDLY && frequency != MINUTELY &&
      frequency != HOURLY && frequency != DAILY && frequency != WEEKLY &&
      frequency != MONTHLY && frequency != YEARLY &&
      frequency != NO_RECURRENCE) {
    throw invalid_argument("Invalid frequency");
  }
  frequency = freq;
}


/**
 * Utility function -- case-insensitive comparison.
 * XXX: Use Unicode comparison functions instead, once we have those?
 */
static int stringCmpi(const string& s1, const string& s2) {
  string::const_iterator p1, p2;

  for (p1 = s1.begin(), p2 = s2.begin(); p1 != s1.end() && p2 != s2.end();
       p1++, p2++) {
    if (toupper(*p1) != toupper(*p2)) {
      // Report which was lexically  greater:
      return (int)(toupper(*p1)) - (int)(toupper(*p2));
    }
  }

  // If they match up to the detected eos, say 
  // which was longer. Return 0 if the same.
  return(s2.size() - s1.size());
}


/**
 * Set the frequency of the recurrence.
 * @param freq A string representing the recurrence frequency (one of
 *            "SECONDLY", "MINUTELY", "HOURLY", "DAILY", "WEEKLY", "MONTHLY",
 *            "YEARLY", or "" representing no recurrence).
 * @throws invalid_argument if the frequency isn't one of
 *         the above values.
 */
void Recurrence::setFrequency(const string& freq) {
  if (stringCmpi(freq, "SECONDLY") == 0) {
    frequency = SECONDLY;
  }
  else if (stringCmpi(freq, "MINUTELY") == 0) {
    frequency = MINUTELY;
  }
  else if (stringCmpi(freq, "HOURLY") == 0) {
    frequency = HOURLY;
  }
  else if (stringCmpi(freq, "DAILY") == 0) {
    frequency = DAILY;
  }
  else if (stringCmpi(freq, "WEEKLY") == 0) {
    frequency = WEEKLY;
  }
  else if (stringCmpi(freq, "MONTHLY") == 0) {
    frequency = MONTHLY;
  }
  else if (stringCmpi(freq, "YEARLY") == 0) {
    frequency = YEARLY;
  }
  else if (freq == "") {
    frequency = NO_RECURRENCE;
  }
  else {
    throw invalid_argument("Invalid frequency \"" + freq + "\"");
  }
}


/**
 * Get the interval of the recurrence.
 * @return The recurrence interval.
 */
int Recurrence::getInterval() const {
  return interval;
}


/**
 * Get the interval of the recurrence, as a string.
 * @return A string representing the recurrence interval, or the empty string
 *         if no interval is defined.
 */
string Recurrence::getIntervalString() const {
  if (interval <= 1) {
    return "";
  }

  ostrstream obuf;
  obuf << getInterval() << ends;

  string ret(obuf.str());
  obuf.rdbuf()->freeze(0);
  return ret;
}

/**
 * Set the interval of the recurrence.
 * @param intr The recurrence interval.
 */
void Recurrence::setInterval(int intr) {
  interval = intr;
}


/**
 * Set the interval of the recurrence, from a string.
 * 
 * @param intr The recurrence interval, as a string, or the empty string.
 * @throws invalid_argument if the argument isn't a valid interval.
 */
void Recurrence::setInterval(const string& intr)
{
  if (intr.empty()) {
    interval = 1;
    return;
  }
  
  int i;
  char dummy;

  istrstream ibuf(intr.c_str());

  ibuf >> i;
  if (ibuf.fail()) {
    throw invalid_argument("Bad interval string \"" + intr + "\"");
  }
  
  ibuf >> dummy; // Try to read one extra char, to make sure we're at EOF.
  if (!ibuf.eof()) {
    throw invalid_argument("Bad interval string \"" + intr + "\"");
  }
  
  if (i < 1) {
    throw invalid_argument("Bad interval value " + intr);
  }

  interval = i;
}


/**
 * Get the upper bound of the recurrence.
 * If <code>count</code> has been set, this computes
 * <code>until</code> from <code>count</code>.
 * @return The upper bound.  <code>NULL</code> if unbounded.
 * @see #computeUntilFromCount()
 */
Calendar* Recurrence::getUntil() const {
  computeUntilFromCount();
  
  return (until != NULL ? new Calendar(*until) : NULL);
}


/**
 * Get a string representing the upper bound of the recurrence, or the empty
 * string if unbounded.
 * If <code>count</code> has been set, this computes <code>until</code> from
 * <code>count</code>.
 * @return A string representation of the upper bound.  "" if unbounded.
 */
string Recurrence::getUntilString() const {
  computeUntilFromCount();

  if (until == NULL) {
    return "";
  }

  return generateDateTimeString(*until);
}


/**
 * Set the upper bound of the recurrence.
 * @param u The upper bound, or <code>NULL</code> if unbounded.
 * @throws invalid_argument if non-NULL and <code>count</code>
 *                                  is set.
 */
void Recurrence::setUntil(const Calendar* u) {
  if (until != NULL) {
    delete until;
  }

  if (u == NULL) {
    until = NULL;
    return;
  }

  if (count != 0) {
    throw invalid_argument("Cannot set both count and until");
  }

  until = new Calendar(*u);
}


/**
 * Set the upper bound of a recurrence, based on a string.
 * @param A string representation of the upper bound, or "" if unbounded.
 * @throws invalid_argument if non-"" and <code>count</code>
 *                          is set.
 * @throws invalid_argument If the given string does not describe a valid
 *                          Date-Time.
 */
void Recurrence::setUntil(const string& u) {
  if (u == "") {
    setUntil(NULL);
  }
  else {
    Calendar tmp_u = parseDateTime(u);
    setUntil(&tmp_u);
  }
}


/**
 * Get the repeat count of the recurrence.
 * @return The repeat count.  0 if it is unset; -1 if it is
 *         unknown (because 'until' is set).
 */
int Recurrence::getCount() const {
  if (count == 0 && until != NULL) {
    return -1;
  }
  return count;
}


/**
 * Get the repeat count of the recurrence, as a string.
 * @return A string representation of the repeat count.  "" if it is 
 *         unset or unknown (because 'until' is set).
 */
string Recurrence::getCountString() const {
  if (count == 0) {
    return "";
  }

  ostrstream obuf;
  obuf << getCount() << ends;

  string ret(obuf.str());
  obuf.rdbuf()->freeze(0);
  return ret;
}


/**
 * Set the repeat count of the recurrence.
 * @param c The repeat count, or <code>0</code> if unbounded.
 * @throws invalid_argument If a non-computed <code>until</code>
 *         has been set.
 * @see #computeUntilFromCount()
 */
void Recurrence::setCount(int c) {
  if (c != 0 && count == 0 && until != NULL) {
    throw invalid_argument("Cannot set both count and until");
  }
  if (c != 0 || count != 0) {
    until = NULL;
  }
  count = c;
}

/*
 * Set the repeat count of the recurrence.
 * @param c The repeat count, or <code>0</code> if unbounded.
 * @throws invalid_argument If a non-computed <code>until</code>
 *         has been set.
 * @throws invalid_argument if the argument isn't a valid interval.
 */
void Recurrence::setCount(const string& c)
{
  if (c.empty()) {
    setCount(0);
    return;
  }
  
  int cnt;
  char dummy;

  istrstream ibuf(c.c_str());

  ibuf >> cnt;
  if (ibuf.fail()) {
    throw invalid_argument("Bad count string \"" + c + "\"");
  }

  ibuf >> dummy; // Try to read one extra character, to make sure we're at EOF.
  if (!ibuf.eof()) {
    throw invalid_argument("Bad count string \"" + c + "\"");
  }
  if (cnt < 1) {
    throw invalid_argument("Bad count value " + c);
  }

  setCount(cnt);
}


/**
 * Get the first day of the week for this recurrence.
 *
 * @return The day of the week (one of {@link Calendar#SUNDAY},
 *         {@link Calendar#MONDAY}, {@link Calendar#TUESDAY},
 *         {@link Calendar#WEDNESDAY}, {@link Calendar#THURSDAY},
 *         {@link Calendar#FRIDAY}, or {@link Calendar#SATURDAY}).
 */
Calendar::Weekday Recurrence::getWeekStart() const {
  return dtStart->getFirstDayOfWeek();
}


/**
 * Get the first day of the week for this recurrence, as a string.
 *
 * @return The first day of the week (one of "SU", "MO", "TU", "WE",
 *         "TH", "FR", "SA").
 */
string Recurrence::getWeekStartString() const {
  return DayAndPosition::generateDayOfWeek(getWeekStart());
}

/**
 * Set the first day of the week for this recurrence.
 *
 * @param weekstart The day of the week (one of {@link Calendar#SUNDAY},
 *         {@link Calendar#MONDAY}, {@link Calendar#TUESDAY},
 *         {@link Calendar#WEDNESDAY}, {@link Calendar#THURSDAY},
 *         {@link Calendar#FRIDAY}, or {@link Calendar#SATURDAY}).
 */
void Recurrence::setWeekStart(Calendar::Weekday weekstart) {
  dtStart->setFirstDayOfWeek(weekstart);
}


/**
 * Set the first day of the week for this recurrence.
 *
 * @param weekstart The day of the week, as a string (one of "SU", "MO", "TU"
 *                  "WE", "TH", "FR", "SA").
 * @throws invalid_argument If the string is not a valid weekday.
 */
void Recurrence::setWeekStart(const string& weekstart) {
  if (weekstart.empty())
    return;

  Calendar::Weekday ws = DayAndPosition::parseDayOfWeek(weekstart);

  if (ws == Calendar::NO_WEEKDAY) {
    throw invalid_argument("Bad week start \"" + weekstart + "\"");
  }

  setWeekStart(ws);
}


/**
 * Get the array of seconds-of-the-minute constraints for this recurrence. 
 *
 * @return The array of seconds of the minute.
 */
vector<int> Recurrence::getBySecond() const {
  vector<int> b(bySecond);
  
  return b;
}


/**
 * Get a string representing the seconds-of-the-minute constraints for
 * this recurrence.
 *
 * @return The list of seconds of the minute, comma-separated.  "" if this 
 *         recurrence has no such constraints.
 */
string Recurrence::getBySecondString() const {
  return generateIntSetString(getBySecond());
}


/**
 * Set the array of seconds-of-the-minute constraints for this recurrence. 
 *
 * @param b The array of seconds of the minute.
 */
void Recurrence::setBySecond(const vector<int>& b) {
  bySecond = b;
}


/**
 * Set the list of seconds-of-the-minute constraints for this recurrence.
 *
 * @param b The seconds-of-the-minute constraints, as a comma-separated string.
 * @throws invalid_argument If the string is malformed, or if a second
 *         is outside the range 0 - 60.
 */
void Recurrence::setBySecond(const string& b) {
  setBySecond(parseIntSet(b, "second", 0, 60, false));
}


/**
 * Get the array of minutes-of-the-hour constraints for this recurrence. 
 *
 * @return The array of minutes of the hour.
 */
vector<int> Recurrence::getByMinute() const {
  vector<int> b(byMinute);
  return b;
}


/**
 * Get a string representing the minutes-of-the-hour constraints for
 * this recurrence.
 *
 * @return The list of minutes of the hour, comma-separated.  "" if this 
 *         recurrence has no such constraints.
 */
string Recurrence::getByMinuteString() const {
  return generateIntSetString(getByMinute());
}


/**
 * Set the array of minutes-of-the-hour constraints for this recurrence. 
 *
 * @param b The array of minutes of the hour.
 */
void Recurrence::setByMinute(const vector<int>& b) {
  byMinute = b;
}


/**
 * Set the list of minutes-of-the-hour constraints for this recurrence.
 *
 * @param b The minutes-of-the-hour constraints, as a comma-separated string.
 * @throws invalid_argument If the string is malformed, or if a minute
 *         is outside the range 0 - 59.
 */
void Recurrence::setByMinute(const string& b) {
  setByMinute(parseIntSet(b, "minute", 0, 59, false));
}


/**
 * Get the array of hours-of-the-day constraints for this recurrence. 
 *
 * @return The array of hours of the day.
 */
vector<int> Recurrence::getByHour() const {
  vector<int> b(byHour);

  return b;
}


/**
 * Get a string representing the hours-of-the-day constraints for
 * this recurrence.
 *
 * @return The list of hours of the day, comma-separated.  "" if this 
 *         recurrence has no such constraints.
 */
string Recurrence::getByHourString() const {
  return generateIntSetString(getByHour());
}


/**
 * Set the array of hours-of-the-day constraints for this recurrence. 
 *
 * @param b The array of hours of the day.
 */
void Recurrence::setByHour(const vector<int>& b){
  byHour = b;
}


/**
 * Set the list of hours-of-the-day constraints for this recurrence.
 *
 * @param b The hours-of-the-day constraints, as a comma-separated string.
 * @throws invalid_argument If the string is malformed, or if an hour
 *         is outside the range 0 - 23.
 */
void Recurrence::setByHour(const string& b) {
  setByHour(parseIntSet(b, "hour", 0, 23, false));
}


/**
 * Get the array of day-of-the-week and position constraints for this
 * recurrence. 
 *
 * @return The array of days of the week and positions.
 */
vector<DayAndPosition> Recurrence::getByDay() const {
  // XXX I *think* this will do a proper copy.
  vector<DayAndPosition> b(byDay);

  return b;
}


/**
 * Get a string representing the day-of-the-week and position constraints for
 * this recurrence.
 *
 * @return The list of day-of-the-week and position constraints,
 *         comma-separated.  "" if this recurrence has no such constraints.
 */
string Recurrence::getByDayString() const {
  return generateDaySetString(getByDay());
}


/**
 * Set the array of day-of-the-week and position constraints for this
 * recurrence. 
 *
 * @param b The array of days of the week and positions.
 */
void Recurrence::setByDay(const vector<DayAndPosition>& b) {
  // XXX I *think* this will do a proper copy.
  byDay = b;
}


/**
 * Set the list of day-of-the-week and position constraints for this
 * recurrence.
 *
 * @param b The hours-of-the-day constraints, as a comma-separated string.
 * @throws invalid_argument If the string is malformed, or if a day or
 *         position is outside its valid range.
 */
void Recurrence::setByDay(const string& b) {
  setByDay(parseDaySet(b));
}


/**
 * Get the array of day-of-the-month constraints for this recurrence. 
 *
 * @return The array of days of the month.
 */
vector<int> Recurrence::getByMonthDay() const {
  vector<int> b(byMonthDay);

  return b;
}


/**
 * Get a string representing the day-of-the-month constraints for
 * this recurrence.
 *
 * @return The list of days of the month, comma-separated.  "" if this 
 *         recurrence has no such constraints.
 */
string Recurrence::getByMonthDayString() const {
  return generateIntSetString(getByMonthDay());
}


/**
 * Set the array of day-of-the-month constraints for this recurrence. 
 *
 * @param b The array of days of the month.
 */
void Recurrence::setByMonthDay(const vector<int>& b) {
  byMonthDay = b;
}


/**
 * Set the list of day-of-the-month constraints for this recurrence.
 *
 * @param b The day-of-the-month constraints, as a comma-separated string.
 * @throws invalid_argument If the string is malformed, or if a day
 *         is outside the range 1 - 31 or -31 to -1.
 */
void Recurrence::setByMonthDay(const string& b) {
  setByMonthDay(parseIntSet(b, "monthday", 1, 31, true));
}


/**
 * Get the array of day-of-the-year constraints for this recurrence. 
 *
 * @return The array of days of the year.
 */
vector<int> Recurrence::getByYearDay() const {
  vector<int> b(byYearDay);

  return b;
}


/**
 * Get a string representing the day-of-the-year constraints for
 * this recurrence.
 *
 * @return The list of days of the year, comma-separated.  "" if this 
 *         recurrence has no such constraints.
 */
string Recurrence::getByYearDayString() const {
  return generateIntSetString(getByYearDay());
}


/**
 * Set the array of day-of-the-year constraints for this recurrence. 
 *
 * @param b The array of days of the year.
 */
void Recurrence::setByYearDay(const vector<int>& b) {
  byYearDay = b;
}


/**
 * Set the list of day-of-the-year constraints for this recurrence.
 *
 * @param b The day-of-the-year constraints, as a comma-separated string.
 * @throws invalid_argument If the string is malformed, or if a day
 *         is outside the range 1 - 366 or -366 to -1.
 */
void Recurrence::setByYearDay(const string& b) {
  setByYearDay(parseIntSet(b, "yearday", 1, 366, true));
}


/**
 * Get the array of week-number constraints for this recurrence. 
 *
 * @return The array of week numbers.
 */
vector<int> Recurrence::getByWeekNo() const {
  vector<int> b(byWeekNo);

  return b;
}


/**
 * Get a string representing the week-number constraints for
 * this recurrence.
 *
 * @return The list of week numbers, comma-separated.  "" if this 
 *         recurrence has no such constraints.
 */
string Recurrence::getByWeekNoString() const {
  return generateIntSetString(getByWeekNo());
}


/**
 * Set the array of week-number constraints for this recurrence. 
 *
 * @param b The array of week numbers.
 */
void Recurrence::setByWeekNo(const vector<int>& b) {
  byWeekNo = b;
}


/**
 * Set the list of week-number constraints for this recurrence.
 *
 * @param b The week-number constraints, as a comma-separated string.
 * @throws invalid_argument If the string is malformed, or if a week number
 *         is outside the range 1 - 53 or -53 to -1.
 */
void Recurrence::setByWeekNo(const string& b) {
  setByWeekNo(parseIntSet(b, "weekno", 1, 53, true));
}


/**
 * Get the array of month constraints for this recurrence.  Months are
 * specified using the constants from {@link Calendar}
 * ({@link Calendar#JANUARY}, etc.).
 *
 * @return The array of months.
 * @see Calendar#JANUARY
 * @see Calendar#FEBRUARY
 * @see Calendar#MARCH
 * @see Calendar#APRIL
 * @see Calendar#MAY
 * @see Calendar#JUNE
 * @see Calendar#JULY
 * @see Calendar#AUGUST
 * @see Calendar#SEPTEMBER
 * @see Calendar#OCTOBER
 * @see Calendar#NOVEMBER
 * @see Calendar#DECEMBER
 */
vector<int> Recurrence::getByMonth() const {
  vector<int> b(byMonth);

  return b;
}


/**
 * Get a string representing the month constraints for this recurrence. 
 * Months are counted as January = 1.
 *
 * @return The list of months, comma-separated.  "" if this 
 *         recurrence has no such constraints.
 */
string Recurrence::getByMonthString() const {
  vector<int> b = getByMonth();

  for (vector<int>::size_type i = 0; i < b.size(); i++) {
    b[i] += 1 - Calendar::JANUARY;
  }
  return generateIntSetString(b);
}


/**
 * Set the array of month constraints for this recurrence.  Months should
 * be specified using the constants from {@link Calendar}
 * ({@link Calendar#JANUARY}, etc.).
 *
 * @param b The array of months.
 * @see Calendar#JANUARY
 * @see Calendar#FEBRUARY
 * @see Calendar#MARCH
 * @see Calendar#APRIL
 * @see Calendar#MAY
 * @see Calendar#JUNE
 * @see Calendar#JULY
 * @see Calendar#AUGUST
 * @see Calendar#SEPTEMBER
 * @see Calendar#OCTOBER
 * @see Calendar#NOVEMBER
 * @see Calendar#DECEMBER
 */
void Recurrence::setByMonth(const vector<int>& b) {
  byMonth = b;
}


/**
 * Set the list of month constraints for this recurrence.  Months are counted
 * as January = 1.
 *
 * @param b The month constraints, as a comma-separated string.
 * @throws invalid_argument If the string is malformed, or if a week number
 *         is outside the range 1 - 12.
 */
void Recurrence::setByMonth(const string& b) {
  vector<int> bm = parseIntSet(b, "month", 1, 12, false);

  for (vector<int>::size_type i = 0; i < bm.size(); i++) {
    bm[i] += Calendar::JANUARY - 1;
  }
  setByMonth(bm);
}


/**
 * Test if the specified time falls within a repetition of this
 * recurrence.  See
 * <a href="../../../algorithm.txt"><tt>algorithm.txt</tt></a> for details
 * of how this works. 
 *
 * @param current The "current" time, i.e. the time which is being
 *                considered.
 * @return Whether it falls within a recurrence.
 */
bool Recurrence::isInRecurrence(const Calendar& current) const {
  return isInRecurrence(current, false);
}


/**
 * Debugging interface to {@link #isInRecurrence(java.util.Calendar)}.
 *
 * @param current The "current" time, i.e. the time which is being
 *                considered.
 * @param debug   Whether to print debugging information.
 * @return Whether it falls within a recurrence.
 * @see #isInRecurrence(java.util.Calendar)
 */
bool Recurrence::isInRecurrence(const Calendar& current, bool debug) const {
  Calendar myCurrent = current;

  setCalendarWeekStart(myCurrent, dtStart->getFirstDayOfWeek());

  if (myCurrent.getTime() < dtStart->getTime()) {
    // The current time is earlier than the start time.
    if (debug) { cerr << "current < start" << endl; }
    return false;
  }

  if (myCurrent.getTime() <
      dtStart->getTime() + duration->getInterval()) {
    // We are within "duration" of dtStart.
    if (debug) { cerr << "within duration of start" << endl; }
    return true;
  }
  
  /* If we have 'count' and not 'until', compute 'until' now. */
  computeUntilFromCount();
  
  Calendar* candidate = internalGetCandidateStartTime(myCurrent, false, debug);
  /* Check whether internalGetCandidateStartTime could find a candidate */
  if (candidate == NULL) {
    if (debug) { cerr << "No candidate start times" << endl; }
    return false;
  }
  if (candidate->getTime() > myCurrent.getTime()) {
    delete candidate;
    throw
      runtime_error("Internal error: candidate > current");
  }

  /* Loop over ranges for the duration. */
  while (candidate->getTime() + duration->getInterval() >
         myCurrent.getTime()) {
    if (candidateIsInRecurrence(*candidate, debug)) {
      delete candidate;
      return true;
    }
    
    time_t oldCandidateTime = candidate->getTime();
    
    /* Roll back to one second previous, and try again. */
    candidate->add(Calendar::SECOND, -1);
    
    /* Make sure we haven't rolled back to before dtStart. */
    if (candidate->getTime() < dtStart->getTime()) {
      if (debug) { cerr << "No candidates after dtStart" << endl; }
      delete candidate;
      return false;
    }
    
    Calendar* oldCandidate = candidate;
    candidate = internalGetCandidateStartTime(*oldCandidate, false, debug);
    delete oldCandidate;
    
    /* Check whether internalGetCandidateStartTime could find a candidate */
    if (candidate == NULL) {
      if (debug) { cerr << "No candidate start times" << endl; }
      return false;
    }
    if (candidate->getTime() >= oldCandidateTime) {
      delete candidate;
      throw
        runtime_error("Internal error: candidate >= old candidate");
    }
  }

  if (debug) { cerr << "No matching candidates" << endl; }
  delete candidate;
  return false;
}


/**
 * Check whether a particular candidate start time falls within a
 * reptition of the recurrence.  Support function for
 * {@link #isInRecurrence}.
 * @param candidate The candidate start time to be checked.
 * @param debug     Whether to print debugging information.
 * @return   whether it falls within a recurrence
 */
bool Recurrence::candidateIsInRecurrence(const Calendar& candidate,
                                         bool debug) const {
  if (until != NULL &&
      candidate.getTime() > until->getTime()) {
    // After "until"
    if (debug) { cerr << "after until" << endl; }
    return false;
  }
  
  if (getRecurrenceCount(candidate) % interval != 0) {
    // Not a repetition of the interval
    if (debug) { cerr << "not an interval rep" << endl; }
    return false;
  }
  
  if (!matchesBySecond(candidate) || !matchesByMinute(candidate) ||
      !matchesByHour(candidate) || 
      !matchesByDay(candidate) || !matchesByMonthDay(candidate) ||
      !matchesByYearDay(candidate) ||
      !matchesByWeekNo(candidate) || !matchesByMonth(candidate)) {
    // Doesn't match a by* rule
    if (debug) { cerr << "doesn't match a by*" << endl; }
    return false;
  }
  if (debug) { cerr << "All checks succeeded" << endl; }
  return true;
}


/**
 * Get the minimum interval for this recurrence, based on 
 * the by* parameters and the frequency.
 * 
 * @return One of {@link #SECONDLY}, {@link #MINUTELY}, {@link #HOURLY},
 * {@link #DAILY}, {@link #WEEKLY}, {@link #MONTHLY},
 * {@link #YEARLY}, or {@link #NO_RECURRENCE} (infinite).
 */
Recurrence::Frequency Recurrence::getMinimumInterval() const {
  if (frequency == SECONDLY || bySecond.size() > 0) {
    return SECONDLY;
  }
  if (frequency == MINUTELY || byMinute.size() > 0) {
    return MINUTELY;
  }
  if (frequency == HOURLY || byHour.size() > 0) {
    return HOURLY;
  }
  if (frequency == DAILY || byDay.size() > 0 || byMonthDay.size() > 0
      || byYearDay.size() > 0) {
    return DAILY;
  }
  else if (frequency == WEEKLY || byWeekNo.size() > 0) {
    return WEEKLY;
  }
  else if (frequency == MONTHLY || byMonth.size() > 0) {
    return MONTHLY;
  }
  else if (frequency == YEARLY) {
    return YEARLY;
  }
  else if (frequency == NO_RECURRENCE) {
    return NO_RECURRENCE;
  }
  else {
    // Shouldn't happen
    throw
      runtime_error("Internal error: Unknown frequency value");
  }
}


/**
 * Get the candidate start time for this recurrence, for a given date.
 * The candidate start time is the previous/next repitition of the minimum
 * interval, counting from dtStart.  See
 * <a href="../../../algorithm.txt"><tt>algorithm.txt</tt></a> for more
 * details.
 *
 * @param current    The "current" time, i.e. the time which is being
 *                   considered.
 * @param is_forward Whether to find the next (true) or previous (false)
 *                   candidate start time.
 * @param debug      Whether to print debugging information.
 *
 * @return The candidate start time, or NULL if there is none.
 */
Calendar* Recurrence::getCandidateStartTime(const Calendar& current,
                                            bool is_forward, bool debug) const
{
  Calendar myCurrent = current;
  setCalendarWeekStart(myCurrent, dtStart->getFirstDayOfWeek());
  
  computeUntilFromCount();
  
  return internalGetCandidateStartTime(current, is_forward, debug);
}


/**
 * Internal version of getCandidateStartTime.  Doesn't check various
 * parameters. 
 *
 * @param current    The "current" time, i.e. the time which is being
 *                   considered.
 * @param is_forward Whether to find the next (true) or previous (false)
 *                   candidate start time.
 * @param debug      Whether to print debugging information.
 *
 * @return The candidate start time, or NULL if there is none.
 *
 * @see #getCandidateStartTime(Calendar, bool, bool)
 */
Calendar* Recurrence::internalGetCandidateStartTime(const Calendar& current,
                                                    bool is_forward,
                                                    bool debug) const
{
  // XXX Testing
  // Force current to be recalculated.
  // current.clear(Calendar::WEEK_OF_YEAR);
  
  if (dtStart->getTime() > current.getTime()) {
    if (debug) {
      cerr << "Current time before DtStart" << endl;
    }
    if (is_forward) {
      return new Calendar(*dtStart);
    }
    else {
      return NULL;
    }
  }
  
  /* If 'until' has not yet been computed, this next statement is a no-op.
   * That's OK. */
  if (is_forward && until != NULL &&
      until->getTime() > current.getTime()) {
    if (debug) {
      cerr << "Current time after Until" << endl;
    }
    return NULL;
  }

  if (frequency == NO_RECURRENCE) {
    if (is_forward) {
      return NULL;
    }
    else {
      return new Calendar(*dtStart);
    }
  }

  Frequency minInterval = getMinimumInterval();

  Calendar* candidate = new Calendar(current);

  if (minInterval == SECONDLY) {
    return candidate;
  }
  adjust_constant_length_field(Calendar::SECOND, dtStart, candidate,
                               is_forward);

  if (minInterval == MINUTELY) {
    return candidate;
  }
  adjust_constant_length_field(Calendar::MINUTE, dtStart, candidate,
                               is_forward);

  if (minInterval == HOURLY) {
    return candidate;
  }
  adjust_constant_length_field(Calendar::HOUR_OF_DAY, dtStart, candidate,
                               is_forward);
  
  switch (minInterval) {
  case DAILY:
    /* No more adjustments needed */
    break;
  case WEEKLY:
    adjust_constant_length_field(Calendar::DAY_OF_WEEK, dtStart, candidate,
                                 is_forward);
    break;
  case MONTHLY:
    adjust_day_of_month(dtStart, candidate, is_forward);
    break;
  case YEARLY:
    adjust_day_of_year(dtStart, candidate, is_forward);
    break;
  default:
    // Shouldn't get here
    delete candidate;
    throw
      runtime_error("Internal error: unknown frequency value, shouldn't get here");
    
  }
  
  return candidate;
}


/** 
 * Adjust the time specified by <code>candidate</code> so that the field
 * <code>field</code> equals that of <code>start</code>.  If
 * <code>is_forward</code>, increase the time, otherwise reduce it.
 * Adjust other fields appropriately.  Only works for constant-length fields
 * (e.g., {@link Calendar#SECOND}, {@link Calendar#MINUTE},
 * {@link Calendar#HOUR_OF_DAY}, or {@link Calendar#DAY_OF_WEEK}).
 *
 * @param field The calendar field to reduce.
 * @param start The calendar to get the field from
 * @param candidate The calendar to adjust
 * @param is_forward Whether to adjust the time forward.
 * @throws invalid_argument <code>field</code> is not a
 *                                  constant-length field
 */
void Recurrence::adjust_constant_length_field(Calendar::Field field,
                                              const Calendar* start,
                                              Calendar* candidate,
                                              bool is_forward)
{
  if (start->getMaximum(field) != start->getLeastMaximum(field) ||
      start->getMinimum(field) != start->getGreatestMinimum(field)) {
    throw invalid_argument("Not a constant length field");
  }
  
  int fieldLength =
    (start->getMaximum(field) - start->getMinimum(field) + 1);
  
  int delta = start->get(field) - candidate->get(field);
  
  if (is_forward && delta < 0) {
    delta += fieldLength;
  }
  else if (!is_forward && delta > 0) {
    delta -= fieldLength;
  }
  
  candidate->add(field, delta);
}


/** 
 * Adjust the time specified by <code>candidate</code> until its day of
 * the month equals that of <code>start</code>.  If <code>is_forward</code>,
 * increase the time, otherwise reduce it.  Adjust other fields
 * appropriately.
 *
 * <p>Months can vary in length.  The iCal spec says:
 * "If BYxxx rule part values are found which are beyond the available
 * scope (ie, BYMONTHDAY=30 in February), they are simply ignored."
 * Assume this applies to times derived from DTStart as well (pending
 * response from iCal WG mailing list), and roll back/forward through
 * previous months until we find a match.
 *
 * <p>Example of how this works: consider the current date October 3, and
 * the start date January 31.  Assume !is_forward.  start - candidate
 * yields 28, which is more than 0, so we subtract the number of days in
 * the previous month (September), which is 30.  The resulting value, 1,
 * of the month day, is not equal to the start month day (31), so we
 * subtract the number of days in the second-previous month (31).  This
 * gives us August 31; this agrees, so we stop.
 *
 * <p>When we roll forward, by contrast, we adjust by the number of days
 * in the <i>current</i> month, not the next month.  (What matters is the
 * number of days in the month whose end is being crossed.)
 *
 * <p>Because there are never two consecutive months with fewer than 31
 * days, this algorithm will try at most two months (in the Gregorian
 * calendar).
 *
 * @param start The calendar with the target day of the month
 * @param candidate The calendar to adjust
 * @param is_forward Whether to adjust the time forward.
 */
void Recurrence::adjust_day_of_month(const Calendar* start,
                                     Calendar* candidate,
                                     bool is_forward)
{
  Calendar tempCal(*candidate);

  int delta = start->get(Calendar::DATE) - candidate->get(Calendar::DATE);

  if (is_forward && delta < 0) {
    delta += candidate->getActualMaximum(Calendar::DATE);
  }
  else if (!is_forward && delta > 0) {
    tempCal.add(Calendar::MONTH, -1);
    delta -= tempCal.getActualMaximum(Calendar::DATE);
  }

  candidate->add(Calendar::DATE, delta);
    
  while (start->get(Calendar::DATE) != candidate->get(Calendar::DATE)) {
    if (is_forward) {
      delta = candidate->getActualMaximum(Calendar::DATE) - 1;
    }
    else {
      tempCal.add(Calendar::MONTH, -1);
      delta = - tempCal.getActualMaximum(Calendar::DATE);
    }

    candidate->add(Calendar::DATE, delta);
  }
}


/** 
 * Adjust the time specified by <code>candidate</code> until its day of
 * the year equals that of <code>start</code>.   If <code>is_forward</code>,
 * increase the time, otherwise reduce it.  Adjust other fields
 * appropriately.
 *
 * <p>The day of year is calculated based on (month, day) pairs, not
 * the numeric day of the year (which is different after February
 * in leap years).
 *
 * <p>For start dates of February 29, this algorithm will loop through at
 * most seven years before it finds a matching year (in the Gregorian
 * calendar).  For candidates between 1901 and 2099, it will loop through
 * at most three years.  Start dates which do not fall on a leap day will
 * not loop at all.
 *
 * @param start The calendar with the target day of the month
 * @param candidate The calendar to adjust
 */
void Recurrence::adjust_day_of_year(const Calendar* start,
                                    Calendar* candidate,
                                    bool is_forward)
{
  if (!is_forward &&
      (start->get(Calendar::MONTH) > candidate->get(Calendar::MONTH)
       || (start->get(Calendar::MONTH) == candidate->get(Calendar::MONTH)
           && start->get(Calendar::DATE) > candidate->get(Calendar::DATE)))) {
    /**
     * Start date is later in the year than candidate date.  Roll
     * the candidate back to the previous year.
     */
    candidate->add(Calendar::YEAR, -1);
  }

  if (is_forward &&
      (start->get(Calendar::MONTH) < candidate->get(Calendar::MONTH)
       || (start->get(Calendar::MONTH) == candidate->get(Calendar::MONTH)
           && start->get(Calendar::DATE) < candidate->get(Calendar::DATE)))) {
    /**
     * Start date is earlier in the year than candidate date.  Roll
     * the candidate forward to the next year.
     */
    candidate->add(Calendar::YEAR, 1);
  }

  /* Set the candidate date to the start date. */
  candidate->set(Calendar::MONTH, start->get(Calendar::MONTH));
  candidate->set(Calendar::DATE, start->get(Calendar::DATE));
  
  int increment = is_forward ? 1 : -1;
  
  /**
   * If the candidate date doesn't equal the start date (which can happen
   * if the start date doesn't exist in this year, i.e. Feb 29)
   * try the next/previous year until we get a match.
   */
  while (start->get(Calendar::MONTH) != candidate->get(Calendar::MONTH)
         || start->get(Calendar::DATE) != candidate->get(Calendar::DATE)) {
    candidate->add(Calendar::YEAR, increment);
    
    candidate->set(Calendar::MONTH, start->get(Calendar::MONTH));
    candidate->set(Calendar::DATE, start->get(Calendar::DATE));
  }
}


/**
 * The last time we're willing to search for <code>count</code> to resolve
 * it to <code>until</code>.  If a recurrence exceeds this instant before
 * it's filled its counts, consider it infinite, or throw an exception,
 * depending on the value of strict_count_bounds.
 *
 * <p>The default value of this parameter is 2000000000L, or
 * Wednesday, May 18, 2033 03:33:20.000 UTC.  This is comfortably less
 * than time_t (2^31 - 1), so we don't have to worry about rounding on
 * systems which implement Calendar's support functions on a 
 * 32-bit time_t.
 *
 * @see #computeUntilFromCount()
 * @see #strict_count_bounds
 */
time_t Recurrence::max_count_time = 2000000000L;


/**
 * The maximum number of candidate start times we're willing to consider
 * for <code>count</code> to resolve it to <code>until</code>.  If a
 * recurrence exceeds this instant before it's filled its counts, consider
 * it infinite, or throw an exception, depending on the value of
 * strict_count_bounds.
 *
 * <p>The default value of this parameter is 10000.
 *
 * @see #computeUntilFromCount()
 * @see #strict_count_bounds
 */
int Recurrence::max_count_loops = 10000;


/**
 * Whether to throw an exception if a conversion from <code>count</code>
 * to <code>until</code> exceeds its defined bounaries, or simply to
 * assume the count is effectively infinite. 
 *
 * <p>The default value of this parameter is false.
 *
 * @see #computeUntilFromCount()
 * @see #max_count_time
 * @see #max_count_loops
 */
bool Recurrence::strict_count_bounds = false;


/**
 * If <code>count</code> is set and <code>until</code> is not, calculate
 * <code>until</code> from <code>count</code>.  <code>until</code> will be
 * set to one second after the <code>count</code>'th start time of the
 * interval.
 *
 * If we do not reach <code>count</code> recurrences before we reach
 * <code>max_count_time</code>, or before we have tried
 * <code>max_count_loops</code> candidate start times, set
 * <code>until</code> to <code>max_count_time</code>, or throw an
 * exception, depending on the state of <code>strict_count_bounds</code>.
 *
 * @throws invalid_argument If we run off the end, and
 *         <code>strict_count_bounds</code> is in effect.
 * @see #max_count_time
 * @see #max_count_loops
 * @see #strict_count_bounds
 */
void Recurrence::computeUntilFromCount() const
{
  // XXX casting away const!
  // This is so we can call this from const methods, which are const
  // except for this one aspect.
  const_cast<Recurrence*>(this)->computeUntilFromCount(false);
}


/**
 * Debugging interface to {@link #computeUntilFromCount()}.
 *
 * @throws invalid_argument If we run off the end, and
 *         <code>strict_count_bounds</code> is in effect.
 * @see #computeUntilFromCount()
 */
void Recurrence::computeUntilFromCount(bool debug)
{
  if (until != NULL) {
    return;
  }
  
  if (count == 0) {
    until = NULL;
    return;
  }
  
  int starts_found = 0;
  int loops;
  
  Calendar* candidate = new Calendar(*dtStart);
  Calendar* oldCandidate;

  for (loops = 0;
       loops < max_count_loops &&
         candidate->getTime() < max_count_time;
       loops++, candidate->add(Calendar::SECOND, 1)) {
    long oldCandidateTime = candidate->getTime();
    
    oldCandidate = candidate;
    candidate = internalGetCandidateStartTime(*oldCandidate, true, false);
    delete oldCandidate;

    if (candidate == NULL) {
      if (debug) { cerr << "No next candidate start" << endl; }
      break;
    }

    if (candidate->getTime() < oldCandidateTime) {
      delete candidate;
      throw
        runtime_error("Internal error: candidate < old candidate");
    }
      
    if (candidateIsInRecurrence(*candidate, debug)) {
      starts_found++;
      if (starts_found == count) {
        candidate->add(Calendar::SECOND, 1);
        until = candidate;
        return;
      }
    }
  }

  /* If we get here, we've fallen or broken out of the loop */
  delete candidate;
  if (debug) { cerr << "Could not resolve count to until" << endl; }
  if (strict_count_bounds) {
    throw
      invalid_argument("Could not resolve count to until");
  }
  else {
    until = new Calendar(*dtStart);
    until->setTime(max_count_time);
    return;
  }
}


/**
 * Get the number of instances of the current recurrence frequency
 * that have occured between this recurrence's dtStart, and the
 * time specified by <code>candidate</code>.
 *
 * @param candidate The end time of the period
 * @return The number of recurrences
 */
long Recurrence::getRecurrenceCount(const Calendar& candidate) const
{
  switch (frequency) {
  case NO_RECURRENCE:
    return 0;
    
  case SECONDLY:
  case MINUTELY:
  case HOURLY:
  case DAILY:
    return (long)(getFieldNumber(candidate, frequency) -
                  getFieldNumber(*dtStart, frequency));
    
  case WEEKLY:
    {
      Calendar tempCand = candidate;
      tempCand.setFirstDayOfWeek(dtStart->getFirstDayOfWeek());
      return (long)(getWeekNumber(tempCand) - getWeekNumber(*dtStart));
    }
    
  case MONTHLY:
    return (long)(getMonthNumber(candidate) - getMonthNumber(*dtStart));
    
  case YEARLY:
    return (long)(candidate.get(Calendar::YEAR) -
                  dtStart->get(Calendar::YEAR));
    
  default:
    throw runtime_error("Internal error: bad frequency");
  }
}


/**
 * Get the field number corresponding to the date of this calendar
 * (field intervals - seconds, minutes, hours, or days - since the Epoch).
 *
 * @param cal The calendar to calculate.
 * @param field The field to calculate - one of {@link #SECONDLY},
 * {@link #MINUTELY}, {@link #HOURLY}, or  {@link #DAILY}.
 * 
 * @return The field number.
 */
long Recurrence::getFieldNumber(const Calendar& cal, Frequency field)
{
  Calendar tempCal = cal;
  long multiple = 1; /* ticks per sec */

  if (field > SECONDLY) {
    tempCal.set(Calendar::SECOND, 0);
    multiple *= 60; /* secs per min */
  }
  if (field > MINUTELY) {
    tempCal.set(Calendar::MINUTE, 0);
    multiple *= 60; /* mins per hour */
  }
  if (field > HOURLY) {
    tempCal.set(Calendar::HOUR_OF_DAY, 0);
    multiple *= 24; /* hours per day */
  }
    
  return tempCal.getTime() / multiple;
}


/**
   * Get the week number corresponding to the date of this calendar
   * (weeks since the Epoch).  This depends on the calendar's
   * current {@link #getWeekStart} setting.
   *
   * @param cal The calendar to calculate.
   * @return The week number.
   * @see #setWeekStart
   * @see Calendar#setFirstDayOfWeek
   */
long Recurrence::getWeekNumber(const Calendar& cal)
{
  Calendar tempCal = cal;

  // Set to midnight
  tempCal.set(Calendar::SECOND, 0);
  tempCal.set(Calendar::MINUTE, 0);
  tempCal.set(Calendar::HOUR_OF_DAY, 0);
  
  // Roll back to the first day of the week
  int delta = tempCal.getFirstDayOfWeek() -
    tempCal.get(Calendar::DAY_OF_WEEK);
    
  if (delta > 0) { delta -= 7; }
  
  // tempCal now points to the first instant of this week.
  
  // Calculate the "week epoch" -- the weekstart day closest to January 1,
  // 1970 (which was a Thursday)
  
  long weekEpoch = 
    (tempCal.getFirstDayOfWeek() - Calendar::THURSDAY) *
    24 * 60 * 60;
  
  return (tempCal.getTime() - weekEpoch) /
    (7 * 24 * 60 * 60);
}


/**
 * Get the month number corresponding to the date of this calendar
 * (months since the Epoch).
 *
 * @param cal The calendar to calculate.
 * @return The month number.
 */
long Recurrence::getMonthNumber(const Calendar& cal)
{
  return (cal.get(Calendar::YEAR) - 1970) * 12 +
    (cal.get(Calendar::MONTH) - Calendar::JANUARY);
}


/**
 * Determine if <code>candidate</code> matches the recurrence's
 * {@link #byDay} rules.
 * @param candidate The candidate date to check.
 * @return Whether it matches.
 */
bool Recurrence::matchesByDay(const Calendar& candidate) const
{
  if (byDay.size() == 0) {
    /* No byDay rules, so it matches trivially */
    return true;
  }

  unsigned int i;

  for (i = 0; i < byDay.size(); i++) {
    if (matchesIndividualByDay(candidate, byDay[i])) {
      return true;
    }
  }
  
  return false;
}
  
/**
 *
 * Determine if <code>candidate</code> matches one individual
 * {@link DayAndPosition}.
 * @param candidate the candidate date to check
 * @param pos the candidate position
 * @return whether it matches
 */
bool Recurrence::matchesIndividualByDay(const Calendar& candidate,
                                        const DayAndPosition& pos) const
{
  if (pos.getDayOfWeek() != candidate.get(Calendar::DAY_OF_WEEK)) {
    return false;
  }
  int position = pos.getDayPosition();
  
  if (position == 0) {
    return true;
  }
  
  Calendar::Field field;
  
  /**
   * Calendar has a DAY_OF_WEEK_IN_MONTH field which
   * holds the "positive" value of this function within the week.
   * It hasn't got a negative version of it, though, or a
   * DAY_OF_WEEK_IN_YEAR field, so calculate these manually.
   */
  
  switch (frequency) {
  case MONTHLY:
    field = Calendar::DAY_OF_MONTH;
    break;
  case YEARLY:
    field = Calendar::DAY_OF_YEAR;
    break;
  default:
    throw logic_error("byday has a day position " 
                      "in non-MONTHLY or YEARLY recurrence");
  }
  
  if (position > 0) {
    int day_of_week_in_field = 
      ((candidate.get(field) - 1) / 7) + 1;
    return (position == day_of_week_in_field);
  }
  else {
    /* position < 0 */
    int negative_day_of_week_in_field = 
      ((candidate.getActualMaximum(field) - candidate.get(field)) / 7) + 1;
    return (-position == negative_day_of_week_in_field);
  }
}

/**
 * Determine if field <code>field</code> of <code>candidate</code> matches
 * an entry in <code>array</code>
 *
 * @param array The by* array to check
 * @param field The Calendar field to check
 * @param candidate The Candidate date to check
 * @param allowNegative Whether negative numbers should count as
 *                      N-from-the-end
 * @return whether it matches.
 */
bool Recurrence::matchesByField(const vector<int>& array,
                                Calendar::Field field,
                                const Calendar& candidate,
                                bool allowNegative) const
{
  if (array.size() == 0) {
    /* No rules, so it matches trivially */
    return true;
  }

  unsigned int i;

  for (i = 0; i < array.size(); i++) {
    int val;
    if (allowNegative && array[i] < 0) {
      // byMonthDay = -1, in a 31-day month, means 31
      int max = candidate.getActualMaximum(field);
      val = (max + 1) + array[i];
    }
    else {
      val = array[i];
    }
    if (val == candidate.get(field)) {
      return true;
    }
  }
  
  return false;
}
  
/**
 * Determine if <code>candidate</code> matches the recurrence's
 * {@link #bySecond} rules.
 * @param candidate the Candidate date to check
 * @return whether it matches.
 */
bool Recurrence::matchesBySecond(const Calendar& candidate) const
{
  return matchesByField(bySecond, Calendar::SECOND, candidate, false);
}
  
/**
 * Determine if <code>candidate</code> matches the recurrence's
 * {@link #byMinute} rules.
 * @param candidate the Candidate date to check
 * @return whether it matches.
 */
bool Recurrence::matchesByMinute(const Calendar& candidate) const
{
  return matchesByField(byMinute, Calendar::MINUTE, candidate, false);
}

/**
 * Determine if <code>candidate</code> matches the recurrence's
 * {@link #byHour} rules.
 * @param candidate the Candidate date to check
 * @return whether it matches.
 */
bool Recurrence::matchesByHour(const Calendar& candidate) const
{
  return matchesByField(byHour, Calendar::HOUR_OF_DAY, candidate, false);
}
  
/**
 * Determine if <code>candidate</code> matches the recurrence's
 * {@link #byMonthDay} rules.
 * @param candidate the Candidate date to check
 * @return whether it matches.
 */
bool Recurrence::matchesByMonthDay(const Calendar& candidate) const
{
  return matchesByField(byMonthDay, Calendar::DATE, candidate, true);
}

/**
 * Determine if <code>candidate</code> matches the recurrence's
 * {@link #byYearDay} rules.
 * @param candidate the Candidate date to check
 * @return whether it matches.
 */
bool Recurrence::matchesByYearDay(const Calendar& candidate) const
{
  return matchesByField(byYearDay, Calendar::DAY_OF_YEAR, candidate, true);
}
  
/**
 * Determine if <code>candidate</code> matches the recurrence's
 * {@link #byWeekNo} rules.
 * @param candidate the Candidate date to check
 * @return whether it matches.
 */
bool Recurrence::matchesByWeekNo(const Calendar& candidate) const
{
  return matchesByField(byWeekNo, Calendar::WEEK_OF_YEAR, candidate, true);
}

/**
 * Determine if <code>candidate</code> matches the recurrence's
 * {@link #byMonth} rules.
 * @param candidate the Candidate date to check
 * @return whether it matches.
 */
bool Recurrence::matchesByMonth(const Calendar& candidate) const
{
  return matchesByField(static_cast<vector<int> >(byMonth),
                        Calendar::MONTH, candidate, false);
}

/**
 * Set the given calendar's week start value to <code>start</code>, and
 * set its minimal days in first week to 4.
 * Clear its <code>WEEK_OF_MONTH</code> and <code>WEEK_OF_YEAR</code>
 * fields - setting the values doesn't do this, so old values can persist.
 * Note: this works for calendars set by the usual iCalendar/CPL syntax,
 * but can fale if someone actually specified the current time with 
 * <code>WEEK_OF_MONTH</code> or <code>WEEK_OF_YEAR</code>.
 *
 * @param cal The calendar to reset
 * @param start The start day of the wee
 */
void Recurrence::setCalendarWeekStart(Calendar& cal, Calendar::Weekday start)
{
  cal.setMinimalDaysInFirstWeek(4);
  cal.setFirstDayOfWeek(start);
  cal.clear(Calendar::WEEK_OF_MONTH);
  cal.clear(Calendar::WEEK_OF_YEAR);
}



/**
 * Parse the given Date-Time string into a Calendar object.  Internal function
 * used as a support function for set*(std::string). <p>
 *
 * The syntax of a Date-Time is given in RFC 2445 as follows:
 * <pre>
 *   date-time  = date "T" time ;As specified in the date and time
 *                              ;value definitions
 *
 *   date               = date-value
 *
 *   date-value         = date-fullyear date-month date-mday
 *   date-fullyear      = 4DIGIT
 *
 *   date-month         = 2DIGIT        ;01-12
 *   date-mday          = 2DIGIT        ;01-28, 01-29, 01-30, 01-31
 *                                      ;based on month/year
 *
 *   time               = time-hour time-minute time-second [time-utc]
 *
 *   time-hour          = 2DIGIT        ;00-23
 *   time-minute        = 2DIGIT        ;00-59
 *   time-second        = 2DIGIT        ;00-60
 *   ;The "60" value is used to account for "leap" seconds.
 *
 *   time-utc   = "Z"
 * </pre>
 * 
 *
 * @param  s  The string representation of the date-time.
 * @returns The date-time
 * @throws invalid_argument If the given string does not describe a valid
 *         Date-Time
 */
Calendar Recurrence::parseDateTime(const std::string& str)
{
  int y, mo_int, d, h, m, sec;
  Calendar::Month mo;
  bool zulu = false;

  /* The grammar of date-time defines it to be either 15 (local) or 16 (zulu)
   * characters. */
  if (str.size() < 15 || str.size() > 16 ||
      (str.size() == 16 && str[15] != 'Z')) {
    throw invalid_argument("Bad date-time string: incorrect length");
  }
  if (str[8] != 'T') {
    throw invalid_argument("Bad date-time string: missing \'T\'");
  }
  if (str.size() == 16 && str[15] == 'Z') {
    zulu = true;
  }

  y  = getFixedSizeInt(str, 0, 4, "year");
  mo_int = getFixedSizeInt(str, 4, 2, "month");
  d  = getFixedSizeInt(str, 6, 2, "day of month");
  
  h   = getFixedSizeInt(str, 9, 2, "hour");
  m   = getFixedSizeInt(str, 11, 2, "minute");
  sec = getFixedSizeInt(str, 13, 2, "second");

  /* Make mo Calendar::JANUARY-based rather than 1-based */
  mo = static_cast<Calendar::Month>(mo_int + Calendar::JANUARY - 1);

  Calendar ret(y, mo, d, h, m, sec);

  if (zulu) {
    ret.setTimeZone("");
  }

  /* Validate fields by normalizing and making sure they've stayed
   * the same.
   */
  ret.getTime(); // Force normalization

  if (ret.get(Calendar::SECOND) != sec) {
    throw invalid_argument("Bad date-time string: second out of range");
  }
  if (ret.get(Calendar::MINUTE) != m) {
    throw invalid_argument("Bad date-time string: minute out of range");
  }
  if (ret.get(Calendar::HOUR_OF_DAY) != h) {
    throw invalid_argument("Bad date-time string: hour out of range");
  }
  if (ret.get(Calendar::DAY_OF_MONTH) != d) {
    throw invalid_argument("Bad date-time string: day of month out of range");
  }
  if (ret.get(Calendar::MONTH) != mo) {
    throw invalid_argument("Bad date-time string: month out of range");
  }
  if (ret.get(Calendar::YEAR) != y) {
    throw invalid_argument("Bad date-time string: year out of range");
  }

  return ret;
}


/**
 * Support function for parseDateTime.  Get a fixed-size integer from
 * the given string.  Throw an exception if the string isn't entirely
 * numeric.
 * 
 * @param str The string from which to extract the integer
 * @param pos The position of the integer
 * @param ext The extent of the integer
 */
long Recurrence::getFixedSizeInt(const std::string& str,
                                 size_t pos, size_t ext,
                                 const char * fieldname)
{
  char buf[80]; /* Should be well more than the number of digits in a long. */
  if (ext >= sizeof(buf)) {
    throw logic_error("getFixedSizeInt: ext too large");
  }
  
  str.copy(buf, ext, pos);
  buf[ext] = '\0';

  char *endpos;

  long ret = strtol(buf, &endpos, 10);
  if (endpos != buf + ext) {
    throw invalid_argument("Bad date-time string: non-numeric "
                           + string(fieldname));
  }
  return ret;
}


/**
 * Parse the given comma-separated list of integers into a vector<int>.
 * used as a support function for setByXXX(const std::string&), other than
 * setByDay.<p>
 *
 * @param  s    The string representation of the integer list.
 * @param  name The name of the integer field (for exceptions)
 * @param  min  The minimum integer value allowed
 * @param  max  The maximum integer value allowed
 * @param  neg_ok Whether negative values are allowed
 * @returns The list of integers
 * @throws invalid_argument If the given string does not describe a valid
 *         list of integers complying with the above requirements
 */
vector<int> Recurrence::parseIntSet(const string& str, const char *name,
                                    int min, int max, bool neg_ok)
{
  vector<int> ret;

  string::const_iterator s = str.begin(), end = str.end();
  int value = 0;
  bool negative_value = false;

  while (s != end) {
    if (!isdigit(*s) && *s != '+' && *s != '-') {
      throw invalid_argument("Bad by" + string(name)
                             + " list: unexpected character \'"
                             + string(s, s+1) + "\'");
    }

    negative_value = false;
    if (*s == '-') {
      negative_value = true;
    }

    if (*s == '-' || *s == '+') {
      s++;
      if (s == end || !isdigit(*s)) {
        throw invalid_argument("Bad by" + string(name) +
                               " list: \'+\' or \'-\' without number");
      }
    } 

    for (value = 0; s != end && isdigit(*s); s++) {
      value = value * 10 + (*s - '0');
    }

    if (value < min || value > max) {
      ostrstream buf;
      buf << "Out-of-range " << name << " ";
      if (negative_value) { buf << "-"; }
      buf << value << " in by" << name << " list" << ends;
      invalid_argument err(buf.str());
      buf.rdbuf()->freeze(0);
      throw err;
    }
    if (negative_value) { value = -value; }
    ret.push_back(value);

    if (s != end) {
      if (*s != ',') {
        throw invalid_argument("Bad by" + string(name) +
                               " list: unexpected character \'" + 
                               string(s, s+1) + "\'");
      }
      s++;
    }
  }

  return ret;
}


/**
 * Parse the given comma-separated list of integers into a
 * vector<DayAndPosition>.  Used as a support function for
 * setByDay(const std::string&).
 *
 * @param  s    The string representation of the day list.
 * @returns The list of day-and-positions
 * @throws invalid_argument If the given string does not describe a valid
 *         list of day-and-positions
 */
vector<DayAndPosition> Recurrence::parseDaySet(const string& str)
{
  string::size_type i;
  vector<DayAndPosition> ret;

  for (i = 0; i < str.size(); i++) {
    string::size_type comma = str.find(',', i);
    string thisdaypos(str, i,
                      (comma == string::npos ? string::npos : comma - i));
    DayAndPosition dap(thisdaypos);
    ret.push_back(dap);
    i = (comma == string::npos ? str.size() - 1 : comma);
  }

  return ret;
}


/**
 * Generate a standard Date-Time string for the given calendar.
 *
 * @param cal The calendar to generate.
 * @returns A string representation of the calendar.
 */
string Recurrence::generateDateTimeString(const Calendar& cal)
{
  ostrstream obuf;

  obuf.fill('0');

  obuf.width(4);
  obuf << cal.get(Calendar::YEAR);

  obuf.width(2);
  // Count from 1, not from Calendar::JANUARY
  obuf << cal.get(Calendar::MONTH) - Calendar::JANUARY + 1;

  obuf.width(2);
  obuf << cal.get(Calendar::DAY_OF_MONTH);

  obuf << "T";

  obuf.width(2);
  obuf << cal.get(Calendar::HOUR_OF_DAY);
  
  obuf.width(2);
  obuf << cal.get(Calendar::MINUTE);
  
  obuf.width(2);
  obuf << cal.get(Calendar::SECOND);

  string ret(obuf.str());
  obuf.rdbuf()->freeze(0);
  return ret;
}


/**
 * Generate a comma-separated int-set string (as used in BY* arguments).
 *
 * @param A vector of ints
 * @returns A comma-separated list of ints
 */
string Recurrence::generateIntSetString(const vector<int>& set) {
  vector<int>::size_type i;

  ostrstream obuf;

  for (i = 0; i < set.size(); i++) {
    if (i != 0) {
      obuf << ",";
    }
    obuf << set[i];
  }

  string ret(obuf.str());
  obuf.rdbuf()->freeze(0);
  return ret;
}


/**
 * Generate a comma-separated DayAndPosition set string (as used in
 * BYDAY arguments).
 *
 * @param A vector of DayAndPositions
 * @returns A comma-separated list of DayAndPositions
 */
string Recurrence::generateDaySetString(const vector<DayAndPosition>& set) {
  vector<DayAndPosition>::size_type i;

  ostrstream obuf;

  for (i = 0; i < set.size(); i++) {
    if (i != 0) {
      obuf << ",";
    }
    obuf << set[i].getString();
  }

  string ret(obuf.str());
  obuf.rdbuf()->freeze(0);
  return ret;
}


/**
 * Return a string representation of this recurrence. This method 
 * is intended to be used only for debugging purposes.
 * The returned string may be empty but may not be <code>NULL</code>.
 * 
 * @return  A string representation of this recurrence.
 */
string Recurrence::toString() const {
  unsigned int i;
  
  ostrstream buffer;
  
  buffer << "Recurrence";
  buffer << "[dtStart=";
  buffer << dtStart != NULL ? dtStart->toString() : "NULL";
  buffer << ",duration=";
  buffer << duration != NULL ? duration->toString() : "NULL";
  buffer << ",frequency=";
  buffer << frequency;
  buffer << ",interval=";
  buffer << interval;
  buffer << ",until=";
  buffer << until != NULL ? until->toString() : "NULL";
  buffer << ",count=";
  buffer << count;
  buffer << ",bySecond=";
  buffer << stringizeIntArray(bySecond);
  buffer << ",byMinute=";
  buffer << stringizeIntArray(byMinute);
  buffer << ",byHour=";
  buffer << stringizeIntArray(byHour);
  buffer << ",byDay=";
  buffer << "[";
  for (i = 0; i < byDay.size(); i++) {
    if (i != 0) buffer << ",";
    buffer << byDay[i].toString();
  }
  buffer << "]";
  buffer << ",byMonthDay=";
  buffer << stringizeIntArray(byMonthDay);
  buffer << ",byYearDay=";
  buffer << stringizeIntArray(byYearDay);
  buffer << ",byWeekNo=";
  buffer << stringizeIntArray(byWeekNo);
  buffer << ",byMonth=";
  buffer << stringizeIntArray(byMonth);
  buffer << ']';
  
  buffer << ends;
  
  string ret(buffer.str());
  buffer.rdbuf()->freeze(0);
  
  return ret;
}

string Recurrence::stringizeIntArray(const vector<int>& a) {
  unsigned int i;

  ostrstream buffer;
  
  buffer << "[";
  for (i = 0; i < a.size(); i++) {
    if (i != 0) buffer << ",";
    buffer << a[i];
  }
  buffer << "]";

  string ret(buffer.str());
  buffer.rdbuf()->freeze(0);
  
  return ret;
}
