/* recur.h - Represent an iCalendar (subset) recurrence rule. */
/* Copyright 2001 Columbia University.  All rights reserved. */

/* This is a manual translation of the Java Cal-code package. */

#ifndef RECUR_H_
#define RECUR_H_

#include <time.h>
#include <string>
#include <vector>
#include <iostream>

#include "duration.h"
#include "calendar.h"
#include "dayandposition.h"

class Recurrence {
public:
  typedef enum {
    SECONDLY, MINUTELY, HOURLY, DAILY, WEEKLY, MONTHLY, YEARLY,
    NO_RECURRENCE
  } Frequency; 
  
  static time_t max_count_time;
  static int max_count_loops;
  static bool strict_count_bounds;
  
protected:
  Calendar* dtStart;
  Duration* duration;
  Recurrence::Frequency frequency;
  int interval;
  Calendar* until;
  int count;

  std::vector<int> bySecond;
  std::vector<int> byMinute;
  std::vector<int> byHour;
  std::vector<DayAndPosition> byDay;
  std::vector<int> byMonthDay;
  std::vector<int> byYearDay;
  std::vector<int> byWeekNo;
  std::vector<int> byMonth;

  void init();
  
public:
  Recurrence();
  Recurrence(const Calendar& start, const Duration& dur);
  Recurrence(const Calendar& start, const Duration& dur,
             Recurrence::Frequency freq);

  // Copy constructor
  Recurrence(const Recurrence& other);

  // Destructor
  ~Recurrence();

  // Assignment operator
  Recurrence& operator=(const Recurrence& r);

  Calendar getDtStart() const;
  std::string getDtStartString() const;
  void setDtStart(const Calendar& start);
  void setDtStart(const std::string& str);
  
  Duration getDuration() const;
  std::string getDurationString() const;
  void setDuration(const Duration& d);
  void setDuration(const std::string& str);
  
  Calendar getDtEnd() const;
  std::string getDtEndString() const;
  void setDtEnd(const Calendar& end);
  void setDtEnd(const std::string& str);
  
  Recurrence::Frequency getFrequency() const;
  std::string getFrequencyString() const;
  void setFrequency(Recurrence::Frequency freq);
  void setFrequency(const std::string& str);
  
  int getInterval() const;
  std::string getIntervalString() const;
  void setInterval(int intr);
  void setInterval(const std::string& str);
  
  Calendar* getUntil() const;
  std::string getUntilString() const;
  void setUntil(const Calendar* u);
  void setUntil(const std::string& str);
  
  int getCount() const;
  std::string getCountString() const;
  void setCount(int c);
  void setCount(const std::string& str);
  
  Calendar::Weekday getWeekStart() const;
  std::string getWeekStartString() const;
  void setWeekStart(Calendar::Weekday weekstart);
  void setWeekStart(const std::string& str);
  
  std::vector<int> getBySecond() const;
  std::string getBySecondString() const;
  void setBySecond(const std::vector<int>& b);
  void setBySecond(const std::string& str);
  
  std::vector<int> getByMinute() const;
  std::string getByMinuteString() const;
  void setByMinute(const std::vector<int>& b);
  void setByMinute(const std::string& str);
  
  std::vector<int> getByHour() const;
  std::string getByHourString() const;
  void setByHour(const std::vector<int>& b);
  void setByHour(const std::string& str);
  
  std::vector<DayAndPosition> getByDay() const;
  std::string getByDayString() const;
  void setByDay(const std::vector<DayAndPosition>& b);
  void setByDay(const std::string& str);
  
  std::vector<int> getByMonthDay() const;
  std::string getByMonthDayString() const;
  void setByMonthDay(const std::vector<int>& b);
  void setByMonthDay(const std::string& str);
  
  std::vector<int> getByYearDay() const;
  std::string getByYearDayString() const;
  void setByYearDay(const std::vector<int>& b);
  void setByYearDay(const std::string& str);
  
  std::vector<int> getByWeekNo() const;
  std::string getByWeekNoString() const;
  void setByWeekNo(const std::vector<int>& b);
  void setByWeekNo(const std::string& str);
  
  std::vector<int> getByMonth() const;
  std::string getByMonthString() const;
  void setByMonth(const std::vector<int>& b);
  void setByMonth(const std::string& str);
  
  bool isInRecurrence(const Calendar& current) const;
  bool isInRecurrence(const Calendar& current, bool debug) const;

protected:
  bool candidateIsInRecurrence(const Calendar& candidate, bool debug) const;
  Recurrence::Frequency getMinimumInterval() const;
  
public:
  Calendar* getCandidateStartTime(const Calendar& current, bool is_forward,
                                  bool debug) const;
  
protected:
  Calendar* internalGetCandidateStartTime(const Calendar& current,
                                          bool is_forward, bool debug) const;
  static void adjust_constant_length_field(Calendar::Field field,
                                           const Calendar* start,
                                           Calendar* candidate,
                                           bool is_forward);
  static void adjust_day_of_month(const Calendar* start, Calendar* candidate,
                                  bool is_forward);
  static void adjust_day_of_year(const Calendar* start, Calendar* candidate,
                                 bool is_forward);
  
public:
  void computeUntilFromCount() const;
  void computeUntilFromCount(bool debug);
  
protected:
  long getRecurrenceCount(const Calendar& candidate) const;
  static long getFieldNumber(const Calendar& cal, Recurrence::Frequency field);
  static long getWeekNumber(const Calendar& cal);
  static long getMonthNumber(const Calendar& cal);
  
  bool matchesByDay(const Calendar& candidate) const;
  bool matchesIndividualByDay(const Calendar& candiate,
                              const DayAndPosition& pos)
    const;
  bool matchesByField(const std::vector<int>& array, Calendar::Field field,
                      const Calendar& candidate,
                      bool allowNegative) const;
  
  bool matchesBySecond(const Calendar& candidate) const;
  bool matchesByMinute(const Calendar& candidate) const;
  bool matchesByHour(const Calendar& candidate) const;
  bool matchesByMonthDay(const Calendar& candidate) const;
  bool matchesByYearDay(const Calendar& candidate) const;
  bool matchesByWeekNo(const Calendar& candidate) const;
  bool matchesByMonth(const Calendar& candidate) const;
  
  static void setCalendarWeekStart(Calendar& cal,
                                   Calendar::Weekday start);

public:
  static Calendar parseDateTime(const std::string&);
  static std::string generateDateTimeString(const Calendar&);

protected:
  static long getFixedSizeInt(const std::string&, size_t, size_t,
                              const char *);
  static std::vector<int> parseIntSet(const std::string&, const char *,
                                 int, int, bool);
  static std::vector<DayAndPosition> parseDaySet(const std::string&);

  static std::string generateIntSetString(const std::vector<int>&);
  static std::string generateDaySetString(const std::vector<DayAndPosition>&);
  
public:
  std::string toString() const;
  
protected:
  static std::string stringizeIntArray(const std::vector<int>& a);
};
  
  

#endif /* RECUR_H_ */
