/*
 * Copyright (c) 2000, 2001, Columbia University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution. 
 *
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.  
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// Recursion tests.  From the examples in RFC 2445 (only those
// representable in our subset), plus a few others.

#include "recur.h"
#include "tests.h"
#include "tzfile.h"
#include <limits.h>

using namespace std;

#ifdef USE_OLD_ZONENAMES
#define TIMEZONE "US/Eastern"
#else
#define TIMEZONE "America/New_York"
#endif


vector<RecurHolder> tests;

static void addTest(const string& name,
                    int runCount,
                    const string& start,
                    const string& frequency);

static void addTest(const string& name,
                    int runCount,
                    const string& start,
                    const string& frequency,
                    const string& until,
                    const string& count,
                    const string& interval);

static void addTest(const string& name,
                    int runCount,
                    const string& start,
                    const string& frequency,
                    const string& until,
                    const string& count,
                    const string& interval,
                    const string& byDay,
                    const string& byMonthDay,
                    const string& byYearDay,
                    const string& byWeekNo,
                    const string& byMonth,
                    const string& wkst);

static void addTest(const string& name,
                    int runCount,
                    const string& start,
                    const string& duration,
                    const string& frequency,
                    const string& until,
                    const string& count,
                    const string& interval,
                    const string& bySecond,
                    const string& byMinute,
                    const string& byHour,
                    const string& byDay,
                    const string& byMonthDay,
                    const string& byYearDay,
                    const string& byWeekNo,
                    const string& byMonth,
                    const string& wkst,
                    Calendar::Field period,
                    int step,
                    time_t end);

static void addTest(const string& name,
                    int runCount,
                    const Recurrence& recurrence,
                    const Calendar& start);

static void addTest(const string& name,
                    int runCount,
                    const Recurrence& recurrence,
                    const Calendar& start,
                    Calendar::Field period,
                    int step,
                    long end);


static void addTest(const string& name, int runCount, const string& start,
                    const string& frequency) {
  Recurrence recur;
  recur.setDtStart(start);
  recur.setDuration("P1D");
  recur.setFrequency(frequency);
  
  addTest(name, runCount, recur, recur.getDtStart());
}

static void addTest(const string& name, int runCount, const string& start,
                    const string& frequency, const string& until,
                    const string& count, const string& interval) {
  Recurrence recur;
  recur.setDtStart(start);
  recur.setDuration("P1D");
  recur.setFrequency(frequency);
  
  recur.setUntil(until);
  recur.setCount(count);
  recur.setInterval(interval);
  
  recur.computeUntilFromCount();
  
  addTest(name, runCount, recur, recur.getDtStart());
}

static void addTest(const string& name,
                    int runCount,
                    const string& start,
                    const string& frequency,
                    const string& until,
                    const string& count,
                    const string& interval,
                    const string& byDay,
                    const string& byMonthDay,
                    const string& byYearDay,
                    const string& byWeekNo,
                    const string& byMonth,
                    const string& wkst) {
  Recurrence recur;
  recur.setDtStart(start);
  recur.setDuration("P1D");
  recur.setFrequency(frequency);
  
  recur.setUntil(until);
  recur.setCount(count);
  recur.setInterval(interval);
  recur.setByDay(byDay);
  recur.setByMonthDay(byMonthDay);
  recur.setByYearDay(byYearDay);
  recur.setByWeekNo(byWeekNo);
  recur.setByMonth(byMonth);
  recur.setWeekStart(wkst);
  
  recur.computeUntilFromCount();
  
  addTest(name, runCount, recur, recur.getDtStart());
}


static void addTest(const string& name,
                    int runCount,
                    const string& start,
                    const string& duration,
                    const string& frequency,
                    const string& until,
                    const string& count,
                    const string& interval,
                    const string& bySecond,
                    const string& byMinute,
                    const string& byHour,
                    const string& byDay,
                    const string& byMonthDay,
                    const string& byYearDay,
                    const string& byWeekNo,
                    const string& byMonth,
                    const string& wkst,
                    Calendar::Field period,
                    int step,
                    time_t end) {
  Recurrence recur;
  recur.setDtStart(start);
  recur.setDuration(duration);
  recur.setFrequency(frequency);
  
  recur.setUntil(until);
  recur.setCount(count);
  recur.setInterval(interval);
  
  recur.setBySecond(bySecond);
  recur.setByMinute(byMinute);
  recur.setByHour(byHour);
  
  recur.setByDay(byDay);
  recur.setByMonthDay(byMonthDay);
  recur.setByYearDay(byYearDay);
  recur.setByWeekNo(byWeekNo);
  recur.setByMonth(byMonth);
  recur.setWeekStart(wkst);
  
  recur.computeUntilFromCount();
  
  addTest(name, runCount, recur, recur.getDtStart(), period, step, end);
}

static void addTest(const string& name, int runCount,
                    const Recurrence& recurrence,
                    const Calendar& start) {
  addTest(name, runCount, recurrence, start, Calendar::DATE, 1, 1500000000);
}

static void addTest(const string& name, int runCount,
                    const Recurrence& recurrence, 
                    const Calendar& start, Calendar::Field period, int step, long end) {
  Calendar myStart = start;
  myStart.add(Calendar::DATE, -1);

  RecurHolder
    newHolder(recurrence, myStart, runCount, name, period, step, end);
  
  tests.push_back(newHolder);
}

vector<RecurHolder>& getTests()
{
  if (tests.size() != 0) return tests;

  putenv("TZ=" TIMEZONE);

  addTest("My birthday (May 4, every year)",
          29,
          "19730504T111100",
          "YEARLY");

  addTest("Daily for 10 occurences",
          15,
          "19970902T090000",
          "DAILY",
          "", /* until */
          "10", /* count */
          ""); /* interval */

  addTest("Daily until December 24, 1997",
          200,
          "19970902T090000",
          "DAILY",
          "19971224T000000Z",
          "", /* count */
          ""); /* interval */

  addTest("Every other day - forever",
          100,
          "19970902T090000",
          "DAILY",
          "", /* until */
          "", /* count */
          "2"); /* interval */

  addTest("Every 10 days, 5 occurrences",
          10,
          "19970902T090000",
          "DAILY",
          "", /* until */
          "5", /* count */
          "10"); /* interval */

  addTest("Everyday in January, for 3 years (1)",
          100, /* runCount */
          "19980101T090000",
          "YEARLY",
          "20000131T090000Z",
          "", /* count */
          "", /* interval */
          "SU,MO,TU,WE,TH,FR,SA", /* byDay */
          "", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "1", /* byMonth */
          ""); /* wkst */

  addTest("Everyday in January, for 3 years (2)",
          100, /* runCount */
          "19980101T090000",
          "DAILY",
          "20000131T090000Z",
          "", /* count */
          "", /* interval */
          "", /* byDay */
          "", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "1", /* byMonth */
          ""); /* wkst */

  addTest("Weekly for 10 occurrences",
          15,
          "19970902T090000",
          "WEEKLY",
          "", /* until */
          "10", /* count */
          ""); /* interval */

  addTest("Weekly until December 24, 1997",
          20,
          "19970902T090000",
          "WEEKLY",
          "19971224T000000Z",
          "",  /* count */
          ""); /* interval */
  
  addTest("Every other week - forever",
          20,
          "19970902T090000",
          "WEEKLY",
          "", /* until */
          "", /* count */
          "2"); /* interval */
  
  addTest("Weekly on Tuesday and Thursday for 5 weeks (1)",
          15,
          "19970902T090000",
          "WEEKLY",
          "19971007T000000Z",
          "", /* count */
          "", /* interval */
          "TU,TH", /* byDay */
          "", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "", /* byMonth */
          "SU"); /* wkst */

  addTest("Weekly on Tuesday and Thursday for 5 weeks (2)",
          15,
          "19970902T090000",
          "WEEKLY",
          "", /* until */
          "10", /* count */
          "", /* interval */
          "TU,TH", /* byDay */
          "", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "", /* byMonth */
          "SU"); /* wkst */

  addTest("Every other week on Monday, Wednesday and Friday until "
          "December 24, 1997, but starting on Tuesday, September 2, 1997",
          40,
          "19970902T090000",
          "WEEKLY",
          "19971224T000000Z",
          "", /* count */
          "2", /* interval */
          "MO,WE,FR", /* byDay */
          "", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "", /* byMonth */
          "SU"); /* wkst */

  addTest("Every other week on Tuesday and Thursday, for 8 occurrences",
          15,
          "19970902T090000",
          "WEEKLY",
          "", /* until */
          "8", /* count */
          "2", /* interval */
          "TU,TH", /* byDay */
          "", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "", /* byMonth */
          "SU"); /* wkst */

  addTest("Monthly on the 1st Friday for ten occurrences",
          15,
          "19970905T090000",
          "MONTHLY",
          "", /* until */
          "10", /* count */
          "", /* interval */
          "1FR", /* byDay */
          "", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "", /* byMonth */
          ""); /* wkst */

  addTest("Monthly on the 1st Friday until December 24, 1997",
          15,
          "19970905T090000",
          "MONTHLY",
          "19971224T000000Z", /* until */
          "", /* count */
          "", /* interval */
          "1FR", /* byDay */
          "", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "", /* byMonth */
          ""); /* wkst */

  addTest("Every other month on the 1st and last Sunday of the month "
          "for 10 occurrences",
          15,
          "19970907T090000",
          "MONTHLY",
          "", /* until */
          "10", /* count */
          "2", /* interval */
          "1SU,-1SU", /* byDay */
          "", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "", /* byMonth */
          ""); /* wkst */

  addTest("Monthly on the second to last Monday of the month "
          "for 6 months",
          10,
          "19970922T090000",
          "MONTHLY",
          "", /* until */
          "6", /* count */
          "", /* interval */
          "-2MO", /* byDay */
          "", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "", /* byMonth */
          ""); /* wkst */

  addTest("Monthly on the third to the last day of the month, forever",
          30,
          "19970928T090000",
          "MONTHLY",
          "", /* until */
          "", /* count */
          "", /* interval */
          "", /* byDay */
          "-3", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "", /* byMonth */
          ""); /* wkst */

  addTest("Monthly on the 2nd and 15th of the month for 10 occurrences",
          15,
          "19970902T090000",
          "MONTHLY",
          "", /* until */
          "10", /* count */
          "", /* interval */
          "", /* byDay */
          "2,15", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "", /* byMonth */
          ""); /* wkst */


  addTest("Monthly on the first and last day of the month "
          "for 10 occurrences",
          15,
          "19970930T090000",
          "MONTHLY",
          "", /* until */
          "10", /* count */
          "", /* interval */
          "", /* byDay */
          "1,-1", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "", /* byMonth */
          ""); /* wkst */

  addTest("Every 18 months on the 10th thru 15th of the month "
          "for 10 occurrences",
          15,
          "19970910T090000",
          "MONTHLY",
          "", /* until */
          "10", /* count */
          "18", /* interval */
          "", /* byDay */
          "10,11,12,13,14,15", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "", /* byMonth */
          ""); /* wkst */


  addTest("Every Tuesday, every other month",
          30,
          "19970902T090000",
          "MONTHLY",
          "", /* until */
          "", /* count */
          "2", /* interval */
          "TU", /* byDay */
          "", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "", /* byMonth */
          ""); /* wkst */

  addTest("Yearly in June and July for 10 occurrences",
          20,
          "19970610T090000",
          "YEARLY",
          "", /* until */
          "10", /* count */
          "", /* interval */
          "", /* byDay */
          "", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "6,7", /* byMonth */
          ""); /* wkst */

  addTest("Every other year in January, February, and March "
          "for 10 occurrences",
          15,
          "19970310T090000",
          "YEARLY",
          "", /* until */
          "10", /* count */
          "2", /* interval */
          "", /* byDay */
          "", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "1,2,3", /* byMonth */
          ""); /* wkst */

  addTest("Every 3rd year on the 1st, 100th and 200th day "
          "for 10 occurrences",
          15,
          "19970101T090000",
          "YEARLY",
          "", /* until */
          "10", /* count */
          "3", /* interval */
          "", /* byDay */
          "", /* byMonthDay */
          "1,100,200", /* byYearDay */
          "", /* byWeekNo */
          "", /* byMonth */
          ""); /* wkst */

  addTest("Every 20th Monday of the year, forever",
          20,
          "19970519T090000",
          "YEARLY",
          "", /* until */
          "", /* count */
          "", /* interval */
          "20MO", /* byDay */
          "", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "", /* byMonth */
          ""); /* wkst */

  addTest("Monday of week number 20 (where the default start of the "
          "week is Monday), forever",
          20,
          "19970512T090000",
          "YEARLY",
          "", /* until */
          "", /* count */
          "", /* interval */
          "MO", /* byDay */
          "", /* byMonthDay */
          "", /* byYearDay */
          "20", /* byWeekNo */
          "", /* byMonth */
          ""); /* wkst */

  addTest("Every Thursday in March, forever",
          20,
          "19970313T090000",
          "YEARLY",
          "", /* until */
          "", /* count */
          "", /* interval */
          "TH", /* byDay */
          "", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "3", /* byMonth */
          ""); /* wkst */

  addTest("Every Thursday, but only during June, July, and August, forever",
          40,
          "19970605T090000",
          "YEARLY",
          "", /* until */
          "", /* count */
          "", /* interval */
          "TH", /* byDay */
          "", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "6,7,8", /* byMonth */
          ""); /* wkst */

  // XXX we don't support EXDATE.  A real time-switch would handle
  // this with an earlier time-switch representing the exception.

  addTest("Every Friday the 13th, forever",
          20,
          "19970902T090000",
          "MONTHLY",
          "", /* until */
          "", /* count */
          "", /* interval */
          "FR", /* byDay */
          "13", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "", /* byMonth */
          ""); /* wkst */


  addTest("The first Saturday that follows the first Sunday of the month, "
          "forever",
          20,
          "19970913T090000",
          "MONTHLY",
          "", /* until */
          "", /* count */
          "", /* interval */
          "SA", /* byDay */
          "7,8,9,10,11,12,13", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "", /* byMonth */
          ""); /* wkst */

  addTest("Every four years, the first Tuesday after a Monday in November, "
          "forever (U.S. Presidential Election day)",
          20,
          "19961105T090000",
          "P1D",
          "YEARLY",
          "", /* until */
          "", /* count */
          "4", /* interval */
          "", /* bySecond */
          "", /* byMinute */
          "", /* byHour */
          "TU", /* byDay */
          "2,3,4,5,6,7,8", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "11", /* byMonth */
          "", /* wkst */
          Calendar::DATE, /* period */
          1,  /* step */
          0x7ff00000); /* end */

  /* Excluded because of BYSETPOS:
     "The third instance into the month of one of Tuesday, Wednesday, "
     "or Thursday, for the next three months"
  */

  /* Excluded because of BYSETPOS:
     "The 2nd to last weekday of the month"
  */

  addTest("Every 3 hours from 9:00 AM to 5:00 PM on a specific day",
          10,
          "19970902T090000",
          "PT1H",
          "HOURLY",
          "19970902T170000Z", /* until */
          "", /* count */
          "3", /* interval */
          "", /* bySecond */
          "", /* byMinute */
          "", /* byHour */
          "", /* byDay */
          "", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "", /* byMonth */
          "", /* wkst */
          Calendar::HOUR_OF_DAY, /* period */
          1,  /* step */
          1000000000); /* end */

  addTest("Every 15 minutes for 6 occurrences",
          15,
          "19970902T090000",
          "PT1M",
          "MINUTELY",
          "", /* until */
          "6", /* count */
          "15", /* interval */
          "", /* bySecond */
          "", /* byMinute */
          "", /* byHour */
          "", /* byDay */
          "", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "", /* byMonth */
          "", /* wkst */
          Calendar::MINUTE, /* period */
          1,  /* step */
          880000000); /* end.  Trim low (only a few months after dtStart)
                         because we're doing minutely checks. */

  addTest("Every hour and a half for 4 occurrences",
          10,
          "19970902T090000",
          "PT1M",
          "MINUTELY",
          "", /* until */
          "4", /* count */
          "90", /* interval */
          "", /* bySecond */
          "", /* byMinute */
          "", /* byHour */
          "", /* byDay */
          "", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "", /* byMonth */
          "", /* wkst */
          Calendar::MINUTE, /* period */
          1,  /* step */
          880000000); /* end.  Trim low (only a few months after dtStart)
                         because we're doing minutely checks. */

  addTest("Every 20 minutes from 9:00 AM to 4:40 PM every day (1)",
          50,
          "19970902T090000",
          "PT1M",
          "DAILY",
          "", /* until */
          "", /* count */
          "", /* interval */
          "", /* bySecond */
          "0,20,40", /* byMinute */
          "9,10,11,12,13,14,15,16", /* byHour */
          "", /* byDay */
          "", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "", /* byMonth */
          "", /* wkst */
          Calendar::MINUTE, /* period */
          1,  /* step */
          880000000); /* end.  Trim low (only a few months after dtStart)
                         because we're doing minutely checks. */

  addTest("Every 20 minutes from 9:00 AM to 4:40 PM every day (2)",
          50,
          "19970902T090000",
          "PT1M",
          "MINUTELY",
          "", /* until */
          "", /* count */
          "20", /* interval */
          "", /* bySecond */
          "", /* byMinute */
          "9,10,11,12,13,14,15,16", /* byHour */
          "", /* byDay */
          "", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "", /* byMonth */
          "", /* wkst */
          Calendar::MINUTE, /* period */
          1,  /* step */
          880000000); /* end.  Trim low (only a few months after dtStart)
                         because we're doing minutely checks. */

  addTest("An example where the days generated makes a difference "
          "because of WKST",
          8,
          "19970805T090000",
          "WEEKLY",
          "", /* until */
          "4", /* count */
          "2", /* interval */
          "TU,SU", /* byDay */
          "", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "", /* byMonth */
          "MO"); /* wkst */

  addTest("changing only WKST from MO to SU, yields different results...",
          8,
          "19970805T090000",
          "WEEKLY",
          "", /* until */
          "4", /* count */
          "2", /* interval */
          "TU,SU", /* byDay */
          "", /* byMonthDay */
          "", /* byYearDay */
          "", /* byWeekNo */
          "", /* byMonth */
          "SU"); /* wkst */

  /* XXX put the old time zone back */
  return tests;
}
