/*
 * Copyright (c) 2000, 2001, Columbia University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution. 
 *
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.  
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// Recursion tests.  From the examples in RFC 2445 (only those
// representable in our subset), plus a few others.

#include "recur.h"
#include "tests.h"
#include "tzfile.h"
#include <limits.h>

using namespace std;

#ifdef USE_OLD_ZONENAMES
#define TIMEZONE "US/Eastern"
#else
#define TIMEZONE "America/New_York"
#endif


#define VECTORIZE(type, name, definition...) \
  type name ## _a [] = definition; \
  vector<type> name(name ## _a, name ## _a + \
                    sizeof(name ## _a) / sizeof(*name ## _a));

vector<RecurHolder> tests;

static void addTest(string name,
                    int runCount,
                    const Calendar& start,
                    Recurrence::Frequency frequency);

static void addTest(string name,
                    int runCount,
                    const Calendar& start,
                    Recurrence::Frequency frequency,
                    const Calendar* until,
                    int count,
                    int interval);

static void addTest(string name,
                    int runCount,
                    const Calendar& start,
                    Recurrence::Frequency frequency,
                    const Calendar* until,
                    int count, int interval,
                    const vector<DayAndPosition>& byDay,
                    const vector<int>& byMonthDay,
                    const vector<int>& byYearDay,
                    const vector<int>& byWeekNo,
                    const vector<int>& byMonth,
                    Calendar::Weekday wkst);

static void addTest(string name,
                    int runCount,
                    const Calendar& start,
                    const Duration& duration,
                    Recurrence::Frequency frequency,
                    const Calendar* until,
                    int count,
                    int interval,
                    const vector<int>& bySecond,
                    const vector<int>& byMinute,
                    const vector<int>& byHour,
                    const vector<DayAndPosition>& byDay,
                    const vector<int>& byMonthDay,
                    const vector<int>& byYearDay,
                    const vector<int>& byWeekNo,
                    const vector<int>& byMonth,
                    Calendar::Weekday wkst,
                    Calendar::Field period,
                    int step,
                    time_t end);

static void addTest(string name,
                    int runCount,
                    const Recurrence& recurrence,
                    const Calendar& start);

static void addTest(string name,
                    int runCount,
                    const Recurrence& recurrence,
                    const Calendar& start,
                    Calendar::Field period,
                    int step,
                    long end);


static void addTest(string name, int runCount, const Calendar& start,
                    Recurrence::Frequency frequency) {
  Recurrence recur(start,
                   Duration(1, 0, 0, 0),
                   frequency);
  
  addTest(name, runCount, recur, start);
}

static void addTest(string name, int runCount, const Calendar& start,
                    Recurrence::Frequency frequency, const Calendar* until,
                    int count, int interval) {
  Recurrence recur(start,
                   Duration(1, 0, 0, 0),
                   frequency);
  recur.setUntil(until);
  recur.setCount(count);
  recur.setInterval(interval);
  
  recur.computeUntilFromCount();
  
  addTest(name, runCount, recur, start);
}

static void addTest(string name, int runCount, const Calendar& start,
                    Recurrence::Frequency frequency, const Calendar* until, int count, int interval,
                    const vector<DayAndPosition>& byDay, const vector<int>& byMonthDay,
                    const vector<int>& byYearDay, const vector<int>& byWeekNo,
                    const vector<int>& byMonth, Calendar::Weekday wkst) {
  Recurrence recur(start,
                   Duration(1, 0, 0, 0),
                   frequency);
  recur.setUntil(until);
  recur.setCount(count);
  recur.setInterval(interval);
  recur.setByDay(byDay);
  recur.setByMonthDay(byMonthDay);
  recur.setByYearDay(byYearDay);
  recur.setByWeekNo(byWeekNo);
  recur.setByMonth(byMonth);
  if (wkst != Calendar::NO_WEEKDAY) { recur.setWeekStart(wkst); }
  
  recur.computeUntilFromCount();
  
  addTest(name, runCount, recur, start);
}


static void addTest(string name, int runCount, const Calendar& start,
                    const Duration& duration,
                    Recurrence::Frequency frequency, const Calendar* until, int count,
                    int interval,
                    const vector<int>& bySecond, const vector<int>& byMinute, const vector<int>& byHour,
                    const vector<DayAndPosition>& byDay, const vector<int>& byMonthDay,
                    const vector<int>& byYearDay, const vector<int>& byWeekNo,
                    const vector<int>& byMonth, Calendar::Weekday wkst,
                    Calendar::Field period, int step, long end) {
  Recurrence recur(start, duration, frequency);
  recur.setUntil(until);
  recur.setCount(count);
  recur.setInterval(interval);
  
  recur.setBySecond(bySecond);
  recur.setByMinute(byMinute);
  recur.setByHour(byHour);
  
  recur.setByDay(byDay);
  recur.setByMonthDay(byMonthDay);
  recur.setByYearDay(byYearDay);
  recur.setByWeekNo(byWeekNo);
  recur.setByMonth(byMonth);
  if (wkst != Calendar::NO_WEEKDAY) { recur.setWeekStart(wkst); }
  
  recur.computeUntilFromCount();
  
  addTest(name, runCount, recur, start, period, step, end);
}


static void addTest(string name, int runCount, const Recurrence& recurrence,
                    const Calendar& start) {
  addTest(name, runCount, recurrence, start, Calendar::DATE, 1, 1500000000);
}

static void addTest(string name, int runCount, const Recurrence& recurrence,
                    const Calendar& start, Calendar::Field period, int step, long end) {
  Calendar myStart = start;
  myStart.add(Calendar::DATE, -1);

  RecurHolder
    newHolder(recurrence, myStart, runCount, name, period, step, end);
  
  tests.push_back(newHolder);
}

vector<RecurHolder>& getTests()
{
  if (tests.size() != 0) return tests;

  char *oldTZ = getenv("TZ");
  putenv("TZ=" TIMEZONE);

  addTest("My birthday (May 4, every year)",
          29,
          Calendar(1973, Calendar::MAY, 4, 11, 11),
          Recurrence::YEARLY);

  addTest("Daily for 10 occurences",
          15,
          Calendar(1997, Calendar::SEPTEMBER, 2, 9, 0),
          Recurrence::DAILY,
          NULL, /* until */
          10, /* count */
          1); /* interval */

  addTest("Daily until December 24, 1997",
          200,
          Calendar(1997, Calendar::SEPTEMBER, 2, 9, 0),
          Recurrence::DAILY,
          &Calendar(1997, Calendar::DECEMBER, 24),
          0,
          1); /* interval */

  addTest("Every other day - forever",
          100,
          Calendar(1997, Calendar::SEPTEMBER, 2, 9, 0),
          Recurrence::DAILY,
          NULL, /* until */
          0, /* count */
          2);

  addTest("Every 10 days, 5 occurrences",
          10,
          Calendar(1997, Calendar::SEPTEMBER, 2, 9, 0),
          Recurrence::DAILY,
          NULL, /* until */
          5, /* count */
          10); /* interval */

  VECTORIZE(DayAndPosition,testbyday1_allWeek, {
      DayAndPosition(Calendar::SUNDAY, 0),
          DayAndPosition(Calendar::MONDAY, 0),
          DayAndPosition(Calendar::TUESDAY, 0),
          DayAndPosition(Calendar::WEDNESDAY, 0),
          DayAndPosition(Calendar::THURSDAY, 0),
          DayAndPosition(Calendar::FRIDAY, 0),
          DayAndPosition(Calendar::SATURDAY, 0)});
      
  VECTORIZE(int, testbymonth1_January, { Calendar::JANUARY });

  addTest("Everyday in January, for 3 years (1)",
          100, /* runCount */
          Calendar(1998, Calendar::JANUARY, 1, 9, 0),
          Recurrence::YEARLY,
          &Calendar(2000, Calendar::JANUARY, 31, 9, 0),
          0, /* count */
          1, /* interval */
          testbyday1_allWeek, /* byDay */
          vector<int>(), /* byMonthDay */
          vector<int>(), /* byYearDay */
          vector<int>(), /* byWeekNo */
          testbymonth1_January, /* byMonth */
          Calendar::NO_WEEKDAY); /* wkst */

  addTest("Everyday in January, for 3 years (2)",
          100, /* runCount */
          Calendar(1998, Calendar::JANUARY, 1, 9, 0),
          Recurrence::DAILY,
          &Calendar(2000, Calendar::JANUARY, 31, 9, 0),
          0, /* count */
          1, /* interval */
          vector<DayAndPosition>(), /* byDay */
          vector<int>(), /* byMonthDay */
          vector<int>(), /* byYearDay */
          vector<int>(), /* byWeekNo */
          testbymonth1_January, /* byMonth */
          Calendar::NO_WEEKDAY); /* wkst */

  addTest("Weekly for 10 occurrences",
          15,
          Calendar(1997, Calendar::SEPTEMBER, 2, 9, 0),
          Recurrence::WEEKLY,
          NULL, /* until */
          10, /* count */
          1); /* interval */

  addTest("Weekly until December 24, 1997",
          20,
          Calendar(1997, Calendar::SEPTEMBER, 2, 9, 0),
          Recurrence::WEEKLY,
          &Calendar(1997, Calendar::DECEMBER, 24), /* until */
          0,  /* count */
          1); /* interval */
  
  addTest("Every other week - forever",
          20,
          Calendar(1997, Calendar::SEPTEMBER, 2, 9, 0),
          Recurrence::WEEKLY,
          NULL, /* until */
          0,  /* count */
          2); /* interval */
  
  VECTORIZE(DayAndPosition, testbyday2_TuTh, {
      DayAndPosition(Calendar::TUESDAY, 0),
      DayAndPosition(Calendar::THURSDAY, 0)});

    addTest("Weekly on Tuesday and Thursday for 5 weeks (1)",
            15,
            Calendar(1997, Calendar::SEPTEMBER, 2, 9, 0),
            Recurrence::WEEKLY,
            &Calendar(1997, Calendar::OCTOBER, 7),
            0, /* count */
            1, /* interval */
            testbyday2_TuTh, /* byDay */
            vector<int>(), /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::SUNDAY); /* wkst */

    addTest("Weekly on Tuesday and Thursday for 5 weeks (2)",
            15,
            Calendar(1997, Calendar::SEPTEMBER, 2, 9, 0),
            Recurrence::WEEKLY,
            NULL,
            10, /* count */
            1, /* interval */
            testbyday2_TuTh, /* byDay */
            vector<int>(), /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::SUNDAY); /* wkst */

    VECTORIZE(DayAndPosition, testbyday3_MoWeFr, {
      DayAndPosition(Calendar::MONDAY, 0),
      DayAndPosition(Calendar::WEDNESDAY, 0),
      DayAndPosition(Calendar::FRIDAY, 0)});

    addTest("Every other week on Monday, Wednesday and Friday until "
            "December 24, 1997, but starting on Tuesday, September 2, 1997",
            40,
            Calendar(1997, Calendar::SEPTEMBER, 2, 9, 0),
            Recurrence::WEEKLY,
            &Calendar(1997, Calendar::DECEMBER, 24),
            0, /* count */
            2, /* interval */
            testbyday3_MoWeFr, /* byDay */
            vector<int>(), /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::SUNDAY); /* wkst */

    addTest("Every other week on Tuesday and Thursday, for 8 occurrences",
            15,
            Calendar(1997, Calendar::SEPTEMBER, 2, 9, 0),
            Recurrence::WEEKLY,
            NULL,
            8, /* count */
            2, /* interval */
            testbyday2_TuTh, /* byDay */
            vector<int>(), /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::SUNDAY); /* wkst */

    VECTORIZE(DayAndPosition, testbyday4_1Fr, {
      DayAndPosition(Calendar::FRIDAY, 1)});

    addTest("Monthly on the 1st Friday for ten occurrences",
            15,
            Calendar(1997, Calendar::SEPTEMBER, 5, 9, 0),
            Recurrence::MONTHLY,
            NULL, /* until */
            10, /* count */
            1, /* interval */
            testbyday4_1Fr, /* byDay */
            vector<int>(), /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::NO_WEEKDAY); /* wkst */

    addTest("Monthly on the 1st Friday until December 24, 1997",
            15,
            Calendar(1997, Calendar::SEPTEMBER, 5, 9, 0),
            Recurrence::MONTHLY,
            &Calendar(1997, Calendar::DECEMBER, 24),
            0, /* count */
            1, /* interval */
            testbyday4_1Fr, /* byDay */
            vector<int>(), /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::NO_WEEKDAY); /* wkst */

    VECTORIZE(DayAndPosition, testbyday5_1Su_m1Su, {
      DayAndPosition(Calendar::SUNDAY, 1),
      DayAndPosition(Calendar::SUNDAY, -1)});

    addTest("Every other month on the 1st and last Sunday of the month "
            "for 10 occurrences",
            15,
            Calendar(1997, Calendar::SEPTEMBER, 7, 9, 0),
            Recurrence::MONTHLY,
            NULL,
            10, /* count */
            2, /* interval */
            testbyday5_1Su_m1Su, /* byDay */
            vector<int>(), /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::NO_WEEKDAY); /* wkst */

    VECTORIZE(DayAndPosition, testbyday6_m2Mo, {
      DayAndPosition(Calendar::MONDAY, -2)});

    addTest("Monthly on the second to last Monday of the month "
            "for 6 months",
            10,
            Calendar(1997, Calendar::SEPTEMBER, 22, 9, 0),
            Recurrence::MONTHLY,
            NULL,
            6, /* count */
            1, /* interval */
            testbyday6_m2Mo, /* byDay */
            vector<int>(), /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::NO_WEEKDAY); /* wkst */

    VECTORIZE(int, testbymonthday1_m3, {-3});

    addTest("Monthly on the third to the last day of the month, forever",
            30,
            Calendar(1997, Calendar::SEPTEMBER, 28, 9, 0),
            Recurrence::MONTHLY,
            NULL,
            0, /* count */
            1, /* interval */
            vector<DayAndPosition>(), /* byDay */
            testbymonthday1_m3, /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::NO_WEEKDAY); /* wkst */

    VECTORIZE(int, testbymonthday2_2_15, {2, 15});

    addTest("Monthly on the 2nd and 15th of the month for 10 occurrences",
            15,
            Calendar(1997, Calendar::SEPTEMBER, 2, 9, 0),
            Recurrence::MONTHLY,
            NULL,
            10, /* count */
            1, /* interval */
            vector<DayAndPosition>(), /* byDay */
            testbymonthday2_2_15, /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::NO_WEEKDAY); /* wkst */


    VECTORIZE(int, testbymonthday3_1_m1, {1, -1});

    addTest("Monthly on the first and last day of the month "
            "for 10 occurrences",
            15,
            Calendar(1997, Calendar::SEPTEMBER, 30, 9, 0),
            Recurrence::MONTHLY,
            NULL,
            10, /* count */
            1, /* interval */
            vector<DayAndPosition>(), /* byDay */
            testbymonthday3_1_m1, /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::NO_WEEKDAY); /* wkst */

    VECTORIZE(int, testbymonthday4_10_to_15, {10, 11, 12, 13, 14, 15});

    addTest("Every 18 months on the 10th thru 15th of the month "
            "for 10 occurrences",
            15,
            Calendar(1997, Calendar::SEPTEMBER, 10, 9, 0),
            Recurrence::MONTHLY,
            NULL,
            10, /* count */
            18, /* interval */
            vector<DayAndPosition>(), /* byDay */
            testbymonthday4_10_to_15, /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::NO_WEEKDAY); /* wkst */


    VECTORIZE(DayAndPosition, testbyday7_Tu, {
      DayAndPosition(Calendar::TUESDAY, 0)});

    addTest("Every Tuesday, every other month",
            30,
            Calendar(1997, Calendar::SEPTEMBER, 2, 9, 0),
            Recurrence::MONTHLY,
            NULL,
            0, /* count */
            2, /* interval */
            testbyday7_Tu, /* byDay */
            vector<int>(), /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::NO_WEEKDAY); /* wkst */

    VECTORIZE(int, testbymonth2_JunJul, {Calendar::JUNE, Calendar::JULY});

    addTest("Yearly in June and July for 10 occurrences",
            20,
            Calendar(1997, Calendar::JUNE, 10, 9, 0),
            Recurrence::YEARLY,
            NULL,
            10, /* count */
            1, /* interval */
            vector<DayAndPosition>(), /* byDay */
            vector<int>(), /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            testbymonth2_JunJul, /* byMonth */
            Calendar::NO_WEEKDAY); /* wkst */

    VECTORIZE(int, testbymonth3_JanFebMar, {Calendar::JANUARY,
                                    Calendar::FEBRUARY,
                                    Calendar::MARCH});

    addTest("Every other year in January, February, and March "
            "for 10 occurrences",
            15,
            Calendar(1997, Calendar::MARCH, 10, 9, 0),
            Recurrence::YEARLY,
            NULL,
            10, /* count */
            2, /* interval */
            vector<DayAndPosition>(), /* byDay */
            vector<int>(), /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            testbymonth3_JanFebMar, /* byMonth */
            Calendar::NO_WEEKDAY); /* wkst */

    VECTORIZE(int, testbyyearday1_1_100_200, {1, 100, 200});

    addTest("Every 3rd year on the 1st, 100th and 200th day "
            "for 10 occurrences",
            15,
            Calendar(1997, Calendar::JANUARY, 1, 9, 0),
            Recurrence::YEARLY,
            NULL,
            10, /* count */
            3, /* interval */
            vector<DayAndPosition>(), /* byDay */
            vector<int>(), /* byMonthDay */
            testbyyearday1_1_100_200, /* byYearDay */
            vector<int>(), /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::NO_WEEKDAY); /* wkst */

    VECTORIZE(DayAndPosition, testbyday8_20Mo, {
      DayAndPosition(Calendar::MONDAY, 20)});

    addTest("Every 20th Monday of the year, forever",
            20,
            Calendar(1997, Calendar::MAY, 19, 9, 0),
            Recurrence::YEARLY,
            NULL,
            0, /* count */
            1, /* interval */
            testbyday8_20Mo, /* byDay */
            vector<int>(), /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::NO_WEEKDAY); /* wkst */

    VECTORIZE(DayAndPosition, testbyday9_Mo, {
      DayAndPosition(Calendar::MONDAY, 0)});

    VECTORIZE(int, testbyweekno1_20, {20});

    addTest("Monday of week number 20 (where the default start of the "
            "week is Monday), forever",
            20,
            Calendar(1997, Calendar::MAY, 12, 9, 0),
            Recurrence::YEARLY,
            NULL,
            0, /* count */
            1, /* interval */
            testbyday9_Mo, /* byDay */
            vector<int>(), /* byMonthDay */
            vector<int>(), /* byYearDay */
            testbyweekno1_20, /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::NO_WEEKDAY); /* wkst */

    VECTORIZE(DayAndPosition, testbyday10_Th, {
      DayAndPosition(Calendar::THURSDAY, 0)});

    VECTORIZE(int, testbymonth4_Mar, {Calendar::MARCH});

    addTest("Every Thursday in March, forever",
            20,
            Calendar(1997, Calendar::MARCH, 13, 9, 0),
            Recurrence::YEARLY,
            NULL,
            0, /* count */
            1, /* interval */
            testbyday10_Th, /* byDay */
            vector<int>(), /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            testbymonth4_Mar, /* byMonth */
            Calendar::NO_WEEKDAY); /* wkst */

    VECTORIZE(int, testbymonth5_JunJulAug, {Calendar::JUNE, Calendar::JULY,
                                    Calendar::AUGUST});

    addTest("Every Thursday, but only during June, July, and August, forever",
            40,
            Calendar(1997, Calendar::JUNE, 5, 9, 0),
            Recurrence::YEARLY,
            NULL,
            0, /* count */
            1, /* interval */
            testbyday10_Th, /* byDay */
            vector<int>(), /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            testbymonth5_JunJulAug, /* byMonth */
            Calendar::NO_WEEKDAY); /* wkst */

    // XXX we don't support EXDATE.  A real time-switch would handle
    // this with an earlier time-switch representing the exception.

    VECTORIZE(DayAndPosition, testbyday11_Fr, {
      DayAndPosition(Calendar::FRIDAY, 0)});

    VECTORIZE(int, testbymonthday5_13, {13});

    addTest("Every Friday the 13th, forever",
            20,
            Calendar(1997, Calendar::SEPTEMBER, 2, 9, 0),
            Recurrence::MONTHLY,
            NULL,
            0, /* count */
            1, /* interval */
            testbyday11_Fr, /* byDay */
            testbymonthday5_13, /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::NO_WEEKDAY); /* wkst */


    VECTORIZE(DayAndPosition, testbyday12_Sa, {
      DayAndPosition(Calendar::SATURDAY, 0)});

    VECTORIZE(int, testbymonthday6_7_to_13, {7, 8, 9, 10, 11, 12, 13});

    addTest("The first Saturday that follows the first Sunday of the month, "
            "forever",
            20,
            Calendar(1997, Calendar::SEPTEMBER, 13, 9, 0),
            Recurrence::MONTHLY,
            NULL,
            0, /* count */
            1, /* interval */
            testbyday12_Sa, /* byDay */
            testbymonthday6_7_to_13, /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::NO_WEEKDAY); /* wkst */

    VECTORIZE(int, testbymonth6_Nov, {Calendar::NOVEMBER});
    VECTORIZE(int, testbymonthday7_2_to_8, {2, 3, 4, 5, 6, 7, 8});

    addTest("Every four years, the first Tuesday after a Monday in November, "
            "forever (U.S. Presidential Election day)",
            20,
            Calendar(1996, Calendar::NOVEMBER, 5, 9, 0),
            Duration(1, 0, 0, 0), 
            Recurrence::YEARLY,
            NULL,
            0, /* count */
            4, /* interval */
            vector<int>(), /* bySecond */
            vector<int>(), /* byMinute */
            vector<int>(), /* byHour */
            testbyday7_Tu, /* byDay */
            testbymonthday7_2_to_8, /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            testbymonth6_Nov, /* byMonth */
            Calendar::NO_WEEKDAY, /* wkst */
            Calendar::DATE, /* period */
            1,  /* step */
            0x7ff00000); /* end */

    VECTORIZE(DayAndPosition, testbyday13_TuSu, {
      DayAndPosition(Calendar::TUESDAY, 0),
      DayAndPosition(Calendar::SUNDAY, 0)});

    /* Excluded because of BYSETPOS:
       "The third instance into the month of one of Tuesday, Wednesday, "
       "or Thursday, for the next three months"
       */

    /* Excluded because of BYSETPOS:
       "The 2nd to last weekday of the month"
       */

    addTest("Every 3 hours from 9:00 AM to 5:00 PM on a specific day",
            10,
            Calendar(1997, Calendar::SEPTEMBER, 2, 9, 0),
            Duration(0, 1, 0, 0),
            Recurrence::HOURLY,
            &Calendar(1997, Calendar::SEPTEMBER, 2, 17, 0),
            0, /* count */
            3, /* interval */
            vector<int>(), /* bySecond */
            vector<int>(), /* byMinute */
            vector<int>(), /* byHour */
            vector<DayAndPosition>(), /* byDay */
            vector<int>(), /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::NO_WEEKDAY, /* wkst */
            Calendar::HOUR_OF_DAY, /* period */
            1,  /* step */
            1000000000); /* end */

    addTest("Every 15 minutes for 6 occurrences",
            15,
            Calendar(1997, Calendar::SEPTEMBER, 2, 9, 0),
            Duration(0, 0, 1, 0),
            Recurrence::MINUTELY,
            NULL,
            6, /* count */
            15, /* interval */
            vector<int>(), /* bySecond */
            vector<int>(), /* byMinute */
            vector<int>(), /* byHour */
            vector<DayAndPosition>(), /* byDay */
            vector<int>(), /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::NO_WEEKDAY, /* wkst */
            Calendar::MINUTE, /* period */
            1,  /* step */
            880000000); /* end.  Trim low (only a few months after dtStart)
                           because we're doing minutely checks. */

    addTest("Every hour and a half for 4 occurrences",
            10,
            Calendar(1997, Calendar::SEPTEMBER, 2, 9, 0),
            Duration(0, 0, 1, 0),
            Recurrence::MINUTELY,
            NULL,
            4, /* count */
            90, /* interval */
            vector<int>(), /* bySecond */
            vector<int>(), /* byMinute */
            vector<int>(), /* byHour */
            vector<DayAndPosition>(), /* byDay */
            vector<int>(), /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::NO_WEEKDAY, /* wkst */
            Calendar::MINUTE, /* period */
            1,  /* step */
            880000000); /* end.  Trim low (only a few months after dtStart)
                           because we're doing minutely checks. */

    VECTORIZE(int, testbyhour1_9_to_16, {9, 10, 11, 12, 13, 14, 15, 16});
    VECTORIZE(int, testbyminute1_0_20_40, {0, 20, 40});

    addTest("Every 20 minutes from 9:00 AM to 4:40 PM every day (1)",
            50,
            Calendar(1997, Calendar::SEPTEMBER, 2, 9, 0),
            Duration(0, 0, 1, 0),
            Recurrence::DAILY,
            NULL,
            0, /* count */
            1, /* interval */
            vector<int>(), /* bySecond */
            testbyminute1_0_20_40, /* byMinute */
            testbyhour1_9_to_16, /* byHour */
            vector<DayAndPosition>(), /* byDay */
            vector<int>(), /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::NO_WEEKDAY, /* wkst */
            Calendar::MINUTE, /* period */
            1,  /* step */
            880000000); /* end.  Trim low (only a few months after dtStart)
                           because we're doing minutely checks. */

    addTest("Every 20 minutes from 9:00 AM to 4:40 PM every day (2)",
            50,
            Calendar(1997, Calendar::SEPTEMBER, 2, 9, 0),
            Duration(0, 0, 1, 0),
            Recurrence::MINUTELY,
            NULL,
            0, /* count */
            20, /* interval */
            vector<int>(), /* bySecond */
            vector<int>(), /* byMinute */
            testbyhour1_9_to_16, /* byHour */
            vector<DayAndPosition>(), /* byDay */
            vector<int>(), /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::NO_WEEKDAY, /* wkst */
            Calendar::MINUTE, /* period */
            1,  /* step */
            880000000); /* end.  Trim low (only a few months after dtStart)
                           because we're doing minutely checks. */

    addTest("An example where the days generated makes a difference "
            "because of WKST",
            8,
            Calendar(1997, Calendar::AUGUST, 5, 9, 0),
            Recurrence::WEEKLY,
            NULL,
            4, /* count */
            2, /* interval */
            testbyday13_TuSu, /* byDay */
            vector<int>(), /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::MONDAY); /* wkst */

    addTest("changing only WKST from MO to SU, yields different results...",
            8,
            Calendar(1997, Calendar::AUGUST, 5, 9, 0),
            Recurrence::WEEKLY,
            NULL,
            4, /* count */
            2, /* interval */
            testbyday13_TuSu, /* byDay */
            vector<int>(), /* byMonthDay */
            vector<int>(), /* byYearDay */
            vector<int>(), /* byWeekNo */
            vector<int>(), /* byMonth */
            Calendar::SUNDAY); /* wkst */

    /* XXX put the old time zone back */
    return tests;
  }

