// timezone.cpp -- object wrapper around struct tzinfo

#include "timezone.h"
#include "calendar.h"

#include <cstdlib>
#include <strstream>

using namespace std;

char * TimeZone::default_tzname;

/*
 * TimeZone::TimeZone()
 * Construct a new timezone object, based on the current default timezone.
 */
TimeZone::TimeZone() {
  init(getDefault());
} /* TimeZone::TimeZone() */


/*
 * TimeZone::TimeZone(const char *)
 * Construct a new timezone object, based on the given time zone name.
 */
TimeZone::TimeZone(const char *name) {
  init(name);
} /* TimeZone::TimeZone(const char *) */


/*
 * TimeZone::~TimeZone()
 * Destruct a timezone object.
 */
TimeZone::~TimeZone() {
  delete[] tzname;
  tz_free(tz);
} /* TimeZone::~TimeZone() */


/*
 * TimeZone::init(const char *)
 * Initialize a timezone object.
 */
void TimeZone::init(const char * name)
{
  if (name == NULL) {
    tzname = NULL;
  }
  else {
    tzname = new char[strlen(name) + 1];
    strcpy(tzname, name);
  }

  int err;
  err = tz_prep(&tz, tzname);
  if (err != 0) {
    tz = NULL;
    
    ostrstream buf;
    buf << "Error setting time zone: " << (tzname ? tzname : "(System)");
    buf << ": " << strerror(err) << ends;
    
    timezone_error e(buf.str());
    buf.rdbuf()->freeze(0);
    throw e;
  }
} /* TimeZone::init(const char *) */


/*
 * TimeZone::getDefault()
 * Get the current default time zone
 * XXX: thread-safety
 */
const char * TimeZone::getDefault()
{
  if (default_tzname != NULL) {
    return default_tzname;
  }
  else {
    return getenv("TZ");
  }
} /* TimeZone::getDefault() */


/*
 * TimeZone::setDefault()
 * Set the current default time zone.
 * XXX: thread safety
 */
void TimeZone::setDefault(const char * name)
{
  if (default_tzname != NULL) {
    delete[] default_tzname;
  }

  if (name == NULL) {
    default_tzname = NULL;
  }
  else {
    default_tzname = new char[strlen(name) + 1];
    strcpy(default_tzname, name);
  }
} /* TimeZone::setDefault() */
