/*
** This file is in the public domain.  Contributed by Jonathan Lennox
** <lennox@cs.columbia.edu>
*/

#include "private.h"
#include "tzthread-posix.h"
#include "struct_tzinfo.h"

#include <pthread.h>

/*
** POSIX Thread versions of the TZ thread-safety operations.
*/

static pthread_mutex_t _tz_lcl_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t _tz_gmt_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t _tz_strftime_mutex = PTHREAD_MUTEX_INITIALIZER;

static int get_lclptr_key P((pthread_key_t *));

#ifdef ALL_STATE
static struct tzinfo *	_tz_lclptr;
static struct tzinfo *	_tz_gmtptr;

struct tzinfo*
_tz_get_lclptr(process_wide)
int *	process_wide;
{
	pthread_key_t lclptr_key;
	struct tzinfo *ret;

	if (get_lclptr_key(&lclptr_key) != 0) {
		return NULL;
	}
	if ((ret = pthread_getspecific(lclptr_key)) != NULL) {
		*process_wide = 0;
		return ret;
	}

	*process_wide = 1;

	if (_tz_lclptr == NULL) {
		_tz_lclptr = (struct tzinfo *) malloc(sizeof *_tz_lclptr);
	}
	return _tz_lclptr;
}


struct tzinfo*
_tz_get_gmtptr P((void))
{
	/*
	** Fast-path the common case. Nothing sets or alters gmtptr
	** once it's set.
	*/
	if (_tz_gmtptr != NULL) {
		return _tz_gmtptr;
	}

	_tz_lock_gmtptr();
	/* 
	** Check the pointer again, in case another thread just 
	** allocated it.
	*/
	if (_tz_gmtptr != NULL) {
		_tz_unlock_gmtptr();
		return _tz_gmtptr;
	}

	_tz_gmtptr = (struct tzinfo *) malloc(sizeof *_tz_gmtptr);
	_tz_unlock_gmtptr();

	return _tz_gmtptr;
}

#else /* !defined ALL_STATE */

static struct tzinfo _tz_lclmem;
static struct tzinfo _tz_gmtmem;

struct tzinfo *
_tz_get_lclptr(process_wide)
int *	process_wide;
{
	pthread_key_t lclptr_key;
	struct tzinfo *ret;

	if (get_lclptr_key(&lclptr_key) != 0) {
		return NULL;
	}
	if ((ret = pthread_getspecific(lclptr_key)) != NULL) {
		*process_wide = 0;
		return ret;
	}

	*process_wide = 1;

	return &_tz_lclmem;
}


struct tzinfo *
_tz_get_gmtptr P((void))
{
	return &_tz_gmtmem;
}

#endif /* defined ALL_STATE */

void _tz_lock_lclptr P((void))
{
	pthread_mutex_lock(&_tz_lcl_mutex);
}


void _tz_lock_gmtptr P((void))
{
	pthread_mutex_lock(&_tz_gmt_mutex);
}


extern void _tz_unlock_lclptr P((void))
{
	pthread_mutex_unlock(&_tz_lcl_mutex);
}


extern void _tz_unlock_gmtptr P((void))
{
	pthread_mutex_unlock(&_tz_gmt_mutex);
}

void _tz_lock_strftime_cache P((void))
{
	pthread_mutex_lock(&_tz_strftime_mutex);
}


void _tz_unlock_strftime_cache P((void))
{
	pthread_mutex_unlock(&_tz_strftime_mutex);
}



struct tm*
_tz_get_tm_buf P((void))
{
	static pthread_mutex_t tmbuf_mutex = PTHREAD_MUTEX_INITIALIZER;
	static pthread_key_t tmbuf_key;
	static int tmbuf_initialized = 0;
	struct tm *p_tm;
	int err;

	if (!tmbuf_initialized) {
		pthread_mutex_lock(&tmbuf_mutex);
		if (!tmbuf_initialized) {
			if ((err = pthread_key_create(&tmbuf_key, free))
					!= 0) {
				errno = err;
				pthread_mutex_unlock(&tmbuf_mutex);
				return NULL;
			}
			tmbuf_initialized = 1;
		}
		pthread_mutex_unlock(&tmbuf_mutex);
	}

	p_tm = pthread_getspecific(tmbuf_key);
	if (p_tm == NULL) {
		if ((p_tm = (struct tm *) malloc(sizeof(struct tm))) == NULL) {
			return NULL;
		}
		if ((err = pthread_setspecific(tmbuf_key, p_tm)) != 0) {
			errno = err;
			free(p_tm);
			return NULL;
		}
	}
	return p_tm;
}


char* _tz_get_asctime_buf P((void))
{
	static pthread_mutex_t asctime_buf_mutex = PTHREAD_MUTEX_INITIALIZER;
	static pthread_key_t asctime_buf_key;
	static int asctime_buf_initialized = 0;
	char *asctime_buf;
	int err;

	if (!asctime_buf_initialized) {
		pthread_mutex_lock(&asctime_buf_mutex);
		if (!asctime_buf_initialized) {
			if ((err = pthread_key_create(&asctime_buf_key, free))
					!= 0) {
				errno = err;
				pthread_mutex_unlock(&asctime_buf_mutex);
				return NULL;
			}
			asctime_buf_initialized = 1;
		}
		pthread_mutex_unlock(&asctime_buf_mutex);
	}

	asctime_buf = pthread_getspecific(asctime_buf_key);
	if (asctime_buf == NULL) {
		if ((asctime_buf = (char *) malloc(ASCTIME_BUF_SIZE))
				== NULL) {
			return NULL;
		}
		if ((err = pthread_setspecific(asctime_buf_key, asctime_buf))
				!= 0) {
			errno = err;
			free(asctime_buf);
			return NULL;
		}
	}
	return asctime_buf;
}


int _tz_set_thread_lclptr(tz)
const struct tzinfo * tz;
{
	pthread_key_t lclptr_key;
	int err;
	struct tzinfo *old, *new;

	if ((err = get_lclptr_key(&lclptr_key)) != 0) {
		return err;
	}
	if ((old = pthread_getspecific(lclptr_key)) != NULL) {
	  free(old);
	}

	new = (struct tzinfo *) malloc(sizeof(struct tzinfo));
	if (new == NULL) {
	  return ENOMEM;
	}

	*new = *tz;

	if ((err = pthread_setspecific(lclptr_key, new)) != 0) {
	  free(new);
	  return err;
	}
	return 0;
}


int _tz_clear_thread_lclptr P((void))
{
	pthread_key_t lclptr_key;
	int err;
	struct tzinfo *old;

	if ((err = get_lclptr_key(&lclptr_key)) != 0) {
		return err;
	}
	if ((old = pthread_getspecific(lclptr_key)) != NULL) {
	  free(old);
	}

	return pthread_setspecific(lclptr_key, NULL);
}


static int get_lclptr_key(lclptr_key_p)
pthread_key_t *		lclptr_key_p;
{
	static pthread_mutex_t lclptr_key_mutex = PTHREAD_MUTEX_INITIALIZER;
	static pthread_key_t lclptr_key;
	static int lclptr_key_initialized = 0;

	int err;

	if (!lclptr_key_initialized) {
		pthread_mutex_lock(&lclptr_key_mutex);
		if (!lclptr_key_initialized) {
			if ((err = pthread_key_create(&lclptr_key, free))
					!= 0) {
				pthread_mutex_unlock(&lclptr_key_mutex);
				return err;
			}
			lclptr_key_initialized = 1;
		}
		pthread_mutex_unlock(&lclptr_key_mutex);
	}

	*lclptr_key_p = lclptr_key;
	return 0;
}
