#ifndef TZTHREAD_POSIX_H
#define TZTHREAD_POSIX_H

/*
** This file is in the public domain.  Contributed by Jonathan Lennox
** <lennox@cs.columbia.edu>
*/

/*
** Thread primitives for tzcode.  TZcode requires the following primitive
** functions to be defined (probably as macros):
** void LOCK_LCLPTR(void)	- serialize references to the process-wide or
**				  thread-specific local timezone object.
** void LOCK_GMTPTR(void)	- serialize references to write to the global
**				  UTC timezone object.
** void UNLOCK_LCLPTR(void)	- Unlock the local timezone object.
** void UNLOCK_GMTPTR(void)	- Unlock the UTC timezone object.
** void LOCK_STRFTIME_CACHE(void)
**				- Serialize references to strftime()'s locale
**				  cache.
** void UNLOCK_STRFTIME_CACHE(void)
**				- Unlock strftime()'s locale cache.
** 
** struct tzinfo * GET_LCLPTR(int *process)
**				- Return the current local timezone object.
**				  The argument 'process' is filled in with
**				  a boolean value indicating whether the
**				  returned object is process-wide (true)
**				  or thread-specific (false).
**				  This will only be called when the local
**				  pointer is locked.
**				  The thread primitive is not responsible for
**				  filling in the value, only for assuring that
**				  sizeof(struct tzinfo) bytes are available.
**				  This function returns NULL on failure.
**
** struct tzinfo * GET_GMTPTR(void)
**				- Return the UTC timezone object.
**				  This *can* be called when the GMT pointer
**				  is not locked.  This function is responsible
**				  for any resulting serialization necessary.
**				  Any attempts to write to the object *will*
**				  lock the object first, however.
**				  The thread primitive is not responsible for
**				  filling in the value, only for assuring that
**				  sizeof(struct tzinfo) bytes are available.
**				  This function returns NULL on failure.
**
** struct tm *GET_TM_BUF(void)	- Get a pointer to a struct tm, suitable for 
**				  use as a return value for the localtime()
**				  family of functions.  In a threaded
**				  environment, this buffer should be thread-
**				  local.
**
** int SET_THREAD_LCLPTR(struct tzinfo *)
**				- Set the thread-specific local timezone
**				  object to the argument.  Calls to
**				  GET_LCLPTR() from the current thread should
**				  from now on return this value.
**				  If a previous local pointer was set, free
**				  it first.
**				  Return an errno error code on failure, or 0.
**
** int CLEAR_THREAD_LCLPTR(void)
**				- Clear the thread-specific local timezone
**				  object to the argument.  Calls to
**				  GET_LCLPTR() from the current thread should
**				  from now on return the global pointer.
**				  If a previous local pointer was set, free
**				  it first.
**				  Return an errno error code on failure, or 0.
*/

#include "private.h"

/*
** POSIX Thread versions of the TZ thread-safety operations.
*/

extern struct tzinfo *	_tz_get_lclptr P((int *));
extern struct tzinfo *	_tz_get_gmtptr P((void));

#define GET_LCLPTR _tz_get_lclptr
#define GET_GMTPTR _tz_get_gmtptr

extern void _tz_lock_lclptr P((void));
extern void _tz_lock_gmtptr P((void));

extern void _tz_unlock_lclptr P((void));
extern void _tz_unlock_gmtptr P((void));

#define LOCK_LCLPTR    _tz_lock_lclptr
#define LOCK_GMTPTR    _tz_lock_gmtptr

#define UNLOCK_LCLPTR    _tz_unlock_lclptr
#define UNLOCK_GMTPTR    _tz_unlock_gmtptr

extern void _tz_lock_strftime_cache P((void));
extern void _tz_unlock_strftime_cache P((void));

#define LOCK_STRFTIME_CACHE _tz_lock_strftime_cache
#define UNLOCK_STRFTIME_CACHE _tz_unlock_strftime_cache

extern struct tm* _tz_get_tm_buf P((void));

#define GET_TM_BUF _tz_get_tm_buf

extern char* _tz_get_asctime_buf P((void));

#define GET_ASCTIME_BUF _tz_get_asctime_buf

extern int _tz_set_thread_lclptr P((const struct tzinfo *));
extern int _tz_clear_thread_lclptr P((void));

#define SET_THREAD_LCLPTR _tz_set_thread_lclptr
#define CLEAR_THREAD_LCLPTR _tz_clear_thread_lclptr

#endif /* !defined TZTHREAD_POSIX_H */


