#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>     /* printf() */
#include <unistd.h>    /* sleep */
#include <stdlib.h>
#include <pthread.h>
#include <string.h>
#include <limits.h>   /* INT_MAX */

#include "http.h"
#include "pthread-util.h"
#include "servers.h"
#include "timespec.h"
#include "tcp.h"
/*
#include "ssl_tcp.h"
*/
#include "request-sip.h"
#include "rr_support.h"
#include "config.h"  
#include "msgflow.h"
#include "sipsnmp_interface.h"

/* The thread which handles a single SIP client, i.e. one branch.
 * Implemented as a state machine.

   Return values:
   FALSE - next location should be tried
   TRUE - next location should not be tried

   States:
   client_initial
   client_retransmitting_request
   client_awaiting_responses
   client_sent_final
   client_retransmitting_cancel
   client_awaiting_cancel_response
   client_got_cancel_final

   Events:
   1xx recevied for request
   2xx received for request
   3xx - 6xx received for request

   1xx received for cancel
   2xx - 6xx received for cancel

   ACK sent by parent
   message_shutdown_client sent by parent
   message_cancel_branch sent by parent

   socket write fails (or TCP socket closed by server)
   timeout expires

   State transitions:
   initial:
   (n.b. cannot receive any normal events before jumping to another state)
   lookup destinations for specified URL.
   some found: send -> client_retransmitting_request
   none found: generate 404 & send to parent -> client_sent_final

   client_retransmitting_request:
        timeout = time to next retransmission
   1xx received: send to parent -> client_awaiting_responses
   2xx received: send to parent, record -> client_sent_final
   3xx - 6xx received: send to parent, ACK if invite -> client_sent_final

   response for cancel: ignore -> client_retransmitting_request

   ACK sent: ignore -> client_retransmitting_request
   message_shutdown_client sent: exit from state machine, TRUE
   message_cancel_branch sent: send CANCEL; generate 487 & send to parent
   	-> client_retransmitting_cancel

   socket write fails: exit from state machine, FALSE
   timeout expires: do next retransmission.  If none, exit from state
   machine, FALSE (XXX: send CANCEL?)

   client_awaiting_responses:
        timeout:
          if req. was INVITE
             time until "Expires" of request, or infinite (or pref)
          if req. was any other
             time until "Expires"
   1xx received: send to parent -> client_awaiting_responses
   2xx received: send to parent, record -> client_sent_final
   3xx - 6xx received: send to parent, ACK if invite -> client_sent_final

   response for cancel: ignore -> client_awaiting_responses

   ACK sent: ignore -> client_awaiting_responses

   message_shutdown_client sent: exit from state machine, status TRUE
   message_cancel_branch sent: send CANCEL; generate 487 & send to parent
   	  -> client_retransmitting_cancel

   socket write fails: generate 500 & send to parent;
        -> client_sent_final
   timeout expires: send CANCEL; generate 408 & send to parent;
        -> client_retransmitting_cancel


   client_sent_final:
        timeout is infinite
   1xx received: ignore -> client_sent_final
   2xx received: send to parent, record -> client_sent_final
   3xx - 6xx received: ACK if invite -> client_sent_final

   response for cancel: ignore -> client_sent_final

   ACK sent: if sent 200, send to branch -> client_sent_final

   message_shutdown_client sent: exit from state machine, status TRUE
   message_cancel_branch sent: ignore -> client_sent_final

   socket write fails: ignore, not a problem -> client_sent_final
   timeout expires: ignore, log error -> client_sent_final


   client_retransmitting_cancel:
       timeout = time until next cancel retransmission
   1xx for request: ignore -> client_retransmitting_cancel
   2xx for request: send to parent, record -> client_sent_final
   3xx - 6xx for request: ACK if invite -> client_sent_final

   1xx for cancel: -> client_awaiting_cancel_response
   2xx - 6xx for cancel: -> client_got_cancel_final

   ACK sent: ignore, -> client_retransmitting_cancel
   message_shutdown_client: exit from state machine, TRUE
   message_cancel_branch: ignore -> client_retransmitting_cancel (log
   	error?)

   socket write fails: -> client_got_cancel_final (???)
   timeout expires: do next retransmission.  If none,
   	-> client_got_cancel_final (???)


   client_awaiting_cancel_response:
        timeout = T2 (slow retransmit) or infinite afterward
   1xx for request: ignore -> client_awaiting_cancel_response
   2xx for request: send to parent, record -> client_sent_final
   3xx - 6xx for request: ACK if invite -> client_sent_final

   1xx for cancel: ignore -> client_awaiting_cancel_response
   2xx - 6xx for cancel: -> client_got_cancel_final

   ACK sent: ignore -> client_awaiting_cancel_response
   message_shutdown_client: exit from state machine, TRUE
   message_cancel_branch: ignore -> client_awaiting_cancel_response (log
   	error?)

   socket write fails: -> client_got_cancel_final (???)
   timeout expires: do next retransmission.  If none,
   	-> client_got_cancel_final (???)

   
   client_got_cancel_final:
        timeout = infinite
   1xx for request: ignore -> client_got_cancel_final
   2xx for request: send to parent, record -> client_sent_final
   3xx - 6xx for request: ACK if invite -> client_sent_final

   1xx for cancel: ignore -> client_got_cancel_final
   2xx - 6xx for cancel: ignore -> client_got_cancel_final

   ACK sent: ignore -> client_got_cancel_final
   message_shutdown_client: exit from state machine, TRUE
   message_cancel_branch: ignore -> client_got_cancel_final (log
   	error?)

   socket write fails: ignore
   timeout expires: ignore, log error
   */


/* The states described above */
typedef enum {
  client_initial,
  client_retransmitting_request, client_awaiting_responses, client_sent_final,
  client_retransmitting_cancel, client_awaiting_cancel_response,
  client_got_cancel_final,
  client_done
} client_state_t;


/* The events described above */
typedef enum {
  event_request_1xx, event_request_2xx, event_request_3xx_6xx,
  event_cancel_1xx, event_cancel_2xx_6xx,
  event_ack, event_shutdown, event_cancel,
  event_sockerr, event_timeout,
  event_unknown
} client_event_t;


/* Forward declarations of local functions */
/* XXX: sort these */
static int AllowedToForward(request_t *r);
static void DecrementMaxForwards(headers_t *hdrs);
static int ClientSendToOneLocation(branch_t *branch, server_t location);
static int ClientSendStatelessToOneLocation(request_t *r, message_t *m,
                                           int s, DString *uri_str);
static client_event_t ClientCategorizeMessage(branch_t *branch, message_t *m);

static int ClientStateMachine(branch_t *branch, int s, int is_invite,
                              int socktype);

static void ClientBuildProxiedRequest(DString *msg, message_t *m,
                                      branch_t *branch, int socktype);

static void ClientCreateFollowupRequest(DString *msg, branch_t *branch,
                                        parser_t *p, char *method,
                                        int socktype);
static void CopyAndMungeVias(headers_t *hdrs, parser_header_t *src,
                             sip_via_t *parent_via);
static void PrintRequestLine(DString *req_line,
                             const char *method, const DString *uri_str,
                             const char *proto, const char *version);
static void AddLocalVia(headers_t *hdrs, char *protocol, char *version,
                        int socktype, unsigned long hash, int branchID);
static server_t *ClientGetLocations(uri_t uri, int type);
static int CreateClientSocket(server_t location);
static void CloseClientSocket(int sock);
static void SendRequest(int s, DString *msg, branch_t *branch);
static void SendMessageToParent(message_t *m, branch_t *branch);
static void ClientACK(int s, message_t *m, branch_t *branch, int socktype);
static void ClientProxyACK(int s, message_t *m, branch_t *branch,
                           int socktype);

static void SendPseudoResponseToParent(int status, branch_t *branch);

typedef struct statemach_info statemach_info;

/* Forward declarations for functions which depend on statemach_info */
static client_event_t WaitNextEvent(branch_t *branch, message_t **m,
                                    statemach_info *state);
static void CalculateMaxRetransmits(statemach_info *state);
static double CalculateNextDelay(statemach_info *state);
static void DoClientStateTransitions(client_event_t event, message_t *message,
                                     statemach_info *state);


static void DoRRTransitions(client_event_t event, message_t *message,
                            statemach_info *state);
static void DoARTransitions(client_event_t event, message_t *message,
                            statemach_info *state);
static void DoSFTransitions(client_event_t event, message_t *message,
                            statemach_info *state);
static void DoRCTransitions(client_event_t event, message_t *message,
                            statemach_info *state);
static void DoACRTransitions(client_event_t event, message_t *message,
                             statemach_info *state);
static void DoGCFTransitions(client_event_t event, message_t *message,
                             statemach_info *state);

static void StartCancel(statemach_info *state);

/*
 * Reasons why an errno is a "normal" message failure.  I.e., it's
 * failure to reach a host, rather than internal failure.
 */
#define ERRNO_NORMAL_FAILURE(err) \
   ((err) == ECONNREFUSED || \
    (err) == ECONNRESET || \
    (err) == ETIMEDOUT || \
    (err) == EHOSTUNREACH || \
    (err) == ENETUNREACH)

/* XXX: EADDRNOTAVAIL also? */



/* 
 * client()
 * Client thread responsible for one client branch.  Called with
 * branch_t argument, which points to the parent request.  client()
 * looks up the locations corresponding to the given URI, then starts a
 * client state machine.  If the state machine returns FALSE, tries the
 * next location.  If no responses were found, sends a 404 pseudo-response
 * to the parent.
 */
void *client(void *arg)
{
  branch_t *branch = (branch_t *) arg;
  request_t *r = branch->parent;
  server_t *locations = NULL;  /* list of servers */
  int si;               /* destination index in server table */
  struct message_t *m, *m_next;
  int done = 0;
  DString uri_str;      /* destination URI as string */

  pthread_mutex_lock(&r->mutex);

  DStringInit(&uri_str);

  URI_SPrint(&branch->uri, &uri_str);

  debug("client", "r=%p, branch %d: starting, URI=%s\n",
        r, branch->branch, DStringValue(&uri_str));

  /* If the request had a MaxForwards: 0 header, return "too many hops" and
   * don't bother doing anything else. */
  if (!AllowedToForward(r)) {
    SendPseudoResponseToParent(483, branch);
    goto cleanup;
  }

  if ( r->outbound_proxy.host != NULL ) {
    /* Outbound proxy is set, send the request to this,
     * irrespective of the request URI.
     * This is needed only for user agenets. */
    locations = ClientGetLocations(r->outbound_proxy, r->preferred_type);
  }
  else {
    /* Connect to the request uri specified */
    locations = ClientGetLocations(branch->uri, r->preferred_type);
  }

  debug("client", "r=%p, branch %d: ClientGetLocations completed\n",
        r, branch->branch);

  for (si = 0; !done && locations[si].host != NULL; si++) {
    r->remote_sin = locations[si].sin; /* Set remote addr for parent */
    done = ClientSendToOneLocation(branch, locations[si]);
  }

  /* Check to make sure we sent sucessfully (i.e., there were locations
     returned, and at least one of them got through). */
  if (!done) {
    SendPseudoResponseToParent(404, branch);
  }

  if (done < 0) {
    /* Some error connecting -- should this be 400-class? */
    SendPseudoResponseToParent(500, branch);
  }

  /* Free storage and exit. */

cleanup:
  DStringFree(&uri_str);
  if (locations) {
    for (si = 0; locations[si].host != NULL; si++) {
      free(locations[si].host);
    }
    free(locations);
  }

  for (m = branch->mlist; m != NULL; m = m_next) {
    m_next = m->next;
    MessageDelete(m);
  }
  branch->mlist = NULL;

  MessageDelete(branch->orig);

  pthread_mutex_unlock(&r->mutex);

  return NULL; 
} /* client */


/* 
 * client_stateless()
 * This is a stateless sending of a message to remote.
 * The mutex is already locked. The mutex is not released
 * when the function exits. This is a regular function as
 * opposed to a thread. 
 * return 0 on success or appropriate error code.
 * The possible error codes are
 *  483: If MaxForward is reached.
 *  404: If no contact found.
 *       Or the first contact is not UDP
 *       Or could not connect to the first contact location.
 *  400: Socket error while send() was called.
 */
int client_stateless(request_t *r, uri_t *uri, message_t *m)
{
  server_t *locations = NULL;  /* list of servers */
  int si;               /* destination index in server table */
  DString uri_str;      /* destination URI as string */
  int status = 0;       /* return value */

  /* Format the request URI as a string */
  DStringInit(&uri_str);
  URI_SPrint(uri, &uri_str);

  debug("client_stateless", "r=%p, starting, URI=%s\n",
        r, DStringValue(&uri_str));

  /* If the request had a MaxForwards: 0 header, return "too many hops" and
   * don't bother doing anything else. */
  if (!AllowedToForward(r)) {
    status = 483;
    goto cleanup;
  }

  if ( r->outbound_proxy.host != NULL ) {
    /* Outbound proxy is set, send the request to this,
     * irrespective of the request URI.
     * This is needed only for user agenets. */
    locations = ClientGetLocations(r->outbound_proxy, SOCK_DGRAM);
  }
  else {
    /* Connect to the request uri specified */
    locations = ClientGetLocations(*uri, SOCK_DGRAM);
  }

  debug("client_stateless", "r=%p, ClientGetLocations completed\n",
        r);

  /* 
   * Stateless proxy is done only for 
   * UDP locations 
   */
  if (locations[0].host != NULL &&
      locations[0].type == SOCK_DGRAM) {
    int s;   /* Socket id */ 

    debug("client_stateless", "Found %s\n", locations[0].host);
    r->remote_sin = locations[0].sin; /* Set remote addr for parent */

    /* create and connect socket */
    if ((s = CreateClientSocket(locations[0])) < 0) {
      debug("client_stateless", "Could not connect to location\n");
      status = 404;
    }
    else {
      status = ClientSendStatelessToOneLocation(r, m, s,
                                                &uri_str);
      CloseClientSocket(s);
    } /* Client socket creation failed */
  } /* No location found */
  else {
    debug("client_stateless", "No UDP contact location found\n");
    status = 404;
  }

  /* Free storage and exit. */

cleanup:
  DStringFree(&uri_str);
  if (locations) {
    for (si = 0; locations[si].host != NULL; si++) {
      free(locations[si].host);
    }
    free(locations);
  }

  return status; 
} /* client_stateless */


/*
 * AllowedToForward()
 * Return true if the request r can legally be forwarded; i.e., 
 * if it has a Max-Forwards header its value is positive.
 */
static int AllowedToForward(request_t *r)
{
  parser_header_t *h;
  char *end_ptr;
  long max_forwards;

  h = RFC822_Header(r->p.headers, "Max-Forwards");
  if (h == NULL || h->elements == NULL || h->elements->element == NULL) {
    /* No Max-Forwards header; proxying is okay. */
    return 1;
  }

  errno = 0;
  max_forwards = strtol(h->elements->value, &end_ptr, 10);
  if (errno != 0 || end_ptr[0] != '\0') {
    /* Mal-formed Max-Forwards header; assume value of '0'. */
    return 0;
  }
  if (max_forwards == 0) {
    return 0;
  }

  return 1;
} /* AllowedToForward */


/*
 * DecrementMaxForwards()
 * Decrement the Max-Forwards header of the outgoing request,
 * if there is one.
 */
static void DecrementMaxForwards(headers_t *hdrs)
{
  char *end_ptr;
  char *max_forwards_str;
  void *dummy = NULL;
  long max_forwards;

  dummy = NULL;
  max_forwards_str = HeaderGet(hdrs, "Max-Forwards", 1, &dummy);
  if (max_forwards_str[0] == '\0') {
    return;
  }

  errno = 0;
  max_forwards = strtol(max_forwards_str, &end_ptr, 10);
  if (errno != 0 || end_ptr[0] != '\0') {
    /* Mal-formed Max-Forwards header; just drop it. */
    HeaderDelete(hdrs, "Max-Forwards");
    return;
  }

  HeaderPrintf(hdrs, 1, "Max-Forwards", "%ld", max_forwards - 1);
} /* DecrementMaxForwards */


/*
 * ClientSendToOneLocation()
 * Send the request to one location: open the socket, send the initial
 * request, and start the client state machine.  Returns when either the
 * transaction is finished, or we've determined we can't send to this
 * location.  Returns 0 if we should try the next location (normal failure);
 * true values mean don't continue (negative is error, positive is success).
 */
static int ClientSendToOneLocation(branch_t *branch, server_t location)
{
  int s;                /* Socket id */ 
  char *method;
  int is_invite;
  int ret;
  
  /* create and connect socket */
  if ((s = CreateClientSocket(location)) < 0) {
    if (s == -1) return 0;
    return -1;
  }
  
  /* Check if this was an invite request */
  method = HTTP_Method(&branch->orig->p);
  is_invite = (strcasecmp(method, "INVITE") == 0);
  free(method);

  ret = ClientStateMachine(branch, s, is_invite, location.type);

  CloseClientSocket(s);

  return ret;
} /* ClientSendToOneLocation */

/*
 * ClientSendStatelessToOneLocation()
 * Send the request statelessly to one location.
 * Returns 0 on success or an error code on
 * failure.
 * Possible error codes:
 *  400 : is the send() fails.
 */
static int ClientSendStatelessToOneLocation(request_t *r, message_t *m,
                                           int s, DString *uri_str)
{
  /* Connected to the remote socket */
  DString reqmsg;
  char *method;
  char *version;
  char *protocol;
  headers_t hdrs;
  headers_t tmp_hdrs;
  int n;
  int status = 0; /* return value */

  DStringInit(&reqmsg);

  /* Perform initial transmission */
  method   = HTTP_Method(&m->p);
  version  = HTTP_Version(&m->p);
  protocol = HTTP_Protocol(&m->p);
  
  hdrs.headers = NULL;
  hdrs.last = NULL;
  tmp_hdrs.headers = NULL;
  tmp_hdrs.last = NULL;

  /* if the record_route var is set, and the method is invite,
   * call our AddRecordRoute function. ignore the branch param
   * as this is stateless proxy.
   */
  if (config.RecordRoute && !strcmp(method,"INVITE")) 
    AddRecordRoute_Request(m, r->hash, 0);
    
  PrintRequestLine(&reqmsg, method, uri_str, protocol, version);
  
  AddLocalVia(&hdrs, protocol, version, 
              SOCK_DGRAM, 0,
              0);
  CopyAndMungeVias(&hdrs, m->p.headers, &r->via);
  
  /* If the request to be proxied is missing a Content-Length
   * header field, add one.  Needed for UDP -> TCP proxying. */
  if (RFC822_Header(m->p.headers, "Content-Length") == NULL) {
    HeaderPrintf(&hdrs, 0, "Content-Length", "%d", m->body.content_length);
  }
  
  /* This is a kludge; I should write a HeaderCopyAllParsedHeadersBut()
   * function for efficiency.*/
  HeaderCopyParsedHeaders(&tmp_hdrs, m->p.headers);
  HeaderCopyAllBut(&hdrs, &tmp_hdrs, "Via");
  
  DecrementMaxForwards(&hdrs);
  
  /* Delete first "Route" header */
  HeaderDeleteFirst(&hdrs, "Route");
      
  HeaderMessage(&reqmsg, &hdrs);
  DStringAppend(&reqmsg, "\r\n", 2);
  DStringAppend(&reqmsg, m->body.content, m->body.content_length);
  
  free(method);
  free(version);
  free(protocol);
  HeaderFree(&tmp_hdrs);
  HeaderFree(&hdrs);


  if (SnmpProcessMessage) {
    sipsnmp_extra_data data;
    /* Read the method out of the msg string.
     * XXX this is a hack: should pass the method as a string to this
     * function.
     */
    sscanf(reqmsg.string, "%19s", data.request_type);
        
    (*SnmpProcessMessage)(r, SNMP_TYPE_REQ_OUT, &data);
  }
      
  debug("ClientSendStatelessToOneLocation", 
        "Sending out message \n%s\n",
        reqmsg.string);
  if ((n = send(s, reqmsg.string, reqmsg.length, 0)) < 0) {
    if (!(ERRNO_NORMAL_FAILURE(errno))) {
      error("ClientSendToOneLocation", "Error sending proxied request\n");
    }
    else {
      debug("ClientSendToOneLocation",
            "Normal error sending proxied request: %s (%d)\n",
            strerror(errno), errno);
    }
    status = 400; /* socket error */
  }

  /* Free the message and close the socket */
  DStringFree(&reqmsg);

  return status;
} /* ClientSendStatelessToOneLocation */

/* Info for the state machine */
struct statemach_info {
  client_state_t state; /* The current state of the state machine */
  struct timeval start;
  DString req_msg;      /* initial outgoing request */
  DString can_msg;      /* CANCEL outgoing request */
  int req_trans_count, can_trans_count; /* Counts of number of transmissions
                                           performed so far */
  int req_trans_max, can_trans_max; /* Counts of number of transmissions to
                                       be performed */
  double delay;         /* seconds to wait until timeout */

  int client_return;    /* the return value to be sent if we enter
                           client_done */

  /* Read-only parts: */
  int sock;             /* socket to write requests to */
  int is_invite;        /* whether the initial request was an invite */
  int socktype;         /* the type of the branch request (SOCK_STREAM or
                           SOCK_DGRAM */
  branch_t *branch;     /* this branch */ 
};



/*
 * ClientStateMachine()
 * The core loop of the client state machine.  Get a message/event, then
 * dispatch appropriately with the state transition graph.
 */
static int ClientStateMachine(branch_t *branch, int s, int is_invite,
                              int socktype)
{
  statemach_info state;

  message_t *m;            /* next message from queue */
  client_event_t event;    /* next event to be processed */
  client_state_t oldstate; /* the state of the state machine before the 
                            * most recent transition */

  DStringInit(&state.req_msg);
  DStringInit(&state.can_msg);

  state.sock = s;
  state.is_invite = is_invite;
  state.socktype = socktype;
  state.branch = branch;

  CalculateMaxRetransmits(&state);

  /* Set up the initial state */
  state.state = client_retransmitting_request;
  oldstate = client_initial;
  event = event_unknown;

  /* Perform initial transmission */
  ClientBuildProxiedRequest(&state.req_msg, branch->orig,
                            branch, state.socktype);
  SendRequest(s, &state.req_msg, branch);
  state.req_trans_count++;

  /* The core loop of the state machine */
  while (state.state != client_done) {
    /* On state transitions and timeouts, recalculate the delay. */
    if (event == event_timeout || oldstate != state.state) {
      state.delay = CalculateNextDelay(&state);
      debug("ClientStateMachine", "r=%p, branch=%d, calc timeout %f\n",
            state.branch->parent, state.branch->branch, state.delay);
    }

    oldstate = state.state;

    event = WaitNextEvent(branch, &m, &state);

    if (event == event_unknown) {
      /* Debug message should have already been printed */
      state.client_return = 1;
      state.state = client_done;
      break;
    }

    DoClientStateTransitions(event, m, &state);

    if (m) {
      MessageDelete(m);
    }
  }

  DStringFree(&state.req_msg);
  DStringFree(&state.can_msg);

  return state.client_return;
} /* ClientStateMachine */


/*
 * WaitNextEvent()
 * Wait for the next message on the branch message queue, or a timeout,
 * then classify it.
 */
static client_event_t WaitNextEvent(branch_t *branch, message_t **m,
                                    statemach_info *state)
{
  struct timespec next; /* next timeout */
  int c;                /* pthread_cond_timedwait back value */
  request_t *r = branch->parent;

  /* Get the next message from the message queue */
  while ((*m = GetNextMessage(&branch->mlist)) == NULL) {
    /* No messages waiting; go into a timedwait */
    next = timespec_delay(state->start, state->delay);
    c = pthread_cond_timedwait(&branch->response_received, &r->mutex,
                               &next);
    if (c) {
      if (c == ETIME || c == ETIMEDOUT) {
        *m = NULL;
        return event_timeout;
      }
      else {
        error("WaitNextEvent",
              "error while waiting for client event: %s\n", strerror(c));
        *m = NULL;
        return event_unknown;
      }
    }
  }
    
  return ClientCategorizeMessage(branch, *m);
} /* WaitNextEvent */


/*
 * DoClientStateTransitions()
 * Dispatch the state transitions for the client state machine to each
 * individual state's event handlers.
 */
static void DoClientStateTransitions(client_event_t event, message_t *message,
                                     statemach_info *state)
{
  switch (state->state) {
  case client_retransmitting_request:
    DoRRTransitions(event, message, state);
    break;
  case client_awaiting_responses:
    DoARTransitions(event, message, state);
    break;
  case client_sent_final:
    DoSFTransitions(event, message, state);
    break;
  case client_retransmitting_cancel:
    DoRCTransitions(event, message, state);
    break;
  case client_awaiting_cancel_response:
    DoACRTransitions(event, message, state);
    break;
  case client_got_cancel_final:
    DoGCFTransitions(event, message, state);
    break;
  case client_initial:
  case client_done:
  default:
    debug("DoClientStateTransitions",
          "Entered client state machine in unknown or invalid state %d\n",
          state->state);
    state->client_return = -1;
    state->state = client_done;
    break;
  }
} /* DoClientStateTransitions */


/*
 * DoRRTransitions()
 * The state transitions when we're in the 'Retransmitting Request' state.
 */
static void DoRRTransitions(client_event_t event, message_t *message,
                            statemach_info *state)
{
  switch (event) {
  case event_request_1xx:
    SendMessageToParent(message, state->branch);
    state->state = client_awaiting_responses;
    break;
  case event_request_2xx:
    SendMessageToParent(message, state->branch);
    state->branch->sent_200 = 1;
    state->state = client_sent_final;
    break;
  case event_request_3xx_6xx:
    SendMessageToParent(message, state->branch);
    if (state->is_invite) {
      ClientACK(state->sock, message, state->branch, state->socktype);
    }
    state->state = client_sent_final;
    break;
  case event_cancel_1xx:
  case event_cancel_2xx_6xx:
  case event_ack:
    /* These shouldn't happen; someone must be sending us something
       spurious.  Ignore. */
    break;
  case event_shutdown:
    state->client_return = 1;
    state->state = client_done;
    break;
  case event_cancel:
    SendPseudoResponseToParent(487, state->branch);
    StartCancel(state);
    state->state = client_retransmitting_cancel;
    break;
  case event_sockerr:
    state->client_return = 0;
    state->state = client_done;
    break;
  case event_timeout:
    if (state->req_trans_count < state->req_trans_max) {
      /* Retransmit, and set the next delay. */
      SendRequest(state->sock, &state->req_msg, state->branch);
      state->req_trans_count++;
    } else {
      state->client_return = 0;
      state->state = client_done;
    }
    break;
  case event_unknown:
  default:
    error("DoRRTransitions", "Unknown event %d seen\n", event);
    state->client_return = 1;
    state->state = client_done;
    break;
  }
} /* DoRRTransitions */


/*
 * DoARTransitions()
 * The state transitions when we're in the 'Awaiting Responses' state.
 */
static void DoARTransitions(client_event_t event, message_t *message,
                             statemach_info *state)
{
  switch (event) {
  case event_request_1xx:
    SendMessageToParent(message, state->branch);
    /* No state transition */
    break;
  case event_request_2xx:
    SendMessageToParent(message, state->branch);
    state->branch->sent_200 = 1;
    state->state = client_sent_final;
    break;
  case event_request_3xx_6xx:
    SendMessageToParent(message, state->branch);
    if (state->is_invite) {
      ClientACK(state->sock, message, state->branch, state->socktype);
    }
    state->state = client_sent_final;
    break;
  case event_cancel_1xx:
  case event_cancel_2xx_6xx:
  case event_ack:
    /* These shouldn't happen; someone must be sending us something
       spurious.  Ignore. */
    break;
  case event_shutdown:
    state->client_return = 1;
    state->state = client_done;
    break;
  case event_cancel:
    SendPseudoResponseToParent(487, state->branch);
    StartCancel(state);
    state->state = client_retransmitting_cancel;
    break;
  case event_sockerr:
    SendPseudoResponseToParent(500, state->branch);
    state->state = client_sent_final;
    break;
  case event_timeout:
    if (!state->is_invite && state->req_trans_count < state->req_trans_max) {
      /* Retransmit. */
      SendRequest(state->sock, &state->req_msg, state->branch);
      state->req_trans_count++;
    } else {
      SendPseudoResponseToParent(408, state->branch);
      StartCancel(state);
      state->state = client_retransmitting_cancel;
    }
    break;
  case event_unknown:
  default:
    error("DoARTransitions", "Unknown event %d seen\n", event);
    state->client_return = 1;
    state->state = client_done;
    break;
  }
} /* DoARTransitions */


/*
 * DoSFTransitions()
 * The state transitions when we're in the 'Sent Final' state.
 */
static void DoSFTransitions(client_event_t event, message_t *message,
                             statemach_info *state)
{
  switch (event) {
  case event_request_1xx:
    /* ignore */
    break;
  case event_request_2xx:
    SendMessageToParent(message, state->branch);
    state->branch->sent_200 = 1;
    /* No state transition */
    break;
  case event_request_3xx_6xx:
    if (state->is_invite) {
      ClientACK(state->sock, message, state->branch, state->socktype);
    }
    /* no state transition */
    break;
  case event_cancel_1xx:
  case event_cancel_2xx_6xx:
    /* Responses to a dead cancel (probably); ignore */
    break;
  case event_ack:
    if (state->branch->sent_200) {
      ClientProxyACK(state->sock, message, state->branch, state->socktype);
    }
    /* no state transition */
    break;
  case event_shutdown:
    state->client_return = 1;
    state->state = client_done;
    break;
  case event_cancel:
    /* Too late to cancel; ignore */
    break;
  case event_sockerr:
    /* Transaction is finished; doesn't matter */
    break;
  case event_timeout:
    /* Just a timer ticking over -- ignore */
    break;
  case event_unknown:
  default:
    error("DoSFTransitions", "Unknown event %d seen\n", event);
    state->client_return = 1;
    state->state = client_done;
    break;
  }
} /* DoSFTransitions */


/*
 * DoRCTransitions()
 * The state transitions when we're in the 'Retransmitting Cancel' state.
 */
static void DoRCTransitions(client_event_t event, message_t *message,
                             statemach_info *state)
{
  switch (event) {
  case event_request_1xx:
    /* ignore */
    break;
  case event_request_2xx:
    SendMessageToParent(message, state->branch);
    state->branch->sent_200 = 1;
    state->state = client_sent_final;
    break;
  case event_request_3xx_6xx:
    if (state->is_invite) {
      ClientACK(state->sock, message, state->branch, state->socktype);
    }
    state->state = client_sent_final;
    break;
  case event_cancel_1xx:
    state->state = client_awaiting_cancel_response;
    break;
  case event_cancel_2xx_6xx:
    state->state = client_got_cancel_final;
    break;
  case event_ack:
    /* ignore */
    break;
  case event_shutdown:
    state->client_return = 1;
    state->state = client_done;
    break;
  case event_cancel:
    /* parent canceled us twice; ignore */
    break;
  case event_sockerr:
    state->state = client_got_cancel_final;
    break;
  case event_timeout:
    if (state->can_trans_count < state->can_trans_max) {
      /* Retransmit. */
      SendRequest(state->sock, &state->can_msg, state->branch);
      state->can_trans_count++;
    } else {
      state->state = client_got_cancel_final;
    }
    break;
  case event_unknown:
  default:
    error("DoRCTransitions", "Unknown event %d seen\n", event);
    state->client_return = 1;
    state->state = client_done;
    break;
  }
} /* DoRCTransitions */


/*
 * DoACRTransitions()
 * The state transitions when we're in the 'Awaiting Cancel Response' state.
 */
static void DoACRTransitions(client_event_t event, message_t *message,
                             statemach_info *state)
{
  switch (event) {
  case event_request_1xx:
    /* ignore */
    break;
  case event_request_2xx:
    SendMessageToParent(message, state->branch);
    state->branch->sent_200 = 1;
    state->state = client_sent_final;
    break;
  case event_request_3xx_6xx:
    if (state->is_invite) {
      ClientACK(state->sock, message, state->branch, state->socktype);
    }
    state->state = client_sent_final;
    break;
  case event_cancel_1xx:
    /* ignore */
    break;
  case event_cancel_2xx_6xx:
    state->state = client_got_cancel_final;
    break;
  case event_ack:
    /* ignore */
    break;
  case event_shutdown:
    state->client_return = 1;
    state->state = client_done;
    break;
  case event_cancel:
    /* parent canceled us twice; ignore */
    break;
  case event_sockerr:
    state->state = client_got_cancel_final;
    break;
  case event_timeout:
    if (state->can_trans_count < state->can_trans_max) {
      /* Retransmit. */
      SendRequest(state->sock, &state->can_msg, state->branch);
      state->can_trans_count++;
    } else {
      state->state = client_got_cancel_final;
    }
    break;
  case event_unknown:
  default:
    error("DoACRTransitions", "Unknown event %d seen\n", event);
    state->client_return = 1;
    state->state = client_done;
    break;
  }
} /* DoACRTransitions */


/*
 * DoGCFTransitions()
 * The state transitions when we're in the 'Got Cancel Final' state.
 */
static void DoGCFTransitions(client_event_t event, message_t *message,
                             statemach_info *state)
{
  switch (event) {
  case event_request_1xx:
    /* ignore */
    break;
  case event_request_2xx:
    SendMessageToParent(message, state->branch);
    state->branch->sent_200 = 1;
    state->state = client_sent_final;
    break;
  case event_request_3xx_6xx:
    if (state->is_invite) {
      ClientACK(state->sock, message, state->branch, state->socktype);
    }
    state->state = client_sent_final;
    break;
  case event_cancel_1xx:
    /* ignore */
    break;
  case event_cancel_2xx_6xx:
    /* ignore */
    break;
  case event_ack:
    /* ignore */
    break;
  case event_shutdown:
    state->client_return = 1;
    state->state = client_done;
    break;
  case event_cancel:
    /* parent canceled us twice; ignore */
    break;
  case event_sockerr:
    /* ignore */
    break;
  case event_timeout:
    /* Just a timer ticking over -- ignore */
    break;
  case event_unknown:
  default:
    error("DoGCFTransitions", "Unknown event %d seen\n", event);
    state->client_return = 1;
    state->state = client_done;
    break;
  }
} /* DoGCFTransitions */


/*
 * SendRequest()
 * Send the request in 'msg' to the socket 's'.  If there's an error,
 * enqueue a socket_write_error on branch's message queue.
 */
static void SendRequest(int s, DString *msg, branch_t *branch)
{
  int n;

  debug("SendRequest", "Proxying request to socket %d:\n%s\n",
        s, msg->string);
 
  if (SnmpProcessMessage) {
    sipsnmp_extra_data data;
    /* Read the method out of the msg string.
     * XXX this is a hack: should pass the method as a string to this
     * function.
     */
    sscanf(msg->string, "%19s", data.request_type);
    
    (*SnmpProcessMessage)(branch->parent, SNMP_TYPE_REQ_OUT, &data);
  }

  DRAW_MESSAGE_FLOW({
    char method[80];
    sscanf(msg->string, "%s", method);
    DrawMessageFlow("SIPAPI", "This", "Remote", method);
  });
 

  if ((n = send(s, msg->string, msg->length, 0)) < 0) {
    /* If there was a problem, put a message_socket_error in our own
       queue */
    if (!(ERRNO_NORMAL_FAILURE(errno))) {
      error("ClientSendToOneLocation", "Error sending proxied request\n");
    }
    else {
      debug("ClientSendToOneLocation",
	    "Normal error sending proxied request: %s (%d)\n",
	    strerror(errno), errno);
    }
    MessageTrivialEnqueue(&branch->mlist, message_socket_error,
                          nobody_handled);
  }
} /* SendRequest */


/*
 * CalculateMaxRetransmits()
 * Calculate the maximum number of times the request, and its CANCEL
 * message, should be retransmitted, based on the type of request and the
 * type of connection this is.  Also initialize the current transmission
 * count to zero.
 */
static void CalculateMaxRetransmits(statemach_info *state)
{
  state->req_trans_count = 0;
  state->can_trans_count = 0;

  if (state->socktype == SOCK_STREAM) {
    state->req_trans_max = 1;
    state->can_trans_max = 1;
  }
  else {
    if (state->is_invite) {
      state->req_trans_max = 7;
    }
    else {
      state->req_trans_max = 11;
    }
    state->can_trans_max = 11;
  }
} /* CalculateMaxRetransmits */


#define CHECK_MAX_OKAY(ct,mx) \
do { \
  if ((ct) > (mx)) { \
    debug("CalculateNextDelay", "Called when count > max (%d > %d)\n", \
          (ct), (mx)); \
    return timeout; \
  } \
  if ((ct) < 1) { \
    debug("CalculateNextDelay", "Called when count < 1 (%d)\n", \
          (ct)); \
    (ct)++; \
  } \
} while(0)


/*
 * CalculateNextDelay()
 * Calculate the delay until the next timeout for the current state.
 * This is called after the state machine runs, in two circumstances: after
 * a timeout has expired, and when the state machine has entered a new
 * state.
 * config.timeout is used to represent "infinite" times.  XXX: is this correct
 * when in "ringing" states?
 */
static double CalculateNextDelay(statemach_info *state)
{
  const double T1 = 0.5; /* s */
  const double T2 = 4.0; /* s */

  time_t expires;
  double timeout;

  int *count = NULL;
  int *max = NULL;

  gettimeofday(&state->start, NULL);

  expires = SIP_Expires(&state->branch->orig->p);
  if (expires == 0) {
    timeout = (double)config.timeout;
  }
  else if (expires == 1) {
    timeout = 0;
  }
  else {
    timeout = (double)(expires - state->start.tv_sec);
    if (timeout < 0) timeout = 0;
  }

  if (state->socktype == SOCK_STREAM) {
    return timeout;
  }

  if (state->state == client_retransmitting_cancel ||
      state->state == client_awaiting_cancel_response) {
    count = &state->can_trans_count;
    max = &state->can_trans_max;
  }
  else if (state->state == client_retransmitting_request ||
           state->state == client_awaiting_responses) {
    count = &state->req_trans_count;
    max = &state->req_trans_max;
  }

  /* Reliability for BYE, CANCEL, OPTIONS, REGISTER requests: (1) */
  if ((!state->is_invite && state->state == client_retransmitting_request)
      || (state->state == client_retransmitting_cancel)) {
    double delay;

    CHECK_MAX_OKAY(*count, *max);

    /* If the count is 1, delay until next retrans is 0.5 s, etc. */
    delay = (1 << (*count - 1)) * T1;
    /* Cap at T2. */
    if (delay > T2) {
      delay = T2;
    }
    if (delay > timeout) {
      *max = *count;
      delay = timeout;
    }
    return delay;
  }

  /* Reliability for BYE, CANCEL, OPTIONS, REGISTER requests: (2) */
  if ((!state->is_invite && state->state == client_awaiting_responses)
      || (state->state == client_awaiting_cancel_response)) {
    CHECK_MAX_OKAY(*count, *max);

    if (T2 > timeout) {
      *max = *count;
      return timeout;
    }

    return T2;
  }

  /* Reliability for INVITE requests */
  if (state->is_invite && state->state == client_retransmitting_request) {
    double delay;

    CHECK_MAX_OKAY(*count, *max);

    delay = (1 << (*count - 1)) * T1;

    if (delay > timeout) {
      *max = *count;
      delay = timeout;
    }
    return delay;
  }

  /* Remaining states should be:
   * client_awaiting_responses if is_invite
   * client_sent_final
   * client_got_cancel_final
   *
   * complain if it's anything else. */
  if (!((state->is_invite && state->state == client_awaiting_responses) ||
        state->state == client_sent_final ||
        state->state == client_got_cancel_final)) {
    debug("CalculateNextDelay", "In surprising state %d (is_invite = %d)\n",
          state->state, state->is_invite);
  }

  return timeout;
} /* CalculateNextDelay */


/*
 * StartCancel()
 * Called when transitioning to the "Cancel" part of the state machine.
 * Builds & sends the first CANCEL message.
 * Does not send a 487 to the parent, or transition to the
 * client_retransmitting_cancel state; these are the caller's
 * responsibility.  (XXX: would it be clearer if those happened here?)
 */
static void StartCancel(statemach_info *state)
{
  /* Perform initial CANCEL transmission */
  ClientCreateFollowupRequest(&state->can_msg, state->branch, 
                              &state->branch->orig->p, "CANCEL",
                              state->socktype);
  SendRequest(state->sock, &state->can_msg, state->branch);
  state->can_trans_count++;
} /* StartCancel */


/*
 * ClientCreateFollowupRequest()
 * Build a locally-generated followup request (ACK or CANCEL).
 * We include only the mandatory message headers.  To: From: and Call-ID: are 
 * copied verbatim from the parsed message in 'p'; CSeq is copied with the new
 * method tag; an appropriate local Via is inserted.  The body is empty.
 * 'p' should be the request being ACKed, for an ACK, or the original
 * request, for a CANCEL.  (XXX: tags received in 1xx responses?)
 */
static void ClientCreateFollowupRequest(DString *msg, branch_t *branch,
                                        parser_t *p, char *method,
                                        int socktype)
{
  headers_t hdrs; /* Followup request headers */ 
  char *version  = HTTP_Version(p);
  char *protocol = HTTP_Protocol(p);
  sip_cseq_t cseq = SIP_CSeq(p);
  DString uri_str;

  DStringInit(&uri_str);
  URI_SPrint(&branch->uri, &uri_str);

  hdrs.headers = NULL;
  hdrs.last = NULL;

  /* 
   * Create request message by copying all headers and body, 
   * adding the Via with the proxy in it 
   */
  DStringInit(msg);

  PrintRequestLine(msg, method, &uri_str, protocol, version);

  AddLocalVia(&hdrs, protocol, version, socktype,
              branch->parent->hash, branch->branch);

  HeaderCopyOneParsedHeader(&hdrs, p->headers, "To");
  HeaderCopyOneParsedHeader(&hdrs, p->headers, "From");
  HeaderCopyOneParsedHeader(&hdrs, p->headers, "Call-ID");
  HeaderPrintf(&hdrs, 1, "Cseq", "%lu %s", (long)cseq.seq, method);
  HeaderAppend(&hdrs, "Content-Length", "0"); /* No body */
  /* XXX Do we want any others -- Date, User-Agent, etc? */

  HeaderMessage(msg, &hdrs);
  DStringAppend(msg, "\r\n", 2);   /* Empty body. */

  free(version);
  free(protocol);
  DStringFree(&uri_str);
  HeaderFree(&hdrs);
} /* ClientCreateFollowRequest */


/*
 * PrintRequestLine()
 * Print (append) a request line into a DString.
 * Don't use DStringPrintf() because absurdly long URIs or methods can
 * overrun PRINTF_MAX.
 */
static void PrintRequestLine(DString *req_line,
                             const char *method, const DString *uri_str,
                             const char *protocol, const char *version)
{
  DStringAppend(req_line, method, -1);
  DStringAppend(req_line, " ", 1);
  DStringAppendD(req_line, uri_str);
  DStringAppend(req_line, " ", 1);
  DStringAppend(req_line, protocol, -1);
  DStringAppend(req_line, "/", 1);
  DStringAppend(req_line, version, -1);
  DStringAppend(req_line, "\r\n", 2);
} /* PrintRequestLine */


/*
 * AddLocalVia()
 * Add a Via header to hdrs referring to the local machine.
 */
static void AddLocalVia(headers_t *hdrs, char *protocol, char *version,
                        int socktype, unsigned long hash, int branchID)
{
  HeaderPrintf(hdrs, 0, "Via", "%s/%s/%s %s:%d; branch=%lu-%d",
               protocol, version,
               (socktype == SOCK_STREAM) ? "TCP" : "UDP",
               config.ProxyName, config.port,
               hash, branchID);
} /* AddLocalVia */


/*
 * CreateViaWithReceived()
 * Given the Via header element described in 'e', build a DString containing
 * the header with 'received' added, in 'munged_via'
 */
static void CreateViaWithReceived(DString *munged_via, parser_element_t *e,
                                  struct in_addr *received)
{
  parser_parameter_t *p;
  
  DStringInit(munged_via);
  DStringAppend(munged_via, e->value, -1);

  for (p = e->parameters; p != NULL; p = p->next) {
    if (strcasecmp(p->attribute, "received") == 0)
      continue;
    DStringAppend(munged_via, "; ", -1);
    DStringAppend(munged_via, p->attribute, -1);
    if (p->value) {
      DStringAppend(munged_via, "=", -1);
      if (p->value_type == PT_token) {
        DStringAppend(munged_via, p->value, -1);
      }
      else {
        DStringAppend(munged_via, "\"", -1);
        DStringAppendQuoted(munged_via, p->value, "\\\"");
        DStringAppend(munged_via, "\"", -1);
      }
    }
  }

  DStringAppend(munged_via, "; received=", -1);
  DStringAppend(munged_via, inet_ntoa(*received), -1);

  if (e->comment) {
    DStringAppend(munged_via, " ", -1);
    /* XXX Is this correct?  Or do we need to add back the parens? */
    DStringAppend(munged_via, e->comment, -1);
  }
}


/*
 * CopyAndMungeVias()
 * Copy Via headers from src to hdrs.  If parent_via->received !=
 * INADDR_ANY, add a "received" header to the top header.
 * TODO: Support 'Hide' header. 
 */
static void CopyAndMungeVias(headers_t *hdrs, parser_header_t *src,
                             sip_via_t *parent_via)
{
  if (parent_via->received.s_addr == INADDR_ANY) {
    HeaderCopyOneParsedHeader(hdrs, src, "Via");
  }
  else {
    parser_header_t *h;
    parser_element_t *e;
    DString munged_via;

    h = RFC822_Header(src, "Via");

    if (h == NULL || (e = h->elements) == NULL || e->element == NULL)
      return;

    CreateViaWithReceived(&munged_via, e, &parent_via->received);
    HeaderAppend(hdrs, "Via", DStringValue(&munged_via));
    DStringFree(&munged_via);

    for (e = e->next; e != NULL && e->element != NULL; e = e->next) {
      HeaderAppend(hdrs, "Via", e->element);
    }
  }
} /* CopyAndMungeVias */


/*
 * ClientACK()
 * Send ACK to downstream server.
 */
static void ClientACK(int s, message_t *m, branch_t *branch, int socktype) {
  DString msg; /* outgoing ACK */

  ClientCreateFollowupRequest(&msg, branch, &m->p, "ACK", socktype);

  SendRequest(s, &msg, branch);

  DStringFree(&msg);
} /* ClientACK */


/*
 * ClientProxyACK()
 * Send ACK received from upstream server on to downstream server.
 */
static void ClientProxyACK(int s, message_t *m, branch_t *branch,
                           int socktype)
{
  DString msg; /* outgoing ACK */

  DStringInit(&msg);

  ClientBuildProxiedRequest(&msg, m, branch, socktype);

  SendRequest(s, &msg, branch);

  DStringFree(&msg);
} /* ClientProxyACK */


/*
 * ClientBuildProxiedRequest()
 * Build a proxied request the client will send out.
 * Copy all headers and body, adding the Via with the proxy in it,
 * and update the URI appropriately.
 */
static void ClientBuildProxiedRequest(DString *msg, message_t *m,
                                      branch_t *branch, int socktype)
{
  char *method   = HTTP_Method(&m->p);
  char *version  = HTTP_Version(&m->p);
  char *protocol = HTTP_Protocol(&m->p);
  headers_t hdrs;       /* client request */
  headers_t tmp_hdrs;   /* Intermediate staging area */
  DString uri_str;

  hdrs.headers = NULL;
  hdrs.last = NULL;
  tmp_hdrs.headers = NULL;
  tmp_hdrs.last = NULL;

  DStringInit(&uri_str);
  URI_SPrint(&branch->uri, &uri_str);

  /* if the record_route var is set, and the method is invite,
   * call our AddRecordRoute function.
   */
  if (config.RecordRoute && !strcmp(method,"INVITE")) {
    if (!branch->parent->hash) 
      debug("ClientBuildProxiedRequest","Hash not found...\n");
    
    AddRecordRoute_Request(m, branch->parent->hash, branch->branch);
  }

  /* XXX trim off unnecessary parameters (e.g. "maddr") from URI. */
  PrintRequestLine(msg, method, &uri_str, protocol, version);

  AddLocalVia(&hdrs, protocol, version, socktype, branch->parent->hash,
              branch->branch);
  CopyAndMungeVias(&hdrs, m->p.headers, &branch->parent->via);

  /* If the request to be proxied is missing a Content-Length
   * header field, add one.  Needed for UDP -> TCP proxying. */
  if (RFC822_Header(m->p.headers, "Content-Length") == NULL) {
    HeaderPrintf(&hdrs, 0, "Content-Length", "%d", m->body.content_length);
  }

  /* This is a kludge; I should write a HeaderCopyAllParsedHeadersBut()
   * function for efficiency.*/
  HeaderCopyParsedHeaders(&tmp_hdrs, m->p.headers);
  HeaderCopyAllBut(&hdrs, &tmp_hdrs, "Via");

  DecrementMaxForwards(&hdrs);

  /* Delete first "Route" header */
  HeaderDeleteFirst(&hdrs, "Route");

  HeaderMessage(msg, &hdrs);
  DStringAppend(msg, "\r\n", 2);
  DStringAppend(msg, m->body.content, m->body.content_length);

  free(method);
  free(version);
  free(protocol);
  DStringFree(&uri_str);
  HeaderFree(&tmp_hdrs);
  HeaderFree(&hdrs);
} /* ClientBuildProxiedRequest */



/*
 * ClientGetLocations()
 * Get the list of locations to try to send this request to.
 */
server_t *ClientGetLocations(uri_t uri, int type)
{
  int transport = 0;        /* transport protocol to use */
  server_t *server_list;
  char *maddr;

  if (uri.host == NULL ||
     strcasecmp(uri.scheme, "sip") != 0) {
    server_list = (server_t *) calloc(1, sizeof(server_t));
    return server_list;
  }

  if (URI_Parameter(uri, "transport")) {
    if (strcasecmp(URI_Parameter(uri, "transport"), "udp") == 0) {
      transport = SOCK_DGRAM;
    }
    else if (strcasecmp(URI_Parameter(uri, "transport"), "tcp") == 0) {
      transport = SOCK_STREAM;
    }
    else if (strcasecmp(URI_Parameter(uri, "transport"), "ssl") == 0) {
      transport = SOCK_SSL;
    }
  }
  else {
    if ( type == SOCK_DGRAM || type == SOCK_STREAM || type == SOCK_SSL ) {
      transport = type;
    }
    /* else just ignore the value,
     * servers() takes care of transport = 0 */
  }

  /* if the maddr param is present, and it is NOT us, then next hop is given
   * by maddr. else, instead of looping, get the next hop from the host 
   * part (else). Useful for Route policy, though this loop should
   * never happen.
   */
  if ((maddr = URI_Parameter(uri, "maddr")) != NULL && 
      strcmp(maddr, config.ProxyName)) {
    debug("ClientGetLocations",
	  "calling servers(\"%s\", \"sip\", %d, %d)\n",
	  maddr, transport, uri.port ? uri.port : default_sip_port);

    server_list = servers(maddr, "sip", transport,
                          uri.port ? uri.port : default_sip_port);
  }
  else {
    /* Negative means use it as the default, but don't insist on it.
     * Positive means insist on it, so don't use SRV. */
    debug("ClientGetLocations",
	  "calling servers(\"%s\", \"sip\", %d, %d)\n",
	  uri.host, transport, uri.port ? uri.port : -default_sip_port);
    server_list = servers(uri.host, "sip", transport,
                          uri.port ? uri.port : -default_sip_port);
  }
  return server_list;
} /* ClientGetLocations */


/*
 * CreateClientSocket()
 * Create the socket for the child.  Returns the number of the socket
 * on success; negative on failure.  -1 is normal failure (no server there,
 * try the next one); other negative is unexpected failure.
 */
static int CreateClientSocket(server_t location) 
{
  int s;
  struct socket_info *si;
  int thr_err;          /* thread error value */
  pthread_t th;         /* TCP thread */
  struct sockaddr_in sin;
 
  debug("CreateClientSocket", "opening %s connection to %s:%d\n",
        location.type == SOCK_STREAM ? "TCP" : "UDP",
        inet_ntoa(location.sin.sin_addr), ntohs(location.sin.sin_port));

  if ((s = socket(PF_INET, location.type, 0)) < 0) {
    error("CreateClientSocket", "Error creating socket\n");
    return -2;
  }
  
  if (config.PrimaryIP != NULL) { /* Use this IP not INADDR_ANY */
    memset(&sin, 0, sizeof(sin));
    sin.sin_family = AF_INET;
    sin.sin_port = htons(0);    /* dont care */
    sin.sin_addr.s_addr = inet_addr(config.PrimaryIP);
	
    debug("CreateClientSocket", "Using %s as source address\n", 
     	  config.PrimaryIP);

    if (bind(s, (struct sockaddr*)&sin, sizeof(sin)) < 0) 
      error("CreateClientSocket", "bind fails...\n");
  }

  if (connect(s, (struct sockaddr *)&location.sin,
              sizeof(location.sin)) < 0) {
    if (ERRNO_NORMAL_FAILURE(errno)) {
      debug("CreateClientSocket",
	    "Normal error connecting socket: %s (%d)\n",
	    strerror(errno), errno);
      closesocket(s);
      return -1;
    }
    error("CreateClientSocket", "Error connecting socket\n");
    closesocket(s);
    return -2;
  } 
  /* Set up TCP receiver thread for socket. */
  if (location.type == SOCK_STREAM) {
    si = (struct socket_info *) malloc(sizeof(struct socket_info));
    if (si == NULL) {
      error("CreateClientSocket", "malloc\n");
      closesocket(s);
      return -2;
    }
    si->socket = s;
    si->sin = location.sin;
    
    debug("pthread_create", "ReceiveTCPrequests thread\n");
    if ((thr_err = pthread_create_detached(&th, NULL, ReceiveTCPRequests,
                                           (void *)si))) {
      error("CreateClientSocket",
            "Couldn't Create TCP thread: %s\n", strerror(thr_err));
      closesocket(s);
      return -2;
    }
  }
  return s;
} /* CreateClientSocket */


/*
 * CloseClientSocket()
 * Close the socket for the child.
 * XXX: this is only a sketch.  Need to figure out the right way to
 * close a TCP socket, since the ReceiveTCPRequests thread is sitting in a
 * read() on the socket.  This will deadlock on FreeBSD! 
 */
static void CloseClientSocket(int s)
{
  closesocket(s);
} /* CloseClientSocket */


/*
 * SendMessageToParent()
 * The specified message was "interesting", so send it to the parent.
 */
static void SendMessageToParent(message_t *m, branch_t *branch)
{
  debug("SendMessageToParent",
        "Request %p, branch %d: Sending response %p to parent\n",
        branch->parent, branch->branch, m);

  MessageEnqueue(&branch->parent->mqueue, message_real_message, &m->p,
                 &m->body, m->source, branch, branch_handled);

  pthread_cond_broadcast(&branch->parent->cond);

  /* Update branch "sent final" state if appropriate */
  /* XXX: having this here is a bit of an ugly hack */
  if (m->p.type == http_response &&
      m->p.status >= 200 && m->p.status < 700) {
    branch->sent_final = 1;
  }
} /* SendMessageToParent */


/*
 * SendPseudoRespnoseToParent()
 * Send a pseudo-response -- just a response status code, and nothing else
 * -- to the parent.  If the parent wants to use it as the real response,
 * it must generate the appropriate headers itself.
 */
static void SendPseudoResponseToParent(int status, branch_t *branch)
{
  debug("SendMessageToParent",
        "Request %p, branch %d: Sending pseudo-response %d to parent\n",
        branch->parent, branch->branch, status);

  MessageEnqueuePseudoResponse(&branch->parent->mqueue, branch,
                               branch_handled, status);

  pthread_cond_broadcast(&branch->parent->cond);

  /* Update branch "sent final" state if appropriate */
  /* XXX: duplicating this from above is even more of an ugly hack */
  if (status >= 200 && status < 700) {
    branch->sent_final = 1;
  }
} /* SendPseudoResponseToParent */


/* ClientCategorizeMessage()
 * Given a message from the parent, categorize it as a client event.
 */
static client_event_t ClientCategorizeMessage(branch_t *branch, message_t *m)
{
  switch (m->type) {
  case message_shutdown_client:
    return event_shutdown;
  case message_cancel_branch:
    return event_cancel;
  case message_socket_error:
    return event_sockerr;
  case message_real_message:
    switch (m->p.type) {
    case http_request:
      if (strcasecmp(m->p.method, "ACK") == 0) {
        return event_ack;
      }
      else {
        /* No other method types should show up here (including CANCEL; it's
         * represented by message_cancel_branch). */
        error("ClientCategorizeMessage",
              "Request %p, branch %d: Surprising request method %s",
              branch->parent, branch->branch, m->p.method); 
        return event_unknown;
      }
      break;
    case http_response:
      if (m->p.status < 100 || m->p.status >= 700) {
        error("ClientCategorizeMessage",
              "Request %p, branch %d: Surprising response status %d",
              branch->parent, branch->branch, m->p.status); 
        return event_unknown;
      }
      else {
        sip_cseq_t cseq;
        cseq = SIP_CSeq(&m->p);
        if (strcasecmp(cseq.method, branch->orig->p.method) == 0) {
          /* Response corresponds to method of original request */
          if (m->p.status < 200) {
            return event_request_1xx;
          }
          else if (m->p.status < 300) {
            return event_request_2xx;
          }
          else {
            return event_request_3xx_6xx;
          }
        }
        else if (strcasecmp(cseq.method, "CANCEL") == 0) {
          if (m->p.status < 200) {
            return event_cancel_1xx;
          }
          else {
            return event_cancel_2xx_6xx;
          }
        }
        else {
          error("ClientCategorizeMessage",
                "Request %p, branch %d: Surprising cseq method %s",
                branch->parent, branch->branch, cseq.method); 
          return event_unknown;
        }
      }
      break;
    default:
      error("ClientCategorizeMessage",
            "Request %p, branch %d: Surprising message parse type %d",
            branch->parent, branch->branch, m->p.type); 
      return event_unknown;
      break;
    }
  default:
    error("ClientCategorizeMessage",
          "Request %p, branch %d: Surprising message type %d",
          branch->parent, branch->branch, m->type); 
    return event_unknown;
    break;
  }
  error("ClientCategorizeMessage",
        "Request %p, branch %d: Message not categorized",
        branch->parent, branch->branch);
  return event_unknown;

} /* ClientCategorizeMessage */


/* External functions for querying about client state */
/* Abstracted so the internal structure of a branch can change. */

/* 
 * ClientSent200()
 * Has this client sent a 200 to its parent?
 */
int ClientSent200(branch_t *branch)
{
  return branch->sent_200;
} /* ClientSent200 */


/* External functions for querying about client state */
/* 
 * ClientSentFinal()
 * Has this client sent a final response to its parent?
 */
int ClientSentFinal(branch_t *branch)
{
  return branch->sent_final;
} /* ClientSentFinal */
