/*
http.c - HTTP header parser

Copyright 1998-1999 by Columbia University; all rights reserved 
*/

#ifdef __linux__
#define _GNU_SOURCE /* Needed for strptime declaration from <time.h> */
#endif

#include <stdio.h>
#include <stdlib.h>  /* atoi() */
#include <string.h>  /* strchr(), strncpy() */
#include <ctype.h>   /* isspace(), isalpha() */
#include <time.h>

#include "rfc822.h"
#include "http.h"
#include "base64.h"
#include "types.h"
#include "sysdep.h"
#include "strdupn.h"
#include "error.h"
#include "md5.h"

/*
* HTTP_Parse()
* Parse HTTP-like request into basic objects:  first line components,
* header.  Invoke RFC822_Parse to parse header lines.  Does not process
* body, since the length of the body is not known.  It depends on the
* protocol.  All protocols may have Content-Length, but in the absence
* of that field, HTTP assumes that the body extends to the end of the
* message (EOF), while RTSP assumes a content length of zero. Returns 0
* if ok, -1 if syntax error.
*/
int HTTP_Parse(char *msg, parser_t *pp)
{
  parser_t p;
  unsigned char *s;
  enum {P_space, P_protocol, P_version, P_status,
        P_reason, P_URI, P_header, P_done} 
       state = P_space, next = P_protocol;
  unsigned char *start = msg;          /* start of atom */
  int topline_complete = 0;

  p.type          = http_unknown_type;
  p.method        = NULL;              /* init the status line */
  p.URI_string    = NULL;
  memset(&p.URI, '\0', sizeof(uri_t));

  p.reason        = NULL;
  p.protocol      = NULL;
  p.version       = NULL;
  p.status        = -1;
  p.header_body   = NULL;
  p.headers       = NULL;              /* assume no headers */
  p.header_length = 0;

  for (s = msg; *s && state != P_done; s++) {        
    switch(state) {
    /* Space between atoms. */
    case P_space:
      if (!isspace((int)*s)) {
        state = next;
        start = s;
        s--;
      }
      break;

    /* protocol (RTSP/1.0 or SIP/2.0) or method (GET) */
    case P_protocol:
      if (*s =='/') {
        if (p.type == http_unknown_type)
          p.type = http_response;
        p.protocol = strdupn(start, s - start);
        state = P_version;
        start = s+1;
      }
      else if (isspace((int)*s)) {
        if (p.type != http_unknown_type) goto syntax;
        p.type = http_request;
        p.method = strdupn(start, s - start);
        state = P_space;
        next  = P_URI;
      }
      else if (!isalpha(*s) && !(*s == '-') && !(*s == '_')) goto syntax;
      break;
      
    /* Version */
    case P_version:
      if (isspace((int)*s)) {
        p.version = strdupn(start, s - start);
        state = P_space;
        next  = (p.type == http_response) ? P_status : P_header;
        s--;
      }
      else if (!(*s == '.') && !isdigit(*s)) goto syntax;
      break;
      
    /* status (number) */
    case P_status:
      if (isspace((int)*s)) {
        p.status  = atoi(start);
        state     = P_space;
        next      = P_reason;
      }
      else if (!isdigit(*s)) goto syntax;
      break;
      
    /* Reason phrase */
    case P_reason:
      if ((*s =='\n') || (*s == '\r')) {
        p.reason = strdupn(start, s - start);
        state = P_space; 
        next  = P_header;
      }
      break;
      
    /* request URI */
    case P_URI:
      if (isspace((int)*s)) {
        p.URI_string = strdupn(start, s - start);
        p.URI = URI_Parse(p.URI_string);
        state = P_space; 
        next  = P_protocol;
      }
      break;

    /* remainder of header, until blank line */
    case P_header:
      topline_complete = 1;
      p.header_body = s;
      if (RFC822_Parse(s, &p.header_length, &p.headers) < 0) {
        goto syntax;
      }
      p.header_length += (char *)s - msg;   /* account for request line */
      state = P_done;
      break;

    case P_done:
      break;
    }
  }  
  *pp = p;
  return 0;

syntax:
  if (!topline_complete) p.type = http_unknown_type;
  *pp = p;
  return -1;
} /* HTTP_Parse */


/*
 * HTTP_Duplicate()
 * Duplicate a parsed HTTP structure.
 */
void HTTP_Duplicate(parser_t *dst, parser_t *src)
{
  dst->type          = src->type;
  strassign(&dst->method, src->method);
  strassign(&dst->URI_string, src->URI_string);

  URI_Free(&dst->URI); /* Free the dst URI. */
  dst->URI = URI_Copy(src->URI);

  strassign(&dst->reason, src->reason);
  strassign(&dst->protocol, src->protocol);
  strassign(&dst->version, src->version);
  dst->status        = src->status;

  /* XXX Is this correct?  Nothing seems to use header_body outside
   * HTTP_Parse.  Why does it exist at all? */
  dst->header_body = NULL;

  RFC822_Duplicate(&dst->headers, src->headers);
  dst->header_length = src->header_length;
} /* HTTP_Duplicate */


/*
 * HTTP_Free()
 * Free resources allocated.
 */
void HTTP_Free(parser_t *p)
{
  if (p->method)      free(p->method);
  if (p->URI_string)  free(p->URI_string);
  if (p->reason)      free(p->reason);
  if (p->protocol)    free(p->protocol);
  if (p->version)     free(p->version);
  RFC822_Free(p->headers);
  URI_Free(&p->URI);
} /* HTTP_Free */


/* 
 * HTTP_Method
 * Return request method.
 */
char *HTTP_Method(parser_t *p)
{
  return (p && p->method) ? strdup(p->method) : NULL;
} /* HTTP_Method */


/*
 * HTTP_URI()
 * Return URI, as components.
 */
uri_t HTTP_URI(parser_t *p)
{
  return p->URI;
} /* HTTP_URI */


/*
 * HTTP_Status:
 * Return status code.
 */
int HTTP_Status(parser_t *p)
{
  return p ? p->status : 0;
} /* HTTP_Status */


/*
 * HTTP_Reason
 * Return reason phrase.
 */
char *HTTP_Reason(parser_t *p)
{
  return (p && p->reason) ? strdup(p->reason) : NULL;
} /* HTTP_Reason */


/*
 * HTTP_Protocol
 * Return protocol name.
 */
char *HTTP_Protocol(parser_t *p)
{
  return (p && p->protocol) ? strdup(p->protocol) : NULL;
} /* HTTP_Protocol */


/* 
 * HTTP_Version()
 * Return version number (major.minor form), NULL if none.
 */
char *HTTP_Version (parser_t *p)
{
  return (p && p->version) ? strdup(p->version) : NULL;
} /* HTTP_Version */


/* 
 * HTTP_ContentLength()
 * Content-Length: 12345
 * L: 12345
 * Return length of message body; 100,000,000 if no content length.
 */
int HTTP_ContentLength(parser_t *p)
{
  parser_header_t *h = RFC822_Header(p->headers, "Content-Length");
  int len = INFINITE_CONTENT_LENGTH;

  if (h) len = atoi(h->elements->value);
  return len;
} /* HTTP_ContentLength */


/*
 * HTTP_ContentType()
 * Content-Type: base/sub
 * C: base/sub
 * Return message body MIME type and subtype.
 */
http_content_type_t HTTP_ContentType(parser_t *p)
{
  parser_header_t *h = RFC822_Header(p->headers, "Content-Type");
  http_content_type_t c = {"", ""};
  char *slash, *v;
  
  if (h) {
    v = h->elements->value;
    if ((slash = strchr(v, '/'))) {
      if (slash - v < MAX_MIME_TYPE) {
        strncpy(c.type, v, slash - v);
        c.type[slash - v] = '\0';
      }
      if ((strlen(v) - (slash - v)) < MAX_MIME_TYPE) {
        strcpy(c.subtype, slash+1);
      }
    }
  }
  return c;
} /* HTTP_ContentType */


/*
 * Parse_HTTP_Date()
 * Parse an 'HTTP-date' (RFC 1123) date string:
 * "Wed, 15 Oct 1997 16:30:25 GMT".  This is the same as a SIP-date.
 * 
 * Note: we do *not* accept the deprecated date formats (asctime or RFC 850).
 * This is correct for SIP; wrong for HTTP; and arguably incorrect for RTSP.
 * The value 'p' must be a pointer to a valid 'struct tm'.
 * Returns the passed-in pointer on success, NULL on failure.
 *
 * ('p' may be altered even if NULL is returned; its value in this case is
 * undefined.)
 */
struct tm* Parse_HTTP_Date(char *str, struct tm* p)
{
  char *remains;

  if (str == NULL) {
    return NULL;
  }

  /* Some strptime implementations do not initialize all fields.
   * Initialize d here.
   */
  memset(p, 0, sizeof(struct tm));
  
  /* The %Z switch in strptime is unreliable -- in particular, our
   * replacement strptime() function doesn't support it. */
  if ((remains = strptime(str, "%a, %d %b %Y %T", p)) == NULL) {
    /* Failed */
    memset(p, 0, sizeof(struct tm));
    return NULL;
  }
  /* The rest of the string must be some string of whitespace,
   * followed by exactly the timezone "GMT", capitalized as such. */
  while (isspace((unsigned char)*remains)) {
    remains++;
  }
  if (strcmp(remains, "GMT") != 0) {
    /* Failed */
    memset(p, 0, sizeof(struct tm));
    return NULL;
  }

  /* (Should we perform any sanity checks on the returned struct tm?) */
  return p;
} /* Parse_HTTP_Date */


/*
 * HTTP_Date
 * Date: Wed, 15 Oct 1997 16:30:25 GMT
 * RFC822 date
 */
struct tm HTTP_Date(parser_t *p)
{
  parser_header_t *h = RFC822_Header(p->headers, "Date");
  struct tm d;

  if (!h || Parse_HTTP_Date(h->elements->value, &d) == NULL) {
    /* Doesn't exist, or parse failed.  Return an all-zeros (bogus) date. */
    memset(&d, 0, sizeof(d));
  }

  return d;
} /* HTTP_Date */


/*
 * HTTP_Server()
 * Server: Apache
 * Return server description, NULL.
 */
http_server_t HTTP_Server(parser_t *p)
{
  parser_header_t *h = RFC822_Header(p->headers, "Server");
  http_server_t s = {""};

  if (h) s.type = h->elements->value;
  return s;
} /* HTTP_Server */


/*
 * HTTP_UserAgent:
 * User-Agent: Netsoft
 * Return client description.
 */
http_agent_t HTTP_UserAgent(parser_t *p)
{
  parser_header_t *h = RFC822_Header(p->headers, "User-Agent");
  http_agent_t a = {""};

  if (h) a.type = h->elements->value;
  return a;
} /* HTTP_UserAgent */


/*
 * HTTP_AuthenticationMethod()
 * Given the string 'm' representing an authentication method, return
 * its http_authentication_method_t.
 */
http_authentication_method_t HTTP_AuthenticationMethod(const char *m)
{
  if (strcompare(m, "Basic", 1) == 0) {
    return auth_basic;
  }
  else if (strcompare(m, "Digest", 1) == 0) {
    return auth_digest;
  }
  else if (strcompare(m, "PGP", 1) == 0) {
    return auth_pgp;
  }
  else {
    return auth_unknown;
  }
  /* auth_none isn't handled here, since it's not a real authentication
     method. */
} /* HTTP_AuthenticationMethod */


/*
 * HTTP_AuthenticateParse()
 * Fill in the fields of auth, based on the content of h,
 * which should be a WWW-Authenticate or Proxy-Authenticate header.
 */
static void HTTP_AuthenticateParse(http_authenticate_t *auth,
                                    parser_header_t *h,
                                    parser_t *p) 
{
  parser_element_t *e;
  parser_parameter_t *prm;
  
  const char *paraValue;

  if (auth == NULL) return; /* Shouldn't happen */

  HTTP_AuthenticateInit(auth);

  if (h == NULL || (e = h->elements) == NULL) {
    /* No WWW-Authenticate header in message */
    return;
  }

  e = h->elements;

  auth->method = HTTP_AuthenticationMethod(e->value);

  /*
   * If the authentication method is "basic"
   */
  for (prm = e->parameters; prm != NULL; prm = prm->next) {
    paraValue = prm->value ? prm->value : "";
    
    if (strcompare(prm->attribute, "realm", 1) == 0) {
      strassign(&auth->realm, paraValue);
    }
    else if (strcompare(prm->attribute, "nonce", 1) == 0) {
      strassign(&auth->nonce, paraValue);
    }
    else if (strcompare(prm->attribute, "algorithm", 1) == 0) {
      strassign(&auth->algorithm, paraValue);
    }
    else if (strcompare(prm->attribute, "opaque", 1) == 0) {
      strassign(&auth->opaque, paraValue); 
    }   
    else if (strcompare(prm->attribute, "domain", 1) == 0) {
      strassign(&auth->domain, paraValue);
    } 
    else if (strcompare(prm->attribute, "stale", 1) == 0) {
      strassign(&auth->stale, paraValue);
    }
  }
  
  strassign(&auth->http_method, p->method);

  return;
} /*  HTTP_AuthenticateParse */

/*
* WWW-Authenticate: Basic realm="WallyWorld"
*/
http_authenticate_t HTTP_Authenticate(parser_t *p)
{
  parser_header_t *h = RFC822_Header(p->headers, "WWW-Authenticate");
  http_authenticate_t auth;

  memset(&auth, 0, sizeof(auth));

  HTTP_AuthenticateParse(&auth, h, p);

  return auth;
} /* HTTP_Authenticate */

/*
* Proxy-Authenticate: Basic realm="WallyWorld"
*/
http_authenticate_t HTTP_ProxyAuthenticate(parser_t *p)
{
  parser_header_t *h = RFC822_Header(p->headers, "Proxy-Authenticate");
  http_authenticate_t auth;

  memset(&auth, 0, sizeof(auth));

  HTTP_AuthenticateParse(&auth, h, p);

  return auth;
} /* HTTP_Authenticate */


/*
 * HTTP_AuthorizationParse()
 * Fill in the fiels of *auth with proper values found in *h
 * which should be a ProxyAuthorization or Authorization header.
 * p is used to get the HTTP method, and to determine the From header
 * (needed for pgp authentication).
 */
static void HTTP_AuthorizationParse(http_authorization_t *auth,
                                    parser_header_t *h,
                                    parser_t *p) 
{
  parser_element_t *e;
  parser_parameter_t *prm;
  
  char *lasts, *user_pass;
  const char *paraValue;

  if (auth == NULL) return; /* Shouldn't happen */

  HTTP_AuthorizationInit(auth);

  if (h == NULL || (e = h->elements) == NULL) {
    /* No (Proxy-)Authorization header in request */
    return;
  }

  e = h->elements;

  auth->method = HTTP_AuthenticationMethod(e->value);

  /*
   * If the authentication method is "basic"
   */
  if (auth->method == auth_basic) {
    strtok_r(e->element, " \t", &lasts);
    user_pass = base64_decode_string(strtok_r(NULL, " \t", &lasts));
    strassign(&auth->userid, strtok_r(user_pass, ":", &lasts));
    strassign(&auth->password, strtok_r(NULL, ":", &lasts));
    free(user_pass);

    strassign(&auth->http_method, p->method);
  } 
  else if (auth->method == auth_digest) {
    for (prm = e->parameters; prm != NULL; prm = prm->next) {
      paraValue = prm->value ? prm->value : "";
      
      if (strcompare(prm->attribute, "username", 1) == 0) {
        strassign(&auth->userid, paraValue);
      } 
      else if (strcompare(prm->attribute, "realm", 1) == 0) {
        strassign(&auth->realm, paraValue);
      }
      else if (strcompare(prm->attribute, "nonce", 1) == 0) {
        strassign(&auth->nonce, paraValue);
      }
      else if (strcompare(prm->attribute, "algorithm", 1) == 0) {
        strassign(&auth->algorithm, paraValue);
      }
      else if (strcompare(prm->attribute, "opaque", 1) == 0) {
        strassign(&auth->opaque, paraValue); 
      }   
      else if (strcompare(prm->attribute, "uri", 1) == 0) {
        strassign(&auth->digest_uri, paraValue);
      } 
      else if (strcompare(prm->attribute, "response", 1) == 0) {
        strassign(&auth->response_digest, paraValue);
      }
      else if (strcompare(prm->attribute, "digest", 1) == 0) {
        strassign(&auth->entity_digest, paraValue);
      }
    }

    strassign(&auth->http_method, p->method);
  }
#ifdef USE_PGP_SDK
  else if (auth->method == auth_pgp) == 0) {
    for (prm = e->parameters; prm != NULL; prm = prm->next) {
      paraValue = prm->value ? prm->value : "";

      if (strcompare(prm->attribute, "version", 1) == 0) {
        strassign(&auth->version, paraValue);
      }
      else if (strcompare(prm->attribute, "signed-by", 1) == 0) {  
        strassign(&auth->signed_by, paraValue); 
      }
      else if (strcompare(prm->attribute, "signature", 1) == 0) {  
        strassign(&auth->signature, paraValue); 
      }
      else if (strcompare(prm->attribute, "realm", 1) == 0) {
        strassign(&auth->realm, paraValue);
      }
      else if (strcompare(prm->attribute, "nonce", 1) == 0) {
        strassign(&auth->nonce, paraValue);
      }
      else if (strcompare(prm->attribute, "algorithm", 1) == 0) {
        strassign(&auth->algorithm, paraValue);
      }
      else if (strcompare(prm->attribute, "opaque", 1) == 0) {
        strassign(&auth->opaque, paraValue); 
      }   
    }
    {
      /* If the authentication method is pgp then fill in the userid
       * with the email found in the From header.
       */
      char *signerStr = NULL;
      DString msg;
      parser_header_t *h1;
      h1 = RFC822_Header(p->headers, "From");
      if (h1 && h1->elements && h1->elements->element) {
        signerStr = strstr(h1->elements->element, ":");
        signerStr++;
        DStringInit(&msg);
        DStringAppend(&msg, signerStr, -1);
        if (auth->userid) free(auth->userid);
        auth->userid = strdup(DStringTrim(&msg));
        DStringFree(&msg);
      }
    }
  }
#endif /* USE_PGP_SDK */
  /* Otherwise leave the structure unfilled, except for the method field. */
  return;
} /*  HTTP_AuthorizationParse */


/*
* Authorization: Basic QWxhZGRpbjpvcGVuIHNlc2FtZQ==
*/
http_authorization_t HTTP_Authorization(parser_t *p)
{
  http_authorization_t auth;
 
  parser_header_t *h = RFC822_Header(p->headers, "Authorization");

  memset(&auth, 0, sizeof(auth));

  HTTP_AuthorizationParse(&auth, h, p);

  return auth;
} /* HTTP_Authorization */


/*
* ProxyAuthorization:
*/
http_authorization_t HTTP_ProxyAuthorization(parser_t *p)
{
  http_authorization_t auth;
  
  parser_header_t *h = RFC822_Header(p->headers, "Proxy-Authorization");

  memset(&auth, 0, sizeof(auth));

  HTTP_AuthorizationParse(&auth, h, p);

  return auth;
} /* HTTP_ProxyAuthorization */


/* 
 * HTTP_AuthorizationInit()
 * Initialize the values of an http_authorization_t.
 * Calls HTTP_AuthorizationFree().
 */
void HTTP_AuthorizationInit(http_authorization_t *auth)
{
  if (auth == NULL) return;
  
  HTTP_AuthorizationFree(auth);  

  memset(auth, 0, sizeof(auth));
} /* HTTP_AuthorizationInit */


/* 
 * HTTP_AuthorizationFree()
 * Free the values of an http_authorization_t.
 * Calls HTTP_AuthorizationFree().
 */
void HTTP_AuthorizationFree(http_authorization_t *aa)
{
  free_valid(aa->http_method);
  free_valid(aa->userid);
  free_valid(aa->password);
  free_valid(aa->realm);
  free_valid(aa->nonce);
  free_valid(aa->digest_uri);
  free_valid(aa->response_digest);
  free_valid(aa->entity_digest);
  free_valid(aa->opaque);
  free_valid(aa->algorithm);
  free_valid(aa->version);
  free_valid(aa->signed_by);
  free_valid(aa->signature);
} /* HTTP_AuthorizationFree */ 


/* 
 * HTTP_AuthenticateInit()
 * Initialize the values of an http_authenticate_t.
 * Calls HTTP_AuthenticateFree().
 */
void HTTP_AuthenticateInit(http_authenticate_t *auth)
{
  if (auth == NULL) return;
  
  HTTP_AuthenticateFree(auth);  

  memset(auth, 0, sizeof(auth));
} /* HTTP_AuthenticateInit */


/* 
 * HTTP_AuthenticateFree()
 * Free the values of an http_authenticate_t.
 * Calls HTTP_AuthenticateFree().
 */
void HTTP_AuthenticateFree(http_authenticate_t *aa)
{
  free_valid(aa->http_method);
  free_valid(aa->realm);
  free_valid(aa->domain);
  free_valid(aa->nonce);
  free_valid(aa->opaque);
  free_valid(aa->stale);
  free_valid(aa->algorithm);
} /* HTTP_AuthenticateFree */ 


/*
 * HTTP_BuildAuthorizationHeader()
 * Build an Authorization header corresponding to 'auth', for the given
 * information.
 * It is the caller's responsibility to free the returned string.
 */
char *HTTP_BuildAuthorizationHeader(http_authenticate_t *auth,
                                    const char *userid,
                                    const char *password,
                                    const char *uri)
{
  DString h;
  DString v;
  char *ret = NULL;
  
  DStringInit(&h);

  if (auth->method == auth_basic) {
    char *base64_user_pass;
    /* Build the basic authorization header */
    DStringInit(&v);
    
    DStringPrintf(&v, "%s:%s", userid, password);
    base64_user_pass = base64_encode_string(DStringValue(&v));
    if ( base64_user_pass != NULL ) {
      DStringPrintf(&h, "Basic %s", base64_user_pass);
      free_valid(base64_user_pass);
    }
    
    DStringFree(&v);
  }
  else if (auth->method == auth_digest) {
    /* Build the digest authorization header */
    /* Create the response field */
    char *A1, *HA1;
    char *A2, *HA2;
    char *KD, *HKD;
    char *response = NULL;
    
    A1 = (char *)malloc(strlen(userid) + strlen(auth->realm) +
                        strlen(password) + 3);
    A2 = (char *)malloc(strlen(auth->http_method) + 
                        strlen(uri) + 2);
    
    sprintf(A1, "%s:%s:%s", userid, auth->realm, password);
    sprintf(A2, "%s:%s", auth->http_method, uri);
    HA1 = MD5L(A1);
    HA2 = MD5L(A2); 
    //#define DEBUG_DIGEST
         
#ifdef DEBUG_DIGEST
         debug("authenticateDigestResponse",
               "A1 = \"%s\", A2 = \"%s\", HA1 = \"%s\", HA2 = \"%s\"\n",
               A1, A2, HA1, HA2);
#endif
    
    free(A1);
    free(A2);
    
    KD = (char *)malloc(strlen(HA1) + strlen(HA2) + strlen(auth->nonce) + 3);
    sprintf(KD, "%s:%s:%s", HA1, auth->nonce, HA2);
    HKD = MD5L(KD);
    
#ifdef DEBUG_DIGEST
    debug("authenticateDigestResponse",
          "KD = \"%s\", HKD = \"%s\"\n", KD, HKD);
#endif
    
    free(HA1);
    free(HA2);
    free(KD);
    
    response = HKD;
    
    /* Build the header value */
    DStringPrintf(&h, "Digest username=\"%s\", realm=\"%s\", nonce=\"%s\", uri=\"%s\", response=\"%s\", opaque=\"%s\", algorithm=\"%s\"",
                  userid ? userid : "",
                  auth->realm ? auth->realm : "",
                  auth->nonce ? auth->nonce : "",
                  uri ? uri : "",
                  response ? response : "",
                  auth->opaque ? auth->opaque : "",
                  auth->algorithm ? auth->algorithm : "");
    
    free(HKD);
  }
  else if (auth->method != auth_none) {
    debug("HTTP_BuildAuthorizationHeader", "unknown auth method %d\n",
          auth->method);
  }

  /* Return the string containing the authorization header */
  ret = DStringValue(&h);
  if ( ret == NULL || *ret == '\0') {
    DStringFree(&h);
    ret = NULL;
  }
  return ret;
} /* HTTP_BuildAuthorizationHeader */
