/*
log.c

Create and add entries to one or more server log file(s).  Any log file
can be a file, a process (pipe) or syslog.  The log file format can be
defined by the user when creating the log file, with each file possibly
having a different format. The log format is ignored for SQL output.

Syslog file 'names' have the format '@facility.level'.
SQL database files are identified by sql://user:password@hostname/database.

Copyright 1998-2001 by Columbia University; all rights reserved 
*/

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>     /* getenv() */
#include <time.h>       /* time() */
#include <unistd.h>     /* getpid() */
#include <sys/types.h>
#include <sys/wait.h>   /* WEXITSTATUS, etc. */
#include <sys/socket.h>
#include <string.h>
#include <signal.h>     /* SIG* */
#include <netinet/in.h>
#include <arpa/inet.h>  /* inet_ntoa() */
#include <pthread.h>
#include <syslog.h>     /* syslog() */
#include "log.h"
#include "error.h"
#include "dstring.h"
#include "request-common.h"
#include "request-sip.h"
#include "strdupn.h"
#include "sysdep.h"


extern double gettimeofdayf(void);

typedef enum {
  Log_Invalid, Log_Literal, 
  Log_ProcessId, Log_Env, Log_Time, 
  Log_RemoteUser, Log_RemoteHost, Log_RemoteIP,
  Log_RequestLine, Log_RequestURI, Log_RequestHeader, 
  Log_ResponseStatus, Log_ResponseHeader, Log_ResponseTime
} log_t;

typedef enum {
  LogDest_File, LogDest_Pipe, LogDest_Syslog, LogDest_SQL
} log_destination_t;

typedef struct {
  log_t type;
  char *value;
} log_item_t;

/* for each log file or log program */
static struct logger_t {
  log_destination_t log_dest;  /* type of log destination (file, pipe, syslog) */
  FILE *file;                  /* log file pointer */
  char *name;                  /* name of log file or URL */
  MYSQL mysql;                 /* database pointer */
  int priority;                /* syslog priority value */
  int items;                   /* number of log items */
  log_item_t *itemlist;        /* list of log items */
  int methods;                 /* number of methods in methodlist */
  char **methodlist;           /* list of methods to be logged */
  int status;                  /* number of status entries */
  short *statuslist;           /* list of status codes to be logged */
} *logger = NULL;

static int logger_count = 0;
/* One lock for all logging; no need to be finer-grained. */
static pthread_mutex_t log_mutex = PTHREAD_MUTEX_INITIALIZER;

/* 
* new_item()
* Add to list of log elements.
*/
static void new_item(struct logger_t *logger, log_t type,
                     const char *value, int size)
{
  /* Ignore empty literals. */
  if ((type == Log_Literal && size == 0) || type == Log_Invalid) return;

  logger->itemlist = realloc(logger->itemlist, 
    (logger->items+1) * sizeof(log_item_t)); 
  logger->itemlist[logger->items].type  = type;
  logger->itemlist[logger->items].value = (value && size) ? 
    strdupn(value, size) : NULL;
  logger->items++;
} /* new_item */


/*
* parse_item()
* Parse a single log item (text after a %).
* %a  remote IP-address
* %e  environment variable
* %h  remote host name (currently also the IP address)
* %i  request header
* %o  the output (response) header with name header
* %P  process id
* %r  request line
* %s  status of response
* %t  current time
* %T  response time
* %U  request URI
* %u  remote user if authentication is used, "-" otherwise. 
*/
static const char *parse_item(struct logger_t *logger, const char *s)
{
  enum {initial, braces, directive} state = initial;
  const char *value = NULL;   /* value in {} */
  int size = 0;
  struct {
    log_t c[2];
  } map[] = {
    /* upper, lower */
    {{Log_Invalid,      Log_RemoteIP}},       /* 'a' */
    {{Log_Invalid,      Log_Invalid}},        /* 'b' */
    {{Log_Invalid,      Log_Invalid}},        /* 'c' */
    {{Log_Invalid,      Log_Invalid}},        /* 'd' */
    {{Log_Invalid,      Log_Env}},            /* 'e' */
    {{Log_Invalid,      Log_Invalid}},        /* 'f' */
    {{Log_Invalid,      Log_Invalid}},        /* 'g' */
    {{Log_Invalid,      Log_RemoteHost}},     /* 'h' */
    {{Log_Invalid,      Log_RequestHeader}},  /* 'i' */
    {{Log_Invalid,      Log_Invalid}},        /* 'j' */
    {{Log_Invalid,      Log_Invalid}},        /* 'k' */
    {{Log_Invalid,      Log_Invalid}},        /* 'l' */
    {{Log_Invalid,      Log_Invalid}},        /* 'm' */
    {{Log_Invalid,      Log_Invalid}},        /* 'n' */
    {{Log_Invalid,      Log_ResponseHeader}}, /* 'o' */
    {{Log_Invalid,      Log_ProcessId}},      /* 'p' */
    {{Log_Invalid,      Log_Invalid}},        /* 'q' */
    {{Log_Invalid,      Log_RequestLine}},    /* 'r' */
    {{Log_Invalid,      Log_ResponseStatus}}, /* 's' */
    {{Log_ResponseTime, Log_Time}},           /* 't' */
    {{Log_RequestURI,   Log_RemoteUser}},     /* 'u' */
    {{Log_Invalid,      Log_Invalid}},        /* 'v' */
    {{Log_Invalid,      Log_Invalid}},        /* 'w' */
    {{Log_Invalid,      Log_Invalid}},        /* 'x' */
    {{Log_Invalid,      Log_Invalid}},        /* 'y' */
    {{Log_Invalid,      Log_Invalid}},        /* 'z' */
  };

  for (; *s; s++) {
    switch (state) {
    case initial:
      if (*s == '{') {
        state = braces;
        value = s+1;
      }
      else {
        s--;
        state = directive;
      }
      break;
    case braces:
      if (*s == '}') {
        state = directive;
      }
      else size++;
      break;
    case directive:
      new_item(logger,
               map[(int)toupper((int)*s) - 'A'].c[islower((int)*s) != 0],
        value, size);
      goto done;
    }
  }
done:
  return s;
} /* parse_item */


/*
 * AddLogInternal()
 * Create a new request logger, but do not open it.  (Used so that we
 * can chdir() between the time the config file is written, and the time
 * the files are opened/pipes are executed.) The 'when' argument is a
 * list of response codes (three digits), response code classes (single
 * digit) or method names that cause this log item to be invoked.
 * Return a pointer to the new logger, or NULL on failure. 
 */
static struct logger_t *AddLogInternal(const char *filename,
                                       const char *format,
                                       const char *when)
{
  struct logger_t *l; /* current logger */
  const char *s;      /* character pointer */
  char c;             /* current character */
  DString literal;

  static struct {
    char *label;
    int facility;
  } facility[] = {
    {"kern",    LOG_KERN},
    {"user",    LOG_USER},
    {"mail",    LOG_MAIL},
    {"daemon",  LOG_DAEMON},
    {"auth",    LOG_AUTH},
    {"lpr",     LOG_LPR},
    {"news",    LOG_NEWS},
    {"uucp",    LOG_UUCP},
    {"cron",    LOG_CRON},
    {"local0",  LOG_LOCAL0},
    {"local1",  LOG_LOCAL1},
    {"local2",  LOG_LOCAL2},
    {"local3",  LOG_LOCAL3},
    {"local4",  LOG_LOCAL4},
    {"local5",  LOG_LOCAL5},
    {"local6",  LOG_LOCAL6},
    {"local7",  LOG_LOCAL7},
    {0, 0}
  };
  static struct {
    char *label;
    int severity;
  } severity[] = {
    {"emergency",LOG_EMERG},
    {"alert",    LOG_ALERT},
    {"critical", LOG_CRIT},
    {"error",    LOG_ERR},
    {"warning",  LOG_WARNING},
    {"notice",   LOG_NOTICE},
    {"info",     LOG_INFO},
    {"debug",    LOG_DEBUG},
    {0, 0}
  };

  if (!filename) {
    error("LogInit", "Invalid file name");
    return NULL;
  }
  logger = realloc(logger, (logger_count+1) * sizeof(struct logger_t));
  l = &logger[logger_count];

  if (filename[0] == '|') {
    l->log_dest = LogDest_Pipe;
  }
  else if (filename[0] == '@') {
    int i;
    char *severity_string, *facility_string;
    char *lasts;

    l->log_dest = LogDest_Syslog;
    l->priority = 0;

    /* facility */
    facility_string = strtok_r((char *)&filename[1], ".", &lasts);
    if (!facility_string) {
      error("LogInit", "Invalid facility field for syslog.");
      return NULL;
    }
    for (i = 0; facility[i].label; i++) {
      if (!strcmp(facility[i].label,facility_string)) {
        l->priority |= facility[i].facility;
        break;
      }
    }
    if (!facility[i].label) {
      error("LogInit", "Invalid facility name %s.", facility_string);
      return NULL;
    }

    /* severity */
    severity_string = strtok_r((char *)NULL, ".", &lasts);
    if (!severity_string) {
      error("LogInit", "Invalid severity field for syslog.");
      return NULL;
    }
    for (i = 0; severity[i].label; i++) {
      if (!strcmp(severity[i].label,severity_string)) {
        l->priority |= severity[i].severity;
        break;
      }
    }
    if (!severity[i].label) {
      error("LogInit", "Invalid severity name %s.", severity_string);
      return NULL;
    }
  }
  else if (strncmp(filename, "sql://", 6) == 0) {
    l->log_dest = LogDest_SQL;
  }
  else {
    l->log_dest = LogDest_File;
  }

  /* Provide default if no format specified. */
  if (!format || format[0] == '\0') {
    format = "%h %u %t \"%r\" %s";
  }

  /* Extend list of loggers. */
  l->name     = strdup(filename);
  l->items    = 0;
  l->itemlist = NULL;
  l->file     = NULL;

  /* Create logger entry. */
  l->name       = strdup(filename);
  l->items      = 0;
  l->itemlist   = NULL;
  l->file       = NULL;
  l->methods    = 0;
  l->methodlist = NULL;
  l->status     = 0;
  l->statuslist = NULL;

  /* Parse format and create itemlist. */
  DStringInit(&literal);
  for (s = format; *s; s++) {
    c = *s;
    if (c == '%') {
      new_item(l, Log_Literal, DStringValue(&literal), DStringLength(&literal));
      DStringTrunc(&literal);
      s = parse_item(l, s+1);
    }
    else {
      if (c == '\\') {
        c = *(++s);
        if      (c == 'n') c = '\n';
        else if (c == 't') c = '\t';
      }
      DStringAppend(&literal, &c, 1);
    }
  }

  new_item(l, Log_Literal, DStringValue(&literal), DStringLength(&literal));
  DStringFree(&literal);

  /* if 'when' is not NULL, add condition list */
  l->status  = 0;
  l->methods = 0;
  if (when != NULL) {
    char *w = strdup(when), *s, *lasts;

    s = strtok_r(w, ",", &lasts);
    while (s) {
      if (isdigit((int)s[0])) {
        l->statuslist = realloc(l->statuslist, (l->status+1) * sizeof(l->statuslist[0])); 
        l->statuslist[l->status++] = atoi(s); 
      }
      else {
        l->methodlist = realloc(l->methodlist, (l->methods+1) * sizeof(l->methodlist[0])); 
        l->methodlist[l->methods++] = strdup(s);
      }
      s = strtok_r(NULL, ",", &lasts);
    }
    free(w);
  }
  
  logger_count++;

  return l;
} /* AddLogInternal */


/*
 * OpenLogInternal()
 * Open an already-created logger (if it isn't already open).
 */
static void OpenLogInternal(struct logger_t *l)
{
  FILE *file = NULL;  /* log file to be used */
  uri_t u;

  switch (l->log_dest) {
  case LogDest_Pipe:
    if (!(file = popen(&l->name[1], "w"))) {
      error("LogInit", "Could not open log pipe %s", l->name);
      return;
    }
    break;

  case LogDest_Syslog:
    file = NULL;
    break;

  case LogDest_SQL:
    u = URI_Parse((const char *)l->name);
    mysql_init(&l->mysql);
    if (!(mysql_real_connect(&l->mysql, u.host, u.user, u.password, u.path, 0, NULL, 0))) {
      error("LogInit", "Could not open log SQL database %s: %s",
        l->name, mysql_error(&l->mysql));
      return;
    }      
    URI_Free(&u);
    break;

  case LogDest_File:
    if (!(file = fopen(l->name, "a+"))) {
      error("LogInit", "Could not open log file %s", l->name);
      return;
    }
    setvbuf(file, NULL, _IOLBF, 0);
    break;
  }

  l->file = file;
} /* OpenLogInternal */


/*
 * LogInit()
 * Create and open a new request logger. Same as AddLog(), but also
 * opens the log destination.
 */
void LogInit(const char *filename, const char *format, const char *when)
{
  struct logger_t *l;

  pthread_mutex_lock(&log_mutex);

  l = AddLogInternal(filename, format, when);
  if (l != NULL) {
    OpenLogInternal(l);
  }

  pthread_mutex_unlock(&log_mutex);
} /* LogInit */


/*
 * AddLog()
 * Add a new log.  Wrapper around AddLogInternal() which doesn't
 * return internal data.
 */
void AddLog(const char *filename, const char *format, const char *when)
{
  pthread_mutex_lock(&log_mutex);
  AddLogInternal(filename, format, when);
  pthread_mutex_unlock(&log_mutex);

  return;
} /* AddLog */


/*
 * OpenLogs()
 * Open all not-yet-opened logs.
 */
void OpenLogs(void)
{
  struct logger_t *l;
  int n;

  pthread_mutex_lock(&log_mutex);

  for (n = 0; n < logger_count; n++) {
    l = &logger[n];
    if (l->file == NULL) {
      OpenLogInternal(l);
    }
  }

  pthread_mutex_unlock(&log_mutex);

  return;
} /* OpenLogs */


/*
 * ReportAbnormalPipeExit()
 * If a pipe exits with an abnormal status code, report it.
 */
void ReportAbnormalPipeExit(int n, char *logname, int status)
{
  if (status == -1) {
    error("LogClose", "Could not close log process %d: \"%s\"\n",
          n, logname);
  }
  else if (WIFSIGNALED(status)) {
    int signal = WTERMSIG(status);
    /* Signals caught by exitThread() are ok */
    if (signal != SIGTERM && signal != SIGINT && signal != SIGQUIT) {
      error("LogClose",
            "Log process %d (%s) died on signal %d%s\n",
            n, logname,
            WTERMSIG(status),
            WCOREDUMP(status) ? " (core dumped)" : "");
    }
  }
  else if (!WIFEXITED(status)) {
    error("LogClose",
          "Log process %d (%s) did not exit normally, status = 0x%x\n", 
          n, logname, status);
  }
  else if (WEXITSTATUS(status) != 0) {
    error("LogClose",
          "Log process %d (%s) exited with non-zero exit code %d\n",
          n, logname, WEXITSTATUS(status));
  }
} /* ReportAbnormalPipeExit */


/*
* LogClose()
* Close log files; free storage.
*/
void LogClose(void)
{
  int k, n, status;

  pthread_mutex_lock(&log_mutex);
  for (n = 0; n < logger_count; n++) {
    for (k = 0; k < logger[n].items; k++) {
      free(logger[n].itemlist[k].value);
    }
    free(logger[n].itemlist);
    for (k = 0; k < logger[n].methods; k++) {
      free(logger[n].methodlist[k]);
    }
    free(logger[n].methodlist);
    free(logger[n].statuslist);
    if (logger[n].file != NULL) {
      if (logger[n].log_dest == LogDest_Pipe) {
        if ((status = pclose(logger[n].file)) != 0) {
          ReportAbnormalPipeExit(n, logger[n].name, status);
        }
      }
      else if (logger[n].log_dest == LogDest_File) {
        if (fclose(logger[n].file) != 0) {
          error("LogClose", "Could not close log file %d (%s)\n",
                n, logger[n].name);
        }
      }
      else if (logger[n].log_dest == LogDest_SQL) {
        mysql_close(&logger[n].mysql);
      }
    }
    free(logger[n].name);
  }
  free(logger);
  logger = NULL;
  logger_count = 0;
  pthread_mutex_unlock(&log_mutex);
} /* LogClose */


/*
* Create a string log entry.
*/
static DString log_string(struct logger_t *l, request_t *sr)
{
  request_common_t *r = (request_common_t *)sr;
  DString entry;
  int i;
  char *value;
  parser_t *p = &r->p; 
  parser_element_t *e;
  parser_header_t *h;
  void *hp;
  extern int gmtoff(void);

  DStringInit(&entry);
  for (i = 0; i < l->items; i++) {
    value = l->itemlist[i].value;
    switch(l->itemlist[i].type) {
      case Log_Env:
        DStringAppend(&entry, getenv(value), -1);
        break;

      case Log_Invalid:
        break;

      case Log_Literal:
        DStringAppend(&entry, value, -1);
        break;

      case Log_ProcessId:
        DStringPrintf(&entry, "%ld", (long)getpid());
        break;

      case Log_RemoteHost:
      case Log_RemoteIP:
        DStringAppend(&entry, inet_ntoa(r->sin.tcp_s.sin_addr), -1);
        break;

      case Log_RemoteUser:
        DStringAppend(&entry, r->authorization.userid ? r->authorization.userid : "-", -1);
        break;

      case Log_RequestHeader:
        h = RFC822_Header(p->headers, value);
        if (h) {
          for (e = h->elements; e; e = e->next) {
            DStringAppend(&entry, e->value, -1); 
            if (e->next) DStringAppend(&entry, ",", 1);
          }
        }
        break;

      case Log_RequestLine:
        DStringPrintf(&entry, "%s %s %s/%s", 
          p->method, p->URI_string, p->protocol, p->version);
        break;

      case Log_RequestURI:
        DStringAppend(&entry, p->URI_string, -1);
        break;

      case Log_ResponseHeader:
        hp = NULL;
        do {
          DStringAppend(&entry, HeaderGet(&r->headers, value, 1, &hp), -1);
        } while (hp);
        break;

      case Log_ResponseStatus:
        DStringPrintf(&entry, "%d", r->status);
        break;

      case Log_ResponseTime: /* real time taken from request to final resp. */
        DStringPrintf(&entry, "%.2f", gettimeofdayf() - r->time_received);
        break;

      case Log_Time:
        DStringCftime(&entry, value ? value : "[%d/%b/%Y:%H:%M:%S", time(0), /*localtime*/ 1);
        /* append time zone for CLF */
        if (!value) {
          int offset = gmtoff(); /* minutes */
          DStringPrintf(&entry, " %c%.2d%.2d]", offset < 0 ? '-' : '+',
          abs(offset)/60, abs(offset)%60);
        }
        break;
    }
  } /* for items */
  return entry;
} /* log_string */


/*
* Create an SQL log entry.
*/
static void log_sql(struct logger_t *l, request_t *sr)
{
  request_common_t *r = (request_common_t *)sr;
  DString cmd;
  char *sip_from = URI_AsString(sr->from.uri);
  char *sip_to   = URI_AsString(sr->to.uri);

  DStringInit(&cmd);
  DStringPrintf(&cmd, "INSERT INTO requestlog \
    SET request_time=NOW(), method='%s', sip_status='%d', \
    request_url='%s', sip_callid='%s', sip_from='%s', sip_to='%s'",
    r->p.method, r->status, r->p.URI_string, sr->callID, sip_from, sip_to);
 
  if((mysql_query(&l->mysql, (const char*)DStringValue(&cmd))) != 0) {
    error("Log", "SQL log entry %s failed: %s\n", DStringValue(&cmd), mysql_error(&l->mysql));
  }
  free(sip_from);
  free(sip_to);

  DStringFree(&cmd);
} /* log_sql */


/*
* Log()
* Log matching request to all loggers.
*/
void Log(request_t *sr)
{
  request_common_t *r = (request_common_t *)sr;
  int i, n;
  struct logger_t *l;
  int match = -1;

  pthread_mutex_lock(&log_mutex);
  for (n = 0; n < logger_count; n++) {
    l = &logger[n];
    if ((l->log_dest == LogDest_File || l->log_dest == LogDest_Pipe) && l->file == NULL) 
      continue;

    /* check if method or status matches one entry in list */ 
    if (l->methods == 0 && l->status == 0) {
      match = 1;
    } else if (l->methods > 0) {
      match = 0;
      for (i = 0; i < l->methods; i++) {
        if (!strcmp(l->methodlist[i], r->p.method)) {
          match = 1;
          break;
        }
      }
    }

    if (match < 0 && l->status > 0) {
      match = 0;
      for (i = 0; i < l->status; i++) {
        if (r->status == l->statuslist[i] || r->status/100 == l->statuslist[i]) {
          match = 1;
          break;
        }
      }      
    }
    if (match) {
      if (l->log_dest == LogDest_SQL) {
        log_sql(l, sr);
      } else {
        DString entry = log_string(l, sr);

        if (l->log_dest == LogDest_File || l->log_dest == LogDest_Pipe) {
          fputs(DStringValue(&entry), l->file);
          fputs("\n", l->file);
        } else if (l->log_dest == LogDest_Syslog) {
          syslog(l->priority, DStringValue(&entry));
        }
        DStringFree(&entry);
      }
    }
  } /* for logger */
  pthread_mutex_unlock(&log_mutex);
} /* Log */
