/*
 * presence.c - handle SUBSCRIBE request, send NOTIFY request
  
  Copyright 1998-2000 by Columbia University; all rights reserved 
*/

#include <time.h>
#include <stdio.h>
#include <stdlib.h>   /* qsort() */
#include <fcntl.h>
#include <string.h>   /* strdup() */
#include <sys/types.h>
#include <regex.h>    /* regexec() */
#include <pthread.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "rfc822.h"
#include "dstring.h"
#include "uri.h"
#include "register.h"
#include "request-sip.h"
#include "config.h"
#include "strdupn.h"  /* free_valid() */
#include "http.h"
#include "log.h"
#include "sysdep.h"
#include "authenticateSIP.h"
#include "policy.h"
#include "message.h"
#include "dbapi.h" /* MYSQL Api's */
#include "method.h"
#include "proxy.h"
#include "presence.h"
#include "host2ip.h"
#include "pthread-util.h"
#include "thread_pool.h"
#include "dbaccess.h"
#include "canonicalize.h"
#include "querydb.h"
#include "rr_support.h"
#include "sipd-support.h"

/* Local subscribe structure */

struct subscribe_info {
  int subscription_added;
  registration_t *reg;
};

/* Subscribe Policy functions */
static int subscribe_init(request_t *r, registration_t *reg, int reg_status);
static int subscribe_handle_request(request_t *r, message_t *m);
static int subscribe_handle_response(request_t *r, message_t *m, branch_t *b);
static int subscribe_timeout_expired(request_t *r);
static int subscribe_cleanup(request_t *r);

struct notify_info {
  int received_final; /* whether we received a "final" response to request */
};

/* Notify Policy functions */
static int notify_init(request_t *r, registration_t *reg, int reg_status);
static int notify_handle_request(request_t *r, message_t *m);
static int notify_handle_response(request_t *r, message_t *m, branch_t *b);
static int notify_timeout_expired(request_t *r);
static int notify_cleanup(request_t *r);

/* local functions */
static void GetSubscribeContacts(registration_t* reg);
static request_t* CreateNotifyRequest(sub_table_entry* info, 
                                      char * status);


/* The structure containing the subscribe "default" policy */
policy_t subscribe_policy =
{
  "subscribe",
  &subscribe_init,
  &subscribe_handle_request,
  &subscribe_handle_response,
  &subscribe_timeout_expired,
  &subscribe_cleanup
};

/* The structure containing the notify policy */
policy_t notify_policy =
{
  "notify",
  &notify_init,
  &notify_handle_request,
  &notify_handle_response,
  &notify_timeout_expired,
  &notify_cleanup
};


/* subscribe policy functions */

/*
 * subscribe_init()
 * Called when a transaction being handled by subscribe_policy is started
 */
static int subscribe_init(request_t *r, registration_t *reg, int reg_status)
{
  struct subscribe_info *info;
  int i = 0;
  message_t* m;
  
  debug("subscribe_init","starting subscribe policy\n");
  
  if (!(info = 
	(struct subscribe_info*) calloc(1, sizeof(struct subscribe_info))))
    return 1;

  info->reg = reg;
  r->policy_info = (void *)info;
  policy_set_timeout(r, (double)config.timeout);
  
  GetSubscribeContacts(info->reg);
  
  /* no contacts which handle SUBSCRIBE requests */
  if (!info->reg->contacts) { 
    debug("subscribe_init", "No contacts which handle SUBSCRIBE\n");
    policy_send_new_response(r, 405, NULL);
    policy_set_timeout(r,(double)config.timeout);
  }
  
  /*
   * start the local proxying (i.e., proxying to local registrations) with 
   * the highest q value first.
   */
  else { 
    for (i = (info->reg->contacts - 1); i >= 0; i--) { 
      m = GetRequestAsMessage(r);
      policy_proxy_original_request_to(r, &info->reg->contact[i].uri, m, NULL);
      MessageDelete(m);
    }
  }
  
  return 0;
} /* subscribe_init */


/*
 * subscribe_handle_request()
 * Called when a follow-up request (CANCEL) comes in from the client.
 * Dispatch the request as appropriate
 */
static int subscribe_handle_request(request_t *r, message_t *m)
{
  struct subscribe_info *info = (struct subscribe_info *) r->policy_info;
  
  /* 
   * We were canceled.  Cancel all the branches that aren't done, and
   * send back 487.
   */
  if (strcmp(m->p.method, "CANCEL") == 0) {
    
    CancelUnfinishedBranches(r);
    
    if (!info->subscription_added) {
      policy_send_new_response(r, 487, NULL);
      policy_set_timeout(r, (double)config.timeout);
    }
    
    m->handled = finished_handling;
  }
  else {
    /* I don't think any other methods should be showing
     * up here. 
     */
    error("subscribe_handle_request",
	  "Request with surprising method %s on message queue\n",
	  m->p.method);
    m->handled = finished_handling;
  }
  
  return 0;
} /* subscribe_handle_request */


static int subscribe_handle_response(request_t *r, message_t *m, branch_t *b)
{
  struct subscribe_info *info = (struct subscribe_info *) r->policy_info;
  
  /* 
   * If we received a 200 response, and we haven't added the subscription to 
   * the database yet, then add it to the database (if Expires: non-zero).
   */
  if (((m->p.status >= 200) && (m->p.status < 300)) &&
      (!info->subscription_added)) {
    
    /* Expires: 0, we're done, but we don't need to add the subcription to the
     * database. 
     */
    if (r->expires == 1) {
      
      info->subscription_added = 1;
      policy_forward_response(r, m);
      
      /* If we get multiple 200's, we may cancel more than once here; but
       * that's okay. 
       */
      CancelUnfinishedBranches(r);
      
      policy_set_timeout(r, (double)config.timeout);
      m->handled = finished_handling;
    }  /* try to add the subscription to the database. */
    else { 
      if (!add_subscribed_user(r)) { 
        error("subscribe_handle_response",
	      "error adding subscription to database:%s",
	      getMysqlError());
        policy_send_new_response(r, 500, NULL);
        policy_set_timeout(r, (double)config.timeout);
        m->handled = finished_handling;
      }
      else {
        info->subscription_added = 1;
        policy_forward_response(r, m);
        
        /* If we get multiple 200's, we may cancel more than once here; but
	 * that's okay. 
	 */
        CancelUnfinishedBranches(r);
        
        policy_set_timeout(r, (double)config.timeout);
        m->handled = finished_handling;
      }
    } /* if r->expires == 1 */
  } /* check if all branches are finished */
  else if (m->p.status >= 300) { 
    /* send out this message's final response */
    if ((r->branches_outstanding == 0) && (!info->subscription_added)) {
      policy_send_new_response(r, m->p.status, NULL);
      policy_set_timeout(r, (double)config.timeout);
      m->handled = finished_handling;
    }
  }
  
  return 0;
} /* subscribe_handle_response */


/*
 * subscribe_timeout_expired()
 * If we've received a local 200 and we've added the subscription to the
 * database, we're done; otherwise, refresh the timeouts -- 
 * we're waiting for a local 200 response from a branch.
 */
static int subscribe_timeout_expired(request_t *r)
{
  
  struct subscribe_info* info = (struct subscribe_info*)r->policy_info;
  
  /* 
   * if we've already added the subscription to the database or a final 
   * response was generated.
   */
  if ((info->subscription_added) || (r->sent_final)) 
    policy_done(r);
  else /* reset the timeout */
    policy_set_timeout(r, (double)config.timeout);
  
  return 0;
} /* subscribe_timeout_expired */


/*
 * subscribe_cleanup()
 * Called when cleaning up the subscribe policy
 * Free subscribe-specific memory
 */
static int subscribe_cleanup(request_t *r)
{
  free(r->policy_info);
  return 0;
} /* subscribe_cleanup */



/* notify policy functions */

/*
 * notify_init()
 * Called when a transaction being handled by notify_policy is started.  A 
 * notify_policy can only be started after a successful REGISTER request.
 */
static int notify_init(request_t *r, registration_t *reg, int reg_status)
{
  struct notify_info *info;
  message_t* m;
  uri_t request_uri;
  
  debug("notify_init", "starting notify policy\n");
  
  if(!(info = 
       (struct notify_info*) calloc(1, sizeof(struct notify_info))))
    return 1;  
  
  r->policy_info = (void*)info;
  
  request_uri = HTTP_URI(&r->p);  //dont free request_uri
  
  m = GetRequestAsMessage(r);
  policy_proxy_original_request_to(r, &request_uri, m, NULL);
  MessageDelete(m);
  
  policy_set_timeout(r, (double)config.timeout);
  
  return 0;
} /* notify_init */


/*
 * notify_handle_request()
 * Just return a 0, since we are creating the request, this method shouldn't
 * be called.
 */
static int notify_handle_request(request_t *r, message_t *m)
{
  return 0;
} /* notify_handle_request */


static int notify_handle_response(request_t *r, message_t *m, branch_t *b)
{
  struct notify_info *info = (struct notify_info *) r->policy_info;
  
  if (m->p.status >= 200) { /* final response received */
    info->received_final = 1;
    policy_set_timeout(r, (double)config.timeout);
    m->handled = finished_handling;
    
    /* this is a hack, but we have no choice, as otherwise, the policy would
     * send a complaint back (500 error). 
     */
    r->sent_final = 1; 
    
    policy_done(r);
  }
  
  return 0;
} /* notify_handle_response */


/*
 * notify_timeout_expired()
 * if the request received a final response, then we're done.  Otherwise, 
 * refresh the timeout.
 */
static int notify_timeout_expired(request_t *r)
{
  struct notify_info* info = (struct notify_info*)r->policy_info;
  
  /* 
   * if we received a final response to our request, then we're done.
   */
  if (info->received_final)
    policy_done(r);
  else /* reset the timeout */
    policy_set_timeout(r, (double)config.timeout);
  
  return 0;
} /* notify_timeout_expired */


/*
 * notify_cleanup()
 * Called when cleaning up the notify policy
 * Free notify-specific memory
 */
static int notify_cleanup(request_t *r)
{
  struct notify_info* info = (struct notify_info*)r->policy_info;
  
  free(info);
  
  return 0;
} /* notify_cleanup */


/* 
 * for all the contacts in reg, update registration_t with only the 
 * contacts which have method = "SUBSCRIBE" or 'any'.  Look up the user in the 
 * database based on user@host as the key.
 */ 
static void GetSubscribeContacts(registration_t* reg)
{
  int i, found;
  
  for (i = 0; i < reg->contacts; i++) {
    
    if (reg->contact[i].methods) { 
      
      found = (strstr(reg->contact[i].methods, "any") != NULL) ||
	(strstr(reg->contact[i].methods,"SUBSCRIBE") != NULL);
      
      if (!found) {
	debug("GetSubscribeContacts", 
	      "Contact:%s does not allow subscriptions for presence\n",
	      URI_AsString(reg->contact[i].uri));
	
	contact_free(&reg->contact[i]);
	reg->contacts--;
	continue;
      }
    } 
    /* XXX: The else part is commented right now, because none of our UA's
     * send the method param. for initial versions, we will consider
     * this to be OK, to help testing.
     */
    /*
      else { // methods of contact is NULL, remove 
      
      debug("GetSubscribeContacts", 
      "Contact:%s does not allow subscriptions for presence\n",
      URI_AsString(reg->contact[i].uri));
      
      contact_free(&reg->contact[i]);
      reg->contacts--;
      }
    */
    
  }
  return;
} /* GetSubscribeContacts */


static int 
CreateSocket(struct sockaddr_in *sin, char *host, unsigned short port) 
{
  int sock;
  
  memset(sin, 0, sizeof(struct sockaddr_in));
  
  /* Create UDP socket for request. */
  if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    error("CreateSocket", "socket call fails...\n");
    return -1;
  }
    
  sin->sin_family      = AF_INET;
  sin->sin_port        = htons(port);
  sin->sin_addr        = host2ip(host);
  
  return sock;

} /* CreateSocket */


/*
 * Creates a NOTIFY request for the sub_table_entry passed in.
 */
static request_t* CreateNotifyRequest(sub_table_entry* info, 
                                      char * status)
{
  request_t* r;
  int sock;
  struct sockaddr_in sin;
  
  uri_t uri_from;
  uri_t uri;
  parser_header_t* h;
   
  int use_contact = 0; /* whether to use contact or from in request-URI */
  int total_length, content_length;

  DString sipmsg;

  sock_t dum_s;
 
  /* get the URI to be used for sending the request */
  if (info->contact) { /* try "contact" first */
    uri_from = URI_Parse(info->contact);
    use_contact = 1;
  }
  else {
    uri_from = URI_Parse(info->sub_from);
    use_contact = 0;
  }
  
  sock = CreateSocket(&sin, uri_from.host, (unsigned short)uri_from.port);
  
  r = RequestCreate(SOCK_DGRAM, sock, dum_s);
  //r = RequestCreate(SOCK_DGRAM, sock, sin);
  dum_s.tcp_s = sin;
  
  /* format the request, however, in this case the sub_to becomes the from
   * (from) and the sub_from becomes the remote_user (to, contact). 
   */
  DStringInit(&sipmsg);

  if (use_contact) 
    DStringPrintf(&sipmsg, "NOTIFY %s SIP/2.0\r\n", info->contact);
  else 
    DStringPrintf(&sipmsg, "NOTIFY %s SIP/2.0\r\n", info->sub_from);
  
  /* should not add Via, since it is automatically added */
  DStringPrintf(&sipmsg, "CSeq: %u NOTIFY\r\n", 2);
  
  //if (use_contact) //this should ideally give contact of sip_from...see spec
  // DStringPrintf(&sipmsg, "Contact: <%s>\r\n",info->contact);
  
  DStringPrintf(&sipmsg, "From: sip:%s\r\n", info->sub_to);
  
  DStringPrintf(&sipmsg, "Call-ID: %s\r\n", info->call_id);
  
  {
    DString current_time;
    time_t time_now = time(0);
    DStringInit(&current_time);
    DStringCftime(&current_time, "%a, %d %b %Y %H:%M:%S GMT", time_now, 0);
    DStringPrintf(&sipmsg, "Date: %s\r\n", DStringValue(&current_time));
    DStringFree(&current_time);
  }

  DStringPrintf(&sipmsg, "To: %s\r\n", info->sub_from);
  
  /* Expires is defaulted to one hour */
  DStringPrintf(&sipmsg, "Expires: %d\r\n", 3600);
  DStringAppend(&sipmsg, "Content-Type: application/xml\r\n\r\n", -1);

  if (info->record_route) {
    /* record_route is actually a set of well constructed Route headers.
     */
    DStringPrintf(&sipmsg, "%s", info->record_route);
  }

  /* Fill content-body and type fields */
  {
    DString content_body;
    DStringInit(&content_body);
    DStringAppend(&content_body, "<?xml version=\"1.0\"?>\r\n", -1);
    DStringAppend(&content_body, "<events>\r\n", -1);
    DStringPrintf(&content_body, "<event>%s</event>\r\n</events>\r\n", status);
    content_length = DStringLength(&content_body);
    DStringPrintf(&sipmsg, "Content-Length: %u\r\n", content_length);
    DStringAppend(&sipmsg, DStringValue(&content_body),-1); 
    DStringFree(&content_body);
  }
  
  debug("CreateNotifyRequest",
	"Constructed message: %s\n",DStringValue(&sipmsg));

  total_length = DStringLength(&sipmsg);

  /* parse the request */
  if (HTTP_Parse(DStringValue(&sipmsg), &r->p) < 0 ) {
    error("CreateNotifyRequest", "parse error\n");
    goto create_error;
  }
  
  uri = HTTP_URI(&r->p);
  
  if (uri.scheme == NULL ||
      (!strcmp(uri.scheme, "sip") && !strcmp(uri.scheme, "tel"))) {
    error("CreateNotifyRequest", "uri.scheme is not sip or tel\n");
    goto create_error;
  }
   
  if (content_length > 0 
      && (r->p.header_length + content_length > total_length)) {
    DString body;
    char *msg = DStringValue(&sipmsg);
    DStringInit(&body);
    DStringAppend(&body, &msg[r->p.header_length], 
		  total_length - r->p.header_length);
    r->request_body.content = DStringValue(&body);
    r->request_body.content_length = total_length - r->p.header_length;
  }
  
  if ((h = RFC822_Header(r->p.headers, "Content-Type")) && h->elements) {
    /* request_body.content_type points to the internal
     * parser structure element. */
    r->request_body.content_type = h->elements->value;
  }
  
  /* ---- Now fill in the relevant fields of r */
  r->version = HTTP_Version(&r->p);
  if (!SIP_Has_Mandatory_Headers(&r->p)) {
    error("CreateNotifyRequest", "does not have mandatory headers\n");
    goto create_error;
  }
  
  r->cseq         = SIP_CSeq(&r->p);
  r->callID       = NULL;
  strassign(&r->callID, SIP_CallID(&r->p));
  r->contact      = SIP_Contact(&r->p);
  
  r->expires      = SIP_Expires(&r->p);
  /* Via headers are added automatically */
  r->to           = SIP_To(&r->p);
  r->from         = SIP_From(&r->p);
  
  
  r->sent_200 = 0; /* initialize to 0 */
  r->sent_final = 0;
  
  r->hash = ComputeRequestHash(r);
        
  /* XXX: Ideally we should lock the mutex before putting it in the
   * RequestAdd list. Here, we dont do this because in Win32 only the
   * creating thread can unlock the mutex. Since the thread we create here,
   * is short-lived, this results in a deadlock.
   */
  // pthread_mutex_lock(&r->mutex);
  
  debug("CreateNotifyRequest",
	"Request 0x%p: Adding request to list\n", r);
  
  if(RequestAdd(r) == 0) {
    debug("CreateNotifyRequest","Request could not be added\n");
    goto create_error;
  }      
  
  goto done;

 create_error:
  RequestFree(r);
  r = NULL;
 done:
  if (r)
    Log(r);
  return r;
} /* CreateNotifyRequest */


/* 
 * add a user into the subscription table.  Returns 1 if successful, 0
 * if not successful.
 */
int add_subscribed_user(request_t* r)
{
  sub_table_entry* entry;
  
  entry = (sub_table_entry*)calloc(1, sizeof(sub_table_entry));
  
  entry->sub_to = strdup(r->destination_user->user);   /* free on exit */
  entry->sub_from = URI_AsString(SIP_From(&r->p).uri); /* free on exit */
  entry->call_id = SIP_CallID(&r->p);                  /* must not be freed */
  entry->record_route = BuildUASRouteSet(&r->p);       /* free on exit */
  
  if (r->contact && r->contact[0].uri.scheme) { /* Expires parameter for contact */
    
    if (r->contact[0].expires != 0) 
      entry->expires = r->contact[0].expires;
    else /* Expires header */
      entry->expires = r->expires;
    
    /* Contact */
    entry->contact = URI_AsString(r->contact[0].uri);
    
    sub_insert(config.database, entry, NULL, NULL);
    
    free_valid(entry->contact);
    
    entry->contact = NULL;
  }
  else { /* no contact header */
  
    entry->expires = r->expires;
  
    sub_insert(config.database, entry, NULL, NULL);
  
  }

  /* dont use free_sub_table_entry as the call_id shouldnt be freed */

  free_valid(entry->sub_to);
  free_valid(entry->sub_from);
  free_valid(entry->record_route);
  free_valid(entry);

  return 1;
} /* add_subscribed_user */

/*
 * Checks to see if a subscription for presence current exists in the database.
 * A subscription exists if the to (subscribee) and from (subscriber) exist in 
 * the database.  Returns 1, if a subscription exists, or 0, if it doesn't.
 */
static int subscription_exists(char * to, uri_t from_uri)
{
  char * from;
  sub_table_entry* entry;
  int retval = 0;

  from = URI_AsString(from_uri);
  
  entry = sub_get_by_fromto(config.database, from, to, NULL, NULL);
  
  if (entry) {
    retval = 1;
    free_valid(entry);
  }
    
  free_valid(from);

  return retval;  
} /* subscription_exists */
  

/*
 * Remove a particular subscriber for a subscribee from the database.  Only 
 * called when r->contact->expires == 1 (or 0), meaning Contact:"some uri";
 * expires = 0.  Returns 1, if successful, 0 if unsuccessful.
 */
static int remove_subscribed_user(char *subscribee, uri_t subscriber)
{
  char *from = NULL;
  int retval = 1;
    
  from = URI_AsString(subscriber);
  
  delete_sub_by_fromto(config.database, from, subscribee, NULL, NULL);
    
  free_valid(from);

  return retval;
} /* remove_subscribed_user */
  
/* 
 * make sure Event header contains proper token of "presence" 
 * assume only one Event header per SUBSCRIBE request.  If
 * the header is there, add it to the response.
 */
static int containsProperEvent(request_t *r)
{
  parser_header_t *h = RFC822_Header(r->p.headers, "Event");
  parser_element_t *e = h->elements;
    
  HeaderAppend(&r->headers, "Event", e->value);
    
  if (strcmp(e->value,"presence") == 0)
    return 1;
  else
    return 0;
}
  
/*
  * SetupAndValidateTarget()
  * Set r->destination_user, based on 'uri'.  Validate that
  * r->destination_user is authorized to perform a SUBSCRIBE request.  If not,
  * or if the user doesn't exist, set response status and return false.
  */
static int SetupAndValidateTarget(request_t *r, const uri_t *destination)
{
  db_err db_ret;
  const char* method = "SUBSCRIBE";
    
  r->destination_user = CanonicalizeAndLookupURI(r, destination, 0,
						 "Request-URI", "subscription",
						 r->source_user);
  if (r->destination_user == NULL) {
    return 0;
  }
    
  db_ret = method_allowed_for_user(method, r->destination_user);
    
  if (db_ret != DB_OK) {
    //    request_set_dberr_status(r, db_ret, r->destination_user->user,
    //                           405, NULL);
    /* XXX: add Allow: headers for DB_NO */
    return 0;
  }
  
  return 1;
} /* SetupAndValidateTarget */
  
/*
 * SetupSource()
 * Set r->source_user, based on 'source'.  If the user doesn't exist,
 * return false and set the response status.
 */
static int SetupSource(request_t *r, const uri_t *source)
{
  r->source_user = CanonicalizeAndLookupURI(r, source, 1,
					    "From URI", "subscription",
					    NULL);
  
  if (r->source_user == NULL) {
    return 0;
  }
  
  return 1;
} /* SetupAndValidateSource */

/*
 * SUBSCRIBE()
 * Handle SUBSCRIBE request, adding and deleting subscription entries to/from
 * database.  
 */
void SUBSCRIBE(request_t * r)
{
  char *userId = NULL;      /* subscribee (local user) */
  registration_t reg;       /* user registration */
  uri_t request_uri, from_uri; 
  int status; 
  canon_result target_canon;
  
  memset(&reg, 0, sizeof(reg));
  memset(&target_canon, 0, sizeof(target_canon));
  
  ResponseBuildDefault(r); 
  
  r->policy = NULL;
  
  request_uri = HTTP_URI(&r->p);
    
  /* Set up r->destination_user, and check if it's allowed to subscribe */
  if (!SetupAndValidateTarget(r, &request_uri)) {
    ResponseSetStatus(r, 604, NULL);
    goto failure;
  }
    
  /* Set up r-source */
  if (!SetupSource(r, &r->from.uri)) {
    ResponseSetStatus(r, 604, NULL);
    goto failure;
  }
    
  /* Make sure the request contains authentication information for
   * from_userid. */
  if (AuthenticateSIP(r, 1, 0, 0) != auth_success) {
    /* The error status appropriate headers (WWW-Auth..) are already set */
    goto done;
  }
    
  userId = strdup(r->destination_user->user);
    
  /* check to make sure Event header is there and Event parameter is 
     correct */
  if (!(RFC822_Header(r->p.headers,"Event"))) {
    ResponseSetStatus(r, 400, "Event Header missing");
    goto failure;
  }
  else if (!containsProperEvent(r)) {
    ResponseSetStatus(r, 489, "Event type not supported");
    goto failure;
  }
    
  from_uri = SIP_From(&r->p).uri;
    
  /* If the subscription exists and Expires: 0, then remove all subscriptions 
   * from database, otherwise, start the subscribe policy, and forward the 
   * SUBSCRIBE request to all the contacts.  When a subscription for presence
   * does not exist and a SUBSCRIBE request with Expires: 0 is received, it 
   * means one-time presence notification, so the end clients (registrations)
   * should process the one-time presence notification.
   */
  if (r->expires == 1) { 
    if (subscription_exists(userId, from_uri)) {
      if (!remove_subscribed_user(userId, from_uri)) /* database error */
	ResponseSetStatus(r, 500, "Database error");
      else 
	ResponseSetStatus(r, 202, "Accepted");
      goto failure; //send the response
    }
  }
  else { 
    /*
     * Compute expiration times for new subscriptions. 
     */
    if (r->expires == 0 ) {
      /* Set the default expiry time */
      r->expires = time(0) + config.expires;
    }
      
    /* Update the expires on per contact basis */
    if ( r->contact[0].expires == 0 ) {
      /* Per contact expires is not set. Use Expires: header
       * from r->expires 
       */
      r->contact[0].expires = r->expires;
    }
    if (!strcmp(r->contact[0].uri.scheme, "sip") &&
	r->contact[0].expires > time(0) + config.maxexpires) {
      /*
       * expires is more than max allowed for SIP registrations. 
       * Set it to max allowed.
       */
      r->contact[0].expires = time(0) + config.maxexpires;
    }
  } /* if r->expires == 1 */
  
  /**
   * initialize to 300 so if cgi_policy returns "default-action" and the 
   * method is not "SUBSCRIBE", then some redirect or proxy policy will be 
   * found.  (but this should never happen, since if cgi returns "default-
   * action", then the method will be SUBSCRIBE, which will then set the
   * policy to subscribe_default_policy).
   */
  status = 300; /* initialize to 300 */
    
  /* locate this user */
  register_find(userId, &reg);
    
  debug("SUBSCRIBE",
	"r=%p: status=%d, reg.scriptset.count=%d, reg.contacts=%ld\n",
	r, status, reg.scriptset.count, (long)reg.contacts);
  
  if (reg.contacts) {
    /* remove any contact subscriptions from database */
    if (reg.contact[0].expires == 1) {
      if (!remove_subscribed_user(userId, from_uri)) { 
	ResponseSetStatus(r, 500, "Database error");
	goto failure;
      }
      else { /* remove contact from registrations */
	contact_free(&reg.contact[0]);
	reg.contacts--;
      }
    }
  }
  
  /* Check if we have a script specified, if so, run the cgi script */
  if (GetScriptByDisposition(&reg.scriptset, disposition_sip_cgi) != NULL) {
    r->policy = &cgi_policy;
    execute_policy(r, &reg, status);
    goto done;
  }
  /* user exists, but no script.  Proceed with 'default-action' for subscribe */
  else if (reg.contacts >= 0) {
    r->policy = &subscribe_policy;
    execute_policy(r, &reg, status);
    goto done;
  }
  else { /* user doesn't exist */
    DString reason;
    DStringInit(&reason);
    debug("SUBSCRIBE", "No registration information for:%s\n",
	  userId);
    DStringPrintf(&reason, "No registration information for:%s\n", 
		  userId);
    ResponseSetStatus(r, 604, DStringValue(&reason));
    DStringFree(&reason);
    goto failure;
  } /* if register_find */
    
 failure: 
  Log(r);
  ResponseSend(r, 0, 1);
  
 done:
  register_free(&reg); 
  free_valid(userId);
  return;
} /* SUBSCRIBE */


/*
 * Threaded function which executes the notify policy for each new NOTIFY
 * request created.
 */
void* NotifyProcessThread(void* ptr)
{
  request_t* r = (request_t*)ptr;
  
  debug("NotifyProcessThread", "Creating Notify Thread\n");
  r->policy = &notify_policy;
  execute_policy(r, NULL, 200); /* don't need registrations */
  
  return NULL;
}
  

/*
 * Creates a new NOTIFY request for each subscription found in the database and
 * starts a seperate thread for each new request.
 */
void send_out_notify_requests(int expires, char * registered_user)
{
  sub_table_entry** subscriptions = NULL; 
  int count = 0;
  int i;
  char * event_status;
  
  if (expires) /* offline */
    event_status = strdup("offline");
  else
    event_status = strdup("online");
  
  debug("send_out_notify_requests", "creating notify requests\n");
  
  subscriptions = get_all_subscriptions(config.database,
					registered_user, &count, NULL, NULL);  
  
  //create a request and send it out
  for (i = 0; i < count; i++) { 
    request_t* r; /* new NOTIFY request */
    
    r = CreateNotifyRequest(subscriptions[i], event_status);
    
    if (r) {
      r->expires = expires;
      if ((pool_schedule(NotifyProcessThread, (void *)r)) 
          != 0) {
	error("send_out_notify_requests", "couldn't start NOTIFY thread");
	RequestFree(r);
      }
    }
  }
    
  /* Free the sub_table_entry** structs, we don't need them anymore. */
  free_sub_table_entries(subscriptions);
  
  free_valid(event_status);
}
