/*
request.c - Create, search, process, destroy requests.

Copyright 1998-1999 by Columbia University; all rights reserved 
*/

#include <stdio.h>
#include <stdlib.h>     /* calloc() */
#include <assert.h>
#include <pthread.h>
#include <unistd.h>     /* close() */
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <errno.h>
#include "pthread-util.h"
#include "request-sip.h"
#include "sip.h"
#include "strdupn.h"
#include "sysdep.h"
#include "rfc822.h"
#include "config.h"
#include "authenticate.h"
#include "libsip.h"

#define PROTOTYPES         /* for ht.h, to enable ANSI style decl. */
#include "htimpl.h"

#include "log.h"
#include "msgflow.h"
#include "thread_pool.h"
#include "sipsnmp_interface.h"

#include "canonicalize.h" /* for free_canon_result */

extern double gettimeofdayf(void);
extern void ResponseInit();   /* response.c's init */

/*
 * Hash table. declaration of variables, and callbacks for bookkeeping.
 */
static dict_h request_table;  /* hash table for holding request_t's */

static ht_val compute_requestobj_value(dict_obj o);  /* call backs */
static ht_val compute_requestkey_value(dict_key o);
static ht_val compare_requestobj_values(dict_obj o1, dict_obj o2);
static ht_val compare_requestkey_values(dict_key k1, dict_obj o1);

/*
 * Hash table initialization.
 */
static struct ht_args ht_request = { DEFAULT_TABLE_ENTRIES,  
				     DEFAULT_BUCKET_ENTRIES, 
				     (ht_func)compute_requestobj_value, 
				     (ht_func)compute_requestkey_value
};

/* Lock for 'hash table'.  To avoid deadlock, always lock 
 * R_lock before locking an individual request's lock, if you 
 * need to hold both locks simultaneously. 
 */
static pthread_mutex_t R_lock = PTHREAD_MUTEX_INITIALIZER;

/*
 * Init function that must be called once during program startup.
 *
 * for requests, initialize the hashtable. for responses, initialize
 * the string table by calling ResponseInit.
 */
void InitLibSip() {

  static int once = 0;

  if (!once) {
    request_table = ht_create((dict_function)compare_requestobj_values, 
			      (dict_function)compare_requestkey_values, 
			      0, 
			      (int*)0, 
			      &ht_request);
  
    if (!request_table) 
      fatal("init_hashtable","ht_create returns null...\n");
    
    ResponseInit();
    
    once = 1;
  }
  
  return;  

} /* InitLibSip */


/*
 * RequestCreate()
 * Create a new request object.
 */
request_t *RequestCreate(int type, int socket, sock_t sin)
{
  request_t *r = calloc(1, sizeof(request_t));
 
  r->type          = type;
  r->socket        = socket;
  r->time_received = gettimeofdayf();
  
  pthread_mutex_init(&r->mutex, NULL);
  pthread_cond_init(&r->cond, NULL);
  
  return r;
} /* RequestCreate */


/*
 * RequestAdd()
 * If it's valid (i.e., findable by RequestSearch), insert the request
 * into the hash table.  Return 1 if it was added, 0 otherwise. 
 */
int RequestAdd(request_t *r)
{
  int ret = 1;
  unsigned long hash = r->hash;
  /* We could also check if we have a CSeq, but both 0 and (ulong)-1 are
   * valid according to the spec... */
  if (r->callID == NULL) {
    return 0;
  }

  pthread_mutex_lock(&R_lock);

  if (!hash)
    hash = ComputeRequestHash(r);

  if (ht_search(request_table, &hash)) {
    error("RequestAdd","Trying to add a duplicate entry...\n");
    ret = 0;
  }

  if(ret &&  ht_insert(request_table, r) != 0) {
    error("RequestAdd","ht_insert returns non zero...\n");
    ret = 0;
  }
  
  pthread_mutex_unlock(&R_lock);

  /* Note:
   * This must not be done before unlocking the global lock, because:
   *   - the Snmp subsystem has its own locks
   *   - deadlock may occur if its done from inside
   *
   * The SNMP add call is the last thing that happens when a request is
   * added, and the SNMP delete call is the first thing that happens when a
   * request is deleted.
   * 
   * All other SNMP requests do not affect the SNMP transaction list.
   */
  if (ret && SnmpProcessMessage) {
    (*SnmpProcessMessage)(r, SNMP_TYPE_ADD_TRANS, NULL);
  }

  return ret;
} /* RequestAdd */

/*
 * RequestDelete()
 * Remove the request_t structure form the hash table,
 *  without actually freeing the request_t structure.
 * The mutex on the request_t is assumed to be locked.
 * It remains locked when the function exits.
 */
void RequestDelete(request_t *r)
{
  /* 
   * Lock R_lock before r->mutex to avoid
   * circular loop condition for a deadlock. 
   */
  pthread_mutex_unlock(&r->mutex);
  pthread_mutex_lock(&R_lock);
  pthread_mutex_lock(&r->mutex);

  /* Remove request from the request list. */
  debug("RequestDelete", "Freeing 0x%p\n", r);
  
  ht_delete(request_table,(dict_obj)r);

  pthread_mutex_unlock(&R_lock);
} /* RequestDelete */

/*
 * RequestFree()
 * Free the request and the data associated with it.
 * Remove the request from the hash table of requests if it is on it.
 * Note: you should *NOT* have the request locked before calling this.
 * Note: make sure that your program flows allow this to be called from
 * only one location for each request.  The locks do not protect against
 * RequestFree()/RequestFree() race conditions, though we are protected
 * against RequestFree()/RequestSearch() races.
 */
void RequestFree(request_t *r)
{
  struct message_t *m, *m_next;
  int i;

  if (!r) return;
 
  /* Note:
   * The ordering is important.  Don't put the snmp code inside the
   * lock.  The snmp subsystem has its own locks.  Putting it inside the
   * R_lock mutex may cause deadlocks.
   */
  if (SnmpProcessMessage) {
    (*SnmpProcessMessage)(r, SNMP_TYPE_DEL_TRANS, NULL);
  }

  /* Remove request from the request list. */
  /* Lock R_lock before r->mutex. */
  pthread_mutex_lock(&R_lock);
  pthread_mutex_lock(&r->mutex);

  debug("RequestFree", "Freeing 0x%p\n", r);
  
  ht_delete(request_table, (dict_obj) r);

  pthread_mutex_unlock(&R_lock);
  
  /* mark so that it's more obvious if old request get re-used */
  r->status = 0;

  /* Free storage of request. */
  HTTP_Free(&r->p);
  free_valid(r->version);
  /* r->request_body.content_type is a pointer owned by the parse structure, 
     and it was freed by HTTP_Free; so don't free it again here. */
  free_valid(r->request_body.content);

  free_valid(r->callID);
  free_valid(r->via.hostname);
  free_valid(r->via.branch.str);

  URI_Free(&r->to.uri);
  free_valid(r->to.display_name);

  URI_Free(&r->from.uri);
  free_valid(r->from.display_name);

  URI_Free(&r->outbound_proxy);

  SIP_FreeContact(r->contact);
  free_valid(r->user_agent);

  free_valid(r->reason);
  HeaderFree(&r->headers);

  free_valid(r->response_body.content_type);
  free_valid(r->response_body.content);

  HTTP_AuthorizationFree(&r->authorization);
  HTTP_AuthenticateFree(&r->authenticate);

  if (r->source_user != NULL && r->source_user != r->destination_user) {
    free_put_table_entry(r->source_user);
  }
  if (r->destination_user != NULL) {
    free_put_table_entry(r->destination_user);
  }

  free_canon_result(&r->canonicalized_uri);

  /* Free the message queue */
  for (m = r->mqueue; m != NULL; m = m_next) {
    m_next = m->next;
    MessageDelete(m);
  }

  /* Free URIs from each branch */
  for (i = 0; i < r->branch_count; i++) {
    URI_Free(&r->branches[i].uri);

    /* 
     * Free any unprocessed message here. e.g. shut_down.
     * We must free this here since policy_core may post
     * shutdown message to client after the client thread
     * has already stopped. Other fields of branch_t are
     * freed by the client thread.
     */
    for (m = r->branches[i].mlist; m != NULL; m = m_next) {
      m_next = m->next;
      MessageDelete(m);
    }
  }

  /* Free the branch structure */
  free_valid(r->branches);

  pthread_mutex_unlock(&r->mutex); /* Release lock before deleting */
  pthread_mutex_destroy(&r->mutex);
  pthread_cond_destroy(&r->cond);
  free(r);

  return;
} /* RequestFree */


/*
 * RequestPrint()
 * Prints the existing requests on the output stream FILE *fp
 */
void RequestPrint(FILE *fp)
{
  request_t *rp;

  pthread_mutex_lock(&R_lock);
  rp = ht_minimum(request_table);
 
  while (rp) {
    /* Add prints here */

    rp = ht_successor(request_table, rp);
  }
  pthread_mutex_unlock(&R_lock);
 
  return;
} /* RequestPrint */

 
/*
 * RequestSearch()
 * Determine if the request 'r' with hash 'hash' exists in the list of
 * pending requests.  A request is deemed identical if it has the same
 * Request-URI, To, From, Call-ID, and sequence number.  Return the matching
 * request if it exists, otherwise NULL.
 * N.B.: The returned request has its mutex locked!  You must unlock it
 * when you're done with it!  */
request_t *RequestSearch(request_t *r, unsigned long hash)
{
  request_t *rp, *result = NULL;

  pthread_mutex_lock(&R_lock);

  /*
   * Algorithm:
   * if hash is zero, run ComputeHash to get hash.
   * 
   * lookup into hash table using "hash" as the key.
   *   -then check whether the object equals what is needed based on
   *    equivalence rules.
   *   -if not, then iterate until end of list is reached, or a match is found.
   *
   */
  if (!hash) 
    hash = ComputeRequestHash(r);

  rp = (request_t*) ht_search(request_table, (dict_key)&hash);
  
  if (!rp) 
    goto search_done;
  
  while (rp) {
    
    /* strictly speaking this is not needed. */
    if ( rp != r 
	 && (!r->hash || rp->hash == r->hash) 
	 && (rp->p.URI_string == NULL || r->p.URI_string == NULL
	     || URI_Equal_With_Defaults(rp->p.URI, r->p.URI, sip_uri_defaults))
	 && URI_Equal_With_Defaults(rp->from.uri, r->from.uri, sip_uri_defaults)
	 && URI_Equal_With_Defaults(rp->to.uri, r->to.uri, sip_uri_defaults)
	 && strcmp(rp->callID, r->callID) == 0
	 && rp->cseq.seq == r->cseq.seq) {
      pthread_mutex_lock(&rp->mutex);
      debug("RequestSearch", "Got rp->lock\n");
      result = rp;
      break;
    }

    rp = ht_successor(request_table, rp);
  }
 
 search_done:
  pthread_mutex_unlock(&R_lock); 
  debug("RequestSearch", "returning %p for hash value %lu\n",result, hash);
  return result;
} /* RequestSearch */


/*
 * RequestFind()
 * Determine if the request corresponding to message 'm' with hash 'hash' 
 * exists in the hash table. A request is deemed identical if it has the same
 * Request-URI, To, From, Call-ID, and sequence number.  Return the matching
 * request if it exists, otherwise NULL.
 * N.B.: The returned request has its mutex locked!  You must unlock it
 * when you're done with it! 
 * The comparision for ACK is slightly different because the ACK
 * should match the original INVITE and as ACK can be sent end-to-end
 * it is possible that the Request-URI is different.
 */
request_t *RequestFind(parser_t *p, unsigned long hash)
{
  request_t *rp = NULL, *result = NULL;
  sip_addr_t to, from;
  char * callid;
  sip_cseq_t cseq;

  if ( p->headers == NULL ) {
    error("RequestFind", "incorrect parser struct. Probably a loop.\n");
    return NULL;
  }

  to     = SIP_To(p);
  from   = SIP_From(p);
  callid = SIP_CallID(p);
  cseq   = SIP_CSeq(p);

  if ( callid == NULL ) {
    error("RequestFind", "callid in the message is NULL\n");
    return NULL;
  }

  pthread_mutex_lock(&R_lock);
  debug("RequestFind", "Got R_lock\n");
  
  /*
   *
   * The ACK part is iterative, to guarantee the needed semantics as
   * given above. The else part (non-ACK) part, is same as RequestSearch.
   * Look into the hash table for item with value hash. If it is not 
   * there then the item is not found. If the hash value is zero, do 
   * iteratively.
   */
  if ( p->method && !strcmp(p->method, "ACK") ) {
    
    rp = ht_minimum(request_table);     
    
    while (rp) {
      
      if ( URI_Equal_With_Defaults(rp->from.uri, from.uri, sip_uri_defaults)
           && URI_Equal_With_Defaults(rp->to.uri, to.uri, sip_uri_defaults)
           && strcmp(rp->callID, callid) == 0
           && rp->cseq.seq == cseq.seq
           && (rp->p.method == NULL ||
               !strcmp(rp->p.method,"INVITE"))) {
        result = rp;
        debug("RequestFind", "Got rp->lock\n");
        pthread_mutex_lock(&rp->mutex);
        break;
      }
      
      rp = ht_successor(request_table, rp);
    }
  } /* p->method && ... */
  else {
    if (hash) {
      if (!(rp = ht_search(request_table, &hash)))
      	goto find_done;
    }
    else 
      rp = ht_minimum(request_table);
    
    while (rp) {
      if ( (hash == 0 || rp->hash == hash)
           && (rp->p.URI_string == NULL || p->URI_string == NULL
               || URI_Equal_With_Defaults(rp->p.URI, p->URI, sip_uri_defaults))
           && URI_Equal_With_Defaults(rp->from.uri, from.uri, sip_uri_defaults)
           && URI_Equal_With_Defaults(rp->to.uri, to.uri, sip_uri_defaults)
           && strcmp(rp->callID, callid) == 0
           && rp->cseq.seq == cseq.seq) {
        result = rp;
        debug("RequestFind", "Got rp->lock\n");
        pthread_mutex_lock(&rp->mutex);
        break;
      }

      rp = ht_successor(request_table, rp);
    }
  }
   
 find_done:
  pthread_mutex_unlock(&R_lock);  
  URI_Free(&to.uri);
  URI_Free(&from.uri);
  free_valid(to.display_name);
  free_valid(from.display_name);
  
  return result;
} /* RequestFind */

/*
 * SendFailureResponse()
 * For a request which has failed, send back a failure response, if it
 * wasn't an ACK request.  The request's response status should be set.
 * If "built" is true, the request's response headers should also already
 * be set; otherwise the default response headers will be built.
 * NOTE: because malformed requests don't end up on the request list, if
 * this request was an INVITE it can't be ACK'd.
 * HACK: don't retransmit them; let the server retransmit if necessary.
 * Since we're immediately sending a definitive (4xx or 5xx) response,
 * reliability works okay.
 * Technically, this violates RFC 2543.  However, you're not actually
 * obligated to send back responses to malformed requests at all...
 */
static void SendFailureResponse(request_t *r, int built) 
{
  char *method = HTTP_Method(&r->p);
  if (strcasecmp(method, "ACK") == 0) {
    /* Don't respond to mal-formed ACKs; just free them.
     * Everything else gets a response. */
    debug("SendFailureResponse", "r = 0x%p: Rejected ACK\n", r);
    RequestFree(r);
  }
  else {
    if (!built) {
      ResponseBuildDefault(r);
    }
    /* For ResponseSend, we must have the mutex locked. */
    pthread_mutex_lock(&r->mutex);
    ResponseSend(r, 0, 1);
  }

  free_valid(method);
} /* SendFailureResponse */


/*
 * ValidateIncomingRequest()
 * A new request has arrived.  Verify that it is correct and processable; if
 * not, send the appropriate error response, and return false.
 */
static int ValidateIncomingRequest(request_t *r)
{
  char *method = NULL, *version = NULL, *protocol = NULL;
  uri_t uri;                   /* request URI */
  parser_header_t *h;
  parser_element_t *e;
  int handle_locally;
  int built = 0;

  if (r->status >= 400) {
    debug("ValidateIncomingRequest", "r = 0x%p: (r->status = %d) >= 400\n",
          r, r->status);
    goto fail;
  }

  method = HTTP_Method(&r->p);
  version = HTTP_Version(&r->p);
  protocol = HTTP_Protocol(&r->p);

  if (method == NULL) {
    debug("ValidateIncomingRequest", "r = 0x%p: method == NULL\n", r);
    ResponseSetStatus(r, 400, NULL);
    goto fail;
  }

  if (protocol == NULL) {
    debug("ValidateIncomingRequest", "r = 0x%p: protocol == NULL\n", r);
    ResponseSetStatus(r, 400, NULL);
    goto fail;
  }

  if (version == NULL) {
    debug("ValidateIncomingRequest", "r = 0x%p: version == NULL\n", r);
    ResponseSetStatus(r, 400, NULL);
    goto fail;
  }

  if (strcmp(protocol, "SIP") != 0) {
    debug("ValidateIncomingRequest", "r = 0x%p: protocol == %s\n", 
          r, protocol);
    ResponseSetStatus(r, 505, NULL);
    goto fail;
  }

  if (strcmp(version, "2.0") != 0) {
    debug("ValidateIncomingRequest", "r = 0x%p: version == %s\n", 
          r, version);
    ResponseSetStatus(r, 505, NULL);
    goto fail;
  }

  uri = HTTP_URI(&r->p);
  /* sipd supports the 'sip' and 'tel' URI schemes. */
  if (uri.scheme == NULL ||
      (strcmp(uri.scheme, "sip") != 0 && strcmp(uri.scheme, "tel") != 0)) {
    debug("ValidateIncomingRequest",
          "r = 0x%p: uri.scheme = %s is not sip or tel\n",
          r, uri.scheme != NULL ? uri.scheme : "NULL");
    ResponseSetStatus(r, 400, NULL);
    goto fail;
  }

  handle_locally = SIP_HandleRequestLocally(method, uri);
  if (handle_locally) {
    h = RFC822_Header(r->p.headers, "Require");
  }
  else {
    h = RFC822_Header(r->p.headers, "Proxy-Require");
  }

  if (h != NULL) {
    /* TODO: recognize locally-understood extensions. */
    debug("ValidateIncomingRequest", "r = 0x%p: %s present (420)\n",
          r, handle_locally ? "Require" : "Proxy-Require");
    ResponseSetStatus(r, 420, NULL);
    ResponseBuildDefault(r);
    /* Add "Unsupported" headers */
    for (e = h->elements; e != NULL && e->element != NULL; e = e->next) {
      HeaderAppend(&r->headers, "Unsupported", e->element);
    }
    built = 1;
    goto fail;
  }

  /* If some mandatory headers are missing, return 400. */
  if (!SIP_Has_Mandatory_Headers(&r->p)) {
    debug("ValidateIncomingRequest",
          "r = 0x%p: mandatory headers missing\n", r);
    ResponseSetStatus(r, 400, NULL);
    goto fail;
  }

  free_valid(method);
  free_valid(protocol);
  free_valid(version);

  return 1;

fail:
  SendFailureResponse(r, built);

  debug("RequestProcessThread",
        "r = 0x%p: ValidateIncomingRequest failed\n", r);

  free_valid(method);
  free_valid(protocol);
  free_valid(version);

  return 0;
} /* ValidateIncomingRequest */


/*
 * HashString()
 * Hash a null-terminated string into an unsigned long, using the Tcl hash
 * function (multiply by 9 and add).  Helper function for HashString().
 * Question: should this be defined in some other file?
 */
static unsigned long HashString(unsigned long start_hash, char *str)
{
  char *c;
  unsigned long hash = start_hash;

  if (str) {  /* Hash on non-null strings only */
    for (c = str; *c != '\0'; c++) {
      hash = ((hash << 3) + hash);
      hash += (unsigned long) *c;
    }
  }

  return hash;
} /* HashString */


/*
 * HashInt()
 * Hash an unsigned long into an unsigned long, exactly as if we sprintf'd the
 * value being added to the hash into a string with "%lu" and then called
 * HashString() on it.
 * Because of this "precisely the same" requirement, we first unpack
 * the unsigned long in decimal into a buffer, then add the buffer in
 * the opposite order (most-significant-digit first).
 */
static unsigned long HashInt(unsigned long start_hash,
                             unsigned long start_value)
{
  unsigned long hash = start_hash;
  unsigned long value = start_value;
  int buf[sizeof(unsigned long)*3];
  int idx;
  
  for (idx = 0;
       idx < sizeof(unsigned long) * 3 && (value != 0 || idx == 0);
       idx++) {
    buf[idx] = value % 10;
    value /= 10;
  }

  for (idx--; idx >= 0; idx--) {
    hash = ((hash << 3) + hash);
    hash += buf[idx] + '0';
  }

  return hash;
} /* HashInt */


/*
 * HashURI()
 * Hash a uri_t structure into an unsigned long.  The value hashed must
 * be the same if two uri's compare as equal according to
 * URI_Equal_With_Defaults, regardless of their actual form.
 * Currently we accomplish this by calling URI_Canonicalize, URI_SPrint, and
 * hashing the resulting string.
 */
static unsigned long HashURI(unsigned long start_hash, uri_t *uri)
{
  unsigned long hashval = start_hash;

  uri_t canon;
  DString dstr;

  canon = URI_Copy(*uri);
  URI_Canonicalize(&canon, sip_uri_defaults);

  DStringInit(&dstr);
  URI_SPrint(&canon, &dstr);

  hashval = HashString(hashval, DStringValue(&dstr));

  DStringFree(&dstr);
  URI_Free(&canon);

  return hashval;
} /* HashURI */


/*
 * ComputeRequestHash()
 * Compute a request's identifying hash.  Base this on its
 * Request-URI, To, From, Call-ID, and CSeq headers.  Used for loop
 * detection and for fast corrolation of retransmissions and responses.
 */
unsigned long ComputeRequestHash(request_t *r)
{
  unsigned long hashval = 0;

  /* Request-URI */
  hashval = HashURI(hashval, &r->p.URI);
  hashval = HashString(hashval, ":");

  /* To */
  hashval = HashURI(hashval, &r->to.uri);

  hashval = HashString(hashval, ":");

  /* From */
  hashval = HashURI(hashval, &r->from.uri);
  hashval = HashString(hashval, ":");

  /* Call-ID */
  hashval = HashString(hashval, r->callID);
  hashval = HashString(hashval, ":");

  /* CSeq */
  hashval = HashInt(hashval, r->cseq.seq);

  /* Reserve hashval == 0 for special cases */
  if (hashval == 0) {
    hashval = 1;
  }

  return hashval;
} /* ComputeRequestHash */


/*
 * SetRequestVia()
 * Set r->via appropriately, based on the topmost via and the address from
 * which the request was received.  Remember the value for a "received" tag
 * to add to the topmost via if appropriate.
 */
static void SetRequestVia(request_t *r)
{
  r->via          = SIP_Via(&r->p, 1, 1, config.ProxyName, (unsigned short)config.port);

 if(!r->type == SOCK_SSL) {
  if (r->via.hostname == NULL) {
    /* Missing or invalid Via header, fall back to default. */
    debug("SetRequestVia",
      "r = 0x%p: r->via.hostname == NULL\n", r);    
    r->via.sin = r->sin.tcp_s;
    r->via.received = r->sin.tcp_s.sin_addr;
  }
  if (r->via.sin.sin_addr.s_addr != r->sin.tcp_s.sin_addr.s_addr) {
    /* Via has wrong host addr. */
    /* Specify actual addr so we an add "received" tag to outgoing requests */
    debug("SetRequestVia",
          "r = 0x%p: topmost via header does not match received addr (%s/%s)\n",
          r, r->via.hostname, inet_ntoa(r->sin.tcp_s.sin_addr));
    r->via.sin.sin_addr = r->sin.tcp_s.sin_addr;
    r->via.received = r->sin.tcp_s.sin_addr;
  }
  r->sin.tcp_s          = r->via.sin;  /* All other attributes of response addr
                                    (notably port) come from Via. */
 }
} /* SetRequestVia */


/*
 * RequestIsNotLooped()
 * Check if this request has been looped -- i.e., if our proxy name appears
 * in its Via headers, and the hash part of that Via header's branch ID
 * corresponds to this request's request-hash.  Return false if there is a
 * loop, i.e. if this request should no longer be processed.
 */
static int RequestIsNotLooped(request_t *r) 
{
  sip_via_t via;
  int idx;
  int ret = 1;

  /* Check for Via loops */
  for (idx = 1, via = SIP_Via(&r->p, idx, 0, config.ProxyName, (unsigned short)config.port);
       via.hostname != NULL;
       idx++, via = SIP_Via(&r->p, idx, 0, config.ProxyName, (unsigned short)config.port)) {
    if (via.islocal && (via.branch.hash == r->hash)) {
      /* Via loop! */
      /* XXX: check for hash collisions */
      debug("RequestIsNotLooped",
            "r = 0x%p: hash %lu collides, request is looped\n", r, r->hash);

      ResponseSetStatus(r, 482, NULL);
      SendFailureResponse(r, 0);

      ret = 0;
      break;
    }
    free_valid(via.hostname);
    free_valid(via.branch.str);
  }

  return ret;
} /* RequestIsNotLooped */


/*
 * SetEarlyRequestData()
 * Extract some data from r->p into fields of r.
 * These items are needed for ResponseBuildDefault or ResponseSend -- set
 * them early so they're set for failure responses.
 */
static void SetEarlyRequestData(request_t *r)
{
  r->cseq         = SIP_CSeq(&r->p);
  strassign(&r->callID, SIP_CallID(&r->p));
  SetRequestVia(r);
  r->version      = HTTP_Version(&r->p);

  return;
} /* SetEarlyRequestData */


/*
 * SetLateRequestData()
 * Extract the rest of the data -- that not extracted by
 * SetEarlyRequestData() -- needed by r out of r->p.
 */
static void SetLateRequestData(request_t *r)
{
  parser_header_t *h;

  r->contact      = SIP_Contact(&r->p);
  r->expires      = SIP_Expires(&r->p);

  r->to           = SIP_To(&r->p);
  r->from         = SIP_From(&r->p);

  r->hash         = ComputeRequestHash(r);

  if ((h = RFC822_Header(r->p.headers, "Content-Type")))
    r->request_body.content_type = h->elements->value;

  return;
} /* SetLateRequestData */


/*
 * HandleDuplicateRequest()
 * This request is a "duplicate" of another request that already exists in
 * the request list -- it has the same request hash value.  It may be
 * an ACK or CANCEL of that request, or it may be a true retransmission.
 * Handle appropriately, and clean up the request state when done.
 */
static void HandleDuplicateRequest(request_t *r, request_t *rp)
{
  char *method, *pmethod;

  method = HTTP_Method(&r->p);
  pmethod = HTTP_Method(&rp->p);

  debug("HandleDuplicateRequest", "Request 0x%p: duplicate of 0x%p\n",
        r, rp);

  /* If the top Vias of r and rp don't match, we have request merging --
   * send a 485. */ 
  if (!SIP_ViaEqual(&r->via, &rp->via)) {
    debug("HandleDuplicateRequest",
	  "Request 0x%p: not a duplicate of 0x%p (different top Via) "
	  "(method %s)\n",
	  r, rp, method);
    if (strcasecmp(method, "ACK") == 0) {
      r->status = 0; /* Don't send a response to a stray, merged ACK */
    }
    else {
      /* Otherwise report request merging */
      ResponseSetStatus(r, 482, "Request merging");
      ResponseBuildDefault(r);
    }
  }
  else if (strcasecmp(method, "ACK") == 0) {
    /* Method is ACK */
    if (rp->sent_200) {
      /* This request (which is probably behing handled by
       * proxy_policy) has sent a 200, so this ACK probably needs to be
       * handled end-to-end.  Hand it off to the policy core. */
      debug("HandleDuplicateRequest",
            "Request 0x%p: End-to-end ACK (sent 200)\n", r);
      MessageEnqueue(&rp->mqueue, message_real_message, &r->p,
                     &r->request_body, r->source_sin,NULL,
                     nobody_handled);
      /*
       * Note: Do *NOT* set rp->acked here for proxy.  It suppresses response
       * retransmissions, which is wrong if the ACK is an end-to-end
       * ACK.
       * But setting rp->acked is needed for SIP UA because otherwise the
       * policy thread waiting in ResponseSend() won't know that
       * ack has been received.
       * A Callback is defined to implement the selective behaviour.
       */
      SIP_OnEndToEndAck(rp);
    }
    else {
      /* Just set the flag. */
      debug("HandleDuplicateRequest",
            "Request 0x%p: Hop-by-hop ACK (haven't sent 200)\n", r);
      rp->acked = 1;
    }
    pthread_cond_broadcast(&rp->cond);

    /* Don't send a response to an ACK. */
    r->status = 0;
  }
  else if (strcasecmp(method, "CANCEL") == 0) {
    /* Method is CANCEL. */
    debug("HandleDuplicateRequest", "Request 0x%p: CANCEL 0x%p\n",
          r, rp);

    /* Hand it up to the policy core. */
    MessageEnqueue(&rp->mqueue, message_real_message, &r->p,
                   &r->request_body, r->source_sin, NULL, nobody_handled);
    pthread_cond_broadcast(&rp->cond);
    
    /* Send the CANCEL response here, regardless of what the policy
       does with it. */
    ResponseSetStatus(r, 200, NULL);
    ResponseBuildDefault(r);
  }
  else {
    /* Method is BYE, INVITE, OPTIONS, REGISTER, or whatever. */
    /* If we get a retransmitted request with a different method,
     * complain. */
    if (strcasecmp(method, pmethod) != 0) {
      ResponseSetStatus(r, 400,
                        "Method in retransmission does not match "
                        "initial request");
      ResponseBuildDefault(r);
    }
    else {
      debug("HandleDuplicateRequest", "just copying response\n");
      ResponseCopy(r, rp);
    }
  }

  pthread_mutex_unlock(&rp->mutex);
  if (r->status > 0) {
    pthread_mutex_lock(&r->mutex);
    ResponseSend(r, 0, 1); /* Set cleanup status to 1 */
  }
  else {
    RequestFree(r);
  }

  free_valid(method);
  free_valid(pmethod);

  return;
} /* HandleDuplicateRequest */


/*
 * HandleNewRequest()
 * A new request -- one for which a parent wasn't found in the tables --
 * has arrived.  Dispatch it appropriately.  Clean up if necessary.
 */
static void HandleNewRequest(request_t *r)
{
  char *method;

  method = HTTP_Method(&r->p);

  if (strcmp(method, "ACK") == 0) {
    /* ACK of an unknown request. */
    /* Call the application specific processing for ACK.
     * sipd ignores it, but siplib will try to match with an
     * existing call leg.  (The ACK could have been sent end-to-end
     * to this UA, and RequestSearch() wouldn't have returned 
     * the correct matching call leg.)
     */
    SIP_OnIncomingAck(r);
    /* Request is freed by OnIncominAck */
  }
  else if (strcmp(method, "CANCEL") == 0) {
    if (!config.TryStateless) {
      /* CANCEL of an unknown request */
      debug("RequestProcessThread",
            "Request 0x%p: CANCEL of unknown request\n", r);
      
      ResponseSetStatus(r, 481, NULL);
      ResponseBuildDefault(r);
      ResponseSend(r, 0, 1); /* Set cleanup status to 1 */
    }
    else {
      pthread_mutex_lock(&r->mutex);
      /* No need to add in the list */
      SIP_OnIncomingRequest(r);
    }
  }
  else {
    /* 
     * Method is BYE, INVITE, OPTIONS, REGISTER, etc.  Insert into the
     * request list and handle as appropriate.  Once the request
     * goes on the request list it's visible to other threads, so it
     * must be locked before handing it to SIP_OnIncomingRequest().
     */
    pthread_mutex_lock(&r->mutex);
    debug("RequestProcessThread",
          "Request 0x%p: Adding request to list\n", r);
    RequestAdd(r);

    SIP_OnIncomingRequest(r);
  }

  free_valid(method);
} /* HandleNewRequest */


/*
 * RequestProcessThread()
 * Process a SIP request according to the request type. 
 * This is the thread which actually does the processing work.
 * The return value is irrelevant and should be NULL; pthread_create
 * requires a thread's start routine to return a pointer.
 */
static void *RequestProcessThread(void* ptr)
{
  char *method;                /* method of request */
  request_t *rp;               /* matching earlier request */
  request_t *r = (request_t *) ptr;

 
  if (SnmpProcessMessage) {
    (*SnmpProcessMessage)(r, SNMP_TYPE_REQ_IN, NULL);
  }

  SetEarlyRequestData(r);

  debug("RequestProcessThread",
        "Starting to process request 0x%p, callid %s\n",
        r, r->callID != NULL ? r->callID : "NULL");

  if (!ValidateIncomingRequest(r)) {
    return NULL;
  }

  SetLateRequestData(r);

  ResponseSetStatus(r, 100, NULL);

  method = HTTP_Method(&r->p);

  if (!RequestIsNotLooped(r)) {
    return NULL;
  }

  DRAW_MESSAGE_FLOW({
    DrawMessageFlow("SIPAPI", "Remote", "This", method);
  });
 
  /* Does this match an existing request? */
  rp = RequestSearch(r, r->hash);
  debug("RequestProcessThread", "Request r=0x%p: rp=0x%p\n", r, rp);

  if (rp) {
    HandleDuplicateRequest(r, rp);
  }
  else {
    HandleNewRequest(r);
  }
  free_valid(method);
  return NULL;
} /* RequestProcessThread */

  
/*
 * RequestOverloadBailout()
 * If the server is overloaded and pthread_create failed, do the
 * absolute minimum necessary to send back a 500 response to a
 * non-ACK request.  For an ACK or a response, just free the structure.
 */
static void RequestOverloadBailout(request_t *r)
{
  if (r->p.type != http_request || 
      strcasecmp(r->p.method, "ACK") == 0) {
    RequestFree(r);
    return;
  }

  /* Copy the minimal set of mandatory headers needed
   * for ResponseBuildDefault and ResponseSend to work.*/
  SetEarlyRequestData(r);
      
  ResponseSetStatus(r, 500, "Server overloaded");
      
  /* This will free the request */
  SendFailureResponse(r, 0);
} /* RequestOverloadBailout */


/* 
 * RequestProcess()
 * Process a SIP request or response; create the appropriate thread to do
 * the actual processing work.
 */
void RequestProcess(request_t *r)
{
  pthread_t th;                 /* id of new thread */
  int thr_err;                  /* thread error indication */
  void* (*thread_func)(void *); /* Function to start */
  
  int usepool = 0;              /* Whether to use the thread pool? */
  char *method;
  
  printf("*** RequestProcess is called ***\n");
  if (r->p.type == http_request) {
    thread_func = RequestProcessThread;
    
    /* Try to use the thread pool if config.UseThreadPool is 1
     * and the incoming method is REGISTER.
     */
    if(config.UseThreadPool) {
	method = HTTP_Method(&r->p);
	if(method && strcasecmp(method,"REGISTER") == 0) 
	  usepool = 1;
    }
  } else if (r->p.type == http_response) {
    thread_func = ResponseProcessThread;
  } else {
    /* If it's neither a request nor a response, just ignore it. */
    debug("RequestProcess", "Unknown message (neither request nor response\n");
    RequestFree(r);
    return;
  }

 
  if (!usepool) {
    debug("RequestProcess", "Creating processor thread\n");
    thr_err =
      pthread_create_detached(&th, NULL, thread_func, (void *)r);
  }  
  else {
    debug("RequestProcess", "Getting a pool thread\n");
    thr_err = 
      pool_schedule(thread_func, (void *) r);
  }

  if (thr_err != 0) {
    errno = thr_err;
    error("RequestProcess", "couldn't start thread");
    
    /* This will free the request */
    RequestOverloadBailout(r);
  }

} /* RequestProcess */


/*
 * hash table bookkeeping functions.
*/
static ht_val compute_requestkey_value(dict_key o) {
 
  /* o is a pointer to hash. key value is  *o 
   */
  return (unsigned long) *((unsigned long*)o);
  
} /* compute_requestkey_value */

static ht_val compare_requestkey_values(dict_key k1, dict_obj o1) {

  /* k1 is a pointer to hash. o1 is a pointer to request_t 
   */
  return (unsigned long) *((unsigned long*)k1) - 
    (unsigned long)compute_requestobj_value(o1);
  
} /* compare_requestkey_values */


static ht_val compare_requestobj_values(dict_obj o1, dict_obj o2) {
  
  request_t *r, *r1;
  unsigned long hash, hash1;

  r = (request_t*) o1;
  r1 = (request_t*) o2;
  
  /* o1, o2 are pointers to request_t. to compare them, we 
   * need their respective hashes. their hash fields may or
   * may not have the actual hash value. 
   */
  if (r && r1) {
    hash = (r->hash) ? r->hash : ComputeRequestHash (r);
    hash1 = (r1->hash) ? r1->hash : ComputeRequestHash (r1);
    return hash - hash1;
  }
  
  error("compare_requestobj_values","getting null params...\n");
  return -1;

} /* compare_requestobj_values */


static ht_val compute_requestobj_value(dict_obj o) {
  
  request_t *r = (request_t*) o;

  /* o is a pointer to request_t. its hash field may or may not be
   * set to correct value. 
   */
  if (r) 
    return (r->hash) ? r->hash : ComputeRequestHash(r);
  
  error("compute_requestobj_value","getting null params...\n");
  return 0;

} /* compute_requestobj_value */

/*
 * KillBranches.
 *
 * Kill all branch threads. This is an ugly function and should be
 * used under exceptional conditions (reason being it kills threads
 * without caring what state they are in). Before calling this thread,
 * make sure that the main Request and Response handler's are killed.
 */
void KillBranches() {
  
  request_t *rp;
  int b = 0;
  
  pthread_mutex_lock(&R_lock);

  rp = (request_t*) ht_minimum(request_table);
  
  while (rp!=NULL) {
    for (b=0; b<rp->branch_count; b++) {
      if (rp->branches[b].tid != 0) 
	pthread_cancel(rp->branches[b].tid);
    }
    
    rp = (request_t*) ht_successor(request_table, rp);
  }
  
  pthread_mutex_unlock(&R_lock);
  
  return;
} /* KillBranches */


/* KillLibSipThreads      
 *
 * Kills all libsip threads (in an unclean fashion). should be
 * used with extreme caution, only when abrupt libsip termination is needed. 
 */
void KillLibSipThreads() {
  
  KillDetachedThreads();
  
  KillBranches();
  
  return;
  
} /* KillLibSipThreads */
