/* sipd-config - Setup config global and read configuration file. */ 

/* Copyright 1998-2001 by Columbia University.  All rights reserved. */

#include <errno.h>
#include <netdb.h>     /* gethostbyname, etc. */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/stat.h>  /* mkdir() */
#include <regex.h>
#include <sys/param.h> /* MAXHOSTNAMELEN, MAXPATHLEN */
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "error.h"
#include "fullpath.h"
#include "log.h"
#include "strdupn.h"

#include "canonicalize.h"

#include "dbapi.h"

#include "config.h"
#include "gateways.h"
#include "sipd.h"

#include "sipd-config.h"

/* Prototypes for external libcine functions without their own header files.
 * (They really ought to have them.) */
extern uid_t uname2id(const char *name);
extern gid_t gname2id(const char *name);
extern void setusergroup(uid_t uid, gid_t gid);

/* Globals. */
struct commandline_t commandline;
struct config_t config;

/*
 * getcanonicalhostname()
 * Get and return the canonical hostname of this machine.
 * Not thread-safe -- only intended for startup time.
 */
char *getcanonicalhostname()
{
  char buf[MAXHOSTNAMELEN];
  struct hostent *hep;
  struct in_addr addr;

  addr.s_addr = INADDR_ANY;

  gethostname(buf, MAXHOSTNAMELEN);
  
  hep = gethostbyname(buf);
  if (hep == NULL) {
    error("getcanonicalhostname", "gethostbyname failed on %s, h_errno = %d\n",
          buf, h_errno);
    return NULL;
  }

  memcpy(&addr, hep->h_addr, sizeof(addr));

  hep = gethostbyaddr((char *)&(addr.s_addr), sizeof(addr.s_addr), AF_INET);
  if (hep == NULL) {
    error("getcanonicalhostname", "gethostbyaddr failed on %s, h_errno = %d\n",
          inet_ntoa(addr), h_errno);
    return NULL;
  }

  if (commandline.use_numeric_address == 0 &&
      hep->h_name && strchr(hep->h_name, '.') != NULL) {
     return strdup(hep->h_name);
  }
  else {
     return strdup(inet_ntoa(addr));
  }
} /* getcanonicalhostname */


/* 
 * compile_regex()
 * Compile a regular expression, storing it in the specified regex_t. 
 * It allocates the necessary memory and resets the value if the
 * compilation fails.  Because this can be run before the logfile is set
 * up, the error() call might do strange things.
 */
static void compile_regex(regex_t **compiled, char *pattern) {
  int retval;
  *compiled = (regex_t *) malloc(sizeof(regex_t));
  if (*compiled == NULL) return;

  retval = regcomp(*compiled, pattern, REG_EXTENDED|REG_NOSUB|REG_ICASE);
  if (retval) {
    char buf[1024];
    regerror(retval, *compiled, buf, sizeof(buf));
    error("compile_regex",
          "Regular expression \"%s\" compilation failed: %s\n",
          pattern, buf);
    regfree(*compiled);
    free_valid(*compiled);
    *compiled = NULL;
  }
} /* compile_regex */


/*
 * GetTrueFalse()
 * Get a true or false value from the string.  Print an error if it's
 * neither true nor false; in this case, return false anyway.
 */
static int GetTrueFalse(char *name, char *value)
{
  if (!strcmp(value, "true")) {
    return 1;
  }
  else if (!strcmp(value, "false")) {
    return 0;
  }
  else {
    /* print an error for any other value, but still continue */
    error("ConfigInit", "Invalid %s value %s; must be 'true' or 'false'",
          name, value);
    return 0;
  }
} /* GetTrueFalse */


/* 
 * ConfigUser()
 * Determine if user information is to be retrieved from a SQL Database or 
 * a flat text file.  If from a SQL Database, get all the necessary parameters
 * and create the connection to the SQL Database.
 */
static void ConfigUser()
{
  uri_t table_uri;
  db_err db_ret;
  char *db_ret_str;

  if (config.UserConfig == NULL) {
    fatal("ConfigUser", "No UserConfig configured\n");
  }


  /* New-style database configuration */
  config.database = new_cinema_table(config.UserConfig, &db_ret, &db_ret_str);
  if (config.database == NULL) {
    fatal("ConfigUser", "Could not open database %s: %s\n",
          config.UserConfig, db_ret_str);
  }

  /* Old-style database configuration */
  table_uri = URI_Parse(config.UserConfig);

  if (strcasecmp(table_uri.scheme, "sql") == 0 ||
      strcasecmp(table_uri.scheme, "fastsql") == 0) {
    char *password;
    password = table_uri.password;
    if ( password != NULL && !strcmp(password,"NULL")) {
      password = NULL;
    }
    if (initializeConnection(table_uri.user, password,
                             table_uri.path,
                             table_uri.host, table_uri.port) != 0) {
      fatal("ConfigUser", "Could not open database %s\n", config.UserConfig);
    }
  }
  else {
    fatal("ConfigUser",
          "Unknown database scheme %s in UserConfig pseudo-URL %s\n",
          table_uri.scheme, config.UserConfig);
  }

  URI_Free(&table_uri);

} /* ConfigUser */


/*
 * SetupConfigDefaults()
 * Setup the default values for config options.
 */
static void SetupConfigDefaults(void)
{
  config.ServerRoot   = NULL;
  config.AuthMethod   = auth_none;
  config.PrivateKey   = NULL;
  config.Realm        = getcanonicalhostname();
  config.LogCondition = strdup("");
  config.LogFormat    =
    strdup("%h %u %t \"%r\" %s To=%{To}i "
           "From=%{From}i Call-ID=%{Call-ID}i %T");
  config.ErrorLog     = strdup("logs/error_log");
  config.Canonicalize = NULL;
  config.PidFile      = strdup("logs/sipd.pid");
  config.ScriptBase   = strdup("scripts");
  config.ServerName   = strdup("");
  config.UserConfig   = NULL;
  config.expires      = 3600;  /* one hour */
  config.maxexpires   = 86400; /* 24 hours */
  config.domain       = NULL;
  config.uid          = -1;
  config.gid          = -1;
  config.port         = default_sip_port;
  config.SSLport      =  0; /* SSL is not chosen */
  config.timeout      = 30;
  config.cgi_timeout  = 15;
  config.HelpResolve  = 0; /* by default, doesn't help resolve DNS name. */
  config.foreign      = L_reject;
  config.defaultreg   = L_redirect;
  config.DisplayAmbiguous = 0; /* by default doesn't display contacts on
                                  ambiguous response */
  config.ProxyName    = getcanonicalhostname();
  config.ProxyRecursion = 0; /* by default don't recurse on 3xx's while
                                proxying. */
  config.ThirdPartyReg = 1; /* Allow third parties to register */
  config.RegRespContacts = 1; /* contact info is sent */
#ifdef USE_PGP_SDK
  config.PgpPubRing   = strdup("pubring.pkr");
  config.PgpSecRing   = strdup("secring.skr");
  config.PgpPvtPass   = strdup("thePassPhrase");
#endif
  config.NonceLifeTime = 60; /* seconds */
  config.ClockSkew    = 300; /* +- seconds */
  config.GatewayMap   = NULL;
  config.EnumRoot     = NULL;
  config.database     = NULL;
  config.TryStateless = 0;
  config.ThreadPoolSize = 128;
  config.UseThreadPool = 1;
  config.RecordRoute = 1;
} /* SetupConfigDefaults */


/*
 * ReadConfigFile()
 * Open the configuration file, and read its contents into the 'config'
 * global.
 * This function should *not* invoke any actions based on the content of the
 * file, since there may be inter-dependencies among config options.  Instead,
 * invoke them from DoConfigFile(), below, instead.
 */
static void ReadConfigFile(void)
{
  FILE *f;                     /* configuration file */
  int argc;                    /* number of arguments */
  char *argv[10];              /* words in configuration line */
  int i;                       /* loop variable */

  debug("ReadConfigFile", "Reading configuration from %s\n",
        commandline.ConfigFile); 
  if ((f = fopen(commandline.ConfigFile, "r")) == NULL) {
    fatal("ReadConfigFile",
          "Could not open configuration file %s in directory %s",
          commandline.ConfigFile, getcwd(NULL, MAXPATHLEN));
  }

  argc = 5;

  while (ConfigRead(f, &argc, argv)) {
    if (argc < 2) {
    }
    else if (!strcmp(argv[0], "AuthMethod")) {
      config.AuthMethod = HTTP_AuthenticationMethod(argv[1]);
      if (config.AuthMethod == auth_unknown &&
          strcasecmp(argv[1], "none") == 0) {
        config.AuthMethod = auth_none;
      }
    }
    else if (!strcmp(argv[0], "Canonicalize")) {
      strassign(&config.Canonicalize, argv[1]);
    }
    else if (!strcmp(argv[0], "PrivateKey")) {
      strassign(&config.PrivateKey, argv[1]);
    }
    else if (!strcmp(argv[0], "Domain")) {
      compile_regex(&config.domain,argv[1]);
    }
    else if (!strcmp(argv[0], "ForeignDomain")) {
      if      (!strcmp(argv[1], "proxy")) config.foreign = L_proxy;
      else if (!strcmp(argv[1], "redirect")) config.foreign = L_redirect;
      else if (!strcmp(argv[1], "reject")) config.foreign = L_reject;
      else fatal("ConfigInit", "Invalid ForeignDomain value %s", argv[1]);
    }
    else if (!strcmp(argv[0], "DefaultReg")) {
      if      (!strcmp(argv[1], "proxy")) config.defaultreg = L_proxy;
      else if (!strcmp(argv[1], "redirect")) config.defaultreg = L_redirect;
      else fatal("ConfigInit", "Invalid DefaultReg value %s", argv[1]);
    }
    else if (!strcmp(argv[0], "ErrorLog")) {
      strassign(&config.ErrorLog, argv[1]);
    }
    else if (!strcmp(argv[0], "Expires")) {
      config.expires = atoi(argv[1]);
    }
    else if (!strcmp(argv[0], "MaxExpires")) {
      config.maxexpires = atoi(argv[1]);
    }
    else if (!strcmp(argv[0], "Group")) {
      config.gid = gname2id(argv[1]);
    }
    else if (!strcmp(argv[0], "Port")) {
      config.port = atoi(argv[1]);
    }
    else if (!strcmp(argv[0], "SSLPort")) {
      config.SSLport = atoi(argv[1]);
    }
    else if (!strcmp(argv[0], "PidFile")) {
      strassign(&config.PidFile, argv[1]); 
    }
    else if (!strcmp(argv[0], "ScriptBase")) {
      strassign(&config.ScriptBase, argv[1]);
    }
    else if (!strcmp(argv[0], "LogCondition")) {
      strassign(&config.LogCondition, argv[1]);
    }
    else if (!strcmp(argv[0], "LogFormat")) {
      strassign(&config.LogFormat, argv[1]);
    }
    else if (!strcmp(argv[0], "RequestLog")) {
      if (argc < 3) {
        AddLog(argv[1], config.LogFormat, config.LogCondition);
      }
      else if (argc < 4) {
        AddLog(argv[1], argv[2], 0);
      }
      else {
        AddLog(argv[1], argv[2], argv[3]);
      }
    }
    else if (!strcmp(argv[0], "ServerRoot")) {
      strassign(&config.ServerRoot, argv[1]);
    }
    else if (!strcmp(argv[0], "ServerName")) {
      strassign(&config.ServerName, argv[1]);
    }
    else if (!strcmp(argv[0], "TimeOut")) {
      config.timeout = atoi(argv[1]); 
    }
    else if (!strcmp(argv[0], "CGITimeout")) {
      config.cgi_timeout = atoi(argv[1]);
    }      
    else if (!strcmp(argv[0], "User")) {
      config.uid = uname2id(argv[1]);
    }
    else if (!strcmp(argv[0], "HelpResolve")) {
      config.HelpResolve = GetTrueFalse("HelpResolve", argv[1]);
    }
    else if (!strcmp(argv[0], "DisplayAmbiguous")) {
      config.DisplayAmbiguous = GetTrueFalse("DisplayAmbiguous", argv[1]);
    }
    else if (!strcmp(argv[0], "ProxyName")) {
      strassign(&config.ProxyName, argv[1]);
    }
    else if (!strcmp(argv[0], "Realm")) {
      strassign(&config.Realm, argv[1]);
    }
    else if (!strcmp(argv[0], "ProxyRecursion")) {
      config.ProxyRecursion = GetTrueFalse("ProxyRecursion", argv[1]);
    }
    else if (!strcmp(argv[0], "ThirdPartyReg")) {
      config.ThirdPartyReg = GetTrueFalse("ThirdPartyReg", argv[1]);
    }
#ifdef USE_PGP_SDK
    else if (!strcmp(argv[0], "PgpPubRing")) {
      strassign(&config.PgpPubRing, argv[1]);
    }
    else if (!strcmp(argv[0], "PgpSecRing")) {
      strassign(&config.PgpSecRing, argv[1]);
    }
    else if (!strcmp(argv[0], "PgpPvtPass")) {
      strassign(&config.PgpPvtPass, argv[1]);
    } 
#endif
#ifdef USE_SNMP
    else if (!strcmp(argv[0], "AgentxSocketPath")) {
      strassign(&config.AgentxSocketPath, argv[1]);
    }
    else if (!strcmp(argv[0], "StartSnmp")) {
      config.StartSnmp = GetTrueFalse("StartSnmp", argv[1]);
    }
#endif
    else if (!strcmp(argv[0], "NonceLifeTime")) {
      int i = atoi(argv[1]);
      if (i >= 0) {
	config.NonceLifeTime = atoi(argv[1]);
      }
    }
    else if (!strcmp(argv[0], "ClockSkew")) {
      int i = atoi(argv[1]);
      if (i >= 0) {
	config.ClockSkew = atoi(argv[1]);
      }
    }
    else if (!strcmp(argv[0], "GatewayMap")) {
      strassign(&config.GatewayMap, argv[1]);
    }
    else if (!strcmp(argv[0], "EnumRoot")) {
      strassign(&config.EnumRoot, argv[1]);
    }
    else if (!strcmp(argv[0], "UserConfig")) {
      strassign(&config.UserConfig, argv[1]);
    }
    else if (!strcmp(argv[0], "TryStateless")) {
      config.TryStateless = GetTrueFalse("TryStateless", argv[1]);
    }
    else if (!strcmp(argv[0], "ThreadPoolSize")) {
      int i = atoi(argv[1]);
      if (i > 128) {  /* Use a minimum of 128 threads. */
	config.ThreadPoolSize = atoi(argv[1]);
      }
    }
    else if (!strcmp(argv[0], "UseThreadPool")) {
      config.UseThreadPool = GetTrueFalse("UseThreadPool", argv[1]);
    }
    else if (!strcmp(argv[0], "RecordRoute")) {
      config.RecordRoute = GetTrueFalse("RecordRoute", argv[1]);
    }
  

    for (i = 0; i < argc; i++) free(argv[i]);
    argc = 5;
  }
  fclose(f);  /* close configuration file */
} /* ReadConfigFile */


/*
 * DoConfigFile()
 * For items in the configuration file which require actions to be taken,
 * at startup/reload time, invoke them.
 */
static void DoConfigFile(void)
{
  FILE *f;                     /* PID file */
  char *tmp;

  /*
   * Set the actual ServerRoot.
   *
   * If the ServerRoot in the config file is a relative pathname, then it is
   * relative to the directory specified on the command line with -s.  If
   * *that* is a relative pathname, or if it was not specified, it is
   * relative to the directory in which sipd was invoked.
   *
   * If neither -s nor the ServerRoot directive was used, DEFAULT_SERVER_ROOT
   * (normally /usr/local/sipd/) is the server root.  */
  
  if (config.ServerRoot == NULL) {
    if (commandline.ServerRoot == NULL) {
      strassign(&config.ServerRoot, DEFAULT_SERVER_ROOT);
    }
    else {
      strassign(&config.ServerRoot, commandline.ServerRoot);
    }
  }
  else if (commandline.ServerRoot != NULL) {
    /* The actual ServerRoot is relative to the commandline.ServerRoot.
     * This is a no-op (other than a strdup) if ServerRoot is an absolute
     * path. */
    tmp = config.ServerRoot;
    strassign(&config.ServerRoot,
              fullpath(commandline.ServerRoot, tmp));
    /* tmp is freed by the strassign. */
  }

  /* Change directories to the deduced ServerRoot.
   * All relative pathnames in the config file are relative to this directory.
   */
  debug("DoConfigFile", "Changing directory to %s\n", config.ServerRoot);
  if (chdir(config.ServerRoot) != 0) {
    fatal("DoConfigFile", "Couldn't change directory to %s (from %s)\n",
          config.ServerRoot, getcwd(NULL, MAXPATHLEN));
  }

  if (!config.domain) {
    compile_regex(&config.domain, config.ServerName);
  }

  if (strcmp(config.ErrorLog, "syslog") == 0) ErrorInit(0);
  /* without the line below, we can't force log to use stderr */
  else if (strcmp(config.ErrorLog, "stderr") == 0) ErrorInit("stderr");
  else ErrorInit(config.ErrorLog);

  debug("DoConfigFile", "Writing pid file %s\n", config.PidFile);

  if (!(f = fopen(config.PidFile, "w"))) {
    fatal("ConfigInit", "Could not write pid file %s", config.PidFile);
  }
  else {
    fprintf(f, "%d\n", (int)getpid());
    fclose(f);
  }

  if (config.GatewayMap != NULL) {
    if (!(init_gateways(config.GatewayMap))) {
      fatal("ConfigInit",
            "Could not initialize gateway map from map file %s",
            config.GatewayMap);
    }
  }

  /* Open request logs */
  OpenLogs();

  if (config.UserConfig != NULL) {
    ConfigUser(); /* determine if SQL database or flat file */
  }

  /* Make sure that, if we're running as root, "User" and "Group" have been
   * set. */
  if (geteuid() == 0 && (config.uid == -1 || config.gid == -1)) {
    fatal("DoConfigFile",
          "Running as UID 0, but User and Group are not set in %s.",
          commandline.ConfigFile);
  }

  /* Do this after opening files. Probably can't be reversed. */
  setusergroup(config.uid, config.gid);

  /* Don't execute canonicalize as root, though */
  if (canonicalize_init() != 0) {
    fatal("ConfigInit", "Error executing canonicalizer \"%s\"",
          config.Canonicalize != NULL ? config.Canonicalize
          : "(no file name)");
  }
  

} /* DoConfigFile */


/*
 * ConfigInit()
 * Read configuration file.
 * We open the files (configuration, logs) as the user starting the
 * daemon (typically, root) and then switch to the designated user and
 * group.  User location programs are run as that user rather than root.
 * This function is invoked when SIGHUP is received, to re-load the file.
 * (It's not clear that this is thread-safe, though!)
 */
void ConfigInit(int sig)
{
  LogClose();  /* close old logs */

  SetupConfigDefaults();

  ReadConfigFile();

  DoConfigFile();
} /* ConfigInit */


/*
 * ConfigCleanup()
 * Clean up before exiting.
 */
void ConfigCleanup(void)
{
  free_valid(config.Canonicalize);
  free_valid(config.ErrorLog);
  free_valid(config.PidFile);
  free_valid(config.ScriptBase);
  free_valid(config.ServerName);
  free_valid(config.ServerRoot);
  free_valid(config.LogFormat);
  free_valid(config.LogCondition);
#ifdef USE_PGP_SDK
  free_valid(config.PgpPubRing);
  free_valid(config.PgpSecRing);
  free_valid(config.PgpPvtPass);
#endif /* USE_PGP_SDK */
  free_valid(config.PrivateKey);
  free_valid(config.ProxyName);
  free_valid(config.Realm);
  if (config.domain) regfree(config.domain);
  free_valid(config.domain);
  free_valid(config.UserConfig);

  if (config.GatewayMap) {
    free_gateways();
  }
  free_valid(config.GatewayMap);

  /* close old-style SQL Connection (if connection exists) */
  closeConnection(); 
  /* Close new-style database connection */
  if (config.database != NULL) {
    free_table(config.database);
  }

  memset(&config, 0, sizeof(config));

  /* Terminate canonicalizer */
  canonicalize_terminate();

#ifdef USE_PGP_SDK
/*  ===== BEGIN DeInitialize PGPsdk ===== */
  EnginePgpDeInit();
/*  ===== END Initialize PGPsdk ===== */
#endif /* USE_PGP_SDK */
  FreeNonceList(gNonceHead.first);
  gNonceHead.first=NULL;
  gNonceHead.last=NULL;

  LogClose();
  ErrorClose();
} /* ConfigCleanup */


