/*
* SIP test client: 
* [-s host] [-c callid] [-m method] -a [-p clientport] [-P serverport]
*      [-s server] [-u] [-x] [-l count] [-n] file
* -u uses UDP.
* -x uses SSL.
* The client port is chosen dynamically unless specified.
* Sends file1, then waits for response, then sends file2, etc.
* Fills in 
*   $H: own hostname
*   $I: own IP address
*   $T: transport protocol
*   $P: client receive port
*   $M: method, from -m.
*   $C: call-id (local), from -c
*   $B: body, from -b
*   $L: content length, computed from -b
*   $S: server name passed from command line
*
* Does not implement reliability.
*/

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <arpa/inet.h>   /* inet_addr() */
#include <stdlib.h>      /* getopt() */
#include <stdio.h>       /* fprintf() */
#include <string.h>      /* strlen(), strcpy(), memcpy() */
#include <unistd.h>      /* open(), read(), close() */
#include <fcntl.h>       /* open() */
#include <netdb.h>       /* gethostbyname */

#include "host2ip.h"
#include "ssl_common.h"
#include "ssl_client.h"

#define SOCK_SSL SOCK_DGRAM + SOCK_STREAM +1

static void usage(char *prog)
{
  fprintf(stderr, "SIP test client %s (c) Columbia University\n", VERSION);
  fprintf(stderr, "Usage: %s [-a] [-c callid] [-p port] [-m method]\n", prog);
  fprintf(stderr, "\t[-s server] [-b body] [-l loops] [-u] [-n] file\n");
  fprintf(stderr, "\
 -a          send ACK after final response\n\
 -c callid   Call-ID value\n\
 -m method   method (INVITE, ACK, ...)\n\
 -p port     server port (default 5060)\n\
 -P port     client port (default 5060) (UDP only)\n\
 -s server   server address (default localhost)\n\
 -l loops    number of times to repeat (default 1)\n\
 -u          use UDP instead of TCP\n\
 -x          use SSL instead of TCP or UDP\n\
 -n          print only the final response status and exit or timeout after 2 seconds\n\
");
  exit(1);
} /* usage */


/*
 * Get and return the (an) IP address of this machine.
 */
struct in_addr getlocalip()
{
  char buf[MAXHOSTNAMELEN];
  struct hostent *hep;
  struct in_addr addr;

  addr.s_addr = INADDR_ANY;

  gethostname(buf, MAXHOSTNAMELEN);
  
  hep = gethostbyname(buf);
  if (hep == NULL) {
    fprintf(stderr, "getlocalip: gethostbyname failed on %s, h_errno = %d\n",
            buf, h_errno);
  }


  memcpy(&addr, hep->h_addr, sizeof(addr));

  return addr;
}


/* Create outgoing packet. */
static void create_packet(char *template, char *packet, 
                          char *method, int port, int sock_type, char *callid,
                          char *body, int body_length, char *server)
{
  char *s, *d;

  for (s = template, d = packet; *s; s++) {
    if (*s == '$') {
      if (*(s+1) == 'M') {
        strcpy(d, method);
      }
      else if (*(s+1) == 'C') {
        strcpy(d, callid);
      }
      else if (*(s+1) == 'S') {
        strcpy(d, server);
      }
      else if (*(s+1) == 'H') {
        gethostname(d, 1000);
      }
      else if (*(s+1) == 'I') {
        strcpy(d, inet_ntoa(getlocalip()));
      }
      else if (*(s+1) == 'P') {
        sprintf(d, "%d", port);
      }
      else if (*(s+1) == 'T') {
        strcpy(d, sock_type == SOCK_DGRAM ? "UDP" : "TCP");
      }
      else if (*(s+1) == 'B') {
        if (body != NULL) {
          memcpy(d, body, body_length);
          d += body_length;
        }
        *d = '\0';
      }
      else if (*(s+1) == 'L') {
        sprintf(d, "%d", body_length);
      }
      else {
        sprintf(d, "$%c", *(s+1));
      }
      d += strlen(d);
      s++;
    }
    else {
      *d++ = *s;
    }
  }
  *d = '\0';
} /* create_packet */


int main(int argc, char *argv[])
{
  int osock;                   /* request socket */
  int isock;                   /* response socket */ 
  struct sockaddr_in sin, from;
  SSL_LINK *slink;

  int fromlen;
  char template[8192];         /* request template read from file */
  char packet[8192];           /* packet to be sent/received */
  char body[8192];             /* Body to (possibly) be read */
  int serverport = 5060;       /* destination (server) port */
  int clientport = 0;          /* local (UDP client) port for responses */
  int sock_type = SOCK_STREAM; /* TCP socket */
  int c;                       /* character */
  int n;                       /* bytes read */
  int body_length = 0;         /* length of body */
  char *method = "INVITE";     /* request method */
  char *server = "localhost";  /* name of server host */
  char *callid = "1234";       /* Call-ID */
  char *body_file = NULL;      /* File to read the body from */
  int ack = 0;                 /* flag: send ACK */
  int status = 0;              /* response status */
  int fd;                      /* file for template */
  struct timeval now;
  char *orig_callid;
  int loops = 0;
  int only_status = 0;
  fd_set fdset;
  struct timeval tv;

#ifdef WIN32
  /* Since not defined in stdlib.h; implemented in ntutils.lib(getopt.c) */
  extern char *optarg;
  extern int optind;
#endif

  while ((c = getopt(argc, argv, "ab:c:l:m:p:P:s:uxn")) != EOF) {
    switch (c) {
    case 'a':
      ack = 1;
      break;
    case 'b':
      body_file = optarg;
      break;
    case 'c':
      callid = optarg;
      break;
    case 'l':
      loops = atoi(optarg);
      break;
    case 'm':
      method = optarg;
      break;
    case 'p':
      serverport = atoi(optarg);
      break;
    case 'P':
      clientport = atoi(optarg);
      break;
    case 's':
      server = optarg;
      break;
    case 'u':
      sock_type = SOCK_DGRAM;
      break;
    case 'x':
      sock_type = SOCK_SSL;
      printf("*** Use SSL ***\n");
      printf("sock_type=%d\n", sock_type);
      break;
    case 'n' :
      only_status = 1;
      break;
    case '?':
      usage(argv[0]);
      break;
    }
  }

  if (argc <= optind) {
    usage(argv[0]);
  }

  /* Read file. */
  if ((fd = open(argv[optind], O_RDONLY)) < 0) {
    perror(argv[optind]);
    exit(1);
  }
  if ((n = read(fd, template, sizeof(template)-1)) <= 0) {
    perror(argv[optind]);
    exit(1);
  }
  template[n] = '\0';
  close(fd);

  /* Read body file if specified. */
  if (body_file != NULL) {
    if ((fd = open(body_file, O_RDONLY)) < 0) {
      perror(body_file);
      exit(1);
    }
    if ((body_length = read(fd, body, sizeof(body)-1)) <= 0) {
      perror(body_file);
      exit(1);
    }
    body[body_length] = '\0';
    close(fd);
  }

  if(sock_type == SOCK_SSL) {
     printf("*** Connect SSL ***\n");
     slink = tcp_connect(server, serverport);
  } else {
    if (sock_type == SOCK_STREAM) {
       /* Let the OS choose the local port # for this. */
       clientport = 0;
    }

    /* Create socket for incoming requests. */
    if ((isock = socket(PF_INET, sock_type, 0)) < 0) {
       perror("socket");
       exit(1);
    }
    from.sin_family      = AF_INET;
    from.sin_port        = htons(clientport);
    from.sin_addr.s_addr = INADDR_ANY;
    if (bind(isock, (struct sockaddr *)&from, sizeof(from)) < 0) {
       perror("bind");
       exit(1);
    }
    fromlen = sizeof(from);
    if (getsockname(isock, (struct sockaddr *)&from, &fromlen) < 0) {
       perror("getsockname");
       exit(1);
    }
    clientport = ntohs(from.sin_port);

    /* Connect to server. For TCP, reuse same connection. */
    if (sock_type == SOCK_DGRAM) {
       if ((osock = socket(PF_INET, sock_type, 0)) < 0) {
          perror("socket");
          exit(1);
       }
    }
    else osock = isock;
    sin.sin_family  = AF_INET;
    sin.sin_port    = htons(serverport);
    sin.sin_addr    = host2ip(server);
    if (connect(osock, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
       perror("connect");
       exit(1);
    }
  }

  orig_callid = callid;
  do {
    char buf[256];
    if (loops) {
      sprintf(buf, "%s%d", orig_callid, loops);
      callid = buf;
    }

    create_packet(template, packet, method, clientport, sock_type, callid,
                  body, body_length, server);

    /* Send packet. */
    if(sock_type == SOCK_SSL) {
       SSL_write(slink->ssl, packet, strlen(packet));
    } else {
      if (write(osock, packet, strlen(packet)) < 0) {
        perror("write request");
      }
      if ( !only_status ) {
        printf("Written from source=%d to destination=%d:\n%s\n------------------------------------------\n",
           ntohs(from.sin_port), ntohs(sin.sin_port), packet);
      }

    }

    /* read from network and print */
    do {

      if ( !only_status ) {
         if(sock_type == SOCK_SSL) {
            c = SSL_read(slink->ssl, packet, sizeof(packet));
         } else {

           if ((c = recvfrom(isock, packet, sizeof(packet)-1, 0, 
                          (struct sockaddr *)&from, &fromlen)) < 0) {
             perror("recvfrom");
           }
         }
      }
      else {
        /* Use select */
        int response_status = 0;
        FD_ZERO(&fdset);
        FD_SET(isock, &fdset);
        tv.tv_sec = 2;
        tv.tv_usec = 0;
        if ( select(isock + 1, &fdset, NULL, NULL, &tv) > 0 ) {
          if ((c = recvfrom(isock, packet, sizeof(packet)-1, 0, 
                            (struct sockaddr *)&from, &fromlen)) < 0) {
            perror("recvfrom");
          }
          packet[c] = '\0';
          strtok(packet, " ");
          response_status = atoi(strtok(NULL, " "));
          if ( response_status >= 200 ) {
            printf("%d\n", response_status);
            exit(0);
          }
        }
        else {
          /* Exit since no response */
          exit(1);
        }
      }
      packet[c] = '\0';
      gettimeofday(&now, 0);    
      if ( !only_status ) {
        printf("Received t=%ld.%03ld from=%s:%d:\n%s\n------------------------------------------\n",
             (long)now.tv_sec, (long)now.tv_usec/1000,
             inet_ntoa(from.sin_addr), from.sin_port, packet);
      }

      /* If we ACK, check if final response. */
      status = 0;
      sscanf(packet, "%*s %u", &status);
      if (ack && status >= 200 && strcmp(method, "INVITE") == 0) {
        create_packet(template, packet, "ACK", clientport, sock_type, callid,
                      NULL, 0, server);
        if(sock_type == SOCK_SSL) {
            c = SSL_write(slink->ssl, packet, strlen(packet));
        } 
        else {
           if ((c = write(osock, packet, strlen(packet))) < 0) {
              perror("write ACK");
           }
        }
        if ( !only_status ) {
          printf("Sending ACK %d from source=%d to destination=%d:\n%s\n------------------------------------------\n",
               c, ntohs(from.sin_port), ntohs(sin.sin_port), packet);
        }
      }
      if (loops) {
        /* XXX hack */
        usleep(50000);
        c = 0;
      }
    } while (c > 0);

  } while (loops-- > 0);

  return 0;
} /* main */
