/*
 * bidipopen.c: open a bi-directional pipe to a sub-process.
 * Equivalent of popen(), except that popen(..., "r+") isn't implemented on
 * all systems (most notably Solaris).
 */
/* This code is taken from the FreeBSD 3.5 /usr/src/lib/libc/gen/popen.c 
 * and cleaned up slightly to be appropriate for use in cinema. */
/* Copyright 2001 Columbia University.  All rights reserved */

/*
 * Copyright (c) 1988, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software written by Ken Arnold and
 * published in UNIX Review, Vol. 6, No. 8.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#if defined(LIBC_SCCS) && !defined(lint)
static char sccsid[] = "@(#)popen.c	8.3 (Berkeley) 5/3/95";
#endif /* LIBC_SCCS and not lint */

#include <sys/param.h>
#include <sys/wait.h>

#include <errno.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h>

#include "sysdep.h"
#include "bidipipe.h"
#include "error.h"

#ifdef WIN32  /* Use vfork and pipe for unix platforms */

bidipipe *
bidipopen(const char *command, const char *type)
{
  bidipipe *iop;
  int pdes[2], pdes2[2], pid, twoway;
  char **argv = NULL;
  char *str;
  char *lasts = NULL;
  char *original_command;
  int size = 0;
  int index = 0;
  int hStdOut;
  int hStdIn;

  /*
   * Lite2 introduced two-way popen() pipes using socketpair().
   * FreeBSD's pipe() is bidirectional, so we use that.
   */
  /* CU: assume bidirectional pipe() on all supported systems.
   * Confirmed for Solaris, FreeBSD.  Need to check Linux, Tru64.
   * Or autoconf for it, and fall back to socketpair().
   */
  if (strchr(type, '+')) {
    twoway = 1;
    type = "r+";
  } else  {
    twoway = 0;
    if ((*type != 'r' && *type != 'w') || type[1])
      return (NULL);
  }
  
  if (_pipe(pdes, 512, O_BINARY | O_NOINHERIT) < 0)
    return (NULL);
  
  if (_pipe(pdes2, 512, O_BINARY | O_NOINHERIT) < 0) {
    (void)close(pdes[0]);
    (void)close(pdes[1]);
    return (NULL);
  }
  
  if ((iop = malloc(sizeof(bidipipe))) == NULL) {
    (void)close(pdes[0]);
    (void)close(pdes[1]);
    (void)close(pdes2[0]);
    (void)close(pdes2[1]);
    return (NULL);
  }
  
  /* Duplicate the stdin and stdout */
  hStdOut = _dup(_fileno(stdout));
  hStdIn  = _dup(_fileno(stdin));

  /* Duplicate write end of the pipe to stdout and 
  * read end to stdin */
  if (_dup2(pdes2[1], _fileno(stdout)) != 0 ||
    _dup2(pdes[0], _fileno(stdin)) != 0) {
    (void)close(pdes[0]);
    (void)close(pdes[1]);
    (void)close(pdes2[0]);
    (void)close(pdes2[1]);
    free(iop);
    return (NULL);
  }

  /* Close original write end of pipe2 and read end of pipe1 */
  close(pdes2[1]);
  close(pdes[0]);

  /* Parse the command to argv */
  size = 10;
  lasts = NULL;
  argv = (char **)calloc(size, sizeof(char *));
  original_command = strdup(command);
  
  str = strtok_r(original_command, " ", &lasts);
  
  for ( index = 0; str && *str; index ++ ) {
    if ( index >= size ) {
      size += 10;
      argv = (char **)realloc(argv, size * sizeof(char *));
    }
    argv[index] = str;
    str = strtok_r(NULL, " ", &lasts);
  }
  argv[index] = NULL;

  /* Spawn the process */
  if ( argv[0] != NULL ) {
    pid = spawnv(P_NOWAIT, argv[0], argv);
    free(original_command);
    free(argv);
  } 
  else {
    (void)close(pdes[0]);
    (void)close(pdes[1]);
    (void)close(pdes2[0]);
    (void)close(pdes2[1]);
    free(iop);
    free(original_command);
    free(argv);
    return (NULL);
  }
  
  /* Copy original stdout and stdin */
  _dup2(hStdOut, _fileno(stdout));
  _dup2(hStdIn, _fileno(stdin));
  
  close(hStdOut);
  close(hStdIn);
  
  /* Parent; assume fdopen can't fail. */
  iop->file_r = fdopen(pdes2[0], "rb");
  iop->file_w = fdopen(pdes[1], "wb");
  
  /* Link into list of file descriptors. */
  iop->pid = pid;
  
  return (iop);
}

/*
 * pclose --
 *	Pclose returns -1 if stream is not associated with a `popened' command,
 *	if already `pclosed', or waitpid returns an error.
 */
int
bidipclose(bidipipe *iop)
{
  pid_t pid;
  int exit_code = STILL_ACTIVE;
  
  (void)fclose(iop->file_r);
  (void)fclose(iop->file_w);
  
  pid = iop->pid;
  
  while (iop->pid && exit_code == STILL_ACTIVE) {
    GetExitCodeProcess((HANDLE) iop->pid, (unsigned long *)&exit_code);
  };
  
  free(iop);
  return (pid ? exit_code : -1);
}

#else 
/* Unix, i.e. systems with vfork() */

#define PATH_BINSHELL "/bin/sh"

bidipipe *
bidipopen(const char *command, const char *type)
{
  bidipipe *iop;
  int pdes[2], pdes2[2], pid, twoway;
  char *argv[4];
  
  /* If we're two-way, open bidirectionally. */
  if (strchr(type, '+')) {
    twoway = 1;
    type = "r+";
  } else  {
    twoway = 0;
    if ((*type != 'r' && *type != 'w') || type[1])
      return (NULL);
  }
  if (pipe(pdes) < 0)
    return (NULL);

  if (pipe(pdes2) < 0) {
    (void)close(pdes[0]);
    (void)close(pdes[1]);
    return (NULL);
  }
  
  if ((iop = malloc(sizeof(bidipipe))) == NULL) {
    (void)close(pdes[0]);
    (void)close(pdes[1]);
    (void)close(pdes2[0]);
    (void)close(pdes2[1]);
    return (NULL);
  }
  
  argv[0] = "sh";
  argv[1] = "-c";
  argv[2] = (char *)command;
  argv[3] = NULL;
  
  switch (pid = vfork()) {
  case -1:                     /* Error. */
    (void)close(pdes[0]);
    (void)close(pdes[1]);
    (void)close(pdes2[0]);
    (void)close(pdes2[1]);
    free(iop);
    return (NULL);
    /* NOTREACHED */
  case 0:                      /* Child. */
    if (*type == 'r') {
      if (pdes[1] != STDOUT_FILENO) {
        (void)dup2(pdes[1], STDOUT_FILENO);
        (void)close(pdes[1]);
      }
    }
    if (*type == 'w' || twoway) {
      if (pdes2[0] != STDIN_FILENO) {
        (void)dup2(pdes2[0], STDIN_FILENO);
        (void)close(pdes2[0]);
      }
    }
    (void)close(pdes[0]);
    (void)close(pdes2[1]);

    execv(PATH_BINSHELL, argv);
    _exit(127);
    /* NOTREACHED */
  }
  
  /* Parent */
  if (*type == 'r') {
    iop->file_r = fdopen(pdes[0], "r");
    if (iop->file_r == NULL) {
      error("bidipopen", "fdopen(%d, \"r\")\n", pdes[0]);
      (void)close(pdes[0]);
      (void)close(pdes[1]);
      (void)close(pdes2[0]);
      (void)close(pdes2[1]);
      free(iop);
      return (NULL);
    }      
    (void)close(pdes[1]);
  } 
  if (*type == 'w' || twoway ) {
    iop->file_w = fdopen(pdes2[1], "w");
    if (iop->file_w == NULL) {
      error("bidipopen", "fdopen(%d, \"w\")\n", pdes2[1]);
      (void)close(pdes[0]);
      (void)close(pdes[1]);
      (void)close(pdes2[0]);
      (void)close(pdes2[1]);
      free(iop);
      return (NULL);
    }      
    (void)close(pdes2[0]);
  }
  
  /* Link into list of file descriptors. */
  iop->pid = pid;

  return (iop);
}

/*
 * pclose --
 *	Pclose returns -1 if stream is not associated with a `popened' command,
 *	if already `pclosed', or waitpid returns an error.
 */
int
bidipclose(bidipipe *iop)
{
  int pstat;
  pid_t pid;
  
  if (iop->file_r) {
    (void)fclose(iop->file_r);
  }
  if ( iop->file_w) {
    (void)fclose(iop->file_w);
  }
  
  do {
    pid = waitpid(iop->pid, &pstat, 0);
  } while (pid == -1 && errno == EINTR);
  
  free(iop);
  
  return (pid == -1 ? -1 : pstat);
}
#endif
