/*
* Error handling and debugging functions.
*/

#include <stdio.h>     /* setvbuf() */
#include <stdarg.h>
#include <stdlib.h>    /* abort() */
#include <string.h>    /* strerror() */
#include <sys/time.h>  /* gettimeofday() */
#include <errno.h>
#include <syslog.h>
#include <time.h>      /* strftime() */
#include "sysdep.h"    /* for the other operating systems */

int debug_flag = 0;

FILE *logfile   = NULL; /* Name of logfile */
FILE *debugfile = NULL; /* Name of debug/trace output file */

/*
* debug()
* Print debug message to stdout or log file.
*/
void debug(char *function_name, char *format, ...)
{
  static int init_flag = 0;
  struct timeval tv;
  va_list ap;

  if(!debug_flag)
     return;

  /* Default debug trace goes to stdout */
  if ( debugfile == NULL )
  {
    debugfile = stdout;
  }

  va_start(ap, format);
  if (debug_flag) {
    if (!init_flag) {
      setvbuf(debugfile, NULL, _IOLBF, 0);
      init_flag = 1;
    }

    gettimeofday(&tv, NULL);

    /* print out name of function causing error */
    if ( function_name != NULL ) {
      (void) fprintf(debugfile, "%9.3f %s(): ", 
        tv.tv_sec + tv.tv_usec/1e6, function_name);
    }

    /* print out remainder of message */
    (void) vfprintf(debugfile, format, ap);

    if ( debugfile != stdout )
    {
      fflush(debugfile);
    }
  }

  va_end(ap);
} /* debug */


/*
* fatal()
* Print error message, using the format 'format', identified by the
* function name 'function_name' and abort process.
*/
void fatal(char *function_name, char *format, ...)
{
  char tmp[1024], *s = tmp;
  int e = errno;
  va_list ap;
  va_start(ap, format);

  /* timestamp */
  tmp[0] = '\0'; 
  if (logfile) {
    time_t t = time(0);

    strftime(s, sizeof(tmp), "%x %X: ", localtime(&t));
    s += strlen(s);
  }

  /* print out name of function causing error */
  sprintf(s, "Fatal error in %s: ", function_name);
  s += strlen(s);

  if (e > 0) {
    sprintf(s, "%s (%d): ", strerror(e), e);
    s += strlen(s);
  }

  /* print out remainder of message */
  vsprintf(s, format, ap);
  s += strlen(s);
  va_end(ap);

  if (logfile) {
    if (tmp[strlen(tmp)-1] != '\n') {
      sprintf(s, "\n");
      s += strlen(s);
    }
    fputs(tmp, logfile);
  }
  else {
    syslog(LOG_ERR, "%s", tmp);
  }

  abort();
} /* fatal */


/*
* error()
* Print error message to stderr.
*/
void error(char *function_name, char *format, ...)
{
  char tmp[1024], *s = tmp;
  int e = errno;
  va_list ap;
  va_start(ap, format);

  /* timestamp */
  tmp[0] = '\0'; 
  if (logfile) {
    time_t t = time(0);

    strftime(s, sizeof(tmp), "%x %X: ", localtime(&t));
    s += strlen(s);
  }

  /* print out name of function causing error */
  sprintf(s, "Error in %s: ", function_name);
  s += strlen(s);

  if (e > 0) {
    sprintf(s, "%s (%d): ", strerror(e), e);
    s += strlen(s);
  }

  /* print out remainder of message */
  vsprintf(s, format, ap);
  s += strlen(s);

  va_end(ap);

  if (logfile) {
    if (tmp[strlen(tmp)-1] != '\n') {
      sprintf(s, "\n");
      s += strlen(s);
    }
    fputs(tmp, logfile);
  }
  else {
    syslog(LOG_ERR, "%s", tmp);
  }

  va_end(ap);
} /* error */


/*
* ErrorInit()
* Initialize error log to send output to file 'filename'.  If 'file' is
* 0, use syslog.  If 'file' is "stderr", use that.
*/
void ErrorInit(char *filename)
{
  if (!filename) {
    openlog("sipd", LOG_PID, LOG_DAEMON);
    logfile = NULL;
  }
  else if (strcmp(filename, "stderr") == 0) {
    logfile = stderr;
  }
  else if ((logfile = fopen(filename, "a+")) == NULL) {
    logfile = stderr;
    fatal("ErrorInit", "Couldn't open error log %s", filename);
  }
} /* ErrorInit */


/*
* Close error log file or syslog.
*/
void ErrorClose(void)
{
  if (logfile) fclose(logfile);
  else closelog();
} /* ErrorClose */


/*
* Initialize debugging log to send output to 'filename'.  If 'file' is
* 0, use stdout.
*/
void DebugInit(char *filename)
{
  if (!filename) {
    debugfile = stdout;
  }
  else if (strcmp(filename, "stdout") == 0) {
    debugfile = stdout;
  }
  else if (strcmp(filename, "stderr") == 0) {
    debugfile = stderr;
  }
  else if ((debugfile = fopen(filename, "w+")) == NULL) {
    debugfile = stdout;
    fatal("ErrorInit", "Couldn't open debug log %s", filename);
  }
} /* ErrorInit */


/*
* Close log file or syslog.
*/
void DebugClose(void)
{
  if (debugfile) fclose(debugfile);
} /* ErrorClose */
