/*
 * glob2regex.c
 * Convert a 'glob' pattern (of the form accepted by gmatch) to a 'regex'
 * extended pattern (of the form accepted by Posix regcomp()/regexec() with
 * the REG_EXTENDED flag).
 */

/* Copyright 2001 Columbia University.  All rights reserved. */

#include <sys/types.h>
#include "sysdep.h"
#include <regex.h>
#include <string.h>

#include "dstring.h"
#include "glob2regex.h"

static int is_backslashable(char c);

/*
 * glob2regex()
 * Convert a 'glob' pattern to a 'regex' extended (v8) pattern.
 * Return a pointer to the converted regex in regex_p.
 * On failure, return a regex error code.
 *
 * Not all syntax errors in the glob will be detected.  In this case, they
 * will be converted to equivalent syntax errors in the regex.
 *
 * The argument paren_prefix is a special-case to handle the specific
 * needs of the canonicalize program's dialplan module.  In this case,
 * if a prefix of the glob pattern is parenthesized, everything *other*
 * than that prefix in the pattern is remembered as a backreference,
 * for later substitution.
 *
 * Because parts of the glob within the prefix may have been translated into
 * parenthesized subexpressions, the argument suffix_offset is a pointer to 
 * the indexed subexpression which will contain the actual suffix.  If there
 * was no such prefix, this will be the value 0, which is the match line for
 * the entire string.  This argument is ignored if paren_prefix is false, and
 * may be NULL in this case.
 *
 * Glob patterns, with the exception of the special parenthesized prefix
 * aspect, are documented in the file gmatch.c in this directory.
 * Note though that the leading ^ of glob patterns (match negation) is not
 * handled here; it must be handled by the calling function.
 *
 * Posix regular expressions are documented at
 * <http://www.opengroup.org/onlinepubs/007908799/xbd/re.html>.
 */
int glob2regex(const char *glob, char **regex_p, int paren_prefix,
               int *suffix_offset)
{
  DString regex;
  int found_prefix = 0;
  int brace_count = 0;
  int open_brace_count = 0;
  const char *glob_b;

  DStringInit(&regex);

  /* Opening anchor */
  DStringAppend(&regex, "^", -1);

  if (paren_prefix) {
    if (*glob == '(') {
      found_prefix = 1;
      glob++;
    }
    else {
      *suffix_offset = 0;
    }
  }

  for (; *glob != '\0'; glob++) {
    switch (*glob) {
    case '?':
      DStringAppend(&regex, ".", -1);
      break;

    case '*':
      DStringAppend(&regex, ".*", -1);
      break;

    case '{':
      DStringAppend(&regex, "(", -1);
      brace_count++;
      open_brace_count++;
      break;

    case '}':
      if (brace_count <= 0) {
        /* Too many '}'s, so this must be a literal '}'. */
        DStringAppend(&regex, "\\}", -1);
      }
      else {
        DStringAppend(&regex, ")", -1);
        brace_count--;
      }
      break;

    case ',':
      if (brace_count <= 0) {
        /* Not inside a {} group, so this must be a literal ','. */
        DStringAppend(&regex, ",", -1);
      }
      else {
        DStringAppend(&regex, "|", -1);
      }
      break;

    case '[':
      /* Scan the bracket group.  Bracket groups are terminated by
         a ']' in a position other than the first. */
      DStringAppend(&regex, "[", -1);
      for (glob_b = glob+1; *glob_b != ']' || glob_b == glob+1; glob_b++) {
        if (*glob_b == '\0') {
          DStringFree(&regex);
          return REG_EBRACK;
        }
        DStringAppend(&regex, glob_b, 1);
      }
      DStringAppend(&regex, "]", -1);
      
      glob = glob_b;
      break;

    case '\\':
      glob++;

      if (*glob == '\0') {
        /* This is the last character of glob. */
        DStringFree(&regex);
        return REG_EESCAPE;
      }

      if (is_backslashable(*glob)) {
        DStringAppend(&regex, "\\", -1);
      }
      /* Append the single character from glob. */
      DStringAppend(&regex, glob, 1);
      break;

    case ')':
      if (paren_prefix && found_prefix) {
        DStringAppend(&regex, "(", -1);
        *suffix_offset = open_brace_count + 1;
        found_prefix = 2;
        break;
      }   
      /* else FALLTHRU */

    default:
      /* Backslash, if necessary. */
      if (is_backslashable(*glob)) {
        DStringAppend(&regex, "\\", -1);
      }
      /* Append the single character from the glob */
      DStringAppend(&regex, glob, 1);
      break;
    }
  }

  if (paren_prefix) {
    if (found_prefix == 2) {
      DStringAppend(&regex, ")", -1);
    }
    else if (found_prefix) {
      DStringFree(&regex);
      return REG_EPAREN;
    }
  }

  /* Closing anchor */
  DStringAppend(&regex, "$", -1);

  *regex_p = DStringValue(&regex);
  return 0;
}


/*
 * is_backslashable()
 * According to the POSIX regex spec, only a special character may be
 * preceded by a backslash; backslash preceding an ordinary character
 * is undefined (and may be used by an implementation for extensions).
 * Return true if the character c may be preceded by a backslash.
 */
static int is_backslashable(char c)
{
  return (strchr("^.[$()|*+?{\\", c) != NULL);
} /* is_backslashable */


/*
 * glob2regex_regerror()
 * A wrapper for regerror() which translates error codes appropriately.
 * The only error for which this is necessary is REG_EPAREN, since
 * glob2regex maps parens to braces.  Everything else is either
 * syntactically the same or isomorphic, I believe.
 */
size_t glob2regex_regerror(int errcode, const regex_t *preg,
                           char *errbuf, size_t errbuf_size)
{
  if (errcode == REG_EPAREN) {
    return regerror(REG_EBRACE, NULL, errbuf, errbuf_size);
  }
  else {
    return regerror(errcode, preg, errbuf, errbuf_size);
  }
}
