#ifndef PARSER_HTTP_H
#define PARSER_HTTP_H
#include "rfc822.h"
#include "uri.h"
#include <stdio.h>

/*
 * The type of a parsed message.
 */
typedef enum {
  http_unknown_type = 0,
  http_request, 
  http_response
} http_message_type_t;

/*
 * Structure that represents complete message.
 */
typedef struct parser_t {
  http_message_type_t type; /* Any: message type */
  char *method;             /* Request: method */
  char *URI_string;         /* Request: request-URI, whole string */
  uri_t URI;                /* Request: URI components */
  int status;               /* Response: response code */
  char *reason;             /* Response: reason phrase */
  char *protocol;           /* R/R: protocol name */
  char *version;            /* R/R: protocol version */
  char *header_body;        /* pointer to headers and body */
 
  parser_header_t *headers; /* header list */
  int header_length;        /* length of header, including CRLF/CRLF */
} parser_t;

typedef struct http_content_type {    /* message body protocol */
  char type[MAX_MIME_TYPE];
  char subtype[MAX_MIME_TYPE];
} http_content_type_t;

typedef struct http_accept {
  char type[MAX_MIME_TYPE];
} http_accept;

typedef struct http_agent {           /* client identification */
  char *type;
} http_agent_t;

typedef struct http_server {          /* server identification */
  char *type;
} http_server_t;

typedef enum {
  auth_unknown,
  auth_basic,
  auth_digest,
  auth_pgp,
  auth_none
} http_authentication_method_t;

typedef struct http_authenticate {
  char *http_method;
  http_authentication_method_t method;
  char *realm;
  char *domain;
  char *nonce;
  char *opaque;
  char *stale;
  char *algorithm;
} http_authenticate_t;

typedef struct http_authorization {
  char *http_method;
  http_authentication_method_t method;
  char *userid;
  char *password;
  char *realm;
  char *nonce;
  char *digest_uri;
  char *response_digest;
  char *entity_digest;
  char *opaque;
  char *algorithm;
  unsigned char valid; /* flag: header syntax is correct */
  char *version;   /* The pgp version */
  char *signed_by; /* The key id that created the signature */
  char *signature; /* The pgp signature */
} http_authorization_t;

/* Functions to initialize, duplicate, and close parser. */
extern int      HTTP_Parse (char *, parser_t *);
extern void     HTTP_Duplicate(parser_t *dst, parser_t *src);
extern void     HTTP_Free  (parser_t *);

/* HTTP request and status line information */
extern char * HTTP_Method   (parser_t *);
extern uri_t  HTTP_URI      (parser_t *);
extern int    HTTP_Status   (parser_t *);
extern char * HTTP_Reason   (parser_t *);
extern char * HTTP_Protocol (parser_t *);
extern char * HTTP_Version  (parser_t *);

extern char *               HTTP_From          (parser_t *);
extern int                  HTTP_ContentLength (parser_t *);
extern http_content_type_t  HTTP_ContentType   (parser_t *);
extern http_server_t        HTTP_Server        (parser_t *);
extern http_agent_t         HTTP_UserAgent     (parser_t *);
extern http_authentication_method_t
                            HTTP_AuthenticationMethod(const char *m);
extern http_authenticate_t  HTTP_Authenticate  (parser_t *);
extern http_authenticate_t  HTTP_ProxyAuthenticate(parser_t *p);
extern http_authorization_t HTTP_Authorization (parser_t *);
extern http_authorization_t HTTP_ProxyAuthorization (parser_t *);
extern void                 HTTP_AuthorizationInit(http_authorization_t *);
extern void                 HTTP_AuthorizationFree(http_authorization_t *);
extern void                 HTTP_AuthenticateInit(http_authenticate_t *);
extern void                 HTTP_AuthenticateFree(http_authenticate_t *);
extern char *               HTTP_BuildAuthorizationHeader(
                                     http_authenticate_t *auth,
                                     const char *userid,
                                     const char *password,
                                     const char *uri);
extern struct tm            HTTP_Date          (parser_t *p);
extern struct tm *          Parse_HTTP_Date    (char *str, struct tm* p);

#define INFINITE_CONTENT_LENGTH 100000000

#endif





