/*
pthread_util.c - utility wrappers for pthread functions

Copyright 2000 by Columbia University; all rights reserved 
*/

#include <pthread.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include "error.h"
#include "pthread-util.h"
#include "sysdep.h"

/* uncomment this if you need the fast thread cleanup. by default
commented, as this behavior is not needed for most cinema components.
*/
/*
#define SIPLIB_EXTENSIONS  1
*/ 
#ifdef SIPLIB_EXTENSIONS

#define MAXTHREADS 1024

typedef struct {
  int active;     
  pthread_t myid;
} thread_id;

static thread_id thread_table[MAXTHREADS];
static pthread_mutex_t thread_mutex = PTHREAD_MUTEX_INITIALIZER;


typedef void* (*DFN)(void*);

typedef struct {
  DFN actual_function;
  void *param;
} thread_data;

static void add_thread(pthread_t t) {
  
  int i = 0;

  pthread_mutex_lock(&thread_mutex);

  for (i=0; i<MAXTHREADS; i++) {
    if (!thread_table[i].active) {
      thread_table[i].active = 1;
      thread_table[i].myid = t;
      break;
    }
  }

  if (i == MAXTHREADS) 
    error("add_thread", "Warning: Too many active threads...\n");
    
  pthread_mutex_unlock(&thread_mutex);

  return;

} /* add_thread */


static void del_thread(pthread_t t) {

  int i = 0;

  pthread_mutex_lock(&thread_mutex);

  for (i = 0; i<MAXTHREADS; i++) {
    if (thread_table[i].active && 
	pthread_equal(thread_table[i].myid, t)) {
      thread_table[i].active = 0;
      break;
    }
  }

  pthread_mutex_unlock(&thread_mutex);

  return;

} /* del_thread */


/*
   * instead of using cancellation handler's which cause
 * compatibility problems in Windows, simply use the "one
 * level indirection" approach to keep track of what functions
 * are executing.
 */
static void* dummyhandler(void *param) {
  
  thread_data* tdata = (thread_data*) param;

  add_thread(pthread_self());

  (*(tdata->actual_function))(tdata->param);

  del_thread(pthread_self());

  free(tdata);
  
  return NULL;
  
} /* dummyhandler */



/*
 * pthread_create_detached() for libsip++
 * -see the pthread_create_detached comment in the else part.
 *
 * In addition, if the SIPLIB macro is defined, pushes
 * a cancellation cleanup handler. This cancellation cleanup
 * handler is needed to ensure that all detached threads are 
 * killed by the KillDetachedThreads function.
 */
int pthread_create_detached(pthread_t *thr,
                            const pthread_attr_t *attr,
                            void *(*start_routine)(void *),
                            void *arg)
{
  pthread_attr_t myattr;
  thread_data *tdata;
  int thr_err, destroy_thr_err;

  if (attr != NULL) {
    return EINVAL;
  }

  if ((thr_err = pthread_attr_init(&myattr)) != 0) {
    return thr_err;
  }

  if ((thr_err = pthread_attr_setdetachstate(&myattr,
                                             PTHREAD_CREATE_DETACHED)) != 0) {
    goto done;
  }

  tdata = (thread_data*) calloc(1, sizeof(thread_data));
 
  if (tdata) {
    tdata->actual_function = start_routine;
    tdata->param = arg;
   
    /* tdata will be freed by dummyhandler...*/
    thr_err = pthread_create(thr, &myattr, dummyhandler, tdata);
  }
  
 done:
  /* Ignore any return values of this.  The only thing it's defined as
   * returning is EINVAL, which shouldn't ever happen here. */
  if ((destroy_thr_err = pthread_attr_destroy(&myattr)) != 0) {
    debug("pthread_create_detached", "pthread_attr_destroy: %s (%d)\n",
          strerror(destroy_thr_err), destroy_thr_err);
  }

  return thr_err;
    
} /* pthread_create_detached */


void KillDetachedThreads() {
  
  int i = 0;

  pthread_mutex_lock(&thread_mutex);
  
  for (i = 0; i<MAXTHREADS; i++) {
    if (thread_table[i].active) { 
      pthread_cancel(thread_table[i].myid);
      thread_table[i].active = 0;
    }
  }

  pthread_mutex_unlock(&thread_mutex);
  
  return;

} /* KillDetachedThreads */


#else /* for sipd */

/*
 * pthread_create_detached()
 * Create a new pthread in detached mode.  This means that nothing is going
 * to want to call pthread_join() on the created thread, so the system
 * can free the resources associated with the thread as soon as it exits.
 *
 * Note: if a pthread is not in detached mode (either by use of the attr
 * at create time, or by calling pthread_detach()), you will leak resources
 * if you don't call pthread_join() on it!  Eventually, you will run out
 * of pthreads, and crash.
 *
 * Note also that on Unix, WaitForMultipleObjects() is a wrapper around
 * pthread_join().  Thus, if you want to WaitForMultipleObjects() on a 
 * thread, you must not detach it.
 *
 * This function deliberately has the same calling conventions as
 * pthread_create(), so it's a drop-in replacement.  However, so things
 * don't get confused, we refuse to allow the incoming 'attr' parameter to
 * be set.  If you're doing something complicated, call
 * pthread_attr_setdetachstate() yourself.
 */
int pthread_create_detached(pthread_t *thr,
                            const pthread_attr_t *attr,
                            void *(*start_routine)(void *),
                            void *arg)
{
  pthread_attr_t myattr;
  int thr_err, destroy_thr_err;

  if (attr != NULL) {
    return EINVAL;
  }

  if ((thr_err = pthread_attr_init(&myattr)) != 0) {
    return thr_err;
  }

  if ((thr_err = pthread_attr_setdetachstate(&myattr,
                                             PTHREAD_CREATE_DETACHED)) != 0) {
    goto done;
  }

  thr_err = pthread_create(thr, &myattr, start_routine, arg);

done:
  /* Ignore any return values of this.  The only thing it's defined as
   * returning is EINVAL, which shouldn't ever happen here. */
  if ((destroy_thr_err = pthread_attr_destroy(&myattr)) != 0) {
    debug("pthread_create_detached", "pthread_attr_destroy: %s (%d)\n",
          strerror(destroy_thr_err), destroy_thr_err);
  }

  return thr_err;
    
} /* pthread_create_detached */



/*
 * KillDetachedThreads
 *
 * Not implemented for sipd or rtspd etc;
 */
void KillDetachedThreads() {

  return;  
  
} /* KillDetachedThreads */

#endif
