/*
* Expands a SIP domain name into a list of suitable servers, in the
* following order:
*  (0) IP address
*  (1) SRV records      (negative preferences starting at -65536)
*  (3) A and RR records (assigned a preference of 65536)
*
* Based on DomainSearch by Christian Zahl (1996);
*   Cleaned up by Henning Schulzrinne, 1997-07-13;
*   Extended to handle SRV records and port, 1998-08-03
*
* Copyright 1998-1999 by Columbia University; all rights reserved 
*/

#include "sysdep.h"

#include <stdio.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#if HAVE_ARPA_NAMESER_COMPAT_H
#include <arpa/nameser_compat.h> /* T_SRV, etc., on recent BIND */
#endif
#include <resolv.h>
#include <sys/socket.h>
#include <netdb.h>
#include <ctype.h>
#include <stdlib.h>   /* qsort(), drand48() */
#include <string.h>
#include <pthread.h>

#include "rr.h"       /* from resparse package */
#include "servers.h"

extern unsigned long random32(int);

/* extern int res_search(char *dname, int class, int type, */
/*   unsigned char *answer, int anslen); */

/*
* Insert host if address if not a duplicate.  Update preference if lower
* one found.
*/
static server_t *server_insert(server_t *server, char *host, int type,
  struct sockaddr_in sin, int preference, int weight, int *entries)
{
  int i;

  for (i = 0; i < *entries; i++) {
    if (server[i].sin.sin_addr.s_addr == sin.sin_addr.s_addr
        && server[i].sin.sin_port == sin.sin_port
        && server[i].type == type) {
      if (preference < server[i].preference)
        server[i].preference = preference; 
      break;
    }
  }
  /* no matching entry found , thus insert */
  if (i == *entries) {
    (*entries)++;
    server = server ? realloc(server, sizeof(server_t) * (*entries)) :
                    malloc(sizeof(server_t));
    server[i].host = host ? strdup(host) : NULL;
    server[i].sin  = sin;
    server[i].type = type;
    server[i].preference = preference;
    server[i].weight = weight;
  }
  return server;
} /* server_insert */


/*
 * Sort function:  by preference (lowest first) and then weight (highest
 * first).  For equal preference, UDP wins.
 */
static int server_sort(const server_t *s1, const server_t *s2)
{
  if (s1->preference > s2->preference)
    return 1;
  if (s1->preference < s2->preference)
    return -1;
  if (s1->weight < s2->weight)
    return 1;
  if (s1->weight > s2->weight)
    return -1;
  if (s1->type == SOCK_STREAM) return 1;
  return 0;
} /* server_sort */


/*
* Find the address of a host with domain name 'domain' and add it to the
* address list.
*/
static server_t *server_addr(server_t *server, 
  char *domain, int type, int port, 
  int preference, int weight, int *entries, int depth)
{
  struct hostent *h;
  char msg[PACKETSZ];
  s_rr **rr;
  res_response *res;
  struct sockaddr_in sin;
  int i;

  /* Allow nesting to 5 levels. */
  if (depth > 5) return server;
  sin.sin_family = AF_INET;
  sin.sin_port = htons(port);

  /*
   * Query for any CNAMEs and recurse.
   */

  if ((res_search(domain, C_IN, T_CNAME, msg, sizeof (msg)) != -1)
       && ((res = res_parse((char *)&msg)) != NULL)) {
    rr = res->answer;
    for (i = 0; i < res->header.ancount; i++) {
      if ((rr[i]->type == T_CNAME)
          && (rr[i]->class == C_IN))
        server = server_addr(server, rr[i]->rdata.string, type,
                    port, preference, weight, entries, depth + 1);
    }
    res_free(res);
  }
  /*
   * Try A RRs next if no CNAME.
   */
  else if ((res_search(domain, C_IN, T_A, msg, sizeof (msg)) != -1)
             && ((res = res_parse((char *)&msg)) != NULL)) {
    rr = res->answer;
    for (i = 0; i < res->header.ancount; i++) {
      if ((rr[i]->type == T_A)
          && (rr[i]->class == C_IN)) {
        sin.sin_addr = rr[i]->rdata.address;
        server = server_insert(server, rr[i]->name, type,
                   sin, preference, weight, entries);
      }
    }
    res_free(res);
  }
  /*
   * If that fails, try gethostbyname (this also uses local resources
   *   such as NIS or /etc/hosts).
   */
  else if (((h = gethostbyname (domain)) != NULL)
            && (h->h_addrtype == AF_INET)) {
    for (i = 0; h->h_addr_list[i] != NULL; i++) {
      sin.sin_addr = *(struct in_addr *)(h->h_addr_list[i]); 
      server = server_insert(server, h->h_name, type,
                 sin, preference, weight, entries);
    }
  }
  return server;
} /* server_addr */


/*
* Add to address list for 'domain'.  First, try for SRV RRs, then A RRs.
* Update 'entries' with number of entries.
*/
server_t *server_hosts(char *domain, char *service, int type, int port,
  int *entries)
{
  char msg[PACKETSZ];
  char name[1024];
  s_rr **rr;
  res_response *res;
  int i;
  server_t *server = NULL;
  struct sockaddr_in sin;

  /*
   * Check if the domain is just a numeric IP address.
   */

  if ((sin.sin_addr.s_addr = inet_addr(domain)) != -1) {
    sin.sin_family = AF_INET;
    sin.sin_port = htons(abs(port));
    if (type == 0 || type == SOCK_DGRAM) {
      server = server_insert(server, domain, SOCK_DGRAM,
         sin, 0, 0, entries);
    }
    if (type == 0 || type == SOCK_STREAM) {
      server = server_insert(server, domain, SOCK_STREAM,
         sin, 0, 0, entries);
    }
    return server;
  }

  /*
   * Look for SRV RRs of type 'sip' in the domain, first UDP, then TCP.
   */
#ifdef T_SRV
  if ((type == 0 || type == SOCK_DGRAM) && port < 0) {
    res = NULL;

    /* New-style (RFC 2782) format */
    sprintf(name, "_%s._udp.%s", service, domain);
    if (res_search(name, C_IN, T_SRV, msg, sizeof (msg)) != -1) {
      res = res_parse((char *)&msg);
    }
    if (res == NULL) {
      /* Old-style (RFC 2052) format */
      sprintf(name, "%s.udp.%s", service, domain);

      if (res_search(name, C_IN, T_SRV, msg, sizeof (msg)) != -1) {
        res = res_parse((char *)&msg);
      }
    }
    if (res != NULL) {
      rr = res->answer;
      for (i = 0; i < res->header.ancount; i++) {
        if ((rr[i]->type == T_SRV) && (rr[i]->class == C_IN)
            && (strcasecmp(name, rr[i]->name) == 0))
          server = server_addr(server, rr[i]->rdata.srv.target,
                               SOCK_DGRAM, rr[i]->rdata.srv.port,
                               rr[i]->rdata.srv.priority - 65536, 
                               rr[i]->rdata.srv.weight * drand48(),
                               entries,
                               0);
      }
      res_free(res);
    }
  }
  /* TCP */
  if ((type == 0 || type == SOCK_STREAM) && port < 0) {
    res = NULL;

    sprintf(name, "_%s._tcp.%s", service, domain);
    if (res_search(name, C_IN, T_SRV, msg, sizeof (msg)) != -1) {
      res = res_parse((char *)&msg);
    }
    if (res == NULL) {
      /* Old-style (RFC 2052) format */
      sprintf(name, "%s.tcp.%s", service, domain);
      
      if (res_search(name, C_IN, T_SRV, msg, sizeof (msg)) != -1) {
        res = res_parse((char *)&msg);
      }
    }
    if (res != NULL) {
      rr = res->answer;
      for (i = 0; i < res->header.ancount; i++) {
        if ((rr[i]->type == T_SRV) && (rr[i]->class == C_IN)
            && (strcasecmp(name, rr[i]->name) == 0))
          server = server_addr(server, rr[i]->rdata.srv.target,
                               SOCK_STREAM, rr[i]->rdata.srv.port, 
                               rr[i]->rdata.srv.priority - 65536, 
                               rr[i]->rdata.srv.weight * drand48(),
                               entries,
                               0);
      }
      res_free(res);
    }
  }
#endif

  /*
   * Finally, we take any host with that name, but give it the lowest
   * preference. We use UDP first, then TCP.
   */
  if (type == 0 || type == SOCK_DGRAM) {
    server = server_addr(server, domain, SOCK_DGRAM,  
      abs(port), 65536, 0, entries, 0);
  }
  if (type == 0 || type == SOCK_STREAM) {
    server = server_addr(server, domain, SOCK_STREAM,
      abs(port), 65536, 0, entries, 0);
  }
  return server;
} /* server_hosts */


/*
* External interface; return list of server entries for 'domain', with
* last entry having a host value of NULL.  The service is a protocol
* such as 'sip' or 'rtsp'.  The 'port' argument is used when the DNS
* entry doesn't contain one.  A negative number indicates the default
* port.  (This is indicated explicitly since SRV records are skipped
* unless the default port is used.) The 'type' restricts whether UDP or
* TCP addresses are sought.  If 0 both are used.
*/

/*
 * Use a mutex lock to protect MT-unsafe res_search().
 * Since all occurances of res_search are called from within
 * servers() the lock is set and unset in this function.
 */
static pthread_mutex_t S_lock = PTHREAD_MUTEX_INITIALIZER;

server_t *servers(char *domain, char *service, int type, int port)
{
  struct sockaddr_in sin;
  server_t *server = NULL;
  int entries = 0;
  static int seed = 0;

  pthread_mutex_lock(&S_lock);

  if (!seed) seed = random32(0);

  memset(&sin, '\0', sizeof(sin));

  server = server_hosts(domain, service, type, port, &entries);
  qsort(server, entries, sizeof(server_t),
        (int (*)(const void*, const void*)) server_sort);

  server = server_insert(server, NULL, 0, sin, 0, 0, &entries);  /* insert terminator */
  
  pthread_mutex_unlock(&S_lock);

  return server;
} /* servers */
