/* cpp_pthread_mutex.h -- object-based pthread mutex locking for C++.
 * pthread_mutex_obj calls pthread_mutex_(init|destroy) as a constructor
 * and destructor.  pthread_mutex_stack_obj, when allocated on the stack,
 * automatically locks when instantiated.  If you instantiate it on the stack,
 * you'll automatically unlock when leaving the enclosing block by any means,
 * including a thrown exception.)
 */

/* Copyright 2001 Columbia University.  All rights reserved. */

#ifndef STACK_LOCK_H_
#define STACK_LOCK_H_

#ifndef __cplusplus
#error "stack_lock.h is only useful for C++ code"
#endif

extern "C" {
#include "pthread.h"
}

class pthread_mutex_obj {
private:
  pthread_mutex_t the_lock;

  void throw_from_errno(int err);
public:
  pthread_mutex_obj();
  ~pthread_mutex_obj();

  void lock();
  void unlock();
};

inline pthread_mutex_obj::pthread_mutex_obj() {
  int err = pthread_mutex_init(&the_lock, NULL);
  if (err != 0) throw_from_errno(err);
}

inline pthread_mutex_obj::~pthread_mutex_obj() {
  pthread_mutex_destroy(&the_lock);
  // Don't throw exceptions from a destructor
}

inline void pthread_mutex_obj::lock() {
  int err = pthread_mutex_lock(&the_lock);
  if (err != 0) throw_from_errno(err);
}

inline void pthread_mutex_obj::unlock() {
  int err = pthread_mutex_unlock(&the_lock);
  if (err != 0) throw_from_errno(err);
}


class pthread_mutex_stack_obj {
private:
  pthread_mutex_obj& the_lock;
public:
  pthread_mutex_stack_obj(pthread_mutex_obj &);
  ~pthread_mutex_stack_obj();
};

inline
pthread_mutex_stack_obj::pthread_mutex_stack_obj(pthread_mutex_obj& lock)
  : the_lock(lock) {
  the_lock.lock();
}

inline pthread_mutex_stack_obj::~pthread_mutex_stack_obj() {
  the_lock.unlock();
}
   

#endif /* STACK_LOCK_H_ */
