/*
tcp.c - handle incoming TCP connections and requests

Copyright 1998-1999 by Columbia University; all rights reserved 
*/

#include <stdio.h>     /* FILE */
#include <string.h>    /* strerror */
#include <pthread.h>   /* pthread_ */
#include <stdlib.h>    /* malloc/free */
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>    /* close */
#include "request-common.h"
#include "error.h"
#include "pthread-util.h"
#include "sysdep.h" 
#include "thread_pool.h"

#include "tcp.h"

static int tcpsock;         /* socket for incoming TCP connection requests */

/*
 * ReceiveTCPRequests()
 * Process incoming TCP requests and send out responses.
 * Messages without Content-Length are assumed to have no body. (This
 * differs from HTTP/1.0 conventions.)
 */
void *ReceiveTCPRequests(void *arg)
{
  struct socket_info *si = (struct socket_info *)arg;
  enum {Header, CR, CRNL, CRNLCR, NL, Body} state = Header;
  int content_length;      /* Content-Length field */
  int cl = 0;              /* bytes read in content (body) */
  int c;                   /* current character */
  char ch;                 /* same, as char */
  DString header, body;    /* header and body */
  FILE_SOCKET *f;          /* socket, as FILE */
  request_common_t *r;     /* request parameters */


  fd_set read_fd;
  struct timeval timeout;
  int retval;
  sock_t dum_s;

  if ( si == NULL ) {
    error("ReceiveTCPRequests", "param is NULL\n");
    return NULL;
  }

  f = fdopen_socket(si->socket, "r");
  if (f == NULL) {
    error ("ReceiveTCPRequests", "Couldn't fdopen socket %d for reading",
           si->socket);
    free(si);
    return NULL;
  }

  c = EOF; /* To avoid warning */  
  /* Once around the loop for each message. */
  do {
    DStringInit(&header);
    DStringInit(&body);
    state          = Header;
    content_length = INFINITE_CONTENT_LENGTH;
    cl             = 0;
    r              = NULL;
    while ((cl < content_length) && ((c = getc_socket(f)) != EOF)) {
      ch = c;
      DStringAppend(state == Body ? &body : &header, &ch, 1);
      switch (state) {
      case Header:
        if (c == '\r') state = CR;
        else if (c == '\n') state = NL;
        break;
      case CR:
        if (c == '\n') state = CRNL;
        else state = Header;
	break;
      case CRNL:
        if (c == '\r') state = CRNLCR;
        else state = Header;
        break;
      case NL:
      case CRNLCR:
        if (c == '\n') {
          if (header.length > 4) {
            debug("ReceiveTCPRequests",
                  "Incoming TCP message on socket %d from %s:%d:\n%s\n",
                  si->socket,
                  inet_ntoa(si->sin.sin_addr), ntohs(si->sin.sin_port),
                  DStringValue(&header));
            r = (request_common_t *)
              RequestCreate(SOCK_STREAM, si->socket, dum_s);
              //RequestCreate(SOCK_STREAM, si->socket, si->sin);
            r->sin.tcp_s = si->sin;
            r->server = si->server; /* To know original server */

            if (HTTP_Parse(DStringValue(&header), &r->p) < 0) {
	      /* If this was a request whose headers didn't parse, we want
	       * to try to send a 400 response to it.  On the other hand,
	       * there's not much we can do with responses that didn't parse,
	       * so just mark them as the unknown message type so that the
	       * server ignores them and frees their memory.
	       */
	      if (r->p.type == http_request) {
                ResponseSetStatus((request_t *)r, 400, NULL);
	      } else {
		r->p.type = http_unknown_type;
	      }
            }
            content_length = HTTP_ContentLength(&r->p);
            if (missing_content_length_means_zero) { /* RTSP */
              /* INFINITE_CONTENT_LENGTH is the "default" value returned by
                 the parser, implying content length was not specified */
              if (content_length == INFINITE_CONTENT_LENGTH) {
		content_length = 0;
              }
            }
          }
          else content_length = -1;
          DStringFree(&header);
          state = Body;
        }
        else state = Header;
        break; 
      case Body:
        cl++;
        break;
      }
    }
    /* We now have a complete message, with header and content. */
    if (r) {
      r->request_body.content = DStringValue(&body);
      r->request_body.content_length = content_length;
      RequestProcess((request_t *)r);

      FD_ZERO(&read_fd);
      FD_SET(si->socket, &read_fd);
      timeout.tv_sec = 120;
      timeout.tv_usec = 0;
      
      retval = select(si->socket+1, &read_fd, NULL, NULL, &timeout);
        
      if (retval < 0) {
	error("ReceiveTCPRequests", "select returns an error\n");
	break;
      }
        
      if (!retval && !FD_ISSET(si->socket, &read_fd)) {
	debug("ReceiveTCPRequests", 
	      "No activity on socket %d. Freeing resources\n", si->socket);
 	break;
      }
    } /* if (r) */
  } while (c != EOF);

  fclose_socket(f);
  debug("ReceiveTCPRequests", "Connection from %s:%d closed (socket %d)\n",
	inet_ntoa(si->sin.sin_addr), ntohs(si->sin.sin_port), si->socket);
  
  free(si);
  
  return NULL;
} /* ReceiveTCPRequests */


/*
* TCP request receiver. 
*/
void *ReceiveTCP(void *arg)
{
  struct tcp_thread_arg *ptr = (struct tcp_thread_arg *)arg; 
  void *server = ptr->server;
  char *primaryIP = ptr->primaryIP;
  unsigned short port = (ptr->port);  /* TCP port. This is int and not short */
  struct socket_info *si;  /* accept socket information */
  struct sockaddr_in sin;  /* bind socket address */
  int sinlen;              /* length of address */
  int thr_err;             /* thread error condition */
  int opt;                 /* opt for setting setsockopt */

  /* Create TCP listening socket. */
  if ((tcpsock = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
    error("ReceiveTCP", "socket\n");
    return NULL;
  }

  /* Set "reuseaddr" (and "reuseport", if it exists) so that we don't
   * have to wait for TIME_WAIT to expire if we crash the server while
   * connections are open. */
  opt = 1;
  if (setsockopt(tcpsock, SOL_SOCKET, SO_REUSEADDR, (char *)&opt, sizeof(opt)) < 0) {
    error("RecieveTCP", "setsockopt(SO_REUSEADDR)");
    return NULL;
  }

#ifdef SO_REUSEPORT
  opt = 1;
  if (setsockopt(tcpsock, SOL_SOCKET, SO_REUSEPORT, (char *)&opt, sizeof(opt)) < 0) {
    error("ReceiveTCP", "setsockopt(SO_REUSEPORT)");
    return NULL;
  }
#endif

  /* Bind to server address. */
  sin.sin_family      = AF_INET;
  sin.sin_port        = htons(port);
  if (primaryIP) {
    sin.sin_addr.s_addr = inet_addr(primaryIP);
    if (sin.sin_addr.s_addr == -1) {
      error("ReceiveTCP", "Bad address %s specified...\n", primaryIP);
      exit(0);
    }
  }
  else {    
    sin.sin_addr.s_addr = INADDR_ANY;
  }

  if (bind(tcpsock, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
    error("ReceiveTCP", "bind on %hu\n", port);
    return NULL;
  }
  if (listen(tcpsock, 1000) < 0) {
    return NULL;
  }

  if ( port == 0 ) {
    int sinlen = sizeof(sin);
    if ( getsockname(tcpsock, (struct sockaddr *)&sin, &sinlen) == 0 ) {
      port = ntohs(sin.sin_port);
      ptr->port = port;
      debug("ReceiveTCP", "bind to port %hu\n", port);
    }
    else {
      error("ReceiveTCP", "getsockname\n");
    }
  }

  /* Block waiting for connection. */
  while (1) {
    si = (struct socket_info *) malloc(sizeof(struct socket_info));
    if (si == NULL) {
      error("ReceiveTCP", "malloc\n");
      return NULL;
    }
    
    si->server  = server;
      
    sinlen = sizeof(si->sin);
    if ((si->socket = accept(tcpsock, (struct sockaddr *)&si->sin, &sinlen)) < 0) {
      error("ReceiveTCP", "accept\n");
      free(si);
      return NULL;
    }

    debug("ReceiveTCP", "Connection from %s:%d (socket %d)\n",
      inet_ntoa(si->sin.sin_addr), ntohs(si->sin.sin_port), si->socket);

    if ((thr_err = pool_schedule(ReceiveTCPRequests, (void *)si))) {
      error("ReceiveTCP", "unable to get thread from pool\n");
      closesocket(si->socket);
      free(si);
    }
    
    /* {
     pthread_t th;
     if ((thr_err = pthread_create_detached(&th, NULL, ReceiveTCPRequests,
	  (void *)si))) {
	  error("ReceiveTCP", "thread: %s\n", strerror(thr_err));
	  closesocket(si->socket);
	  free(si);
     }
     }
    */
    
  }

  /*
   * XXX ptr must not be free'd because otherwise libsip++ code may access
   * it without knowing that the thread terminated abnormally and
   * ptr is freed or not. Since this is a one time memory leak, it
   * is tolerable.
   */
  /* free(ptr); */

  return NULL;
} /* ReceiveTCP */

/*
 * CloseTCP
 *  -Close the tcp socket.
 *  -Todo: Kill all sleeping threads in ReceiveTCPRequests.
*/
void CloseTCP() {
 closesocket(tcpsock);
}
