/*
  thread_pool.c - Implementation of a simple thread pool - with a 
G/G/T/N queue.

  version 2. Revised the pool to be G/G/T/N instead of G/G/T/1, and made
it more lighter.

Synopsis:

InitThreadPool sets up a pool of threads based on the maxthreads
argument.  The maxqueuesize parameter controls the number of outstanding
waiting requests. 

pool_schedule puts the job into the queue. It returns error if the queue
is already full (equal to maxqueuesize). 

pool_get_status returns current status of the pool - Viz. total threads,
queue sizes (max and current).

DestroyThreadPool will cancel all threads. This is useful
only when the system wants to exit immediately. In normal exit
scenarios, there is no need to call this function (this function is
specifically provided to support cleanup, when libsip (++) is 
used inside a Windows DLL, that needs fast cleanup.

Copyright 2000 by Columbia University; all rights reserved 
*/

#include <pthread.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include "error.h"
#include "thread_pool.h"
#include "sysdep.h"

extern int errno;

typedef void* (*FPTR) (void*);  


typedef struct per_thread_data 
{
  int cookie;                    /* Unique Identifier */
  
  pthread_t thread_data;         /* this thread's identifier */
  
} ThreadData;


/*
 * Work information. FPTR is the function to execute, function_args
 * is the argument to pass to this function.
 */
typedef struct work_queue_element {
  
  FPTR fptr;
  
  void *function_args;
  
  struct work_queue_element *p_next;
  
} WorkQueueElement;


static int init = 0;                          /* is init over? */

/*
 * This is the routine which is executed by all our threads, 
 * and which executes the actual user function when needed.
 */
static void* pool_routine (void*);

static ThreadData *threadPoolTable = NULL;    /* thread pool table */

static WorkQueueElement *p_queueHead = NULL, 
  *p_queueLast = NULL;                        /* worker queue */


static int max_threads = 0;
static int current_queuesize = 0;    /* work queue size */
static int max_queuesize = 0;        /* maximum queue size */


/* if no jobs to do, threads will be sleeping. this will be used
 * to wake up one.
 */
static pthread_cond_t queueCond = PTHREAD_COND_INITIALIZER;     


/* All access to p_queueHead, p_queueLast, queueCond will be 
 * synched through this.
 */
static pthread_mutex_t queueMutex = PTHREAD_MUTEX_INITIALIZER;  

/*
 * InitThreadPool
 *  Sets up a thread pool with maxthreads, and a maximum
 * queue size = maxthreads.
*/
int InitThreadPool(int maxthreads) 
{

  return InitThreadPool2 (maxthreads, maxthreads);

} /* InitThreadPool */

/*
 * InitThreadPool2
 *
 * Sets up our thread pool, with as many threads as needed.
 *
 * Returns 0 on success, non-zero on error. errno is set.
 */
int InitThreadPool2(int maxthreads, int maxqueue)
{
  int i, stat;
  
  if (!init)  
    init = 1;
  else {
    debug("InitThreadPool","Calling second time...buggy\n");
    return 0;
  }
  
  if (maxthreads <= 0) {
    errno = EINVAL;
    error("InitThreadPool", "Bad value for maxthreads %d\n",maxthreads);
    return -1;
  }
  
  max_threads = maxthreads;
  max_queuesize = maxqueue;

  threadPoolTable = (ThreadData*)calloc(maxthreads,sizeof(ThreadData));
  
  if (!threadPoolTable) {
    error("InitThreadPool", "Unable to allocate memory...\n");
    return -1;
  }
  
  for (i=0;i<maxthreads;i++) {
    
    threadPoolTable[i].cookie = i;
    
    if((stat = pthread_create(&threadPoolTable[i].thread_data, 
			      NULL, 
			      pool_routine, 
			      (void*)&threadPoolTable[i])) != 0) {
      error("InitThreadPool", "pthread_create failed...no more threads will be added to the pool\n");
      
      if (i>0) { /* atleast one thread is there in the pool -
		  * dont exit if 'maxthreads' threads were not
		  * creatable. just flag an error
		  */
	max_threads = i;
       	break;
      }
      
      return stat;
    }
    
  }     
  
  debug("InitThreadPool2","Created %d threads successfully...\n",maxthreads);
  
  return 0;
  
} /* InitThreadPool2 */


/*
 * Executes the function start_routine in a separate thread.
 *
 * Returns 0 if successfull. Returns non-zero if too many requests
 * are waiting in the queue.
 */
int pool_schedule(void *(*start_routine)(void *), void *arg)
{
  /*
   * Algorithm:
   *
   * Add it to our queue if current queue size < max queue size. 
   * Then wake up one waiting thread.
   */
  WorkQueueElement *p_temp;
  
  if(!init) {
    error("pool_schedule", "Call before Init...\n");
    return EINVAL;
  }
  
  pthread_mutex_lock(&queueMutex);
 
  errno = 0;
  debug("Queue Size: ", "%d\n",current_queuesize);
  
  if(current_queuesize >= max_queuesize) {
    pthread_mutex_unlock(&queueMutex);
    return EBUSY;
  }
  
  p_temp = (WorkQueueElement*)calloc(1,sizeof(WorkQueueElement));
  p_temp->fptr = start_routine;
  p_temp->function_args = arg;
  
  if (p_queueLast) {
    p_queueLast->p_next = p_temp;
    p_queueLast = p_temp;
  }
  else {
    p_queueHead = p_queueLast = p_temp;
  }

  ++current_queuesize;

  /*
   * We dont need to use pthread_cond_broadcast here, as when woken
   * up our worker takes a job from the queue, and then checks if
   * the queue is non-empty. if the queue is non-empty, it wakes
   * up another worker also.
   */
  debug("pool_schedule","Now scheduling a thread for job %p\n",
	p_temp->fptr);

  pthread_cond_signal(&queueCond);
  
  pthread_mutex_unlock(&queueMutex);
  
  return 0;
  
} /* pool_schedule */

/* 
 * pool_get_status
 *
 * Return total threads in pool, number of busy threads, the max queue
 * size, and current queue size.
 *
 * Locking should not be done as this function needs to be super-fast.
 * A +/- minor error due to races are acceptable.
 */
int pool_get_status(int *p_maxthreads, int *p_max_queuesize, int *p_currentqueuesize)
{
  if (p_maxthreads)  
    *p_maxthreads = max_threads;
  
  if (p_max_queuesize) 
    *p_max_queuesize = max_queuesize;
  
  if (p_currentqueuesize) 
    *p_currentqueuesize = current_queuesize;
  
  return 0;
  
} /* pool_get_status */

/*
 * DestroyThreadPool
 *
 * Not implemented as i dont see any reasons for calling this. 
 */
int DestroyThreadPool()
{
  
  int i;

  /*
   * We will cancel all the active threads here.
   */
  pthread_mutex_lock(&queueMutex);
 
 for (i=0; i<max_threads; i++) 
   pthread_cancel(threadPoolTable[i].thread_data); 
 
 max_threads = 0;
 
 pthread_mutex_unlock(&queueMutex);
 
 return 0;
}


/*
 * pool_routine
 *
 * This is our worker. Whenever work is available in the queue, takes
 * it, and executes it. loops infinitely.
 */
static void* pool_routine(void *arg) 
{
  ThreadData *p_myData; 
  WorkQueueElement *p_job;
  
  if(!arg) {
    error("pool_routine","Bad parameter to call...\n");
    return NULL;
  }
  
  p_myData = (ThreadData*)arg;
  
  /*
   * Algorithm:
   *
   * Whenever a job is scheduled, the caller will signal us. 
   * When this thread is done with its work, we will check if more
   * work is availble, in case we will continue.
   */
  
  while(1) {
    
    pthread_mutex_lock(&queueMutex);
    
    if(current_queuesize <= 0)         /* dont wait if there is work to do */ 
      pthread_cond_wait(&queueCond, &queueMutex);
    
    if (!p_queueHead) {               /* protection against spurious wakeups */
      pthread_mutex_unlock(&queueMutex);
      continue;
    }
    
    p_job = p_queueHead;              /* Remove a job */
    p_queueHead = p_job->p_next;
    p_job->p_next = NULL;
    
    --current_queuesize;
  
    if(!p_queueHead)  
      p_queueLast = NULL;  /* no more work in queue */
    else
      pthread_cond_signal(&queueCond);
    
    pthread_mutex_unlock(&queueMutex);
    
    /*error("pool_routine","%d\n",p_myData->cookie); */
 
    debug("pool_routine","starting execution of %p in thread %d\n",
	  p_job->fptr,
	  p_myData->cookie);
    
    (*(p_job->fptr))(p_job->function_args);

    debug("pool_routine","finished execution of %p in thread %d\n",
	p_job->fptr, p_myData->cookie);

    free(p_job);
       
  } /* while */
  
  
  /* never reached.....
   */
  
  return 0;
  
} /* pool_routine */



