/*
udp.c - handle incoming UDP requests

Copyright 1998-1999 by Columbia University; all rights reserved 
*/

#include <stdio.h>      /* perror() */
#include <stdlib.h>     /* free() */
#include <string.h>     /* strstr, ... */
#include <memory.h>     /* memset() */
#include <unistd.h>     /* close() */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "thread_pool.h"
#include "request-common.h"
#include "error.h"
#include "sysdep.h"
#include "udp.h"
#include <sys/time.h>
#define MAX_REQUEST 8192  /* large enough for UDP packet */

typedef struct udp_worker_data {
  
  int sock_response;

  struct sockaddr_in sin;
  
  void* server;

  char *msg;

  int msglen;

} UDPWorkerData;

static void* ProcessUDPPacket(void *arg);

/*
 * getMessageType
 *
 * Tentatively decide whether the incoming message is a sip/rtsp
 * request or response.
 */
static int getMessageType(const char *buf) 
{
  char line[1024];
  char *p;
  int tocopy = sizeof(line);
  
  /* copy the first line */
  if(!(p = strchr(buf,'\n'))) 
    return http_response;
  
  if((p - buf) < tocopy)
    tocopy = p-buf;
  
  strncpy(line, buf, tocopy);
  
  if ((p = strstr(line, "SIP"))) 
    return (p == line) ? http_response : http_request;
  
  if ((p = strstr(line,"RTSP"))) 
    return (p == line) ? http_response : http_request;
  
  /*
   * better to err on the wrong side.
   */
  return http_response;
  
} /* getMessageType */

/* make udpsock a static global so that the CloseUDP function can
 * close it.
*/
static int sock;          /* socket for incoming requests */
static int sock_response; /* socket for responses */

/*
 * ReceiveUDP()
 * UDP SIP receiver:  get message from socket and pass to
 * ProcessRequest(). One message may contain several requests.
 */
void *ReceiveUDP(void *arg)
{
  struct udp_thread_arg *ptr = (struct udp_thread_arg *)arg;
  void *server = ptr->server;
  char *primaryIP = ptr->primaryIP; /* If non null, overrides INADDR_ANY */
  unsigned short port = ptr->port;  /* UDP port; This is int and not short */
  struct sockaddr_in sin;
  int sinlen, type;          
  int maxqueuesize, curqueuesize;  /* to decide whether to parse or drop */
  UDPWorkerData *workerData;
  
 
#ifdef IP_ADD_MEMBERSHIP
  struct ip_mreq mreq;             /* multicast group */
#endif

  /* Create UDP send socket. */
  if ((sock_response = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    perror("UDP socket for responses");
  }

  /* Create UDP receive socket for requests. */
  if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    return NULL;
  }
 
  /* Bind to default server port. */
  memset(&sin, 0, sizeof(sin));
  sin.sin_family      = AF_INET;

  if (primaryIP) {
    sin.sin_addr.s_addr = inet_addr(primaryIP);
    if (sin.sin_addr.s_addr == -1) {
	error("ReceiveUDP", "Bad address %s passed...\n", primaryIP);
	exit(0);
    }
  } 
  else { 
    sin.sin_addr.s_addr = htonl(INADDR_ANY);
  }
  sin.sin_port        = htons(port);

  /* Accept packets on multicast address sip.mcast.net as well. */
#ifdef IP_ADD_MEMBERSHIP
#ifdef BIND_MCAST
  sin.sin_addr.s_addr = inet_addr("224.0.1.75");
#endif
  mreq.imr_multiaddr.s_addr = inet_addr("224.0.1.75");
  mreq.imr_interface.s_addr = htonl(INADDR_ANY);
  if (setsockopt(sock, IPPROTO_IP,
    IP_ADD_MEMBERSHIP, (char *)&mreq, sizeof(mreq)) < 0)
      error("ReceiveUDP", "IP_ADD_MEMBERSHIP\n");
#endif
 
  if (bind(sock, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
    error("ReceiveUDP", "bind on %hu\n", port);
    return NULL;
  }
 
  if ( port == 0 ) {
    int sinlen = sizeof(sin);
    if ( getsockname(sock, (struct sockaddr *)&sin, &sinlen) == 0 ) {
      port = ntohs(sin.sin_port);
      ptr->port =  port;
      debug("ReceiveUDP", "bind to port %hu\n", port);
    }
    else {
      error("ReceiveUDP", "getsockname\n");
    }    
  }

  workerData = (UDPWorkerData*)calloc(1,sizeof(UDPWorkerData));

  workerData->msg = (char*)malloc(MAX_REQUEST * sizeof(char));

  for (;;) {
  
    sinlen = sizeof(workerData->sin);
    workerData->sock_response = sock_response;
    workerData->server = server;
    
    if ((workerData->msglen = 
	 recvfrom(sock, workerData->msg, MAX_REQUEST, 0, 
		  (struct sockaddr *)&workerData->sin, &sinlen)) < 0) {
      error("ReceiveUDP", "recvfrom\n");
      break;
    } 
    
    workerData->msg[workerData->msglen] = '\0';
    
    debug("ReceiveUDP", "received UDP packet (length %d) from %s:%d:\n%s\n",
          workerData->msglen, inet_ntoa(workerData->sin.sin_addr), 
	  ntohs(workerData->sin.sin_port), workerData->msg);

    /*
     * Check if this is a request or response. We check for responses
     * by seeing if this message starts with SIP or RTSP.
     *
     * If this is a response we always schedule it in a new thread,
     * which does parsing, and response processing.
     *
     * If this is a request, we schedule it only if the current_queue_length
     * is less than max_queue_length/2 (the reasoning is: max_queue_length
     * is set to 2*total_threads. even if all threads are active, this will
     * ensure that a thread picks up this job, once it has completed what
     * it is doing currently - which typically will be less than 10ms).
     */
    pool_get_status(NULL, &maxqueuesize, &curqueuesize);
    
    if ((type = getMessageType(workerData->msg)) != http_response 
	&& (curqueuesize >= (maxqueuesize >> 1))) 
    {
      error("udp","dropping request\n");
      continue;
   }
    
    /*
     * what happens if pool_schedule returns < 0 (i.e., EBUSY) ???
     * in this case we drop requests. for responses, we serve it
     * iteratively (i.e., instead of calling ProcessUDPRequest in 
     * a new thread, we call it in this thread itself.
     */
    if (pool_schedule(ProcessUDPPacket, (void*) workerData) != 0) {
      if (type == http_request)  {
        error("udp","dropping packet2\n");
	continue;
      }
      else
	ProcessUDPPacket((void*)workerData);
    }
    
    if (!(workerData = (UDPWorkerData*) calloc(1, sizeof(UDPWorkerData)))) {
      error("ReceiveUDP", "Unable to allocate memory for worker data\n");
      break;
    }
    
    if (!(workerData->msg = (char*) malloc(MAX_REQUEST * sizeof(char)))) { 
      error("ReceiveUDP", "Unable to allocate buffer space\n");
      break;
    }
    
  } /* for */

  closesocket(sock);
  closesocket(sock_response);
  
  /*
   * XXX ptr must not be free'd because otherwise libsip++ code may access
   * it without knowing that the thread terminated abnormally and
   * ptr is freed or not. Since this is a one time memory leak, it
   * is tolerable.
   */
  /* free(ptr); */
  
  return NULL;
} /* ReceiveUDP */

void* ProcessUDPPacket(void *arg) 
{

  UDPWorkerData* workerData = (UDPWorkerData*) arg;

  request_common_t *r;     /* incoming request */
  int offset = 0;          /* offset into message */  
  int content_length = 0;  /* content_length */
  DString body;            /* space for message body */
  char* msg;

  sock_t dum_s;

  /* 
   * Extract all requests from message.  For invalid messages, the
   * header length is zero; we must exit in that case to avoid looping
   * forever.
   */
  msg = workerData->msg;

  do {
  
    r = (request_common_t *) RequestCreate(SOCK_DGRAM, 
					   workerData->sock_response, 
					   dum_s);
					   //workerData->sin);
    r->sin.tcp_s = workerData->sin;

    r->server = workerData->server; /* server/port at which it was recvd */

    if (HTTP_Parse(&msg[offset], &r->p) < 0) {
        /* If this was a request whose headers didn't parse, we want
         * to try to send a 400 response to it.  On the other hand,
         * there's not much we can do with responses that didn't parse,
         * so just mark them as the unknown message type so that the
         * server ignores them and frees their memory.
         */
        if (r->p.type == http_request) {
          ResponseSetStatus((request_t *)r, 400, NULL);
        } else {
          r->p.type = http_unknown_type;
        }
      }
      offset += r->p.header_length;
      content_length = HTTP_ContentLength(&r->p);
      if (content_length > 0) {
        if (offset + content_length > workerData->msglen) 
	  content_length = workerData->msglen - offset;
        DStringInit(&body);
        DStringAppend(&body, &msg[offset], content_length);
        r->request_body.content = DStringValue(&body);
        r->request_body.content_length = content_length;
      }
      offset += content_length;
      
      RequestProcess((request_t *)r);
      
  } while ((offset+1 < workerData->msglen) && (r->p.header_length > 0));

  debug("ProcessUDPPacket", "Finished parse and schedule...\n");
  
  free(workerData->msg);
  workerData->msg = NULL;
  free(workerData);
  
  return NULL;
}

/*
 * CloseUDP
 * -Should be called during cleanup.
*/
void CloseUDP() {

  closesocket(sock);

  closesocket(sock_response);

}
