/*
uri.c - parse a URI

Parse a URI of form scheme://user@host:port/path?search or
scheme:user@host;parameter?header=value&header=value.

Copyright 1998-1999 by Columbia University; all rights reserved 
*/

#include "uri.h"
#include "dstring.h"
#include "strdupn.h"
#include <ctype.h>     /* tolower() */
#include <stdlib.h>    /* atoi() */
#include <stdio.h>     /* sscanf() */
#include <string.h>    /* strcmp() */
#include <sys/types.h>
#include <netinet/in.h>  /* struct in_addr */
#include <arpa/inet.h>   /* inet_addr() */

#include "sysdep.h"

typedef enum {U_start, U_scheme, U_user, U_password, U_host, U_port, 
  U_path, U_pname, U_pvalue, U_search, U_hname, U_hvalue, U_done} state_t;

static struct {
  char *delimiter;
  state_t next[5];
} transition[] = {
  {" ",    {U_start}},                                  /* U_start */
  {":",    {U_user}},                                   /* U_scheme */
  {":@/;?",{U_password,U_host,U_path,U_pname,U_hname}}, /* U_user */
  {"@/;?", {U_host,U_path,U_pname,U_hname}},            /* U_password */
  {":/;?", {U_port,     U_path,   U_pname,  U_hname}},  /* U_host */
  {"/;?",  {U_path,     U_pname,  U_hname}},            /* U_port */
  {"?",    {U_search}},                                 /* U_path */
  {"=;?",  {U_pvalue,   U_pname,  U_hname}},            /* U_pname */
  {";?",   {U_pname,    U_hname}},                      /* U_pvalue */
  {"",     {U_done}},                                   /* U_search */
  {"=&",   {U_hvalue,   U_hname}},                      /* U_hname */
  {"&",    {U_hname}}                                   /* U_hvalue */
};


/*
 * encoded()
 * Convert hex-encoded character.
 */
static char encoded(unsigned const char **s)
{
  int c_value = 0;

  sscanf((*s)+1, "%2x", &c_value);
  *s += 2;
  return c_value;
} /* encoded */


/*
* Compute the next state based on the current state, the delimiting
* character and the general URI delimiter.
*/
static state_t nextstate(state_t current, char c, char delimiter)
{
  char *s;
  int n;

  if (current > U_start && (c == delimiter || c == ' ')) {
    return U_done;
  }
  else {
    s = strchr(transition[current].delimiter, c); 
    n = s - transition[current].delimiter; 
    return transition[current].next[n];
  }
} /* nextstate */


/*
 * Parse a generic URI and normalize it (remove %xx, convert scheme and
 * host name to lower case).  Recognize * as valid. Recognize <URI>.
 * URI returned must be freed by caller.
 * Strip leading white space.
 */
uri_t URI_Parse(const char *url)
{
  uri_t u;
  unsigned const char *s;
  DString v;
  char c;
  int len = strlen(url);
  int parameters = 0;    /* number of parameters */
  int headers = 0;       /* number of headers */
  int delimiter = 0;     /* URI enclosed in <> or "" */
  state_t state = U_start;

  memset(&u, '\0', sizeof(uri_t));

  if (!strcmp(url, "*")) {
    u.scheme = strdup("*");
    return u;
  }

  DStringInit(&v);
  for (s = url; ((char *)s - url) <= len && state < U_done; s++) {
    c = *s;

    /* Check if URL ends; normally, terminated by space, but also
       handles URLs enclosed in <> or "". */
    if (isspace((int)c)) c = ' ';

    if (state == U_start && c != ' ') {
      if (c == '<' || c == '"') {
        delimiter = (c == '<') ? '>' : c;
      }
      else s--;
      state = U_scheme;
    }
    else if (strchr(transition[state].delimiter, c) || c == delimiter || c == ' ') {
      switch (state) {
        case U_start:
          break;

        case U_scheme:
          if (*s && *(s+1) && *(s+2) && *(s+1) == '/' && *(s+2) == '/') {
            /* Skip over // before user/host */
            s += 2;
          }
          u.scheme = DStringValue(&v);
          break;

        case U_user:
          u.user = DStringValue(&v);
          break;

        case U_password:
          u.password = DStringValue(&v);
          break;

        case U_host:
          u.host = DStringValue(&v);
          break;

        case U_port:
          if (DStringValue(&v)) u.port = atoi(DStringValue(&v));
          DStringFree(&v);
          break;

        case U_path:
          u.path = DStringValue(&v);
          break;

        case U_search:
          u.search = DStringValue(&v);
	  u.param = strdup(u.search);
          break;
       
        case U_pname:
          parameters++;
          if (parameters == 1) {
            u.parameters = calloc(2, sizeof(uri_nv_t));
          }
          else {
            u.parameters = realloc(u.parameters, (parameters+1)*sizeof(uri_nv_t));
            u.parameters[parameters].name  = NULL;
            u.parameters[parameters].value = NULL;
          }
          u.parameters[parameters-1].name  = DStringValue(&v);
          u.parameters[parameters-1].value = NULL;
          break;

        case U_pvalue:
          u.parameters[parameters-1].value = DStringValue(&v);
          break;

        case U_hname:
          headers++;
          if (headers == 1) {
            u.headers = calloc(2, sizeof(uri_nv_t));
          }
          else {
            u.headers = realloc(u.headers, (headers+1)*sizeof(uri_nv_t));
            u.headers[headers].name  = NULL;
            u.headers[headers].value = NULL;
          }
          u.headers[headers-1].name  = DStringValue(&v);
          u.headers[headers-1].value = NULL;
          break;

        case U_hvalue:
          u.headers[headers-1].value = DStringValue(&v);
          break;

        case U_done:
          break;
      } /* switch */
      state = nextstate(state, c, delimiter);
      DStringInit(&v);
    } /* if */
    else {
      if (c == '%') c = encoded(&s);
      DStringAppend(&v, &c, 1);
    }
  } /* for */

  /* If no host name, this was really 'scheme:[//]host:port'. */
  if (!u.host) {
    u.host = u.user;
    u.user = NULL;
    if (u.password) {
      u.port = atoi(u.password);
      free(u.password);
      u.password = NULL;
    }
  }

  return u;
} /* URI_Parse */


static const uri_nv_t dummy_uri_nv[] = {{NULL, NULL}};

static const uri_defaults_t dummy_uri_defaults = {
  NULL,
  0,
  dummy_uri_nv
};


/*
 * get_scheme_defaults()
 * Given a NULL-scheme-terminated list of uri defaults, return the one whose
 * scheme corresponds to 'scheme', or else a pointer to dummy.
 */
static const uri_defaults_t*
get_scheme_defaults(const uri_defaults_t *defaults, const char *scheme)
{
  int i;

  if (defaults == NULL)
    return &dummy_uri_defaults;

  for (i = 0; defaults[i].scheme != NULL; i++) {
    if (strcompare(scheme, defaults[i].scheme, 1) == 0) {
      return &defaults[i];
    }
  }

  return &dummy_uri_defaults;
} /* get_scheme_defaults */


/*
 * cmp_nv()
 * Compare two uri_nv_t's.
 */
static int cmp_nv(const uri_nv_t *nv1, const uri_nv_t *nv2) 
{
  return strcompare(nv1->name, nv2->name, 1);
} /* cmp_nv */


/*
 * nvlen()
 * Return the size of a NULL-terminated nvset.
 */
static int nvlen(const uri_nv_t *set)
{
  int i;

  for (i = 0; set[i].name != NULL; i++) 
    ; /* Do nothing */

  return i;
} /* nvlen */


/*
 * copy_and_sort_nvs()
 * Copy a NULL-terminated set of uri_nv_t's, and sort the copy.
 * If size == -1, compute nvlen ourselves.
 */
static uri_nv_t *copy_and_sort_nvs(const uri_nv_t *orig, int size)
{
  uri_nv_t *copy;

  if (size == -1) {
    size = nvlen(orig);
  }

  copy = (uri_nv_t *) malloc(sizeof(uri_nv_t) * (size + 1));
  if (copy == NULL) return NULL;

  memcpy(copy, orig, sizeof(uri_nv_t) * (size + 1));

  qsort((void *)copy, size, sizeof(uri_nv_t),
        (int(*)(const void*, const void*))cmp_nv);

  return copy;
} /* copy_and_sort_nvs */


/*
 * URI_NV_Equal()
 * Return 1 if URI name-value sets nv1 and nv2 are equal, otherwise 0.
 * Order of name-value pairs is irrelevant; names are case-insensitive.
 */
int URI_NV_Equal(const uri_nv_t *nv1, const uri_nv_t *nv2,
                 const uri_nv_t *defaults)
{
  int nv1_size, nv2_size, defaults_size;
  uri_nv_t *nv1_copy, *nv2_copy, *defaults_copy;
  int i, j, k;
  int ret = 1;

  if (nv1 == NULL && nv2 == NULL)
    return 1;

  if (nv1 == NULL)      nv1 = dummy_uri_nv;
  if (nv2 == NULL)      nv2 = dummy_uri_nv;
  if (defaults == NULL) defaults = dummy_uri_nv;

  nv1_size = nvlen(nv1);
  nv2_size = nvlen(nv2);
  defaults_size = nvlen(defaults);

  if (defaults_size == 0 && nv1_size != nv2_size)
    return 0;

  if (nv1_size == 0 && nv2_size == 0)
    return 1;

  nv1_copy      = copy_and_sort_nvs(nv1, nv1_size);
  nv2_copy      = copy_and_sort_nvs(nv2, nv2_size);
  defaults_copy = copy_and_sort_nvs(defaults, defaults_size);

  /* Triple traversal.  i walks nv1, j walks nv2, k walks defaults. */
  i = 0; j = 0; k = 0;
  while (i < nv1_size || j < nv2_size) {
    int rel, reldef;
    rel = cmp_nv(&nv1_copy[i], &nv2_copy[j]);
    if (rel == 0) {
      /* Same name for nv1 and nv2 */
      if (strcompare(nv1_copy[i].value, nv2_copy[j].value, 0) != 0) {
        ret = 0;
        break;
      }
      i++; j++;
    }
    else {
      int *thisidx;
      uri_nv_t **thisnv;

      if (k >= defaults_size) {
        /* No more defaults left; nv1 and nv2 are different */
        ret = 0;
        break;
      }

      /* Work with whichever nv's name is lexicographically earlier */
      if (rel < 0) {
        thisidx = &i;
        thisnv = &nv1_copy;
      }
      else /* rel > 0 */ {
        thisidx = &j;
        thisnv = &nv2_copy;
      }
      reldef = cmp_nv(&(*thisnv)[*thisidx], &defaults_copy[k]);
      if (reldef == 0) {
        /* (*thisnv)[*thisidx] is a name with a default. */
        if (strcompare((*thisnv)[*thisidx].value,
                       defaults_copy[k].value, 0) != 0) {
          /* nv's value doesn't match default; nv1 and nv2 are different */
          ret = 0;
          break;
        }
        (*thisidx)++; k++;
      }
      else if (reldef < 0) {
        /* (*thisnv)[*thisidx] matches neither the other nv, nor the defaults.
         * nv1 and nv2 are different */ 
        ret = 0;
        break;
      }
      else /* reldev > 0 */ {
        /* default doesn't appear in either nv1 or nv2.  Try next default, 
         * leave nv1 and nv2 as they are. */
        k++;
      }
    }
  }

  free(nv1_copy);
  free(nv2_copy);
  free(defaults_copy);
  return ret;
} /* URI_NV_Equal */


/*
 * URI_Equal_With_Defaults()
 * Return 1 if URI's u1 and u2 are equal, otherwise 0.
 * u1 and u2 must have been parsed with URI_Parse, or otherwise have had all
 * %-encodings decoded.
 * Matches if URI 'u2' has a null search string and 'u1' has a non-null
 * search string. 
 * Schemes, hostnames, and parameter and header names are compared
 * case-insensitively; everything else is case-sensitive.
 * Order of headers and parameters is irrelevant.
 * If default_port is not 0, consider port == 0 to be equivalent to
 * default_port.
 * If param_defaults is not NULL, use it as default values of parameters.
 */
int URI_Equal_With_Defaults(const uri_t u1, const uri_t u2, 
                            const uri_defaults_t *defaults)
{
  const uri_defaults_t *scheme_defaults;

  if ((strcompare(u1.scheme, u2.scheme, 1) != 0) ||
      (strcompare(u1.user,   u2.user,   0) != 0) ||
      (strcompare(u1.host,   u2.host,   1) != 0) ||
      (strcompare(u1.path,   u2.path,   0) != 0))
    return 0;

  scheme_defaults = get_scheme_defaults(defaults, u1.scheme);

  if (!((u1.port == u2.port) ||
        (u1.port == 0 && u2.port == scheme_defaults->default_port) ||
        (u2.port == 0 && u1.port == scheme_defaults->default_port)))
    return 0;

  if (u2.search != NULL &&
      strcompare(u1.search, u2.search, 0) != 0)
    return 0;

  if (!URI_NV_Equal(u1.parameters, u2.parameters,
                    scheme_defaults->param_defaults))
    return 0;

  if (!URI_NV_Equal(u1.headers, u2.headers, NULL))
    return 0;

  return 1;
} /* URI_Equal_With_Defaults */


/*
 * URI_Equal()
 * A wrapper for URI_Equal_With_Defaults() that doesn't use any
 * scheme defaults.
 */
int URI_Equal(const uri_t u1, const uri_t u2)
{
  return URI_Equal_With_Defaults(u1, u2, NULL);
} /* URI_Equal */


/*
 * downcase()
 * convert a string (in place) to all-lower-case
 */
static void downcase(char *str)
{
  if (str == NULL) return;

  for ( ; *str != '\0'; str++) {
    *str = tolower((int)*str);
  }
} /* downcase() */


/*
 * URI_Canonicalize()
 * Convert a uri_t (in place) to a canonical representation -- all
 * case-insensitive fields are made all-lower-case, default parameters are
 * dropped, parameters and headers are in alphabetical order.
 */
void URI_Canonicalize(uri_t *up, const uri_defaults_t *defaults)
{
  const uri_defaults_t *scheme_defaults =
    get_scheme_defaults(defaults, up->scheme);

  downcase(up->scheme);
  downcase(up->host);

  if (up->port == scheme_defaults->default_port)
    up->port = 0;

  if (up->parameters) {
    int p_len;
    uri_nv_t *defaults_copy;
    int i, j, k;

    if (defaults->param_defaults == NULL) {
      defaults_copy = copy_and_sort_nvs(dummy_uri_nv, -1);
    }
    else {
      defaults_copy = copy_and_sort_nvs(defaults->param_defaults, -1);
    }

    p_len = nvlen(up->parameters);

    qsort((void *)up->parameters, p_len, sizeof(uri_nv_t),
          (int(*)(const void*, const void*))cmp_nv);
    /* Delete default parameters, and downcase names */
    i = 0; j = 0; k = 0;
    while (i < p_len) {
      int rel = cmp_nv(&up->parameters[i], &defaults_copy[k]);
      if (rel == 0) {
        if (strcompare(up->parameters[i].value, defaults_copy[k].value, 0)
            == 0) {
          /* Delete this entry; jump to next default */
          free_valid(up->parameters[i].name);
          free_valid(up->parameters[i].value);
          i++; k++;
          continue;
        }
      }
      else if (rel > 0) {
        /* Next default */
        k++;
        continue;
      }

      /* rel < 0, or values aren't equal */
      /* Keep this entry.  Copy downward if we've deleted anything. */
      if (j < i) {
        up->parameters[j].name = up->parameters[i].name;
        up->parameters[j].value = up->parameters[i].value;
      }
      downcase(up->parameters[j].name);
      i++; j++;
    }
    if (j < i) {
      /* Put new final character in place, then re-allocate memory
       * appropriately */
      up->parameters[j].name = NULL;
      up->parameters[j].value = NULL;
      up->parameters = realloc(up->parameters, (j + 1) * sizeof(uri_nv_t));
    }

    free(defaults_copy);
  }    

  if (up->headers) {
    int h_len;
    int i;

    h_len = nvlen(up->headers);

    qsort((void *)up->headers, h_len, sizeof(uri_nv_t),
          (int(*)(const void*, const void*))cmp_nv);

    /* Downcase names */
    for (i = 0; i < h_len; i++) {
      downcase(up->headers[i].name);
    }
  }
} /* URI_Canonicalize */


/*
 * Create a copy of the URI content.
 * Return copy.
 */
uri_t URI_Copy(const uri_t u)
{
  uri_t uu;
  int i;
  int parameters, headers;

  memset(&uu, '\0', sizeof(uri_t));

  if (u.scheme)   uu.scheme   = strdup(u.scheme);
  if (u.user)     uu.user     = strdup(u.user);
  if (u.password) uu.password = strdup(u.password);
  if (u.host)     uu.host     = strdup(u.host);
  uu.port = u.port;
  if (u.path)     uu.path     = strdup(u.path);
  if (u.param)    uu.param    = strdup(u.param);
  if (u.search)   uu.search   = strdup(u.search);
  if (u.parameters) {
    for (i = 0; u.parameters[i].name != NULL; i++) {
      ; /* Do nothing */
    }
    parameters = i;
    uu.parameters = (uri_nv_t *) calloc(parameters + 1, sizeof(uri_nv_t));
    for (i = 0; u.parameters[i].name != NULL; i++) {
      strassign(&uu.parameters[i].name, u.parameters[i].name);
      strassign(&uu.parameters[i].value, u.parameters[i].value);
    }
  }
  if (u.headers) {
    for (i = 0; u.headers[i].name != NULL; i++) {
      ; /* Do nothing */
    }
    headers = i;
    uu.headers = (uri_nv_t *) calloc(headers + 1, sizeof(uri_nv_t));
    for (i = 0; u.headers[i].name != NULL; i++) {
      strassign(&uu.headers[i].name, u.headers[i].name);
      strassign(&uu.headers[i].value, u.headers[i].value);
    }
  }
  return uu;
} /* URI_Copy */


/*
* URI_Free()
* Free URI parameter.
*/
void URI_Free(uri_t *u)
{
  int i;

  if (u->scheme) {
    free(u->scheme);
    u->scheme = NULL;
  }
  if (u->host) {
    free(u->host);
    u->host = NULL;
  }
  if (u->user) {
    free(u->user);
    u->user = NULL;
  }
  if (u->password) {
    free(u->password);
    u->password = NULL;
  }
  if (u->path) {
    free(u->path);
    u->path = NULL;
  }
  if (u->param) {
    free(u->param);
    u->param = NULL;
  }
  if (u->search) {
    free(u->search);
    u->search = NULL;
  }
  if (u->headers) {
    for (i = 0; u->headers[i].name; i++) {
      free(u->headers[i].name);
      if (u->headers[i].value) free(u->headers[i].value);
    }
    free(u->headers);
    u->headers = NULL;
  }
  if (u->parameters) {
    for (i = 0; u->parameters[i].name; i++) {
      free(u->parameters[i].name);
      if (u->parameters[i].value) free(u->parameters[i].value);
    }
    free(u->parameters);
    u->headers = NULL;
  }
} /* URI_Free */


/*
 * URI_Print()
 * Print a URI to stdout.
 */
void URI_Print(uri_t u)
{
  DString u_str;

  DStringInit(&u_str);
  URI_SPrint(&u, &u_str);

  puts(DStringValue(&u_str));

  DStringFree(&u_str);
} /* URI_Print */


/*
 * URI_AsString()
 * Return null terminated string representation of uri.
 * Caller should free the memory of returned string.
 */
char *URI_AsString(uri_t u)
{
  char *ret = NULL;
  DString u_str;

  DStringInit(&u_str);
  URI_SPrint(&u, &u_str);

  ret = DStringValue(&u_str);
  if ( ret != NULL ) {
    ret = strdup(ret);
  }

  DStringFree(&u_str);

  return ret;
} /* URI_Print */


/* 
 * Non-alphanumeric characters that may nonetheless appear unescaped in
 * a URI.  RFC 2396, sec.  2.3; also RFC 2543, sec.  C.1; also RFC 2732.
 */
static const char *uri_unreserved = "-_.!~*\'()";

/* 
 * Non-alphanumeric characters that may appear in a URI parameter
 * Somewhat pessimistic interpretation; over-quoting is okay.
 */
static const char *param_unreserved = "-_.!~*\'():@&+$[]";

/*
 * Non-alphanumeric characters that may appear in a URI header
 * Also a somewhat pessimistic interpretation.
 */
static const char *header_unreserved = "-_.!~*\'()?:@+$[]";

/* Non-alphanumeric characters that may appear in a URI path */
static const char *path_unreserved   = "-_.!~*\'();/@&=+$[]";

/* Non-alphanumeric characters that may appear in a URI query string */
static const char *query_unreserved  = "-_.!~*\'();/?:@&=+$[]";


/*
* URI_SPrint() 
* Returns a dynamic string with the URI in u_dstr.  Assumes
* u_dstr has been properly initialized.
*/
void URI_SPrint(const uri_t *u, DString *u_dstr)
{
  int i;

  DStringTrunc(u_dstr); /* truncate, so we don't append to previous content */

  if (u->scheme) {
    DStringAppend(u_dstr, u->scheme, -1);
    if (strcmp(u->scheme, "*") == 0) {
      /* Special case */
      return;
    }
    DStringAppend(u_dstr, ":", -1);
  }
  else {
#if 0 /* NO-ERROR if no scheme is present XXX 
      * This is needed for SIP API. Since H323 addresses may
      * not have a scheme. */
    /* Error: a URI without a scheme?  XXX Log an error. */
    DStringAppend(u_dstr, "xxx-error-no-scheme:", -1);
#endif
  }

  if (u->path) {
    DStringAppend(u_dstr, "//", -1);
  }

  if (u->user) {
    DStringAppendEscaped(u_dstr, u->user, uri_unreserved);
    if (u->password) {
      DStringAppend(u_dstr, ":", -1);
      DStringAppendEscaped(u_dstr, u->password, uri_unreserved);
    }
    DStringAppend(u_dstr, "@", -1);
  }

  if (u->host) {
    DStringAppend(u_dstr, u->host, -1);

    if (u->port) {
      DStringPrintf(u_dstr, ":%hu", u->port);
    }
  }

  if (u->path) {
    DStringAppend(u_dstr, "/", -1);
    DStringAppendEscaped(u_dstr, u->path, path_unreserved);
    if (u->search) {
      DStringAppend(u_dstr, "?", -1);
      DStringAppendEscaped(u_dstr, u->search, query_unreserved);
    }
  }
  else {
    for (i = 0; u->parameters && u->parameters[i].name; i++) {
      DStringAppend(u_dstr, ";", -1);
      DStringAppendEscaped(u_dstr, u->parameters[i].name, param_unreserved);
      if (u->parameters[i].value) {
        DStringAppend(u_dstr, "=", -1);
        DStringAppendEscaped(u_dstr, u->parameters[i].value, param_unreserved);
      }
    }
    for (i = 0; u->headers && u->headers[i].name; i++) {
      DStringAppend(u_dstr, (i == 0) ? "?" : "&", -1);
      DStringAppendEscaped(u_dstr, u->headers[i].name, header_unreserved);
      DStringAppend(u_dstr, "=", -1);
      if (u->headers[i].value != 0) {
        DStringAppendEscaped(u_dstr, u->headers[i].value, header_unreserved);
      }
    }
  }

  return;
} /* URI_SPrint */


/*
* URI_Parameter()
* Find URI parameter with name 'pname'.
* Return NULL if no match, otherwise value.
*/
char *URI_Parameter(uri_t u, char *pname)
{
  int i;
  uri_nv_t *p = u.parameters;

  if (p == NULL)
    return NULL;

  for (i = 0; p[i].name; i++) {
    if (strcasecmp(p[i].name, pname) == 0) return p[i].value;
  }
  return NULL;
} /* URI_Parameter */


/*
 * URI_SetParameter()
 * is used to set the parameters for the URI, typically the
 * transport parameter.
 */
void  URI_SetParameter(uri_t *u, const char *pname, const char *pvalue)
{
  int i;
  uri_nv_t *p = u->parameters;
  int size = 0;

  if ( !pname )
    return;

  if (p != NULL) {
    for (i = 0; p[i].name; i++, size++) {
      if (strcasecmp(p[i].name, pname) == 0) {
        if ( pvalue == NULL ) {
          /* This param has to be deleted */
          free_valid(p[i].value);
          p[i].value = NULL;
        }
        else {
          strassign(&p[i].value, pvalue);
        }
        return;
      }
    }
  }

  if ( pvalue != NULL ) {
    u->parameters = realloc(u->parameters, (size+2) * sizeof(uri_nv_t));
    if ( u->parameters ) {
      p = u->parameters;
      p[size+1].name = NULL;
      p[size+1].value = NULL;
      p[size].name = strdup(pname);
      p[size].value = pvalue ? strdup(pvalue) : NULL;
    }
  }
  return;
}
