/* -*- C++ -*- */
/* cinema_dummy_table.h -- A dummy database implementation, intended as a
 * sample from which real database implementations can be written. */
   
/* Copyright 2001 Columbia University.  All rights reserved. */

#include "cinema_dummy_table.h"

#include <strstream>

#include "uri.h"

using namespace std;

/*
 * cinema_dummy_table::cinema_dummy_table()
 * Construct an empty dummy_table.
 */
cinema_dummy_table::cinema_dummy_table()
{
  /* Do nothing */
}

/*
 * cinema_dummy_table::cinema_dummy_table(string)
 * Construct a cinema dummy table, based on the given URI given as a string.
 */
cinema_dummy_table::cinema_dummy_table(string uri)
{
  table_uri = URI_Parse(uri.c_str());
  if (table_uri.scheme != get_table_scheme()) {
    URI_Free(&table_uri);
    ostrstream s;
    s << "Invalid table scheme for " << get_table_scheme() << " table" << ends;
    throw cinema_invalid_arg(s);
  }
} /* cinema_dummy_table::cinema_dummy_table(string) */


/*
 * cinema_dummy_table::cinema_dummy_table(uri_t&)
 * Construct a cinema dummy table, based on the given URI given as a uri_t.
 */
cinema_dummy_table::cinema_dummy_table(uri_t& uri)
{
  if (uri.scheme != get_table_scheme()) {
    ostrstream s;
    s << "Invalid table scheme for " << get_table_scheme() << " table" << ends;
    throw cinema_invalid_arg(s);
  }

  table_uri = URI_Copy(uri);
} /* cinema_dummy_table::cinema_dummy_table(uri_t) */


/*
 * cinema_dummy_table::~cinema_dummy_table()
 * Delete a cinema dummy table, freeing its URI.
 */
cinema_dummy_table::~cinema_dummy_table()
{
  URI_Free(&table_uri);
} /* cinema_dummy_table::~cinema_dummy_table() */


/*
 * cinema_dummy_table::get_key(const uri_t *)
 * Get a database key suitable for the dummy table, based on a URI.
 */
char*
cinema_dummy_table::get_key(const uri_t *uri) const
{
  ostrstream s;
  s << "Unknown URI format (URI conversions not implemented in " <<
    get_table_scheme() << " tables)" << ends;
  throw cinema_invalid_arg(s);
} /* cinema_dummy_table::cinema_dummy_table(const uri_t *) */


/*
 * cinema_dummy_table::get_uri(const char*)
 * Get a canonical URI which represents the given database key.
 */
uri_t
cinema_dummy_table::get_uri(const char *key) const
{
  ostrstream s;
  s << "Unknown key format (key conversions not implemented in " <<
    get_table_scheme() << " tables)" << ends;
  throw cinema_invalid_arg(s);
} /* cinema_dummy_table::get_uri(const char*) */


/*
 * cinema_dummy_table::get_aliases_by_alias(const char *)
 * Get an alias entry for the given database key.
 */
alias_table_entry*
cinema_dummy_table::get_aliases_by_alias(const char *key)
{
  ostrstream s;
  s << "No such entry (Alias table not implemented in " <<
    get_table_scheme() << " tables)" << ends;
  throw cinema_notfound(s);
} /* cinema_dummy_table::get_aliases_by_alias(const char *) */


/*
 * cinema_dummy_table::get_put_by_user(const char *)
 * Get a PUT entry for the given database key.
 */
put_table_entry*
cinema_dummy_table::get_put_by_user(const char *key)
{
  ostrstream s;
  s << "No such entry (PUT table not implemented in " <<
    get_table_scheme() << " tables)" << ends;
  throw cinema_notfound(s);
} /* cinema_dummy_table::get_put_by_user(const char *) */


/*
 * cinema_dummy_table::get_contacts_by_user(const char *)
 * Get a list of contact entries for the given database key.
 */
contact_table_entry** 
cinema_dummy_table::get_contacts_by_user(const char *user, int *count)
{
  ostrstream s;
  s << "No such entry (Contact table not implemented in " <<
    get_table_scheme() << " tables)" << ends;
  throw cinema_notfound(s);
} /* cinema_dummy_table::get_contacts_by_user(const char *) */


/*
 * cinema_dummy_table::insert_contact(const char *)
 * Insert the given contact entry into the database.  If a contact already
 * exists with the given user/contact pair, replace it.
 */
void
cinema_dummy_table::insert_contact(const contact_table_entry *entry)
{
  ostrstream s;
  s << "Contacts not implemented in " << get_table_scheme() << 
    " tables" << ends;
  throw cinema_internal_err(s);
} /* cinema_dummy_table::insert_contact(const char *) */


/*
 * cinema_dummy_table::delete_contacts_by_user(const char *)
 * Delete all contacts for the given user from the database.
 * If no such contacts exist, this is a no-op.
 */
void
cinema_dummy_table::delete_contacts_by_user(const char *user)
{
  /* No contacts ever exist in a dummy table. */
  return;
} /* cinema_dummy_table::delete_contacts_by_user(const char *) */


/*
 * cinema_dummy_table::delete_contact_by_contact(const char *, const char *)
 * Delete the given user/contact from the database table.
 * If no such contact exists, this is a no op.
 */
void
cinema_dummy_table::delete_contact_by_contact(const char *user,
                                              const char *contact)
{
  /* No contacts ever exist in the dummy table. */
  return;
} /* cinema_dummy_table::delete_contact_by_contact(const char *,
   *                                               const char *) */


/*
 * cinema_dummy_table::delete_expired_contacts_by_user(const char *,
 *                                                     time_t)
 * Delete all expired contacts (those whose expires value is earlier than
 * 'expire_time') for the given user from the database.  If no such contacts
 * exist, this is a no-op.
 */
void
cinema_dummy_table::delete_expired_contacts_by_user(const char *user,
                                                    time_t expire_time)
{
  /* No contacts ever exist in a dummy table. */
  return;
} /* cinema_dummy_table::delete_expired_contacts_by_user(const char *,
   *                                                     time_t) */



sub_table_entry* cinema_dummy_table::get_sub_by_fromto(const char* from, const char* to)
{
  /* No subscriptions ever exist in a dummy table */
  return NULL;
}

void cinema_dummy_table::insert_sub(const sub_table_entry *entry) 
{
  ostrstream s;
  s << "Subscription_insert not implemented in " << get_table_scheme() << 
    " tables" << ends;
  throw cinema_internal_err(s);
}

int cinema_dummy_table::num_subscriptions(const char *to)
{
  /* no subscriptions in a dummy table */
  return 0;
}

sub_table_entry **cinema_dummy_table::get_all_subscriptions(const char* to, int *count)
{

  if (count) *count = 0;
  return NULL;
}

void cinema_dummy_table::delete_sub_by_fromto(const char* from, const char *to)
{
  return;
}
