/* -*- C++ -*- */
/* cinema_fastsql_cache.cpp -- 
 * A mysql database implementation with a fast (hopefully)
 * main-memory cache, and managing routines. 
 *
 * At the beginning this code was very compact and nice.
 * but, now due to 4 tables being supported with complicated
 * refresh semantics, it has become kind of bloated. Todo: Cut
 * out code that is already in cinema_mysql_table, try to
 * aggregate cinema_mysql_table inside this, and reduce code.
 *
 * A very nice design abstraction can be used to reduce atleast
 * 50% of code. Todo again. {sankaran}
 */

/* Copyright 2001 Columbia University.  All rights reserved. 
 */

#ifdef __linux__
#define _GNU_SOURCE /* Needed for strptime declaration from <time.h> */
#endif

#include <strstream>
#include <string>
#include <vector>

#include <cstdlib>
#include <ctime>

#include <sys/time.h>

extern "C" {

#include <pthread.h>
#include <errno.h>

#include "sysdep.h"
  
#ifndef HAVE_TIMEGM
#include "timegm.h"
#endif
  
#include "dstring.h"
#include "uri.h"
#include "strdupn.h"
#include "error.h"
}

/*
 *Max time_t value for infinite time
 */
#ifndef MAX_TIME_T
#define MAX_TIME_T  ((time_t) 0x7fffffff)
#endif

#include "mysqlconn.h"

#include "cinema_mysql_table.h"
#include "cinema_fastsql_cache.h"

static struct timeval start_time, end_time;
static long time_elapsed;

#define START() \
 gettimeofday(&start_time, NULL);
   
#define END(X) \
 gettimeofday(&end_time, NULL); \
 time_elapsed = (end_time.tv_sec - start_time.tv_sec) * 1000; \
 time_elapsed += ((end_time.tv_usec - start_time.tv_usec) / 1000); \
 debug("cache_profiler","%s took %ld ms\n",(X),time_elapsed);

using namespace std;


inline int compute_hash(const char *key)
{
  int   hash, i, len;
  
  if(!key)
    {
      error("compute_hash", "hash function called with null value");
      return 0;
    }
  len = strlen(key);

  for (hash=0, i=0; i<len; ++i)
    {
      hash += key[i];
      hash += (hash << 10);
      hash ^= (hash >> 6);
    }
  hash += (hash << 3);
  hash ^= (hash >> 11);
  hash += (hash << 15);
  return (hash & 0xffffffff);
} 

#define DEFAULT_BUCKET_ENTRIES	15
#define PARAMNAME  "REFRESH"

/*
 * pthread_ variables, declarations for managing the refresh thread.
 */
static pthread_t tid;
static pthread_mutex_t worker_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t worker_cond = PTHREAD_COND_INITIALIZER;
static volatile int runWorkerThread = 1;
static void SetupWorkerThread();            /* Sets and starts the cache refresh thread */
static void* RunCacheRefreshThread(void*);  /* the cache refresh thread's worker function */

/*
 * This is our singleton cache.
 */
static cinema_fastsql_cache* cacheTable;

/*
 * Time after which the cache gets refreshed.
 */
static int refreshPeriod;

/*
 * atexit handler
 */
static void abnormal_exit();


/*
 * Copy functions - we return only copies of our cache objects.
 * caller should free them after use.
 */
static void copy_alias(alias_table_entry *to, const alias_table_entry *from);
static void copy_put(put_table_entry *to, const put_table_entry *from);
static void copy_contacts(contact_table_entry *to, const contact_table_entry *from);
static void copy_sub(sub_table_entry *to, const sub_table_entry *from);

/*
 * These functions are the callbacks for the hash table  to compute
 * various values. 
*/
static ht_val compute_key_value(dict_key o);
static ht_val compare_putobj_values(dict_obj o1, dict_obj o2);
static ht_val compare_aliasobj_values(dict_obj o1, dict_obj o2);
static ht_val compare_contact2obj_values(dict_obj o1, dict_obj o2);
static ht_val compare_subobj_values(dict_obj o1, dict_obj o2);
static ht_val compute_putobj_value(dict_obj o);
static ht_val compute_aliasobj_value(dict_obj o);
static ht_val compute_contact2obj_value(dict_obj o);
static ht_val compute_subobj_value(dict_obj o);
static ht_val compare_putkey_values(dict_key k1, dict_obj o1);
static ht_val compare_aliaskey_values(dict_key k1, dict_obj o1);
static ht_val compare_contact2key_values(dict_key k1, dict_obj o1);
static ht_val compare_subkey_values(dict_key k1, dict_obj o1);

/*
 * factory_mutex is for the class factory. if two threads call
 * get_ref at the same time this serializes them. right now,
 * in the current sipd model this wont happen. but, just in case.
 */
static pthread_mutex_t factory_mutex = PTHREAD_MUTEX_INITIALIZER;



/*
 * init_fastsql_module
 *
 * Set up the singleton cache object if it has not been done before.
 * 
 * Parse the fastsql url, check if the REFRESH param is there, if
 * yes, set the global variable to suitable value. 
 *
 * Create the lazy writer/reader thread that takes care of periodically
 * calling the cache refresh functions.
 *
 * Thread safety is needed as the singleton construction should be
 * done only once, hence use pthread_mymutex_stack_obj.
 */
void cinema_fastsql_cache::init_fastsql_module(string uri)
{
  
  pthread_mymutex_stack_obj lock(&factory_mutex, __LINE__); 
  
  if (cacheTable) return;
  
  char *refresh = strstr((char*)uri.c_str(),PARAMNAME);
  
  if (refresh) {
    refresh = strchr(refresh,'=');
    if(refresh) {
      ++refresh;
      if (refresh)
	refreshPeriod = atoi(refresh);
      else
	debug("init_fastsql_module", "malformed param ignored");
    }
  }
  
  if (refreshPeriod <= 0) refreshPeriod = 3600;
  
  debug("init_fastsql_module", "Cache refresh set to %d\n", refreshPeriod);
  
  START();
  
  cacheTable = new cinema_fastsql_cache(uri, refreshPeriod);
  
  END("Cache loading time");
  
  SetupWorkerThread();
  
  return;
  
} /* init_fastsql_module */


/*
 * init_fastsql_module : See above.
 */
void cinema_fastsql_cache::init_fastsql_module(uri_t &uri)
{
  
  pthread_mymutex_stack_obj lock(&factory_mutex, __LINE__);
  
  if (cacheTable) return;
  
  if (uri.search) { 
    char *refresh = strstr((char*)uri.search,PARAMNAME);
      
    if (refresh) {
      refresh = strchr(refresh,'=');
      if (refresh) {
	++refresh;
	if (refresh)
	  refreshPeriod = atoi(refresh);
	else
	  debug("init_fastsql_module", "malformed param ignored");
      }
    }
  }
  
  if (refreshPeriod <= 0) refreshPeriod = 3600;
  
  debug("init_fastsql_module", "Cache refresh set to %d\n", refreshPeriod);
  
  START();
  
  cacheTable = new cinema_fastsql_cache(uri, refreshPeriod);

  END("Cache loading time");
  
  SetupWorkerThread();
  
  return;
  
} /* init_fastsql_module */


/*
 * get_fastsql_ref
 *
 * -Call init_fastsql_module, and then return the cacheTable global
 * variable.
 */
const cinema_fastsql_cache* 
cinema_fastsql_cache::get_fastsql_ref(string uri)
{
  
  cinema_fastsql_cache::init_fastsql_module(uri);
  
  return (const cinema_fastsql_cache*)cacheTable;
  
} /* get_fastsql_ref */


/*
 * get_fastsql_ref
 */
const cinema_fastsql_cache* 
cinema_fastsql_cache::get_fastsql_ref(uri_t& uri)
{
  cinema_fastsql_cache::init_fastsql_module(uri);
  
  return (const cinema_fastsql_cache*)cacheTable;
  
} /* get_fastsql_ref */


/*
 * close_fastsql_module
 *
 * Call this during exit to cleanup everything.
 */
void cinema_fastsql_cache::close_fastsql_module()
{
  
  pthread_mymutex_stack_obj lock(&factory_mutex, __LINE__);
  /*
   * Algorithm:
   * -Stop the refresh thread.
   * -Cleanup the cache.
   * -Free the object.
   * -This should be done in a single threaded fashion.
   */
  
  /*
   * already called. do nothing
   */
  if(!runWorkerThread) return;
  
  runWorkerThread = 0;
  
  /*
   * signal the worker thread.
   */
  pthread_cond_signal(&worker_cond);
  
  {
    pthread_t tid_arr[1];
    tid_arr[0]  = tid;
    WaitForMultipleObjects(1,tid_arr, TRUE, INFINITE);
  }
  
  debug("close_cached_table_module",
	"Finished operation...thread terminated, now freeing cache");

  delete cacheTable;

  debug("close_cached_table_module", "...Cached freed\n");

  cacheTable = NULL;
  
  return;
  
} /* close_cached_table_module */


/*
 * free_fastsql_ref
 *
 * For reference counting. 
 */
void 
cinema_fastsql_cache::free_fastsql_ref(const cinema_fastsql_cache* notused)
{
  return;
} /* cinema_fastsql_cache::free_fastsql_ref */

/*
 * refresh_cache
 *
 * Externally visible "forced refresh".
 *
 */ 
void cinema_fastsql_cache::refresh_cache()
{
  
  /*
   * Signal the worker thread. It will do the needful.
   */
  debug("refresh_cache", "signalling worker to refresh...\n");
  
  pthread_cond_signal(&worker_cond);
  
  /*
    if(cacheTable)
    {
    cacheTable->propagateUpdates();
    
    cacheTable->refresh();
    }
  */

} /* refresh_cache */

/*
 * SetupWorkerThread.
 *
 * -creates a thread to periodically refresh the cache.
 */
void SetupWorkerThread()
{  
  int e;
  
  pthread_mutex_init(&worker_mutex,NULL);
  
  pthread_cond_init(&worker_cond,NULL);
  
  if (!cacheTable) {
    error("setupWorkerThread","Bad call...initialize the cache first!!!\n");
    return;
  }
  
  tid = 0;
  
  if ((e = pthread_create(&tid, NULL, 
			  RunCacheRefreshThread, (void*)0)) != 0) {
    fatal("init_cached_table_module", 
	  "Cannot create cache refreshthread: %s\n",strerror(e));
  }
  
  debug("fastsql_module","Worker thread started...init over\n");
  
  /*
   * This is for abnormal exits.
   */
  atexit(abnormal_exit);
  
  return;
  
} /* SetupWorkerThread */

/*
 * abnormal_exit
 *
 * On abnormal exits (like pressing ctrl-c for ex), call the
 * close_fastsql_module for cleanup.
 */
void abnormal_exit()
{

  cinema_fastsql_cache::close_fastsql_module();
  
} /* abnormal_exit */


/*
 * RunCacheRefreshThread()
 * Start and run the cache refresh thread.
 */
static void* RunCacheRefreshThread(void* notused)
{
  struct timespec abstime;
  
  struct timeval timeval;
  
  struct timeval lastRefresh = {0, 0};
  
  int e;
  
  /*
   * Refresh or write-back the contact info to the database every
   * 120 seconds. Refresh or read-the-updates to put and alias tables
   * every refreshPeriod seconds.
   */
  while (runWorkerThread) {
    gettimeofday(&timeval,NULL);
    
    abstime.tv_sec = timeval.tv_sec + 120;
    
    abstime.tv_nsec = 0;
    
    pthread_mutex_lock(&worker_mutex);
    
    e = pthread_cond_timedwait(&worker_cond,&worker_mutex,&abstime);
    
    pthread_mutex_unlock(&worker_mutex);
    
    debug("RunWorkerThread",
	  "cond_timedwait returned: %d : %s\n",e,strerror(errno));
    
    if (e<0 && errno==EINVAL) {
      error("RunCacheRefreshThread",strerror(e));
      break;
    }
    
    if (runWorkerThread && cacheTable) {
      
      cacheTable->propagateUpdates();
	  
      if (lastRefresh.tv_sec + refreshPeriod > abstime.tv_sec) {
	
	cacheTable->refresh();
	lastRefresh.tv_sec = abstime.tv_sec;
	errno = 0;
	debug("RunCacheRefreshThread","Refreshed cache...\n");
      }   
	
    }
  }
  
  debug("RunWorkerThread","Exiting...\n");
  
  return NULL;
  
} /* RunCacheRefreshThread */


/*
 * cinema_fastsql_cache::cinema_fastsql_cache(string)
 * Construct a fastsql table, based on the given URI given as a string.
 */
cinema_fastsql_cache::cinema_fastsql_cache(string uri, int refresh_period) 
  : refresh_period(refresh_period)
{
  
  table_uri = URI_Parse((char*)uri.c_str());
  
  if (table_uri.scheme != table_scheme()) {
    URI_Free(&table_uri);
    throw cinema_invalid_arg("Invalid table scheme for mysql cached table");
  }
  
  try {
    init_connection();
  }
  catch(...) {
    URI_Free(&table_uri);
    throw;
  }
  
} /* cinema_fastsql_cache::cinema_fastsql_cache(string) */


/*
 * cinema_fastsql_cache::cinema_fastsql_cache(uri_t&)
 * Construct a fastsql table, based on the given URI given as a uri_t.
 */
cinema_fastsql_cache::cinema_fastsql_cache(uri_t& uri, int refresh_period) 
  : refresh_period(refresh_period)
{
  
  if (uri.scheme != table_scheme()) {
    throw cinema_invalid_arg("Invalid table scheme for mysql cached table");
  }
  
  table_uri = URI_Copy(uri);
  
  init_connection();
  
} /* cinema_fastsql_cache::cinema_fastsql_cache(uri_t) */


/*
 * cinema_fastsql_cache::~cinema_fastsql_cache()
 * Destruct a cinema mysql table.
 */
cinema_fastsql_cache::~cinema_fastsql_cache()
{
  
  /*
   * no need to do thread sync.
   */
  
  debug("cinema_fastsql_cache", "destructor called...");
  delete conn;
  URI_Free(&table_uri);
  
  /*
   * free the put, alias, contact, and subscription caches.
   */
  pthread_mutex_lock(&put_lock);
  
  {
    put_table_entry *p_entry, *p_entry1;
    
    p_entry = (put_table_entry*)ht_minimum(put_table);
    
    while (p_entry) {
      p_entry1 = (put_table_entry*)
	ht_successor(put_table, (dict_obj)p_entry);
      ht_delete(put_table,(dict_obj)p_entry);
      free_put_table_entry(p_entry);
      p_entry = p_entry1;
    }
  }
  
  ht_destroy(put_table);
  pthread_mutex_unlock(&put_lock);
  
  pthread_mutex_lock(&alias_lock);
  {
    alias_table_entry *p_entry, *p_entry1;
    
    p_entry = (alias_table_entry*)ht_minimum(alias_table);
    
    while (p_entry) {
      p_entry1 = (alias_table_entry*)
	ht_successor(alias_table, (dict_obj)p_entry);
      ht_delete(alias_table,(dict_obj)p_entry);
      free_alias_table_entry(p_entry);
      p_entry = p_entry1;
    }
  }
  
  ht_destroy(alias_table);
  pthread_mutex_unlock(&alias_lock);
  
  pthread_mutex_lock(&contacts_lock);
  {
    contact_table_entry2 *p_entry, *p_entry1;
    
    p_entry = (contact_table_entry2*)ht_minimum(contacts_table);
    
    while (p_entry) {
      p_entry1 = (contact_table_entry2*)
	ht_successor(contacts_table, (dict_obj)p_entry);
      ht_delete(contacts_table, (dict_obj)p_entry);
      free_contact_table_entry(p_entry->contact);
      free_valid(p_entry);
      p_entry = p_entry1;
    }
  }
  
  ht_destroy(contacts_table);
  pthread_mutex_unlock(&contacts_lock);
  
  
  pthread_mutex_lock(&sub_lock);
  {
    sub_table_entry2 *p_entry, *p_entry1;
    
    p_entry = (sub_table_entry2*)ht_minimum(sub_table);
    
    while (p_entry) {
      p_entry1 = (sub_table_entry2*)
	ht_successor(sub_table, (dict_obj)p_entry);
      ht_delete(sub_table,(dict_obj)p_entry);
      free_sub_table_entry(p_entry->sub);
      free_valid(p_entry);
      p_entry = p_entry1;
    }
  }
  
  ht_destroy(sub_table);
  pthread_mutex_unlock(&put_lock);
  
  pthread_mutex_destroy(&put_lock);
  pthread_mutex_destroy(&alias_lock);
  pthread_mutex_destroy(&contacts_lock);
  pthread_mutex_destroy(&sub_lock);
  
  return;
  
} /* cinema_fastsql_cache::~cinema_fastsql_cache() */


/*
 * cinema_fastsql_cache::init_connection()
 * Initialize conn, based on table_uri.  Throw an exception
 * on error.
 */
void
cinema_fastsql_cache::init_connection()
{
  
  if (table_uri.password != NULL &&
      strcasecmp(table_uri.password, "NULL") == 0) {
    /* Backward misfeature-compatibility -- treat the literal password
     * "NULL" as the null password.
     * This was a workaround for an earlier, broken URI parser.
     */
    free(table_uri.password);
    table_uri.password = NULL;
  }
  
  /* if no database is specified default to sip */
  if (!table_uri.path)
	strassign(&table_uri.path, "sip");

  try {
    conn =  
      new MysqlConnection(table_uri.user, table_uri.password,
                          table_uri.path, table_uri.host, table_uri.port);
    
    pthread_mutex_init(&conn_lock,NULL);
    pthread_mutex_init(&put_lock,NULL);
    pthread_mutex_init(&alias_lock,NULL);
    pthread_mutex_init(&contacts_lock,NULL);
    pthread_mutex_init(&sub_lock, NULL);
    
    /*
     * Call the primary cache loader here 
     */
    init_hashtables();

    init_cache();

  }
  catch (exception& e) {
    URI_Free(&table_uri);
    throw cinema_internal_err(string("Couldn't create fastsql table: ") +
                              e.what());
  }
  
  return;

} /* cinema_fastsql_cache::init_connection() */


void cinema_fastsql_cache::load_or_update_put(bool firsttime) 
{
  
  int succ, i;

  put_table_entry *p_putrow, *p_oldrow;
  
  char query[512] = "select user, hash_value,realm,sip_groups,"
    "auth,algorithm,sip_methods,remote_user,busy,"
    "noresponse,um_timeout,max_msgsize_kb,"
    "last_modified,gwclass from put";
  
  /*
   * Algorithm:
   *
   * If this is the firsttime, load the put table. For each record,
   * create a put_table_entry, and add it to the put_table hashtable.
   *
   * If this is not the firsttime, get the last_refresh time,  and use that
   * as the argument to the where clause of the query's last_modified
   * field. Update or add the entries returned.
   *
   * These need to be done in a fashion that access the database in a 
   * mutul. exclusive fashion. All modifications to the put_table 
   * hashtable should be m.e. again.
   */

  if (firsttime) {
    debug("load_or_update_put","first time load happening...");
    strcat(query, " order by user");
    succ = conn->rawQueryTable(query);
  }
  else {
    char date[32];
    struct tm tm;      
      
    time_t tim;
    time(&tim);
            
    tim -= refresh_period;
    localtime_r(&tim, &tm);

    strftime(date,sizeof(date),"%Y%m%d%H%M%S",&tm);
    strcat(query, " where last_modified>='");
    strcat(query, date);
    strcat(query,"' order by user");
    
    succ = conn->rawQueryTable(query);
  }
  
  try { 
    sanity_check_result(succ, false, 14, NULL , NULL, NULL);
  }catch(...)
    {
      conn->freeResultSet();
      throw;
    }
  
  for (i=0;i<conn->getNumRowsReturned();i++) {
    p_putrow = this->deserialize_next_put_row();
    pthread_mutex_lock(&put_lock);
    
    /*
     * Since the cache manager returns only copies of the
     * objects in its cache, we can safely delete it and
     * replace it with the new object.
     *
     * Also, this ht_search is protected by the put_lock,
     * so thread safe also.
     */
    p_oldrow = (put_table_entry*)
      ht_search(put_table, (dict_key)p_putrow->user);
    
    if (p_oldrow) {
      ht_delete(put_table, (dict_obj)p_oldrow);
	 
      if (ht_insert( put_table, p_putrow) != 0) {
	pthread_mutex_unlock(&put_lock);
	error("cacheRefresh","ht_insert returns non zero...\n");
	throw bad_alloc();
      }
      
      free_put_table_entry(p_oldrow);
    }
    else {
	  
      if (ht_insert(put_table,p_putrow) != 0)  {	
	pthread_mutex_unlock(&put_lock);
	throw cinema_internal_err
	  (string("ht_insert for put returns an error in init_cache "));
      }      
    }
    
    pthread_mutex_unlock(&put_lock);
    
  }
  
  conn->freeResultSet();
  
  debug("init_cache", "init_cache loaded %d rows for put\n",i);
  
} /* load_or_update_put */


/*
 * load_or_update_aliases
 *
 * Same as above, but for put table.
 */
void cinema_fastsql_cache::load_or_update_aliases(bool firsttime) 
{

  int succ,i;
  alias_table_entry *p_aliasrow, *p_old;
  char query[512];
  
  if (firsttime) {
      debug("load_or_update_aliases","first time load happening...");
      succ = conn->rawQueryTable("select * from aliases order by alias");
  }
  else {
    char date[32];
    struct tm tm;      
    time_t tim;
      
    time(&tim);
    tim -= refresh_period;
    localtime_r(&tim, &tm);
 
    strftime(date, sizeof(date), "%Y%m%d%H%M%S",&tm);
    sprintf(query,
	    "select * from aliases where last_modified>='%s' order by alias",date);
    
    succ = conn->rawQueryTable(query);
  }
 
  try { 
    sanity_check_result(succ, false, 3, NULL, NULL, "aliases");
  }catch(...)
    { 
      conn->freeResultSet();
      throw;
    }
  
  for (i=0; i<conn->getNumRowsReturned(); i++) {
      
      p_aliasrow = (alias_table_entry*)deserialize_next_alias_row();
      pthread_mutex_lock(&alias_lock);
      p_old = (alias_table_entry*)
	ht_search(alias_table, (dict_key)p_aliasrow->alias);
      
      if (p_old) {
	ht_delete(alias_table, (dict_obj)p_old);
	
	if (ht_insert(alias_table, p_aliasrow) != 0)
	  error("aliasHandler", "ht_insert returns non zero\n");
	  	  
	free_alias_table_entry(p_old);
      }
      else {
	if (ht_insert(alias_table,p_aliasrow) != 0) {
	  pthread_mutex_unlock(&alias_lock);
	      
	  throw cinema_internal_err(
				    string("ht_insert for alias returns an error in init_cache"));
	}
      }
      
      pthread_mutex_unlock(&alias_lock);
  }
  
  conn->freeResultSet();
  
  debug("init_cache", "init_cache loaded/updated %d rows for aliases\n",i);
  
} /* load_or_update_alias_table */


/*
 * load_contacts
 *
 * Loads the contact_table hash table with the records from the
 * contacts table of sip database.
 */
void cinema_fastsql_cache::load_or_update_contacts(bool firsttime) 
{   
  
  int succ, i, found;
  contact_table_entry* p_contactsrow;
  contact_table_entry2 *p_temp, *p_old;
  
  /*
   * Algorithm:
   * -Read the contact table, for each record construct one
   *  contact_table_entry object. Wrap it inside contact_table_entry2
   * object, put it in the hashtable contact_table.
   *
   * This needs to be done in a thread safe manner.
   */

  char query[512] = "select user, contact, expires, q, action,"
    "last_modified, display_name, sip_methods "
    " from contacts";
  
  if (firsttime) {
      strcat(query," order by user");
      debug("load_contacts","first time load happening...");
      
      succ = conn->rawQueryTable(query);
      
      sanity_check_result(succ, false, 8, "user", NULL, "contacts");
      
      for (i = 0; i < conn->getNumRowsReturned(); i++) {
	
	p_contactsrow = deserialize_next_contact_row();
	
	p_temp = (contact_table_entry2*)
	  calloc(1,sizeof(contact_table_entry2));
	
	if (!p_temp) throw bad_alloc();
	
	p_temp->contact = p_contactsrow;
	
	pthread_mutex_lock(&contacts_lock);
	
	if (ht_insert(contacts_table,p_temp) != 0)
	  throw cinema_error(string("load_or_update_contacts ht_insert returns an error"));
	
	pthread_mutex_unlock(&contacts_lock);
      }
      
      conn->freeResultSet();
      
      debug("init_cache", "init_cache loaded %d rows for contacts\n",i); 
      
      return;
  }
  
  /* this is the else part - comes when firsttime == false */
  char date[32];
  struct tm tm;      
  
  localtime_r(&last_refresh, &tm);
  strftime(date,sizeof(date),"%Y%m%d%H%M%S",&tm);
  
  strcat(query," where last_modified>'");
  strcat(query,date);
  strcat(query,"' order by user");

  succ = conn->rawQueryTable(query);
  
  try {
    sanity_check_result(succ, false, 8, "user", NULL, "contacts");
  }catch(...)
    {
      conn->freeResultSet();
      throw;
    }

  for (i = 0; i < conn->getNumRowsReturned(); i++) {
    
    p_contactsrow = deserialize_next_contact_row();
    pthread_mutex_lock(&contacts_lock);
    found = 0;
    
    p_old = (contact_table_entry2*)
      ht_search(contacts_table, (dict_key)p_contactsrow->user);
      
    while (p_old) {
	    
      if (strcasecmp(p_old->contact->contact, 
		     p_contactsrow->contact) == 0) {
	found = 1;
	break;
      }
	
      p_old = (contact_table_entry2*)ht_successor(contacts_table,
						  p_old);
      if (p_old && (strcasecmp(p_old->contact->user,
			       p_contactsrow->user) != 0))
	break;
    }
	
    p_temp = (contact_table_entry2*)
      calloc(1, sizeof(contact_table_entry2));
    
    if (!p_temp) {
      pthread_mutex_unlock(&contacts_lock);
      throw bad_alloc();
    }
    
    p_temp->contact = p_contactsrow;
    
    if (found)  {
      ht_delete(contacts_table, (dict_obj)p_old);
      free_contact_table_entry(p_old->contact);
      p_old->contact = NULL;
      free_valid(p_old);
    }
   
    if (ht_insert(contacts_table, p_temp) != 0)
      error("update_contacts","ht_insert fails...");
       
    pthread_mutex_unlock(&contacts_lock);
  }
  
  conn->freeResultSet();
  
  debug("load_or_update_contacts", "updated %d rows for contacts\n",i);
  
  return;
  
} /* load_or_update_contacts_table */

/*
 * load_or_update_sub
 *
 * Loads the sub_table hash table with the records from the
 * subscription table of sip database.
 */
void cinema_fastsql_cache::load_or_update_sub(bool firsttime) 
{   
  
  int succ, i, found;
  sub_table_entry* p_subrow;
  sub_table_entry2 *p_temp, *p_old;
  
  /*
   * Algorithm:
   * -Read the subscription table, for each record construct one
   *  sub_table_entry object. Wrap it inside sub_table_entry2
   * object, put it in the hashtable sub_table.
   *
   * This needs to be done in a thread safe manner.
   */
  
  char query[512] = "select sub_to, sub_from, expires, contact"
    ", call_id, record_route, last_modified from subscription";
  
  if (firsttime) {
    strcat(query," order by sub_to");
    debug("load_subscription","first time load happening...");
      
    succ = conn->rawQueryTable(query);
      
    sanity_check_result(succ, false, 7, "sub_to", NULL, "subscription");
      
    for (i = 0; i < conn->getNumRowsReturned(); i++) {
	
      p_subrow = deserialize_next_sub_row();
	
      p_temp = (sub_table_entry2*)
	calloc(1,sizeof(sub_table_entry2));
	
      if (!p_temp) throw bad_alloc();
	
      p_temp->sub = p_subrow;
      pthread_mutex_lock(&sub_lock);
	
      if (ht_insert(sub_table,p_temp) != 0)
	throw cinema_error(string("load_or_update_syb ht_insert returns an error"));
      
      pthread_mutex_unlock(&sub_lock);
    }

    conn->freeResultSet();
    
    debug("init_cache", "init_cache loaded %d rows for subscription\n",i); 
    
    return;
  }
  
  /* this is the else part - comes when firsttime == false */
  
  char date[32];
  struct tm tm;      
  
  localtime_r(&last_refresh, &tm);
  strftime(date,sizeof(date),"%Y%m%d%H%M%S",&tm);
            
  strcat(query," where last_modified>'");
  strcat(query,date);
  strcat(query,"' order by sub_to");

  succ = conn->rawQueryTable(query);
  
  try {
    sanity_check_result(succ, false, 7, "sub_to", NULL, "subscription");
  }catch(...)
    {
      conn->freeResultSet();
      throw;
    }

  for (i = 0; i < conn->getNumRowsReturned(); i++) {
    
    p_subrow = deserialize_next_sub_row();
    pthread_mutex_lock(&sub_lock);
    found = 0;
    
    p_old = (sub_table_entry2*)
      ht_search(sub_table, (dict_key)p_subrow->sub_to);
    
    while (p_old) {
      
      if (strcasecmp(p_old->sub->sub_from, 
		     p_subrow->sub_from) == 0) {
	found = 1;
	break;
      }
      
      p_old = (sub_table_entry2*)ht_successor(sub_table,
					      p_old);
      if (p_old && (strcasecmp(p_old->sub->sub_to,
			       p_subrow->sub_to) != 0))
	break;
    }
    
    p_temp = (sub_table_entry2*)
      calloc(1, sizeof(sub_table_entry2));
    
    if (!p_temp)  {
      pthread_mutex_unlock(&sub_lock);
      throw bad_alloc();
    }
    
    p_temp->sub = p_subrow;
    
    if (found)  {
      ht_delete(sub_table, (dict_obj)p_old);
      free_sub_table_entry(p_old->sub);
      p_old->sub = NULL;
      free_valid(p_old);
    }
    
    if (ht_insert(sub_table, p_temp) != 0) 
      error("update_sub","ht_insert fails...");
 
    pthread_mutex_unlock(&sub_lock);
  }
  
  conn->freeResultSet();
  
  debug("load_or_update_sub", "updated %d rows for subscription\n",i);
  
  return;
  
} /* load_or_update_sub_table */


/*
 * init_cache
 *
 * Initialize the cache. 
 */
void cinema_fastsql_cache::init_cache() 
{

  pthread_mymutex_stack_obj lock(&conn_lock, __LINE__);
  
  load_or_update_put(true);
  load_or_update_aliases(true);
  load_or_update_contacts(true);
  load_or_update_sub(true);
  
  time(&last_refresh); // time we did our first load
  
  return;
  
} /* init_cache */ 


/*
 * cinema_fastsql_cache::deserialize_next_sub_row()
 * Get the next MYSQL_ROW from the connection, and convert it to an
 * sub_table_entry*.
 * The previous query must have been for the subscription table.
 */
sub_table_entry*
cinema_fastsql_cache::deserialize_next_sub_row()
{
  sub_table_entry* result = (sub_table_entry *)
    calloc(1, sizeof(sub_table_entry));
  
  if (result == NULL) throw bad_alloc();
  
  MYSQL_ROW row = conn->getNextRow();
  
  strassign(&result->sub_to, row[0]);
  strassign(&result->sub_from, row[1]);
  
  if (row[2] != NULL) {
    result->expires = cinema_mysql_table::sql_datetime_to_time(row[2]);
  }
  else {
    result->expires = MAX_TIME_T;
  }
  
  strassign(&result->contact, row[3]);
  strassign(&result->call_id, row[4]);
  strassign(&result->record_route, row[5]);

  result->last_modified = cinema_mysql_table::sql_timestamp14_to_time(row[6]);
  
  return result;
} /* cinema_fastsql_cache::deserialize_next_sub_row */

/*
 * deserialize_next_alias_row
 * 
 * Exact as cinema_mysql_table.
 * cinema_fastsql_cache::deserialize_next_alias_row()
 * Get the next MYSQL_ROW from the connection, and convert it to an
 * alias_table_entry*
 * The previous query must have been for the alias table.
 */
alias_table_entry*
cinema_fastsql_cache::deserialize_next_alias_row()
{
  alias_table_entry* result = (alias_table_entry *)
    calloc(1, sizeof(alias_table_entry));
  
  if (result == NULL) throw bad_alloc();
  
  MYSQL_ROW row = conn->getNextRow();
  
  strassign(&result->alias, row[0]);
  
  strassign(&result->primary_user, row[1]);
  
  result->last_modified = cinema_mysql_table::sql_timestamp14_to_time(row[2]);
  
  return result;
  
} /* cinema_fastsql_cache::deserialize_next_alias_row */


/*
 * same as cinema_mysql_table except for message_template field
 * which is not here.
 * 
 * cinema_fastsql_cache::deserialize_next_put_row()
 *
 * Get the next MYSQL_ROW from the connection, and convert it to an
 * put_table_entry*.
 *
 * The previous query must have been for the PUT table.
 */
put_table_entry*
cinema_fastsql_cache::deserialize_next_put_row()
{
  put_table_entry* result = (put_table_entry *)
    calloc(1, sizeof(put_table_entry));
  
  if (result == NULL) throw bad_alloc();
  
  MYSQL_ROW row = conn->getNextRow();

  strassign(&result->user, row[0]);

  strassign(&result->hash_value, row[1]);
  
  strassign(&result->realm, row[2]);

  result->sip_groups = cinema_mysql_table::parse_sip_groups(row[3]);

  result->auth = cinema_mysql_table::parse_auth(row[4]);

  strassign(&result->algorithm, row[5]);
  
  result->sip_methods = cinema_mysql_table::parse_sip_methods(row[6]);

  result->remote_user = strvec_to_c_strs(split_string(row[7], " "));

  strassign(&result->busy, row[8]);
  
  strassign(&result->noresponse, row[9]);
  
  if (row[10] != NULL) {
    istrstream r10(row[10]);
    r10 >> result->um_timeout;
  }
  else {
    result->um_timeout = 0; /* XXX ??? */
  }

  istrstream r11(row[11]);
  r11 >> result->max_msgsize_kb;

  result->last_modified = cinema_mysql_table::sql_timestamp14_to_time(row[12]);

  strassign(&result->gwclass, row[13]);
  
  return result;

} /* cinema_fastsql_cache::deserialize_next_put_row */


/*
 * cinema_mysql_table::deserialize_next_contact_row()
 * Get the next MYSQL_ROW from the connection, and convert it to an
 * contact_table_entry*.
 * The previous query must have been for the contacts table.
 */
contact_table_entry*
cinema_fastsql_cache::deserialize_next_contact_row()
{
  contact_table_entry* result = (contact_table_entry *)
    calloc(1, sizeof(contact_table_entry));
  
  if (result == NULL) throw bad_alloc();

  MYSQL_ROW row = conn->getNextRow();

  strassign(&result->user, row[0]);
  
  strassign(&result->contact, row[1]);

  if (row[2] != NULL) {
    result->expires = cinema_mysql_table::sql_datetime_to_time(row[2]);
  }
  else {
    result->expires = MAX_TIME_T;
  }

  if (row[3] != NULL) {
    istrstream r3(row[3]);
    r3 >> result->q;
  }
  else {
    result->q = 1.0;
  }

  result->action = cinema_mysql_table::parse_action(row[4]);

  result->last_modified = cinema_mysql_table::sql_timestamp14_to_time(row[5]);

  strassign(&result->display_name, row[6]);

  strassign(&result->sip_methods, row[7]);

  return result;

} /* cinema_fastsql_cache::deserialize_next_contact_row() */


/*
 * refresh
 *
 * Calling this method, makes this object to refresh its internal
 * data.
 */
void cinema_fastsql_cache::refresh() 
{

  pthread_mymutex_stack_obj lock(&conn_lock, __LINE__);

  START();
  
  try {
      load_or_update_put(false);
      load_or_update_aliases(false);
      
  }catch(exception &e)
    {
      error("refresh","cache refresh failed %s",(char*)e.what());
    }
  
  END("put_alias_refresh");
  
} /* refresh */

  
/*
 *
 * propagateUpdates
 *
 * Calling this method, will make the object to flush dirty data back
 * to the database.
 */
void cinema_fastsql_cache::propagateUpdates()
{
  /*
   * Algorithm:
   * Acquire the contacts, subscription and database mutexes.
   *
   * -First load all the entries that were modified after we
   * updated. These modifications would have been done by sipua
   * for example.
   *
   * -Then,
   * Iterate the contacts hashed table, and flush the dirty data.
   *   -needful is three cases:
   *   -insert, update or replace, and delete.
   *
   * There are two tricky items to take care here:
   * -When deleting an element, we remove it from hash table. 
   * this means that we cant  use the ht_successor. so, 
   * get the succ. before deleting.
   *
   * -I dont want the hashtable for contacts to be locked 
   * while we are doing queries. since this is the only place that d
   * does changes to the hash table (Viz. deletes),
   * i release the contacts mutex, run the query and get it back. 
   * during this interval, somebody may even insert a contact entry. 
   * but this doesnt matter, as before making the next ht_successor 
   * call, i get the mutex back.
   
   * same as above for subscription table.
   */

  int total = 0, succ;

  pthread_mymutex_stack_obj lock(&conn_lock, __LINE__);
  
  debug("propagateUpdates", "entered propagate updates");

  /*
   * first call load_or_update to get the modifications done
   * directly to the database since we last loaded entries.
   */
  START();
  
  load_or_update_contacts(false);
  
  load_or_update_sub(false);
  
  time(&last_refresh);     /* this is the current refresh time */
  
  { /* contact_table_scope */
    contact_table_entry2 *p_entry, *p_entry1;
    pthread_mutex_lock(&contacts_lock);

    p_entry = (contact_table_entry2*)ht_minimum(contacts_table);
  
    while (p_entry) {
      
      /* if due to some race, an item is both deleted and
       * modified, modified takes preference.
       */
      if (p_entry->modified || p_entry->toinsert) {
	/* First, serialize the entry. */
	/* This isn't split out to a separate function because 
	 * it's a lot cleaner
	 * to keep the "names" and "values" arrays on the stack. */
	const char* names[8];
	const char* values[8];
	int num_columns;
	
	char expires_buf[32];
	char last_modified_buf[32];
	ostrstream q_buf;
	  
	names[0] = "user";
	values[0] = p_entry->contact->user;
	  
	names[1] = "contact";
	values[1] = p_entry->contact->contact;
	  
	num_columns = 2;
	  
	if (p_entry->contact->expires != 0) {
	  cinema_mysql_table::time_to_sql_datetime(
						   p_entry->contact->expires,
						   expires_buf);
	  names[num_columns] = "expires";
	  values[num_columns] = expires_buf;
	  num_columns++;
	}
	  
	if (p_entry->contact->q >= 0.) {
	  q_buf.precision(2);
	  q_buf << p_entry->contact->q << ends;
	  names[num_columns] = "!q";
	  values[num_columns] = q_buf.str();
	  num_columns++;
	}
	  
	if (p_entry->contact->action != contact_sip_action_NULL) {
	  names[num_columns] = "action";
	  switch (p_entry->contact->action) {
	  case contact_sip_action_Proxy:
	    values[num_columns] = "Proxy";
	    break;
	      
	  case contact_sip_action_Redirect:
	    values[num_columns] = "Redirect";
	    break;
	      
	  default:
	    values[num_columns] = "NULL";
	    break;
	  }
	  num_columns++;
	}
	  
	{
	  struct tm tm;
	    
	  localtime_r(&last_refresh, &tm);
	    
	  strftime(last_modified_buf,sizeof(last_modified_buf),
		   "%Y%m%d%H%M%S",&tm);	    
	    
	  names[num_columns] = "last_modified";
	  values[num_columns] = last_modified_buf;
	  num_columns++;
	    
	}
	  
	if (p_entry->contact->display_name != NULL) {
	  names[num_columns] = "display_name";
	  values[num_columns] = p_entry->contact->display_name;
	  num_columns++;
	}
	  
	if (p_entry->contact->sip_methods != NULL) {
	  names[num_columns] = "sip_methods";
	  values[num_columns] = p_entry->contact->sip_methods;
	  num_columns++;
	}
	  
	pthread_mutex_unlock(&contacts_lock);
	  
	succ = conn->replaceRow("contacts", names, values, num_columns);
	  
	q_buf.rdbuf()->freeze(0); /* Done with the buffer */
	  
	if (succ != 0) {
	  ostrstream s;
	    
	  s << "Error writing to MySQL contacts database: "
	    << conn->getLastError() << ends;
	  
	  throw cinema_internal_err(s);
	  
	}
	  
	++total;

	pthread_mutex_lock(&contacts_lock);
	  
	p_entry->modified = p_entry->toinsert = 0;

	p_entry->deleted = 0; /* make sure it is not entered */
      }

      if (p_entry->deleted) {
	/*
	 * While we do database queries, we dont want others to get
	 * blocked. note also that conn_lock provides one level above us.
	 */
	{
	  const char *names[] = {"user", "contact"};
	  const char* values[2];
	  
	  values[0] = p_entry->contact->user;
	  values[1] = p_entry->contact->contact;
	  
	  pthread_mutex_unlock(&contacts_lock);

	  succ = conn->deleteRow("contacts", names, values, 2);
	    
	}
	  
	if (succ != 0) {
	  ostrstream s;
	    
	  s << "Error writing to MySQL contacts database: "
	    << conn->getLastError() << ends;
	  
	  throw cinema_internal_err(s);
	}
	  
	++total;
	  
	pthread_mutex_lock(&contacts_lock);
	  
	p_entry1 = (contact_table_entry2*)
	  ht_successor(contacts_table, p_entry);
	  
	ht_delete(contacts_table, p_entry);
	  
	free_contact_table_entry(p_entry->contact);
	  
	p_entry->contact = NULL;

	free_valid(p_entry);
	  
	p_entry = p_entry1;
	
	continue; /* dont find successor, as already done */
      }
      
      p_entry = (contact_table_entry2*)
	ht_successor(contacts_table, p_entry);
      
    } /* while */
  
    pthread_mutex_unlock(&contacts_lock);
    
  } /* contact_table_scope */
  
  { /* subscription_table_scope */
    sub_table_entry2 *p_entry, *p_entry1;
    pthread_mutex_lock(&sub_lock);

    p_entry = (sub_table_entry2*)ht_minimum(sub_table);
  
    while (p_entry) {
      
      /* if due to some race, an item is both deleted and
       * modified, modified takes preference.
       */
      if (p_entry->modified || p_entry->toinsert) {
	/* First, serialize the entry. */
	/* This isn't split out to a separate function because 
	 * it's a lot cleaner
	 * to keep the "names" and "values" arrays on the stack. */
	const char* names[7];
	const char* values[7];
	int num_columns;
	  
	char expires_buf[32];
	char last_modified_buf[32];
	 
	names[0] = "sub_to";
	values[0] = p_entry->sub->sub_to;
	  
	names[1] = "sub_from";
	values[1] = p_entry->sub->sub_from;
	  
	num_columns = 2;
	  
	if (p_entry->sub->expires != 0) {
	  cinema_mysql_table::time_to_sql_datetime(
						   p_entry->sub->expires,
						   expires_buf);
	  names[num_columns] = "expires";
	  values[num_columns] = expires_buf;
	  num_columns++;
	}
	  
	{
	  struct tm tm;

	  localtime_r(&last_refresh, &tm);

	  strftime(last_modified_buf,sizeof(last_modified_buf),
		   "%Y%m%d%H%M%S",&tm);	    
	    
	  names[num_columns] = "last_modified";
	  values[num_columns] = last_modified_buf;
	  num_columns++;
	    
	}

	if (p_entry->sub->contact != NULL) {
	  names[num_columns] = "contact";
	  values[num_columns] = p_entry->sub->contact;
	  num_columns++;
	}
	  
	if (p_entry->sub->record_route != NULL) {
	  names[num_columns] = "record_route";
	  values[num_columns] = p_entry->sub->record_route;
	  num_columns++;
	}

	if (p_entry->sub->call_id != NULL) {
	  names[num_columns] = "call_id";
	  values[num_columns] = p_entry->sub->call_id;
	  num_columns++;
	}

	pthread_mutex_unlock(&sub_lock);
	  
	succ = conn->replaceRow("subscription", names, values, num_columns);

	if (succ != 0) {
	  ostrstream s;
	    
	  s << "Error writing to MySQL subscription database: "
	    << conn->getLastError() << ends;
	  throw cinema_internal_err(s);
	}
		  
	++total;
	  
	pthread_mutex_lock(&sub_lock);
	  
	p_entry->modified = p_entry->toinsert = 0;

	p_entry->deleted = 0; /* make sure it is not entered */
      }

      if (p_entry->deleted) {
	/*
	 * While we do database queries, we dont want others to get
	 * blocked. note also that conn_lock provides one level above us.
	 */
	{
	  const char *names[] = {"sub_to", "sub_from"};
	  const char* values[2];
	  
	  values[0] = p_entry->sub->sub_to;
	  values[1] = p_entry->sub->sub_from;
	  
	  pthread_mutex_unlock(&sub_lock);

	  succ = conn->deleteRow("subscription", names, values, 2);
	    
	}
	
	if (succ != 0) {
	  ostrstream s;
	    
	  s << "Error writing to MySQL contacts database: "
	    << conn->getLastError() << ends;
	  
	  throw cinema_internal_err(s);
	}
	  
	++total;
	  
	pthread_mutex_lock(&sub_lock);
	  
	p_entry1 = (sub_table_entry2*) ht_successor(sub_table, p_entry);
	  
	ht_delete(sub_table, p_entry);
	  
	free_sub_table_entry(p_entry->sub);
	  
	p_entry->sub = NULL;

	free_valid(p_entry);
	  
	p_entry = p_entry1;
	
	continue; /* dont find successor, as already done */
      }
      
      p_entry = (sub_table_entry2*) ht_successor(sub_table, p_entry);
      
    } /* while */
    
    pthread_mutex_unlock(&sub_lock);
    
  } /* sub_table_scope */

  END("propagateUpdates");
  
  debug("propagateUpdates",
	"a total of %d updates were propagated...\n",total);
  
  return;
  
} /* propagateUpdates */


/*
 * get_aliases_by_alias
 *
 */
alias_table_entry* 
cinema_fastsql_cache::get_aliases_by_alias(const char* alias)
{
  pthread_mymutex_stack_obj lock(&alias_lock, __LINE__);
  
  alias_table_entry* a, *a1;
  
  /*
   * Search the hash table for alias entries in a thread safe manner.
   * If the entry is found, return a copy.
   */
  a = (alias_table_entry*)ht_search(alias_table, (dict_key)alias);
  
  if (a) {
    a1 = (alias_table_entry*)calloc(1,sizeof(alias_table_entry));
      
    if (!a1) throw bad_alloc();
    
    copy_alias(a1,a);
    
    return a1;
  }
  
  {
  ostrstream s;
  s << "aliases : ";
  s << "alias" << " " << alias;
  s << ": not found" << ends;
  
  throw cinema_notfound(s);
  }
} /* get_aliases_by_alias */


/*
 * get_put_by_user
 */
put_table_entry* 
cinema_fastsql_cache::get_put_by_user(const char* user)
{
  pthread_mymutex_stack_obj lock(&put_lock, __LINE__);
  
  put_table_entry *p, *p1;
  
  /*
   * Search the hash table for put entries in a thread safe manner.
   * If the entry is found, return a copy.
   */
  p = (put_table_entry*)ht_search(put_table, (dict_key)user);
  
  if (p) {
    p1 = (put_table_entry*)calloc(1,sizeof(put_table_entry));
    
    if (!p1) throw bad_alloc();
      
    copy_put(p1,p);
    
    return p1;
  }

  {  
  ostrstream s;
  s << "put : ";
  s << "user" << " " << user;
  s << ": not found" << ends;
  throw cinema_notfound(s);
  }
} /* get_put_by_user */

/*
 * get_sub_by_fromto
 *
 * Note: i am getting dumb day by day. instead of using from and to
 * as hash keys, i am using to as the key, and then am iterating
 * on from. Todo: Fix. 
 */
sub_table_entry* 
cinema_fastsql_cache::get_sub_by_fromto(const char* from, const char *to)
{
  pthread_mymutex_stack_obj lock(&sub_lock, __LINE__);
  
  sub_table_entry2 *p, *p1;
  sub_table_entry* ret;
  
  /*
   * Search the hash table for sub entries in a thread safe manner.
   * If the entry is found, return a copy.
   */
  if( !(p = (sub_table_entry2*)ht_search(sub_table, (dict_key)to)))
    goto not_found;
  
  while (p) {
    if (strcmp(p->sub->sub_from, from) == 0) {
      
      if(!(ret = (sub_table_entry*)calloc(1,sizeof(sub_table_entry))))
	throw bad_alloc();
      
      copy_sub(ret,p->sub);
      
      return ret;
    }

    p1 = (sub_table_entry2*)ht_successor(sub_table, (dict_obj)p);

    if (p1 && strcmp(p1->sub->sub_to, to) == 0) {
      p = p1;
      continue;
    }

    break;
  }
  
 not_found:
  {
    ostrstream s;
    s << "subscription : ";
    s << "sub_to" << " " << to;
    s << " sub_from" << " " << from;
    s << ": not found" << ends;
    throw cinema_notfound(s);
  }
  
} /* get_sub_by_fromto */


/*
 * delete_sub_by_fromto
 *
 * Note: i am getting dumb day by day. instead of using from and to
 * as hash keys, i am using to as the key, and then am iterating
 * on from. Todo: Fix. 
 */
void
cinema_fastsql_cache::delete_sub_by_fromto(const char* from, const char *to)
{
  pthread_mymutex_stack_obj lock(&sub_lock, __LINE__);
  
  sub_table_entry2 *p, *p1;
  /*
   * Search the hash table for sub entries in a thread safe manner.
   * If the entry is found, set the deleted flag to 1.
   */
  if( !(p = (sub_table_entry2*)ht_search(sub_table, (dict_key)to)))
    return;
  
  while (p) {
    if (strcmp(p->sub->sub_from, from) == 0) {
      p->deleted = 1;
      break;
    }
    
    p1 = (sub_table_entry2*)ht_successor(sub_table, (dict_obj)p);
    
    if (p1 && strcmp(p1->sub->sub_to, to) == 0) {
      p = p1;
      continue;
    }
    
    break;
  }
return;  
} /* delete_sub_by_fromto */


/*
 * cinema_fastsql_cache::num_subscriptions
 * Return the number of subscriptions for this user.
 */
int cinema_fastsql_cache::num_subscriptions(const char *to)
{

  pthread_mymutex_stack_obj lock(&sub_lock, __LINE__);
  
  sub_table_entry2 *p, *p1;
  int count = 0;

  /*
   * Search the hash table for sub entries in a thread safe manner.
   * If the entry is found, iterate and count.
   */
  if(!(p = (sub_table_entry2*)ht_search(sub_table, (dict_key)to)))
    goto not_found;
  
  while (p) {

    if (strcmp(p->sub->sub_to, to) == 0) {
      
      ++count;
      
      p1 = (sub_table_entry2*)ht_successor(sub_table, (dict_obj)p);
      
      if (p1 && strcmp(p1->sub->sub_to, to) == 0) {
        p = p1;
        continue;
      }
      
    }
    break;
  }
  
 not_found:
  return count;
  
} /* cinema_fastsql_cache::num_subscriptions */


sub_table_entry **cinema_fastsql_cache::get_all_subscriptions(const char* to, int *count)
{
  pthread_mymutex_stack_obj lock(&sub_lock, __LINE__);
  int i, cnt = 0;

  vector<sub_table_entry*> v;
  
  vector<sub_table_entry*>::iterator viter;

  sub_table_entry2 *p_test, *p_test1;
  sub_table_entry **base;

  /*
   * Algorithm
   * -for the given sub_to, first locate all contacts from the hashtable.
   * -then allocate a sub_table_entry* [] array, copy all objects into
   * it.
   * -make sure that there is one last pointer in it, which is null.
   *
   * return it. this should be done in a thread safe manner.
   */
  v.reserve(10);

  if (count) *count = 0;  
  
  p_test = (sub_table_entry2*)
    ht_search(sub_table, (dict_key)to);
  
  if (!p_test) {
    base = (sub_table_entry**)calloc(1,sizeof(sub_table_entry)); 
    if (!base) throw bad_alloc();
    return base;
  }
  
  while (1) {
    
    if (!p_test->deleted) {
      v.insert(v.end(),p_test->sub);
      ++cnt;
    }

    p_test1 = (sub_table_entry2*)
      ht_successor(sub_table,(dict_obj)p_test);
    
    if (p_test1 && (compare_subobj_values(p_test1, p_test) == 0)) {
      p_test = p_test1;
      continue;
    }
      
    break;
  }
  
  if(!(base = (sub_table_entry**) calloc(cnt+1, sizeof(sub_table_entry*))))
    throw bad_alloc();
  
  viter = v.begin();
  
  for (i=0;i<cnt;i++) {

      base[i] = (sub_table_entry*)calloc(1,sizeof(sub_table_entry));
      if (!base[i]) throw bad_alloc();
      copy_sub(base[i],*viter);
      viter++;
  }
  
  v.clear();
  
  if (count) *count = cnt;
 
  debug("get_all_subscriptions", "returning %d entries for %s\n",
	*count, to); 
  
  return base;
  
} /* cinema_fastsql_cache::get_all_subscriptions */


/*
 * helper function for creating copy of contact objects.
 */
static void 
copy_contacts(contact_table_entry *to, const contact_table_entry *from)
{
  
  strassign(&to->user, from->user);
  strassign(&to->contact, from->contact);
  to->expires = from->expires;
  to->q = from->q;
  to->action = from->action;
  to->last_modified = from->last_modified;
  strassign(&to->display_name, from->display_name);
  strassign(&to->sip_methods, from->sip_methods);
  return;

} /* copy_contacts */


/*
 * helper function for creating copy of sub_table_entry objects.
 */
static void 
copy_sub(sub_table_entry *to, const sub_table_entry *from)
{
  
  strassign(&to->sub_to, from->sub_to);
  strassign(&to->sub_from, from->sub_from);
  to->expires = from->expires;
  strassign(&to->contact, from->contact);
  strassign(&to->call_id, from->call_id);
  strassign(&to->record_route, from->record_route);
  to->last_modified = from->last_modified;
  return;
  
} /* copy_sub */

/*
 * copy_put
 * helper function for creating copies of put objects.
 */
static void copy_put(put_table_entry *to, const put_table_entry *from)
{
  
  int i,j;
  
  strassign(&to->user,from->user);
  strassign(&to->hash_value, from->hash_value);
  strassign(&to->realm, from->realm);
  to->sip_groups = from->sip_groups;
  to->auth = from->auth;
  strassign(&to->algorithm, from->algorithm);
  to->sip_methods = from->sip_methods;
  
  for (i=0;from->remote_user[i]!=NULL;i++);

  if(i>0) {
    to->remote_user = (char**)calloc(i+1,sizeof(char*));
  
    if (!to->remote_user) throw bad_alloc();

    for (j=0;from->remote_user[j]!=NULL;j++) {
      strassign(&to->remote_user[j],from->remote_user[j]);
    }
  }
  
  strassign(&to->busy, from->busy);
  strassign(&to->noresponse, from->noresponse);
  to->um_timeout = from->um_timeout;
  to->max_msgsize_kb = from->max_msgsize_kb;
  to->last_modified = from->last_modified;
  strassign(&to->gwclass, from->gwclass);
  return;

} /* copy_put */


/*  
 * copy_alias
 *  helper for copying alias objects.
 */
static void copy_alias(alias_table_entry *to, const alias_table_entry *from)
{
  
  strassign(&to->alias,from->alias);
  strassign(&to->primary_user, from->primary_user);
  to->last_modified = from->last_modified;
  return;

} /* copy_alias */


/*
 * get_contacts_by_user
 */
contact_table_entry** 
cinema_fastsql_cache::get_contacts_by_user(const char *user, int *count)
{
  
  int i = 0, cnt = 0;

  pthread_mymutex_stack_obj lock(&contacts_lock, __LINE__);
  contact_table_entry **base;
  contact_table_entry2 *p_test, *p_test1;
  vector<contact_table_entry*> v;
  vector<contact_table_entry*>::iterator viter;

  /*
   * Algorithm
   * -for the given user, first locate all contacts from the hashtable.
   * -then allocate a contact_table_entry* [] array, copy all objects into
   * it.
   * -make sure that there is one last pointer in it, which is null.
   *
   * return it. this should be done in a thread safe manner.
   */
  
  v.reserve(10);
 
  if (count) *count = 0;  
  
  p_test = (contact_table_entry2*)
    ht_search(contacts_table, (dict_key)user);
  
  if (!p_test) {
    base = (contact_table_entry**)calloc(1,sizeof(contact_table_entry)); 
    if (!base) throw bad_alloc();
    return base;
  }
  
  while (1) {
    
      if (!p_test->deleted) {
      	v.insert(v.end(),p_test->contact);
	++cnt;
      }

      p_test1 = (contact_table_entry2*)
	ht_successor(contacts_table,(dict_obj)p_test);
      
      if (p_test1 && (compare_contact2obj_values(p_test1, p_test) == 0)) {
	p_test = p_test1;
	continue;
      }
      
      break;
  }
  
  base = (contact_table_entry**) calloc(cnt+1, sizeof(contact_table_entry*));
  
  if (!base) throw bad_alloc();
  
  viter = v.begin();
  
  for (i=0;i<cnt;i++) {
    
    base[i] = (contact_table_entry*)calloc(1,sizeof(contact_table_entry));
  
    if (!base[i]) throw bad_alloc();
    
    copy_contacts(base[i],*viter);
      
    viter++;
  }
  
  v.clear();
  
  if (count) *count = cnt;
 
  debug("get_contacts_by_user", "returning %d contacts for %s\n",
	*count, user); 

  return base;
  
} /* get_contacts_by_user */


/*
 * insert_contact
 */
void cinema_fastsql_cache::insert_contact(const contact_table_entry *entry)
{
  
  /* Algorithm:
   *  -Try to locate an entry which has the same user and contact.
   *  -If it is not found, then add it, set the insert flag.
   *  -If it is found, update information, set the modified flag.
   *  -This should be done in a thread-safe manner.
   */
  pthread_mymutex_stack_obj lock(&contacts_lock, __LINE__);
  contact_table_entry2 *p_test, *p_test1;
  contact_table_entry *p_new;
  int found = 0;
  
  p_test = (contact_table_entry2*)ht_search(contacts_table, entry->user);
  
  if (!p_test) goto insert_new_entry;
  
  /* there is atleast one user whose user field matches
   * the entry->user. iterate and find contacts.
   */
  while (1) {
      if (strcasecmp(entry->contact, p_test->contact->contact) == 0) {
	found = 1;
	break;
      }
      
      p_test1 = (contact_table_entry2*) ht_successor(contacts_table, p_test);
      
      if (p_test1  && 
	  (strcasecmp(p_test1->contact->user, p_test->contact->user) == 0)) {
	p_test = p_test1;
	continue;
      }
      else
	break;
  }
  
  if (found == 1) {

    p_test->contact->expires = entry->expires;
    p_test->contact->q = entry->q;
    p_test->contact->action = entry->action;
    time(&p_test->contact->last_modified);
      
    if (p_test->contact->sip_methods && entry->sip_methods
	&& strcasecmp(p_test->contact->sip_methods, entry->sip_methods) != 0) {
      
      free_valid(p_test->contact->sip_methods);
      strassign(&p_test->contact->sip_methods, entry->sip_methods);
    }
    
    p_test->modified = 1;
    p_test->deleted = 0;     
    return;
    
  }
  
 insert_new_entry:
  p_new = (contact_table_entry*) calloc(1,sizeof(contact_table_entry));

  if (!p_new) throw bad_alloc();

  copy_contacts(p_new,entry);
  
  p_test = (contact_table_entry2*) calloc(1,sizeof(contact_table_entry2));
  
  if (!p_test) throw bad_alloc();

  p_test->toinsert = 1;
  p_test->contact = p_new;
  
  if (ht_insert(contacts_table, p_test) != 0)
    error("insert_contact","ht_insert fails...");
  
  return;
  
} /* insert_new_contact */

/*
 * insert_sub
 */
void cinema_fastsql_cache::insert_sub(const sub_table_entry *entry)
{
  
  /* Algorithm:
   *  -Try to locate an entry which has the same from and to.
   *  -If it is not found, then add it, set the insert flag.
   *  -If it is found, update information, set the modified flag.
   *  -This should be done in a thread-safe manner.
   */
  pthread_mymutex_stack_obj lock(&sub_lock, __LINE__);
  sub_table_entry2 *p_test, *p_test1;
  sub_table_entry *p_new;
  int found = 0;
  
  p_test = (sub_table_entry2*)ht_search(sub_table, entry->sub_to);
  
  if (!p_test) goto sub_insert_new_entry;
  
  /* there is atleast one sub_to whose user field matches
   * the entry->sub_to. iterate and find sub_from.
   */
  while (1) {
    if (!strcasecmp(entry->sub_from, p_test->sub->sub_from)) {
      found = 1;
      break;
    }
      
    p_test1 = (sub_table_entry2*) ht_successor(sub_table, p_test);
      
    if (p_test1  && 
	(!strcasecmp(p_test1->sub->sub_to, p_test->sub->sub_to))) {
      p_test = p_test1;
      continue;
    }
    else
      break;
  }
  
  if (found == 1) {
    p_test->sub->expires = entry->expires;
    
    time(&p_test->sub->last_modified);
      
    free_valid(p_test->sub->contact);
    free_valid(p_test->sub->call_id);
    free_valid(p_test->sub->record_route);

    strassign(&p_test->sub->contact, entry->contact);
    strassign(&p_test->sub->call_id, entry->call_id);
    strassign(&p_test->sub->record_route, entry->record_route);
    
    p_test->modified = 1;
    p_test->deleted = 0;     
    
    return;
    
  }
  
 sub_insert_new_entry:
  if (!(p_new = (sub_table_entry*) calloc(1,sizeof(sub_table_entry))))
    throw bad_alloc();
  
  copy_sub(p_new,entry);
  
  if (!(p_test = (sub_table_entry2*) calloc(1,sizeof(sub_table_entry2))))
    throw bad_alloc();
  
  p_test->toinsert = 1;
  
  p_test->sub = p_new;
  
  if (ht_insert(sub_table, p_test) != 0)
    error("insert_sub","ht_insert fails...");
  
  return;
  
} /* insert_new_sub */


/*
 * delete_contacts_by_user
 */
void cinema_fastsql_cache::delete_contacts_by_user(const char *user)
{

  contact_table_entry2 *p_test, *p_test1;
  pthread_mymutex_stack_obj lock(&contacts_lock, __LINE__);
  int total = 0;

  /* 
   * Algorithm:
   * -Iterate and delete. (delete here means that we set the deleted flag to 1,
   * and it will be actually deleted in the next pass of cache refresh.
   */
  p_test = (contact_table_entry2*) ht_search(contacts_table,(dict_key)user);

  while (p_test) {
    
    ++total;
    p_test->deleted = 1;

    p_test1 = (contact_table_entry2*) ht_successor(contacts_table, p_test);
    
    
    if (p_test1 && (strcasecmp(p_test1->contact->user, p_test->contact->user) == 0)) {
      p_test = p_test1;
      continue;
    }
    
    break;
  }
  
  debug("delete_contact_by_user",
	"marked %d contacts for late-delete - user %s\n", total, user);
  
  return;
  
} /* delete_contacts_by_user */

/*
 * delete_contact_by_contact
 */
void cinema_fastsql_cache::delete_contact_by_contact(const char *user,
						     const char *contact)
{
  
  contact_table_entry2 *p_test, *p_test1;
  pthread_mymutex_stack_obj lock(&contacts_lock, __LINE__);
  int total = 0;
  
  /* 
   * Algorithm:
   * -Iterate and delete. (delete here means that we set the deleted flag to 1,
   * and it will be actually deleted in the next pass of cache refresh.
   */
  p_test = (contact_table_entry2*) ht_search(contacts_table, (dict_key)user);
  
  while (p_test) {
      if (strcasecmp(p_test->contact->contact, contact) == 0) {
	p_test->deleted = 1;
	++total;
	break;
      }
      
      p_test1 = (contact_table_entry2*) ht_successor(contacts_table, p_test);
      
      if (p_test1 && 
	  (strcasecmp(p_test1->contact->user, p_test->contact->user) == 0)) {
	p_test = p_test1;
	continue;
      }
      
      break;
  }
  
  debug("delete_contact_by_contact",
	"marked %d contacts for late-delete - user %s, contact %s\n",
	total, user, contact);
  
  return;
  
} /* delete_contact_by_contact */

void 
cinema_fastsql_cache::delete_expired_contacts_by_user(const char *user, 
						      time_t expire_time)
{
  
  contact_table_entry2 *p_test, *p_test1;
  pthread_mymutex_stack_obj lock(&contacts_lock, __LINE__);
  int total = 0;
  
  /* 
   * Algorithm:
   * -Iterate and delete. (delete here means that we set the deleted flag to 1,
   * and it will be actually deleted in the next pass of cache refresh.
   */
  
  p_test = (contact_table_entry2*) ht_search(contacts_table,(dict_key)user);
  
  while (p_test) {
    
    if (p_test->contact->expires <= expire_time) {
      p_test->deleted = 1;
      ++total;
    }
    
    p_test1 = (contact_table_entry2*) ht_successor(contacts_table, p_test);
    
    
    if (p_test1 && 
	(strcasecmp(p_test1->contact->user, p_test->contact->user) == 0)) {
      p_test = p_test1;
      continue;
    }
    
    break;
  }
  
  debug("delete_expired_contacts_by_user",
	"marked %d contacts for late-delete - user %s\n",
	total, user);
  
  return;
  
} /* delete_expired_contacts_by_user */


/*
 * Taken from cinema_mysql_table.
 *
 * cinema_fastsql_cache::sanity_check_result(int, bool, unsigned int,
 *                                         const char*, const char*,
 *                                         const char*)
 * After a call to queryTable, we want to make sure that the result is
 * reasonable.  Check it.
 * Parameters:
 *   succ: the return code of cinema_mysql_table::query_table().
 *   primary: whether the query was for a primary key (i.e. we expect exactly
 *            one row in the result)
 *   columns: the number of columns we expect in the result
 *   keyname: the name of the key with which the query was made (used for
 *            strings in exceptions)
 *   keyval: the value of the key with which the query was made (likewise)
 *   table: the name of the table against which the query was made (likewise)
 * Throws a cinema_error if any of the checks fail.
 */
void
cinema_fastsql_cache::sanity_check_result(int succ, bool primary,
					  unsigned int columns,
					  const char* keyname,
					  const char* keyval,
					  const char* table)
{
  ostrstream s;
  
  if (succ != 0) {
    s << "Error reading from MySQL " << table << " database: "
      << conn->getLastError() << ends;
    throw cinema_internal_err(s);
  }
  
  if (primary && conn->getNumRowsReturned() > 1) {
    s << "MySQL Sanity check failed! ";
    s << conn->getNumRowsReturned() << " rows returned in table ";
    s << table << "for primary key \'" << keyname << "\' = " << keyval << ends;
    
    throw cinema_internal_err(s);
  }
  
  if (conn->getColumnCount() != columns) {
    s << "MySQL Sanity check failed!  Asked " << columns << " questions, got ";
    s << conn->getColumnCount() << " answers!" << ends;
    
    throw cinema_internal_err(s);
  }
} /* cinema_fastsql_cache::sanity_check_result */



/*
 * cinema_fastsql_cache::get_key(const uri_t *)
 * Get a database key suitable for the mysql table, based on a URI.
 * For SIP, this is currently just user@host; other URI schemes use
 * the whole URI.  XXX TODO?: make SIP the whole URI too.  Will
 * require updating the database.
 * Use DString rather than std::string so we can use URI_SPrint.
 */
char*
cinema_fastsql_cache::get_key(const uri_t *uri) const
{
  DString userid;

  DStringInit(&userid);

  if (strcasecmp(uri->scheme, "sip") != 0) {
    URI_SPrint(uri, &userid);
  }
  else {
    if (uri->user == NULL && uri->host == NULL) {
      /* Shouldn't happen */
      DStringFree(&userid);
      throw cinema_invalid_arg("URI argument has neither user nor host");
    }
    if (uri->user != NULL) {
      DStringAppend(&userid, uri->user, -1);
      DStringAppend(&userid, "@", -1);
    }
    if (uri->host != NULL) {
      DStringAppend(&userid, uri->host, -1);
    }
  }
  
  return DStringValue(&userid);
} /* get_key */


/*
 * cinema_fastsql_cache::get_uri(const char*)
 * Get a canonical URI which represents the given database key.
 */
uri_t
cinema_fastsql_cache::get_uri(const char *key_cp) const
{
  string key(key_cp);

  if (key.find(':') == string::npos) {
    /* Keys without a scheme are user@host, and indicate sip */
    key = "sip:" + key;
  }

  uri_t uri = URI_Parse((char*)key.c_str());

  return uri;
} /* cinema_fastsql_cache::get_uri(const char*) */

/* 500 users for time being, we will use count later */
struct ht_args hput = { 500,  
			DEFAULT_BUCKET_ENTRIES, 
			(ht_func)compute_putobj_value, 
			(ht_func)compute_key_value
};

/* 500 aliases for time being, we will use count later */
struct ht_args halias = { 500,  
			  DEFAULT_BUCKET_ENTRIES, 
			  (ht_func)compute_aliasobj_value, 
			  (ht_func)compute_key_value
};

/* 2000 contacts for time being, we will use count later */
struct ht_args hcontacts = { 2000, 
			     DEFAULT_BUCKET_ENTRIES, 
			     (ht_func)compute_contact2obj_value, 
			     (ht_func)compute_key_value
};

struct ht_args hsubs = { 500,
			 DEFAULT_BUCKET_ENTRIES,
			 (ht_func)compute_subobj_value,
			 (ht_func)compute_key_value
};

static char query_count_put[] = "select count(*) from put";

/*
 * init_hashtables.
 *
 */
void cinema_fastsql_cache::init_hashtables()  
{
  int succ, put_row;
  MYSQL_ROW row;
  
  succ = conn->rawQueryTable(query_count_put);
  
  if (succ < 0 || conn->getNumRowsReturned()!=1)
    throw cinema_error(string("unable to estimate put table size..."));
  
  row = conn->getNextRow();
  
  if (row[0] != NULL) {
    istrstream r(row[0]);
    r >> put_row;
    debug("init_hashtables","found %d rows for put\n",put_row);
  }
  else
    throw cinema_error(string("unable to estimate put table size..."));

  conn->freeResultSet();

  if(put_row < 30)
    put_row = 30;
  
  hput.ht_table_entries = put_row * 2;
  
  hsubs.ht_table_entries = halias.ht_table_entries = put_row * 2 * 3;
  
  hcontacts.ht_table_entries = put_row * 2 * 5;
  
  put_table = ht_create((dict_function)compare_putobj_values, 
			(dict_function)compare_putkey_values, 
			0, (int*)0, &hput);
  
  
  alias_table = ht_create((dict_function)compare_aliasobj_values, 
			  (dict_function)compare_aliaskey_values, 
			  0, (int*)0, &halias);
  
  
  contacts_table = ht_create((dict_function)compare_contact2obj_values, 
			     (dict_function)compare_contact2key_values, 
			     0, (int*)0, &hcontacts);
  
  sub_table = ht_create((dict_function)compare_subobj_values,
			(dict_function)compare_subkey_values,
			0, (int*)0, &hsubs);

  if (!put_table || !alias_table || !contacts_table || !sub_table) 
    throw cinema_error(string("ht_create failed..."));
  
  return;
}


static ht_val compute_key_value(dict_key o)
{
  
  return compute_hash((const char*)o);
  
}

static ht_val compare_putobj_values(dict_obj o1, dict_obj o2)
{

  return compute_putobj_value(o1) - compute_putobj_value(o2);

}

static ht_val compare_aliasobj_values(dict_obj o1, dict_obj o2)
{

  return compute_aliasobj_value(o1) - compute_aliasobj_value(o2);

}

static ht_val compare_contact2obj_values(dict_obj o1, dict_obj o2)
{
  
  return compute_contact2obj_value(o1) - compute_contact2obj_value(o2);

}


static ht_val compare_subobj_values(dict_obj o1, dict_obj o2)
{
  
  return compute_subobj_value(o1) - compute_subobj_value(o2);

}

static ht_val compute_putobj_value(dict_obj o)
{
  put_table_entry *p_entry;

  p_entry = (put_table_entry*)o;

  if (p_entry && p_entry->user)
    return compute_key_value(p_entry->user);
  else
    {  
      error("compute_putobj_value", "dict_obj or its values null...\n");
      return (ht_val)0;
    }
}


static ht_val compute_subobj_value(dict_obj o)
{
  sub_table_entry2 *p_entry;
  
  p_entry = (sub_table_entry2*)o;

  if (p_entry && p_entry->sub && p_entry->sub->sub_to)
    return compute_key_value(p_entry->sub->sub_to);
  else
    {  
      error("compute_subobj_value", "dict_obj or its values null...\n");
      return (ht_val)0;
    }
}

static ht_val compute_aliasobj_value(dict_obj o)
{
  alias_table_entry *p_entry;

  p_entry = (alias_table_entry*)o;

  if (p_entry && p_entry->alias)
    return compute_key_value(p_entry->alias);
  else
    {  
      error("compute_aliasobj_value", "dict_obj or its values null...\n");
      return (ht_val)0;
    }
}

static ht_val compute_contact2obj_value(dict_obj o)
{
  contact_table_entry2 *p_entry;

  p_entry = (contact_table_entry2*)o;

  if (p_entry && p_entry->contact && p_entry->contact->user)
    return compute_key_value(p_entry->contact->user);
  else
    {
      error("compute_contact2obj_value", "dict_obj or its values null...\n");
      return (ht_val)0;
    }
}

static ht_val compare_putkey_values(dict_key k1, dict_obj o1)
{
  
  ht_val t =  compute_putobj_value(o1);
  
  ht_val t1 = compute_key_value(k1);

  return t1 - t;
}


static ht_val compare_aliaskey_values(dict_key k1, dict_obj o1)
{

	
  ht_val t =  compute_aliasobj_value(o1);
  
  ht_val t1 = compute_key_value(k1);

  return t1 - t;
}

static ht_val compare_contact2key_values(dict_key k1, dict_obj o1)
{

  ht_val t =  compute_contact2obj_value(o1);
  
  ht_val t1 = compute_key_value(k1);

  return t1 - t;
}

static ht_val compare_subkey_values(dict_key k1, dict_obj o1)
{

  ht_val t =  compute_subobj_value(o1);
  
  ht_val t1 = compute_key_value(k1);
  
  return t1 - t;
}


/*
  //testcode
  #include <unistd.h>
  #include "cinema_fastsql_table.h"
  extern "C" 
  {
  void make_changes(void)
  {
  int i;
  cinema_fastsql_table a(string("a"));

  alias_table_entry *a1 = a.get_aliases_by_alias("7042@cs.columbia.edu");
  printf("%s %s\n",a1->alias, a1->primary_user);

  alias_table_entry *b = a.get_aliases_by_alias("xw71@columbia.edu");
  printf("%s %s\n",b->alias, b->primary_user);

  free_alias_table_entry(a1);
  free_alias_table_entry(b);

  put_table_entry *p = a.get_put_by_user("sankaran@cs.columbia.edu");
  printf("%s %s\n",p->realm, p->user);

  put_table_entry *p1 = a.get_put_by_user("foobarfoo@cs.columbia.edu");
  printf("%s %s\n",p1->realm, p1->user);

  free_put_table_entry(p1);
  free_put_table_entry(p);

  contact_table_entry **c = a.get_contacts_by_user("hgs@cs.columbia.edu");
  
  for(i = 0; c[i]!=NULL; i++)
  {
  printf("%s %s\n",c[i]->user,c[i]->contact);
  }


  contact_table_entry **c1 = a.get_contacts_by_user("kns10@cs.columbia.edu");
  
  for(i = 0; c1[i]!=NULL; i++)
  {
  printf("%s %s\n",c1[i]->user,c1[i]->contact);
  }

  contact_table_entry **c2 = a.get_contacts_by_user("hgs1@cs.columbia.edu");
  
  if(c2)
  {
  for(i = 0; c2[i]!=NULL; i++)
  {
  printf("%s %s\n",c2[i]->user,c2[i]->contact);
  }
  }

  free_contact_table_entries(c);
  free_contact_table_entries(c1);
  if(c2)
  free_contact_table_entries(c2);

  
  contact_table_entry c3;
  c3.action = contact_sip_action_Proxy;
  c3.contact = "sip:sankaran@thalys.cs.columbia.edu";
  c3.display_name = "narayanan sankaran";
  time(&c3.expires); 
  c3.expires += 2000;
  c3.q = 0.5;
  c3.user = "sankaran@cs.columbia.edu";
  c3.sip_methods = "any";

  a.insert_contact(&c3);

  cinema_fastsql_cache::refresh_cache();

  a.insert_contact(&c3);

  cinema_fastsql_cache::refresh_cache();
  
  {
  useconds_t u = 10000000U;

  debug("make_changes", "sleeping for 10 seconds...\n");
  usleep(u);
  }
 
  a.delete_contacts_by_user("rvg@cs.columbia.edu");

  {
  useconds_t u = 10000000U;
    
  debug("make_changes", "sleeping for 10 seconds...\n");
  usleep(u);
  }
 
  a.delete_contact_by_contact("sankaran@cs.columbia.edu","sip:sankaran@thalys.cs.columbia.edu");

  a.delete_expired_contacts_by_user("hgs@cs.columbia.edu", c3.expires - 2000);

  
  cinema_fastsql_cache::refresh_cache();

  {
  useconds_t u = 10000000U;

  debug("make_changes", "sleeping for 10 seconds...\n");
  usleep(u);
  }
   
  return;
   
  }
  }

*/


