/* cinema_fastsql_cache.h 
 *   Caching (in-memory) database support provider for CINEMA.
 *
 *   Currently, supports put, aliases, contacts tables of the mysql
 *   database.
 *
 *   Use the url fastsql://user:password@hostname/databasename?REFRESH=seconds
 *   as invocation URI.
 *
 *   Copyright 2001 Columbia University.  All rights reserved. 
 */
#ifndef CINEMA_FASTSQL_CACHE_H_
#define CINEMA_FASTSQL_CACHE_H_

#include <ctime>  
#include <string>

extern "C" {
#include <pthread.h>
#include "sysdep.h"

/* this should be inside "C", otherwise Win32 creates build problems */
#define PROTOTYPES         /* for ht.h, to enable ANSI style decl. */
#include "ht.h"
}

#include "mysqlconn.h"
#include "dbaccess.h"
#include "cinema_table.h"

/*
 * The first part defines wrappers around table definitions.
 * To put them into a hashtable, and maintain them we need a uniform
 * mechanism to keep track of 
 *   1) whether a write happened before a cache refresh, so that we can
 *      do a write-through in the next refresh cycle.
 *   2) whether a particular row was invalidated, so that we need to delete
 *      it.
 *
 */
extern "C" {
  
  /*
   * This wrapper structure is used to keep track of contact entries
   * that change in between cache refresh cycles. This is what goes
   * into our hash table for contacts. ---- Todo: Define an abstract
   * struct called wrapper, and make the pointer as a void ptr, instead
   * of defining so many identical structs.
   */
  typedef struct contact_table_entry2_struct 
  {
    contact_table_entry* contact;  /* the actual contact ds */
    int modified;                  /* was this modified ? */
    int toinsert;                  /* is this new, that should be inserted? */
    int deleted;                   /* should this be deleted? */
  } contact_table_entry2;
  
  
  typedef struct sub_table_entry2_struct 
  {
    sub_table_entry* sub;          /* actual sub ds */
    int modified;                  /* was this modified ? */
    int toinsert;                  /* is this new, that should be inserted? */
    int deleted;                   /* should this be deleted? */
  } sub_table_entry2;  
};


/*
 *
 * Most of the code here is taken from cinema_mysql_table. But
 * this class does not inherit from cinema_table as i want a singleton
 * , which means i cant have public constructors but only class factory.
 * 
 * I also donot want anybody to invoke "delete" on this object, which
 * means i cannot return a public-visible pointer or object reference.
 *
 * Hence, this gets wrapped inside cinema_fastsql_table.
 * that wrapper contains this singleton, and provides
 * the cinema_table interface. It delegates all methods to this class.
 * 
 */

class cinema_fastsql_cache
{
  
 public:
  
  /*
   * Call this function to get a reference to the cinema_mysql_table
   * object. 
   */
  static const cinema_fastsql_cache* get_fastsql_ref(std::string uri);
  
  static const cinema_fastsql_cache* get_fastsql_ref(uri_t& uri);
  
  /*
   * After using the const pointer got from get, call this to free it.
   * 
   */
  static void free_fastsql_ref(const cinema_fastsql_cache*);
  
  /*
   * Call this during exit to cleanup everything.
   */
  static void close_fastsql_module();
  
  /*
   * Forced cache refresh.
   */
  static void refresh_cache();
  
  /*
   * These methods are very similar to cinema_table and cinema_mysql_table.
   * In fact, some of the code is straight copy and paste from there.
   */
  virtual char* get_key(const uri_t *uri) const;
  
  virtual uri_t get_uri(const char *key) const;
  
  virtual alias_table_entry *get_aliases_by_alias(const char* alias);
  
  virtual put_table_entry *get_put_by_user(const char* user);
  
  virtual contact_table_entry** get_contacts_by_user(const char *user,
                                                     int *count = NULL);
  
  virtual void insert_contact(const contact_table_entry *entry);
  
  virtual void delete_contacts_by_user(const char *user);
  
  virtual void delete_contact_by_contact(const char *user,
                                         const char *contact);
  
  virtual void delete_expired_contacts_by_user(const char *user, 
                                               time_t expire_time);

  virtual sub_table_entry *get_sub_by_fromto(const char* from, const char* to);
  
  virtual int num_subscriptions(const char *to);
  
  virtual sub_table_entry **get_all_subscriptions(const char* to, int *count);
  
  virtual void insert_sub(const sub_table_entry *entry);  
  
  virtual void delete_sub_by_fromto(const char* from, const char *to);
    
  TABLE_SCHEME("fastsql");
  
 protected:
  MysqlConnection* conn;
  
  pthread_mutex_t conn_lock;                       /* database lock */
  
  /* our hashtables at some stage, i may go for dict_h[]
   * if number of tables is more
   */
  dict_h put_table, alias_table, contacts_table, sub_table;
 
   
  pthread_mutex_t put_lock, alias_lock, contacts_lock, sub_lock; 
  /* locks for each */
  
  /*
   * see cinema_mysql_table for what each does.
   */
  void sanity_check_result(int, bool, unsigned int, const char*, const char*,
                           const char*);
  
  alias_table_entry *deserialize_next_alias_row();
  
  put_table_entry *deserialize_next_put_row();
  
  contact_table_entry* deserialize_next_contact_row();
  
  sub_table_entry* deserialize_next_sub_row();  
  
 private:
  uri_t table_uri;
  
  time_t last_refresh;     /* for reloads */
  
  int refresh_period;   /* time period between put/alias/contacts 
			 * table refreshes in seconds
			 */
  
  void init_connection();
  
  void init_cache();
  
  void init_hashtables();
  
  /*
   * Call this function to initialize this module.
   * Apart from initial loading of data, it will also start a thread
   * that takes care of periodic refresh data. 
   */
  static void init_fastsql_module(std::string uri);
  
  static void init_fastsql_module(uri_t &uri);
  
  void load_or_update_put(bool firsttime);
  
  void load_or_update_aliases(bool firsttime);
  
  void load_or_update_contacts(bool firsttime);
  
  void load_or_update_sub(bool firsttime);
  
 protected:
  /*
   * The constructor and destructor for the singleton instance.
   *
   * refresh_period is in seconds.
   */
  cinema_fastsql_cache(std::string uri, int refresh_period);
  
  cinema_fastsql_cache(uri_t& uri, int refresh_period);
  
  virtual ~cinema_fastsql_cache();  
  
 public:
  /* making these two public because the cache refresh thread
   * needs to access them. this is a bad design, and will fix
   * this by an adapter later.
   */
  
  /*
   * Calling this method, makes this object to refresh its internal
   * data (for put and alias, see the database and put in).
   */
  virtual void refresh();
  
  /*
   * Calling this method, will make the object to flush dirty data back
   * to the database.
   */
  virtual void propagateUpdates();
  
};

/*
 * A helper class that takes care of locking when a function is
 * entered, and unlocking when it is left.
 */
class pthread_mymutex_stack_obj {
 
 private:
  pthread_mutex_t* the_lock;
  int lineno;

 public:
  pthread_mymutex_stack_obj(pthread_mutex_t *, int lineno);
  ~pthread_mymutex_stack_obj();

};

inline
pthread_mymutex_stack_obj::pthread_mymutex_stack_obj(
						     pthread_mutex_t* lock, 
						     int lineno)
  : the_lock(lock),lineno(lineno) 
{
 
  /* debug("mymutex_lock","locking %d\n",lineno);
   */
  
  pthread_mutex_lock(the_lock);

} /* pthread_mymutex_stack_obj(pthread_mutex_t* lock , int lineno) */

inline pthread_mymutex_stack_obj::~pthread_mymutex_stack_obj() 
{
  
  /* debug("mymutex_unlock", "unlocking %d\n", lineno);
   */
  
  pthread_mutex_unlock(the_lock);
  
} /* ~pthread_mymutex_stack_obj() */



#endif /* CINEMA_FASTSQL_CACHE_H_ */
