/* -*- C++ -*- */
/* cinema_file_table.cpp -- A file database implementation, intended as a
 * file based information retrieval. */
   
/* Copyright 2001 Columbia University.  All rights reserved. */

#ifdef __linux__
#define _GNU_SOURCE /* Needed for strptime declaration from <time.h> */
#endif

#include <strstream>
#include <string>
#include <vector>

#include <cstdlib>
#include <ctime>

extern "C" {

#include <stdio.h>

#include "sysdep.h"
#include "dstring.h"
#include "uri.h"
#include "strdupn.h"

}

#include "cinema_dummy_table.h"
#include "cinema_file_table.h"

using namespace std;

/*
 * cinema_file_table::cinema_file_table(string)
 * Construct a cinema file table, based on the given URI given as a string.
 */
cinema_file_table::cinema_file_table(std::string uri)
{
  table_uri = URI_Parse(uri.c_str());
  if (table_uri.scheme != get_table_scheme()) {
    URI_Free(&table_uri);
    throw cinema_invalid_arg("Invalid table scheme for dummy table");
  }
} /* cinema_file_table::cinema_file_table(string) */


/*
 * cinema_file_table::cinema_file_table(uri_t&)
 * Construct a cinema file table, based on the given URI given as a uri_t.
 */
cinema_file_table::cinema_file_table(uri_t& uri)
{
  if (uri.scheme != get_table_scheme()) {
    throw cinema_invalid_arg("Invalid table scheme for dummy table");
  }

  table_uri = URI_Copy(uri);
} /* cinema_file_table::cinema_file_table(uri_t) */


/*
 * cinema_file_table::~cinema_file_table()
 * Delete a cinema file table, freeing its URI.
 */
cinema_file_table::~cinema_file_table()
{
  URI_Free(&table_uri);
} /* cinema_file_table::~cinema_file_table() */


/*
 * cinema_file_table::get_key(const uri_t *)
 * Get a database key suitable for the file table, based on a URI.
 */
char*
cinema_file_table::get_key(const uri_t *uri) const
{
  DString userid;

  DStringInit(&userid);

  if (strcasecmp(uri->scheme, "sip") != 0) {
    URI_SPrint(uri, &userid);
  }
  else {
    if (uri->user == NULL && uri->host == NULL) {
      /* Shouldn't happen */
      DStringFree(&userid);
      throw cinema_invalid_arg("URI argument has neither user nor host");
    }
    if (uri->user != NULL) {
      DStringAppend(&userid, uri->user, -1);
      DStringAppend(&userid, "@", -1);
    }
    if (uri->host != NULL) {
      DStringAppend(&userid, uri->host, -1);
    }
  }
  
  return DStringValue(&userid);
} /* cinema_file_table::cinema_file_table(const uri_t *) */


/*
 * cinema_file_table::get_uri(const char*)
 * Get a canonical URI which represents the given database key.
 */
uri_t
cinema_file_table::get_uri(const char *key_cp) const
{
  string key(key_cp);

  if (key.find(':') == string::npos) {
    /* Keys without a scheme are user@host, and indicate sip */
    key = "sip:" + key;
  }

  uri_t uri = URI_Parse(key.c_str());

  return uri;
} /* cinema_file_table::get_uri(const char*) */


/*
 * cinema_file_table::get_aliases_by_alias(const char *)
 * Get an alias entry for the given database key.
 */
alias_table_entry*
cinema_file_table::get_aliases_by_alias(const char *key)
{
  throw cinema_notfound("No such entry");
} /* cinema_file_table::get_aliases_by_alias(const char *) */


/*
 * cinema_mysql_table::parse_sip_groups(const char *)
 * Parse a sip_groups string into a bitwise 'or' of put_sip_group values.
 */
int
parse_sip_groups(const char* orig_groups)
{
  int ret = 0;

  vector<string> group_list = split_string(orig_groups, " ");
  vector<string>::iterator this_group;

  for (this_group = group_list.begin();
       this_group != group_list.end();
       this_group++) {
    if (strcasecmp(this_group->c_str(), "admin") == 0) {
      ret |= put_sip_group_admin;
    }
    else if (strcasecmp(this_group->c_str(), "cgi") == 0) {
      ret |= put_sip_group_cgi;
    }
    else if (strcasecmp(this_group->c_str(), "voicemail") == 0) {
      ret |= put_sip_group_voicemail;
    }
  }

  return ret;
} /* parse_sip_groups(const char *) */


/*
 * parse_sip_methods(const char *)
 * Parse a sip_methods string into a bitwise 'or' of put_sip_method values.
 */
int
parse_sip_methods(const char* orig_methods)
{
  int ret = 0;

  vector<string> group_list = split_string(orig_methods, " ");
  vector<string>::iterator this_group;

  for (this_group = group_list.begin();
       this_group != group_list.end();
       this_group++) {
    if (strcasecmp(this_group->c_str(), "INVITE") == 0) {
      ret |= put_sip_method_INVITE;
    }
    else if (strcasecmp(this_group->c_str(), "REGISTER") == 0) {
      ret |= put_sip_method_REGISTER;
    }
    else if (strcasecmp(this_group->c_str(), "any") == 0) {
      ret |= put_sip_method_any;
    }
  }

  return ret;
} /* parse_sip_methods(const char *) */

/*
 * parse_auth(const char *)
 * Parse an auth string into a put_auth enum.
 */
put_auth
parse_auth(const char* auth, put_auth def)
{
  if (auth == NULL) {
    /* Fail closed */
    return def;
  }
  else if (strcasecmp(auth, "required") == 0) {
    return put_auth_required;
  }
  else if (strcasecmp(auth, "request") == 0) {
    return put_auth_request;
  }
  else if (strcasecmp(auth, "never") == 0) {
    return put_auth_never;
  }
  else {
    /* Fail closed. */
    return def;
  }
} /* parse_auth(const char *) */


/*
 * Deserialize the row containing PUT information
 * into a structure of type put_table_entry.
 */
put_table_entry *
cinema_file_table::deserialize_put_row(const char *row)
{
  int elem_index;

  put_table_entry* result = (put_table_entry *)
    calloc(1, sizeof(put_table_entry));

  vector<string> elem = split_string(row, ":");
  vector<string>::iterator this_elem;

  result->auth = put_auth_never;
  result->algorithm = strdup("MD5");

  for (elem_index = 0, this_elem = elem.begin();
       this_elem != elem.end();
       this_elem ++, elem_index++) {
    switch(elem_index) {
    case 0 : /* user */
      strassign(&result->user, this_elem->c_str());
      break;
    case 1 : /* realm */
      strassign(&result->realm, this_elem->c_str());
      break;
    case 2 : /* hash_value */
      strassign(&result->hash_value, this_elem->c_str());
      break;
    case 3 : /* sip_groups */
      result->sip_groups = parse_sip_groups(this_elem->c_str());
      break;
    case 4 : /* auth */
      result->auth = parse_auth(this_elem->c_str(),
                                put_auth_never);
      break;
    case 5 : /* algorithm */
      strassign(&result->algorithm, this_elem->c_str());
      break;
    case 6 : /* sip_methods */
      result->sip_methods = parse_sip_methods(this_elem->c_str());
      break;
    case 7 : /* remote_user */
      result->remote_user = strvec_to_c_strs(
                  split_string(this_elem->c_str(), " "));
      break;
    case 8 : /* busy */
      strassign(&result->busy, this_elem->c_str());
      break;
    case 9 : /* noresponse */
      strassign(&result->noresponse, this_elem->c_str());
      break;
    case 10 : /* message_template */
      strassign(&result->message_template, this_elem->c_str());
      break;
    case 11 : /* um_timeout */
      {
        istrstream r11(this_elem->c_str());
        r11 >> result->um_timeout;
      }
      break;
    case 12 : /* masx_msgsize_kb */
      {
        istrstream r12(this_elem->c_str());
        r12 >> result->max_msgsize_kb;
      }
      break;
    case 13 : /* last_modified */
      {
        istrstream r13(this_elem->c_str());
        r13 >> result->last_modified;
      }
      break;
    case 14 : /* gwclass */
      strassign(&result->gwclass, this_elem->c_str());
      break;
    }
  }

  if ( result->sip_methods == 0 ) {
    result->sip_methods = 
      put_sip_method_any | put_sip_method_INVITE |
      put_sip_method_REGISTER;
  }

  return result;
}


/*
 * cinema_file_table::get_put_by_user(const char *)
 * Get a PUT entry for the given database key.
 */
put_table_entry*
cinema_file_table::get_put_by_user(const char *key)
{
  FILE *fp;
  char buffer[1024];
  int n;

  /* Since no modification hence mutex lock is
   * not needed. */

  /* Open the database file */
  if ( key == NULL || *key == '\0' ) {
    throw cinema_notfound("No such empty user");
  }
  if ( table_uri.path == NULL ||
       *table_uri.path == '\0' ||
       (fp = fopen(table_uri.path, "r")) == NULL ) {
    throw cinema_notfound("No such file");
  }

  n = strlen(key);
  while (fgets(buffer, sizeof(buffer), fp)) {
    if ( !strncmp(key, buffer, n) ) {
      /* Match is found */
      put_table_entry *result = deserialize_put_row(buffer);
      return result;
    }
  }
  return NULL;

} /* cinema_file_table::get_put_by_user(const char *) */


sub_table_entry *cinema_file_table::get_sub_by_fromto(const char* from, const char* to)
{
  return NULL;
}

void cinema_file_table::insert_sub(const sub_table_entry *entry)
{
  throw cinema_error("Not implemented...");
}

int cinema_file_table::num_subscriptions(const char *to)
{
  /* not implemented */
  return 0;
}

sub_table_entry **cinema_file_table::get_all_subscriptions(const char* to, int *count)
{
  
  if (count)
    *count = 0;
  
  return NULL;
}


void cinema_file_table::delete_sub_by_fromto(const char* from, const char *to)
{
  /* Not Implemented */
  return;
}
