/* -*- C++ -*- */
/* cinema_table.h -- Virtual base class representing cinema database
 * tables. */
   
/* Copyright 2001 Columbia University.  All rights reserved. */

#ifndef CINEMA_TABLE_H_
#define CINEMA_TABLE_H_

#include <stdexcept>
#include <string>
#include <strstream>
#include <vector>

#include <ctime>

#include "dbaccess.h"

extern "C" {
#include "uri.h"
}

/* Macro which sets up table scheme names.  Call with literal string,
 * usually. (Can be anything which can be converted to std::string.) */
/* get_table_scheme() is virtual so you can get the scheme for a specific
 * table. table_scheme() is static so you can get the scheme for a class. */
#define TABLE_SCHEME(s) \
virtual const std::string get_table_scheme() const { return s; } \
static const std::string table_scheme() { return s; }


/* Pure-virtual base class for all database tables */
class cinema_table
{
public:
  virtual char* get_key(const uri_t *uri) const = 0;
  virtual uri_t get_uri(const char *key) const = 0;

  virtual ~cinema_table() { }

  virtual alias_table_entry *get_aliases_by_alias(const char* key) = 0;
  

  virtual put_table_entry *get_put_by_user(const char* key) = 0;
  

  virtual contact_table_entry** get_contacts_by_user(const char *user,
                                                     int *count = NULL) = 0;

  virtual void insert_contact(const contact_table_entry *entry) = 0;

  virtual void delete_contacts_by_user(const char *user) = 0;

  virtual void delete_contact_by_contact(const char *user,
                                         const char *contact) = 0;

  virtual void delete_expired_contacts_by_user(const char *user, 
                                               time_t expire_time) = 0;
  
  virtual sub_table_entry *get_sub_by_fromto(const char* from, const char* to)
    = 0;
  
  virtual int num_subscriptions(const char *to) = 0;
  
  virtual sub_table_entry **get_all_subscriptions(const char* to, int *count)
    = 0;
  
  virtual void insert_sub(const sub_table_entry *entry) = 0;

  virtual void delete_sub_by_fromto(const char* from, const char *to) 
    = 0;

  static cinema_table* create(std::string table_uri);
  
  TABLE_SCHEME("base--do-not-instantiate")
};


/* Cinema database exceptions */
/* There should be one exception for every db_err value in dbaccess.h. */

/* Base class of cinema database exceptions */
class cinema_error : public std::runtime_error {
public:
  cinema_error (const std::string& what_arg): runtime_error(what_arg) { }
  cinema_error (std::ostrstream& what_argstr):
    runtime_error(what_argstr.str())
  { what_argstr.rdbuf()->freeze(0); }
  virtual db_err get_errcode() const { return DB_UNKNOWN; }
};

/* The specified entry was not found in the database */
class cinema_notfound : public cinema_error {
public:
  cinema_notfound (const std::string& what_arg): cinema_error(what_arg) { }
  cinema_notfound (std::ostrstream& what_argstr):
    cinema_error(what_argstr) { }
  virtual db_err get_errcode() const { return DB_NOTFOUND; }
};

/* The function was passed an invalid argument */
class cinema_invalid_arg : public cinema_error {
public:
  cinema_invalid_arg (const std::string& what_arg): cinema_error(what_arg) { }
  cinema_invalid_arg (std::ostrstream& what_argstr):
    cinema_error(what_argstr) { }
  virtual db_err get_errcode() const { return DB_INVALIDARG; }
};

/* The database had an internal error */
class cinema_internal_err : public cinema_error {
public:
  cinema_internal_err (const std::string& what_arg): cinema_error(what_arg) { }
  cinema_internal_err (std::ostrstream& what_argstr):
    cinema_error(what_argstr) { }
  virtual db_err get_errcode() const { return DB_ERR; }
};


/*
 * Common utility functions for mysql, fastsql and file tables.
 */
extern std::vector<std::string>
split_string(const char* c_string, const char *separators);

extern void free_cstrs(char **array);

extern char **
strvec_to_c_strs(const std::vector<std::string> vec);

#endif /* CINEMA_TABLE_H_ */
