#include <stdlib.h>

#ifdef USE_MYSQL

#include <stdexcept>
#include <strstream>
using namespace std;

#include "sysdep.h"
#include "mysqlconn.h"
#include "dbapi.h"

extern "C" {
#include "error.h"
#include "dstring.h"
#include "strdupn.h"
#include "pthread.h"
}

/* use to turn on debugging of SQL statements */
#define DEBUG_DB

pthread_mutex_t D_lock = PTHREAD_MUTEX_INITIALIZER;
#define LOCK()   pthread_mutex_lock(&D_lock)
#define UNLOCK() pthread_mutex_unlock(&D_lock)

MysqlConnection* MysqlConnection::mysqlConnection = NULL;

MysqlConnection::MysqlConnection(const char * user, const char * passwd, 
				 const char * database, const char * host, 
				 int port)
{
  mysql_init(&mysql);

  res = NULL;

  if (!(mysql_real_connect(&mysql, host, user, passwd, database, port, NULL,0))){
    /* Close the database */
    mysql_close(&mysql);

    ostrstream s;
    s << "Error connecting to mysql database";
    if (database != NULL) {
      s << " " << database;
    }
    if (host != NULL) {
      s << " on host " << host;
    }
    if (port != 0) {
      s << " at port " << port;
    }
    if (user != NULL) {
      s << " as user " << user;
    }
    s << ": " << mysql_error(&mysql) << ends;
    
    runtime_error e(s.str());
    s.rdbuf()->freeze(0);
    throw e;
  }

  debug("MysqlConnection", "Connected to mysql on host:%s as user:%s\n",
	host, user);

  res = NULL;
}

MysqlConnection::~MysqlConnection()
{
  /* free the memory for the result set */
  freeResultSet();

  mysql_close(&mysql);

  debug("~MysqlConnection", "Connection to mysql database closed\n");
}

int
MysqlConnection::queryTable(const char * table, const char * key, 
                            const char * value, const char * columns,
                            const char *secondkey, const char *secondvalue)
{
  int retval = 0;
  DString query;
  DStringInit(&query);

  /* create query string */
  DStringAppend(&query, "select ", 7);
  DStringAppend(&query, (const char*) columns, strlen(columns));
  DStringAppend(&query, " from ", 6);
  DStringAppend(&query, (const char*) table,strlen(table));
  if ( key != NULL && *key != 0 ) {
    DStringAppend(&query, " where ",7);
    DStringAppend(&query, (const char*) key, strlen(key));
    DStringAppend(&query, " = '", 4);
    DStringAppend(&query, (const char*) value,strlen(value));
    DStringAppend(&query, "'", 1);

    if ( secondkey != NULL ) {
      DStringAppend(&query, " and ", -1);
      DStringAppend(&query, (const char*) secondkey, -1);
      DStringAppend(&query, " = '", 4);
      DStringAppend(&query, (const char*) secondvalue, -1);
      DStringAppend(&query, "'", 1);
    }
  }
  
#ifdef DEBUG_DB  
  debug("queryTable", "%s\n", DStringValue(&query));
#endif

  if ((mysql_query(&mysql, (const char*) DStringValue(&query))) != 0) {
    error("MysqlConnection::queryTable",
	  "Failed on selected from %s table: Error: %s\n",
	  table, mysql_error(&mysql));
    DStringFree(&query);
    return -1;
  }

  res = mysql_store_result(&mysql);
  DStringFree(&query);
  return retval;
} /* queryTable */

/* 
 * public function to query(SELECT from) a table.
 * Returns 0 if successful, -1 if not. 
 */
int 
MysqlConnection::rawQueryTable(const char * sqlQuery) {

  int retval = 0;
  DString query;
  DStringInit(&query);

  if(!sqlQuery) return -1;

  /* create query string */
  DStringAppend(&query, sqlQuery, strlen(sqlQuery));
  
#ifdef DEBUG_DB  
  debug("rawQueryTable", "%s\n", DStringValue(&query));
#endif

  if ((mysql_query(&mysql, (const char*) DStringValue(&query))) != 0) {
    error("MysqlConnection::rawQueryTable",
	  "Query %s failed: Error: %s\n",
	  sqlQuery, mysql_error(&mysql));
    DStringFree(&query);
    return -1;
  }

  res = mysql_store_result(&mysql);
  DStringFree(&query);
  return retval;

} /* rawQueryTable */

int 
MysqlConnection::deleteRow(const char * table, const char * key, 
			   const char * value)
{
  int retval = 0;
  DString del;
  DStringInit(&del);

  /* create delete string */
  DStringAppend(&del, "delete from ", 12);
  DStringAppend(&del, (const char*) table, strlen(table));
  DStringAppend(&del, " where ", 7);
  DStringAppend(&del, (const char*) key, strlen(key));
  DStringAppend(&del, " = '", 4);
  DStringAppend(&del, (const char*) value, strlen(value));
  DStringAppend(&del, "'", 1);

#ifdef DEBUG_DB  
  debug("deleteRow", "%s\n", DStringValue(&del));
#endif

  if ((mysql_query(&mysql, (const char*) DStringValue(&del))) != 0) {
    error("deleteRow", "Failed on delete from %s table: Error: %s\n",
	  table, mysql_error(&mysql));
    retval = -1;
  }

  DStringFree(&del);
  return retval;
} /* deleteRow */

int 
MysqlConnection::deleteRow(const char * table, const char ** columns, 
                           const char ** values, int num_columns)
{
  int retval = 0;
  int i;
  DString del;
  const char * test; /* used to determine float or int column values */
  DStringInit(&del);

  DStringAppend(&del, "delete from ", 12);
  DStringAppend(&del, (const char*) table, strlen(table));
  DStringAppend(&del, " where ", 7);  

  for (i = 0; i < num_columns; i++) {
    test = columns[i];
    if (*test == '#') { /* int value for column */
      int int_value;
      DString columnValue;
      DStringInit(&columnValue);

      DStringAppend(&del, (test+1), strlen(test)-1);

      DStringAppend(&del, "=", 1);
      int_value = atoi(values[i]);
      DStringPrintf(&columnValue, "%d", int_value);
      DStringAppendD(&del, &columnValue);
      DStringFree(&columnValue);
    }      
    else if (*test == '!') { /* float value for column */
      float float_value;
      DString columnValue;
      DStringInit(&columnValue);

      DStringAppend(&del, (test+1), strlen(test)-1);

      DStringAppend(&del, "=", 1);
      float_value = (float) atof(values[i]);
      DStringPrintf(&columnValue, "%f", float_value);
      DStringAppendD(&del, &columnValue);
      DStringFree(&columnValue);

    }
    else { /* string value for column */
      DStringAppend(&del, (const char*) columns[i], strlen(columns[i]));
      DStringAppend(&del, "=", 1);
      DStringAppend(&del, "'", 1);
      DStringAppend(&del, (const char*) values[i], strlen(values[i]));
      DStringAppend(&del, "'", 1);
    }

    if (i < (num_columns-1))
      DStringAppend(&del, " and ", 5);
  } /* for */

#ifdef DEBUG_DB  
  debug("deleteRow", "%s\n", DStringValue(&del));
#endif

  if ((mysql_query(&mysql, (const char*) DStringValue(&del))) != 0)
  {
    error("deleteRow", "Failed on delete from %s table: Error: %s\n",
	  table, mysql_error(&mysql));
    retval = -1;
  }

  DStringFree(&del);
  return retval;
} /* deleteRow */

int
MysqlConnection::insertOrReplaceRow(const char * table, const char *command,
                                    const char ** columns,
                                    const char ** values,
                                    int num_columns) 
{
  int retval = 0;
  int i;
  DString insert;
  const char * test; /* used to determine float or int column values */
  int *float_values; /* bit map of float values */
  int *int_values; /* bit map of int values */

  float_values = new int[num_columns];
  int_values = new int[num_columns];

  DStringInit(&insert);

  /* initialize bit arrays */
  for (i = 0; i < num_columns; i++) {
    float_values[i] = 0;
    int_values[i] = 0;
  }
    
  DStringAppend(&insert, command, -1);
  DStringAppend(&insert, " into ", -1);
  DStringAppend(&insert, (const char*) table, strlen(table));
  DStringAppend(&insert, " (", 2);

  for (i = 0; i < num_columns; i++) {
    test = columns[i];
    if ((*test == '#') || (*test == '!')) {
      if (*test == '#')
	int_values[i] = 1;
      else
	float_values[i] = 1;

      DStringAppend(&insert, test+1, strlen(test)-1);
    }
    else {
      DStringAppend(&insert, columns[i], strlen(columns[i]));
    }
    if (i < (num_columns-1) )
      DStringAppend(&insert, ",", 1);
  }

  DStringAppend(&insert, ") values(", 9);

  for (i = 0; i < num_columns; i++) {
    if (int_values[i] == 1) { /* int value for column */
      int int_value;
      DString columnValue;
      DStringInit(&columnValue);

      int_value = atoi(values[i]);
      DStringPrintf(&columnValue, "%d", int_value);
      DStringAppendD(&insert, &columnValue);
      DStringFree(&columnValue);
    }
    else if (float_values[i] == 1) { /* float value for column */
      float float_value;
      DString columnValue;
      DStringInit(&columnValue);

      float_value = (float) atof(values[i]);
      DStringPrintf(&columnValue, "%f", float_value);
      DStringAppendD(&insert, &columnValue);
      DStringFree(&columnValue);
    }
    else { /* string value for column */
      DStringAppend(&insert, "'", 1);
      DStringAppend(&insert, (const char*) values[i], strlen(values[i]));
      DStringAppend(&insert, "'", 1);
    }
    if (i < (num_columns-1) )
      DStringAppend(&insert, ",", 1);
  }

  DStringAppend(&insert, ")", 1);

#ifdef DEBUG_DB  
  debug("insertRow", "%s\n", DStringValue(&insert));
#endif

  if ((mysql_query(&mysql, (const char*) DStringValue(&insert))) != 0) {
    error("insertRow", "Failed on insert into %s table: Error: %s\n",
	  table, mysql_error(&mysql));
    retval = -1;
  }

  DStringFree(&insert);

  delete[] int_values;
  delete[] float_values;

  return retval;
  
} /* insertRow */

int
MysqlConnection::insertRow(const char * table, const char ** columns,
                           const char ** values,
                           int num_columns) 
{
  return insertOrReplaceRow(table, "insert", columns, values, num_columns);
}

int
MysqlConnection::replaceRow(const char * table,
                           const char ** columns, const char ** values,
                           int num_columns) 
{
  return insertOrReplaceRow(table, "replace", columns, values, num_columns);
}

int
MysqlConnection::updateRow(const char * table, const char * key,
			   const char * value, const char ** columns,
                           const char ** values, 
			   int num_columns,
                           const char *secondkey, const char *secondvalue,
                           int start_update_from) 
{
  int retval = 0;
  int i;
  DString update;
  const char * test; /* used to determine float or int column values */
  DStringInit(&update);

  DStringTrunc(&update);
  DStringAppend(&update, "update ", 7);
  DStringAppend(&update, (const char*) table, strlen(table));
  DStringAppend(&update, " set ", -1);
  
  for (i = start_update_from; retval >= 0 && i < num_columns; i++) {

    //DStringAppend(&update, " ", 1);
    test = columns[i];
    if (*test == '#') { /* int value for column */
      int int_value;
      DString columnValue;
      DStringInit(&columnValue);

      DStringAppend(&update, test+1, strlen(test)-1);

      DStringAppend(&update,"=",1);
      int_value = atoi(values[i]);
      DStringPrintf(&columnValue, "%d", int_value);
      DStringAppendD(&update, &columnValue);
      DStringFree(&columnValue);
    }      
    else if (*test == '!') { /* float value for column */
      float float_value;
      DString columnValue;
      DStringInit(&columnValue);

      DStringAppend(&update, test+1, strlen(test)-1);

      DStringAppend(&update, "=", 1);
      float_value = (float) atof(values[i]);
      DStringPrintf(&columnValue, "%.2f", float_value);
      DStringAppendD(&update, &columnValue);
      DStringFree(&columnValue);

    }
    else { /* string value for column */
      DStringAppend(&update, (const char*) columns[i], strlen(columns[i]));
      DStringAppend(&update, "=", 1);
      DStringAppend(&update, "'", 1);
      DStringAppend(&update, (const char*) values[i], strlen(values[i]));
      DStringAppend(&update, "'", 1);
    }

    if (i < (num_columns-1) ) {
      DStringAppend(&update, ",", 1);
    }
  }

  DStringAppend(&update, " where ", 7);
  DStringAppend(&update,  (const char*) key, strlen(key));
  DStringAppend(&update," = '", 4);
  DStringAppend(&update, (const char*) value, strlen(value));
  DStringAppend(&update, "'", 1);
  
  if ( secondkey != NULL ) {
    DStringAppend(&update, " and ", -1);
    DStringAppend(&update,  (const char*) secondkey, -1);
    DStringAppend(&update," = '", 4);
    DStringAppend(&update, (const char*) secondvalue, -1);
    DStringAppend(&update, "'", 1);
  }
  
#ifdef DEBUG_DB  
  debug("updateRow", "%s\n", DStringValue(&update));
#endif
  
  if ((mysql_query(&mysql, (const char*) DStringValue(&update))) != 0) {
    error("updateRow", "Failed on update %s table: Error: %s\n",
          table, mysql_error(&mysql));
    retval = -1;
  }

  DStringFree(&update);
  return retval;
  
} /* updateRow */

int
MysqlConnection::getNumRowsReturned()
{
  if(res)
    return mysql_num_rows(res);
  else
    return 0;
}

char * 
MysqlConnection::getLastError()
{
  return mysql_error(&mysql);
}

MYSQL_ROW
MysqlConnection::getNextRow()
{
  currentRow = mysql_fetch_row(res);
  return currentRow;
}

unsigned int
MysqlConnection::getColumnCount()
{
  return mysql_num_fields(res);
}

unsigned long*
MysqlConnection::getColumnLengths()
{
  return mysql_fetch_lengths(res);
}

unsigned int
MysqlConnection::getRowMaxLength()
{
  unsigned int totalLength = 0;
  unsigned int i;
  unsigned int num_fields = getColumnCount();
  MYSQL_FIELD* fields;

  fields = mysql_fetch_fields(res);

  for (i = 0; i < num_fields; i++)
  {
    totalLength += fields[i].length;
  }

  return totalLength;
}


void
MysqlConnection::freeResultSet()
{
  /* free the memory for the result set */
  if (res) {
    mysql_free_result(res);
    res = NULL;
  }
}

void
MysqlConnection::createConnection(const char * user, const char * passwd, 
				  const char * database, 
				  const char * sqlServer, int port)
{
  if (!mysqlConnection) {
    mysqlConnection = new MysqlConnection(user, passwd, database, 
					  sqlServer, port);
  }
}

void 
MysqlConnection::destroyConnection()
{
  if (mysqlConnection) {
    delete mysqlConnection;
    mysqlConnection = NULL;
  }
}

/* externed "C" functions from dbapi.h */
/* Return 0 on success, -1 on failure. */
int initializeConnection(const char * user, const char * passwd, 
                         const char * database, const char * sqlServer, 
                         int port)
{
  int ret = 0;

  LOCK();
  try {
    MysqlConnection::createConnection(user, passwd, database, sqlServer,
                                      port);
  }
  catch (exception& e) {
    error("initializeConnection", "%s\n", e.what());
    ret = -1;
  }
  catch (...) {
    error("initializeConnection", "Unknown error\n");
    ret = -1;
 }

  UNLOCK();

  return ret;
}


void closeConnection()
{
  LOCK();
  MysqlConnection::destroyConnection();
  UNLOCK();
}

ResultTable query(const char * table, const char * key, const char * value, 
		  const char * columns, int * rowsReturned)
{
  ResultTable rt = NULL;
  MysqlConnection * conn;
  int rowsRetrieved = 0;

  LOCK();

  conn  = MysqlConnection::getMySQLConnection();
  if (conn == NULL) {
    error("query", "MySQL Connection not configured/initialized\n");
    *rowsReturned = -1;
    UNLOCK();
    return NULL;
  }

  /* first query table */
  int succ = conn->queryTable(table, key, value, columns);

  if (succ == 0) { /* only build table if successful */ 
    rowsRetrieved = conn->getNumRowsReturned();
    if (rowsRetrieved > 0) { /* make sure some rows were returned */
      rt = (ResultTable) malloc(rowsRetrieved*sizeof(MYSQL_ROW));
      int columnCount = (int)conn->getColumnCount();
      
      for (int i = 0; i < rowsRetrieved; i++) {
	MYSQL_ROW row = conn->getNextRow();
	if (row) { /* copy each row into result table */
	  unsigned long* lengths = conn->getColumnLengths();
	  long rowLength = 0;
	  for (int x = 0; x < columnCount; x++) {
	    rowLength += lengths[x];
	  }

	  if (rowLength > 0) { /* only allocate and copy if there is data */
	    rt[i] = (MYSQL_ROW) calloc(rowLength, sizeof(char *));
	    for (int j = 0; j < columnCount; j++) {
	      if (row[j])
		strassign(&rt[i][j], row[j]);
	      else
		rt[i][j] = NULL; /* make sure to NULL values */
	    }
	  }
	  else { /* null out all columns */
	    unsigned int maxLength = conn->getRowMaxLength();
	    rt[i] = (MYSQL_ROW) calloc(maxLength, sizeof(char*));
	    for (int j = 0; j < columnCount; j++) {
	      rt[i][j] = NULL;
	    }
	  }
	}
      }
    }
    /* free up memory */
    conn->freeResultSet();
  }
  else {
    *rowsReturned = -1; /* indicates database error */
  }
  *rowsReturned = rowsRetrieved;

  UNLOCK();

  return rt;
}

ResultTable rawQuery(const char * sqlQuery, int * rowsReturned)
{
  ResultTable rt = NULL;
  MysqlConnection * conn;
  int rowsRetrieved = 0;

  LOCK();

  conn  = MysqlConnection::getMySQLConnection();
  if (conn == NULL) {
    error("query", "MySQL Connection not configured/initialized\n");
    *rowsReturned = -1;
    UNLOCK();
    return NULL;
  }

  /* first query table */
  int succ = conn->rawQueryTable(sqlQuery);

  if (succ == 0) { /* only build table if successful */ 
    rowsRetrieved = conn->getNumRowsReturned();
    if (rowsRetrieved > 0) { /* make sure some rows were returned */
      rt = (ResultTable) calloc(rowsRetrieved,sizeof(MYSQL_ROW));
      int columnCount = (int)conn->getColumnCount();
      
      for (int i = 0; i < rowsRetrieved; i++) {
	MYSQL_ROW row = conn->getNextRow();
	if (row) { /* copy each row into result table */
	  unsigned long* lengths = conn->getColumnLengths();
	  long rowLength = 0;
	  for (int x = 0; x < columnCount; x++) {
	    rowLength += lengths[x];
	  }

	  if (rowLength > 0) { /* only allocate and copy if there is data */
	    rt[i] = (MYSQL_ROW) calloc(rowLength, sizeof(char *));
	    for (int j = 0; j < columnCount; j++) {
	      if (row[j])
		strassign(&rt[i][j], row[j]);
	      else
		rt[i][j] = NULL; /* make sure to NULL values */
	    }
	  }
	  else { /* null out all columns */
	    unsigned int maxLength = conn->getRowMaxLength();
	    rt[i] = (MYSQL_ROW) calloc(maxLength, sizeof(char*));
	    for (int j = 0; j < columnCount; j++) {
	      rt[i][j] = NULL;
	    }
	  }
	}
      }
    }
    /* free up memory */
    conn->freeResultSet();
    *rowsReturned = rowsRetrieved;
  }
  else {
    *rowsReturned = -1; /* indicates database error */
  }

  UNLOCK();

  return rt;
}

int removeAllData(const char * table, const char * key, const char * value)
{
  int ret;
  LOCK();
  MysqlConnection * conn = MysqlConnection::getMySQLConnection();
  if (conn == NULL) {
    error("removeAllData", "MySQL connection not configured/initialized\n");
    UNLOCK();
    return -1;
  }
  ret =  conn->deleteRow(table, key, value);
  UNLOCK();
  return ret;
}

int removeData(const char * table, const char ** columns,
               const char ** values, int num_columns)
{
  int ret;
  LOCK();
  MysqlConnection * conn = MysqlConnection::getMySQLConnection();
  if (conn == NULL) {
    error("removeData", "MySQL connection not configured/initialized\n");
    UNLOCK();
    return -1;
  }

  ret = conn->deleteRow(table, columns, 
                        values, num_columns);
  UNLOCK();
  return ret;
}

int insert(const char * table, const char ** columns, const char ** values, 
	   int num_columns)
{
  int ret;
  LOCK();
  MysqlConnection * conn = MysqlConnection::getMySQLConnection();
  if (conn == NULL) {
    error("insert", "MySQL connection not configured/initialized\n");
    UNLOCK();
    return -1;
  }

  ret = conn->insertRow(table, columns, values,
                        num_columns);
  UNLOCK();
  return ret;
}

int updateOnly(const char * table, const char * key, const char * value, 
	   const char ** columns, const char ** values, int num_columns)
{
  int ret;
  LOCK();
  MysqlConnection * conn = MysqlConnection::getMySQLConnection();
  if (conn == NULL) {
    error("updateOnly", "MySQL connection not configured/initialized\n");
    UNLOCK();
    return -1;
  }

  ret = conn->updateRow(table, key, value,
                        columns, values,
                        num_columns);
  UNLOCK();
  return ret;
}

int updateOrInsert(const char * table, const char * key, const char * value, 
                   const char *secondkey, const char *secondvalue,
                   const char ** columns, const char ** values,
                   int num_columns, int start_update_from)
{
  int ret = 0;
  LOCK();
  MysqlConnection * conn = MysqlConnection::getMySQLConnection();
  if (conn == NULL) {
    error("update", "MySQL connection not configured/initialized\n");
    UNLOCK();
    return -1;
  }

  int succ;

  succ = conn->queryTable(table, key, value, columns[0], 
                          secondkey, secondvalue);
  if ( succ < 0 ) {
    debug("updateOrInsert", "queryTable failed\n");
    ret = -1;
  }
  else {
    /* Get the count of rows */
    succ = conn->getNumRowsReturned();
    conn->freeResultSet();

    if ( succ == 0 ) {
      ret = conn->insertRow(table, columns, values, num_columns);
      
    }
    else if ( succ > 0 ) {
      ret = conn->updateRow(table, key, value,
                            columns, values,
                            num_columns, 
                            secondkey, secondvalue,
                            start_update_from);
    }
    else {
      ret = -1;
    }
  }

  UNLOCK();
  return ret;
}

int isConnectionEstablished()
{
  int ret;
  LOCK();
  if (MysqlConnection::getMySQLConnection() != NULL)
    ret = 1;
  else
    ret = 0;
  UNLOCK();
  return ret;
}

char * getMysqlError()
{
  char * e;
  LOCK();
  MysqlConnection * conn = MysqlConnection::getMySQLConnection();
  if (conn == NULL) {
    UNLOCK();
    return "MySQL connection not configured/initialized";
  }

  e = conn->getLastError();
  UNLOCK();
  return e;
}

void freeResultTable(ResultTable table, int row_count, int column_count)
{
  LOCK();
  if (table) {
    for (int i = 0; i < row_count; i++) {
      if (table[i]) {
	for (int j = 0; j < column_count; j++) {
	  if (table[i][j]) 
	    free_valid(table[i][j]);
	}
        free_valid(table[i]);
      }
    }
    
    free_valid(table); 
  }
  UNLOCK();
}

#endif /* USE_MYSQL */

