#ifndef _MYSQL_CONN_H_
#define _MYSQL_CONN_H_

#include<stdio.h>
#include<mysql.h>

/* mysql database interface */

class MysqlConnection
{
 public:
  /* 
   * constructor creates a connection to the a database.  
   */
  MysqlConnection(const char * user, const char * passwd, const char * database, 
		  const char * host, int port);

  /* destructor */
  virtual ~MysqlConnection();

  /* 
   * public function to query(SELECT from) a table based on key, values 
   * and columns.  Returns 0 if successful, -1 if not successful.
   */
  int queryTable(const char * table, const char * key, const char * value, 
		 const char * columns, const char *secondkey = NULL,
                 const char *secondvalue = NULL); 

  /* 
   * public function to query(SELECT from) a table.
   * Returns 0 if successful, -1 if not. 
   */
  int rawQueryTable(const char * sqlQuery);

  /*
   * public function to delete from a table using a deletion key.  Returns
   * 0 if successful, -1 if not successful.
   */
  int deleteRow(const char * table, const char * key, const char * value);

  /*
   * public function to delete from a table using multiple conditions using 
   * AND to link conditions.  Returns 0 if successful, -1 if not successful.
   */
  int deleteRow(const char * table, const char ** columns, const char ** values, 
		int num_columns);
  /*
   * public function to insert into a table based on columns and values.  The
   * value type is assumed to be string unless a "1" is set in the position in
   * in the int_values or float_values array.  Returns 0 if succesful, -1 if not
   * successful.
   */
  int insertRow(const char * table, const char ** columns,
                const char ** values, int num_columns);

  /*
   * public function to replace into a table based on columns and values.  The
   * value type is assumed to be string unless a "1" is set in the position in
   * in the int_values or float_values array.  Returns 0 if succesful, -1 if not
   * successful.
   */
  int replaceRow(const char * table, const char ** columns,
                const char ** values, int num_columns);

  /*
   * public function to insert into a table based on columns and values.  The
   * value type is assumed to be string unless a "1" is set in the position in
   * in the int_values or float_values array.  Returns 0 if successful, -1 if not
   * successful.
   */
  int updateRow(const char * table, const char * key, const char * value,
		const char ** columns, const char ** values, int num_columns,
                const char *secondkey = NULL,
                const char *secondvalue = NULL,
                int start_update_from = 0);

  /*
   * public function to insert into a table based on columns and values.  The
   * value type is assumed to be string unless a "1" is set in the position in
   * in the int_values or float_values array.  Returns 0 if successful, -1 if not
   * successful.  This is a conditional update where first_key = XXX and second
   * key = YYY
   */
  int updateRow(const char * table, const char * firstKey, 
		const char * firstValue, const char * secondKey, 
		const char * secondValue,
                const char ** columns, const char ** values, 
		int num_columns);

  /*
   * public function to return the number of row retrieved from last executed 
   * query.
   */
  int getNumRowsReturned();

  /* 
   * public function to get the number of columns
   */
  unsigned int getColumnCount();

  /*
   * public function to get the column lengths of the current row
   */
  unsigned long* getColumnLengths();

  /*
   * public function to count the maximum length of all the fields
   */
  unsigned int getRowMaxLength();

  /*
   * public function to return last MYSQL error 
   */
  char * getLastError();

  /*
   * free memory of last result set 
   */
  void freeResultSet();

  /*
   * get the next row from the set of retrieved rows from the last executed
   * query.
   */
  MYSQL_ROW getNextRow();

  /* 
   * return singleton.  
   */
  static MysqlConnection* getMySQLConnection() { return mysqlConnection; }

  /* 
   * static function to allocate memory for singleton. 
   */
  static void createConnection(const char * user, const char * passwd, 
			       const char * database, const char * sqlServer, 
			       int port);

  /*
   * static function to destroy memory for singleton.
   */
  static void destroyConnection();

protected:
  /*
   * protected function to insert or replace into a table based on columns
   * and values.  The value type is assumed to be string unless a "1" is set
   * in the position in in the int_values or float_values array.  Returns 0
   * if succesful, -1 if not successful.
   * insertRow and replaceRow are wrappers around this function.
   */
  int insertOrReplaceRow(const char * table, const char *command,
                         const char ** columns, const char ** values, 
                         int num_columns);

 private:

  static MysqlConnection* mysqlConnection; /* singleton */
  MYSQL mysql; /* handle to database connection */
  MYSQL_RES * res; /* result of query (contains rows) */
  MYSQL_ROW currentRow; /* current row data in query result */
};

#endif
