/***************************** FILE HEADER *****************************
#               Copyright (c) 1999 by Columbia University              *
#                        All Rights Reserved                           *
#***********************************************************************
#***********************************************************************
#
#  File Name    : main.h
#
#  Module       : H323 Library API implementation.
#
#  Author(s)    : Kundan Singh (kns10@cs.columbia.edu)
#
#  Created      : 8 Oct 1999
#
#  Description  : This file implements the API for H323 Library.
#                 This maintains a mapping between API objects
#                 and openh323 objects.
#                 
#  Notes        : 
#
#***********************************************************************
#                Modification History (Reverse Order)
#-----------------------------------------------------------------------
#   Date         Author              Description
#-----------------------------------------------------------------------
#              |             | 
#              |             | 
#-----------------------------------------------------------------------
*/

/***********************************************************************
*****************************  INCLUDES  *******************************
*/
#include <ptlib.h>   /* general purpose portal windows library */
#include "main.h"    /* openh323 interface */

#include "iptelapi.h"/* generic iptel API */
#include "h323api.h" /* h323 API */
#include "h323pdu.h" /* h323 related protocols data units */
#include "h225.h"    /* H225 related definitions */

/* XXX use mutex for multithreading */

/*
 * InformAPIThread: Thread to inform the API about an incoming call.
 * This is a seperate thread so that the main module does not
 * block while the Application/User deciOnIndes whether to accept the call
 * or not. This thread just calls the OnNewIncomingCall() method
 * of a newly created H323Call object.
 */
class InformAPIThread : public PThread
{
    PCLASSINFO(InformAPIThread, PThread)
  public:
    InformAPIThread(H323Call *call,H323_Endpoint *server);
    void Main();

  protected:
    H323Call      *c;  /* Associated CallObject */
    H323_Endpoint *s;  /* Associated Endpoint */
};

#define new PNEW  /* required by ptlib */

///////////////////////////////////////////////////////////////
/////////////// H323 API Implementation ///////////////////////
//
// This part of the file imlplements the API methods defined
// in h323api.h 
//
///////////////////////////////////////////////////////////////

///////// H323_Endpoint ///////////////////////////////////////

H323_Endpoint::H323_Endpoint(bool IsServer, bool IsRegistrar)
: IptelEndpoint(IsServer,IsRegistrar) /* No registrar */
{
  /* Initialse a new Open H323 endpoint */
  PTRACE(3, "H323_Endpoint::H323_Endpoint(isServer)");

  H323SubGateway *ep;
  ep = new H323SubGateway;

  endpoint = ep;

  if ( ep == NULL )
  {
    PTRACE(2,"H323_Endpoint() can not initialize H323SubGateway");
    endpoint = NULL;
    return;
  }

  /* If this is a server then allocate a H323TCP Listener also */
  if ( IsServer )
  {
    /* Listen at well known port for server case */
    H323ListenerTCP * listener = new H323ListenerTCP(
                            *((H323SubGateway *)endpoint));
    if (!ep->StartListener(listener)) 
    {
      PTRACE(2, "H323_Endpoint() can not start listener");
      //delete ep;
      //delete listener; 
      listener = NULL;
      endpoint = NULL;
      PTRACE(3, "Returning from constructor");
      isserver = false;
      return;
    }
  }

  ListeningAddress = "127.0.0.1";
  ListeningAddress.SetPort(1720 /* default port */);

  ep->endpoint = this;

  if ( IsRegistrar ) {
    ep->StartGatekeeperServer();
  }

}

H323_Endpoint::H323_Endpoint(const IptelIPAddress & listeningAddress,
                             bool IsRegistrar)
: IptelEndpoint(listeningAddress,IsRegistrar) /* No registrar */
{
  PTRACE(3, "H323_Endpoint::H323_Endpoint(listeningAddress)");

  /* Initialse a new Open H323 endpoint */
  H323SubGateway *ep; /* Interface to OpenH323 */
  ep = new H323SubGateway;
  endpoint = ep;
  if ( ep == NULL )
  {
    PTRACE(2, "H323_Endpoint() can not create H323subGateway");
    return;
  }

  /*  
   * Listen at well known port for server case 
   * Effect is undefined if port is invalid 
   */
  H323ListenerTCP * listener = 
    new H323ListenerTCP(
                   *((H323SubGateway*)endpoint),
                   listeningAddress.GetPort());
  if (!ep->StartListener(listener)) 
  {
    PTRACE(2, "H323_Endpoint() can not start listener");
    delete ep;
    return;
  }

  ListeningAddress = "127.0.0.1";
  ListeningAddress.SetPort(listeningAddress.GetPort());

  ep->endpoint = this;

  if ( IsRegistrar ) {
    ep->StartGatekeeperServer();
  }

}

H323_Endpoint::~H323_Endpoint()
{
  PTRACE(3, "H323_Endpoint::~H323_Endpoint()");
  if ( endpoint != NULL) {
    ((H323SubGateway *)endpoint)->endpoint = NULL;
    delete (H323SubGateway *)endpoint;
  }
  endpoint = NULL;
}

IptelCall * H323_Endpoint::AllocateNewCallObject(const char *callid)
{
  /* Callid is just ignored since not needed for H.323 */
  PTRACE(3, "H323_Endpoint::AllocateNewCallObject");
  return new H323Call(*this);
}

int H323_Endpoint::Register(
      const IptelEndpoint::RegistrationInfo & reginfo)
{
  PTRACE(3, "H323_Endpoint::Register");

  /* Check if already present in the map */
  H323SubGateway *gw =  (H323SubGateway *)endpoint;

  /* if registrar is present then do only
   * RRQ. Otherwise GRQ and RRQ both. */

  if ( gw != NULL ) 
  {
    PTRACE(3, "H323\tDiscovering gatekeeper.");
    char * host = reginfo.registrar.GetHost();
   
    if ( host == NULL ) {
      if (gw->DiscoverGatekeeper(
              new H323TransportUDP(
                  *((H323SubGateway *)endpoint)))) {
        PTRACE(3, "H323\tGatekeeper found");
        cout << "\nGatekeeper found\n" << endl;
        return 0;
      }
      else {
        PTRACE(3, "H323\tGatekeeper not found.");
        return -1;
      }
    }
    else {
      PString gatekeeperAddress;
      gatekeeperAddress = PString("ip:") + host;
      free(host);

      PTRACE(3, "H323\tUsing the prespecified gatekeeper at " << gatekeeperAddress);
      if (gw->SetGatekeeper(
              gatekeeperAddress,
              new H323TransportUDP(
                  *((H323SubGateway *)endpoint)))) {
        PTRACE(3, "H323\tSpecified Gatekeeper found");
        cout << "\nSpecified Gatekeeper found\n" << endl;
        return 0;
      }
      else {
        PTRACE(3, "H323\tSpecified Gatekeeper not found.");
        return -1;
      }
    }
  }
  else {
    PTRACE(3, "H323\tNo H323 endpoint created");
    return -1;
  }
}

int H323_Endpoint::Unregister(
      const IptelEndpoint::RegistrationInfo & reginfo)
{
  PTRACE(2, "Unregister() not implemented");
  return 0;
}

int H323_Endpoint::OnLocationRequest(
      const IptelAddress & alias,
      IptelAddress & transport)
{
  PTRACE(2, "OnLocationRequest() not implemented");
  return -1; /* Default implementation does not succeed */
}

int H323_Endpoint::OnIncomingRegistration(
      const IptelEndpoint::RegistrationInfo & reginfo)
{
  PTRACE(2, "OnIncomingRegistration() not implemented");
  return 0; /* Default implementation always succeeds */
}

void H323_Endpoint::OnIncomingUnregistration(
      const IptelEndpoint::RegistrationInfo & reginfo)
{
  PTRACE(2, "OnIncomingUnregistration() not implemented");
  return; 
}

int H323_Endpoint::GetRegistration(            
                    const IptelAddress & addr,
                    IptelEndpoint::RegistrationInfo & reginfo)
{
  PTRACE(2, "GetRegistration()");
  if ( endpoint == NULL ) {
    PTRACE(1, "No endpoint");
    return 0;
  }

  H323SubGateway *ep = (H323SubGateway *)endpoint;
  H225_AliasAddress alias;
  PTRACE(3, "Address string for " << addr);
  PString name = addr.GetAddressString();
  
  PTRACE(3, "Address string is " << name);
  H323SetAliasAddress(name,alias);

  H225_TransportAddress trans;

  if ( ep != NULL ) {
    PTRACE(3, "GetRegistration() checking if RRQ was received");
    if ( ep->RegistrationExists(alias, trans) ) {
      PTRACE(3, "Registration found");
      H323TransportAddress taddr(trans);
      PTRACE(3, "Registration found as " << taddr);
      name = (const char *) taddr;
      reginfo.contactAddress = name;
      PTRACE(3, "Registration as contact " << reginfo.contactAddress);
      return 1;
    }
    else {
      PTRACE(3, "No registration found");
      return 0;
    }
  }
  else {
    PTRACE(1, "GetRegistration() no endpoint found");
    return 0;
  }

  return 0;
}
bool H323_Endpoint::LocationRequest(
                     const IptelAddress & dest,
                     IptelAddress & ipaddress)
{
  if ( endpoint == NULL ) {
    PTRACE(1, "No endpoint");
    return false;
  }

  H323SubGateway *ep = (H323SubGateway *)endpoint;

  PString name = dest.GetAddressString();
  H323TransportAddress trans;

  if ( ep != NULL ) {
    if ( ep->LocationRequest(name, trans) ) {
      name = trans;
      ipaddress = name;
      
      return true;
    }
    PTRACE(3, "LocationRequest() failed.");
    return false;
  }
  else {
    PTRACE(1, "LocationRequest() no endpoint found");
    return false;
  }
}             

ostream & operator<<(ostream & o,H323_Endpoint &ep)
{
    o << "Endpoint(H323): " 
      << ep.LocalName << " "
      << (ep.isserver ? "Server" : "Client") << ","
      << (ep.isregistrar ? "Registrar, " : " ");
    o << "Listening at " << ep.ListeningAddress << " ";
    if ( ep.endpoint != NULL ) {
      ((H323SubGateway *)ep.endpoint)->PrintRegistrationsOn(o);
    }
    return o;
}

/////////// H323 Call /////////////////////////////////////////

H323Call::H323Call()
: IptelCall()
{
  PTRACE(3, "H323Call::H323Call");
  /* Check if already present in the map */
  endpoint = NULL;
  connection = NULL;

  /* Allocate OpenH323 Endpoint */
  if ( (endpoint = new H323SubGateway ) == NULL )    {
    PTRACE(2, "H323Call() can not initialize H323SubGateway");
  }
  else {
    ((H323SubGateway *)endpoint)->call = this;
  }
}

H323Call::H323Call(H323_Endpoint &epoint)
: IptelCall()
{
  PTRACE(3, "H323Call::H323Call");
  H323_Endpoint *server = &epoint;
  connection = NULL;
  endpoint = NULL;

  /* Allocate OpenH323 Endpoint */
  H323SubGateway *ep ;
  if ( (ep = (H323SubGateway *)server->endpoint) == NULL ) {
    PTRACE(2, "H323Call(): server->endpoint is NULL");
    endpoint = new H323SubGateway;
  }
  else {
    endpoint = ep;
  }

  if ( endpoint == NULL ) {
    PTRACE(2, "H323_Call() can not create H323SubGateway");
  }
  else {
    ((H323SubGateway *)endpoint)->call = this;
  }
}

H323Call::~H323Call()
{
  PTRACE(3, "H323Call::~H323Call");
  if ( connection != NULL ) {
    ((H323SubGatewayConnection *)connection)->call = NULL;
  }
  connection = NULL;
  if ( endpoint != NULL ) {
    ((H323SubGateway *)endpoint)->call = NULL;
  }
  endpoint = NULL;
}

bool H323Call::IsActive()
{
  PTRACE(3, "H323Call::IsActive");
  /* Returns isactive if Connection is present else false */
  if ( connection == NULL ) {
    PTRACE(2, "IsActive() No connection");
    return false;
  }
  else {
    if ( isactive ) {
      PTRACE(3, "IsActive() returning TRUE");
    }
    else {
      PTRACE(3, "IsActive() returning FALSE");
    }

    return isactive;
  }
}

int   H323Call::Initiate(const IptelAddress & to)
{
    /* Set remote address */
  PTRACE(3, "H323Call::Initiate");
  SetRemoteAddress(to);
  return Initiate();
}

int   H323Call::Initiate(const IptelAddress & to,
                         const IptelAddress & from)
{
  PTRACE(3, "H323Call::Initiate");
  /* Set local and remote addresses */
  SetSelfAddress(from);
  SetRemoteAddress(to);
  return Initiate();
}

int   H323Call::Initiate()
{
  PTRACE(3, "H323Call::Initiate");
  H323SubGateway *ep = (H323SubGateway *)endpoint;;
  if ( ep != NULL )
  {
    /* XXX send ARQ to Gatekeeper if registered */

    IptelAddress to = GetRemoteAddress();
    IptelAddress from = GetSelfAddress();

    /* Create a new Transport Connection */
    H323TransportTCP * transport = new H323TransportTCP(*
       ((H323SubGateway *)endpoint));

    /* Check if Proxy address is specified then
    * use that for TCP connection */
    IptelAddress   proxy_dummy;
    proxy_dummy =  GetProxyAddress();

    if (proxy_dummy.IsIpAddress()) 
    {
      proxy_dummy.SetScheme("ip");
      proxy_dummy.SetUser(NULL);
      proxy_dummy.SetPassword(NULL);

      PString str;
      PTRACE(3, "Getting address string");

      str = proxy_dummy.GetAddressString();
      PTRACE(3, "Initiate() setting proxy to " << str);
      if (!transport->SetRemoteAddress(str)) {
        PTRACE(2, "Initiate() set remote address failure");
        delete transport;
        errorno = IPTELERR_INVALID_ADDRESS;
        return -1;
      }
    }
    else if (!transport->SetRemoteAddress( "ip:" +
                                          PString(to.GetAddressString()))) {
      delete transport;
      PTRACE(2, "Initiate() Set remote Address failed");
      errorno = IPTELERR_INVALID_ADDRESS;
      return -1;
    }

    /* Set the local name */
    ep->SetLocalUserName(from.GetAddressString());

    if (!ep->MakeCall(PString(to.GetAddressString()),
           transport, ep->currentCallToken)) {
      delete transport;
      PTRACE(2, "Initiate() Makecall() failed");
      errorno = IPTELERR_OPENH323_CORE;
      return -1;
    }
  }
  else
  {
    errorno = IPTELERR_OPENH323_INTF;
    PTRACE(2, "Initiate() No endpoint present");
    return -1;
  }
  return 0;
}

int   H323Call::Hangup(int /* reason */,const char * /*strReason*/)
{
  H323SubGateway *ep ;
  PTRACE(3, "H323Call::Hangup()");

  if ( (ep = (H323SubGateway *)endpoint ) != NULL)
  {
    if (!ep->currentCallToken) 
    {
      if (!ep->ClearCall(ep->currentCallToken))
      {
        PTRACE(2, "Hangup() ClearCall() failed");
        errorno = IPTELERR_OPENH323_CORE;
        return -1;
      }
      else {
        PTRACE(3, "H323Call::Hangup() complete");
      }
    }
    else {
      PTRACE(2, "Hangup() call token is already NULL");
    }
  }
  else
  {
    PTRACE(2, "Hangup() no endpoint present");
    errorno = IPTELERR_OPENH323_INTF;
    return -1;
  }
  return 0;
}

int   H323Call::Alert()
{
  /* Not implemented currently */
  PTRACE(3, "Alert() not implemented");
  return -1;
}

void H323Call::OnNewIncomingCall(IptelEndpoint & /* ep ignored */)
{
  /* Just Accept the call (by default ) */
  PTRACE(3, "OnNewIncomingCall() accepting by default");
  Accept();
  return;
}

int  H323Call::Accept()
{
  PTRACE(3, "H323Call::Accept");
  H323SubGateway *ep ;
  if ( (ep = (H323SubGateway *)endpoint) != NULL )
  {
    isactive = TRUE;
    PTRACE(3, "Accept(): Signaling Incoming call");
    ep->incomingCall.Signal(); /* Call is Answered */
    return 0;
  }
  else {
    PTRACE(2, "Accept(): No endpoint found");
  }
  return -1;
}

int  H323Call::Reject()
{ 
  PTRACE(3, "H323Call::Reject");
  isactive = FALSE;
  H323SubGateway *ep;
  if ( (ep = (H323SubGateway *)endpoint) != NULL )
  {
    PTRACE(2, "Reject(): Signaling IncomingCall");
    ep->incomingCall.Signal(); /* Call is Answered */
    return 0;
  }
  else {
    PTRACE(2, "Reject(): No endpoint found");
  }
  return -1;
}

void   H323Call::OnAlerting()
{
  /* Not implemented */
  PTRACE(2, "OnAlerting(): not implemented");
  return;
}

int  H323Call::WaitForResponse(int timeout /* in millisec */)
{
  PTRACE(3, "H323Call::WaitForResponse");
  H323SubGateway *ep;
  if ( (ep = (H323SubGateway *)endpoint) != NULL )
  {
    if ( timeout == -1 ) {
      PTRACE(3, "WaitForResponse(): waiting for INFINITE time");
      ep->responseFlag.Wait();
      PTRACE(3, "WaitForResponse(): waiting completed");
      return 0;
    }
    else {
      int oRet;
      PTimeInterval interval(timeout);
      PTRACE(3, "WaitForResponse(): waiting for interval");
      oRet =  ep->responseFlag.Wait(interval) ? 0 : 1; 
      PTRACE(3, "WaitForResponse(): waiting completed");
      /* return 1 for timeout */
      return oRet;
    }
  }
  else
  {
    PTRACE(2, "WaitForResponse(): No endpoint found");
    return -1; /* Error */
  }
}

void   H323Call::OnCallEstablished()
{
  PTRACE(3, "H323Call::OnCallEstablished");
  isactive = TRUE;
  H323SubGateway *ep;
  if ( (ep = (H323SubGateway *)endpoint) != NULL )
  {
    PTRACE(3, "OnCallEstablished(): Signaling ResponseFlag");
    ep->responseFlag.Signal();
  }
  else {
    PTRACE(2, "OnCallEstablished(): No endpoint found");
  }
  return;
}

int  H323Call::OnCallRejected()
{
  PTRACE(3, "H323Call::OnCallRejected");
  isactive = FALSE;
  H323SubGateway *ep;
  if ( (ep = (H323SubGateway *)endpoint) != NULL )
  {
    PTRACE(3, "OnCallRejected(): Signaling OLC flag");
    ep->responseFlag.Signal();
  }
  else {
    PTRACE(2, "OnCallRejected(): No endpoint found");
  }
  return 0;
}

int  H323Call::OnHangup()
{
  PTRACE(3, "H323Call::OnHangup");
  isactive = FALSE;
  H323SubGateway *ep;
  if ( (ep = (H323SubGateway *)endpoint) != NULL )
  {
    /*
    cout << "OpenH323: Signaling ResponseFlag" << endl;
    endpoint->responseFlag.Signal();
    cout << "OpenH323: Signaling OLC Flag" << endl;
    endpoint->olcFlag.Signal();
    */
    PTRACE(2, "OnHangup(): closing call");
    ep->responseFlag.Signal(); /* This should be done based on the
                                * call status */
    //Incorrect hence commented ep->endpoint = NULL;
  }
  else {
    PTRACE(2, "OnHangup(): No endpoint found");
  }
  endpoint = NULL;
  if ( connection != NULL ) {
    ((H323SubGatewayConnection *)connection)->call = NULL;
  }
  else {
    PTRACE(2, "OnHangup(): no connection present");
  }
  connection = NULL;

  return 0;
}

void   H323Call::OnError(int /* errorno is ignored */)
{
  PTRACE(3, "H323Call::OnError");
  H323SubGateway *ep;
  if ( (ep = (H323SubGateway *)endpoint) != NULL )
  {
    /* Signal all semaphores */
    cout << "OpenH323: OnError() called" << endl;
    //cout << "OpenH323: Signaling ResponseFlag" << endl;
    ep->responseFlag.Signal();
    //cout << "OpenH323: Signaling OLC flag" << endl;
    ep->olcFlag.Signal();
    //cout << "OpenH323: Signaling Incoming Call" << endl;
    ep->incomingCall.Signal();
  }

  endpoint = NULL;
  if ( connection != NULL ) {
    ((H323SubGatewayConnection *)connection)->call = NULL;
  }
  else {
    PTRACE(2, "OnHangup(): no connection present");
  }
  connection = NULL;

  //cout << "Error: " << errno << ":" << H323Call::ErrorString(errno) << endl;
  return;
}

int  H323Call::SetProxyAddress(const IptelAddress & proxyAddr)
{
  PTRACE(3, "H323Call::SetProxyAddress");
  proxy = proxyAddr;
  return 0;
}

IptelAddress     H323Call::GetProxyAddress()
{
  PTRACE(3, "H323Call::GetProxyAddress");
  return proxy;
}

int H323Call::WaitForTermination(int /* timeout is always -1 */)
{
  /* Not implemented */
  PTRACE(2, "WaitForTermination(): Not implemented");
  return -1;
}

void H323Call::OnOpenLogicalChannelAck()
{
  PTRACE(3, "H323Call::OnOpenLogicalChannelAck");
  isactive = TRUE;
  H323SubGateway *ep;
  if ( (ep = (H323SubGateway *) endpoint) != NULL )
  {
    PTRACE(3, "OnOpenLogicalChannelAck(): Signaling OLC flag");
    ep->olcFlag.Signal();
  }
  else {
    PTRACE(2, "OnOpenLogicalChannelAck() no endpoint found");
  }
  return;
}

int    H323Call::WaitForOpenLogicalChannelAcks(int timeout)
{
  PTRACE(3, "H323Call::WaitForOpenLogicalChannelAcks");
  H323SubGateway *ep;
  if ( (ep = (H323SubGateway *) endpoint) != NULL )
  {
    int oReturn = 0;
    while ( remotesession == NULL ) {
      if ( timeout == -1 ) {
        PTRACE(3, "WaitForOLCAck() waiting for INFINITE time");
        ep->olcFlag.Wait();
        PTRACE(3, "WaitForOLCAck() completed");
      }
      else {
        PTimeInterval interval(timeout);
        PTRACE(3, "WaitForOLCAck() waiting for interval");
        oReturn = (ep->olcFlag.Wait(interval) ? 0 : 1); 
        PTRACE(3, "WaitForOLCAck() completed");
        /* return 1 for timeout */
      }
    }
    PTRACE(3, "WaitForOLCAck() exiting");
    return oReturn;
  }
  else
  {
    PTRACE(3, "WaitForOLCAck() endpoint not found");
    return -1;
  }
}

///////////////////////////////////////////////////////////////
/////////////// Modification in OpenH323 //////////////////////
//
// This part of the file implements the derived classes of
// openh323 (H323SubGateway and H323SubGatewayConnection)
// declared in main.h header file.
//
///////////////////////////////////////////////////////////////

//////// SubGateway ///////////////////////////////////////////

H323SubGateway::H323SubGateway()
  : H323EndPoint()
{
  PTRACE(3, "H323SubGateway::H323SubGateway");
  endpoint = NULL;
  call = NULL;
  registrations.AllowDeleteObjects();
}

H323SubGateway::~H323SubGateway()
{
  PTRACE(3, "H323SubGateway::~H323SubGateway");

  if ( endpoint != NULL ) {
    ((H323_Endpoint *)endpoint)->endpoint = NULL;
  }
  endpoint = NULL;
  call = NULL;
}

H323Connection * H323SubGateway::CreateConnection(
                                 unsigned callReference)
{
  /* Second param is FALSE to indicate that Fast start is
   * always present */
  PTRACE(3, "H323SubGateway::CreateConnection()");
  return new H323SubGatewayConnection(*this, callReference, FALSE);
}


void H323SubGateway::OnConnectionEstablished(H323Connection & cn
                                             /* connection is ignored */,
                                             const PString & token)
{
  PTRACE(3, "H323SubGateway::OnConnectionEstablished");
  currentCallToken = token;
  
  /* Inform API */
  H323Call *c;
  if ( (c = (H323Call *)((H323SubGatewayConnection*)&cn)->call) != NULL )
  {
    PTRACE(3, "H323SubGateway::OnConnectionEstablished()");
    c->OnCallEstablished();
  }
  else {
    PTRACE(3, "H323SubGateway::OnConnectionEstablished() no call");
  }
}


void H323SubGateway::OnConnectionCleared(H323Connection & cn
                                         /* connection is ignored */,
                                         const PString &)
{
  PTRACE(3, "H323SubGateway::OnConnectionCleared");
  /* Nullify callToken */
  currentCallToken = "";

  /* Inform API */
  H323Call *c;
  if ( (c = (H323Call *)((H323SubGatewayConnection *)&cn)->call) != NULL )
  {
    PTRACE(3, "H323SubGateway::OnConnectionCleared()");
    c->OnHangup();
    call = NULL;
    c->endpoint = NULL; /* XXX is this correct */
  }
  else {
    PTRACE(3, "H323SubGateway::OnConnectionCleared() no call");
  }
}


BOOL H323SubGateway::RegistrationExists(
     const H225_AliasAddress & alias,
     H225_TransportAddress & trans)
{
  PTRACE(3, "G323SubGateway\tRegistrationExists");

  if ( registrations.GetSize() <= 0 ) {
    return FALSE;
  }

  for ( int i = ((int) registrations.GetSize())-1; i>=0; i--) {
    H323RegistrationInfo *reg = (H323RegistrationInfo *)
      registrations.GetAt(i);
    if ( reg == NULL ) {
      continue;
    }
    if ( reg->IsEqual(alias) ) {
      reg->SetUpTransportAddress(trans);
      return TRUE;
    }
  }
  return FALSE;
}

BOOL H323SubGateway::RegistrationExists(
     const H225_ArrayOf_AliasAddress & aliases,
     const H225_ArrayOf_TransportAddress & trans,
     H225_EndpointIdentifier & endpointId)
{
  BOOL oFound = FALSE;
  BOOL oFlag1 = FALSE;

  for (int i = ((int)registrations.GetSize())-1; i>=0; i--) {
    PTRACE(3, "Loop with  i = " << i );
    H323RegistrationInfo *reg = (H323RegistrationInfo *) 
      registrations.GetAt(i);
    if ( reg == NULL ) {
      PTRACE(3, "Element is NULL");
      continue;
    }
    if ( reg->IsEqual(trans[0]) ) {

      oFlag1 = TRUE;

      reg->GetEndpointIdentifier(endpointId);

      /* Same transport found */
      /* Check for all aliases */
      for ( int j = 0; j< aliases.GetSize(); j ++) {
        oFound = FALSE;

        for (int k = ((int)registrations.GetSize())-1; k>=0; k--) {
          PTRACE(3, "Loop with j = " << j << " and i = " << k );
          H323RegistrationInfo *reg2 = (H323RegistrationInfo *) 
            registrations.GetAt(k);
          if ( reg2 == NULL ) {
            PTRACE(3, "Element is NULL");
            continue;
          }
          if ( reg2->IsEqual(aliases[j], trans[0]) ) {
            oFound = TRUE;
            break;
          }
        }
        if ( !oFound ) {
          H323RegistrationInfo *reg2 = new 
            H323RegistrationInfo(endpointId, aliases[j],trans[0]);
          registrations.Append(reg);
        }
      }
      
    }
  }

  return oFlag1;
    
}

BOOL H323SubGateway::OnIncomingRegistration(
     const H225_ArrayOf_AliasAddress & aliases,
     const H225_ArrayOf_TransportAddress & trans,
     H225_EndpointIdentifier & endpointId)
{
  static int epId = 1;

  PTRACE(3, "H323SubGateway\tOnIncomingRegistration");

  if ( RegistrationExists(aliases, trans, endpointId)) {
    /* Check if the registration already exists then
     * set  endpoint To appropriate value and return TRUE */
    return TRUE;
  }

  PString endpointIdStr;
  endpointIdStr.sprintf("Terminal %d", epId);
  epId ++;

  endpointId = (const char *) endpointIdStr;

  PTRACE(3, "Endpoint ID is " << endpointIdStr);

  /* Only one transport address is used */
  if ( trans.GetSize() <= 0 ) {
    PTRACE(2, "H323SubGateway\tRRQ: No transport address, rejecting");
    return FALSE;
  }

  for ( PINDEX i = 0; i < aliases.GetSize(); i++ ) {

    PTRACE(3, "Entered in loop");

    PTRACE(3, "Alias " << i << " is \n" << aliases[i]);

    H225_AliasAddress & alias = aliases[i];

    PTRACE(3, "New registration Info created");

    H323RegistrationInfo *reg = new H323RegistrationInfo(
                          endpointId, alias, trans[0]);
    
    /* On new registrations to be called here */
    H323_Endpoint *ep;
    BOOL oAppend = FALSE;
    if ( (ep = (H323_Endpoint *)endpoint) != NULL) {
      IptelEndpoint::RegistrationInfo reginfo;

      PString str;
      str = PIPSocket::GetHostName();
      reginfo.registrar = (const char *) str;
      str = H323GetAliasAddressString(alias);
      reginfo.localName = str;
      H323TransportAddress taddr(trans[0]);
      str = (const char *) taddr;
      reginfo.contactAddress = str;
      /*
       * EXPIRATION and q_value are not set XXX 
       */

      if ( ep->OnIncomingRegistration(reginfo) >=0 ) {
        /* Accept the registration */
        PTRACE(3, "API accepted the registration");
        oAppend = TRUE;
      }
      else {
        /* Rejected the registration */
        PTRACE(3, "API rejected the registration");
        oAppend = FALSE;
      }
    }
    else {
      /* No endpoint. Just add into the list */
      PTRACE(3, "No endpoint present, adding");
      oAppend = TRUE;
    }

    if ( oAppend ) {
      PTRACE(3, "Adding in list of registrations");
      registrations.Append(reg);
    }
    else {
      delete reg;
    }
  }

  PTRACE(3, "OnIncomingRegistration completed");
  return TRUE; /* GK is not implemented */
}

void H323SubGateway::InformUnregistrationToAPI(
     const H225_AliasAddress & alias,
     const H225_TransportAddress & trans
     )
{
  H323_Endpoint *ep;
  PTRACE(3, "Trying to inform API about unregistration");

  if ( (ep = (H323_Endpoint *)endpoint) != NULL) {
    IptelEndpoint::RegistrationInfo reginfo;
    
    PString str;
    str = PIPSocket::GetHostName();
    reginfo.registrar = (const char *) str;
    str = H323GetAliasAddressString(alias);
    reginfo.localName = str;
    H323TransportAddress taddr(trans);
    str = (const char *) taddr;
    reginfo.contactAddress = str;
    ep->OnIncomingUnregistration(reginfo);
  }
}

void H323SubGateway::OnIncomingUnregistration(
     const H225_EndpointIdentifier & endpointId,
     const H225_ArrayOf_AliasAddress & aliases,
     const H225_ArrayOf_TransportAddress & trans
     )
{
  PTRACE(3, "H323SubGateway\tOnIncomingUnregistration[1]");

  for ( int j = 0; j< aliases.GetSize(); j ++) {
    PTRACE(3, "H323SubGateway\t[1]");
    for (int i = ((int)registrations.GetSize())-1; i>=0; i--) {
      PTRACE(3, "Loop with j = " << j << " and i = " << i );
      H323RegistrationInfo *reg = (H323RegistrationInfo *) 
        registrations.GetAt(i);
      if ( reg == NULL ) {
        PTRACE(3, "Element is NULL");
        continue;
      }
      if ( reg->IsEqual(endpointId, aliases[j], trans[0]) ) {
        InformUnregistrationToAPI(reg->GetAliasAddress(), 
                                  reg->GetTransportAddress());
        PTRACE(3, "Removing");
        registrations.Remove(reg);
      }
    }
  }

  PTRACE(3, "H323SubGateway\tOnIncomingUnregistration[1] done");

  return; /* GK is not implemented */
}

void H323SubGateway::OnIncomingUnregistration(
     const H225_EndpointIdentifier & endpointId,
     const H225_ArrayOf_TransportAddress & trans
     )
{
  PTRACE(3, "H323SubGateway\tOnIncomingUnregistration[2]");

  for (int i = ((int)registrations.GetSize())-1; i>=0; i--) {
    H323RegistrationInfo *reg = (H323RegistrationInfo *) 
      registrations.GetAt(i);
    if ( reg->IsEqual(endpointId, trans[0]) ) {
      InformUnregistrationToAPI(reg->GetAliasAddress(), 
                                reg->GetTransportAddress());
      PTRACE(3, "Removing");
      registrations.Remove(reg);
    }
  }

  PTRACE(3, "H323SubGateway\tOnIncomingUnregistration[2] done");

  return; /* GK is not implemented */
}

void H323SubGateway::OnIncomingUnregistration(
     const H225_ArrayOf_AliasAddress & aliases,
     const H225_ArrayOf_TransportAddress & trans
     )
{
  PTRACE(3, "H323SubGateway\tOnIncomingUnregistration[3]");

  for ( int j = 0; j < aliases.GetSize();  j++) {
    for (int i = ((int)registrations.GetSize())-1; i>=0; i--) {
      H323RegistrationInfo *reg = (H323RegistrationInfo *) 
        registrations.GetAt(i);
      if ( reg->IsEqual(aliases[j], trans[0]) ) {
        InformUnregistrationToAPI(reg->GetAliasAddress(), 
                                  reg->GetTransportAddress());
        PTRACE(3, "Removing");
        registrations.Remove(reg);
      }
    }
  }

  PTRACE(3, "H323SubGateway\tOnIncomingUnregistration[3] done");

  return; /* GK is not implemented */
}

void H323SubGateway::OnIncomingUnregistration(
     const H225_ArrayOf_TransportAddress & trans
     )
{
  PTRACE(3, "H323SubGateway\tOnIncomingUnregistration[4]");

  for (int i = ((int)registrations.GetSize())-1; i>=0; i--) {
    H323RegistrationInfo *reg = (H323RegistrationInfo *) 
      registrations.GetAt(i);
    if ( reg->IsEqual(trans[0]) ) {
      InformUnregistrationToAPI(reg->GetAliasAddress(), 
                                reg->GetTransportAddress());
      PTRACE(3, "Removing");
      registrations.Remove(reg);
    }
  }

  PTRACE(3, "H323SubGateway\tOnIncomingUnregistration[4] done");

  return; /* GK is not implemented */
}

void H323SubGateway::PrintRegistrationsOn(ostream & strm) const
{
  if ( registrations.GetSize() > 0 ) {
    strm << "\n-----Registrations:-----";
    for ( PINDEX i = 0; i<registrations.GetSize(); i++) {
      strm << "\n" << i+1 << " ) Item " << endl;
      strm << registrations[i];
    }
    strm << "------- End --------------" << endl;
  }
}

BOOL H323SubGateway::OnIncomingAdmission(BOOL answerCall,
                         const H225_ArrayOf_AliasAddress & destInfo,
                         H225_TransportAddress & callSignal)
{
  PTRACE(2, "H323SubGateway\tOnIncomingAdmission: GK is not implemented");
  /* XXX should call an API to update default callSignal */
  return TRUE; /* Give self address */
}

BOOL H323SubGateway::OnLocationRequest(
      const H225_ArrayOf_AliasAddress & destInfo,
      H225_TransportAddress & callSignal)
{
  H323_Endpoint *ep;
  PTRACE(3, "Trying to inform API about LRQ");

  /* XXX search in local database first */

  if ( (ep = (H323_Endpoint *)endpoint) != NULL) {

    for ( int i = 0; i<destInfo.GetSize(); i++ ) {
      PString str;
      str = H323GetAliasAddressString(destInfo[i]);
      IptelAddress name, trans;
      name = (const char *) str;
      
      PTRACE(3, "Trying " << name );
      int oRet;

      if ( (oRet = ep->OnLocationRequest(name, trans)) >= 0 ) {
        
        if ( oRet == 0 ) {
          /* Use local transport address */
          PIPSocket::Address addr;
          if ( PIPSocket::GetHostAddress(addr) ) {
            /* IP address */
            PString n = addr.AsString();
            trans = (const char *) n;
          }
          else {
            PString n = PIPSocket::GetHostName(); /* Host name */
            trans = (const char *) n;
          }
        }

        H323TransportAddress taddr = trans.GetAddressString();
        taddr.SetPDU(callSignal);
        PTRACE(3, "Alias: " << name << " found as " << 
               (const char *) taddr);
        return TRUE;
      }
      PTRACE(3, "Alias: " << name << " not found");
    }
  }

  PTRACE(3, "OnLocationRequest failed");
  return FALSE; /* Give self address. */
}


//////// SubGatewayConnection /////////////////////////////////

H323SubGatewayConnection::H323SubGatewayConnection(
                          H323SubGateway & epoint,
                          unsigned callReference,
                          BOOL /* _noFastStart */)
  : H323Connection(epoint, callReference),
    myEndpoint(epoint)
{
  PTRACE(3, "H323SubGatewayConnection::H323SubGatewayConnection");
  H323Call *c        = NULL;
  H323SubGateway *ep = &epoint;

  PTRACE(3, "H323SubGatewayConnection()");

  call = NULL;
  c = (H323Call *) ep->call;

  if ( c == NULL ) {
    H323_Endpoint *server = NULL;
    if ( (server = (H323_Endpoint *) ep->endpoint ) != NULL )  {
      PTRACE(3, "Allocating new Call from server");
      /* Use server to allocate a call object */
      c = (H323Call *) server->AllocateNewCallObject(NULL);
    }
    else {
      PTRACE(3, "Allocating new call for default");
      /* Allocate the default call object */
      c = new H323Call();
    }
  }
  else {
    PTRACE(2, "H323SubGatewayConnection(): Call already presesnt");
  }

  //endpoint = ep;
  call = c;
  c->connection = this;
}

H323SubGatewayConnection::~H323SubGatewayConnection()
{
  PTRACE(3, "H323SubGatewayConnection::~H323SubGatewayConnection");
  if ( call != NULL ) {
    ((H323Call *)call)->connection = NULL;
  }

  call = NULL;
}

H323Connection::AnswerCallResponse H323SubGatewayConnection::OnAnswerCall(
                                    const PString & /* caller */,
                                    const H323SignalPDU & setupPDU/*setupPDU*/,
                                    H323SignalPDU & /*connectPDU*/)
{
  PTRACE(3, "H323SubGatewayConnection::OnAnswerCall");
  H323_Endpoint *server = (H323_Endpoint *) myEndpoint.endpoint;

  if ( server != NULL )
  {
    H323Call *c;
    /* Allocate Call if not present */
    if ( (c = (H323Call *) call) == NULL )
    {
      PTRACE(2, "AnswerCallResponse() allocating new call object");
      c = (H323Call *)server->AllocateNewCallObject(NULL);
      call = c;
      c->connection = this;
    }

    /* Set remote Name */
    IptelAddress remoteName;
    remoteName = remotePartyName; 
    c->SetRemoteAddress(remoteName); /* build remoteName from setupPDU */

    /* Update Other Call parameters received in incoming call */
    extern void SetParametersForIncomingCall(H323Call *,
                         const H323SignalPDU & setupPDU);
    SetParametersForIncomingCall(c,setupPDU);

    /* OnNewIncomingCall */
    /* Create a thread to call call->OnNewIncomingCall */
    /* Wait for API to respond */
    H323SubGateway *ep ;
    if ( (ep = (H323SubGateway*)c->endpoint) != NULL)
    {
      /* Set the local endpoint name */
      IptelAddress selfName = c->GetSelfAddress();

      ep->SetLocalUserName(selfName.GetAddressString());

      /* Inform API */
      ep->informAPIThread = new InformAPIThread(c,server);
      ep->incomingCall.Wait();
    }
    else {
      PTRACE(2, "AnswerCallResponse(): no endpoint found");
    }

    if ( c->IsActive() )
    {
      PTRACE(2, "AnswerCallResponse(): active");
      return AnswerCallNow;
    }
    else
    {
      PTRACE(2, "AnswerCallResponse(): not active, denying");
      return AnswerCallDenied;
    }
  }
  else
  {
    /* No server present - reject */
    PTRACE(2, "AnswerCallResponse(): no server present, rejecting");
    return AnswerCallDenied;
  }

  return AnswerCallNow;
}

BOOL H323SubGatewayConnection::OnOpenLogicalChannelAck(
      unsigned SessionId,
      PIPSocket::Address & remoteAddr,WORD dataPort,WORD dataControlPort)
{
  PTRACE(3, "H323SubGatewayConnection::OnOpenLogicalChannelAck");
  /* Only audio is supported 
   * does not work with NetSpeak::webphone */
  //if ( SessionId == RTP_Session::DefaultAudioSessionID )
  {
    H323Call *c;
    if ( (c = (H323Call *)call) != NULL )
    {
      SessionDescription *session = c->GetRemoteSessionDescription();
      if ( session == NULL )
      {
        MediaInfo media;
        session = new SessionDescription;
        session->SetSessionId((SESSIONID) SessionId); /* XXX This might be incorrect */
      }
      MediaInfo media;
      media.SetRxAddress(IptelIPAddress(
                           remoteAddr[0],
                           remoteAddr[1],
                           remoteAddr[2],
                           remoteAddr[3])
                          );
      media.SetDirection(MediaInfo::Receive);
      media.SetDataType(MediaInfo::Audio);
      media.AddFormat(0 /* n-kns10 assuming PCMU */);
      media.SetRxPort(dataPort);
      session->AddSession(media);

      if ( c->GetRemoteSessionDescription() == NULL )
      {
        c->SetRemoteSessionDescription(session);
      }

      c->OnOpenLogicalChannelAck();

      return TRUE;
    }
    else
    {
      /* No call exists; return false */
      PTRACE(2, "OnOLCAck(): No call exists");
      return FALSE;
    }
  }
  return FALSE;
}

BOOL H323SubGatewayConnection::OnAlerting(const H323SignalPDU & /*alertingPDU*/,
                                  const PString & /* username */)
{
  PTRACE(3, "H323SubGatewayConnection::OnAlerting");
  H323Call *c;
  if ( (c = (H323Call *) call) != NULL)
  {
    c->OnAlerting();
    return TRUE;
  }
  else
  {
    /* No call exists; return false */
    PTRACE(2, "OnAlerting() no call exists");
    return FALSE;
  }
}

BOOL H323SubGatewayConnection::GetLogicalChannelAddress(H323Channel::Directions dir,
                                  PIPSocket::Address &ipnum,
 								                  WORD &port,
                                  int DataType)
{
  PTRACE(3, "H323SubGatewayConnection::GetLogicalChannelAddress");
  H323Call *c;
  if ( (c = (H323Call *)call ) == NULL)
  {
    /* No call exists */
    PTRACE(2, "GetLogicalChannelAddress() No acll exists");
    return FALSE;
  }
  else
  {
    IptelIPAddress sessionAddress;

    SessionDescription *sessionList = c->GetSelfSessionDescription();
    if ( sessionList )
    {
      for ( int i=0; i<sessionList->GetMediaCount(); i++) 
      {
        MediaInfo & session = (*sessionList)[i];
        /* Only audio is supported */
        if ( ! ( (session.GetDataType() == MediaInfo::Audio) &&
                 (DataType == H323Capability::e_Audio) ) )
             continue;
        /* Only Unidirectional sessions are supported */
        /*
         * To be modified n-kns10.
        if ( ! ( (dir == H323Channel::IsTransmitter ) &&
                 (session->direction == MediaInfo::Transmit)  ||
                 (dir == H323Channel::IsReceiver ) && 
                 (session->direction == MediaInfo::Receive) ) )
                 continue;
         */
        IptelIPAddress selfAddress = session.GetRxAddress();
        unsigned short selfPort = session.GetRxPort();
        unsigned char a,b,c,d;
        selfAddress.GetIpAddress(a,b,c,d);
        ipnum = PIPSocket::Address(a,b,c,d);
        port = selfPort;

        if ( selfAddress.IsValid() && (port != 0))
        {
          return TRUE;
        }
        else
        {
          PTRACE(2, "GetLogicalChannelAddress(): Address not valid or port0");
          return FALSE;
        }
      }
    }
    else
    {
      PTRACE(2, "GetLogicalChannelAddress(): No session set");
      return FALSE;
    }
  }

  return FALSE;
}

BOOL H323SubGatewayConnection::PluginCapabilities()
{
  PTRACE(3, "H323SubGatewayConnection::PluginCapabilities");
  H323Call *c;
  if ( (c = (H323Call *)call ) == NULL)
  {
    PTRACE(2, "PluginCapabilities() no call presesnt");
    return FALSE;
  }
  else
  {
    SessionDescription *sessionList = c->GetSelfSessionDescription();
    if ( sessionList )
    {
      for ( int i=0; i<sessionList->GetMediaCount(); i++) 
      {
        MediaInfo & session = (*sessionList)[i];

        /* Only audio is supported */
        if ( !(session.GetDataType() == MediaInfo::Audio))
             continue;

        /* Only Unidirectional sessions are supported */
        if ( ! ((session.direction == MediaInfo::Transmit)  ||
                (session.direction == MediaInfo::Receive) ) )
                 continue;

        FormatInfo & format = session.GetFormatInfo();

        int pt = format.GetPayloadType();

        switch (pt)
        {
        case 0: /* PCMU */
          {
            PTRACE(3, "PluginCapabilities() adding PCMU");
            H323Capability * ulaw64 = 
              new H323_G711Capability(
                                      H323_G711Capability::muLaw, 
                                      H323_G711Capability::At64k);
            endpoint.AddCapability(ulaw64);
            endpoint.SetCapability(0, 0, ulaw64);
          }
          break;

        case 8: /* A Law 64 K */
          {
            PTRACE(3, "PluginCapabilities() adding PCMA");
            H323Capability * alaw64 = 
              new H323_G711Capability(
                                      H323_G711Capability::ALaw, 
                                      H323_G711Capability::At64k);
            endpoint.AddCapability(alaw64);
            endpoint.SetCapability(0, 0, alaw64);
          }
          break;
        default: /* Unknown */
          /* Currently we do not add any other capability */
          PTRACE(3, "PluginCapabilities() invalid for payloadtype");
          break;
        }
      }
    }
    else
    {
      PTRACE(2, "Plugin Capabilities(): Session list is NULL");
      return FALSE;
    }
  }

  return TRUE;
}

void SetParametersForIncomingCall(H323Call *call,
                                  const H323SignalPDU & setupPDU)
{
  PTRACE(3, "SetParametersForIncomingCall");
  const H225_Setup_UUIE & setup = setupPDU.m_h323_uu_pdu.m_h323_message_body;
    
    /* Set self Address */
    if (setup.HasOptionalField(H225_Setup_UUIE::e_destinationAddress) &&
        setup.m_destinationAddress.GetSize() > 0) {
      PTRACE(3, "SetParametersForIncomingCall setting dest");
      PString selfPartyName;
      selfPartyName = H323GetAliasAddressString(setup.m_destinationAddress[0]);
      if (setup.m_destinationAddress.GetSize() > 1) {
        PTRACE(3, "SetParametersForIncomingCall setting dest [2]");
        selfPartyName += " (";
        for (int i = 1; i < setup.m_destinationAddress.GetSize(); i++) {
          if (i > 1)
            selfPartyName += ", ";
          selfPartyName += H323GetAliasAddressString(setup.m_destinationAddress[i]);
        }
        selfPartyName += ')';
      }
      IptelAddress selfName;
      selfName = selfPartyName; 
      call->SetSelfAddress(selfName); /* build remoteName from setupPDU */
      PTRACE(3, selfPartyName);
    }
    else {
      PTRACE(3, "No Destination set");
    }


    /* If Fast start elemets are present then update SessionDescription */
    if (setup.HasOptionalField(H225_Setup_UUIE::e_fastStart)) 
    {
      PTRACE(3, "H225\tFast start is present");

      SessionDescription *session;
      session = new SessionDescription;

      // Extract capabilities from the fast start OpenLogicalChannel structures
      for (int i = 0; i < setup.m_fastStart.GetSize(); i++) 
      {
        H245_OpenLogicalChannel open;
        PPER_Stream strm = (PBYTEArray)setup.m_fastStart[i];
        if (open.Decode(strm)) {
          PTRACE(4, "H225\tFast start open:\n  " << setprecision(2) << open);
          unsigned error;

          const H245_H2250LogicalChannelParameters * param;
          const H245_DataType * dataType;
          H323Channel::Directions direction;

          if (open.HasOptionalField(H245_OpenLogicalChannel::e_reverseLogicalChannelParameters)) {
            if (open.m_reverseLogicalChannelParameters.m_multiplexParameters.GetTag() !=
                      H245_OpenLogicalChannel_reverseLogicalChannelParameters_multiplexParameters
                                                              ::e_h2250LogicalChannelParameters) {
              PTRACE(2, "H323\tCreateLogicalChannel - reverse channel, H225.0 only supported");
              continue; /* in the for loop */
            }

            PTRACE(3, "H323\tCreateLogicalChannel - reverse channel");
            dataType = &open.m_reverseLogicalChannelParameters.m_dataType;
            param = &(const H245_H2250LogicalChannelParameters &)
                              open.m_reverseLogicalChannelParameters.m_multiplexParameters;
            direction = H323Channel::IsTransmitter;
          }
          else {
            continue; /* Only Rx values are stored */
            if (open.m_forwardLogicalChannelParameters.m_multiplexParameters.GetTag() !=
                      H245_OpenLogicalChannel_forwardLogicalChannelParameters_multiplexParameters
                                                              ::e_h2250LogicalChannelParameters) {
              PTRACE(2, "H323\tCreateLogicalChannel - forward channel, H225.0 only supported");
              continue; /* in the for loop */
            }

            PTRACE(3, "H323\tCreateLogicalChannel - forward channel");
            dataType = &open.m_forwardLogicalChannelParameters.m_dataType;
            param = &(const H245_H2250LogicalChannelParameters &)
                              open.m_forwardLogicalChannelParameters.m_multiplexParameters;
            direction = H323Channel::IsReceiver;
          }

          // Only support unicast channels so *must* have a media control channel
          if (!param->HasOptionalField(H245_H2250LogicalChannelParameters::e_mediaControlChannel)) {
            PTRACE(2, "H323\tCreateLogicalChannel - multicast not supported");
            continue;
          }

          MediaInfo media;
          media.SetDirection(
            (direction == H323Channel::IsReceiver) ? 
            MediaInfo::Transmit : MediaInfo::Receive);

          if ((dataType->GetTag() == H245_DataType::e_audioData) )
          {
            media.SetDataType(MediaInfo::Audio);

            const H245_AudioCapability & audioCaps =  (*dataType);
            BOOL oFlag = FALSE;
            int payload;

            switch(audioCaps.GetTag())
            {
              case H245_AudioCapability::e_g711Ulaw64k:
                payload = 0;
                break;

              default:
                oFlag = TRUE;
                break;
            }
            if ( oFlag )
              continue;
            media.AddFormat(payload);
          }
          else if ((dataType->GetTag() == H245_DataType::e_videoData))
          {
            media.SetDataType(MediaInfo::Video);
            continue; /* Not supported currently */
          }
          else
          {
            PTRACE(3, "H323\tNon Audio/Video channel Not Supported");
            continue;
          }

          /* Assumtion is that RTP port is one less than RTCP port */
          PIPSocket::Address remoteAddr;
          WORD remoteDataPort,remoteControlPort;
          unsigned int errorCode;
          extern BOOL ExtractTransport(const H245_TransportAddress &,
                             PIPSocket::Address &, WORD &,
                             unsigned & errorCode);
          if (!ExtractTransport(param->m_mediaControlChannel, 
                remoteAddr, remoteControlPort, errorCode))
          {
            PTRACE(3, "H323\tMediaControlChannel address not present");           
            continue;
          }
          remoteDataPort = remoteControlPort - 1;

          media.SetRxAddress(IptelIPAddress(
                               remoteAddr[0],
                               remoteAddr[1],
                               remoteAddr[2],
                               remoteAddr[3])
                            );
          media.SetRxPort(remoteDataPort);
          
          session->AddSession(media);
        }
      }
      call->SetRemoteSessionDescription(session);
    }
}

////////////////// InformAPIThread definition /////////////////////

InformAPIThread::InformAPIThread(H323Call *call,H323_Endpoint *server)
: PThread(10000) /* Stack size is 10000 */
{
  PTRACE(3, "InformAPIThread::InformAPIThread");
  c = call;
  s = server;
  Resume();
}

void InformAPIThread::Main()
{
  PTRACE(3, "InformAPIThread::Main");
  c->OnNewIncomingCall(*s);
}

///// End of thread definition ////////////////////////////////////

H323RegistrationInfo::H323RegistrationInfo(H225_EndpointIdentifier & e,
                                           H225_AliasAddress & a,
                                           H225_TransportAddress & t)
{
  endpointId = new H225_EndpointIdentifier(e);
  *endpointId = e;

  alias = new H225_AliasAddress(a);
  *alias = a;

  transport = new H225_TransportAddress(t);
  *transport = t;

  PTRACE(3, "H323Registration created");
}

/* 
 * XXX delete all internal pointers in
 * ~H323RegistrationInfo() desctructor
 */

BOOL H323RegistrationInfo::IsEqual(const H225_EndpointIdentifier & e, 
                                   const H225_AliasAddress &a,
                                   const H225_TransportAddress &t)
{
  return ( *endpointId == e && *alias == a && *transport == t );
}
BOOL H323RegistrationInfo::IsEqual(const H225_EndpointIdentifier & e, 
                                   const H225_TransportAddress &t)
{
  return ( *endpointId == e && *transport == t);
}

BOOL H323RegistrationInfo::IsEqual(const H225_AliasAddress & a, 
                                   const H225_TransportAddress &t)
{
  return ( *alias == a && *transport == t );
}

BOOL H323RegistrationInfo::IsEqual(const H225_TransportAddress & t)
{
  return ( *transport == t );
}

BOOL H323RegistrationInfo::IsEqual(const H225_AliasAddress & a)
{
  return ( *alias == a );
}

BOOL H323RegistrationInfo::SetUpTransportAddress(H323TransportAddress
                                                 & t)
{
 t = *transport;
 return TRUE;
}

BOOL H323RegistrationInfo::SetUpTransportAddress(H225_TransportAddress
                                                 & t)
{
  t = *transport;
  return TRUE;
}

BOOL H323RegistrationInfo::SetUpAliasAddress(H225_AliasAddress & a)
{
  a = *alias;
  return TRUE;
}
BOOL H323RegistrationInfo::GetEndpointIdentifier(H225_EndpointIdentifier & e)
{
  e = *endpointId;
  return TRUE;
}

const H225_AliasAddress & H323RegistrationInfo::GetAliasAddress()
{
  return *alias;
}

const H225_TransportAddress & H323RegistrationInfo::GetTransportAddress()
{
  return *transport;
}

ostream & operator<<(ostream & o,H323RegistrationInfo &r)
{
  o << *r.endpointId;
  o << *r.alias;
  o << *r.transport;
  return o;
}

bool H323_Endpoint::IsLocalHost(const PString & host)
{
  return PIPSocket::IsLocalHost(host) ? true : false;
}

bool H323_Endpoint::GetHostAddress(const PString & host, PString &
                                   addr)
{
  BOOL oRet;
  PIPSocket::Address a;
  oRet = PIPSocket::GetHostAddress(host, a);
  if ( oRet ) {
    addr = a.AsString();
    return true;
  }
  else {
    return false;
  }
  
}


/******************* End of File ******************************************
*/

