/***************************** FILE HEADER *****************************
#               Copyright (c) 1999 by Columbia University              *
#                        All Rights Reserved                           *
#***********************************************************************
#***********************************************************************
#
#  File Name    : main.h
#
#  Module       : H323 Library API.
#
#  Author(s)    : Kundan Singh (kns10@cs.columbia.edu)
#
#  Created      : 8 Oct 1999
#
#  Description  : This file defines the API for H323 Library.
#                 Most of the classes are derived from iptelapi.h
#                
#  Notes        : 
#
#***********************************************************************
#                Modification History (Reverse Order)
#-----------------------------------------------------------------------
#   Date         Author              Description
#-----------------------------------------------------------------------
#              |             | 
#              |             | 
#-----------------------------------------------------------------------
*/

#ifndef _H323_API_H_
#define _H323_API_H_

/*
 * Includes 
 */
#include "iptelapi.h"

/*
 * H323Address, H323MediaInfo and H323SessionDescription are
 * not required. Instead, IptelAddress, MediaInfo and SessionDescription
 * are used.
 */

class H323_Endpoint : public IptelEndpoint
{
  public:
    /* 
     * Registrar (Gatekeeper) is currently not implemented 
     * and H323_Endpoint has a Server (i.e. incoming call
     * listener) by default. If the listening Address is
     * not provided in the constructor then INADDR_ANY:1720
     * is used.
     */
    H323_Endpoint( bool IsServer = true, bool IsRegistrar = false );
    H323_Endpoint( const IptelIPAddress & listeningAddress, 
                   bool IsRegistrar = false);

    virtual ~H323_Endpoint();
    
    /*
     * Derived classes should override this method to return
     * appropriate CallObject.
     * Default behaviour returns H323Call object.
     */
    virtual IptelCall *AllocateNewCallObject(const char *callid);

    /*
     * This is called to register with an external gatekeeper.
     * The gatekeeper address (reginfo.registrar) should be
     * unique. The first time Register is called, the library
     * discovers the gatekeeper and registers with it.
     * Gatekeeper discovery is not performed if the 
     * registrar is specified in reginfo.
     * Application should leave reginfo.registrar as empty
     * to allow the automatic discovery of the gatekeeper.
     */
    virtual int Register(
      const IptelEndpoint::RegistrationInfo & reginfo);

    /* 
     * This is called to unregister with the gatekeeper.
     */
    virtual int Unregister(
      const IptelEndpoint::RegistrationInfo & reginfo);

    /*
     * This callback is called when the system receives an
     * LRQ request. The implementation should
     * get the contact transport information
     * reginfo. The callback should return >=0 if
     * the location is found and -1 if not.
     * The caller (library) sends LCF/LRJ accordingly.
     * If return value is 0 then local contactAddress of
     * this endpoint is used. Otherwise if return > 0 then
     * the value stored in transport is used as
     * contact address.
     */
    virtual int OnLocationRequest(
                                  const IptelAddress & alias,
                                  IptelAddress & transport);

    /* Registrations */
    virtual int GetRegistration(
                   const IptelAddress & addr,
                   IptelEndpoint::RegistrationInfo & reginfo);
    virtual bool LocationRequest(
                   const IptelAddress & dest,
                   IptelAddress & ipaddress);              
    virtual int OnIncomingRegistration(
                   const IptelEndpoint::RegistrationInfo & reginfo);
    virtual void OnIncomingUnregistration(
                   const IptelEndpoint::RegistrationInfo & reginfo);

    /* Generic methods */
    static bool IsLocalHost(const PString & host);
    static bool GetHostAddress(const PString & host, PString & addr);

    /** operator << for printing onto output stream */
    friend ostream & operator<<(ostream & o, H323_Endpoint & ep);

    /*
     * Internal pointer for mapping between API and openh323
     */
    void *endpoint;
};


/*
 * H323Call: is derived from IptelCall and it models a H323 Call.
 * Most of the virtual methods of IptelCall are implemented here 
 */
class H323Call : public IptelCall
{
  public:
    H323Call();
    H323Call(H323_Endpoint &ep);
    virtual ~H323Call();

    /* Overrides from IptelCall */
    virtual bool  IsActive();
    virtual int   Initiate();
    virtual int   Initiate(const IptelAddress & to);
    virtual int   Initiate(const IptelAddress & to,
                           const IptelAddress & from);
    virtual int   Hangup(
                  int reason = 0, /* 0 for normal termination */
                  const char *strReason = "Normal Termination");
    virtual int    Alert();
    virtual int    Accept();
    virtual int    Reject();
    virtual void   OnNewIncomingCall(IptelEndpoint & ep);
    virtual void   OnAlerting();
    virtual int    WaitForResponse(int timeout = -1);
    virtual void   OnCallEstablished();
    virtual int    OnCallRejected();
    virtual int    OnHangup();
    virtual int    WaitForTermination(int timeout = -1);
    virtual int    SetProxyAddress(const IptelAddress & proxy);
    virtual IptelAddress     GetProxyAddress();
    virtual void   OnError(int errorno);

    /* additional methods */
    virtual void   OnOpenLogicalChannelAck();
    virtual int    WaitForOpenLogicalChannelAcks(int timeout = -1);

    /*
     * Internal pointer for mappoing between API and openh323 
     */
    void *connection;
    void *endpoint;

  protected :
    IptelAddress proxy;
};

#endif /* _H323_API_H_ */

/************************* End of File *****************************
*/
