/***************************** FILE HEADER *****************************
#               Copyright (c) 1999 by Columbia University              *
#                        All Rights Reserved                           *
#***********************************************************************
#***********************************************************************
#
#  File Name    : main.h
#
#  Module       : Open H323 Library.
#
#  Author(s)    : Kundan Singh (kns10@cs.columbia.edu)
#
#  Created      : 14 Oct 1999
#
#  Description  : Interface for OpenH323 and the Library API.
#                 This file defines two inportant classes:
#                 H323SubGateway and H323SubGatewayConnection
#                 derived from H323Endpoint and H323Connection
#                 of openh323.
#                
#  Notes        : 
#
#***********************************************************************
#                Modification History (Reverse Order)
#-----------------------------------------------------------------------
#   Date         Author              Description
#-----------------------------------------------------------------------
#              |             | 
#              |             | 
#-----------------------------------------------------------------------
*/

#ifndef _H323SUBGW_MAIN_H
#define _H323SUBGW_MAIN_H

#include "h323.h"

class H225_AliasAddress;

/*
 * Registration information.
 */
class H323RegistrationInfo : public PObject
{
  PCLASSINFO(H323RegistrationInfo, PObject)
  public:
  H323RegistrationInfo(H225_EndpointIdentifier &,
                       H225_AliasAddress &,
                       H225_TransportAddress &);

  BOOL IsEqual(const H225_EndpointIdentifier & e, 
               const H225_AliasAddress &a,
               const H225_TransportAddress &t);
  BOOL IsEqual(const H225_EndpointIdentifier & e, 
               const H225_TransportAddress &t);
  BOOL IsEqual(const H225_AliasAddress & a, 
               const H225_TransportAddress &t);
  BOOL IsEqual(const H225_TransportAddress & t);
  BOOL IsEqual(const H225_AliasAddress & a);

  BOOL SetUpTransportAddress(H323TransportAddress & t);
  BOOL SetUpTransportAddress(H225_TransportAddress & t);
  BOOL SetUpAliasAddress(H225_AliasAddress & a);
  BOOL GetEndpointIdentifier(H225_EndpointIdentifier & e);
  const H225_AliasAddress & GetAliasAddress();
  const H225_TransportAddress & GetTransportAddress();

  /** operator << for printing onto output stream */
  friend ostream & operator<<(ostream & o, H323RegistrationInfo & r);

  protected:
    H225_AliasAddress  *alias; /* XXX make this protected */
    H225_TransportAddress  *transport;
    H225_EndpointIdentifier  *endpointId;
};

PLIST(H323RegistrationInfoList, H323RegistrationInfo);

/**
 * H323SubGateway: This is derived from H323Endpoint of openh323.
 * This is the entry point into openh323.
 */
class H323SubGateway : public H323EndPoint
{
    PCLASSINFO(H323SubGateway, H323EndPoint);

  public:
    H323SubGateway();
    ~H323SubGateway();

    // overrides from H323EndPoint
    virtual H323Connection * CreateConnection(unsigned callReference);
    virtual void OnConnectionEstablished(H323Connection &, const PString &);
    virtual void OnConnectionCleared(H323Connection &, const PString &);

    virtual BOOL OnIncomingRegistration(
            const H225_ArrayOf_AliasAddress & aliases,
            const H225_ArrayOf_TransportAddress & trans,
            H225_EndpointIdentifier & endpointId);
    
    virtual void OnIncomingUnregistration(
            const H225_EndpointIdentifier & endpointId,
            const H225_ArrayOf_AliasAddress & aliases,
            const H225_ArrayOf_TransportAddress & trans);
    virtual void OnIncomingUnregistration(
            const H225_EndpointIdentifier & endpointId,
            const H225_ArrayOf_TransportAddress & trans);
    virtual void OnIncomingUnregistration(
            const H225_ArrayOf_AliasAddress & aliases,
            const H225_ArrayOf_TransportAddress & trans);
    virtual void OnIncomingUnregistration(
            const H225_ArrayOf_TransportAddress & trans);
    virtual void InformUnregistrationToAPI(
                                           const H225_AliasAddress & alias,
                                           const H225_TransportAddress & trans
                                           );

    BOOL RegistrationExists(
                            const H225_AliasAddress & alias,
                            H225_TransportAddress & trans);
    BOOL RegistrationExists(
                         const H225_ArrayOf_AliasAddress & aliases,
                         const H225_ArrayOf_TransportAddress & trans,
                         H225_EndpointIdentifier & endpointId);


    virtual BOOL OnIncomingAdmission(BOOL answerCall,
                         const H225_ArrayOf_AliasAddress & destInfo,
                         H225_TransportAddress & callSignal);

    virtual BOOL OnLocationRequest(const H225_ArrayOf_AliasAddress & destInfo,
                                   H225_TransportAddress & callSignal);

    void PrintRegistrationsOn(ostream & strm) const;

    PString    currentCallToken;  /* Call token */
    PSyncPoint responseFlag;      /* remote has responded the Q931 setup */
    PSyncPoint olcFlag;           /* H245 OLC is complete */
    PSyncPoint incomingCall;      /* Incoming call is answered by API */
    void *     informAPIThread;

    /*
     * Internal pointer for mapping between API and openh323 
     */
    void *endpoint; /* H323_Endpoint */
    void *call;     /* H323Call */

  protected:
    PSyncPoint exitFlag;          /* Exit the application */
    PSyncPoint answerFlag;        /* Call is answered */

    H323RegistrationInfoList  registrations; /* Incoming RRQs */

  friend class H323SubGatewayConnection;
};

/** 
 * H323SubGatewayConnection: This is derived from H323Connection.
 * This is an entry point to openh323 
 */

class H323SubGatewayConnection : public H323Connection
{
    PCLASSINFO(H323SubGatewayConnection, H323Connection);

  public:
    H323SubGatewayConnection(H323SubGateway &, unsigned, BOOL);
    ~H323SubGatewayConnection();

    // overrides from H323Connection
    AnswerCallResponse OnAnswerCall(const PString &, const H323SignalPDU &, H323SignalPDU &);

    BOOL H323SubGatewayConnection::OnOpenLogicalChannelAck(
        unsigned sessionId,
        PIPSocket::Address & remoteAddr,WORD dataPort,WORD dataControlPort);

    BOOL OnAlerting(const H323SignalPDU &, const PString &);

	  BOOL GetLogicalChannelAddress(H323Channel::Directions dir,
                                    PIPSocket::Address &ipnum,
                								    WORD &port,
                                    int datatype);
    BOOL PluginCapabilities();

    /*
     * Internal pointer for mapping between API and openh323
     */
    void *call; /* H323Call */

  protected:
    H323SubGateway & myEndpoint;
};

#endif  // _H323SUBGW_MAIN_H


// End of File ///////////////////////////////////////////////////////////////
