/***************************** FILE HEADER *****************************
#               Copyright (c) 2000 by Columbia University              *
#                        All Rights Reserved                           *
#***********************************************************************
#***********************************************************************
#
#  File Name    : buffer.c
#
#  Module       : Audio Conference Server
#
#  Author(s)    : Kundan Singh (kns10@cs.columbia.edu)
#                 Gautam Nair  (gnair@cs.columbia.edu)
#
#  Created      : 28 Apr 2000
#
#  Description  : Buffer management routines for realtime softwares.
#                
#  Notes        : 
#
#***********************************************************************
#                Modification History (Reverse Order)
#-----------------------------------------------------------------------
#   Date         Author              Description
#-----------------------------------------------------------------------
#              |             | 
#              |             | 
#-----------------------------------------------------------------------
*/

#include <stdlib.h>
#include <pthread.h>

#include "error.h"
#include "stdtypes.h"
#include "buffer.h"

/* Buffer management  */

/* List of free buffers */
media_buffer_t *free_list = NULL;
pthread_mutex_t free_list_mutex;
int  free_list_first_time = 1;
int  allocated_buffer_count = 0;
long allocated_buffer_size = 0;

/* 
 * If the size of the requirement is less 
 * or equal to
 * the size of buffer return 1, else 0
 */
int ComparableSize(int requirement, int maxsize)
{
  if ( requirement <= maxsize ) { /* XXX use <= */
    return 1;
  }
  else {
    return 0;
  }
}

/*
 * Allocate a new buffer.
 * Search the free list first,
 * if not found then allocate 
 */
media_buffer_t *NewBuffer(int size)
{
  media_buffer_t *p = NULL;

  if ( size <=0 ) {
    debug("NewBuffer", "Size is %d\n", size);
    return NULL;
  }

  if ( free_list_first_time ) {

    /*
     *  Allocate some number of buffers in the
     * initialization
     */
    free_list_first_time = 0;
    pthread_mutex_init(&free_list_mutex, NULL);
    {
      int i;
      int ms = 1500;
      for (i=0; i<0; i++) { /* XXX */
        p = AllocateObject(media_buffer_t);
        p->data = AllocateObjects(char, ms);
        p->maxsize = ms;
        p->size = 0;
        p->next = free_list;
        free_list = p;
        allocated_buffer_count ++;
        allocated_buffer_size += size;
      }
    }
  }

  /* make sure size is positive */
  if ( size < 1) {
    debug("NewBuffer", "Size is %d, using 1\n", size);
    size = 1;
  }

  LOCK(free_list_mutex);

#ifdef DEBUG
  debug("NewBuffer", "called with size %d, total=%d\n", size, 
        allocated_buffer_size);
#endif

  if ( free_list != NULL ) {
    media_buffer_t *i;

    if ( ComparableSize(size, free_list->maxsize) ) {
      p = free_list;
      free_list = free_list->next;
      p->next = NULL;
      p->size = 0;
      UNLOCK(free_list_mutex);
      return p;
    }
    /* Look into free list */
    for ( i=free_list; i->next; ) {
      p = i->next;
      if ( ComparableSize(size, p->maxsize) ) {
        i->next = i->next->next;
        p->next = NULL;
        p->size = 0;
        UNLOCK(free_list_mutex);
        return p;
      }
      else {
        i= i->next;
      }
    }
  }
#ifdef DEBUG
  debug("NewBuffer", "allocating new\n");
#endif

  /* Allocate a new element */
  p = AllocateObject(media_buffer_t);
  if ( p == NULL ) {
    debug("NewBuffer", "calloc Failed media_buffer_t %d\n", size);
    UNLOCK(free_list_mutex);
    return NULL;
  }
  p->data = AllocateObjects(char,size);
  if ( p->data == NULL ) {
    debug("NewBuffer", "calloc Failed for size %d\n", size);
    UNLOCK(free_list_mutex);
    return NULL;
  }
  p->maxsize = size;
  p->size = 0;
  p->next = NULL;

  allocated_buffer_count ++;
  allocated_buffer_size += size;

  UNLOCK(free_list_mutex);
  return p;
}

/*
 * Put a buffer back to free list
 */
void FreeBuffer(media_buffer_t *p)
{
  if ( p == NULL ) {
    debug("FreeBuffer", "called with NULL\n");
  }

  if ( free_list_first_time ) {
    free_list_first_time = 0;
    pthread_mutex_init(&free_list_mutex, NULL);
  }

  LOCK(free_list_mutex);

#ifdef DEBUG
  debug("FreeBuffer", "called with size %d, %d, total=%d\n",
        p->size,p->maxsize, allocated_buffer_size);
#endif

  if ( allocated_buffer_count > 100 ) {
    /* Free all buffers if more than 100 are already allocated */
    debug("FreeBuffer", "Freeing all in free list\n");

    allocated_buffer_count--;
    allocated_buffer_size -= p->maxsize;
    DeallocateObject(p->data);
    DeallocateObject(p);

    /* Free all from free_buffer_list */
    for ( p=free_list; p; ) {
      media_buffer_t *i = p->next;
      DeallocateObject(p->data);
      DeallocateObject(p);
      p = i;
    }
    free_list = NULL;
    allocated_buffer_count = 0;
    allocated_buffer_size = 0;

    UNLOCK(free_list_mutex);
    return;
  }

  p->size = 0;
  p->next = free_list;
  free_list = p;
  UNLOCK(free_list_mutex);
  return;
}

/*
 * Enqueue a buffer to a list
 */
void BufferEnqueue(media_buffer_t **q, media_buffer_t *buf)
{
  media_buffer_t *i;

  buf->next = NULL;

  if ( *q == NULL ) {
    *q = buf;
    return;
  }

  /* Enqueue based on the playout pointer, buf->param XXX */
  if ( buf->param < (*q)->param) {
    buf->next = (*q);
    return;
  }
  else {
    for ( i= *q; i->next; ) {
      if ( buf->param < i->next->param) {
        buf->next = i->next;
        i->next = buf;
        return;
      }
      else {
        i = i->next;
      }
    }
    buf->next = NULL;
    i->next = buf;
  }
  return;
}

/*
 * Delete the queue 
 */
void DeleteQueue(media_buffer_t **q)
{
  media_buffer_t *p;
  media_buffer_t *n;
  for ( p = *q; p != NULL; ) {
    n = p->next;
    p->oRemove = 0;
    p->next = NULL;
    FreeBuffer(p);
    p = n;
  }
  *q = NULL;

  return;
}


/************************** END OF FILE ********************************
*/

