/***************************** FILE HEADER *****************************
#               Copyright (c) 2000 by Columbia University              *
#                        All Rights Reserved                           *
#***********************************************************************
#***********************************************************************
#
#  File Name    : data.c
#
#  Module       : Audio/video Conference Server
#
#  Author(s)    : Kundan Singh (kns10@cs.columbia.edu)
#
#  Description  : Data switching logic.
#                
#  Notes        : 
#
#***********************************************************************
#                Modification History (Reverse Order)
#-----------------------------------------------------------------------
#   Date         Author              Description
#-----------------------------------------------------------------------
#              |             | 
#              |             | 
#-----------------------------------------------------------------------
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "error.h"
#include "timespec.h"
#include "list.h"
#include "sdp.h"
#include "stdtypes.h"
#include "rtpmapi.h"

#ifndef WIN32
#define closesocket close
#endif

extern list_t glConference; /* list of conferences */
extern list_t glUsers;      /* List of users */
extern struct in_addr host2ip(char *);

/*
 * Start listen port tries to allocate the consecutive port numbers
 * for RTP and RTCP and bind the appropriate sockets.
 * If the port is specified, it is preferred for
 * RTP port. All port numbers are in host byte order.
 * returns the allocated RTP port number. RTCP is RTP+1.
 * returns 0 in case of error. The *rtp_sock and *rtcp_sock
 * hold the socket ID of the RTP and RTCP sockets.
 * Set port to 0 if no preferred port.
 */
static unsigned short start_listen_port(unsigned short port,
                                 int * data_sock)
{
  struct sockaddr_in sin;
  int sinlen;
  int max_retry = 10;

  if ( data_sock == NULL  ) {
    debug("start_listen_port", "One of the params are NULL\n");
    return 0;
  }

  /* While max retry */
  while ( --max_retry >= 0 ) {    

    /* Allocate the RTP and RTCP sockets */
    if (((*data_sock) = socket(AF_INET, SOCK_DGRAM, 0)) < 0 ) {
      debug("start_listen_port", "data_socket failed\n");
      return 0;
    }    
    
    /* Bind the RTP socket to the port */
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = INADDR_ANY;
    sin.sin_port = htons(port);
    if ( (bind(*data_sock, (struct sockaddr *)&sin, sizeof(sin))) < 0) {
      if ( port == 0 ) {
        /* If port was already 0 (any),
         * some serious error */
        closesocket(*data_sock);
        return 0;
      }
      else {
        /* We tried an non-0 port.
         * Try a 0 port now. */
        port = 0;
        debug("start_listen_port", "Requested port not bound. Trying any.\n");
        continue;
      }
    }

    /* Extract the port number now. If it were 0. */
    if ( port == 0 ) {
      sinlen = sizeof(sin);
      if ( getsockname(*data_sock, (struct sockaddr *)&sin, &sinlen) < 0 ) {
        debug("start_listen_port", "getsockname error\n");
      }
      port = ntohs(sin.sin_port);
    }
    else {
      /* An non-0 value was specified for the 
       * port. Continue with rtcp binding */
    }

    /* Everytthing is OK till now. */
    return port;
  }

  debug("start_listen_port", "Max retry exhausted\n");
  return 0;
}


/*
 * Create the data socket for RTP and RTCP. Set appropriate values in
 * the call leg structure.
 * return 0 for success and < 0 for failure.
 */
int RTPMDataCreateSocket(rtpm_call_leg_t *leg,
                          media_line_t   *media)
{
  media_line_t media_rx;
  int namelen;
  int ipx;
  char hostname[200];
  unsigned short port;

  debug("RTPMCreateDataSocket", "called\n");

  port = start_listen_port(ntohs(media->sin.sin_port),
                           &leg->data_sock);

  if ( port == 0 ) {
    leg->data_sock = -1;
    return -1;
  }

  if ( getsockname(leg->data_sock,
                   (struct sockaddr *) &leg->data_local_sin,
                   &namelen) < 0 ) {
    error("RTPMCreateDataSocket", "getsockname failed\n");
    closesocket(leg->data_sock);
    leg->data_sock = -1;
    return -4; /* getsockname failed */
  }

  /* Set the media_rx */
  media_rx.type = media->type;
  media_rx.ttl  = media->ttl; /* if needed */

  ListCreate(&media_rx.payloads, 1, 1);

  for ( ipx = 0; ipx < ListCount(&media->payloads); ipx++) {
    payload_type_t *px = ListElement(&media->payloads, ipx);
    payload_type_t pr;
    if ( !px )
      continue;
    pr.pt = px->pt;
    pr.description = px->description ? strdup(px->description) :
      NULL;
    ListAddElement(&media_rx.payloads, &pr, sizeof(pr));
  }
  media_rx.sin.sin_port = leg->data_local_sin.sin_port; /* Do not call htons*/
  media_rx.sin.sin_family = AF_INET;
  gethostname(hostname,sizeof(hostname));
  media_rx.sin.sin_addr = host2ip(hostname);
  ListAddElement(&leg->sdp_rx.media_line_t, 
                 &media_rx, sizeof(media_rx));

  /* Set the call leg parameters */
  memcpy(&leg->data_remote_sin, &media->sin, sizeof(media->sin));
  leg->data_remote_sin.sin_family = AF_INET;

  return 0;
}

/*
 * Delete the data socket.
 */
int RTPMDataDeleteSocket(rtpm_call_leg_t *leg)
{
  close(leg->data_sock);
  memset(&leg->data_local_sin, 0, sizeof(leg->data_local_sin));
  memset(&leg->data_remote_sin, 0, sizeof(leg->data_remote_sin));
  return 0;
}

/*
 * Returns 1 if the sender can send the particular RTP packet 
 * to the receiver. Else returns 0.
 * This matches the payloadtype with the allowed SDP payload types
 * of the receiever. The buffer is updated to reflect the new
 * payload type of the receiver.
 */
int RTPMDataCanSend(rtpm_call_leg_t *sender,
                    rtpm_call_leg_t *receiver,
                    char *buffer,
                    int length)
{
  /* Always send */
  return 1;
}

/*
 * Thread for processing incoming data RTP/RTCP packets
 */
void * RTPMDataThread(void *param)
{
  int context = (int) param;
  rtpm_call_leg_t *leg;
  struct timeval initialtime, socktv;
  struct timespec next;
  fd_set selection;
  int maxfd = 0;
  int sock_rtp;
  char rtp_buffer[1500]; int rtp_length;
  int  packet_received; /* 1 for RTP, 2 for RTCP, 3 for both */
  struct sockaddr_in remote_sin;
  int sin_len;
  rtpm_conf_t *conf;
  int k;
  rtpm_call_leg_t *user;
  int *userid;
  int pt;

  debug("DataThread", "Entering\n");

  leg = ListElement(&glUsers, context);
  if ( leg == NULL ) {
    error("RTPMDataThread", "No user with context %d\n", context);
    return NULL;
  }

  LOCK(leg->mutex);

  sock_rtp = leg->data_sock;
  maxfd = sock_rtp;

  /* Loop while not terminated */
  gettimeofday(&initialtime, NULL);
  while (!leg->oTerminate) {
    next = timespec_delay(initialtime, 20/1000.0 ) ; /* 10 millisec */
    COND_TIMEDWAIT(leg->cond, leg->mutex, next);
    if ( leg->oTerminate ) {
      break; /* from while loop */
    }

    gettimeofday(&initialtime, NULL);

    FD_ZERO(&selection);
    FD_SET(sock_rtp, &selection);

    socktv.tv_sec = 0;
    socktv.tv_usec = 0; /* 0 millisec */

    /* Recv UDP packets from RTP/RTCP socket */
    while ( select(maxfd + 1, &selection, NULL, NULL, &socktv) > 0 ) {
      packet_received = 0;
      if ( FD_ISSET (sock_rtp, &selection)) {
        rtp_length = recvfrom(sock_rtp, rtp_buffer, sizeof(rtp_buffer), 0,
                 (struct sockaddr *) &remote_sin, &sin_len);
        if ( rtp_length > 0 ) {

          packet_received = 1;

          debug("RTPMDataThread", "Received %d bytes of data\n",
                rtp_length);

        }
      }

      FD_ZERO(&selection);
      FD_SET(sock_rtp, &selection);

      socktv.tv_sec = 0;
      socktv.tv_usec = 0; /* 0 millisec */

      if ( packet_received != 0 ) {
        /* Release the mutex */
        UNLOCK(leg->mutex);

        /* Get the conference mutex. */
        conf = ListElement(&glConference, leg->confid);

        if ( conf ) {
          LOCK(conf->mutex);
        }

        /* Get the call leg mutex again */
        leg = ListElement(&glUsers, context);
        if ( leg == NULL ) {
          error("RTPMDataThread", "No user with context %d, exiting\n", 
                context);
          if ( conf ) {
            UNLOCK(conf->mutex);
          }
          return NULL;
        }
        LOCK(leg->mutex);
        
        /* Now we have both leg and conf mutexes */
        if ( conf ) {
          /* Iterate through all the participants list */

          for ( k=0; k<ListCount(&conf->participants); k++) {
            userid = ListElement(&conf->participants, k);
            if ( !userid ) {
              //debug("RTPMDataThread", "returned NULL from participants\n");
              continue;
            }
            user   = ListElement(&glUsers, *userid); /* participant */
            if ( !user ) {
              debug("RTPMDataThread", "returned NULL from user list\n");
              continue;
            }
            
            if ( *userid == context || user == leg) {
              /* Same user */
              continue;
            }

            /*
            debug("RTPMDataThread", "Participant %d of %d\n", *userid,
                  ListCount(&conf->participants));
            */

            LOCK(user->mutex);

            if ( user->data_sock >= 0 ) {
              //debug("RTPMDataThread", "This has data\n");

              if ( (packet_received != 0) && 
                   RTPMDataCanSend(leg,user,rtp_buffer,rtp_length)) {
                remote_sin = user->data_remote_sin;

                
                rtp_length = sendto(user->data_sock,
                       rtp_buffer, rtp_length, 0,
                       (struct sockaddr *)&remote_sin,
                       sizeof(remote_sin));
                
                
                debug("RTPMDataThread", "tx data size %d to %s:%d\n",
                      rtp_length,
                      inet_ntoa(remote_sin.sin_addr), 
                      ntohs(remote_sin.sin_port));
                
              }
            }

            UNLOCK(user->mutex);
          } /* Iterate through all participants */

          UNLOCK(conf->mutex);
        }
      }
    }
  }

  UNLOCK(leg->mutex);

  debug("DataThread", "Exiting\n");

  return NULL;
}

/************************** END OF FILE ********************************
*/

