/***************************** FILE HEADER *****************************
#               Copyright (c) 2000 by Columbia University              *
#                        All Rights Reserved                           *
#***********************************************************************
#***********************************************************************
#
#  File Name    : rtpmapi.c
#
#  Module       : Audio Conference Server
#
#  Author(s)    : Kundan Singh (kns10@cs.columbia.edu)
#                 Gautam Nair  (gnair@cs.columbia.edu)
#
#  Created      : 25 Apr 2000
#
#  Description  : RTP Mixer interface for audio mixer.
#                
#  Notes        : 
#
#***********************************************************************
#                Modification History (Reverse Order)
#-----------------------------------------------------------------------
#   Date         Author              Description
#-----------------------------------------------------------------------
#              |             | 
#              |             | 
#-----------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "error.h"
#include "timespec.h"
#include "list.h"

#ifdef USE_RTP
#include "rtp_api.h"    /* RTP context */
#include "rtp_highlevel.h"
#else
#error "Can not build this module without RTP. configure --with-rtp option"
#endif

#include "rtpmapi.h"
#include "stdtypes.h"
#include "sdp.h"
#include "sched.h"
#include "sendrecv.h"
#include "transcode.h"
#include "video.h"
#include "data.h"

/* Control macros */
#undef FIXED_PORTNO
//#define FIXED_PORTNO

/* Global variables */

list_t glConference; /* list of conferences */
list_t glUsers;      /* List of users */
int oInitialized = 0;/* Whether initialized or not */

#ifdef INCLUDE_SSRC
/* Mutex for RTP CReate and Destroy */
pthread_mutex_t rtp_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif

extern struct in_addr host2ip(char *);
extern unsigned long random32(void);


/* 
 * RTPMAddCallLeg is used to create a RTP Mixer call leg
 * for particular user. This is called
 * when a new SIP call arrives.
 */
int RTPMAddCallLeg(const char *confname, const char *source)
{
  rtpm_call_leg_t leg, *legp;
  rtpm_conf_t     *conf = NULL;
  int confidx;
  int useridx;
  int oFound;
  rtpm_call_leg_t *user;
  int pid;

  debug("RTPMAddCallLeg","Conf=%s,user=%s\n", 
        confname ? confname : "<NULL>", 
        source ? source : "<NULL>");

  /* Do one time initializations, create mutexes */
  if ( !oInitialized ) {
    pthread_t rtp_schedule_thread;

    oInitialized = 1;
    InitSched();
    ListCreate(&glConference, 0, 1);
    ListCreate(&glUsers, 5, 1);

    if ( pthread_create_detached(&rtp_schedule_thread, NULL, 
                        RTPScheduleThread, 
                        NULL) != 0 ) {
      error("RTPMAddCallLeg",
            "pthread_create_detached for RTPMScheduleThread failed");
      return -1;
    }

  }

  /* Add the new participant in the list of users */
  memset(&leg, 0, sizeof(leg));
  useridx = ListAddElement(&glUsers, &leg, sizeof(leg));
  legp = ListElement(&glUsers, useridx);
  /* Initialize other fields */
  legp->participant = strdup(source);
  SDPInit(&legp->sdp_tx);
  SDPInit(&legp->sdp_rx);
  legp->rtp_context = -1;
  pthread_mutex_init(&legp->mutex, NULL);
  pthread_cond_init(&legp->cond, NULL);
  legp->rxbuflist = NULL;
  legp->delay.first_time = 1;
  legp->oTerminate = 0;
  legp->tsinc = 0; 
  legp->ssrc = 0; /* Initialize to 0 */
#ifdef INCLUDE_SSRC
  legp->name = strdup(source);
#endif
  legp->video_sock = -1;
  legp->video_sock_rtcp = -1;
  memset(&legp->video_local_sin, 0, sizeof(legp->video_local_sin));
  memset(&legp->video_remote_sin, 0, sizeof(legp->video_remote_sin));
  legp->data_sock = -1;
  memset(&legp->data_local_sin, 0, sizeof(legp->data_local_sin));
  memset(&legp->data_remote_sin, 0, sizeof(legp->data_remote_sin));

  LOCK(legp->mutex);

  /* Create the thread for the session to listen to packet */
  /* Note that the leg.rx_thread id is not stored */
  if ( pthread_create(&legp->rx_thread, NULL, 
                      RTPReceiveThread, 
                      (void *) useridx) != 0 ) {
    error("RTPMAddCallLeg",
          "pthread_create for RTPMReceiveThread failed");
    UNLOCK(legp->mutex);
    return -1;
  }

  /* Iterate through all the existing conferences */
  oFound = 0;
  for (confidx=0; confidx < ListCount(&glConference); confidx ++ ) {
    conf = ListElement(&glConference, confidx);
    if ( !conf ) 
      continue;
    /* Check if the conference exists */
    if ( !strcmp(conf->conference, confname) ) {
      UNLOCK(legp->mutex);
      LOCK(conf->mutex);
      LOCK(legp->mutex);
      oFound = 1;
      break;
    }
  }

  if ( !oFound ) {

    /* Create the conference element, since this is first participant */
    conf = AllocateObject(rtpm_conf_t);
    confidx = ListAddElement(&glConference, conf,
                             sizeof(rtpm_conf_t));
    DeallocateObject(conf);

    conf = ListElement(&glConference, confidx);

    pthread_mutex_init(&conf->mutex, NULL);
    pthread_cond_init(&conf->cond, NULL);
    UNLOCK(legp->mutex);
    LOCK(conf->mutex);
    LOCK(legp->mutex);

    conf->oTerminate = 0;

    if ( ListCreate(&conf->participants, 5, 1) != 0 ) {
      debug("RTPMAddCallLeg", "ListCreate for participants Failed\n");
    }

    conf->conference = strdup(confname);
    conf->play_ptr = 0;
#ifdef INCLUDE_SSRC
    ListCreate(&conf->user_ssrcs, 3, 1);
#endif
    /* XXX note that the tx_thread id is not stored */
    if ( pthread_create(&conf->tx_thread, NULL,
                        RTPSendThread, (void *)confidx) != 0 ) {
      error("RTPMAddCallLeg", 
            "pthread_create failed for RTPSendThread\n");
    }

  }
  else {
    debug("RTPMAddCallLeg", "Conference already exists\n");
  }

  /* Existing conference, new participant */
  if ( (pid = ListAddElement(&conf->participants, &useridx,
                             sizeof(useridx))) < 0 ) {
    debug("RTPMAddCallLeg", "ListAddElement to participants "
          "list of conf failed\n");
  }
  debug("RTPMAddCallLeg", "participants count %d, %d, %d\n",
        ListSize(&conf->participants), pid, useridx);
  
  user = ListElement(&glUsers, useridx);
  if ( user != NULL ) {
    user->confid = confidx;
  }
  else {
    debug("RTPAddCallLeg", "ListElement() failed for %d\n", useridx);
  }

  UNLOCK(legp->mutex);
  UNLOCK(conf->mutex);

  return useridx;
}

/*
 * RTPMRemoveCallLeg: is called when a person
 * leaves the conference. The context is freed
 * If this is the last member then the
 * conference context is also removed.
 */
int RTPMRemoveCallLeg(int context)
{
  /* XXX remove the callleg, and if needed the conference */
  rtpm_call_leg_t *user;
  rtpm_conf_t *conf;
  void *value = NULL;
  media_buffer_t *ptr;
  int j;

  user = ListElement(&glUsers, context);
  if ( user == NULL ) {
    debug("RTPMRemoveCallLeg", "No call leg %d\n", context);
    return -1;
  }

  LOCK(user->mutex);
  user->oTerminate = 1; /* terminate the thread */
  COND_SIGNAL(user->cond);
  UNLOCK(user->mutex);

  debug("RTPMRemoveCallLeg", "Waiting for rx thread\n");
  pthread_join(user->rx_thread, &value);
  debug("RTPMRemoveCallLeg", "Rx thread terminated\n");

  if ( user->video_sock >= 0 ) {
    /* Not good to test video_sock while waiting for
     * video_thread. */
    debug("RTPMRemoveCallLeg", "Waiting for video thread\n");
    pthread_join(user->video_thread, &value);
    debug("RTPMRemoveCallLeg", "Video thread terminated\n");
  }
  if ( user->data_sock >= 0 ) {
    /* Not good to test video_sock while waiting for
     * video_thread. */
    debug("RTPMRemoveCallLeg", "Waiting for data thread\n");
    pthread_join(user->data_thread, &value);
    debug("RTPMRemoveCallLeg", "Data thread terminated\n");
  }

  LOCK(user->mutex);

  if ( user->video_sock >= 0 ) {
    RTPMVideoDeleteSocket(user);
  }
  if ( user->data_sock >= 0 ) {
    RTPMDataDeleteSocket(user);
  }

  if ( user->rtp_context != -1 ) {
#ifdef INCLUDE_SSRC
    LOCK(rtp_mutex);
#endif
    debug("RTPMRemoveCallLeg", "RTPDestroy for %d\n", 
          user->rtp_context);
    RTPDestroy(user->rtp_context);
#ifdef INCLUDE_SSRC
    UNLOCK(rtp_mutex);
#endif
  }

  /* Get the conference context */  
  conf = ListElement(&glConference, user->confid);
  if ( conf == NULL ) {
    debug("RTPMRemoveCallLeg", "User %d does not belong to any conf\n",
          context);
  }
  else {
    int i, oFound = 0;

    UNLOCK(user->mutex);
    LOCK(conf->mutex);
    LOCK(user->mutex);

    /* Remove this from the list of participants */
    for ( i = 0; i < ListCount(&conf->participants); i++) {
      int *userid = ListElement(&conf->participants, i);
      if ( !userid ) {
        continue;
      }
      if ( *userid == context ) {
        oFound = 1;
        break;
      }
    }
    if ( oFound ) {
      debug("RTPMRemoveCallLeg", "Removing %d participant from conf %d\n",
            i, user->confid);
      /* Delete from the list of participants */
      ListDeleteElement(&conf->participants, i);
    }
    else {
      debug("RTPMRemoveCallLeg", "User %d does not belong to conf %d\n",
            context, user->confid);
    }

    /* Delete from conf->user_ssrcs */
    if ( user->rtp_context != -1 ) {
      int u;
      int oFound = 0;
      rtpm_ssrc_info_t *p;

      for ( u=0; u<ListCount(&conf->user_ssrcs); u++) {
        p =
          ListElement(&conf->user_ssrcs, u);
        if ( !p) {
          continue;
        }
        if ( p->cid == user->rtp_context ) {
          oFound = 1;
          break;
        }
      }
      DeallocateObject(p->name);
      ListDeleteElement(&conf->user_ssrcs, u);
    }

#ifdef INCLUDE_SSRC
    for ( j=0; (user->rtp_context != -1) &&
            (j<ListCount(&conf->user_ssrcs)); j++) {
      rtpm_ssrc_info_t *ssrc;
      rtperror the_err = RTP_OK;

      ssrc = ListElement(&conf->user_ssrcs, j);
      if ( !ssrc ) {
        continue;
      }
      if ( (ssrc->ssrc == user->ssrc) ||
           (ssrc->cid < 0)) {
        /* same source */
        debug("RTPSetSelfSession", 
              "SSRC %u == %u, skipping\n",
              ssrc->ssrc, user->ssrc);
        continue;
      }

      /* Remove from the contributors list */
      the_err = 
        RTPSessionRemoveFromContributorList(ssrc->cid,
                                            user->ssrc,
                                            NULL);

      if ( the_err != RTP_OK ) {
        debug("RTPMRemoveCallLeg", 
              "Removing contributor (cid%d,ssrc%u) Error:%s\n",
              ssrc->cid, user->ssrc, RTPDebugStr());
      }
      else {
        debug("RTPSessionRemoveContributorList",
              "Removing SSRC %u as a contributor for context %d\n",
              ssrc->ssrc, user->rtp_context);
      }
    }
#endif

    if ( ListSize(&conf->participants) <= 0) {
      /* Last member left, remove the conference */
      conf->oTerminate = 1;
      COND_SIGNAL(conf->cond);
      UNLOCK(conf->mutex);

      debug("RTPMRemoveCallLeg", "Waiting for tx thread\n");
      pthread_join(conf->tx_thread, &value);
      debug("RTPMRemoveCallLeg", "Tx thread terminated\n");

      LOCK(conf->mutex);
      /* Free all the structure members */
      ListDelete(&conf->participants);

#ifdef INCLUDE_SSRC
      {
        int u;
        for ( u=0; u<ListCount(&conf->user_ssrcs); u++) {
          rtpm_ssrc_info_t *p =
            ListElement(&conf->user_ssrcs, u);
          if ( !p) {
            continue;
          }
          DeallocateObject(p->name);
        }
        ListDelete(&conf->user_ssrcs);
      }
#endif

      DeallocateObject(conf->conference);

      UNLOCK(conf->mutex);
      pthread_mutex_destroy(&conf->mutex);
      pthread_cond_destroy(&conf->cond);

      ListDeleteElement(&glConference, user->confid /* XXX is this
                                                     correct */);

    }
    else {
      debug("RTMPRemoveCallLeg",
            "Participants count is %d\n", ListSize(&conf->participants));
      UNLOCK(conf->mutex);
    }
  }

  /* Free the context */
  DeallocateObject(user->participant);
  SDPFree(&user->sdp_tx);
  SDPFree(&user->sdp_rx);

  /* Free unused buffers */
  for ( ptr = user->rxbuflist; ptr; ) {
    media_buffer_t *temp = ptr->next;
    FreeBuffer(ptr);
    ptr = temp;
  }
#ifdef INCLUDE_SSRC
  DeallocateObject(user->name);
#endif

  UNLOCK(user->mutex);
  pthread_mutex_destroy(&user->mutex);
  pthread_cond_destroy(&user->cond);

  ListDeleteElement(&glUsers, context);

  return 0;
}

/*
 * This function is called to set the SDP for the
 * person.
 */
int RTPMSetSelfSessionDescription(int context, 
                                  session_description_t s)
{
  rtperror the_err;
  rtpm_call_leg_t *leg;
  int i;
  media_line_t *media;
  int audio_added = 0;

  debug("RTPMSetSelfSessionDescription", "called with %d\n", context);

  leg = ListElement(&glUsers,context);
  if ( leg == NULL ) {
    debug("RTPMSetSelfSessionDescription", "No suchcall leg %d\n",
          context);
    return -1;
  }

  LOCK(leg->mutex);

  /* Copy the SDP to both Tx and Rx */
  SDPFree(&leg->sdp_tx);
  SDPCopy(&leg->sdp_tx, &s);

  SDPFree(&leg->sdp_rx);
  SDPInit(&leg->sdp_rx);
  //SDPCopy(&leg->sdp_rx, &leg->sdp_tx);

  debug("RTPMSetSelfSessionDescription", "iterate through all media\n");

  /* Check if already set XXX */

  /*
   * Scan through all the m= linein SDP. Only one is
   * supported right now.
   */
  for (i=0; i<ListCount(&s.media_line_t); i++) {
    media = ListElement(&s.media_line_t, i);
    if ( !media ) {
      continue;
    }
    debug("RTPMSetSelfSessionDescription", "media index %d\n", i);

    if ( media == NULL ) {
      debug("RTPMSetSelfSessionDescription", "No media for %d\n", i);
      UNLOCK(leg->mutex);
      return -1;
    }

    if ( media->type != RTPM_MEDIA_AUDIO ) {
      debug("RTPMSetSelfSessionDescription", "Non audio mode\n");
      if ( media->type == RTPM_MEDIA_VIDEO ) {
        /* Create a new thread for processing the
         * video data */
        /*
         * Allocate the socket for the video RTP/RTCP data.
         */
        if ( RTPMVideoCreateSocket(leg, media) == 0 ) {
          /* The socket is created with the
           * appropriate parameters set. */
          if ( pthread_create(&leg->video_thread, NULL, 
                              RTPMVideoThread, 
                              (void *) context) != 0 ) {
            error("RTPMSetSelfSession",
                  "pthread_create failed\n");
          }
          else {
            /* Thread is created */
            debug("RTPMSetSelfSession", "Video thread created\n");
          }
        }
        else {
          error("RTPMSetSelfSession", "RTPVideoCreateSocket failed\n");
        }
      }
      else if ( media->type == RTPM_MEDIA_DATA ) {
        /* Create a new thread for processing the
         * application data */
        /*
         * Allocate the socket for the data RTP/RTCP data.
         */
        if ( RTPMDataCreateSocket(leg, media) == 0 ) {
          /* The socket is created with the
           * appropriate parameters set. */
          if ( pthread_create(&leg->data_thread, NULL, 
                              RTPMDataThread, 
                              (void *) context) != 0 ) {
            error("RTPMSetSelfSession",
                  "pthread_create failed for data\n");
          }
          else {
            /* Thread is created */
            debug("RTPMSetSelfSession", "Data thread created\n");
          }
        }
        else {
          error("RTPMSetSelfSession", "RTPMDataCreateSocket failed\n");
        }
      }
      else {
        debug("RTPMSetSelfSession", "Type is %d\n", media->type);
      }

      /* Continue with the next media */
      continue;
    }
    else {
      debug("RTPMSetSelfSession", "Audio mode\n");
      if ( audio_added ) {
        debug("RTPMSetSelfSession", "Audio is already added\n");
        continue;
      }
    }

    audio_added = 1;

#ifdef INCLUDE_SSRC
    LOCK(rtp_mutex);
#endif

    /* Create the RTP cpntext */
    debug("RTPCreate", "called\n");
    the_err = RTPCreate(&leg->rtp_context);
    debug("RTPCreate", "Created context %d for leg %d\n",
          (int) (leg->rtp_context), context);

#ifdef INCLUDE_SSRC
    UNLOCK(rtp_mutex);
#endif

    if ( the_err != RTP_OK ) {
      print_rtperror("RTPMSetSelfSessionDescription", the_err);
      UNLOCK(leg->mutex);
      return -1;
    }

    /* Set the SSRC */
    leg->ssrc = random32();

    the_err = RTPMemberInfoSetSSRC(leg->rtp_context, 0, leg->ssrc);
    if ( the_err != RTP_OK ) {
      print_rtperror("RTPMemberInfoSetSSRC", the_err);
      UNLOCK(leg->mutex);
      return -1;
      /* No need to terminate */
    }
    else {
      debug("RTPMemberInfoSetSSRC", "set to %u\n", leg->ssrc);
    }
      
    debug("RTPSessionSetReceiveAddr", "called\n");
    {
#ifdef FIXED_PORTNO
      static unsigned short port = 8000;
      the_err = RTPSessionSetReceiveAddr(leg->rtp_context,
                                         NULL, port);

      port += 2;
#else
      the_err = RTPSessionSetReceiveAddr(leg->rtp_context,
                                         NULL, 0 );
#endif
    }

    if ( the_err != RTP_OK ) {
      print_rtperror("RTPMSetSelfSessionDescription", the_err);
      UNLOCK(leg->mutex);
      return -1;
    } 

    debug("RTPSessionAddSendAddr", "called with %s:%d\n",
          inet_ntoa(media->sin.sin_addr),ntohs(media->sin.sin_port));
    the_err = RTPSessionAddSendAddr(leg->rtp_context,
                                    inet_ntoa(media->sin.sin_addr),
                                    ntohs(media->sin.sin_port),
                                    media->ttl /* ignored for unicast */
                                    );

    if ( the_err != RTP_OK ) {
      print_rtperror("RTPMSetSelfSessionDescription", the_err);
      UNLOCK(leg->mutex);
      return -1;
    }

    debug("RTPOpenConnection", "called\n");
    the_err = RTPOpenConnection(leg->rtp_context);

    if ( the_err != RTP_OK ) {
      print_rtperror("RTPMSetSelfSessionDescription", the_err);
      UNLOCK(leg->mutex);
      return -1;
    }

    {
      media_line_t media_rx;
      int ipx;
      char address[100];
      char hostname[200];
      u_int16 port;
      debug("RTPSessionGetReceiveAddr", "called\n");
      /* This must be called after RTPOpenConnection */
      the_err = RTPSessionGetReceiveAddr(leg->rtp_context,
                                       address,
                                       &port);
      debug("RTPMSetSelfSessionDescription",
            "Listening at %s:%u\n",
            address, port);

      media_rx.type = media->type;
      media_rx.ttl  = media->ttl;
      ListCreate(&media_rx.payloads, 1, 1);
      for ( ipx = 0; ipx < ListCount(&media->payloads); ipx++) {
        payload_type_t *px = ListElement(&media->payloads, ipx);
        payload_type_t pr;
        if ( !px )
          continue;
        debug("RTPMSetSelfSessionDdescription", "Adding %d\n", px->pt);
        pr.pt = px->pt;
        pr.description = px->description ? strdup(px->description) :
          NULL;
        ListAddElement(&media_rx.payloads, &pr, sizeof(pr));
      }
      media_rx.sin.sin_port = htons(port);
      media_rx.sin.sin_family = AF_INET;
      gethostname(hostname,sizeof(hostname));
      media_rx.sin.sin_addr = host2ip(hostname);
      ListAddElement(&leg->sdp_rx.media_line_t, 
                     &media_rx, sizeof(media_rx));
    }
  }

  debug("RTPMSetSelfSessionDescription", "returning\n");

  COND_SIGNAL(leg->cond);

  UNLOCK(leg->mutex);

  /* Update the conference structure */

#ifdef INCLUDE_SSRC
  {
    rtpm_conf_t *conf;
    int j;
    rtpm_ssrc_info_t ssrcinfo;

    conf = ListElement(&glConference, leg->confid);
    if ( conf ) {
      char confname[257]; /* Max is 256 */
      LOCK(conf->mutex);
      LOCK(leg->mutex);

      ssrcinfo.cid = leg->rtp_context;
      ssrcinfo.ssrc = leg->ssrc;
      ssrcinfo.name = leg->name ? strdup(leg->name) : NULL;
      
      if ( conf->conference ) {
        strcpy(confname, "conference:");
        strncpy(&confname[strlen(confname)], conf->conference,
                sizeof(confname) - strlen(confname) - 2);
        confname[sizeof(confname) - 1] = '\0'; 

      }
      else {
        strcpy(confname, "conference: unknown");
      }

      /* Add the SDES */
      if ( leg->name && leg->rtp_context >= 0) {
        the_err = 
          RTPMemberInfoSetSDES(leg->rtp_context, 0, RTP_MI_NAME, 
                               confname);
        if ( the_err != RTP_OK ) {
          print_rtperror("RTPMemberInfoSetSDES", the_err);
        }
      }
      else {
        debug("RTPMSetSelfSession", "No user name\n");
      }

      if ( ssrcinfo.ssrc !=0 &&
           ListAddElement(&conf->user_ssrcs,
                          &ssrcinfo, 
                          sizeof(ssrcinfo)) < 0 ) {

        debug("RTPMSetSelfSession",
              "ListAddElement failed %d\n", __LINE__);
      }
      else {
        debug("RTPMSetSession", "Added SSRC %u\n", ssrcinfo.ssrc);
      }

      for ( j=0; leg->rtp_context >= 0 &&
              j<ListCount(&conf->user_ssrcs); j++) {
        rtpm_ssrc_info_t *ssrc;
        person p_person = 0, i_person;
        member_iterator iter;

        ssrc = ListElement(&conf->user_ssrcs, j);
        if ( !ssrc ) {
          continue;
        }
        if ( ssrc->ssrc == leg->ssrc ) {
          /* same source */
          debug("RTPSetSelfSession", 
                "SSRC %u == %u, skipping\n",
                ssrc->ssrc, leg->ssrc);
          continue;
        }

        /* Add to contributors list */

        the_err = 
          RTPSessionAddToContributorList(leg->rtp_context,
                                         ssrc->ssrc);
        if ( the_err != RTP_OK ) {
          debug("RTPSetSelfSession",  "%s\n", RTPDebugStr());
        }
        else {
          debug("RTPSessionAddToContributorList",
                "Adding SSRC %u as a contributor for context %d\n",
                ssrc->ssrc, leg->rtp_context);
        }


        the_err = RTPSessionGetMemberList (leg->rtp_context,
                                           &iter);
        if (the_err != RTP_OK) {
          debug ("RTPSetSelfDescription", 
                 "RTPSessionGetMemberList Error\n");
        }

        /*
        the_err = RTPCurrentMember (leg->rtp_context,
                                    &iter,
                                    &i_person);
        if (the_err != RTP_OK) {
          debug ("RTPSetSelfDescription", 
                 "RTPSessionGetMemberList Error\n");
        }
        */

        while ( 1 ) {
          unsigned int ssrc2;

          the_err = RTPNextMember(leg->rtp_context,
                                  &iter,
                                  &i_person);
          if ( the_err != RTP_OK ) {
            break;
          }

          the_err = RTPMemberInfoGetSSRC(leg->rtp_context, i_person,
                               (u_int32 *) &ssrc2);
          if ( the_err != RTP_OK ) {
            debug("RTPSetSelfSession", "%s\n", RTPDebugStr());
          }

          if ( ssrc2 == ssrc->ssrc ) {
            p_person = i_person;
            break;
          }

        }

        if ( p_person != 0 ) {
          if ( ssrc->name ) {
            the_err = RTPMemberInfoSetSDES(leg->rtp_context, 
                                           p_person,
                                           RTP_MI_NAME,
                                           ssrc->name);
            if ( the_err != RTP_OK) {
              debug("RTPSetSession", "%s\n", RTPDebugStr());
            }
            else {
              debug("RTPSetSelfSession",
                    "SDES set for context %d, person %d to %s\n",
                    leg->rtp_context, p_person, ssrc->name);
            }
          }
          else {
            debug("RTPSetSelfSession", 
                  "Member name is not present for %d(ssrc=%lu)\n",
                  j, ssrc->ssrc);
          }
        }
        else {
          debug("RTPSetSession", "Person could not be found %d\n",
                __LINE__);
        }

        p_person = 0;

        the_err =
          RTPSessionAddToContributorList(ssrc->cid,
                                         leg->ssrc);
        if ( the_err != RTP_OK ) {
          debug("RTPSetSelfSession",  "%s\n", RTPDebugStr());
        }
        else {
          debug("RTPMSetSelfSession", "Adding %u as contrib to context %d\n",
                leg->ssrc, ssrc->cid);
        }


        the_err = RTPSessionGetMemberList (ssrc->cid,
                                           &iter);
        if (the_err != RTP_OK) {
          debug ("RTPSetSelfDescription", 
                 "RTPSessionGetMemberList Error\n");
        }

        /*
        the_err = RTPCurrentMember (ssrc->cid,
                                    &iter,
                                    &i_person);
        if (the_err != RTP_OK) {
          debug ("RTPSetSelfDescription", 
                 "RTPSessionGetMemberList Error\n");
        }
        */

        while ( 1 ) {
          unsigned int ssrc2;

          the_err = RTPNextMember(ssrc->cid,
                                  &iter,
                                  &i_person);
          if ( the_err != RTP_OK ) {
            break;
          }

          the_err = RTPMemberInfoGetSSRC(ssrc->cid, i_person,
                               (u_int32 *) &ssrc2);
          if ( the_err != RTP_OK ) {
            debug("RTPSetSelfSession", "%s\n", RTPDebugStr());
          }

          if ( ssrc2 == leg->ssrc ) {
            p_person = i_person;
            break;
          }
        }

        if ( p_person != 0 ) {
          if ( leg->name ) {
            the_err = RTPMemberInfoSetSDES(ssrc->cid, p_person,
                                           RTP_MI_NAME,
                                           leg->name);
            if ( the_err != RTP_OK) {
              debug("RTPSetSession", "%s\n", RTPDebugStr());
            }
            else {
              debug("RTPSetSession", "Setting SDES %s to context %d, "
                    "person %d\n",
                    leg->name, ssrc->cid, p_person);
            }
          }
        }
        else {
          debug("RTPSetSession", "No person name, Line %d\n",
                __LINE__);
        }
      }

      UNLOCK(leg->mutex);
      UNLOCK(conf->mutex);
    }
    else {
      debug("RTPMSetSelfSession", "No conference for user\n");
    }
  }
#endif

  return 0;
}

/*
 * Get the Mixer SDP for the call leg
 */
session_description_t RTPMGetRemoteSessionDescription(int context)
{
  session_description_t sdp;
  rtpm_call_leg_t *leg;

  SDPInit(&sdp);

  leg = ListElement(&glUsers, context);
  if (leg != NULL) {
    LOCK(leg->mutex);
    /* Update the IP address and port numbers */
    SDPCopy(&sdp, &leg->sdp_rx);
    UNLOCK(leg->mutex);

    /* Update the media paylod types */
    SDPDefault(&sdp);
  }
  else {
    debug("RTPMGetRemoteSession", "No call leg\n");
  }

  return sdp;
}

/************************** END OF FILE ********************************
*/
