/***************************** FILE HEADER *****************************
#               Copyright (c) 2000 by Columbia University              *
#                        All Rights Reserved                           *
#***********************************************************************
#***********************************************************************
#
#  File Name    : rtpmapi.h
#
#  Module       : Audio Conference Server
#
#  Author(s)    : Kundan Singh (kns10@cs.columbia.edu)
#                 Gautam Nair  (gnair@cs.columbia.edu)
#
#  Created      : 28 Apr 2000
#
#  Description  : RTP Audio Mixer API.
#                
#  Notes        : 
#
#***********************************************************************
#                Modification History (Reverse Order)
#-----------------------------------------------------------------------
#   Date         Author              Description
#-----------------------------------------------------------------------
#              |             | 
#              |             | 
#-----------------------------------------------------------------------
*/

#ifndef RTPMAPI_H
#define RTPMAPI_H

#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h> /* struct sockaddr */
#include <netinet/in.h> /* struct sockaddr_in */

#ifdef __cplusplus
extern "C" 
{
#endif

#include "list.h"       /* generic list_t */

#ifdef USE_RTP
#include "rtp_api.h"    /* RTP context */
#else
#error "Can not build this module without RTP. configure --with-rtp option"
#endif

#include "sdp.h"
#include "buffer.h"

#ifdef __cplusplus
}
#endif

/* -------------------- RTP Audio Mixer API ---------------------- */

/* Context to refer to a call leg */
typedef int rtpmcontext;

#ifdef __cplusplus
extern "C" 
{
#endif

/* To add a call leg (new participant) for a conference */
int RTPMAddCallLeg(const char *confname, const char *source);
/* This returns a unique context info, -1 on error */

/* To remove a participant */
int RTPMRemoveCallLeg(int context);
/* This returns 0 on success and -1 on failure */

/* To set the SDP of the participant */
int RTPMSetSelfSessionDescription(int context, 
                                  session_description_t s);
/* This returns 0 on success and -1 on failure.
 * session_description_t should be freed by the caller. */

/* To get the SDP of the mixer for a specific participant.
 * the SDP will be (almost) same as that of the 
 * participant */
session_description_t RTPMGetRemoteSessionDescription(int context); 
/* This returns 0 on success and -1 on failure.
 * The return value should be freed by the caller */

/* Set the packetization time in ms
 * This may work only when there is no active conferences.
 */
void RTPMSetPacketizationTime(int time_in_millisec);

/*
 * Send back the audio to the participant who generate that audio
 */
  void RTPMSendBackAudioToParticipant(int true_or_false);

#ifdef __cplusplus
}
#endif

/* -------------------- Internal structures ------------------ */

/* Structure to hold information about a participant */
typedef struct rtpm_call_leg_t {
  int confid;         /* Index in conference list */
  char * participant; /* Name of the participant */
  session_description_t sdp_tx; /* Session description of the
                                   participant */
  session_description_t sdp_rx; /* Session description of the
                                   mixer for the participant */ 
  context rtp_context;  /* returned by RTPCreate() */
  pthread_t rx_thread;  /* thread to process the call leg */
  pthread_mutex_t mutex; /* mutex and condition variables */
  pthread_cond_t cond;
  media_buffer_t *rxbuflist; /* received decoded buffers */
  delay_calc_t    delay;
  int oContributed;    /* Whether this contributed to RTP audio */
  int oTerminate;      /* To terminate the thread */
  unsigned long tsinc; /* Timestamp increment */
  unsigned long ssrc;  /* SSRC for the user */
#ifdef INCLUDE_SSRC
  char   *name;        /* RTP specific SDES */
#endif

  pthread_t video_thread; /* Thread for video data processing */
  int       video_sock;   /* socket for video data */
  int       video_sock_rtcp; /* socket for video RTCP data */
  struct sockaddr_in video_local_sin; /* Local IP address */
  struct sockaddr_in video_remote_sin; /* remote IP address */

  pthread_t data_thread; /* Thread for data processing, text, chat */
  int       data_sock;
  struct sockaddr_in data_local_sin;
  struct sockaddr_in data_remote_sin;
} rtpm_call_leg_t;

typedef struct {
  context       cid;  /* RTP context */
  unsigned long ssrc; /* SSRC */
  char         *name; /* CNAME */
} rtpm_ssrc_info_t;

/* Structure to hole information about a conference */
typedef struct rtpm_conf_t {
  char * conference; /* Name of the conference */
  list_t participants; /* List of indices of participants */
  pthread_t tx_thread; /* Thread to send the RTP packets */
  pthread_mutex_t mutex;
  pthread_cond_t cond;
  unsigned long play_ptr; /* relative position from where to play
                          * Same as local sampling clock */
  int oTerminate;     /* To terminate the thread */
#ifdef INCLUDE_SSRC
  list_t        user_ssrcs;   /* SSRCs of other users */
#endif
} rtpm_conf_t;

#endif

/* -------------------- Design and pseudo code --------------- */
#if 0

list_t conf;  /* List of conferences */
list_t users; /* List of various users in different conferences.
               * Index in this list is used as rtpm_context_t */


//-----
int RTPMAddCallLeg(const char *confname, const char *source);
{
  Create a new rtpm_call_leg_t and add it in (list_t)users;
  If the (list_t)conf does not have (char *)confname then
    create a new (rtpm_conf_t) and add it in (conf);
  call RTPCreate(&callleg->rtp_context);
  Start the calllegthread->t();
  return the index of rtpm_call_leg_t in (users).
}

//-----
int RTPMRemoveCallLeg(int context);
{
  Destroy the associated calllegthread->t();
  Call RTPDestroy() if applicable.
  Remove the element from the conf list_t. If last element in the
    conf list then remove the conf list_t element also.
  Remove the element from users list_t.
  Free all other data structures.
}

//-----
int RTPMSetSelfSessionDescription(int context, 
                                  session_description_t s);
{
  store s as callleg->sdp_tx.
  copy s as callleg->sdp_rx.
  XXX update the audio mixer info.
}

//-----
session_description_t RTPMGetRemoteSessionDescription(int context); 
{
  return callleg->sdp_rx;
}

rtpm_call_leg_thread() {
  /* This thread processes the packets to and from the 
   * particular call leg user */
}

#endif

/* -------------------- End of file -------------------------- */
