/***************************** FILE HEADER *****************************
#               Copyright (c) 2000 by Columbia University              *
#                        All Rights Reserved                           *
#***********************************************************************
#***********************************************************************
#
#  File Name    : sched.c
#
#  Module       : Audio Conference Server
#
#  Author(s)    : Kundan Singh (kns10@cs.columbia.edu)
#                 Gautam Nair  (gnair@cs.columbia.edu)
#
#  Created      : 28 Apr 2000
#
#  Description  : RTP Scheduler. Used in RTP library.
#                
#  Notes        : 
#
#***********************************************************************
#                Modification History (Reverse Order)
#-----------------------------------------------------------------------
#   Date         Author              Description
#-----------------------------------------------------------------------
#              |             | 
#              |             | 
#-----------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <pthread.h>

#ifdef USE_RTP
#include "rtp_api.h"
#include "rtp_highlevel.h"
#else
#error "Can not build this module without RTP. configure --with-rtp option"
#endif

#include "sched.h"
#include "stdtypes.h"

//#undef INCLUDE_SSRC

/* RTP related functions --- */

#define MAX_SCHEDULE 1024

#ifndef FALSE
#define FALSE 0
#define TRUE 1
#endif
 
/* Below are functions that implement the scheduler */

typedef struct {
  context cid;
  int Active;
  struct timeval activate_time;
  rtp_opaque_t opaque;
} sched_item;

/* Allow scheduling of up to MAX_SCHEDULE items */
sched_item sched[MAX_SCHEDULE];
#ifdef INCLUDE_SSRC
pthread_mutex_t sched_mutex;  
extern pthread_mutex_t rtp_mutex;
#endif

/* Scheduler initialization routine */
int InitSched()
{
  int i;
  for (i=0; i<MAX_SCHEDULE; i++){
    sched[i].Active = FALSE;
  }
#ifdef INCLUDE_SSRC
  pthread_mutex_init(&sched_mutex, NULL);
#endif
  return 0;
}


/* FindInactive locates an empty slot on the scheduler array so
   that a new item can be scheduled there */
int FindInactive(){
  int i;
  for (i=0; i < MAX_SCHEDULE;i++){
    if (!sched[i].Active){
      sched[i].Active = TRUE;
      return i;
    }
  }
  return -1;
}


/* Returns an item that is ready to be activated.  Note that this does
   not return the items in any particular chronological order. */
int FindReady(context *cid, rtp_opaque_t *opaque){
  int i;
  struct timeval curtime;

  gettimeofday(&curtime, NULL);
  for (i=0; i<MAX_SCHEDULE; i++){
    if (sched[i].Active && (sched[i].activate_time.tv_sec < curtime.tv_sec ||
	(sched[i].activate_time.tv_sec == curtime.tv_sec &&
	 sched[i].activate_time.tv_usec < curtime.tv_usec))){
      *cid = sched[i].cid;
      *opaque = sched[i].opaque;
      sched[i].Active = FALSE;
      return i;
    }
  }

  return -1;
}

/* RTPScheduleThread thread to continuosly schedule the thread */
void *RTPScheduleThread (void * param)
{
  int i;
  struct timeval tinytime;
  context cid;
  rtp_opaque_t opaque;

  while ( 1 ) {

#ifdef INCLUDE_SSRC
    LOCK(sched_mutex);
#endif

    i= FindReady(&cid, &opaque);

#ifdef INCLUDE_SSRC
    UNLOCK(sched_mutex);
#endif

    if ( i != -1) {
      //debug("RTPScheduleThread", "executing..\n");
#ifdef INCLUDE_SSRC
      LOCK(rtp_mutex);
#endif

      RTPExecute(cid, opaque);

#ifdef INCLUDE_SSRC
      UNLOCK(rtp_mutex);
#endif
    }
    else {
      /* Wait for sometime say 100 ms */
      tinytime.tv_sec = 0;
      tinytime.tv_usec = 100000; /* 100 ms */
      select(0, NULL, NULL, NULL, &tinytime);
    }
  }

  return NULL;
}

/* The scheduling routine */
void RTPSchedule(context cid, rtp_opaque_t opaque, struct timeval *tp)
{
  //struct timeval curtime;
  int i;

  if ( tp == NULL ) {
    return;
  }

#ifdef INCLUDE_SSRC
  LOCK(sched_mutex);
#endif

  i = FindInactive();
  if (i < 0){
    printf("error: no more room in scheduler\n");
#ifdef INCLUDE_SSRC
    UNLOCK(sched_mutex);
#endif
    return;
  }
  sched[i].cid = cid;
  sched[i].Active = TRUE;
  sched[i].activate_time = *tp;
  sched[i].opaque = opaque;

#ifdef INCLUDE_SSRC
  UNLOCK(sched_mutex);
#endif

  /*
  gettimeofday(&curtime, NULL);

  debug("RTPSchedule", "Scheduled for %f seconds\n", (float) 
         (tp->tv_sec - curtime.tv_sec) +
	 (float) (tp->tv_usec - curtime.tv_usec) / 1000000);
  */

  return;
}


/************************** END OF FILE ********************************
*/

