/***************************** FILE HEADER *****************************
#               Copyright (c) 2000 by Columbia University              *
#                        All Rights Reserved                           *
#***********************************************************************
#***********************************************************************
#
#  File Name    : sdp.c
#
#  Module       : Audio Conference Server
#
#  Author(s)    : Kundan Singh (kns10@cs.columbia.edu)
#                 Gautam Nair  (gnair@cs.columbia.edu)
#
#  Created      : 28 Apr 2000
#
#  Description  : SDP parser and generator in C.
#                
#  Notes        : 
#
#***********************************************************************
#                Modification History (Reverse Order)
#-----------------------------------------------------------------------
#   Date         Author              Description
#-----------------------------------------------------------------------
#              |             | 
#              |             | 
#-----------------------------------------------------------------------
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "error.h"
#include "list.h"
#include "sdp.h"
#include "stdtypes.h"

/* SDP related functions */

/* Initialize SDP struct */
int SDPInit(session_description_t *sdp)
{
  if ( sdp == NULL ) {
    return -1;
  }
  ListCreate(&sdp->media_line_t, 1, 1);
  return 0;
}

/* Copy an SDP struct to another */
int SDPCopy(session_description_t *dest,
            session_description_t *src)
{
  int i,j;

  if ( dest == NULL || src == NULL ) {
    debug("SDPCopy", "%s is NULL", dest == NULL ? "dest" : "src");
    return -1;
  }

  /* Free XXX dest if already present */
  ListDelete(&dest->media_line_t); /* Destroy the previous list */

  ListCreate(&dest->media_line_t, ListCount(&src->media_line_t), 1);
  for ( i=0; i<ListCount(&src->media_line_t); i++) {
    media_line_t *m = ListElement(&src->media_line_t, i);
    media_line_t m2 = *m;
    if ( !m ) 
      continue;
    ListCreate(&m2.payloads, ListCount(&m->payloads), 1);
    for (j=0; j<ListCount(&m->payloads); j++) {
      payload_type_t *p = ListElement(&m->payloads, j);
      payload_type_t p2 = *p;
      if ( !p ) 
        continue;
      if ( p->description != NULL ) {
        p2.description = strdup(p->description);
      }
      ListAddElement(&m2.payloads, &p2, sizeof(p2));
    }
    ListAddElement(&dest->media_line_t, &m2, sizeof(m2));
  }

  return 0;
}

/*
 * Build the default SDP depending on the
 * algorithms supported.
 * Returns 0 on success and -1 on failure.
 */

int SDPDefault(session_description_t *dest)
{
  media_line_t *m2;
  payload_type_t p2;
  int i;
  int found = 0;

  if ( dest == NULL
       || ListCount(&dest->media_line_t) < 1) {
    debug("SDPDefault", "dest or dest->media is NULL\n");
    return -1;
  }

  for ( i=0; i<ListCount(&dest->media_line_t); i++) {
    m2 = ListElement(&dest->media_line_t, i);
    if ( !m2 ) 
      continue;
    if ( found == 0 &&
         m2->type == RTPM_MEDIA_AUDIO ) {
      /* First audio media Found */
      found = 1;

      debug("SDPDefault", "Audio port is %u\n",
            ntohs(m2->sin.sin_port));
      /* Delete existing payloads and create new list */
      ListDelete(&m2->payloads);

      ListCreate(&m2->payloads, 4, 1);
      
      p2.pt = 0; p2.description = strdup("pcmu/8000");
      ListAddElement(&m2->payloads, &p2, sizeof(p2));
      
      p2.pt = 8; p2.description = strdup("pcma/8000");
      ListAddElement(&m2->payloads, &p2, sizeof(p2));
      
      p2.pt = 3; p2.description = strdup("gsm/8000");
      ListAddElement(&m2->payloads, &p2, sizeof(p2));
      
      p2.pt = 5; p2.description = strdup("dvi/8000");
      ListAddElement(&m2->payloads, &p2, sizeof(p2));
      
    }
    else if ( m2->type == RTPM_MEDIA_VIDEO ) {
      /* Since video is supported, do not change the modes. */      
    }
    else if ( m2->type == RTPM_MEDIA_DATA ) {
      /* Since data is supported, do not change the modes. */
    }
    else {
      /* No other media type is supported */
      debug("SDPDefault", "Second audio or unknown media, set port = 0\n");
      m2->sin.sin_port = 0; /* put on hold */
    }
  }

  return 0;
}


/* Free an SDP struct internal members */
int SDPFree(session_description_t *sdp)
{
  int i,j;

  if ( sdp == NULL ) {
    debug("SDPFree","called with NULL\n");
    return -1;
  }

  for ( i=0; i<ListCount(&sdp->media_line_t); i++ ) {
    media_line_t *m = ListElement(&sdp->media_line_t, i);
    if ( !m ) 
      continue;
    for ( j=0; j<ListCount(&m->payloads); j++) {
      payload_type_t *p = ListElement(&m->payloads, j);
      if ( !p ) 
        continue;
      if ( p->description ) {
        DeallocateObject(p->description);
      }
    }
    ListDelete(&m->payloads);
  }
  ListDelete(&sdp->media_line_t);

  return 0;
}

/* Print an SDP struct for trace */
int SDPPrint(session_description_t *sdp)
{
  int i,j;

  if ( sdp == NULL ) {
    printf("SDP--NULL\n");
    return -1;
  }

  printf("SDP--\n");
  for ( i=0; i<ListCount(&sdp->media_line_t); i++) {
    media_line_t *m = ListElement(&sdp->media_line_t, i);
    if ( !m ) 
      continue;
    printf(" %s %s:%u -- ", 
           m->type == RTPM_MEDIA_AUDIO ? "audio" : 
           (m->type == RTPM_MEDIA_VIDEO ? "video" : 
            (m->type == RTPM_MEDIA_DATA ? "data" : "unknown" )),
           inet_ntoa(m->sin.sin_addr),
           ntohs(m->sin.sin_port));
    for ( j=0; j<ListCount(&m->payloads); j++) {
      payload_type_t *p;
      p= ListElement(&m->payloads, j);
      if ( !p ) 
        continue;
      printf(" %d[%s],", p->pt,
             p->description ? p->description : "none");
             
    }
    printf("\n");
  } 
  printf("\n");
  return 0;
}

/*
 * rudimentary parser for SDP
 */
session_description_t *SDPParse(const char *str, int size)
{
  char *t;
  session_description_t *sdp = NULL;
  media_line_t   media, mymedia;
  char *lasts, *line;
  int last_media_index = -1;

  if ( str == NULL ) {
    return NULL;
  }

  sdp = AllocateObject(session_description_t);
  ListCreate(&sdp->media_line_t, 1,1);

  if ( size <= 0 ) {
    t = strdup(str);
  }
  else {
    t = AllocateObjects(char, size+1);
    memcpy(t, str, size);
    t[size] = 0;
  }

  lasts = NULL;
  while ((line = strtok_r(t, "\r\n", &lasts)) != NULL) {
    t = NULL;
    if ( !strncmp(line, "c=", 2) ) {
      /* c=IN IP4 a.b.c.d/ttl/numofaddress */
      char *l2 = NULL;
      char *ptr;

      debug("SDPParse", "ConnectionLine: %s\n", line);

      line += 2;

      strtok_r(line, " ", &l2); /* ignore IN */
      strtok_r(NULL, " ", &l2); /* ignore IP4 */
      ptr = strtok_r(NULL, " ", &l2); /* get the address */
      if (ptr != NULL ) {
        char *addr, *ttlstr;
        l2 = NULL;
        addr = strtok_r(ptr, "/", &l2); /* get IP address */
        ttlstr = strtok_r(NULL, "/ ", &l2); /* get TTL */
        media.sin.sin_addr.s_addr = inet_addr(addr);
        media.sin.sin_port = 0; /* set to 0 */
        media.ttl = ttlstr ? atoi(ttlstr) : 127;

        if ( last_media_index != -1 ) {
          media_line_t *m = 
            ListElement(&sdp->media_line_t, last_media_index);

          debug("SDPParse", "Adding c= to m= line\n");

          if ( m != NULL ) {
            m->sin.sin_addr.s_addr = media.sin.sin_addr.s_addr;
          }
        }
      }
    } 
    else if ( !strncmp(line, "m=", 2)) {
      /* m=audio port RTP/AVP 0 4 8 */
      char *l2 = NULL;
      char *ptr;
      char *mediastr, *portstr;

      debug("SDPParse", "MediaLine: %s\n", line);

      line += 2;

      mediastr = strtok_r(line, " ", &l2); /* audio/video */
      if ( mediastr && (!strcasecmp(mediastr, "audio")
                        || !strcasecmp(mediastr, "video")
                        || !strcasecmp(mediastr, "text"))) {
        /* Only audio is parsed */
        debug("SDPParse", "audio/video\n");

        portstr = strtok_r(NULL, " ", &l2); /* port */
        strtok_r(NULL, " ", &l2); /* ignore RTP/AVP */
        /* Create the media to be added */
        mymedia = media;
        mymedia.sin.sin_port = htons(portstr ? atoi(portstr) : 0);
        if ( !strcasecmp(mediastr, "audio")) {
          mymedia.type = RTPM_MEDIA_AUDIO;
        } 
        else if (!strcasecmp(mediastr, "video")) {
          mymedia.type = RTPM_MEDIA_VIDEO;
        }
        else if ( !strcasecmp(mediastr, "text")) {
          mymedia.type = RTPM_MEDIA_DATA;
        }
        else {
          mymedia.type = RTPM_MEDIA_UNKNOWN;
        }
        
        debug("SDPParse", "media addr %s:%u\n",
              inet_ntoa(mymedia.sin.sin_addr),
              ntohs(mymedia.sin.sin_port));
        
        ListCreate(&mymedia.payloads, 3, 1);

        while ( (ptr = strtok_r(NULL, " ", &l2)) != NULL) {
          /* ptr is a payload type */
          payload_type_t pt;
          if ( mymedia.type == RTPM_MEDIA_AUDIO ||
               mymedia.type == RTPM_MEDIA_VIDEO ) {
            pt.description = NULL;
            pt.pt = atoi(ptr);
          }
          else {
            /* For data, text, or unknown */
            pt.description = strdup(ptr);
            pt.pt = -1; /* Put is -1 for unknown */
          }
          debug("SDPParse", "payload %d\n", pt.pt);

          ListAddElement(&mymedia.payloads, &pt, sizeof(pt));
        }

        last_media_index = 
          ListAddElement(&sdp->media_line_t, &mymedia, sizeof(mymedia));

        /*
        debug("SDPParse", "media count %d\n",
              ListCount(&sdp->media_line_t));
        */
      }
    }
    else if ( ! strncmp(line, "a=rtpmap:", 9)) {
      /* a=rtpmap:96 pcmu/8000 */
      char *l2=NULL;
      char *ptstr, *descr;
      int ptint;

      //debug("SDPParse", "AttribLine: %s\n", line);

      line += 9;
      ptstr = strtok_r(line, " ", &l2);
      descr = strtok_r(NULL, " ", &l2);
      if ( last_media_index != -1) {
        media_line_t *m = ListElement(&sdp->media_line_t, 
                                      last_media_index);
        if ( m != NULL ) {
          payload_type_t *pt;
          int idx;
          ptint = ptstr ? atoi(ptstr) : -1;
          for (idx = 0; idx < ListCount(&m->payloads); idx ++ ) {
            pt = ListElement(&m->payloads, idx);
            if ( !pt )
              continue;
            if ( pt && pt->pt == ptint ) {
              if ( pt->description ) {
                DeallocateObject(pt->description);
              }
              pt->description = strdup(descr);
            }
          }
        }
      }
    }
    else {
      /* ignore */
    }
  }

  SDPPrint(sdp);

  return sdp;
}

/*
 * Get the default predefined payload type description 
 * Define a new format here
 */
static char *pcmu = "pcmu/8000";
static char *pcma = "pcma/8000";
static char *dvi  = "dvi/8000";
static char *gsm  = "gsm/8000";
static char *h261 = "h261/90000";

char *GetPayloadTypeString(int pt) 
{
  switch (pt) {
  case 0:
    return pcmu;
  case 3:
    return gsm;
  case 5:
    return dvi;
  case 8:
    return pcma;
  case 31:
    return h261;
  default:
    debug("GetPayloadType", "pt= %d not defined\n", (int) pt);
    
    return NULL;
  }
}

/* 
 * Return the RTP payload type from sdp,
 * This returns the first payload type
 * for audio if *pt is -1. Else it
 * returns the appropriate payload format
 * for the payload type.
 */
char *GetPayloadType(int *pt, session_description_t *sdp)
{
  int i,j;

  for ( i=0; i < ListCount(&sdp->media_line_t); i++) {
    media_line_t *l = ListElement(&sdp->media_line_t, i);
    if ( !l ) 
      continue;
    for (j=0; j< ListCount(&l->payloads); j++) {
      payload_type_t *p = ListElement(&l->payloads, j);
      if ( !p ) 
        continue;
      if ( *pt != -1 ) {
        if ( p->pt == (int) *pt ) {
          if ( p->description != NULL ) {
            return p->description;
          }
          else {
            return GetPayloadTypeString(*pt);
          }
        }
      }
      else if ( l->type == RTPM_MEDIA_AUDIO ) {
        *pt = p->pt;
        if ( p->description != NULL ) {
          return p->description;
        }
        else {
          return GetPayloadTypeString(p->pt);
        }
      }
    }
  }

  debug("GetPayloadType", "returned unknown for pt= %d\n", (int) *pt);

  return NULL; /* unknown */
}


/* 
 * Return the RTP payload type from sdp,
 * returns the appropriate payload format
 * for the payload type. Return -1 for unknown.
 */
int GetPayloadTypeFromDescr(char *pt, session_description_t *sdp)
{
  int i,j;

  for ( i=0; i < ListCount(&sdp->media_line_t); i++) {
    media_line_t *l = ListElement(&sdp->media_line_t, i);
    if ( !l ) 
      continue;
    for (j=0; j< ListCount(&l->payloads); j++) {
      payload_type_t *p = ListElement(&l->payloads, j);
      if ( !p ) 
        continue;
      if ( pt != NULL ) {
        char *description;
        if ( p->description ) {
          description = p->description;
          }
        else {
          description = GetPayloadTypeString(p->pt);
        }
        if ( description && !strcasecmp(description, pt)) {
          return p->pt;
        }
      }
    }
  }

  return -1; /* unknown */
}


/************************** END OF FILE ********************************
*/

