/***************************** FILE HEADER *****************************
#               Copyright (c) 2000 by Columbia University              *
#                        All Rights Reserved                           *
#***********************************************************************
#***********************************************************************
#
#  File Name    : transcode.c
#
#  Module       : Audio Conference Server
#
#  Author(s)    : Kundan Singh (kns10@cs.columbia.edu)
#                 Gautam Nair  (gnair@cs.columbia.edu)
#
#  Created      : 26 Apr 2000
#
#  Description  : Transcoder to be used in audio conference server.
#                
#  Notes        : 
#
#***********************************************************************
#                Modification History (Reverse Order)
#-----------------------------------------------------------------------
#   Date         Author              Description
#-----------------------------------------------------------------------
#              |             | 
#              |             | 
#-----------------------------------------------------------------------
*/

#include <stdio.h>
#include "error.h"
#include "buffer.h"
#include "types.h"
#include "g72x.h"
#include "audio_descr.h"
#ifdef USE_GSM
#include "gsm.h"
#else
#warning "Not using GSM codec"
#endif

extern media_buffer_t *NewBuffer(int size);

/*
 * Transcoder from any alogorithm to L16 and
 * from L16 to any algorithm
 * Currently only pcmu, pcma, g721 and 
 * dvi_adpcm are supported.
 */

media_buffer_t *MediaTranscode (const char *inputalgo, 
				const char *outputalgo,
				char *data,
				int   data_len) 
{
  media_buffer_t *retbuf;
  int             count;


  if (inputalgo == NULL ||
      outputalgo == NULL ||
      data == NULL ||
      (data_len <= 0)) {
    return NULL;
  }

  //SSS
  if ((strcasecmp (inputalgo, "pcmu/8000") == 0) && 
      (strcasecmp (outputalgo, "pcma/8000") == 0)) {
    /* convert from pcmu to pcma */
    retbuf = NewBuffer (data_len); /* output data length is the same */
    retbuf->size = data_len;
    for (count = 0; count < data_len; count ++) {
      retbuf->data[count] = ulaw2alaw (data[count]);
    }
    return (retbuf);
  }

  if ((strcasecmp (inputalgo, "pcma/8000") == 0) && 
      (strcasecmp (outputalgo, "pcmu/8000") == 0)) {
    /* convert from pcma to pcmu */
    retbuf = NewBuffer (data_len); /* output data length is the same */
    retbuf->size = data_len;
    for (count = 0; count < data_len; count ++) {
      retbuf->data[count] = alaw2ulaw (data[count]);
    }
    return (retbuf);
  }
 
  if ((strcasecmp (inputalgo, "l16") == 0) && 
      (strcasecmp (outputalgo, "pcma/8000") == 0)) {
    /* convert from l16 to pcma */
    short *pShort;
    retbuf = NewBuffer (data_len / 2); /* output data length is half */
    retbuf->size = data_len / 2;
    pShort = (short *)data;
    for (count = 0; count < data_len / 2; count ++) {
      retbuf->data[count] = linear2alaw (*pShort);
      pShort ++;
    }
    return (retbuf);
  }

  if ((strcasecmp (inputalgo, "pcma/8000") == 0) && 
      (strcasecmp (outputalgo, "l16") == 0)) {
    /* convert from pcma to l16 */
    short *pShort;
    retbuf = NewBuffer (data_len * 2); /* output data length is twice */
    retbuf->size = data_len * 2;
    pShort = (short *)retbuf->data;
    for (count = 0; count < data_len; count ++) {
      *pShort = alaw2linear (data[count]);
      pShort ++;
    }
    return (retbuf);
  }

  if ((strcasecmp (inputalgo, "l16") == 0) && 
      (strcasecmp (outputalgo, "pcmu/8000") == 0)) {
    /* convert from l16 to pcmu */
    short *pShort;
    
    retbuf = NewBuffer (data_len / 2); /* output data length is half */
    retbuf->size = data_len / 2;
    pShort = (short *)data;
    for (count = 0; count < data_len / 2; count ++) {
      retbuf->data[count] = linear2ulaw (*pShort);
      pShort ++;
    }
    return (retbuf);
  }

  if ((strcasecmp (inputalgo, "pcmu/8000") == 0) && 
      (strcasecmp (outputalgo, "l16") == 0)) {
    /* convert from pcmu to l16 */
    short *pShort;
    
    retbuf = NewBuffer (data_len * 2); /* output data length is twice */
    retbuf->size = data_len * 2;
    pShort = (short *)retbuf->data;
    for (count = 0; count < data_len; count ++) {
      *pShort = ulaw2linear (data[count]);
      pShort ++;
    }
    return (retbuf);
  }

  if ((strcasecmp (inputalgo, "l16") == 0) &&
      (strcasecmp (outputalgo, "g721/8000") == 0)) {
    /* convert from l16 to g721 */
    struct g72x_state state;
    short *pShort;
    
    g72x_init_state (&state);
    /* output data length is 1/4 */
    retbuf = NewBuffer (data_len / 4);
    retbuf->size = data_len / 4;
    pShort = (short *) data;
    for (count = 0; count < data_len / 4; count ++) {
      retbuf->data[count] = g721_encoder (*pShort,
					  AUDIO_ENCODING_LINEAR,
					  &state);
      ((retbuf->data[count]) <<= 4);
      retbuf->data[count] |= g721_encoder (*pShort,
					   AUDIO_ENCODING_LINEAR,
					   &state);
      pShort ++;
    }
    return (retbuf);
  }

  if ((strcasecmp (inputalgo, "g721/8000") == 0) &&
      (strcasecmp (inputalgo, "l16") == 0)) {
    /* convert from g721 to l16 */
    struct g72x_state  state;
    short  *pShort;

    g72x_init_state (&state);
    /* output data length is 4 times */    
    retbuf = NewBuffer (data_len * 4); 
    retbuf->size = data_len * 4;
    pShort = (short *)retbuf->data;
    for (count = 0; count < data_len; count ++) {
      *pShort = g721_decoder (data [count], 
			      AUDIO_ENCODING_LINEAR,
			      &state);
      pShort ++;
    }
    return (retbuf);	      
  }

  {
    static void *ret = NULL;
    if ( ret == NULL ) {
      ret = dvi_adpcm_init(NULL, 0);
    }

    if ((strcasecmp (inputalgo, "l16") == 0) &&
        (strcasecmp (outputalgo, "dvi/8000") == 0)) {
      /* convert from l16 to dvi */
      audio_descr_t au_descr;
      int           retsize;
      
      au_descr.encoding    = AE_L16;
      au_descr.sample_rate = 8000;
      au_descr.channels    = 1;
      retbuf = NewBuffer (1500);
      dvi_adpcm_encode ((char *) data, data_len, &au_descr, 
                        retbuf->data, &retsize, ret, 1);
      retbuf->size = retsize;
      return (retbuf);
    }
    
    if ((strcasecmp (inputalgo, "dvi/8000") == 0) &&
        (strcasecmp (outputalgo, "l16")) == 0) {
      /* convert from dvi to l16 */
      audio_descr_t au_descr;
      int           retsize;
      
      au_descr.encoding    = AE_L16;
      au_descr.sample_rate = 8000;
      au_descr.channels    = 1;
      retbuf = NewBuffer (1500);
      dvi_adpcm_decode ((char *) data, data_len, &au_descr,
                        retbuf->data, &retsize, ret);
      retbuf->size = retsize;
      return (retbuf);    
    }
  }

#ifdef USE_GSM
  /* GSM 06.10 RPE-LTP, 8KHz 13.2 kb/s, payload type = 3 */
  {
    static gsm ret = 0; /* XXX this should be part of per 
                         * call leg variables, and not static */
    if ( ret == 0 ) {
      ret = gsm_create();
    }

    if ((strcasecmp (inputalgo, "l16") == 0) &&
        (strcasecmp (outputalgo, "gsm/8000") == 0)) {
      /* convert from l16 to gsm */
      int outcount;
      retbuf = NewBuffer ((data_len + 159)/ 320 * 33);
      for ( count=0, outcount=0; 
            (count + 319) < data_len; 
            count += 320, outcount += 33) {
        gsm_encode(ret, (data + count), (retbuf->data + outcount));
      }
      retbuf->size = outcount;
      return (retbuf);
    }
    
    if ((strcasecmp (inputalgo, "gsm/8000") == 0) &&
        (strcasecmp (outputalgo, "l16")) == 0) {
      /* convert from gsm to l16 */
      int outcount;
      retbuf = NewBuffer ((data_len + 32)/33 * 320);
      for ( count=0, outcount=0; 
            (outcount + 32) < data_len; 
            count += 320, outcount += 33) {
        gsm_decode(ret, (data + outcount), (retbuf->data + count));
      }
      retbuf->size = count;
      return (retbuf);
    }
  }
#endif
    
  /* didn't find any matches for the codec */
  debug("MediaTranscode", "returning NULL\n");
  return (NULL);

} /* MediaTranscode */

/************************** END OF FILE ********************************
*/

