/***************************** FILE HEADER *****************************
#               Copyright (c) 2000 by Columbia University              *
#                        All Rights Reserved                           *
#***********************************************************************
#***********************************************************************
#
#  File Name    : video.c
#
#  Module       : Audio/video Conference Server
#
#  Author(s)    : Kundan Singh (kns10@cs.columbia.edu)
#
#  Description  : Video switching logic.
#                
#  Notes        : 
#
#***********************************************************************
#                Modification History (Reverse Order)
#-----------------------------------------------------------------------
#   Date         Author              Description
#-----------------------------------------------------------------------
#              |             | 
#              |             | 
#-----------------------------------------------------------------------
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "error.h"
#include "timespec.h"
#include "list.h"
#include "sdp.h"
#include "stdtypes.h"
#include "rtpmapi.h"

#define INTEROPERATE_WITH_SIPC 1 // set self port number as remote

#ifndef WIN32
#define closesocket close
#endif

extern list_t glConference; /* list of conferences */
extern list_t glUsers;      /* List of users */
extern struct in_addr host2ip(char *);

/*
 * Start listen port tries to allocate the consecutive port numbers
 * for RTP and RTCP and bind the appropriate sockets.
 * If the port is specified, it is preferred for
 * RTP port. All port numbers are in host byte order.
 * returns the allocated RTP port number. RTCP is RTP+1.
 * returns 0 in case of error. The *rtp_sock and *rtcp_sock
 * hold the socket ID of the RTP and RTCP sockets.
 * Set port to 0 if no preferred port.
 */
static unsigned short start_listen_port(unsigned short port,
                                 int * rtp_sock,
                                 int * rtcp_sock)
{
  struct sockaddr_in sin;
  int sinlen;
  int max_retry = 10;

  if ( rtp_sock == NULL || rtcp_sock == NULL ) {
    debug("start_listen_port", "One of the params are NULL\n");
    return 0;
  }

  /* While max retry */
  while ( --max_retry >= 0 ) {    

    /* Allocate the RTP and RTCP sockets */
    if (((*rtp_sock) = socket(AF_INET, SOCK_DGRAM, 0)) < 0 ) {
      debug("start_listen_port", "rtp_socket failed\n");
      return 0;
    }    
    if (((*rtcp_sock) = socket(AF_INET, SOCK_DGRAM, 0)) < 0 ) {
      closesocket(*rtp_sock);
      debug("start_listen_port", "rtcp socket failed\n");
      return 0;
    }
    
    /* Bind the RTP socket to the port */
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = INADDR_ANY;
    sin.sin_port = htons(port);
    if ( (bind(*rtp_sock, (struct sockaddr *)&sin, sizeof(sin))) < 0) {
      if ( port == 0 ) {
        /* If port was already 0 (any),
         * some serious error */
        closesocket(*rtp_sock);
        closesocket(*rtcp_sock);
        return 0;
      }
      else {
        /* We tried an non-0 port.
         * Try a 0 port now. */
        port = 0;
        debug("start_listen_port", "Requested port not bound. Trying any.\n");
        continue;
      }
    }

    /* Extract the port number now. If it were 0. */
    if ( port == 0 ) {
      sinlen = sizeof(sin);
      if ( getsockname(*rtp_sock, (struct sockaddr *)&sin, &sinlen) < 0 ) {
        debug("start_listen_port", "getsockname error\n");
      }
      port = ntohs(sin.sin_port);
    
      if ( port % 2 == 0 ) {
        /* Port is even number. Good */
      }
      else {
        /* Port is some odd number. Bad. */
        port = port + 1; /* make it even */
        closesocket(*rtp_sock);
        closesocket(*rtcp_sock);
        /* Continue with the next attempt */
        debug("start_listen_port", "Bound to odd port\n");
        continue;
      }
    }
    else {
      /* An non-0 value was specified for the 
       * port. Continue with rtcp binding */
    }

    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = INADDR_ANY;
    sin.sin_port = htons(port + 1);
    if ( (bind(*rtcp_sock, (struct sockaddr *)&sin, sizeof(sin))) < 0) {
      /* RTCP bind failed. Try again. */
      port = 0;
      closesocket(*rtp_sock);
      closesocket(*rtcp_sock);
      debug("start_listen_port", "rtcp bind failed\n");
      continue;
    }

    /* Everytthing is OK till now. */
    return port;
  }

  debug("start_listen_port", "Max retry exhausted\n");
  return 0;
}


/*
 * Create the video socket for RTP and RTCP. Set appropriate values in
 * the call leg structure.
 * return 0 for success and < 0 for failure.
 */
int RTPMVideoCreateSocket(rtpm_call_leg_t *leg,
                          media_line_t   *media)
{
  media_line_t media_rx;
  int namelen;
  int ipx;
  char hostname[200];
  unsigned short port;

  debug("RTPMCreateVideoSocket", "called\n");

  port = start_listen_port(ntohs(media->sin.sin_port),
                           &leg->video_sock,
                           &leg->video_sock_rtcp);

  if ( port == 0 ) {
    leg->video_sock = -1;
    leg->video_sock_rtcp = -1;
    return -1;
  }

  if ( getsockname(leg->video_sock,
                   (struct sockaddr *) &leg->video_local_sin,
                   &namelen) < 0 ) {
    error("RTPMCreateVideoSocket", "getsockname failed\n");
    closesocket(leg->video_sock);
    leg->video_sock = -1;
    return -4; /* getsockname failed */
  }

  /* Set the media_rx */
  media_rx.type = media->type;
  media_rx.ttl  = media->ttl; /* if needed */

  ListCreate(&media_rx.payloads, 1, 1);

  for ( ipx = 0; ipx < ListCount(&media->payloads); ipx++) {
    payload_type_t *px = ListElement(&media->payloads, ipx);
    payload_type_t pr;
    if ( !px )
      continue;
    pr.pt = px->pt;
    pr.description = px->description ? strdup(px->description) :
      NULL;
    ListAddElement(&media_rx.payloads, &pr, sizeof(pr));
  }
  media_rx.sin.sin_port = leg->video_local_sin.sin_port; /* Do not call htons*/
  media_rx.sin.sin_family = AF_INET;
  gethostname(hostname,sizeof(hostname));
  media_rx.sin.sin_addr = host2ip(hostname);
  ListAddElement(&leg->sdp_rx.media_line_t, 
                 &media_rx, sizeof(media_rx));

  /* Set the call leg parameters */
  memcpy(&leg->video_remote_sin, &media->sin, sizeof(media->sin));
  leg->video_remote_sin.sin_family = AF_INET;

#ifdef INTEROPERATE_WITH_SIPC
  /* Since vic can accept only one port set the remote port
   * same as self */
  leg->video_remote_sin.sin_port = media_rx.sin.sin_port;
#endif

  return 0;
}

/*
 * Delete the video socket.
 */
int RTPMVideoDeleteSocket(rtpm_call_leg_t *leg)
{
  close(leg->video_sock);
  memset(&leg->video_local_sin, 0, sizeof(leg->video_local_sin));
  memset(&leg->video_remote_sin, 0, sizeof(leg->video_remote_sin));
  return 0;
}

/*
 * Returns 1 if the sender can send the particular RTP packet 
 * to the receiver. Else returns 0.
 * This matches the payloadtype with the allowed SDP payload types
 * of the receiever. The buffer is updated to reflect the new
 * payload type of the receiver.
 */
int RTPMVideoCanSend(rtpm_call_leg_t *sender,
                     rtpm_call_leg_t *receiver,
                     char *buffer,
                     int length)
{

  int pt = 0;
  char *ptstr;

  if ( length < 12 ) {
    /* Min RTP packet length */
    debug("RTPMVideoCanSend", "Length is %d\n", length);
    return 0; /* No */
  }

  pt = (buffer[1] & 0x07f); /* Set the payload type */

  ptstr = GetPayloadType(&pt, &sender->sdp_rx);

  if (ptstr != NULL ) {
    pt = GetPayloadTypeFromDescr(ptstr, &receiver->sdp_tx);
    if ( pt != -1 ) {
      /* Update the payload type in buffer */
      unsigned char p = (unsigned char) ((buffer[1] & 0x080) | (pt & 0x007f));
      buffer[1] = p;
      return 1; /* Yes */
    }
  }
  return 0; /* No */
}

/*
 * Thread for processing incoming video RTP/RTCP packets
 */
void * RTPMVideoThread(void *param)
{
  int context = (int) param;
  rtpm_call_leg_t *leg;
  struct timeval initialtime, socktv;
  struct timespec next;
  fd_set selection;
  int maxfd = 0;
  int sock_rtp;
  int sock_rtcp;
  char rtp_buffer[1500]; int rtp_length;
  char rtcp_buffer[1500]; int rtcp_length;
  int  packet_received; /* 1 for RTP, 2 for RTCP, 3 for both */
  struct sockaddr_in remote_sin;
  int sin_len;
  rtpm_conf_t *conf;
  int k;
  rtpm_call_leg_t *user;
  int *userid;
  int pt;

  debug("VideoThread", "Entering\n");

  leg = ListElement(&glUsers, context);
  if ( leg == NULL ) {
    error("RTPMVideoThread", "No user with context %d\n", context);
    return NULL;
  }

  LOCK(leg->mutex);

  sock_rtp = leg->video_sock;
  sock_rtcp = leg->video_sock_rtcp;
  maxfd = sock_rtp > sock_rtcp ? sock_rtp : sock_rtcp;

  /* Loop while not terminated */
  gettimeofday(&initialtime, NULL);
  while (!leg->oTerminate) {
    next = timespec_delay(initialtime, 20/1000.0 ) ; /* 10 millisec */
    COND_TIMEDWAIT(leg->cond, leg->mutex, next);
    if ( leg->oTerminate ) {
      break; /* from while loop */
    }

    gettimeofday(&initialtime, NULL);

    FD_ZERO(&selection);
    FD_SET(sock_rtp, &selection);
    FD_SET(sock_rtcp, &selection);

    socktv.tv_sec = 0;
    socktv.tv_usec = 0; /* 0 millisec */

    /* Recv UDP packets from RTP/RTCP socket */
    while ( select(maxfd + 1, &selection, NULL, NULL, &socktv) > 0 ) {
      packet_received = 0;
      if ( FD_ISSET (sock_rtp, &selection)) {
        rtp_length = recvfrom(sock_rtp, rtp_buffer, sizeof(rtp_buffer), 0,
                 (struct sockaddr *) &remote_sin, &sin_len);
        if ( rtp_length > 0 ) {

          packet_received |= 0x001;

          /*
          debug("RTPMVideoThread", "Received %d bytes of video (%d)\n",
                rtp_length, (rtp_buffer[1] & 0x7f));
          */
        }
      }
      if ( FD_ISSET (sock_rtcp, &selection)) {

        rtcp_length = recvfrom(sock_rtcp, rtcp_buffer, sizeof(rtcp_buffer), 0,
                  (struct sockaddr *) &remote_sin, &sin_len);
        if ( rtcp_length > 0 ) {

          packet_received |= 0x002;

          /*
          debug("RTPMVideoThread", "Received %d bytes of RTCP for video\n",
                rtcp_length);
          */
        }
      }

      FD_ZERO(&selection);
      FD_SET(sock_rtp, &selection);
      FD_SET(sock_rtcp, &selection);
      socktv.tv_sec = 0;
      socktv.tv_usec = 0; /* 0 millisec */

      if ( packet_received != 0 ) {
        /* Release the mutex */
        UNLOCK(leg->mutex);

        /* Get the conference mutex. */
        conf = ListElement(&glConference, leg->confid);

        if ( conf ) {
          LOCK(conf->mutex);
        }

        /* Get the call leg mutex again */
        leg = ListElement(&glUsers, context);
        if ( leg == NULL ) {
          error("RTPMVideoThread", "No user with context %d, exiting\n", 
                context);
          if ( conf ) {
            UNLOCK(conf->mutex);
          }
          return NULL;
        }
        LOCK(leg->mutex);
        
        /* Now we have both leg and conf mutexes */
        if ( conf ) {
          /* Iterate through all the participants list */

          for ( k=0; k<ListCount(&conf->participants); k++) {
            userid = ListElement(&conf->participants, k);
            if ( !userid ) {
              //debug("RTPMVideoThread", "returned NULL from participants\n");
              continue;
            }
            user   = ListElement(&glUsers, *userid); /* participant */
            if ( !user ) {
              debug("RTPMVideoThread", "returned NULL from user list\n");
              continue;
            }
            
            if ( *userid == context || user == leg) {
              /* Same user */
              continue;
            }

            /*
            debug("RTPMVideoThread", "Participant %d of %d\n", *userid,
                  ListCount(&conf->participants));
            */

            LOCK(user->mutex);

            if ( user->video_sock >= 0 ) {
              //debug("RTPMVideoThread", "This has video\n");

              /* Video is active, probably */
              pt = (rtp_buffer[1] & 0x0ff); /* Store the payload type */
              if ( (packet_received & 0x001) && 
                   RTPMVideoCanSend(leg,user,rtp_buffer,rtp_length)) {
                remote_sin = user->video_remote_sin;

                
                rtp_length = sendto(user->video_sock,
                       rtp_buffer, rtp_length, 0,
                       (struct sockaddr *)&remote_sin,
                       sizeof(remote_sin));
                
                /*
                debug("RTPMVideoThread", "tx video (%d) size %d to %s:%d\n",
                      (rtp_buffer[1] & 0x7f), rtp_length,
                      inet_ntoa(remote_sin.sin_addr), 
                      ntohs(remote_sin.sin_port));
                */
              }
              rtp_buffer[1] = (pt & 0x0ff); /* restore payload type */

              if ( packet_received & 0x002 ) {
                /* Send RTCP */
                remote_sin = user->video_remote_sin;
                remote_sin.sin_port = /* change for RTCP */
                  htons(ntohs(remote_sin.sin_port)+1);
                
                rtcp_length = sendto(user->video_sock_rtcp,
                       rtcp_buffer, rtcp_length, 0,
                       (struct sockaddr *)&remote_sin, sizeof(remote_sin));

                
                debug("RTPMVideoThread", "tx video RTCP size %d to %s:%d\n",
                      rtcp_length,
                      inet_ntoa(remote_sin.sin_addr), 
                      ntohs(remote_sin.sin_port));
                

              }
            }

            UNLOCK(user->mutex);
          } /* Iterate through all participants */

          UNLOCK(conf->mutex);
        }
      }
    }
  }

  UNLOCK(leg->mutex);

  debug("VideoThread", "Exiting\n");

  return NULL;
}

/************************** END OF FILE ********************************
*/

