/***************************** FILE HEADER *****************************
#               Copyright (c) 1999 by Columbia University              *
#                        All Rights Reserved                           *
#***********************************************************************
#***********************************************************************
#
#  File Name    : iptelapi.cpp
#
#  Module       : Iptel Interface Implementation
#
#  Author(s)    : Kundan Singh (kns10@cs.columbia.edu)
#
#  Created      : 29 Sep 1999
#
#  Description  : Methods defined in iptelapi.h
#                
#  Notes        : 
#
#***********************************************************************
#                Modification History (Reverse Order)
#-----------------------------------------------------------------------
#   Date         Author              Description
#-----------------------------------------------------------------------
#              |             | 
#              |             | 
#-----------------------------------------------------------------------
*/

#include "stdio.h"    //NULL,sprintf,sscanf
#include "stdlib.h"   //standard lib
#include "string.h"   //strlen,strchr...
#include "stdarg.h"   // variable length args.
#include "iptelapi.h" //IPTEL API 

#ifdef __cplusplus
extern "C" {
#endif

#include "sysdep.h"
#include "dstring.h"
#include "uri.h"
#include "sip.h"
#include "strdupn.h"   /* free_valid */

#ifdef __cplusplus
}
#endif

/* 
 * For easy migration to a different memory manager
 */
#define AllocateObject(type,count)    \
    (((count) == 1) ? (new type) : (new type[count]));
#define DeallocateObject(obj)  if (obj) delete (obj);
#define ReallocateArray(obj,type,oldcount,newcount)\
  {                                                 \
    type * newObj = new type[newcount];             \
    int min = ((oldcount) <= (newcount)) ? (oldcount) : (newcount); \
    if ( (newObj != NULL) && (oldcount) > 0 && obj != NULL )\
    {                                               \
      memcpy(newObj,obj,sizeof(type) * (min));      \
      delete obj;                                   \
    }                                               \
    if ( newObj )                                   \
      obj = newObj;                                 \
  }
#define DuplicateString(newPtr,ptr)                 \
  {                                                 \
    newPtr = new char[strlen(ptr)+1];               \
    strcpy(newPtr,ptr);                             \
  }

/*
 * For easy migration to a different memory manager
 */
#define DeallocatePointer(ptr) if (ptr) { free(ptr); ptr=NULL; }

/*
 * For portability with Linux and other platforms 
 */
#define IPADDRESS_A1(s) (unsigned char)((((unsigned long) ntohl(s)) >> 24) & 0x0ff) 
#define IPADDRESS_A2(s) (unsigned char)((((unsigned long) ntohl(s)) >> 16) & 0x0ff) 
#define IPADDRESS_A3(s) (unsigned char)((((unsigned long) ntohl(s)) >> 8) & 0x0ff) 
#define IPADDRESS_A4(s) (unsigned char)((((unsigned long) ntohl(s))) & 0x0ff) 

#define IPADDRESS_ALL(a1,a2,a3,a4) \
htonl((((unsigned long)a1) << 24) | \
(((unsigned long)a2) << 16) | \
(((unsigned long)a3) << 8)  | \
(((unsigned long)a4)))

/////////////////////// Extern prototypes /////////////////
extern sip_addr_t URI_ParseNew(const char *v);
extern IptelAddress SipAddressToIptelAddress(sip_addr_t addr);

/////////////////////// Address ///////////////////////////

IptelAddress::IptelAddress()
{
  /* Empty address */
  Address = NULL;
}

IptelAddress::IptelAddress(const char *address)
{
  Address = NULL;

  if ( address == NULL ) {
    return;
  }

  Address = strdup(address);
}

IptelAddress::IptelAddress(const IptelAddress & address)
{
  Address = NULL;
  if ( address.Address == NULL ) {
    return;
  }

  Address = strdup(address.Address);
}

IptelAddress::~IptelAddress()
{
  DeallocatePointer(Address);
}

IptelAddress& IptelAddress::operator=(const char * address)
{
  DeallocatePointer(Address); // optional if garbage-collection
  if ( address == NULL ) {
    return *this;
  }
  Address = strdup(address);
  return (*this);
}

IptelAddress& IptelAddress::operator=(const IptelAddress & address)
{
  DeallocatePointer(Address); // optional if garbage-collection
  if ( address.Address == NULL ) {
    return *this;
  }

  Address = strdup(address.Address);
  return (*this);
}

IptelAddress& IptelAddress::operator=(const IptelIPAddress & address)
{
  DeallocatePointer(Address); // optional if garbage-collection
  Address = (char *) address; /* This allocates the memory */
  return (*this);
}

bool IptelAddress::operator==(const IptelAddress & address) const
{
  if ( Address == NULL ) {
    return (!address.IsEmpty());
  }

  return strcmp(address.Address,Address) == 0;
}

bool IptelAddress::operator==(const char * address) const
{
  if ( address == NULL || *address == '\0' ) {
    return !IsEmpty();
  }

  return !strcmp(address,Address);
}

IptelAddress::operator IptelIPAddress () const
{
  IptelIPAddress address;
  address = GetAddressString();
  return address;
}

const char *IptelAddress::GetAddressString() const
{
  if ((Address != NULL))
  {
    return Address;
  }
  else
  {
    /* XXX is this correct, to return "unknown" if the
     * address is not valid */
    return "unknown";
  }
}

bool IptelAddress::IsEmpty() const
{
  if ((Address != NULL)) {
    if ( *Address == '\0') {
      return true;
    }
    else {
      return false;
    }
  }
  else {
    return true;
  }
}

bool IptelAddress::IsTelephoneNumber() const
{
  if ( Address == NULL ) {
    return false;
  }
  int size = strlen(Address);
  for ( int i=0; i<size; i++) {
    char c = *(Address + i);
    if ( ! (
            c >= '0' && c <= '9' ||
            c == '+' || c == '-' || 
            c == '(' || c == ')' ||
            c == '#' || c == '*' ||
            c == ',' ) ) {
      return false;
    }
  }
  return true;
}

bool IptelAddress::IsIpAddress() const
{
  if ( Address == NULL) {
    return false;
  }

  char * h = GetHost();
  if ( h ) {
    IptelAddress newaddr = h;
    if ( newaddr.IsTelephoneNumber() ) {
      /* telephone number is not IP address */
      return false;
    }

    /* Typecast it to an IptelIPAddress and see
     * if the resulting address is valid or not */
    IptelIPAddress a = h;
    
    if ( a.IsValid() ) {
      DeallocatePointer(h);
      return true;
    }
    else {
      DeallocatePointer(h);
      return false;
    }
  }
  else {
    return false;
  }
}

/* URI related API */

char *IptelAddress::GetScheme() const
{
  if ( IsEmpty() ) 
    return NULL;

  sip_addr_t uri;
  char *oRet;

  uri = URI_ParseNew(Address);
  oRet =  uri.uri.scheme ? strdup(uri.uri.scheme) : NULL;
  free_valid(uri.display_name);
  URI_Free(&uri.uri);
  return oRet;
}

void IptelAddress::SetScheme(const char *s)
{
  sip_addr_t uri;

  uri   = URI_ParseNew(Address ? Address : (char *) "");
  if ( uri.uri.scheme ) {
    free(uri.uri.scheme);
  }
  uri.uri.scheme = s ? strdup(s) : NULL;

  if ( Address ) {
    free(Address);
  }

  Address = strdup((const char *) SipAddressToIptelAddress(uri));

  free_valid(uri.display_name);
  URI_Free(&uri.uri);
}


char *IptelAddress::GetUser() const
{
  if ( IsEmpty() ) 
    return NULL;

  sip_addr_t uri;
  char *oRet;

  uri = URI_ParseNew(Address);
  oRet =  uri.uri.user ? strdup(uri.uri.user) : NULL;
  free_valid(uri.display_name);
  URI_Free(&uri.uri);
  return oRet;
}


void IptelAddress::SetUser(const char *s)
{
  sip_addr_t uri;
  uri   = URI_ParseNew(Address ? Address : (char *)"");
  free_valid(uri.uri.user);

  uri.uri.user = s ? strdup(s) : NULL;

  if ( Address ) {
    free(Address);
  }

  Address = strdup((const char *) SipAddressToIptelAddress(uri));

  free_valid(uri.display_name);
  URI_Free(&uri.uri);
}

char *IptelAddress::GetURI() const
{
  DString str;
  if ( IsEmpty() ) 
    return NULL;

  sip_addr_t uri;
  char *oRet;

  uri = URI_ParseNew(Address);
  DStringInit(&str);
  URI_SPrint(&uri.uri, &str);
  oRet =  DStringValue(&str);
  /* Do not free str */
  free_valid(uri.display_name);
  URI_Free(&uri.uri);
  return oRet;
}

char *IptelAddress::GetDisplayName() const
{
  if ( IsEmpty() ) 
    return NULL;

  sip_addr_t uri;
  char *oRet;

  uri = URI_ParseNew(Address);
  oRet =  uri.display_name;
  /* Do not free display_name !! */
  URI_Free(&uri.uri);
  return oRet;
}


void IptelAddress::SetDisplayName(const char *s)
{
  sip_addr_t uri;
  uri   = URI_ParseNew(Address ? Address : (char *)"");
  free_valid(uri.display_name);

  uri.display_name = s ? strdup(s) : NULL;

  if ( Address ) {
    free(Address);
  }

  Address = strdup((const char *) SipAddressToIptelAddress(uri));

  free_valid(uri.display_name);
  URI_Free(&uri.uri);
}

char *IptelAddress::GetPassword() const
{
  if ( IsEmpty() ) 
    return NULL;

  sip_addr_t uri;
  char *oRet;

  uri = URI_ParseNew(Address);
  oRet =  uri.uri.password ? strdup(uri.uri.password) : NULL;
  free_valid(uri.display_name);
  URI_Free(&uri.uri);
  return oRet;
}


void IptelAddress::SetPassword(const char *s)
{
  sip_addr_t uri;

  uri   = URI_ParseNew(Address ? Address : (char *) "");
  if ( uri.uri.password ) {
    free(uri.uri.password);
  }
  uri.uri.password = s ? strdup(s) : NULL;

  if ( Address ) {
    free(Address);
  }

  Address = strdup((const char *) SipAddressToIptelAddress(uri));

  free_valid(uri.display_name);
  URI_Free(&uri.uri);
}

char *IptelAddress::GetHost() const
{
  if ( IsEmpty() ) 
    return NULL;

  sip_addr_t uri;
  char *oRet;

  uri = URI_ParseNew(Address);
  oRet =  uri.uri.host ? strdup(uri.uri.host) : NULL;
  free_valid(uri.display_name);
  URI_Free(&uri.uri);
  return oRet;
}

void IptelAddress::SetHost(const char *s)
{
  sip_addr_t uri;

  uri   = URI_ParseNew(Address ? Address : (char *) "");
  if ( uri.uri.host ) {
    free(uri.uri.host);
  }
  uri.uri.host = s ? strdup(s) : NULL;

  if ( Address ) {
    free(Address);
  }

  Address = strdup((const char *) SipAddressToIptelAddress(uri));

  free_valid(uri.display_name);
  URI_Free(&uri.uri);
}

int  IptelAddress::GetPort() const
{
  if ( IsEmpty() ) 
    return 0;

  sip_addr_t uri;
  int oRet;

  uri = URI_ParseNew(Address);
  oRet =  (uri.uri.port > 0) ? uri.uri.port : 0;
  free_valid(uri.display_name);
  URI_Free(&uri.uri);
  return oRet;
}


void IptelAddress::SetPort(int s)
{
  sip_addr_t uri;

  uri   = URI_ParseNew(Address ? Address : (char *) "");
  uri.uri.port = (s > 0 ? s : 0);

  if ( Address ) {
    free(Address);
  }

  Address = strdup((const char *) SipAddressToIptelAddress(uri));

  free_valid(uri.display_name);
  URI_Free(&uri.uri);
}

char *IptelAddress::GetPath() const
{
  if ( IsEmpty() ) 
    return NULL;

  sip_addr_t uri;
  char *oRet;

  uri = URI_ParseNew(Address);
  oRet =  uri.uri.path ? strdup(uri.uri.path) : NULL;
  free_valid(uri.display_name);
  URI_Free(&uri.uri);
  return oRet;
}

void IptelAddress::SetPath(const char *s)
{
  sip_addr_t uri;

  uri   = URI_ParseNew(Address ? Address : (char *) "");
  if ( uri.uri.path ) {
    free(uri.uri.path);
  }
  uri.uri.path = s ? strdup(s) : NULL;

  if ( Address ) {
    free(Address);
  }

  Address = strdup((const char *) SipAddressToIptelAddress(uri));

  free_valid(uri.display_name);
  URI_Free(&uri.uri);
}

bool IptelAddress::IsParameter(const char *pname) const
{
  if ( IsEmpty() || pname == NULL ) 
    return false;

  sip_addr_t uri;
  bool oRet;

  uri = URI_ParseNew(Address);
  oRet = (URI_Parameter(uri.uri, (char *) pname) == NULL) ? false : true;
  free_valid(uri.display_name);
  URI_Free(&uri.uri);
  return oRet;
}

char *IptelAddress::GetParameter(const char *pname) const
{
  if ( IsEmpty() || pname == NULL ) 
    return NULL;

  sip_addr_t uri;
  char *oRet;

  uri = URI_ParseNew(Address);
  oRet = URI_Parameter(uri.uri, (char *) pname);
  oRet =  oRet ? strdup(oRet) : NULL;
  free_valid(uri.display_name);
  URI_Free(&uri.uri);
  return oRet;
}

void IptelAddress::SetParameter(const char *pname,
                                const char *value)
{
  //XXX implement this method.
  //err << "IptelAddress::SetParameter() is not implemented" << endl;
}

/////////////////////// IP Address /////////////////////////

IptelIPAddress::IptelIPAddress()
{
  Address.sin_family = AF_INET;
  Address.sin_port   = 0;  /* 0 */
  Address.sin_addr.s_addr = INADDR_ANY; /* 0.0.0.0 */
}

IptelIPAddress::IptelIPAddress(unsigned char a,
               unsigned char b, 
               unsigned char c,
               unsigned char d,
               unsigned short port)
{
  Address.sin_family = AF_INET;
  Address.sin_port   = htons(port);
  Address.sin_addr.s_addr = IPADDRESS_ALL(a,b,c,d);
}

IptelIPAddress::IptelIPAddress(unsigned long ip,unsigned short port)
{
  Address.sin_family = AF_INET;
  Address.sin_port   = htons(port);
  Address.sin_addr.s_addr = ip;
}

IptelIPAddress::IptelIPAddress(unsigned char ip[4],unsigned short port)
{
  Address.sin_family = AF_INET;
  Address.sin_port   = htons(port);
  Address.sin_addr.s_addr = IPADDRESS_ALL(ip[0],ip[1],ip[2],ip[3]);
}

IptelIPAddress::IptelIPAddress(const sockaddr & saddr)
{
  Address = *((sockaddr_in*) &saddr);
}

IptelIPAddress::IptelIPAddress(const sockaddr_in & saddr)
{
  Address = saddr;
}

IptelIPAddress::IptelIPAddress(const struct in_addr & saddr_in)
{
  Address.sin_family = AF_INET;
  Address.sin_addr = saddr_in;
  Address.sin_port = 0;
}

IptelIPAddress::IptelIPAddress(const char *address)
{
  int ipa[4], port=0;

  /* XXX check if it has dots in the address.
   * Other wise a local extension (tel:) will
   * also be accepted.
   */
  if ( sscanf(address,
    "%d.%d.%d.%d:%d",&ipa[0],&ipa[1],&ipa[2],&ipa[3], &port) >= 4)
  {
    Address.sin_family = AF_INET;
    Address.sin_port = htons(port);
    Address.sin_addr.s_addr = IPADDRESS_ALL(ipa[0],ipa[1],ipa[2],ipa[3]);
  }
  else
  {
    Address.sin_family = AF_INET;
    Address.sin_port   = 0;  /* 0 */
    Address.sin_addr.s_addr = INADDR_ANY; /* 0.0.0.0 */
  }
}

IptelIPAddress::IptelIPAddress(const IptelAddress & address)
{
  IptelIPAddress::IptelIPAddress(address.GetAddressString());
}

IptelIPAddress::IptelIPAddress(const IptelIPAddress & address)
{
  Address = address.Address;
}

IptelIPAddress::~IptelIPAddress()
{
  /* Do nothing */
}

IptelIPAddress & IptelIPAddress::operator=(const IptelIPAddress & address)
{
  Address = address.Address;
  return *this;
}

IptelIPAddress & IptelIPAddress::operator=(const sockaddr & address)
{
  Address = *((sockaddr_in*) &address);
  return *this;
}

IptelIPAddress & IptelIPAddress::operator=(const sockaddr_in & address)
{
  Address = address;
  return *this;
}

IptelIPAddress & IptelIPAddress::operator=(const IptelAddress & address)
{
  IptelIPAddress ipaddr(address);
  Address = ipaddr.Address;
  return *this;
}

IptelIPAddress & IptelIPAddress::operator=(const char * address)
{
  IptelIPAddress ipaddr(address);
  Address = ipaddr.Address;
  return *this;
}

bool IptelIPAddress::operator==(const IptelIPAddress & address) const
{
  return (address.Address.sin_port == Address.sin_port) &&
    (address.Address.sin_family == Address.sin_family) &&
    (address.Address.sin_addr.s_addr == Address.sin_addr.s_addr);
}

IptelIPAddress::operator IptelAddress() const
{
  char *str = *this;
  IptelAddress addr(str);
  DeallocatePointer(str);
  return addr;
}

IptelIPAddress::operator sockaddr_in() const
{
  return Address;
}

IptelIPAddress::operator in_addr() const
{
  return Address.sin_addr;
}

IptelIPAddress::operator unsigned long() const
{
  return Address.sin_addr.s_addr;
}

unsigned char IptelIPAddress::operator[] (int index) const
{
  if ( index < 0 || index >=4 ) 
    return 0;
  switch(index) {
    case 0: return IPADDRESS_A1(Address.sin_addr.s_addr);
    case 1: return IPADDRESS_A2(Address.sin_addr.s_addr);
    case 2: return IPADDRESS_A3(Address.sin_addr.s_addr);
    case 3: return IPADDRESS_A4(Address.sin_addr.s_addr);
  }
  return 0;
}

IptelIPAddress::operator char *() const
{
  DString str;
  DStringInit(&str);
  DStringPrintf(&str,"%u.%u.%u.%u",
                IPADDRESS_A1(Address.sin_addr.s_addr),
                IPADDRESS_A2(Address.sin_addr.s_addr),
                IPADDRESS_A3(Address.sin_addr.s_addr),
                IPADDRESS_A4(Address.sin_addr.s_addr));
  if ( Address.sin_port != 0 ) {
    DStringPrintf(&str, ":%u", 
                  GetPort());
  }
  char *value = strdup(DStringValue(&str));
  DStringFree(&str);
  return value;
}

unsigned short IptelIPAddress::GetPort() const
{
  return ntohs(Address.sin_port);
}

void IptelIPAddress::SetPort(unsigned short port)
{
  Address.sin_port = htons(port);
}

int IptelIPAddress::GetIpAddress(sockaddr_in & saddr) const
{
  saddr = Address;
  return 0;
}

int IptelIPAddress::GetIpAddress(sockaddr & saddr) const
{
  saddr = *((sockaddr*) &Address);
  return 0;
}

int IptelIPAddress::GetIpAddress(unsigned char ip[4]) const
{
  ip[0] = IPADDRESS_A1(Address.sin_addr.s_addr);
  ip[1] = IPADDRESS_A2(Address.sin_addr.s_addr);
  ip[2] = IPADDRESS_A3(Address.sin_addr.s_addr);
  ip[3] = IPADDRESS_A4(Address.sin_addr.s_addr);
  return 0;
}

int IptelIPAddress::GetIpAddress(unsigned char & a, 
                            unsigned char & b,
                            unsigned char & c,
                            unsigned char & d) const
{
  a = IPADDRESS_A1(Address.sin_addr.s_addr);
  b = IPADDRESS_A2(Address.sin_addr.s_addr);
  c = IPADDRESS_A3(Address.sin_addr.s_addr);
  d = IPADDRESS_A4(Address.sin_addr.s_addr);
  return 0;
}

void IptelIPAddress::SetIpAddress(const sockaddr_in & saddr)
{
  Address = saddr;
  return;
}

void IptelIPAddress::SetIpAddress(const sockaddr & saddr)
{
  Address = *((sockaddr_in*)&saddr);
  return;
}

void IptelIPAddress::SetIpAddress(unsigned char ip[4])
{
  Address.sin_addr.s_addr = 
    IPADDRESS_ALL(ip[0],ip[1],ip[2],ip[3]);
  return;
}

void IptelIPAddress::SetIpAddress(unsigned char  a,
                            unsigned char b,
                            unsigned char c,
                            unsigned char d)
{
  Address.sin_addr.s_addr = 
    IPADDRESS_ALL(a,b,c,d);
  return;
}

void IptelIPAddress::SetIpAddress(unsigned long ip)
{
  Address.sin_addr.s_addr = ip;
  return;
}

bool IptelIPAddress::IsValid() const
{
  if ( IPADDRESS_A1(Address.sin_addr.s_addr) == 0)
  {
    /* 0.x.x.x  is invalid */
    return false;
  }
  else
  {
    /* Any other address is valid */
    return true;
  }
}

bool IptelIPAddress::IsMulticast() const
{
  return (((unsigned long)(Address.sin_addr.s_addr & 0xf0000000)) == 
    (unsigned long) 0xf0000000);
}

/////////////////// SESSION INFO //////////////////////////////

/* Formats access */
/*
 * Pay load table
 */
struct PayloadTable
{
  const char *name;
  int clockrate;
  int channel_count;
};
#define UNASSIGNED_PT {"unassigned", 0, 0}

/* Copied from RFC 1890 */
static struct PayloadTable ptTable[] =
{
  { "pcmu", 8000, 1 }, /* payloadtype = 0 */
  { "1016", 8000, 1 }, /* 1 */
  { "g721", 8000, 1 },
  { "gsm", 8000, 1 },  /* 3 */
  UNASSIGNED_PT,
  { "dvi", 8000, 1 },  /* 5 */
  { "dvi", 16000, 1 },
  { "lpc", 8000, 1 },
  { "pcma", 8000, 1 }, /* 8 */
  { "G722", 8000, 1 },
  { "L16", 44100, 2 },
  { "L16", 44100, 1 },  /* 11 */
  UNASSIGNED_PT,
  UNASSIGNED_PT,
  { "mpa", 90000, 1 },  /* 14 */
  { "G728", 8000, 1 },
  UNASSIGNED_PT,        /* 16 */
  UNASSIGNED_PT,
  UNASSIGNED_PT,
  UNASSIGNED_PT,
  UNASSIGNED_PT,
  UNASSIGNED_PT,
  UNASSIGNED_PT,
  UNASSIGNED_PT,        /* 23 */
  UNASSIGNED_PT,        /* 24 */
  { "CelB", 90000, 1 }, /* 25 */
  { "JPEG", 90000, 1 }, /* 26 */
  UNASSIGNED_PT,        /* 27 */
  { "nv",   90000, 1 }, /* 28 */
  UNASSIGNED_PT,        /* 29 */
  UNASSIGNED_PT,        /* 30 */
  { "H261", 90000, 1 }, /* 31 */
  { "MPV",  90000, 1 }, /* 32 */
  { "MP2T", 90000, 1 }, /* 33 */

  /* Add new here */

  /* Fill the payloadTable indexed to payloadtype */
};
static int    PayloadTableSize = (sizeof(ptTable)/sizeof(ptTable[0]));

#define INIT_ALL_FORMAT_PARAMS { \
    payloadtype = -1; \
    description = NULL;\
    clockrate = 0; \
    channelcount = 1; \
}

FormatInfo::FormatInfo()
{
  INIT_ALL_FORMAT_PARAMS;
}

FormatInfo::FormatInfo(int pt)
{
  INIT_ALL_FORMAT_PARAMS;
  if ( (pt >= 0 ) && (pt < PayloadTableSize) ) {
    payloadtype = pt;
    description = strdup(ptTable[pt].name);
    clockrate = ptTable[pt].clockrate;
    channelcount = ptTable[pt].channel_count;
  }
}

FormatInfo::FormatInfo(const char *mimetype)
{
  INIT_ALL_FORMAT_PARAMS;
  if ( mimetype == NULL ) {
    return;
  }
  if ( !strcasecmp(mimetype, "audio/basic") ) {
    payloadtype = 0;
    description = strdup("pcmu");
    clockrate = 8000;
    channelcount = 1;
  }
  /* Add other mime types here */
}

FormatInfo::FormatInfo(const char *def, int rate, int cnl)
{
  INIT_ALL_FORMAT_PARAMS;
  if ( def == NULL ) {
    return;
  }
  for ( int i=0; i<PayloadTableSize; i++) {
    if ( !strcasecmp(def, ptTable[i].name) 
         && ((rate == 0) || (rate == ptTable[i].clockrate) ) 
         && (cnl == ptTable[i].channel_count)) {
      payloadtype = i;
      break;
    }
  }
  description = strdup(def);
  clockrate = rate;
  channelcount = cnl;
}

FormatInfo::FormatInfo(int pt, const char *mimetype)
{
  INIT_ALL_FORMAT_PARAMS;
  payloadtype = pt;
  if ( mimetype == NULL ) {
    return;
  }
  if ( !strcasecmp(mimetype, "audio/basic") ) {
    description = strdup("pcmu");
    clockrate = 8000;
    channelcount = 1;
  }
}

FormatInfo::FormatInfo(int pt, const char *def, int rate, int cnl)
{
  INIT_ALL_FORMAT_PARAMS;
  payloadtype = pt;
  description = (def != NULL ? strdup(def) : NULL);
  clockrate = rate;
  channelcount = cnl;

  if ( (pt >= 0 ) && (pt < PayloadTableSize) ) {
    if ( description == NULL ) {
      description = strdup(ptTable[pt].name);
    }
    if ( clockrate == 0 ) {
      clockrate = ptTable[pt].clockrate;
    }
    if ( channelcount <= 1 ) {
      channelcount = ptTable[pt].channel_count;
    }
  }
}

FormatInfo::FormatInfo(const FormatInfo & format)
{
  payloadtype = format.payloadtype;
  //:-( DeallocatePointer(description);
  description = (format.description != NULL) ?
    strdup(format.description) : NULL;
  clockrate = format.clockrate;
  channelcount = format.channelcount;
}

FormatInfo::~FormatInfo()
{
  DeallocatePointer(description);
}

void FormatInfo::SetFormatName(const char *desc)
{
  
  DeallocatePointer(description);
  description = (desc != NULL ? strdup(desc) : NULL);
}

FormatInfo & FormatInfo::operator=(const FormatInfo & format)
{
  payloadtype = format.payloadtype;
  DeallocatePointer(description);
  description = (format.description != NULL) ?
    strdup(format.description) : NULL;
  clockrate = format.clockrate;
  channelcount = format.channelcount;
  return *this;
}

bool FormatInfo::operator==(const FormatInfo format) const
{
  if ( (format.payloadtype == -1) ||
       (payloadtype == -1) ||
       (format.description != NULL && description != NULL) ) {
    /* one of them has -1 as payload type (unset) or
     both have defined descriptions. */
    if ( (format.description == NULL) && (description == NULL)) {
      return true; /* Both are empty */
    }
    else if ( (format.description != NULL) && (description != NULL) &&
              !strcasecmp(format.description, description)) {
      /* description matches */
      if ( clockrate == 0 || format.clockrate == 0) {
        /* One of the clock rate is not set. Matched !! */
        return true;
      }
      else if ( clockrate == format.clockrate) {
        return ( channelcount == format.channelcount);
      }
      else {
        /* Clock rate are not matching */
        return false;
      }
    }
    else {
      return false; /* no description of one of them */
    }
  }
  else {
    return (payloadtype == format.payloadtype);
  }
}

//////////// MediaInfo //////////////////////

MediaInfo::MediaInfo()
{
  direction = MediaInfo::Receive;
  datatype  = MediaInfo::Audio;
  transport = MediaInfo::RTP_AVP_Transport;
  network   = MediaInfo::NetworkIN;
  addresstype= MediaInfo::AddressIP4;
  
  tx_bandwidth = 0;
  rx_bandwidth = 0;
  rx_ttl = 127; /* default */
}

MediaInfo::MediaInfo(MediaInfo::Direction dir,
                     MediaInfo::DataType dat,
                     MediaInfo::TransportType trans,
                     MediaInfo::NetworkType net,
                     MediaInfo::AddressType addr)
{
  direction = dir;
  datatype  = dat;
  transport = trans;
  network   = net;
  addresstype= addr;
  tx_bandwidth = 0;
  rx_bandwidth = 0;
  rx_ttl = 127; /* default */
}

MediaInfo::MediaInfo(const MediaInfo &media)
{
  direction = media.direction;
  datatype = media.datatype;
  transport = media.transport;
  formats = media.formats;
  addresstype = media.addresstype;
  network = media.network;
  tx = media.tx;
  rx = media.rx;
  rx_ttl = media.rx_ttl;
  tx_bandwidth = media.tx_bandwidth;
  rx_bandwidth = media.rx_bandwidth;
  headers = media.headers;
  parameters = media.parameters;
}

int MediaInfo::RemoveFormat(const FormatInfo & format)
{
  /* Search for the format and remove all 
  * matching formats. */
  while(true) {
    int index = GetFormatIndex(format);
    if ( index >= 0 ) {
      formats.RemoveAt(index);
    }
    else {
      break;
    }
  }
  return 0;
}

int MediaInfo::RemoveAllFormats()
{
  formats.RemoveAll();
  return 0;
}

int MediaInfo::GetFormatIndex(const FormatInfo & format) const
{
  int formatcount = formats.GetSize();
  for (int i=0; i<formatcount; i++) {
    if (formats[i] == format) {
      return i;
    }
  }
  return -1;
}

MediaInfo & MediaInfo::operator=(const MediaInfo & media)
{
  direction = media.direction;
  datatype = media.datatype;
  transport = media.transport;
  formats = media.formats;
  network = media.network;
  addresstype = media.addresstype;
  tx = media.tx;
  rx = media.rx;
  rx_ttl = media.rx_ttl;
  tx_bandwidth = media.tx_bandwidth;
  rx_bandwidth = media.rx_bandwidth;
  headers = media.headers;
  parameters = media.parameters;
  return *this;
}

bool MediaInfo::operator==(const MediaInfo & media) const
{
  return ((media.direction == MediaInfo::UnknownDirection) ||
          (media.direction == direction)) &&
         ((media.datatype == MediaInfo::UnknownData) ||
          (media.datatype == datatype)) && 
         ((media.transport == MediaInfo::UnknownTransport) ||
          (media.transport == transport)) &&
         ((media.network == MediaInfo::UnknownNetwork) ||
          (media.network == network)) && 
         ((media.addresstype == MediaInfo::UnknownAddress) ||
          (media.addresstype == addresstype));
}

///////////////////// Session Description ///////////////

SessionDescription::SessionDescription()
{
  sessionId = (SESSIONID) -1; /* some invalid session id */
}

SessionDescription::SessionDescription(const SessionDescription
                                       & session)
{
  sessions = session.sessions;
  tx = session.tx;
  rx = session.rx;
  sessionId = session.sessionId;
  headers = session.headers;
  //cout << "Session copied  to " << *this;
}

SessionDescription & SessionDescription::operator=(const SessionDescription
                                           & session)
{
  sessions = session.sessions;
  tx = session.tx;
  rx = session.rx;
  sessionId = session.sessionId;
  headers = session.headers;
  //cout << "Session assigned to " << *this;
  return *this;
}

void SessionDescription::DeleteSession(const MediaInfo & session)
{
  if ( session.GetDirection() == MediaInfo::UnknownDirection &&
       session.GetDataType() == MediaInfo::UnknownData ) {
    /* Remove all */
    sessions.RemoveAll();
  }
  else {
    /* Remove specific */
    int index = FindSessionIndex(session);
    if ( index >=0 ) {
      sessions.RemoveAt(index);
    }
  }
}

int SessionDescription::FindSessionIndex(const MediaInfo & session) const
{
  int formatcount = sessions.GetSize();
  for (int i=0; i<formatcount; i++) {
    if (sessions[i] == (const MediaInfo) session) {
      return i;
    }
  }
  return -1;  
}

bool SessionDescription::FindSession(MediaInfo & mediainfo,
      MediaInfo::Direction dir,
      MediaInfo::DataType  data,
      MediaInfo::TransportType transport) const
{ 
  int i;
  MediaInfo m(dir,data,transport);
  if ( (i = FindSessionIndex(m)) >=0 ) {
    mediainfo = sessions[i];
    return true;
  }
  else {
    return false;
  }
}

SESSIONID SessionDescription::GetSessionId()
{
  if ( sessionId == ((SESSIONID) -1)) {
    sessionId = (SESSIONID) rand();
  }
  return sessionId;
}

SessionDescription *SessionDescription::Clone()
{
  SessionDescription *s;
  s = AllocateObject(SessionDescription,1);
  s->tx = tx;
  s->rx = rx;
  s->sessionId = sessionId;
  s->sessions = sessions;
  return s;
}

/////////////////// IptelEndpoint ////////////////

IptelEndpoint::IptelEndpoint( bool IsServer, 
                   bool IsRegistrar )
{
  isserver = IsServer;
  isregistrar = IsRegistrar;
}

IptelEndpoint::IptelEndpoint( const IptelIPAddress & listening_address, 
                   bool IsRegistrar)
{
  isserver         = true;
  isregistrar      = IsRegistrar;
  ListeningAddress = listening_address;
}

IptelEndpoint::IptelEndpoint( bool IsServer, 
                   const IptelAddress   & registration_address)
{
  isserver         = IsServer;
  isregistrar      = true;
  RegistrationAddress = registration_address;
}

IptelEndpoint::IptelEndpoint( const IptelIPAddress & listening_address, 
                   const IptelIPAddress & registration_address)
{
  isserver         = true;
  isregistrar      = true;
  ListeningAddress = listening_address;
  RegistrationAddress = registration_address;
}

IptelEndpoint::~IptelEndpoint()
{
  /* Nothing */
}
    
int IptelEndpoint::GetRegistrationCount() const
{
  return 0; /* not implemented */
}

int IptelEndpoint::GetRegistration(IptelEndpoint::RegistrationInfo reginfo[],
                    int /* size */) const
{
  return -1; /* not implemented */
}
int IptelEndpoint::GetRegistrationCount(IptelAddress /*localName*/) const
{
  return -1; /* not implemented */
}
int IptelEndpoint::GetRegistration(IptelAddress /*localName*/,
                IptelEndpoint::RegistrationInfo reginfo[],
                int /*size*/) const
{
  return -1; /* not implemented */
}

int IptelEndpoint::OnIncomingRegistration(const IptelEndpoint::RegistrationInfo /* reginfo */)
{
  return -1; /* Not implemented */
}

void IptelEndpoint::OnIncomingUnregistration(const IptelEndpoint::RegistrationInfo /* reginfo */)
{
  return; /* Not implemented */
}

/* Ignore all registration and unregistratio request */
int IptelEndpoint::Register()
{
  //cout << "IptelEndpoint::Register[0] is not implemented" << endl;
  return -1;
}

int IptelEndpoint::Unregister()
{
  //cout << "IptelEndpoint::Unregister is not implemented" << endl;
  return -1;
}

int IptelEndpoint::Register(const IptelEndpoint::RegistrationInfo &)
{
  //cout << "IptelEndpoint::Register[1] is not implemented" << endl;
  return -1;
}

int IptelEndpoint::Unregister(const IptelEndpoint::RegistrationInfo &)
{
  //cout << "IptelEndpoint::Unregister is not implemented" << endl;
  return -1;
}

int IptelEndpoint::Register(const IptelAddress &)
{
  //cout << "IptelEndpoint::Register[2] is not implemented" << endl;
  return -1;
}

int IptelEndpoint::Register(const IptelAddress &,const IptelAddress &)
{
  //cout << "IptelEndpoint::Register[3] is not implemented" << endl;
  return -1;
}

int IptelEndpoint::Unregister(const IptelAddress &)
{
  //cout << "IptelEndpoint::Unregister is not implemented" << endl;
  return -1;
}

int IptelEndpoint::Unregister(const IptelAddress &,const IptelAddress &)
{
  //cout << "IptelEndpoint::Unregister is not implemented" << endl;
  return -1;
}

////////////////// IptelCall //////////////////////

IptelCall::IptelCall()
  : remote("unknown"),self("unknown")
{
  isactive = false;
  selfsession = NULL;
  remotesession = NULL;
  errorno = 0;
  bandwidth = 0;
}

IptelCall::~IptelCall()
{
  if ( selfsession ) {
    delete selfsession;
  }
  if ( remotesession ) {
    delete remotesession;
  }

  /* Nothing */
}

int  IptelCall::GetLastError()
{
  return errorno;
}

void IptelCall::SetLastError(int e)
{
  errorno = e;
}

void IptelCall::ClearError()
{
  errorno = 0;
}

const char *IptelCall::strerror(int e)
{
  switch(e)
  {
  #define MAPERRORSTRING(err,string) case err: return string;

  MAPERRORSTRING( IPTELERR_NO_ERROR,      "IPTEL-No error");
  MAPERRORSTRING( IPTELERR_NOT_SUPPORTED, "IPTEL-Not supported");
  MAPERRORSTRING( IPTELERR_MEMORY_FAILURE,"IPTEL-Memory problem");
  MAPERRORSTRING( IPTELERR_NETWORK_ERROR, "IPTEL-Network problem");
  MAPERRORSTRING( IPTELERR_PROTOCOL_ERROR,"IPTEL-Protocol error");
  MAPERRORSTRING( IPTELERR_INVALID_ADDRESS,"IPTEL-Ivalid address");
  MAPERRORSTRING( IPTELERR_OPENH323_INTF, "IPTEL-Error in OpenH323 interface");
  MAPERRORSTRING( IPTELERR_OPENH323_CORE, "IPTEL-Error in OpenH323 core");
  /* Add new values above this line */
  MAPERRORSTRING( IPTELERR_UNDEFINED,     "IPTEL-Error not defined");

  #undef MAPERRORSTRING

  default:
    return "IPTEL-Text error string not present for this error value";
  }
}

int IptelCall::SetSelfSessionDescription(SessionDescription *session)
{
  if ( selfsession != session ) {
    DeallocateObject(selfsession);
    selfsession = session;
  }
  return 0;
}

SessionDescription * IptelCall::GetSelfSessionDescription()
{
  return selfsession;
}

int IptelCall::SetRemoteSessionDescription(SessionDescription *session)
{
  if ( remotesession != session ) {
    DeallocateObject(remotesession);
    remotesession = session;
  }
  return 0;
}

SessionDescription * IptelCall::GetRemoteSessionDescription()
{
  return remotesession;
}


/********************* Debugging Methods *******************************
*/

ostream & operator<<(ostream & o,IptelAddress & addr)
{
  return o << addr.GetAddressString();
}

ostream & operator<<(ostream & o,IptelIPAddress & addr)
{
  IptelAddress ipaddr;
  ipaddr = addr;
  return o << ipaddr;
}

ostream & operator<<(ostream & o,FormatInfo &f)
{
  if ( f.GetPayloadType() != -1 ) {
    o << "(" << f.GetPayloadType() << ")";
  }
  if ( f.GetFormatName() != NULL ) {
    o << f.GetFormatName();
  }
  o << "/" << f.GetClockRate();
  if ( f.GetChannelCount() != 1) {
    o << "/" << f.GetChannelCount();
  }
  return o;
}
ostream & operator<<(ostream & o,MediaInfo &m)
{
  switch(m.datatype)
  {
  case MediaInfo::Audio: o << "Audio" ; break;
  case MediaInfo::Video: o << "Video" ; break;
  case MediaInfo::Application : o << "Application" ; break;
  case MediaInfo::Data : o << "Data" ; break;
  case MediaInfo::Control: o << "Control" ; break;
  default: o << "Type?" ; break;
  }

  o << " ";

  switch(m.direction) 
  {
  case MediaInfo::Transmit: o << "Tx"; break;
  case MediaInfo::Receive: o << "Rx"; break;
  case MediaInfo::TransmitAndReceive : o << "Tx&Rx"; break;
  default: o << "Dir?" ; break;
  }

  IptelIPAddress ipaddr = m.GetRxAddress();
  
  if ( ipaddr.IsValid())
  {
    o << ipaddr << " ";
  }

  for ( int i=0; i<m.GetFormatCount(); i++) {
    FormatInfo & format = m[i];
    o << format;
  }
  return o;
}

ostream & operator<<(ostream & o,SessionDescription &s)
{
  SessionDescription *session = &s;

  if ( session == NULL )
  {
    o << "SessionDescription ?" << endl;
  }
  else
  {
    for ( int i=0; i<s.GetMediaCount(); i++) {
      MediaInfo & media  = s[i];
      o << media << "; ";
    }
  }
  return o;
}

ostream & operator<<(ostream & o,IptelEndpoint &ep)
{
    o << "Endpoint: " 
      << ep.LocalName << " "
      << (ep.isserver ? "Server" : "Client") << ","
      << (ep.isregistrar ? "Registrar, " : " ") 
      << "Listening at " << ep.ListeningAddress << " "
      ;
    return o;
}

ostream & operator<<(ostream & o,IptelCall &c)
{
  o << "Self: " << c.self << "; " 
    << "Remote:" << c.remote << "; "
    << (c.isactive ? "Active" : "Idle") << "; ";
  return o;
}

/************************** END OF FILE ********************************
*/
