/***************************** FILE HEADER *****************************
#               Copyright (c) 1999-2000 by Columbia University         *
#                        All Rights Reserved                           *
#***********************************************************************
#***********************************************************************
#
#  File Name    : sipapi.h
#
#  Module       : SIP API specification
#
#  Author(s)    : Kundan Singh (kns10@cs.columbia.edu)
#
#  Created      : 14 Oct 1999
#
#  Description  : Header for SIP C++ library
#                
#  Notes        : 
#
#***********************************************************************
#                Modification History (Reverse Order)
#-----------------------------------------------------------------------
#   Date         Author              Description
#-----------------------------------------------------------------------
# $Log: sipapi.h,v $
# Revision 1.9  2001/01/07 20:57:48  kns10
# Lots of bug fixes [kns]
#
# Revision 1.8  2000/12/17 20:30:30  kns10
#
# Few bug fixes.
#
# [kns]
#
# Revision 1.7  2000/11/15 18:39:43  kns10
#
# Changes in libsip++. Mostly related to
# (1) Authentication (Basic and Digest) in API for INVITE
#     (incoming and outgoing) and
#     REGISTER (outgoing only).
# (2) Send Contact in 200 OK of INVITE, always.
# (3) Work with non-5062 port for SIP listening object.
# (4) Other cosmetic chages: Build distribution, add Path in
#      IptelAddress for rtsp/http like URLs, etc.
#
# [kns]
#
# Revision 1.6  2000/09/26 20:04:22  kns10
#
# Implemented registration refresh logic for UA.
# (strptime issue not yet resolved)
#
# module.mk : changed MKDIST to MKDIST_WIN32 for win32 dist
# sipapi.h : Registration callbacks
# sipintf.cpp : Registration refresh logic.
# sipreg.cpp : Registration refresh logic.
# user.cpp : Registration refresh logic.
#
# [kns]
#
# Revision 1.5  2000/09/18 19:23:35  kns10
#
# Restructured the libsip++. Some of the *big* files are split into
# multiple files. user.c is renamed as user.cpp. Useless files are
# removed.
#
# [kns]
#
#
# Modified Files:
#  	iptelapi.cpp iptelapi.h libsipapi.dsp module.mk sipapi.cpp
#  	sipapi.h sipintf.cpp
# Added Files:
#  	README ads.cpp ads.h footer.html sipcall.cpp sipcall.h
#  	sipreg.cpp sipsdp.cpp user.cpp
# Removed Files:
#  	ccppmap.cpp ccppmap.h user.c
#
#-----------------------------------------------------------------------
*/

#ifndef _SIP_API_H_
#define _SIP_API_H_

/**@name SIP Library API, derived from Generic API. */
//@{

#include <stdio.h>    /* FILE * */
#include "iptelapi.h"
#ifdef MEDIACONTROL_SIPCOMM			/* MVakil 11-17-2000 */
#include "IMediaControl.h"     /* MVakil 9-30-2000 for MediaControl */
#endif

/** 
 * This Class represents the library. 
 * Global attributes, for example, Trace enable, timeouts etc
 * can be set using this class.
 */
class SIPLibrary
{
  public:
    /** 
     * Library initializer.
     * It initializes sockets, host name, IP address etc.
     * This should be the first method to be called in the library.
     * This is automatically called when an endpoint
     * is created.
     *
     * @return 0 on success and -1 on failure.
     */
    static int Initialize();   

    /** Timeouts in seconds for request, responses */
    static int GeneralTimeout; /* General Timeout value in seconds */

    /** Timeout in seconds for CGI script execution */
    static int CgiTimeout;     /* CGI timeout in seconds */

    /** Trace level. 0 is no trace, 3 is max info */                           
    static int TraceLevel;     /* Set/Get the trace level. 
                                * 0 is no trace, max is 3 */

    /** Trace file. NULL disables the trace. default is stderr
     Applications should not use this variable. */
    static FILE *TraceFile;    /* Fileptr for trace info.
                                * You can use NULL to disable.
                                * default is stderr. */

    /** Set the debug info file. Default is stdout/stderr. 
     * Call to this functional automatically sets
     * debug_flag to ON. */
    static void DebugOn(const char * filename);

    /* debug_flag to ON without changing debug file. */
    static void DebugOn();

    /** debug_flag to Off. */
    static void DebugOff();
    
    /** Set the error info file. You can specify it to be "stderr"
     * to printout errors on stderr. */
    static void ErrorOn(const char * filename);

    /** Set the file for printing out message flow. NULL disables it
     */
    static void MessageFlowOn(FILE *fp);

    /** Debug() print the trace information */
    static void Debug(char *function_name, char *format, ...);

    /** Fatal() prints the Fatal error information and aborts */
    static void Fatal(char *function_name, char *format, ...);

    /** Error() prints the error information */
    static void Error(char *function_name, char *format, ...);
    
    /** Number of endpoints instantiated */
    static int EndpointCount;

    /** Number of calls instantiated */
    static int CallCount;

    /** Use Numeric (dotted decimal) address for HostName */
    static bool oUseNumericAddress;

    /** Hostname of local host. This should not be modified by appln */
    static char *HostName;      /* Host name */

    /** IPv4 address of local host. This should not be modified by appln. */
    static struct in_addr HostInAddr;  /* default Net Interface address */

    /** Interface to a generic Host2IP function */
    static struct in_addr Host2IP(const char *host);

    /**Check the license file.
     Returns 0 if checked successfully.
     Returns > 0 if license if not valid, expired, or no license file 
                 specified.
     Returns < 0 if licensing is disabled. */
    static int CheckLicense(const char *filename);

    /**Sets the name of the application. This name is used
     as ServerName in SIP messages. */
    static void SetName(const char *name);

    /**Random number generator */
    static unsigned long Random32();

    /**User agent tag. This is used in To header for response.
     * The motivation to make is accessible to API is to
     * allow restoration by the application after a crash
     * recovery. A simple application does not need to look
     * into this paramter. */
    static char * UAID;
};

/*
 * IptelAddress, SIPMediaInfo and SIPSessionDescription are
 * not required 
 */

// Forward declaration
class SIPCall;
class SIPMsgEngine;
class SIPMessage;

/*
 * SIPRegistrationInfo adds over the IptelRegistrationInfo.
 * This includes some SIP specific stuff, e.g., 
 * address of registering party, expiry, callid etc.
 */
class SIPRegistrationInfo : public IptelEndpoint::RegistrationInfo
{
public:
  /** Default constructor, uses SOCK_STREAM (TCP) */
  SIPRegistrationInfo();

  /** Copy constructor */
  SIPRegistrationInfo(const SIPRegistrationInfo &r);

  /** Default destructor */
  ~SIPRegistrationInfo();

  /**Copy operator */
  SIPRegistrationInfo & operator=(const SIPRegistrationInfo &r);

  /**comparision */
  bool operator==(const SIPRegistrationInfo &r) const;

  /*
   * All these parameters are public, to provide easy access
   * from the C code in the library.
   */

  /** Address of registering party. If absent then same as localName */
  IptelAddress fromAddress;
  /** Duration of the registration. Used for incoming registration. */
  long         expiry; /* Duration for registration */
  /** Preference value, 1.0 is most preferred, 0 for least. */
  float        q_value;
  /** protocol specific registration identifier. */
  char         *callid; 
  /** transport protocol to be used, SOCK_STREAM (TCP), or SOCK_DGRAM (UDP)*/
  int          type; /* socket type */
  /** cseq number, not of particular interest to API */
  unsigned long lastcseq; 
  /** Any other protocol specific information */
  void         *info; 
   
  /** Outbound proxy */
  IptelAddress  outbound_proxy;

  /**Authorization header */
  char         *authorization;
};

/**SIPRegistrationInfoList is a list of SIP registrations */
DEFINE_AS_LIST(SIPRegistrationInfoList, SIPRegistrationInfo);


/**
 * SIP Message Content: This class defines a typical *non-SDP*
 * message body. SDP message bodies should be accessesd
 * using Get/Set[Self|Remote]SessionDescription() methods.
 * This is used for sending or receibing non-SDP
 * contents in 3xx-6xx responses.
 */
class SIPMessageBody 
{
 public:
  /** Default constructor */
  SIPMessageBody();

  /** Construct a valid content */
  SIPMessageBody(const char *content,
                 const char *content_type        = "text/plain",
                 int   content_length            = 0,
                 const char *content_disposition = "render",
                 const char *content_handling    = "optional");
  /** Copy constructor */
  SIPMessageBody(const SIPMessageBody & body);

  /** Default destuctor */
  ~SIPMessageBody();

  /** Copy operator */
  SIPMessageBody & operator=(const SIPMessageBody & body);

  /** Member access operators */
  void SetContent(const char *body);
  void SetContentType(const char *type);
  void SetContentLength(int length) { content_length = length; }
  void SetContentDisposition(const char *disposition);
  void SetContentHandling(const char *handling);
  const char *GetContent() const { return content; }
  int   GetContentLength() const { return content_length; }
  const char *GetContentType() const { return content_type; }
  const char *GetContentDisposition() const { return content_disposition; }
  const char *GetContentHandling() const { return content_handling; }

 private:
  char * content;
  int    content_length;
  char * content_type;
  char * content_disposition;
  char * content_handling;
};

/**
 * This represents a SIP endsystem (entity), for example, a SIP
 * user agent. 
 *
 * Every application will need to instantiate an
 * object of this type or a derived object of this class.
 * Example, in a signaling gateway, the SIP side can
 * be represented by an object of this type.
 *
 * @see #IptelEndpoint#.
 */
class SIPEndpoint : public IptelEndpoint
{
  public:
    /** Constructor.
     * Registrar is currently not implemented and always has Server.
     *
     * @param IsServer should be true.
     * @param IsRegistrar should be false.
     */
    SIPEndpoint( bool IsServer    = false, 
                 bool IsRegistrar = false );
    SIPEndpoint( const IptelIPAddress & listening_address, 
                 bool IsRegistrar = false);
    /*
    SIPEndpoint( bool IsServer, 
                 IptelAddress   & registration_address);
    SIPEndpoint( IptelIPAddress & listening_address, 
                 IptelIPAddress & registration_address);
    */

    /** Destroy the SIP endpoint */
    virtual ~SIPEndpoint();
    
    /**
     * Allocate a new callobject of type SIPCall.
     * The derived classes should return the appropriate
     * call object, for example, MySIPCall. This function
     * is used to allocate a new call object on an incoming
     * call.
     * @param callid identifies the call. If this is NULL
     *  then create a new call, otherwise represents an
     *  existing call, e.g., for new call leg in a conference.
     */
    virtual IptelCall *AllocateNewCallObject(const char *callid);

    /**@name Registration */
    //@{
    /** 
     * Register with the remote SIP registrar.
     *
     * @param reginfo specifies the contact address,
     * registring address, registrar address, q value and
     * so on.
     *
     * @return 0 on success and -1 on failure.
     * 
     * @see #Unregister#
     */
    virtual int Register(
      const SIPRegistrationInfo & reginfo);

    /**
     * Register with a sip-cgi script.
     *
     * @param reginfo specifies the contact address,
     * registering address, registrar address, q value
     * and so on.
     * @param cgi_file specifies the file name of the CGI script.
     * If this is NULL then no cgi-script is uploaded, only normal
     * registrartion is performed.
     * @param mimetype optional parameter specifies the
     * mime type for the cgi script. Default is 
     * application/x-perl. Note that the script is assumed to be
     * a text file.
     */
    virtual int Register(
      const SIPRegistrationInfo & reginfo,
      const char *cgi_file,
      const char *mime_type = "application/x-perl");

    /**
     * Register with a given user name (sipurl).
     */
    virtual int Register(const IptelAddress & user);

    /**
     * Register with the given user name and contact address.
     */
    virtual int Register(const IptelAddress & user,
                         const IptelAddress & contact);

    /**
     * Unregister with the remote SIP registar.
     * Unregistration is implemented using
     * REGISTER with Expires 0. 
     *
     * @param reginfo contains
     * the registered address, and registrar address.
     *
     * @return 0 on success and -1 on failue.
     *
     * @see #Register# 
     */
    virtual int Unregister(
      const SIPRegistrationInfo & reginfo);

    /**
     * Callback to indicate that the registration is 
     * successful 
     */
    virtual int OnRegistrationSuccess();


    /**
     * Callback to indicate that the registration is 
     * failed 
     */
    virtual int OnRegistrationFailed(int status=0, const char *reason=NULL);


    /**
     * Callback to indicate that the registration is 
     * expired and could not be refreshed. 
     */
    virtual int OnRegistrationExpired();


    /**
     * Callback to indicate that authorization is needed for
     * outgoing registration.
     * @param method is "basic" or "digest"
     * @param realm is the WWW-Authenticate realm field.
     * @userid is set by the application if available.
     * @password is set by the application if available.
     * Both userid and password are freed by the caller.
     * @return 0 on success and -1 on failure (do not have
     * password for the realm).
     */
    virtual int SIPEndpoint::OnAuthorizationNeeded(const char *method,
                                       const char *realm,
                                       char *& userid,
                                       char *& password);

    /** 
     * Callback for incoming REGISTER message.
     * This is a callback invoked by the stack to inform the
     * application about a new incoming registration.
     * This is called only when SIPEndpoint is a registrar.
     * Default behaviour for a registrar accepts the registration.
     *
     * @param reginfo gives the registring name,
     * contact address, q value,and so on for incoming
     * registration.
     * 
     * @return  Implementation should return
     *   >=0 for accepting and <0 for rejecting the registration.
     */
    virtual int OnIncomingRegistration(
      const SIPRegistrationInfo & reginfo);
    //@}

    /**@name Authorization related methods  */
    //@{
    /**
     * Set the authorization header for next outgoing message,
     * e.g., for registrartion. This is used for basic authorization.
     */
    virtual void SetAuthorization(const char *user, const char *passwd);

    /**
     * This method build the basic authorization header
     * from the password and user name.
     */
    static char * BuildAuthorizationHeader(const char *user,
                                           const char *passwd);
    //@}

    /**@name Session description related methods */
    //@{
    /**
     * Get the default SessionDescription.
     * Library maintains a default session description.
     *
     * @return returns the session description
     * for a default session, for example, PCMU 8kHz.
     */
    static SessionDescription * GetDefaultSessionDescription();

    /**
     * Set the default session description.
     * The default session description is unique for the whole application.
     * This is used for responding to incoming OPTIONS
     * methods. Note that by default the library defines a
     * default session description, for calling this method
     * bye the application is optional.
     *
     * @param session SDP to be set as default.
     */
    static void SetDefaultSessionDescription(SessionDescription &session);

    /**
     * Get the session description of the request URI
     * identified by to address, without establishing
     * the SIP call. This uses the OPTIONS
     * method to retrieve the remote SDP, if possible.
     * @param self indicates the local name to be used.
     * This is optional.
     * @param remote is the remote URI and the To address.
     */
    virtual SessionDescription *GetRemoteSessionDescription(
                 const IptelAddress & remote,
                 const IptelAddress & self);
    //@}

    /**@name Methods for generic messages like MESSAGE, DO, SUBSCRIBE.*/
    //@{
#if 0
    /** 
     * This callback is called to allocate a new message engine
     * associated with a particular SIP method, and/or callid.
     * This should be implemented by the application to
     * return the appropriate SIPMsgEngine (or derived class)
     * object for the appropriate method (MESSAGE, DO, SUBSCRIBE,..)
     * @return a pointer to the newly allocated SIPMsgEngine.
     * The default implementation returns NULL.
     */
    virtual SIPMsgEngine *AllocateNewMessageEngine(const char *method,
                                                   const char *callid);

    /**
     * Call this method to deallocate a particular
     * SIPMsgEngine. Application must implement this
     * to delete the SIPMessageEngine object.
     * The library could have called delete operator on
     * the object, but since some future implementation
     * may want to keep the object this
     * callback is needed.
     */
    virtual void DeallocateSIPMsgEngine(SIPMsgEngine *);
#endif
    //@}

    /**@name Other misc methods, used internally */
    //@{
    /**
     * Set the outgoing header for requests, e.g. REGISTER.
     */
    virtual void SetOutgoingRequestHeader(const char *h, const char *v,
                                          bool append = false);

    //@}

  protected:
    /* generic info */
    //IptelEndpoint::isserver
    //IptelEndpoint::isregistrar 

    /* Server related data */
    //IptelEndpoint::ListeningAddress;

    /* Registration related info */
    //IptelEndpoint::RegistrationAddress;
    //IptelEndpoint::LocalName;
    //IptelEndpoint::Registrar; 
    //IptelEndpoint::contactAddress;
    //IptelEndpoint::reg_list,reg_count;
    pthread_mutex_t  mutex;  /* For muti-thread protection */
    unsigned long lastcseq;
    
  public:
    void Lock();
    void Unlock();
    GenericHeaders h_orq;  /* outgoing request headers */

  private:
    pthread_t tid[2]; /* Listening threads for TCP and UDP */
    SIPRegistrationInfoList regs; /* list of outgoing registration */
    /* To search an existing reg or a new registration */
    SIPRegistrationInfo & CreateRegistration(
                          const SIPRegistrationInfo & reginfo);
    static SessionDescription *default_session;
    char *userid; /* For authorization */
    char *passwd;
};


/**
 * A SIP call.
 * This represents a SIP call between self and remote.
 * Various methods and attributes allow changing call
 * state and retrieving call information.
 * A call is associated with an #IptelEndpoint#. It is
 * required that the endpoint be of type
 * #SIPEndpoint# or a derived class.
 */
class SIPCall : public IptelCall
{
  public:
    /**
     * Default constructor.
     * This attaches the SIP call with a default #SIPEndpoint#.
     */
    SIPCall(const char *callid = NULL);

    /** 
     * Create a new call for the endpoint.
     * This attaches the SIP call with the endpoint
     * specified. Some of the attributes can be
     * borrowed (for example, self name) from the
     * associated endpoint.
     */
    SIPCall(SIPEndpoint & ep, const char *callid = NULL);

    /** Destructor to destroy the object */
    virtual ~SIPCall();

    /**@name Call state modification and indication */
    //@{
    /**
     * Place an outgoing call with already set
     * self and remote addresses.
     *
     * @return 0 on successful initiation and -1 on failure.
     * Not that even if the method returns 0 the actual call may fail,
     * and OnCallRejected or OnCallEstablished will be called as 
     * appropriate.
     */
    virtual int   Initiate();

    /**
     * Place an outgoing call. 
     * To initiate a new outgoing call with the given
     * remote and self addresses. 
     *
     * @param to Specifies the self (from) address. This is of the
     * form of "From:" SIP header.
     * @param from specifies the remote (to) address.
     * This is of the form of "To:" SIP header.
     *
     * @return 0 on successful initiation and -1 on failure. 
     */
    virtual int   Initiate(const IptelAddress & to,
                           const IptelAddress & from);

#if 0 /* Currently this is not supported */
    /**
     * Place an outgoing call to any of the given addresses.
     * All the addresses are tried in parallel as specified by
     * forking proxy behaviour. The q values, if present 
     * are used to choose the best response.
     * Note that once the call is accepted by one of the
     * branch all other branches are cancelled.
     *
     * @param to specifies a list of destinations to be tried in parallel.
     * Individual addresses in to list are of the form of 
     * "Contact:" SIP header. Note that only "sip:" URI schemes
     * are used. Others are ignored.
     * 
     * @return 0 on successful initiation and -1 on failure.
     */
    virtual int   Initiate(const IptelAddressList & to);
#endif

    /**
     * Terminate a call.
     *
     * @param reason is the status code.
     * @param strReason is the reason string. 
     *
     * @return 0 on success and -1 on failure.
     */
    virtual int   Hangup(
          int reason = 0, /* 0 for normal termination */
          const char *strReason = "Normal Termination");

    /**
     * Indicate that the local phone is ringing to remote.
     * This sends 180 Ringing response to remote for
     * an incoming call.
     *
     * @return 0 on success and -1 on failue.
     */
    virtual int    Alert();
    
    /**
     * Accept an incoming call.
     *
     * @return 0 on success and -1 on failure.
     */
    virtual int    Accept();

    /**
     * Reject an incoming call.
     *
     * @return 0 on success and -1 on failue.
     */
    virtual int    Reject();

    /**
     * Reject an incoming call with appropriate reason
     * 
     * @param status is the status code and .
     * @param reason is the reason string to be  sent
     *        in the SIP response to the caller.
     *
     * @return 0 on success and -1 on failue.
     */
    virtual int    Reject(int status,const char *reason=NULL);

    /**
     * Redirect an incoming call with appropriate reason.
     * 
     * @param newaddress is the "Contact:" address of the new
     * destination where you want the call to be redirected.
     *
     * @param reason specifies the reason for redirection.
     * If this is NULL then default is "Redirect".
     *
     * @return 0 on success and -1 on failure.
     */
    virtual int Redirect(const IptelAddress newaddress,
                         const char *reason=NULL);

    /**
     * Authenticate a particular call with Bappropriate Authentication
     * Methods used are Basic, and Digest. This call does not reject the
     * incoming call, but just checks if the authentication/Authorization
     * headers are valid or not. It is upto the application
     * to accept or reject a call.
     *
     * @param method is type of authentication needed."basic" or "digest"
     *
     * @param userfile is the password file for basic and digest auth.
     *
     * @return true on success and false on failure.
     */
    virtual bool Authenticate(const char *method,
                              const char *userfile);

    /** 
     * Set BASIC authorization header. This will be used for
     * subsequent requests for this call.
     * 
     * @param user is the user id
     *
     * @param passwd is the password to be used for basic authentication.
     */
    virtual void SetAuthorization(const char *user, const char *passwd);

    /**
     * Callback when an Authorization is needed for outgoing call.
     * Called when 401 is received in response to an
     * outgoing INVITE (or other call related method).
     * Default implementation sets the user and pass if
     * it is available (set using SetAuthorization earlier).
     *
     * @param method is basic or digest
     * @param realm is the authorization realm.
     * @param user should be set by the implementation.
     * @param pass should be set by the implementation.
     * @return returns 0 on success and -1 if user and pass are
     * not set or not known.
     */
    virtual int SIPCall::OnAuthorizationNeeded(const char *method,
                                               const char *realm,
                                               char *& user,
                                               char *& pass);

    /**
     * Callback when a ProxyAuthorization is needed for outgoing call.
     * Called when 407 is received in response to an
     * outgoing INVITE (or other call related method).
     * Default implementation sets the user and pass if
     * it is available (set using SetAuthorization earlier).
     *
     * @param method is basic or digest
     * @param realm is the authorization realm.
     * @param user should be set by the implementation.
     * @param pass should be set by the implementation.
     * @return returns 0 on success and -1 if user and pass are
     * not set or not known.
     */
    virtual int SIPCall::OnProxyAuthorizationNeeded(const char *method,
                                               const char *realm,
                                               char *& user,
                                               char *& pass);

    /**
     * Callback on new incoming call.
     * The library indicates an incoming call to the
     * application by calling this method with the
     * associated IptelEndpoint. The IptelEndpoint
     * muct have been created as a server listening
     * for incoming calls. Application should Accept or
     * Reject the incoming call either in the implementation
     * of this call back function or in a separate thread.
     * Default behaviour accepts any incoming call.
     * 
     * @param ep is the associated endpoint object
     * with the call.
     *
     * @see #Accept#, #Reject#.
     */
    virtual void   OnNewIncomingCall(IptelEndpoint & ep);

    /**
     * Indication that remote phone is ringing.
     * When a 180 response is received the Library 
     * calls this callback to indicate remote ringing to the
     * application. Application can generate local
     * ring tones as appropriate.
     * Default behaviour does nothing.
     */
    virtual void   OnAlerting();

    /**
     * Waiting for a final response from remote for outgoing call.
     * Application can call this method to block till 
     * a final response is received for an outgoing call.
     *
     * @param timeout in seconds can be specified. Default is -1 for
     * INFINITE timeout.
     */
    virtual int    WaitForResponse(int timeout = -1);

    /**
     * Callback on successful call establishment.
     * The Library calls this method to indicate successful call
     * establishment. Default behaviour does nothing.
     */
    virtual void   OnCallEstablished();

    /**
     * Callback on call rejected by remote.
     * The Library calls this method to indicate that the
     * remote has rejected your call.
     * @return 0 if you want the library to free the call object.
     * otherwise return non-zero if your application is responsible
     * for freeing the call object.
     */
    virtual int   OnCallRejected();

    /**
     * Callback on call hung up.
     * The library calls this method to indicate that 
     * an ongoing call was hungup by remote.
     * @return 0 if you want the library to free the call object.
     * otherwise the application is responsible for
     * freeing the call object.
     */
    virtual int   OnHangup();

    /**
     * Block till call is terminated.
     * Application can call this method to block till
     * the call is terminated by remote. Note that
     * if you are using this method then OnCallRejected and
     * OnHangup should return non-zero to avoid bus-error.
     *
     * @param timeout in seconds can be specified.
     * Default is -1 for INFINITE timeout.
     *
     * @return 0 on success and -1 on failure.
     */
    virtual int    WaitForTermination(int timeout = -1);

    /**
     * This callback is called on error.
     * Application can act on the error, for example,
     * by printing the error message and terminating the call.
     * 
     * @param errorno indicates the error which occurred.
     * @see IptelCall::strerror.
     */
    virtual void   OnError(int errorno);

    /**
     * Is it an outgoing or incoming call.
     * @return true if this is the originator of the call.
     * false otherwise.
     */
    virtual bool   IsOriginator() const { return isoriginator;}

    /** 
     * Return the callid string.
     * Callid string is useful for signaling gateway
     * applications where you want to map
     * the SIP callid with the PSTN call, for example.
     *
     * @return returns the call id string.
     */
    virtual const char *GetCallId() const {return callid; }

    /**
     * Sends INVITE message to the same destination.
     * Typically used when session description is changed
     * during a call, for example in H323v1.0->GW->SIP.
     * Note that  unlike forking of INVITE in earlier request,
     * the re-INVITE is sent to only one destination.
     * This can be called either by the originator or the
     * receiver of the original INVITE request.
     *
     * @return 0 on success and -1 on failure.
     */
    virtual int  Reinvite();

    //@}

    /**@name Session Management */
    //@{
    /** 
     * Formats an SDP packet and return the formatted
     * object. The length stores the length of the
     * returned string. Application is responsible for
     * freeing the memory of the returned string.
     *
     * @param length If length is not null, it stores the 
     * length of the text string returned excluding the
     * terminating '\0'.
     * 
     * @return A null ('\0') terminated SDP text string. Application
     * should free the memory for this string.
     */

    static char *FormatSDP(SessionDescription & session, int *length = NULL); 

    /** 
     * Deformats an SDP packet into a SessionDescription Object.
     * Deformats the SDP packet from the text
     * buffer passed. If length is <= 0 then
     * sdp is assumed to be nul terminated, otherwise
     * length is the length of the string pointed to by
     * sdp without the terminating '\0'. It is assumed that
     * session is empty.
     *
     * @param sdp SDP string to be deformatted.
     *
     * @param length Length of SDP string.
     * 
     * @return the SessionDescription pointer.
     */
    static SessionDescription * DeformatSDP(const char *sdp,int length); 
    //@}

    /**@name Misc methods */
    //@{
    /**
     * Set the address of the default outbound proxy.
     * All outgoing calls are sent to this address
     * irrespective of the specified destination address.
     *
     * @param proxy specifies the proxy address.
     *
     * @return 0 on success and -1 on failure.
     */
    virtual int    SetProxyAddress(const IptelAddress & proxy);

    /** 
     * Retrieve the proxy address.
     *
     * @return The proxy address set earlier by SetProxyAddress.
     */
    virtual IptelAddress     GetProxyAddress();
    /**
     * Setting the requestURI. 
     * Default is SIP-URL part of To address.
     * @return 0 on success and -1 on failure.
     */
    virtual int  SetRequestURI(const IptelAddress & requri);

    /** 
     * Retrieving the request URI.
     * @return The request URI.
     */
    virtual IptelAddress     GetRequestURI();

    /**
     * Set the self address.
     */
    virtual int  SetSelfAddress(const IptelAddress & self);

    /**
     * Set the remote address.
     */
    virtual int  SetRemoteAddress(const IptelAddress & remote);

    /**
     * Get the local address.
     */
    virtual IptelAddress GetSelfAddress();

    /** 
     * Get the remote address.
     */
    virtual IptelAddress GetRemoteAddress();

    /** 
     * Retrieve the SessionDescription (local).
     */
    virtual SessionDescription *GetSelfSessionDescription();

    /** 
     * Retrieve the SessionDescription (remote).
     */
    virtual SessionDescription *GetRemoteSessionDescription();

    /**
     * Set the local Session Description.
     */
    virtual int SetSelfSessionDescription(SessionDescription * s);

    /**
     * Set the remote session description.
     */
    virtual int SetRemoteSessionDescription(SessionDescription * s);

    /** 
     * Set the content type and the content body if different
     * from default (application/sdp).
     * @param contentType application/sdp, application/html.
     * @param contentBody appropriate body, NULL terminated string.
     */
     virtual void SetContentInfo(const char * contentType,
                                 const char * contentBody);

    /** Set the status value and string to be used for next response */
    void SetStatus(int status, const char *string);

    /**Get the status value */
    int  GetStatus();
    
    /** Get the status string */
    const char *GetStatusString();

    /** Set the subject header */
    void SetSubject(const char *subject) {
      SetOutgoingRequestHeader("Subject", subject); }

    /** Retrieve the subject header, return NULL if none. */
    const char *GetSubject() { 
      const char *s = GetIncomingRequestHeader("Subject");
      return s ? s : GetOutgoingRequestHeader("Subject"); }

    /** 
     * Get the header value.
     * @param h header type, for example, From, To, etc.
     * @return Returns the associated header value from
     * the last request or response.
     */
    virtual const char *GetHeader(const char *h);

    /**
     * Get the parameter value for the SIP header.
     * @param headerName is the header name (From,To,etc) and .
     * @param paramName is the parameter name (tag,timeout,etc).
     * @return the parameter value of a header.  
     * If it is not present, 
     * return an empty string.
     */
    virtual const char * GetParameter(const char * headerName, 
                                      const char * paramName);
    /**
     * Get the header value for last incoming request.
     */
    virtual const char *GetIncomingRequestHeader(const char *h);

    /**
     * Get the header value for the last incoming response
     */
    virtual const char *GetIncomingResponseHeader(const char *h);

    /** 
     * Get the header value for outgoing request.
     */
    virtual const char *GetOutgoingRequestHeader(const char *h);

    /**
     * Get the header value for outgoing response.
     */
    virtual const char *GetOutgoingResponseHeader(const char *h);

    /**
     * Set the header value for Outgoing Resquest.
     */
    virtual void SetOutgoingRequestHeader(const char *h, const char *v,
                                          bool append = false);

    /**
     * Set the header value for Outgoing response 
     */
    virtual void SetOutgoingResponseHeader(const char *h, const char *v,
                                            bool append = false);

    /**
     * Message body, set for the next response, request
     */
    virtual void SetMessageBody(const SIPMessageBody & body);

    /**
     * Callback when a new message body (non-SDP) is
     * received typically in 3xx-6xx responses.
     */
    virtual void OnNewMessageBody(const SIPMessageBody & body);

    /**
     * Callback when the remote session is changed.
     * Note that the session description should not be freed by the 
     * application.
     * @return the application should return 0 if accepts the
     * session change and -1 if not.
     */
    virtual int OnRemoteSessionChanged(const SessionDescription & sdp);

    /**
     * SetThe transport type to TCP or UDP 
     */
    virtual void SetTransportType(int type);

  protected :
    /** false for incoming call. true for outgoing */
    bool isoriginator;
    //@}

  public:
    SIPEndpoint *       endpoint; /* associated endpoint reference */
    char                *callid;  /* SIP call ID */
    char                ErrorString[128]; /* Last Error string */
    int                 LastStatus; /* Last status received or sent */

    /* For receieving responses */
    int                 response_initialized;
    pthread_cond_t      response_cond;
    pthread_mutex_t     response_mutex;
    int                 response_received;  /* 0 means NO and 1 means yes */

    int                 mark_done;   /* whether the callthread knows that 
                                        object is being deleted. */
    char                *userid;
    char                *passwd;

#ifdef MEDIACONTROL_SIPCOMM			/* MVakil 11-17-2000 */
  private:							/* MVakil 9-30-2000 for MediaControl */
	void SetUpSelfSessionDescription();
	MediaSessionInfo	*m_pMSInfo;
	void OpenMediaStream();
	void CloseMediaStream();
  public:							/* MVakil 10-28-2000 */
	HWAVEOUT GetWaveOutHandle();
	bool SetMuteState(BOOL mute);
#endif

};

/* ---------------- Additional code for generic message interface --- */

#if 0

/**
 * SIPMessage: encapsulates a SIP request or response.
 * This is used to maintain a mapping between SIP call and
 * pending requests/responses.
 */
class SIPMessage
{
 public:
  /**@name Constructors and destructors */
  //@{
  /**Construct a request */
  SIPMessage(const char *method);

  /**Construct a request using headers from an
   * existing request */
  SIPMessage(SIPMessage &request, const char *method);

  /**Construct a response */
  SIPMessage(int status, const char *reason=NULL);

  /**Construct a response for the specified request */
  SIPMessage(SIPMessage &request, int status, const char *reason = NULL);

  /**Copy constructor */
  SIPMessage(const SIPMessage &msg);

  /**Destroy the message */
  virtual ~SIPMessage();

  /**Copy operator */
  SIPMessage & operator=(const SIPMessage & msg);

  //@}

  /**@name Getters and Setters for different message params */
  //@{

  /**Whether the message is a request or not */
  bool IsRequest();

  /** Whether the message is a response or not */
  bool IsResponse();

  /** Set any header to the specified value */
  void SetHeader(const char *header, const char *value);
  /** Append the header to the existing set */
  void AppendHeader(const char *header, const char *value);
  /** Remove the header from the existing headers */
  void Remove(const char *header);
  /** Get the header value */
  const char * GetHeader(const char *header);

  /** Get the To header value */
  IptelAddress GetTo();
  /**Set the To header value */
  void SetTo(const IptelAddress & addr);

  /** Get the From header value */
  IptelAddress GetFrom();
  /** Set the From header value */
  void SetFrom(const IptelAddress & addr);


  //@}

 private:
};


/**
 * SIPMsgEngine: a generic message engine which can be used to
 * implement any new method support in SIP. This is 
 * useful for implementing newly created extensions like
 * MESSAGE, DO, SUBSCRIBE, NOTIFY. XXX You can even 
 * implement HTTP request/response (client/server)
 * using this.
 */
class SIPMsgEngine 
{
 public:
    /**@name constructors and destructors */
    //@{
    /**
     * Default constructor.
     * This attaches the SIP call with a default #SIPEndpoint#.
     */
    SIPMsgEngine(const char *method, const char *callid = NULL);

    /** 
     * Create a new msg engine for the endpoint.
     * This attaches the SIP call with the endpoint
     * specified. Some of the attributes can be
     * borrowed (for example, self name) from the
     * associated endpoint.
     */
    SIPMsgEngine(SIPEndpoint & ep, 
            const char *method, 
            const char *callid = NULL);

    /** Destructor to destroy the object */
    virtual ~SIPMsgEngine();

    //@}

    /**@name message send and receive */
    //@{
    /**
     * Send an outgoing message.
     *
     * @param msg is the SIP Message to be sent to
     * remote. It is assumed that the
     * self and remote addresses are set in
     * in the SIPMsgEngine object.
     *
     * @return 0 on successful initiation and -1 on failure.
     * Not that even if the method returns 0 the actual call may fail,
     * and OnSendFailed or OnSendSuccessful will be called as 
     * appropriate.
     */
    virtual int   SendRequest(SIPMessage & msg);

    /**
     * Send a response for an incoming message.
     * 
     * @param msg is the SIPMessage (of type response).
     *
     * @return 0 on success and -1 on failure.
     */
    virtual int   SendResponse(SIPMessage & msg);


    /**
     * Redirect an incoming request with appropriate reason.
     * 
     * @param newaddress is the "Contact:" address of the new
     * destination where you want the call to be redirected.
     *
     * @param reason specifies the reason for redirection.
     * If this is NULL then default is "Redirect".
     *
     * @return 0 on success and -1 on failure.
     */
    virtual int Redirect(const IptelAddress newaddress,
                         const char *reason=NULL);

    /**
     * Authenticate a particular call with appropriate Authentication
     * Methods used are Basic, and Digest. This call does not reject the
     * incoming call, but just checks if the authentication/Authorization
     * headers are valid or not. It is upto the application
     * to accept or reject a call.
     *
     * @param method is type of authentication needed."basic" or "digest"
     *
     * @param userfile is the password file for basic and digest auth.
     *
     * @return true on success and false on failure.
     */
    virtual bool Authenticate(const char *method,
                              const char *userfile);

    /** 
     * Set BASIC authorization header. This will be used for
     * subsequent requests for this msg engine.
     * 
     * @param user is the user id
     *
     * @param passwd is the password to be used for basic 
     * or digest authentication.
     */
    virtual void SetAuthorization(const char *user, const char *passwd);

    /**
     * Callback when an Authorization is needed for outgoing call.
     * Called when 401 is received in response to an
     * outgoing request (or other call related method).
     * Default implementation sets the user and pass if
     * it is available (set using SetAuthorization earlier).
     *
     * @param method is basic or digest
     * @param realm is the authorization realm.
     * @param user should be set by the implementation.
     * @param pass should be set by the implementation.
     * @return returns 0 on success and -1 if user and pass are
     * not set or not known.
     */
    virtual int OnAuthorizationNeeded(const char *method,
                                      const char *realm,
                                      char *& user,
                                      char *& pass);

    /**
     * Callback when a ProxyAuthorization is needed for outgoing call.
     * Called when 407 is received in response to an
     * outgoing INVITE (or other call related method).
     * Default implementation sets the user and pass if
     * it is available (set using SetAuthorization earlier).
     *
     * @param method is basic or digest
     * @param realm is the authorization realm.
     * @param user should be set by the implementation.
     * @param pass should be set by the implementation.
     * @return returns 0 on success and -1 if user and pass are
     * not set or not known.
     */
    virtual int OnProxyAuthorizationNeeded(const char *method,
                                           const char *realm,
                                           char *& user,
                                           char *& pass);

    /**
     * Callback on new incoming message.
     * The library indicates an incoming message to the
     * application by calling this method with the
     * associated IptelEndpoint. The IptelEndpoint
     * must have been created as a server listening
     * for incoming calls. Application should Accept or
     * Reject the incoming call either in the implementation
     * of this call back function or in a separate thread.
     * Default behaviour accepts any incoming call.
     * 
     * @param ep is the associated endpoint object
     * with the call.
     *
     */
    virtual void   OnIncomingRequest(SIPMessage & msg);

    /**
     * Waiting for a final response from remote for outgoing request.
     * Application can call this method to block till 
     * a final response is received for an outgoing call.
     *
     * @param timeout in seconds can be specified. Default is -1 for
     * INFINITE timeout.
     */
    virtual int    WaitForResponse(int timeout = -1);

    /**
     * Callback on successful call response from remote
     * for the last sent message.
     * The Library calls this method to indicate successful msg sent
     * Default behaviour does nothing.
     * @return 0 if you want the library to free the
     * call object otherwise return non-zeron if your
     * application intend to use the object later also.
     */
    virtual int   OnSuccessfulResponse(SIPMessage & msg);

    /**
     * Callback on request rejected by remote.
     * The Library calls this method to indicate that the
     * remote has rejected your call.
     * @return 0 if you want the library to free the call object.
     * otherwise return non-zero if your application is responsible
     * for freeing the call object.
     */
    virtual int   OnFailureResponse(SIPMessage & msg);

    /**
     * Is it an outgoing or incoming call.
     * @return true if this is the originator of the call.
     * false otherwise.
     */
    virtual bool   IsOriginator() const { return isoriginator;}

    /** 
     * Return the callid string.
     * Callid string is useful for signaling gateway
     * applications where you want to map
     * the SIP callid with the PSTN call, for example.
     * This can also be used to associate the
     * object with other types of objects.
     *
     * @return returns the call id string.
     */
    virtual const char *GetCallId() const {return callid; }

    //@}

    /**@name Misc methods */
    //@{
    /**
     * Set the address of the default outbound proxy.
     * All outgoing calls are sent to this address
     * irrespective of the specified destination address.
     *
     * @param proxy specifies the proxy address.
     *
     * @return 0 on success and -1 on failure.
     */
    virtual int    SetProxyAddress(const IptelAddress & proxy);

    /** 
     * Retrieve the proxy address.
     *
     * @return The proxy address set earlier by SetProxyAddress.
     */
    virtual IptelAddress     GetProxyAddress();

    /**
     * Set the self address.
     */
    virtual int  SetSelfAddress(const IptelAddress & self);

    /**
     * Set the remote address.
     */
    virtual int  SetRemoteAddress(const IptelAddress & remote);

    /**
     * Get the local address.
     */
    virtual IptelAddress GetSelfAddress();

    /** 
     * Get the remote address.
     */
    virtual IptelAddress GetRemoteAddress();

    /**
     * SetThe transport type to TCP or UDP 
     */
    virtual void SetTransportType(int type);

  protected :
    /** false for incoming call. true for outgoing */
    bool isoriginator;
    //@}

  public:
    SIPEndpoint *       endpoint; /* associated endpoint reference */
    char                *callid;  /* SIP call ID */

    /* For receieving responses */
    int                 response_initialized;
    pthread_cond_t      response_cond;
    pthread_mutex_t     response_mutex;
    int                 response_received;  /* 0 means NO and 1 means yes */

    int                 mark_done;   /* whether the callthread knows that 
                                        object is being deleted. */
    char                *userid;
    char                *passwd;

    /* Local and remote addresses */
    IptelAddress        self;
    IptelAddress        remote;
    IptelAddress        proxy;
};

#endif

//@}

#endif /* _SIP_API_H_ */

/************************** END OF FILE ********************************
*/
