/***************************** FILE HEADER *****************************
#               Copyright (c) 1999 by Columbia University              *
#                        All Rights Reserved                           *
#***********************************************************************
#***********************************************************************
#
#  File Name    : sipcall.h
#
#  Module       : SIP call related information.
#
#  Author(s)    : Kundan Singh (kns10@cs.columbia.edu)
#
#  Created      : 28 July 2000
#
#  Description  : This file defines the C sipcall_t structure.
#                
#  Notes        : 
#
#***********************************************************************
#                Modification History (Reverse Order)
#-----------------------------------------------------------------------
#   Date         Author              Description
#-----------------------------------------------------------------------
#              |             | 
#              |             | 
#-----------------------------------------------------------------------
*/
 
/***********************************************************************
******************* INCLUDES *******************************************
*/

#ifndef SIP_CALL_H
#define SIP_CALL_H  1

#include <stdio.h>    /* FILE */
#include "ads.h"
#include "iptelapi.h"
#include "sipapi.h"

#ifdef __cplusplus
extern "C" {
#endif

#include "request-sip.h"
#include "message.h"
#include "headers.h"

#ifdef __cplusplus
}
#endif

/*
 * Call state for sip call 
 */
typedef enum sipcallstate_t {
  SIPCALL_IDLE,               /* Call is idle */
  SIPCALL_OUTGOINGCALL,       /* sent INVITE waiting for response */
  SIPCALL_INCOMINGCALL,       /* received INVITE, waiting for API message */
  SIPCALL_INCOMINGCALL_ACK,   /* sent 2xx for incoming call, waiting for 
                               * from remote */
  SIPCALL_ACTIVE,             /* call is active, ACK is sent or received */
  SIPCALL_TERMINATING,        /* Call is being terminated, either remote 
                               * or senf. */
  /* add new states here */

  SIPCALL_STATE_COUNT
} sipcallstate_t;

//forward declaration.
struct sipcall_t;

/**
 * sipcall_t structure stores all information related to call state.
 */
typedef struct sipcall_t {
  sipcallstate_t   state;         /* call state */

  /* SIP related */
  char             *callid;       /* call id string */
  char             *self;         /* self address */
  char             *remote;       /* remote address */
  char             *self_tag;     /* To, From tag for self */
  char             *remote_tag;   /* To or from tag for remote */
  int              lastcseq_tx;   /* last sequence number used for og req. */
  int              lastcseq_rx;   /* last cseq from remote */

  uri_t            proxy;       /* address of the outbound proxy, port 
                                   * if zeron then use 5060. */

  /* FSM related */
  message_t        *mqueue;       /* message queue */
  pthread_t        tid;           /* thread id */
  pthread_mutex_t  mutex;         /* for lock */
  pthread_cond_t   cond;          /* for signaling */
  int              done;          /* 0 means call thread is active */

  parser_t         p;             /* generic headers, parser for first request*/
  body_t           body;          /* last body received or sent */
  uri_t            remote_uri;    /* remote address where all, requests should go */

  parser_t         p_irq;         /* generic parser, headers for incoming req */
  parser_t         p_irs;         /* generic parser, headers for incoming res */
  headers_t        h_orq;         /* generic headers for outgoing requests */
  headers_t        h_ors;         /* gereric headers for outgoing response */

  /* C++ object */
  SIPCall          *callobj;      /* C++ call object */
  char             *self_sdp;     /* Self session description */
  char             *remote_sdp;   /* remote SDP */

  int              type;          /* SOCK_STREAM, SOCK_DGRAM */

  int              incoming_reinvite; /* 1 if pending incoming
                                         reinvite */

  int              unauth_count;  /* Number of times unauth received */
  char             *userid;       /* Authorization info */
  char             *password;     /* Authorization info */


  /* Route header for the Record-Route implementation */
  parser_header_t *route;

  /* next element in the list */
  struct sipcall_t * next;
} sipcall_t;

/* Functions are defined in sipcall.c */

/* CallCreation, deletion, and thread, FSM handling */
sipcall_t *CallCreate(const char *id);
void CallAdd(sipcall_t *c);
void CallFree(sipcall_t *r);
void CallPrint(FILE *fp);
sipcall_t *CallSearch(const char *callid);
void *CallThread(void *cid);
int  CallHandleRequest(sipcall_t *c, message_t *m);
int  CallHandleResponse(sipcall_t *c, message_t *m);
void CallComplaint(sipcall_t *c);
int  CallCleanup(sipcall_t *c);
void CallDone(sipcall_t *c);
void CallSetStatus(sipcall_t *c, int status, const char *reason);

/* Internal call leg */
int CallCreateLeg();

/* Internal request related */
request_t * FormatRequest(const char *method,
                  int type,
                  sipcall_t *c,
                  int cseq,
                  body_t * body);
int CallCreatePolicyThread(request_t *r, uri_t uri, uri_t proxy);

/* returns 0 on success */
int CallAuthenticate(sipcall_t *call, const char *method, const char *passwdfile);
char * CallMatchSessionWithSource(const char *src_sdp,
                                  const char *new_sdp,
                                  char **error);
IptelAddress CallExtractContact(sipcall_t *c, bool outgoing, int response_port=0);

#endif
