/***************************** FILE HEADER *****************************
#               Copyright (c) 1999 by Columbia University              *
#                        All Rights Reserved                           *
#***********************************************************************
#***********************************************************************
#
#  File Name    : sipintf.cxx
#
#  Module       : SIP API implementation.
#
#  Author(s)    : Kundan Singh (kns10@cs.columbia.edu)
#
#  Created      : 14 Oct 1999
#
#  Description  : Core SIP to application
#                
#  Notes        : 
#
#***********************************************************************
#                Modification History (Reverse Order)
#-----------------------------------------------------------------------
#   Date         Author              Description
#-----------------------------------------------------------------------
#              |             | 
#              |             | 
#-----------------------------------------------------------------------
*/

/***********************************************************************
******************* INCLUDES *******************************************
*/

#ifdef __linux__
#define _GNU_SOURCE /* Needed for strptime() declaration from <time.h> */
#endif

#include "stdio.h"
#include "iptelapi.h" /* IptelAddress,Session etc */
#include "sipapi.h"   /* SIPCall, Endpoint etc */
#include "msgflow.h"
#include "sipcall.h"
#include "sipreg.h"
#include "sipintf.h"

#ifdef __cplusplus
extern "C" {
#endif

#include <stdlib.h>  /* free */
#include <unistd.h>  /* gethostname */
#include <stdio.h>
#include <string.h>  /* strdup */
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <arpa/inet.h>  /* inet_ntoa,inet_addr */
#include <netdb.h>      /* MAXHOSTNAMELEN */
#include <errno.h>      /* errno */
#include <time.h>
#include <pthread.h>

#include "udp.h"
#include "tcp.h"
#include "request-sip.h" /* request_t */
#include "log.h"     /* Log() */
#include "config.h"    /* registration_t */
#include "policy.h"  /* policy_t */
#include "strdupn.h" /* free_valid */
#include "netdb.h"   /* gethost...*/
#include "sysdep.h"
#include "md5.h"
#include "error.h"
#include "config.h"
#include "authenticateSIP.h"
#include "timespec.h"
#include "md5.h"
#include "libsip.h"  /* SIP_OnIncoming... */
#include "querydb.h"
#include "pthread-util.h"

#ifdef __cplusplus
}
#endif


/***********************************************************************
******************* Common Macros and Types for Portability ************
*/
/* Memory management */
#define AllocateObject(type,count)    \
    (((count) == 1) ? (new type) : (new type[count]));
#define DeallocateObject(obj)  if (obj) delete (obj);

#define DEFAULT_Q_VALUE ((float)0.0)

/* Default listening port for SIP, XXX copied from sipapi.cpp ? */
#define DEFAULT_SIP_PORTNO  5060

/***********************************************************************
******************* Functions Prototypes *******************************
*/
extern "C"
{
#ifdef DO_LICENSING
  extern char *fullpath(char *root, char *filename);
#endif
  extern void set_pwinfo_file(const char *file);

}

IptelAddress SipAddressToIptelAddress(sip_addr_t addr);
IptelAddress SipContactToIptelAddress(sip_contact_t *addr);
IptelAddress SipUriToIptelAddress(uri_t sip_addr);
char *getcanonicalhostname();
extern bool CallCompareContentType(const char *type,
                            parser_t *p);

/***********************************************************************
******************* Globals and statics ********************************
*/
struct config_t config;
/***********************************************************************
******************* Functions ******************************************
*/

/* Dummy function */
char *build_userid(const uri_t *uri)
{
  DString userid;

  DStringInit(&userid);

  if (strcasecmp(uri->scheme, "sip") != 0) {
    URI_SPrint(uri, &userid);
  }
  else {
    if (uri->user == NULL && uri->host == NULL) {
      /* Shouldn't happen */
      DStringFree(&userid);
      return NULL;
    }
    if (uri->user != NULL) {
      DStringAppend(&userid, uri->user, -1);
      DStringAppend(&userid, "@", -1);
    }
    if (uri->host != NULL) {
      if ( strstr(uri->host, config.ServerName) ) {
        DStringAppend(&userid, config.ServerName, -1);
      }
      else {
        DStringAppend(&userid, uri->host, -1);
      }
    }
  }
  
  return DStringValue(&userid);
} /* build_userid */

IptelAddress SipAddressToIptelAddress(sip_addr_t sip_addr)
{
  IptelAddress addr;
  
  DString uri;
  DStringInit(&uri);

  if ( sip_addr.display_name != NULL && *sip_addr.display_name)
  {
    DStringPrintf(&uri,"\"%s\" <",sip_addr.display_name);
  }

  char *ptr = URI_AsString(sip_addr.uri);
  if ( ptr && *ptr ) {
    DStringPrintf(&uri, "%s", ptr);
  }
  else {
    /* XXX is this correct to sent URI as unknown */
    DStringPrintf(&uri, "unknown");
  }
  free_valid(ptr);

  if ( sip_addr.display_name != NULL && *sip_addr.display_name)
  {
    DStringPrintf(&uri,">");
  }
 
  addr = DStringValue(&uri);

  DStringFree(&uri);

  return addr;
}

IptelAddress SipContactToIptelAddress(sip_contact_t sip_addr)
{
  /* sip_addr is an array */
  IptelAddress addr;
  if ( sip_addr.uri.scheme != NULL ) {
    if ( sip_addr.nameaddr  != NULL && *sip_addr.nameaddr) {
      /* If nameaddr format is available use that */
      addr = sip_addr.nameaddr;
    }
    else {
      char *ptr = URI_AsString(sip_addr.uri);
      if ( ptr != NULL && *ptr ) { 
        addr = ptr;
      }
      else {
        addr = "unknown";
      }
      free_valid(ptr);
    }
  }
  else {
    /* addr is empty */
  }
  return addr;
}

IptelAddress SipUriToIptelAddress(uri_t sip_addr)
{
  IptelAddress addr;
  char *ptr = URI_AsString(sip_addr);

  if ( ptr != NULL && *ptr ) {
    addr = ptr;
  }
  else {
    addr = "unknown";
  }
  free_valid(ptr);

  return addr;
}

/*
 * List of possible URI scheme supported in the SIP (Iptel) API.
 * XXX needs modification.
 */

static char * PossibleSchemes[] = 
{
  "ip", "http", "https", "rtsp", "rtspu",
  "tel", "h323", "sip", "mailto"
};

/*
 * Parse name-addr (To, From, Contact, Route) addresses:
 * '[name] <url>' or 'url'.
 * url can be possibly without a scheme.
 */
sip_addr_t URI_ParseNew(const char *v)
{
  sip_addr_t a;
  char *url = NULL;
  int oFound = 0;
  int i;
  DString uristr;
  DString name, addr;

  /* Check if the string is empty */
  if (!v || !(*v)) 
    goto badaddr;

  /* Split the address into possible name and url(addr) */
  DStringInit(&name);
  DStringInit(&addr);
  if (!SIP_NameAddr((char *) v, &name, &addr))
    goto badaddr;

  /* Set the display name to name portion */

  a.display_name = DStringValue(&name);

  /* Parse the URL portion */
  url = DStringValue(&addr);
  oFound = 0;

  /* Try parsing */
  a.uri  = URI_Parse(url);

  /* check if the scheme is valid */
  if ( a.uri.scheme ) {
    for ( i = sizeof(PossibleSchemes)/sizeof(PossibleSchemes[0])-1;
          i >=0; i--) {
      /* Are schemes case sensitive XX */
      if ( !strcasecmp(a.uri.scheme, PossibleSchemes[i])) {
        oFound = 1;
        break;
      }
    }
  }

  if ( !oFound ) {
    /* Scheme is invalid or probably absent */
    url = strdup(a.uri.scheme);
    URI_Free(&a.uri);
    /* Fake a scheme of "sip:" for parsing */

    DStringInit(&uristr);
    DStringAppend(&uristr, "sip:", -1);
    DStringAppend(&uristr, url, -1);
  
    a.uri = URI_Parse(DStringValue(&uristr));
    DStringFree(&uristr);

    /* Remove the faked scheme sip: */
    if ( a.uri.scheme ) {
      free(a.uri.scheme);
      a.uri.scheme = NULL;
    }
    if ( url ) {
      free(url);
    }
  }

  DStringFree(&addr); /* to free the url */

  /* "name" remains, sip_addr_t.display_name is dynamic */

  return a;

badaddr:
  memset(&a.uri, 0, sizeof(a.uri));
  a.display_name = NULL;

  return a;
}

/* Extract Contact from the call */
IptelAddress CallExtractContact(sipcall_t *c, bool outgoing, int response_port)
{
  void *hp;
  char *contact = NULL;

  /* Check h_orq:Outgoing Request */
  if ( outgoing ) {
    hp = NULL;
    if ( (contact = HeaderGet(&c->h_orq, "Contact", 1, &hp)) != NULL &&
         *contact != '\0') {
      return contact;
    }
  } 
  else {
    /* check h_ors: Outgoing Response */
    hp = NULL;
    if ((contact = HeaderGet(&c->h_ors, "Contact", 1, &hp)) != NULL &&
        *contact != '\0') {
      return contact;
    }
  }

  IptelAddress cont = (const char *) c->self;
  debug("CallExtractContact", "Contact is %s\n", cont.GetAddressString());
  cont.SetHost(SIPLibrary::HostName);

  if ( response_port == 0 ) {
    if ( c->callobj && c->callobj->endpoint ) {
      IptelIPAddress response_at;
      response_at =
        c->callobj->endpoint->GetServerAddress();
      response_port = response_at.GetPort();
      debug("CallExtractContact", "Response port is %d\n", response_port);
    }
    else {
      response_port = DEFAULT_SIP_PORTNO;
    }
  }

  debug("CallExtractContact", "Response port (%d)\n", response_port);
  if ( response_port != DEFAULT_SIP_PORTNO &&
       response_port != 0) {
    cont.SetPort((int) response_port);
  }

  return cont;
}

/*-------------- sipcall_t related methods ----------------*/

/*
 * Format a request_t structure using the infomation
 * given. The mutex of the returned structure is locked !!
 */
request_t * FormatRequest(const char *method,
                  int type,
                  sipcall_t *c,
                  int cseq,
                  body_t * body)
{
  request_t *r = NULL;
  struct sockaddr_in sin;
  IptelIPAddress response_at;
  DString sipmsg, uri;

  sin.sin_family = AF_INET;
  sin.sin_port   = 0;
  sin.sin_addr.s_addr = INADDR_ANY;
  r = RequestCreate(type, /* no-socket */  -1,sin);
  r->preferred_type = type; /* Set this to be same as type */

  if ( c->callobj && c->callobj->endpoint ) {
    response_at = c->callobj->endpoint->GetServerAddress();
  }
  else {
    response_at = SIPLibrary::HostInAddr;
    response_at.SetPort(DEFAULT_SIP_PORTNO);
  }
  r->sin = (sockaddr_in) response_at;

  /* Format the request */
  DStringInit(&uri);

  DStringInit(&sipmsg);

  if ( c->route && c->route->elements ) {
    parser_element_t *e;
    for ( e = c->route->elements; e; e = e->next) {
      debug("RecordRoute", "Element: %s\n", e->value ? e->value : "NULL");
    }
  }

  if ( c->route != NULL && c->route->elements != NULL &&
       c->route->elements->value) {
    sip_addr_t firstroute;

    debug("RecordRoute", "Route is not null\n");

    memset(&firstroute, 0, sizeof(firstroute));

    firstroute = sip_addr(c->route->elements->value);
    if ( firstroute.uri.scheme != NULL ) {
      URI_SPrint(&firstroute.uri, &uri);
      debug("RecordRoute", "Added first route: %s\n",
            DStringValue(&uri));
    }
    else {
      debug("RecordRoute", "Strange!! first route is empty\n");

      if ( c->p.URI.scheme == NULL ||
           *(c->p.URI.scheme) == '\0') {
        IptelAddress remote = c->remote;
        char * uristr = remote.GetURI();
        DStringAppend(&uri,uristr ? uristr : "<none>",-1);
        free_valid(uristr);
      }
      else {
        URI_SPrint(&c->p.URI, &uri);
      }
    }
    free_valid(firstroute.display_name);
    URI_Free(&firstroute.uri);
  }
  else {
    if ( c->p.URI.scheme == NULL ||
         *(c->p.URI.scheme) == '\0') {
      IptelAddress remote = c->remote;
      char * uristr = remote.GetURI();
      DStringAppend(&uri,uristr ? uristr : "<none>",-1);
      free_valid(uristr);
    }
    else {
      URI_SPrint(&c->p.URI, &uri);
    }
  }

  DStringPrintf(&sipmsg, "%s %s SIP/2.0\r\n", method,
                DStringValue(&uri));
  /* should not add Via, since it is automatically added */
  if ( !c->self_tag ) {
    DStringPrintf(&sipmsg, "From: %s\r\n", c->self ? c->self : "Unknown");
  }
  else {
    DStringPrintf(&sipmsg, "From: %s; tag=%s\r\n", c->self ? c->self : "Unknown",
                                                   c->self_tag);
  }

  if ( !c->remote_tag ) {
    DStringPrintf(&sipmsg, "To: %s\r\n", c->remote ? c->remote : "Unknown");
  }
  else {
    DStringPrintf(&sipmsg, "To: %s; tag=%s\r\n", c->remote ? c->remote : "Unknown",
                                                 c->remote_tag);
  }

  void *hp = NULL;
  char *subject = HeaderGet(&c->h_orq, "Subject", 1, &hp);
  if ( subject && *subject ) {
    DStringPrintf(&sipmsg, "Subject: %s\r\n", subject);
  }
  else {
    /* Remove this */
    DStringPrintf(&sipmsg, "Subject: test call from libsip++\r\n");
  }

  DStringPrintf(&sipmsg, "CSeq: %d %s\r\n", cseq, method);
  DStringPrintf(&sipmsg, "Call-ID: %s\r\n", c->callid);

  hp = NULL;
  char *authorization = HeaderGet(&c->h_orq, "Authorization", 1, &hp);
  if ( authorization && *authorization ) {
    DStringPrintf(&sipmsg, "Authorization: %s\r\n", authorization);
  }

  hp = NULL;
  authorization = HeaderGet(&c->h_orq, "Proxy-Authorization", 1, &hp);
  if ( authorization && *authorization ) {
    DStringPrintf(&sipmsg, "Proxy-Authorization: %s\r\n", authorization);
  }

  IptelAddress contactaddr = CallExtractContact(c,true,response_at.GetPort());
  DStringPrintf(&sipmsg, "Contact: <%s>\r\n",
                (const char *) contactaddr);

  hp = NULL;
  char *acceptstr = NULL;

  if ( !strcmp(method, "INVITE") ) {
    /* INVITE must have application/sdp as Accept XXX */
    if ((acceptstr = HeaderGet(&c->h_orq, "Accept", 1, &hp)) != NULL  &&
      *acceptstr) {
      DStringPrintf(&sipmsg, "Accept: application/sdp, %s\r\n", acceptstr);
    }
    else {
      DStringPrintf(&sipmsg, "Accept: application/sdp\r\n");
    }
  }
  else {
    if ((acceptstr = HeaderGet(&c->h_orq, "Accept", 1, &hp)) != NULL  &&
      *acceptstr) {
      DStringPrintf(&sipmsg, "Accept: %s\r\n", acceptstr);
    }
  }

  hp = NULL;
  char *requirestr = NULL;
  if ((requirestr = HeaderGet(&c->h_orq, "Require", 1, &hp)) != NULL  &&
    *requirestr) {
    DStringPrintf(&sipmsg, "Require: %s\r\n", requirestr);
  }

  hp = NULL;
  char *useragentstr = NULL;
  if ((useragentstr = HeaderGet(&c->h_orq, "User-Agent", 1, &hp)) != NULL  &&
    *useragentstr) {
    DStringPrintf(&sipmsg, "User-Agent: %s\r\n", useragentstr);
  }

  if ( c->route ) {
    parser_element_t *e;
    /* Add the Route header */
    debug("RecordRoute", "Adding Route header\n");
    /* Add a dummy route header which will be ignored any way */
    DStringPrintf(&sipmsg, "Route: <%s>\r\n", 
                  (const char *) contactaddr);
 
    for ( e = c->route->elements; e; e=e->next) {
      debug("RecordRoute", "Adding %s\n", e->element);
      DStringPrintf(&sipmsg, "Route: %s\r\n", e->element);
    }
  }

  /* XXX add other headers here like, Accept, Require etc.
   * Or have a loop to add all the headers if not
   * already added.
   */

  if ( (body == NULL) || (body->content_type == NULL) ||
       (body->content == NULL) ) {
    DStringPrintf(&sipmsg, "Content-Length: 0\r\n");
    DStringAppend(&sipmsg, "\r\n", -1);
  }
  else {
    if ( body->content_length <= 0) {
      body->content_length = strlen(body->content);
    }
    DStringPrintf(&sipmsg, "Content-Type: %s\r\n", body->content_type);
    DStringPrintf(&sipmsg, "Content-Length: %u\r\n", body->content_length);
    DStringAppend(&sipmsg, "\r\n", -1);
    DStringAppend(&sipmsg, body->content, body->content_length);
  }

  /* parse the request */
  if (HTTP_Parse(DStringValue(&sipmsg), &r->p) < 0 ) {
    error("FormatRequest", "parse error\n");
    CallSetStatus(c, 400, "Can not format a valid request");
    pthread_mutex_unlock(&r->mutex);
    RequestFree(r);
    r = NULL;
  }
  else {
    char *msg = DStringValue(&sipmsg);
    int offset = r->p.header_length;
    int content_length = HTTP_ContentLength(&r->p);
    if (content_length > 0) {
      DString body_str;
      DStringInit(&body_str);
      DStringAppend(&body_str, &msg[offset], content_length);
      r->request_body.content = DStringValue(&body_str);
      r->request_body.content_length = content_length;
      parser_header_t *h;
      if ((h = RFC822_Header(r->p.headers, "Content-Type"))
          && h->elements) {
        /* request_body.content_type points to the internal
         * parser structure element. */
        r->request_body.content_type = h->elements->value;
      }
    }

    /* ---- Now fill in the relevant fields of r */
    uri_t uri;
    r->version = HTTP_Version(&r->p);
    uri = HTTP_URI(&r->p);
    if ( uri.scheme == NULL ||
      ( !strcmp(uri.scheme, "sip") && !strcmp(uri.scheme, "tel"))) {
      error("FormatRequest", "uri.scheme is not sip or tel\n");
      CallSetStatus(c, 400, "URI scheme is not sip or tel");
      pthread_mutex_unlock(&r->mutex);
      RequestFree(r);
      r = NULL;
    }
    else {
      if ( !SIP_Has_Mandatory_Headers(&r->p)) {
        error("FormatRequest", "does not have mandatory headerd\n");
        CallSetStatus(c, 400, "Does not have mandatory headers");
        pthread_mutex_unlock(&r->mutex);
        RequestFree(r);
        r = NULL;
      }
      else {
        r->cseq         = SIP_CSeq(&r->p);
        r->callID       = NULL;
        strassign(&r->callID, SIP_CallID(&r->p));
        r->contact      = SIP_Contact(&r->p);

        r->expires      = SIP_Expires(&r->p);
        /* Via headers are added automatically */
        r->to           = SIP_To(&r->p);
        r->from         = SIP_From(&r->p);

        parser_header_t *h;
        if ((h = RFC822_Header(r->p.headers, "Content-Type"))
            && h->elements) {
          /* request_body.content_type points to the internal
           * parser structure element. */
          r->request_body.content_type = h->elements->value;
        }

        ResponseSetStatus(r,200,NULL); 
        ResponseBuildDefault(r); /* to build the default response */

        /* --- Add the request in the list --- */
        pthread_mutex_lock(&r->mutex);      
        RequestAdd(r);
      }
    }
  }

  DStringFree(&sipmsg);
  DStringFree(&uri);

  return r;
}

/*
 * Format a request_t structure using the infomation
 * given for a REGISTER request. 
 * REGISTER request is different because there is no
 * call associated with this.
 * This can be further extended to other call independent
 * method like OPTIONS, SUBSCRIBE etc.
 * The mutex of the returned structure is locked !!
 */
request_t * FormatRegistration(const char *method,
                  int type,
                  const char *callid,
                  unsigned int cseq,
                  body_t * body,
                  sipreg_t *c)
{
  request_t *r = NULL;
  struct sockaddr_in sin;
  IptelIPAddress response_at;
  DString sipmsg;
  char *mymethod = NULL;

  if ( !strcmp(method, "REGISTER") ) {
    mymethod = "REGISTER";
  }
  else if ( !strcmp(method, "UNREGISTER")) {
    mymethod = "REGISTER";
  }
  else if ( !strcmp(method, "REFRESH")) {
    mymethod = "REGISTER";
  }
  else {
    /* define OPTIONS, SUBSCRIBE here */
    debug("FormatRegistration", "Invalid method %s\n", method);
    return NULL;
  }

  sin.sin_family = AF_INET;
  sin.sin_port   = 0;
  sin.sin_addr.s_addr = INADDR_ANY;
  r = RequestCreate(type, /* no-socket */  -1,sin);
  r->preferred_type = type;
  r->server = (void *) c->regobj; /* Does this work ? */

  response_at = c->regobj->GetServerAddress();

  r->sin = (sockaddr_in) response_at;

  /* Format the request */
  DStringInit(&sipmsg);
  char *uristr = URI_AsString(c->remote_uri);
  DStringPrintf(&sipmsg, "%s %s SIP/2.0\r\n", mymethod,
                          uristr);
  free_valid(uristr);

  /* should not add Via, since it is automatically added */
  DStringPrintf(&sipmsg, "From: %s\r\n", c->self);
  DStringPrintf(&sipmsg, "To: %s\r\n", c->remote);

  DStringPrintf(&sipmsg, "CSeq: %u %s\r\n", cseq, mymethod);
  DStringPrintf(&sipmsg, "Call-ID: %s\r\n", callid);
  if ( !strcmp(method, "UNREGISTER") ) {
    /* Unregistration has expires as 0 */
    DStringPrintf(&sipmsg, "Expires: 0\r\n");
  }
  else {
    /* Fill in the expires header XXX for REGISTER */
    if ( c->contact.expires > 0 ) {
      DStringPrintf(&sipmsg, "Expires: %d\r\n", c->contact.expires);
      
      struct timeval now;
      
      gettimeofday(&now, NULL);
      if ( c->contact.expires > 10 ) {
        c->expireson = timespec_delay(now, (double) (
                                      c->contact.expires - 10));
      }
      else {
        c->expireson = timespec_delay(now, (double) (
                                      c->contact.expires));
      }
        
    }
  }

  if ( c->regobj->h_orq & "Authorization") {
    DStringPrintf(&sipmsg, "Authorization: %s\r\n", 
       (const char *) c->regobj->h_orq["Authorization"]);
  }

  if ( !strcmp(method, "REGISTER") ||
       !strcmp(method, "REFRESH") ) {
    char *contactstr = URI_AsString(c->contact.uri);
    DStringPrintf(&sipmsg, "Contact: <%s>\r\n", contactstr);
    free_valid(contactstr);
  }
  else if ( !strcmp(method, "UNREGISTER") ) {
    DStringPrintf(&sipmsg, "Contact: *\r\n");
  }

  /* XXX add other headers here like, Accept, Require etc.
   * Or have a loop to add all the headers if not
   * already added.
   */

  if ( (body == NULL) || (body->content_type == NULL) ||
       (body->content == NULL) ) {
    DStringPrintf(&sipmsg, "Content-Length: 0\r\n");
    DStringAppend(&sipmsg, "\r\n", -1);
  }
  else {
    if ( body->content_length <= 0) {
      body->content_length = strlen(body->content);
    }
    DStringPrintf(&sipmsg, "Content-Type: %s\r\n", body->content_type);
    DStringPrintf(&sipmsg, "Content-Length: %u\r\n", body->content_length);
    DStringAppend(&sipmsg, "\r\n", -1);
    DStringAppend(&sipmsg, body->content, body->content_length);
  }

  /* parse the request */
  if (HTTP_Parse(DStringValue(&sipmsg), &r->p) < 0 ) {
    error("FormatRegistration", "parse error\n");
    pthread_mutex_unlock(&r->mutex);
    RequestFree(r);
    r = NULL;
  }
  else {
    char *msg = DStringValue(&sipmsg);
    int offset = r->p.header_length;
    int content_length = HTTP_ContentLength(&r->p);
    if (content_length > 0) {
      DString body_str;
      DStringInit(&body_str);
      DStringAppend(&body_str, &msg[offset], content_length);
      r->request_body.content = DStringValue(&body_str);
      r->request_body.content_length = content_length;
      parser_header_t *h;
      if ((h = RFC822_Header(r->p.headers, "Content-Type"))
          && h->elements) {
        /* request_body.content_type points to the internal
         * parser structure element. */
        r->request_body.content_type = h->elements->value;
      }
    }

    /* ---- Now fill in the relevant fields of r */
    uri_t uri;
    r->version = HTTP_Version(&r->p);
    uri = HTTP_URI(&r->p);
    if ( uri.scheme == NULL ||
      ( !strcmp(uri.scheme, "sip") && !strcmp(uri.scheme, "tel"))) {
      error("FormatRequest", "uri.scheme is not sip or tel\n");
      pthread_mutex_unlock(&r->mutex);
      RequestFree(r);
      r = NULL;
    }
    else {
      if ( !SIP_Has_Mandatory_Headers(&r->p)) {
        error("FormatRequest", "does not have mandatory headerd\n");
        pthread_mutex_unlock(&r->mutex);
        RequestFree(r);
        r = NULL;
      }
      else {
        r->cseq         = SIP_CSeq(&r->p);
        r->callID       = NULL;
        strassign(&r->callID, SIP_CallID(&r->p));
        r->contact      = SIP_Contact(&r->p);

        r->expires      = SIP_Expires(&r->p);
        /* Via headers are added automatically */
        r->to           = SIP_To(&r->p);
        r->from         = SIP_From(&r->p);

        parser_header_t *h;
        if ((h = RFC822_Header(r->p.headers, "Content-Type"))
            && h->elements) {
          r->request_body.content_type = h->elements->value;
        }

        ResponseSetStatus(r,200,NULL); 
        ResponseBuildDefault(r); /* to build the default response */

        /* --- Add the request in the list --- */
        pthread_mutex_lock(&r->mutex);      
        RequestAdd(r);
      }
    }
  }

  DStringFree(&sipmsg);

  return r;
}

request_t * FormatOptions(const char *method, int type,
                          const char *callid,
                          int cseq,
                          const IptelAddress & to,
                          const IptelAddress & from,
                          body_t *body,
                          const SIPEndpoint & endpoint)
{

  request_t *r = NULL;
  struct sockaddr_in sin;
  IptelIPAddress response_at;
  DString sipmsg;

  sin.sin_family = AF_INET;
  sin.sin_port   = 0;
  sin.sin_addr.s_addr = INADDR_ANY;
  r = RequestCreate(type, /* no-socket */  -1,sin);
  r->preferred_type = type;
  r->server = (void *) &endpoint; /* Does this work ? */

  response_at = endpoint.GetServerAddress();

  r->sin = (sockaddr_in) response_at;

  /* Format the request */
  IptelAddress uri;
  char *ptr = to.GetURI();
  uri = ptr;
  free_valid(ptr);

  /* request uri <sip:...> */
  uri.SetScheme("sip");

  IptelAddress src = from;
  if ( src.IsEmpty() ) {
    src = SIPLibrary::HostName;
    src.SetScheme("sip");
    /*
     * Should we set the user name also
     */
  }

  DStringInit(&sipmsg);
  DStringPrintf(&sipmsg, "%s %s SIP/2.0\r\n", method,
                          (const char *) uri);
  /* should not add Via, since it is automatically added */
  DStringPrintf(&sipmsg, "From: %s\r\n", (const char *) src);
  DStringPrintf(&sipmsg, "To: %s\r\n", (const char *) to);

  DStringPrintf(&sipmsg, "CSeq: %u %s\r\n", cseq, method);
  DStringPrintf(&sipmsg, "Call-ID: %s\r\n", callid);

  if ( endpoint.h_orq & "Authorization") {
    DStringPrintf(&sipmsg, "Authorization: %s\r\n", 
       (const char *) endpoint.h_orq["Authorization"]);
  }

  if ( (body == NULL) || (body->content_type == NULL) ||
       (body->content == NULL) ) {
    DStringPrintf(&sipmsg, "Content-Length: 0\r\n");
    DStringAppend(&sipmsg, "\r\n", -1);
  }
  else {
    if ( body->content_length <= 0) {
      body->content_length = strlen(body->content);
    }
    DStringPrintf(&sipmsg, "Content-Type: %s\r\n", body->content_type);
    DStringPrintf(&sipmsg, "Content-Length: %u\r\n", body->content_length);
    DStringAppend(&sipmsg, "\r\n", -1);
    DStringAppend(&sipmsg, body->content, body->content_length);
  }

  /* parse the request */
  if (HTTP_Parse(DStringValue(&sipmsg), &r->p) < 0 ) {
    error("FormatRegistration", "parse error\n");
    pthread_mutex_unlock(&r->mutex);
    RequestFree(r);
    r = NULL;
  }
  else {
    char *msg = DStringValue(&sipmsg);
    int offset = r->p.header_length;
    int content_length = HTTP_ContentLength(&r->p);
    if (content_length > 0) {
      DString body_str;
      DStringInit(&body_str);
      DStringAppend(&body_str, &msg[offset], content_length);
      r->request_body.content = DStringValue(&body_str);
      r->request_body.content_length = content_length;
      parser_header_t *h;
      if ((h = RFC822_Header(r->p.headers, "Content-Type"))
          && h->elements) {
        /* request_body.content_type points to the internal
         * parser structure element. */
        r->request_body.content_type = h->elements->value;
      }
    }

    /* ---- Now fill in the relevant fields of r */
    uri_t uri;
    r->version = HTTP_Version(&r->p);
    uri = HTTP_URI(&r->p);
    if ( uri.scheme == NULL ||
      ( !strcmp(uri.scheme, "sip") && !strcmp(uri.scheme, "tel"))) {
      error("FormatRequest", "uri.scheme is not sip or tel\n");
      pthread_mutex_unlock(&r->mutex);
      RequestFree(r);
      r = NULL;
    }
    else {
      if ( !SIP_Has_Mandatory_Headers(&r->p)) {
        error("FormatRequest", "does not have mandatory headerd\n");
        pthread_mutex_unlock(&r->mutex);
        RequestFree(r);
        r = NULL;
      }
      else {
        r->cseq         = SIP_CSeq(&r->p);
        r->callID       = NULL;
        strassign(&r->callID, SIP_CallID(&r->p));
        r->contact      = SIP_Contact(&r->p);

        r->expires      = SIP_Expires(&r->p);
        /* Via headers are added automatically */
        r->to           = SIP_To(&r->p);
        r->from         = SIP_From(&r->p);

        parser_header_t *h;
        if ((h = RFC822_Header(r->p.headers, "Content-Type"))
            && h->elements) {
          r->request_body.content_type = h->elements->value;
        }

        ResponseSetStatus(r,200,NULL); 
        ResponseBuildDefault(r); /* to build the default response */

        /* --- Add the request in the list --- */
        pthread_mutex_lock(&r->mutex);      
        RequestAdd(r);
      }
    }
  }

  DStringFree(&sipmsg);

  return r;
}


/*
 * Structure for passing the param to this thread
 */
struct user_policy_param {
  request_t *r;
  uri_t uri;
};

/*
 * Thread which calls the user_policy core.
 */
void *UserAgentThread(void *arg)
{
  struct user_policy_param *param = (struct user_policy_param *)arg;

  if ( param == NULL ) {
    debug("UserAgentThread", "param is NULL\n");
    return NULL;
  }

  policy_info_t reg;
  request_t *r = param->r;
  uri_t uri_contact = param->uri;

  free(param);
  param = NULL;

  /* grab the lock first !! */
  pthread_mutex_lock(&r->mutex);
  r->policy = &user_policy; /* set the policy to user, even if already set */

  /* -- Handover the request to user_policy thread --- */
  r->policy = &user_policy; 

  debug("UserAgentThread",
        "r=%p: policy %s chosen\n", r, r->policy->name);

  reg.contacts = 1; /* only one contact i.e. TO address */
  reg.contact  = (sip_contact_t *)calloc(2, sizeof(sip_contact_t));
  reg.contact[0].expires = (time_t) -1; /* Not used */
  reg.contact[0].q       = 0.5; /* Not used */
  reg.contact[0].comment = NULL; /* Not used */
  reg.contact[0].action  = L_unspecified; /* Not used */
  reg.contact[0].uri = URI_Copy(uri_contact);
  /* Last element is set to NULL scheme */
  reg.contact[1].uri.scheme = NULL;

  URI_Free(&uri_contact);

  /* Call the policy_core to take over */
  execute_policy(r, &reg, 200 /* indicating no error */);

  return NULL;
}

/*
 * Create a new thread for user_policy.
 * It is assumed that the mutex is locked.
 */
int CallCreatePolicyThread(request_t *r, uri_t uri,
                           uri_t proxy)
{
  struct user_policy_param *param;

  extern void *UserAgentThread(void *arg);
  pthread_t tid;

  param = (struct user_policy_param *) calloc(1, sizeof(struct user_policy_param));
  param->r = r;
  param->uri = URI_Copy(uri);

  if ( proxy.scheme != NULL ) {
    /* Copy the proxy address to the uri */
    debug("CallCreatePolicyThread", "Using an outbound proxy\n");
    r->outbound_proxy = URI_Copy(proxy);
  }
  /* do not free uri !! */

  debug("pthread_create_detached", "UserAgentThread\n");
  int oRet = pthread_create_detached(&tid, NULL, 
                                     UserAgentThread, (void *) param);

  /* Unlock the mutex once done. */
  pthread_mutex_unlock(&r->mutex);
  return oRet;
}

/*
 * Authenticates the SIP incoming request.
 * returns 0 if successful,
 * -1 on error,
 * and 1 if non-authenticated or
 * Authorization header is missing.
 */
int CallAuthenticate(sipcall_t *c, const char * method, const char *file)
{
  int oRet = -1;
  char *userid;

  request_t *r = NULL;

  /* Find a matching previous request */
  r = RequestFind(&c->p, 0);

  if ( r == NULL ) {
    debug("CallAuthenticate", "no request found\n");
    return oRet;
  }

  /* Note that the r->mutex is locked ! */
  if ( r->policy != &user_policy ) {
    debug("CallAuthenticate","Not a User Call\n");
    pthread_mutex_unlock(&r->mutex);
    return oRet;
  }

  /* XXX lock the config mutex here */
  set_pwinfo_file(file);
  if ( !strcasecmp(method, "digest") ) {
    config.AuthMethod    = auth_digest;
  }
  else if ( !strcasecmp(method, "none")) {
    config.AuthMethod    = auth_none;
  }
  else if ( !strcasecmp(method, "basic")) {
    config.AuthMethod    = auth_basic;
  }
  else if ( !strcasecmp(method, "pgp")) {
    config.AuthMethod    = auth_pgp;
  }
  else {
    config.AuthMethod = auth_none;
  }

  config.NonceLifeTime = 60; /* seconds */
  config.PrivateKey    = "1"; /* key */
  config.Realm         = getcanonicalhostname();
  /* XXX set realm to cs.columbia.edu */
  if ( config.Realm && strstr(config.Realm, "cs.columbia.edu") ) {
    strassign(&config.Realm, "cs.columbia.edu");
  }

  config.ClockSkew     = 300; /* +- seconds */

  userid = build_userid(&r->p.URI);

  oRet =  AuthenticateSIP( r,  
                           0 /* no proxy auth */, 
                           1 /* Build headers */, 
                           0 /* Bad password is not OK */);

  free_valid(userid);

  /* XXX unlock theconfig mutex here */

  pthread_mutex_unlock(&r->mutex);

  return oRet;
}

/* 
 * Dummy implementation of CanHandleRequest.
 * Currently we just accept the request in the
 * UA library.
 */
extern "C" int SIP_CanHandleRequest(request_t *r)
{
  return 1;
}

extern "C" void SIP_OnEndToEndAck(request_t *r)
{
  /* Set the acked field as true */
  r->acked = 1;
}

/*
 * Implementation of SIP_OnIncomingAck().
 */
extern "C" void SIP_OnIncomingAck(request_t *r)
{
  request_t *rp;

  /* Find the associated request */
  if ( (rp = RequestFind(&r->p, r->hash) )!= NULL) {
    debug("SIP_OnIncomingAck", "ACK\n");
    /* Method is ACK */
    if (rp->sent_200) {
      /* This request (which is probably a proxy) has sent a 200, so
       * this ACK probably needs to be handled end-to-end.  Hand it off to
       * the policy core. */
      debug("RequestProcessThread", "Already sent 200\n");
      MessageEnqueue(&rp->mqueue, message_real_message, &r->p,
                     &r->request_body, r->source_sin,NULL,
                     nobody_handled);
      /* XXX the rp->acked should be set to 1 regardless
       * of whether rp->sent_200 is true or not.
       * Jonathan, is this correct ? */
      rp->acked = 1;
    } else {
      /* just set flag. */
      debug("RequestProcessThread", "not sent 200\n");
      rp->acked = 1;
    }
    pthread_cond_broadcast(&rp->cond);
    /* Don't send a response to an ACK. */
    r->status = 0;
    
    pthread_mutex_unlock(&rp->mutex);
  }
  else {
    debug("RequestProcessThread", "ACK for unknown request\n");
    /* ACK of an unknown request -- ignore. */
  }

  /* Free the request */
  RequestFree(r);
  return;
}

/*
 * This is called when an incoming (new) request
 * arrives. This module creates a new call thread
 * if needed and informs the call thread
 * about the request.
 * the request policy_core is invoked to
 * takeover.
 */
extern "C" void SIP_OnIncomingRequest(request_t *r)
{
  char *callid;
  sipcall_t *c = NULL;
  uri_t uri;                   /* request URI */
  char *method;                /* request method */
  SIPEndpoint *pobjEndpoint;

  pobjEndpoint = (SIPEndpoint *)r->server;
  method = HTTP_Method(&r->p);
  uri = URI_Copy(HTTP_URI(&r->p));

  callid = strdup(SIP_CallID(&r->p));

  r->policy = &user_policy; /* set the default policy */

  if ( !strcmp(method, "REGISTER") ) {
    SIPRegistrationInfo reginfo;
    sip_addr_t to = SIP_To(&r->p);

    reginfo.localName = SipAddressToIptelAddress(to);
    free_valid(to.display_name);
    URI_Free(&to.uri);

    reginfo.registrar = SipUriToIptelAddress(uri);
    if ( r->contact ) {
      reginfo.contactAddress = SipContactToIptelAddress(r->contact[0]);
    }
    else {
      /* XXX check if appropriate status response is returned */
      debug("OnIncomingRequest", "REGISTER does not have a contact\n");
    }
    reginfo.expiry    = r->expires;
    reginfo.q_value   = (r->contact) ? r->contact[0].q : DEFAULT_Q_VALUE;

    /* 
     * XXX should we have a fromAddress int RegistrationInfo
     * to allow thirt party registration.
     */

    if ( pobjEndpoint != NULL && 
         pobjEndpoint->IsRegistrar() )
    {
      /* provisional responses are sent only for REGISTER */
      ResponseSetStatus(r, 100, NULL); 
      ResponseSetBody(r, NULL, NULL, 0);
      ResponseBuildDefault(r);
      ResponseSend(r, 0, 0);

      if ( pobjEndpoint->OnIncomingRegistration(reginfo) != 0 ) {
        debug("OnIncomingRequest", "registration unacceptable\n");
        ResponseSetStatus(r, 600, "registration unacceptable");
        r->policy = &failure_policy;
        execute_policy(r, NULL /* reg is not used */, r->status);
      }
      else {
        /* Registration is accepted */
        /* XXX */
        ResponseSetStatus(r, 500, "Registration not implemented");
        r->policy = &failure_policy;
        execute_policy(r, NULL /* reg is not used */, r->status);
      }
    }
    else {
      debug("OnIncomingRequest", "Not a registrar\n");
      ResponseSetStatus(r, 501, "Not a registrar");
      r->policy = &failure_policy;
      execute_policy(r, NULL /* reg is not used */, r->status);
    }
  }
  else if ( !strcmp(method, "OPTIONS") ) {
    /* XXX Fill in default options, SDP */
    HeaderAppend(&r->headers, "Allow", 
      "ACK, BYE, CANCEL, INVITE, OPTIONS, REGISTER");

    c = CallSearch(callid);

    if ( c != NULL ) {
      pthread_mutex_unlock(&c->mutex); /* to avoid dead-lock */

      /* Let the policy_core and call thread interact */
      execute_policy(r, /* incoming request */ NULL, 200);
    }
    else if ( pobjEndpoint != NULL ) {
      /* ask the endpoint object for default SDP */
      /* Send the succssful response to remote */
      SessionDescription *session = 
        SIPEndpoint::GetDefaultSessionDescription();
      char *self_sdp = NULL;
      if ( session ) {
        self_sdp = SIPCall::FormatSDP(*session);
      }

      /* Check for Accept header if present and send accordingly */
      if ( self_sdp ) {
        if ( CallCompareContentType("application/sdp", &r->p) == true ) { 
          ResponseBuildDefault(r);
          ResponseSetBody(r, self_sdp, "application/sdp",
                             strlen(self_sdp));
          ResponseSetStatus(r, 200, NULL);
          ResponseSend(r, 0, 1);
        }
        else {
          /* Should we always accept application/sdp ? */
          ResponseBuildDefault(r);
          HeaderAppend(&r->headers, "Accept", (char *) "application/sdp");
          /* Should it be 406 or 415 !!! Should we add
           * Other headers like Accept-Encoding and Accept-Language.
           */
          ResponseSetStatus(r, 415, "You do not accept application/sdp !");
          ResponseSend(r, 0, 1);
        }
      }
      else {
        ResponseBuildDefault(r);
        ResponseSetStatus(r, 200, "No SDP available");
        ResponseSend(r, 0, 1);
      }
      /* Free mem for sdp */
      free_valid(self_sdp);
      if ( session != NULL ) {
        delete session;
      }
    }
    else {
      debug("OnIncomingRequest", "No endpoint available\n");
      ResponseSetStatus(r, 500, NULL);
      r->policy = &failure_policy;
      execute_policy(r, NULL /* reg is not used */, r->status);
    }
  } 
  else if ( !strcmp(method, "ACK")
         || !strcmp(method, "BYE") 
         || !strcmp(method, "INVITE")
         || !strcmp(method, "CANCEL") ) { /* Other INVITE, BYE, ACK, etc */

    /* Send 100 if not ACK */
    if ( strcmp(method, "ACK") != 0) {
      /* Searching may take a while, thus, send a provisional (1xx) response. */
      /* provisional responses are sent only for BYE, INVITE */
      ResponseSetStatus(r, 100, NULL); 
      ResponseSetBody(r, NULL, NULL, 0);
      ResponseBuildDefault(r);
      ResponseSend(r, 0, 0);
    }

    c = CallSearch(callid);

    if ( c != NULL && !strcmp(method, "INVITE") && c->callobj == NULL ) {
      /*
       *  This call object is already in IDLE state.
       * This is possible when the call was rejected due too
       * 401 Unauthorized error and the remote has retried the call
       * with the same callid.
       */
      pthread_mutex_unlock(&c->mutex);
      c = NULL;
    }

    if ( c == NULL && !strcmp(method, "INVITE") ) {
      /* This is for a new call, create call object */
      if ( pobjEndpoint != NULL ) {
        if ( !strcmp(method, "INVITE")) {
          c = CallCreate(callid);
          if ( c != NULL ) {
            SIPCall * pobjCall;
            pobjCall = (SIPCall *) pobjEndpoint->AllocateNewCallObject(callid);
            c->callobj = pobjCall;
            
            /* Unlock the mutex to avoid deadlock */
            pthread_mutex_unlock(&c->mutex);
          }
          else {
            debug("OnIncomingRequest", "Can not create call object\n");
            ResponseBuildDefault(r);
            ResponseSetStatus(r, 500, "Could not create call object");
            ResponseSend(r, 0, 1);
          
            free_valid(method);
            free_valid(callid);
            URI_Free(&uri);
            debug("OnIncomingRequest","Exiting\n");
            return;  
          }
        }
      }
    }

    if ( c != NULL ) {
      /* unlock the call thread mutex to avoid deadlock */
      pthread_mutex_unlock(&c->mutex);
    }

    /* Create the policy thread */
    execute_policy(r, /* incoming request */ NULL, 200);
  }
  else {
    /* Method is some unknown method.
     * Should we return 405 Method Not Allowed.
     * As I understand, 501 is more appropriate here.
     * 405 will be used if, e.g., you are trying
     * to perform INVITE on an http://... URI or
     * say GET for an sip:... URI.
     */
    debug("OnIncomingRequest", "Invalid method %s\n", method);
    ResponseBuildDefault(r);
    ResponseSetStatus(r, 501, NULL);
    HeaderAppend(&r->headers, "Allow", "ACK, BYE, CANCEL, INVITE, OPTIONS, REGISTER");
    ResponseSend(r, 0, 1);
  }

  free_valid(method);
  free_valid(callid);
  URI_Free(&uri);
  debug("OnIncomingRequest","Exiting\n");
  return;  
}

/*
 * This method does the comparision of the
 * source and new SDP for matching m= appropriately.
 * The new SDP stringis returned.
 * If there is no match then NULL is returned
 * and the error is set appropriately.
 * The caller should free error 
 */
char * CallMatchSessionWithSource(const char *src_sdp,
                                  const char *new_sdp,
                                  char **error)
{
  SessionDescription *sdp1, *sdp2, *sdp;
  int oGlobalMatch = 0;
  DString e;

  
  debug("CallMatchSessionWithSource", "SDP1----\n%s\nSDP2-----\n%s\n",
        src_sdp ? src_sdp : "NULL",
        new_sdp ? new_sdp : "NULL");
  

  sdp1 = SIPCall::DeformatSDP(src_sdp, 0);
  if ( src_sdp == NULL || sdp1 == NULL ) {
    return new_sdp ? strdup(new_sdp) : NULL;
  }

  sdp2 = SIPCall::DeformatSDP(new_sdp, 0);
  if ( sdp2 == NULL ) {
    delete sdp1; sdp1=NULL;
    if ( *error ) {
      *error = strdup("399 No SDP was by this endpoint");
    }
    return NULL;
  }

  DStringInit(&e);
  int added = 0;

  /* Both sdp1 and sdp2 are valid */
  /* Build new sdp based on sdp2 */
  sdp = sdp2->Clone();
  sdp->DeleteSession(); /* This should delete all sessions */
  for ( int i=0; i<sdp1->GetMediaCount(); i++ ) {
    MediaInfo & media1 = (*sdp1)[i];
    MediaInfo media2;

    debug("CallMatchSessionWithSource",
          "Trying to find session\n");
    if ( sdp2->FindSession(media2, media1.direction, media1.datatype,
                           media1.transport)) {
      int oMatch = 1;

      debug("CallMatchSessionWithSource",
            "session found\n");

      if ( media1.transport != media2.transport ) {
        oMatch = 0;
        DStringPrintf(&e, "%s302 Incompatible transport protocol",
          added ? "\n" : "");
        added = 1;
      }

      if ( media1.network != media2.network ) {
        oMatch = 0;
        DStringPrintf(&e, "%s300 Incompatible network protocol",
          added ? "\n" : "");
        added = 1;
      }
      
      if ( media1.addresstype != media2.addresstype ) {
        oMatch = 0;
        DStringPrintf(&e, "%s301 Incompatible network address formats",
          added ? "\n" : "");
        added = 1;
      }

      if ( media1.GetRxAddress().IsMulticast() &&
        media2.GetRxAddress().IsMulticast() == false) {
        oMatch = 0;
        DStringPrintf(&e, "%s330 Multicast not available",
          added ? "\n" : "");
        added = 1;
      }

      if ( oMatch ) {
        /* Matching found */
        /* Match the formats */
        MediaInfo media = media2;
        media.RemoveAllFormats(); /* This should remove all formats */

        debug("CallMatchSessionWithSource",
              "Match found\n");

        for ( int j=0; j<media1.GetFormatCount(); j++) {
          FormatInfo & format1 = media1[j];
          if ( media2.IsFormatPresent(format1) ) {
            media.AddFormat(format1);
          }
        }
        if ( media.GetFormatCount() <= 0 ) {
          /* There is no format for this media */
          /* Set the port to 0 */
          debug("CallMatchSessionWithSource", "No matching format, set port=0\n");
          media.SetRxPort(0); 
        }
        else {
          /* Atleast one matching format */
          oGlobalMatch = 1;
        }

        sdp2->DeleteSession(media2);
        sdp->AddSession(media);
      }
      else {
        debug("CallMatchSessionWithSource", "Match rejected: %s\n",
              DStringValue(&e));
        MediaInfo media = media1;
        media.SetRxPort(0);
        media.RemoveAllFormats();
        sdp->AddSession(media);
      }
    }
    else {
      /* This session is not present */
      debug("CallMatchSessionWithSource",
            "session not found\n");
      media2 = media1;
      media2.SetRxPort(0);
      media2.RemoveAllFormats();
      sdp->AddSession(media2);
    }
  }

  char *sdp_str = NULL;
  if ( oGlobalMatch && (sdp != NULL ) ) {
    sdp_str = SIPCall::FormatSDP(*sdp, NULL);
    /*
    debug("CallMatchSessionWithSource", "Formatted SDP---\n%s\n", 
          sdp_str);
    */
  }
  else {
    debug("CallMatchSessionWithSource", "Could not format %s\n",
          oGlobalMatch ? "SDP is null" : "No match" );
    if ( error ) {
      *error = DStringValue(&e);
      if ( !*error || !**error ) {
        *error = strdup("399 There was a mismatch in SDP");
      }
      /* Do not free e */
    }
  }

  if ( sdp ) {
    delete sdp;
  }
  if ( sdp1 ) {
    delete sdp1;
  }
  if ( sdp2 ) {
    delete sdp2;
  }
  return sdp_str;
}


/*
 * Get and return the canonical hostname of this machine.
 * Not thread-safe -- only intended for startup time.
 */
char *getcanonicalhostname()
{
  char buf[MAXHOSTNAMELEN];
  struct hostent *hep;
  struct in_addr addr;

  gethostname(buf, MAXHOSTNAMELEN);
  
  hep = gethostbyname(buf);
  if (hep == NULL) {
    error("getcanonicalhostname", "gethostbyname failed on %s, h_errno = %d\n", 
          buf, h_errno);
    return NULL;
  }

  addr.s_addr = inet_addr("127.0.0.1");

  if ( hep->h_length >= sizeof(addr)) {
    memcpy(&addr, hep->h_addr, sizeof(addr));
  }

  hep = gethostbyaddr((char *)&(addr.s_addr), 
                      sizeof(addr.s_addr), AF_INET);
  if (hep == NULL) {
    error("getcanonicalhostname", "gethostbyaddr failed on %s, h_errno = %d\n", 
          inet_ntoa(addr),
          h_errno);
    return NULL;
  }

  if ( hep->h_length >= sizeof(addr)) {
    memcpy(&addr, hep->h_addr, sizeof(addr));
  }

  if ( !SIPLibrary::oUseNumericAddress &&
       hep->h_name && strchr(hep->h_name, '.') != NULL ) {
     return strdup(hep->h_name);
  }
  else {
     return strdup(inet_ntoa(addr));
  }
}


/* ------------------- Checking license ---------------------
 */

#ifdef DO_LICENSING

static char *hostname;
static char *internal_code = "irt_libsip++_1.0";

struct reg_info {
  char *domain;
  char *exp_str;
  time_t t_expire;
  char *md5;
};


/* parse_registration()
 * Parses a registration string: of the form "domain:YYYY-MM-DD:md5" or
 * "domain:-:md5" for unlimited-time licenses.
 * Return false if it could not be parsed, otherwise fills in reg_info.
 * reg_info will contain pointers into registration string, which must be
 * writeable.
 * Not thread-safe; should only be called at program startup.
 */
static int parse_registration(char *registration, struct reg_info *reg_info)
{
  char *lasts;
  struct tm tm_expire;

  reg_info->domain   = strtok_r(registration, ":", &lasts);
  reg_info->exp_str  = strtok_r(NULL, ":", &lasts);
  reg_info->md5      = strtok_r(NULL, ":\n", &lasts);

  if (reg_info->domain == NULL || reg_info->exp_str == NULL || 
      reg_info->md5 == NULL) {
    fprintf(stderr, "check_license: Missing registration field\n");
    return 0;
  }
  
  if (strcmp(reg_info->exp_str, "-") == 0) {
    reg_info->t_expire = 0;
  }
  else {
    /* Some strptime() implementations (glibc, FreeBSD) don't initialize
       all the fields of struct tm, so set it to zero first. */
    memset(&tm_expire, '\0', sizeof(tm_expire));
    if (strptime(reg_info->exp_str, "%Y-%m-%d", &tm_expire) == NULL) {
      fprintf(stderr, "check_license: Invalid expiration date %s\n",
              reg_info->exp_str);
      return 0;
    }
    
    reg_info->t_expire = mktime(&tm_expire);
    if (reg_info->t_expire == (time_t)-1) {
      fprintf(stderr, "check_license: Invalid expiration date %s\n",
              reg_info->exp_str);
      return 0;
    }
  }
  return 1;
} /* parse_registration */


/* check_reg_info()
 * Confirm that the registration information corresponds
 * to this host, and is up-to-date.  Returns false if not.
 */
static int check_reg_info(struct reg_info *reg_info)
{
  time_t t;
  char *domain_start;

  if (reg_info->t_expire != 0) {
    time(&t);
    if (t > reg_info->t_expire) {
      fprintf(stderr,
              "check_license: expiration date %s (%ld) is in the past\n",
              reg_info->exp_str, (long)reg_info->t_expire);
      return 0;
    }
  }

  domain_start = hostname + strlen(hostname) - strlen(reg_info->domain);
  if (strcmp(domain_start, reg_info->domain) != 0) {
    fprintf(stderr, "check_license: host %s not in domain %s\n",
            hostname, reg_info->domain);
    return 0;
  }

  return 1;
} /* check_registration */


/* check_md5()
 * Validate the the MD5 hash of the registration information.
 */
static int check_md5(struct reg_info *reg_info)
{
  int ret = 0, n;
  char buf[256], *md5;

#ifdef HAVE_SNPRINTF
  n = snprintf(buf, sizeof(buf), "%s:%s:%s", reg_info->domain,
               reg_info->exp_str, internal_code);
#else
  if (strlen(reg_info->domain) + strlen(reg_info->exp_str) +
      strlen(internal_code) >= sizeof(buf)) {
    fprintf(stderr, "check_license: snprintf error or overflow\n");
  }
  n = sprintf(buf, "%s:%s:%s", reg_info->domain,
              reg_info->exp_str, internal_code);
#endif
  
  if (n < 0 || n >= sizeof(buf)) {
    fprintf(stderr, "check_license: snprintf error or overflow\n");
  }
  md5 = MD5(buf);
  if (strcmp(md5, reg_info->md5) == 0) {
    ret = 1;
  }
  else {
    fprintf(stderr, "check_license: hash %s is not valid\n", reg_info->md5);
  }
  free(md5); /* this is malloc'ed storage */
  return ret;
}


/* 
 * Returns 1 if the running system has a valid registration code.
 */
static int registration_valid(const char * filename)
{
  struct reg_info reg_info;
  char code[BUFSIZ];
  FILE *regs;

  if ( filename == NULL ) {
    error("check_license", "No license file specified\n");
    return 0;
  }

  if ((regs = fopen(filename, "r"))
      == NULL) { 
    error("check_license", "Cannot open license file %s\n", 
          filename);
    return 0;
  }
  
  while (fgets(code, BUFSIZ, regs) != NULL) {
    if (parse_registration(code, &reg_info) &&
        check_md5(&reg_info) &&
        check_reg_info(&reg_info)) {
      fclose(regs);
      return 1;
    }
  }
  fclose(regs);
  return 0;
} /* registration_valid */


/* 
 * if match, continue, else terminate program
 */
int SIPLibrary::CheckLicense(const char *filename)
{
  /* hostname is a static variable */
  hostname = getcanonicalhostname();
  if (hostname == NULL) {
    error("CheckLicense", "getcanonicalhostname");
  }

  if (!registration_valid(filename)) {
    return 1;
  }
  /* everything ok, continue */
  free_valid(hostname);
  hostname = NULL; /* so that next call does not cause a memory leak */
  return 0;
} /* check_registration */

#else /* ! DO_LICENSING */

/* 
 * check_registration()
 * Dummy registration check if we configured with licensing turned off.
 */
int SIPLibrary::CheckLicense(const char *filename)
{
  return -1;
} /* check_registration */

#endif

/********************* END OF FILE *************************************
*/
