/***************************** FILE HEADER *****************************
#               Copyright (c) 1999 by Columbia University              *
#                        All Rights Reserved                           *
#***********************************************************************
#***********************************************************************
#
#  File Name    : sipapi.cxx
#
#  Module       : SIP API implementation.
#
#  Author(s)    : Kundan Singh (kns10@cs.columbia.edu)
#
#  Created      : 14 Oct 1999
#
#  Description  : Implementation of most of the methods are in this file.
#                
#  Notes        : 
#
#***********************************************************************
#                Modification History (Reverse Order)
#-----------------------------------------------------------------------
#   Date         Author              Description
#-----------------------------------------------------------------------
#              |             | 
#              |             | 
#-----------------------------------------------------------------------
*/
 
/***********************************************************************
******************* INCLUDES *******************************************
*/

#include <time.h>

#include "iptelapi.h" /* IptelAddress,Session etc */
#include "sipapi.h"   /* SIPCall, Endpoint etc */
#include "msgflow.h"  /* For msg flow trace */
#include "sipcall.h"
#include "sipreg.h" 

#ifdef __cplusplus
extern "C" {
#endif

#include <sys/errno.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <time.h>
#include <sys/time.h>
#include "sysdep.h"
#include "strdupn.h" /* free_valid */
#include "request-sip.h"
#include "policy.h"
#include "timespec.h"
#include "uri.h"
#include "pthread-util.h"

#ifdef __cplusplus
}
#endif

#include "log.h"
#include "msgflow.h"

/* Default listening port for SIP */
#define DEFAULT_SIP_PORTNO  5060

static sipreg_t *Rg = NULL;    /* list of calls */
static pthread_mutex_t Rg_lock = PTHREAD_MUTEX_INITIALIZER;

/***********************************************************************
******************* Extern C Function protypes *************************
*/

/* XXX should we use SIPLibrary::Random32 in all c++ files ? */
extern "C" unsigned long random32();

extern "C" struct tm *gmtime_r(const time_t *clock, struct tm *res);

/* Local prototypes */
void * SipRegThread(void *);

extern IptelAddress SipAddressToIptelAddress(sip_addr_t addr);
extern sip_addr_t   URI_ParseNew(const char *);
request_t * FormatRegistration(const char *method,
                  int type,
                  const char *callid,
                  unsigned int cseq,
                  body_t * body,
                  sipreg_t *c);

/*
* SipRegCreate()
* Create a new sipreg object and the sipreg thread.
* It also adds the sipreg in the list.
*/
sipreg_t *SipRegCreate(const char *id)
{
  int e; /* error */
  sipreg_t *c = (sipreg_t *) calloc(1, sizeof(sipreg_t));
  char *callid = NULL;

  c->state = SIPREG_IDLE;
  if ( id == NULL ) {
    /* create a new callid */
    const char *host = SIPLibrary::HostName;
    DString NewCallId;
    DStringInit(&NewCallId);
    DStringPrintf(&NewCallId,"%u@%s", (unsigned int) random32(), 
                  (host != NULL) ? 
                  host : "unknown");
    /* We use strdup() instead of directly taking the value
    * of DString() to save space. Note that DString()
    * may allocate some additional bytes */
    c->callid = strdup(DStringValue(&NewCallId));
    callid = strdup(c->callid);
    DStringFree(&NewCallId);  
  }
  else {
    c->callid = strdup(id);
    callid = strdup(c->callid);
  }

  c->self      = NULL;
  c->remote    = NULL;
  memset(&c->contact, 0, sizeof(c->contact));
  c->contact.expires = 3600;
  memset(&c->expireson, 0, sizeof(c->expireson));

  c->lastcseq_tx  = 0;
  c->mqueue    = NULL;
  //c->tid       = NULL; /* does not work for unix */
  c->next      = NULL;

  pthread_mutex_init(&c->mutex, NULL);
  pthread_cond_init(&c->cond, NULL);
  c->done  = 0;

  c->regobj = NULL;
  c->type  = SOCK_DGRAM; /* default is SIP over TCP */  
  c->unauth_count = 0;
  c->userid = NULL;
  c->password = NULL;

  pthread_mutex_lock(&c->mutex);

  SipRegAdd(c);

  /* Create the thread */
  if ((e = pthread_create_detached(&c->tid, NULL, 
                          SipRegThread, (void *) strdup(callid))) != 0) {
    error("SipRegCreate","pthread_create_detached failed\n");
  }

  if ( e != 0 ) {
    SipRegFree(c);
    c = NULL;
  }

  /*
  if ( c ) {
    pthread_mutex_unlock(&c->mutex);
  }
  */

  free(callid);

  return c;
} /* CallCreate */

/* 
* Add the reg in the regs list,
* it is assumed that the reg mutex is locked
*/
void SipRegAdd(sipreg_t *c)
{
  /* Add the new reg in the begining */
  pthread_mutex_lock(&Rg_lock);
  c->next = Rg;
  Rg = c;
  pthread_mutex_unlock(&Rg_lock);
} /* CallAdd */

/*
* SipRegFree()
* If present, remove request from linked list of requests.
* It is assumed that the mutex is locked.
* If the reg thread is active, this should only
* be called from that thread, to avoid any
* race condition !
*/
void SipRegFree(sipreg_t *c)
{
  sipreg_t *cp;
  message_t *m, *m_next;

  if (!c) return;

  debug("SipRegFree", "Freeing %p\n", c);

  /* set this to terminating no new reg is tried. */
  /* This is against the normal state transition if the state
  * is already IDLE, since you are transitioning from 
  * IDLE to TERMINATING. But that's OK since we are terminating
  * the call thread (FSM) anyway.
  */
  c->state = SIPREG_TERMINATING; 

  pthread_mutex_unlock(&c->mutex);
  pthread_mutex_lock(&Rg_lock);
  pthread_mutex_lock(&c->mutex);

  /* Remove from list. */
  if (c == Rg) {
    Rg = Rg->next;
  }
  else {
    for (cp = Rg; cp && cp->next != c; cp = cp->next) ;
    if (cp) cp->next = c->next;
  }
  pthread_mutex_unlock(&Rg_lock);

  /* Free the message queue */
  for (m = c->mqueue; m != NULL; m = m_next) {
    m_next = m->next;
    MessageDelete(m);
  }
  /* We set all the pointers to NULL,
  * after freeing to avoid any misuse of freed
  * memory !!! */

  c->mqueue = NULL;
  c->regobj = NULL;
  free_valid(c->callid);     c->callid     = NULL;
  free_valid(c->self);       c->self       = NULL;
  free_valid(c->remote);     c->remote     = NULL;
  URI_Free(&c->remote_uri);
  URI_Free(&c->proxy);
  URI_Free(&c->contact.uri);
  free_valid(c->contact.nameaddr);
  free_valid(c->contact.comment);

  /* free the body */
  free_valid(c->body.content); c->body.content = NULL;
  free_valid(c->body.content_type); c->body.content_type = NULL;
  /* Free any request or response parsed or unparsed headers */
  HTTP_Free(&c->p_irs);
  HTTP_Free(&c->p);
  HeaderFree(&c->h_orq);

  free_valid(c->userid);
  free_valid(c->userid);

  pthread_mutex_unlock(&c->mutex); /* Release lock before deleting */
  pthread_mutex_destroy(&c->mutex);
  pthread_cond_destroy(&c->cond);
  free(c);

  debug("SipRegFree", "Exiting\n");
  return;
} /* SipRegFree */

/*
* SipRegPrint()
* This prints the existing calls on the output stream FILE *fp
* If fp is NULL then prints to stdout.
*/
void SipRegPrint(FILE *fp)
{
  sipreg_t *cp;

  if ( !fp ) {
    fp = stdout;
  }

  pthread_mutex_lock(&Rg_lock);
  for (cp = Rg; cp; cp = cp->next) {
    /*  Print per call info here XXX */
  }
  pthread_mutex_unlock(&Rg_lock);
 
  return;
} /* SipRegPrint */

 
/*
 * SipRegSearch()
 * Determine if the request 'c' with callid 'callid' exists in the list of
 * calls. 
 * N.B.: The returned reg has its mutex locked!  You must unlock it
 * when you're done with it!  
 */
sipreg_t *SipRegSearch(const char *callid)
{
  sipreg_t *cp, *result = NULL;

  if ( !callid )
    return NULL;

  pthread_mutex_lock(&Rg_lock);
  for (cp = Rg; cp; cp = cp->next) {
    if (!strcmp(cp->callid, callid) ) {
      result = cp;
      pthread_mutex_lock(&cp->mutex);
      pthread_mutex_unlock(&Rg_lock); /* Release Rg_lock to avoid deadlock */
      return result;
    }
  }
  pthread_mutex_unlock(&Rg_lock);

  return NULL; /* not found */
} /* SipRegSearch */


/*
 * SipRegFind()
 * Determine if the request 'c' with remote address  exists in the list of
 * calls. 
 */
sipreg_t *SipRegFind(const char *remoteaddr)
{
  sipreg_t *cp, *result = NULL;

  if ( !remoteaddr )
    return NULL;

  pthread_mutex_lock(&Rg_lock);
  for (cp = Rg; cp; cp = cp->next) {
    if (!strcmp(cp->remote, remoteaddr)) {
      result = cp;
      pthread_mutex_lock(&cp->mutex);
      if ( cp->regobj == NULL ) {
        pthread_mutex_unlock(&cp->mutex);
        continue;
      }
      pthread_mutex_unlock(&Rg_lock); /* Release Rg_lock to avoid deadlock */
      return result;
    }
  }
  pthread_mutex_unlock(&Rg_lock);

  return NULL; /* not found */
} /* SipRegSearch */

/*
* Remove the endpoint reference from all the registration
* objeccts in the queue
*/
void RemoveEndpointFromAllReg(SIPEndpoint *ep)
{
  sipreg_t *cp;

  if ( !ep )
    return;

  pthread_mutex_lock(&Rg_lock);
  for (cp = Rg; cp; cp = cp->next) {
      pthread_mutex_lock(&cp->mutex);
        if ( cp->regobj == ep ) {
              cp->regobj = NULL;
        }
      pthread_mutex_unlock(&Rg_lock); /* Release Rg_lock to avoid deadlock */
  }
  pthread_mutex_unlock(&Rg_lock);

  return; 
}


/*
 * SipRegDone();
 * This sets the flag for call thread to terminate
 * and signals the condition.
 * It is assumed that the mutex is locked.
 */
void SipRegDone(sipreg_t *c)
{
  c->done = 1;
  pthread_cond_broadcast(&c->cond);
}

/*
 * SipRegComplaint()
 */
void SipRegComplaint(sipreg_t *c)
{
  error("SipRegComplaint", "Somthing wrong happened\n");
}

/*
 * SipRegCleanup() is empty XXX 
 */
int SipRegCleanup(sipreg_t *c)
{
  return 0;
}

/*
 * Set the reg status 
 */
void SipRegSetStatus(sipreg_t *c, int status, const char *reason)
{
  c->p.status = status;
  strassign(&c->p.reason, reason);
}

/*
 * SipRegThread().
 * This implements the reg thread 
 */
void *SipRegThread(void *cid)
{
  char *callid = (char *)cid;
  sipreg_t *c = NULL;
  struct message_t *m, *m_next, **prev_next;
  int thr_err;
  int first_time = 1;

  if ( cid == NULL) {
    error("SipRegThread", "cid is NULL\n");
    return NULL;
  }

  debug("SipRegThread", "called with %s\n", callid);

  if ( (c = SipRegSearch(callid)) == NULL) {
    debug("SipRegThread", "Call is not in the list\n");
    free_valid(callid);
    return NULL;
  }

  free_valid(callid);

  /* c->mutex is now locked. */

  while (!c->done) {
    /* Do not wait on the cond for the first time,
       to avoid lost wakeup problem. The
       API thread may signal cond before it is started.
    */
    if ( !first_time ) {
      if ( c->regobj == NULL ) {
        /* No associated call obj wait for 40 seconds */
        struct timeval now;

        gettimeofday(&now, NULL);
        struct timespec timeout;
        timeout = timespec_delay(now, 40.0);

        thr_err = pthread_cond_timedwait(&c->cond, &c->mutex,
                                         &timeout);
        debug("SipRegThread", "Wait exiting when regobj is NULL\n");
      }
      else if ( c->state != SIPREG_ACTIVE ) {        
        /* Not active, wait for 60 seconds */
        struct timeval now;

        gettimeofday(&now, NULL);
        struct timespec timeout;
        timeout = timespec_delay(now, 60.0);

        thr_err = pthread_cond_timedwait(&c->cond, &c->mutex, &timeout);
        debug("SipRegThread", "Wait existing in NON-ACTIVE state.\n");
      } 
      else {
        if ( c->expireson.tv_sec > 0 || c->expireson.tv_nsec > 0 ) {
          debug("SipRegThread", "Timed wait in ACTIVE state\n");
          thr_err = pthread_cond_timedwait(&c->cond, &c->mutex,
                                           &c->expireson);
        }
        else {
          thr_err = pthread_cond_wait(&c->cond, &c->mutex);
        }
        debug("SipRegThread", "Wait existing in ACTIVE state\n");
      }

      if (thr_err) {
        if (thr_err != ETIME && thr_err != ETIMEDOUT) {
          error("SipRegThread", "Error waiting for proxy events: %s (%d)",
                strerror(thr_err), thr_err);
          continue;
        }
        else {
          /* The timeout expired. */
          if ( c->regobj == NULL || c->state != SIPREG_ACTIVE ) {
            /* Don't trigger another timeout unless one is set */
            goto cleanup;
          }
          else {
            /* Trigger the timeout message */
            c->expireson.tv_sec = 0;
            c->expireson.tv_nsec = 0;
            MessageEnqueuePseudoRequest(&c->mqueue, "REFRESH", 
              nobody_handled, NULL);
            pthread_cond_broadcast(&c->cond);
            /* Fall through to message processing */
          }
        }
      }
    }
    first_time = 0;

    /* Walk the message queue, and send requests not already handled */
    /* If the handler for any message calls SipRegDone(), stop
       immediately. */
    for (m = c->mqueue; m != NULL && !c->done; m = m->next) {
      if (m->handled >= policy_handled)
        continue;

      if (m->type != message_real_message && m->type != message_pseudo_message &&
          m->type != message_api_message ) {
        error("SipRegThread",
              "Surprising message type %d on mqueue", m->type);
        m->handled = finished_handling;
        continue;
      }

      if (m->p.type == http_request) {
        if (SipRegHandleRequest(c, m) != 0) {
          SipRegComplaint(c);
          goto cleanup;
        }
      }
      else if (m->p.type == http_response) {
        if (SipRegHandleResponse(c, m) != 0) {
          SipRegComplaint(c);
          goto cleanup;
        }
      }
      else {
        error("SipRegThread",
              "Surprising message parse type %d on mqueue", m->p.type);
        m->handled = finished_handling;
      }
    }

    /* Remove all nodes marked finished_handling from the queue. */
    prev_next = &c->mqueue;
    for (m = c->mqueue; m != NULL; m = m_next) {
      m_next = m->next;
      if (m->handled == finished_handling) {
        *prev_next = m->next;
        MessageDelete(m);
      } else {
        prev_next = &m->next;
      }
    }
  }

cleanup:  /* Jump here when any policy function returns non-zero */

  SipRegCleanup(c); /* should this be after cleanup: label XXX.
                   * what about in proxy.c and user.cpp */

  /* XXX Alert the requests to shut down */

  /* XXX Alert the API call objects to terminate */

  debug("SipRegThread", "c=%p finished\n", c);

  c->regobj = NULL;

  SipRegFree(c);

  return NULL;
}

int SipRegUpdateExpiry(sipreg_t *c, message_t *m)
{
  time_t t = SIP_Expires(&m->p);

  if ( t > 1 ) {
    struct timeval now;
    time_t t2 = time(0);
    struct tm tm2;

    //gmtime_r(&t2, &tm2);

    //t2 = mktime(&tm2);

    long v = t - t2;
    
    gettimeofday(&now, NULL);
    
    if ( v > 10 ) {
      c->expireson = timespec_delay(now, (double) (
                                                   v - 10 ));
    }
    else if ( v > 2 ) { /* Min 2 seconds */
      c->expireson = timespec_delay(now, (double) (
                                                   v ));
    }
    else {
      c->expireson = timespec_delay(now, (double) (
                                                   10 ));
    }

    debug("SipRegUpdateExpiry", "Timeout delay is %ld for expiry\n", v);

    /* We must signal the mutex so that the thread uses the new
     * timeout for waiting */
    pthread_cond_broadcast(&c->cond);
  }
  else {
    debug("SipRegUpdateExpiry", "Could not update the expiry (%ld)\n",
          (long) t);
  }

  return 0;
}

int SipRegHandleRequest(sipreg_t *c, message_t *m)
{
  request_t *rp = NULL;

  if ( c->state != SIPREG_IDLE ) {
    rp = RequestFind(&c->p, 0);
  }

  /* State machine processing */
  switch ( c->state ) {
    case SIPREG_IDLE: /* No active call */
      debug("CallHandleRequest", "received %s in IDLE state\n", m->p.method);

      if ( strcmp( m->p.method, "REGISTER") == 0 ||
           strcmp( m->p.method, "UNREGISTER") == 0 ) {
          /*
           * outgoing reg.
           * Invoke the request thread/policy for new REGISTER.
           */
          if ( strcmp(m->p.method, "REGISTER") == 0 ) {
            c->state = SIPREG_OUTGOING_1;
          }
          else {
            /* Probably terminating a previous boot's registration */
            c->state = SIPREG_TERMINATING;
          }

          request_t *r;

          r = FormatRegistration(m->p.method, c->type, 
            c->callid, ++c->lastcseq_tx, 
            c->body.content_type ? &c->body : NULL,
            c);

          URI_Free(&c->remote_uri);
          c->remote_uri = URI_Copy(r->p.URI);

          /* Should store the outgoing request here */
          HTTP_Duplicate(&c->p, &r->p);

          CallCreatePolicyThread(r, c->remote_uri, c->proxy);
      }
      else {
        /* If this is a remote request then send back appropriate response. */
        debug("SipRegHandleRequest", "Message in unexpected state\n");
      }
      break;

    case SIPREG_OUTGOING_1: /* waiting for final response from remote  */
    case SIPREG_OUTGOING_2: /* Waiting for final response for reREG */
    case SIPREG_OUTGOING_3: /* Waiting for final response for refresh */
      debug("SipRegHandleRequest", "received %s in OUTGOING1 state\n", m->p.method);
      if ( strcmp( m->p.method, "UNREGISTER") == 0) {
          c->state = SIPREG_TERMINATING;

          if ( rp == NULL ) {
            debug("SipRegHandleRequest", "Ignoring UNREG no original req\n");
          }
          else {
            /* use existing request */
            MessageEnqueuePseudoRequest(&rp->mqueue, "CANCEL", 
              nobody_handled, NULL);
            pthread_cond_broadcast(&rp->cond);
          }
      }
      else {
        debug("SipRegHandleRequest", "Message in unexpected state\n");
      }
      break;

    case SIPREG_ACTIVE:
      debug("SipRegHandleRequest", "received %s in ACTIVE state\n", m->p.method);
      if ( strcmp( m->p.method, "REGISTER") == 0) {
        /* Inform the API that the call is closed */
        c->state = SIPREG_OUTGOING_2;

        request_t *r;
        
        r = FormatRegistration(m->p.method, c->type, 
                               c->callid, ++c->lastcseq_tx, 
                               c->body.content_type ? &c->body : NULL,
                               c);

        URI_Free(&c->remote_uri);
        c->remote_uri = URI_Copy(r->p.URI);
        
        /* Should store the outgoing request here */
        HTTP_Duplicate(&c->p, &r->p);
        
        CallCreatePolicyThread(r, c->remote_uri, c->proxy);
      }
      else if ( strcmp( m->p.method, "UNREGISTER") == 0) {
        c->state = SIPREG_TERMINATING;
        request_t *r;
        
        r = FormatRegistration(m->p.method, c->type, 
                               c->callid, ++c->lastcseq_tx, 
                               c->body.content_type ? &c->body : NULL,
                               c);
        
        URI_Free(&c->remote_uri);
        c->remote_uri = URI_Copy(r->p.URI);
        
        /* Should store the outgoing request here */
        HTTP_Duplicate(&c->p, &r->p);
        
        CallCreatePolicyThread(r, c->remote_uri, c->proxy);
      }
      else if ( strcmp( m->p.method, "REFRESH") == 0 ) {
        c->state = SIPREG_OUTGOING_3;

        request_t *r;
        
        r = FormatRegistration("REGISTER", c->type, 
                               c->callid, ++c->lastcseq_tx, 
                               c->body.content_type ? &c->body : NULL,
                               c);

        URI_Free(&c->remote_uri);
        c->remote_uri = URI_Copy(r->p.URI);
        
        /* Should store the outgoing request here */
        HTTP_Duplicate(&c->p, &r->p);
        
        CallCreatePolicyThread(r, c->remote_uri, c->proxy);
      }
      else {
        debug("SipRegHandleRequest", "Message in unexpected state\n");
      }
      break;

    case SIPREG_TERMINATING: /* call is being terminated */
      debug("SipRegHandleRequest", "received %s in TERMINATING state\n", m->p.method);
      debug("SipRegHandleRequest", "Message ignored\n");
      break;

    default:
      error("SipRegHandleRequest", "surprising state %d\n", c->state);
      break;
  }

  if (  rp != NULL ) {
    pthread_mutex_unlock(&rp->mutex);
  }

  m->handled = finished_handling;

  return 0;
}

/*
 * Handle the response(s) from remote
 */
int SipRegHandleResponse(sipreg_t *c, message_t *m)
{
  request_t *rp = NULL;

  if ( m->type == message_real_message ) {
    rp = RequestFind(&m->p, 0 /* hash is not known */);

    /* Update the incoming response */
    HTTP_Duplicate(&c->p_irs, &m->p);
  }
  else if ( m->type == message_pseudo_message ) {
    rp = RequestFind(&c->p, 0);
  }
  else {
    debug("SipRegHandleResponse", "No original request\n");
  }

  if ( m->p.status >= 100 && m->p.status < 700) {
    /* Contact address is ignored */

    if ( m->p.status < 200 ) {
      debug("SipRegHandleResponse", 
            "ignoring non 100 provisional response\n");
    }
    else {
      /* State machine processing */
      switch ( c->state ) {
      case SIPREG_IDLE: /* No active call */
        debug("SipRegHandleResponse", "received %d in IDLE state\n",
              m->p.status);
        debug("SipRegHandleResponse", "Response is ignored\n");
        break;

      case SIPREG_OUTGOING_1:
      case SIPREG_OUTGOING_2:
        /*
         * Possible messages : 1xx-6xx for an outgoing UNREGISTER.
         */
        if ( m->p.status < 300 ) {
          /* Successful response */
          c->state = SIPREG_ACTIVE;

          SipRegUpdateExpiry(c, m);

          c->unauth_count = 0;

          if ( c->regobj ) {
            pthread_mutex_unlock(&c->mutex); /* This is important to avoid deadlock */
            c->regobj->OnRegistrationSuccess();
            pthread_mutex_lock(&c->mutex);
          }
          else {
            debug("SipRegHandleResponse", "No call object\n");
          }
        }
        else if ( m->p.status < 700 ) {

          c->state = SIPREG_IDLE;

          if ( c->regobj ) {
            int e = -1; /* Error */
            http_authenticate_t auth;
            char *userid = NULL;
            char *password = NULL;

            memset(&auth, 0, sizeof(auth));

            if ( (m->p.status == 401 || m->p.status == 407) 
                 && (c->unauth_count < 1)) {
              char *auth_method;
              /* Authorization needed */

              auth = HTTP_Authenticate(&m->p);
              //strassign(&auth.opaque, "86712313163187632617");
              strassign(&auth.http_method, "REGISTER");

              c->unauth_count ++;

              switch(auth.method) {
              case auth_basic:
                auth_method = strdup("basic");
                break;
              case auth_digest:
                auth_method = strdup("digest");
                break;
              case auth_none:
                auth_method = strdup("none");
                break;
              case auth_pgp:
                auth_method = strdup("pgp");
                break;
              default:
                auth_method = strdup("none");
                break;
              }

              pthread_mutex_unlock(&c->mutex); /* This is important to
                                                avoid deadlock */
              e = c->regobj->OnAuthorizationNeeded(auth_method,
                                               auth.realm,
                                               userid,
                                               password);
              pthread_mutex_lock(&c->mutex);

              free_valid(auth_method);

              strassign(&c->userid, userid);
              strassign(&c->password, password);
            }

            if ( e != 0 ) {
              pthread_mutex_unlock(&c->mutex); /* This is important to
                                                avoid deadlock */
              c->regobj->OnRegistrationFailed(m->p.status,
                                              m->p.reason);
              pthread_mutex_lock(&c->mutex);
            }

            if ( e == 0 ) {
              char * authorization;
              char *uristr;

              uristr = URI_AsString(c->remote_uri);

              authorization = HTTP_BuildAuthorizationHeader(
                                            &auth,
                                            userid,
                                            password,
                                            uristr);
              free_valid(uristr);
              
              if ( authorization ) {
                /* Set header here */
                if ( m->p.status == 401 ) {
                  c->regobj->h_orq -= (const char *) "Authorization";
                  c->regobj->h_orq +=
                    GenericHeaderValue("Authorization", 
                                       authorization);
                }
                else {
                  c->regobj->h_orq -= (const char *) "Proxy-Authorization";
                  c->regobj->h_orq +=
                    GenericHeaderValue("Proxy-Authorization", 
                                       authorization);
                }
                free_valid(authorization);

                /*
                 * outgoing reg.
                 * Invoke the request thread/policy for new REGISTER.
                 */
                c->state = SIPREG_OUTGOING_1;

                request_t *r;

                r = FormatRegistration("REGISTER", c->type, 
                                       c->callid, ++c->lastcseq_tx, 
                                       c->body.content_type ? &c->body : NULL,
                                       c);

                URI_Free(&c->remote_uri);
                c->remote_uri = URI_Copy(r->p.URI);

                /* Should store the outgoing request here */
                HTTP_Duplicate(&c->p, &r->p);

                CallCreatePolicyThread(r, c->remote_uri, c->proxy);

              }
              else {
                /* Failed */
                c->regobj = NULL;
              }
            }
            else {
              /* Set the registration object to NULL,
                 So that this SipReg is deleted */
              c->regobj = NULL;
            }
            
            HTTP_AuthenticateFree(&auth);
          }
          else {
            debug("SipRegHandleResponse", "No call object\n");
          }
        }        
        break;

      case SIPREG_OUTGOING_3:
        /*
         * Possible messages : 1xx-6xx 
         */
        if ( m->p.status < 300 ) {
          /* Successful response */
          c->state = SIPREG_ACTIVE;
          SipRegUpdateExpiry(c, m);
          c->unauth_count = 0;
        }
        else if ( m->p.status < 700 ) {

          c->state = SIPREG_IDLE;
          
          if ( (c->regobj != NULL ) &&
               (m->p.status == 401 || m->p.status == 407) 
               && (c->unauth_count < 1)) {          
            /* Authorization needed */

            http_authenticate_t auth;
            memset(&auth, 0, sizeof(auth));
            
            auth = HTTP_Authenticate(&m->p);
            strassign(&auth.http_method, "REGISTER");
            //strassign(&auth.opaque, "86712313163187632617");

            c->unauth_count ++;

            char * authorization;
            char *uristr;
            uristr = URI_AsString(c->remote_uri);
            
            authorization = HTTP_BuildAuthorizationHeader(
                                            &auth,
                                            c->userid,
                                            c->password,
                                            uristr);
            free_valid(uristr);
              
            if ( authorization ) {
                /* Set header here */
              if ( m->p.status == 401 ) {
                  c->regobj->h_orq -= (const char *) "Authorization";
                  c->regobj->h_orq +=
                    GenericHeaderValue("Authorization", 
                                       authorization);
              }
              else {
                  c->regobj->h_orq -= (const char *) "Proxy-Authorization";
                  c->regobj->h_orq +=
                    GenericHeaderValue("Proxy-Authorization", 
                                       authorization);
              }

              free_valid(authorization);
              
              /*
               * outgoing reg.
               * Invoke the request thread/policy for new REGISTER.
               */
              c->state = SIPREG_OUTGOING_3;

              request_t *r;

              r = FormatRegistration("REGISTER", c->type, 
                                     c->callid, ++c->lastcseq_tx, 
                                     c->body.content_type ? &c->body : NULL,
                                     c);
              
              URI_Free(&c->remote_uri);
              c->remote_uri = URI_Copy(r->p.URI);
              
              /* Should store the outgoing request here */
              HTTP_Duplicate(&c->p, &r->p);
              
              CallCreatePolicyThread(r, c->remote_uri, c->proxy);
              
            }
            else {
              /* Failed */
              pthread_mutex_unlock(&c->mutex); /* This is important to avoid deadlock */
              c->regobj->OnRegistrationExpired();
              pthread_mutex_lock(&c->mutex);
            }
            HTTP_AuthenticateFree(&auth);
          }
          else {
          
            if ( c->regobj ) {
              pthread_mutex_unlock(&c->mutex); /* This is important to avoid deadlock */
              c->regobj->OnRegistrationExpired();
              pthread_mutex_lock(&c->mutex);
            }
            else {
              debug("SipRegHandleResponse", "No call object\n");
            }
          }

          c->regobj = NULL;
        }        
        break;

      case SIPREG_ACTIVE:
        /* Ignore any response in active state */
        debug("SipRegHandleResponse", "received %d in ACTIVE state\n", m->p.status);
        debug("SipRegHandleResponse", "ignoring the response in this state, ignored\n");
        break;

      case SIPREG_TERMINATING: /* call is being terminated */
        debug("SipRegHandleResponse", "received %d in TERMINATING state\n", m->p.status);
        /* Remove the request */
        /* set the state to IDLE irrespective or response received */
        c->state = SIPREG_IDLE;
        c->regobj = NULL;
        /* XXX what if 401 or 407 is received */
        break;

      default:
        error("SipRegHandleResponse", "surprising state %d\n", c->state);
        break;
      }
    }
  }

  if (  rp != NULL ) {
    pthread_mutex_unlock(&rp->mutex);
  }

  m->handled = finished_handling;

  return 0;
}


/***********************************************************************
**************************** Function definitions **********************
*/
int SIPEndpoint::Register(const IptelAddress & user)
{
  SIPRegistrationInfo reginfo;
  reginfo.localName = user;
  char *host = user.GetHost();
  if ( host != NULL ) {
    reginfo.registrar = host;
    free(host);
  }
  char *userid = user.GetUser();
  reginfo.contactAddress = SIPLibrary::HostName;

  reginfo.contactAddress.SetScheme("sip");
  if ( userid != NULL ) {
    reginfo.contactAddress.SetUser(userid);
    free(userid);
  }

  IptelIPAddress listening_at = GetServerAddress();
  int listening_port = listening_at.GetPort();

  if ( listening_port != 0 && listening_port != DEFAULT_SIP_PORTNO) {
    reginfo.contactAddress.SetPort(listening_port);
  }

  return Register(reginfo);
}

int SIPEndpoint::Register(const IptelAddress & user,
                          const IptelAddress & contact)
{
  SIPRegistrationInfo reginfo;
  reginfo.localName = user;
  char *host = user.GetHost();
  if ( host != NULL ) {
    reginfo.registrar = host;
    free(host);
  }
  reginfo.contactAddress = contact;

  return Register(reginfo);
}

int SIPEndpoint::Register(const SIPRegistrationInfo & reginfo)
{
  return Register(reginfo, 
           NULL /* no cgi-file */, 
           NULL /* no mime type */);
}


SIPRegistrationInfo::SIPRegistrationInfo()
{
  expiry = 0;
  q_value = 0.0;
  type = SOCK_DGRAM;
  callid = NULL;
  lastcseq = 0;
  info = NULL;
  authorization = NULL;
}

SIPRegistrationInfo::SIPRegistrationInfo(const SIPRegistrationInfo &r)
{
  localName = r.localName;
  registrar = r.registrar;
  contactAddress = r.contactAddress;
  fromAddress = r.fromAddress;
  expiry = r.expiry;
  q_value = r.q_value;
  type = r.type;
  callid = r.callid ? strdup(r.callid) : NULL;
  lastcseq = r.lastcseq;
  info = r.info;
  authorization = r.authorization ? strdup(r.authorization) : NULL;
  outbound_proxy = r.outbound_proxy;
}

SIPRegistrationInfo & SIPRegistrationInfo::operator=(const SIPRegistrationInfo &r)
{
  localName = r.localName;
  registrar = r.registrar;
  contactAddress = r.contactAddress;
  fromAddress = r.fromAddress;
  expiry = r.expiry;
  q_value = r.q_value;
  type = r.type;
  callid = r.callid ? strdup(r.callid) : NULL;
  lastcseq = r.lastcseq;
  info = r.info;
  authorization = r.authorization ? strdup(r.authorization) : NULL;
  outbound_proxy = r.outbound_proxy;
  return *this;
}

bool SIPRegistrationInfo::operator==(const SIPRegistrationInfo &r) const
{
  /* Currently we compare based on only the registered name.
  * Should we use other parameters also. */
  if ( localName == r.localName ) {
    return true;
  }
  else {
    return false;
  }
}

SIPRegistrationInfo::~SIPRegistrationInfo()
{
  free_valid(callid);
  free_valid(authorization);
  /* Do not free info !! */
}

sipreg_t *FindOrCreateRegistrationObject(SIPEndpoint & ep,
                                         const SIPRegistrationInfo &
                                         reginfo,
                                         const char *cgi_file,
                                         const char *mimetype)
{
  /* Search the local data base for the registration info */
  sipreg_t *r = NULL;

  r = SipRegFind(reginfo.localName.GetAddressString());
  if ( r == NULL ) {
    /* This is a new registration */
    r = SipRegCreate(NULL);
    if ( r == NULL ) {
      return NULL;
    }
    else {
      /* Set the endpoint object as the regobj of sipreg_t */
      //pthread_mutex_lock(&r->mutex);
      r->regobj = &ep;
      r->type = reginfo.type;
    }
  }

  /* Set the registration object */
  if ( r->regobj != NULL ) {
    r->regobj = &ep;
  }
  else {
    debug("FindOr...", "regobj exists\n");
  }

  /*
   * Set the self name and remote names.
   */

  /* Outbound proxy */
  if ( !reginfo.outbound_proxy.IsEmpty() ) {
    URI_Free(&r->proxy);
    r->proxy = URI_Parse((char *) reginfo.outbound_proxy.GetAddressString());
  }

  /* Request URI */
  IptelAddress uri;
  if ( reginfo.registrar.IsEmpty() ) {
    char * ptr = reginfo.localName.GetHost();
    uri = ptr;
    free_valid(ptr);
  }
  else {
    IptelAddress registrar = reginfo.registrar;
    char *ptr = reginfo.registrar.GetURI();
    uri = ptr;
    free_valid(ptr);
  }
  uri.SetUser(NULL);
  uri.SetDisplayName(NULL);
  uri.SetScheme("sip");
  URI_Free(&r->remote_uri);
  r->remote_uri = URI_Parse((char *)((const char *) uri));

  /* From address */
  strassign(&r->self, reginfo.fromAddress.IsEmpty() ?
            (const char *) reginfo.localName :
            (const char *) reginfo.fromAddress);

  /* To address */
  IptelAddress to = reginfo.localName;
  char *ptr = to.GetScheme();
  if ( ptr && !strcasecmp(ptr,"sip") && !strcasecmp(ptr,"tel")) {
    /*Valid name */
  }
  else {
    to.SetScheme("sip");
  }
  free_valid(ptr);

  strassign(&r->remote, (const char *) to);

  /* Set the contact header */
  IptelAddress contact = reginfo.contactAddress;
  if ( !contact.IsEmpty() ) {
    URI_Free(&r->contact.uri);
    sip_addr_t a =  URI_ParseNew((const char *) contact);
    r->contact.uri = a.uri;
    free_valid(a.display_name);      
  }

  r->contact.expires = reginfo.expiry;
  r->contact.q = reginfo.q_value;
  r->contact.action = L_proxy;

  /* Build a default body, or from the params */
  body_t body;
  memset(&body, 0, sizeof(body));
  if ( cgi_file != NULL && mimetype != NULL) {
    /* Open the file and read the content */
    int length = 0;
    FILE *fp = NULL;
    char buf[1024];
    DString sipmsg;

    debug("Register", "CGI specified %s, %s\n",
          cgi_file ? cgi_file : "<No file>",
          mimetype ? mimetype : "<No mime type>");

    /* rb is need for Win32, Unix works with rb only,
     * although b is ignored. */
    DStringInit(&sipmsg);
    if ((fp = fopen(cgi_file, "rb")) != NULL) {      
      while ((length = fread(buf, sizeof(char), sizeof(buf)-1, fp)) >
             0) {
        buf[length] = '\0';
        DStringAppend(&sipmsg, buf, length);
      } 

      /* Assign the content_length as the length of
       * string, because we do not
       * process the \0 in a script. */
      body.content_length = strlen(DStringValue(&sipmsg));
      if ( body.content_length > 0 ) {
        strassign(&body.content, DStringValue(&sipmsg));
        strassign(&body.content_type, mimetype);
      }
    }
    else {
      error("Register", "could not open %s, ignoring cgi\n", cgi_file);
    }
    DStringFree(&sipmsg);
  }

  if ( body.content_type != NULL ) {
    if ( body.content_length == 0 ) {
      body.content_length = strlen(r->body.content);
    }
    free_valid(r->body.content);
    free_valid(r->body.content_type);
    r->body = body; /* Copy the pointers */
  }

  return r;
}

int SIPEndpoint::Register(const SIPRegistrationInfo & reginfo,
                          const char *cgi_file,
                          const char *mimetype)
{
  DRAW_MESSAGE_FLOW({
    DrawMessageFlow("SIPAPI","API","This","Register");
  });

  sipreg_t *r = NULL;

  r = FindOrCreateRegistrationObject(*this,reginfo,cgi_file,mimetype);

  if ( r != NULL ) {
    /* Set authorization header if needed */
    debug("Register", "Sending REGISTER message to sipreg\n");
    MessageEnqueueApiRequest(&r->mqueue, "REGISTER",
                             nobody_handled);
    pthread_cond_broadcast(&r->cond);
    pthread_mutex_unlock(&r->mutex);
  }
  else {
    return -1;
  }

  return 0;
}

int SIPEndpoint::Unregister(
      const SIPRegistrationInfo & reginfo)
{
  /* TODO- to be defined */
  DRAW_MESSAGE_FLOW({
    DrawMessageFlow("SIPAPI","API","This","Unregister");
  });

  sipreg_t *r = NULL;
  
  r = FindOrCreateRegistrationObject(*this,reginfo,NULL,NULL);

  if ( r ) {
    /* Set authorization header if needed */
    debug("Register", "Sending UNREGISTER message to sipreg\n");
    MessageEnqueueApiRequest(&r->mqueue, "UNREGISTER",
                             nobody_handled);
    pthread_cond_broadcast(&r->cond);
    pthread_mutex_unlock(&r->mutex);
  }
  else {
    return -1;
  }
  return 0;
}


/* Dummy callbacks */
int SIPEndpoint::OnRegistrationSuccess()
{
  debug("OnRegistrationSuccess","Registration success--dummy\n");
  return 0;
}

int SIPEndpoint::OnRegistrationFailed(int status, const char *reason)
{
  debug("OnRegistrationFailed","Registration failed--dummy\n");
  return 0;
}

int SIPEndpoint::OnAuthorizationNeeded(const char *method,
                                       const char *realm,
                                       char *& user,
                                       char *& pass)
{
  if ( userid != NULL && passwd != NULL ) {
    user = strdup(userid);
    pass = strdup(passwd);
    return 0;
  }
  else {
    return -1;
  }
}

int SIPEndpoint::OnRegistrationExpired()
{
  debug("OnRegistrationExpired","Registration expired--dummy\n");
  return 0;
}


/************************** END OF FILE ********************************
*/
