/*
 *  authenticatePGP.c
 *  
 *  PGP authentication of SIP requests.
 *  Bit rotten; probably won't compile.
 *  
 *  December 1999
 *  Triantafyllos Alexiou
 *  ta154@columbia.edu t.alexiou@ieee.org  alexiou@hyper.gr
 */

/* Copyright 1999 Columbia University.  All rights reserved. */

#include <stdio.h>
#include <stdlib.h>     /* calloc() */
#include <assert.h>
#include <unistd.h>     /* close() */
#include <string.h>
#include "request-sip.h"
#include "sip.h"
#include "strdupn.h"
#include "sysdep.h"
#include "rfc822.h"
#include "http.h"
#include "authenticateSIP.h"
#include "log.h"
#include "config.h"
#include "dbapi.h"

#include "md5.h"
#include "enginepgp.h"
#include "pgpErrors.h"
#include "headers.h"

#include "authenticatePGP.h"

#ifndef HAVE_TIMEGM
#include "timegm.h"
#endif



/*
 * AuthSIP_NonceIsGood()
 * It returns a non zero value if the nonce is acceptable.
 * The authorization structure should be filled in 
 * before calling this function.
 */
static int AuthSIP_PGPNonceIsGood( request_t *r )
{
  time_t packetTime_UTC, now_UTC, timeDiff_UTC;
  struct tm packetTM_UTC;
  char *callIdStr, *cSeqStr, *dateStr, *nonceStr;
  int nonceOkFlag = 0;
  DString msg;
  
  callIdStr = RFC822_Header(r->p.headers, "Call-Id")->elements->value;
  cSeqStr = RFC822_Header(r->p.headers, "CSeq")->elements->value;
  dateStr = RFC822_Header(r->p.headers, "Date")->elements->value;

  if (r->authorization.nonce && ( strlen( r->authorization.nonce ) > 0))
    if (callIdStr && cSeqStr && dateStr) {  
      DStringInit(&msg);
      DStringAppend(&msg, callIdStr, -1);
      DStringAppend(&msg, cSeqStr, -1);
      DStringAppend(&msg, dateStr, -1);
      nonceStr= DStringValue(&msg);	
      
      printf(":%s:\n:%s:\n", r->authorization.nonce, nonceStr );
      
      if (strcmp(r->authorization.nonce, DStringValue(&msg)) == 0) {
	/* The client used the Call-IdCSeqDate as nonce */	
	
	if (Parse_HTTP_Date(dateStr, &packetTM_UTC) == NULL)
	   error("AuthSIP_PGPNonceIsGood", "Parse_HTTP_Date() failed\n");
	
	if ((packetTime_UTC = timegm(&packetTM_UTC)) == (time_t)-1) 
	  error("AuthSIP_PGPNonceIsGood", "timegm() failed\n");
	
        if ((time_t)-1 == time(&now_UTC))
	  error("AuthSIP_PGPNonceIsGood", "time() failed\n");
       
	timeDiff_UTC = now_UTC - packetTime_UTC + timezone - 3600;
	
	if (timeDiff_UTC <= config.ClockSkew && timeDiff_UTC >=
            -config.ClockSkew)
	  nonceOkFlag = 
            !NonceFoundElseAddTimeOrdered(&gCallIdCseqDateNonceHead, 
                                          nonceStr,
                                          packetTime_UTC,
                                          config.ClockSkew);
	DStringFree(&msg);
      }
      else {
	nonceOkFlag = NonceFound(&gNonceHead, r->authorization.nonce, 1);
      }
    }
  return nonceOkFlag;  
} /* AuthSIP_PGPNonceIsGood */


/*
 * AuthSIP_PGPMakeSignedText()
 * Return through msg the text upon which the signature is verified.
 * On error returns a non zero value.
 */
static int AuthSIP_PGPMakeSignedText(DString *msg, request_t *r)
{
  struct header_t *aHeader, *lastHeader, *lastAuthHeader = NULL;
  int error;
  headers_t genHeaders;

  error = kPGPError_BadSignature;
  genHeaders.headers = genHeaders.last=NULL;
  HeaderCopyParsedHeaders(&genHeaders, r->p.headers);

  aHeader = genHeaders.headers;
  if (!aHeader)
    goto done;
  
  do { 
    if (strcasecmp(aHeader->name, "Authorization") == 0)
      lastAuthHeader = aHeader;
    aHeader = aHeader->next;
  } while (aHeader != NULL);
  
  if (lastAuthHeader) {
    lastHeader = genHeaders.last;
    aHeader = lastAuthHeader->next;
    if (!aHeader)
      goto done;
   
    lastAuthHeader->next = NULL;
    genHeaders.last = lastAuthHeader;
    HeaderFree(&genHeaders);
    genHeaders.last = lastHeader;
    genHeaders.headers = aHeader;
        
    DStringFree(msg);
    DStringAppend(msg, r->authorization.realm, -1);
    DStringAppend(msg, r->authorization.nonce, -1);
    DStringAppend(msg, r->authorization.http_method, -1);
    DStringAppend(msg, r->p.protocol, -1);
    DStringAppend(msg, "/", -1);
    DStringAppend(msg, r->p.version, -1);
    HeaderMessage(msg, &genHeaders);
  }
  else {
    error = kPGPError_BadSignature; 
    goto done;
    /* We found no Authorization Header; just send WWW-Authenticate header. */
  }
  error = 0;
 done:
  HeaderFree(&genHeaders);
  return error;
} /* AuthSIP_PGPMakeSignedText */
 

/*
 * AuthSIP_PGPAuthenticate()
 * Create the WWW-Authenticate Header for PGP
 */
static void AuthSIP_PGPAuthenticate(request_t *r, int thePgpError,
                                    int doProxyAuth)
{
  DString msg;
  char* rspHdr = "WWW-Authenticate";

  if(doProxyAuth)
    rspHdr = "Proxy-Authenticate";

  switch (thePgpError) { 
  case kPGPError_NoErr:
    break;
  case kPGPError_OutOfMemory:
  case kPGPError_KeyRevoked:
  case kPGPError_KeyExpired:
  case kPGPError_KeyDisabled:
  case kPGPError_KeyInvalid:
  case kPGPError_ItemNotFound:
  case kPGPError_KeyUnusableForSignature:
    /* ResponseSetStatus(r, 400, "Key Error in general");
       goto done;
       break; */
  case kPGPError_BadSignature:
  default:
    break;
  }
  
  if(!(r->from.uri.user) || !strlen(r->from.uri.user) 
     || !(r->from.uri.host)  || !strlen(r->from.uri.host))
    return;
  /* Since I cannot make a valid realm to send back */
 
  DStringInit(&msg);
  DStringAppend(&msg, r->from.uri.user, -1);
  DStringAppend(&msg, "@", -1);
  DStringAppend(&msg, r->from.uri.host, -1);
  r->authenticate.realm = strdup(msg.string);  
  DStringFree(&msg);
  
  r->authenticate.method = strdup("pgp");
  r->authenticate.algorithm = strdup("");
  
  {
    char *duplNonce, *tempStr;
    time_t timeNow = time(NULL);
    size_t len, passwdLen;
    passwdLen = (config.PgpPvtPass ? strlen(config.PgpPvtPass) : 0);
    
    tempStr = malloc(sizeof(time_t) + passwdLen + 1);
    
    if (tempStr) {
      len = sizeof(time_t);
      memcpy(tempStr, &timeNow, len);
      if (config.PgpPvtPass) {
        memcpy(tempStr+len, config.PgpPvtPass, passwdLen);
        len += passwdLen;
      }
      memcpy(tempStr+len, "\0", 1);
      r->authenticate.nonce = MD5(tempStr);
      free(tempStr);
    }
    else {
      r->authenticate.nonce = NULL;
    }
    if (r->authenticate.nonce && strcmp(r->authenticate.nonce, "")) {
      duplNonce = strdup(r->authenticate.nonce);
      if (AddNonce(&gNonceHead, duplNonce, config.NonceLifeTime)) {
	free_valid(duplNonce);
      }
    }
    else {
      r->authenticate.nonce=strdup("");
    }
  }
  HeaderPrintf(&r->headers, 0, rspHdr, 
	       "%s realm=\"%s\",\r\n"
               "\tnonce=\"%s\"",
	       r->authenticate.method, r->authenticate.realm, 
	       r->authenticate.nonce); 
} /* AuthSIP_PGPAuthenticate */


/*
 * AuthSIP_PGP()
 * Do the PGP authentication. If theSIPAuthMode is non zero
 * then proxy authentication is assumed.
 */
int AuthSIP_PGP(request_t *r, int doProxyAuth)
{
  int error = kPGPError_BadSignature;
  char *rspHdr = "WWW-Authenticate";
  int rspHdrStatus = 401;
  char *prepSgn = NULL;
  DString msg;
  
  if (doProxyAuth) {
    /* if non zero than proxy mode requested, so set all proxy flags */
    rspHdr = "Proxy-Authenticate";
    rspHdrStatus = 407;
  }

  if (doProxyAuth) 
    r->authorization = HTTP_ProxyAuthorization(&r->p);
  else
    r->authorization = HTTP_Authorization(&r->p);
  
  if (strcasecmp(r->authorization.method, "pgp") == 0) {
    if (AuthSIP_PGPNonceIsGood(r)) {
      char *signerStr;  
      if(AuthSIP_PGPMakeSignedText(&msg, r)) {
	if (doProxyAuth)
	  ResponseBuildDefault(r);
	ResponseSetStatus(r, 400, NULL);
	goto done;
      }
      /*
       * Currently we disallow signers other than those in From header.
       * Ignore r->authorization.signed_by. The signer is
       * r->authorization.userid.  HTTP_Authorization fills in this field
       * properly from the From Header.
       */
      signerStr = r->authorization.userid;
      
      if (msg.string && msg.length) {
	prepSgn =
          EnginePgpPrepareOneLineArmoredSignature(gPgpContext,
                  r->authorization.signature,
                  strlen(r->authorization.signature));
	
	error = EnginePgpOVerifySignatureWithPubKey(gPgpContext,
                  gPgpKeySet, msg.string, msg.length, prepSgn,
                  prepSgn ? strlen(prepSgn) : 0, signerStr, NULL, 1, 1);

	if (prepSgn) PGPFreeData(prepSgn);
	prepSgn = NULL;
      }
      else {
	error = kPGPError_BadSignature; /* We did not find a signature. */
      }
      DStringFree(&msg);
    }
    else {
      error = kPGPError_BadSignature; /* The nonce is not valid */
    }
  }
  else {
    /* The Client did not supply pgp authentication */
    error = kPGPError_BadSignature; 
  }
  
  if (error) {
    if (doProxyAuth) {
      ResponseBuildDefault(r);
    }
    ResponseSetStatus(r, rspHdrStatus, NULL);
    AuthSIP_PGPAuthenticate(r, error, doProxyAuth); 
  }
done:
  return error; 
} /* AuthSIP_PGP() */

