/*
 * policy-core.c - execute a policy handler.
 *
 * This executes the low-level core logic of a SIP transaction.  The
 * high-level core logic -- anything which isn't inevitable in the
 * protocol -- goes in a policy handler instead, which this calls.
 *
 * Copyright 1998-2001 by Columbia University; all rights reserved 
 */

#include <sys/errno.h>
#include <string.h>
#include <stdlib.h>
#include <pthread.h>
#include <time.h>
#include <sys/time.h>

#include "policy.h"
#include "message.h"
#include "timespec.h"
#include "strdupn.h"
#include "pthread-util.h"
#include "config.h"
#include "sysdep.h"
#include "log.h"

#include "rr_support.h"

static void SendComplaint(request_t *r);

/* execute_policy()
 * This is the core code which executes a policy.  Call the policy's init
 * function, then wait for events and dispatch them to the policy.  Finish
 * when the policy calls policy_done().
 */
void execute_policy(request_t *r, policy_info_t *reg, int reg_status)
{
  struct message_t *m, *m_next, **prev_next;
  int thr_err;
  pthread_t *tid;
  int i, n;

  debug("execute_policy", "Called\n");

  if (r->policy->init(r, reg, reg_status) != 0) {
    SendComplaint(r);
    goto cleanup;
  }

  while (!r->done) {
    if (r->timeout.tv_sec != 0)
      thr_err = pthread_cond_timedwait(&r->cond, &r->mutex, &r->timeout);
    else
      thr_err = pthread_cond_wait(&r->cond, &r->mutex);
    debug("execute_policy","wait exiting\n");
    if (thr_err) {
      if (thr_err != ETIME && thr_err != ETIMEDOUT) {
        error("policy_core", "Error waiting for proxy events: %s (%d)",
              strerror(thr_err), thr_err);
        /*
         * Cleanup the transaction, if the pthread_cond_[timed]wait
         * returns non-timeout error. This is a critical
         * error and should not be ignored.
         */
        goto cleanup;
      }
      else {
        /* The timeout expired. */
        /* Don't trigger another timeout unless one is set */
        r->timeout.tv_sec = 0;
        if (r->policy->timeout_expired(r) != 0) {
          SendComplaint(r);
          goto cleanup;
        }
      }
    }

    /* Walk the message queue, and send requests not already handled */
    /* We could keep a pointer to the last unhandled request,
     * but I don't expect there to be that many. */
    /* If the handler for any message calls policy_done(), stop
       immediately. */
    for (m = r->mqueue; m != NULL && !r->done; m = m->next) {      
      if (m->handled >= policy_handled)
        continue;

      if (m->type != message_real_message && m->type != message_pseudo_message) {
        error("execute_policy",
              "Surprising message type %d on mqueue", m->type);
        m->handled = finished_handling;
        continue;
      }

      if (m->p.type == http_request) {
        if (r->policy->handle_request(r, m) != 0) {
          SendComplaint(r);
          goto cleanup;
        }
      }
      else if (m->p.type == http_response) {
        /* For messages from call thread to userapi policy
         * the  branch is NULL because there is no
         * actual branch */
        if ( m->branch ) {
          if (m->p.status >= 200 &&
              ClientSentFinal(m->branch) && !m->branch->marked_done) {
            debug("execute_policy", "sent final or done for response\n");
            /* If this branch is done (i.e. it's received a final response),
             * decrement branches_outstanding if we haven't done so before
             * for this branch. This is done only for final responses. */
            r->branches_outstanding--;
            m->branch->marked_done = 1;
          }
        }

        if (r->policy->handle_response(r, m, m->branch) != 0) {
          SendComplaint(r);
          goto cleanup;
        }

      }
      else {
        error("execute_policy",
              "Surprising message parse type %d on mqueue", m->p.type);
        m->handled = finished_handling;
      }
    }

    /* Remove all nodes marked finished_handling from the queue. */
    prev_next = &r->mqueue;
    for (m = r->mqueue; m != NULL; m = m_next) {
      m_next = m->next;
      if (m->handled == finished_handling) {
        *prev_next = m->next;
        MessageDelete(m);
      } else {
        prev_next = &m->next;
      }
    }
  }

  r->policy->cleanup(r);
  /* If we haven't sent a final response by this point, something's
   * definitely wrong. */
  if (!r->sent_final) {
    SendComplaint(r);
  }

cleanup:  /* Jump here when any policy function returns non-zero */

  /* Alert the branches to shut down */
  for (i = 0; i < r->branch_count; i++) {
    if (!r->branches[i].shutdown) {
      MessageTrivialEnqueue(&r->branches[i].mlist, message_shutdown_client,
                            nobody_handled);
      pthread_cond_broadcast(&r->branches[i].response_received);
    }
  }

  /* Wait for the branch threads.  Build list of them: */
  n = 0;
  if (r->branch_count > 0) {
    tid = (pthread_t *) calloc(r->branch_count, sizeof(pthread_t));
    if (tid == NULL) {
      error("execute_policy", "Couldn't allocate tid");
    }
    else {
      for (i = 0; i < r->branch_count; i++) {
        if (r->branches[i].tid != 0) {
          tid[n++] = r->branches[i].tid;
        }
      }
    }
  }
  else {
    tid = NULL;
  }

  /* Need to free mutex before waiting for branches, or we deadlock. */
  /* NOTE: we have to make sure that RequestFree is *only* called from
   * here for transactions that have policies, or we have a race condition
   * which can lead to threads reading from freed memory. */
  pthread_mutex_unlock(&r->mutex);

  /* Wait for branches to complete */
  debug("execute_policy", "waiting for %d branch%s\n", n,
        (n != 1) ? "es" : "");

  if ( n > 0 ) {
    WaitForMultipleObjects(n, tid, TRUE, INFINITE);
  }

  debug("execute_policy", "r=%p: policy %s finished\n", r, r->policy->name);

  free_valid(tid);
  RequestFree(r);
} /* execute_policy */


/* SendComplaint()
 * Something went wrong in a policy.  If we haven't already sent a final
 * response, send a 500 error message. 
 */
static void SendComplaint(request_t *r)
{
  if (!r->sent_final) {
    policy_send_new_response(r, 500, "Internal Policy Error");
  }
} /* SendComplaint */


/* Functions a policy can call. */

/* policy_done()
 * A policy calls this when it is finished with the transaction.  It tells
 * the core to clean things up.
 */
void policy_done(request_t *r)
{
  r->done = 1;
} /* policy_done */


/* policy_set_timeout()
 * A policy calls this to have the timeout callback called when the
 * specified time elapses.
 */
void policy_set_timeout(request_t *r, double delay)
{
  struct timeval now;

  gettimeofday(&now, NULL);
  r->timeout = timespec_delay(now, delay);
} /* policy_set_timeout */
 


/* policy_send_modified_new_response()
 * A policy calls this when it wants to send a locally-generated response
 * back to originating client.
 * status should usually not be 200-class, since this server isn't equipped
 * to behave as a UAC.  (XXX policy faking UAC?  Possible for CGI.)
 * reason may be NULL, in which case the standard reason will be filled in.
 * delta_headers contains headers to be added to, removed from, or changed
 * in the default set of response headers.
 * This response will be remembered until something else changes it.
 * Retransmissions of the request will receive retransmissions of the
 * most recent response.
 */
void policy_send_modified_new_response(request_t *r, int status, char *reason,
                                       headers_t *new_headers,
                                       headers_t *del_headers,
                                       body_t *body)
{
  int retransmit;
  char *method;

  method = HTTP_Method(&r->p);
  /* For new responses to INVITE, we retransmit all final responses. */
  retransmit = (status >= 200 && strcmp(method, "INVITE") == 0);
  free_valid(method);

  ResponseSetStatus(r, status, reason);
  ResponseBuildDefault(r);

  if (new_headers != NULL) {
    HeaderMerge(&r->headers, new_headers);
  }
  if (del_headers != NULL) {
    HeaderDeleteSet(&r->headers, del_headers);
  }

  /* Order here is important.  ResponseSetBody sets or overrides
   * Content-Type and Content-Length. */
  if (body != NULL) {
    ResponseSetBody(r, body->content, body->content_type,
                    body->content_length);
  }
  else {
    ResponseSetBody(r, NULL, NULL, 0);
  }

  if (status >= 200) {
    r->sent_final = 1;
    if (status < 300) {
      r->sent_200 = 1;
    }
  }

  ResponseSend(r, retransmit, 0);

  if (r->sent_final) {
    Log(r);
  }
} /* policy_send_modified_new_response */


/*
 * policy_send_new_response()
 * A wrapper around policy_send_modified_new_response for the
 * simple case when the caller doesn't want to modify the response.
 */
void policy_send_new_response(request_t *r, int status, char *reason)
{
  policy_send_modified_new_response(r, status, reason, NULL, NULL, NULL);
} /* policy_send_new_response */


/*
 * policy_forward_response()
 * Forward a received response upstream.
 * Set various flags as appropriate.
 * Generally the same characteristics apply here as apply in
 * send_new_response.
 */
void policy_forward_response(request_t *r, message_t *m)
{
  int retransmit;
  char *method;

  method = HTTP_Method(&r->p);

  /* For forwarded responses to INVITE, we retransmit 300-600's, but not
     200's; those are ACK'd end-to-end.
   */
  retransmit = (m->p.status >= 300 && strcmp(method, "INVITE") == 0);


  if (m->p.status >= 200) {
    r->sent_final = 1;
    if (m->p.status < 300) {
      r->sent_200 = 1;
    }
  }

  HeaderFree(&r->headers);

  /* RecordRoute transformation goes here..
   */
  if (config.RecordRoute && strcmp(method,"INVITE") == 0) {
    ModifyRRElement(m, r->hash);
  }

  free_valid(method);
  
  HeaderCopyParsedHeaders(&r->headers, m->p.headers);
  ResponseSetStatus(r, m->p.status, m->p.reason);
  ResponseSetBody(r, m->body.content, m->body.content_type,
                  m->body.content_length);
  /* Strip top Via field.  N.B.: we send our reply to the parent
   * address that originated the request, whether or not it matches
   * the next Via field.  XXX?
   */
  HeaderDeleteFirst(&r->headers, "Via");
  ResponseSend(r, retransmit, 0);

  if (r->sent_final) {
    Log(r);
  }
} /* policy_forward_response */


/*
 * policy_forward_user_response()
 * Forward a received response upstream.
 * Set various flags as appropriate.
 * Generally the same characteristics apply here as apply in
 * send_new_response.
 * It is assumed that the call thread has already
 * updated the headers as needed.
 *
 */
void policy_forward_user_response(request_t *r, message_t *m)
{
  int retransmit;
  char *method;

  method = HTTP_Method(&r->p);

  /* For forwarded responses to INVITE, we retransmit all responses.
   * This is the only difference between this function and
   * policy_forward_respose(). Should these two be merged ? XXX.
   */
  /* KKK should we have all responses to INVITE retransmitted */
  retransmit = (/*m->p.status >= 300 && */ strcmp(method, "INVITE") == 0);
  free_valid(method);

  if (m->p.status >= 200) {
    r->sent_final = 1;
    if (m->p.status < 300) {
      r->sent_200 = 1;
    }
  }

  /* Do not rewrite the r->headers */
  //HeaderFree(&r->headers);
  //HeaderCopyParsedHeaders(&r->headers, m->p.headers);
  ResponseSetStatus(r, m->p.status, m->p.reason);
  /* Body is already set 
  ResponseSetBody(r, m->body.content, m->body.content_type,
                  m->body.content_length);
  */
  /* Do not stript the top via */
  //HeaderDeleteFirst(&r->headers, "Via");
  ResponseSend(r, retransmit, 0);

  if (r->sent_final) {
    Log(r);
  }
} /* policy_forward_user_response */


/*
 * increase_branch_space()
 * Internal function to increase the branch space for a request
 * Increases the branch space by at least one; grows the space used
 * exponentially for efficiency reasons.
 * Returns non-zero on failure.
 */
static int increase_branch_space(request_t *r)
{
#if 0
  /* XXX This doesn't work!  We pass the pointer to the branch_t to the
   * client; if we subsequently realloc(), the client's pointer points to
   * freed memory, and we blow up real good.  Sigh. */
  int old_branch_space = r->branch_space;

  /* This should only execute exactly once in normal circumstances,
   * but put it in a while loop for paranoia's sake. */
  while (r->branch_space <= r->branch_count) {
    if (r->branch_space == 0) {
      r->branch_space = 1;
    }
    else {
      r->branch_space *= 2;
    }
  }

  r->branches = (branch_t *) realloc(r->branches,
                                     r->branch_space * sizeof(branch_t));
  if (r->branches == NULL) {
    return 1;
  }

  /* Zero the memory for the newly-allocated branches (only) */
  memset(&r->branches[old_branch_space], '\0',
         (r->branch_space - old_branch_space) * sizeof(branch_t));

  return 0;
#else
  /* XXX Quick and dirty alternative. */
  if (r->branch_count == MAX_BRANCHES) {
    return 1;
  }
  if (r->branch_space == 0) {
    r->branches = calloc(MAX_BRANCHES, sizeof(branch_t));
    r->branch_space = MAX_BRANCHES;
  }
  return 0;
#endif
} /* increase_branch_space */


/* policy_proxy_original_request_to()
 * Proxy the original request of the transaction to the specified
 * URI.  This starts the client branch, and updates the request
 * data structures.
 */
void policy_proxy_original_request_to(request_t *r, uri_t *uri,
                                      message_t *m, void *policy_info)
{
  int branch_num;
  int thr_err;

  if (r->branch_space <= r->branch_count) {
    if (increase_branch_space(r) != 0) {
      error("policy_proxy_original_request_to",
            "couldn't increase branch space");
      return;
    }
  }

  branch_num = r->branch_count;
  r->branch_count++;

  r->branches[branch_num].branch      = branch_num;
  r->branches[branch_num].parent      = r;
  r->branches[branch_num].uri         = URI_Copy(*uri);
  r->branches[branch_num].orig        = MessageDuplicate(m);

  r->branches[branch_num].sent_200    = 0;
  r->branches[branch_num].sent_final  = 0;
  r->branches[branch_num].policy_info = policy_info;

  pthread_cond_init(&r->branches[branch_num].response_received, NULL);

  debug("policy_proxy_original_request_to",
        "Request %p: starting client %d, branch = 0x%p\n",
        r, branch_num, &r->branches[branch_num]);
        
  debug("pthread_create", "Creating client thread\n");
  /* NB: we do not create this thread detached, since we wait for it
   * when execute_policy() returns. */
  if ((thr_err = pthread_create(&r->branches[branch_num].tid, NULL, 
                                client, (void *)&r->branches[branch_num]))) {
    error("pthread_proxy_original_requst_to", "branch thread %d: %s (%d)\n",
          branch_num, strerror(thr_err), thr_err);
  } else {
    r->branches_outstanding++;
  }
} /* policy_proxy_original_request_to */


/* policy_proxy_subsequent_request_to()
 * Proxy a subsequent request (ACK, typically) on an existing branch.
 */
void policy_proxy_subsequent_request_to(request_t *r, branch_t *b,
                                        message_t *m)
{
  MessageEnqueue(&b->mlist, message_real_message, &m->p, &m->body,
                 m->source, b, policy_handled);
  pthread_cond_broadcast(&b->response_received);
} /* policy_proxy_subsequent_request_to */


/* policy_cancel_branch()
 * Tell an existing branch to cancel itself.
 */
void policy_cancel_branch(request_t *r, branch_t *b)
{
  MessageTrivialEnqueue(&b->mlist, message_cancel_branch, policy_handled);
  pthread_cond_broadcast(&b->response_received);
} /* policy_proxy_subsequent_request_to */


