/*
 * proxy-support.c - useful functions for proxying a SIP request.
 * used by sipd/proxy.c, sipd/cgi.c, sipd/route.c, and libsip++/user.cpp.
 *
 * Copyright 1998-2001 by Columbia University; all rights reserved 
 */

#include <string.h>

#include "policy.h"
#include "config.h"
#include "proxy-support.h"


/*
 * PickBestResponse()
 * The message queue mq contains responses with status 3xx-6xx. 
 * Pick the best response of these, if there is one.
 * "Best" is defined as the first response in the lowest message class.
 */
message_t *PickBestResponse(message_t *mq)
{
  int best_status = 1000;
  message_t *m, *best = NULL;
  
  for (m = mq; m != NULL; m = m->next) {
    if (m->handled != policy_handled)
      continue;

    if (m->p.status < 300 || m->p.status > 699) {
      error("PickBestResponse",
            "message with status %d got through to PickBestResponse somehow.",
            m->p.status);
      continue;
    }
    
    /* Find the earliest response of the lowest class. 
     * We keep early ones because they are likely to correspond to
     * branches with high Q values. (XXX: I think?)
     */
    if ((best_status / 100) > (m->p.status / 100)) {
      best_status = m->p.status;
      best = m;
    }
  }
  return best;
} /* PickBestResponse */


/*
 * SendBestResponse()
 * Send the best response, as chosen by the above function, upstream.
 * If there isn't one, send an appropriate new response.
 * If the best response is a pseudo-response, handle it properly.
 */
void SendBestResponse(request_t *r)
{
  message_t *best = PickBestResponse(r->mqueue);
  char *reason = NULL;
  
  if (best == NULL) {
    int status;
    /* If no responses arrived, it means the response timed out.
     * If some responses arrived, but we didn't find any, it means
     * we (e.g.) had two 300-class responses which pointed to each other.
     * The status is 408 in the former case, and 482 in the latter.
     */
    if (r->branches_outstanding == r->branch_count) {
      status = 408;
    }
    else {
      status = 482;
      reason = "Redirect loop";
    }
    policy_send_new_response(r, status, reason);
  }
  else if (best->type == message_pseudo_message) {
    policy_send_new_response(r, best->p.status, NULL);
  }
  else {
    policy_forward_response(r, best);
  }
} /* SendBestResponse */


/*
 * ForwardACKToBranches()
 * Forward an end-to-end ACK message to the appropriate branches.
 * XXX: branches should keep track of To: tags.  Right now we just
 * forward to every branch which has received a 200 response, and let
 * clients worry about whether this ACK is for them.
 */
void ForwardACKToBranches(request_t *r, message_t *m)
{
  int i;

  for (i = 0; i < r->branch_count; i++) {
    if (ClientSent200(&r->branches[i])) {
      debug("ForwardACKToBranches", "Forwarding ACK to branch %d\n",
            r->branches[i].branch);
      policy_proxy_subsequent_request_to(r, &r->branches[i], m);
    }
  }
} /* ForwardACKToBranches */


/*
 * CancelUnfinishedBranches()
 * Cancel all the unfinished branches of the request.
 */
void CancelUnfinishedBranches(request_t *r)
{
  int i;
  for (i = 0; i < r->branch_count; i++) {
    if (!ClientSentFinal(&r->branches[i])) {
      debug("CancelUnfinishedBranches", "Cancelling branch %d\n",
            r->branches[i].branch);
      policy_cancel_branch(r, &r->branches[i]);
    }
  }
} /* CancelUnfinishedBranches */


/*
 * Do3xxRecursion()
 * We have received a 3xx response.  Go through it and see if it has any
 * Contact headers with SIP URI's.  If it does, immediately start new
 * child branches to those URI's, and return 1.  If not, return 0.
 * XXX: update registry on 301, 305?
 */
static int Do3xxRecursion(request_t *r, message_t *m)
{
  int found_something = 0, already_tried;
  sip_contact_t *contact;
  int i, j;
  message_t *req_copy;
  
  contact = SIP_Contact(&m->p);
  if (!contact)
    return 0;

  for (i = 0; contact[i].uri.scheme != NULL; i++) {
    if (strcmp(contact[i].uri.scheme, "sip") == 0) {
      found_something = 1;
      /* Check if some other branch has already visited this URI */
      already_tried = 0;
      for (j = 0; j < r->branch_count; j++) {
        if (URI_Equal_With_Defaults(r->branches[j].uri, contact[i].uri,
                                    sip_uri_defaults) == 0) {
          already_tried = 1;
          break;
        }
      }
      if (!already_tried) {
        req_copy = GetRequestAsMessage(r);
        policy_proxy_original_request_to(r, &contact[i].uri, req_copy, NULL);
        MessageDelete(req_copy);
      }
    }
  }

  SIP_FreeContact(contact);
  return found_something;
} /* Do3xxRecursion */


/*
 * ProxyDispatchResponse()
 * The message 'm' contains a new response.  Dispatch it as appropriate.
 */
void ProxyDispatchResponse(request_t *r, message_t *m)
{
  if (m->p.status < 100 || m->p.status >= 700) {
    error("ProxyDispatchResponse",
          "Response with surprising status %d received\n", m->p.status);
    m->handled = finished_handling;
  }
  else if (m->p.status >= 200 && m->p.status < 300) {
    /* We received a 2xx response. */
    /* Send it immediately, and cancel all other branches. */
    policy_forward_response(r, m);

    /* If we get multiple 200's, we may cancel more than once here; but
     * that's okay. */
    CancelUnfinishedBranches(r);

    policy_set_timeout(r, (double)config.timeout);
    m->handled = finished_handling;
  }
  else {
    /* If we've already sent a final response, don't bother to send or keep
     * any non-2xx responses. */
    if (r->sent_final) {
      m->handled = finished_handling;
    }
    else {
      if (m->p.status < 200) {
        /* 1xx response received.  If non-100, proxy it. */
        if (m->p.status != 100) {
          policy_forward_response(r, m);
        }
        m->handled = finished_handling;
      }
      /* 2xx responses handled above */
      else if (m->p.status < 400) {
        /* We received a 3xx response.  If ProxyRecursion is turned on,
         * try to do 3xx recursion.  If not, or if that function says it
         * can't do any recursion on this response, keep it in the message
         * queue for later consideration. */ 
        if (config.ProxyRecursion && Do3xxRecursion(r, m)) {
          m->handled = finished_handling;
        }
        else {
          m->handled = policy_handled;
        }
      }
      else if (m->p.status >= 400) {
        /* 4xx - 6xx response received */
        /* Keep this message around for later decisions of which response to
         * send. */
        m->handled = policy_handled;
      }
    }
  }
} /* ProxyDispatchResponse */
