/*
 * querydb.c: Wrapper functions around dpapi.h to do the specific
 * functions needed by libsip and sipd
 */

/* Copyright 2000, 2001 Columbia University.  All rights reserved. */

#ifdef __linux__
#define _GNU_SOURCE /* Needed for strptime declaration from <time.h> */
#endif

#include <string.h>

#include "dstring.h"
#include "error.h"
#include "strdupn.h"
#include "uri.h"

#include "querydb.h"
#include "timegm.h"

#if 0 /* Old file-based code.  Should move into libdb++. */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>

#include "dict.h"
#include "ht.h"
#include "config.h"

#ifndef WIN32
#warning querydb.c contains only dummy functions without MySQL, configure with --use-mysql
#endif

/* Dummy function */
char *build_userid(const uri_t *uri)
{
  DString userid;

  DStringInit(&userid);

  if (strcasecmp(uri->scheme, "sip") != 0) {
    URI_SPrint(uri, &userid);
  }
  else {
    if (uri->user == NULL && uri->host == NULL) {
      /* Shouldn't happen */
      DStringFree(&userid);
      return NULL;
    }
    if (uri->user != NULL) {
      DStringAppend(&userid, uri->user, -1);
      DStringAppend(&userid, "@", -1);
    }
    if (uri->host != NULL) {
      if ( strstr(uri->host, config.ServerName) ) {
        DStringAppend(&userid, config.ServerName, -1);
      }
      else {
        DStringAppend(&userid, uri->host, -1);
      }
    }
  }
  
  return DStringValue(&userid);
} /* build_userid */


/* Dummy function */
db_err method_allowed_for_user(const char *method, const char *user)
{
  debug("method_allowed_for_user",
        "Dummy function called, configure with --use-mysql\n");
  return DB_ERR;
} /* method_allowed_for_user */


/* Dummy function */
db_err user_wants_request_authentication(const char *userid, int *required)
{
  debug("user_wants_request_authentication",
        "Dummy function called, configure with --use-mysql\n");
  return DB_ERR;
} /* method_allowed_for_user */


/* Dummy function */
db_err may_register_for(const char *from_userid, const char *to_userid)
{
  debug("may_register_for",
        "Dummy function called, configure with --use-mysql\n");
  return DB_ERR;
} /* may_register_for */


/*
 * File variables for the HashTable of password.
 */
#define STALETIME 60
#define HT_ENTRY 1000
#define BUCKET_ENTRY 10

static char * pwdFilename = NULL;
static char * pwds = NULL;
static dict_h pwdHandle;
static struct stat pwdStat;
static int   table_initialized = 0;

/*
 * Used to get the value of the string for hashing
 * The value is calculated as the summary of ASCII value of each 
 * character in the name string
 */
static ht_val auth_ht_value(dict_obj o)
{
  char *s;
  ht_val r=0;

  s = (char *)o;
  while (*s != ':') {
    r += *s;
    ++s;
  }

  return r;
} /* auth_ht_value */


/*
 *  Used to compare the value in hash table
 *  We need to search the record based on the name
 *  If the name match the string, we return match.
 *  The further work for matching the password will be done outside
 *  hashtable so that we can distinguish the different errors.
 *
 *  return -1 : The first object's name part less than the second one.
 *  return 0  : The first object's name part equal the second one.
 *  return 1  : The first object's name part greater than the second one.
 *
 *  suppose the userID is unique in the file 
 *
 */
static int auth_ht_comp(dict_obj obj1, dict_obj obj2)
{
  char *s, *t;

  s = (char *)obj1;
  t = (char *)obj2;

  while ((*s != ':') && (*t != ':')) {
    if (*s > *t)
      return 1;
    if (*s < *t)
      return -1;
    ++s;
    ++t;
  }

  if (*s != ':')
    return 1;
  if (*t != ':')
    return -1;
  return 0;
} /* auth_ht_comp() */

/*
 * Set the password file. 
 */
void set_pwinfo_file(const char *file)
{
  strassign(&pwdFilename, file);
}

/*
 * Used to see whether the password file is modified or not
 * return 0 : not modified
 * return 1 : modified
 */
static int authenticateBasicRefresh (char *pwdFilename)
{
  struct stat buf;

  /* If not loaded, return 1 */
  if (pwds == NULL)
    return 1;

  /* If the last modification time of file is later than the file loaded 
   * return 1 */
  stat(pwdFilename, &buf);
  if (buf.st_mtime != pwdStat.st_mtime) 
    return 1;

  return 0;
} /* authenticateBasicRefresh() */

/*
 * Used to read the password file's content into memory block
 * so that the query speed can be improved
 * read all the content of the file into global pointer *pwds,
 * then we just need to do matching in memory, no more I/O needed
 * Create hash table for searching.
 *  
 * Question: What if the password file is changed when 
 *           the server is running
 * Solution: Each time, when match failed, first to see whether the 
 *           file is modified based on the last modification time, if
 *           it is modified, then reload the file to the memory,
 *           and do matching again. By this way, for an unauthorized 
 *           user, they need to wait longer to get the response. But
 *           for valid user, the response will be quicker than do matching
 *           by reading from file.
 *
 * return 0  : OK
 *  return -1 : password file open error
 */

static int authenticateBasicInit(char *pwdFilename)
{
  FILE *fp;
  struct ht_args argsp = {HT_ENTRY, BUCKET_ENTRY, auth_ht_value, auth_ht_value};
  char *s, *t;

  fp = fopen(pwdFilename, "rt");
  if (fp == NULL) {
    return -1;
  }

  /* read all the content of the password file into *pwds */
  stat(pwdFilename, &pwdStat);
  if (pwds == NULL) {
    pwds = (char *)malloc(pwdStat.st_size*sizeof(char)+1);
  } else {
    pwds = (char *)realloc((void *)pwds, pwdStat.st_size*sizeof(char)+1);
    ht_destroy(pwdHandle);
  }
  fread(pwds, pwdStat.st_size, sizeof(char), fp);
  pwds[pwdStat.st_size] = '\0';

  /* create hash table and insert the record */
  pwdHandle = ht_create(auth_ht_comp, auth_ht_comp, DICT_UNIQUE_KEYS, (int *)0, &argsp);
  s = pwds;
  while (*s != '\0') {
    t = s;
    while (*t != '\n') 
      ++t;
    /* If only '\n' in a seperate line, skip to next line */
    if (t == s+1) {
      s = t+1;
      continue;
    }
    /* Else insert the record into hash table */
    *t = '\0';
    debug("authenticateBasicInit", "Adding %s\n", s ? s : "NULL");
    ht_insert(pwdHandle, (dict_obj)s);
    s = t+1;
  }
    
  fclose(fp);
  return 0;
} /* authenticateBasicInit() */

/*
 * Use file based authentication if needed.
 * No hashtable here. Just read the file and 
 * try to match the userid.
 */
db_err get_pwinfo_for_user(const char *userid, 
                           char **hash_value,
                           char **realm,
                           char **algorithm)
{
  char *uid_t;
  char *r;
  char *last;

  /* Initialize the return values first */
  if ( hash_value ) {
    *hash_value = NULL;
  }
  if ( realm ) {
    *realm = NULL;
  }
  if ( algorithm ) {
    *algorithm = NULL;
  }

  if ( userid == NULL ) {
    error("get_pwinfo_for_user", "userid is NULL\n");
    return DB_ERR;
  }

  if ( pwdFilename == NULL ) {
    debug("get_pwinfo_for_user", "No password file set\n");
    return DB_ERR;
  }

  /* In the password file is modified then 
   * reread the password file. */
  if ( authenticateBasicRefresh(pwdFilename) ) {
    debug("get_pwinfo_for_user", "Initializing the hash table\n");
    if ( authenticateBasicInit(pwdFilename) != 0 ) {
      debug("get_pwinfo_for_user", "Could not open password file %s\n", 
            pwdFilename);
      return DB_ERR;
    }
  }

  if ( pwdHandle == NULL ) {
    debug("get_pwinfo_for_user", "Database error.\n");
    return DB_ERR;
  }

  uid_t = (char *)malloc(strlen(userid) + 2);
  strcpy(uid_t, userid);
  strcat(uid_t, ":");
  r = (char *)ht_search(pwdHandle, (dict_obj)uid_t);
  free(uid_t);
  if (r == NULL) {
    debug("get_pwinfo_for_user", "ht_search for user %s failed\n",
          userid);
    return DB_ERR;
  }

  r = strdup(r);

  /* get the pasword string */
  last = NULL;
  /* Skip userid */
  strtok_r(r, ":", &last);
  /* Extract realm */
  if ( realm ) {
    *realm = strtok_r(NULL, ":", &last);
    if ( *realm ) {
      *realm = strdup(*realm);
    }
  }
  else {
    strtok_r(NULL, ":", &last);
  }
  /* Extract hash */
  if ( hash_value ) {
    *hash_value = strtok_r(NULL, ":", &last);
    if ( *hash_value ) {
      *hash_value = strdup(*hash_value);
    }
  }
  else {
    strtok_r(NULL, ":", &last);
  }

  if ( algorithm ) {
    /* Always MD5 */
    *algorithm = strdup("MD5");
  }

  free_valid(r);

  debug("get_pwinfo_for_user", "Returning %s,%s,%s,%s\n",
        userid, realm ? *realm : "NULL" , 
        hash_value ? *hash_value : "NULL" ,
        algorithm ? *algorithm : "NULL");

  return DB_OK;
} /* get_realm_for_user */
  

#else /* !0 -- real code */


/*
 * method_allowed_for_user()
 * Test if the method 'method' is allowed for the user 'user'.
 */
db_err method_allowed_for_user(const char *method,
                               const put_table_entry *user)
{
  put_sip_method flag_to_test;

  if (strcmp(method, "INVITE") == 0 ||
      strcmp(method, "BYE") == 0 ||
      strcmp(method, "ACK") == 0 ||
      strcmp(method, "CANCEL") == 0) {
    flag_to_test = put_sip_method_INVITE;
  }
  else if (strcmp(method, "REGISTER") == 0) {
    flag_to_test = put_sip_method_REGISTER;
  }
  else {
    flag_to_test = put_sip_method_any;
  }

  if ((user->sip_methods & flag_to_test) != 0 || 
      (user->sip_methods & put_sip_method_any) != 0) {
    return DB_OK;
  }

  return DB_NO;
} /* method_allowed_for_user */



/*
 * user_wants_request_authentication()
 * Check whether the user 'user' wants incoming requests destined for them
 * to be authenticated.  Return DB_OK (yes, it's wanted) or DB_NO on
 * normal behavior.
 * The parameter 'required', if non-NULL, is used to indicate whether the
 * user *insists* on the authentication, or simply requests it and will allow
 * processing to proceed without it if the authentication fails once.  This
 * is filled in with a boolean value if the funcion returns DB_OK.
 */
db_err user_wants_request_authentication(const put_table_entry *user,
                                         int *required)
{
  if (user->auth == put_auth_required) {
    if (required != NULL) { *required = 1; }
    return DB_OK;
  }

  else if (user->auth == put_auth_request) {
    if (required != NULL) { *required = 0; }
    return DB_OK;
  }

  else if (user->auth == put_auth_never) {
    return DB_NO;
  }

  else {
    error("user_wants_request_authentication",
          "unknown user->auth value \"%d\" for user->user == \"%s\"",
          user->auth, user->user);
    return DB_ERR;
  }
} /* user_wants_request_authentication */


/*
 * may_register_for()
 * Check whether the user described by 'from_userid' is allowed to register
 * for the user described by 'to_userid'.
 * If the userids are the same, return DB_OK, otherwise check the database.
 */
db_err may_register_for(const put_table_entry *from,
                        const put_table_entry *to)
{
  char **cur_user;

  if (from == to ||
      strcmp(from->user, to->user) == 0) {
    return DB_OK;
  }

  /* If the entry is NULL no third-party registrations are allowed */
  if (to->remote_user == NULL) {
    return DB_NO;
  }

  for (cur_user = to->remote_user; *cur_user != NULL; cur_user++) {
    if (strcmp(*cur_user, from->user) == 0) {
      return DB_OK;
    }
  }

  return DB_NO;
} /* allow_third_party_reg */


/*
 * Set the password file. Dummy implementation.
 */
void set_pwinfo_file(const char *file)
{
  debug("set_pwinfo_file", "Dummy implementation of password file\n");
}

#endif /* !0 */

/*
 * request_set_dberr_status()
 * A database request for 'userid' has returned a value other than DB_OK.
 * Set the response's status and reason appropriately.
 * If the response was DB_NO, use the caller-supplied status and reason.
 * Other statuses and reasons are standardized.
 * XXX: should DB_NOTFOUND offer a URL to add users to the database?
 */
void request_set_dberr_status(request_t *r, db_err db_ret,
                              const char *userid,
                              int status_db_no,
                              const char *reason_db_no)
{
  DString reason;
  DStringInit(&reason);

  switch (db_ret) {
  case DB_OK:
    /* Shouldn't have gotten here */
    error("request_set_dberr_status", "Called with db_ret = DB_OK!\n");
    ResponseSetStatus(r, 500, NULL);
    break;

  case DB_NO:
    ResponseSetStatus(r, status_db_no, reason_db_no);
    break;

  case DB_NOTFOUND:
    /* No such user in database */
    DStringPrintf(&reason, "No user %s at this server",
                  userid);
    ResponseSetStatus(r, 404, DStringValue(&reason));
    break;

  case DB_ERR:
  case DB_INVALIDARG:
  default:
    ResponseSetStatus(r, 500, "Database error");
    break;
  }

  DStringFree(&reason);
} /* request_set_dberr_status */


/*
 * time2dbstring()
 * Convert the time_t to a MySQL string representation
 * YYYY-MM-DD HH:MM:SS 
 * Infinite time (MAX_TIME_T) corresponds to
 * 0000-00-00 00:00:00
 * All others are converted as per GMT 
 */
char* time2dbstring(time_t time)
{
  struct tm tm;

  DString expires;
  DStringInit(&expires);

  if ( time == MAX_TIME_T ) {
    /* This time_t never expires so we store 
     * the keyword 0000-00-00 00:00:00 in db
     * which means the same */
    DStringAppend(&expires, "0000-00-00 00:00:00", -1);
  }
  else {
    /* Convert the time to db string */
    gmtime_r(&time, &tm);
    DStringStrftime(&expires, "%Y-%m-%d %H:%M:%S", &tm);
  }

  return expires.string;
} /* time2dbstring */


/*
 * dbstring2time()
 * Convert db string representation of time to time_t.
 * db string: YYYY-MM-DD HH:MM:SS 
 * Infinite time (MAX_TIME_T) corresponds to
 * 0000-00-00 00:00:00
 * All others are converted as per GMT 
 */
time_t dbstring2time(const char *string)
{
  struct tm tm;
  time_t expire;
  
  if ( !strcmp(string, "0000-00-00 00:00:00") ) {
    /* If the string is unspecified; i.e. all zeros then 
     * registration is always active */
    expire = MAX_TIME_T; /* Max Possible */
  }
  else {
    /* Some strptime() implementations (glibc, FreeBSD) don't initialize
       all the fields of struct tm, so set it to zero first. */
    memset(&tm, '\0', sizeof(tm));
    strptime(string, "%Y-%m-%d %H:%M:%S", &tm);
    expire = timegm(&tm); /* convert GMT struct tm to time_t */
  }
  
  return expire;
} /* dbstring2time */
