/*
 * rr_support.c - Record Route support.
 *  Some pieces of the code were moved from sipd/route.c to this
 * file. Modified them to be bis03 compliant.
 */
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>     /* printf() */
#include <unistd.h>    /* sleep */
#include <stdlib.h>
#include <pthread.h>
#include <string.h>
#include <limits.h>   /* INT_MAX */

#include "config.h"  

#include "sysdep.h"
#include "strdupn.h"
#include "http.h"
#include "pthread-util.h"
#include "servers.h"
#include "timespec.h"
#include "tcp.h"
#include "request-sip.h"

#include "rr_support.h"

/*
 * AddRecordRoute_Request()
 * Add a Record-Route header, referring to the local machine,
 * to the request described by the message m.
 */
void AddRecordRoute_Request(message_t *m, unsigned long hash, int branchId)
{
  parser_header_t *h;
  parser_element_t *e;
  parser_header_t *hcontact;
  uri_t rr_uri;
  
  h = RFC822_Header(m->p.headers, "Record-Route");
  
  if (h == NULL) {
    /* Add ourselves as the sole RR header. */
    /* XXX: does this show up in the right place in the message? 
     * I think it will be at the bottom. */
    h = (parser_header_t *)malloc(sizeof(parser_header_t));
    if (h == NULL) {
      /* XXX log error */
      return;
    }
    memset(h, 0, sizeof(parser_header_t));
    strassign(&h->name, "Record-Route");
    h->next = m->p.headers;
    m->p.headers = h;
  }

  if( !(hcontact = RFC822_Header(m->p.headers, "Contact"))) 
      hcontact = RFC822_Header(m->p.headers, "From");

  if (hcontact && hcontact->elements->value)  {
       rr_uri  = URI_Parse(hcontact->elements->value);
       e = BuildRRElement(&rr_uri, hash, branchId);

      if (e == NULL) {
        /* XXX log error */
        /* XXX will send empty record-route header outbound if no previous
         * RR header. */
        return;
      }
  
     e->next = h->elements;
     h->elements = e;
  }
  
  URI_Free(&rr_uri);
  return;
} /* AddRecordRoute_Request */

/*
 * BuildRRElement()
 * Build a Record-Route header element referring to the local machine,
 * based on the Contact or From headers.
 */
parser_element_t*
BuildRRElement(const uri_t *u, unsigned long hash, int branchId)
{
  DString rr, qrr;
  uri_t rr_uri;
  uri_nv_t *maddr = NULL;
  int i = 0;
  parser_element_t *e;
  int buflen = 0;
  char* newuser;
  char buf[256] = { 0 };

  rr_uri = URI_Copy(*u);
  
  /* Algorithm:
   *  Take the uri, and append to the user part ourname-hash-branchid.
   * ignore hash and branchid if they are zero. ourname is added to
   * help identify the rr header inserted by us, when a response
   * comes back. 
   * XXX: the maddr parameter itself could be used to identify which
   * one was inserted by us. Todo: Fix.
   */
  
  /* hash == 0 is a case, where concating the tid doesnt make sense */
  if (hash!=0) 
    sprintf(buf, "-%s-%lu-%d", config.ProxyName, hash, branchId);
  /*else
    sprintf(buf, "-%s", config.ProxyName); */

  newuser = rr_uri.user;
  buflen = strlen(buf)  + 20; /* 20 is guard space */
  
  if (newuser) 
    buflen += strlen(newuser);

  rr_uri.user = (char*) malloc(buflen);
  if (newuser) {
    strcpy(rr_uri.user, newuser);
    strcat(rr_uri.user, buf);
    free(newuser);
  }
  else
    strcpy(rr_uri.user, buf);
  
  /* Find if rr_uri has an maddr address */
  if (rr_uri.parameters) {
    for (i = 0; rr_uri.parameters[i].name != NULL; i++) {
      if (strcasecmp(rr_uri.parameters[i].name, "maddr") == 0) {
        maddr = &rr_uri.parameters[i];
        break;
      }
    }
  }
  
  if (maddr == NULL) {
    /* If not, create one */
    uri_nv_t *tmp_ptr;
    tmp_ptr = realloc(rr_uri.parameters, sizeof (uri_nv_t) * (i + 2));
    if (tmp_ptr == NULL) {
      URI_Free(&rr_uri);
      return NULL;
    }
    rr_uri.parameters = tmp_ptr;

    memset(&rr_uri.parameters[i+1], 0, sizeof(uri_nv_t));

    maddr = &rr_uri.parameters[i];
    maddr->name = strdup("maddr");
    maddr->value = NULL;
  }
  
  /* Set maddr value to server name */
  strassign(&maddr->value, config.ProxyName);
  
  /* Set port to local port */
  rr_uri.port = config.port;

  DStringInit(&rr);
  DStringInit(&qrr);

  /* Print RR URI */
  URI_SPrint(&rr_uri, &rr);
  URI_Free(&rr_uri);

  /* Make into name-addr form */
  DStringAppend(&qrr, "<", 1);
  DStringAppendD(&qrr, &rr);
  DStringAppend(&qrr, ">", 1);

  DStringFree(&rr);

  /* Build parser element */
  e = (parser_element_t *) malloc(sizeof(parser_element_t));
  if (e == NULL) {
    DStringFree(&qrr);
    return NULL;
  }

  memset(e, 0, sizeof(parser_element_t));

  strassign(&e->element, DStringValue(&qrr));
  strassign(&e->value, DStringValue(&qrr));

  DStringFree(&qrr);
  
  return e;
} /* BuildRRElement */


/*
 * ModifyRRElement
 * 
 * Called by the policy handler when a 200 Response is received
 * for INVITE. Do the transformation as specified in bis03 {
 * change our RR header by modifying the contact from caller to callee}.
 */
void ModifyRRElement(message_t *m, unsigned long hash) {

  parser_header_t *h;
  parser_element_t *e;
  parser_header_t *hcontact;
  parser_element_t *save, *temp;
    
  char buf[256];
  uri_t rr_uri;

  if((h = RFC822_Header(m->p.headers, "Record-Route")) == NULL) {
  /* What is happening? there should have been a record-route inserted
   * by us, at the very least.
   */
    return;
  }
  
  if (hash)
    sprintf(buf, "-%s-%lu", config.ProxyName, hash);
  
  /* h != NULL */
  temp = save = h->elements;
  while (temp) {
    if (temp->value) {
      if (strstr(temp->value, buf)) {
        /* this is the element to modify */
        break;
      }
    }
    if(temp!=save) save = temp;
    temp = temp->next;
  }
  
  if (!temp) {
    /* try for the stateless case */
    sprintf(buf,"-%s", config.ProxyName);
  
    temp = save = h->elements;
    while (temp) {
      if (temp->value) {
        if (strstr(temp->value, buf)) /* this is the element to modify */
	  break;
      }
      if (temp!=save) save = temp;
      temp = temp->next;
    }
    
    return;
  }
  
  if (!(hcontact = RFC822_Header(m->p.headers, "Contact"))) 
    hcontact = RFC822_Header(m->p.headers, "From");
  
  if (hcontact && hcontact->elements->value)  {
    rr_uri  = URI_Parse(hcontact->elements->value);
    e = BuildRRElement(&rr_uri, 0, 0);
    
    if (e == NULL) {
      /* XXX log error */
      /* XXX will send empty record-route header outbound if no previous
       * RR header. */
      return;
    }
    
    /*
     * now delete that element only, taking care of the fact that
     * only one element may be there.
     */
    if (save != temp) { 
      save->next = e;
      if (h->last_element == temp)
        h->last_element = e;
      e->next = temp->next;
    }
    else 
    { /* first element deletion */
      if (h->elements == h->last_element)
        h->elements = h->last_element = e;
      else {
        e->next = h->elements->next;
        h->elements = e;
      }
    }
    
    temp->next = NULL;
    if (temp->element) free(temp->element);
    RFC822_ParameterFree(temp->parameters);
    free(temp);
    
    URI_Free(&rr_uri);
  }  
  
  return;
  
} /* ModifyRRElement */


/*
 * Given a set of Record-Route headers, builds a route set for a UAS.
 * Used by the presence agent.
*/
char* BuildUASRouteSet(parser_t *p) 
{

  parser_header_t *h = RFC822_Header(p->headers, "Record-Route");
  parser_element_t *e;
  parser_parameter_t *pp;
  
  DString d;
  
  DStringInit(&d);

  if (!h) goto contact;
  
  e = h->elements;

  while (e) {
    
    DStringPrintf(&d,"Route: %s", e->value);

    for (pp = e->parameters; pp; pp = pp->next) 
      DStringPrintf(&d, " ; %s=%s",pp->attribute, pp->value);

    DStringAppend(&d, "\r\n", -1);

    e = e->next;
  }
  
contact:
  /*  Now Append the contact header */
  h = RFC822_Header(p->headers, "Contact");
  
  if (h && h->elements) {
    
    e = h->elements;
    
    DStringPrintf(&d,"Route: %s\n", e->value);

    for (pp = e->parameters; pp; pp = pp->next) 
      DStringPrintf(&d, " ; %s=%s",pp->attribute, pp->value);

    DStringAppend(&d, "\r\n", -1);
  }

  return DStringValue(&d);
}





