#ifdef __SIPSNMP_MIB__
/*
 * commoncfg.c
 *
 *  This file implements the get/set interface functions for SipCommonCfg Objects.
 *
 * Configuration parameters are sent out/set in the functions here.
 *
 * For a discussion on general design used through out this module, see the 
 * design doc in the libsipsnmp root directory.
 *
 * Note that this file is intended to be included inside sipMIB.c and hence
 * will not be compiled directly.
 */
 
/*
 * var_sipCommonCfgTable(): Cfg objects including protocol version, 
 * uptime, etc - see MIB definition for more details
 */
GET_FUNCTION(var_sipCommonCfgTable)
{
  
  const SipCommonCfgInterface *p_cfgIntf;
  
  p_cfgIntf = &p_sipEntity->sipCommonCfgTable.sipCommonCfgInterface;
  
  /* 
   * Algorithm
   *  Locate the appropriate variable in our SipEntity data structure,
   * and get its value. Do mutex locking as appropriate
   */

  /* table size is 1  for this subtree*/
  if (header_simple_table(vp,name,length,exact,var_len,write_method, 1) 
      == MATCH_FAILED )
    return NULL;
  
  LOCK();
  
  switch(vp->magic) {
    
  case SIPPROTOCOLVERSION:
    strncpy(string,p_cfgIntf->sipProtocolVersion,SPRINT_MAX_LEN);
    *var_len = strlen(string);
    goto cfg_strret;
    
  case SIPSERVICEOPERSTATUS:
    long_ret = p_cfgIntf->sipServiceOperStatus;
    break;
    
  case SIPSERVICEADMINSTATUS:
    *write_method = write_sipServiceAdminStatus;
    long_ret = p_cfgIntf->sipServiceAdminStatus;
    break;
    
  case SIPSERVICESTARTTIME:
    
    {
      /*
       * the start time is in time_t. we need to convert that into snmp's TimeTicks
       * which is relative time since server start in hundredth of a second.
       *
       * this means, find the difference between current time, and the time_t
       * value contained in sipServiceStartTime, and multiply it by 100
       */
      struct timeval t;
      gettimeofday(&t,NULL);
      
      long_ret = (t.tv_sec - p_cfgIntf->sipServiceStartTime) * 100;
    }
    break;
    
  case SIPSERVICELASTCHANGE:
    
    {
      /* same as sipServiceStartTime, for our impl */
      struct timeval t;
      gettimeofday(&t,NULL);
      
      long_ret = (t.tv_sec - p_cfgIntf->sipServiceLastChange) * 100;
    }
    break;
    
  case SIPORGANIZATION:
    
    *write_method = write_sipOrganization;
    strncpy(string,p_cfgIntf->sipOrganization,SPRINT_MAX_LEN);
    *var_len = strlen(string);
    goto cfg_strret;
    
  case SIPMAXTRANSACTIONS:
    
    long_ret = p_cfgIntf->sipMaxTransactions;
    break;
    
  case SIPINVITETIMEOUT:
    
    *write_method = write_sipInviteTimeout;
    long_ret = p_cfgIntf->sipInviteTimeOut;
    break;
    
  case SIPHIDEOPERATION:
    
    *write_method = write_sipHideOperation;
    long_ret = p_cfgIntf->sipHideMode;
    break;
    
  case SIPUSERLOCATIONSERVERADDR:

    *write_method = write_sipUserLocationServerAddr;
    strncpy(string,p_cfgIntf->sipUserLocationServerAddr,SPRINT_MAX_LEN);
    *var_len = strlen(string);
    goto cfg_strret;
    
  case SIPCANUSECOMPACT:
        
    long_ret = p_cfgIntf->sipCanUseCompact;
    break;
    
  case SIPREQUESTURIHOSTMATCHING:

    *write_method = write_sipRequestUriHostMatching;
    long_ret = p_cfgIntf->sipRequestURIMatch;
    break;
    
    
  default:
    ERROR_MSG(DEFERR);
    goto cfg_errret;
  }

  UNLOCK();
  return (unsigned char *) &long_ret;
  
 cfg_strret:
  UNLOCK();
  return (unsigned char *) string;
  
 cfg_errret:
  UNLOCK();
  return NULL;
}

/*
 * var_sipPortTable(): Return a list of ports that we use.
 * 
 */
GET_FUNCTION(var_sipPortTable)
{
  
  const SipCommonCfgTable *p_cfg = &p_sipEntity->sipCommonCfgTable;
  const SipCommonPort *p_head, *p_toserve;
  
  int oidlen = SIPMIB_OIDLEN + sipMIB_variables[SIPTRANSPORT].namelen -1;
  
  /* 
   * Algorithm
   *  Locate the appropriate variable in our SipCommonPortTable linked list,
   *  get its value. Do mutex locking as appropriate
   */
  
  LOCK();

  p_toserve = p_head = p_cfg->pSipCommonPortTable;
  
  if(p_cfg->nPortTable <= 0 || !p_head) 
    goto port_errret;
  
  /* sipMIB's oid is .1.3.6.1.2.1.9998 - which is length 7 defined in macro
     SIPMIB_OIDLEN in sipMIB.c
     after this, we need to append the oid of the elements in this subtree.
     to get this we find the length from sipMIB_variables[SIPTRANSPORT].namelen
     which is 5. 7 + 5 = 12, which means that from 0 to 11 we have the oid.
     
     this can also be viewed as below:
     vp->name[9] contains sipPortTable (2)
     vp->name[10] contains sipPortEntry (1)
     we need to set (or consider)
     vp->name[11] as 2 for sipTransport
     vp->name[11] as 3 for sipRowStatus
     
     this function will get called only when the above is true
     
     then	
     vp->name[12] contains applnIndex (vp->name[oidlen])
     vp->name[13] contains portIndex (vp->name[oidlen+1])
  */ 	
  
  if(!exact) {
    
    if(vp->namelen != oidlen) goto port_errret; 
    
    if(vp->name[oidlen - 3] != 2) goto port_errret;

    if(vp->name[oidlen - 3] == 2 && name[oidlen - 3]  < 2) {
      name[oidlen - 3] = 2;
      name[oidlen - 2] = vp->name[oidlen-2];
      name[oidlen-1] =  vp->name[oidlen-1];
      *length = vp->namelen;
    }

    if(*length <= oidlen) { /* concat applnIndex and portIndex */
      
      memcpy(name,vp->name,vp->namelen * sizeof(oid)); 
      name[vp->namelen] = 1;
      name[vp->namelen + 1] = 5060;
      p_toserve = p_head;
      *length = vp->namelen + 2;
      goto portgoto;
    }
    
    if(*length == oidlen+1) { /* applnIndex is already there */
      if(name[*length - 1] != 1) goto port_errret;
      
      name[*length] = 5060;
      *length = *length +1;
      goto portgoto;
    } 
    
    if(*length != oidlen+2) goto port_errret;
    
    if(name[*length - 3] != vp->name[*length - 3]) {
      name[*length - 3] = vp->name[*length - 3];
      name[*length - 2] = 1;
      name[*length - 1] = 5060;
      p_toserve = p_head;
    }
    else {
      
      int index = name[*length - 1];
      /* find it */
      while(p_head && p_head->sipPort != index)
	p_head = p_head->pNext;
      
      if(!p_head || !p_head->pNext)
	goto port_errret;
      
      p_head = p_head->pNext;
      
      name[*length - 1]  = p_head->sipPort;
      
      p_toserve = p_head; 
    }
    goto portgoto;
  }
  
  if(exact) { 
    int index1= name[*length -2];
    int index = name[*length - 1];
    
    if(*length != oidlen+2 || index1!=1) goto port_errret;
    
    while(p_head && p_head->sipPort != index)
      p_head = p_head->pNext;
    
    if(!p_head)
      goto port_errret;
    
    p_toserve = p_head;
  }
 portgoto: 
  *var_len = sizeof(long);
  
  switch(vp->magic) {
    
  case SIPTRANSPORT:
    *write_method = write_sipTransport;
    long_ret = p_toserve->sipTransport; 
    goto port_retlong;

  case SIPPORTSTATUS:
    *write_method = write_sipPortStatus;
    long_ret = p_toserve->sipPortStatus; 
    goto port_retlong;

  default:
    ERROR_MSG(DEF_ERR);
  }
  
 port_errret:
  UNLOCK();
  return NULL;
 port_retlong:
  UNLOCK();
  return (unsigned char*) &long_ret;
}

/*
 * var_sipUriSupportedTable():
 */
GET_FUNCTION(var_sipUriSupportedTable)
{


  const SipCommonCfgTable *p_cfgTable = &p_sipEntity->sipCommonCfgTable;

  const SipCommonURI *p_head, *p_toserve;
  
  int oidlen = SIPMIB_OIDLEN + sipMIB_variables[SIPURISUPPORTED].namelen - 1;

  LOCK();
  
  p_head = p_toserve = p_cfgTable->pSipCommonURITable;
  
  if(p_cfgTable->nURITable <= 0 || !p_head) 
    goto uri_errret;
  
  if(!exact) {
    
    if(vp->namelen != oidlen) goto uri_errret;


    if(vp->name[oidlen - 3] != 3) goto uri_errret;

    if(vp->name[oidlen - 3] == 3 && name[oidlen - 3]  < 3) {
      name[oidlen - 3] = 3;
      name[oidlen - 2] = vp->name[oidlen - 2];
      name[oidlen - 1] =  vp->name[oidlen -1];
      *length = vp->namelen;
    }


    if(*length <= oidlen) { /* concat applnIndex and uriIndex */
      
      memcpy(name,vp->name,vp->namelen * sizeof(oid)); 
      name[vp->namelen] = 1;
      name[vp->namelen + 1] = 1;
      p_toserve = p_head;
      *length = vp->namelen + 2;
      goto urigoto;
    }
    
    if(*length == oidlen+1) {
      if(name[*length - 1] != 1) goto uri_errret;
      
      name[*length] = 1;
      *length = *length +1;
      goto urigoto;
    } 
    
    if(*length != oidlen+2) goto uri_errret;
    
    if(name[*length - 3] != vp->name[*length - 3]) {
      name[*length - 3] = vp->name[*length - 3];
      name[*length - 2] = 1;
      name[*length - 1] = 1;
      p_toserve = p_head;
    }
    else {
      
      int index = name[*length - 1];
      /* find it */
      while(p_head && p_head->sipUriSupportedIndex != index)
	p_head = p_head->pNext;
      
      if(!p_head || !p_head->pNext)
	goto uri_errret;
      
      p_head = p_head->pNext;
      
      name[*length - 1]  = p_head->sipUriSupportedIndex;
      
      p_toserve = p_head; 
    }
    goto urigoto;
  }
  
  if(exact) { 
    int index1= name[*length -2];
    int index = name[*length - 1];
    
    if(*length != oidlen+2 || index1!=1) goto uri_errret;
    
    while(p_head && p_head->sipUriSupportedIndex != index)
      p_head = p_head->pNext;
    
    if(!p_head)
      goto uri_errret;
    
    p_toserve = p_head;
  }
  
 urigoto: 
  *var_len = sizeof(long);

  switch(vp->magic) {
    
  case SIPURISUPPORTED:
    
    strncpy(string,p_toserve->sipUriSupported,SPRINT_MAX_LEN);
    *var_len = strlen(string);
    UNLOCK();
    return (unsigned char *) string;
    
  default:
    ERROR_MSG(DEF_ERR);
  }
  
 uri_errret:
  UNLOCK();
  return NULL;
}

/*
 * var_sipFtrSupportedTable():
 *   Handle this table separately from the scalar value case.
 *   The workings of this are basically the same as for var_sipMIB above.
 */
GET_FUNCTION(var_sipFtrSupportedTable)
{
  
  const SipCommonCfgTable *p_cfgTable = &p_sipEntity->sipCommonCfgTable;
  const SipCommonFtr *p_head, *p_toserve;

  int oidlen = SIPMIB_OIDLEN + sipMIB_variables[SIPFTRSUPPORTED].namelen - 1;

  LOCK();
  
  p_head = p_toserve = p_cfgTable->pSipCommonFtrTable;
  
  if(p_cfgTable->nFtrTable <= 0 || !p_head) 
    goto ftr_errgoto;
  
  if(!exact) {
    
    if(vp->namelen != oidlen) goto ftr_errgoto; 
    
    if(vp->name[oidlen-3] != 4) goto ftr_errgoto;    
    
    if(vp->name[oidlen-3] == 4 && name[oidlen-3]  < 4) {
      name[oidlen-3] = 4;
      name[oidlen-2] = vp->name[oidlen-2];
      name[oidlen-1] =  vp->name[oidlen-1];
      *length = vp->namelen;
    }

    if(*length <= oidlen) { /* concat applnIndex and uriIndex */
      
      memcpy(name,vp->name,vp->namelen * sizeof(oid)); 
      name[vp->namelen] = 1;
      name[vp->namelen + 1] = 1;
      p_toserve = p_head;
      *length = vp->namelen + 2;
      goto ftrgoto;
    }
    
    if(*length == oidlen+1) {
      if(name[*length - 1] != 1) goto ftr_errgoto;
      
      name[*length] = 1;
      *length = *length +1;
      goto ftrgoto;
    } 
    
    if(*length != oidlen+2) goto ftr_errgoto;
    
    if(name[*length - 3] != vp->name[*length - 3]) {
      name[*length - 3] = vp->name[*length - 3];
      name[*length - 2] = 1;
      name[*length - 1] = 1;
      p_toserve = p_head;
    }
    else {
      
      int index = name[*length - 1];
      /* find it */
      while(p_head && p_head->sipFtrSupportedIndex != index)
	p_head = p_head->pNext;
      
      if(!p_head || !p_head->pNext)
	goto ftr_errgoto;
      
      p_head = p_head->pNext;
      
      name[*length - 1]  = p_head->sipFtrSupportedIndex;
      
      p_toserve = p_head; 
    }
    goto ftrgoto;
  }
  
  if(exact) { 
    
    int index1= name[*length -2];
    int index = name[*length - 1];
    
    if(*length != oidlen+2 || index1!=1) goto ftr_errgoto;
    
    while(p_head && p_head->sipFtrSupportedIndex != index)
      p_head = p_head->pNext;
    
    if(!p_head)
      goto ftr_errgoto;
    
    p_toserve = p_head;
  }

 ftrgoto: 
  *var_len = sizeof(long);

  switch(vp->magic) {

  case SIPFTRSUPPORTED:
    
    strncpy(string,p_toserve->sipFtrSupported,SPRINT_MAX_LEN);
    *var_len = strlen(string);
    UNLOCK();
    return (unsigned char *) string;
      
  default:
    ERROR_MSG(DEF_ERR);
  }
 ftr_errgoto:
  UNLOCK();
  return NULL;
}

static unsigned char cfg_storage[STANDARDSTRLEN];
static unsigned int cfg_int;

typedef int (*FPTR)(SipCommonCfgInterface*,u_int);
typedef int (*SPTR)(SipCommonCfgInterface*,const char*);

static int simple_cfgwrite_string(int  action,u_char   *var_val,
				  u_char   var_val_type,size_t   var_val_len,
				  u_char   *statP,oid      *name,
				  size_t   name_len,SPTR finvoke, 
				  SipCommonCfgInterface *p_cfg)
{
  int value;

  switch ( action ) 
    {
    case RESERVE1:
      if (var_val_type != ASN_OCTET_STR)          
	return SNMP_ERR_WRONGTYPE;
      
      if (var_val_len > sizeof(cfg_storage))
	return SNMP_ERR_WRONGLENGTH;
      
      break;
      
    case RESERVE2:  /* todo: check */
      strncpy(cfg_storage,var_val,STANDARDSTRLEN);
      break;
      
      
    case COMMIT:
      value = finvoke(p_cfg,cfg_storage);
      if(value < 0) return SNMP_ERR_BADVALUE;
      break;
      
    }
  return SNMP_ERR_NOERROR;
}

static int simple_cfgwrite_int(int  action,u_char   *var_val,
			       u_char   var_val_type,size_t   var_val_len,
			       u_char   *statP, oid      *name,
			       size_t   name_len,FPTR finvoke,
			       SipCommonCfgInterface *p_cfg)
{
  int val;

  switch ( action ) 
    {
    case RESERVE1:
      if (var_val_type != ASN_INTEGER)          
	return SNMP_ERR_WRONGTYPE;
      
      if (var_val_len > sizeof(long*))
	return SNMP_ERR_WRONGLENGTH;
      
      break;
      
    case RESERVE2:
      cfg_int = (unsigned int)*(long*)var_val;
      break;
      
    case COMMIT:
      
      val = finvoke(p_cfg,cfg_int);
      if(val < 0) return SNMP_ERR_BADVALUE;
      break;
      
    }
  return SNMP_ERR_NOERROR;
}

SET_FUNCTION(write_sipServiceAdminStatus)
{
  
  SipCommonCfgInterface *p_intf = (SipCommonCfgInterface*)
    &p_sipEntity->sipCommonCfgTable.sipCommonCfgInterface;
  
  return 
    simple_cfgwrite_int(action,var_val,
                        var_val_type,
			var_val_len,
			statP,
			name,name_len,
			p_intf->setSipServiceAdminStatus,
			p_intf);
  
}

SET_FUNCTION(write_sipOrganization)
{
  
  SipCommonCfgInterface *p_intf = (SipCommonCfgInterface*)
    &p_sipEntity->sipCommonCfgTable.sipCommonCfgInterface;
  
  return SNMP_ERR_READONLY;
  
  return simple_cfgwrite_string(action,
				var_val,var_val_type,var_val_len,
                                statP,
				name,name_len,
				p_intf->setSipOrganization,p_intf);
}

SET_FUNCTION(write_sipInviteTimeout)
{
  
  SipCommonCfgInterface *p_intf = (SipCommonCfgInterface*)
    &p_sipEntity->sipCommonCfgTable.sipCommonCfgInterface;
  
  return simple_cfgwrite_int(action,var_val,var_val_type,var_val_len,statP,
			     name,name_len,
			     p_intf->setSipInviteTimeOut,p_intf);
  
}

SET_FUNCTION(write_sipHideOperation)
{
  SipCommonCfgInterface *p_intf = (SipCommonCfgInterface*) 
    &p_sipEntity->sipCommonCfgTable.sipCommonCfgInterface;

  return SNMP_ERR_READONLY;
  
  return simple_cfgwrite_int(action,var_val,var_val_type,var_val_len,statP,
			     name,name_len,p_intf->setSipHideMode,p_intf);
  
}


SET_FUNCTION(write_sipUserLocationServerAddr)
{
  
  SipCommonCfgInterface *p_intf = (SipCommonCfgInterface*)
    &p_sipEntity->sipCommonCfgTable.sipCommonCfgInterface;
 
  return SNMP_ERR_READONLY;
  
  return simple_cfgwrite_string(action,var_val,var_val_type,var_val_len,statP,
				name,name_len,
				p_intf->setSipUserLocationServerAddr,p_intf);
  
}

SET_FUNCTION(write_sipRequestUriHostMatching)
{
  SipCommonCfgInterface *p_intf = (SipCommonCfgInterface*)
    &p_sipEntity->sipCommonCfgTable.sipCommonCfgInterface;
 
  return SNMP_ERR_READONLY;
  
  return simple_cfgwrite_int(action,var_val,var_val_type,var_val_len,statP,
			     name,name_len,
			     p_intf->setSipRequestURIMatch,p_intf);
}

SET_FUNCTION(write_sipTransport)
{
  /* not implemented currently */
  return SNMP_ERR_NOCREATION;
}

SET_FUNCTION(write_sipPortStatus)
{
  /* not implemented currently */
  return SNMP_ERR_NOCREATION;
}


#endif



