/**********************************************************************
 *
 *               libsipsnmp - snmp support provider for sip.
 *
 *       sipentity_commonds.h : Interface definitions of general SNMP
 *                              vars common to all sip entities
 *
 *      Copyright(c) 2000-2001 Columbia University ; all rights reserved
 *
 *           Initial version by: Sankaran N, February 2001.
 *
 *********************************************************************
*/

#ifndef SIPENTITY_COMMONDS_H__
#define SIPENTITY_COMMONDS_H__

#include <sys/time.h>

#ifndef STANDARDSTRLEN        /* snmp octet string, typical length */
#define STANDARDSTRLEN  256
#endif

#ifndef SERVER_NAME
#define SERVER_NAME "Columbia-SIP-Server"
#endif


/*
 *
 * Main config object, common for all sip entities.
 *
 */
typedef struct sip_common_cfg_interface {
  
#define VERSIONSTRLEN	12
#define SIPVERSION "SIP/2.0"
  char sipProtocolVersion[VERSIONSTRLEN];
  
  char *sipOrganization;   
  int (*setSipOrganization)(struct sip_common_cfg_interface* p_this,const char* p_buf); 
  
#define OSTATUS_UP         1       /* i can process sip messages */
#define OSTATUS_DOWN       2       /* i can't process sip messages now */
#define OSTATUS_TESTING    3       /* i am being tested */
#define OSTATUS_UNKNOWN    4       /* undeterminable */
#define OSTATUS_HALTED     5       /* i have been told not to process sip messages,now */
#define OSTATUS_RESTARTING 6       /* i am restarting, and soon you can expect up stats*/
  u_int sipServiceOperStatus;
  
#define ASTATUS_SHUTDOWN    1    /* immediately kill the app, dont restart */
#define ASTATUS_RESET       2    /* restart the app. and reinitialize data  */
#define ASTATUS_RESTART     3    /* restart the app, no data loss */
#define ASTATUS_STOP        4    /* dont accept new requests */
#define ASTATUS_RTS         5    /* return to service */
  
  u_int sipServiceAdminStatus;
  int (*setSipServiceAdminStatus)(struct sip_common_cfg_interface* p_this,u_int newStatus);
  
  time_t sipServiceStartTime;    /* see definition of time_t, note that
				    the snmp part needs to convert into
				    relative time in 100th of a second as
				    per ASN.1/SNMPSMI semantics
				 */
  
  time_t sipServiceLastChange;
  
  u_int sipInviteTimeOut;
  int (*setSipInviteTimeOut)(struct sip_common_cfg_interface* p_this,u_int newTimeOutSecs);
  
#define HIDE_NONE     1
#define HIDE_DEFAULT  HIDE_NONE
#define HIDE_HOP      2
#define HIDE_ROUTE    3 
  u_int sipHideMode;
  int (*setSipHideMode)(struct sip_common_cfg_interface* p_this,u_int newMode);  
  

  u_int sipMaxTransactions;
 
#define COMPACT_SUPPORTED_TRUE       1
#define COMPACT_SUPPORTED_FALSE      2
  u_int sipCanUseCompact;
  
  /* the ipv4 or hostname of the primary name server */
  char sipUserLocationServerAddr[STANDARDSTRLEN];
  int (*setSipUserLocationServerAddr)(struct sip_common_cfg_interface* p_this,
				      const char* newAddr);

#define REQUEST_URI_HOSTMATCH_TRUE   1
#define REQUEST_URI_HOSTMATCH_FALSE  2
  u_int sipRequestURIMatch;
  int (*setSipRequestURIMatch)(struct sip_common_cfg_interface* p_this,
			       u_int newValue);
  
} SipCommonCfgInterface;

/*
 *
 * Port Table 
 *
 */
typedef struct sip_common_port {
  
  u_int sipPort;             
  
#define TRANSPORT_UDP     1
#define TRANSPORT_TCP     2
#define TRANSPORT_BOTH    3
  u_int sipTransport;
  
  u_int sipPortStatus;     
  
  struct sip_common_port *pNext;   /* linked list */
  
} SipCommonPort;


/*
 *
 * URL's other that sip that are supported by sip entities.
 * 
 * URI Table
 *
 */
typedef struct sip_common_uri {
  
  u_int sipUriSupportedIndex;
  
  char sipUriSupported[STANDARDSTRLEN];
  
  struct sip_common_uri *pNext;   /* linked list */
  
} SipCommonURI;


/*
 *
 * Extensions - ex. com.example.billing
 * 
 */
typedef struct sip_common_ftr {
  
  u_int sipFtrSupportedIndex;
  
  char sipFtrSupported[STANDARDSTRLEN]; 
  
  struct sip_common_ftr *pNext;   /* linked list*/
  
} SipCommonFtr;


/*
 *
 * Summary statistics - in/out counters for requests sent,received,
 * and responses sent/received. 
 *
 */
typedef struct sip_common_summary_stats {
  
  int   sipSummaryInRequests;
  
  int   sipSummaryOutRequests;
  
  int   sipSummaryInResponses;
  
  int   sipSummaryOutResponses;
  
  int   sipSummaryTotalTransactions;   /* non-1xx final response */
  
} SipCommonSummaryStats;


#define STATS_IN  0                     
#define STATS_OUT 1
#define STATS_MAXTYPE  2

typedef struct sip_common_method_stats {

#define STATS_INVITE    0                    /* do not change the #def-number ordering */

#define STATS_ACK       1
  
#define STATS_BYE       2
  
#define STATS_CANCEL    3
  
#define STATS_OPTIONS   4
  
#define STATS_REGISTER  5
  
#define STATS_MAX       6
  
  int sipStats[STATS_MAX][STATS_MAXTYPE];
  
} SipCommonMethodStats;


/*
 *
 * 1xx,2xx,... responses sent - aggregate count.
 *
 * Do not change the order.
 *
 */
typedef struct sip_common_statuscode_stats {
  
#define CLASS_INFO          0   /* 1xx class */

#define CLASS_SUCCESS       1   /* 2xx class */

#define CLASS_REDIR         2

#define CLASS_REQFAIL       3

#define CLASS_SERVFAIL      4

#define CLASS_GLOBALFAIL    5

#define CLASS_MAX           6
  
  int sipStatsInfoClass[CLASS_MAX][STATS_MAXTYPE];           
  
} SipCommonStatusCodeStats;

  
/*
 *
 * A sip transaction.
 *
 */
typedef struct sipcommon_transaction_mib {
  
  u_int  transaction_id;   
  
  int isValid;         /* standard 'c' convention of 0 as false, 1 as true */
  
  time_t timedeleted; /* the time at which this xact was deleted (as it got
			 over). Note that if timedelted == 0, p_transact
			 is valid. otherwise it is not (is null). 
			 
			 though timedeleted may be non-zero, isValid may be
			 1, until our timer detects that it is good for
			 deletion.
		      */
  
  void* p_transact;  /* implementation specific rep */
  
  char sipTransTo[256];
  
  char sipTransFrom[256];
  
  char sipTransCallId[256];
  
  u_int sipTransCSeq;
  
#define TRANS_NULL           1
#define TRANS_INITIAL        2
#define TRANS_CALLING        3
#define TRANS_RINGING        4
#define TRANS_COMPLETED      5
#define TRANS_CALLPROCEEDING 6
#define TRANS_FAILURE        7
#define TRANS_SUCCESS        8
#define TRANS_CONFIRMED      9
   
  char sipTransExpiry[24]; /* yyyy-mm-dd hh:mm:ss */
  
  char sipTransCallingPartyContentType[256];
  
  char sipTransCalledPartyContentType[256];
 
  /* Added on 12/April/2001 for performance profiling
   *
   * For register time_start, time_200 sent are used.
   * For invite all are used, though the most important thing
   * i am interested right now is time_start to time_100sent.
  */
  struct timeval time_start;

  struct timeval time_100sent;

  struct timeval time_180sent;

  struct timeval time_finalsent;
  
  int finalCode;
  /* 12/April/2001 End */

  struct sipcommon_transaction_mib* pNext;
 
   
} SipCommonTransactionMib;

/* must have the same type as the p_transact pointer.
 * implementation specific.
*/
extern void updateTransactionTimes(void* p_req);

/*
 *
 * Given a transaction_id, returns the element.
 *
 * These accessor functions remind me the following quote-
 *
 * "Any sufficiently complicated C program contains an ad hoc 
 *  informally-specified bug-ridden slow implementation of 
 *  half of Common Lisp" 
 *
 */
extern const SipCommonTransactionMib* 
getTransactionById(const SipCommonTransactionMib *p_head, 
		   u_int transaction_id);


/*
 * 
 * call this function to get 'n' th valid element of a transaction list.
 * do not iterate directly, as it may contain holes
 *
 */
extern const SipCommonTransactionMib* 
getTransactionByIndex(const SipCommonTransactionMib *p_head, 
		      int index);

/*
 *
 * Given a transaction, find its current state (one of the TRANS_XXX)
 * values. The p_numOutStandingBranches is in-out, must be non-null,
 * and on return will contain the number of branches that are active
 *
 */
extern u_int getTransactionState(const SipCommonTransactionMib *p_this,
				 u_int* p_numOutStandingBranches);


/*
 *
 * Status codes - extension support 
 *
 */
typedef  struct sip_common_scode_special {
  
  u_int sipCommonStatusCodeValue;         /* which status code to monitor */
  
  u_int sipCommonStatusCodeIns;           /* in-responses with this scode */
  
  u_int sipCommonStatusCodeOuts;          /* out-responses with this scode */
  
  u_int sipCommonStatusCodeSendNotif;     /* truthvalue  TRU or FAL - whether to send traps?*/
  
  u_int sipCommonStatusCodeStatus;        /* SNMP row status - active, etc; */
  
  struct sip_common_scode_special *pNext; /* linked list */
  
} SipCommonSCodeSpecial;


/*
 * 
 * Extension methods (non rfc2543 methods) - special monitoring support
 *
 */
typedef struct sip_common_extmethod_stats {
  
  u_int  sipStatsExtMethodIndex; 
  
  char   sipStatsExtMethodName[20];          /* use sizeof in strncpy, name of method */
  
  u_int  sipStatsExtMethodIns;               /* ins? */
  
  u_int  sipStatsExtMethodOuts;              /* outs? */

  u_int  sipStatsExtMethodRowStatus;         /* snmp row status, active...*/
  
  struct sip_common_extmethod_stats *pNext;  /* linked list */
  
} SipCommonExtMethodStats; 


/*
 * 
 * SNMP_Trap object.
 *
 * This is the object that should be used while sending traps (notifications).
 *
 */
typedef struct sip_common_notif {
  
  u_int sipStatusCodeNotifValue;                   /* the status code that caused this? */
  
  char sipStatusCodeNotifTo[STANDARDSTRLEN];       /* To: */
   
  char sipStatusCodeNotifFrom[STANDARDSTRLEN];     /* From: URI */

  char sipStatusCodeNotifCallId[STANDARDSTRLEN];   /* Call-Id: */
  
  u_int sipStatusCodeNotifCSeq;                    /* CSeq value */ 
  
} SipCommonNotif;   


/*
 *
 * send_trap sends a snmpv2 notification to all configured hosts.
 *
 * The host configuration is independent of our code, and it is
 * done by administrator who runs snmpd.
 *
 */
extern void send_trap(const SipCommonNotif* p_notifData);

/*
 *
 * SipCommonCfgTable - common cfg objects to all sip entities.
 *
*/
typedef struct sip_common_cfg_table {

  SipCommonCfgInterface  sipCommonCfgInterface;
  
  int nPortTable;                      /* number of elements in port table */
  SipCommonPort *pSipCommonPortTable;  /* linked list of port entries for this entity */
  int (*addPortElement)(const SipCommonPort *p_new);
  int (*delPortElement)(SipCommonPort *p_which);
  
  int nURITable;                       /* number of elements in URI Table */
  SipCommonURI *pSipCommonURITable;    /* linked list of uri entries for this entity */
  
  int nFtrTable;                       /* number of elements in Ftr table */
  SipCommonFtr *pSipCommonFtrTable;    /* linked list of ftr entries for this entity */

} SipCommonCfgTable;


/*
 * Summary statistics - common summary for all entities.
 */
typedef struct sip_common_summ_stats_table {


  SipCommonSummaryStats sipCommonSummaryStats;
  
  SipCommonMethodStats  sipCommonMethodStats;
  
  SipCommonStatusCodeStats  sipCommonStatusCodeStats;
  
  int nCommonSCodeSpecial;
  SipCommonSCodeSpecial *pSCSCS_head; /* sorry for this name */
  
  int (*addSCodeToMonitor)(struct sip_common_summ_stats_table *p_this,
			   const SipCommonSCodeSpecial *p_new);

  int (*delSCodeToMonitor)(struct sip_common_summ_stats_table *p_this,
			   int statusCode);
  
  int nCommonExtMethodStats;
  
  SipCommonExtMethodStats *pSCEMS_head; /* sorry for this naming scheme */
  
  int (*addMethodToMonitor)(struct sip_common_summ_stats_table *p_this,
			    const SipCommonExtMethodStats *p_new);

  int (*delMethodToMonitor)(struct sip_common_summ_stats_table *p_this,
			    int index);
  
  
} SipCommonSummStatsTable;


/*
 * 
 * Search the Extension method linked list for a particular item
 * 
 */
extern const SipCommonExtMethodStats* 
getExtMethodEntry(const SipCommonExtMethodStats *p_head,
		  int extMethodIndex);

/* 
 *
 * Search the extension method linked list for a particular item
 *
 */
extern const SipCommonExtMethodStats* 
getExtMethodEntryByName(const SipCommonExtMethodStats *p_head, 
			const char* methodName);

/*
 *
 * Search the status code linked list for a particular item
 *
 */
extern const SipCommonSCodeSpecial* 
getSCodeEntry(const SipCommonSCodeSpecial *p_head,
	      int scodeValue);


#endif   /* sipentity_commonds.h */
