#ifdef USE_SNMP
/*
 * registrar_impl.c
 *
 * Support functions for SIP registrars
 * 
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <time.h>

/*
 *Max time_t vaule for infinite time
 */
#define MAX_TIME_T  ((time_t) 0x7fffffff)    
#define MAX_TIME_STR  "9999-12-31,23:59:59"
#define HAVE_TIMEGM
#ifndef HAVE_TIMEGM
#include "timegm.h"
#endif    

#include "sysdep.h"

#include "dbapi.h"               /* database accessor api definitions */

#include "sipsnmp_interface.h"   
#include "sipentity_ds.h"        /* interface data structure definitions */

#include "sipsnmp_utils.h"       /* string2time and other helpers */


/*
 * Instead of going to database everytime for fetching contacts,
 * we cache it for sometime in the ContactCache linked list.
 */
typedef struct contact_cache {
  
  SipRegUserEntry* userEntry;  /* which user? */
  
  SipRegContactEntry* contactListHead; /* start of contact list */
  
  time_t timeSaved; /* time cached */
  
  int numContacts;  /* num contacts in the ll pointed to by contactListHead */
  
  struct contact_cache* pNext; /* next user's entry */
  
} ContactCache;

static ContactCache *p_contactCacheHead;


/*
 * getNumRegistedUsers
 *
 *  -Counts and returns the number of distinct users registered to this
 *   system (a distinct user is defined by his email id - thus as of now,
 *   for the marta database, hgs@cs and schulzrinne@cs are different
 *
 * Parameters
 *  -A registrar object, whose p_userList member will be filled with
 *  SipRegUserEntry records, one for each user, as a singly linked list.
 *
 * Return values
 *  -1 on error, 0 on success
 *
 */
int 
getNumRegisteredUsers(SipRegistrar *p_registrar) 
{
  
  static struct timeval last_queried;
  static int last_rows;
  
  int query_rows_ret = 0,i;

  struct timeval curtime;

  ResultTable query_result; /* two-dimensional table of MYSQL result */
  
  SipRegUserEntry* p_temp, *p_prev;
 
  assert(p_registrar);

  /* Algorithm:
   *  -if this is the first time or the cached entry is not valid,
   *   make sql requests, fill appropriate members.
   *  -after the request is made, cache the user count for 600 seconds.
   *
   *  -fill the user list of the p_registrar object with the user data.
   * 
   *  -note that though the p_registrar object is a parameter, i have used
   *  a single static member for caching results. this is because, 
   *  only one p_registrar object is expected to be there at any time (or
   *  the data will be identical).  
   */
  
  gettimeofday(&curtime,NULL);
  
  if(curtime.tv_sec - last_queried.tv_sec < 600) 
    return last_rows;
  else /* we are setting last_rows at curtime */
    last_queried.tv_sec = curtime.tv_sec;
  
  p_temp = (SipRegUserEntry*)p_registrar->p_userList;
  
  query_result = rawQuery("select user from put order by user",
			  &query_rows_ret);
    

  /* now check if the each user is in our global ds. if not, add
   *
   * To do this, first set the isValid of all elements to FALSE.
   * Then update each as it is found
   *
   * the current algo is inefficient as it is o(n2) search in MM,
   * Todo: fix
   */
  while(p_temp) {

    p_temp->isValid = 0;
    
    p_temp = p_temp->pNext;

  }  
  
  for(i = 0;i<query_rows_ret;i++) {
    
    p_temp = p_registrar->p_userList;
    
    p_prev = NULL;
    
    while(p_temp) {
      if(!strcmp(p_temp->sipUserUri,query_result[i][0])) 
	{
	  p_temp->isValid = 1;
	  break;
	}
      p_prev = p_temp;
      p_temp = p_temp->pNext;
    }

    if(p_temp) continue;
    
    /* this element is new, insert at front */
    {
      SipRegUserEntry* p_temp1 = 
	(SipRegUserEntry*)calloc(1,sizeof(SipRegUserEntry));

      if(!p_temp1) 
	{
	  error("getNumRegisteredUsers","calloc failed...");
	  return -1;
	}

      strcpy(p_temp1->sipUserUri,query_result[i][0]);

      p_temp1->isValid = 1;

      if(p_prev) {
        p_temp1->sipUserIndex = p_prev->sipUserIndex + 1;
        p_prev->pNext = p_temp1;
      }
      else {
        p_temp1->sipUserIndex = 1;
        p_registrar->p_userList = p_temp1;
      }

    }

  } /* for loop */
  
  last_rows = query_rows_ret; /* cache result */

  return query_rows_ret;
  
} /* getNumRegisteredUsers */


/*
 * getUser
 *
 *  Finds the user in p_registrar->sipUserIndex whose userIndex
 *  matches sipUserIndex
 *
 * Return valeus:
 *  -the pointer to SipRegUserEntry if a match is found, or not.
 *
 * Worst case Complexity : o(n)
 */    
const SipRegUserEntry* 
getUser(const SipRegistrar *p_registrar, int sipUserIndex) 
{
  
  SipRegUserEntry* p_user;
  
  /*
   * Algorithm:
   * -Get the number of users, if none were found, return NULL.
   *
   * -Otherwise search and return.
   */
  
  if(getNumRegisteredUsers((SipRegistrar*)p_registrar) <= 0) {
    debug("getUser","No users were found");
    return NULL;
  }

  p_user = p_registrar->p_userList;
  
  while(p_user) {
  
    if(p_user->sipUserIndex == sipUserIndex)
      return (const SipRegUserEntry*) p_user;
    
    p_user = p_user->pNext;
  
  }
  
  return NULL;

} /* getUser */


/*
 * getContact
 *
 * Given a contact list for a particular user in p_contact, finds the
 * contact entry with index contactIndex
 *
 * Return values
 *  NULL if not found, SipRegContactEntry* if present.
 *
 * Worst case complexity: o(n)
 */
const SipRegContactEntry* 
getContact(const SipRegContactEntry* p_contact, 
	   int contactIndex) 
{
  
  SipRegContactEntry* p_temp = (SipRegContactEntry*)p_contact;
  
  /* Algorithm:
   *   -Linear search
   */  

  while(p_temp) {
  
    if(p_temp->sipContactIndex == contactIndex)
      return (const SipRegContactEntry*) p_temp;
    
    p_temp = p_temp->pNext;
  
  }
 
  return NULL;

} /* getContact */


/*
 * getContactList
 *
 * Given a valid user (SipRegUserEntry*), locates and returns the head pointer
 * to a linked list of all contact entries.
 *
 * Important note:
 *  Free the elements after using the contactList by a call to 
 *  freeContactList
 *  
 *
 * Parameters 
 *  p_rows - out param, number of contact entries found.
 *
 */
SipRegContactEntry* 
getContactList(const SipRegUserEntry* p_user, int* p_rows) 
{
  
  int query_rows_ret, i;

  char *p = NULL;
  
  ResultTable query_result;
  
  SipRegContactEntry *p_head, *p_temp = NULL, *p_save= NULL;

  ContactCache* p_cachetemp;

  int toadd = 0;

  time_t expire;
  
  time_t t = time(0);
  
  if(!p_user || !p_rows) return NULL;

  /* 
   * first check if this user is in the ContactCache linked list.
   * If yes, and if last refresh is not past 300 seconds, return it.
   * 
   * This may lead to returning stale data occasionally, but we can
   * live with it.
   */
  p_cachetemp = p_contactCacheHead;
  
  while(p_cachetemp) {
    if(p_cachetemp->userEntry == (SipRegUserEntry*) p_user) 
      break;
    
    p_cachetemp = p_cachetemp->pNext;
    
  }
  
  if(p_cachetemp) {  /* a cache entry is there */
    
    time_t curTime;
    
    time(&curTime);
    
    if(curTime - p_cachetemp->timeSaved < 300) {
      *p_rows = p_cachetemp->numContacts;
      return p_cachetemp->contactListHead;
    }
    else {
      
      if(p_cachetemp->contactListHead)
	free(p_cachetemp->contactListHead);
      else 
	error("getContactList", "buggy code...listhead null");
      
      p_cachetemp->contactListHead = NULL;
      
      p_cachetemp->timeSaved = curTime;
      
      p_cachetemp->numContacts = 0;
      
      toadd = 1;
    }
  }
  else {
   
    p_cachetemp = (ContactCache*)calloc(1,sizeof(ContactCache));
    
    if(!p_cachetemp) {
      error("getContactList","calloc failed...");
      *p_rows = -1;
      return NULL;
    }

    time(&p_cachetemp->timeSaved);
    
    p_cachetemp->userEntry = (SipRegUserEntry*)p_user;
    p_cachetemp->pNext = p_contactCacheHead;
    
    p_contactCacheHead = p_cachetemp;
    
    toadd = 1;
  }
  
  /* fetch the contact records for this user.
   * Algorithm:
   *  -use a select query to get all contact records for this user.
   *  -construct a linked list containing only those entries which have
   *   not expired.
   *  -some query_result[i][j] elements may be null, so take care.
   *  -convert date variables from the database format to the needed
   *   format as defined in the SipRegistrar interface.
   *  -return the head pointer, as well as number of valid rows found
   */
  query_result = query("contacts", "user",
		       p_user->sipUserUri,
		       "contact,display_name,expires,action,q,last_modified",
                       &query_rows_ret);      
  
  *p_rows = 0;
  
  if(query_rows_ret <= 0) {
    *p_rows = query_rows_ret;
    return NULL;
  }
  
  p_head = p_temp = 
    (SipRegContactEntry*)calloc(query_rows_ret,sizeof(SipRegContactEntry));
  
  if(!p_temp) {
    error("getContactList","calloc failed...errno: %d\n",errno);
    return NULL;
  }
  
  /* Not expired. */
  debug("getContactList", 
	"constructing contact list...%d entries found for %s", 
	query_rows_ret,
	p_user->sipUserUri);
  
  
  for(i=0;i<query_rows_ret;i++) {
    
    if(query_result[i][2]) {
      
      expire = string2time(query_result[i][2]);

      if (expire <= t) { /* Expired entry. */
	
	debug("getContactList", 
	      "entry Expired %lu > %lu for %s\n",
	      expire, t, 
	      p_user->sipUserUri);
	
	continue;
      
      }
      
      /* not expired, copy it */
      strncpy(p_temp->sipContactExpiry,
	      query_result[i][2],
	      sizeof(p_temp->sipContactExpiry));
      
      /* we need 1999-12-12,02:02:02 while the db stores the , as a space */
      if((p = (char*)strchr(p_temp->sipContactExpiry,' ')))  
	*p=','; 
    }
    else {

      strncpy(p_temp->sipContactExpiry,
	      MAX_TIME_STR,
	      sizeof(p_temp->sipContactExpiry));
    }
    
    if(query_result[i][0]) 
      strncpy(p_temp->sipContactURI,
		  query_result[i][0],
		  sizeof(p_temp->sipContactURI));
    else
      continue;
      
    
    if(query_result[i][1])
      strncpy(p_temp->sipContactDisplayName,
	      query_result[i][1],
	      sizeof(p_temp->sipContactDisplayName));
    else
      {
	strncpy(p_temp->sipContactDisplayName,query_result[i][0],
		sizeof(p_temp->sipContactDisplayName));
      }
    
    if(!query_result[i][3] || !strcmp(query_result[i][3],"Proxy"))
      p_temp->sipContactAction = SIPSERVERACTION_PROXY;
    else
      p_temp->sipContactAction = SIPSERVERACTION_REDIRECT;
    
    
    if(query_result[i][4]) 
      strncpy(p_temp->sipContactPreference,
	      query_result[i][4],
	      sizeof(p_temp->sipContactPreference));
    else
      strcpy(p_temp->sipContactPreference,"0");

    if(query_result[i][5]) {
      
      struct tm tm;

      /* the database stores this as 19991231235959 */
      strptime(query_result[i][5],"%Y%m%d%H%M%S",&tm);
      
      p_temp->sipContactLastUpdated = mktime(&tm);
      
    }
    else 
      p_temp->sipContactLastUpdated = 0;
    
    p_temp->sipContactRetryAfter = 0;
    
    if(!p_save)  /* first element */
      p_save = p_temp;
    else {
      p_save->pNext = p_temp;
      p_save = p_temp;
    }
    
    p_temp->sipContactIndex = *p_rows+1;
    
    p_temp++;
   
    *p_rows = *p_rows + 1;
  }
  
  if(toadd) {
    p_cachetemp->contactListHead = p_head;
    
    p_cachetemp->numContacts = *p_rows;
  }
  return p_head;
  
} /* getContactList */


/*
 * frees all members of the contact list
 */
void 
freeContactList(SipRegContactEntry* p_rce) 
{

  /*
   * Nothing to do as we cache entries, and the cache refresh
   * module in getContactList frees appropriate entries
   */
  return;
  
} /* freeContactList */


#endif /* registrar_impl.c, USE_SNMP */

