#ifndef SIPENTITY_DS_H__
#define SIPENTITY_DS_H__

/**********************************************************************
 *
 *               libsipsnmp - snmp support for sip
 *
 *       sipentity_ds.h : Generic Management interface for SIP
 *
 *
 *      Copyright(c) 2000-2001 Columbia University ; all rights reserved
 *
 *           Initial version by: Sankaran N, February 2001.
 *
 **********************************************************************/
/*
 * This file and all included header files with _entity in them define a 
 * network management interface (mapped from the SIP MIB). The idea
 * is to decouple the sip code from the snmp code. 
 *
 * The motivations behind this type of architecture, the threading model,
 * extensibility aspects 
 * etc; can be found in the design doc present in this directory. 
*/

#define __SIPSERVER__

#ifdef __cplusplus
extern "C" { 
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <pthread.h>

  
/*
 * This file defines a set of data structures that actually define the network
 * management data. All sip-manageable entities need to include this file, and
 * implement the interfaces defined here.
 *
 * The meaning of **entities** is "an independent
 * process" that need to be fitted into the sip network management 
 * framework of a system.
 *
 * Examples are sipd and ua servers.
 *
*/


/* 
 *
 * typical sizes of uri,etc dont put this <256. 
 * SNMPv2 defines 256 as a typical size for octet strings.
 *
 */
#define STANDARDSTRLEN	256 
  

  /*
   * 
   * The snmp definition of boolean is 1 for true and 2 for false.
   * This differs from the standard 'C' conventions, and so
   * the following enum:
   *
   */  
  enum { TRU = 1, FAL = 2} REALITY;   
  

#define TRUTHVALUE u_int reality  
#define TRUTH reality


  /*
   * 
   * some helper definitions
   *
   */
#define RETTYPE_LONG	1
#define RETTYPE_STR	2
  

  /*
   * SNMP rowstatus textual convention
   */
#define SNMP_ROWSTATUS_ACTIVE         1
#define SNMP_ROWSTATUS_NOTINSERVICE   2
#define SNMP_ROWSTATUS_NOTREADY       3
#define SNMP_ROWSTATUS_CREATEANDGO    4
#define SNMP_ROWSTATUS_CREATEANDWAIT  5
#define SNMP_ROWSTATUS_DESTROY        6


  /* 
   *
   * SNMP Server Action textual convention
   *
   */
#define SIPSERVERACTION_REDIRECT      1
#define SIPSERVERACTION_PROXY         2
  

  /*
   * Data structures common to all sip entities such as registrar,
   * proxy etc;
   */
#include "sipentity_commonds.h"
  

  /*
   * portions of interface common to proxy and registrar servers
   * there should be another #def for ua server's, but i am not including
   * them rightnow.
   */

#ifdef __SIPSERVER__
#include "sipentity_serverds.h"
#include "sipentity_registrards.h"
#endif


  /* 
   * This is our main data structure.
   * Each sip_entity object represents a single network-manageable
   * sip entity in this system.
   *
   */
  typedef struct sip_entity { 
    
    
    pthread_mutex_t sipsnmp_lock;         /* common mutex for 
					   * SipCommonCfgTable,
					   * SipCommonSummStatsTable
					   */
    
    SipCommonCfgTable sipCommonCfgTable;  /* main cfg params */
    
    SipCommonSummStatsTable sipCommonSummStatsTable;
    
    pthread_mutex_t sipsnmp_translock;    /* mutex for p_transactList */ 
    
    int nSipCurrentTransactions;          /* currently active transactions */

    SipCommonTransactionMib *p_transactList;
    
    int sipNumUnsupportedUris;         /* number of 420 responses
					*  why it is here? ask me
					* (see design doc)
					*/
    
#ifdef __SIPSERVER__
    
    pthread_mutex_t sipsnmp_reglock;   /* mutex for sipRegistrar */
    
    SipServerEntry sipServer;         
    
    SipRegistrar sipRegistrar;
    
#endif
    
    /*
      #ifdef __UASERVER__ 
      
      conceptually its here, tho not implemented 
      
      SipUAS sipUAS;
    
      #endif
    */
    
  } SipEntity;
  
  
  /*
   *
   * Expose a one SipEntity pointer for each sip entity.
   * The intent is that the snmp code depends only on the
   * SipEntity pointer to get/set data, thereby making the
   * sip parts of nm code independent from snmp part.
   *
   * Rightnow, we have only one sip entity, therefore one ptr,
   * though i preferred something like a array of pointers with
   * one element.
   */
  extern const SipEntity* p_sipEntity; 

  
  /* 
   * 
   * This function should be called during init- of this module, to 
   * initialize the SipEntity* pointers, and their data. 
   * 
   */
  extern void initSipEntity();  
  
  
  /* 
   * 
   * Snmp-code
   *  -reading or writing p_sipEntity or its assoc'd objects
   * must be done by first acquiring the appropriate mutex, 
   * and then unlocking is done after it is over.
   *
   * 
   * Several locks are provided to implement fine grained locking, as it
   * improves performance. It is also semantically correct, as these
   * locks protect disjoint regions of code and data.
   *
   */ 


  /*
   * the typecast below is to get rid of the const 
   */  
#define LOCK()   pthread_mutex_lock((pthread_mutex_t*)&p_sipEntity->sipsnmp_lock)
#define UNLOCK() pthread_mutex_unlock((pthread_mutex_t*)&p_sipEntity->sipsnmp_lock) 

 
  /*
   *
   * Separate lock for registrar.
   *
   */
#define REG_LOCK()   pthread_mutex_lock((pthread_mutex_t*)&p_sipEntity->sipsnmp_reglock);
#define REG_UNLOCK() pthread_mutex_unlock((pthread_mutex_t*)&p_sipEntity->sipsnmp_reglock);

  
  /*
   *
   *separate lock for transact list as it may involve queries to the database.
   *
   */
#define TRANS_LOCK()   pthread_mutex_lock((pthread_mutex_t*)&p_sipEntity->sipsnmp_translock);
#define TRANS_UNLOCK() pthread_mutex_unlock((pthread_mutex_t*)&p_sipEntity->sipsnmp_translock);
  

  
#ifdef __cplusplus
};
#endif


#endif         /* sipentity_ds.h */

