/**********************************************************************
 *
 *                 libsipsnmp - snmp support provider.
 *
 *    sipsnmp_main.c - Main initialization routine for this subsystem.
 *
 *
 *      Copyright(c) 2000-2001 Columbia University ; all rights reserved
 *
 * Revision History 
 *           Additions to change Agentx Socket path, and graceful
 *           error handling: Sankaran, N, April 2001.
 *           Initial version by: Sankaran N, February 2001.
 *
 **********************************************************************/  

#ifdef USE_SNMP

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <errno.h>

#include <ucd-snmp/ucd-snmp-config.h>
#include <ucd-snmp/ucd-snmp-includes.h>
#include <ucd-snmp/ucd-snmp-agent-includes.h>
#include <ucd-snmp/snmp_logging.h>
#include <ucd-snmp/callback.h>

#include "error.h"
#include "request-sip.h"

#include "sipMIB.h"          /* the mib get/set handlers */
#include "sipsnmp_interface.h"
#include "sipentity_ds.h"    /* our main sip2snmp data structures */

#ifdef WIN32
#define sleep(x) Sleep((x) * 1000)
#endif

/*
 * Forward declarations.
 *
 */
static int dummy = 0;
static int full = 0;


extern int errno;/*
 *
 * callback function that gets called whenever the snmp library encounters an 
 * error. delegate this to our error handlers.
 *
 */
static int error_callback(int majorID, int minorID, void* serverarg, void* clientarg);


/*
 *
 * cleanup handler - call during exit, to deregister from snmpd.
 * if this is not done, the next init wont be successfull and we
 * will be forced to restart snmpd, and then do this.
 * 
 */
static void cleanup(void);

/* 
 *
 * Algorithm:
 *  Check the argument and determine whether we are running as subagent 
 *  or master. 
 *
 *  Initialize the following:
 *    UCD-SNMP Error handling through callbacks, so that our function 
 *    (error_callback) becomes UCD-SNMP's error handler. We will then delegate
 *    it to our error function.
 *
 *    Call init_agent to initialize agentx/snmp library.
 *
 *    Call our init code (application code). this is in initSipEntity, which 
 *    initializes our global data structures.
 *
 *    Call the mib init code which is in initSipMIB. This will install the
 *    get/set handlers in the snmp deamon's code as well as register the mib
 *    subtree mibII-SIP...
 *
 *    Then call init_agent (mandatory to bring the system up)
 *     and init_master_agent (this is optional).
 *
 *    Finally, enter the message processing loop. 
 *
 *   This function is a straight forward copy from the ucd-snmp tutorial, and
 *   also includes a patch for sending proper traps, thanks to dave shields.
 *
 */
void* sipsnmp_main (void* arg) {
  
  int agentx_subagent= SUB_AGENT,timeout;
  const char* agentx_socket_path=NULL;
  
  SnmpData *p_snmpData = (SnmpData*)arg;
  
  if(p_snmpData) {
    agentx_subagent = p_snmpData->sub_agent;
    agentx_socket_path = p_snmpData->agentx_socket_path;
    free(p_snmpData);
  }

  if(!p_snmpData || !agentx_socket_path) {
    error("sipsnmp_main","You must specify the AF_UNIX socket path. libsipsnmp not started\n");
    return NULL;
  }
  
  assert(agentx_subagent == 1 || agentx_subagent == 0);

  debug("sipsnmp_main","starting the sipsnmp agent...\n");

  debug("sipsnmp_main","contacting snmpd at %s\n",agentx_socket_path);

  init_callbacks();
  
  snmp_disable_stderrlog();
   
  snmp_register_callback(SNMP_CALLBACK_LIBRARY,
			 SNMP_CALLBACK_LOGGING,
			 error_callback,
			 (void*)&dummy);
  
  snmp_enable_calllog();  
 
  ds_set_boolean(DS_APPLICATION_ID, DS_AGENT_NO_ROOT_ACCESS, 1); 
  ds_set_string(DS_APPLICATION_ID,DS_AGENT_X_SOCKET,agentx_socket_path);

  if (agentx_subagent) 
    ds_set_boolean(DS_APPLICATION_ID, DS_AGENT_ROLE, 1);
  
  initSipEntity();

  /* initialize the agent library 
   */
  timeout = 5;  
  
  while(init_agent("sipMIB") != 0) {
    error("sipsnmp_main", "check whether the snmpd is running : will retry after some time");
    sleep(timeout);
    timeout *= 2;
    if(timeout > 1800) timeout  = 1800;
  }
  
  init_sipMIB();
  
  init_snmp("sipMIB");	
  
  atexit(cleanup); 
  
  if (!agentx_subagent)
    init_master_agent(161, NULL, NULL);  /* open port 161 (UDP:snmp) */
  
  debug("sipsnmp_main","started successfully...");
  
  full = 1;
  
  while(1) 
    agent_check_and_process(1); /* or 0 */
  
  
  snmp_shutdown("sipMIB");
  
} /* sipsnmp_main */


/*
 *
 * This callback function gets called whenever snmp subsystem encounters
 * an error.
 * 
 * we will delegate this to our error function.
 *
 */
static int 
error_callback(int majorID, int minorID, void* serverarg, void* clientarg) {
  
  struct snmp_log_message* p_slm = (struct snmp_log_message*)serverarg;
  
  
  /* we have p_slm->priority and p_slm->message here which need to be passed 
   *
   * forget error messages at startup, which means as long as full is false.
   */
  errno = 0;
  if(full && p_slm->msg)
    error("sipsnmp",(char*)p_slm->msg);
  
  return 0;

} /* error_callback */


/*
 * 
 * deregister our module from the snmpd code.
 *
 */
static void cleanup(void) {

  snmp_shutdown("sipMIB");

} /* cleanup */

#endif
