/**********************************************************************
 *
 *               libsipsnmp - snmp support provider.
 *
 *               sipsnmp_utils.c - Helper functions
 *
 *
 *      Copyright(c) 2000-2001 Columbia University ; all rights reserved
 *
 *           Initial version by: Sankaran N, February 2001.
 *
 **********************************************************************/

#ifdef USE_SNMP

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#ifndef HAVE_TIMEGM
#include "timegm.h"
#endif

#include "sipsnmp_utils.h"

/*
 *
 * time2string()
 * Convert the time_t to a string representation
 * YYYY-MM-DD HH:MM:SS (adapted from code in register.c)
 * Infinite time (MAX_TIME_T) corresponds to
 * 0000-00-00 00:00:00
 * All others are converted as per GMT 
 *
 */
int time2string(time_t time, char* resultBuf, int buflen)
{
  struct tm tm;
  
  assert(resultBuf && buflen > 20);
  
  if ( time == MAX_TIME_T ) {
    /* This time_t never expires so we store 
     * the keyword 0000-00-00 00:00:00 in db
     * which means the same 
     */
    strncpy(resultBuf,"9999-12-31 23:59:59",buflen);
  }
  else {
    /* Convert the time to db string */
    gmtime_r(&time, &tm);
    strftime(resultBuf,buflen,"%Y-%m-%d %H:%M:%S", &tm);
  }
  
  return 0;
} /* time2string */

    
/*
 *
 * Convert string representation of time to time_t.
 * string: YYYY-MM-DD HH:MM:SS
 * Infinite time (MAX_TIME_T) corresponds to
 * 0000-00-00 00:00:00
 * All others are converted as per GMT
 *
 */
time_t string2time(const char *string)
{
  struct tm tm;
  time_t expire;
  
  assert(string);
  
  if ( !strcmp(string, "0000-00-00 00:00:00") ) {
    /* If the string is unspecified; i.e. all zeros then
     * registration is always active 
     */
    expire = MAX_TIME_T; /* Max Possible */
  }
  else {
    
    strptime(string, "%Y-%m-%d %H:%M:%S", &tm);
    
    expire = timegm(&tm); /* convert GMT struct tm to time_t */ 
  }
  
  return expire;
  
} /* string2time */


/*
 *
 * returns the ip address (or domain name) of the primary
 * name server by querying the /etc/resolv.conf file. 
 * Todo: Win32 equivalent
 *
 * the first parameter is a buffer to contain the result,
 * the second param is the length of buffer (must be >24)
 *
 */
void get_primary_ns(char *buf, int len) 
{
  
  char local[512];
  char *p;
  FILE *fp = fopen("/etc/resolv.conf","r");
  
  assert(buf && len > 24);
  
  buf[0] = 0;
  if(fp) {
    fgets(local,512,fp);
    local[strlen(local)-1] = '\0';
    
    while(!feof(fp)) {
      p = (char*)strchr(local,' ');
      if(p) {
	*p = '\0';
	if(strcmp(local,"nameserver") == 0) {
	  strncpy(buf,p+1,len);
	  fclose(fp);
	  return;
	}
      }
      fgets(local,512,fp);
      local[strlen(local)] = '\0';
    }
    fclose(fp);
  }	
  return;	   

} /* get_primary_ns */

#endif
