#ifdef USE_SNMP
/*
 * snmp_sp.c
 *
 * Snmp support provider. Function handlers for various sip
 * -set functions, init function, and callback functions are 
 * implemented here.
 *
 * For a general discussion on the design philosophy see the design doc.
 */
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <limits.h>
#include <pthread.h>
#include <sys/time.h>

#include "dbapi.h"
#include "config.h"
#include "sysdep.h"
#include "request-sip.h"
#include "sipsnmp_interface.h"
#include "sipentity_ds.h"
#include "sipsnmp_utils.h"

/* 
 * this is the only instance of SipEntity for this system.
 * It is exposed to outside (snmp get/set handlers) through the 
 * p_sipEntity pointer.
 */
static SipEntity sip_entity;
const  SipEntity* p_sipEntity = &sip_entity;
static SipEntity *p_entity; /* = (SipEntity*)p_sipEntity; */
/* get rid of the const */

/*
 * This is the only port we listen to right now, by default
 */
static SipCommonPort sipCommonPort = { 5060, 
				       TRANSPORT_BOTH, 
				       SNMP_ROWSTATUS_ACTIVE, 
				       NULL 
};


/*
 * Forward declarations
 */
static void snmpProcessMessage
(struct request_t* p_req,int type,const sipsnmp_extra_data *p_extra_data);

static void processRegisterMessage(struct request_t *p_req, int type);

extern void processTransaction(struct request_t *p_req,
			       int type,
			       const sipsnmp_extra_data *p_extra_data);

/*
 * These functions try to acheive OO in 'C'. They are the handlers for the
 * various function pointers defined in the sip-snmp interface header files.
 *
 * The initSipEntity function installs them in our global DS
 */

typedef const char* pcchar;
#define PTR1(X,Y) static int X (SipCommonCfgInterface*, Y);

PTR1(setSipOrganization, pcchar);
PTR1(setSipUserLocationServerAddr, pcchar);

PTR1(setSipServiceAdminStatus,u_int);
PTR1(setSipInviteTimeOut,u_int) ;
PTR1(setSipHideMode,u_int);

PTR1(setSipRequestURIMatch,TRUTHVALUE);

static int addPortElement(const SipCommonPort *p_new);
static int delPortElement(SipCommonPort *p_which);
  
static int addMethodToMonitor(struct sip_common_summ_stats_table *p_this,
			      const SipCommonExtMethodStats *p_new);

static int delMethodToMonitor(struct sip_common_summ_stats_table *p_this,
			      int index);


static int addSCodeToMonitor(struct sip_common_summ_stats_table *p_this,
			     const SipCommonSCodeSpecial *p_new);

static int delSCodeToMonitor(struct sip_common_summ_stats_table *p_this,
			     int statusCode);



#ifdef __SIPSERVER__

#define PTR2(X,Y) static int X (SipServerCfgEntry*, Y);

PTR2(setSipServerHost,pcchar);

PTR2(setSipInformationTryingIntervalMsecs,u_int);
PTR2(setSipServerDfltAction,u_int); 
PTR2(setSipServerRespectUAAction,u_int);
PTR2(setSipInformationAlwaysSendTrying, TRUTHVALUE);

#define PTR3(X,Y) static int X(SipProxyCfgEntry*, Y);

PTR3(setSipPgpPrivateKey,const unsigned char*);
PTR3(setSipProxyAuthRealm,pcchar);

PTR3(setSipProxyAuthMethod,u_int);
PTR3(setSipProxyAuthPgpAlgorithm,u_int);
PTR3(setSipProxyNonceLifeTimeMsecs, u_int);

PTR3(setSipRequestMaxExpires,u_int);

#define PTR4(X) PTR3(X,TRUTHVALUE)
PTR4(setSipProxyStateful);
PTR4(setSipProxySendsCancel);
PTR4(setSipProxyForwardAll1xx); 
PTR4(setSipProxyRecursion);
PTR4(setSipProxyProvideAlternatives);
PTR4(setSipProxyRecordRoute);
PTR4(setSipProxyStateTimeout);
PTR4(setSipHideRespect);
 
static int addProxyAuthEntry(const SipRxProxyAuthEntry*);
static int delProxyAuthEntry(SipRxProxyAuthEntry*);

#define PTR5(X) static int X(SipRegCfgEntry*, u_int);
#define PTR6(X) static int X(SipRegCfgEntry*, TRUTHVALUE);

PTR5(setSipRegContactDfltExpiryDuration);
PTR5(setSipRegMaxContactExpiryDate);

PTR6(setSipRegAllowThirdParty);
PTR6(setSipRegRespHasContacts);


#endif

/*
static char* server_name = SERVER_NAME;
*/
static char organization[] = "Columbia University";

#ifdef WIN32
#define PTHREAD_THREADS_MAX 64
#else
#ifndef PTHREAD_THREADS_MAX
#define PTHREAD_THREADS_MAX _POSIX_THREAD_THREADS_MAX
#endif
#endif


/*
 * Main init function.
 *
 * Called by the initialization code during startup.
 *
 * Initializes global data.
 */
void initSipEntity() {

  SipCommonCfgTable *p_table;
  
  SipCommonSummStatsTable *p_stats;
  
  SipCommonCfgInterface *p_temp;
  
  /* Algorithm:
   * -Initialize global config parameters.
   * -Initialize function pointers by installing function handlers.
   * -Initialize appropriate mutexes.
   */

  SipCommonCfgInterface init = { 
    SIPVERSION,
    organization,setSipOrganization,
    OSTATUS_UP,
    ASTATUS_RTS,setSipServiceAdminStatus,
    0,0,
    config.timeout, setSipInviteTimeOut,
    HIDE_DEFAULT,setSipHideMode,
    PTHREAD_THREADS_MAX,
    COMPACT_SUPPORTED_TRUE,
    "ipv4:",setSipUserLocationServerAddr,
    REQUEST_URI_HOSTMATCH_FALSE, setSipRequestURIMatch
  };

  p_entity = (SipEntity*)p_sipEntity;  
  p_table = &p_entity->sipCommonCfgTable;
  p_stats = &p_entity->sipCommonSummStatsTable;
  p_temp = &p_table->sipCommonCfgInterface;

  memcpy(p_temp,(const void*)&init,sizeof(SipCommonCfgInterface));
  
  pthread_mutex_init(&p_entity->sipsnmp_lock,NULL);
  pthread_mutex_init(&p_entity->sipsnmp_reglock,NULL);
  pthread_mutex_init(&p_entity->sipsnmp_translock, NULL); 
  
  SnmpProcessMessage = snmpProcessMessage;

  {
    struct timeval t;
    char p[256];
    
    gettimeofday(&t,NULL);
    p_temp->sipServiceStartTime = 
      p_temp->sipServiceLastChange =t.tv_sec;
    
    get_primary_ns(p,sizeof(p));
    if(*p) 
      strcat(p_temp->sipUserLocationServerAddr,p);
  }
  
  p_temp->sipInviteTimeOut = 120;
  
  p_table->nPortTable = 1;
  
  p_table->pSipCommonPortTable = &sipCommonPort;
  
  {
    SipCommonURI *p_uri;
       
    p_uri = (SipCommonURI*)calloc(1,sizeof(SipCommonURI));
  
    p_uri->sipUriSupportedIndex = 1;
    strcpy(p_uri->sipUriSupported,"tel");
   
    p_table->nURITable = 1; 
    p_table->pSipCommonURITable = p_uri;
    
  }
  
  p_table->nFtrTable = 0;
  
  p_table->pSipCommonFtrTable = NULL;
 

  p_table->addPortElement =  addPortElement;
  p_table->delPortElement = delPortElement;
  
  p_stats->addMethodToMonitor = addMethodToMonitor;
  p_stats->delMethodToMonitor = delMethodToMonitor;
  
  p_stats->delSCodeToMonitor = delSCodeToMonitor;
  p_stats->addSCodeToMonitor = addSCodeToMonitor;
  
  
#ifdef __SIPSERVER__
  {
    
   
    SipProxyServerEntry *p_proxy = &p_entity->sipServer.sipProxyServerEntry;
    SipProxyCfgEntry *p_proxyCfg = &p_proxy->sipProxyCfgEntry;
   
    {
      SipServerCfgEntry *p_cfg = &p_entity->sipServer.sipServerCfgEntry;
      SipServerCfgEntry server_cfg =
      {
	config.ServerName, setSipServerHost,
	TRU, setSipInformationAlwaysSendTrying,
	100, setSipInformationTryingIntervalMsecs,
	"1.0",
	DFLT_PROXY, setSipServerDfltAction,
	TRU, setSipServerRespectUAAction
      };
      memcpy(p_cfg,(void*)&server_cfg,sizeof(server_cfg));
    }
    
    /* Proxy specific part starts here */
    
    {
      SipProxyCfgEntry proxy_cfg = {
	config.maxexpires,  setSipRequestMaxExpires,
	TRU, setSipProxyStateful,
	TRU, setSipProxySendsCancel,
	TRU, setSipProxyForwardAll1xx,
	TRU, setSipProxyRecursion,
	TRU, setSipProxyProvideAlternatives,
	TRU, setSipProxyRecordRoute,
	TRU, setSipProxyStateTimeout,
	PROXY_AUTH_BASIC, setSipProxyAuthMethod,
	PROXY_AUTH_ALGOMD5, setSipProxyAuthPgpAlgorithm,
        config.Realm, setSipProxyAuthRealm,
	config.NonceLifeTime,  setSipProxyNonceLifeTimeMsecs,
	0, setSipPgpPrivateKey,
	TRU, setSipHideRespect
	};
      
      if(!config.ProxyRecursion)
	proxy_cfg.sipProxyRecursion = FAL;
      
      memcpy(p_proxyCfg, (void*)&proxy_cfg, sizeof(proxy_cfg));
    }
    
    p_proxy->addProxyAuthEntry = addProxyAuthEntry;
    p_proxy->delProxyAuthEntry = delProxyAuthEntry;
          
    /* Proxy specific part ends here */
    
    /* registrar specific part starts here */
    {
      SipRegCfgEntry *p_reg =  &p_entity->sipRegistrar.sipRegCfgEntry;
      SipRegCfgEntry regcfg= { 
	TRU, 
	setSipRegAllowThirdParty,
	0, 
	setSipRegContactDfltExpiryDuration,
	0,
	setSipRegMaxContactExpiryDate,
	TRU, 
	setSipRegRespHasContacts,
	4294967294UL    /* so many users, maximum possible */
      };
      
      regcfg.sipRegContactDfltExpiryDuration = config.expires;
      regcfg.sipRegMaxContactExpiryDate = config.maxexpires;
      memcpy(p_reg,(void*)&regcfg,sizeof(regcfg));
    }
  }


#endif

  return;

} /* initSipEntity */


static int setSipOrganization(SipCommonCfgInterface* p_intf,
			      const char* p_buf) 
{
  
  return 0;

} /* setSipOrganization */


static int setSipServiceAdminStatus(SipCommonCfgInterface* p_intf,
				    u_int newStatus) 
{
  
  return 0;
  
} /* setSipServiceAdminStatus */


static int setSipInviteTimeOut(SipCommonCfgInterface* p_intf,
			       u_int newTimeOutSecs) 
{
  
  if(!p_intf) 
    return -1;
  
  if(newTimeOutSecs > 0) {
    p_intf->sipInviteTimeOut = newTimeOutSecs;
    config.timeout = p_intf->sipInviteTimeOut;
  }
  
  return 0;

} /* setSipInviteTimeOut */


static int setSipHideMode(SipCommonCfgInterface* p_intf,
			  u_int newMode) 
{
  
  assert(p_intf);
  p_intf->sipHideMode = newMode;
  return 0;

} /* setSipHideMode */


static int setSipUserLocationServerAddr(SipCommonCfgInterface* p_intf,
					const char* p_newAddr) 
{
  
  return 0;
  
} /* setSipHideUserLocationServerAddr */


static int setSipRequestURIMatch(SipCommonCfgInterface* p_intf,
				 TRUTHVALUE) 
{
  
  assert(p_intf);

  p_intf->sipRequestURIMatch = TRUTH;
  
  return 0;

} /* setSipRequestURIMatch */


static int addPortElement(const SipCommonPort *p_new)
{
  
  return -1; /* we support only one port */
  
} /* addPortElement */


static int delPortElement(SipCommonPort *p_which) 
{
  
  return -1; /* nothing to delete */
  
} /* delPortElement */


/*
 * Add a new status code to monitor and send traps for 
 */
static int addSCodeToMonitor(SipCommonSummStatsTable *p_this,
			     const SipCommonSCodeSpecial *p_new) 
{
  
  /* Add this at the end, set the index */
  
  SipCommonSCodeSpecial *p_save,*p_temp, *p_act;
  
  if(!p_this || !p_new) return -1;
  
  p_save = NULL;
  
  p_temp = p_this->pSCSCS_head;
  
  while(p_temp) {
    
    if(p_temp->sipCommonStatusCodeValue ==
       p_new->sipCommonStatusCodeValue)  
      
      return -1;   /* its already there */
    
    p_save = p_temp;
    
    p_temp = p_temp->pNext;
  }
  
  p_act = (SipCommonSCodeSpecial*)calloc(1,sizeof(SipCommonSCodeSpecial));
  
  memcpy(p_act,p_new,sizeof(SipCommonSCodeSpecial));
  
  p_this->nCommonSCodeSpecial++;
  
  if(p_save)  /* some elements are there */
    p_save->pNext = p_act;
  else 
    p_this->pSCSCS_head = p_act;
  
  return 0;
  
} /* addSCodeToMonitor */


/* delete a scode 
*/
static int delSCodeToMonitor(SipCommonSummStatsTable *p_this,
			     int statusCode) 
{
  
  SipCommonSCodeSpecial *p_temp, *p_del = NULL;
  
  if(!p_this || !p_this->pSCSCS_head)
    return -1;
    
  p_temp = p_this->pSCSCS_head;
  
  if(p_temp->sipCommonStatusCodeValue == statusCode) { /* first elem del */
    
    p_del = p_temp;
    
    p_this->pSCSCS_head = p_temp->pNext;
    
    goto delete_psave;
    
  }

  while(p_temp->pNext) {
    
    if(p_temp->pNext->sipCommonStatusCodeValue == statusCode) {
      p_del = p_temp->pNext;
      p_temp->pNext = p_temp->pNext->pNext;
      goto delete_psave;
    }
    
    p_temp = p_temp->pNext;
  }
  
  return -1;
  
 delete_psave:
  p_del->pNext = NULL;
  free(p_del);
  p_this->nCommonSCodeSpecial--;
  
  return 0;

} /* delSCodeToMonitor */

/*
 * Find a status code entry by traversing the linked list
 * pointed to by p_head
 */
const SipCommonSCodeSpecial* 
getSCodeEntry(const SipCommonSCodeSpecial *p_head,
	      int scodeValue) 
{
  
  const SipCommonSCodeSpecial *p_temp;
  
  if(!p_head) return NULL;
  
  p_temp = p_head;
  
  while(p_temp) {
    if(p_temp->sipCommonStatusCodeValue == scodeValue)
      return p_temp;
    
    p_temp = p_temp->pNext;
  }
  
  return NULL;
  
} /* getSCodeEntry */


/* 
 * Monitor an extension method (which is not a rfc2543 method)
 */
static int 
addMethodToMonitor(SipCommonSummStatsTable *p_this,
		   const SipCommonExtMethodStats *p_new) 
{
  
  /* Add this at the end, set the index */
  SipCommonExtMethodStats *p_save=NULL,*p_temp, *p_act;
  
  int index = 0;

  if(!p_this || !p_new) return -1;
  
  p_temp = p_this->pSCEMS_head;
  
  while(p_temp) {
    
    index = p_temp->sipStatsExtMethodIndex;
    
    if(!strcasecmp(p_temp->sipStatsExtMethodName,
		  p_new->sipStatsExtMethodName)) 
      return -1;
    
    p_save = p_temp;
    
    p_temp = p_temp->pNext;
  }
  
  p_act = (SipCommonExtMethodStats*)calloc(1,sizeof(SipCommonExtMethodStats));

  memcpy(p_act,p_new,sizeof(SipCommonExtMethodStats));
  
  p_act->sipStatsExtMethodIndex = index+1;

  p_this->nCommonExtMethodStats++;

  if(p_save)  /* some elements are there */
    p_save->pNext = p_act;
  else 
    p_this->pSCEMS_head = p_act;
  
  return 0;

} /* addMethodToMonitor */


/*
 * Delete a possibly active extension method
 */
static int 
delMethodToMonitor(SipCommonSummStatsTable *p_this,
		   int index) 
{
  
  SipCommonExtMethodStats *p_temp, *p_del = NULL;
  
  if(!p_this || !p_this->pSCEMS_head)
    return -1;
  
  p_temp = p_this->pSCEMS_head;
  
  if(p_temp->sipStatsExtMethodIndex == index) { /* first elem del */
    
    p_del = p_temp;
    
    p_this->pSCEMS_head = p_temp->pNext;
    
    goto delete_psave;
    
  }

  while(p_temp->pNext) {
    
    if(p_temp->pNext->sipStatsExtMethodIndex == index) {
      p_del = p_temp->pNext;
      p_temp->pNext = p_temp->pNext->pNext;
      goto delete_psave;
    }
    
    p_temp = p_temp->pNext;
  }
  
  return -1;
  
 delete_psave:
  p_del->pNext = NULL;
  free(p_del);
  p_this->nCommonExtMethodStats--;
  
  return 0;

} /* delMethodToMonitor */

/*
 * Given the name of an extension method, locate the SipCommonExtMethodStats
 * object from the linked list containing it. The head of the linked list
 * is at p_head
 */
const SipCommonExtMethodStats* 
getExtMethodEntryByName(const SipCommonExtMethodStats *p_head, 
			const char* methodName) 
{
  
  const SipCommonExtMethodStats *p_temp;
  
  if(!p_head) 
    return NULL;
  
  p_temp = p_head;
  
  while(p_temp) {
    
    if(!strcmp(p_temp->sipStatsExtMethodName,methodName))
      return p_temp;
    
    p_temp = p_temp->pNext;
  }
  
  return NULL;

} /* getExtMethodEntryByName */


/*
 * Given an extMethodIndex, return the corresponding object from the
 * linked list whose head is p_head
 */
const SipCommonExtMethodStats* 
getExtMethodEntry(
		  const SipCommonExtMethodStats *p_head,
		  int extMethodIndex)
{
  
  const SipCommonExtMethodStats *p_temp;

  if(!p_head) 
    return NULL;
  
  p_temp = p_head;
  
  while(p_temp) {
 
    if(p_temp->sipStatsExtMethodIndex == extMethodIndex)
      return p_temp;

    p_temp = p_temp->pNext;
  }
  
  return NULL;

} /* getExtMethodEntry */


#ifdef __SIPSERVER__

static int 
setSipServerHost(struct sip_server_cfg* p_cfg,
		 const char* newName) 
{
  
  return 0;
  
} /* setSipServerHost */


static int 
setSipInformationAlwaysSendTrying(struct sip_server_cfg* p_cfg, 
				  TRUTHVALUE) 
{

  p_cfg->sipInformationAlwaysSendTrying = TRUTH;
  
  return 0;

} /* setSipInformationAlwaysSendTrying */


static int 
setSipInformationTryingIntervalMsecs(struct sip_server_cfg* p_cfg,
				     u_int milliSecs) 
{

  if(milliSecs>=0 && milliSecs<=60000)
    p_cfg->sipInformationTryingIntervalMsecs = milliSecs;
  
  return 0;

} /* setSipInformationTryingIntervalMsecs */


static int 
setSipServerDfltAction(struct sip_server_cfg* p_cfg,
		       u_int action) 
{
  
  if(action == DFLT_REDIRECT || action == DFLT_PROXY)
    p_cfg->sipServerDfltAction = action;

  return 0;

} /* setSipServerDfltAction */


static int 
setSipServerRespectUAAction(struct sip_server_cfg* p_cfg,
			    TRUTHVALUE)
{
  
  p_cfg->sipServerRespectUAAction = TRUTH;
  
  return 0;

} /* setSipServerRespectUAAction */


static int 
setSipRequestMaxExpires(struct sip_proxy_cfg_entry* p_cfg,
			u_int secs) 
{
  
  p_cfg->sipRequestMaxExpires = secs;
  
  return 0;
 
} /* setSipRequestMaxExpires */


static int 
setSipProxyStateful(struct sip_proxy_cfg_entry* p_cfg,
		    TRUTHVALUE) 
{

  p_cfg->sipProxyStateful = TRUTH;
  
  return 0;

} /* setSipProxyStateful */


static int 
setSipProxySendsCancel(struct sip_proxy_cfg_entry* p_cfg,
		       TRUTHVALUE) 
{
  
  p_cfg->sipProxySendsCancel = TRUTH;
  
  return 0;

} /* setSipProxySendsCancel */


static int 
setSipProxyForwardAll1xx(struct sip_proxy_cfg_entry* p_cfg,
			 TRUTHVALUE) 
{
  
  p_cfg->sipProxyForwardAll1xx = TRUTH;
  
  return 0;

} /* setSipProxyForwardAll1xx */


static int 
setSipProxyRecursion(struct sip_proxy_cfg_entry* p_cfg,
		     TRUTHVALUE) 
{
  
  p_cfg->sipProxyRecursion = TRUTH;
  
  return 0;

} /* setSipProxyRecursion */

static int 
setSipProxyProvideAlternatives(struct sip_proxy_cfg_entry* p_cfg,
			       TRUTHVALUE) 
{

  p_cfg->sipProxyProvideAlternatives = TRUTH;
  
  return 0;
  
} /* setSipProxyProvideAlternatives */


static int 
setSipProxyRecordRoute(struct sip_proxy_cfg_entry* p_cfg,
		       TRUTHVALUE) 
{
    p_cfg->sipProxyRecordRoute = TRUTH;
    
    return 0;

} /* setSipProxyRecordRoute */


static int 
setSipProxyStateTimeout(struct sip_proxy_cfg_entry* p_cfg,
			u_int secs) 
{

    if(secs>=0 && secs<=255)
      p_cfg->sipProxyStateTimeout = secs;
    
    return 0;

} /* setSipProxyStateTimeout */


static int 
setSipProxyAuthMethod(struct sip_proxy_cfg_entry* p_cfg,
		      u_int value) 
{

  if(value >= PROXY_AUTH_NONE && value<=PROXY_AUTH_PGP) 
    p_cfg->sipProxyAuthMethod = value;
  
  return 0;

} /* setSipProxyAuthMethod */


static int 
setSipProxyAuthPgpAlgorithm(struct sip_proxy_cfg_entry* p_cfg,
			    u_int algo) 
{
  
  if(algo > PROXY_AUTH_ALGOOTHER && algo <= PROXY_AUTH_ALGOSHA1) 
    p_cfg->sipProxyAuthPgpAlgorithm = algo;
  
  return 0;

} /* setSipProxyAuthPgpAlgorithm */


static int 
setSipProxyAuthRealm(struct sip_proxy_cfg_entry* p_cfg,
		     const char* newRealm) 
{
  
  return 0;

} /* setSipProxyAuthRealm */


static int 
setSipProxyNonceLifeTimeMsecs(struct sip_proxy_cfg_entry* p_cfg,
			      u_int msecsValue) 
{
  
  if(p_cfg->sipProxyNonceLifeTimeMsecs <= 65535)
    p_cfg->sipProxyNonceLifeTimeMsecs = msecsValue;
  
  return 0;
  
} /* setSipProxyNonceLifeTimeMsecs */


static int 
setSipPgpPrivateKey(struct sip_proxy_cfg_entry* p_cfg,
		    const unsigned char* key) 
{

  return 0;

} /* setSipPgpPrivateKey */


static int 
setSipHideRespect(struct sip_proxy_cfg_entry* p_cfg,
		  TRUTHVALUE) 
{

  p_cfg->sipHideRespect = TRUTH;
  
  return 0;

} /* setSipHideRespect */


static int 
addProxyAuthEntry(const SipRxProxyAuthEntry* p_new) 
{
  
  return -1;
  
} /* addProxyAuthEntry */

static int 
delProxyAuthEntry(SipRxProxyAuthEntry* p_ae) 
{
  
  return -1;
 
} /* delProxyAuthEntry */


static int 
setSipRegAllowThirdParty(SipRegCfgEntry* p_cfg,
			 TRUTHVALUE)
{
  
  if(!p_cfg) return -1;
  
  p_cfg->sipRegAllowThirdParty = TRUTH;
  
  if(p_cfg->sipRegAllowThirdParty == TRU)
    config.ThirdPartyReg = 1;
  else
    config.ThirdPartyReg = 0;
  
  return 0;
  
} /* setSipRegAllowThirdParty */

static int 
setSipRegContactDfltExpiryDuration(SipRegCfgEntry* p_cfg,
				   u_int timeSecs) 
{
  
  if(!p_cfg) return -1;
  
  p_cfg->sipRegContactDfltExpiryDuration = timeSecs;
  
  if(timeSecs > 0) 
    config.expires = timeSecs;
  
  return 0;

} /* setSipRegContactDfltExpiryDuration */


static int 
setSipRegMaxContactExpiryDate(SipRegCfgEntry* p_cfg,
					 u_int durationSecs) 
{
  if(!p_cfg) return -1;
  
  p_cfg->sipRegMaxContactExpiryDate = durationSecs;
  
  if(durationSecs > 0) 
    config.maxexpires = durationSecs;
  
  return 0;
  
} /* setSipRegMaxContactExpiryDate */


static int setSipRegRespHasContacts(SipRegCfgEntry* p_cfg,
				    TRUTHVALUE) 
{

  if(!p_cfg) return -1;
  
  p_cfg->sipRegRespHasContacts = TRUTH;
  
  if(p_cfg->sipRegRespHasContacts == TRU)
    config.RegRespContacts = 1;
  else
    config.RegRespContacts = 0;

  return 0;

} /* setSipRegRespHasContacts */


#endif

/*
 * Given a status code say 400, and a direction say in or out,
 * increment the appropriate class counter. Then see if a trap
 * needs to be sent for this status code. If yes, then send a trap
 * and return. 
 */
static void 
incClassCounter(struct request_t *r, int statusCode, int dir) 
{
  
  SipCommonSummStatsTable *p_summ;
  SipCommonStatusCodeStats *p_stats;
  SipCommonSCodeSpecial *p_temp;
  
  int sCode;

  if(dir < 0 || dir>=STATS_MAX) {
    error("incClassCounter","buggy code Dir %d StatusCode %d...\n",
	  dir,statusCode);
    return;
  }
  
  sCode = statusCode/100 - 1;
  
  p_summ = &p_entity->sipCommonSummStatsTable;
  p_stats = &p_summ->sipCommonStatusCodeStats;
  
  if(sCode>=0 && sCode < CLASS_MAX)
    p_stats->sipStatsInfoClass[sCode][dir]++;
  
  /* now search the special status code list, and check for the
   * particular item
   */  
 
  p_temp = (SipCommonSCodeSpecial*)
    getSCodeEntry(p_summ->pSCSCS_head,statusCode);

  if(p_temp && 
     p_temp->sipCommonStatusCodeStatus == SNMP_ROWSTATUS_ACTIVE)
    {
      
      if(dir == STATS_IN) 
	p_temp->sipCommonStatusCodeIns++;
      else
	p_temp->sipCommonStatusCodeOuts++;
      
      
      if(p_temp->sipCommonStatusCodeSendNotif == TRU) 
	{
	  
	  /* construct a SipCommonNotif struct which is the
	   * desired data structure for traps, and then call our
	   * trap function
	   */
	  SipCommonNotif notif;
	  
	  if(r->to.uri.user && r->to.uri.host) 
	  {
	    strcpy(notif.sipStatusCodeNotifTo,r->to.uri.user);
	    strcat(notif.sipStatusCodeNotifTo,"@");
	    strcat(notif.sipStatusCodeNotifTo,r->to.uri.host);
	 }
	  else
	    strcpy(notif.sipStatusCodeNotifTo, " ");
	  
	  if(r->from.uri.user && r->from.uri.host) 
	  {
	    strcpy(notif.sipStatusCodeNotifFrom,r->from.uri.user);
	    strcat(notif.sipStatusCodeNotifFrom,"@");
	    strcat(notif.sipStatusCodeNotifFrom,r->from.uri.host);
	  }
	  else
	    strcpy(notif.sipStatusCodeNotifFrom, " ");
	  
	  notif.sipStatusCodeNotifValue = p_temp->sipCommonStatusCodeValue;
	  	  
	  if(r->callID) 
	    strncpy(notif.sipStatusCodeNotifCallId,
		    r->callID,
		    sizeof(notif.sipStatusCodeNotifCallId));
	  else
	    strcpy(notif.sipStatusCodeNotifCallId,"");
	  
	  notif.sipStatusCodeNotifCSeq = r->cseq.seq;
	  
	  send_trap(&notif);
	  
	}
    }

  return;

} /* incClassCounter */


/*
 * Given a method name (rfc2543 method or an extension method), and a 
 * direction flag such as in or out, increment the appropriate counters.
 */ 
static void 
incMessCounter(struct request_t *r, char* methodName, int dir) 
{
  
  int messType = -1;
  
  SipCommonSummStatsTable *p_summ;
  SipCommonMethodStats *p_cfg;
  SipCommonExtMethodStats *p_temp;

  p_summ = &p_entity->sipCommonSummStatsTable;
  p_cfg  = &p_summ->sipCommonMethodStats;
  
  if(!methodName || strlen(methodName) <= 0 || dir<0 || dir>=STATS_MAX) {
    debug("incMessCounter", "buggy code - call has bad params");
    return;
  }
  
  if(!strcmp(methodName,"INVITE"))
    messType = STATS_INVITE;      
  if(!strcmp(methodName,"REGISTER"))
    messType = STATS_REGISTER;      
  if(!strcmp(methodName,"BYE"))
    messType = STATS_BYE;      
  if(!strcmp(methodName,"OPTIONS"))
    messType = STATS_OPTIONS;      
  if(!strcmp(methodName,"ACK"))
    messType = STATS_ACK;      
  if(!strcmp(methodName,"CANCEL"))
    messType = STATS_INVITE;      
  
  if(messType != -1) {
    p_cfg->sipStats[messType][dir]++;
    return;
  }

  /* if messType == -1 case, search the ext method list 
   * if this entry is present, increment appropriately
   */
       
  p_temp = (SipCommonExtMethodStats*)
    getExtMethodEntryByName(p_summ->pSCEMS_head,
			    methodName);
  
  if(p_temp && p_temp->sipStatsExtMethodRowStatus == SNMP_ROWSTATUS_ACTIVE)
    {
      if(dir == STATS_IN)
	p_temp->sipStatsExtMethodIns++;
      else
	p_temp->sipStatsExtMethodOuts++;
    }
  
  return;

} /* incMethodCounter */


/*
 * snmpProcessMessage
 *
 * Entry point function. 
 * Depending on the 'type' of the call, perform appropriate actions. 
 *
 * See sipsnmp_interface.h for interface semantics
 */
void 
snmpProcessMessage(struct request_t *p_req,
		   int type, 
		   const sipsnmp_extra_data *p_extra_data) 
{

  request_common_t *r = (request_common_t*) p_req; 
 
  parser_t *p;
  
  if(!p_req || !r) {
    error("SnmpProcessMessage", "bad call: arg1 null");
    return;
  }
  
  p = &r->p;

  /* ordering of the type == checks is important 
   */
  
  if(type == SNMP_TYPE_ADD_TRANS || type == SNMP_TYPE_DEL_TRANS) 
    {
      processTransaction(p_req,type,p_extra_data);
      return;
    }
  
  if(p->method && !strcmp(p->method,"REGISTER")) 
    {
      processRegisterMessage(p_req,type);  
      return;
    }
  
  LOCK();
  {
    char* method;
    
    SipCommonSummaryStats *p_summ = 
      &p_entity->sipCommonSummStatsTable.sipCommonSummaryStats;
       
    if(type == SNMP_TYPE_REQ_IN) {
      p_summ->sipSummaryInRequests++;
      method = HTTP_Method(&p_req->p);
      incMessCounter(p_req,method,STATS_IN);  
    }
    
    if(type == SNMP_TYPE_REQ_OUT) {
      p_summ->sipSummaryOutRequests++;
      if(p_extra_data) {
	method = (char*)p_extra_data->request_type;
	incMessCounter(p_req,method,STATS_OUT);
      }
    } 

    if(type == SNMP_TYPE_RES_IN) {
      p_summ->sipSummaryInResponses++;
      if(p_extra_data) {
	method = (char*)p_extra_data->request_type;
	incMessCounter(p_req,method,STATS_IN);
	incClassCounter(p_req,r->p.status, STATS_IN);
      }      
    } 
    
    if(type == SNMP_TYPE_RES_OUT) {
     
      /*
       * A response is being sent. Set the time counters properly.
      */
      p_summ->sipSummaryOutResponses++;
      
      method = HTTP_Method(&p_req->p);
      
      incMessCounter(p_req,method,STATS_IN);
      
      incClassCounter(p_req,r->status, STATS_OUT);
      
      if(r->status == 400) { /* check for unsupported uri scheme */
	
	uri_t uri = HTTP_URI(&r->p);
	
	/* sipd supports the 'sip' and 'tel' URI schemes. */
	
	if (uri.scheme == NULL ||
	    (strcmp(uri.scheme, "sip") != 0 
	     && strcmp(uri.scheme, "tel") != 0)) 
	  {
	    SipEntity *p_temp = (SipEntity*) p_sipEntity;
	    p_temp->sipNumUnsupportedUris++;
	  }
      }
      
      if(r->status >= 200) 
	p_summ->sipSummaryTotalTransactions++;
    }
  }
  
  UNLOCK();
 
  if(type == SNMP_TYPE_RES_OUT) 
  {     /* Note:
	 * i want to avoid doing TRANS_LOCK inside LOCK and hence one more if
	 * after UNLOCK.  The general idea is all accesses to the transaction state
	 * data structures are done inside TRANS_XXLOCK, all accesses to the 
	 * registration state data structures inside REG_XXLOCK and others inside
	 * LOCK.  
	*/
	TRANS_LOCK();
	updateTransactionTimes((void*)p_req);
	TRANS_UNLOCK();
  } 
  return; 
  

} /* snmpProcessMessage */



/*
 * processRegisterMessage
 *
 * Manipulate counters, datum for register requests/responses.
 *
 * Parameters
 *  -request_t for the request/response in question.
 *  -type (see sipsnmp_impl.h)
 *
 * Algorithm:
 *  -increment all message counters based on message type.
 *  -we also need to increment the per-user authentication failure count.
 *   use the fine grained lock, to avoid making other threads lock for a long
 *   time.
*/
static void 
processRegisterMessage(struct request_t *p_req, int type) 
{
  
  request_common_t *r = (request_common_t *)p_req;
  parser_t *p = &r->p;
  
  SipCommonSummStatsTable *p_statsTable = 
    &p_entity->sipCommonSummStatsTable; 
  
  SipCommonSummaryStats* p_stats = 
    &p_statsTable->sipCommonSummaryStats;
  
  SipRegistrar* p_reg = &p_entity->sipRegistrar;
  
  
  /* these types have no business here */
  if(type==SNMP_TYPE_RES_IN || type == SNMP_TYPE_REQ_OUT) {
    debug("processRegisterMessage","bad type in call : buggy code");
    return;
  }
  
  /* add one incoming, one outgoing common counters */
  
  LOCK();
  
  if(type == SNMP_TYPE_REQ_IN) {
    
    incMessCounter(p_req,p->method,STATS_IN);
    
    p_stats->sipSummaryInRequests++;
    
    goto lock_done; /* we want to avoid a big else part */
  
  }

  /* this part is for SNMP_TYPE_OUT */
  incMessCounter(p_req,p->method,STATS_OUT);
  
  
  /* in register we send only responses with code */
  
  /* this is for 100, 201, 401, and other errors */
  
  incClassCounter(p_req,r->status,STATS_OUT);
  
  p_stats->sipSummaryOutResponses++;
 
  /* non 1xx indicates a transaction - which means 
   * for register add the totalTrans counter
   */
  if(r->status<200) 
    goto lock_done;
  
  p_stats->sipSummaryTotalTransactions++;
  
  if(r->status == 200) {
    p_reg->sipRegAcceptedRegistrations++;
    goto lock_done;
  } 
  
  /* increment the sipAuthFailure count, rejected reg count */
  p_reg->sipRegRejectedRegistrations++;
  
  /* here, we need the finer lock, so remove the coarse lock */
  UNLOCK();
 
  TRANS_LOCK();
  updateTransactionTimes((void*)p_req);
  TRANS_UNLOCK();
 
  REG_LOCK();
  
  /* this part is for error responses, r->status > 200 */
  {
    char from[256];

    SipRegUserEntry *p_ulist = p_entity->sipRegistrar.p_userList;
    
    strcpy(from,p_req->from.uri.user);
    strcat(from,"@");
    strcat(from,p_req->from.uri.host);
    
    
    if(!p_ulist) { /* try loading from the db */
      getNumRegisteredUsers((SipRegistrar*)&p_sipEntity->sipRegistrar);  
      
      if(!(p_ulist = p_sipEntity->sipRegistrar.p_userList)) 
	goto reglock_done;
    }
    /* increment the sipUserAuthFailures field for this user*/
    
    while(p_ulist) {
      if(!strcmp(p_ulist->sipUserUri,from)) {
	debug("processRegisterMessage",
	      "incrementing auth failure for user %s\n",from);
	
	p_ulist->sipUserAuthFailures++;
	
	p_ulist->isValid = 1;
	
	break;
      }
      
      p_ulist = p_ulist->pNext;
    }
  }
  
 reglock_done:
  
  REG_UNLOCK();
  return;
  
 lock_done:
  
  UNLOCK();
  return;
  
} /* processRegisterMessage */

#endif /* snmp_sp.c, USE_SNMP */
