#ifdef USE_SNMP
/*
 * This file implements helper functions for per-transaction tables,
 * and related counters
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include "request-sip.h"
#include "sipentity_ds.h"
#include "sipsnmp_interface.h"
#include "sipsnmp_utils.h"

static int next_trans = 0;
static int numHoles = 0;
static int delayedDelete = 0;
static time_t lastListRefresh;
extern int errno;
/*
 * Given a request_t struct, convert it into SipCommonTransactionMib.
 */
static int 
convertRequestToMib(const struct request_t *p_req,
		    SipCommonTransactionMib* result) 
{
  
  if(!p_req || !result) {
    debug("convertTransToMIB","bad parameters in call");
    return -1;
  }  
  
  memset(result,0,sizeof(SipCommonTransactionMib));
  
  result->transaction_id = ++next_trans;

  result->p_transact = (void*)p_req;
  
  result->isValid = 1;
  
  if(p_req->to.uri.user && p_req->to.uri.host) 
  {
    strcpy(result->sipTransTo,p_req->to.uri.user);
    strcat(result->sipTransTo,"@");
    strcat(result->sipTransTo,p_req->to.uri.host);
  }
  if(p_req->from.uri.user && p_req->from.uri.host) 
  {
    strcpy(result->sipTransFrom,p_req->from.uri.user);
    strcat(result->sipTransFrom,"@");
    strcat(result->sipTransFrom,p_req->from.uri.host);
  }
  if(p_req->callID) 
    strncpy(result->sipTransCallId,
	    p_req->callID,
	    sizeof(result->sipTransCallId));
  
  if(p_req->response_body.content_type) {
    strncpy(result->sipTransCalledPartyContentType,
	    p_req->response_body.content_type,
	    sizeof(result->sipTransCalledPartyContentType));
  }
  if(p_req->request_body.content_type) {
    strncpy(result->sipTransCallingPartyContentType,
	    p_req->request_body.content_type,
	    sizeof(result->sipTransCallingPartyContentType));
  }
  
  result->sipTransCSeq = p_req->cseq.seq;

  if(time2string(p_req->expires,
		 result->sipTransExpiry,
		 sizeof(result->sipTransExpiry) ) < 0)
    {
      debug("convertTransToMIB", "time2String error");
      result->sipTransExpiry[0] = '\0';
    }
  
  return 0;
  
} /* convertRequestToMib */


/*
 * Given a SipCommonTransactionMib object, searches the request_t
 * struct, gets the request_t object, and return the current 
 * transaction state, along with the number of outstanding branches
 * still pending 
 */
u_int 
getTransactionState(const SipCommonTransactionMib *p_this,
		    u_int* numOutStandingBranches) 
{

  struct request_t *p_req;
  
  u_int sipTransState = TRANS_NULL;

  if(!p_this || !numOutStandingBranches) {
    
    error("getTransactionState","bad call : params null");
    
    return TRANS_NULL;
  }
  
  if(!p_this->isValid) {
    
    error("getTransactionState","bad call : invalid transaction object");
    
    return TRANS_NULL;
  }

  if(p_this->timedeleted) {
    
    *numOutStandingBranches = 0;
    
    return TRANS_COMPLETED;
  }

  if(!p_this->p_transact) {
    error("getTransactionState",
	  "buggy code: p_transact pointer null, unexpected");
    
    *numOutStandingBranches = 0;
    
    return TRANS_NULL;
  }
  
  p_req = (struct request_t*)p_this->p_transact;
  
  if(!p_req->status) {
    sipTransState = TRANS_INITIAL;
    goto udone;
  }

  if(p_req->status == 100) {
    sipTransState = TRANS_CALLPROCEEDING;
    goto udone;
  }
  
  if(p_req->status > 100 && p_req->status < 200) {
    sipTransState = TRANS_RINGING;
    goto udone;
  }
  
  if(p_req->status == 200) {
    
    if(p_req->sent_final)
      sipTransState = TRANS_CONFIRMED;
    else
      sipTransState = TRANS_SUCCESS;

    goto udone;
  }
  
  if(p_req->status >= 300) {
    sipTransState = TRANS_FAILURE;
    goto udone; 
  }

  if(p_req->acked)
    sipTransState = TRANS_COMPLETED;
  
 udone:

  *numOutStandingBranches = 
    p_req->branches_outstanding;
  
  return sipTransState;

} /* getTransactionState */


void updateTransactionTimes(void *p_r)
{

 	SipCommonTransactionMib *p_head;
       
	 struct request_t *p_req = (struct request_t*)p_r;
 	
	if(!p_req)
		return;

	/* do a linear search and see whether we have the p_req ptr
	*/
	p_head = p_sipEntity->p_transactList;

	while(p_head)
	{
		if(p_head->p_transact == p_r)
		{

			struct timeval curtime;
			gettimeofday(&curtime,NULL);
			if(p_req->status == 100)
				p_head->time_100sent = curtime;
			else 
			{
				if(p_req->status == 180)
					p_head->time_180sent = curtime;
			}	
			if(p_req->status >= 200)
			{
				p_head->time_finalsent = curtime;
				p_head->finalCode = p_req->status;
			}
			return;
		}
		p_head = p_head->pNext;
	}

} /* updateTransactionTimes */



/*
 * processTransaction
 *
 * -Add a new transaction or delete one.
 */
void processTransaction(struct request_t *p_req,
			int type,
			const sipsnmp_extra_data *p_extra_data)
{
  
  static int once = 0;
  
  SipCommonTransactionMib *p_head;
  
  if(!once) {
    time(&lastListRefresh);
    once = 1;
  }
  
  /* Algorithm:
   * -If a new transaction happens, add it to our global link list.
   * -If a transaction is deleted, mark it as deleted, and remove 
   * the external dependencies part. keep the local data alive.
   *
   * -Periodically, cleanup the deleted items. 
   * 
   * -While cleaning up the deleted items, remove the data, but keep the
   * object in the transaction linked list. This will create "holes"
   * in the linked list, where new objects can be inserted.
   */


  TRANS_LOCK();
  
  if(type == SNMP_TYPE_ADD_TRANS) {
   
    if(numHoles > 0) { /* o(N) insertion */

      p_head = p_sipEntity->p_transactList;
      
      while(p_head) {
	
	if(p_head->isValid == 0) {

	  convertRequestToMib(p_req,p_head);

	  gettimeofday(&p_head->time_start,NULL);
	  
	  numHoles--;
	  
	  goto transact_add;
	}
	
	p_head = p_head->pNext;
      
      }

      /* control should never come here */
      debug("processTransaction",
	    "buggy code...control reaches unexp. places");
      goto err_done;
    }
    else 
      {
	/* calloc and add at beginning */
	SipEntity *p_entity = (SipEntity*)p_sipEntity;

	p_head = (SipCommonTransactionMib*)
	  malloc(sizeof(SipCommonTransactionMib));
	
	convertRequestToMib(p_req,p_head);

	gettimeofday(&p_head->time_start, NULL);
	
	/* add it to head o(1) insertion, but we have to be careful */
	
	if(!p_entity->p_transactList) {
	  p_entity->p_transactList = p_head;
	  goto transact_add;
	}
	
	/* this place is the non empty list case */
	p_head->pNext = (SipCommonTransactionMib*)  
	  p_entity->p_transactList;
	
	p_entity->p_transactList = p_head;
	
      }
    
    goto transact_add;
  }
  
  if(type == SNMP_TYPE_DEL_TRANS) {
    
    /* Algorithm:
       -Find the entry, and set the time_t flag as well as invalidate
       the request_t pointer.
       -Set the state to completed.
    */
    
    p_head = p_sipEntity->p_transactList;
    
    while(p_head) {
    
      if(p_head->p_transact == (void*)p_req) {
	
	p_head->p_transact = NULL;
	
	time(&p_head->timedeleted);

	if(1) { /* for time being set this to 1, later we will make config */
	  
	  struct timeval curTime;
	  int hundred = 0, h80 = 0, firstFinal = 0;

	  gettimeofday(&curTime,NULL);
	
	  if(p_head->time_finalsent.tv_sec == 0)
	  {
		p_head->time_finalsent = curTime;
		p_head->finalCode = p_req->status;
	  }

	  if(p_head->time_100sent.tv_sec)
       	  {
		hundred = ((p_head->time_100sent.tv_sec - p_head->time_start.tv_sec) * 1000) + 
			  ((p_head->time_100sent.tv_usec - p_head->time_start.tv_usec) / 1000);
	  }
	  if(p_head->time_180sent.tv_sec)
	  {
		h80 = ((p_head->time_180sent.tv_sec - p_head->time_start.tv_sec) * 1000) + 
			  ((p_head->time_180sent.tv_usec - p_head->time_start.tv_usec) / 1000);
	  }

		firstFinal = ((p_head->time_finalsent.tv_sec - p_head->time_start.tv_sec) * 1000) + 
			  ((p_head->time_finalsent.tv_usec - p_head->time_start.tv_usec) / 1000);
			   
	  errno = 0;
	  debug("SnmpTransact", "[CallId %s - 100 %d 180 %d %d %d ms]",
						p_head->sipTransCallId,
						hundred,
						h80,
						p_head->finalCode,
						firstFinal);
	}
	
	++delayedDelete;
	
	goto transact_del;
      
      }

      p_head = p_head->pNext;
    }

    goto transact_del;
   
    /*
    if(!p_head) {
       XXX: this is not buggy as we have more dels than adds due to the
       * fact that not all request_t ptrs go to RequestAdd in libsip,
       * but all of them go to RequestFree
      error("processTransaction","Buggy code: an expected element not found in the transact list");
      
      goto err_done ;
    }
   */
 
  }
  
 transact_add:
  
  {
    SipEntity* p_entity = (SipEntity*)p_sipEntity;

    p_entity->nSipCurrentTransactions++;
  }
  TRANS_UNLOCK();

  return;
  
 transact_del:

  /* iterate through the linked list and remove deleted items every
   * five minutes - this is not strict five minutes, but suits us fine.
   * the worst case time for which an entry will live here is 5*2 = 10 
   * minutes because
   * -we cleanup every 5 minutes.
   * -and during cleanup we remove elements which have been in the completed
   * state for atleast 5 minutes -> roughly 10 minutes worst case.
   */
  {
    time_t curTime;
    
    time(&curTime);
    
    if((curTime - lastListRefresh > 300) && delayedDelete > 0) {
     
      p_head = (SipCommonTransactionMib*)p_sipEntity->p_transactList;
      
      while(p_head) {
	
	if(p_head->isValid == 1 &&
	   p_head->timedeleted!=0 && 
	   curTime - p_head->timedeleted > 300) {
	
	  p_head->isValid = 0;
	
	  numHoles++;
	  
	  delayedDelete--;
	 
	  { /* the const pointer is a pain sometimes */
	    
	    SipEntity *p_entity = (SipEntity*)p_sipEntity;
	    p_entity->nSipCurrentTransactions--;
	  
	  }
	
	}
	
	p_head = p_head->pNext;
      }
      
      lastListRefresh = curTime;
    }
  }

 err_done:

  TRANS_UNLOCK();  

  return;
  
} /* processTransaction */


/*
 * Given a transaction_id, return the element from the linked list,
 * which has that id.
 */
const SipCommonTransactionMib* 
getTransactionById(const SipCommonTransactionMib *p_head, 
		   u_int transaction_id) 
{
 
  SipCommonTransactionMib *p_temp = 
    (SipCommonTransactionMib*)p_head;
  
  while(p_temp && p_temp->transaction_id != transaction_id)
    p_temp = p_temp->pNext;
  
  return p_temp;

} /* getTransactionById */
  
/*
 * Given an index (logical number starting 1), return the element at 
 * index'th position from p_head
 */
const SipCommonTransactionMib* 
getTransactionByIndex(const SipCommonTransactionMib *p_head, 
		      int index) 
{
  
  int cur = 0; /* index starts at 1 */
  
  SipCommonTransactionMib *p_temp = (SipCommonTransactionMib*)p_head;
  
  while(p_temp) {
    
    if(p_temp->isValid == 1) {  
      cur++;
      
      if(cur == index)
	return p_temp;
    }

    p_temp = p_temp->pNext;
  
  }

  return p_temp;

} /* getTransactionByIndex */


#endif   /* trans_impl.c */
