/*
 * RTSP Authentication Module
 *
 * Author: Chen Zhou and Henning Schulzrinne
 * Department of Computer Science
 * Columbia University
 * Last updated March 30th, 1998
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include <sys/time.h>

#include "global.h"
#include "authen.h"

/* local function prototypes */
static int read_root_config(char *, realm_t *);
static realm_t *load_rtaccess(char *, char *);
static int isin_domain(char *, char *);
static int isin_group(char *, char *);

/*
 * Basically, the server creates a pool of realms for authentication.
 * When it first starts it should create a "root" realm from the 
 * server configuration file. Then as each access occurs, specific
 * ".rtaccess" files gets read. Future access checkes against
 * this pool of realms, as well as its parent until it hits the root
 * 
 */

/* 
 * read in the server config file and set up the global root realm
 *
 * returns: the root realm structure
 */
realm_t *Authen_Init(char *conf)
{
    realm_t *s;
    int c;
    struct timeval t;
    
    s = (realm_t *) malloc(sizeof(realm_t));
    if (s == NULL)
    {
        printf("not enough memory to allocate root realm\n");
        return NULL;
    }
        
    /* pull stuff off the server configuration file */
    c = read_root_config(conf, s);

    s->parent = NULL;   /* root has no parent */
    s->next = NULL;     /* nothing yet */
    gettimeofday(&t, NULL);
    s->time = t.tv_sec;
    s->path = "/";
    

    return s;
}

/* 
 * given a realm, display its content 
 */
void Authen_Show_Realm(realm_t *r)
{
    printf("time = %d\n", r->time);
    printf("name = %s\n", r->title);
    printf("path = %s\n", r->path);
    printf("rootpath = %s\n", r->rootpath);
    printf("override = %d\n", r->override);
    printf("hostorder = %d\n", r->hostorder);
    printf("userorder = %d\n", r->userorder);
    printf("grouporder = %d\n", r->grouporder);
}

/* 
 * given a Request, and the realm pool.
 * check the realm pool, if the request realm does
 * not exist. fetch the corresponding .rtaccess file
 * (a .htaccess -> .rtaccess converter will be availalbe soon)
 *
 * returns: the updated realm pool
 *
 */
realm_t *Authen_ReadAccess(request_t *req, realm_t *r)
{
    realm_t *a;
    realm_t *end;
    char *path;
    char *cp;
    char here[MAX_PATH];
    char rtpath[MAX_PATH];
    int pl;
    int i;
    int tmp;
    
    strcpy(here, "");         /* init */
    strcpy(rtpath, ""); 
    tmp = 0;
    end = r;
    
    path = strdup(req->path); /* copy in the path for faster access */
    printf("%s\n", path);
    
    /* acquire the number of levels */
    pl = 0;
    for (i = 0; i < strlen(path); i++)
        if (path[i] == '/')
            pl++;
    printf("levels: %d\n", pl);
    
    if (pl <= 1)    /* root level file, no need additonal .rtaccess */
        return r;
    
    /* does the server allow override */
    if ( r->override == 0)
    {
        printf("server does not allow access override\n");
        return r;
    }
    else if ( r->override == 1)
    {
        printf("server permit access override\n");
        
        /* the the first level */
        strcpy(here, "/");
        cp = strtok(path, "/");
        strcat(here, cp);
        strcat(here, "/");
        sprintf(rtpath, "%s.rtaccess", here);
        a = load_rtaccess(rtpath, r->rootpath);
        if (a != NULL)
        {
            a->parent = end;
            a->next = NULL;
            end->next = a;
            end = a;
            tmp++;
        } /* do nothing if can't load, move on to next level */
    
        for (i = 0; i < (pl - 2); i ++)
        {
            cp = strtok(NULL, "/");
            strcat(here, cp);
            strcat(here, "/");
            strcpy(rtpath, ""); /* just to be safe */
            sprintf(rtpath, "%s.rtaccess", here);
            a = load_rtaccess(rtpath, r->rootpath);
            if (a != NULL)      /* link em up */
            {
                a->parent = end;
                a->next = NULL;
                end->next = a;
                end = a;
                tmp++;
            }
        }

    } /* end override test */

    printf("tmp = %d\n", tmp);
    return r;
}

/* 
 * given the path of an .rtaccess file read in the
 * parameters, and return a realm ready to be added
 * to the realm pool
 */
static realm_t *load_rtaccess(char *rt, char *droot)
{
    realm_t *v;
    char path[MAX_PATH];
    FILE *fp;
    char *tmp;
    
    sprintf(path, "%s%s", droot, rt);
    printf("loading access control file = %s\n", path);
    
    fp = fopen(path, "r");
    if (fp == NULL)
    {
        printf("no such .rtaccess file \n");
        return NULL;
    }
    else
    {
        v = (realm_t *)malloc(sizeof(realm_t));
        if (v == NULL)
        {
            printf("out of memory. unable load %s\n", path);
            return NULL;
        }
        else    /* fill it up */
        {
            int l = strlen(droot);      /* root path */
            tmp = (char *)malloc(l*sizeof(char));
            strncpy(tmp, droot, l);
            v->rootpath = tmp;
            
                
            /* more to come */
            return v;
        }
    }
    
    

}

/*
 * given Request, realm pool. check to see if the 
 * request is authorized
 * 
 * return code:
 * 1 = permission granted
 * 0 = access denied
 */
int Authen_CheckPoint(request_t *req, realm_t *r, int where)
{
    
    
    /* 1) check to see if host is authorized */
    
    
    /* 2) check to see if user is authorized */
    /* 3) check to see if user belong to the authroized group */
    return 0;
}


/* 
 * given an ip, and a rule
 * return 1: if the two match or the host falls with in the domain
 *        2: everything else     
 *              
 */
static int isin_domain(char *rule, char *ip)
{
    int i;
    return 2;    
}

/*
 * given a user, and a group
 * return 1: if user is in group
 *        0: if user is NOT in group
 */
static int isin_group(char *group, char *user)
{
    int i;
    return 2;
}

/* 
 * read the authen.conf file, testing only function
 * return  1: success
 *         0: failed
 */
static int read_root_config(char *fname, realm_t *fill)
{
    FILE *fp;
    char VAR[100];
    char VAL[100];
    char *tmp;
    
    fp = fopen(fname, "r");
    while (fscanf(fp, "%s %s\n", VAR, VAL) != EOF )
    {
        if ( strcmp(VAR, "RootRealmName") == 0 )
        {
            int l = strlen(VAL);
            tmp = (char *)malloc(l*sizeof(char));
            strncpy(tmp, VAL, l);
            fill->title = tmp;
 
        }
        else if ( strcmp(VAR, "RootPath") == 0 )
        {
            int l = strlen(VAL);
            tmp = (char *)malloc(l*sizeof(char));           
            strncpy(tmp, VAL, l);
            fill->rootpath = tmp;
        }
        else if ( strcmp(VAR, "RootOverRide") == 0 )
        {
            if ( strcmp(VAL, "yes") == 0 )
            {
                fill->override = 1;
            }
            else if ( strcmp(VAL, "no") == 0 )
            {
                fill->override = 0;
            }
            else
            {
                printf("unknown override parameters\n");
            }
        }
        else if ( strcmp(VAR, "RootHostOrder") == 0 )
        {
            if ( strcmp(VAL, "deny,allow") == 0 )
            {
                fill->hostorder = 1;
            }
            else if ( strcmp(VAL, "allow,deny") == 0 )
            {
                fill->hostorder = 2;
            }
            else
            {
                printf("unknown host order parameters\n");
            }
        }
        else if ( strcmp(VAR, "RootUserOrder") == 0 )
        {
            if ( strcmp(VAL, "deny,allow") == 0 )
            {
                fill->userorder = 1;
            }
            else if ( strcmp(VAL, "allow,deny") == 0 )
            {
                fill->userorder = 2;
            }
            else
            {
                printf("unknown user order parameters\n");
            }
        }
        else if ( strcmp(VAR, "RootGroupOrder") == 0 )
        {
            if ( strcmp(VAL, "deny,allow") == 0 )
            {
                fill->grouporder = 1;
            }
            else if ( strcmp(VAL, "allow,deny") == 0 )
            {
                fill->grouporder = 2;
            }
            else
            {
                printf("unknown group order parameters\n");
            }
        }
        else
        {
            printf("%s unknow Variable\n", VAR);
        }
    }
    
    fclose(fp);
    return 1;
}
