/*
* Describe()
* Handles DESCRIBE request for a particular URL.  Return status; if
* description file available, set 'fd' to file descriptor.  If not, set
* 'fd' to -1 and set 'descr' to dynamically allocated description.  Set
* 'type' to point to content type.
*/

#include <stdio.h>
#include <pthread.h>
#include <stdlib.h> /* malloc(), free() */
#include <unistd.h> /* close() */
#include <string.h> /* strcasecmp() */
#include "rtspd.h"
#include "mime.h"
#include "snd.h"
#include "wav.h"
#include "qt.h"
#include "mpv.h"
#include "mp3.h"
#include "rtpfile.h"

static struct {
  char *type;
  void *(*open)(char *path, char *track, int rw); 
  char *(*describe)(void *, char *type);
  int   (*close)(void *);
} filehandlers[] = {
  {"audio/basic", SndOpen, SndDescribe, SndClose},
  {"audio/x-wav", WavOpen, WavDescribe, WavClose},
  {"video/quicktime",QTOpen, QTDescribe, QTClose},
  {"video/mpeg",MpvOpen, MpvDescribe, MpvClose},
  {"audio/mpeg",Mp3Open, Mp3Describe, Mp3Close},
  {"application/x-rtp",RtpOpen, RtpDescribe, RtpClose},
/*
  {"audio/x-pn-realaudio", RealMediaOpen, RealMediaDescribe, RealMediaClose},
  {"video/x-pn-realvideo", RealMediaOpen, RealMediaDescribe, RealMediaClose},
*/
  {NULL,          0,       0,           0}
};


/*
* Find a description for file 'path' of MIME type 'type', where 'path'
* is MIME type 'mt' (index into filehandlers).  Set 'fd' to file and
* 'descr' to description, if no file.
*/
static int DescribeTry(request_t *r, char *path, char *type, 
                       int mt, int *fd)
{
  char *descr = NULL;
  int size = -1;
  *fd = MIME_Open(path, type, &size);

  /* If no pre-stored file, compute on the fly. */
  if (*fd < 0 && filehandlers[mt].open) {
    void *v;
    if ((v = (*filehandlers[mt].open)(path, NULL, 0))) {
      descr = (*filehandlers[mt].describe)(v, type);
      (*filehandlers[mt].close)(v);
    }
  }

  if ( descr == NULL && *fd < 0) {
    ResponseSetStatus(r, 404, "File not found");
    return 404; /* Not found */
  }
  else if ( *fd >= 0 && descr == NULL ) {
    ResponseSetStatus(r, 200, NULL);
    /* Set the content-type */
    HeaderReplaceOrAppend(&r->headers, "Content-Type", type);
    HeaderDelete(&r->headers, "Content-Length");
    HeaderPrintf(&r->headers, 0, "Content-Length", "%d", size > 0 ?
                 size : 0);
    return 200; /* File is part of *fd */
  }
  else if ( descr != NULL ) {
    DString body;
    DString u_dstr;

    DStringInit(&body); 
    DStringInit(&u_dstr);

    URI_SPrint(&r->p.URI, &u_dstr);
    DStringPrintf(&body,"%sa=control:%s\r\n",descr,DStringValue(&u_dstr));
    
    HeaderReplaceOrAppend(&r->headers, "Content-Base", DStringValue(&u_dstr));

    DStringFree(&u_dstr);
    
    ResponseSetStatus(r, 200, NULL);
    ResponseSetBody(r, DStringValue(&body), type, -1);
    
    free(descr);
    DStringFree(&body);
    return 200; /* OK */
  }
  else {
    ResponseSetStatus(r, 500, "Illegal case in DescribeTry");
    return 500;
  }

} /* DescribeTry */

int Describe(request_t *r, int *fd)
{
  parser_header_t *h;
  parser_element_t *e;
  static char *accept[] = {"application/sdp", NULL};
  char *mediatype;   /* audio/basic, etc. */
  int i;
  int mt;
  int status = 200; /* Success */

  snprintf(r->path, sizeof(r->path)-1, "%s/%s", 
           config.DocumentRoot, r->p.URI.path);

  /* Find MIME type for media file. */
  if (!(mediatype = MIME_FilenameToType(r->path))) {
    ResponseSetStatus(r, 415, "Can not map file name to MIME type" );
    return 415;
  }

  /* Extract the appropriate filehandlers for the
   * MIME type */
  for (mt = 0; filehandlers[mt].open; mt++) {
    if (!strcasecmp(mediatype, filehandlers[mt].type)) {
      break;
    }
  }

  /*
   * Loop through acceptable types. Ignore q= parameter for now and
   * assume that types are listed in order of preference.
   */
  *fd = -1;
  if ((h = RFC822_Header(r->p.headers, "Accept"))) {
    for (e = h->elements; e; e = e->next) {
      status = DescribeTry(r, r->path, e->value, mt, fd);
      if ( status < 300 ) {
        break;
      }
      /* Else continue for next accepted mime type */
    }
  }
  /* No Accept: header; thus, use default type list. */
  else {
    for (i = 0; accept[i]; i++) {
      status = DescribeTry(r, r->path, accept[i], mt, fd);
      if ( status < 300 ) {
        break;
      }
      /* Else next element in accepted list */
    }
  }

  if ( status != 200 ) {
    debug("Describe", "Status is %d\n", status);
  }

  return status;

} /* Describe */
