/*
* Create list of MIME-type to extension mappings.
*/
#include <limits.h>
#include <stdio.h>    /* fopen() */
#include <stdlib.h>   /* malloc() */
#include <fcntl.h>    /* O_RDONLY */
#include <string.h>   /* strrchr() */
#include <errno.h>    /* errno */
#include <sys/types.h>
#include <sys/stat.h>  /* fstat() */

#include "dict.h"
#include "ht.h"
#include "strdupn.h"
#include "error.h"

#include "config.h"

#include "sysdep.h"

#define MIME_MAX 500
#define DEFAULT_DIR "/usr/local/lib/"

typedef struct mentry_t {
  char *key;
  void *value;
} mentry_t;

typedef struct msuffix_t {
  char *suffix;
  struct msuffix_t *next;
} msuffix_t;

static dict_h mime_suffix2type, mime_type2suffix;
static int dict_errno;
/* XXX thread-safety? */

static ht_val hash_mentry(dict_obj o);
static ht_val hash_key(dict_key k);
static int mentry_compare(dict_obj obj1, dict_obj obj2);
static int mentry_key_compare(dict_key key, dict_obj obj);

/*
* MIME_Init()
* Initialize MIME mapping tables. Two tables: one mapping from suffix to
* MIME type, the other from MIME type to one or more suffixes. Return 0 on
* success, -1 on failure.
*/
int MIME_Init(char *root, char *filename)
{
  FILE *f;                  /* file to read MIME definitions from */
  int argc = MIME_MAX;      /* number of words in configuration line */
  char *argv[MIME_MAX];
  mentry_t *m;              /* MIME table entry */
  msuffix_t *s, *s2;        /* MIME suffix entry */
  char full_path[PATH_MAX]; /* path name of MIME file */
  int i;
  struct ht_args argsp = {0, 0, &hash_mentry, &hash_key};

  sprintf(full_path, "%s/%s", root, filename);
  if ((f = fopen(full_path, "r")) == NULL) {
    sprintf(full_path, "%s/%s", DEFAULT_DIR, "mime.types");
    if ((f = fopen(full_path, "r")) == NULL) return -1;
  }

  /* Suffix table.  Each suffix has exactly one type. */
  mime_suffix2type = ht_create(&mentry_compare, &mentry_key_compare,
                               DICT_UNIQUE_KEYS|DICT_UNORDERED,
                               &dict_errno, &argsp);
  if (mime_suffix2type == NULL) {
    error("MIME_Init", "Couldn't create suffix table: error %d",
          dict_errno);
    return -1;
  }

  /* Type table.  Each type has a list of suffixes. */
  mime_type2suffix = ht_create(&mentry_compare, &mentry_key_compare,
                               DICT_UNIQUE_KEYS|DICT_UNORDERED,
                               &dict_errno, &argsp);
  if (mime_type2suffix == NULL) {
    error("MIME_Init", "Couldn't create type table: error %d",
          dict_errno);
    return -1;
  }

  while (!feof(f)) {
    argc = MIME_MAX;
    if (!ConfigRead(f, &argc, argv)) break;
    s = NULL;
    /* Create entry for each suffix */
    for (i = 1; i < argc; i++) {
      /* Suffix table. */
      m = (mentry_t *) calloc(1, sizeof(mentry_t));
      if (m == NULL) {
        error("MIME_Init", "Couldn't allocate mentry");
        return -1;
      }

      strassign(&(m->key), argv[i]);
      strassign((char **)&(m->value), argv[0]);
      if (ht_insert(mime_suffix2type, (dict_obj) m) != DICT_OK) {
        errno = 0;
        error("MIME_Init",
              "Error inserting type in suffix table: %d", dict_errno);
        continue;
      }

      s2 = (msuffix_t *) calloc(1, sizeof(msuffix_t));
      if (s2 == NULL) {
        error("MIME_Init", "Couldn't allocate msuffix");
        return -1;
      }

      strassign(&(s2->suffix), argv[i]);
      s2->next = s;
      s = s2;
    }
    m = (mentry_t *) calloc(1, sizeof(mentry_t));
    if (m == NULL) {
      error("MIME_Init", "Couldn't allocate mentry");
      return -1;
    }
    strassign(&(m->key), argv[0]);
    m->value = (void *) s;

    if (ht_insert(mime_type2suffix, (dict_obj) m) != DICT_OK) {
      errno = 0;
      error("MIME_Init",
            "Error inserting type in type table: %d", dict_errno);
      continue;
    }
  }

  fclose(f);
  return 0;
} /* MIME_Init */


/*
* MIME_FilenameToType()
* Map file name suffix to MIME type. Return NULL if no match.
*/
char *MIME_FilenameToType(char *filename)
{
  dict_obj o;
  char *suffix;

  suffix = strrchr(filename, '.');
  if (suffix) {
    o = ht_search(mime_suffix2type, suffix+1);
    if (o != NULL) {
      return (char *)((mentry_t *)o)->value;
    }
  }
  return NULL;
} /* MIME_FilenameToType */


/*
* MIME_Open()
* Given a MIME type and a path, find the first file constructed from the
* path and the suffix entry for that MIME type.  Return file descriptor or
* NULL, if no match or if file cannot be read.
*/
int MIME_Open(char *path, char *type, int *size)
{
  dict_obj o;
  int fd = -1;
  char full_path[PATH_MAX];
  msuffix_t *s;
  struct stat statbuf;

  debug("MIME_Open", "called for type %s: %s\n", type, path);
  
  o = ht_search(mime_type2suffix, type);
  if (o != NULL) {
    for (s = (msuffix_t *)((mentry_t *)o)->value; s != NULL && fd < 0;
         s = s->next) {
      sprintf(full_path, "%s.%s", path, s->suffix);
      fd = open(full_path, O_RDONLY);
      if ( fd < 0 ) { /* Added to allow files names as aaa.sdp */
        char *suff;
        strcpy(full_path, path);
        suff = strrchr(full_path, '.');
        if ( suff && *suff && strcmp(suff, s->suffix)==0) {
          fd = open(path, O_RDONLY);
          if ( fd >= 0 ) {
           debug("MIME_Open", "Successfully opened\n");
          }
        }
      }
    }
  }

  if ( size ) {
    *size = 0;
  }

  if ( fd >= 0 && fstat(fd, &statbuf) == 0 ) {
    if ( size ) {
      *size = statbuf.st_size;
    }
  }

  return fd;
} /* MIME_Open */


/*
 * hash_mentry()
 * hash a mentry (passed as a dict_obj, which is really a void *).  This
 * really means hash its key. 
 */
static ht_val hash_mentry(dict_obj o)
{
  mentry_t *m = (mentry_t *) o;

  return hash_key((dict_key) m->key);
} /* hash_message */


/*
 * hash_key()
 * hash a mentry key (passed as a dict_key, which is really a void *).
 * We do the standard technique for string hashes, lifted from Tcl: multiply
 * by 9 and add.
 */
static ht_val hash_key(dict_key k) {
  int result = 0;
  char *c = (char *) k;

  if (c == NULL) return 0;

  while (*c != '\0') {
    result += (result << 3) + *c;
    c++;
  }
  return result;
} /* hash_key */


/*
 * msg_compare()
 * Compare the keys of two messages for order -- return -1, 0, 1.
 * Since they're strings, just use strcasecmp on the keys.
 */
static int mentry_compare(dict_obj obj1, dict_obj obj2)
{
  mentry_t *m1, *m2;
  m1 = (mentry_t *) obj1;
  m2 = (mentry_t *) obj2;

  return strcasecmp((char *) m1->key, (char *) m2->key);
} /* msg_compare */


/*
 * msg_key_compare()
 * Compare a key with a message for order -- return -1, 0, 1.
 * Since they're strings, just use strcasecmp on the keys.
 */
static int mentry_key_compare(dict_key key, dict_obj obj) 
{
  char *k = (char *) key;
  mentry_t *m = (mentry_t *) obj;

  return strcasecmp(k, (char *) m->key);
} /* msg_key_compare */
