/* 
 * Copyright (c) 2000 by Columbia University
 *
 * file: mparse.h
 * description: mpeg video mparser
 * author: Stephen Chou
 * email: schou@cs.columbia.edu
 */
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
//#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <sys/stat.h>
//#include <bits/types.h>
//#include <sys/select.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <stdio.h>
#include <errno.h>
#include <signal.h>

#ifndef MPARSE_H_
#define MPARSE_H_

#define PATTERN_SIZE 100
#define PATH_SIZE 256
#define MAX_FRAMES 54000 // half an hour video at 30fps;
#define MAX_SLICES 100

/* Start codes. */

#define READ_BLOCK_SIZE 512
#define MAX_TIMER_ADJUST (100 * SPEEDUP_INV_SCALE)

#define SEQ_END_CODE 0x000001b7
#define SEQ_START_CODE 0x000001b3
#define GOP_START_CODE 0x000001b8
#define PICTURE_START_CODE 0x00000100
#define SLICE_MIN_START_CODE 0x00000101
#define SLICE_MAX_START_CODE 0x000001af
#define EXT_START_CODE 0x000001b5
#define USER_START_CODE 0x000001b2

// mpeg video exteded header
typedef struct rtpmpgext {
#ifdef BIG_ENDIAN
    unsigned mbz : 5; // not used
    unsigned t : 1;   // mpeg2 header extension present
    unsigned tr : 10; // temporal reference
    unsigned an : 1;  // active n bit
    unsigned n : 1;   // new picture header
    unsigned s : 1;   // sequence header present
    unsigned b : 1;   // beginning of slice
    unsigned e : 1;   // end of slice
    unsigned p : 3;   // picture type
    unsigned fbv : 1; // full pel backward vector
    unsigned bfc : 3; // backward f code
    unsigned ffv : 1; // full pel forward vector
    unsigned ffc : 3; // forward f code

#define SetTR(x,v) (x)->tr = (v);
#define GetTR(x) (x)->tr
#else
    unsigned tr0 : 2; // temporal reference
    unsigned t : 1;   // mpeg2 header extension present
    unsigned mbz : 5; // not used

    unsigned tr1 : 8; // temporal reference

    unsigned p : 3;   // picture type
    unsigned e : 1;   // end of slice
    unsigned b : 1;   // beginning of slice
    unsigned s : 1;   // sequence header present
    unsigned n : 1;   // new picture header
    unsigned an : 1;  // active n bit

    unsigned ffc : 3; // forward f code
    unsigned ffv : 1; // full pel forward vector
    unsigned bfc : 3; // backward f code
    unsigned fbv : 1; // full pel backward vector

#define SetTR(x,v) (x)->tr0 = (v) >> 8; (x)->tr1 = (v) & 0xff;
#define GetTR(x)   ((x)->tr0 << 8) | ((x)->tr1 & 0xff)
#endif
} rtpmpgext_t;


// MPEG Sequence header in BIG-Endian sequence
typedef struct SeqHeaderInfo {
    unsigned startcode; // 32-bits
    unsigned h_size : 12; // width
    unsigned v_size : 12; // height
    unsigned aspect_ratio : 4; // aspect ratio
    unsigned bit_rate : 18; // bit rate
    unsigned markerbit : 1; // ?
    unsigned vbv_buffersize : 10; // VBV Buffersize
} SeqHeaderInfo_t;

// MPEG GOP header in BIG-Endian sequence
typedef struct GOPHeaderInfo {
    unsigned startcode;
    unsigned drop_flag : 1; // drop flag
    unsigned tc_hours : 5; // hour component of time code
    unsigned tc_minutes : 6; // minute component of time code
    unsigned markerbit : 1;
    unsigned tc_seconds : 6; // second coponent of time code
    unsigned tc_pictures : 6; // pcture count
    unsigned cgbl_flag : 2; // close gop and broken link flags
} GOPHeaderInfo_t;

// MPEG PicHeader header in BIG-Endian sequence
typedef struct PicHeaderInfo {
    unsigned startcode;
    unsigned tempref : 10; // temporal reference
    unsigned pictype : 3; // picture type
    unsigned vbv_delay : 16; // vbv buffer delay value
    unsigned pel_fv : 1; // pel forw vector
    unsigned forw_rcode : 3; // forward rcode
    unsigned pel_bv : 1; // pel backward vector
    unsigned back_rcode : 3; // back_rcode
} PicHeaderInfo_t;

// system header size and offset for .Info file
struct SystemHeader {
  long offset;
  int size;
};

// GOP  header info for .info
struct GopTable {
  int systemHeader;
  long offset;
  int headerSize;
  int size;
  int totalFrames;
  int previousFrames;
  long firstIoffset;
};

// Frame table .info
struct FrameTable {
  char type;
  unsigned short size;
};

// number of slices
struct SliceOffsetTable {
    int numslices;
    long sliceoffset[MAX_SLICES];
};


// Mpeg video class
class Mpeg1video {
private:
    SystemHeader* systemHeader; 
    FrameTable* frameTable; 
    GopTable* gopTable;
    SliceOffsetTable *sliceTable;

    static double pictureRateTable[]; // picture rate conversion table

    char videoFile[PATH_SIZE]; // mpeg file name
    FILE *fp;                  // mpeg file pointer

    int curFrame; // current frame number
    int curGroup; // current gop number

    SeqHeaderInfo_t curSeqHeader; // current seq header
    GOPHeaderInfo_t curGopHeader; // current gop header
    PicHeaderInfo_t curPicHeader; // current picture header

    long curFrameStart;  // file offset of beginning of current frame
    long curFrameOffset; // file offset of data to read next
    long curFrameSize;   // current frame size

    bool sendEndSeqHeader; // frame reset, send end of sequence
    bool sendSeqHeader;    // frame reset, resend sequence header
    bool endSeq;           // end of a sequence

    // some staticss of mpeg frame
    int fileSize;
    int maxS, maxG, maxI, maxP, maxB;
    int minS, minG;
    int minI, minP, minB;
    int numS, numG, numF, numI, numP, numB;
    int averageFrameSize;
    int horizontalSize;
    int verticalSize;
    int pelAspectRatio;
    int pictureRate;
    int vbvBufferSize;

    int patternSize;
    char pattern[PATTERN_SIZE];

    int ReadInfoFromFile(void);
    void WriteInfoToFile(void);
    int init_MPEG1_video_file(void);
    void ParseSeqHead(SeqHeaderInfo_t* ci, 
		      long offset, 
		      int size, 
		      unsigned char* buf);
    void ParseGOPHead(GOPHeaderInfo_t* gop, 
		      long offset, 
		      int size, 
		      unsigned char* buf);
    long ParsePictureHead(PicHeaderInfo_t* info, 
			  long offset, 
			  int size, 
			  unsigned char* buf,
			  int remsize);
    void ReadFrame(unsigned char* buf, int size);

    void write_bytes(int s, char * data, int len) {
	write(s, data, len);
    }
    void read_bytes(int s, char * buf, int size) {
	read(s, buf, size);
    }

    void read_int(int s, int * data) {
	read_bytes(s, (char *)data, sizeof(int));
	*data = ntohl(*data);
    }
    void read_short(int s, short * data) {
	read_bytes(s, (char *)data, sizeof(short));
	*data = ntohs(*data);
    }
    void read_byte(int s, char * data) {
	read_bytes(s, data, 1);
    }

    void write_short(int s, short data) {
	short val = htons(data);
	write_bytes(s, (char *)&val, sizeof(short));
    }

    void write_byte(int s, char data) {
	write_bytes(s, &data, 1);
    }

    void write_int(int s, int data) {
	int val = htonl(data);
	write_bytes(s, (char *)&val, sizeof(int));
    }

    void NextFrame();
    long ReadFrame(unsigned char*buf, int size,
		   long offset, int remainsize);


    void CreateSliceTable();
    void DeleteSliceTable();

public:
    Mpeg1video();
    void Open(char* streamname);
    void Close();
    double FrameRate() { return pictureRateTable[pictureRate -1]; }
    int TotalFrames() { return numF; }

    bool ReadNext(unsigned char* buffer, int* size, 
		  int* gop, int* frame);
    void SeekFrame(int frame);
    void PrintFrames();
};

#endif // MPARSE_H_
