/* 
 * Copyright (c) 2000 by Columbia University
 *
 * file: mpv.cpp
 * description: mpeg video interface module
 * author: Stephen Chou
 * email: schou@cs.columbia.edu
 */


#include <stdio.h>    /* fseek() */
#include <fcntl.h>    /* O_ */
#include <stdlib.h>   /* calloc(), free() */
#include <string.h>   /* strncmp() */
#include <math.h>     /* floor() */
#include <unistd.h> /*fstat*/
#include <sys/stat.h> /*fstat*/

extern "C" {
#include "types.h"
#include "dstring.h"  /* Dstring..() */
#include "error.h"    /* debug() */
#include "getusername.h" /* getusername() */
}
#include "mparse.h"
#include "mpv.h"


/*
* MPVOpen()
* Open a sound file and return pointer to class object.
* MPEG stream is readonly. Track is ignored.
*/
void* MpvOpen(char *path, char *track, int rw)
{
    Mpeg1video* mpg = new Mpeg1video;
    debug("MpvOpen", "mpv=%lx track=%s, rw=%d, path=%s\n", 
	  (long)mpg, track, rw, path);
    mpg->Open(path);
    return (void*)mpg;
}

/*
* MPVOpen()
* Return a description of mpg file. Return NULL if the type is not
* understood.
*/
char* MpvDescribe(void* mpv, char* type)
{
    DString d;
    char username[USERNAME_MAX];
    Mpeg1video* mpg = (Mpeg1video*)mpv;

    getusername(username, sizeof(username));
    debug("MpvDescribe", "mpv=%lx type=%s\n", 
	  (long)mpv, type);

    DStringInit(&d);
    if (!strcmp(type, "application/sdp")) {
	DStringPrintf(&d, "v=0\r\n");
	DStringPrintf(&d, "o=%s %lu %lu IN IP4 %s\r\n",
		      username[0] != '\0' ? username : "-", time(0), time(0), "127.0.0.1");
	DStringPrintf(&d, "t=0 0\r\n");
	DStringPrintf(&d, "a=length:npt=%.3f\r\n", 
		      mpg->TotalFrames() / mpg->FrameRate());
	DStringPrintf(&d, "a=framerate:%.3f\r\n", 
			mpg->FrameRate());
	DStringPrintf(&d, "m=video %d RTP/MPV %d\r\n", 0, RTP_PT_MPV);
    }
    return d.string;
}

/*
* MpvSeek()
* Position within sound file to given time value.  Return actual
* position, accounting for any block structure in file, if successful,
* -1 on error.
*/
int MpvSeek(void *mpv, double npt)
{
    Mpeg1video* mpg = (Mpeg1video*)mpv;
    int framenum = (int)(npt * mpg->FrameRate());
    debug("MpvSeek", "mpv=%lx npt=%f, est frame=%d\n", 
	  (long)mpv, npt, framenum);
    mpg->SeekFrame(framenum);
    return 0;
}

/*
* MpvRead()
* Read a block of snd data into 'buffer' and determine next playout
* time.  Set 'pt' to RTP MPV payload type and 'tsinc' to frame size.
* Return number of bytes read.  Return 0 on EOF and -1 on
* error.
*/
int MpvRead(void *mpv, char *buffer, double *npt, 
	      int *pt, int *tsinc)
{
    bool isnewframe;
    int size = 1400;
    int gop;
    int frame;
    Mpeg1video* mpg = (Mpeg1video*)mpv;

    debug("MpvRead", "mpv=%lx npt=%f\n",
	  (long)mpv, *npt);
    isnewframe = mpg->ReadNext((unsigned char*)buffer, &size, &gop, &frame);
    *pt = RTP_PT_MPV; // defined in mpv.h
    *tsinc = size;
    if (isnewframe) {
	// This is end of packet. Advance npt by inverse of framerate
	*npt += (1 / mpg->FrameRate());

    } else {
	// seq or gop header, or transmission of remaining frame
	*npt = *npt + 0; // leave npt unchanged
    }
    debug("MpvRead", "gop=%d frame=%d npt=%f, pt=%d, tsinc=%d, hdr=%lx\n",
	  gop, frame,
	  *npt, *pt, *tsinc, (unsigned long) ntohl(*((long*)(&buffer[4]))));
    return size;
}

/*
* MpvWrite()
* return -1. Not supported
*/
int MpvWrite(void *mpv, int pt, u_int32 ts,
	     u_int16 seq, char *buffer,
	     int size, double *npt)
{
    // not supported
    debug("MpvWrite", "mpv=%lx npt=%f, pt=%d, size=%d\n",
	  (long)mpv, *npt, pt, size);
    return -1;
}

/*
* MpvClose()
* Close the file and release all resources.
*/
int MpvClose(void *mpv)
{
    Mpeg1video* mpg = (Mpeg1video*)mpv;
    debug("MpvClose", "mpv=%lx\n", (long)mpv);
    mpg->Close();
    delete mpg;
    return 0;
}

