/*
* Play()
* Add a play request to the queue, sorted by effective date.  If the
* PLAY request has no range and queue is empty, re-start playing at the
* current time.  Return status.
*/

#include "rtspd.h"
#include <pthread.h>

int PlayRecord(int op, Stream *st, request_t *r)
{
  request_t *rr;
  int status = 500;

  /* Check if this stream allows the method. */
  if (!(r->transport.mode & op)) {
    ResponseSetStatus(r, status = 405, "Method is not allowed");
    return status;
  }

  /* Check if range is valid. */
  if ( r->range.end >= 0) { /* -1.0 is for indicating to the end */
    if (r->range.begin > r->range.end) {
      ResponseSetStatus(r, status = 457, "Begin is after end");
      return status;
    }
  }

  /* Count request usage. */
  r->reference++;

  /* Lock stream state. */
  pthread_mutex_lock(&st->lock);  
  debug("Play","mutex(st->lock) locked\n");

  /* Empty queue or new first element: signal to stream thread. */
  st->state = op == T_PLAY ? S_PLAY : S_RECORD;

  if ( op == T_PLAY ) {
    /* Set the range to be returned in response */
    debug("PlayRecord", "range is not NULL for play\n");
    r->range.type = R_npt;
    if ( r->range.begin > 0 ) {
      debug("PlayRecord", "range begin is set to %g\n", 
            r->range.begin);
    }
    else {
      r->range.begin = (st->npt < 0) ? 0.0 : st->npt;
      debug("PlayRecord", "range begin is set to npt %g\n",
            r->range.begin);
    }
    if ( r->range.end > 0 ) {
      debug("PlayRecord", "range end set to %g\n", r->range.end);
    }
    else {
      if ( st->seek != NULL && st->fp != NULL ) {
        r->range.end = (*st->seek)(st->fp, -1.0 /* some negative value
                                               */);
        debug("PlayRecord","range end is set to file end %g\n", r->range.end);
      }
      else {
        debug("PlayRecord","range end is not set\n");
      }
    }
  }

  if (!st->requests || r->range.effective < st->requests->range.effective) {
    r->next      = st->requests;
    st->requests = r;
  }

  /* Non-empty queue: insert according to effective time. */

  else {

    for (rr = st->requests; rr; rr = rr->next) {
      if (!rr->next || (rr->next->range.effective > r->range.effective)) {
        r->next  = rr->next;
        rr->next = r;
        break;
      }
    }

  }

  /*
   * If this request was inserted at the front of the queue, we need to
   * signal to the stream thread, as it may be waiting on this event.
   */
  if (st->requests == r) {
    pthread_cond_signal(&st->queue_change);
    debug("Play", "Signaled for request %p on stream %p\n", r, st);
  }

  /* Set the response */
  ResponseSetStatus(r, status = 200, NULL);
  
  /* Set the other headers */
  ResponseSetRange(&r->headers, &r->range);

  /* Unlock stream state. */
  debug("Play","mutex(st->lock) unlocked\n");
  pthread_mutex_unlock(&st->lock);  

  return status;
} /* Play */

