/*
* PlayService()
* Send packets for a single PLAY request 'r'.
*/


#include <pthread.h>
#include "rtspd.h"

#include <unistd.h>  /* gethostname(), etc. */

#include <sys/param.h> /* MAXHOSTNAMELEN */
#include <netdb.h>     /* MAXHOSTNAMELEN */

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 64
#endif

#include "rtp.h"     /* rtp/rtcp headers, rtp constants, etc. */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <string.h>
#include "rtspd.h"
#include "rtp_api.h"  /* conflict with rtplib/rtp_api.h */
#include "getusername.h"



#define dumppacket(fp,caption,buf,len) {          \
  int i; int j;                                   \
  if (caption) fprintf(fp, "\n\n---%s---",caption); \
  for ( i=0; i<len; i++) {                        \
    if ( i % 4 == 0 ) { fprintf(fp,"\n"); }       \
    else { fprintf(fp," | "); }                   \
    for ( j=0x080; j > 0; j >>= 1) {              \
      if ( j == 0x08 ) fprintf(fp," ");           \
      if ( j & ((int)buf[i]) ) fprintf(fp,"1");   \
      else fprintf(fp,"0");                       \
    }                                             \
  }                                               \
}

/* limit on the length of a CNAME */
#define MAX_USER_HOST 255
/* initial number of session members (assumed 2) */
#define INIT_MEMBERS 2 
/* init. RTCP bandwidth (5% of total assumed session bw of 64kbits/sec) */
#define INIT_RTCP_BW 410 

char *get_user_and_host();
int estimate_rtcp_bw(double total_time, u_int32 rtcp_octets_sent);
void get_rtcp_sdes_buff(char *rtcp_sdes_buff,
			int *p_sdes_buff_length, 
			u_int32 ssrc); 
void get_rtcp_sr_buff(char *rtcp_sr_buff,
		       u_int32 ssrc, u_int32 pkt_count, 
		       u_int32 oct_count, u_int rtp_ts);
/* slightly modified, Henning's function from appendix A.7 of RFC 1889 */
double rtcp_interval(int members, 
		     int senders, 
		     double rtcp_bw, 
		     int we_sent, 
		     int packet_size, 
		     int *avg_rtcp_size, 
		     int initial);

#define CONST_HEADER_LENGTH 12

extern u_int32 random32(int type);

void InitPacket(rtp_hdr_t* rtp_hdr, int seq, int ts, int ssrc, int pt)
{
  rtp_hdr->version = RTP_VERSION;
  rtp_hdr->p = 0;
  rtp_hdr->x = 0;
  rtp_hdr->cc = 0; /* NOTE: where do we get csrc? */
  rtp_hdr->m  = 0;
  rtp_hdr->pt = pt;
  rtp_hdr->seq  = htons(seq);
  rtp_hdr->ts   = htonl(ts);
  rtp_hdr->ssrc = htonl(ssrc);
}


void PlayService(Stream *st, request_t *r)
{
  double mynpt,initialnpt;     /* normal play time (0.0...) */
  struct timespec next;        /* next playout time */
  char buffer[8192];           /* packet buffer */
  int e;                       /* error condition */
  int pt;                      /* RTP payload type */
  int tsinc;                   /* RTP timestamp increment */

  struct timeval temp_time; /* temporary time holder */
  double 
    rtcp_initial_time,      /* ititial time, used for total time calculation */
    rtcp_current_time,      /* current time, in seconds */ 
    rtcp_time_interval,     /* time, in sec., AFTER which to send next RTCP paket */
    rtcp_send_time;         /* time, in sec., AT which to send next RTCP packet */
  
  int avg_pkt_size;         /* set to avg. pkt. size by rtcp_interval() */

  int pkt_count = 0, oct_count = 0, rtcp_oct_count = 0; 
  
  int sdes_buff_length; /* set by call to get_rtcp_sdes_buff() */

  /* this buffer will be used for sendinig compound RTCP packets */
  char rtcp_send_buff[512];
  
  /* length of the compound RTCP packet.  It will be equal to 
   * 28 + sdes_buff_length (where 28 is the fixed length of our
   * SR packets, i.e. 7 * 4)
   */
  int send_buff_length; 

  int buf_length=0;
  int packet_length=0;
  int start_seq = random32(1); 

  debug("PlayService", "Servicing request %p at %.3f\n", r,
    r->range.effective);

  {
    struct timeval t;
    gettimeofday(&t, NULL);
    debug("PlayService", "current time = %lu,begin=%g, end=%g\n", 
          (long unsigned int) t.tv_sec, r->range.begin,
          r->range.end);
  }

  /* get the sdes packet buffer into the 2nd part of rtcp_send_buff */
  get_rtcp_sdes_buff(rtcp_send_buff + 28, &sdes_buff_length, st->ssrc);

  /* at this time, we know the total size of our compound 
     RTCP packet to be sent. */
  send_buff_length = sdes_buff_length + 28;

  gettimeofday(&temp_time, NULL);
  rtcp_initial_time = temp_time.tv_sec + temp_time.tv_usec / 1000000.0;

  rtcp_time_interval = rtcp_interval(INIT_MEMBERS,
				     1 /* total senders */,
				     INIT_RTCP_BW,
				     1 /* have we sent? */,
				     0, /* i.e. we haven't sent RTCP yet */
				     &avg_pkt_size,
				     1 /* initial call? */);

  /* this is the time at which we should send the first RTCP packet */
  rtcp_send_time = rtcp_initial_time /* + rtcp_time_interval */;
  /* XXX first RTCP packet is sent immediately */

  /* Loop for packet retrieval. */
  mynpt = 0.0;
  initialnpt = st->npt;

  do {
    rtp_hdr_t *rtp_hdr;
    char buf[10000];

    if ((buf_length = (*st->read)(st->fp, buffer, &mynpt, &pt, &tsinc)) <= 0) {
      debug("PlayService", "File over\n");
      break;
    }

    st->npt = initialnpt + mynpt;

    next.tv_sec  = (time_t)(r->range.effective + mynpt);
    next.tv_nsec = (long)(((r->range.effective + mynpt) - next.tv_sec) * 1e9);
    
    /*
    debug("PlayService", "Waiting for packet for %p at %ld.%09ld\n",
         r, next.tv_sec, next.tv_nsec);
    */
    //debug("PlayService","W(%g)..\n", st->npt);

    {
      struct timeval t;
      gettimeofday(&t, NULL);
      if ( next.tv_sec < t.tv_sec) {
        /*
         * No wait.
         * XXX hack for FreeBSD where pthread_cond_timedwait()
         * does not return immediately for expired timespec param.
         * and waits for approx 200 millisec.
         */
      }
      else {
        e = pthread_cond_timedwait(&st->pause, &st->lock, &next);
      }
    }

    //debug("PlayService","..D\n");
    if (st->state == S_PAUSE || st->state == S_TEARDOWN) {
      debug("PlayService", "%s\n", 
            st->state == S_PAUSE ? "Paused" : "Teardown");
      break;
    }

    /* XXX output packet */
    rtp_hdr = (rtp_hdr_t *) buf;
    InitPacket(rtp_hdr, start_seq++, st->ts, st->ssrc, pt);

    packet_length = CONST_HEADER_LENGTH + rtp_hdr->cc * (sizeof (u_int32))
      + buf_length /* + sizeof (long) */;

    memcpy(&buf[CONST_HEADER_LENGTH],buffer,buf_length);

    /* send the RTP packet */
    if ((st->transport).transport == T_TCP) {
      if (send(st->rtp.socket[0], buf, packet_length,0) < 0) { 
        perror("playservice.c: RTP TCP socket write"); 
      }
    }
    else {
      if ((st->transport).transport == T_UDP) {
        /* static int firsttime = 0; */

        struct sockaddr_in saddr;
        saddr.sin_family = AF_INET;
        saddr.sin_addr   = st->transport.destination;
        saddr.sin_port   = htons(st->transport.client_port[0]);

        /*
        if ( firsttime < 4 ) {
          dumppacket(stderr,"RTP", buf, 12);
          firsttime++;
        }
        */
        /*
        fprintf(stderr,"S %s:%d\n", inet_ntoa(saddr.sin_addr), 
                ntohs(saddr.sin_port));
        */

        if (sendto(st->rtp.socket[0], buf, packet_length,0,
          (struct sockaddr *)&saddr,sizeof(saddr)) < 0) {
	        perror("playservice.c: RTP UDP socket write"); 
        }
      }
    }
    pkt_count++;
    oct_count+=buf_length;

    gettimeofday(&temp_time, NULL);
    rtcp_current_time = temp_time.tv_sec + temp_time.tv_usec / 1000000.0;    
  
    /* is it time to send an RTCP packet yet? */
    if (rtcp_current_time >= rtcp_send_time) { 

      /* increment counter of total RTCP octets sent by us.
       * It is used below to estimate rtcp_bw (in the call to
       * estimate_rtcp_bw() )
       */
      rtcp_oct_count += send_buff_length;
      
      rtcp_time_interval = 
          rtcp_interval(INIT_MEMBERS,
		      1 /* total senders */,
		      estimate_rtcp_bw(rtcp_current_time - rtcp_initial_time, 
				       rtcp_oct_count), 
		      1 /* have we sent? */,
		      send_buff_length +
		      20 /* IP */ +
		      (T_UDP==(r->transport).transport) ?
		      8 : 20, /* 8 for UDP, 20 for TCP */
		      &avg_pkt_size,
		      0 /* initial call? */);

      /* this is the time at which we should send the next RTCP packet */
      rtcp_send_time = rtcp_current_time + rtcp_time_interval;

      /* below we compose and send the compound RTCP packet */

      /* get the SR packet buffer and put it into
       * the first portion of rtcp_send_buff
       * (remember, rtcp_sdes_buff has been placed 
       * into the 2nd portion) 
       */ 
      get_rtcp_sr_buff(rtcp_send_buff, st->ssrc, pkt_count, oct_count, st->ts);

      /* send the rtcp_send_buff */
      if ((st->transport).transport == T_TCP) {
        if (send(st->rtp.socket[1], rtcp_send_buff, send_buff_length,0) <
        0) { 
	        perror("playservice.c: RTCP TCP socket write"); 
        }
      }
      else {
        if ((st->transport).transport == T_UDP) {
          struct sockaddr_in saddr;
          saddr.sin_family = AF_INET;
          saddr.sin_addr   = st->transport.destination;
          saddr.sin_port   = htons(st->transport.client_port[1]);

          /*dumppacket(stderr,"RTCP",  rtcp_send_buff,
            send_buff_length);*/

          debug("PlayService", "RTCP\n");

          if (sendto(st->rtp.socket[1], rtcp_send_buff,
              send_buff_length, 0,
              (struct sockaddr *)&saddr,sizeof(saddr)) < 0) {
	          perror("playservice.c: RTP UDP socket write"); 
          }
        }
      } /* else */
    } /* if time to send RTCP packet */

    st->ts += tsinc;

  } while ((st->npt < r->range.end) || (r->range.end <= 0));

  debug("PlayService", "Exiting\n");

} /* PlayService */


/* 
 * Allocates and returns a string of the format "username@hostname" with
 * a total length of at most 255 characters (as required by RTCP).
 */
char *get_user_and_host()
{
  char hostname[MAXHOSTNAMELEN + 1]; /* additional room for a '\0' */
  int diff, hostlen;
  char *user_host = (char *) malloc(MAX_USER_HOST + 1);

  getusername(user_host, MAX_USER_HOST);
  if ('\0' == user_host[0]) /* if not found, then a fixed username */
    sprintf(user_host, "username_not_found");

  if (gethostname(hostname, MAXHOSTNAMELEN) < 0)
    perror("user_and_host: gethostname");
 
  hostname[MAXHOSTNAMELEN] = '\0';

  /* "user@host" must be at most 255 characters long */
  diff = MAX_USER_HOST - 1 - strlen(user_host);
  hostlen = strlen(hostname);

  if (hostlen >= diff) /* hostname is longer than allowed */
    hostname[diff] = '\0'; /* make it shorter */
 
  sprintf(user_host + strlen(user_host), "@%s", hostname);
  return user_host;
}


/* this function fills up the rtcp_sdes_buff passed to it.
 * It also sets the buffer's length, pointed to by p_sdes_buff_length
 */
void get_rtcp_sdes_buff(char *rtcp_sdes_buff,
			int *p_sdes_buff_length, 		    
			u_int32 ssrc) 
{
  char *temp_cname; /* temporary pointer to CNAME returned by 
		       get_user_and_host() function */
  int temp_length; 
  rtcp_t *p_rtcp_sdes_pkt = (rtcp_t *) rtcp_sdes_buff;

  /* populate the header and SSRC fields of SDES packet */
  p_rtcp_sdes_pkt->common.version = RTP_VERSION;
  p_rtcp_sdes_pkt->common.pt      = RTCP_SDES;
  p_rtcp_sdes_pkt->common.p       = 0;
  p_rtcp_sdes_pkt->common.count   = 1; /* 1 sdes chunk */
  /* the length field will be put in later */

  p_rtcp_sdes_pkt->r.sdes.src     = htonl(ssrc);

  /* populate the type and length fields of the first and only SDES item */

  p_rtcp_sdes_pkt->r.sdes.item->type = RTCP_SDES_CNAME; 

  temp_cname = get_user_and_host(); /* get the cname string */ 
  p_rtcp_sdes_pkt->r.sdes.item->length = strlen(temp_cname);

  /* now, populate the data field of the SDES item (i.e. actual CNAME) */

  memcpy(p_rtcp_sdes_pkt->r.sdes.item->data, 
	 temp_cname,
	 p_rtcp_sdes_pkt->r.sdes.item->length);

  free(temp_cname);

  /* length of SDES pkt payload =
     [ 
     p_rtcp_sdes_pkt->r.sdes.item->length 
     + sizeof(p_rtcp_sdes_pkt->r.sdes.item->length) 
     + sizeof(p_rtcp_sdes_pkt->r.sdes.item->type) 
     + sizeof(p_rtcp_sdes_pkt->r.sdes.src) 
     ] bytes 
  */
  temp_length = p_rtcp_sdes_pkt->r.sdes.item->length + 1 + 1 + 4; 

  /* round temp_length to the next multiple of 4 */
  temp_length = (temp_length + 3) & 0xfffc;

  /* finally, put in, as promised, the length field of SDES packet;
     it is equal to the number of 32-bit words in SDES pkt payload
     (i.e. not including the SDES packet header word) */
  p_rtcp_sdes_pkt->common.length = htons(temp_length / 4);

  /* total_length of the SDES packet =
     4 bytes for header, and temp_length for the rest */
  *p_sdes_buff_length = 4 + temp_length; /* the calling function will
					       see this value */

  /* below we just pad the packet to the next word boundary */
  
  /* skip to the end of the buffer */
  rtcp_sdes_buff += (10 + p_rtcp_sdes_pkt->r.sdes.item->length); 

  memset(rtcp_sdes_buff, 0, (*p_sdes_buff_length - 10 
			     - p_rtcp_sdes_pkt->r.sdes.item->length));
} 



/* This function fills up the rtcp_sr_pkt that is passed to it.
 * We don't return its length because in our case it's fixed
 * at 4 * 7 == 28 bytes 
 */
void get_rtcp_sr_buff(char *rtcp_sr_buff,
		       u_int32 ssrc, u_int32 pkt_count, 
		       u_int32 oct_count, u_int rtp_ts) {
  struct timeval tv;

  rtcp_t *rtcp_sr_pkt = (rtcp_t *) rtcp_sr_buff;

  rtcp_sr_pkt->common.version = RTP_VERSION;
  rtcp_sr_pkt->common.pt      = RTCP_SR;
  rtcp_sr_pkt->common.p       = 0;
  rtcp_sr_pkt->common.count   = 0; /* i.e. no rr blocks */
  rtcp_sr_pkt->common.length  = htons(6); /* i.e. there are 6 other 32-bit fields */

  if (gettimeofday(&tv, NULL) < 0) 
    perror("get_rtcp_sr_buff: gettimeofday");

  rtcp_sr_pkt->r.sr.ssrc      = htonl(ssrc);
  rtcp_sr_pkt->r.sr.ntp_sec   = htonl((u_int32)tv.tv_sec);
  rtcp_sr_pkt->r.sr.ntp_frac  = htonl((u_int32)tv.tv_usec);
  rtcp_sr_pkt->r.sr.rtp_ts    = htonl(rtp_ts);
  rtcp_sr_pkt->r.sr.psent     = htonl(pkt_count);
  rtcp_sr_pkt->r.sr.osent     = htonl(oct_count);
}



/*
 * This function returns the 
 * interval (in seconds) between sending of RTCP packets.
 *   members = estimated # of session members, incl. ourselves.
 *   senders = # of active senders (in our case, 1)
 *   rtcp_bw = total bandwidth to be used for RTCP packets by all members
 *             of this session (in octets/sec).  This should be 5% of the
 *             "session bandwidth" parameter supplied to the application at 
 *              startup.
 *   we_sent = Flag that is true if we have sent data during the last two
 *             RTCP intervals. If the flag is true, the compound RTCP packet
 *             just sent contained an SR packet.
 *   packet_size = The size of the compound RTCP packet just sent, in
 *       octets, including the network encapsulation (e.g., 28 octets for
 *       UDP over IP).
 *   avg_rtcp_size = Pointer to estimator for compound RTCP packet size;
 *       initialized and updated by this function for the packet just
 *       sent, and also updated by an identical line of code in the RTCP
 *       receive routine for every RTCP packet received from other
 *       participants in the session.
 *   initial = Flag that is true for the first call upon startup to
 *       calculate the time until the first report should be sent.
 *
 */
double rtcp_interval(int members, 
		     int senders, 
		     double rtcp_bw, 
		     int we_sent, 
		     int packet_size, 
		     int *avg_rtcp_size, 
		     int initial) 
{ 
  struct timeval tv; /* for the seed, to be use in calling srand() */

  /* 
   * Minimum time between RTCP packets from this site (in seconds). 
   * This time prevents the reports from `clumping' when sessions 
   * are small and the law of large numbers isn't helping to smooth 
   * out the traffic.  It also keeps the report interval from 
   * becoming ridiculously small during transient outages like a 
   * network partition. 
   */ 
  double const RTCP_MIN_TIME = 5.; 
  /* 
   * Fraction of the RTCP bandwidth to be shared among active 
   * senders.  (This fraction was chosen so that in a typical 
   * session with one or two active senders, the computed report 
   * time would be roughly equal to the minimum report time so that 
   * we don't unnecessarily slow down receiver reports.) The 
   * receiver fraction must be 1 - the sender fraction. 
   */ 
  double const RTCP_SENDER_BW_FRACTION = 0.25; 
  double const RTCP_RCVR_BW_FRACTION = (1-RTCP_SENDER_BW_FRACTION); 
  /* 
   * Gain (smoothing constant) for the low-pass filter that 
   * estimates the average RTCP packet size (see Cadzow reference). 
   */ 
  double const RTCP_SIZE_GAIN = (1./16.); 
  
  double t;                   /* interval */ 
  double rtcp_min_time = RTCP_MIN_TIME; 
  int n;                      /* no. of members for computation */ 
  
  /* 
   * Very first call at application start-up uses half the min 
   * delay for quicker notification while still allowing some time 
   * before reporting for randomization and to learn about other 
   * sources so the report interval will converge to the correct 
   * interval more quickly.  The average RTCP size is initialized 
   * to 128 octets which is conservative (it assumes everyone else 
   * is generating SRs instead of RRs: 20 IP + 8 UDP + 52 SR + 48 
   * SDES CNAME). 
   */ 
  if (initial) { 
    rtcp_min_time /= 2; 
    *avg_rtcp_size = 128; 
  } 
  
  /* 
   * If there were active senders, give them at least a minimum 
   * share of the RTCP bandwidth.  Otherwise all participants share 
   * the RTCP bandwidth equally. 
   */ 
  n = members; 
  if (senders > 0 && senders < members * RTCP_SENDER_BW_FRACTION) { 
    if (we_sent) { 
      rtcp_bw *= RTCP_SENDER_BW_FRACTION; 
      n = senders; 
    } else { 
      rtcp_bw *= RTCP_RCVR_BW_FRACTION; 
      n -= senders; 
    } 
  } 
  
  /* 
   * Update the average size estimate by the size of the report 
   * packet we just sent. 
   */ 
  *avg_rtcp_size += (packet_size - *avg_rtcp_size)*RTCP_SIZE_GAIN; 
  
  /* 
   * The effective number of sites times the average packet size is 
   * the total number of octets sent when each site sends a report. 
   * Dividing this by the effective bandwidth gives the time 
   * interval over which those packets must be sent in order to 
   * meet the bandwidth target, with a minimum enforced.  In that 
   * time interval we send one report so this time is also our 
   * average time between reports. 
   */ 
  t = (*avg_rtcp_size) * n / rtcp_bw; 
  if (t < rtcp_min_time) t = rtcp_min_time; 
  
  /* 
   * To avoid traffic bursts from unintended synchronization with 
   * other sites, we then pick our actual next report interval as a 
   * random number uniformly distributed between 0.5*t and 1.5*t. 
   */ 

  /* NOTE: I had to change the code below slightly because Henning 
     used drand48() and it didn't work. -maksim */
  gettimeofday(&tv, NULL);
  srand(tv.tv_usec);

  return t * ((rand() / 31768.0)  + 0.5); 
} 



/* This function takes the total elapsed time (in s) since we started 
 * sending rtcp data and the number of octets of RTCP data that was
 * sent. 
 */
int estimate_rtcp_bw(double total_time, u_int32 rtcp_octets_sent) {
  
  /* for now, we only count the total octets of RTCP data sent by _us_.
   * When an RTCP listener gets implemented (which would listen for RTCP
   * RR packets from other session members), then we should take into
   * account their bandwidth as well
   */

  return (int) (rtcp_octets_sent/total_time); 
}
