/*
* Read RMF (RealMedia) files.
*
* RealMedia (RMF) files consist of chunks.
*/

#include <stdio.h>      /* fseek() */
#include <fcntl.h>      /* O_ */
#include <malloc.h>     /* calloc(), free() */
#include <string.h>     /* strncmp() */
#include <stdlib.h>

#define REALMEDIA_INTERNAL
#include "realmedia.h"
#include "rtp_pt.h"
#include "dstring.h"    /* DString..() */

typedef unsigned short u_int16;
typedef unsigned long u_int32;
typedef unsigned int u_int8;

typedef struct {
    u_int32 object_id;           /* object ID for Media Properties Header */
    u_int32 size;                /* size of Media Properties Header (bytes) */
    u_int16 object_version;      /* version of header */
    u_int32 stream_number;       /* identifies a media stream */
    u_int32 max_bit_rate;        /* max rate to deliver over network */ 
    u_int32 avg_bit_rate;        /* avg rate to deliver over network */
    u_int32 max_packet_size;     /* largest packet size (bytes) */
    u_int32 avg_packet_size;     /* avg packet size (bytes) */
    u_int32 start_time;          /* time offset in (ms) added to timestamp */
    u_int32 preroll;             /* time offset in (ms) subtract to timestamp */
    u_int32 duration;            /* duration of stream in (ms) */
    u_int8  stream_name_size;    /* length of stream name (bytes) */
    char    *stream_name;        /* stream name */
    u_int8  mime_type_size;      /* length of mime_type (bytes) */
    char    *mime_type;          /* MIME style */
    u_int8  type_specific_len;   /* length of type_specific_data (bytes) */
    char    *type_specific_data; /* type_specific_data */
} mdpr;                          /* RealMedia Media Properties Header */

typedef struct {
    u_int16 object_version;      /* version of header */
    u_int16 length;              /* length of packet (bytes) */
    u_int16 stream_number;       /* 16-bit alias */
    u_int32 timestamp;           /* timestamp of packet */
    u_int8  reserved;            /* not used */
    u_int16 flags;               /* properties of packet */
    char    *data;               /* media data */
} data_packet;                   /* RealMedia Data Packet Header */

typedef struct {
  FILE      *f;                  /* file pointer */
  mdpr      stream_a;            /* media audio stream headers */
  mdpr      stream_v;            /* media video stream headers */
  struct {
    u_int32 object_id;           /* object ID for RealMedia files */
    u_int32 size;                /* size of RealMedia header (bytes) */
    u_int16 object_version;      /* version of header */
    u_int32 file_version;        /* version of RealMedia file in PN format */
    u_int32 num_headers;         /* number of headers in header section */
  } rmf;                         /* Realmedia file header */
  struct {
    u_int32 object_id;           /* object ID for Properties Header */
    u_int32 size;                /* size of Properties Header (bytes) */
    u_int16 object_version;      /* version of header */
    u_int32 max_bit_rate;        /* max rate to deliver over network */
    u_int32 avg_bit_rate;        /* avg rate to deliver over network */
    u_int32 max_packet_size;     /* largest packet size (bytes) */
    u_int32 avg_packet_size;     /* avg packet size (bytes) */
    u_int32 num_packets;         /* number of packets in media data */
    u_int32 duration;            /* duration of file in (ms) */
    u_int32 preroll;             /* duration to pre-buffer before play (ms) */
    u_int32 index_offset;        /* offset from start to index header */ 
    u_int32 data_offset;         /* offset from start to data (bytes) */
    u_int16 num_streams;         /* number of media streams in file */
    u_int16 flags;               /* characteristics of RealMedia file */
  } prop;                        /* Realmedia properties header */
  struct {
    u_int32 object_id;           /* object ID for Content Header */
    u_int32 size;                /* size of Content Header (bytes) */
    u_int16 object_version;      /* version of header */
    u_int16 title_len;           /* length of title (bytes) */
    char    *title;              /* title */
    u_int16 author_len;          /* length of author (bytes) */
    char    *author;             /* author */
    u_int16 copyright_len;       /* length of copyright (bytes) */
    char    *copyright;          /* copyright */
    u_int16 comment_len;         /* length of comment (bytes) */
    char    *comment;            /* comment */
  } cont;                        /* RealMedia Content Description Header */
  struct {
    u_int32 object_id;           /* object ID for Data Header */
    u_int32 size;                /* size of Data Header (bytes) */
    u_int16 object_version;      /* version of header */
    u_int32 num_packets;         /* number of pckets in data chunk */
    u_int32 next_data_header;    /* offset from start to next data chunk */
  } data;                        /* RealMedia Content Data Chunk Header */
  data_packet *packets;          /* RealMedia Packets */
} RealMedia;



/*
* RealMediaOpen()
* Open a .rm file and parse its header.
*/
RealMedia *RealMediaOpen(char *path, char *track, int rw)
{
  RealMedia *realmedia;           /* .rm object */
  FILE *f;                        /* .rm file pointer */
  char magic[4];
  char *mode;                     /* file mode */
  char *buffer;                   /* memory space to store input */
  int section = 0;                /* indicates the header section */
  int len;                        /* length of field */

  if      (rw == O_RDWR)   mode = "r+";
  else if (rw == O_RDONLY) mode = "r";
  else if (rw == O_WRONLY) mode = "w";

  realmedia = calloc(1, sizeof(RealMedia));
  if ((realmedia->f = fopen(path, mode)) < 0) {
    free(realmedia);
    return NULL;
  }
  f = realmedia->f;

  buffer = calloc(4, sizeof(char));
  if ((fread(buffer, sizeof(char), 4, f) != 4) || 
      (strncmp(buffer, ".RMF", 4))) {
    goto error;
  }

  while(!feof(f)) {
    realmedia->rmf.object_id = ".RMF";
    if ((fread(buffer, sizeof(char), 4, f)) == 4) {
      if (strncmp(buffer, "DATA", 4) == 0 || section == 4) {
	section = 4;                         /* section=4 is the DATA header */
        realmedia->data.object_id = "DATA";

        /* Additional Parsing code for DATA section */

      }
      else if (strncmp(buffer, "CONT", 4) == 0 || section == 3) {
	section = 3;                         /* section=3 is the CONT header */
        realmedia->cont.object_id = "CONT";

        fseek(f, 43, SEEK_CUR);
        buffer = calloc(2, sizeof(char));
        fread(buffer, sizeof(char), 1, f);
        len = (int)buffer[0];
        buffer = calloc(len+1, sizeof(char));
        fread(buffer, sizeof(char), len, f); /* title */
        realmedia->cont.title_len = len;
        realmedia->cont.title = buffer;

        fseek(f, 1, SEEK_CUR);
        buffer = calloc(2, sizeof(char));
        fread(buffer, sizeof(char), 1, f);
        len = (int)buffer[0];
        buffer = calloc(len+1, sizeof(char));
        fread(buffer, sizeof(char), len, f); /* author */
        realmedia->cont.author_len = len;
        realmedia->cont.author = buffer;

        fseek(f, 2, SEEK_CUR);
        buffer = calloc(184, sizeof(char));
        fread(buffer, sizeof(char), 184, f); /* comment */
        realmedia->cont.comment_len = 184;
        realmedia->cont.comment = buffer;

        buffer = calloc(4, sizeof(char));    /* setup next header section */

        /* Additional Parsing code for CONT section */

      }
      else if (strncmp(buffer, "MDPR", 4) == 0 || section == 2) {
        section = 2;                         /* section=2 is the MDPR header */

        realmedia->stream_a.object_id = buffer;
        fseek(f, 36, SEEK_CUR);              /* skipping unknown data */

        buffer = calloc(2, sizeof(char));
        fread(buffer, sizeof(char), 1, f);
        len = (int)buffer[0];
        buffer = calloc(len+1, sizeof(char));
        fread(buffer, sizeof(char), len, f); /* stream_name */
        realmedia->stream_a.stream_name_size = len;
        realmedia->stream_a.stream_name = buffer;

        buffer = calloc(2, sizeof(char));
        fread(buffer, sizeof(char), 1, f);
        len = (int)buffer[0];
        buffer = calloc(len+1, sizeof(char));
        fread(buffer, sizeof(char), len, f); /* mime_type */
        realmedia->stream_a.mime_type_size = len;
        realmedia->stream_a.mime_type = buffer;

        fseek(f, 77, SEEK_CUR);
        buffer = calloc(5, sizeof(char));
        fread(buffer, sizeof(char), 4, f);     
        realmedia->stream_v.object_id = buffer;
        fseek(f, 36, SEEK_CUR);

        buffer = calloc(2, sizeof(char));
        fread(buffer, sizeof(char), 1, f);
        len = (int)buffer[0];
        buffer = calloc(len+1, sizeof(char));
        fread(buffer, sizeof(char), len, f); /* stream_name */
        realmedia->stream_v.stream_name_size = len;
        realmedia->stream_v.stream_name = buffer;

        buffer = calloc(2, sizeof(char));
        fread(buffer, sizeof(char), 1, f);
        len = (int)buffer[0];
        buffer = calloc(len+1, sizeof(char));
        fread(buffer, sizeof(char), len, f); /* mime_type */
        realmedia->stream_v.mime_type_size = len;
        realmedia->stream_v.mime_type = buffer;

        fseek(f, 38, SEEK_CUR);              /* setup next header section */
        buffer = calloc(4, sizeof(char));

        /* Additional Parsing code for MDPR section */

      }
      else if (strncmp(buffer, "PROP", 4) == 0 || section == 1) {
        section = 1;                         /* section=1 is the PROP header */
        realmedia->prop.object_id = "PROP";

        /* Additional Parsing code for PROP section */

      }
    }
  }

  return realmedia;

error:
    free(realmedia);
    return NULL;

} /* RealMediaOpen */



/*
* RealMediaDescribe()
* Return a description of a sound file. Return NULL if the type is not
* understood.
*/
char *RealMediaDescribe(RealMedia *realmedia, char *type)
{
  DString d;
  
  return d.string;
} /* RealMediaDescribe */



/*
* RealMediaSeek()
* Position within REALMEDIA file to given time value.  Return 0 
* if successful, -1 on error.
*/
double RealMediaSeek(RealMedia *realmedia, double npt)
{
  long offset;      /* bytes of offset into file */
  int blocks;       /* packet number */
  int stat;         /* fseek return value */

  /* positioning to the nearest block */
  offset = realmedia->prop.avg_bit_rate * npt;
  blocks = offset / realmedia->prop.avg_packet_size;
  stat = fseek(realmedia->f, realmedia->prop.data_offset + offset, SEEK_SET);

  return stat;
} /* RealMediaSeek */



/*
* RealMediaRead()
* Read a block of realmedia data and determine next playout time. Return
* number of bytes read. Return 0 on EOF and -1 on error.
*/
int RealMediaRead(RealMedia *realmedia, char *buffer, int *cpt, int *npt)
{
  int offset;       /* bytes of offset into file */
  int blocks;       /* packet number */
  int n;            /* bytes read */

  offset = realmedia->prop.avg_bit_rate * *cpt;

  /* adjusting the offset value to the minimum starting offset */
  if (offset < realmedia->prop.data_offset) {
    offset = realmedia->prop.data_offset;     
    fseek(realmedia->f, offset, SEEK_SET);
  }
  blocks = offset / realmedia->prop.avg_packet_size;

  n = fread(buffer, sizeof(char), realmedia->packets[blocks].length, 
            realmedia->f);

  *npt += realmedia->packets[blocks].length / realmedia->prop.avg_bit_rate;

  return n;
} /* RealMediaRead */



/*
* RealMediaClose()
* Close the .rm file and release all resources.
*/
int RealMediaClose(RealMedia *realmedia)
{
  fclose(realmedia->f);
  free(realmedia);
  return 0;
} /* RealMediaClose */







