/*
* request.c
* Create, search, process, destroy requests.
*
* Currently, pending requests are in a linked list; for efficiency, they
* should be in a hash table.
*/

#include <stdio.h>  
#include <stdlib.h>     /* atoi() */
#include <string.h>     /* strstr() */
#include <unistd.h>     /* access() */
#include <errno.h>      /* errno */
#include "rtspd.h"      /* config. */
#include "stream.h"
#include "strdupn.h"

/* Function prototypes. */

/*
 * GetMethodsAllowed: Return the string containing
 * the comma separated list of methods allowed.
 */
const char *GetMethodsAllowed() 
{
  static char *methods 
    = "ANNOUNCE, DESCRIBE, SETUP, PLAY, RECORD, TEARDOWN, OPTIONS";
  return methods;
}

/*
* RequestCreate()
* Create a new request object.
*/
request_t *RequestCreate(int type, int socket, struct sockaddr_in sin)
{
  request_t *r = calloc(1, sizeof(request_t));

  r->type    = type;
  r->socket  = socket;
  r->sin     = sin;
  r->path[0] = '\0';

  return r;
} /* RequestCreate */


/*
* ProcessRequest()
* Process RTSP messages according to the request type. 
* Return 0 on success.
*/
void RequestProcess(request_t *r)
{
  int session_id;              /* numeric session identifier */
  char *method;                /* method of request */
  Stream *st;                  /* current stream */
  Stream *nextst;
  int matches = 0;             /* streams that match */
  rtsp_session_t session;
  double range_npt = 0.0;
  int fd = -1;

  /* Get request information. */
  method        = HTTP_Method(&r->p);
  if ( method == NULL ) {
    /* Possibly a parse error */
    error("RequestProcess", "method is NULL. Parse error.\n");
    RequestFree(r);
    return;
  }

  r->seq        = RTSP_CSeq(&r->p);
  r->version    = HTTP_Version(&r->p);
  session       = RTSP_Session(&r->p);
  session_id    = session.id ? strtoul(session.id, NULL, 10) : -1;
  r->range      = RTSP_Range(&r->p);
  r->transport  = RTSP_Transport(&r->p);

  /* Get the Session * from session_id */
  r->session    = SessionGet(session_id);

  if (r->p.URI.path) {
    if (*r->p.URI.path == '~') {
      /* Map the ~user to appropriate rtsp directory.
       * Currently this is ignored */
    }
    else {
      snprintf(r->path, sizeof(r->path),
                  "%s/%s", config.DocumentRoot, r->p.URI.path);
    }
  }

  debug("ProcessRequest", "%s %s from %s:%d %p\n", 
    method, r->path, inet_ntoa(r->sin.sin_addr), r->sin.sin_port, r);

  /* Build the default response headers */
  ResponseBuildDefault(r);
  ResponseSetStatus(r, 501, NULL); /* Default failure status */

  /* 
   * Make sure this is a valid URI. In particular, it cannot start with
   * .. after the first slash to avoid security problems. For now, we
   * don't allow .. at all.
   */
  if (r->p.URI.path && strstr(r->p.URI.path, "..")) {
    ResponseSetStatus(r, 404, "Path contains .. that is Forbidden");
  }
  /* XXX to be changed. Does not work if SETUP is sent without ANNOUNCE for
     recording.
     else if ( strcmp(r->p.method,"ANNOUNCE") &&
     strcmp(r->p.method,"RECORD") && 
     r->path[0] && access(r->path, R_OK)) {
     status = 403;
     } 
  */
  else if (!HTTP_Protocol(&r->p) ) {
    ResponseSetStatus(r, 505, "No protocol set; use RTSP");
  }
  else if (strcmp(HTTP_Protocol(&r->p), "RTSP") ) {
    DString errorreason;
    DStringInit(&errorreason);
    DStringPrintf(&errorreason, "Can not support %s; use RTSP", HTTP_Protocol(&r->p));
    ResponseSetStatus(r, 505, DStringValue(&errorreason));
    DStringFree(&errorreason);
  }
  else if (strcmp(HTTP_Version(&r->p), "1.0")) {
    DString errorreason;
    DStringInit(&errorreason);
    DStringPrintf(&errorreason, "Can not support RTSP version %s; use 1.0", 
                  HTTP_Version(&r->p));
    ResponseSetStatus(r, 505, DStringValue(&errorreason));
    DStringFree(&errorreason);
  }
  else if (!strcmp(method, "SETUP")) {
    /* Setup sets the return status code. */
    Setup(r);
  }
  else if (!strcmp(method, "DESCRIBE")) {
    /* Describe sets the return status code as well as the
     * content body if any */
    Describe(r, &fd);
  }
  else if (!strcmp(method, "ANNOUNCE")) {
    /*
     * Announce sets the return status and interprets any
     * content body 
     */
    Announce(r);
  }
  else if (!strcmp(method, "SET_PARAMETER")) {
    ResponseSetStatus(r, 501, "SET_PARAMETER is not implemented");
    HeaderReplaceOrAppend(&r->headers, "Public",
                          GetMethodsAllowed());
    /* Do nothing; just return success. */
  }
  else if (!strcmp(method, "OPTIONS")) {
    ResponseSetStatus(r, 200, NULL);
    HeaderReplaceOrAppend(&r->headers, "Public",
                          GetMethodsAllowed());
  }
  else if (!strcmp(method, "DELETE")) {
    /* Delete sets the status code as needed */
    Delete(r);
  }
  /* All subsequent cases need session id */
  else if (!session.id) {
    ResponseSetStatus(r, 454, "No session identifier");
  }
  else if (!r->session) {
    ResponseSetStatus(r, 454, "Session in empty");
  }
  else {

    /* Iterate through all the streams of this session
     * and call the appropriate method */
    for (st = r->session->streams; st; st = nextst) {
      /* Get the stream for the URI. */
      nextst = StreamGet(st->next, r->p.URI);

      matches++;
      if (!strcmp(method, "PLAY")) {
        PlayRecord(T_PLAY, st, r);
      }
      else if (!strcmp(method, "RECORD")) {
        PlayRecord(T_RECORD, st, r);
      }
      else if (!strcmp(method, "PAUSE")) {
        Pause(st, r);
      }
      else if (!strcmp(method, "TEARDOWN")) {
        Teardown(st, r);      
      }
      else {
        DString errorreason;
        DStringInit(&errorreason);
        DStringPrintf(&errorreason, "Method %s is not supported", method);
        ResponseSetStatus(r, 505, DStringValue(&errorreason));
        DStringFree(&errorreason);
        HeaderReplaceOrAppend(&r->headers, "Public",
                              GetMethodsAllowed());
      }
    }

    if (matches == 0) {
      ResponseSetStatus(r, 454, "Invalid stream");
    }
    else if (matches != 1) {
      error("RequestProcess", "Matches more than one stream");
    }
  }

  /* Send the response to client */
  ResponseSend(r, fd);
  if ( fd >= 0 ) {
    close(fd);
  }

  return;
} /* ProcessRequest */
