#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>  /* fstat() */
#include <string.h>
#include "rtspd.h"
#include "dstring.h"   /* DString... */
#include "strdupn.h"

static char *server_name = "Columbia RTSP Server 1.0";
static char *m[699];

void ResponseInit(void)
{
  m[100] = "Continue";
  m[200] = "OK";
  m[201] = "Created";
  m[250] = "Low on Storage Space";
  m[300] = "Multiple Choices";
  m[301] = "Moved Permanently";
  m[302] = "Moved Temporarily";
  m[303] = "See Other";
  m[304] = "Not Modified";
  m[305] = "Use Proxy";
  m[400] = "Bad Request";
  m[401] = "Unauthorized";
  m[402] = "Payment Required";
  m[403] = "Forbidden";
  m[404] = "Not Found";
  m[405] = "Method Not Allowed";
  m[406] = "Not Acceptable";
  m[407] = "Proxy Authentication Required";
  m[408] = "Request Time-out";
  m[410] = "Gone";
  m[411] = "Length Required";
  m[412] = "Precondition Failed";
  m[413] = "Request Entity Too Large";
  m[414] = "Request-URI Too Large";
  m[415] = "Unsupported Media Type";
  m[416] = "File can not be created";
  m[451] = "Parameter Not Understood";
  m[452] = "Conference Not Found";
  m[453] = "Not Enough Bandwidth";
  m[454] = "Session Not Found";
  m[455] = "Method Not Valid in This State";
  m[456] = "Header Field Not Valid for Resource";
  m[457] = "Invalid Range";
  m[458] = "Parameter Is Read-Only";
  m[459] = "Aggregate operation not allowed";
  m[460] = "Only aggregate operation allowed";
  m[500] = "Internal Server Error";
  m[501] = "Not Implemented";
  m[502] = "Bad Gateway";
  m[503] = "Service Unavailable";
  m[504] = "Gateway Time-out";
  m[505] = "RTSP Version not supported";
} /* ResponseInit */


/*
 * GetStandardReason()
 * Get the standard reason phrase for this status code
 */
const char *GetStandardReason(int status)
{
  if (status < 100 || status >= 600) {
    return "Unknown status";
  }
  if (m[status] != NULL) {
    return m[status];
  }
  else {
    /* Round to the nearest x00 status. */
    status = ((int)(status / 100)) * 100;
    return m[status];
  }
} /* GetStandardReason */

/* 
 * ResponseBuildDefault()
 * Given a request, create the standard response headers for it and add
 * them to the header structure.
*/
int ResponseBuildDefault(request_t *r)
{
  parser_header_t *h;

  /* If there are previous headers here, delete them. */
  HeaderFree(&r->headers);

  /* Session. */
  if (r->session) {
    HeaderPrintf(&r->headers, 0, "Session", "%lu;timeout=%d",
      r->session->session_id, config.timeout);
  }
  HeaderPrintf(&r->headers, 0, "Cseq", "%lu", (long)r->seq);
  if ( server_name != NULL ) {
    HeaderPrintf(&r->headers, 0, "Server", "%s/%s", server_name,
                 VERSION);
  }
  if ((h = RFC822_Header(r->p.headers, "Timestamp")) && h->elements) {
    HeaderPrintf(&r->headers, 0, "Timestamp", "%s", h->elements->value);
  }
  
  ResponseSetBody(r, NULL, NULL, 0);
  return 0;
} /* ResponseBuildDefault */


/*
* ResponseSetBody()
* Set body and body content properties for request or response. If
* content_length is < 0, use length of body.
*/
void ResponseSetBody(request_t *r, const char *body,
                     const char *content_type, int content_length)
{
  strassign(&r->response_body.content,      body);
  strassign(&r->response_body.content_type, content_type);
  r->response_body.content_length = content_length;

  /* Content-Type, Content-Length */
  if (r->response_body.content_length < 0)
    r->response_body.content_length =
      r->response_body.content ? strlen(r->response_body.content) : 0;
  if (content_type && r->response_body.content_length > 0) {
    HeaderReplaceOrAppend(&r->headers, "Content-Type",
                          r->response_body.content_type);
  }
  else {
    HeaderDelete(&r->headers, "Content-Type");
  }
  HeaderPrintf(&r->headers, 1, "Content-Length", "%d",
    r->response_body.content_length);
} /* ResponseSetBody */


/*
* ResponseSetStatus()
* Set the status and reason phrase.
*/
void ResponseSetStatus(request_t *r, int status, const char *reason)
{
  r->status = status;
  strassign(&r->reason, reason);
} /* ResponseSetStatus */

/*
* ResponseSend()
* Generate a response message and send.
*/
int ResponseSend(request_t *r, int fd)
{
  DString msg;         /* to make this MT-safe, new string for each call */
  parser_header_t *h;  /* request header */
  int n;               /* bytes read or written */
  int blk_size = 0;    /* optimal block size for reading response file */

  /* Invalid code. */
  if ((r->status >= 600) || (r->status < 100) ) { 
    error("Response", "invalid status %d", r->status);
    ResponseSetStatus(r, 500, "Invalid status");
  }

  DStringInit(&msg);

  /* Status line. */
  DStringPrintf(&msg, "RTSP/%s %d %s\r\n",
    r->version ? r->version : "1.0",  r->status,
    r->reason ? r->reason : GetStandardReason(r->status));
  
  /* Append the headers */
  HeaderMessage(&msg, &r->headers);
  DStringAppend(&msg, "\r\n", -1);

  /* Append the body */
  if ( r->response_body.content != NULL &&
       r->response_body.content_length > 0 ) {
    DStringAppend(&msg, r->response_body.content,
                  r->response_body.content_length);    
  }

  /* Send the message */
  if (r->type == SOCK_DGRAM) {
    n = sendto(r->socket, msg.string, msg.length, 0,
      (struct sockaddr *)&r->sin, sizeof(r->sin));
  }
  else {
    n = send(r->socket, msg.string, msg.length,0);
  }
  if (n < 0) {
    error("Response", "write");
    return -1;
  }

  debug("Response", "Response for %p:\n%s", r, msg.string);

  /* If available, copy file or 'body' argument to output socket. */
  if (fd >= 0 ) {
    /* Get the input from the file */
    char buffer[4096];

    if (blk_size > sizeof(buffer)) {
      blk_size = sizeof(buffer);
    }

    do {
      n = read(fd, buffer, blk_size);

      if (n > 0) {
        if (r->type == SOCK_DGRAM) {
          /* Is it correct to split RTSP message over two UDP packets
             XXX */
          n = sendto(r->socket, buffer, n, 0,
              (struct sockaddr *)&r->sin, sizeof(r->sin));
        }
        else {
          n = send(r->socket, buffer, n,0);
        }
        debug("Response", "Body for %p: (size %d)\n", r, n);
      }
    } while (n > 0);
  }

  if (r->status < 0) {
    error("Response", "write");
    return -1;
  }
  return 0;
} /* Response */


/*
 * Set the response header Range: 
 */
void ResponseSetRange(headers_t *h, rtsp_range_t *range)
{
  char buffer[200];    /* for time strings */
  if ((range->type != -1 ) && (range->effective > 0)) {
    DString msg;
    DStringInit(&msg);

    if (range->type == R_npt) {
      DStringAppend(&msg, "npt=", -1);
      if (range->begin >= 0) {
        npttostr(range->begin, buffer);
        DStringAppend(&msg, buffer, -1);
      }
      if (range->end >= 0) {
        DStringAppend(&msg, "-", -1);
        npttostr(range->end, buffer);
        DStringAppend(&msg, buffer, -1);
      }
      else {
        if ( range->end == -1.0 ) {
          /* Till the end of the file. XXX should
           * this be change to < 0.0 instead of == -1.0 */
          DStringAppend(&msg, "-", -1);
        }
      }
    }
    else if (range->type == R_clock) {
      DStringAppend(&msg, "clock=", -1);
      if (range->begin >= 0) {
        utctostr(range->begin, buffer);
        DStringAppend(&msg, buffer, -1);
      }
      if (range->end >= 0) { 
        DStringAppend(&msg, "-", -1);
        utctostr(range->end, buffer);
        DStringAppend(&msg, buffer, -1);
      }
    }
    else if (range->type == R_smpte) {
      DStringAppend(&msg, "smpte=", -1);
      if (range->begin >= 0) {
        smptetostr(range->begin, buffer);
        DStringAppend(&msg, buffer, -1);
      }
      if (range->end >= 0) {
        DStringAppend(&msg, "-", -1);
        smptetostr(range->end, buffer);
        DStringAppend(&msg, buffer, -1);
      }
    }
    utctostr(range->effective, buffer);
    DStringPrintf(&msg, ";time=%s", buffer);

    HeaderReplaceOrAppend(h, "Range", DStringValue(&msg));

    DStringFree(&msg);
  }
  else {
    debug("ResponseSetRange", "No range specified\n");
  }
}

/*
 * Set the Transport: header 
 */
void ResponseSetTransport(headers_t *h, rtsp_transport_t *transport)
{
  DString msg;
  DStringInit(&msg);

  DStringPrintf(&msg, "%s/%s","rtp","avp");
  DStringPrintf(&msg, ";%s", transport->multicast ? "multicast" : "unicast");

  if (transport->destination.s_addr != INADDR_ANY ) {
    DStringPrintf(&msg, ";destination=%s",
                  inet_ntoa(transport->destination));
  }

  if ( transport->multicast ) {
    DStringPrintf(&msg, ";port=%d-%d",
                  transport->server_port[0],transport->server_port[1]);
  }
  else {
    DStringPrintf(&msg, ";server_port=%d-%d;client_port=%d-%d",
                  transport->server_port[0],transport->server_port[1],
                  transport->client_port[0],transport->client_port[1]);
  }

  HeaderReplaceOrAppend(h, "Transport", DStringValue(&msg));

  DStringFree(&msg);
}
