/*
* Read and write RTP (.rtp) files.
*/

#define RTPFILE_INTERNAL 1
#include <stdio.h>         /* SEEK_CUR and SEEK_SET */
#include "sysdep.h"
#include "dstring.h"
#include "rtp_pt.h"
#include <unistd.h>        /* lseek() */
#include <string.h>        /* strcmp() */
#include <unistd.h>        /* open() */
#include <stdlib.h>        /* calloc() */
#include <fcntl.h>         /* O_RDONLY, O_RDWR, O_WRONLY */
#include <sys/types.h>
#include <netinet/in.h>    /* ntohl() */
#include <sys/stat.h>
#include "error.h"         /* debug () */
#include "getusername.h"   /* getusername () */

#define MAX_FORMATS 100

typedef struct {
  FILE * fp; /* File */
  int rw; /* Mode. O_RDONLY, O_WRONLY */
  long start;
  int ts;
} RtpFile;

typedef struct {
  int pt; /* Payload type */
  int tsinc; /* Timestamp increment from previous packet. */
  int seq;   /* Sequence number */
  double npt; /* npt timestamp */
  int length; /* Length ofdata */
  char buffer[1500]; /* buffer */

} RtpFilePacket;

double RtpSeek(RtpFile *rtp, double npt);

/* Put the packet in the file */
int PutRtpFilePacket(RtpFile *rtp, int pt, int ts, int seq, double
                     npt, char *buffer, int size)
{
  int tsinc;
  int length;

  /* Payload type */  
  pt = htonl(pt);
  fwrite(&pt, 4, 1, rtp->fp);

  /* tsinc */
  if ( rtp->ts == 0 ) {
    tsinc = htons(160); /* Just set to some number */
  }
  else {
    tsinc = htonl(ts - rtp->ts);
  }
  rtp->ts = ts;
  fwrite(&tsinc, 4, 1, rtp->fp);

  /* seq */
  seq = htonl(seq);
  fwrite(&seq, 4, 1, rtp->fp);

  /* npt */
  fwrite(&npt, sizeof(npt), 1, rtp->fp);

  /* Length */
  length = htonl(size);
  fwrite(&length, 4, 1, rtp->fp);

  /* Buffer */
  if ( size > 0 ) {
    fwrite(buffer, 1, size > 1500 ? 1500 : size, rtp->fp);
  }

  return 0;
}

int GetNextRtpPacket(FILE *rtp, RtpFilePacket *packet ) 
{
  if ( fread(&packet->pt, 4, 1, rtp)<=0 ||
       fread(&packet->tsinc,4,1,rtp) <= 0 ||
       fread(&packet->seq,4,1,rtp) <= 0 ||
       fread(&packet->npt,sizeof(packet->npt),1,rtp) <= 0 ||
       fread(&packet->length,4,1,rtp) <= 0 ) {
    return -1;
  }

  if ( packet->length > 0 ) {
    if ( fread(packet->buffer, 1, packet->length > sizeof(packet->buffer) ?
               sizeof(packet->buffer) : packet->length, rtp) <=0)
      {
        return -1;
      }
  }

  packet->pt = ntohl(packet->pt);
  packet->tsinc = ntohl(packet->tsinc);
  packet->seq = ntohl(packet->seq);
  packet->length = ntohl(packet->length);

  return 0;
}


RtpFile *RtpOpen(char *path, char *track, int rw)
{
  char *mode;
  RtpFile *rtp;           /* rtp file object */

  rtp = calloc(1, sizeof(RtpFile));

  if ( rw == O_RDONLY ) {
    mode = "rb";
  }
  else if ( rw == O_WRONLY) {
    mode = "wb";
  }
  else {
    mode = "rb+";
  }

  if ((rtp->fp = fopen(path, mode)) < 0) {
    debug("RtpOpen", "Could not open file %s in mode %s\n", path, mode);
    free(rtp);
    return NULL;
  }

  rtp->rw = rw;

  if ( rw == O_RDONLY ) {
    char dummy[1024];
    long size;

    /* For read, read the top line first */
    if ( fgets(dummy, sizeof(dummy)-1, rtp->fp) == NULL) {      
      fclose(rtp->fp);
      free(rtp);
      return NULL;
    }

    dummy[sizeof(dummy)-1] = '\0';
    if ( strncmp(dummy, "#!rtpfile", strlen("#!rtpfile")) != 0) {
      debug("RtpOpen", "Not an RTP file\n");
      fclose(rtp->fp);
      free(rtp);
      return NULL;
    }
         
    if ( fread(&size, sizeof(size), 1, rtp->fp) <= 0 ) {
      debug("RtpOpen", "No size for information\n");
      fclose(rtp->fp);
      free(rtp);
      return NULL;
    }

    if ( size > 0 && size <= sizeof(dummy)) {
      /* Skipping size bytes */
      fread(dummy, 1, size, rtp->fp);
    }
    else {
      debug("RtpOpen", "Invalid size %ld\n", size );
    }
    
  }
  else if (rw == O_WRONLY) {
    if (fputs("#!rtpfile 127.0.0.1/8000\n", rtp->fp)<=0) {
      fclose(rtp->fp);
      free(rtp);
      return NULL;
    }

    if ( track != NULL ) {
      /* Write the track info in the file */
      long size = ntohl(strlen(track)+1);
      fwrite(&size, 4, 1, rtp->fp);
      fwrite(track, 1, strlen(track)+1, rtp->fp);
    }
    else {
      long size = 0;
      fwrite(&size, 4, 1, rtp->fp);
    }

  }

  rtp->start = ftell(rtp->fp);
  if ( rtp->start < 0 ) {
    error("RtpOpen", "ftell() failed\n");
    fclose(rtp->fp);
    free(rtp);
    return NULL;
  }

  return rtp;

} /* RtpOpen */


/*
* RtpDescribe()
* Return a description of a sound file. Return NULL if the type is not
* understood.
*/
char *RtpDescribe(RtpFile *rtp, char *type)
{
  DString d;
  char username[USERNAME_MAX];
  int audio_formats[MAX_FORMATS+1];
  int i;
  long pos;
  RtpFilePacket  packet;
  double npt;

  if ( rtp == NULL || rtp->rw != O_RDONLY) {
    debug("RtpDescribe", "RTP is null or mode is not readonly\n");
    return NULL;
  }

  audio_formats[0] = -1;

  /* Get the file size */
  npt =  RtpSeek(rtp, -1.);

  pos = ftell(rtp->fp);

  fseek(rtp->fp, rtp->start, SEEK_SET);
  while( GetNextRtpPacket(rtp->fp, &packet ) == 0 ) {
    int oFound = 0;
    int j;

    int pt = packet.pt;
    for ( j=0; audio_formats[j] >= 0 && j < MAX_FORMATS; j++ ) {
      if ( audio_formats[j] == pt ) {
        oFound = 1;
      }
    }
    if ( !oFound ) {
      audio_formats[j] = pt;
      audio_formats[j+1] = -1;
    }
  }

  fseek(rtp->fp, pos, SEEK_SET);

  getusername(username, sizeof(username));

  DStringInit(&d);
  if (!strcmp(type, "application/sdp")) {
    int oFound = 0;

    DStringPrintf(&d, "v=0\r\n");
    DStringPrintf(&d, "o=%s %lu %lu IN IP4 %s\r\n",
      username[0] != '\0' ? username : "-", (long) time(0), (long) time(0), 
      "127.0.0.1");
    /* s= line must be present */
    DStringPrintf(&d, "s=(no subject)\r\n");
    DStringPrintf(&d, "t=0 0\r\n");

    DStringPrintf(&d, "a=length:npt=%.3f\r\n", npt);
    DStringPrintf(&d, "m=audio %d RTP/AVP", 0);
    for ( i=0; audio_formats[i] >= 0 && i < MAX_FORMATS; i++) {
      DStringPrintf(&d, " %d", audio_formats[i]);
      debug("RtpDescribe", "Adding format %d\n", audio_formats[i]);
      oFound = 1;
    }
    if ( !oFound ) {
      DStringPrintf(&d, " 0");
    }
    DStringAppend(&d, "\r\n", -1);
  }
  return d.string;
} /* SndDescribe */


/*
* RtpSeek()
* Position within sound file to given time value.  Return actual
* position, accounting for any block structure in file, if successful,
* -1 on error. Return duration if
* npt is negative.
*/
double RtpSeek(RtpFile *rtp, double npt)
{
  long lastpos;
  double lastnpt;

  RtpFilePacket packet;

  if ( rtp == NULL ) {
    return 0.0;
  }

  fseek(rtp->fp, rtp->start, SEEK_SET);

  if ( npt == 0. ) {
    /* If start of the file is requested retrun now. */
    return 0.;
  }

  lastpos = rtp->start;
  if ( GetNextRtpPacket(rtp->fp, &packet) < 0 ) {
    /* No packet available */
    return 0.;
  }
  lastnpt = packet.npt;

  while (npt < 0. || npt > lastnpt) {
    if (GetNextRtpPacket(rtp->fp, &packet) < 0 ) {
      break;
    }
    lastnpt = packet.npt;
    lastpos = ftell(rtp->fp);
  }
  fseek(rtp->fp, lastpos, SEEK_SET);

  return lastnpt;
} /* RtpSeek */

/*
* RtpRead()
* Read a block of snd data into 'buffer' and determine next playout
* time.  Set 'pt' to RTP payload type and 'tsinc' to time stamp
* increment.  Return number of bytes read.  Return 0 on EOF and -1 on
* error.
*/
int RtpRead(RtpFile *rtp, char *buffer, double *npt, int *pt, int *tsinc)
{
  RtpFilePacket packet;

  if ( rtp == NULL || rtp->rw != O_RDONLY) {
    debug("RtpRead", "Null or not read mode\n");
    return -1;
  }

  if ( GetNextRtpPacket(rtp->fp, &packet) < 0 ) {
    debug("RtpRead", "Getnext returned -ve\n");
    return 0;
  }

  if (npt) {
    *npt = packet.npt;
  }
  if (pt) {
    *pt = packet.pt;
  }
  if ( tsinc ) {
    *tsinc = packet.tsinc;
  }
  if (buffer && packet.length > 0) {
    memcpy(buffer, packet.buffer, packet.length);
  }

  /* Return length */
  return packet.length;
} /* RtpRead */



/*
* RtpWrite()
* Write a block of snd data.  Return number of bytes written.  Return 0
* on EOF and -1 on error.
*/
int RtpWrite(RtpFile *rtp, int pt, int ts, int seq, char *buffer,
  int size, double *npt)
{
  if ( rtp == NULL || rtp->rw != O_WRONLY) {
    debug("RtpWrite", "Null or not write mode\n");
    return -1;
  }

  /* XXX how do we set npt */
  *npt += 160./8000.;

  PutRtpFilePacket(rtp, pt, ts, seq, *npt, buffer, size);

  return size;

} /* RtpWrite */


/*
* RtpClose()
* Close the file and release all resources.
*/
int RtpClose(RtpFile *rtp)
{
  if ( rtp == NULL ) {
    return -1;
  }

  fclose(rtp->fp);

  free(rtp);

  return 0;
} /* RtpClose */
