/*
* Columbia University RTSP Server
*
* Author: Maria Papadopouli and Henning Schulzrinne
* Department of Computer Science
* Columbia University  
* Last updated: October 17th, 1997
*/

/**
 * @name rtspd.c
 * RTSP Server. Main routine for rtspd server.
 * We use the following threads:
 * - TCP accept, creates TCP connection thread
 * - one per TCP connection to receive requests
 * - receive UDP requests and enter PLAY, PAUSE onto queue
 * - one per stream to get P/P requests from the queue and serve packets
 *   created on SETUP; it waits on a condition variable for new requests
 * - a cleanup thread that periodically checks for stale connections [NOT YET]
 *
 * Requests are created by the network interface and either queued or
 * destroyed when processed.
 */

#include <stdlib.h>
#include <unistd.h>    /* close() */
#include <stdio.h>     /* fopen() */
#include <string.h>    /* strdup() */
#include <pthread.h>   /* pthread_ */
#include <errno.h>     /* errno */
#include <pwd.h>       /* getpwnam() */
#include <grp.h>       /* getgrnam() */
#include <netdb.h>     /* gethostname, etc */
#include <sys/param.h> /* MAXHOSTNAMELEN */
#include <signal.h>    /* signal */
#ifndef WIN32
#include <sys/resource.h>  /* setrlimit,getrlimit for unix */
#endif

#include "stream.h"
#include "dstring.h"
#include "rtspd.h"
#include "mime.h"
#include "config.h"
#include "thread_pool.h" /* for libcine, InitThreadPool */
#include "udp.h"
#include "tcp.h"

#ifndef HAVE_TIMEGM
#include "timegm.h"
#endif

#define MAX_REQUEST 8192  /* large enough for UDP packet */

/** server config information, e.g., server root */
struct config_t config;
extern int debug_flag;   /* debug level */
const int missing_content_length_means_zero = 1;

/*
* Read configuration file.
*/
static void Config(char *root)
{
  FILE *f;                     /* configuration file */
  char full_path[PATH_MAX];
  int argc;                    /* number of arguments */
  char *argv[10];              /* words in configuration line */

  if (!root) root = ".";
  sprintf(full_path, "%s/rtspd.conf", root);
  if (!(f = fopen(full_path, "r"))) {
    fatal("Config", "Cannot open configuration file %s", full_path);
  }

  config.CNAME        = "";
  if (!config.port)
    config.port       = 554;
  config.timeout      = 60;
  config.ServerName   = "dynasty.cs.columbia.edu";
  config.ServerRoot   = ".";
  config.DocumentRoot = "";
  config.ErrorLog     = "error_log";
  config.TransferLog  = "access_log";
  config.uid          = 0;
  config.gid          = 0;

  argc = 10;
  while (ConfigRead(f, &argc, argv)) {
    if (argc < 2) { 
    }
    else if (!strcmp(argv[0], "CNAME")) {
      config.CNAME = argv[1];
    }
    else if (!strcmp(argv[0], "DocumentRoot")) {
      config.DocumentRoot = argv[1];
      if (access(config.DocumentRoot, R_OK) < 0) {
        fatal("Config", "DocumentRoot %s", config.DocumentRoot);
      }
    }
    else if (!strcmp(argv[0], "ErrorLog")) {
      config.ErrorLog = argv[1];
      if (access(config.ErrorLog, W_OK) < 0) {
        fatal("Config", "ErrorLog %s", config.ErrorLog);
      }
    }
    else if (!strcmp(argv[0], "Group")) {
      struct group *grp;

      if (*argv[1] == '#') {
        config.gid = atoi(argv[1]+1);
      }
      else {
        grp = getgrnam(argv[1]);
        if (grp) config.gid = grp->gr_gid;
        else fatal("Config", "Invalid group %s\n", argv[1]);
      }
    }
    else if (!strcmp(argv[0], "Port")) {
      config.port = atoi(argv[1]);
    }
    else if (!strcmp(argv[0], "ServerName")) {
      config.ServerName = argv[1];
    }
    else if (!strcmp(argv[0], "ServerRoot")) {
      config.ServerRoot = argv[1];
      if (access(config.ServerRoot, W_OK) < 0) {
        fatal("Config", "ServerRoot %s", config.ServerRoot);
      }
    }
    else if (!strcmp(argv[0], "TransferLog")) {
      config.TransferLog = argv[1];
      if (access(config.TransferLog, W_OK) < 0) {
        fatal("Config", "TransferLog %s", config.TransferLog);
      }
    }
    else if (!strcmp(argv[0], "User")) {
      struct passwd *pwd;

      if (*argv[1] == '#') {
        config.uid = atoi(argv[1]+1);
      }
      else {
        pwd = getpwnam(argv[1]);
        if (pwd) config.uid = pwd->pw_uid;
        else fatal("Config", "Invalid user %s\n", argv[1]);
      }
    }
    argc = 10;
  }
} /* Config */

/*
 * Get and return the canonical hostname of this machine.
 * Not thread-safe -- only intended for startup time.
 */
char *getcanonicalhostname()
{
  char buf[MAXHOSTNAMELEN];
  struct hostent *hep;
  struct in_addr addr;

  gethostname(buf, MAXHOSTNAMELEN);
  
  hep = gethostbyname(buf);
  if (hep == NULL) {
    error("getcanonicalhostname", "gethostbyname failed on %s, h_errno = %d\n",
          buf, h_errno);
    return NULL;
  }

  memcpy(&addr, hep->h_addr, sizeof(addr));

  hep = gethostbyaddr((char *)&(addr.s_addr), sizeof(addr.s_addr), AF_INET);
  if (hep == NULL) {
    error("getcanonicalhostname", "gethostbyaddr failed on %s, h_errno = %d\n",
          inet_ntoa(addr), h_errno);
    return NULL;
  }

  return strdup(hep->h_name);
}

/*
* usage()
* Usage information. Exits program with status code 1.
*/
void usage(char *prog)
{
  fprintf(stderr,
  "Usage: %s [-p port] [-s root] [-v [-o tracefile]][-X]\n", prog);
  exit(1);
} /* usage */


/*
* Main function for server.
*/
int main(int argc, char *argv[])
{
  int e;                          /* thread error condition */
  extern char *optarg;            /* command line argument value */
  extern int optind;
  int c;                          /* argv character */
  pthread_t tid[2];               /* thread identifiers */
  int isdaemon = 1;
#ifndef WIN32
  struct rlimit rbuf;
#endif
  struct tcp_thread_arg *tcp_arg;
  struct udp_thread_arg *udp_arg;
  extern void ResponseInit(void);
  extern void setupExitThread(void);
  extern void DebugInit(char *filename);
  extern void do_daemon(void);

  startupSocket();

  ErrorInit("stderr");

  /* Parse command line options. */
  while ((c = getopt(argc, argv, "s:p:vo:X")) != EOF) {
    switch (c) {
      case 's':
        config.ServerRoot = optarg;
        break;
      case 'p':
        config.port = atoi(optarg);
        break;
      case 'v':
        debug_flag = 1;
        break;
      case 'o':
        if ( debug_flag ) {
           DebugInit(optarg);
        }
        break;
      case 'X':
        isdaemon = 0;
        break;
      case '?':
        usage(argv[0]);
        break;
    } 
  }

  Config(config.ServerRoot);

  if (isdaemon) {
    ErrorInit(NULL); /* syslog */
    do_daemon();
  }

  if ( debug_flag ) {
    ErrorInit("stderr"); /* use standard output */
  }
  else {
    ErrorInit(NULL); /* use syslog */
  }

  /* Should we capture SIPHUP for re-reading the config file XXX */
  signal(SIGPIPE, SIG_IGN); 

  setupExitThread();

  InitThreadPool(128);

  ResponseInit();
  if (MIME_Init(config.ServerRoot, "mime.types") < 0) {
    fatal("main",
      "%s/mime.types and default MIME types file not readable", 
          config.ServerRoot);
  }

#ifndef WIN32
  /* Set the open file limit. Works for solaris */
  memset(&rbuf, 0, sizeof(rbuf));
  if ( getrlimit(RLIMIT_NOFILE, &rbuf) == 0 ) {
    rbuf.rlim_cur = rbuf.rlim_max;
    if ( setrlimit(RLIMIT_NOFILE, &rbuf) != 0 ) {
      error("main", "Could not set the OPEN_MAX limit to %lu\n",
            rbuf.rlim_cur);
    }
  }
  
  /* Set the core file rlimit. Works for solaris */
  memset(&rbuf, 0, sizeof(rbuf));
  if ( getrlimit(RLIMIT_CORE, &rbuf) == 0 ) {
    rbuf.rlim_cur = rbuf.rlim_max;
    if ( rbuf.rlim_cur > 10000000L ) {
      rbuf.rlim_cur = 10000000L;
    }
    if ( setrlimit(RLIMIT_CORE, &rbuf) != 0 ) {
      error("main", "Could not set the RLIMIT_CORE limit to %lu\n",
            rbuf.rlim_cur);
    }
  }
#endif

  debug("main", "RTSP server %s (built %s %s) listening on port %d\n",
    VERSION, __DATE__, __TIME__, config.port);

  /* Allocate tcp_arg and create TCP thread */
  tcp_arg = (struct tcp_thread_arg *)malloc(sizeof(struct tcp_thread_arg));
  tcp_arg->server = NULL; /* Not used, since single server */
  tcp_arg->port   = config.port;
  if ((e = pthread_create(&tid[0], NULL, 
                                   ReceiveTCP, (void *)tcp_arg)) != 0) {
    fatal("main", "Cannot create TCP thread: %s\n", strerror(e));
  }

  /* Allocate udp_arg and create UDP thread */
  udp_arg = (struct udp_thread_arg *)malloc(sizeof(struct udp_thread_arg));
  udp_arg->server = NULL; /* Not used, since single server */
  udp_arg->port   = config.port;
  if ((e = pthread_create(&tid[1], NULL, 
                                   ReceiveUDP, (void *)udp_arg)) != 0) {
    fatal("main", "Cannot create UDP thread: %s\n", strerror(e));
  }

 
  /* Wait for thread completion. */
  WaitForMultipleObjects( 2, tid, TRUE, INFINITE);

  return 0;
} /* main */
