/*
* Thread that picks up PLAY and RECORD requests and delivers packets.
*/
#include <pthread.h>
#include <errno.h>
#include <sys/time.h>   /* gettimeofday() */
#include <fcntl.h>      /* O_RDWR */
#include <unistd.h>     /* unlink() */
#include <string.h>     /* strcpy() */
#include "rtspd.h"
#include "mime.h"
#include "snd.h"
#include "wav.h"
#include "qt.h"
#include "mpv.h"
#include "mp3.h"
#include "rtpfile.h"

static struct {
  char *type;
  void *(*open)(char *path, char *track, int rw); 
  int   (*seek)(void *, double npt);
  int   (*read)(void *, char *buffer, double *npt, int *pt, int *tsinc);
  int   (*write)(void *, int pt, u_int32 ts, u_int16 seq, char *buffer,
                int size, double *npt);
  int   (*close)(void *);
} filehandlers[] = {
  {"audio/basic", SndOpen, SndSeek, SndRead, SndWrite, SndClose},
  {"audio/x-wav", WavOpen, WavSeek, WavRead, WavWrite, WavClose},
  {"video/quicktime",   QTOpen,  QTSeek,  QTRead,  QTWrite,  QTClose},
  {"video/mpeg",   MpvOpen,  MpvSeek,  MpvRead,  MpvWrite,  MpvClose},
  {"audio/mpeg",   Mp3Open,  Mp3Seek, Mp3Read, Mp3Write, Mp3Close},
  {"application/x-rtp",   RtpOpen,  RtpSeek, RtpRead, RtpWrite, RtpClose},
  {NULL,          0,       0,       0,        0,        0}
};


/*
* Service()
* Service play and record requests.
*/
void *Service(void *arg)
{
  Stream *st = (Stream *)arg;  /* current stream */
  request_t *r;                  /* request to be serviced */
  struct timespec start;       /* time to wake up for start of service */
  struct timeval now;          /* current time */
  int e;                       /* error condition */
  int service_started = 0;     /* whether RECORD/PLAY is received */

  pthread_mutex_lock(&st->lock);
  debug("Service","mutex(st->lock) locked\n");

  for (;;) {
    /* Wait until there is a new request. */
    debug("Service", "Waiting for signal for stream %p\n", st);
    while (!st->requests && st->state != S_TEARDOWN) {
      debug("Service","cond(st->lock) wait\n");
      pthread_cond_wait(&st->queue_change, &st->lock);
      debug("Service","cond(st->lock) wait exiting\n");
    } 
    if (st->state == S_TEARDOWN)  {

      /* SETUP-TEARDOWN is used for DELETING a file if mode is record */
      if ( (service_started == 0) &&
           ( st->transport.mode == T_RECORD)) {
            sprintf(st->filename, "%s/%s", config.DocumentRoot, st->URI.path);
            if ( unlink( st->filename ) < 0 ) {
              debug("Service","Error deleting the file");
          }
      }
      break;
    }

    service_started = 1; /* either RECORD/PLAY is received */

    /* Wait until request becomes effective. */
    do {
      r = st->requests;
      start.tv_sec  = (time_t)r->range.effective;
      start.tv_nsec = (r->range.effective - start.tv_sec) * 1e9;
      //debug("Service", "Waiting for request %p at %d\n", r, start.tv_sec);
      debug("Service","cond(st->lock) timedwait\n");
      e = pthread_cond_timedwait(&st->queue_change, &st->lock, &start);
      debug("Service","cond(st->lock) timedwait exiting\n");
    } while ((st->state != S_TEARDOWN) && (e != ETIMEDOUT));
    if (st->state == S_TEARDOWN) break;

    /* Service a valid, current request. */
    r = st->requests;

    /* Get actual starting time of request. */
    gettimeofday(&now, NULL);
    r->range.effective = now.tv_sec + now.tv_usec / 1e6;

    /* Open file if name is different from previous request. */
    if (strcmp(st->filename, r->path)) {
      int f;       /* file type index */
      int rw;      /* read or write */
      char *type;  /* MIME type */

      /* Close old file. */
      if (st->fp) {
        (*st->close)(st->fp);
        st->fp = NULL;
      } 

      /*
       * Open file in appropriate read/write mode.  Currently, we do
       * this based on the global session characteristics.  We should
       * probably make this dependent on the request (PLAY, RECORD), but
       * would then have to track which mode the file has been opened
       * in. 
       */
      if (r->transport.mode == (T_PLAY | T_RECORD)) rw = O_RDWR;
      else if (r->transport.mode == T_PLAY)         rw = O_RDONLY;
      else if (r->transport.mode == T_RECORD)       rw = O_WRONLY;
      else {
        debug("Service", "Unknown r->transport.mode %d\n", r->transport.mode);
        rw = O_RDONLY;
      }

      /*
       * Open file; use search string to identify tracks within
       * aggregate files.
       */
      strcpy(st->filename, r->path);
      type = MIME_FilenameToType(r->path);
      for (f = 0; filehandlers[f].open; f++) {
        if (!strcmp(type, filehandlers[f].type)) {
          if ( rw != O_WRONLY ) {
            st->fp = (*filehandlers[f].open)(r->path, NULL, rw);
          }
          else {
            /* st->subject stores the info */
            st->fp = (*filehandlers[f].open)(r->path, st->subject, rw);
          }
          st->close = filehandlers[f].close;
          st->seek  = filehandlers[f].seek;
          st->read  = filehandlers[f].read;
          st->write = filehandlers[f].write;
          break;
        }
      }
    }

    /* If Range, set current position. */
    if (r->range.begin > 0.) st->npt = r->range.begin;

    /* Position within file. 
    if ( st->npt < 0 ) {
      // If npt is -1 then make it 0 
      st->npt = 0;
    }
    */
    if ((*st->seek)(st->fp, ((st->npt >= 0) ? st->npt : 0)) >= 0) {
      if (!strcmp(r->p.method, "PLAY")) PlayService(st, r);
      else RecordService(st, r);
    }

    /* Remove serviced request from queue. */
    st->requests = st->requests->next;
    RequestFree(r);
  }
  debug("Service", "Exiting stream %p\n", st);
  debug("Service","mutex(st->lock) unlocked\n");
  pthread_mutex_unlock(&st->lock);

  return NULL;
} /* Service */
