/*
* session.c
*
* A session consists of zero or more streams.  For efficiency (and since
* we get to pick the session numbers), we maintain an expanding table of
* pointers to sessions rather than a linear linked list.  We dynamically
* allocate each session entry since the base address may change when a
* new stream is created.  For security, the session identifier is a
* 32-bit identifier:  the most significant 20 bits are random, the lower
* 12 bits are the index in the table.  The random part could be extended
* as necessary.
*/

#include "stream.h"
#include <stdio.h>
#include <stdlib.h>     /* calloc() */
#include <pthread.h>
 
#define SESSIONS_INITIAL 20

static Session **session_list; /* base of sessions array */
static int entries;            /* currently available entries */
static int used;               /* in use */
static pthread_mutex_t lock;   /* global lock for table */ 

/*
* SessionCreate()
* Create a new session. Return pointer to stream object.
*/
Session *SessionCreate(void)
{
  int n;
  Session *s;
  extern u_int32 random32(int);

  /* expand table */
  if (entries == 0) {
    pthread_mutex_init(&lock, NULL);
    pthread_mutex_lock(&lock);
    session_list = calloc(SESSIONS_INITIAL, sizeof(Session *));
    entries = SESSIONS_INITIAL;
    n = 0;
  }
  else if (used >= entries) {
    pthread_mutex_lock(&lock);
    entries *= 1.5;
    session_list = realloc(session_list, entries * sizeof(Session *));
    memset(&session_list[used], 0, (entries-used)*sizeof(Session *));
    n = used+1;
  }  
  else {
    pthread_mutex_lock(&lock);
    for (n = 0; n < entries; n++) {
      if (session_list[n] == NULL) break;
    } 
    if ( n>=entries ) {
      error("SessionCreate", 
            "n<entries is not true; n=%d, entries=%d\n",
            n, entries);
      pthread_mutex_unlock(&lock);
      return NULL;
    }
  }
  used++;
  pthread_mutex_unlock(&lock);

  s = session_list[n] = calloc(1, sizeof(Session));
  pthread_mutex_init(&s->lock, NULL);
  pthread_mutex_lock(&s->lock);
  s->session_id = (random32((long)pthread_self()) & 0xfffff000) | n;
  time(&s->last_request);
  pthread_mutex_unlock(&s->lock);

  return s;
} /* SessionCreate */


/*
* SessionGet()
* Map a session identifier to a stream by taking the lower 12 bits as a
* table index.
*/
Session *SessionGet(u_int32 session_id)
{
  Session *s;
  int n = session_id & 0xfff;

  if ((n < 0) || (n >= entries))  {
    debug("SessionGet", "returning NULL, n=%d, entries=%d\n", n, entries);
    return NULL;
  }
  s = session_list[n];
  if (s->session_id != session_id) {
    debug("SessionGet", "returning NULL, s->sid=%ld, sid=%ld\n", 
          s->session_id, session_id);
    return NULL;
  }
  time(&s->last_request);
  if ( s== NULL ) {
    debug("SessionGet", "returning NULL\n");
  }
  return s;
} /* SessionGet */


/*
* SessionAttach()
* Attach stream pointer 'st' to end of list for session 's'.
*/
void SessionAttach(Session *s, Stream *st)
{
  Stream *spt;

  if (s->streams) {
    for (spt = s->streams; spt; spt = spt->next) {
      if (spt->next == NULL) {
        spt->next = st;
        break;
      }
    } 
  }
  else s->streams = st;
} /* SessionAttach */


/*
* SessionDetach()
* Detach stream pointer 'st' from session 's'.
*/
void SessionDetach(Session *s, Stream *st)
{
  Stream *spt;

  if (st == s->streams) {
    s->streams = s->streams->next;
  }
  else {
    for (spt = s->streams; spt; spt = spt->next) {
      if (spt->next == st) {
        spt->next = spt->next->next;
        break;
      }
    }
  }
} /* SessionDetach */


/*
* SessionFree()
* Release a stream and its resources.
*/
int SessionFree(int session_id)
{
  if ((session_id < 0) || (session_id >= entries)) return -1;
  pthread_mutex_destroy(&session_list[session_id]->lock);
  pthread_mutex_lock(&lock);
  free(session_list[session_id]);
  session_list[session_id] = NULL;
  pthread_mutex_unlock(&lock);
  debug("SessionFree", "Session %u freed\n", session_id);
  return 0;
} /* SessionFree */


/*
* SessionPrint()
* Show list of sessions.
*/
void SessionPrint(void)
{
  Session *s;
  int n;

  for (n = 0; n < entries; n++) {
    if ((s = session_list[n])) {
      printf("Session %d\n", n);
      StreamPrint(s->streams);
    }
  }
} /* SessionPrint */
