/*
* Setup()
* Create a new stream, after checking that this is not a duplicate request.
*/

#include "rtspd.h"
#include <string.h>
#include <pthread.h>

/* Function prototype for local functions */
static unsigned short start_listen_port(unsigned short port,
                                        int * rtp_sock,
                                        int * rtcp_sock);

int Setup(request_t *r)
{
  Stream *st;
  int e;
  int status = 500;

  /* Need to create a new session. */
  if (!r->session) {
    r->session = SessionCreate();
    if ( r->session == NULL ) {
      debug("Setup", "Can not create the session\n");
      return status;
    }
  }

  /* Check if we already got a setup for this stream. */
  pthread_mutex_lock(&r->session->lock);
  st = StreamByOwner(r->session, r->p.URI, r->sin, r->seq);
  if (!st) {
    /* Create new stream and mark request as belonging to the new stream. */
    st = StreamCreate(r->p.URI);
    pthread_mutex_lock(&st->lock);
    debug("Setup","mutex(st->lock) locked\n");
    SessionAttach(r->session, st);
    st->seq       = r->seq;
    st->sin       = r->sin;
    st->transport = r->transport;
    st->state     = S_SETUP;

    if ( (r->request_body.content_length > 0) &&
         (r->request_body.content != NULL) ) {
      debug("Setup", "detected following message body\n%s\n",
          r->request_body.content);
      strassign(&st->subject, r->request_body.content);
    }
    else {
      strassign(&st->subject, "No information");
    }

    debug("Setup", "Stream %p in Session %ld created by request %p\n", st,
      r->session->session_id, r);

    if (st->transport.transport == T_TCP) {
      /* XXX We do not process RTP embedded in RTSP/TCP connection */
      error("Setup", "Embedded RTP in RTSP/TCP is not processed\n");
      ResponseSetStatus(r, status = 500, "RTP/TCP not supported");

    } else { /* T_UDP */ 

      if ( (st->transport.server_port[0] = 
            start_listen_port(0, 
                              &st->rtp.socket[0], 
                              &st->rtp.socket[1])) != 0 ) {
        st->transport.server_port[1] =
          st->transport.server_port[0] + 1;

        /* Just store the destination */
        if ( st->transport.destination.s_addr == INADDR_ANY) {
          /*
           * destination is not present in SETUP.
           * Set it to st->sin.
           */
          st->transport.destination = st->sin.sin_addr;
        }

        /* Copy the server port */
        r->transport.server_port[0] = st->transport.server_port[0];
        r->transport.server_port[1] = st->transport.server_port[1];
        
        /* Start a thread that handles the packet delivery for this stream. */
        if ((e = pthread_create(&st->tid, NULL, Service, (void *)st)) != 0) {
          ResponseSetStatus(r, status=500, "Could not create thread");
          error("Setup:thread", "%s\n", strerror(e));
        }
        else {
          ResponseSetStatus(r, status = 200, NULL);

          HeaderDelete(&r->headers, "Session");
          HeaderPrintf(&r->headers, 0, "Session", "%lu;timeout=%d",
                       r->session->session_id, config.timeout);

          ResponseSetTransport(&r->headers, &r->transport);
        }
      }
      else {
        debug("Setup","start_listen_socket() failed\n");
        ResponseSetStatus(r, status = 500, "Socket failed");
      }
    }

    debug("Setup","mutex(st->lock) unlocked\n");
    pthread_mutex_unlock(&st->lock);
  }
  else {
    error("Setup", "stream exists\n");
    ResponseSetStatus(r, status = 500, "Stream already exists");
  }

  pthread_mutex_unlock(&r->session->lock);

  return status;
} /* Setup */

/*
 * Start listen port tries to allocate the consecutive port numbers
 * for RTP and RTCP and bind the appropriate sockets.
 * If the port is specified, it is preferred for
 * RTP port. All port numbers are in host byte order.
 * returns the allocated RTP port number. RTCP is RTP+1.
 * returns 0 in case of error. The *rtp_sock and *rtcp_sock
 * hold the socket ID of the RTP and RTCP sockets.
 * Set port to 0 if no preferred port.
 * This is copied from libmixer/video.c file.
 */
static unsigned short start_listen_port(unsigned short port,
                                 int * rtp_sock,
                                 int * rtcp_sock)
{
  struct sockaddr_in sin;
  int sinlen;
  int max_retry = 10;

  if ( rtp_sock == NULL || rtcp_sock == NULL ) {
    debug("start_listen_port", "One of the params are NULL\n");
    return 0;
  }

  /* While max retry */
  while ( --max_retry >= 0 ) {    

    /* Allocate the RTP and RTCP sockets */
    if (((*rtp_sock) = socket(AF_INET, SOCK_DGRAM, 0)) < 0 ) {
      debug("start_listen_port", "rtp_socket failed\n");
      return 0;
    }    
    if (((*rtcp_sock) = socket(AF_INET, SOCK_DGRAM, 0)) < 0 ) {
      closesocket(*rtp_sock);
      debug("start_listen_port", "rtcp socket failed\n");
      return 0;
    }
    
    /* Bind the RTP socket to the port */
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = INADDR_ANY;
    sin.sin_port = htons(port);
    if ( (bind(*rtp_sock, (struct sockaddr *)&sin, sizeof(sin))) < 0) {
      if ( port == 0 ) {
        /* If port was already 0 (any),
         * some serious error */
        closesocket(*rtp_sock);
        closesocket(*rtcp_sock);
        return 0;
      }
      else {
        /* We tried an non-0 port.
         * Try a 0 port now. */
        port = 0;
        debug("start_listen_port", "Requested port not bound. Trying any.\n");
        continue;
      }
    }

    /* Extract the port number now. If it were 0. */
    if ( port == 0 ) {
      sinlen = sizeof(sin);
      if ( getsockname(*rtp_sock, (struct sockaddr *)&sin, &sinlen) < 0 ) {
        debug("start_listen_port", "getsockname error\n");
      }
      port = ntohs(sin.sin_port);
    
      if ( port % 2 == 0 ) {
        /* Port is even number. Good */
      }
      else {
        /* Port is some odd number. Bad. */
        port = port + 1; /* make it even */
        closesocket(*rtp_sock);
        closesocket(*rtcp_sock);
        /* Continue with the next attempt */
        debug("start_listen_port", "Bound to odd port\n");
        continue;
      }
    }
    else {
      /* An non-0 value was specified for the 
       * port. Continue with rtcp binding */
    }

    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = INADDR_ANY;
    sin.sin_port = htons(port + 1);
    if ( (bind(*rtcp_sock, (struct sockaddr *)&sin, sizeof(sin))) < 0) {
      /* RTCP bind failed. Try again. */
      port = 0;
      closesocket(*rtp_sock);
      closesocket(*rtcp_sock);
      debug("start_listen_port", "rtcp bind failed\n");
      continue;
    }

    /* Everytthing is OK till now. */
    return port;
  }

  debug("start_listen_port", "Max retry exhausted\n");
  return 0;
}


#if 0
  sin = st->sin;
  sin.sin_port = htons(st->transport.client_port[0]);
  if((connect(st->rtp.socket[0], (struct sockaddr *) &sin, sizeof
    (struct sockaddr))) < 0) {
    error("Setup","RTP UDP connect");
    ResponseSetStatus(r, 500, "Could not connect RTP socket");
    return -1;
  }
  else {
    debug("Setup","RTP socket connected to %s:%d\n",
      inet_ntoa(sin.sin_addr),ntohs(sin.sin_port));
  }

  sin = st->sin;
  sin.sin_port = htons(st->transport.client_port[0]+1);
  if ((connect(st->rtp.socket[1], (struct sockaddr *)&sin, sizeof
    (struct sockaddr))) < 0) {
        error("Setup","RTCP UDP connect");
        ResponseSetStatus(r, 500, "Could not connect RTCP socket");
        return -1;
  }
  else {
    debug("Setup","RTCP socket connected to %s:%d\n",
      inet_ntoa(sin.sin_addr),ntohs(sin.sin_port));
  }
#endif
