/*
* Read and write Sun sound (.au, .snd) files.
*/

#define SND_INTERNAL
#include <stdio.h>         /* SEEK_CUR and SEEK_SET */
#include "sysdep.h"
#include "snd.h"
#include "dstring.h"
#include "rtp_pt.h"
#include <unistd.h>        /* lseek() */
#include <string.h>        /* strcmp() */
#include <unistd.h>        /* open() */
#include <stdlib.h>        /* calloc() */
#include <fcntl.h>         /* O_RDONLY, O_RDWR, O_WRONLY */
#include <sys/types.h>
#include <netinet/in.h>    /* ntohl() */
#include <sys/stat.h>
#include "error.h"         /* debug () */
#include "getusername.h"   /* getusername () */

typedef unsigned short u_int16;
typedef unsigned long  u_int32;

/*  
 * Used internally only; from multimedia/audio_filehdr.h;
 * all assumed to be in big endian order?
 * The info following is supposed to be zero-terminated.
 */   
typedef struct {
  u_int32 magic;          /* magic number */
  u_int32 hdr_size;       /* size of this header, with info (in bytes) */
  u_int32 data_size;      /* length of data (optional) */
  u_int32 encoding;       /* data encoding format */
  u_int32 sample_rate;    /* samples per second */
  u_int32 channels;       /* number of interleaved channels */
} audio_filehdr_t;

#define AUDIO_FILE_MAGIC 0x2e736e64   /* .snd */
/* Define the encoding fields */
#define AUDIO_FILE_ENCODING_MULAW_8       (1) /* 8-bit ISDN u-law */
#define AUDIO_FILE_ENCODING_LINEAR_8      (2) /* 8-bit linear PCM */
#define AUDIO_FILE_ENCODING_LINEAR_16     (3) /* 16-bit linear PCM */
#define AUDIO_FILE_ENCODING_LINEAR_24     (4) /* 24-bit linear PCM */
#define AUDIO_FILE_ENCODING_LINEAR_32     (5) /* 32-bit linear PCM */
#define AUDIO_FILE_ENCODING_FLOAT         (6) /* 32-bit IEEE floating point */
#define AUDIO_FILE_ENCODING_DOUBLE        (7) /* 64-bit IEEE floating point */
#define AUDIO_FILE_ENCODING_ADPCM_G721   (23) /* 4-bit CCITT g.721 ADPCM */
#define AUDIO_FILE_ENCODING_ADPCM_G722   (24) /* CCITT g.722 ADPCM */
#define AUDIO_FILE_ENCODING_ADPCM_G723_3 (25) /* CCITT g.723 3-bit ADPCM */
#define AUDIO_FILE_ENCODING_ADPCM_G723_5 (26) /* CCITT g.723 5-bit ADPCM */
#define AUDIO_FILE_ENCODING_ALAW_8       (27) /* 8-bit ISDN A-law */

typedef struct {
  int fd;                /* file descriptor */
  audio_filehdr_t hdr;   /* file header */
  int data_offset;       /* offset of data within file */
  int blk_size;          /* bytes to read at once */
  char *name;            /* name of encoding */
  double BytesPerSec;    /* bytes per second */
  int BitsPerSample;     /* bits per sample */
  double duration;       /* duration of audio, in seconds */
  int pt;                /* RTP payload type */
  char *info;            /* file information */
  u_int32 ts0;           /* initial timestamp */
  int     rw;            /* Mode O_RDONLY or O_WRONLY */
} Snd;


/*
* SndOpen()
* Open a sound file and return pointer to object.
* Track contains the information about the
* file in case of recording.
*/
Snd *SndOpen(char *path, char *track, int rw)
{
  int mode;
  int i;
  int len;            /* length of information field */
  Snd *snd;           /* snd file object */
  static struct {     /* table to map to RTP payload type and name */
    int encoding;
    int sample_rate;
    int channels;
    int BitsPerSample;
    int pt;
    char *name;
  } pt[] = {
    {AUDIO_FILE_ENCODING_MULAW_8,       8000, 1,  8, RTP_PT_PCMU,    "PCMU"},
    {AUDIO_FILE_ENCODING_MULAW_8,          0, 0,  8, 96,             "PCMU"},
    {AUDIO_FILE_ENCODING_ALAW_8,        8000, 1,  8, RTP_PT_PCMA,    "PCMA"},
    {AUDIO_FILE_ENCODING_ALAW_8,           0, 0,  8, 96,             "PCMA"},
    {AUDIO_FILE_ENCODING_LINEAR_8,         0, 0,  8, 96,             "L8"},
    {AUDIO_FILE_ENCODING_LINEAR_16,    44100, 2, 16, RTP_PT_L16_2,   "L16"},
    {AUDIO_FILE_ENCODING_LINEAR_16,    44100, 1, 16, RTP_PT_L16_1,   "L16"},
    {AUDIO_FILE_ENCODING_LINEAR_16,        0, 0, 16, 96,             "L16"},
    {AUDIO_FILE_ENCODING_LINEAR_24,        0, 0, 24, 96,             "L24"},
    {AUDIO_FILE_ENCODING_LINEAR_32,        0, 0, 32, 96,             "L32"},
    {AUDIO_FILE_ENCODING_FLOAT,            0, 0, 32, 96,             "FLOAT"},
    {AUDIO_FILE_ENCODING_DOUBLE,           0, 0, 64, 96,             "DBL"},
    {AUDIO_FILE_ENCODING_ADPCM_G721,    8000, 1,  4, RTP_PT_G726_32, "G726-32"}, 
    {AUDIO_FILE_ENCODING_ADPCM_G721,       0, 0,  4, 96,             "G726-32"}, 
    {AUDIO_FILE_ENCODING_ADPCM_G723_3,     0, 0,  3, 96,             "G726-24"},
    {AUDIO_FILE_ENCODING_ADPCM_G723_5,     0, 0,  5, 96,             "G726-40"},
    {-1,                                   0, 0,  0, 96,             ""}
  };

  snd = calloc(1, sizeof(Snd));
  mode = (rw == O_RDONLY) ? O_RDONLY : (O_RDWR | O_TRUNC);
  debug("SndOpen", "Mode is %x, %x, %x, %x\n",
        O_RDONLY, O_WRONLY, O_RDWR, mode);
#ifdef O_BINARY
  debug("SndOpen", "Adding O_BINARY\n");
  mode = mode | O_BINARY;
#endif    
  if ((snd->fd = open(path, mode)) < 0) {
    if ( rw == O_WRONLY ) {
      /* For recording create the file */
      int modeold;
      modeold = umask(0);
      snd->fd = open(path, O_WRONLY | O_CREAT, 0755); /* XXX what should be the
                                                        permissions bits */
      umask(modeold);
      close(snd->fd);

      snd->fd = open(path, mode);
      debug("SndOpen", "Opened record file %d inmode %x\n", snd->fd, mode);
    }
  }
  if ( snd->fd < 0 ) {
    free(snd);
    return NULL;
  }

  snd->rw = rw;
  snd->data_offset = lseek(snd->fd, 0, SEEK_CUR);
  if ( rw == O_RDONLY ) {
    if (read(snd->fd, &snd->hdr, sizeof(snd->hdr)) < (int)sizeof(snd->hdr))
      goto error;
    fprintf(stderr, "Read from file %x\n", snd->hdr.encoding);
  }
  else if (rw == O_WRONLY) {
    snd->info = strdup(track ? track : "No information"); /* No info stored */
    snd->hdr.channels    = htonl(1);
    snd->hdr.data_size   = htonl(0);
    snd->hdr.encoding    = htonl(AUDIO_FILE_ENCODING_MULAW_8);
    snd->hdr.hdr_size    = htonl(sizeof(snd->hdr) + strlen(snd->info) + 1);
    snd->hdr.magic       = htonl(AUDIO_FILE_MAGIC);
    snd->hdr.sample_rate = htonl(8000);
    write(snd->fd, &snd->hdr, sizeof(snd->hdr));
    write(snd->fd, snd->info, strlen(snd->info) + 1);
  }

  /*
   * Check magic number in file header to make sure we got the right
   * file type.
   */
  snd->hdr.magic = ntohl(snd->hdr.magic);
  /* If wrong magic number, restore file pointer. */
  if (snd->hdr.magic != AUDIO_FILE_MAGIC) {
    lseek(snd->fd, snd->data_offset, SEEK_SET);
    goto error;
  }
  /* Convert network byte order to host byte order. */
  snd->hdr.encoding    = ntohl(snd->hdr.encoding);
  snd->hdr.hdr_size    = ntohl(snd->hdr.hdr_size);
  snd->hdr.data_size   = ntohl(snd->hdr.data_size);
  snd->hdr.sample_rate = ntohl(snd->hdr.sample_rate);
  snd->hdr.channels    = ntohl(snd->hdr.channels);
  snd->data_offset     = snd->hdr.hdr_size;
  snd->blk_size        = 160;
  snd->pt              = 96; 
  fprintf(stderr,"headers are %d, %d %d\n", 
     snd->hdr.encoding, snd->hdr.sample_rate, snd->hdr.channels);
  snd->hdr.sample_rate = 8000; //SSS

  /* Find RTP payload type. */
  for (i = 0; pt[i].encoding >= 0; i++) {
    if ((pt[i].encoding == snd->hdr.encoding)
        && (pt[i].sample_rate == 0 || pt[i].sample_rate == snd->hdr.sample_rate)
        && (pt[i].channels    == 0 || pt[i].channels    == snd->hdr.channels)) {
     fprintf(stderr,"Found at index %d\n", i);
      snd->pt            = pt[i].pt;
      snd->name          = pt[i].name;
      snd->BitsPerSample = pt[i].BitsPerSample;
      snd->BytesPerSec   = pt[i].BitsPerSample * snd->hdr.sample_rate / 8;
      snd->duration      = snd->hdr.data_size * 8. / 
                           (snd->hdr.sample_rate * snd->hdr.channels * pt[i].BitsPerSample);
      break;
    }
  }

  fprintf(stderr,"Payload type is %d\n", snd->pt);

  /* Read file info, if it exists. */
  if ( rw == O_RDONLY) {
    len = snd->hdr.hdr_size - sizeof(snd->hdr);
    if (len < 0) goto error;
    if ((snd->info = calloc(len+1, sizeof(char))) == 0) goto error;
    if (len > 0) {
      if (read(snd->fd, snd->info, len) < len) goto error;
    }
  }

  return snd;

error:
  free(snd);
  return NULL;
} /* SndOpen */


/*
* SndDescribe()
* Return a description of a sound file. Return NULL if the type is not
* understood.
*/
char *SndDescribe(Snd *snd, char *type)
{
  DString d;
  char username[USERNAME_MAX];

  if ( snd == NULL ) {
    return NULL;
  }

  getusername(username, sizeof(username));

  DStringInit(&d);
  if (!strcmp(type, "application/sdp")) {
    DStringPrintf(&d, "v=0\r\n");
    DStringPrintf(&d, "o=%s %lu %lu IN IP4 %s\r\n",
      username[0] != '\0' ? username : "-", (long) time(0), (long) time(0), 
      "127.0.0.1");
    if (snd->info) {
      
      DStringPrintf(&d, "s=%s\r\n", snd->info);
    }
    else {
      /* s= line must be present */
      DStringPrintf(&d, "s=(no subject)\r\n");
    }
    DStringPrintf(&d, "b=CT:%.1f\r\n", snd->BytesPerSec * 8 / 1000.);
    DStringPrintf(&d, "t=0 0\r\n");
    DStringPrintf(&d, "a=length:npt=%.3f\r\n", snd->duration);
    DStringPrintf(&d, "m=audio %d RTP/AVP %d\r\n", 0, snd->pt);
    if (snd->pt > 95) {
      DStringPrintf(&d, "a=rtpmap:%d %s/%ld/%ld\r\n", snd->pt, snd->name,
      snd->hdr.sample_rate, snd->hdr.channels);
    }
  }
  return d.string;
} /* SndDescribe */


/*
* SndSeek()
* Position within sound file to given time value.  Return actual
* position, accounting for any block structure in file, if successful,
* -1 on error. Return duration if
* npt is negative.
*/
double SndSeek(Snd *snd, double npt)
{
  off_t pos;
  if ( snd == NULL ) {
    return 0.0;
  }
  if ( npt < 0.0 ) {
    return snd->duration;
  }

  pos = snd->data_offset + snd->BytesPerSec * npt;
  pos = lseek(snd->fd, pos, SEEK_SET);
  if ( snd->BytesPerSec > 0 ) {
    if ( pos > snd->data_offset ) {
      return (pos - snd->data_offset) / snd->BytesPerSec;
    }
    else {
      return 0.0;
    }
  }
  return 0.0;
} /* SndSeek */

/*
* SndRead()
* Read a block of snd data into 'buffer' and determine next playout
* time.  Set 'pt' to RTP payload type and 'tsinc' to time stamp
* increment.  Return number of bytes read.  Return 0 on EOF and -1 on
* error.
*/
int SndRead(Snd *snd, char *buffer, double *npt, int *pt, int *tsinc)
{
  int n;

  if ( snd == NULL ) {
    return -1;
  }

  n = read(snd->fd, buffer, snd->blk_size);

  lseek(snd->fd,0,SEEK_CUR);
  *pt    = snd->pt;
  *tsinc = (snd->blk_size * 8) / (snd->BitsPerSample * snd->hdr.channels);
  *npt   += *tsinc / (double)snd->hdr.sample_rate;
  if ( (n > 0) && (n < snd->blk_size)) {
    memset(&buffer[n],0,snd->blk_size - n);
    n = snd->blk_size;
  }
  return n;
} /* SndRead */


/*
 * SilencePacket()
 * Detects if the packet is a silence packet.
 * Returns 1 if it is a silence packet.
 * returns 0 if it is not.
 * It decodes the PCMU samples and tries to
 * figure out the energy value.
 */
static int SilencePacket(char *buffer, int size)
{
  int i;
  int above_threadhold_count = 0;

  return 0; /* No silence suppression now */

  for ( i=0; i<size; i++) 
  {
	    int		t;
      int val;
      unsigned char u_val = buffer[i];

	    /* Complement to obtain normal u-law value. */
	    u_val = ~u_val;

      #define	SIGN_BIT	(0x80)		/* Sign bit for a A-law byte. */
      #define	QUANT_MASK	(0xf)		/* Quantization field mask. */
      #define	NSEGS		(8)		/* Number of A-law segments. */
      #define	SEG_SHIFT	(4)		/* Left shift for segment number. */
      #define	SEG_MASK	(0x70)		/* Segment field mask. */
      #define	BIAS		(0x84)		/* Bias for linear code. */

	    /*
	     * Extract and bias the quantization bits. Then
	     * shift up by the segment number and subtract out the bias.
	     */
	    t = ((u_val & QUANT_MASK) << 3) + BIAS;
	    t <<= ((unsigned)u_val & SEG_MASK) >> SEG_SHIFT;
	    val = ((u_val & SIGN_BIT) ? (BIAS - t) : (t - BIAS));

      #define THREADHOLD 256
      #define MINSAMPLES  10

      /* val contains the decoded PCMU sample */
      if ( abs(val) >= THREADHOLD ) 
      {
        above_threadhold_count ++;
      }
  }

  if ( above_threadhold_count >= MINSAMPLES ) 
  {
    return 0;
  }
  else 
  {
    return 1;
  }

}

/*
* SndWrite()
* Write a block of snd data.  Return number of bytes written.  Return 0
* on EOF and -1 on error.
*/
int SndWrite(Snd *snd, int pt, int ts, int seq, char *buffer,
  int size, double *npt)
{
  int n;

  if ( snd == NULL ) {
    return -1;
  }

  if (snd->ts0 == 0) snd->ts0 = ts;
  if ( !SilencePacket(buffer,size) ) {
    n = write(snd->fd, buffer, size);

    snd->hdr.data_size += n; /* Size is in host byte order */
  }
  else {
    n = size; /* XXX fake that everything is written properly */
  }

  //*npt += size/(8000.0); // XXX  = (ts - snd->ts0) / (double)snd->hdr.sample_rate;
  return n;
} /* SndWrite */


/*
* SndClose()
* Close the file and release all resources.
*/
int SndClose(Snd *snd)
{
  if ( snd == NULL ) {
    return -1;
  }

  /* Update the file size for RECORD */
  if ( (snd->fd > 0) && (snd->rw == O_WRONLY) ) {
    int data_size;
    off_t dummy;

    /* get current data size */
    data_size = snd->hdr.data_size;
    snd->hdr.data_size = htonl(snd->hdr.data_size);

    if ( (dummy = lseek(snd->fd, 8, SEEK_SET)) == (off_t )-1) {
      debug("SndClose", "lseek failed\n");
    }
    debug("SndClose", "Writing size at offset %u\n", (unsigned int) dummy);
    write(snd->fd, &snd->hdr.data_size, 4);
    lseek(snd->fd, 0, SEEK_END);
    debug("SndClose", "Closing the record file with %d bytes\n", 
          data_size);
  }

  close(snd->fd);

  if (snd->info) {
    free(snd->info);
  }

  free(snd);
 
  return 0;
} /* SndClose */
