/*
* stream.c
*
* Manage individual streams within a session.  Streams are attached as a
* simple linked list to their session.
*/

#include "stream.h"
#include <stdio.h>
#include <stdlib.h>     /* calloc() */
#include <pthread.h>
#include <unistd.h>     /* close() */

extern u_int32 random32(int type);

/*
* RequestFree()
* Free request data if no more references to this request exist.
*/
void RequestFree(request_t *r)
{
  if (--r->reference <= 0) {
    HTTP_Free(&r->p);
    free(r);
  }
} /* RequestFree */

 
/*
* StreamCreate()
* Create a new stream. Return pointer to stream object.
*/
Stream *StreamCreate(uri_t uri)
{
  Stream *st;

  st = calloc(1, sizeof(Stream));

  st->URI           = URI_Copy(uri);
  st->npt           = -1.;
  st->state         = S_INVALID;
  st->rtp.socket[0] = -1;
  st->rtp.socket[1] = -1;
  st->subject = NULL;
  st->ts = random32(2);
  st->ssrc = random32(3);

  pthread_mutex_init(&st->lock, NULL);
  pthread_cond_init(&st->queue_change, NULL);
  pthread_cond_init(&st->pause, NULL);

  return st;
} /* StreamCreate */


/*
* StreamGet()
* Map a URI to the next stream in list pointed to by 'st' by using the
* search string, if available.  A URI of '*' matches the first stream.
*/
Stream *StreamGet(Stream *st, uri_t uri)
{
  if (*uri.scheme == '*') {
    return st;
  }
  else {
    for (; st; st = st->next) {
      if ((*st->URI.scheme == '*') || URI_Equal(st->URI, uri)) break;
    }
  }
  return st;
} /* StreamGet */


/*
* StreamByOwner()
* Is there a stream with the same URI?
*/
Stream *StreamByOwner(Session *s, uri_t uri, struct sockaddr_in sin, long seq)
{
  Stream *st;

  for (st = s->streams; st; st = st->next) {
    if (URI_Equal(uri, st->URI)) break;
  }
  return st;
} /* StreamByOwner */


/*
* StreamFree()
* Release a stream, its requests and its resources.
*/
int StreamFree(Stream *st)
{
  request_t *r, *r_next;

  for (r = st->requests; r; r = r_next) {
    r_next = r->next;
    RequestFree(r);
  }
  /* Release file system handler. */
  if (st->fp) {
    (*st->close)(st->fp);
    st->fp = NULL;
  }
  if ( st->subject ) {
    free(st->subject);
  }
  pthread_mutex_destroy(&st->lock);
  pthread_cond_destroy(&st->queue_change);
  pthread_cond_destroy(&st->pause);
  free(st);
  return 0;
} /* StreamFree */


/*
* StreamPrint()
* Print current list of streams to stdout for debugging.
*/
void StreamPrint(Stream *st)
{
  request_t *r;

  for (; st; st = st->next) {
    printf("  Stream %p: URI=%s port=%d\n", st,
      st->URI.path ? st->URI.path : st->URI.scheme, 
      st->transport.client_port[0]);

    for (r = st->requests; r; r = r->next) {
      printf("    Request %p for %s: %.3f %.3f...%.3f\n", 
        r, r->p.URI_string,
        r->range.effective, r->range.begin, r->range.end);
    }
  }
} /* StreamPrint */
