#ifndef STREAM_H
#define STREAM_H
#include <pthread.h>
#include <limits.h>
#include "rtsp.h"
#include "rtp_api.h"  /* conflict with rtplib/rtp_api.h */
#include "request-rtsp.h"
#include "uri.h"

struct Stream;
struct Session;

/*
* Each Session may consist of several streams.  A stream identifies a
* network association. A stream has a timeline and is played sequentially.
*/
typedef struct Stream
{
  pthread_mutex_t lock;            /* lock this data structure */ 
  pthread_cond_t queue_change;     /* condition variable: new request */
  pthread_cond_t pause;            /* condition variable: pause request */
  pthread_t tid;                   /* thread it for this stream */

  struct Session *session;         /* session that owns this stream */
  long seq;                        /* last request (for duplicate detection) */
  struct sockaddr_in sin;          /* origin of last request (for UDP) */
  enum {S_INVALID, S_SETUP, S_PLAY, S_RECORD, S_PAUSE, S_TEARDOWN} state;
  uri_t URI;                       /* URI for this stream, * if any */
  rtsp_transport_t transport;      /* transport parameters */
  double npt;                      /* current normal play time */
  char *subject;

  /* PLAY request queue, with currently playing request first. */
  request_t *requests;
  void *fp;                        /* opaque streaming file object */
  char filename[PATH_MAX];         /* current file name */
  int (*close)(void *);            /* function to close file */
  int (*seek)(void *, double npt); /* function for seeking */
  int (*read)(void *snd, char *buffer,  /* ... reading */
    double *npt, int *pt, int *tsinc);
  int (*write)(void *snd, int pt,       /* ... writing */
    u_int32 ts, u_int16 seq,
    char *buffer, int size, double *npt);

  /* RTP-related information */
  struct {
    int socket[2];                 /* sockets for RTP/RTCP data */
  } rtp;

  /* Timestamp and SSRC */
  int ts;
  int ssrc;

  struct Stream   *next;           /* next in list */
} Stream;


/*
* A session contains one or more streams. It is identified by a numeric
* session identifier issued by the server. 
*/
typedef struct Session
{
  pthread_mutex_t lock;            /* lock this data structure */ 
  u_int32 session_id;              /* session identifier */
  time_t last_request;             /* time of last request for session */
  Stream *streams;                 /* linked list of streams */
} Session;

extern void     RequestFree(request_t *r);

extern Session *SessionCreate(void);
extern int      SessionFree(int session_id);
extern Session *SessionGet(u_int32 session_id);
extern void     SessionAttach(Session *s, Stream *st);
extern void     SessionDetach(Session *s, Stream *st);
extern void     SessionPrint(void);

extern Stream  *StreamCreate(uri_t uri);
extern int      StreamFree(Stream *st);
extern Stream  *StreamGet(Stream *list, uri_t uri);
extern Stream  *StreamByOwner(Session *s, uri_t uri,
                  struct sockaddr_in sin, long seq);
extern void     StreamPrint(Stream *st);

extern void debug(char *function_name, char *format, ...);
extern void fatal(char *function_name, char *format, ...);
extern void error(char *function_name, char *format, ...);
#endif
