/*
* Read and write RIFF (WAV) files. For now, we write WAV files with a
* single data chunk. WAV files cannot record RTP sessions where there
* are more than one sender or where the sender switches data format or
* packetization interval.
*
* WAV (RIFF) files consist of chunks. 
struct {
  char magic[4] = "RIFF";   magic constant
  u_int32 length;           total length of file - 8
  char type[4]  = "WAVE";   designates as WAVE file
  union {
    struct {
      char type[4] = "fmt ";      type of chunk
      u_int32 length;             length in bytes
      u_int16 wFormatTag;         data format
      u_int16 wChannels;          number of channels
      u_int32 wSamplesPerSec;     samples per second per channel
      u_int32 wAvgBytesPerSec;    estimate of bytes per second
      u_int16 wBlockAlign;        byte alignment of a basic sample
      u_int16 wBitsPerSample;     bits per sample
      u_int16 wExtSize;           header extension size
      u_int16 wSamplesPerBlock;   samples per block
    } fmt_chunk;
    struct {
      char type[4] = "fact";      type of chunk
      u_int32 length;             length in bytes
      u_int32 dwFileSize;         number of samples
    } fact_chunk;
  } chunk[];
}
*/

#include <stdio.h>    /* fseek() */
#include <fcntl.h>    /* O_ */
#include <string.h>   /* strncmp() */
#include <stdlib.h>   /* calloc(), free() */


#define WAV_INTERNAL
#include "wav.h"
#include "rtp_pt.h"
#include "dstring.h"  /* DString..() */
#include "getusername.h" /* getusername() */
#include "error.h"    /* debug() */

typedef unsigned short u_int16;
typedef unsigned long u_int32;

typedef struct {
  FILE *f;                      /* file pointer */
  struct {
    u_int16 wFormatTag;         /* data format */
    u_int16 wChannels;          /* number of channels */
    u_int32 wSamplesPerSec;     /* samples per second per channel */
    u_int32 wAvgBytesPerSec;    /* estimate of bytes per second */
    u_int16 wBlockAlign;        /* bytes per block */
    u_int16 wBitsPerSample;     /* bits per sample */
    u_int16 wExtSize;           /* DVI: size of header extension */
    u_int16 wSamplesPerBlock;   /* DVI: number of samples in one block */
    u_int32 wFileSize;          /* bytes in file */
    char *name;                 /* name of encoding */
    double duration;            /* duration of audio, in seconds */
    int pt;                     /* RTP payload type */
    long data_offset;           /* start of 'data' chunk within file */
  } fmt;
  int pt;                       /* RTP payload type */
  char *name;                   /* name of encoding */
  double duration;              /* play duration */
  u_int32 ts0;                  /* initial timestamp */
} Wav;

#define WAVE_FORMAT_UNKNOWN     (0x0000)
#define WAVE_FORMAT_PCM         (0x0001)
#define WAVE_FORMAT_ADPCM       (0x0002)
#define WAVE_FORMAT_ALAW        (0x0006)
#define WAVE_FORMAT_MULAW       (0x0007)
#define WAVE_FORMAT_OKI_ADPCM   (0x0010)
#define WAVE_FORMAT_DVI_ADPCM   (0x0011)
#define WAVE_FORMAT_DIGISTD     (0x0015)
#define WAVE_FORMAT_DIGIFIX     (0x0016)
#define IBM_FORMAT_MULAW        (0x0101)
#define IBM_FORMAT_ALAW         (0x0102)
#define IBM_FORMAT_ADPCM        (0x0103)


/*
* readl16()
* Read 16-bit integer in little endian format.
*/
static u_int16 freadl16(FILE *f)
{
  unsigned char uc1, uc2;
  uc1 = getc(f); 
  uc2 = getc(f);
  return (uc2 << 8) | uc1;
} /* readl16 */


/*
* readl32()
* Read 32-bit integer in little endian format.
*/
static u_int32 freadl32(FILE *f)
{
  unsigned char u[4] = {0,0,0,0};

  if (fread(u, sizeof(char), 4, f) == 4) {
    return ((u_int32)u[3] << 24) | ((u_int32)u[2] << 16) |
           ((u_int32)u[1] <<  8) | ((u_int32)u[0]);
  }
  return 0;
} /* readl32 */


/*
* WavOpen()
* Open a .wav file and parse its header.
*/
Wav *WavOpen(char *path, char *track, int rw)
{
  Wav *wav;           /* .wav object */
  char magic[4];      /* chunk identifier text */
  unsigned int len;   /* length of chunk */
  char *mode;         /* file mode */
  int i;              /* loop variable */
  static struct {     /* table to map to RTP payload type and name */
    int encoding;
    int sample_rate;
    int channels;
    int bits_per_sample;
    int pt;
    char *name;
  } pt[] = {
    {WAVE_FORMAT_PCM,       8000, 1,  8, RTP_PT_PCMU,    "PCMU"},
    {WAVE_FORMAT_PCM,          0, 0,  0, 96,             "PCMU"},
    {WAVE_FORMAT_ADPCM,     8000, 1,  8, RTP_PT_DVI4_8,  "DVI4"},
    {WAVE_FORMAT_ADPCM,    11025, 1,  8, RTP_PT_DVI4_11, "DVI4"},
    {WAVE_FORMAT_ADPCM,    16000, 1,  8, RTP_PT_DVI4_16, "DVI4"},
    {WAVE_FORMAT_ADPCM,    22050, 1,  8, RTP_PT_DVI4_22, "DVI4"},
    {WAVE_FORMAT_ADPCM,        0, 0,  0, 96,             "DVI4"},
    {WAVE_FORMAT_ALAW,      8000, 1,  8, RTP_PT_PCMA,    "PCMA"},
    {WAVE_FORMAT_ALAW,         0, 0,  0, 96,             "PCMA"},
    {WAVE_FORMAT_MULAW,     8000, 1,  8, RTP_PT_PCMU,    "PCMU"},
    {WAVE_FORMAT_MULAW,        0, 0,  0, 96,             "PCMU"},
    {WAVE_FORMAT_OKI_ADPCM, 8000, 1,  8, RTP_PT_DVI4_8,  "DVI4"},
    {WAVE_FORMAT_DIGISTD,      0, 0,  0, 96,             "DIGISTD"},
    {WAVE_FORMAT_DIGIFIX,      0, 0,  0, 96,             "DIGIFIX"},
    {IBM_FORMAT_MULAW,      8000, 1,  8, RTP_PT_PCMU,    "PCMU"},
    {IBM_FORMAT_ALAW,       8000, 1,  8, RTP_PT_PCMA,    "PCMA"},
    {IBM_FORMAT_ADPCM,      8000, 1,  8, RTP_PT_DVI4_8,  "DVI4"},
    {-1,                       0, 0,  0, 96,             ""}
  };

  if      (rw == O_RDWR)   mode = "rb+";
  else if (rw == O_RDONLY) mode = "rb";
  else if (rw == O_WRONLY) mode = "wb";
  else {
    debug("WavOpen", "Unknown rw value %d\n", rw);
    mode = "rb";
  }

  wav = calloc(1, sizeof(Wav));
  if ((wav->f = fopen(path, mode)) == NULL) {
    free(wav);   
    return NULL;
  }

  /* Initial magic cookies in file. */
  if ((fread(magic, 1, 4, wav->f) != 4) || (strncmp(magic, "RIFF", 4))) {
    goto error;
  }
  len = freadl32(wav->f);
  if (((len = fread(magic, 1, 4, wav->f)) != 4) || (strncmp(magic, "WAVE",
4))) {
    goto error;
  }

  while (!feof(wav->f)) {
    if ((fread(magic, 1, 4, wav->f)) == 4) {
      len = freadl32(wav->f);
      if (!strncmp(magic, "fmt ", 4)) {
        wav->fmt.wFileSize          = 0;
        wav->fmt.wFormatTag         = freadl16(wav->f);
        wav->fmt.wChannels          = freadl16(wav->f);
        wav->fmt.wSamplesPerSec     = freadl32(wav->f);
        wav->fmt.wAvgBytesPerSec    = freadl32(wav->f);
        wav->fmt.wBlockAlign        = freadl16(wav->f);
        wav->fmt.wBitsPerSample     = freadl16(wav->f);
        wav->fmt.wSamplesPerBlock   = 1;
        wav->fmt.wExtSize           = 0;
        if (wav->fmt.wFormatTag == WAVE_FORMAT_DVI_ADPCM) {
          wav->fmt.wExtSize         = freadl16(wav->f);
          wav->fmt.wSamplesPerBlock = freadl16(wav->f);
          len -= 4;
        }

        len -= 16;  /* account for structure read */
        /* Map to RTP payload type. */
      }
      else if (!strncmp(magic, "fact", 4)) {
        wav->fmt.wFileSize = freadl32(wav->f);
        len -= 4;
      }
      else if (!strncmp(magic, "data", 4)) {
        /*
         * Uncompressed data doesn't have the 'fact' chunk, so we need
         * to compute this ourselves.
         */
        if (wav->fmt.wFileSize == 0) 
          wav->fmt.wFileSize = (len*8) / (wav->fmt.wBitsPerSample * wav->fmt.wChannels);
        wav->fmt.data_offset = ftell(wav->f);
        break;  /* skip seeking */
      }
      fseek(wav->f, len, SEEK_CUR);
    } /* if */
  } /* while */

  /* Find RTP payload type. */
  for (i = 0; pt[i].encoding >= 0; i++) {
    if ((pt[i].encoding == wav->fmt.wFormatTag)
        && (pt[i].sample_rate == 0 || pt[i].sample_rate == wav->fmt.wSamplesPerSec)
        && (pt[i].channels    == 0 || pt[i].channels    == wav->fmt.wChannels)) {
      wav->pt       = pt[i].pt;
      wav->name     = pt[i].name;
      wav->duration = wav->fmt.wFileSize / (double)wav->fmt.wSamplesPerSec;
      break;
    }
  }
/*
  wav->fmt.wFormatTag = WAVE_FORMAT_ADPCM;
  wav->pt   = RTP_PT_DVI4_22;
  wav->name = "None";
*/
  return wav;

error:
  free(wav);
  return NULL;
} /* WavOpen */


/*
* WavDescribe()
* Return a description of a sound file. Return NULL if the type is not
* understood.
*/
char *WavDescribe(Wav *wav, char *type)
{
  DString d;
  char username[USERNAME_MAX];

  getusername(username, sizeof(username));

  DStringInit(&d);
  if (!strcmp(type, "application/sdp")) {
    DStringPrintf(&d, "v=0\r\n");
    DStringPrintf(&d, "o=%s %lu %lu IN IP4 %s\r\n",
      username[0] != '\0' ? username : "-", (long)time(0), (long)time(0),
                  "127.0.0.1");
    DStringPrintf(&d, "b=CT:%.1f\r\n", wav->fmt.wAvgBytesPerSec * 8 / 1000.);
    DStringPrintf(&d, "t=0 0\r\n");
    DStringPrintf(&d, "a=length:npt=%.3f\r\n", wav->duration);

    DStringPrintf(&d, "m=audio %d RTP/AVP %d\r\n", 0, wav->pt);
    /* . XXX Refer WavRead() to get correct ptime value.
    DStringPrintf(&d, "a=ptime:%.3f\r\n", wav->fmt.wSamplesPerBlock / wav->fmt.wSamplesPerSec);
    */
    if (wav->pt > 95) {
      DStringPrintf(&d, "a=rtpmap:%d %s/%ld/%d\r\n", wav->pt, wav->name,
      wav->fmt.wSamplesPerSec, wav->fmt.wChannels);
    }
  }
  return d.string;
} /* WavDescribe */


/*
* WavSeek()
* Position within WAV file to given time value. Return actual position
* if successful, -1 on error.
*/
double WavSeek(Wav *wav, double npt)
{
  long offset;   /* byte offset into file */
  int blocks;    /* frame or block number */

  offset = wav->fmt.wAvgBytesPerSec * npt;
  /* 
   * For block-based audio such as DVI, we need to position to the
   * nearest block boundary.
   */
  blocks = offset / wav->fmt.wBlockAlign;
  offset = blocks * wav->fmt.wBlockAlign;
  npt = blocks * wav->fmt.wSamplesPerBlock / (double)wav->fmt.wSamplesPerSec;

  fseek(wav->f, offset + wav->fmt.data_offset, SEEK_SET);
  return npt;
} /* WavSeek */


/*
* WavRead()
* Read a block of wav data and determine next playout time. Return
* number of bytes read. Return 0 on EOF and -1 on error.
*/
int WavRead(Wav *wav, char *buffer, double *npt, int *pt, int *tsinc)
{
  int n = wav->fmt.wBlockAlign;       /* bytes to read */
  int s = wav->fmt.wSamplesPerBlock;  /* samples per block */

#if 1
  if (n < 32) {  /* r-kns10 */
    n = 160;  /* XXX make adjustable */
    s = (n * 8) / wav->fmt.wBitsPerSample;
  }
#endif

  n = fread(buffer, sizeof(char), n, wav->f);

  *pt    = wav->pt;
  *tsinc = s / wav->fmt.wChannels;
  *npt   += (*tsinc / (double)wav->fmt.wSamplesPerSec);

  /* XXX RTP packetization here (e.g., DVI conversion). */

  return n; 
} /* WavRead */


/*
* WavWrite()
* Write a block of wav data.  Return number of bytes written.  Return 0
* on EOF and -1 on error.  Set npt to time corresponding to beginning of
* block written.
*/
int WavWrite(Wav *wav, int pt, u_int32 ts, u_int16 seq, char *buffer,
  int size, double *npt)
{
  int n;

  /* XXX RTP conversion to storage format (e.g., DVI). May need to
     adjust block size. Probably won't work if block size varies. */

  if (wav->ts0 == 0) wav->ts0 = ts;
  ts -= wav->ts0;

  n = fwrite(buffer, sizeof(char), size, wav->f);
  *npt = (ts - wav->ts0) / (double)wav->fmt.wSamplesPerSec;
  return n;
} /* WavWrite */


/*
* WavClose()
* Close the .wav file and release all resources.
*/
int WavClose(Wav *wav)
{
  fclose(wav->f);
  free(wav);
  return 0;
} /* WavClose */
