#!/bin/sh
# The next line is executed by /bin/sh, but not Tcl \
exec tclsh $0 ${1+"$@"}

# Using the Sun Calendar manager, determine whether the callee is
# currently busy.  If busy, return a response that depends on whether
# the calendar entry is public, semi-private (time only), or private. 
# If public, try to find an email-like URL or a phone number and return
# it along with the current activity and the ending time for the
# appointment.  If semi-private, only return the ending-time. If
# private, only indicate that the user is busy.

# Limitations:
# - Relies on locally stored file in /usr/spool/calendar
#   rather than RPC since the latter isn,t documented.

# Ideas:
# - Incorporate priority value, so that only urgent calls interrupt
#   urgent appointments.


# Returns endtime of appointment if currently busy, 0 otherwise
proc busy {now starttime period nth ntimes duration exceptions} {
  if {$ntimes == 0} {
    set ntimes 1
  }
  for {set i 1} {$i <= $ntimes} {incr i} {
    if {[lsearch $exceptions $i] < 0} {
      set endtime [expr $starttime + $duration]
      if {$now >= $starttime && $now <= $endtime} {
        return $endtime
      }
    }
    switch $period {
      "single" {
      }
      "weekly" {
        incr starttime 604800
      }
      "tueThur" {
        set weekday [clock format $starttime -format "%w"]
        if {$weekday == 2} {
          incr starttime 172800
        } else {
          incr starttime 432000
        }
      }
      "monWedFri" {
        set weekday [clock format $starttime -format "%w"]
        if {$weekday < 5} {
          incr starttime 172800
        } else {
          incr starttime 259200 
        }
      }
      "monThruFri" {
        set weekday [clock format $starttime -format "%w"]
        if {$weekday < 5} {
          incr starttime 86400
        } else {
          incr starttime 259200 
        }
      }
      "everyNthDay" {
        incr starttime [expr $nth * 86400]
      }
      "everyNthWeek" {
        incr starttime [expr $nth * 86400 * 7]
      }
      "everyNthMonth" {
        set day   [clock format $starttime -format "%d"]
        set month [clock format $starttime -format "%m"]
        set year  [clock format $starttime -format "%Y"]
        set time  [clock format $starttime -format "%H:%M %Z"]
        regsub {0([1-9])} $month {\1} month
        incr month $nth
        set starttime [clock scan "$month/$day/$year $time"]
      }
      "biweekly" {
        incr starttime 1209600 
      }
      "monthly" {
        set day   [clock format $starttime -format "%d"]
        set month [clock format $starttime -format "%m"]
        set year  [clock format $starttime -format "%Y"]
        set time  [clock format $starttime -format "%H:%M %Z"]
        regsub {0([1-9])} $month {\1} month
        incr month
        if {$month > 12} {
          set month 1
          incr year
        }
        set starttime [clock scan "$month/$day/$year $time"]
      }
      "nthWeekday" {
        # add four weeks and check if the month has changed
        set month  [clock format $starttime -format "%m"]
        set day    [clock format $starttime -format "%d"]
        regsub {0([0-9])} $day {\1} day
        incr starttime 2419200
        set nextmonth [clock format $starttime -format "%m"]
        set nextday   [clock format $starttime -format "%d"]
        regsub {0([0-9])} $nextday {\1} nextday
        # if not, add another week
        if {$nextmonth == $month || [expr ($nextday-1)/7 < ($day-1)/7]} {
          incr starttime 604800
        }
      }
      "yearly" {
        set day   [clock format $starttime -format "%d"]
        set month [clock format $starttime -format "%m"]
        set year  [clock format $starttime -format "%Y"]
        set time  [clock format $starttime -format "%H:%M %Z"]
        incr year
        set starttime [clock scan "$month/$day/$year $time"]
      }
    }
  }
}

# get time to check
if {[llength $argv] > 0} {
  set now [clock scan [lindex $argv 0]]
} else {
  set now [clock seconds]
}

# Open calendar file after extracting local user
if {[info exists env(REQUEST_URI)]} {
  set uri $env(REQUEST_URI)
  regexp {([A-z.-]+)@} $uri user
} else {
  set user $env(USER)
}
set cal [open "/var/spool/calendar/callog.$user" r]

# Read appointment file
while {[gets $cal line] >= 0} {
  if {[regexp {\(add "([A-Z].*[0-9])"} $line dummy date] > 0} {
    set starttime [clock scan [lindex $line 1]]

    # extract appointment parameters
    set i 0
    set duration 0
    set ntimes   0
    set nth      0
    set period   "single"
    foreach e $line {
      incr i
      switch $e {
        "key:"        {set key  [lindex $line $i]}
        "what:"       {set what [lindex $line $i]}
        "privacy:"    {set privacy [lindex $line $i]}
        "duration:"   {set duration [lindex $line $i]}
        "period:"     {set period [lindex $line $i]}
        "nth:"        {set nth [lindex $line $i]}
        "ntimes:"     {set ntimes [lindex $line $i]}
      }
    }
    set exceptions {}
    regexp {exceptions: \(([0-9 ]+)\)} $line dummy exceptions

    # check if we are currently in this appointment
    set endtime [busy $now $starttime $period $nth $ntimes $duration $exceptions]
    if {$endtime > 0} {
      set appt($key) [list $starttime $endtime $privacy $what]
    }
  } elseif {[regexp {\(remove } $line dummy] > 0} {
    regexp {key: ([0-9]+)} $line dummy key
    if {[info exists appt($key)]} {
      unset appt($key)
    }
  }
}

if {[llength [array names appt]] == 0} {
  return 0
}

# Find longest duration of "covering" appointment, but use information
# of most recent appointment.
set endtime 0
set closest [expr $now + 1000000]
foreach a [array names appt] {
  if {[lindex $appt($a) 1] > $endtime} {
    set endtime [lindex $appt($a) 1]
  }
  if {[lindex $appt($a) 0] < $closest} {
    set best $a
    set closest [lindex $appt($a) 0]
  }
}

set endtime [lindex $appt($best) 1]
set privacy [lindex $appt($best) 2]
set what    [lindex $appt($best) 3]

# Check if description contains a SIP URL; if so, save and remove
set where ""
if {[regexp {[A-z0-9.-]+@[A-z0-9.-]+} $what where]} {
  regsub {[A-z0-9.-]+@[A-z0-9.-]+} $what "" what
}

# maybe a phone number?
set reg {[0-9][0-9 -]+}
if {[regexp $reg $what phone]} {
  regsub $reg $what "" what
  regsub -all {[ ]} $phone "-" phone
} else {
  set phone ""
}

# maybe a web page?
set reg {http://[A-z0-9./~-]+}
if {[regexp $reg $what http]} {
  regsub $reg $what "" what
} else {
  set http ""
}

regsub -all "\n" $what " " what
set what [string trim $what]

switch $privacy {
  "public"    {
    puts "SIP/2.0 302 Temporarily elsewhere: $what"
    if {$where != ""} { 
      puts "Contact: <sip:$where>"
    }
    if {$phone != ""} {
      puts "Contact: <phone:$phone>"
    }
    if {$http != ""} {
      puts "Contact: <$http>"
    }
    puts "Retry-After: [clock format $endtime -format "%a, %d %b %Y %T GMT" -gmt 1]"
  }
  "semiprivate" {
    puts "SIP/2.0 486 Busy"
    puts "Retry-After: [clock format $endtime -format "%a, %d %b %Y %T GMT" -gmt 1]"
  } 
  "private" {
    puts "SIP/2.0 486 Busy"
  }
}
