/***************************** FILE HEADER *****************************
#               Copyright (c) 1999 by Columbia University              *
#                        All Rights Reserved                           *
#***********************************************************************
#***********************************************************************
#
#  File Name    : main.cpp
#
#  Module       : Main file for testing SIP H323 gateway.
#
#  Author(s)    : Kundan Singh (kns10@cs.columbia.edu)
#
#  Created      : 16 Oct 1999
#
#  Description  : This file implements a SIP-H.323 gateway using the
#                 SIP and H.323 libraries.
#                  
#                
#  Notes        : 
#
#***********************************************************************
#                Modification History (Reverse Order)
#-----------------------------------------------------------------------
#   Date         Author              Description
#-----------------------------------------------------------------------
#              |             | 
# 27 June 2000 | Kundan      | Major reorganization + user reg.
#-----------------------------------------------------------------------
*/

/* Common includes */
#include <iostream.h>
#include <stdio.h>

#ifdef WIN32
#include <winsock2.h>
#include <time.h>
#else
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif

/* Portable library include */
#include <ptlib.h>

/* APIs for H323 and SIP */
#include "iptelapi.h"
#include "sipapi.h"
#include "h323api.h"

/** File globals for different IP addresses */
static PString h323Address;      // H323 endpoint
static PString sipAddress;       // SIP endpoint
static PString h323GKAddress;    // H.323 GK.
static PString sipServerAddress; // SIP registration server.
static PString domainName;       // Domain name for registration.
static bool    colocatedWithGK    = FALSE; 
static bool    colocatedWithProxy = FALSE;
static bool    mediaSwitchingHere = FALSE;

/* Global needed for SIP library */

/* Global for single call version */
bool oLicensed = false;
int nCallsCount = 0;

/* Program control macros */
#define NO_REINVITE  /* For disabling RE-INVITES.
                     * Note that MSF comes into picture only
                     * when reINVITEs are disabled. */

#ifdef NO_REINVITE

class MediaSwitchingThread : public PThread
{

public:
  MediaSwitchingThread();
  ~MediaSwitchingThread();
  void   Main();
  void SetRedirectAddress(struct sockaddr_in & addr);
  unsigned short GetSelfPort();
  void Terminate();

private:
  static unsigned short beginat; /* Port allocation starts from here
                                  */
  static bool firsttime;
  unsigned short selfport;
  struct sockaddr_in remoteaddr;
  int rtp; /* rtp socket */
  int rtcp; /* rtcp socket */
  BOOL oTerminate;
};

/* Start allocating RTP ports for MSF after 13000 */
unsigned short MediaSwitchingThread::beginat = 13000;
bool MediaSwitchingThread::firsttime = true;

#endif

extern "C"   int debug_flag;

extern "C"
{
  /*
   * SIPLibrary debugging related functions 
   */
  void DebugInit(char *);
  void ErrorInit(char *);
  void pthread_log_print(int type);
  #ifdef WIN32
  void pthread_log_print(int type);
  #endif
  int MessageFlowFile(FILE *fp); 
  void RequestPrint(FILE *fp);
}

class MySIPCall; // forward declaration.
class MyH323Call; // forward declaration.

/**
 * Thread to handle SIP requests, this
 * is needed because SIP library does not use PThread
 * while pwlib->Sockets need PThread::Current()
 * All pwlib functions should be called from within a PThread.
 * Since the callback SIPCall::OnIncomingCall is called
 * from a normal thread (pthread), another thread SIPThread
 * which is a subclass of PThread is invoked to process a
 * SIPCall::OnIncomingCall().
 */

class SIPThread : public PThread
{
public:
  SIPThread(MySIPCall & call);
  ~SIPThread() {};
  void   Main();
  void   OnHangup();

  /* 
   * IsWaiting() returns TRUE if SIP call is not hungup 
   */
  void SetWaiting(BOOL wait);

  /* connections-- mutex */
  PSyncPoint flag;

private:
  /* Associated SIP call */
  MySIPCall & call;

  /* Remote and self SIP addresses */
  IptelAddress remote;
  IptelAddress self;

  /* waiting == TRUE means SIP call is not hung up */
  BOOL            waiting;
};

/*
 * Thread to call H323 hangup if SIP call is closed 
 */
class SIPHangupThread : public PThread
{
public:
  SIPHangupThread();
  ~SIPHangupThread() { };
  void   Main();
  /* Associated H323 call, XXX passthis is constructor and make private */
  MyH323Call * h323;
};

/**
 * MySIPUserAgent and MySIPCall
 */

class MySIPUserAgent : public SIPEndpoint
{
public:
  MySIPUserAgent(bool IsServer = true,bool IsRegistrar=false);
  virtual ~MySIPUserAgent() { };
  virtual IptelCall *AllocateNewCallObject(const char *callid);
};


class MySIPCall : public SIPCall
{
public:
  MySIPCall(const char *callid = NULL) : SIPCall(callid) { sip=NULL; };
  MySIPCall(SIPEndpoint & ep, const char *callid=NULL) : SIPCall(ep,callid) { sip=NULL; };
  virtual ~MySIPCall() { };
  virtual void   OnNewIncomingCall(IptelEndpoint & ep);
  virtual int    OnHangup();
  /* Make sure that the library does not free the call object when it
   * is rejected */
  virtual int    OnCallRejected() { SIPCall::OnCallRejected(); return -1; }
  SIPThread      *sip; /* SIP object associated with this call */

  MyH323Call *h323call;
};

/**
 * MyH323Server and MyH323Call
 */
class MyH323Server : public H323_Endpoint
{
public:
  MyH323Server(bool IsServer = true, bool IsRegistrar = true);
  MyH323Server(const IptelIPAddress &, bool IsRegistrar = true);
  virtual ~MyH323Server() { };
  virtual IptelCall *AllocateNewCallObject(const char *callid);

  /* Gatekeeper/user registration related methods */
  virtual int OnLocationRequest(
                                const IptelAddress & alias,
                                IptelAddress & transport);
  virtual int OnIncomingRegistration(
               const IptelEndpoint::RegistrationInfo & reginfo);
  virtual void OnIncomingUnregistration(
               const IptelEndpoint::RegistrationInfo & reginfo);
};

class MyH323Call : public H323Call
{
public:
  MyH323Call();
  MyH323Call(MyH323Server &);
  virtual ~MyH323Call() { };
  virtual void   OnNewIncomingCall(IptelEndpoint & ep);
  virtual int    OnHangup();
  /* Make sure that the call object is not deleted when call is rejected */
  virtual int    OnCallRejected() { H323Call::OnCallRejected(); return -1; }
  virtual int    Hangup();

  MySIPCall *sipcall;
  /* 
   * IsWaiting() returns TRUE if SIP call is not hungup 
   */
  void SetWaiting(BOOL wait);

  /* connections-- mutex */
  PSyncPoint flag;

private:
  MediaSwitchingThread *msf;
};

/////////////// SIP and H323 endpoints /////////////////////
MySIPUserAgent *SIP = NULL;
MyH323Server   *H323 = NULL;

//////////////// Method defn for SIP ///////////////////////

MySIPUserAgent::MySIPUserAgent(bool IsServer,bool IsRegistrar)
: SIPEndpoint(IsServer,IsRegistrar)
{
}

IptelCall *MySIPUserAgent::AllocateNewCallObject(const char *callid)
{
  return new MySIPCall(*this, callid);
}


void MySIPCall::OnNewIncomingCall(IptelEndpoint & /* ep ignored */)
{
  h323call = NULL;

  sip = new SIPThread(*this);
  sip->Resume();
}

SIPThread::SIPThread(MySIPCall &c)
  : PThread(10000), call(c)
{
}

void SIPThread::SetWaiting(BOOL wait)
{
  waiting = wait;
}

/**
 * Processing Incoming call from SIP side.
 */

void SIPThread::Main()
{
  waiting = TRUE;

  cout << "Call from SIP side: " << call << endl;

  if ( !oLicensed && nCallsCount >= 1) {
    call.Reject(400, "No more call can be accepted by GW");
    call.sip = NULL;
    return;
  } 

  nCallsCount = 1;

  call.h323call = new MyH323Call;
  call.h323call->sipcall = &call;

  IptelAddress src = call.GetRemoteAddress();
  IptelAddress dest = call.GetRequestURI();
  call.h323call->SetSelfAddress(src);

  if ( h323Address != "" ) {
    /* 
     * Preconfigured H323 address.
     * Set the proxy address to this and Initiate the call 
     * The destination address is unchanged. It is upto the
     * destination to interpret the SIP URI.
     */
    /* XXX should we call dest.SetScheme(NULL) */
    cout << "Preconfigured H.323 address " << h323Address << endl;
    call.h323call->SetRemoteAddress(dest);
    call.h323call->SetProxyAddress(IptelAddress(h323Address));
  }
  else {

    /* Set the scheme to NULL */
    dest.SetScheme(NULL);

    /*
     * Remove the Host name is the host name is same as that
     * of the local machine name.
     */
    char *destH = dest.GetHost();
    PString destHost = destH ? destH : "";
    if ( destH ) free(destH);

    if ( H323_Endpoint::IsLocalHost(destHost) ) {
      cout << "Destination  " << dest << " has local host name." << endl;
      char *user = dest.GetUser();
      if ( user != NULL ) {
        dest = user;
        free(user);
        cout << "New Destination; " << dest << endl;
      }
    }

    IptelEndpoint::RegistrationInfo reginfo;

    if ( H323 == NULL ) {
      cerr << "H323 is NULL" << endl;
    }

    if ( colocatedWithGK &&  
         H323->GetRegistration(dest,
                              reginfo) > 0 ) {
      /*
       * H.323 GK is colocated with the SIP323 gateway.
       * Check if the address is present in the
       * database.
       */
      /*
       * Registration found in the GK.
       * Set it to proxy and make the call.
       */
      cout << "Registration found in local GK " <<  endl;
      call.h323call->SetProxyAddress(reginfo.contactAddress);
      call.h323call->SetRemoteAddress(dest);
    }
    else {
      /* No such registration found in the gatekeeper */
      /*
       * Check if the dest name is an IP address.
       * If yes then make the call.
       */
      PString destHostStr = dest.GetAddressString();
      PString addr;

      cout << "Resolving host name for " << destHostStr << endl;

      if ( H323_Endpoint::GetHostAddress(destHostStr, addr)) {
        /* This is a DNS or IP address */
        cout << "Dialing the machine " << destHostStr << endl;
        dest = (const char *) addr;
        cout << "With IP address " << dest << endl;
        call.h323call->SetRemoteAddress(dest);
        /* No need to set proxy address */
      }
      else {
        /* This is not a DNS or IP address */
        /*
         * Try LRQ request 
         */
        IptelAddress addr;
        if ( H323->LocationRequest(dest, addr) ) {
          /* LCF received */
          cout << "LocationRequest successful " << addr << endl;
          call.h323call->SetProxyAddress(addr);
          call.h323call->SetRemoteAddress(dest);
        }
        else {
          /* Should reject the call */
          cerr << "Can not resolve address, rejecting." << endl;
          call.Reject(400, "Not Found");
          delete call.h323call;
          call.h323call = NULL;
          nCallsCount = 0;
          call.sip = NULL;
          return;
        }
      }
    }
  }


  /* Set the session description */

  SessionDescription *session;
  
  session = call.GetRemoteSessionDescription();
  if ( session )
  {
    cout << *session;
    /* Give IP address of this machine */
    session = session->Clone();    

    call.h323call->SetSelfSessionDescription(session);
    //cout << "Session is " << *session << endl;
  }

  dest = call.h323call->GetRemoteAddress();
  src = call.h323call->GetSelfAddress();
  cout << "Initiating H.323 call" << endl;

  call.h323call->Initiate(dest,src);

  cout << "Waiting for response" << endl;

  call.h323call->WaitForResponse(-1);

  if ( (call.h323call != NULL) && (call.h323call->IsActive()) )
  {
    cout << dest << " accepted the call" << endl;

    session = call.h323call->GetRemoteSessionDescription();

    if ( session == NULL )
    {
      call.h323call->WaitForOpenLogicalChannelAcks(-1);
      session = call.h323call->GetRemoteSessionDescription();
    }

    if ( session != NULL )
    {
      cout << *session;

      session = session->Clone();

      call.SetSelfSessionDescription(session);

      call.Accept();
    }
    else
    {
      cout << "Session Description is absent; closing..." << endl;
      call.Reject();
      if ( call.h323call != NULL ) {
        call.h323call->sipcall = NULL;
        call.h323call->Hangup();
        Sleep(5000);
        delete call.h323call;
        call.h323call = NULL;
      }
      nCallsCount = 0;
      call.sip = NULL;
      return;
    }
  }
  else {
    
    cout << dest << " rejected the call" << endl;
    
    call.Reject();
    if ( call.h323call != NULL ) {
      call.h323call->sipcall = NULL;
      delete call.h323call;
      call.h323call = NULL;
    }
    nCallsCount = 0;
    call.sip = NULL;
    return;
  }

  if ( waiting ) { 
    flag.Wait();
    cout << "Wait exiting" << endl;
  }

  if ( call.h323call != NULL )
  {
    call.h323call->sipcall = NULL; /* Make SIP call of H323 NULL */
    call.h323call->Hangup();
    Sleep(5000);
    cout << "Deleting H.323 call " << endl;
    delete call.h323call;
    call.h323call = NULL;
  }

  call.sip = NULL; /* Set the thread ID to NULL */
  nCallsCount = 0;

  /*
   * Should sleep for some time 
   */
  cout << "Thread sleeping for 10 sec." << endl;
  Sleep(10000);
  cout << "Sleep exiting" << endl;
}

int   MySIPCall::OnHangup()
{
  cout << "SIP call closed" << endl;

  nCallsCount = 0;

  if ( sip == NULL ) {
    /* Initiated by H.323 side */
    cout << "Call was initiated by H.323" << endl;
    if ( h323call != NULL ) {
      cout << "Closing H323 call" << endl;
      
      h323call->sipcall = NULL; /* Set SIP call to NULL */
  
      /* Create a thread to call hangup */
      SIPHangupThread *s;
      s = new SIPHangupThread();
      s->h323 = h323call;
      s->Resume();
      /* Sleep(5000), delete h323call; */
      h323call = NULL;
    }
    else {
      cout << "H.323 call does not exists." << endl;
    }
  }
  else {
    /* Initiated by SIP side */
    h323call->sipcall = NULL;
    sip->SetWaiting(FALSE);
    sip->flag.Signal();
    sip = NULL;
  }

  cout << "Calling SIPCall::OnHangup" << endl;
  SIPCall::OnHangup();
  cout << "done" << endl;
  return 0;
}

void SIPHangupThread::Main()
{
  PTRACE(3, "SIPHangupThread() started");
  h323->Hangup();
}

SIPHangupThread::SIPHangupThread()
  : PThread(10000)
{
}

///////////////////// H323 Definitions ///////////////////////
MyH323Server::MyH323Server(bool IsServer, bool IsRegistrar)
: H323_Endpoint(IsServer, IsRegistrar)
{
  /* Do nothing */
}

MyH323Server::MyH323Server(const IptelIPAddress & listeningAddress, bool IsRegistrar)
: H323_Endpoint(listeningAddress, IsRegistrar)
{
  /* Do nothing */
}

IptelCall *MyH323Server::AllocateNewCallObject(const char *callid)
{
  return new MyH323Call(*this);
}

int MyH323Server::OnLocationRequest(
                      const IptelAddress & alias,
                      IptelAddress & transport)
{
  /*
   * XXX send OPTIONS message to SIP network.
   * Wait for response. If 2xx response received then
   * return 0 else return -1;
   */
  cout << "Incoming H.323 LRQ is not implemented" << endl;
  return -1;
}

int MyH323Server::OnIncomingRegistration(
                           const IptelEndpoint::RegistrationInfo & reginfo)
{
  /* If SIP endpoint is created and
   * a SIP registrar is configured then
   * forward the registration to the SIP registrar */

  cout << "On Incoming H.323 registration" << endl;

  if ( SIP != NULL && sipServerAddress != "" ) {
    SIPRegistrationInfo reg;
    reg.registrar = (const char *) sipServerAddress;
    reg.registrar.SetScheme("sip");

    cout << "Registrar " << reg.registrar << endl;

    /* Do not copy registrar */
    /* Set contact as this endpoint */
    reg.contactAddress = SIPLibrary::HostName;

    cout << "Contact " << reg.contactAddress << endl;

    /* For telephone numbers add host name and scheme */
    if ( reginfo.localName.IsTelephoneNumber()) {
      IptelAddress registrar = (const char *) domainName;
      char *host = registrar.GetHost();
      /* XXX should use domain name cs.columbia.edu */
      reg.localName.SetHost(host != NULL ? host : SIPLibrary::HostName);
      if ( host ) {
        free(host);
      }
      reg.localName.SetScheme("sip");
      reg.localName.SetUser(reginfo.localName.GetAddressString());
      cout << "Telephone number " << reg.localName << endl;

    }
    else {
      /* For others just register the alias given */
      reg.localName = reginfo.localName;
      reg.localName.SetScheme("sip");
      char *user = reg.localName.GetUser();
      if ( user == NULL ) {
        char * host = reg.localName.GetHost();
        /* Set host as server name and user as this alias */
        reg.localName.SetUser(host);
        if ( host) free(host);
        IptelAddress registrar = (const char *) domainName;
        host = registrar.GetHost();
        /* XXX should use domain name, cs.columbia.edu */
        if (host ) {
          reg.localName.SetHost(host);
          free(host);
        }
      }
      else {
        free(user);
      }
      cout << "Alias " << reg.localName << endl;
    }

    SIP->Register(reg);
    return 0; /* Return success even if SIP registration fails */
  }
  else {
    if ( SIP == NULL ) {
      cerr << "SIP323:OnIncomingRegistration, " <<
        "No SIP endpoint created yet." << endl;
    }
    else {
      cerr << "No SIP server configured, Registration is local" << endl;
    }
    return 0; /* 
               * just accept the registration without forwrding to SIP
               */
  }
}

void MyH323Server::OnIncomingUnregistration(
                              const IptelEndpoint::RegistrationInfo &
                              reginfo)
{
  /* Send Register() with zero expiry timeout. XXX */
  cout << "Incoming H.323 Unregistration is not implemented" << endl;
  return;
}


MyH323Call::MyH323Call()
: H323Call()
{
  msf = NULL;
  sipcall = NULL;
}

MyH323Call::MyH323Call(MyH323Server &ep)
: H323Call(ep)
{
  msf = NULL;
  sipcall = NULL;
}

void MyH323Call::OnNewIncomingCall(IptelEndpoint & /* ep ignored */)
{
  /* Similar to SIP */

  cout << "Call from H.323 side: " << *this << endl;

  if ( !oLicensed && nCallsCount >= 1 ) {
    /* Only one call supported in unlicensed version */
    Reject();
    return;
  }

  nCallsCount = 1;

  bool h323SessionAvailable = false;

  sipcall = new MySIPCall(*SIP);
  sipcall->h323call = this;

#if (defined NO_REINVITE)
  /* IP address where RTP mapper is running */
  IptelIPAddress thisIpAddr(SIPLibrary::HostInAddr);
#endif

  IptelAddress src           = GetRemoteAddress();
  IptelAddress dest          = GetSelfAddress();
  char *srcuser = src.GetUser();
  if ( srcuser == NULL ) {
     //SSS
     char *srchost = src.GetHost();
     src.SetUser(srchost);
     src.SetHost(SIPLibrary::HostName);
     if ( srchost == NULL ) {
       free(srchost);
     }
  }
  else {
    free(srcuser);
  }

  IptelAddress remoteAddress;
  remoteAddress = sipAddress;
  remoteAddress.SetScheme("sip");

  /* If no address is given for the destination
   * Then set the destination to preconfigured sipAddress, if present. */
  if ( strcmp(dest.GetAddressString(),"unknown") == 0) {
    if ( sipAddress != "") {
      dest = remoteAddress;
      cout << "Setting destination to " << dest << endl;
    }
  }

  if ( sipAddress != "" ) {
    /*
     * SIP address is preconfigured. Set the proxy address
     * (requestURI) to this
     */
    sipcall->SetProxyAddress(remoteAddress);
    cout << "Setting proxy address to " << remoteAddress << endl;
  }

  /*
   *  XXX if dest is same as this GW address then
   *  if the remote Extension address is present use that
   *  else set the dest to remote Address 
   */
  
  if ( dest.IsTelephoneNumber() ) {
    /* 
     * If destination is a telephone number
     * without any host part then
     * make it telephonenumber@server.
     */
    cout << "Destination " << dest << " is a telephone number" <<
      endl;

    if ( sipAddress != "" ) {
      /* Telephone number becomes the user name */
      dest.SetUser(dest.GetAddressString());
      /* Host portion is copied from sipAddress */
      IptelAddress preconfAddress = (const char *) sipAddress;
      dest.SetHost(preconfAddress.GetHost());
    }
    else {
      cout << "There is no preconfigured SIP address, rejecting" <<
        endl;
      Reject();
      cout << "Deleteing SIP call " << endl;
      delete sipcall;
      sipcall = NULL;
      nCallsCount = 0;
      return;
    }
  }

  src.SetScheme("sip"); /* Source address is made as a SIP address */
  dest.SetScheme("sip"); /* destination address is also a SIP address */

  sipcall->SetRemoteAddress(dest);
  sipcall->SetSelfAddress(src);

  cout << "Making SIP call to " << dest << " from " << src << endl;

  /* Set the session description */

  SessionDescription *session;
  
  session = GetRemoteSessionDescription();

  if ( session )
  {
    cout << *session;

    h323SessionAvailable = true;

    /* Give IP address of this machine */
    session = session->Clone();    

    sipcall->SetSelfSessionDescription(session);
  }
  else
  {
#ifdef NO_REINVITE
    /* If RTP is here then fake a media description */
    SessionDescription *session1;
    session1 = SIPEndpoint::GetDefaultSessionDescription();
    session  = new SessionDescription();
    *session = *session1;
    session->DeleteSession(); /* remove all */

    unsigned short port;
    MediaInfo *media=NULL;

    for ( int i=0; i<session1->GetMediaCount(); i++) 
    {
      MediaInfo media = (*session1)[i];
      if ( media.GetDataType() == MediaInfo::Audio &&
           media.GetDirection() == MediaInfo::Receive)
      {
        msf = new MediaSwitchingThread();
        port = msf->GetSelfPort();
        media.SetRxAddress(thisIpAddr);
        media.SetRxPort(port);
        session->AddSession(media);
        break; /* XXX only one media supported */
      }
    }
    delete session1;

    sipcall->SetSelfSessionDescription(session);
    cout << *session;
#endif
    
    cout << "Session description in H323 is absent" << endl;
  }

  dest = sipcall->GetRemoteAddress();

  src = sipcall->GetSelfAddress();

  cout << "Initiating SIP call with src " << src << " and dest " <<
    dest << endl;


  sipcall->Initiate(dest,src);

  sipcall->WaitForResponse(-1);

  if ( sipcall->IsActive() )
  {
    cout << dest << " accepted the call" << endl;

    session = sipcall->GetRemoteSessionDescription();

    if ( session != NULL )
    {
      cout << *session;

      session = session->Clone();

      SetSelfSessionDescription(session);

      cout << "Accepting H323 Call" << endl;

      Accept();

      /* For v1.0 without fast start */
      if ( h323SessionAvailable == false ) 
      {
        if ( (session = GetRemoteSessionDescription() ) == NULL )
        {
          cout << "Waiting for H323 OLC Acks" << endl;

          WaitForOpenLogicalChannelAcks(-1);

          cout << "OLC Acks received" << endl;
          
          session = GetRemoteSessionDescription();
        }

        if ( session != NULL )
        {
#ifdef NO_REINVITE

          for ( int k=0; k<session->GetMediaCount(); k++ ) 
          {
            MediaInfo & media = (*session)[k];
            if ( media.GetDirection() == MediaInfo::Receive)
            {
              struct sockaddr_in saddr;
              
              IptelIPAddress ipaddr = media.GetRxAddress();
              
              ipaddr.GetIpAddress(saddr);
              saddr.sin_family = AF_INET;
              saddr.sin_port = htons(media.GetRxPort());

              if ( msf != NULL ) {
                msf->SetRedirectAddress(saddr);
                msf->Resume();
              }
              break;
            }
          }

#else
          session = session->Clone();

          sipcall->SetSelfSessionDescription(session);

          cout << "Reinviting with new Session Descrition" << endl;

          sipcall->Reinvite();
#endif
        }
        else
        {
          cout << "Session Description is absent in H323 after OLC acks" << endl;
        }
      }
      /* Opened successfully */
      return;
    }
  }


  cout << dest << " rejected the call" << endl;

  nCallsCount = 0;
  Reject();
  delete sipcall;
  sipcall = NULL;

}

int MyH323Call::Hangup()
{
  if ( msf != NULL ) {
    msf->Terminate();
    //delete msf;  should not delete an AutoDeleteThread
    msf = NULL;
  }

  nCallsCount = 0;
  cout << "Closing H.323 call" << endl;

  return H323Call::Hangup();
}

int   MyH323Call::OnHangup()
{
  nCallsCount = 0;

  if ( msf != NULL ) {
    msf->Terminate();
    //delete msf; should not delete an autodeletthread
    msf = NULL;
  }

  cout << "H.323 Call closed" << endl;

  if ( sipcall != NULL )
  {
    if ( sipcall->sip == NULL ) {
      /* Call started by H.323 */
      cout << "Hanging up SIP call" << endl;
      sipcall->h323call = NULL; /* reset back pointer */
      sipcall->Hangup();
      sipcall = NULL;
    }
    else {
      /* Call started by SIP */
      cout << "Signaling SIP thread to hangup" << endl;
      sipcall->h323call = NULL; /* reset back pointer */
      sipcall->sip->SetWaiting(FALSE);
      sipcall->sip->flag.Signal();

      sipcall->Hangup();
      sipcall = NULL;
    }
  }
  H323Call::OnHangup();
  return 0;
}

////////////////// Main routine /////////////////////

/*
 * Test application
 */
class TestApp : public PProcess
{
    PCLASSINFO(TestApp, PProcess)
  public:
    TestApp();
    ~TestApp();

    void Main();
};

PCREATE_PROCESS(TestApp);

TestApp::TestApp()
  : PProcess("Columbia University", "SIP323",
             1, 0, AlphaCode, 1)
{
}

TestApp::~TestApp()
{
}

void TestApp::Main()
{

  PString localname;

  /* Copied from OpenH323 */
  PArgList & args = GetArguments();
  args.Parse("H-h323-address:"
             "S-sip-address:"
             "g-h323-gk-address:"
             "p-sip-server-address:"
             "i-interactive-mode."
             "v-verbose-mode."
             "n-localname:"
             "d-domain-name:"
             "r-no-registration."
             "a-with-colocated-h323-gk."
             "b-with-colocated-sip-proxy."
             "m-with-media-through-this."
             "x-display-H323-trace-on-screen."
             "y-display-SIP-trace-on-screen."
             "l-license-file:"
             "k-dotted-ip."
             "h-help."
          , FALSE);

  cout << GetName() 
       << " Version " << GetVersion(TRUE)
       << " by " << GetManufacturer()
       << " on " << GetOSClass() << ' ' << GetOSName()
       << " (" << GetOSVersion() << '-' << GetOSHardware() << ')' << endl;

  if ( args.HasOption('h') ) {
    /* Print usage and exit */
    cerr << "Usage: " << GetFile().GetFileName() << 
      " [options]\n"
      "Options:\n"
      "   -i           : Interactive mode\n"
      "   -v           : verbose mode\n"
      "   -n <name>    : Name of the gateway, e.g., abc@cs.columbia.edu\n"
      "   -H <address> : Address of preconfigured H.323 terminal (IP or hostname)\n"
      "   -S <address> : Address of preconfigured SIP user agent \"user@domain\"\n"
      "   -g <address> : Address of H.323 gatekeeper (IP or hostname)\n"
      "   -p <address> : Address of SIP server to register with (IP or hostname)\n"
      "   -d <address> : Domain name for registration, e.g,\n"
      "                  cs.columbia.edu, without any scheme or user name\n"
      "   -r           : Do not register with H.323 gatekeeper\n"
      "   -a           : Use In-built H.323 gatekeeper in this gateway\n"
      "                  Presence of -a, ignores -g option. -r is implied\n" 
      "   -b           : Use In-built SIP Proxy server in this gateway \n"
      "   -m           : Have media packet through this gateway\n"
      "   -x  {with -v}: H.323 trace on screen. Default is gwh323.txt.\n"
      "   -y  {with -v}: SIP trace on screen. Default is gwsip.txt.\n"
      "   -l  <file>   : Specify the license file. Default is \"sip323.license\".\n"
      "   -k           : Use dotted-decimal-ip-address instead of host name.\n"
      "   -h           : Print usage and exit\n"
      "\n";
      
    exit(0);
  }


  /* Check for validity of license */
  int e = 
  SIPLibrary::CheckLicense(args.GetOptionString('l', "sip323.license"));

  if ( e > 0 ) {
    cerr << "Your registration file is invalid or has expired.\n"
      "You should obtain a new registration by "
      "sending email to hgs@cs.columbia.edu.\n" << endl;
    exit(0);
  }
  else if ( e < 0 ) {
    cerr << "Unlicensed, single call version." << endl;
  }
  else if ( e == 0 ) {
    oLicensed = true;
  }

  /* This must be before CheckLicense */
  if ( args.HasOption('k') ) {
    SIPLibrary::oUseNumericAddress = true;
  }

  if ( args.HasOption('a') ) {
    colocatedWithGK = TRUE;
  }

  if ( args.HasOption('b') ) {
    cerr << "Signaling Gateway co-located with SIP proxy server is\n"
         << " for future work. Please do not use -b option now." 
         << endl;
      exit(0);
  }

  if ( args.HasOption('m') ) {
    cerr << "Signaling Gateway with media switching fabric is\n"
         << " for future work. Please do not use -m option now." 
         << endl;
      exit(0);
  }

  if ( args.HasOption('v') ) {
    /* --- Enable trace for H323 --- */
    PTrace::SetOptions(PTrace::Timestamp | PTrace::Thread);

    PTrace::SetLevel(4);
    if ( args.HasOption('x') ) {
      PTrace::SetStream(&cerr);
    }
    else {
      PTextFile * output = new PTextFile;
      if (output->Open("gwh323.txt", PFile::WriteOnly)) {
      PTrace::SetStream(output);
      }
      else {
        cout << "Warning: could not open trace output file \""
             << "gwh323.txt" << '"' << endl;
        delete output;
      }
    }

    /* --- Enable trace for SIP  --- */
    debug_flag = 1;
    if ( !args.HasOption('y')) {
      DebugInit("gwsip.txt");
    }
    ErrorInit("stderr");
  }

  SIPLibrary::SetName("SIP-H.323 Gateway, (c) 1999-2000 Columbia University");

  h323Address = args.GetOptionString('H',"");
  H323_Endpoint::GetHostAddress(h323Address, h323Address);
  cout << "H323 preconfigured address is " << h323Address << endl;

  sipAddress = args.GetOptionString('S',"");
  cout << "SIP preconfigured address is " << sipAddress << endl;

  h323GKAddress = args.GetOptionString('g', "");
  sipServerAddress = args.GetOptionString('p', "");

  localname = args.GetOptionString('n', "sip323@cs.columbia.edu");
  cout << "Local name is " << localname << endl;

  /*
   * Get the domain name 
   */
  domainName = args.GetOptionString('d', sipServerAddress /* default
                                                           */);
  cout << "Domain is " << domainName << endl;

  /* Create the listening servers */
  SIP = new MySIPUserAgent(true,false);
  bool isregistrar = args.HasOption('a') ? true : false;
  H323 = new MyH323Server(true, isregistrar); /* Server and GK */
  
  if ( !SIP->IsServer() ) {
    cerr << "Can not create SIP server" << endl;
    exit(1);
  }
  if ( !H323->IsServer() ) {
    cerr << "Can not create H323 server" << endl;
    exit(1);
  }

  cout << "H323 " <<  *H323 << endl;
  cout << "SIP "  << *SIP << endl;

  /* SIP registration */
  if ( args.HasOption('p') ) {
    SIPRegistrationInfo reginfo;
  
    reginfo.localName = strdup((const char *) (PString("sip:") +
                                               localname));
    if ( sipServerAddress == "" ) {
      reginfo.registrar = strdup((const char *) (PString("sip:") +
                                                 localname));
    }
    else {
      reginfo.registrar = strdup((const char *) (PString("sip:") +
                                                 sipServerAddress));
    }
    
    cout << "Hostname " << endl << SIPLibrary::HostName << endl;
    reginfo.contactAddress = strdup((const char *)(PString("sip:") +
                                                   SIPLibrary::HostName));
    
    cout << "Registering to SIP " << reginfo.localName << " at " <<
      reginfo.registrar << " with contact " <<
      reginfo.contactAddress << endl;
    
    SIP->Register(reginfo);    
  }

  if ( !args.HasOption('a') ) {
    IptelEndpoint::RegistrationInfo reginfo;

    if ( !args.HasOption('r') ) {
      reginfo.localName = strdup(localname);
      
      if ( h323GKAddress != "") {
        reginfo.registrar = strdup(h323GKAddress);
      }
      
      reginfo.contactAddress = strdup(SIPLibrary::HostName);
      
      cout << "Registering to H.323 " << reginfo.localName
           << " with contact " <<
        reginfo.contactAddress << endl;
      if ( h323GKAddress != "" ) {
        cout << "H.323 Gatekeeper is " << reginfo.registrar <<  endl;
      }
      
      H323->Register(reginfo);
    }
  }

  cout << "Waiting for connections..." << endl;

  static char *mycommands = 
          "exit | quit\n"
          "help\n"
          "? ( h323 | mutex | cond | request )\n"
          "log (msgflow | debug ) (on | off)\n"
          "register (sip | h323) <address>\n"
          ;
  cout << "Commands supported are : \n" << mycommands << endl;

  // Others to be implemented
  // ,verbose,save filename, dump sip filename, dump h323 filename

  bool oExitFlag = false;
  char command[180];

  if ( args.HasOption('i') ) {
    /* Interactive mode */
   while ( !oExitFlag ) {

    cout << "gw> " << flush;     /* prompt for command */
    strcpy(command,"");
    fscanf(stdin,"%[^\n]",command); /* read command */
    getchar();                      /* remove \n from stdin */

    char * cmd = strtok(command," ");
    if ( cmd == NULL ) continue;
    if ( !strcmp(cmd,"exit") || !strcmp(cmd,"quit")) {
      oExitFlag = true;
    }
    else if ( !strcmp(cmd,"help") ) {
      cout << "Commands supported are : \n" << mycommands << endl;
    }
    else if ( !strcmp(cmd,"?") ) {
      cmd = strtok(NULL," ");
      if ( cmd == NULL ) continue;
      if ( !strcmp(cmd,"mutex") ) {
        #ifdef WIN32
        pthread_log_print(1);
        #endif
      }
      else if ( !strcmp(cmd,"cond") ) {
        #ifdef WIN32
        pthread_log_print(2);
        #endif
      }
      else if ( !strcmp(cmd,"request")) {
        RequestPrint(stderr);
      }
      else if  ( !strcmp(cmd,"h323") ) {
        cout  << *H323 << endl;
      }
      else {
        fprintf(stderr,"Unknown querry option: %s\n", cmd);
      }
    }
    /* ---- LOGging mechanism --- */
    else if ( !strcmp(cmd,"log")) {
      cmd = strtok(NULL," ");
      if ( (cmd == NULL) || 
         !(!strcmp(cmd,"msgflow") || !strcmp(cmd,"debug") 
         || !strcmp(cmd,"error"))) { 
        fprintf(stderr,"mycommands\n");
        continue;
      }
      char *state = strtok(NULL," ");
      if ( !strcmp(cmd,"msgflow") ) {
        if ( state && !strcmp(state,"off")) {
          MessageFlowFile(NULL);
        }
        else {
          MessageFlowFile(stderr);
        }
      }
      else if ( !strcmp(cmd,"msgflow") ) {
        if ( state && !strcmp(state,"off")) {
          debug_flag = 0;
        }
        else {
          debug_flag = 1;
        }
      }
      else if ( !strcmp(cmd,"error") ) {
        fprintf(stderr,"errors' log can not be modified\n");
      }

    }
    
    else {
      fprintf(stderr,"Unknown command: %s\n", cmd);
    }
    
   }
  }
  else {
    while(1) {
      /*
       * There should be a better way to wait till termination. XXX
       */
#ifdef WIN32
      Sleep(10000);
#else
      sleep(1000);
#endif
    }
  }
  exit(1);
}


///////////////////// PacketRedirector /////////////////////////////

#ifdef NO_REINVITE

MediaSwitchingThread::MediaSwitchingThread()
  : PThread(10000)
{
  unsigned short port = beginat;
  beginat += 2;

  selfport = 0;
  remoteaddr.sin_addr.s_addr = 0;


#ifdef WIN32
  if ( firsttime ) {
    WSADATA winsock;
    WSAStartup(0x101, &winsock); /* error checking not done */
    firsttime = false;
  }
#endif
  struct sockaddr_in saddr;

  /* For RTP */
  rtp = ::socket(AF_INET,SOCK_DGRAM,0); 
  saddr.sin_family = AF_INET;
  saddr.sin_port = htons(port);
  saddr.sin_addr.s_addr = INADDR_ANY;

  int oreturn;
  oreturn = ::bind(rtp,(struct sockaddr *) &saddr,
                   sizeof(saddr));
  
  cout << "bind returned " << oreturn << endl;

  /* For RTCP */
  rtcp = ::socket(AF_INET,SOCK_DGRAM,0); 
  saddr.sin_family = AF_INET;
  saddr.sin_port = htons(port+1);
  saddr.sin_addr.s_addr = INADDR_ANY;
  oreturn = ::bind(rtcp,(struct sockaddr *) &saddr,
                     sizeof(saddr));

  selfport = port;
  cout << "MSF started, listening at " << port << endl;
}

MediaSwitchingThread::~MediaSwitchingThread()
{
  /* XXX you need a mutex here */

  if ( !oTerminate ) {
    oTerminate = TRUE;
  }
  cout << "MSF deleted" << endl;
}

void MediaSwitchingThread::SetRedirectAddress(struct sockaddr_in & addr)
{
  remoteaddr = addr;
  cout << "MSF redirect address is " << inet_ntoa(addr.sin_addr) << ":" 
    << ntohs(addr.sin_port) << endl;
}

unsigned short MediaSwitchingThread::GetSelfPort()
{
  return selfport;
}

void MediaSwitchingThread::Terminate()
{
  oTerminate = TRUE;
}

void MediaSwitchingThread::Main()
{
  cout << "MSF thread started" << endl;

  oTerminate = FALSE;

  struct sockaddr_in saddr;
  char buf[1500];
  int length = 0;
  int addr_length;
  int maxfd = (rtp > rtcp ) ? rtp : rtcp;

  fd_set fdsets;
  struct timeval tv;

  while (!oTerminate) {
    FD_ZERO(&fdsets);
    FD_SET(rtp, &fdsets);
    FD_SET(rtcp, &fdsets);

    tv.tv_sec = 0;
    tv.tv_usec = 20000; /* 20 millisec */
    if ( ::select(maxfd+1, &fdsets, NULL, NULL, &tv) > 0 ) {
      if ( FD_ISSET(rtp, &fdsets) ) {
        addr_length = sizeof(struct sockaddr_in);

        if ( (length = ::recvfrom(rtp,buf,sizeof(buf),0,
              (struct sockaddr *)&saddr,&addr_length )) > 0 ) {
          saddr = remoteaddr;
          /*
          cout << "RTP: Forwarding " << length << " bytes to "
               << inet_ntoa(saddr.sin_addr) << ":"
               << ntohs(saddr.sin_port) << endl;
          */
          ::sendto(rtp,buf,length,0,(struct sockaddr
                                     *)&saddr,sizeof(saddr));
#if 0
          /* Playout the file to local device */
          static bool firsttime = true;
          static int au = -1;
          if ( firsttime ) {
            firsttime = false;
            au = open("/dev/audio",O_WRONLY);
          }
          if ( au > 0 ) {
            write(au, buf+12, length);
          }
#endif

        }
        else {
          cout << "received RTP packet of size " << length << endl;
        }
          
      }
      if ( FD_ISSET(rtcp, &fdsets)) {
        addr_length = sizeof(struct sockaddr_in);
        if ( (length = ::recvfrom(rtcp,buf,sizeof(buf),0,
              (struct sockaddr *)&saddr,&addr_length )) > 0 ) {
          saddr = remoteaddr;
          saddr.sin_port = htons( ntohs(saddr.sin_port) + 1);
          /*
          cout << "RTCP: Forwarding " << length << " bytes to "
               << inet_ntoa(saddr.sin_addr) << ":"
               << ntohs(saddr.sin_port) << endl;
          */

          ::sendto(rtcp,buf,length,0,(struct sockaddr
                                     *)&saddr,sizeof(saddr));
        }
        else {
          cout << "received RTCP packet of size " << length << endl;
        }
          
      }
    }
  }

  /* Signal the exit flag */
  /* XXX use mutex to avoid sleep */
#ifdef WIN32
  closesocket(rtp);
  closesocket(rtcp);
  Sleep(1000);
#else
  close(rtp);
  close(rtcp);
  sleep(1);
#endif

  cout << "MSF thread exiting" << endl;
}

#endif

/************************** END OF FILE ********************************
*/
