/***************************** FILE HEADER *****************************
#               Copyright (c) 2000 by Columbia University              *
#                        All Rights Reserved                           *
#***********************************************************************
#***********************************************************************
#
#  File Name    : main.cpp
#
#  Module       : Main file for the SIP conference server.
#
#  Author(s)    : Kundan Singh (kns10@cs.columbia.edu)
#                 Gautam Nair  (gnair@cs.columbia.edu)
#
#  Created      : 21 April 2000
#
#  Description  : SIP conference audio server makes use of
#                 LIBSIP++ and RTP Audio Mixer Library.
#                
#  Notes        : 
#
#***********************************************************************
#                Modification History (Reverse Order)
#-----------------------------------------------------------------------
#   Date         Author              Description
#-----------------------------------------------------------------------
#              |             | 
# Aug 6, 2000  | Kundan      | For new library version.
#-----------------------------------------------------------------------
*/

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "iptelapi.h"
#include "sipapi.h"
#include "rtpmapi.h"

#ifndef WIN32 /* For unix */
#include "unistd.h"
#include "netinet/in.h"
#include "arpa/inet.h"
#include <sys/time.h>
#include <sys/resource.h>  /* setrlimit,getrlimit */
#include <errno.h>
#endif

#ifdef USE_MYSQL
#include "dbapi.h"
#endif

extern "C" {
#include "strdupn.h"
#include "pthread-util.h"
}

/*
 * Define this macro for workaround of plugnsip specific bugs
 */

#define INTEROPERATE_WITH_PLUGNSIP 1 // wait for sometime before accepting
#define SESSION_PRINT  1  //debug prints

static char *conffile = NULL; /* Conf list file */
static bool  oUseMysql = false;

extern "C" {
  void do_daemon(void);
}

/**
 * MySIPUserAgent and MySIPCall
 */
class MySIPUserAgent : public SIPEndpoint
{
public:
  MySIPUserAgent(bool IsServer = true, bool IsRegistrar=false);
  MySIPUserAgent(const IptelIPAddress & ip,bool IsRegistrar=false);
  virtual IptelCall *AllocateNewCallObject(const char *callid);
};

class MySIPCall : public SIPCall
{
private:
  int rtp; /* RTPMixer context for the call leg */
  char * CanAccept(); 
  session_description_t GetRemoteSDP();
  void SetSelfSDP(session_description_t );

public:
  MySIPCall(const char *callid=NULL) : SIPCall(callid) { };
  MySIPCall(SIPEndpoint & ep, const char *callid=NULL) : 
    SIPCall(ep, callid) { };
  virtual ~MySIPCall() { };
  virtual void   OnNewIncomingCall(IptelEndpoint & ep);
  virtual int    OnHangup();
};

//////////////// Method defn for SIP ///////////////////////

/* Dummy constructor */
MySIPUserAgent::MySIPUserAgent(bool IsServer,bool IsRegistrar)
: SIPEndpoint(IsServer,IsRegistrar)
{
}

MySIPUserAgent::MySIPUserAgent(const IptelIPAddress & ip,bool IsRegistrar)
: SIPEndpoint(ip,IsRegistrar)
{
  /* Do nothing here */
}


/* Allocate a new call object of type MySIPCall */
IptelCall *MySIPUserAgent::AllocateNewCallObject(const char *callid)
{
  return new MySIPCall(*this, callid);
}

/*
 * Read the conference list and findout if the
 * call from the particular sender to a particular
 * receiver (destination or conference URI) can
 * be accepted or not.
 * Returns the conference name on success and
 * NULL on failure
 */

static FILE *fp = NULL;
static bool acceptall = false;

char * MySIPCall::CanAccept()
{
  IptelAddress dest    = GetRequestURI();
  IptelAddress src     = GetRemoteAddress();

  if ( acceptall ) {
    char *n;
    n = dest.GetUser(); /* Accept all conference */
    if ( n == NULL ) {
      n = dest.GetHost();
    }
    return n;
  }

  /*
   * Check if the desctination address is a valid conference
   * identifier.
   */
#ifdef USE_MYSQL
  if (!oUseMysql) {
    /* Not using mysql */
#endif
 
  //FILE *fp;
  if ( fp == NULL ) {
    if (( fp = fopen(conffile, "r")) == NULL ) {
      /* No conference list specified */
      cerr << "Could not open " << conffile << "; Error: " << errno <<
        endl;
      //exit(0);
      return NULL;
    }
  }

  char line[80]; /* max 80 char per line is read, rest ignored */
  bool found = false;
  char *linepart;
  char *l1;

  while ((fgets(line, sizeof(line), fp)) != NULL) {
    line[strlen(line)-1] = '\0'; /* \n to \0 */
    //cout << "Line:" << line << endl;
    l1 = NULL;
    linepart = strtok_r(line, " ", &l1);
    if ( linepart == NULL ) {
      //cout << "linepart == NULL" << endl;
      continue;
    }

    IptelAddress listedaddr = linepart;
    IptelAddress givenaddr = dest;

    listedaddr.SetScheme(NULL);
    givenaddr.SetScheme(NULL); /* remove the scheme */

    char *listeduser = listedaddr.GetUser();
    char *givenuser  = givenaddr.GetUser();
    char *listedhost = listedaddr.GetHost();
    char *givenhost  = givenaddr.GetHost();

    /*    
    cout << listeduser << "@" << listedhost << ";"
         << givenuser << "@" << givenhost << ";" << endl;
    */

    if ( listeduser != NULL && givenuser != NULL && 
         listedhost != NULL && givenhost != NULL &&
         !strcasecmp(listeduser, givenuser) &&
         !strcasecmp(listedhost, givenhost)) {
      /*
      cout << listeduser << "@" << listedhost << ";"
           << givenuser << "@" << givenhost << ";" << endl;
      */
      found = true;
      if ( listeduser ) { free(listeduser); }
      if ( givenuser ) { free(givenuser); }
      if ( listedhost ) { free(listedhost); }
      if ( givenhost ) { free(givenhost); }
      break;
    }
    else if ( listeduser == NULL && listedhost != NULL &&
              givenuser != NULL ) {
      if ( !strcasecmp(listedhost, givenuser)) {
        found = true;
        /*
        cout <<  "@" << listedhost << ";"
             << givenuser << "@" <<  ";" << endl;
        */
        if ( listeduser ) { free(listeduser); }
        if ( givenuser ) { free(givenuser); }
        if ( listedhost ) { free(listedhost); }
        if ( givenhost ) { free(givenhost); }
        break;
      }
    }
    if ( listeduser ) { free(listeduser); }
    if ( givenuser ) { free(givenuser); }
    if ( listedhost ) { free(listedhost); }
    if ( givenhost ) { free(givenhost); }
  }

  
  rewind(fp);

  //fclose(fp);

  if ( found ) {
    return strdup(line); /* whatever is in the file */
  }
  else {
    return NULL;
  }

#ifdef USE_MYSQL
  } /* If !use mysql*/
  else {

    /* Check if the name implies adhoc conference. *.adhoc@... */
    char *user = dest.GetUser();
    if ( user != NULL ) {
      char *f = strstr(user, ".adhoc");
      if ( f && !strcmp(f, ".adhoc") ) {
        /* Match found in the username */
        /* Do not free user or f. */
        return user;
      }
    }
   
    /* Try to locate it in the conferences table */
    if ( !isConnectionEstablished() ) {
      /* No connection with mysql server; fail */
      free(user);
      return NULL;
    }

    int ret = 0;
    ResultTable table = query("conferences", "url", user, "url", &ret); 
    freeResultTable(table, ret, 1);
    
    if ( ret <= 0 ) {
      free(user);
      return NULL;
    }
    
    /* If ret > 0 then it is found. return the name */
    return user;
  }
#endif

}

/*
 * OnNewIncomingCall: This is called from the SIP library
 * when an incoming call is received
 */
void MySIPCall::OnNewIncomingCall(IptelEndpoint & ep)
{
  IptelAddress dest    = GetRequestURI();
  IptelAddress src     = GetRemoteAddress();
  char *confname = NULL;

  cout << src << " wants to join " << dest << endl;

  /*
   *  XXX extract the database table for this conference.
   * If the conference URL is not found then return 604.
   * If the conference type is public then skip this step.
   * Else, match the src address agains the participant
   * list for the conference. If matches then continue,
   * otherwise return 400 Forbidden.
   * Check the current_participant_count against
   * max_participant_count and return 500 Conference Full
   * if the limit is reached.
   */
 
  if ( (confname = CanAccept()) == NULL ) {
    /* Can not accept this call leg */
    cout << "Rejecting" << endl;
    Reject(604, "Conference does not exist");
    return;
  }

  /*
   * XXX Check if authentication is required; if yes
   * call the appropriate method. Also, be sure
   * to use the user:passwd as url:password or
   * src:password;
   */

  /*
   * XXX check media allowed field and format
   * the SDP for the conference based on that.
   */

  /* Accept the call */
  /* Set the session description first */
  session_description_t sdp;
  sdp = GetRemoteSDP();

#ifdef SESSION_PRINT
  cout << "Participant's session:" << endl;
  SDPPrint(&sdp);
  cout << flush;
#endif

  /* Add the call leg in the mixer library */
  rtp = RTPMAddCallLeg(confname /* name of conference */,
                       src.GetAddressString());

  free(confname);
  confname = NULL;

  if ( rtp < 0 ) {
    SDPFree(&sdp);
    cout << "Rejecting due to internal failure" << endl;
    Reject(500, "Could not create mixer call leg"); 
    /* could not create the RTP Mixer leg */
    return;
  }

  /* Set the self SDP of mixer */
  if ( RTPMSetSelfSessionDescription(rtp, sdp) < 0) {
    SDPFree(&sdp);
    RTPMRemoveCallLeg(rtp);
    Reject(500, "Error in setting session description");
    return;
  }
  SDPFree(&sdp);

  /* Get the mixer's SDP */
  sdp = RTPMGetRemoteSessionDescription(rtp);

#ifdef SESSION_PRINT
  cout << "Server's session:" << endl;
  SDPPrint(&sdp);
  cout << flush;
#endif

  SetSelfSDP(sdp);
  SDPFree(&sdp);

#ifdef INTEROPERATE_WITH_PLUGNSIP
  /* XXX what if the call is closed during this time */
  sleep(1);
#endif

  /*
   * XXX update the current participant count.
   */

  /* XXX
   * Check the conf users' table and update hte
   * send, receive, and audio video parameters for the
   * callleg. Update the type to dialin.
   * update the status to connected.
   */

  Accept(); /* Accept the call finally */
}

/*
 * A separate thread is used to
 * terminate the call leg in the
 * RTP AUdio mixer library.
 * This is to avoid delay in response to SIP
 * entity
 */
void *RemoveCallLegThread (void *p)
{
  //cout << "RemoveCallLegThread is called" << endl;

  int rtp = (int) p;
  if ( rtp >= 0 ) {
    /*
    if ( debug_flag ) {
      cout << "Calling RTPMRemoveCallLeg from RemoveCallLegThread" <<
        endl;
    }
    */
    RTPMRemoveCallLeg(rtp);
  }
  return NULL;
}

/*
 * This function is called by the 
 * SIP library when the SIP call is
 * closed by remote
 */
int   MySIPCall::OnHangup()
{
  IptelAddress dest    = GetSelfAddress();
  IptelAddress src     = GetRemoteAddress();

  pthread_t tid;

  /* Create a thread to close RTP Mixer Library call leg */
  if ( pthread_create_detached(&tid, NULL, RemoveCallLegThread, (void *) rtp)
       !=0) {
    cerr << "pthread_create_detached failed in OnHangup" << endl;
    RemoveCallLegThread((void *)rtp);
  }

  /* 
   * Terminate the call leg from this user
   * for the appropriate conference 
   */
  cout << src << " leaving conference " << dest << endl;

  /* XXX
   * Update the confusers table to 
   * reflect, status=notconnected. 
   * update the conferences table to reflect 
   * current_participant_count.
   */

  SIPCall::OnHangup();
  return 0; /* We should use Mutex in OnNewIncomingCall to
               protect againts freeing of call object */
}

/*
 * Convert the C++ object to C structure for
 * session description of remote entity 
 */
session_description_t MySIPCall::GetRemoteSDP()
{
  session_description_t sdp;
  SDPInit(&sdp);

  SessionDescription *session = GetRemoteSessionDescription();
  if ( session ) {
    for ( int i=0; i<session->GetMediaCount(); i++ ) {
      MediaInfo & media = (*session)[i];
      media_line_t m;
      ListCreate(&m.payloads,1,1);
      for ( int j=0; j<media.GetFormatCount(); j++) {
        FormatInfo & format = media[j];
        payload_type_t p;
        char description[100];
        int clockrate = format.GetClockRate();
        int channel_count = format.GetChannelCount();
        const char *name = format.GetFormatName();

        p.pt = format.GetPayloadType();
        cout << "GetSDP: " << p.pt << "=>" << (name != NULL ? name : "NULL")
             << endl;
        if ( name != NULL ) {
          if ( channel_count <= 1 ) {
            sprintf(description, "%s/%d", name, clockrate);
          }
          else {
            sprintf(description, "%s/%d/%d", name, clockrate,
                    channel_count);
          }
          p.description = strdup(description);
        }
        else {
          p.description = NULL;
        }
        ListAddElement(&m.payloads, &p, sizeof(p));
      }

      IptelIPAddress addr;
      addr = media.GetRxAddress();
      addr.GetIpAddress(m.sin);

      m.ttl = 255; /* ignored */
      m.sin.sin_port = htons(media.GetRxPort());
      switch (media.GetDataType() ) {
      case MediaInfo::Audio:
        m.type = RTPM_MEDIA_AUDIO;
        break;
      case MediaInfo::Video:
        m.type = RTPM_MEDIA_VIDEO;
        break;
      case MediaInfo::Application:
        m.type = RTPM_MEDIA_DATA;
        break;
      default:
        m.type = RTPM_MEDIA_UNKNOWN;
        break;
      }

      ListAddElement(&sdp.media_line_t, &m, sizeof(m));
    }
  }

  /*
  if ( debug_flag ) {
    cout << "Get remote SDP";
    SDPPrint(&sdp);
    cout << endl;
  }
  */

  return sdp;
}

/*
 * Convert the C++ object to C structure for
 * session description of the local object 
 */
void MySIPCall::SetSelfSDP(session_description_t sdp)
{
  int i,j;

  SessionDescription *session = new SessionDescription;

  for (i = 0; i < ListCount(&sdp.media_line_t); i++) {
    media_line_t *m = (media_line_t *) 
      ListElement(&sdp.media_line_t, i);
    if ( !m )
      continue;

    MediaInfo media;

    media.SetRxAddress(m->sin);
    media.SetDirection(MediaInfo::Receive);
    switch (m->type ) {
    case RTPM_MEDIA_AUDIO:      
      media.SetDataType(MediaInfo::Audio);
      break;
    case RTPM_MEDIA_VIDEO:      
      media.SetDataType(MediaInfo::Video);
      break;
    case RTPM_MEDIA_DATA:
      media.SetDataType(MediaInfo::Application);
      media.SetTransportType(MediaInfo::UDP_Transport);
      break;
    default:      
      media.SetDataType(MediaInfo::UnknownData);
      break;
    }

    media.SetRxPort(ntohs(m->sin.sin_port));

    for ( j=0; j<ListCount(&m->payloads); j++) {
      payload_type_t *p = (payload_type_t *)
        ListElement(&m->payloads, j);
      if ( !p ) 
        continue;
      if ( p->description == NULL ) {
        media.AddFormat(p->pt);
      }
      else {
        char *l = NULL;
        char *description = strdup(p->description);
        char *name = strtok_r(description, "/", &l);
        char *clkstr = strtok_r(NULL, "/", &l);
        char *cnlstr = strtok_r(NULL, "/", &l);
        if ( name == NULL ) {
          media.AddFormat(p->pt);
        }
        else {
          if ( clkstr == NULL ) {
            media.AddFormat(FormatInfo(p->pt, name, 8000) /* deafault */);
          }
          else if ( cnlstr == NULL ) {
            media.AddFormat(FormatInfo(p->pt, name, atoi(clkstr)));
          }
          else {
            media.AddFormat(FormatInfo(p->pt, name, atoi(clkstr),
                            atoi(cnlstr)));
          }
        }
        free(description);
      }
    }

    session->AddSession(media);
  }

  //SDPPrint(&sdp);
  //cout << "C++ Session : " << *session << endl;

  SetSelfSessionDescription(session);

}

/*
 * XXX new methods to be implemented.
 * Implement OnRemoteSessionChanged() method.
 * Makesure that the permissions for audio,video and
 * privilege is checked for the participant.
 *
 * Implement OnNotify() method to process the
 * commands from web GUI.
 * First authenticate the request with digest.
 * Then extract the XML message body and parse it to
 * the command.
 * Following commands are supported:
 * - New Participant Type for a dialout conference:
 *     Initiate a new Call to the participant. Be prepared to
 *     supply any authentication required.
 * - Participant removed or status changed to notconnected.
 *     Send a Hangup to the participant if he was in
 *     the conference.
 * - Change audio, video, send, receive preference for
 *   a participant.
 *     Update the call leg to reflect the change.
 * - Removal or Termination (end-time) of the conference.
 *     Send Hangup to all the participants.
 */


/*
 * usage()
 * Usage information. Exits program with status code 1.
 */
void usage(char *prog) 
{
  fprintf(stderr,"Usage: %s [options]\n"
  " or    %s -h\n"
  "h        : Print usage and exit.\n"
  "v        : Use verbose mode.\n"
  "f file   : Use conference list from this file. Default is conflist.txt\n"
  "X        : Non-daemon mode.\n"
  "i        : Interactive mode, used for debugging.\n"
  "k        : Use numeric (dotted-decimal) for host name.\n"
  "a        : Accept all conference without looking in conf list.\n"
  "m time   : Packatization time for outgoing RTP packets. Default is 40\n"
  "           for 40 milliseconds\n"
  "p port   : Listening SIP port, default is 5060.\n"
  "b        : Send audio back to the participant who generated it.\n"
  "D mysql-host : MySQL server host.\n"
  "U mysql-user : MySQL user id. Default is root.\n"
  "P mysql-password: MySQL user password. Default is tiger.\n"
  "S mysql-database: MySQL database name. Default is sip.\n"
	  , prog, prog);
  exit(1);
} /* usage */

/* Main routine */

int main(int argc, char *argv[])
{
  int c;
  bool isdaemon = true;
  bool isinteractive = false;
  struct rlimit rbuf;
  int packetization_time = 40;
  unsigned short listening_port = 5060;
#ifdef USE_MYSQL
  char *mysql_host = NULL;
  char *mysql_user = strdup("root");
  char *mysql_passwd = strdup("tiger");
  char *mysql_db = strdup("sip");
#endif

  /* Parse command line options */
  while((c = getopt(argc, argv, "f:vhXiam:p:kD:U:P:S:b")) != EOF) {

    switch(c) {
    case 'h' :
      usage(argv[0]);
      exit(0);
      break;
    case 'v' :
      SIPLibrary::DebugOn("stdout");
      break;
    case 'k' :
      SIPLibrary::oUseNumericAddress = true;
      break;
    case 'f' :
      conffile = strdup(optarg);
      break;
    case 'X' :
      isdaemon = false;
      break;
    case 'i' :
      isinteractive = true;
      break;
    case 'a' :
      acceptall = true;
      break;
    case 'm':
      packetization_time = atoi(optarg);
      break;
    case 'p':
      listening_port = atoi(optarg);
      break;
    case 'b':
      RTPMSendBackAudioToParticipant(1 /* 1 for true */);
      break;

    case 'D': /* Database */
    case 'U': 
    case 'P': 
    case 'S': 
#ifdef USE_MYSQL
      {
        switch(c) {
        case 'D': /* Database host */
          strassign(&mysql_host, optarg);
          oUseMysql = true;
          break;
        case 'U': /* Database user */
          strassign(&mysql_user, optarg);
          break;
        case 'P': /* Database password */
          strassign(&mysql_passwd, optarg);
          break;
        case 'S': /* Database name */
          strassign(&mysql_db, optarg);
          break;
        }
      }
      break;
#else
      fprintf(stderr,"Not compiled using mysql; do not use -D -U -P or -S option\n");
      exit(1);
#endif
    default:
      fprintf(stderr,"Unknown option -%c\n", c);
      usage(argv[0]);
      exit(1); /* Not required */
    }
  }

  /* Check for validity of license */
  int e = 
  SIPLibrary::CheckLicense("sipconf.license");

  if ( e > 0 ) {
    cerr << "Your registration file is invalid or has expired.\n"
      "You should obtain a new registration by "
      "sending email to hgs@cs.columbia.edu.\n" << endl;
    exit(0);
  }
  else if ( e < 0 ) {
    cerr << "Unlicensed copy. Contact hgs@cs.columbia.edu for a license." << endl;
    exit(1);
  }
  else if ( e == 0 ) {
    /* Licensed */
  }

  /* Min packetization time is 20 ms, max is 2 seconds */
  if ( packetization_time < 20 ) {
    packetization_time = 20;
  }
  else if ( packetization_time > 2000 ) {
    packetization_time = 2000;
  }

  if ( conffile == NULL ) {
    conffile = strdup("conflist.txt");
  }

  /* Set the open file limit. Works for solaris */
  memset(&rbuf, 0, sizeof(rbuf));
  if ( getrlimit(RLIMIT_NOFILE, &rbuf) == 0 ) {
    rbuf.rlim_cur = rbuf.rlim_max;
    if ( setrlimit(RLIMIT_NOFILE, &rbuf) != 0 ) {
      cerr << "Could not set the OPEN_MAX limit to " << rbuf.rlim_cur
           << endl;
    }
  }
  
  /* Set the core file rlimit. Works for solaris */
  memset(&rbuf, 0, sizeof(rbuf));
  if ( getrlimit(RLIMIT_CORE, &rbuf) == 0 ) {
    rbuf.rlim_cur = rbuf.rlim_max;
    if ( rbuf.rlim_cur > 10000000L ) {
      rbuf.rlim_cur = 10000000L;
    }
    if ( setrlimit(RLIMIT_CORE, &rbuf) != 0 ) {
      cerr << "Could not set the RLIMIT_CORE limit to " << rbuf.rlim_cur
           << endl;
    }
  }

  /* Create the listening servers */
  IptelIPAddress listeningAddress;
  listeningAddress.SetPort(listening_port);


  SIPLibrary::SetName("SIPCONF-Columbia-University-Audio-Conference-Server");
  if ( isdaemon ) {
    do_daemon();
  }
  else {
    SIPLibrary::ErrorOn("stderr");
  }

  MySIPUserAgent sip(listeningAddress, false);

  /* Set the packetization time */
  RTPMSetPacketizationTime(packetization_time);

  cout << "Waiting for connections..." << endl;

#ifdef USE_MYSQL
  if ( mysql_host != NULL ) {
    /* Establish the database connection */
    initializeConnection(mysql_user, 
                         mysql_passwd, 
                         mysql_db, 
                         mysql_host,
                         0);
  }
#endif

  if (isinteractive) {
    bool oExitFlag = false;
    char command[180];
    
    while ( !oExitFlag ) {
      cout << "sipconf> " << flush;     /* prompt for command */
      strcpy(command,"");
      fscanf(stdin,"%[^\n]",command); /* read command */
      getchar();                      /* remove \n from stdin */
      
      char * cmd = strtok(command," ");
      if ( cmd == NULL ) continue;
      if ( !strcmp(cmd,"exit") || !strcmp(cmd,"quit")) {
        oExitFlag = true;
      }
      else {
        cout << "Command " << command << " not implemented" << endl;
      }
    }
  }
  else {
    while ( 1 ) {
      /* There should be a better way to wait for termination */
#if 1
      sleep(10000);
#else
      struct rusage ubuf;
      sleep(5);
      getrusage(RUSAGE_SELF, &ubuf);
      printf("Memory-maxrss(%ld) idrss(%ld)\n",
             ubuf.ru_maxrss,
             ubuf.ru_idrss);
             
#endif
    }

    cout << "Exiting" << endl;
    exit(0);
  }

#ifdef USE_MYSQL
  /*
   * close connection to Mysql Database.
   */
  closeConnection();
#endif

#ifdef USE_MYSQL
  if ( mysql_host ) 
    free(mysql_host);
  if ( mysql_user) 
    free(mysql_user);
  if ( mysql_passwd ) 
    free(mysql_passwd);
  if ( mysql_db)
    free(mysql_db);
#endif

  cout << "Exiting" << endl;
  return 0; /* Not required */
}

/************************** END OF FILE ********************************
*/

