/*
 * cgi - SIP cgi handler
 *
 * Copyright 1998-1999 by Columbia University; all rights reserved 
 * Written by Jonathan Lennox.  Begun November 12, 1998.
 * Re-written for policy handling beginning May 28, 1999.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include <ctype.h>

#include "dict.h"
#include "ht.h"

#include "request-sip.h"
#include "message.h"
#include "config.h"
#include "register.h"
#include "http.h"
#include "libsip.h"
#include "strdupn.h"
#include "policy.h"
#include "sipd-policy.h"
#include "proxy-support.h"
#include "timespec.h"
#include "sysdep.h"

#include "presence.h"
#include "dbapi.h"


/* Local structures */

/* Local info for CGI handling */
struct cgi_info {
  script_t *script;
  char *script_cookie;
  int sent_local_200;
  int again;
  registration_t *reg;
  dict_h responses;
  int dict_errno;
  int subscription_added; /* only one subscription (first 200) added to database */
};


/* Branch local info for CGI handling */
struct cgi_branch_info {
  char *request_token;
};


/* A set of meta-variables, structured as environment variables */
struct metavar_set {
  char **vars;
  int allocated;
  int num;
};

#define METAVAR_SET_INITIALIZER {NULL, 0, 0}


/* The information needed to parse CGI output */
struct parse_info {
  int offset;
  enum {Header, CR, CRNL, CRNLCR, NL, Body} state;
  int content_length;
  int cl;
  DString header, body;
  int parsed;
  int parse_error;
  char *msg_body;
  parser_t parse;
};


/* Policy functions */

static int cgi_init(request_t *r, registration_t *reg, int reg_status);
static int cgi_handle_request(request_t *r, message_t *m);
static int cgi_handle_response(request_t *r, message_t *m, branch_t *b);
static int cgi_timeout_expired(request_t *r);
static int cgi_cleanup(request_t *r);


/* Local functions */

static int chmod_scriptfile(char* script);
static int DoCGIRequestProcessing(request_t *r, message_t *m, int *do_default);
static int DoCGIResponseProcessing(request_t *r, message_t *m, branch_t *b,
                                   int *do_default);
static int AddStandardCGIVars(struct metavar_set *metavars, request_t *r,
                              parser_t *p, body_t *b,
                              sock_t source);
static void AddRegistrations(struct metavar_set *metavars,
                             registration_t *reg);
static int AddMetavar(struct metavar_set *metavars, const char *name,
                      const char *value);
static void FreeMetavars(struct metavar_set *metavars);
static char *canonicalize_header_name(char *name);
static char *canonicalize_header_value(parser_element_t *e);
static int ExecuteCGI(request_t *r, struct metavar_set *metavars,
                      message_t *m, int *do_default);
static pid_t spawn_child(int fds[3], char *script_filename, char *env[]);
static int WriteBody(body_t *b, int *child_in);
static int ReadFromChild(request_t *r, int child_fds[3], message_t *m,
                         int *timed_out, int *do_default);
static int ReapChild(request_t *r, pid_t pid, int timed_out);
static int ReadIntoDString(int fd, DString *buf);
static int ParseChildStderr(request_t *r, DString *readbuf, int *offset);
static void InitParseInfo(struct parse_info *p);
static int ParseChildOutput(request_t *r, DString *readbuf, message_t *m,
                            struct parse_info *p, int closed, int *do_default);
static int DoCGIAction(request_t *r, message_t *m, parser_t *header,
                       body_t *body, int *do_default);
static int HandleAndRemoveCGIHeaders(request_t *r, parser_header_t **hlist,
                                     struct cgi_branch_info *bi,
                                     parser_header_t **to_remove);
static message_t *MergeMessageChanges(message_t *m, parser_t *header,
                                      body_t *body,
                                      parser_header_t *to_delete);
static void RFC822_HeaderReplaceOrAppend(parser_header_t **php,
                                         parser_header_t *orig_add);
static void RFC822_HeaderDelete(parser_header_t **php, char *name);

static void CopyCGIRemove(headers_t *del_headers, parser_header_t *to_delete);
static message_t *LookupResponseToken(request_t *r, message_t *m, char *token);
static void NewResponseToken(request_t *r, message_t *m);
static void CleanOutMessageQueue(request_t *r);

/* Response hashtable */
static ht_val hash_message(dict_obj o);
static ht_val hash_key(dict_key k);
static int msg_compare(dict_obj obj1, dict_obj obj2);
static int msg_key_compare(dict_key key, dict_obj obj);

/* The structure containing the CGI policy */
policy_t cgi_policy =
{
  "cgi",
  &cgi_init,
  &cgi_handle_request,
  &cgi_handle_response,
  &cgi_timeout_expired,
  &cgi_cleanup
};


/* cgi_init()
 * Start handling a CGI-mediated policy
 */
static int cgi_init(request_t *r, registration_t *reg, int reg_status)
{
  struct cgi_info *info;
  int do_default;
  struct ht_args argsp = {0, 0, &hash_message, &hash_key};

  info = (struct cgi_info*) calloc(1, sizeof(struct cgi_info));
  if (info == NULL) {
    cgi_cleanup(r);
    return 1;
  }
  r->policy_info = (void *)info;

  /* Locate and setup CGI scriptfile */
  info->script = GetScriptByDisposition(&reg->scriptset, 
                                        disposition_sip_cgi);
  if (!chmod_scriptfile(info->script->filename)) {
    error("cgi_init", "Couldn't set permissions for script file %s",
          info->script->filename);
    cgi_cleanup(r);
    return 1;
  }

  info->reg = reg;

  /* Create new response hash */
  info->responses = ht_create(&msg_compare, &msg_key_compare,
                              DICT_UNIQUE_KEYS|DICT_UNORDERED,
                              &info->dict_errno, &argsp);
  
  /* initialize to not added */
  info->subscription_added = 0;

  if (info->responses == NULL) {
    error("cgi_init", "Couldn't create response hashtable: error %d",
          info->dict_errno);
  }

  if (DoCGIRequestProcessing(r, GetRequestAsMessage(r), &do_default)) {
    cgi_cleanup(r);
    return 1;
  }

  if (do_default) {

    char * method;
    cgi_cleanup(r);
    method = HTTP_Method(&r->p);
    /* 
     * if SUBSCRIBE request, then we can proxy it to all the local contacts 
     *  (if allowed) using the default subscribe proxy.
     */
    if (strcasecmp(method, "SUBSCRIBE") == 0) 
      r->policy = &subscribe_policy;
    else
      r->policy = find_primary_policy(reg, reg_status);
    free_valid(method);
    /*
    cgi_cleanup(r);
    r->policy = find_primary_policy(reg, reg_status);
    */
    if (r->policy == NULL) {
      ResponseSetStatus(r, 500, NULL);
      r->policy = &failure_policy;
    }
    debug("cgi_init", "r=%p: DEFAULT-ACTION called, using %s policy\n",
          r, r->policy->name);
    return r->policy->init(r, reg, reg_status);
  }

  policy_set_timeout(r, (double)config.timeout);  /* XXX ??? */

  return 0;
} /* cgi_init */


/*
 * cgi_handle_request()
 * This will get called if a) we've sent a 200 response, and
 * so need to handle an ACK, or b) a request should be cancelled.
 * The ACK only goes to the CGI script if the 200 that was sent was
 * locally-generated, i.e. the CGI is emulating a UAS.  Cancel requests go
 * to the script, but they're only informative; the server cancels all the
 * branches and sends back the 487 on its own.
 */
static int cgi_handle_request(request_t *r, message_t *m)
{
  struct cgi_info *info = (struct cgi_info *) r->policy_info;
  int do_default;

  if (strcmp(m->p.method, "CANCEL") == 0) {
    /* 
     * We were canceled.  Cancel all the branches that aren't done, and
     * send back 487.
     */
    CancelUnfinishedBranches(r);

    if (!r->sent_final) {
      policy_send_new_response(r, 487, NULL);
      policy_set_timeout(r, (double)config.timeout);
    }
    
    if (info->again) {
      if (DoCGIRequestProcessing(r, m, &do_default)) {
        cgi_cleanup(r);
        return 1;
      }
    }
    /* Don't care about do_default; default is to do nothing. */

    m->handled = finished_handling;
  }
  else if (strcmp(m->p.method, "ACK") == 0) {
    if (!r->sent_200) {
      error("cgi_handle_request", "ACK arrived at unexpected time");
    }
    else if (info->again && info->sent_local_200) {
      if (DoCGIRequestProcessing(r, m, &do_default)) {
        cgi_cleanup(r);
        return 1;
      }
    }
    else {
      ForwardACKToBranches(r, m);
    }
    m->handled = finished_handling;
  }
  else {
    /* I don't think any other methods should be showing
     * up here. */
    error("cgi_handle_request",
          "Request with surprising method %s on message queue\n",
          m->p.method);
    m->handled = finished_handling;
  }

  return 0;
}


/*
 * cgi_handle_response()
 * This gets called when a CGI-controlled transaction has a response
 * arrive on some branch.
 */
static int cgi_handle_response(request_t *r, message_t *m, branch_t *b)
{
  struct cgi_info *info = (struct cgi_info *) r->policy_info;
  int do_default;

  if (info->again) {
    if (DoCGIResponseProcessing(r, m, b, &do_default)) {
      cgi_cleanup(r);
      return 1;
    }
    m->handled = policy_handled;
  }
  else {
    do_default = 1;
  }

  if (do_default) {
    /**
     * If the response was for a SUBSCRIBE request, it was a 200 response, and
     * we haven't add a subsription yet, then add the subscription to the database
     */
    if (strcasecmp(m->p.method,"SUBSCRIBE") == 0) {
      if ((m->p.status >= 200) && (m->p.status < 300)) { 
	if (!info->subscription_added) {
	  if (!add_subscribed_user(r)) {
	    error("cgi_handle_response",
		  "error adding subscription to database:%s",
		  getMysqlError());
	    policy_send_new_response(r, 500, NULL);
	    policy_set_timeout(r, (double)config.timeout);
	  }
	  else { /* forward message and terminate policy */
	    info->subscription_added = 1;
	    policy_forward_response(r, m);
      
	    /* If we get multiple 200's, we may cancel more than once here; but
	     * that's okay. */
	    CancelUnfinishedBranches(r);

	    policy_set_timeout(r, (double)config.timeout);
	    m->handled = finished_handling;
	  }
	}
      }
      /* check if all branches are finished */
      else if (m->p.status >= 300) {
	  /* send out this message's final response */
	  if ((r->branches_outstanding == 0) && (!info->subscription_added)) {
	    policy_send_new_response(r, m->p.status, NULL);
	    policy_set_timeout(r, (double)config.timeout);
	  }
      }
    }


    ProxyDispatchResponse(r, m);

    if (r->branches_outstanding == 0 && !r->sent_final) {
      CleanOutMessageQueue(r);

      if (!r->sent_final) {
        SendBestResponse(r);
      }
    }
  }

  return 0;
} /* cgi_handle_response */


/*
 * cgi_timeout_expired()
 * If we've sent a final response, we're done; otherwise, refresh the
 * timeouts -- we're waiting for a final response from a branch.
 */
static int cgi_timeout_expired(request_t *r)
{
  if (r->sent_final) {
    policy_done(r);
  }
  else {
    policy_set_timeout(r, (double)config.timeout);
  }
  return 0;
} /* cgi_timeout_expired */


/*
 * cgi_cleanup()
 * Called both when we're done processing a request with CGI, and when we
 * need to bail out.  Clean up after CGI processing.
 */
static int cgi_cleanup(request_t *r)
{
  int i;

  struct cgi_info *info = (struct cgi_info *) r->policy_info;

  if (!info) return 0;

  /*
   * info->script is a pointer into info->reg.scriptset, and so
   * shouldn't be freed here.
   */

  if (info->responses != NULL) {
    ht_destroy(info->responses);
  }

  for (i = 0; i < r->branch_count; i++) {
    free_valid(r->branches[i].policy_info);
    r->branches[i].policy_info = NULL;
  }

  free_valid(info->script_cookie);

  free(info);

  return 0;
}


/* 
 * chmod_scriptfile()
 * Ensure that the scriptfile is readable and executable by its owner.
 * On failure, return false.
 * NOTE: if we ever start using a suexec setup, this function should
 * also chown the file to the appropriate user.
 */
static int chmod_scriptfile(char* filename)
{
  struct stat script_stat;

  if (stat(filename, &script_stat) != 0) {
    return 0;
  }

  if (chmod(filename, S_IRUSR | S_IXUSR | script_stat.st_mode) == -1) {
    return 0;
  }
  return 1;
} /* chmod_scriptfile */


/*
 * DoCGIRequestProcessing()
 * Process a CGI request -- set up its environment variables, invoke
 * its script, and execute the appropriate actions.
 * Return non-zero if something went wrong.
 */
static int DoCGIRequestProcessing(request_t *r, message_t *m, int *do_default) 
{
  /* The meta-variables for the request */
  struct metavar_set metavars = METAVAR_SET_INITIALIZER;    

  /* Standard metavars */
  if (AddStandardCGIVars(&metavars, r, &m->p, &m->body, m->source)) {
    goto failure;
  }

  /* Metavars specific to requests */
  if (AddMetavar(&metavars, "REQUEST_METHOD", m->p.method))
    goto failure;
  if (AddMetavar(&metavars, "REQUEST_URI", m->p.URI_string))
    goto failure;

  /* Execute CGI script */
  if (ExecuteCGI(r, &metavars, m, do_default))
    goto failure;

  FreeMetavars(&metavars);
  return 0;

failure:
  FreeMetavars(&metavars);
  return 1;
} /* DoCGIRequestProcessing */


/*
 * DoCGIResponseProcessing()
 * Process a CGI response, similarly.
 */
static int DoCGIResponseProcessing(request_t *r, message_t *m, branch_t *b,
                                   int *do_default)
{
  char buf[80];
  struct metavar_set metavars = METAVAR_SET_INITIALIZER;    
  struct cgi_info *info = (struct cgi_info *) r->policy_info;
  struct cgi_branch_info *branch_info =
    (struct cgi_branch_info *) b->policy_info;
  message_t *var_src_msg = NULL;
  const char *reason;

  /* Add a response token */
  NewResponseToken(r, m);

  /* Insert the response in the response-token hash */
  if (ht_insert(info->responses, (dict_obj) m) != DICT_OK) {
    errno = 0;
    error("DoCGIResponseProcessing",
          "Error inserting response in response table: %d",
          info->dict_errno);
    goto failure;
  }

  /* The meta-variables for the response */
  if (m->type == message_real_message) {
    var_src_msg = m;
    reason = m->p.reason;
  }
  else if (m->type == message_pseudo_message){
    var_src_msg = GetRequestAsMessage(r);     /* XXX this is wrong... */
    reason = GetStandardReason(m->p.status);
  }
  else {
    errno = 0;
    error("DoCGIResponseProcessing", "Surprising message type %d",
          m->type);
    goto failure;
  }

  /* Standard metavars */
  if (AddStandardCGIVars(&metavars, r, &var_src_msg->p,
                         &var_src_msg->body, var_src_msg->source)) {
    goto failure;
  }

  /* Metavars specific to responses */
  sprintf(buf, "%d", m->p.status);
  if (AddMetavar(&metavars, "RESPONSE_STATUS", buf))
    goto failure;
  if (AddMetavar(&metavars, "RESPONSE_REASON", reason))
    goto failure;
  if (branch_info && branch_info->request_token)
    if (AddMetavar(&metavars, "REQUEST_TOKEN", branch_info->request_token))
      goto failure;
  if (AddMetavar(&metavars, "RESPONSE_TOKEN", (char *) m->policy_info))
    goto failure;

  /* Execute CGI script */
  if (ExecuteCGI(r, &metavars, m, do_default))
    goto failure;

  FreeMetavars(&metavars);
  if (m->type == message_pseudo_message) MessageDelete(var_src_msg);
  return 0;

failure:
  FreeMetavars(&metavars);
  if (m->type == message_pseudo_message && var_src_msg != NULL)
    MessageDelete(var_src_msg);
  return 1;
} /* DoCGIResponseProcessing */


/*
 * AddStandardCGIVars()
 * Add the standard CGI variables -- the ones that don't depend on the type
 * of message triggering this (request, response, etc.)
 */
static int AddStandardCGIVars(struct metavar_set *metavars, request_t *r,
                              parser_t *p, body_t *b,
                              sock_t source)
{
  char buf[80];     /* Big enough to handle numbers, etc.  Not used
                     * for strings. */
  struct cgi_info *info = (struct cgi_info *) r->policy_info;
  parser_header_t *h;

  sprintf(buf, "%d", b->content_length);
  AddMetavar(metavars, "CONTENT_LENGTH", buf);

  if (b->content_length > 0) {
    AddMetavar(metavars, "CONTENT_TYPE", b->content_type);
  }

  AddMetavar(metavars, "GATEWAY_INTERFACE", "SIP-CGI/1.1");

  for (h = p->headers; h != NULL; h = h->next) {
    if (strcmp(h->name, "Authorization") != 0) {
      /* Authorization happens in the server, so don't expose it to the
       * script.  Everything else is OK. */
      char *name, *value;
      name = canonicalize_header_name(h->name);
      value = canonicalize_header_value(h->elements);
      AddMetavar(metavars, name, value);
      free(name);
      free(value);
    }
  }

  if (info->reg != NULL) {
    AddRegistrations(metavars, info->reg);
  }

  //AddMetavar(metavars, "REMOTE_ADDR", inet_ntoa(source.sin_addr));
  /* REMOTE_HOST: we don't do DNS lookups on our source */
 
  AddMetavar(metavars, "SERVER_NAME", config.ProxyName);

  sprintf(buf, "%d", config.port);
  AddMetavar(metavars, "SERVER_PORT", buf);

  sprintf(buf, "%s/%s", p->protocol, p->version);
  AddMetavar(metavars, "SERVER_PROTOCOL", buf);

  sprintf(buf, "%s/%s", server_name, VERSION);
  AddMetavar(metavars, "SERVER_SOFTWARE", buf);

  if (info->script_cookie != NULL) {
    AddMetavar(metavars, "SCRIPT_COOKIE", info->script_cookie);
  }

  /* AUTH_TYPE, REMOTE_USER: we don't currently handle authentication for
   * INVITE, or CGI for REGISTER.  XXX TODO */

  /* REMOTE_IDENT: not supported */

  /* REQUEST_METHOD, REQUEST_URI: specific to requests */

  /* RESPONSE_STATUS, RESPONSE_REASON: specific to responses */

  /* REQUEST_TOKEN, RESPONSE_TOKEN: specific to responses (XXX: or
   * follow-up requests (ACK, CANCEL)???) */

  return 0;
} /* AddStandardCGIVars */


/*
 * FreeMetavars()
 * free the memory pointed to by a metavar structure
 */
static void FreeMetavars(struct metavar_set *metavars)
{
  int i;

  for (i = 0; i < metavars->num; i++) {
    free(metavars->vars[i]);
  }

  free(metavars->vars);
} /* FreeMetavars */


/* 
 * canonicalize_header_name()
 * Convert a SIP header name to a SIP_* name;
 * Caller must free generated name.
 */
static char *canonicalize_header_name(char *name)
{
  char *new_name;
  int name_len;
  int i,j;

  name_len = strlen(name);

  new_name = malloc(name_len + 5); /* "SIP_" plus null byte */

  strcpy(new_name, "SIP_");
  for (i = 4, j = 0; j < name_len; i++, j++) {
    if (name[j] == '-') {
      new_name[i] = '_';
    } else {
      new_name[i] = toupper((int)name[j]);
    }
  }
  new_name[i] = '\0';

  return new_name;
}


/* 
 * canonicalize_header_value()
 * Concatenate a set of SIP header values with ,'s
 * XXX: should this do URL-encoding or anything like that?
 */
static char *canonicalize_header_value(parser_element_t *e)
{
  DString str;
    
  DStringInit(&str);
  for (; e != NULL; e = e->next) {
    if (e->element == NULL) {
      continue;
    }
    if (str.length > 0) {
      DStringAppend(&str, ",", 1);
    }
    DStringAppend(&str, e->element, -1);
  }

  return DStringValue(&str);
} /* canonicalize_header_value */


/*
 * AddRegistrations()
 * Adds the REGISTRATIONS metavar to metavars, based on reg.
 * This logic is (mostly) duplicated from that in redirect.c.
 */
static void AddRegistrations(struct metavar_set *metavars, registration_t *reg)
{
  DString regs;
  int i;
  char t[80];
  struct tm tm;

  DStringInit(&regs);
  for (i = 0; i < reg->contacts && reg->contact[i].nameaddr; i++) {
    if (reg->contact[i].q >= 0) {
      if (regs.length > 0) {
        DStringAppend(&regs, ",", 1);
      }
        
      gmtime_r(&reg->contact[i].expires, &tm);
      strftime(t, sizeof(t), "%a, %b %d %Y %T GMT", &tm);
      
      if (reg->contact[i].q > 0) {
        DStringPrintf(&regs, "%s; expires=\"%s\"; q=%.2f",
                      reg->contact[i].nameaddr,
                      t, reg->contact[i].q);
      }
      else {
        DStringAppend(&regs, reg->contact[i].nameaddr, -1);
      }
    }
  }
  if (regs.length > 0) {
    AddMetavar(metavars, "REGISTRATIONS", DStringValue(&regs));
  }
  
  DStringFree(&regs);
} /* AddRegistrations */


/* 
 * AddMetavar()
 * Add a single metavariable to metavars.  Format: "NAME=value\0".
 * Allocates memory for variable; caller must free it.
 * Keeps the array element one more than the highest assigned var
 * NULL.
 * Returns non-zero on error.
 */
static int AddMetavar(struct metavar_set *metavars, const char *name,
                      const char *value)
{
  int name_len, value_len;

  while (metavars->allocated <= metavars->num + 1) {
    if (metavars->allocated == 0) {
      metavars->allocated = 2;
    }
    else {
      metavars->allocated *= 2;
    }
    metavars->vars = (char **) realloc(metavars->vars,
                                       metavars->allocated * sizeof (char *));
    if (metavars->vars == NULL) {
      error("AddMetavar", "couldn't allocate metavariable array");
      return 1;
    }
  }
    
  name_len = strlen(name);
  value_len = strlen(value);

  metavars->vars[metavars->num] = malloc(name_len + value_len + 2);
  if (!metavars->vars[metavars->num]) {
    error("AddMetavar", "couldn't allocate metavariable buffer");
    return 1;
  }

  strcpy(metavars->vars[metavars->num], name);
  metavars->vars[metavars->num][name_len] = '=';
  strcpy(metavars->vars[metavars->num] + name_len + 1, value);
  metavars->vars[metavars->num][name_len + value_len + 1] = '\0';

  metavars->vars[metavars->num + 1] = NULL;

  metavars->num++;

  return 0;
} /* AddMetavar */


/*
 * ExecuteCGI()
 * Execute a CGI script, and then perform all the policy actions it
 * indicates.
 * Return non-zero on failure.
 */
static int ExecuteCGI(request_t *r, struct metavar_set *metavars,
                      message_t *m, int *do_default)
{
  pid_t child_pid;
  int child_fds[3];
  struct cgi_info *info = (struct cgi_info *) r->policy_info;
  int retval = 0;
  int timed_out = 0;

  info->again = 0;
  *do_default = 1;

  child_pid = spawn_child(child_fds, info->script->filename, metavars->vars);
  if (child_pid < 0) {
    /* Fork failed */
    error("ExecuteCGI", "Couldn't spawn child process");
    return 1;
  }
  /* We're the parent */
  if (WriteBody(&m->body, &child_fds[0]) != 0) {
    retval = 1;
    goto cleanup;
  }
  
  if (ReadFromChild(r, child_fds, m, &timed_out, do_default) != 0) {
    retval = 1;
    /* Don't jump to cleanup:  try to reap the child no matter what happened
     * while reading */
  }

  if (ReapChild(r, child_pid, timed_out) != 0) {
    retval = 1;
  }

cleanup:
  /* Close fd's */

  return retval;
} /* ExecuteCGI */


/* 
 * spawn_child()
 * Set up file descriptors, fork, and exec the child.  Return -1 for
 * failure, otherwise pid of child.
 * For the parent, pass back file descriptors for stdin/out/err to the
 * child.  For the child, doesn't return.
 */
#ifndef WIN32

static pid_t spawn_child(int fds[3], char *script_filename, char *env[])
{
  pid_t pid = -1;
  int in_fds[2] = {-1,-1};
  int out_fds[2] = {-1, -1};
  int err_fds[2] = {-1, -1};
  int save_errno;

  if (pipe(in_fds) < 0) {
    goto cleanup;
  }
 
  if (pipe(out_fds) < 0) {
    goto cleanup;
  }

  if (pipe(err_fds) < 0) {
    goto cleanup;
  }

  if ((pid = fork()) < 0) {
    /* fork failed! */
    goto cleanup;
  }
  else if (pid == 0) {
    /* Child process */
    close(in_fds[1]);
    dup2(in_fds[0], 0); /* 0: stdin */
    close(in_fds[0]);
 
    close(out_fds[0]);
    dup2(out_fds[1], 1); /* 1: stdout */
    close(out_fds[1]);

    close(err_fds[0]);
    dup2(err_fds[1], 2); /* 2: stderr */
    close(err_fds[1]);

    /* Set resource limits? */

    /* exec the child */
    execle(script_filename, script_filename, NULL, env);
    
    /* If we still exist, something went wrong. */
    /* XXX does this error message show up anywhere useful? */
    error("spawn_child", "Couldn't execute script");
    _exit(1);
  }
  /* We're the parent process. */
  close(in_fds[0]);
  fds[0] = in_fds[1];
  
  close(out_fds[1]);
  fds[1] = out_fds[0];
  
  close(err_fds[1]);
  fds[2] = err_fds[0];

  return pid;

cleanup:
  /* Bailout if things screwed up.  Only called before fork or if fork
   * failed, so we don't have to worry about parent/child. */
  save_errno = errno;
  if (in_fds[0] != -1) {
    close(in_fds[0]);
    close(in_fds[1]);
  }
  if (out_fds[0] != -1) {
    close(out_fds[0]);
    close(out_fds[1]);
  }
  if (err_fds[0] != -1) {
    close(err_fds[0]);
    close(err_fds[1]);
  }

  errno = save_errno;
  return -1;
} /* spawn_child */

#else

static pid_t spawn_child(int fds[3], char *script_filename, char *env[])
{
  pid_t pid = -1;
  int in_fds[2] = {-1,-1};
  int out_fds[2] = {-1, -1};
  int err_fds[2] = {-1, -1};
  int hStdOut;
  int hStdIn;
  int hStdErr;
  int save_errno;
  FILE * fp;
  char commandline[8192];
  char **params = NULL;

  /* Open the script file in text mode */
  if ( script_filename == NULL || 
      (fp = fopen(script_filename, "r")) == NULL ) {
    debug("spawn_child", "File %s can not be opened\n",
      script_filename ? script_filename : "NULL");
    return pid; /* error */
  }
  commandline[sizeof(commandline)-1] = '\0';
  if ( fgets(commandline, sizeof(commandline) - 1, fp) != NULL 
    && commandline[0] == '#' && commandline[1] == '!' && commandline[2] != '\0') {
    char *ptr = strdup(&commandline[2]);
    char *l = NULL;
    int index = 0;

    while (1) {
      char *part = strtok_r(ptr, " \r\n", &l);
      ptr = NULL;
      if ( part == NULL ) {
        break; /* break from loop */
      }
      if ( *part == '\0' || *part == ' ' ) {
        continue; /* get next element */
      }
      params = (char **) realloc(params, sizeof(char *) * (index + 2));
      params[index+1] = NULL;
      params[index] = strdup(part);
      index ++;
    }
    if ( index == 0 ) {
      /* could not read  */
      params = (char **) calloc(2, sizeof(char *));
      params[0] = strdup(script_filename);
      /* param[1] is NULL */
    }
    else {
      params = (char **) realloc(params, sizeof(char *) * (index + 2));
      params[index+1] = NULL;
      params[index] = strdup(script_filename);
      index ++;
    }
  }
  else {
    /* could not read the script name or not a script with first line as #!/command/ */
    params = (char **) calloc(2, sizeof(char *));
    params[0] = strdup(script_filename);
    /* param[1] is NULL */
  }

  if (_pipe(in_fds, 512, O_BINARY | O_NOINHERIT) < 0) {
    goto cleanup;
  }
 
  if (_pipe(out_fds, 512, O_BINARY | O_NOINHERIT) < 0) {
    goto cleanup;
  }

  if (_pipe(err_fds, 512, O_BINARY | O_NOINHERIT) < 0) {
    goto cleanup;
  }

  hStdOut = _dup(_fileno(stdout));
  hStdIn  = _dup(_fileno(stdin));
  hStdErr = _dup(_fileno(stderr));

  _dup2(out_fds[1], _fileno(stdout));
  _dup2(in_fds[0], _fileno(stdin));
  _dup2(err_fds[1], _fileno(stderr));

  close(out_fds[1]);
  close(in_fds[0]);
  close(err_fds[1]);

  pid = spawnve(P_NOWAIT, params[0], params, env);
  
  _dup2(hStdOut, _fileno(stdout));
  _dup2(hStdIn, _fileno(stdin));
  _dup2(hStdErr, _fileno(stderr));
  close(hStdOut);
  close(hStdIn);
  close(hStdErr);

  if ( pid < 0 ) {
    error("spawn_child", "Could not create process (%d)\n", pid);
  }
  else {
    debug("spawn_child", "CreateProcess created process %d\n", pid);
  }

  /* We're the parent process. */
  fds[0] = in_fds[1];
  
  fds[1] = out_fds[0];

  fds[2] = err_fds[0];

  return pid;

cleanup:
  /* Bailout if things screwed up.  Only called before fork or if fork
   * failed, so we don't have to worry about parent/child. */
  save_errno = errno;
  if (in_fds[0] != -1) {
    close(in_fds[0]);
    close(in_fds[1]);
  }
  if (out_fds[0] != -1) {
    close(out_fds[0]);
    close(out_fds[1]);
  }
  if (err_fds[0] != -1) {
    close(err_fds[0]);
    close(err_fds[1]);
  }
  errno = save_errno;
  return -1;
} /* spawn_child */


#endif

/*
 * WriteBody()
 * Write the body of a message (if there is one) to the given fd (the
 * child's stdin), then close the socket.  Returns non-zero on error.
 */
static int WriteBody(body_t *b, int *child_in)
{
  int err;
  if (b->content_length > 0) {
    err = write(*child_in, b->content, b->content_length);
    if (err != b->content_length) {
      error("WriteBody", "Couldn't write request body to child %d != %d",
        err, b->content_length);
      return 1;
    }
  }
  err = close(*child_in);
  if (err) {
    error("WriteBody", "Couldn't close child's stdin");
    return 1;
  }
  *child_in = -1;

  return 0;
} /* WriteBody */


/*
 * ReadFromChild()
 * Read data from the child process's stdout and stderr.  If anything
 * comes out on stderr, log it, and return 1.  Otherwise, parse the output
 * of stdout, and dispatch policy events.
 * timed_out is set to 1 if we return because the cgi timeout expired.
 */
static int ReadFromChild(request_t *r, int child_fds[3],
                         message_t *m, int *timed_out, int *do_default)
{
  struct timeval timeout, remaining;
  fd_set fds;
  int maxfd;
  int err = 0, ret, done = 0;
  DString outbuf, errbuf;
  int err_offset = 0;
  int out_closed = 0, err_closed = 0;
  int readfrom = 0; /* Which fds have been read from?  0 - neither; 1 -
                       output; 2 - error.  Error trumps output. */
  struct parse_info out_parse;

  InitParseInfo(&out_parse);

  maxfd = (child_fds[1] > child_fds[2] ? child_fds[1] + 1 : child_fds[2] + 1);

  DStringInit(&outbuf);
  DStringInit(&errbuf);

  gettimeofday(&timeout, NULL);
  timeout.tv_sec += config.cgi_timeout;
  
  while (!done) {
#ifndef WIN32
    FD_ZERO(&fds);
    if (!out_closed)
      FD_SET(child_fds[1], &fds);
    if (!err_closed)
      FD_SET(child_fds[2], &fds);
    remaining = timeval_until(timeout);

    /* select() is guaranteed to block only this thread */
    err = select(maxfd, &fds, NULL, NULL, &remaining);
    if (err <= 0) {
      /* Either error or timeout */
      debug("ReadFromChild", "select returning %d\n", err);
      break;
    }
    if (FD_ISSET(child_fds[2], &fds)) {
      ret = ReadIntoDString(child_fds[2], &errbuf);
      if (ret < 0) {
        error("ReadFromChild", "Couldn't read from child's stderr");
        return 1;
      }
      else if (ret == 0) {
        err_closed = 1;
      }
      else {
        readfrom = 2;
        if (ParseChildStderr(r, &errbuf, &err_offset) != 0) {
          return 1;
        }
      }
    }
    if (FD_ISSET(child_fds[1], &fds)) {
      ret = ReadIntoDString(child_fds[1], &outbuf);
      if (ret < 0) {
        error("ReadFromChild", "Couldn't read child's output");
        return 1;
      }
      else if (ret == 0) {
        out_closed = 1;
      }
      if (readfrom != 2) {
        /* This gets executed when ret == 0 also, because
         * out_closed is newly set now. */
        readfrom = 1;
        if (DStringValue(&outbuf) != NULL) {
          debug("ReadFromChild",
                "r=%p: Read CGI output:\n%s\n", r,
                DStringValue(&outbuf));
        }
        if (ParseChildOutput(r, &outbuf, m, &out_parse, out_closed,
                             do_default) != 0) {
          return 1;
        }
      }
    }

    if (out_closed && err_closed) {
      done = 1;
    }
#else
    ret = ReadIntoDString(child_fds[1], &outbuf);
    if (ret < 0) {
      error("ReadFromChild", "Couldn't read child's output");

      /* Try reading stderr */
      ret = ReadIntoDString(child_fds[2], &errbuf);

      if (ret < 0) {
        error("ReadFromChild", "Couldn't read from child's stderr");
        return 1;
      }
      else if (ret == 0) {
        out_closed = 1;
      }
      else {
        readfrom = 2;
        if (ParseChildStderr(r, &errbuf, &err_offset) != 0) {
          return 1;
        }
      }
    }
    else if (ret == 0) {
      out_closed = 1;
    }
    if (readfrom != 2) {
      /* This gets executed when ret == 0 also, because
       * out_closed is newly set now. */
      readfrom = 1;
      if (DStringValue(&outbuf) != NULL) {
        debug("ReadFromChild",
              "r=%p: Read CGI output:\n%s\n", r,
              DStringValue(&outbuf));
      }
      if (ParseChildOutput(r, &outbuf, m, &out_parse, out_closed,
                           do_default) != 0) {
        return 1;
      }
    }
    if (out_closed ) {
      done = 1;
    }
#endif
  }

  DStringFree(&outbuf);
  DStringFree(&errbuf);

  if (err < 0) {
    error("ReadFromChild", "select error");
    return 1;
  }

  if (!out_closed || !err_closed) {
    *timed_out = 1;
  }

  if (readfrom == 2) {
    return 1;
  }

  else return 0;
} /* ReadFromChild() */


/*
 * ReapChild()
 * It's time to clean up after the process.  See if it's completed; if not,
 * whack it with some signals, then try again.
 */
static int ReapChild(request_t *r, pid_t pid, int timed_out)
{
#ifndef WIN32

  int signals[] = {0, SIGTERM, SIGKILL};
  const int numsignals = sizeof(signals) / sizeof(int);
  int status;
  int i; /* Which signal we're on */
  struct timeval timeout;
  int err;
  pid_t new_pid;

  new_pid = waitpid(pid, &status, WNOHANG);
  if (new_pid == 0) {
    /* The child has not exited */
    if (!timed_out) {
      /* If the child didn't time out (i.e. its sockets were closed normally),
       * it's probably in the process of closing itself, so give it a bit of
       * time before whacking it. */
      /* This is a little bit of a hack; to make the code cleaner, we "kill"
       * the subprocess with signal 0 (this means "test existence of
       * process") */
      i = 0;
    }
    else {
      /* Otherwise, skip signal 0 and go directly to SIGTERM. */
      i = 1;
    }
    for (i = 0; i < numsignals; i++) {
      if (signals[i] != 0) {
        errno = 0;
        error("ReapChild",
              "Terminating runaway CGI process %ld with signal %d",
	      (long)pid, signals[i]);
      }
      err = kill(pid, signals[i]);
      if (err != 0 && errno != ESRCH) {
        error("ReapChild", "Couldn't kill runaway process %ld", (long)pid);
        break;
      }
      else if (err == 0) {
        /* If errno was ESRCH, the process finished already; race condition
           since last waitpid.  Don't delay, but still wait() for the
           process. */

        /* If we just sent a signal, wait for cgi_timeout seconds.
           If we sent a pseudo-signal, wait for only 1 second. */
        if (signals[i] == 0) {
          timeout.tv_usec = 0;
          timeout.tv_sec = 1;
        }
        else {
          timeout.tv_usec = 0;
          timeout.tv_sec = config.cgi_timeout;
        }
      
        /* Wait for the duration of the timeout.  select() is guaranteed to
           block only this thread */
        err = select(0, NULL, NULL, NULL, &timeout);
        if (err != 0) {
          break;
        }
      }

      /* Check for the process again */
      new_pid = waitpid(pid, &status, WNOHANG);
      if (new_pid != 0) {
        break;
      }
    }
  }

  if (new_pid < 0) {
    error("ReapChild", "waitpid: Error waiting for CGI child process");
    return 1;
  }
  else if (new_pid == 0) {
    errno = 0;
    error("ReapChild",
          "Couldn't kill runaway CGI process (even tried SIGKILL)");
    return 1;
  }    
  else if (WIFSIGNALED(status)) {
    errno = 0;
    error("ReapChild", "Child died on signal %d", WTERMSIG(status));
    return 1;
  }
  else if (WEXITSTATUS(status) != 0) {
    errno = 0;
    error("ReapChild", "Child died with exit status %d", 
	  WEXITSTATUS(status));
    return 1;
  }

#endif

  return 0;
} /* ReapChild */
 

/*
 * ReadIntoDString()
 * Read from a FD into a DString.  Return the return value of the read()
 */
static int ReadIntoDString(int fd, DString *buf)
{
  int length;
  char readbuf[BUFSIZ];
  length = read(fd, readbuf, BUFSIZ);
#ifdef WIN32
  if ( length <= 0 ) {
    debug("ReadInfoDString", "read(%d) failed %d\n", fd, GetLastError());
    length = 0;
  }
#endif
  if (length <= 0)
    return length;
  DStringAppend(buf, readbuf, length);
  return length;
} /* ReadIntoDString */


/*
 * ParseChildStderr()
 * Parse the stderr coming out of the child, and do something useful
 * with it -- log it to the sipd error stream.
 */
static int ParseChildStderr(request_t *r, DString *readbuf, int *offset)
{
  char *line_end;
  char *line_start;

  line_start = DStringValue(readbuf) + *offset;

  while ((line_end = strchr(line_start, '\n')) != NULL) {
    errno = 0;
    *line_end = '\0';
    error("ParseChildStderr", "CGI script stderr: %s",
          line_start);

    line_start = line_end + 1;
  }

  *offset = line_start - DStringValue(readbuf);

  return 1;
} /* ParseChildStderr */


/* InitParseInfo()
 * Initialize the parse info structure used by ParseChildOutput().
 */
static void InitParseInfo(struct parse_info *p)
{
  DStringInit(&p->header);
  DStringInit(&p->body);
  p->offset = 0;
  p->state = Header;
  p->content_length = INFINITE_CONTENT_LENGTH;
  p->cl = 0;
  p->parsed = 0;
  p->parse_error = 0;
} /* InitParseInfo */


/*
 * ParseChildOutput()
 * This does the work of parsing the output of the CGI script,
 * and invoking the appropriate policy actions
 */
static int ParseChildOutput(request_t *r, DString *readbuf, message_t *m,
                            struct parse_info *p, int closed, int *do_default)
{
  char c;                   /* current character */

  /* Once around the loop for each message. */
  while (p->offset < readbuf->length) {
    debug("ParseChildOutput", "r=%p: top of loop, offset = %d\n",
          r, p->offset);
    while (p->cl < p->content_length && p->offset < readbuf->length) {
      c = readbuf->string[p->offset];
      p->offset++;
      DStringAppend(p->state == Body ? &p->body : &p->header, &c, 1);
      switch (p->state) {
      case Header:
        if (c == '\r') p->state = CR;
        else if (c == '\n') p->state = NL;
        break;
      case CR:
        if (c == '\n') p->state = CRNL;
        else p->state = Header;
        break;
      case CRNL:
        if (c == '\r') p->state = CRNLCR;
        else p->state = Header;
        break;
      case NL:
      case CRNLCR:
        if (c == '\n') {
          p->parsed = 1;
          if (p->header.length > 4) {
            if (HTTP_Parse(DStringValue(&p->header), &p->parse) < 0) {
              p->parse_error = 1;
            }
            p->content_length = HTTP_ContentLength(&p->parse);
            if (p->content_length == INFINITE_CONTENT_LENGTH) {
              /* CGI processing is subtlely different from regular SIP
               * processing.  The absence of a Content-Length field means that
               * there is no content body associated with this CGI request. */
              p->content_length = 0;
            }
          }
          else {
            p->content_length = -1;
            p->parse_error = 1;
          }
          DStringFree(&p->header);
          p->state = Body;
        }
        else p->state = Header;
        break; 
      case Body:
        p->cl++;
        break;
      }
    }
    if (p->cl < p->content_length) {
      /* This implies that we got here because we ran out of buffer,
       * not because we finished a message. */
      if (!closed) {
        /* More data should be forthcoming; wait for it. */
        return 0;
      }
      else if (!p->parsed) {
        /* Handle the "header lines, no body or separator, EOF" case. 
         * Add a dummy body separator to keep the RFC822 parser happy. */
        /* XXX: repeats code from above. */
        
        DStringAppend(&p->header, "\n\n", 2);
        p->parsed = 1;
        if (p->header.length > 4) {
          if (HTTP_Parse(DStringValue(&p->header), &p->parse) < 0) {
            p->parse_error = 1;
          }
          p->content_length = HTTP_ContentLength(&p->parse);
          if (p->content_length == INFINITE_CONTENT_LENGTH) {
            /* CGI processing is subtlely different from regular SIP
             * processing.  The absence of a Content-Length field means that
             * there is no content body associated with this CGI request. */
            p->content_length = 0;
          }
        }
        else {
          p->content_length = -1;
          p->parse_error = 1;
        }
      }
    }

    /* We now have a complete message, with header and content. */
    if (!p->parse_error) {
      body_t body, *body_p;

      if (p->content_length != 0) {
        parser_header_t *contenttype;
        
        contenttype = RFC822_Header(p->parse.headers, "Content-Type");
        body.content_type = contenttype->elements->element;
        body.content_length = p->content_length;
        body.content = DStringValue(&p->body);
        body_p = &body;
      }
      else {
        body_p = NULL;
      }
      if (DoCGIAction(r, m, &p->parse, body_p, do_default) != 0) {
        p->parse_error = 1;
      }
    }
    else {
      errno = 0;
      error("ParseChildOutput", "Couldn't parse output of CGI script");
    }
    HTTP_Free(&p->parse);

    DStringFree(&p->header);
    DStringFree(&p->body);
    p->state = Header;
    p->content_length = INFINITE_CONTENT_LENGTH;
    p->cl = 0;
    p->parsed = 0;

    /* Reset p->parse to null, so that HTTP_Free on it will do nothing.
     * Otherwise we can potentially double-free the structure. */
    p->parse.method = NULL;
    p->parse.URI_string = NULL;
    p->parse.reason = NULL;
    p->parse.protocol = NULL;
    p->parse.version = NULL;
    p->parse.headers = NULL;
    memset(&p->parse.URI, '\0', sizeof(uri_t));

    if (p->parse_error) {
      return 1;
    }
    p->parse_error = 0;
  }

  return 0;
} /* ParseChildOutput */


/*
 * DoCGIAction
 * *Finally*, actually execute a CGI action after it gets parsed.
 */
static int DoCGIAction(request_t *r, message_t *m, parser_t *header,
                       body_t *body, int *do_default)
{
  struct cgi_info *info = (struct cgi_info *) r->policy_info;
  struct cgi_branch_info branch_info = {NULL};
  parser_header_t *to_delete;
  char *protocol = HTTP_Protocol(header);
  message_t *msg_to_send = NULL;
  message_t *orig_message;

  /* Make sure it wants SIP. */
  if (strcmp(protocol, "SIP") != 0) {
    errno = 0;
    error("DoCGIAction", "CGI invoked unknown protocol %s",
          protocol);
    free(protocol);
    policy_send_new_response(r, 501, NULL);
    goto failure;
  }
  free(protocol);

  /* Generate the delta-headers, and remove any CGI headers */
  HandleAndRemoveCGIHeaders(r, &header->headers, &branch_info, &to_delete);

  if (header->type == http_request) {
    char *method = HTTP_Method(header);
    
    /* This is a "request" */
    if (strcasecmp(method, "CGI-PROXY-REQUEST") == 0) {
      struct cgi_branch_info *bip =
        (struct cgi_branch_info *) malloc(sizeof(struct cgi_branch_info));

      *do_default = 0;
      if (bip == NULL)
        goto failure;

      orig_message = GetRequestAsMessage(r);
      msg_to_send = MergeMessageChanges(orig_message, header, body, to_delete);
      MessageDelete(orig_message);
      memcpy(bip, &branch_info, sizeof(struct cgi_branch_info));

      policy_proxy_original_request_to(r, &header->URI, msg_to_send,
                                       bip);
    }
    else if (strcasecmp(method, "CGI-FORWARD-RESPONSE") == 0) {
      *do_default = 0;

      orig_message = LookupResponseToken(r, m, header->URI_string);
      if (orig_message == NULL)
        goto failure;

      if (orig_message->type == message_pseudo_message) {
        char *reason = HTTP_Reason(&orig_message->p);
        headers_t new_headers = {NULL, NULL}, del_headers = {NULL, NULL};
        
        HeaderCopyParsedHeaders(&new_headers, header->headers);
        CopyCGIRemove(&del_headers, to_delete);
        
        policy_send_modified_new_response(r, HTTP_Status(&orig_message->p),
                                          reason,
                                          &new_headers, &del_headers, body);
        free_valid(reason);
        HeaderFree(&new_headers);
        HeaderFree(&del_headers);
      }
      else {
        msg_to_send = MergeMessageChanges(orig_message, header, body,
                                          to_delete);
        policy_forward_response(r, msg_to_send);
      }
    }
    else if (strcasecmp(method, "CGI-AGAIN") == 0) {
      if (strcasecmp(header->URI_string, "yes") == 0) {
        info->again = 1;
      }
      /* Accept anything that isn't "yes" as "no". */
    }
    else if (strcasecmp(method, "CGI-SET-COOKIE") == 0) {
      strassign(&info->script_cookie, header->URI_string);
    }
    else {
      errno = 0;
      error("DoCGIAction", "CGI script invoked unimplemented method %s",
            method);
      goto failure;
    }
    free(method);
  }
  else {
    /* This is a locally-generated response. */
    char *reason = HTTP_Reason(header);
    headers_t new_headers = {NULL, NULL}, del_headers = {NULL, NULL};

    *do_default = 0;

    HeaderCopyParsedHeaders(&new_headers, header->headers);
    CopyCGIRemove(&del_headers, to_delete);
    policy_send_modified_new_response(r, HTTP_Status(header), reason,
                                      &new_headers, &del_headers, body);
    free_valid(reason);
    HeaderFree(&new_headers);
    HeaderFree(&del_headers);

    if (HTTP_Status(header) >= 200 && HTTP_Status(header) < 300) {
      info->sent_local_200 = 1;
    }
    if (HTTP_Status(header) >= 200) {
      policy_set_timeout(r, 30); /* Wait for final expiration */
    }
  }

  /* If we sent a final response as a result of this, cancel all unfinished
     branches. */
  if (r->sent_final) {
    CancelUnfinishedBranches(r);
  }

  if (msg_to_send != NULL)
    MessageDelete(msg_to_send);
  RFC822_Free(to_delete);
  return 0;

failure:
  if (msg_to_send != NULL)
    MessageDelete(msg_to_send);
  RFC822_Free(to_delete);
  return 1;
} /* DoCGIAction */


/*
 * HandleAndRemoveCGIHeaders 
 * Handle all the CGI headers in hh, removing them from hh in the process.
 */
static int HandleAndRemoveCGIHeaders(request_t *r, parser_header_t **hlist,
                                     struct cgi_branch_info *bi,
                                     parser_header_t **to_remove)
{
  parser_header_t *h, *h_next, **prev_next;

  *to_remove = NULL;
  
  prev_next = hlist;

  for (h = *hlist; h != NULL; h = h_next) {
    h_next = h->next;
    if (strncasecmp(h->name, "CGI-", 4) == 0) {
      /* We match, so remove this entry from hlist. */
      *prev_next = h->next;
      h->next = NULL;

      /* Handle individual headers */
      if (strcasecmp(h->name, "CGI-Request-Token") == 0) {
        /* Only take the first CGI-Request-Token, if several are
         * specified. */
        strassign(&bi->request_token, h->elements->value);
      }
      else if (strcasecmp(h->name, "CGI-Remove") == 0) {
        *to_remove = h;
        /* Don't delete this header. */
        continue;
      }
      /* Delete everything that hasn't continue'd. */
      RFC822_Free(h);
    }
    else {
      /* Headers which don't match "CGI-" are normal headers,
       * and should be preserved.
       * Note that "Expires" is handled by client.c, for all policies, and
       * thus doesn't need to be handled specially here. */
      prev_next = &h->next;
    }
  }
  return 0;
} /* HandleAndRemoveCGIHeaders */


/*
 * CopyCGIRemove
 * Copy a CGI-Remove parser_header_t to a headers_t for
 * policy_send_modified_new_response.
 */
static void CopyCGIRemove(headers_t *del_headers, parser_header_t *to_delete)
{
  parser_header_t *h;
  parser_element_t *e;

  for (h = to_delete; h != NULL; h = h->next) {
    for (e = h->elements; e != NULL && e->element != NULL; e = e->next) {
      HeaderAppend(del_headers, e->element, "deleteme");
    }
  }
} /* CopyCGIRemove */


/*
 * MergeMessageChanges()
 * Merge the changes represented by 'header', 'body', and 'to_delete'
 * into 'm'.
 */
static message_t *MergeMessageChanges(message_t *orig, parser_t *header,
                                      body_t *body,
                                      parser_header_t *to_delete)
{
  message_t *m;
  parser_header_t *h;
  parser_element_t *e;

  m = MessageDuplicate(orig);

  if (HTTP_ContentLength(header) != INFINITE_CONTENT_LENGTH) {
    /* The CGI request specified a Content-Length header; therefore,
     * the body of the CGI request overrides the body of the original
     * message. */
    strassign(&m->body.content_type, body->content_type);
    m->body.content_length = body->content_length;
    strassign(&m->body.content, body->content);
  }
  
  for (h = header->headers; h != NULL; h = h->next) {
    RFC822_HeaderReplaceOrAppend(&m->p.headers, h);
  }

  for (h = to_delete; h != NULL; h = h->next) {
    for (e = h->elements; e != NULL && e->element != NULL; e = e->next) {
      RFC822_HeaderDelete(&m->p.headers, e->element);
    }
  }

  return m;
} /* MergeMessageChanges */


/*
 * RFC822_HeaderReplaceOrAppend()
 * Add the header in 'add' to the header list in 'php'.
 */
static void RFC822_HeaderReplaceOrAppend(parser_header_t **php,
                                         parser_header_t *orig_add)
{
  parser_header_t *h, *add, *orig_add_old_next;
  orig_add_old_next = orig_add->next;
  orig_add->next = NULL;
  RFC822_Duplicate(&add, orig_add);
  orig_add->next = orig_add_old_next;
  
  /* Check if name exists already. */
  h = RFC822_Header(*php, add->name);

  if (!h) {
    /* If not, prepend 'add' to list.  (Note that the parser_header_t is
     * sorted in reverse order, so this corresponds to the end of the
     * message. */
    add->next = *php;
    *php = add;
  }
  else {
    /* Otherwise, put the new entry in place of the old one.  Do this by
     * swapping the values of the structure members of 'h' and 'old', so
     * that the previous header's 'next' pointer still points to the right
     * place.  Then free the old value, along-with the just-allocated
     * 'add'.
     */
    char *tmp_name;
    parser_element_t *tmp_e;

    tmp_name = h->name;
    h->name = add->name;
    add->name = tmp_name;
    
    tmp_e = h->elements;
    h->elements = add->elements; 
    add->elements = tmp_e;

    tmp_e = h->last_element;
    h->last_element = add->last_element;
    add->last_element = tmp_e;

    add->next = NULL;
    RFC822_Free(add);
  }
} /* RFC822_HeaderReplaceOrAppend */


/*
 * RFC822_HeaderDelete()
 * Delete the header with the name 'name' from 'php', if it exists.
 */
static void RFC822_HeaderDelete(parser_header_t **php, char *name)
{
  parser_header_t *h, *h_next, **prev_next;

  prev_next = php;

  for (h = *php; h != NULL; h = h_next) {
    h_next = h->next;
    if (strcasecmp(h->name, name) == 0) {
      *prev_next = h->next;
      h->next = NULL;
      RFC822_Free(h);
    }
    else {
      prev_next = &h->next;
    }
  }
} /* RFC822_HeaderDelete */


/*
 * LookupResponseToken()
 * Find the message corresponding to the response token 'token'.  If none
 * found, return NULL.  'm' is the current response.
 */
static message_t *LookupResponseToken(request_t *r, message_t *m,
                                      char *token)
{
  struct cgi_info *info = (struct cgi_info *) r->policy_info;

  if (strcasecmp(token, "this") == 0) {
    if (m->p.type == http_response) {
      return m;
    }
    else {
      error("LookupResponseToken",
            "\"FORWARD-RESPONSE this\" invoked on non-response trigger!");
      return NULL;
    }
  }
  else {
    dict_obj o;

    o = ht_search(info->responses, (dict_key) token);
    if (o == NULL) {
      debug("LookupResponseToken",
            "FORWARD-RESPONSE called for unknown response token \"%s\"\n",
            (char *) token);
    }
    return (message_t *)o;
  }
} /* LookupResponseToken */


/*
 * NewResponseToken()
 * Create and set a new response token for m.
 * Right now we just print r and m as pointers to a string.
 */
static void NewResponseToken(request_t *r, message_t *m)
{
  char buf[80];
#if HAVE_SNPRINTF
  snprintf(buf, 80, "%p-%p", r, m);
#else
  /* Safe except on architectures with very weird pointers */
  sprintf(buf, "%p-%p", r, m);
#endif

  strassign((char **)&m->policy_info, buf);
}


/*
 * CleanOutMessageQueue()
 * "Clean" the response's message queue -- delete 1xx's, and send any
 * 2xx's.
 */
static void CleanOutMessageQueue(request_t *r)
{
  message_t *m;

  for (m = r->mqueue; m != NULL; m = m->next) {
    if (m->handled != policy_handled)
      continue;

    if (m->p.status < 200) {
      /* We found a 1xx response.  Mark it to be ignored. */
      m->handled = finished_handling;
    }

    else if (m->p.status < 300) {
      /* We found a 2xx response. */
      /* Send it immediately, and cancel all other branches. */
      policy_forward_response(r, m);

      /* If we get multiple 200's, we may cancel more than once here; but
       * that's okay. */
      CancelUnfinishedBranches(r);
      
      policy_set_timeout(r, (double)config.timeout);
      m->handled = finished_handling;
    }
  }
} /* CleanOutMessageQueue */


/*
 * hash_message()
 * hash a message (passed as a dict_obj, which is really a void *).  This
 * really means hash its key. 
 */
static ht_val hash_message(dict_obj o)
{
  message_t *m = (message_t *) o;

  return hash_key((dict_key) m->policy_info);
} /* hash_message */


/*
 * hash_key()
 * hash a message key (passed as a dict_key, which is really a void *).
 * We do the standard technique for string hashes, lifted from Tcl: multiply
 * by 9 and add.
 */
static ht_val hash_key(dict_key k) {
  int result = 0;
  char *c = (char *) k;

  if (c == NULL) return 0;

  while (*c != '\0') {
    result += (result << 3) + *c;
    c++;
  }
  return result;
} /* hash_key */


/*
 * msg_compare()
 * Compare the keys of two messages for order -- return -1, 0, 1.
 * Since they're strings, just use strcasecmp on the keys.
 */
static int msg_compare(dict_obj obj1, dict_obj obj2)
{
  message_t *m1, *m2;
  m1 = (message_t *) obj1;
  m2 = (message_t *) obj2;

  return strcasecmp((char *) m1->policy_info, (char *) m2->policy_info);
} /* msg_compare */


/*
 * msg_key_compare()
 * Compare a key with a message for order -- return -1, 0, 1.
 * Since they're strings, just use strcmp on the keys.
 */
static int msg_key_compare(dict_key key, dict_obj obj) 
{
  char *k = (char *) key;
  message_t *m = (message_t *) obj;

  return strcasecmp(k, (char *) m->policy_info);
} /* msg_key_compare */
