/*
* enginepgp.c
*
* Purpose:
* Add an abstraction layer above PGP sdk to provide with
* higher-level security functions.
* 
* Written-by: Triantafyllos Alexiou
*             ta154@columbia.edu, t.alexiou@ieee.org, alexiou@hyper.gr
*  December 1999
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "enginepgp.h"
#include "pgpPublicKey.h"
#include "pgpRandomPool.h"
#include "config.h"
#include <time.h>

PGPUInt32 gClientAPIVersion= kPGPsdkAPIVersion;
PGPKeySetRef gPgpKeySet=NULL, gPgpSecKeySet=NULL, gPgpPubKeySet=NULL;
PGPFileSpecRef gPgpPubFileSpec=NULL, gPgpSecFileSpec=NULL;
int gPgpDebugFlag=0;

FILE*  gPgpReportFile;
PGPContextRef gPgpContext=NULL;

/*
* ============== BEGIN GENERAL GUIDELINES  ==============
* Allocated memory space that is returned by the following 
* functions to caller, should be de-allocated by the PGPFreeData
* PGPsdk function. The associated PGPContext cannot be freed prior
* to freeing the allocated space.
* ============== END  GENERAL GUIDELINES  ==============
*/


/*
 * InitializePGPsdk()
 * Initialize the PGP SDK code.
 */
void InitializePGPsdk(void)
{    
  extern char *fullpath_r(char *root, char *filename);
  PGPError err;
  char *secFPath, *pubFPath;
  secFPath = fullpath_r(config.ServerRoot, config.PgpSecRing);
  pubFPath = fullpath_r(config.ServerRoot, config.PgpPubRing);
  gPgpDebugFlag = debug_flag;
  err = EnginePgpInit(secFPath,pubFPath, 1); /* debug_flag */
  
  if (secFPath) free(secFPath);
  if (pubFPath) free(pubFPath);
} /* InitializePGPsdk */


/*
* EnginePgpInit()
* It initialiazes a PGPContext ans associates with it the keyrings passed as
* arguments. 
*/
PGPError EnginePgpInit(char *secRingFullPathName, char *pubRingFullPathName, int debugFlag) 
{
#undef _func_name
#define _func_name "EnginePgpInit"
 
  PGPError pgpErr=kPGPError_NoErr;
  gPgpReportFile= stderr;
  if ( !secRingFullPathName || !pubRingFullPathName ) return -1;
  gPgpDebugFlag=debugFlag;
  if ( IsPGPError(pgpErr=PGPsdkInit() ) ) return pgpErr;
  
  pgpErr=PGPGetSDKVersion(&gClientAPIVersion);
  if (IsPGPError(pgpErr)){
    if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":1", pgpErr);
    return pgpErr;
  }  

  if (IsPGPError(pgpErr=PGPNewContext(gClientAPIVersion, &gPgpContext) )) {
    if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":1", pgpErr);
    PGPsdkCleanup();
    return pgpErr;
  }
  if (IsPGPError(pgpErr=PGPNewFileSpecFromFullPath(gPgpContext,secRingFullPathName, 
						   &gPgpSecFileSpec)) ){
    gPgpSecFileSpec=NULL;
    if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":2",pgpErr);
    PGPFreeContext(gPgpContext);gPgpContext=NULL;
    PGPsdkCleanup();
    return pgpErr;
  }
  if (IsPGPError(pgpErr=PGPNewFileSpecFromFullPath(gPgpContext, pubRingFullPathName, 
						   &gPgpPubFileSpec)) ){
    if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":3",pgpErr);
    PGPFreeFileSpec(gPgpSecFileSpec);gPgpSecFileSpec=NULL;
    PGPFreeContext(gPgpContext);gPgpContext=NULL;
    PGPsdkCleanup();
    return pgpErr;
  }
  if (IsPGPError(pgpErr=PGPOpenKeyRingPair(gPgpContext,
					   kPGPKeyRingOpenFlags_Mutable
					   //kPGPKeyRingOpenFlags_Create|
					   // kPGPKeyRingOpenFlags_Trusted|
					   //kPGPKeyRingOpenFlags_Private
					   ,
					   gPgpPubFileSpec,gPgpSecFileSpec,
					   &gPgpKeySet))){
    if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":4",pgpErr);
    PGPFreeFileSpec(gPgpPubFileSpec);gPgpPubFileSpec=NULL;
    PGPFreeFileSpec(gPgpSecFileSpec);gPgpSecFileSpec=NULL;
    PGPFreeContext(gPgpContext);gPgpContext=NULL;
    PGPsdkCleanup();
    return pgpErr;
  }
  return kPGPError_NoErr;
}/*  EnginePgpInit */


/* =============== EnginePgpDeInit =============== */
/*
* EnginePgpDeInit()
* Deinitializes the PGPContext and PGPFrees any memory allocated by
* the EnginePgpInit().
*/
PGPError EnginePgpDeInit() 
{
#undef _func_name
#define _func_name "EnginePgpDeInit"

  PGPError pgpErr=kPGPError_NoErr;
  if (gPgpPubFileSpec){
    if (IsPGPError(pgpErr=PGPFreeFileSpec(gPgpPubFileSpec)))
      if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":1", pgpErr);
    gPgpPubFileSpec=NULL;
  }
  if(gPgpSecFileSpec){
    if (IsPGPError(pgpErr=PGPFreeFileSpec(gPgpSecFileSpec)))
      if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":2", pgpErr);
    gPgpSecFileSpec=NULL;
  }
  if(gPgpSecKeySet){
    if (IsPGPError(pgpErr=PGPFreeKeySet(gPgpSecKeySet)))
      if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":3", pgpErr);
    gPgpSecKeySet=NULL;
  }
  if(gPgpPubKeySet){
    if (IsPGPError(pgpErr=PGPFreeKeySet(gPgpPubKeySet)))
      if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":4", pgpErr);
    gPgpPubKeySet=NULL;
  }
  if(gPgpKeySet){
    if (IsPGPError(pgpErr=PGPFreeKeySet(gPgpKeySet)))
      if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":5", pgpErr);
    gPgpKeySet=NULL;
  }
  if(gPgpContext){
   if (IsPGPError(pgpErr=PGPFreeContext(gPgpContext)))
     if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":6", pgpErr);
   gPgpContext=NULL;
  }
  if (IsPGPError(pgpErr=PGPsdkCleanup()))
    if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":7", pgpErr);
 return pgpErr;
}/* EnginePgpDeInit */


/*
* EnginePgpOSignWithPvtKey()
* Verifies the theSignature on theText. The length of the former is
* provided by theSgnSize and of the latter by theTextSize.
* You can supply a desired signing key in the theSignKey parameter.
* If theSignKey==NULL then it searches in  theKeySet for a key 
* that matches the *theSigner. 
* theHashAlgID specifies the desired hash algorithm is NOT USED.
* theIsArmored is currently NOT USED BUT MAY BE used in the future.
* You should provided it the correct value if known.
* This function will return an error if the found or provided key
* is disabled, revoked or expired.
*/
PGPError EnginePgpOSignWithPvtKey(PGPContextRef thePgpContext,
				  PGPKeySetRef theKeySet, 
				  char* theText, PGPSize theTextSize,
				  char **theSignatureRef,
				  PGPSize *theSgnSizePtr,
				  char* theSigner, char *thePassPhrase,
				  PGPKeyRef theSignKey,
				  PGPHashAlgorithm theHashAlgID,
				  PGPBoolean theIsArmored
				  )
{
#undef _func_name
#define _func_name "EnginePgpOSignWithPvtKey"
  
 PGPError pgpErr=kPGPError_BadSignature; 
 PGPKeyRef aKey;
 PGPOptionListRef aOptionList;
 PGPBoolean aSgnBufferIsProvided;

 if (!theSignKey) {
   pgpErr=EnginePgpSearchForUserKey(thePgpContext,theKeySet,
				    theSigner ,&aKey);
   if (IsPGPError(pgpErr)){
     if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":1", pgpErr);
     return pgpErr;
   }
 } else {
   aKey=theSignKey;
 }

 pgpErr=PGPNewOptionList(thePgpContext, &aOptionList);
 if (pgpErr){
   if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":2", pgpErr);
   return pgpErr;
 }
 pgpErr=PGPAppendOptionList(thePgpContext,aOptionList,
			    //PGPOAskUserForEntropy(thePgpContext,1 ),
			    //PGPOClearSign(thePgpContext,1),
			    PGPOInputBuffer(thePgpContext,theText,theTextSize),
			    PGPOArmorOutput(thePgpContext,theIsArmored),
			    PGPODetachedSig(thePgpContext,
					    PGPOLastOption(thePgpContext)),
			    PGPOSignWithKey(thePgpContext,aKey,
					    PGPOPassphrase(thePgpContext,
							   thePassPhrase),
					    PGPOKeySetRef(thePgpContext,
							  theKeySet ),
					    PGPOLastOption(thePgpContext)),
			    PGPOLastOption(thePgpContext)
			    );
 if (pgpErr){
   if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":3", pgpErr);
   PGPFreeOptionList(aOptionList);
   *theSgnSizePtr=0;
   return pgpErr;
 }
  aSgnBufferIsProvided=0;
 if (*theSignatureRef!=NULL) aSgnBufferIsProvided=1;

 if(!aSgnBufferIsProvided){
   pgpErr=PGPAppendOptionList(thePgpContext, aOptionList,
			      PGPOAllocatedOutputBuffer(thePgpContext,
							(void**) theSignatureRef,
							(PGPSize) (1024*1024),
							theSgnSizePtr),
			      PGPOLastOption(thePgpContext));
   if (pgpErr){
     if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":4", pgpErr);
     PGPFreeOptionList(aOptionList);
     *theSignatureRef=NULL; 
     //just make sure; assumes that no space has been allocate in case of error.
     *theSgnSizePtr=0;
     return pgpErr;
   }
 }else {
   pgpErr=PGPAppendOptionList( thePgpContext, aOptionList,
			       PGPOOutputBuffer(thePgpContext,
						*((void**) theSignatureRef),
						*theSgnSizePtr, theSgnSizePtr),
			       PGPOLastOption(thePgpContext));
   if (pgpErr){
     if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":5", pgpErr);
     PGPFreeOptionList(aOptionList);
     *theSgnSizePtr=0;
     return pgpErr;
   }
 }

 pgpErr=PGPEncode(thePgpContext,aOptionList, PGPOLastOption(thePgpContext));
 if (pgpErr){
   if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":6", pgpErr);
   if(!aSgnBufferIsProvided) *theSignatureRef=NULL;
   *theSgnSizePtr=0;
 }
 PGPFreeOptionList(aOptionList);
 return pgpErr;

}/*  EnginePpgOSignWithPvtKey */


/*
* EnginePgpOVerifySignatureWithPubKeyEvtHandler()
* The PGPDecode function has to have an EventHandler function.
* The handler functions are essential. They should all error.
* Otherwise the PGPDecode may not return an error though there is one.
* For example an invalid signature may be not be catched.
*/
static PGPError
EnginePgpOVerifySignatureWithPubKeyEvtHandler(PGPContextRef thePgpContext,
					      PGPEvent *theEvent,
					      PGPUserValue theUserValue )
{
#undef _func_name
#define _func_name "EnginePgpOVerifySignatureWithPubKeyEvtHandler"
  PGPError pgpErr=kPGPError_BadSignature;
  PGPEventSignatureData sgnData;
  // if (gPgpDebugFlag){
    // EnginePgpPrintEvent(gPgpReportFile, theEvent);
  
  switch ( theEvent->type){
  case kPGPEvent_InitialEvent: 
    *((PGPError*)theUserValue)= kPGPError_BadSignature;
    return kPGPError_NoErr;
    break;
  case kPGPEvent_AnalyzeEvent:
    {
      PGPEventAnalyzeData anlzData=theEvent->data.analyzeData;
      switch(anlzData.sectionType){
      case kPGPAnalyze_DetachedSignature:
	pgpErr=kPGPError_NoErr;
	break;
      case kPGPAnalyze_Encrypted:
      case kPGPAnalyze_Key:
      case kPGPAnalyze_Signed: 
      case kPGPAnalyze_Unknown:
      default:
	pgpErr=kPGPError_BadSignature;
	break;
      }
    }
    break;	  
  case  kPGPEvent_SignatureEvent:
    sgnData=(theEvent->data.signatureData);
    if( sgnData.keyRevoked ) { pgpErr=kPGPError_KeyRevoked;  break; }
    if( sgnData.keyDisabled ){ pgpErr=kPGPError_KeyDisabled; break; }
    if( sgnData.keyExpired ) { pgpErr=kPGPError_KeyExpired;  break; }
    if( sgnData.keyValidity==kPGPValidity_Invalid ||
	 sgnData.keyValidity==kPGPValidity_Unknown ||
	 sgnData.keyMeetsValidityThreshold==FALSE ){
      pgpErr=kPGPError_KeyInvalid;  break;
    }
    if (!sgnData.verified || !sgnData.checked){
      pgpErr=kPGPError_BadSignature;break;
    }   
    pgpErr=kPGPError_NoErr;
    break;
  case kPGPEvent_FinalEvent:
    pgpErr=*((PGPError*)theUserValue);
    return pgpErr;
    //The error returned after a FinalEvent is ignored by PGPDecode.
    break;
  default:
    return kPGPError_NoErr;
    break;
  } 

  *((PGPError*)theUserValue)= pgpErr;
  return pgpErr;

}/*  EnginePgpOVerifySignatureWithPubKeyEvtHandler */


/*
* EnginePgpOVerifySignatureWithPubKey()
* This function signs with theText with a private key.
* The returned signature is  **theSignatureRef and the size of it 
* *theSgnSizePtr. if *theSignatureRef==NULL the function itself allocates 
* space for the signature. if  *theSignatureRef!=NULL the function assumes
* that space of *theSgnSizePtr size for the signature is provided.
* Upon return the  *theSgnSizePtr contains the signature size.
* You can supply a desired signing key in the theSignKey parameter.
* If theSignKey==NULL then it searches in  theKeySet for a key 
* that matches the *theSigner. theHashAlgID specifies the
* desired hash algorithm for signing.
*/
PGPError
EnginePgpOVerifySignatureWithPubKey(PGPContextRef thePgpContext,
				    PGPKeySetRef theKeySet, 
				    char* theText, PGPSize theTextSize,
				    char *theSignature,
				    PGPSize theSgnSize,
				    char* theSigner,
				    PGPKeyRef theSignKey,
				    PGPHashAlgorithm theHashAlgID,
				    PGPBoolean theIsArmored
				    )
{
#undef _func_name
#define _func_name "EnginePgpOVerifySignatureWithPubKey"

  PGPError pgpErr=kPGPError_NoErr; 
  PGPKeyRef aKey;
  PGPOptionListRef aOptionList; 
  PGPError aHandlerErr;
  
  if (!theSignature || !theSgnSize){
    pgpErr=kPGPError_BadSignature;
    if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":1",pgpErr);
    return pgpErr;
  }

  if (!theSignKey) {
    pgpErr=EnginePgpSearchForUserKey(thePgpContext,theKeySet,
				     theSigner ,&aKey);
    if (IsPGPError(pgpErr)){
      if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":2", pgpErr);
      return pgpErr;
    }
  } else {
    aKey=theSignKey;
  }

  pgpErr=PGPNewOptionList(thePgpContext, &aOptionList);
  if (pgpErr){
    if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":3", pgpErr);
    return pgpErr;
  }

  pgpErr=PGPNewSingletonKeySet(aKey,&theKeySet);
  if (pgpErr){
    if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":4", pgpErr);
    return pgpErr;
  }

  pgpErr=PGPAppendOptionList(thePgpContext,aOptionList,
			     PGPOInputBuffer(thePgpContext, theSignature,
					     theSgnSize),
			     PGPODiscardOutput(thePgpContext, 1),
			     PGPODetachedSig(thePgpContext,
					     PGPOInputBuffer(thePgpContext,
							     theText,
							     theTextSize
							     ),
			          PGPOLastOption(thePgpContext)),
			     PGPOEventHandler
			     (thePgpContext,	      
			      EnginePgpOVerifySignatureWithPubKeyEvtHandler,
			      &aHandlerErr),  
			     PGPOKeySetRef(thePgpContext,theKeySet ),
			     PGPOLastOption(thePgpContext));
  
  if (pgpErr){
    if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":5", pgpErr);
    PGPFreeKeySet(theKeySet);
    PGPFreeOptionList(aOptionList);
    return pgpErr;
  }
  pgpErr=PGPDecode(thePgpContext,aOptionList,PGPOLastOption(thePgpContext));
  
  if (aHandlerErr) pgpErr=aHandlerErr;

  if (IsPGPError(pgpErr))
    if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":6", pgpErr);
  PGPFreeKeySet(theKeySet);
  PGPFreeOptionList(aOptionList);
  return pgpErr;

}/* EnginePgpOVerifySignatureWithPubKey */


/*
* EnginePgpOEncryptToUserIDs()
* Encrypts theText by using public keys. The keys can be provided through
* *theRcptsKeySetRefPtr. If this is NULL then their userIDs should be provided
* in theRecipients, which is a NULL-terminated array of pointers to userIDs.
* CURRENTLY ONLY (*theRecipients)[0] IS USED. The keys should be contained in
* the specified theKeySet.  theText contains the plain text and theTextSize
* its size in bytes. if theIsArmored is set than the encrypted output is 
* armored. A buffer for the ciphertext can be provided in theCipherTextRef,
* along with its size in theCipherTextSizePtr, or it is allocated by the 
* function and the size of the ciphertext is returned in theCipherTextSizePtr.
* IF THE BUFFER is allocated by the function, it should be freed by PGPFreeData().
*/
PGPError
EnginePgpOEncryptToUserIDs( PGPContextRef thePgpContext,
			   PGPKeySetRef theKeySet, 
			   char* theText, PGPSize theTextSize,
			   char** theCipherTextRef,
			   PGPSize *theCipherTextSizePtr,
			   char** theRecipients,
			   PGPKeySetRef *theRcptsKeySetRefPtr,
			   PGPBoolean theIsArmored
			   )
{
  
#undef _func_name
#define _func_name "EnginePgpOEncryptToUserIDs"
  
  PGPError pgpErr= -1; //kPGPError_BadSignature; 
  PGPKeySetRef *aKeySetRefPtr= NULL, aSingletonKeySet;
  PGPKeyRef aKey;
  PGPOptionListRef aOptionList;
  PGPBoolean aEncBufferIsProvided;
 
  if ( !theRcptsKeySetRefPtr ) {
   
    if ( !theRecipients ) {
      if (IsPGPError(pgpErr)){
	if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":1", pgpErr);
	return pgpErr;
      }
    }
    aKeySetRefPtr= &aSingletonKeySet;
    
    pgpErr= EnginePgpSearchForUserKey( thePgpContext, theKeySet,
					 theRecipients[0], &aKey );
    if (IsPGPError(pgpErr)){
      if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":1.1", pgpErr);
      return pgpErr;
    }
    
    pgpErr= PGPNewSingletonKeySet( aKey, aKeySetRefPtr );	
      
    if (IsPGPError(pgpErr)){
      if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":1.2", pgpErr);
      return pgpErr;  
    }

  } else {
    aKeySetRefPtr= theRcptsKeySetRefPtr;
  }
 
  pgpErr=PGPNewOptionList(thePgpContext, &aOptionList);
  if (pgpErr){
    if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":2", pgpErr);
    if ( !theRcptsKeySetRefPtr ) 
	PGPFreeKeySet( *aKeySetRefPtr );
    return pgpErr;
  }
  pgpErr=PGPAppendOptionList(thePgpContext, aOptionList,
			     // PGPOAskUserForEntropy(thePgpContext, 1 ),
			     PGPOInputBuffer(thePgpContext,theText, theTextSize),
			     PGPOArmorOutput(thePgpContext,theIsArmored),
			     PGPOCompression( thePgpContext, 1 ),
			     PGPOEncryptToKeySet( thePgpContext, *aKeySetRefPtr ),
			     PGPOLastOption(thePgpContext)
			     );
  if (pgpErr){
    if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":3", pgpErr);
    PGPFreeOptionList(aOptionList);
    *theCipherTextSizePtr=0;
    if ( !theRcptsKeySetRefPtr ) 
      PGPFreeKeySet( *aKeySetRefPtr );
    return pgpErr;
  }
  aEncBufferIsProvided=0;
  if (*theCipherTextRef!=NULL) aEncBufferIsProvided=1;
  
  if(!aEncBufferIsProvided){
    pgpErr=PGPAppendOptionList(thePgpContext, aOptionList,
			       PGPOAllocatedOutputBuffer(thePgpContext,
							 (void**) theCipherTextRef,
							 (PGPSize) (1024*1024),
							 theCipherTextSizePtr ),
			       PGPOLastOption(thePgpContext));

    if (pgpErr){
      if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":4", pgpErr);
      PGPFreeOptionList(aOptionList);
     *theCipherTextRef=NULL; 
     //just make sure; assumes that no space has been allocate in case of error.
     *theCipherTextSizePtr=0;
     if ( !theRcptsKeySetRefPtr ) 
       PGPFreeKeySet( *aKeySetRefPtr );
     return pgpErr;
   }
  }else {
    pgpErr=PGPAppendOptionList( thePgpContext, aOptionList,
				PGPOOutputBuffer(thePgpContext,
						 *((void**) theCipherTextRef),
						 *theCipherTextSizePtr, theCipherTextSizePtr),
				PGPOLastOption(thePgpContext) );
    if (pgpErr){
      if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":5", pgpErr);
      PGPFreeOptionList(aOptionList);
      *theCipherTextSizePtr=0;
      if ( !theRcptsKeySetRefPtr ) 
	PGPFreeKeySet( *aKeySetRefPtr );
      return pgpErr;
    }
  }

  pgpErr=PGPEncode( thePgpContext, aOptionList, PGPOLastOption(thePgpContext) );
  if (pgpErr){
    if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":6", pgpErr);
    if(!aEncBufferIsProvided) *theCipherTextRef=NULL;
    *theCipherTextSizePtr=0;
  }
  PGPFreeOptionList(aOptionList);
  if ( !theRcptsKeySetRefPtr ) 
    PGPFreeKeySet( *aKeySetRefPtr );

  return pgpErr;  
  
} /* EnginePgpOEncyptToUserIDs */


/*
* EnginePgpODecryptWithPvtKeyEvtHandler()
* The PGPDecode function has to have an EventHandler function.
* The handler functions are essential. They should all error.
* Otherwise the PGPDecode may not return an error though there is one.
*/
static PGPError
EnginePgpODecryptWithPvtKeyEvtHandler(PGPContextRef thePgpContext,
				      PGPEvent *theEvent,
				      PGPUserValue theUserValue )
{
#undef _func_name
#define _func_name "EnginePgpOVerifySignatureWithPubKeyEvtHandler"
  
  PGPError pgpErr=kPGPError_UnknownError;
  //if (gPgpDebugFlag)
  // EnginePgpPrintEvent(gPgpReportFile, theEvent);
  
  switch ( theEvent->type){
  case kPGPEvent_InitialEvent: 
    *((PGPError*)theUserValue)= kPGPError_UnknownError;
    return kPGPError_NoErr;
    break;
  case kPGPEvent_BeginLexEvent:
    return kPGPError_NoErr;
    break;
  case kPGPEvent_AnalyzeEvent:
    {
      PGPEventAnalyzeData anlzData=theEvent->data.analyzeData;
      switch(anlzData.sectionType){
      case kPGPAnalyze_Encrypted:
	pgpErr=kPGPError_NoErr;
	break;
      case kPGPAnalyze_DetachedSignature:
      case kPGPAnalyze_Key:
      case kPGPAnalyze_Signed: 
      case kPGPAnalyze_Unknown:
      default:
	pgpErr=kPGPError_UnknownError;
	break;
      }
    }
    break;
  case kPGPEvent_RecipientsEvent:
    return kPGPError_NoErr;
    break;
  case kPGPEvent_EndLexEvent:
    return kPGPError_NoErr;
    break;
    //  case kPGPEvent_PassphraseEvent:
    //pgpErr= kPGPError_MissingPassphrase;
    //break;
  case kPGPEvent_FinalEvent:
    pgpErr=*((PGPError*)theUserValue);
    return pgpErr;
    //The error returned after a FinalEvent is ignored by PGPDecode.
    break;
  default:
    return kPGPError_UnknownError;
    break;
  } 

  *((PGPError*)theUserValue)= pgpErr;
  return pgpErr;

} /* EnginePgpODecryptWithPvtKeyEvtHandler */


/*
* EnginePgpODecryptWithPvtKey()
* It decrypts theCipherText. Its size should be theCipherTextSize. If 
* theRcptsKeySetRefPtr is NULL (the key is not supplied), then the userid should be
* supplied in theRecipient and the key should be contained in theKeySet. theIsArmored
* is a hint that the data is armored, currently not in effect. theCipherText contains
* the cipher text and the theCipherTextSize its size. The plain text is returned in
* *theTextRef and its size in *theTextSizePtr. The plain text buffer can be provided,
* in which case *theTextSizePtr should contain the size or it can be allocated by the
* function; in this case it should be freed by PGPFreeData().
* PGPError
*/
PGPError
EnginePgpODecryptWithPvtKey( PGPContextRef thePgpContext,
			     PGPKeySetRef theKeySet, 
			     char** theTextRef, PGPSize *theTextSizePtr,
			     char* theCipherText,
			     PGPSize theCipherTextSize,
			     char* theRecipient,  char *thePassPhrase,
			     PGPKeySetRef *theRcptsKeySetRefPtr,
			     PGPBoolean theIsArmored
			     )
{
  
#undef _func_name
#define _func_name "EnginePgpODecryptWithPvtKey"
  
  PGPError pgpErr= -1; //kPGPError_BadSignature;
    
  PGPKeySetRef *aKeySetRefPtr= NULL, aSingletonKeySet;
  PGPKeyRef aKey;
  PGPOptionListRef aOptionList;
  PGPBoolean aDecBufferIsProvided;
  PGPError aHandlerErr;

  if ( !theRcptsKeySetRefPtr ) {

    if ( !theRecipient ) {
      if (IsPGPError(pgpErr)){
	if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":1", pgpErr);
	return pgpErr;
      }
    }
    aKeySetRefPtr= &aSingletonKeySet;
    pgpErr= EnginePgpSearchForUserKey( thePgpContext, theKeySet,
					 theRecipient, &aKey );
    if (IsPGPError(pgpErr)){
      if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":1.1", pgpErr);
      return pgpErr;
    }

    pgpErr= PGPNewSingletonKeySet( aKey, aKeySetRefPtr );	
      
    if (IsPGPError(pgpErr)){
      if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":1.2", pgpErr);
      return pgpErr;  
    }

  } else {
    aKeySetRefPtr= theRcptsKeySetRefPtr;
  }
 
  pgpErr=PGPNewOptionList(thePgpContext, &aOptionList);
  if (pgpErr){
    if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":2", pgpErr);
    if ( !theRcptsKeySetRefPtr ) 
       PGPFreeKeySet( *aKeySetRefPtr );
    return pgpErr;
  }
  pgpErr=PGPAppendOptionList(thePgpContext, aOptionList,
			     PGPOInputBuffer(thePgpContext, theCipherText,
					     theCipherTextSize ),
			     PGPOPassphrase(thePgpContext, thePassPhrase),
			     PGPOEventHandler (thePgpContext,	      
					       EnginePgpODecryptWithPvtKeyEvtHandler,
					       &aHandlerErr),
			     PGPOKeySetRef( thePgpContext, *aKeySetRefPtr ),
			     PGPOLastOption(thePgpContext)
			     );
  if (pgpErr){
    if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":3", pgpErr);
    PGPFreeOptionList(aOptionList);
    *theTextSizePtr=0;
    if ( !theRcptsKeySetRefPtr ) 
      PGPFreeKeySet( *aKeySetRefPtr );
    return pgpErr;
  }
  aDecBufferIsProvided=0;
  if (*theTextRef != NULL) aDecBufferIsProvided=1;
  
  if(!aDecBufferIsProvided){
    pgpErr=PGPAppendOptionList(thePgpContext, aOptionList,
			       PGPOAllocatedOutputBuffer(thePgpContext,
							 (void**) theTextRef,
							 (PGPSize) (1024*1024),
							 theTextSizePtr ),
			       PGPOLastOption(thePgpContext));

    if (pgpErr){
      if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":4", pgpErr);
      PGPFreeOptionList(aOptionList);
      *theTextRef=NULL; 
     //just make sure; assumes that no space has been allocate in case of error.
      *theTextSizePtr=0;
      if ( !theRcptsKeySetRefPtr ) 
	PGPFreeKeySet( *aKeySetRefPtr );
     return pgpErr;
   }
  }else {
    pgpErr=PGPAppendOptionList( thePgpContext, aOptionList,
				PGPOOutputBuffer(thePgpContext,
						 *((void**) theTextRef),
						 *theTextSizePtr, theTextSizePtr),
				PGPOLastOption(thePgpContext) );
    if (pgpErr){
      if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":5", pgpErr);
      PGPFreeOptionList(aOptionList);
      *theTextSizePtr=0;
      if ( !theRcptsKeySetRefPtr ) 
	PGPFreeKeySet( *aKeySetRefPtr );
      return pgpErr;
    }
  }

  pgpErr=PGPDecode( thePgpContext, aOptionList, PGPOLastOption(thePgpContext) );
  if (aHandlerErr) pgpErr=aHandlerErr;
  if (pgpErr){
    if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":6", pgpErr);
    if(!aDecBufferIsProvided) *theTextRef=NULL;
    *theTextSizePtr=0;
  }
  PGPFreeOptionList(aOptionList);
  if ( !theRcptsKeySetRefPtr ) 
    PGPFreeKeySet( *aKeySetRefPtr );

  return pgpErr;  
  
} /* EnginePgpODecryptWithPvtKey */


/* 
* EnginePgpPrintEvent()
* It prints to theRptFile information about the event.
*/
PGPError EnginePgpPrintEvent(FILE *theRptFile, PGPEvent *theEvent)
{

#undef _func_name
#define _func_name "EnginePgpPrintEvent" 

  PGPError pgpErr=kPGPError_NoErr;
  char *formatStr="";

  switch (theEvent->type){
    case  kPGPEvent_NullEvent: 
      formatStr=_func_name":Event:kPGPEvent_NullEvent:%u\n";
      fprintf(theRptFile,formatStr,theEvent->type);
      break;
    case  kPGPEvent_InitialEvent:
      formatStr=_func_name" kPGPEvent_InitialEvent:%u\n";
      fprintf(theRptFile,formatStr,theEvent->type);
      break;
    case  kPGPEvent_FinalEvent:
      formatStr=_func_name" kPGPEvent_FinalEvent:%u\n";
      fprintf(theRptFile,formatStr,theEvent->type);
      break;
    case  kPGPEvent_ErrorEvent:
      formatStr=_func_name" kPGPEvent_ErrorEvent:%u\n";
      fprintf(theRptFile,formatStr,theEvent->type);
      break;
    case  kPGPEvent_WarningEvent:
      formatStr=_func_name" kPGPEvent_WarningEvent:%u\n";
      fprintf(theRptFile,formatStr,theEvent->type);
      break;
    case  kPGPEvent_EntropyEvent:
      formatStr=_func_name" kPGPEvent_EntropyEvent:%u\n";
      fprintf(theRptFile,formatStr,theEvent->type);
      break;
    case  kPGPEvent_PassphraseEvent:
      formatStr=_func_name" kPGPEvent_PassphraseEvent:%u\n";
      fprintf(theRptFile,formatStr,theEvent->type);
      {
	PGPEventPassphraseData passphrData= theEvent->data.passphraseData;
	fprintf(theRptFile, "\tfConventional:%i\n", passphrData.fConventional );
	fprintf(theRptFile, "\tkeyset:%p\n", passphrData.keyset );
      }
      break;
    case  kPGPEvent_InsertKeyEvent:
      formatStr=_func_name" kPGPEvent_InsertKeyEvent:%u\n";
      fprintf(theRptFile,formatStr,theEvent->type);
      break;
    case  kPGPEvent_AnalyzeEvent:
      formatStr=_func_name" kPGPEvent_AnalyzeEvent:%u\n";
      fprintf(theRptFile,formatStr,theEvent->type);
       {
	PGPEventAnalyzeData anlzData=theEvent->data.analyzeData;
	  switch(anlzData.sectionType){
	  case kPGPAnalyze_DetachedSignature :
	    fprintf
	      (theRptFile,"\tkPGPAnalyze_DetachedSignature:%u\n",anlzData.sectionType);
	    break;
	  case kPGPAnalyze_Encrypted :
	    fprintf
	      (theRptFile,"\tkPGPAnalyze_Encrypted:%u\n",anlzData.sectionType);
	    break;
	  case kPGPAnalyze_Key :
	    fprintf
	      (theRptFile,"\tkPGPAnalyze_Key:%u\n",anlzData.sectionType);
	    break;
	  case kPGPAnalyze_Signed :
	    fprintf
	      (theRptFile,"\tkPGPAnalyze_Signed:%u\n",anlzData.sectionType);
	    break;
	  case kPGPAnalyze_Unknown :
	      fprintf
		(theRptFile,"\tkPGPAnalyze_Unknown:%u\n",anlzData.sectionType);
	    break;
	  default:
	    break;
	  }
      }
      break;
    case  kPGPEvent_RecipientsEvent:
      formatStr=_func_name" kPGPEvent_RecipientsEvent:%u\n";
      fprintf(theRptFile,formatStr,theEvent->type);
      {
	PGPEventRecipientsData rcptData= theEvent->data.recipientsData;
	fprintf(theRptFile, "\trecipientSet:%p\n", rcptData.recipientSet );
	fprintf(theRptFile, "\tconventionalPassphraseCount:%u\n", 
		rcptData.conventionalPassphraseCount );
	fprintf(theRptFile, "\tkeyCount:%u\n", rcptData.keyCount );
	fprintf(theRptFile, "\tkeyIDArray:%p\n", rcptData.keyIDArray );
      }
      break;
    case  kPGPEvent_KeyFoundEvent:
      formatStr=_func_name" kPGPEvent_KeyFoundEvent:%u\n";
      fprintf(theRptFile,formatStr,theEvent->type);
      break;
    case  kPGPEvent_OutputEvent :
      formatStr=_func_name" kPGPEvent_OutputEvent:%u\n";
      fprintf(theRptFile,formatStr,theEvent->type);
      break;
    case  kPGPEvent_SignatureEvent:
      formatStr=_func_name" kPGPEvent_SignatureEvent:%u\n";
      fprintf(theRptFile,formatStr,theEvent->type);
      {
	PGPEventSignatureData sgnData=theEvent->data.signatureData;
	//fprintf(theRptFile,"\tsigningKeyID:%0X\n",sgnData.signingKeyID);
	fprintf(theRptFile,"\tsigningKey:%p\n",sgnData.signingKey);
	fprintf(theRptFile,"\tchecked:%u\n",sgnData.checked);
	fprintf(theRptFile,"\tverified:%u\n",sgnData.verified); 
	fprintf(theRptFile,"\tkeyRevoked:%u\n",sgnData.keyRevoked);
	fprintf(theRptFile,"\tkeyDisabled:%u\n",sgnData.keyDisabled);
	fprintf(theRptFile,"\tkeyExpired:%u\n",sgnData.keyExpired);
	fprintf(theRptFile,"\tkeyMeetsValidityThreshold:%u\n",
		sgnData.keyMeetsValidityThreshold);
	fprintf(theRptFile,"\tkeyValidity:%u\n",sgnData.keyValidity);
	fprintf(theRptFile,"\tcreationTime:%u\n",sgnData.creationTime);
	/*  }else { */
	/*  fprintf(theRptFile,"\t:NO EVENT DATA FOUND\n"); */
      }
      break;
    case  kPGPEvent_BeginLexEvent:
      formatStr=_func_name" kPGPEvent_BeginLexEvent:%u\n";
      fprintf(theRptFile,formatStr,theEvent->type);
      {
	PGPEventBeginLexData lexData=theEvent->data.beginLexData;
	fprintf(theRptFile,"\tsectionNumber:%u\n",lexData.sectionNumber);
	fprintf(theRptFile,"\tsectionOffset:%u\n",lexData.sectionOffset);
      }
      break;
    case  kPGPEvent_EndLexEvent:
      formatStr=_func_name" kPGPEvent_EndLexEvent:%u\n";
      fprintf(theRptFile,formatStr,theEvent->type);
      {
	PGPEventEndLexData lexData=theEvent->data.endLexData;
	fprintf(theRptFile,"\tsectionNumber:%u\n",lexData.sectionNumber);
      }
      break;
    case  kPGPEvent_RecursionEvent:
      formatStr=_func_name" kPGPEvent_RecursionEvent:%u\n";
      break;
    case  kPGPEvent_DetachedSignatureEvent:
      formatStr=_func_name" kPGPEvent_DetachedSignatureEvent:%u\n";
      fprintf(theRptFile,formatStr,theEvent->type);
      break;
    case  kPGPEvent_KeyGenEvent:
      formatStr=_func_name" kPGPEvent_KeyGenEvent:%u\n";
      fprintf(theRptFile,formatStr,theEvent->type);
      break;
    case  kPGPEvent_KeyServerEvent:
      formatStr=_func_name" kPGPEvent_KeyServerEvent:%u\n";
      fprintf(theRptFile,formatStr,theEvent->type);
      break;
    case  kPGPEvent_KeyServerSignEvent:
      formatStr=_func_name" kPGPEvent_KeyServerSignEvent:%u\n";
      fprintf(theRptFile,formatStr,theEvent->type);
      break;
    default:
      formatStr= _func_name" NOT CODED  EVENT OF TYPE:%u\n";
      fprintf(theRptFile,formatStr,theEvent->type);
    }
fflush(theRptFile);
return pgpErr;
}/*  EnginePgpPrintEvent */


/*
* EnginePgpPrepareOneLineArmoredSignature() 
* Given a one-line armored pgp signature (without a any pgp headers)
* it formats it so that it be valid for verification by the PGPsdk.
*/
char* EnginePgpPrepareOneLineArmoredSignature(PGPContextRef thePgpContext,
						char* theInputSgn, 
						PGPSize theInputSgnSize)
{
#undef _func_name
#define _func_name "EnginePgpTranslateOneLineArmoredSignature"

  char *startSgnPtr, *endSgnPtr, *crcPtr;
  const PGPSize akcrcLen=4, akLnLen=64;
  PGPSize aNeededLength, aWrittenChars;
  char* aTranslatedSgn;
  
  const char* aSgnBeginHeader="-----BEGIN PGP SIGNATURE-----";
  const char* aSgnEndHeader="-----END PGP SIGNATURE-----";
  PGPSize aBeginHdrLen=strlen(aSgnBeginHeader);
  PGPSize aEndHdrLen=strlen(aSgnEndHeader);
 
  if ( theInputSgn==NULL || theInputSgnSize==0  ) {
     if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":1",
				 kPGPError_BadSignature);
    return NULL;
  }

  startSgnPtr=theInputSgn;
  endSgnPtr=theInputSgn+theInputSgnSize-1;
  
  while (isspace((int)*startSgnPtr) && startSgnPtr!=endSgnPtr ){ 
    startSgnPtr++;
  }
  while(isspace((int)*endSgnPtr) && startSgnPtr!=endSgnPtr){
    endSgnPtr--;
  }

  if (startSgnPtr==endSgnPtr){
    if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":2", -1);
   return NULL;
  }
  crcPtr=endSgnPtr-akcrcLen;
  
  if ( *crcPtr != '=') {
    if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":3", -1);
    return NULL;
  }

  endSgnPtr= crcPtr - 1;

  theInputSgnSize= endSgnPtr - startSgnPtr + 1; 
  if ( theInputSgnSize <= akcrcLen + 1 ) {
   if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":4", -1);
    return NULL;
  }

  aNeededLength=( ( theInputSgnSize - ( akcrcLen + 1 ) ) / akLnLen +1)*2+  7 +
    (theInputSgnSize) + aBeginHdrLen +aEndHdrLen ;
  
  
  /*  printf("NEEDED:%u\n",aNeededLength);fflush(stdout); */
  aTranslatedSgn=PGPNewData( thePgpContext, aNeededLength * sizeof(char) );
 
  if (aTranslatedSgn){
    
    memcpy(aTranslatedSgn,aSgnBeginHeader,aBeginHdrLen);
    aWrittenChars=aBeginHdrLen;
    memcpy(aTranslatedSgn+aWrittenChars,"\n",1);
    aWrittenChars++;
    
    while( endSgnPtr>startSgnPtr && endSgnPtr - startSgnPtr > akLnLen ){
      memcpy(aTranslatedSgn+aWrittenChars,startSgnPtr,akLnLen);
      aWrittenChars+=akLnLen;
      startSgnPtr+=akLnLen;
      memcpy(aTranslatedSgn+aWrittenChars,"\n",1);
      aWrittenChars++;
    }
    
    aNeededLength=endSgnPtr-startSgnPtr+1;
    memcpy(aTranslatedSgn+aWrittenChars,startSgnPtr,aNeededLength);
    aWrittenChars+=aNeededLength;
    memcpy(aTranslatedSgn+aWrittenChars,"\n",1);
    aWrittenChars++;

    memcpy(aTranslatedSgn+aWrittenChars,crcPtr,akcrcLen + 1);
    aWrittenChars+=akcrcLen + 1;
    memcpy(aTranslatedSgn+aWrittenChars,"\n",1);
    aWrittenChars+=1;
    memcpy(aTranslatedSgn+aWrittenChars,aSgnEndHeader,aEndHdrLen);
    aWrittenChars+=aEndHdrLen;
    memcpy(aTranslatedSgn+aWrittenChars,"\n\0",2);
    aWrittenChars+=2;
    //printf("WRITTEN:%u\n",aWrittenChars);fflush(stdout);
    return aTranslatedSgn;
  }
 if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":5", -1); 
  return NULL;
}/*  EnginePgpPrepareOneLineArmoredSignature */


/*
* EnginePgpSearchForUserKey()
* Search in theKeySet (which should belong to thePgpContext) for a key that matches the usrID
* (name <email>).
* Currently only the name part of the usrID is used, and if multiple* same-name keys exist in the database, there is no guarantee which one is returned in *theFoundKey.
*/
PGPError EnginePgpSearchForUserKey(PGPContextRef thePgpContext,
				   PGPKeySetRef theKeySet,
				   char*  usrID, PGPKeyRef *theFoundKey
				   )
{
#undef _func_name
#define _func_name "EnginePgpSearchForUserKey"  
  
  char *name, *email;
  PGPError pgpErr= kPGPError_ItemNotFound;
  PGPUInt32 numOfFndKeys;
  PGPFilterRef theEmailFilter=NULL, theNameFilter=NULL, thePgpFilter=NULL;
  PGPKeySetRef theFoundKeySet;
  PGPKeyListRef theKeyList;
  PGPKeyIterRef theKeyIter;
  
  *theFoundKey=NULL;
  
  EnginePgpExtractNameAndEmail(thePgpContext,usrID, &name, &email);
  if(gPgpDebugFlag){
    fprintf(gPgpReportFile,"name:%s:email:%s\n", name?name:"",email?email:"");
    fflush(gPgpReportFile);
  }
  
  if (email){
    pgpErr=PGPNewUserIDEmailFilter(thePgpContext, email, 
				   kPGPMatchEqual,&theEmailFilter);
    if (IsPGPError(pgpErr) && gPgpDebugFlag) {
      Pgp_DEBUG(gPgpReportFile,_func_name":1", pgpErr);
    }
  }
  
  if (name){
    pgpErr=PGPNewUserIDNameFilter(thePgpContext, name,
				 kPGPMatchEqual, &theNameFilter);
    if (IsPGPError(pgpErr) && gPgpDebugFlag) {
      Pgp_DEBUG(gPgpReportFile,_func_name":2", pgpErr);
    }
  }

  if ( theNameFilter && theEmailFilter ) {
    pgpErr=PGPIntersectFilters(theEmailFilter, theNameFilter, &thePgpFilter);
    theEmailFilter=NULL; theNameFilter=NULL;
    if (IsPGPError(pgpErr) && gPgpDebugFlag) {
      Pgp_DEBUG(gPgpReportFile,_func_name":3", pgpErr);
    }
  }
  else
    if( theNameFilter ){
      thePgpFilter= theNameFilter; theNameFilter=NULL;
    } else {
      thePgpFilter= theEmailFilter; theEmailFilter=NULL;
    }
  if (name) PGPFreeData(name);
  if (email) PGPFreeData(email);
  if (!thePgpFilter) return pgpErr;

  pgpErr=PGPFilterKeySet(theKeySet,thePgpFilter, &theFoundKeySet);
  if (IsPGPError(pgpErr)) {
      if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":4", pgpErr);
      if (thePgpFilter) {PGPFreeFilter(thePgpFilter);thePgpFilter=NULL;}
      if (theNameFilter){PGPFreeFilter(theNameFilter);theNameFilter=NULL;}
      if (theEmailFilter){PGPFreeFilter(theEmailFilter);theEmailFilter=NULL;}
      return pgpErr;
  }

  pgpErr=PGPCountKeys(theFoundKeySet, &numOfFndKeys);
  if (IsPGPError(pgpErr)) {
    if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":4.1", pgpErr);
  } else {
    if(gPgpDebugFlag) {
      fprintf(gPgpReportFile,_func_name":4.2:Keys Found:%d\n", numOfFndKeys);
      fflush(stdout);
    }
  }
  if ( !numOfFndKeys ){
    if(gPgpDebugFlag) {
      fprintf(gPgpReportFile,
	      "enginepgp.c:"_func_name":4.2:No User Key\n");fflush(stdout);
    }
    if (thePgpFilter) {PGPFreeFilter(thePgpFilter);thePgpFilter=NULL;}
    if (theNameFilter){PGPFreeFilter(theNameFilter);theNameFilter=NULL;}
    if (theEmailFilter){PGPFreeFilter(theEmailFilter);theEmailFilter=NULL;}
    if (theFoundKeySet){PGPFreeKeySet(theFoundKeySet);theFoundKeySet=NULL;}
    return  kPGPError_ItemNotFound;
  }
  pgpErr=PGPOrderKeySet(theFoundKeySet,kPGPAnyOrdering,&theKeyList);
  if (IsPGPError(pgpErr)) {
    if (gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":5", pgpErr);
    if (thePgpFilter) {PGPFreeFilter(thePgpFilter);thePgpFilter=NULL;}
    if (theNameFilter){PGPFreeFilter(theNameFilter);theNameFilter=NULL;}
    if (theEmailFilter){PGPFreeFilter(theEmailFilter);theEmailFilter=NULL;}
    if (theFoundKeySet){PGPFreeKeySet(theFoundKeySet);theFoundKeySet=NULL;}
    return pgpErr;
  }
  pgpErr=PGPNewKeyIter(theKeyList,&theKeyIter);
  if (IsPGPError(pgpErr)) {
      if (gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":6", pgpErr);
      if (thePgpFilter) {PGPFreeFilter(thePgpFilter);thePgpFilter=NULL;}
      if (theNameFilter){PGPFreeFilter(theNameFilter);theNameFilter=NULL;}
      if (theEmailFilter){PGPFreeFilter(theEmailFilter);theEmailFilter=NULL;}
      if (theFoundKeySet){PGPFreeKeySet(theFoundKeySet);theFoundKeySet=NULL;}
      if(theKeyList){PGPFreeKeyList(theKeyList);theKeyList=NULL;}
      return pgpErr;
  }
  pgpErr=PGPKeyIterNext(theKeyIter, theFoundKey);
  if (IsPGPError(pgpErr)) {
      if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":7", pgpErr);
  }
  if (thePgpFilter) {PGPFreeFilter(thePgpFilter);thePgpFilter=NULL;}
  if (theNameFilter){PGPFreeFilter(theNameFilter);theNameFilter=NULL;}
  if (theEmailFilter){PGPFreeFilter(theEmailFilter);theEmailFilter=NULL;}
  if (theFoundKeySet){PGPFreeKeySet(theFoundKeySet);theFoundKeySet=NULL;}
  if (theKeyList){PGPFreeKeyList(theKeyList);theKeyList=NULL;}
  if (theKeyIter){PGPFreeKeyIter(theKeyIter);theKeyIter=NULL;}
  return pgpErr;

}/*  EnginePgpSearchUserKey */


/*
* EnginePgpPrintKeyProps()
* This function was meant to print the key properties 
*/
static void  EnginePgpPrintKeyProps(FILE *theFile, PGPKeyRef theKey)
{
#undef _func_name
#define _func_name "EnginePgpPrintKeyProps"

  void printKeyProps(FILE* theFile, PGPKeyRef theKey, int propName, char isType ) {

  PGPError pgpErr=kPGPError_NoErr;
  PGPUInt32 propData; 
  PGPBoolean propBool; 
  PGPTime propTime; 
  switch ( isType ){
  case 's':
   break;
  case 'n':pgpErr=PGPGetKeyNumber(theKey, propName, &propData);
    break;
  case 'b':pgpErr=PGPGetKeyBoolean(theKey,propName, &propBool);
    break;
 case 't':pgpErr=PGPGetKeyTime(theKey,propName, &propTime);
   break;
  }  
  fprintf(theFile, "propName:"); 
  if (IsPGPError(pgpErr)){
    fprintf(theFile, "\n");
    if(gPgpDebugFlag)
      Pgp_DEBUG(theFile, _func_name, pgpErr); 
  } else {
    switch ( isType ){
   case 's':
     break;
    case 'n':fprintf(theFile, "%u\n",propData);
      break;
    case 'b':fprintf(theFile, "%c\n",propBool);
      break;
   case 't':fprintf(theFile, "%u\n",propTime);
     break;
   }
  }
  fprintf(theFile,"\n");
  fflush(theFile);
  }

  fprintf(theFile,"========= BEGIN KEY PROPERTIES =========\n");
  printKeyProps( theFile, theKey, kPGPKeyPropFingerprint, 's');
  printKeyProps( theFile, theKey, kPGPKeyPropPreferredAlgorithms, 's');
  printKeyProps( theFile, theKey, kPGPKeyPropAlgID, 'n');
  printKeyProps( theFile, theKey, kPGPKeyPropBits, 'n');
  printKeyProps( theFile, theKey, kPGPKeyPropTrust, 'n');
  printKeyProps( theFile, theKey, kPGPKeyPropValidity, 'n');
  printKeyProps( theFile, theKey, kPGPKeyPropCreation, 't');
  printKeyProps( theFile, theKey, kPGPKeyPropExpiration, 't');
  printKeyProps( theFile, theKey, kPGPKeyPropIsSecret, 'b');
  printKeyProps( theFile, theKey, kPGPKeyPropIsAxiomatic, 'b');
  printKeyProps( theFile, theKey, kPGPKeyPropIsRevoked, 'b');
  printKeyProps( theFile, theKey, kPGPKeyPropIsDisabled, 'b');
  printKeyProps( theFile, theKey, kPGPKeyPropIsNotCorrupt, 'b');
  printKeyProps( theFile, theKey, kPGPKeyPropIsExpired, 'b');
  printKeyProps( theFile, theKey, kPGPKeyPropNeedsPassphrase, 'b');
  printKeyProps( theFile, theKey, kPGPKeyPropCanSign, 'b');
  printKeyProps( theFile, theKey, kPGPKeyPropCanEncrypt, 'b');
  printKeyProps( theFile, theKey, kPGPKeyPropHasUnverifiedRevocation, 'b');
  fprintf(theFile,"========= END  KEY PROPERTIES =========\n");

}/*  EnginePgpPrintKeyProps */


/* 
* EnginePgpReturnHashAlgID()
* This is a utility function that given a  algorithm name,
* it returns a PGPHashAlgorithm which can be passed as an 
* argument to other functions. The supported algorithm names are:
* "MD5". "SHA", "SHADouble" and "RIPEMD160".
* Beaware that if theAlgName is not supported the function
* returns: kPGPHashAlgorithm_Invalid
*/
PGPHashAlgorithm EnginePgpReturnHashAlgID (char *theAlgName) 
{
#undef _func_name
#define _func_name "EnginePgpReturnAlgID"
  if (strcasecmp(theAlgName,"MD5")==0) return kPGPHashAlgorithm_MD5;
  if (strcasecmp(theAlgName,"SHA")==0) return kPGPHashAlgorithm_SHA;
  if (strcasecmp(theAlgName,"RIPEMD160")==0) return kPGPHashAlgorithm_RIPEMD160;
  if (strcasecmp(theAlgName,"SHADouble")==0) return kPGPHashAlgorithm_SHADouble;
  return kPGPHashAlgorithm_Invalid;
}/* EnginePgpReturnAlgID */


/*
* EnginePgpExtractNameAndEmail()
* The function splits *theSigner into name and email part.
* *theSigner is of the form "name <email>" or "name email"
* the name MUST always precede the email, if they exist both
* name can contain spaces and either of them can be missing in which
* case the respective passed-by reference argument will be set to NULL.
* It is an error if one of "<" and ">" is missing, or they appear in opposite
* order.
* if both "<" and ">" are missing *theSigner is searched for the "@"
* character. Then the email is defined from the character which is immediately 
* right of the first space found towards left, until the character 
* which is immediately left to the first space found towards right.
* Handles NULL or zero-length arguments.
*/
void EnginePgpExtractNameAndEmail(PGPContextRef thePgpContext,
			    char *theSigner, char** theName, char** theEmail)
{

  size_t len;
  char  *esPtr, *eePtr, *ePtr, *nsPtr=NULL, *nePtr=NULL, *endPtr;
  /*
   * esPtr: email start Ptr, eePtr: email end Ptr
   * nsPtr: name start Ptr, nsPtr: name end Ptr
   * ePtr: pointer to the "@".
   * endPtr: pointer to the last letter of *theSigner
   * the name MUST always precede the email in *theSigner
   */

  *theName=NULL; *theEmail=NULL;
  if (!theSigner) return;
  len=strlen(theSigner);
  if (!len) return;
  endPtr= theSigner + len - 1;

  esPtr=strstr(theSigner,"<");
  eePtr=strstr(theSigner,">");
  ePtr=strstr(theSigner,"@");
 
  /* printf("esPtr:%p:eePtr:%p:\n",lPtr,eePtr);fflush(stdout); */

  if( (!esPtr && eePtr) || ( esPtr && !eePtr ) ) return;
  if( esPtr > eePtr ) return;
  
  if ( esPtr && eePtr ) {
    if ( eePtr > esPtr + 1  ) { //email should have at least on letter.
      if ( esPtr > theSigner ){
	nePtr=esPtr - 1;
	nsPtr=theSigner;
      } else {
	nePtr=nsPtr= NULL; //there is no name
      }
      esPtr++;
      eePtr--;
    }
  } else {
    if (ePtr) {
      esPtr= ePtr;
      while ( esPtr > theSigner ){
	esPtr--;
	if ( isspace((int) *esPtr ) )  break;
      }
      if ( isspace((int) *esPtr ) ) esPtr++;

      eePtr= ePtr;
      while ( eePtr < endPtr ) {
	eePtr++;
	if ( isspace((int) *eePtr ) )  break;
      }
      if ( isspace((int) *eePtr ) )  eePtr--;
      
      if ( esPtr > theSigner ){
	nsPtr=theSigner;
	nePtr=esPtr - 1;
      } else {
	nePtr=nsPtr= NULL; //there is no name
      }
    } else {
      esPtr=eePtr=NULL; //there is no email;
      nsPtr= theSigner;
      nePtr= endPtr;
    }
  }

  /*
   * The email has no leading or trailing spaces.
   * Let us trim the name.
   */

  while (nsPtr != nePtr) {
    if( !isspace((int) *nsPtr ) ) break;
    nsPtr++;
  }
  while (nsPtr != nePtr) {
    if( !isspace((int) *nePtr ) ) break;
    nePtr--;
  }
  
  if (nsPtr == nePtr && nsPtr &&  isspace((int) *nsPtr ) ) nsPtr=nePtr=NULL;
  
  //Copy whatever is left.
 
  if (esPtr && eePtr) len= eePtr - esPtr +1; else len=0;
  if (len>0){
    *theEmail=PGPNewData(thePgpContext,(len+1)*sizeof(char));
    if (*theEmail !=NULL){
      memcpy(*theEmail,esPtr,len);
      *(*theEmail+len)= (char) 0;
    }
  }

  if (nsPtr && nePtr) len= nePtr - nsPtr + 1; else len=0;
  if (len>0){
    *theName=PGPNewData(thePgpContext,(len+1)*sizeof(char));
    if (*theName !=NULL){
      memcpy(*theName, nsPtr,len);
      *(*theName+len)= (char) 0;
      }
  }
  return;

/*
older version.
The function splits *theSigner into name and email part.
*theSigner is of the form "name <email>".
name can contain spaces and either of them can be missing in which
case the respective passed-by reference argument will be set to NULL.
It is an error if one of "<" and ">" is missing, or they appear in opposite
order.

  size_t len;
  char  *lPtr, *lPtrA, *rPtr, *ePtr;
  *theName=NULL; *theEmail=NULL;
  if (!theSigner) return;
 
  lPtr=strstr(theSigner,"<");
  rPtr=strstr(theSigner,">");
  ePtr=strstr(theSigner,"@");
  // printf("lPtr:%p:rPtr:%p:\n",lPtr,rPtr);fflush(stdout); 

  if( ((!lPtr && rPtr)||(lPtr && !rPtr))) return;
  if( lPtr > rPtr ) return;

  if( lPtr != rPtr) ePtr=NULL; //We have email enclosed in < > pair.

  if (lPtr==rPtr && !ePtr){
    //No "<" and  ">" and "@", thus we have only name.
    len=strlen(theSigner);
    lPtr=theSigner+len;
    rPtr=lPtr;
  } else 
    if ( lPtr==rPtr && ePtr ) {
      //We still have email but not enclosed in <> pair.
      //Adjust lPtr and rPtr to point to the first and last 
      //letter of email respectively.
      lPtr=ePtr;
      while (lPtr > theSigner){
	lPtr--;
	if (isspace((int) *lPtr )) {
	  lPtr++;
	  break;
	}
      }
      rPtr= ePtr;
      while ( *rPtr != (char) 0 ) {
	rPtr++;
	if (isspace((int) *rPtr )) {
	  rPtr--;
	  break;
	}
      }
      if ( *rPtr == (char) 0 ) rPtr--;
      if (lPtr==ePtr || rPtr==ePtr) return;
    }

  while( theSigner < lPtr ) {
    if (isspace((int)*theSigner)) theSigner++;  else  break;
  }
  lPtrA=lPtr;
  while( theSigner < lPtr ) {
    lPtr--;
    if (!isspace((int) *lPtr )) break;
  }
  if (lPtr==lPtrA) len=0; else len=lPtr-theSigner+1;
  if (len>0){
    *theName=PGPNewData(thePgpContext,(len+1)*sizeof(char));
    if (*theName !=NULL){
      memcpy(*theName,theSigner,len);
      *(*theName+len)= (char) 0;
      }
  }
  lPtr=lPtrA;
  lPtrA=rPtr;
  while ( lPtr < rPtr ){
    lPtr++;
    if (!isspace((int) *lPtr )) break;
    }
  while (lPtr<rPtr){
    rPtr--;
    if (!isspace((int) *rPtr )) break;
  }
  if (rPtr==lPtrA) len=0; else len=rPtr-lPtr+1;
  if (len>0){
    *theEmail=PGPNewData(thePgpContext,(len+1)*sizeof(char));
    if (*theEmail !=NULL){
      memcpy(*theEmail,lPtr,len);
      *(*theEmail+len)= (char) 0;
    }
  }
  return;
*/
 
}/*  EnginePgpExtractNameAndEmail */

/*
  { //Test code for EnginePgpExtractNameAndEmail
    char* entry[]= {"<alexiou@hyper.gr>",
		    "Akis Alexiou <alexiou@hyper.gr>",
		    " Triantafyllos Akis Alexiou <alexiou@hyper.gr> ",
		    " Triantafyllos Akis Alexiou <alexiou@hyper.gr> Akis ",
		    "<alexiou@hyper.gr>", 
		    " <alexiou@hyper.gr> ",
		    "Akis Alexiou",
		    "Akis Alexiou@",
		    "alexiou@hyper.gr",
		    "alexiou@hyper.gr ",
		    "@",
		    " @ ",
		    "@alexiou",
		    "Akis @alexiou",
		    "alexiou@hyper.gr Akis Alexiou",
		    " Triantafyllos Akis Alexiou < alexiou@hyper.gr> ",
		    " Triantafyllos Akis Alexiou <alexiou@hyper.gr > Akis ",
		    "   ",
    };
    unsigned int kNumOfEntries=18, k;
    char *name, *email;
    
    for( k=0; k < kNumOfEntries; k++ ) {
      EnginePgpExtractNameAndEmail(gPgpContext,entry[k],&name,&email );
      fprintf(gPgpReportFile,"%3d:%s:\tNAME:%s:\tEMAIL:%s:\n",
	      k, entry[k], name ? name : "NO_NAME", email ? email : "NO_EMAIL");
      fflush(gPgpReportFile);
      if(name) PGPFreeData(name);
      if(email)PGPFreeData(email);
    }
}
*/


/*
* EnginePgpGenerateKey()
* We do not recommend trusting this function for creating high security keys
* due to the its source of random number. We also do not recommend it for
* creating big keys, as it may fail due to lack of adequate entropy
* This function was created in order to generate a key pair for 
* debugging purposes solely 
* The generated key is always returned in *theKey; it is either
* PGPKeyRef or PGPSubKeyRef.
* if theMasterKey == NULL then a master key is generated
* otherwise a subkey, which upon return is also attached to
* theMasterKey.
* Accepted values for thePubKeyAlg are:
* kPGPPublicKeyAlgorithm_ElGamal (a.k.a.Diffie-Hellman)
* kPGPPublicKeyAlgorithm_DSA
* kPGPPublicKeyAlgorithm_RSA
* kPGPPublicKeyAlgorithm_RSAEncryptOnly
* kPGPPublicKeyAlgorithm_RSASignOnly
* theName for subkey generation is neglected
* The changes are committed to the keyrings associated with theKeySet
*/
PGPError 
EnginePgpGenerateKey(PGPContextRef thePgpContext, 
		     PGPKeySetRef theKeySet,
		     void  *theKey,
		     char *theName,PGPSize theNameSize,
		     PGPKeyRef theMasterKey,
		     PGPBoolean theDoFastGen,
		     char *thePassphrase,
		     PGPPublicKeyAlgorithm thePubKeyAlg,
		     PGPUInt32 theKeySize,
		     PGPUInt32 theExpirationDays
		     )
{
#undef _func_name
#define _func_name "EnginePgpGenerateKey"

  PGPSize kIndex;
  PGPError pgpErr=kPGPError_NoErr;
  PGPUInt32 aEntropyNeeded, aEntropyNow;
  
  if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile, _func_name":CREATING ENTROPY:1", 
			      pgpErr);

  kIndex=0;
  while (kIndex!=40000){
    PGPGlobalRandomPoolAddKeystroke((PGPInt32)rand());
    kIndex++;
  }
  if(gPgpDebugFlag) {
    aEntropyNeeded=
      PGPGetKeyEntropyNeeded(thePgpContext,
			     PGPOKeyGenFast(thePgpContext, theDoFastGen ),
			     PGPOKeyGenParams(thePgpContext,
					      thePubKeyAlg, theKeySize),
			     PGPOLastOption( thePgpContext )
			     );
    aEntropyNow=PGPGlobalRandomPoolGetEntropy();
   fprintf(gPgpReportFile,"Ent_Needed:%u:Now:%u\n",aEntropyNeeded,aEntropyNow);
   fflush(gPgpReportFile);
  }

 if (!theMasterKey) {
   pgpErr=PGPGenerateKey( thePgpContext,(PGPKeyRef*) theKey,
			  PGPOKeySetRef(thePgpContext,theKeySet),
			  PGPOKeyGenName(thePgpContext,theName,theNameSize),
			  PGPOKeyGenFast(thePgpContext, theDoFastGen ),
			  PGPOPassphrase(thePgpContext,thePassphrase),
			  PGPOKeyGenParams(thePgpContext,
					    thePubKeyAlg, theKeySize),
			  PGPOExpiration(thePgpContext, theExpirationDays),
			  PGPOLastOption( thePgpContext ));
   if(pgpErr) {
     if (gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile, _func_name":2", pgpErr);
     return pgpErr;
   }
   pgpErr=PGPEnableKey( *((PGPKeyRef*)theKey) );
   if(pgpErr)
     if (gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile, _func_name":3", pgpErr);
   pgpErr=PGPSetKeyAxiomatic( *((PGPKeyRef*)theKey ),1,thePassphrase);
   if(pgpErr)
     if (gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile, _func_name":4", pgpErr);
   pgpErr=PGPSetDefaultPrivateKey(*((PGPKeyRef*)theKey));
   if(pgpErr)
     if (gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile, _func_name":5", pgpErr);	      
   /*
    *pgpErr=PGPSetKeyTrust(*((PGPKeyRef*)theKey),kPGPKeyTrust_Complete);
    *if(pgpErr) 
    * if (gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile, _func_name":6", pgpErr);
    */

 } else {

   pgpErr=PGPGenerateSubKey(thePgpContext,
			    (PGPSubKeyRef*) theKey,
			    PGPOKeyGenMasterKey(thePgpContext,theMasterKey ),
			    PGPOKeyGenFast(thePgpContext, theDoFastGen ),
			    PGPOKeyGenParams(thePgpContext,
					     thePubKeyAlg, theKeySize),
			    PGPOPassphrase(thePgpContext,thePassphrase),
			    PGPOLastOption( thePgpContext ));

   if(pgpErr){
     if (gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile, _func_name":7", pgpErr);
     return pgpErr;
   }
 }

 pgpErr=PGPCommitKeyRingChanges(theKeySet);
 if(pgpErr)
     if (gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile, _func_name":8", pgpErr);
return pgpErr;

}/*  EnginePgpGenerateKey */


/* 
 * ===============  OTHER FUNCTIONS ===============
*/

/*
* EnginePgpSignWithPvtKey()
* This function signs with theText with a private key.
* The returned signature is  **theSignatureRef and the size of it 
* *theSgnSizePtr. The function itself allocates space for the
* signature. In case of error a non zero value is returned and also
* *theSignatureRef=NULL and *theSgnSizePtr=0.
* The  **theSignatureRef is not a C-style string ("\0" terminated).
* If you like to C-style terminate it, you can do it safely at 
* (*theSignatureRef) + (*theSgnSizePtr) byte. You must make no
* assumptions regarding the unused space, after the signature
* in the memory block pointed by *theSignatureRef.
* Bear in mind that C-style termination may be of value only if the
* output were armored. SEE ALSO BELOW
* You can supply a desired signing key in the theSignKey parameter.
* If theSignKey==NULL then it searches in  theKeySet for a key 
* that matches the *theSigner. theHashAlgID specifies the
* desired hash algorithm for signing.
* isArmored is currently NOT SUPPORTED and is provisioned for future 
* versions.
*/
static PGPError 
EnginePgpSignWithPvtKey( PGPContextRef thePgpContext,
			 PGPKeySetRef theKeySet, 
			 char* theText, PGPSize theTextSize,
			 char **theSignatureRef,
			 PGPSize *theSgnSizePtr,
			 char* theSigner, char *thePassPhrase,
			 PGPKeyRef theSignKey,
			 PGPHashAlgorithm theHashAlgID,
			 PGPBoolean isArmored)
{
#undef _func_name
#define _func_name "EnginePgpSignWithPvtKey"

PGPError pgpErr=kPGPError_NoErr;
PGPKeyRef aKey;
PGPPrivateKeyContextRef aPvtKeyContext;
PGPSize maxDecrBufSize, maxEncrBufSize;
PGPHashContextRef aHashContext;

 if (!theSignKey) {
   pgpErr=EnginePgpSearchForUserKey(thePgpContext,theKeySet,
				    theSigner ,&aKey);
   if (IsPGPError(pgpErr)){
     if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":1", pgpErr);
     return pgpErr;
   }
 } else {
   aKey=theSignKey;
 }
 {
   PGPBoolean aPGPBoolean;
   aPGPBoolean=PGPPassphraseIsValid(aKey, thePassPhrase);
   if (!aPGPBoolean){
     if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":1.1:BAD PASSPHRASE",0);
     return pgpErr;
   }
 }
 pgpErr=PGPNewPrivateKeyContext(aKey, kPGPPublicKeyMessageFormat_PGP,
				thePassPhrase,&aPvtKeyContext );
 if (IsPGPError(pgpErr)){
   if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":2", pgpErr);
   return pgpErr;
 }
 pgpErr=PGPGetPrivateKeyOperationSizes(aPvtKeyContext,&maxDecrBufSize,
					&maxEncrBufSize, theSgnSizePtr );
 if (IsPGPError(pgpErr)){
   if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":3", pgpErr);
   return pgpErr;
 }
 if (*theSgnSizePtr == 0){
   if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,
			       _func_name":4:Signing is not availiable", -1);
   PGPFreePrivateKeyContext(aPvtKeyContext);
   return -1;
 }
  *theSignatureRef=NULL;
  *theSignatureRef=PGPNewData(thePgpContext,sizeof(char)*((*theSgnSizePtr)+1));
  if (!(*theSignatureRef)){
    if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,
			       _func_name":5", kPGPError_OutOfMemory);
    PGPFreePrivateKeyContext(aPvtKeyContext);
    *theSgnSizePtr=0;
    return kPGPError_OutOfMemory;
  }

  pgpErr=PGPNewHashContext(thePgpContext, theHashAlgID, &aHashContext);
  if (IsPGPError(pgpErr)) {
      if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":6", pgpErr);
      PGPFreeData(*theSignatureRef);
      *theSignatureRef=NULL;
      *theSgnSizePtr=0;
      PGPFreePrivateKeyContext(aPvtKeyContext);
      return pgpErr;
  }

  pgpErr=PGPContinueHash(aHashContext, theText,theTextSize);
  if (IsPGPError(pgpErr)) {
      if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":7", pgpErr);
      PGPFreeData(*theSignatureRef);
      *theSignatureRef=NULL;
      *theSgnSizePtr=0;
      PGPFreeHashContext(aHashContext);
      PGPFreePrivateKeyContext(aPvtKeyContext);
      return pgpErr;
  }

  pgpErr=PGPPrivateKeySign(aPvtKeyContext,aHashContext,
			   *theSignatureRef,theSgnSizePtr);
  if (IsPGPError(pgpErr)){
    if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":8", pgpErr);  
    PGPFreeData( *theSignatureRef);
  }
  /* It seems that it is freed by PGPPrivateKeySign */
  /* PGPFreeHashContext(aHashContext); */
  
  PGPFreePrivateKeyContext(aPvtKeyContext);
  
return pgpErr;

}/*  EnginePgpSignWithKey */


/*
* EnginePgpVerifySgnWithPubKey()
* This functions verifies theText against *theSignaturePtr.
* The size of *theSignaturePtr and *theText are passed by theTextSize
* and theSgnSize parameters. If theSignKey==NULL it searches the theKeySet 
* for a key that matches the *theSigner. theHashAlgID specifies the desired 
* hash algorithm for signing.
* isArmored is currently NOT SUPPORTED and is provisioned for future 
* versions.
*/
static 
PGPError EnginePgpVerifySgnWithPubKey(PGPContextRef thePgpContext,
				      PGPKeySetRef theKeySet,
				      char *theText, PGPSize theTextSize,
				      char *theSignaturePtr, 
				      PGPSize theSgnSize,
				      char* theSigner,
				      PGPKeyRef theSignKey,
				      PGPHashAlgorithm theHashAlgID,
				      PGPBoolean isArmored)
{
#undef _func_name
#define _func_name "EnginePgpVerifySgnWithPubKey"
  PGPError pgpErr=kPGPError_NoErr;
  PGPPublicKeyContextRef aPublicKeyContext;
  PGPKeyRef aKey;
  PGPHashContextRef aHashContext;

  if (!theText) return -1;

  if (!theSignKey) {
    pgpErr=EnginePgpSearchForUserKey(thePgpContext,theKeySet,
				     theSigner ,&aKey);
    if (IsPGPError(pgpErr)){
      if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":1", pgpErr);
      return pgpErr;
    }
  } else {
    aKey=theSignKey;
  }

  pgpErr=PGPNewPublicKeyContext(aKey, kPGPPublicKeyMessageFormat_PGP,
				&aPublicKeyContext);
  if (IsPGPError(pgpErr)) {
      if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":2", pgpErr);
      return pgpErr;
  }

  pgpErr=PGPNewHashContext(thePgpContext,theHashAlgID,&aHashContext);
  if (IsPGPError(pgpErr)) {
      if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":3", pgpErr);
      PGPFreePublicKeyContext(aPublicKeyContext);
      return pgpErr;
  }
 
  pgpErr=PGPContinueHash(aHashContext, theText, theTextSize);
  if (IsPGPError(pgpErr)) {
      if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":4", pgpErr);
      PGPFreePublicKeyContext(aPublicKeyContext);
      PGPFreeHashContext(aHashContext);
      return pgpErr;
  }
  
  pgpErr=PGPPublicKeyVerifySignature(aPublicKeyContext,aHashContext, 
				     theSignaturePtr, theSgnSize );
  if (IsPGPError(pgpErr))
    if(gPgpDebugFlag) Pgp_DEBUG(gPgpReportFile,_func_name":5", pgpErr);
  PGPFreePublicKeyContext(aPublicKeyContext);
  //  PGPFreeHashContext(aHashContext);
  // it seems that it is deallocated by the PGPPublicKeyVerifySignature
  return pgpErr;

}/* EnginePgpVerifySgnWithPubKey */


/*  ==================== BEGIN SAMPLE KEY-GENERATION CODE ==================== */

/*
  {
    PGPError pgpErr;
    PGPKeyRef theSipdMasterKey;
    PGPSubKeyRef theSipdSubKey;
    char *theSipdMasterKeyName="sipd_cs_columbia_edu";      
    pgpErr=EnginePgpGenerateKey(gPgpContext,
				gPgpKeySet,
				&theSipdMasterKey,
				theSipdMasterKeyName,strlen(theSipdMasterKeyName),
				NULL,
				1,
				"dummykey",
				kPGPPublicKeyAlgorithm_DSA,
				1024,
				kPGPExpirationTime_Never
				);
    Pgp_DEBUG(stdout,"Error from master key generation",pgpErr);
    
    pgpErr=EnginePgpGenerateKey(gPgpContext, 
				gPgpKeySet,
				&theSipdSubKey,
				theSipdMasterKeyName,strlen(theSipdMasterKeyName),
				theSipdMasterKey,
				1,
				"dummykey",
				kPGPPublicKeyAlgorithm_ElGamal,
				1024,
				kPGPExpirationTime_Never
				);
    
    Pgp_DEBUG(stdout,"Error from sub key generation:",pgpErr);
    
    }
*/

/*  ==================== END SAMPLE KEY-GENERATION CODE ==================== */


/*  ======= BEGIN SAMPLE SIGNATURE GENERATION AND VERIFICATION CODE ======== */

/*
    {
      PGPError pgpErr;
      char *theSignature;
      char formatStr[100];
      PGPBoolean isArmored=1;
      PGPSize theSignatureSize;
      char *theSigner="sipd_cs_columbia_edu";
      char *theText="Triantafyllos Alexiou\n";
      char *thePassphrase="dummykey";
      char theTextSize=strlen(theText);
    
      
      pgpErr=EnginePgpOSignWithPvtKey(gPgpContext,
				  gPgpKeySet, 
				  theText, theTextSize,
				  &theSignature,
				  &theSignatureSize,
				  theSigner, thePassphrase,
				  NULL,
				  kPGPHashAlgorithm_SHA,
				  isArmored);
      Pgp_DEBUG(gPgpReportFile,"sipd.c:OSignError:",pgpErr);
   
      sprintf(formatStr,"THE SIGNATURE IS:\n%%%ds\n",theSignatureSize);
      fprintf(gPgpReportFile, formatStr,theSignatureSize?theSignature:"");
      fflush(gPgpReportFile);
    
      pgpErr=EnginePgpOVerifySignatureWithPubKey(gPgpContext,
						 gPgpKeySet, 
						 theText, 
						 theTextSize,
						 theSignature,
						 theSignatureSize,
						 theSigner,
						 NULL,
						 kPGPHashAlgorithm_SHA,
						 isArmored);

      Pgp_DEBUG(gPgpReportFile,"sipd.c:OVerifyError:",pgpErr);

      PGPFreeData(theSignature);
      theSignature=NULL;

        
      fprintf(gPgpReportFile,"\nSTEP 2ond\n");fflush(gPgpReportFile);
      
      pgpErr=EnginePgpSignWithPvtKey( gPgpContext,
				  gPgpKeySet, 
				  theText,theTextSize,
				  &theSignature,
				  &theSignatureSize,
				  theSigner, thePassphrase,
				  NULL,
				  kPGPHashAlgorithm_SHA,isArmored);
      
      Pgp_DEBUG(gPgpReportFile,"sipd.c:SignError:",pgpErr);
      
      pgpErr=EnginePgpVerifySgnWithPubKey(gPgpContext,
					  gPgpKeySet,
					  theText, theTextSize,
					  theSignature, 
					  theSignatureSize,
					  theSigner,
					  NULL,
					  kPGPHashAlgorithm_SHA,isArmored);
      Pgp_DEBUG(gPgpReportFile,"sipd.c:VerifyError:",pgpErr);
      
    }

*/

/*  ======= END SAMPLE SIGNATURE GENERATION AND VERIFICATION CODE ======== */


/*  ======= BEGIN SAMPLE ENCRYPTION-DECRYPTION CODE ======== */
/*
    {

      PGPError pgpErr;
      char formatStr[100];
      PGPBoolean isArmored=1;
      char *theRecipients[]={ "sipd_cs_columbia_edu", NULL };
      char *theText="Triantafyllos Alexiou\n";
      char *thePassphrase="dummykey";
      char theTextSize=strlen(theText);
      char *theCipherText= NULL;
      PGPSize theCipherTextSize=0;
      char *theDecrText= NULL;
      PGPSize theDecrTextSize=0;


      pgpErr=EnginePgpOEncryptToUserIDs( gPgpContext,
					gPgpKeySet,
					theText, theTextSize,
					&theCipherText,
					&theCipherTextSize,
					theRecipients,
					NULL,
					isArmored
					);

      Pgp_DEBUG(gPgpReportFile,"sipd.c:OEncryptError:",pgpErr);
   
      sprintf(formatStr,"THE CIPHERTEXT IS:\n%%%ds\n", theCipherTextSize );
      fprintf(gPgpReportFile, formatStr, theCipherText ? theCipherText :"");
      fflush(gPgpReportFile);
    

      pgpErr=EnginePgpODecryptWithPvtKey( gPgpContext, gPgpKeySet,
					  &theDecrText, &theDecrTextSize,
					  theCipherText, theCipherTextSize,
					  theRecipients[0], 
					  thePassphrase,
					  NULL,
					  1 );


      Pgp_DEBUG(gPgpReportFile,"sipd.c:ODecryptError:",pgpErr);

      sprintf(formatStr,"THE DECRYPTEDTEXT IS:\n%%%ds\n", theDecrTextSize );
      fprintf(gPgpReportFile, formatStr, theDecrText ? theDecrText :"");
      fflush(gPgpReportFile);

      if ( theCipherText ) PGPFreeData( theCipherText );
      theCipherText=NULL;
      if ( theDecrText ) PGPFreeData( theDecrText );
      theDecrText= NULL;

    }
*/
/*  ======= END SAMPLE ENCRYPTION-DECRYPRION CODE ======== */










