/*
expire.c - handles registrations for this trial version of sipd

Copyright 1998-1999 by Columbia University; all rights reserved 
*/

#ifdef __linux__
#define _GNU_SOURCE /* Needed for strptime() declaration from <time.h> */
#endif

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <netdb.h> /* MAXHOSTNAMELEN */
#include <sys/types.h>
#include <sys/param.h> /* MAXHOSTNAMELEN */
#include <unistd.h>
#include <netinet/in.h>  /* struct in_addr */
#include <arpa/inet.h>   /* inet_addr() */
#include "md5.h"
#include "error.h"
#include "sipd.h"
#include "config.h"
#include "sipd-config.h"

#include "sysdep.h"

#ifdef DO_LICENSING

extern char *getcanonicalhostname();

static char *hostname;
static char *internal_code = "irt_sipd_1.0";

struct reg_info {
  char *domain;
  char *exp_str;
  time_t t_expire;
  char *md5;
};


/*
 * noisy_getcanonicalhostname()
 * Get and return the canonical hostname of this machine.
 * Not thread-safe -- only intended for startup time.
 * This is a copy of getcanonicalhostname() from sipd-config.c, but a lot
 * more chatty about what's going on.
 */
static char *noisy_getcanonicalhostname(void)
{
  char buf[MAXHOSTNAMELEN];
  struct hostent *hep;
  struct in_addr addr;
  char *addr_str;

  addr.s_addr = INADDR_ANY;

  gethostname(buf, MAXHOSTNAMELEN);
  
  hep = gethostbyname(buf);
  if (hep == NULL) {
    error("getcanonicalhostname", "gethostbyname failed on %s, h_errno = %d\n",
          buf, h_errno);
    return NULL;
  }

  memcpy(&addr, hep->h_addr, sizeof(addr));

  hep = gethostbyaddr((char *)&(addr.s_addr), sizeof(addr.s_addr), AF_INET);
  if (hep == NULL) {
    error("getcanonicalhostname", "gethostbyaddr failed on %s, h_errno = %d\n",
          inet_ntoa(addr), h_errno);
    return NULL;
  }

  addr_str = inet_ntoa(addr);

  if (commandline.use_numeric_address) {
    return strdup(addr_str);
  }
  else if (hep->h_name == NULL) {
    fprintf(stderr,
            "WARNING: No forward hostname lookup was found for the IP address %s.  "
            "Please check your name resolution and DNS setup.\n"
            "Using \"%s\" as the local hostname for licensing.\n",
            addr_str, addr_str);
    return strdup(addr_str);
  }
  else if (strchr(hep->h_name, '.') == NULL) {
    fprintf(stderr, 
            "WARNING: The resolved local hostname \"%s\" does not appear to be a "
            "fully-qualified domain name.  Please check your name resolution and DNS setup.\n"
            "Using numeric address \"%s\" as the local hostname for licensing.\n",
            hep->h_name, addr_str);
    return strdup(addr_str);
  }
  else {
    return strdup(hep->h_name);
  }
} /* getcanonicalhostname */


/*
 * parse_registration()
 * Parses a registration string: of the form "domain:YYYY-MM-DD:md5" or
 * "domain:-:md5" for unlimited-time licenses.
 * Return false if it could not be parsed, otherwise fills in reg_info.
 * reg_info will contain pointers into registration string, which must be
 * writeable.
 * Not thread-safe; should only be called at program startup.
 */
static int parse_registration(char *registration, struct reg_info *reg_info,
                              const char *filename, int lineno)
{
  char *lasts;
  struct tm tm_expire;

  reg_info->domain   = strtok_r(registration, ":", &lasts);
  reg_info->exp_str  = strtok_r(NULL, ":", &lasts);
  reg_info->md5      = strtok_r(NULL, ":\n", &lasts);

  if (reg_info->domain == NULL || reg_info->exp_str == NULL || 
      reg_info->md5 == NULL) {
    fprintf(stderr,
            "check_license: %s:%d: Malformed license entry (not enough fields).\n",
            filename, lineno);
    return 0;
  }
  
  if (strcmp(reg_info->exp_str, "-") == 0) {
    reg_info->t_expire = 0;
  }
  else {
    /* Some strptime() implementations (glibc, FreeBSD) don't initialize
       all the fields of struct tm, so set it to zero first. */
    memset(&tm_expire, '\0', sizeof(tm_expire));
    if (strptime(reg_info->exp_str, "%Y-%m-%d", &tm_expire) == NULL) {
      fprintf(stderr, "check_license: %s:%d: Invalid expiration date \"%s\".\n",
              filename, lineno, reg_info->exp_str);
      return 0;
    }
    
    reg_info->t_expire = mktime(&tm_expire);
    if (reg_info->t_expire == (time_t)-1) {
      fprintf(stderr, "check_license: %s:%d: Invalid expiration date \"%s\".\n",
              filename, lineno, reg_info->exp_str);
      return 0;
    }
  }
  return 1;
} /* parse_registration */


/* check_reg_info()
 * Confirm that the registration information corresponds
 * to this host, and is up-to-date.  Returns false if not.
 */
static int check_reg_info(struct reg_info *reg_info, const char* filename, 
                          int lineno)
{
  time_t t;
  char *domain_start;

  if (reg_info->t_expire != 0) {
    time(&t);
    if (t > reg_info->t_expire) {
      fprintf(stderr,
              "check_license: %s:%d: expiration date %s (%ld) is in the past.\n",
              filename, lineno, reg_info->exp_str, (long)reg_info->t_expire);
      return 0;
    }
  }

  domain_start = hostname + strlen(hostname) - strlen(reg_info->domain);
  if (strcmp(domain_start, reg_info->domain) != 0) {
    fprintf(stderr,
            "check_license: %s:%d: The local hostname \"%s\" does not match the license domain \"%s\".\n",
            filename, lineno, hostname, reg_info->domain);
    return 0;
  }

  return 1;
} /* check_registration */


/* check_md5()
 * Validate the the MD5 hash of the registration information.
 */
static int check_md5(struct reg_info *reg_info, char *filename, int lineno)
{
  int ret = 0, n;
  char buf[256], *md5;

#ifdef HAVE_SNPRINTF
  n = snprintf(buf, sizeof(buf), "%s:%s:%s", reg_info->domain,
               reg_info->exp_str, internal_code);
#else
  if (strlen(reg_info->domain) + strlen(reg_info->exp_str) +
      strlen(internal_code) >= sizeof(buf)) {
    fprintf(stderr, "check_license: snprintf error or overflow\n");
  }
  n = sprintf(buf, "%s:%s:%s", reg_info->domain,
              reg_info->exp_str, internal_code);
#endif
  
  if (n < 0 || n >= sizeof(buf)) {
    fprintf(stderr, "check_license: snprintf error or overflow\n");
  }
  md5 = MD5(buf);
  if (strcmp(md5, reg_info->md5) == 0) {
    ret = 1;
  }
  else {
    fprintf(stderr,
            "check_license: %s:%d: hash string \"%s\" is not valid "
            "(it does not authorize the domain \"%s\" and expiration date \"%s\").\n",
            filename, lineno, reg_info->md5,
            reg_info->domain, reg_info->exp_str);
  }
  free(md5); /* this is malloc'ed storage */
  return ret;
}


/* 
 * Returns 1 if the running system has a valid registration code.
 */
static int registration_valid(void)
{
  struct reg_info reg_info;
  char code[BUFSIZ];
  FILE *regs;
  int lineno;

  debug("check_license", "Reading license file %s\n",
        commandline.LicenseFile);
  if ((regs = fopen(commandline.LicenseFile, "r")) == NULL) { 
    fatal("check_license", "Cannot open license file %s\n", 
          commandline.LicenseFile);
  }

  lineno = 0;
  
  while (fgets(code, BUFSIZ, regs) != NULL) {
    lineno++;
    if (parse_registration(code, &reg_info, commandline.LicenseFile, lineno) &&
        check_md5(&reg_info, commandline.LicenseFile, lineno) &&
        check_reg_info(&reg_info, commandline.LicenseFile, lineno)) {
      fclose(regs);
      return 1;
    }
  }
  fclose(regs);
  return 0;
} /* registration_valid */


/* 
 * check_registration()
 * Checks for sipd binary registration
 * if match, continue, else terminate program
 */
void check_registration()
{
  hostname = noisy_getcanonicalhostname();
  if (hostname == NULL) {
    error("check_registration", "getcanonicalhostname");
  }

  if (!registration_valid()) {
    fprintf(stderr, "\nYour license file %s does not contain any valid licenses.\n"
            "You can request a new license from the web page\n"
            "<http://www.cs.columbia.edu/~hgs/license/>.\n",
            commandline.LicenseFile);
    exit(1);
  }
  /* everything ok, continue */
  free(hostname);
} /* check_registration */

#else /* ! DO_LICENSING */

/* 
 * check_registration()
 * Dummy registration check if we configured with licensing turned off.
 */
void check_registration()
{
} /* check_registration */

#endif
