/* gateways.c: map a telephone number to a SIP url of an appropriate
 * gateway. */
/* Copyright 2001 Columbia University.  All rights reserved. */

#ifdef __linux__
#define _GNU_SOURCE /* Needed for rwlock declarations from <pthread.h> */
#endif

#include <ctype.h>
#include <errno.h>
#include <pthread.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "bitmap.h"
#include "error.h"
#include "pattern.h"
#include "strdupn.h"

#include "telmap.h"

#include "gateways.h"

/*
 * An entry from the gateway file.
 * Since we always walk the gateways in order, a singly-linked list is a
 * good-enough data structure.
 */
typedef struct gateway_entry {
  pattern_t pattern;

  bitmap_t classes;

  char *result;

  int lineno;

  struct gateway_entry *next;
} gateway_entry;

/* The current gateway file. */
static gateway_entry * gateways = NULL;

/* An entry from the list of classes. */
typedef struct gwclass_entry {
  int number;
  char *name;

  struct gwclass_entry *next;
} gwclass_entry;

/* The current list of gwclasses. */
static gwclass_entry * gwclasses = NULL;

/* The gwclass number counter */
static int gwclass_count = 0;


#ifdef HAVE_PTHREAD_RWLOCK_INIT
static pthread_rwlock_t gateways_lock = PTHREAD_RWLOCK_INITIALIZER;
#else
static pthread_mutex_t gateways_lock = PTHREAD_MUTEX_INITIALIZER;
#endif

/*
 * ReadLockGateways()
 * Lock the gateway list for reading.  While you hold this lock, you must
 * treat the 'gateways' and 'gwclasses' globals as const.
 * XXX: should be more clever for the case when we haven't got Pthread
 * RW-locks.  Pthread mutex's probably block the system too much.
 */
static void ReadLockGateways(void)
{
#ifdef HAVE_PTHREAD_RWLOCK_INIT
  pthread_rwlock_rdlock(&gateways_lock);
#else
  pthread_mutex_lock(&gateways_lock);
#endif
} /* ReadLockGateways() */


/*
 * WriteLockGateways()
 * Lock the gateway list for writing.  While you hold this lock, you may
 * alter the contents of the 'gateways' and 'gwclasses' globals as much as
 * you wish.
 */
static void WriteLockGateways(void)
{
#ifdef HAVE_PTHREAD_RWLOCK_INIT
  pthread_rwlock_wrlock(&gateways_lock);
#else
  pthread_mutex_lock(&gateways_lock);
#endif
} /* WriteLockGateways() */


/*
 * UnlockGateways()
 * Unlock the gateway list.
 */
static void UnlockGateways(void)
{
#ifdef HAVE_PTHREAD_RWLOCK_INIT
  pthread_rwlock_unlock(&gateways_lock);
#else
  pthread_mutex_unlock(&gateways_lock);
#endif
} /* UnlockGateways() */


/*
 * free_gateway()
 * Free a single gateway entry.
 * If this entry is on the global list, you should hold the write lock.
 */
static void free_gateway(gateway_entry *gw)
{
  pattern_free(&gw->pattern);
  free_valid(gw->result);
  free_valid(gw);
} /* free_gateway */


/*
 * free_gateways_core()
 * Free a list of gateway entries.
 * If this list is the global list, you should hold the write lock.
 */
static void free_gateways_core(gateway_entry *gws)
{
  gateway_entry *gw, *gw_next;

  for (gw = gws; gw != NULL; gw = gw_next) {
    gw_next = gw->next;
    free_gateway(gw);
  }
} /* free_gateways_core */


/*
 * free_gwclass()
 * Free a single gwclass entry.
 * If this entry is on the global list, you should hold the write lock.
 */
static void free_gwclass(gwclass_entry *gwclass)
{
  free_valid(gwclass->name);
  free_valid(gwclass);
} /* free_gwclass */


/*
 * free_gwclasses_core()
 * Free a list of gateway classes.
 * If this list is the global list, you should hold the write lock.
 */
static void free_gwclasses_core(gwclass_entry *gwclass_list)
{
  gwclass_entry *gwclass, *gwclass_next;

  for (gwclass = gwclass_list; gwclass != NULL; gwclass = gwclass_next) {
    gwclass_next = gwclass->next;
    free_gwclass(gwclass);
  }
} /* free_gwclasses_core */


/*
 * free_gateways_nolock()
 * Free the gateway list.  You should hold the write lock.
 */
static void free_gateways_nolock(void)
{
  free_gateways_core(gateways);
  gateways = NULL;

  free_gwclasses_core(gwclasses);
  gwclasses = NULL;
  gwclass_count = 0;
} /* free_gateways_nolock */


/*
 * free_gateways()
 * Free the gateway list.  Callable by external modules.
 */
void free_gateways(void)
{
  WriteLockGateways();

  free_gateways_nolock();

  UnlockGateways();
} /* free_gateways */


/*
 * lookup_gwclass()
 * The string 'string' describes a gateway class.  Search for it in the
 * list of classes, and return the class number, or -1.
 * The name may be NULL.
 */
static int lookup_gwclass(const char *gwclass_name)
{
  gwclass_entry *gwclass;

  if (gwclass_name == NULL) return -1;

  for (gwclass = gwclasses; gwclass != NULL; gwclass = gwclass->next) {
    if (strcasecmp(gwclass_name, gwclass->name) == 0) {
      return gwclass->number;
    }
  }

  return -1;
} /* lookup_gwclass */


/*
 * lookup_or_add_gwclass() 
 * The string 'string' describes a gateway class.  Search for it in the
 * list of classes.  If it doesn't exist, add it to the list.
 * The name must not be NULL.
 */
static int lookup_or_add_gwclass(const char *gwclass_name)
{
  int class_num;
  gwclass_entry *new_gwclass;

  class_num = lookup_gwclass(gwclass_name);
  if (class_num != -1) {
    return class_num;
  }

  new_gwclass = (gwclass_entry *) calloc(sizeof(gwclass_entry), 1);
  if (new_gwclass == NULL) {
    error("lookup_or_add_gwclass", "Memory allocation failure");
    return -1;
  }

  new_gwclass->number = gwclass_count++;
  strassign(&new_gwclass->name, gwclass_name);

  new_gwclass->next = gwclasses;
  gwclasses = new_gwclass;

  return new_gwclass->number;
} /* lookup_or_add_gwclass */


/*
 * match_gwclass()
 * Return true if the gateway class 'gwclass' is set in the gateway bitmap
 * 'set'.  If class is non-negative, test the bit in the bitmap, otherwise
 * return false.
 */
static int match_gwclass(bitmap_t set, int gwclass) {
  if (gwclass < 0) {
    return 0;
  }
  return bitisset(set, gwclass);
} /* match_gwclass */


/*
 * parse_gwclass_set()
 * Parse a comma-separated list of gwclasses, and turn them into a bitmap.
 * Fail if we have more than BITMAP_MAX classes.
 * Return an all-0 bitmap on failure.
 */
static bitmap_t parse_gwclass_set(const char *orig_gwclass_set)
{
  bitmap_t ret;
  char *gwclass_set;
  char *gwclass;
  char *lasts;
  int gwclass_num;
  
  clearmap(ret);

  gwclass_set = strdup(orig_gwclass_set);

  for (gwclass = strtok_r(gwclass_set, ",", &lasts); 
       gwclass != NULL;
       gwclass = strtok_r(NULL, ",", &lasts)) {
    if (strcasecmp(gwclass, "*") == 0) {
      /* Special case -- any class. */
      setmap(ret);
      continue;
    }

    gwclass_num = lookup_or_add_gwclass(gwclass);
    if (gwclass_num == -1) {
      clearmap(ret);
      goto done;
    }
    if (gwclass_num > BITMAP_MAX) {
      errno = 0;
      error("parse_gwclass_set",
            "Too many gateway classes: can only have %d\n",
            BITMAP_MAX + 1);
      clearmap(ret);
      goto done;
    }
    setbit(ret, gwclass_num);
  }

done:  
  free_valid(gwclass_set);
  return ret;
} /* parse_gwclass_set */


/*
 * parse_gateway_entry()
 * Parse one gateway entry line, filling in the passed-in gw structure.
 * On error, return false.
 */
static int parse_gateway_entry(gateway_entry *gw, const char* line,
                                int lineno)
{
  const char *p, *n;
  char errbuf[1024];
  char *pattern;
  char *gwclass_set;
  bitmap_t gwclass_set_bits;
  int quote_next = 0;
  int ret;

  /* Find the pattern.  The pattern terminates with the first
   * non-quoted space. */
  for (p = line; quote_next || (*p != ' ' && *p != '\t'); p++) {
    switch (*p) {
    case '\\':
      quote_next = 1;
      break;

    case '\n':
    case '\0':
      /* Fell off the end of the line */
      errno = 0;
      error("parse_gateway_entry", "Malformed gateway entry at line %d\n",
            lineno);
      return 0;

    default:
      quote_next = 0;
      continue;
    }
  }

  pattern = strdupn(line, (p - line));
  if (pattern == NULL) {
    error("parse_gateway_entry", "Memory allocation failure\n");
    return 0;
  }
    
  ret = pattern_compile(&gw->pattern, pattern, errbuf, sizeof(errbuf));
  free_valid(pattern);
  if (!ret) {
    error("parse_gateway_entry", "%s at line %d\n", errbuf, lineno);
    return 0;
  }

  /* Skip over intermediate whitespace */
  while (*p != '\0' && *p != '\n' && isspace((int)*p)) {
    p++;
  }

  /* Find the gateway class set.  It terminates with the next whitespace. */
  n = p;
  while (*n != '\0' && !isspace((int)*n)) {
    n++;
  }

  gwclass_set = strdupn(p, (n - p));
  gwclass_set_bits = parse_gwclass_set(gwclass_set);
  free_valid(gwclass_set);

  if (gwclass_set_bits == 0) {
    errno = 0;
    error("parse_gateway_entry", "Malformed gateway class set at line %d\n",
          lineno);
    pattern_free(&gw->pattern);
    return 0;
  }
  gw->classes = gwclass_set_bits;

  p = n;

  /* Skip over intermediate whitespace again */
  while (*p != '\0' && *p != '\n' && isspace((int)*p)) {
    p++;
  }

  /* Find terminating newline */
  if (*p == '\0' || *p == '\n' || (n = strchr(p, '\n')) == NULL) {
    /* Didn't find a third field before '\n' */
    errno = 0;
    error("parse_gateway_entry", "Malformed gateway entry at line %d\n",
          lineno);
    pattern_free(&gw->pattern);
    return 0;
  }
  gw->result = strdupn(p, (n - p));
  if (gw->result == NULL) {
    error("parse_gateway_entry", "Memory allocation failure\n");
    pattern_free(&gw->pattern);
    return 0;
  }

  gw->lineno = lineno;

  return 1;
} /* parse_gateway_entry */


/*
 * init_gateways()
 * Initialize the gateway list, by reading the gateway file 'gw_file'.
 * May be called more than once -- if so, free the pre-existing gateway
 * information.  (This would typically occur after re-loading the config
 * file.)
 * Return false on error.
 */
int init_gateways(const char *gw_file)
{
  gateway_entry *new_gw, **prev_gws;
  FILE *f;
  char line[1024];
  int line_cnt;

  WriteLockGateways();

  if (gateways != NULL) {
    free_gateways_nolock();
  }
  
  if (!(f = fopen(gw_file, "r"))) {
    error("init_gateways", "%s", gw_file);
    goto bailout;
  }

  prev_gws = &gateways;

  line_cnt = 0;
  while (fgets(line, sizeof(line), f)) {
    line_cnt++;
    if (!line[0] || line[0] == '#' || line[0] == '\n') {
      continue;
    }
    new_gw = (gateway_entry *) malloc(sizeof(gateway_entry));
    if (new_gw == NULL) {
      error("init_gateways", "Memory allocation failure\n");
      goto bailout;
    }
    memset(new_gw, 0, sizeof(gateway_entry));

    if (!parse_gateway_entry(new_gw, line, line_cnt)) {
      free_gateway(new_gw);
      goto bailout;
    }
    
    *prev_gws = new_gw;
    new_gw->next = NULL;
    prev_gws = &new_gw->next;
  }
  fclose(f);

  UnlockGateways();

  return 1;

bailout:
  free_gateways_nolock();

  UnlockGateways();
  
  return 0;
} /* init_gateways() */


/*
 * map_gateway()
 * Given a source URI and a gateway class, map a source URI to a destination
 * URI.  Return a telmap_err describing what's happened.
 */
telmap_err map_gateway(uri_t *uri_dest, const uri_t *uri_src,
                          const char *user_gateway_class)
{
  telmap_err ret;
  int match_ret;
  const gateway_entry *gw;
  int user_gw_classno;
  char errbuf[1024];

  ReadLockGateways();

  ret = TELMAP_NOTFOUND;

  if (gateways == NULL) {
    goto done;
  }

  user_gw_classno = lookup_gwclass(user_gateway_class);

  for (gw = gateways; gw != NULL; gw = gw->next) {
    char *new_uri_str;

    if (match_gwclass(gw->classes, user_gw_classno)) {
      match_ret = pattern_interpolate(uri_src->host, &gw->pattern,
                                      gw->result, &new_uri_str,
                                      errbuf, sizeof(errbuf));
      if (match_ret == 0) {
        continue;
      }
      if (match_ret < 0) {
        errno = 0;
        error("map_gateway", "%s at %d\n", errbuf, gw->lineno);
        ret = TELMAP_ERR;
        goto done;
      }

      /* else match_ret > 0 */
      URI_Free(uri_dest);
      *uri_dest = URI_Parse(new_uri_str);
      free_valid(new_uri_str);
      ret = TELMAP_OK;
      goto done;
    }
    else {
      /* The user isn't in this gateway class.  However, we still want
       * to try to match, so we can return TELMAP_NOPERM rather than
       * TELMAP_NOTFOUND if appropriate. */
      match_ret = pattern_match(uri_src->host, &gw->pattern,
                                errbuf, sizeof(errbuf));
      if (match_ret == 0) {
        continue;
      }
      if (match_ret < 0) {
        errno = 0;
        error("map_gateway", "%s at %d", errbuf, gw->lineno);
        ret = TELMAP_ERR;
        goto done;
      }
      /* else match_ret > 0 */
      ret = TELMAP_NOPERM;
      continue;
    }
  }

done:
  UnlockGateways();

  return ret;
} /* map_gateway */
