/*
 * method.c - handle INVITE and other request methods.
 */

/* Copyright 1998-2001 by Columbia University.  All rights reserved. */

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <regex.h>
#include <errno.h>

#include "sip.h"
#include "config.h"
#include "request-sip.h"
#include "register.h"
#include "dstring.h"
#include "sysdep.h"
#include "strdupn.h"
#include "policy.h"
#include "authenticateSIP.h"
#include "log.h"
#include "libsip.h"
#include "canonicalize.h"
#include "querydb.h"

#include "sipd-policy.h"
#include "sipd-support.h"
#include "method.h"
#include "telmap.h"

/*
 * SendProvisional()
 * Send a provisional (1xx) response to the request.
 * This also sets this value as the response to retransmitted requests.
 */
static void SendProvisional(request_t *r, int status, const char *reason)
{
  ResponseSetStatus(r, status, reason); 
  ResponseSetBody(r, NULL, NULL, 0);
  ResponseBuildDefault(r);
  ResponseSend(r, 0, 0);
} /* SendProvisional */


/*
 * AuthenticateFrom()
 * This request needs to proxy-authenticate its source.
 * Setup r->source_user, and return auth_success if the authentication
 * is correct.  Otherwise, return an appropriate auth_status and set
 * the response status and reason.
 */
static auth_status AuthenticateFrom(request_t *r, int required, int remote_ok)
{
  r->source_user = CanonicalizeAndLookupURI(r, &r->from.uri, remote_ok,
                                            "From URI",
                                            "request needing authentication",
                                            r->destination_user);

  if (r->source_user == NULL) {
    return auth_internal_error;
  }

  return AuthenticateSIP(r, 1, 1, !required);
} /* AuthenticateFrom */


/*
 * DoRegister()
 * Handle a local REGISTER request.
 * A request is a local REGISTER request if it is directed to
 * the registrar URI, i.e. sip:[host name matching config.domain].
 * (Note that the Request-URI has already been through canonicalization
 * once.  However, even a canon_unique URI may be a "foreign" URI by
 * this standard, since canonicalize() may have mapped a local URI to a
 * remote one.)
 * It is an invalid local REGISTER request if it is directed to
 * any non-SIP URI, or if the host part matches config.domain but the
 * user part is non-empty.
 * Otherwise, it should be proxied as normal.
 * Return true if we handled the request here; false if it should proceed.
 */
static int DoRegister(request_t *r, const canon_result *canon_res)
{
  int result;
  uri_t *uri;

  if (canon_res->status != canon_unique) {
    return 0;
  }

  uri = canon_res->u.unique;

  if (strcasecmp(uri->scheme, "sip") != 0) {
    ResponseSetStatus(r, 400, "Invalid Request-URI scheme for REGISTER");
    goto fail;
  }

  if (uri->host == NULL) {
    ResponseSetStatus(r, 400, "Invalid Request-URI for REGISTER");
    goto fail;
  }

  result = regexec(config.domain, uri->host, 0, NULL, 0);

  if (result == REG_NOMATCH) {
    return 0;
  }

  if (result != 0) {
    char buf[1024];
    regerror(result, config.domain, buf, sizeof(buf));
    errno = 0;
    error("canonicalize_hostname",
          "Error matching hostname \"%s\": %s\n",
          uri->host, buf);
    ResponseSetStatus(r, 500, NULL);
    goto fail;
  }

  if (uri->user != NULL && uri->user[0] != '\0') {
    ResponseSetStatus(r, 400, "Invalid Request-URI for REGISTER");
    goto fail;
  }

  /* XXX: check other attributes of URI? */

  /* Invoke REGISTER */
  REGISTER(r);

  return 1;

fail:
  ResponseBuildDefault(r);
  debug("DoRegister", "Request 0x%p: not proceeding: %d %s\n",
        r, r->status, r->reason ? r->reason : GetStandardReason(r->status));
  ResponseSend(r, 0, 1);

  return 1;
} /* DoRegister */


/*
 * DoSubscribe()
 * Handle a local SUBSCRIBE request.
 * A request is a local SUBSCRIBE request if it is directed to
 * the registrar URI, i.e. sip:[host name matching config.domain].
 * (Note that the Request-URI has already been through canonicalization
 * once.  However, even a canon_unique URI may be a "foreign" URI by
 * this standard, since canonicalize() may have mapped a local URI to a
 * remote one.)
 * It is an invalid local SUBSCRIBE request if it is directed to
 * any non-SIP URI, or if the host part matches config.domain but the
 * user part is non-empty.
 * Otherwise, it should be proxied as normal.
 * Return true if we handled the request here; false if it should proceed.
 */
static int DoSubscribe(request_t *r, const canon_result *canon_res)
{
  int result;
  uri_t *uri;

  if (canon_res->status != canon_unique) {
    return 0;
  }

  uri = canon_res->u.unique;

  if (strcasecmp(uri->scheme, "sip") != 0) {
    ResponseSetStatus(r, 400, "Invalid Request-URI scheme for SUBSCRIBE");
    goto fail;
  }

  if (uri->host == NULL) {
    ResponseSetStatus(r, 400, "Invalid Request-URI for SUBSCRIBE");
    goto fail;
  }

  result = regexec(config.domain, uri->host, 0, NULL, 0);

  if (result == REG_NOMATCH) {
    return 0;
  }

  if (result != 0) {
    char buf[1024];
    regerror(result, config.domain, buf, sizeof(buf));
    errno = 0;
    error("canonicalize_hostname",
          "Error matching hostname \"%s\": %s\n",
          uri->host, buf);
    ResponseSetStatus(r, 500, NULL);
    goto fail;
  }

  /* Invoke SUBSCRIBE */
  SUBSCRIBE(r);

  return 1;

fail:
  debug("DoSubscribe", "Request 0x%p: not proceeding: %d %s\n",
        r, r->status, r->reason ? r->reason : GetStandardReason(r->status));
  ResponseSend(r, 0, 1);

  return 1;
} /* DoSubscribe */

/*
 * DoLocalOperations()
 * Check if this request is for a local operation.  Currently the 
 * local operations we support are REGISTER and SUBSCRIBE.
 * If it is for a local operation, invoke the operation, and return true. 
 * Else return false.
 * XXX TODO: OPTIONS to registrar address.
 */
static int DoLocalOperations(request_t *r, const canon_result *canon_res)
{
  char *method;
  int ret;

  method = HTTP_Method(&r->p);

  if (strcasecmp(method, "REGISTER") == 0) {
    ret = DoRegister(r, canon_res);
  }
  else if (strcasecmp(method, "SUBSCRIBE") == 0) {
    ret = DoSubscribe(r, canon_res);
  }
  else {
    ret = 0;
  }
  
  free_valid(method);

  return ret;
} /* DoLocalOperations */


/*
 * DoForeign()
 * 'canonicalize' decided that the request's original Request-URI was
 * a foreign host (it didn't match the config.domain regex).
 * Handle as appropriate for the user: proxy, redirect (with DNS lookup),
 * return 403, or return 407.
 * Return whether to proceed to the "policy" stage, or just to
 * fail immediately.
 */
static int DoForeign(request_t *r, const canon_result *canon_res,
                      int *status, registration_t *reg)
{
  switch (config.foreign) {
  case L_redirect:
    registration_append(reg, *canon_res->u.unique, NULL, 1, L_redirect);
    *status = 302;
    r->policy = &redirect_policy;
    return 1;

  case L_proxy:
    if (AuthenticateFrom(r, 1, 0) != auth_success) {
      return 0;
    }

    registration_append(reg, *canon_res->u.unique, NULL, 0, L_proxy);
    *status = 302;
    r->policy = &proxy_policy;
    return 1;

  case L_unspecified:
  case L_reject:
  default:
    ResponseSetStatus(r, 403, "Relaying not allowed");
    return 0;
  }
} /* DoForeign */


/*
 * DoAmbiguous()
 * 'canonicalize' responded that the request's original Request-URI
 * was ambiguous.  Set up the response appropriately.
 * Return 1: Always proceed to the policy stage.
 */
static int DoAmbiguous(request_t *r, const canon_result *canon_res,
                        int *status, registration_t *reg)
{
  int i;

  *status = 485;

  if (config.DisplayAmbiguous) {
    for (i = 0; i < canon_res->u.multiple.nameaddr_count; i++) {
      registration_append(reg, 
                          canon_res->u.multiple.nameaddrs[i].addr,
                          canon_res->u.multiple.nameaddrs[i].name,
                          0,
                          L_unspecified);
    }
    r->policy = &redirect_policy;
  }
  else {
    r->policy = &failure_policy;
  }

  return 1;
} /* DoAmbiguous */


/*
 * DoTel()
 * 'canonicalize' resolved the request's original Request-URI to a 'tel'
 * URI, and DoUnique() didn't find a corresponding user in the PUT.
 * Thus, resolve this address to a SIP address, based on the number called
 * and the identity of the caller.
 * Return whether to return to the policy stage, or just to fail
 * immediately.
 */
static int DoTel(request_t *r, const canon_result *canon_res,
                 int *status, registration_t *reg)
{
  const char *user_gateway_class;
  telmap_err telmap_ret;
  uri_t uri_dest;

  /* Telephone number routing must be authenticated. */

  if (AuthenticateFrom(r, 1, 0) != auth_success) {
    /* Response code and header fields set by AuthenticateSIP on error */
    return 0;
  }

  if (r->source_user == NULL) {
    user_gateway_class = NULL;
  }
  else {
    user_gateway_class = r->source_user->gwclass;
  }

  telmap_ret = map_tel(&uri_dest, canon_res->u.unique, user_gateway_class);
  switch (telmap_ret) {
  case TELMAP_OK:
    registration_append(reg, uri_dest, NULL, 0, L_proxy);
    URI_Free(&uri_dest);
    *status = 302;
    r->policy = &proxy_policy;
    return 1;

  case TELMAP_NOTFOUND:
    ResponseSetStatus(r, 404, "No route to telephone number available");
    return 0;

  case TELMAP_NOPERM:
    ResponseSetStatus(r, 403,
                      "Insufficient permissions to reach telephone number");
    return 0;

  default:
    ResponseSetStatus(r, 500, "Internal Error in tel routing");
    return 0;
  }
} /* DoTel */


/*
 * DoUnique()
 * 'canonicalize' resolved the request's original Request-URI to a unique,
 * local URI.  Look up the user in the registration database.
 * If not present, return 404.  If authentication requested or required,
 * return 407.  If not registered, return 480.  Otherwise, invoke
 * an appropriate policy.
 * Return whether to return to the policy stage, or just to fail
 * immediately.
 */
static int DoUnique(request_t *r, const canon_result *canon_res,
                    int *status, registration_t *reg)
{
  db_err db_ret;
  char *db_ret_str = NULL;
  auth_status auth_ret;
  char *method;
  char *requri_userid;
  int required;
  int ret;
  DString reason;

  DStringInit(&reason);

  requri_userid = get_table_key_from_URI(config.database, canon_res->u.unique,
                                         &db_ret, &db_ret_str);
  if (requri_userid == NULL) {
    DStringPrintf(&reason, "Can't get database key: %s", db_ret_str);
    ResponseSetStatus(r, 400, DStringValue(&reason));
    ret = 0;
    goto done;
  }

  r->destination_user = put_get_by_user(config.database, requri_userid,
                                        &db_ret, &db_ret_str);

  if (r->destination_user == NULL) {
    if (db_ret_invalid(db_ret)) {
      errno = 0;
      error("DoUnique",
            "Request 0x%p: put_get_by_user(%s) returned %d: %s\n",
            r, requri_userid, db_ret, db_ret_str);
    }

    /* Handle "tel" URIs which aren't registered as users specially. */
    if (db_ret == DB_NOTFOUND &&
        (strcasecmp(canon_res->u.unique->scheme, "tel") == 0)) {
      ret = DoTel(r, canon_res, status, reg);
      goto done;
    }

    request_set_dberr_status(r, db_ret, requri_userid,
                             500, db_ret_str);
    ret = 0;
    goto done;
  }

  method = HTTP_Method(&r->p);

  db_ret = method_allowed_for_user(method, r->destination_user);

  free_valid(method);
  
  if (db_ret != DB_OK) {
    request_set_dberr_status(r, db_ret, requri_userid,
                             405, NULL);
    /* XXX: add Allow: headers for DB_NO */
    ret = 0;
    goto done;
  }

  /* Authenticate the request */
  /* Did the user want the request to be authenticated? */
  db_ret = user_wants_request_authentication(r->destination_user, &required);

  if (db_ret == DB_OK) {
    debug("DoUnique", "Request 0x%p: user %s %s caller authentication\n",
	  r, requri_userid, required ? "required" : "requested");

    auth_ret = AuthenticateFrom(r, required, 1);

    if (!required && auth_ret == auth_wrong_password) {
      /* The request contained an invalid authentication field; the
       * user wants the request to continue in this case.  Send a 183
       * response informing the caller of what's going on.
       */
      SendProvisional(r, 183, "Authentication failed");
    }
    else if (auth_ret != auth_success) {
      /* Authentication failed, and wasn't allowed to fail. */
      /* Response code and header fields set by AuthenticateSIP. */
      return 0;
    }
  }

  /* Query the registration database for the user. */
  register_find(r->destination_user->user, reg);

  if (reg->contacts <= 0 && !HasExecutableScript(&reg->scriptset)) {
    /* User not registered */
    /* XXX Invoke administrative script here? */
    ResponseSetStatus(r, 480, NULL);
    ret = 0;
    goto done;
  }

  *status = 302;

  debug("DoUnique",
        "Request 0x%p: status=%d, reg->scriptset.count=%d, "
        "reg->contacts=%ld\n",
        r, *status, reg->scriptset.count, (long)reg->contacts);
  
  /* If this user has CGI permission, check if there's a script specified */
  if ((r->destination_user->sip_groups & put_sip_group_cgi) &&
      GetScriptByDisposition(&reg->scriptset, 
                             disposition_sip_cgi) != NULL) {
    r->policy = &cgi_policy;
  }
  /* else CPL policy */

  /* Check for Route headers */
  /* XXX this seems wrong -- Script policies should be able to defer to
   * the route policy. */
  else if (RFC822_Header(r->p.headers, "Route") != NULL) {
    r->policy = &route_policy;
  }
  else {
    r->policy = find_primary_policy(reg, *status);
    if (r->policy == NULL) {
      ResponseSetStatus(r, 500, NULL);
      ret = 0;
      goto done;
    }
  }

  ret = 1;
done:
  free_valid(requri_userid);
  free_valid(db_ret_str);
  DStringFree(&reason);

  /* 
   * Route policy takes preference independent of others.
   * i.e, say, the user was not found, but a Route header was there.
   * then, Route it based on the route header. 
   *
   * XXX: This is a hack. if we make the decision whether to use
   * the route header, this is never needed.
   */
  if (!r->policy && RFC822_Header(r->p.headers, "Route") != NULL) {
    free_valid(r->reason);
    r->status = 0;
    r->policy = &route_policy;
    ret = 1;
  }
    
  return ret;
} /* DoUnique */


/*
 * TryStatlessProxy()
 * Try to statlessly proxy the request if possible.
 * If it is successful it frees the request_t 
 * structure and returns NULL.
 * If it could not proxy the request statelessly
 * then it returns the same request_t structure.
 */
static request_t *TryStatelessProxy(request_t *r,
                                    registration_t *reg)
{
  /* Try stateless proxy if allowed */
  if ( r->policy == &proxy_policy &&
       reg->contact != NULL &&
       reg->contact[0].uri.scheme != NULL &&
       reg->contact[1].uri.scheme == NULL) {
    /* Comes here only when there is one location
     * and it has proxy policy */
    
    message_t *m;
    int stateful = 0;

    debug("Method", "Choosing stateless proxy\n");
    
    /* remove the request from the list, so that no
     * retransmission is detected. */
    RequestDelete(r);
    
    /* Format the message and proxy the request */
    m = GetRequestAsMessage(r);
    stateful = client_stateless(r, &reg->contact[0].uri, m);
    MessageDelete(m);
    
    /* Free the request */
    if ( stateful == 0 ) {
      /* Stateless proxy was successful. Free
       * the transaction. */
      pthread_mutex_unlock(&r->mutex);
      RequestFree(r);
      r = NULL;
    }
    else {
      /* stateless proxy failed */
      /* Add the request back to the list*/
      RequestAdd(r);
      
      /* The rest of the code will anyway try the usual way 
       * to handle the request */
    }
  }
  
  return r;
} /* TryStatelessProxy */

/*
 * Method()
 * Process a sipd request.  Canonicalize the request-URI, and based on it
 * categorize the target of the request: foreign, local-user, registrar,
 * or error.
 * Dispatch appropriately.
 * Free the data structures when done.
 */
void Method(request_t *r)
{
  registration_t reg;          /* user locations */
  int status;                  /* the status to use for the response */
  canon_result *canon_res;     /* The result of canonicalization */
  int proceed;                 /* Proceed to the policy stage? */
  uri_t uri;                   /* The request's URI */

  /* Zero reg and canon_res, so that they do not fail when freed. */
  memset(&reg, 0, sizeof(reg));
  canon_res = &r->canonicalized_uri;

  status = -1;

  r->policy = NULL;

  if (!IsAcceptableURI(r, &r->p.URI, "Request-URI")) {
    proceed = 0;
    goto execute;
  }

  uri = HTTP_URI(&r->p);

  canonicalize(&uri, canon_res);

  if (DoLocalOperations(r, canon_res)) {
    goto done;
  }

  /* SendProvisional is broken up. We just build
   * the default response but do not send the response */
  ResponseSetStatus(r, 100, NULL); 
  ResponseSetBody(r, NULL, NULL, 0);
  ResponseBuildDefault(r);

  switch(canon_res->status) {
  case canon_unique:
    /* Determine policy to use.  Check authentication. */
    proceed = DoUnique(r, canon_res, &status, &reg);
    break;

  case canon_remote:
    /* Proxy, redirect, return 403, or return 407, as appropriate. */
    proceed = DoForeign(r, canon_res, &status, &reg);
    break;

  case canon_multiple:
    /* Return 485, with or without contact headers */
    proceed = DoAmbiguous(r, canon_res, &status, &reg);
    break;

  case canon_error:
    /* Canonicalize returned an internal error */
    ResponseSetStatus(r, 500, canon_res->u.error);
    proceed = 0;
    break;

  default:
    /* Shouldn't get here; report an error */
    errno = 0;
    error("Method", "Request 0x%p: Surprising value %d for canon_res->status\n",
          r, canon_res->status);
    ResponseSetStatus(r, 500, NULL);
    proceed = 0;
    break;
  }

execute:
  if (proceed && !r->policy) {
    /* Shouldn't happen -- sanity check */
    debug("Method", "Request 0x%p: proceed is true, but r->policy is NULL",
          r);
    ResponseSetStatus(r, 500, NULL);
    proceed = 0;
  }

  if (!proceed) {
    int retransmit;

    debug("Method", "Request 0x%p: not proceeding: %d %s\n",
          r, r->status, r->reason ? r->reason : GetStandardReason(r->status));
      
    retransmit = (strcasecmp(r->p.method, "INVITE") == 0) ? 1 : 0;
    ResponseSend(r, retransmit, 1); /* Send failure response */
  }

  else {
    debug("Method", "Request 0x%p: %s policy chosen\n", r, r->policy->name);

    /* Try to proxy it statelessly, if the canonicalizer returned
     * one location. */
    if ( config.TryStateless &&
         (canon_res->status == canon_unique ||
          canon_res->status == canon_remote)) {
      r = TryStatelessProxy(r, &reg);
    }

    if ( r!= NULL ) {
      /* This may take a while */
      if ( r->p.method != NULL && strcmp(r->p.method, "ACK") ) {
        ResponseSetStatus(r, 100, NULL);
        ResponseSend(r, 0, 0);
      }

      execute_policy(r, &reg, status);
    }
  }

done:

  register_free(&reg);

  return;
} /* Method */


/* The "core" policy to call. */

/*
 * find_primary_policy()
 * Find the primary policy (i.e. not considering scripts) given the value
 * DoUnique returned.
 */
policy_t *find_primary_policy(registration_t *reg, int status)
{
  if (reg->contacts > 0) {
    if (status >= 300 && status < 400) {
      /* If it's unspecified (can happen with aliases, etc),
       * do the default action. */
      if (reg->contact[0].action == L_unspecified) {
        reg->contact[0].action = config.defaultreg;
      }

      if (reg->contact[0].action == L_redirect) {
        return &redirect_policy;
      }
      else if (reg->contact[0].action == L_proxy) {
        return &proxy_policy;
      }
      else {
        /* Shouldn't happen */
        debug("find_primary_policy", "3xx status, reg 0 action = %d\n",
              reg->contact[0].action);
        return NULL;
      }
    }
    else {
      return &failure_policy;
    }
  }
  else {
    return &failure_policy;
  }
} /* find_primary_policy */
