/*
* proxy.c - proxy a SIP request
*
* Proxy a request to the locations listed in 'reg', assumed to be sorted
* highest-priority first.  All locations are tried in batches.  We try
* the highest-ranked first, then the next batch, etc.  Thus, parallel
* execution happens only if a number of 'q' values are the same.  If
* some address does not succeed within a time-out period, we move on to
* the next batch.
*
* If the parent side CANCELs, the branches are signaled.  Each branch is
* then responsible for retransmitting CANCELs downstream.
*
* Copyright 1998-2001 by Columbia University; all rights reserved 
*/

#include <pthread.h>
#include <time.h>
#include <sys/time.h>
#include <errno.h>     /* ETIME */
#include <stdlib.h>    /* calloc() */
#include <string.h>    /* strerror(), strcmp() */
#include <unistd.h>    /* gettimeofday() */

#include "config.h"
#include "policy.h"
#include "strdupn.h"
#include "sysdep.h"

#include "sipd-policy.h"
#include "register.h"
#include "proxy-support.h"

struct proxy_info {
  enum {sending_requests, awaiting_responses, final_wait} state;
  int pos;
  int more;
  registration_t *reg;
};


/*
 * StartSomeRegBranches()
 * Start some branches based on the registration structure -- those
 * that are within "spread" of the first non-started branch.
 * Returns whether there are still more branches to start after this.
 */
static int StartSomeRegBranches(request_t *r, registration_t *reg, int *pos)
{
  double q;              /* current quality value */
  double spread = 0.05;  /* spread of things to do in parallel */
  int i;
  message_t *m;

  q = reg->contact[*pos].q;

  for (i = *pos; i < reg->contacts && q - reg->contact[i].q <= spread; i++) {
    m = GetRequestAsMessage(r);
    policy_proxy_original_request_to(r, &reg->contact[i].uri, m, NULL);
    MessageDelete(m);
  }
  *pos = i;
  return (i < reg->contacts);
} /* StartSomeRegBranches */


/*
 * NextProxyStep()
 * This is called when it's time to do the "next step" of proxying -- either
 * a timeout has expired, or all currently active proxy branches have
 * returned a final response.  If we have more branches we can still try to
 * start, and we haven't received a 200, do so; otherwise, pick and send the
 * best response. */
static void NextProxyStep(request_t *r)
{
  struct proxy_info *info = (struct proxy_info *)r->policy_info;

  switch (info->state) {
  case sending_requests:
    info->more = StartSomeRegBranches(r, info->reg, &info->pos);
    if (!info->more) {
      info->state = awaiting_responses;
    }
    policy_set_timeout(r, (double)config.timeout);
    break;
  case awaiting_responses:
    SendBestResponse(r);

    /* Once we send the response, go to final wait state. */
    info->state = final_wait;
    policy_set_timeout(r, (double)config.timeout);
    break;
  default:
    error("NextProxyStep", "called with proxy in inappropriate state %d",
          info->state);
    break;
  }
} /* NextProxyStep */


/*
 * proxy_init()
 * Called when a transaction being handled by proxy_policy is started
 * This assumes that we're only called if there's at least one address
 * which it's reasonable to proxy to.  XXX: check what happens if we
 * register with mailto addresses, etc.
 */
static int proxy_init(request_t *r, registration_t *reg, int reg_status)
{
  struct proxy_info *info;

  info = (struct proxy_info*) calloc(1, sizeof(struct proxy_info));
  if (info == NULL)
    return 1;
  r->policy_info = (void *)info;
  info->reg = reg;

  info->state = sending_requests;
  info->more = StartSomeRegBranches(r, reg, &info->pos);
  if (!info->more) {
    info->state = awaiting_responses;
  }
  policy_set_timeout(r, (double)config.timeout);

  return 0;
} /* proxy_init */


/*
 * proxy_handle_request()
 * Called when a follow-up request (ACK or CANCEL) comes in from
 * the client.
 * Dispatch the request as appropriate
 */
static int proxy_handle_request(request_t *r, message_t *m)
{
  struct proxy_info *info = (struct proxy_info *)r->policy_info;

  if (strcmp(m->p.method, "CANCEL") == 0) {
    /* 
     * We were canceled.  Cancel all the branches that aren't done, and
     * send back 487.
     */
    CancelUnfinishedBranches(r);

    if (info->state != final_wait) {
      policy_send_new_response(r, 487, NULL);
      info->state = final_wait;
      policy_set_timeout(r, (double)config.timeout);
    }

    m->handled = finished_handling;
  }
  else if (strcmp(m->p.method, "ACK") == 0) {
    if (!r->sent_200) {
      error("proxy_handle_request", "ACK arrived at unexpected time");
    }
    else {
      ForwardACKToBranches(r, m);
    }
    m->handled = finished_handling;
  }
  else {
    /* I don't think any other methods should be showing
     * up here. */
    error("proxy_handle_request",
          "Request with surprising method %s on message queue\n",
          m->p.method);
    m->handled = finished_handling;
  }

  return 0;
} /* proxy_handle_request */


/*
 * proxy_handle_response()
 * Called when a response comes in from some branch.
 * Dispatch the response as appropriate.
 */
static int proxy_handle_response(request_t *r, message_t *m, branch_t *b)
{
  struct proxy_info *info = (struct proxy_info *)r->policy_info;

  ProxyDispatchResponse(r, m);

  if (r->sent_200) {
    info->state = final_wait;
  }

  if (info->state < final_wait && r->branches_outstanding == 0) {
    NextProxyStep(r);
  }

  return 0;
} /* proxy_handle_response */


/*
 * proxy_timeout()
 * Called when a timeout expires for the proxy
 * If we're sending requests, call NextProxyStep; if we're
 * awaiting responses, just refresh the timeout (the clients will time out
 * and send up a pseudo-response); if we're in final wait, we're done.
 */
static int proxy_timeout(request_t *r)
{
  struct proxy_info *info = (struct proxy_info *)r->policy_info;

  switch (info->state) {
  case sending_requests:
    NextProxyStep(r);
    break;
  case awaiting_responses:
    policy_set_timeout(r, (double)config.timeout);
    break;
  case final_wait:
    policy_done(r);
    break;
  default:
    /* Shouldn't happen */
    return 1;
    break;
  }

  return 0;
}


/*
 * proxy_cleanup()
 * Called when cleaning up the proxy policy
 * Free proxy-specific memory
 */
static int proxy_cleanup(request_t *r)
{
  free(r->policy_info);
  return 0;
} /* proxy_cleanup */


/* The policy structure describing proxying */
policy_t proxy_policy =
{
  "proxy",
  proxy_init,
  proxy_handle_request,
  proxy_handle_response,
  proxy_timeout,
  proxy_cleanup
};
