/*
 * register.c - handle REGISTER request
 *
 * Copyright 1998-2001 by Columbia University.  All rights reserved.
 */

#include <time.h>
#include <stdio.h>
#include <stdlib.h>   /* qsort() */
#include <fcntl.h>
#include <string.h>   /* strdup() */
#include <sys/types.h>
#include <regex.h>    /* regexec() */
#include <pthread.h>
#include <errno.h>
#include <netinet/in.h> 
#include <arpa/inet.h>   /* inet_ntoa() */

#ifndef HAVE_TIMEGM
#include "timegm.h"
#endif

#include "dstring.h"
#include "uri.h"
#include "register.h"
#include "request-sip.h"
#include "config.h"
#include "strdupn.h"  /* free_valid() */
#include "http.h"
#include "log.h"
#include "sysdep.h"
#include "authenticateSIP.h"
#include "libsip.h"
#include "host2ip.h"
#include "dbaccess.h" /* Database APIs */

#include "querydb.h"
#include "canonicalize.h"
#include "sipd-support.h"


/* Internal function declarations */
static sip_contact_t contact_dup(const sip_contact_t *contact);
static int location_compare(const sip_contact_t *s1, const sip_contact_t *s2);
static char *contact_uri_str(const uri_t *uri);

/*
 * remove_expired()
 * Remove expired entries from the database. 
 * Update 'reg', removing these entries.
 * returns: 0 on success (even when nothing to remove).
 *         -1 when failed.
 * XXX: need some way to return an error string upstream.
 */
static int remove_expired(registration_t* reg, const char *user)
{
  int i;
  int retval = 0;
  int some_expired = 0;

  db_err db_ret;
  char *db_ret_str = NULL;

  time_t now = time(0);

  int new_contact_count = reg->contacts;

  for (i = 0; i < reg->contacts; i++) {
    /*
     * Check every registration.  If a row has expired, mark that
     * we need to call contact_delete_expired_contacts().
     */
    if (reg->contact[i].expires < now && reg->contact[i].expires != 0) {
      some_expired = 1;
      reg->contact[i].q = -1;
      new_contact_count--;
    }
  }

  if (some_expired) {
    if (contact_delete_expired_contacts(config.database, user, now,
                                        &db_ret, &db_ret_str) != DB_OK) {
      error("remove_expired",
            "contact_delete_expired_contacts(user=%s, time=%ld): %s\n", 
            user, (long)now, db_ret_str);
      retval = -1; /* error */
    }
  }

  if (new_contact_count < reg->contacts) {
    /* Some contacts have expired.  Reallocate or free the contact list. */
    if (new_contact_count == 0) {
      /* All the contacts have expired. */
      for (i = 0; i < reg->contacts; i++) {
        contact_free(&reg->contact[i]);
      }
      free_valid(reg->contact);
      reg->contact = NULL;
      reg->contacts = 0;
    }
    else {
      int count = 0;
      /* Some contacts have not expired */
      sip_contact_t *old_contacts = reg->contact;

      reg->contact = (sip_contact_t *) calloc(new_contact_count,
                                              sizeof(sip_contact_t));

      for (i = 0; i < reg->contacts; i++) {
        if (old_contacts[i].q < 0) {
          contact_free(&old_contacts[i]);
        }
        else {
          if (count >= new_contact_count) {
            error("remove_expired",
                  "very confused: count %d >= new_contact_count %d\n",
                  count, new_contact_count);
          }
          reg->contact[count] = old_contacts[i];
          count++;
        }
      }

      free_valid(old_contacts);

      reg->contacts = count;
    }
  }

  return retval;
} /* remove_expired */


/*
 * register_deserialize()
 * Convert a list of 'count' contact_table_entry*s (as returned by
 * contacts_get_by_user) to a registration_t, and fill in 'reg'.  Return
 * all, including expired entries.
 * If we copy a string over verbatim, copy the pointer and set it to NULL
 * in the contact_table_entry *, so freeing it does the appropriate thing.
 * Return false on error.
 */
static int register_deserialize(registration_t *reg,
                                contact_table_entry **list,
                                int count)
{
  int i;

  memset((void *)reg, 0, sizeof(registration_t));

  /* XXX: previous code allocated reg.contacts+1 here, with the comment
   * "Extra one is allocated to take care of the expired entry."  What does
   * this mean, and is it still needed? */
  reg->contact = calloc(count, sizeof(sip_contact_t));

  if (reg->contact == NULL) {
    return 0;
  }

  reg->contacts = count;

  for (i = 0; i < count; i++) {
    if (list[i] == NULL) {
      error("register_deserialize", "very confused: list[%d] == NULL\n",
            i);
      memset(&reg->contact[i], 0, sizeof(reg->contact[i]));
      continue;
    }

    reg->contact[i].expires = list[i]->expires;

    switch (list[i]->action) {
    case contact_sip_action_Proxy:
      reg->contact[i].action = L_proxy;
      break;

    case contact_sip_action_Redirect:
      reg->contact[i].action = L_redirect;
      break;

    case contact_sip_action_NULL:
    default:
      reg->contact[i].action = L_unspecified;
      break;
    }

    reg->contact[i].last_modified = list[i]->last_modified;

    reg->contact[i].q = list[i]->q;
      
    reg->contact[i].uri = URI_Parse(list[i]->contact);

    reg->contact[i].nameaddr = SIP_BuildNameAddr(list[i]->display_name,
                                                &reg->contact[i].uri);

    reg->contact[i].methods = list[i]->sip_methods;
    list[i]->sip_methods = NULL;
  }

  /*
   * Sort the entries according to their q values. 
   * (The forking logic assumes this.)
   */
  qsort(reg->contact, reg->contacts, sizeof(sip_contact_t),
        (int(*)(const void *, const void *))location_compare);


  return 1;
} /* register_deserialize */


/* 
 * register_serialize()
 * Translate a sip_contact_t* to a contact_table_entry*.
 * Return NULL on failure.
 */
static contact_table_entry* register_serialize(const sip_contact_t *contact,
                                               const char* userid)
{
  contact_table_entry *entry;
  DString name, addr;

  entry = (contact_table_entry *) calloc(1, sizeof(contact_table_entry));
  if (entry == NULL) {
    return NULL;
  }

  strassign(&entry->user, userid);
  entry->contact = contact_uri_str(&contact->uri);

  entry->expires = contact->expires;

  entry->q = contact->q;

  switch(contact->action) {
  case L_proxy:
    entry->action = contact_sip_action_Proxy;
    break;
  case L_redirect:
    entry->action = contact_sip_action_Redirect;
    break;
  case L_reject:
  case L_unspecified:
  default:
    entry->action = contact_sip_action_NULL;
    break;
  }

  /* Last_modified is handled automatically by the database.  Don't set
   * it. */

  /* Extract the display-name from contact->nameaddr */
  DStringInit(&name);
  DStringInit(&addr);
  if (SIP_NameAddr(contact->nameaddr, &name, &addr) &&
      DStringLength(&name) > 0) {
    entry->display_name = DStringValue(&name);
  }
  else {
    entry->display_name = NULL;
    DStringFree(&name);
  }
  DStringFree(&addr);

  strassign(&entry->sip_methods, contact->methods);

  return entry;
} /* register_serialize */


/*
 * register_find()
 * Check if 'userid' is registered -- i.e., either has entries in the
 * contacts table, or a script.  If so, fill in 'reg' and return true;
 * otherwise, fill in an empty registration value, and return false.
 * If any registrations have expired, call remove_expired to purge them
 * from reg and from the database.
 * XXX: should the latter be done off-line?
 * XXX: need some way to pass error status strings upstream.
 */
int register_find(const char *userid, registration_t *reg)
{
  int retval = 0;    /* return value */
  contact_table_entry** contacts; /* contact entries from database */
  int contact_count; /* # of contacts returned by database */
  db_err db_ret;
  char *db_ret_str;

  contacts = contacts_get_by_user(config.database, userid, &contact_count,
                                  &db_ret, &db_ret_str);

  if (contacts == NULL) {
    if (db_ret != DB_NOTFOUND) {
      error("register_find", "contacts_get_by_user(user=%s): %s",
            userid, db_ret_str);
    }
    free_valid(db_ret_str);
    memset(reg, 0, sizeof(registration_t));
    return 0;
  }

  if (!register_deserialize(reg, contacts, contact_count)) {
    return 0;
  }

  free_contact_table_entries(contacts);

  /* Expunge any expired entries */
  if (remove_expired(reg, userid) < 0) {
    register_free(reg);
    return 0;
  }

  /* If we have some remaining contacts, set the return value to true. */
  retval = (reg->contacts > 0);

  /* Get the user's scriptset. */
  if (!GetScriptset(config.ScriptBase, userid,
                    &reg->scriptset)) {
    debug("register_find", "GetScriptSet(%s, %s)", config.ScriptBase,
          userid);
  }
  else if (reg->scriptset.count > 0) {
    retval = 1;
  }

  return retval;
} /* register_find */


/*
 * location_compare()
 * Compare two sip_contact_t's.  Used in qsort().
 */
static int location_compare(const sip_contact_t *s1, const sip_contact_t *s2)
{
  if (s1->q < s2->q) return 1;
  else if (s1->q > s2->q) return -1;
  else return 0;
} /* location_compare */


/*
 * script_unmodified_since()
 * If the request specified a modification date, make sure that the script
 * of the given type hasn't been changed since that date.
 * On failure, return 0 and set the request's response status.
 */
static int script_unmodified_since(request_t *r, registration_t *reg,
                                   char *disp_str)
{
  parser_header_t *if_unmodified_since;
  time_t mod_time, unmod_time;
  struct tm unmod_tm;

  if_unmodified_since = RFC822_Header(r->p.headers, "If-Unmodified-Since");
  if (if_unmodified_since == NULL) {
    /* No such header field; request is fine. */
    return 1;
  }

  mod_time = GetScriptModificationTime(&reg->scriptset, disp_str);
  if (mod_time == -1) {
    /* No such script in set; request is fine. */
    return 1;
  }
  
  if (Parse_HTTP_Date(if_unmodified_since->elements->value, &unmod_tm) == NULL
      || (unmod_time = timegm(&unmod_tm)) == -1) {
    ResponseSetStatus(r, 400, "Invalid If-Unmodified-Since");
    return 0;
  }

  if (unmod_time < mod_time) {
    ResponseSetStatus(r, 412, "Precondition Failed: Script Modified");
    return 0;
  }

  return 1;
} /* script_unmodified_since */


/*
 * do_script_upload()
 * Check if this REGISTER request is trying to upload a script;
 * and if so, do it.
 * Return false (and set the response status) if anything goes wrong.
 * This should be called before updating the registration database,
 * because a failure here means that the contact addresses shouldn't be
 * updated.
 * TODO: Support MIME multipart.
 * TODO: hooks for script validation
 */
static int do_script_upload(request_t *r, registration_t *reg)
{
  parser_header_t *content_disposition; /* header in request */
  char *action, *mod_date;
  struct tm mod_tm;
  time_t mod_time = -1;

  content_disposition = RFC822_Header(r->p.headers, "Content-Disposition");

  if (content_disposition == NULL) {
    /* No header upload requested -- this is fine. */
    return 1;
  }

  action = RFC822_Attribute(content_disposition, 0, "action");
  mod_date = RFC822_Attribute(content_disposition, 0, "modification-date");

  if (action == NULL) {
    /* No action requested -- this is fine, do nothing. */
    return 1;
  }

  /* Check If-Unmodified-Since */
  if (!script_unmodified_since(r, reg,
                               content_disposition->elements->value)) {
    /* Sub-function will set response status */
    return 0;
  }

  /* Check the modification-date parameter */
  if (mod_date != NULL) {
    if (Parse_HTTP_Date(mod_date, &mod_tm) == NULL ||
        (mod_time = timegm(&mod_tm)) == -1) {
      ResponseSetStatus(r, 400, "Invalid modification-date");
      return 0;
    }
  }    
  
  /* Perform the action */
  if (strcasecmp(action, "store") == 0) {
    if (!AddScriptToSet(&reg->scriptset,
                        content_disposition->elements->value,
                        r->request_body.content_type,
                        r->request_body.content,
                        r->request_body.content_length,
                        mod_time)) {
      ResponseSetStatus(r, 500, "Script storage failed");
      return 0;
    }
    return 1;
  }
  else if (strcasecmp(action, "remove") == 0) {
    if (r->request_body.content_length > 0) {
      /* Non-zero length can only be added */
      ResponseSetStatus(r, 400, "Content-Dispostion remove action "
                        "with non-empty body");
      return 0;
    }
    if (!DeleteScriptFromSet(&reg->scriptset,
                             content_disposition->elements->value)) {
      ResponseSetStatus(r, 500, "Script removal failed");
      return 0;
    }
    return 1;
  }
  else {
    /* Neither "store" nor "remove" */
    /* Should this be some more specific error status? */
    ResponseSetStatus(r, 400, "Unknown Content-Disposition action");
    return 0;
  }
} /* do_script_upload */


/*
 * do_script_download()
 * Insert the user's currently registered scripts, if any, into the body of
 * the REGISTER response.
 * TODO: MIME multipart.  We currently only send the first script.
 * XXX TODO: honor Accept/Accept-Disposition headers in request.
 * XXX TODO: insert Accept/Accept-Disposition headers in response
 */
static void do_script_download(request_t *r, registration_t *reg)
{ 
  DString script, mod;
  struct tm mod_tm;
  DStringInit(&script);
  DStringInit(&mod);

  if (reg->scriptset.count > 0
      && ReadScriptToDString(&reg->scriptset.scripts[0], &script) >= 0) {
    gmtime_r(&reg->scriptset.scripts[0].lastmod, &mod_tm);
    DStringStrftime(&mod, "%a, %d %b %Y %H:%M:%S GMT", &mod_tm);

    HeaderPrintf(&r->headers, 1, "Content-Disposition",
                 "%s; modification-date=\"%s\"",
                 reg->scriptset.scripts[0].disp_str,
                 DStringValue(&mod));
    ResponseSetBody(r, DStringValue(&script),
                    reg->scriptset.scripts[0].content_type,
                    DStringLength(&script));  
  }

  DStringFree(&script);
  DStringFree(&mod);
} /* do_script_download */


/*
 * SetupAndValidateTarget()
 * Set r->destination_user, based on 'uri'.  Validate that
 * r->destination_user is authorized to perform a REGISTER request.  If not,
 * or if the user doesn't exist, set response status and return false.
 */
static int SetupAndValidateTarget(request_t *r, const uri_t *destination)
{
  db_err db_ret;
  const char* method = "REGISTER";

  r->destination_user = CanonicalizeAndLookupURI(r, destination, 0,
                                                 "To URI", "registration",
                                                 r->source_user);
  if (r->destination_user == NULL) {
    return 0;
  }
                                                 
  db_ret = method_allowed_for_user(method, r->destination_user);

  if (db_ret != DB_OK) {
    request_set_dberr_status(r, db_ret, r->destination_user->user,
                             405, NULL);
    /* XXX: add Allow: headers for DB_NO */
    return 0;
  }

  return 1;
} /* SetupAndValidateTarget */


/*
 * SetupAndValidateSource()
 * Set r->source_user, based on 'source', and validate that it is authorized
 * to register for r->destination_user.  If not, or if the user doesn't
 * exist, return false and set the response status.
 */
static int SetupAndValidateSource(request_t *r, const uri_t *source)
{
  db_err db_ret;
  DString reason;
  
  r->source_user = CanonicalizeAndLookupURI(r, source, 1,
                                            "From URI", "registration",
                                            r->destination_user);

  if (r->source_user == NULL) {
    return 0;
  }

  db_ret = may_register_for(r->source_user, r->destination_user);

  if (db_ret_invalid(db_ret)) {
    errno = 0;
    error("ValidateThirdPartyReg",
          "Request 0x%p: may_register_for(%s, %s) returned %d\n",
          r, r->source_user->user, r->destination_user->user, db_ret);
  }

  if (db_ret != DB_OK) {
    DStringInit(&reason);

    DStringPrintf(&reason, "Registration for %s forbidden from %s",
                  r->destination_user->user, r->source_user->user);
    request_set_dberr_status(r, db_ret, r->destination_user->user,
                             403, DStringValue(&reason));

    DStringFree(&reason);
    return 0;
  }

  return 1;
} /* SetupAndValidateSource */


/*
 * ValidateContactsAndSetExpirations()
 * Validate r->contacts -- make sure it's not asking for anything
 * paradoxical, like conflicting action parameters -- and set each contact's
 * expires header if it wasn't set individually.  Also reduce expirations
 * for 'sip' contacts to be at most config.expires.
 * Return false, and set r's response status, on error.
 */
static int ValidateContactsAndSetExpirations(request_t *r,
                                             int contact_count,
                                             sip_action_t current_action)
{
  int i;
  time_t now = time(0);

  if (r->expires == 0) {
    /* Set the default expiry time */
    r->expires = now + config.expires;
  }

  for (i = 0; i < contact_count; i++) {
    /* If the contact doesn't have its own expires parameter, carry over
     * the global expires header field. */
    if (r->contact[i].expires == 0) {
      r->contact[i].expires = r->expires;
    }

    if (strcmp(r->contact[i].uri.scheme, "*") == 0) {
      /* Validate Contact:* for sensibleness */
      if (i > 0) {
        ResponseSetStatus(r, 400,
                          "Contact:* must be the only Contact header"); 
        return 0;
      }
      else if (r->contact[i].expires != 1) {
        ResponseSetStatus(r, 400,
                          "Contact:* is allowed only with Expires: 0"); 
        return 0;
      }
    }
    else if (strcmp(r->contact[i].uri.scheme, "sip") == 0) {
      /* SIP contacts have a maximum expiration time. */
      if (r->contact[i].expires > now + config.maxexpires) {
        r->contact[i].expires = now + config.maxexpires;
      }
    }

    /* If action of registration doesn't agree with current one, fail. */
    if (current_action != L_unspecified) {
      if (r->contact[i].action != L_unspecified &&
          r->contact[i].action != current_action) {
        ResponseSetStatus(r, 409, "Mismatch in action parameter");
        return 0;
      }
    }
    else {
      if (r->contact[i].action != L_unspecified) {
        current_action = r->contact[i].action;
      }
    }
    /*
     * If unspecified, inherit action from current registration/previous
     * contacts.  If no current registration, set to the default
     * reg. action. 
     */
    if (r->contact[i].action == L_unspecified) {
      r->contact[i].action = current_action != L_unspecified ? current_action
        : config.defaultreg;
    }
  }

  return 1;
} /* ValidateContactsAndSetExpirations */


/*
 * DeleteAllRegs()
 * Delete all registrations for this user -- we got a Contact:*;expires=0.
 * Delete both from the database, and from the local 'reg' structure.
 * Return false on failure.
 */
static int DeleteAllRegs(request_t *r, registration_t *reg)
{
  db_err db_ret;
  char *db_ret_str;
  const char *userid = r->destination_user->user;

  /* Delete the registrations from the database. */
  if (contact_delete_contacts(config.database, userid,
                              &db_ret, &db_ret_str) != DB_OK) {
    DString reason;
    DStringInit(&reason);

    errno = 0;
    error("DeleteAllRegs",
          "r=0x%p: contact_delete_contacts(user=%s): %s\n",
          r, userid, db_ret_str);
    DStringPrintf(&reason, "Couldn't delete contacts from database: %s",
                  db_ret_str);
    ResponseSetStatus(r, 500, DStringValue(&reason));

    DStringFree(&reason);
    free_valid(db_ret_str);
    return 0;
  }

  /* Reset reg to 0 contacts. */
  /* Reg shouldn't yet have a script set at this point. */
  register_free(reg);

  return 1;
} /* DeleteAllRegs */


/*
 * DeleteRegByURI()
 * Delete the registration for the contact URI 'uri' for this user --
 * we got a Contact:...;expires=0.
 * Delete it both from the database, and from the local 'reg' structure.
 * Return false on failure.
 */
static int DeleteRegByURI(request_t *r, registration_t *reg,
                          const uri_t *uri)
{
  db_err db_ret;
  char *db_ret_str;
  const char *userid = r->destination_user->user;
  int i;

  char *contact_str;

  /* Delete the registration from the database. */
  contact_str = contact_uri_str(uri);
  if (contact_delete_contact(config.database, userid, contact_str,
                             &db_ret, &db_ret_str) != DB_OK) {
    DString reason;
    DStringInit(&reason);

    errno = 0;
    error("DeleteRegByURI",
          "r=0x%p: contact_delete_contact(user=%s, contact=%s): %s\n",
          r, userid, contact_str, db_ret_str);
    DStringPrintf(&reason, "Couldn't delete contact from database: %s",
                  db_ret_str);
    ResponseSetStatus(r, 500, DStringValue(&reason));

    free_valid(contact_str);
    DStringFree(&reason);
    free_valid(db_ret_str);
    return 0;
  }
  free_valid(contact_str);

  /* Delete the registration from the local reg structure. */
  for (i = 0; i < reg->contacts; i++) {
    if (URI_Equal_With_Defaults(*uri, reg->contact[i].uri, sip_uri_defaults)) {
      contact_free(&reg->contact[i]);
      if (i + 1 < reg->contacts) {
        /* Slide down the remaining contacts */
        memmove(&reg->contact[i], &reg->contact[i+1],
                (reg->contacts - (i + 1)) * sizeof(sip_contact_t));
      }
      /* Reallocate a smaller array */
      reg->contacts--;
      reg->contact = realloc(reg->contact,
                             reg->contacts * sizeof(sip_contact_t));
      break;
    }
  }

  return 1;
} /* DeleteRegByURI */


/*
 * InsertReg()
 * Insert the registration described by 'contact' for this user.
 * Insert it both into the database, and into the local 'reg' structure.
 * Return false on failure.
 */
static int InsertReg(request_t *r, registration_t *reg,
                     const sip_contact_t *contact)
{
  db_err db_ret;
  char *db_ret_str;
  const char *userid = r->destination_user->user;
  contact_table_entry *entry;
  sip_contact_t *new_contacts;
  int i;
  
  entry = register_serialize(contact, userid);

  if (entry == NULL) {
    return 0;
  }
  debug("InsertReg",
          "r=0x%p: contact_insert_contact({user=%s, contact=%s})\n",
          r, entry->user, entry->contact);

  if (contact_insert_contact(config.database, entry,
                             &db_ret, &db_ret_str) != DB_OK) {
    DString reason;
    DStringInit(&reason);

    errno = 0;
    error("InsertReg",
          "r=0x%p: contact_insert_contact({user=%s, contact=%s}): %s\n",
          r, entry->user, entry->contact, db_ret_str);
    DStringPrintf(&reason, "Couldn't insert contact into database: %s",
                  db_ret_str);
    ResponseSetStatus(r, 500, DStringValue(&reason));
    
    free_contact_table_entry(entry);
    DStringFree(&reason);
    free_valid(db_ret_str);
    return 0;
  }
  free_contact_table_entry(entry);

  for (i = 0; i < reg->contacts; i++) {
    if (URI_Equal_With_Defaults(contact->uri, reg->contact[i].uri,
                                sip_uri_defaults)) {
      contact_free(&reg->contact[i]);
      /* Insert the new contact here */
      break;
    }
  }
  if (i == reg->contacts) {
    /* Not found in array.  Allocate a new slot for it. */
    new_contacts = realloc(reg->contact,
                           (reg->contacts + 1) * sizeof(sip_contact_t));
    if (new_contacts == NULL) {
      ResponseSetStatus(r, 500, NULL);
      return 0;
    }

    reg->contact = new_contacts;
    reg->contacts++;
  }

  reg->contact[i] = contact_dup(contact);

  return 1;
} /* InsertReg */


/*
 * MergeContactsAndRegistration()
 * For every contact listed in r->contacts, perform the corresponding
 * database operation, and edit 'reg'.
 * Return false, and set r's ResponseStatus appropriately, if anything
 * goes wrong.
 */
static int MergeContactsAndRegistration(request_t *r, registration_t *reg)
{
  int contact_count;
  int i;

  if (r->contact == NULL) {
    /* Nothing to do */
    return 1;
  }

  /* Count the contacts in the request. */
  for (i = 0; r->contact[i].uri.scheme != NULL; i++) {
    /* Do nothing */;
  }
  contact_count = i;

  if (!ValidateContactsAndSetExpirations(r, contact_count,
                                         reg->contacts > 0
                                         ? reg->contact[0].action
                                         : L_unspecified)) {
    return 0;
  }

  /* For every contact in the list, update the database, and 'reg',
   * appropriately. */
  for (i = 0; i < contact_count; i++) {
    if (strcmp(r->contact[i].uri.scheme, "*") == 0) {
      if (i != 0 || r->contact[i].expires != 1) {
        /* Shouldn't happen -- should have been caught by
           ValidateContactsAndSetExpirations. */
        error("MergeContactsAndRegistration",
              "r = 0x%p: Contact:* with invalid parameters "
              "(i = %d, expires = %ld) got by ValidateContacts\n",
              r, i, (long)r->contact[i].expires);
        return 0;
      }
      /* Contact: *; expires = 0.  Delete all contacts. */
      if (!DeleteAllRegs(r, reg)) {
        return 0;
      }
    }
    else if (r->contact[i].expires == 1) {
      /* expires = 0.  Delete this one contact. */
      if (!DeleteRegByURI(r, reg, &r->contact[i].uri)) {
        return 0;
      }
    }
    else {
      /* Ordinary contact.  Insert it into the database. */
      if (!InsertReg(r, reg, &r->contact[i])) {
        return 0;
      }
    }
  }

  /*
   * Re-sort the entries according to their q values. 
   * (InsertReg may have disrupted the order.)
   */
  qsort(reg->contact, reg->contacts, sizeof(sip_contact_t),
        (int(*)(const void *, const void *))location_compare);

  return 1;
} /* MergeContactsAndRegistration */


/*
 * InsertCurrentContacts()
 * Insert the user's current contacts, based on 'reg', into r's response.
 * Set a response Expires as appropriate: earliest expiration of any contact.
 */
static void InsertCurrentContacts(request_t *r, registration_t *reg)
{
  int i;
  time_t expire_time = 0;  /* Default is no expiration */
  struct tm tm;

  /* Return list of current contacts, and compute earliest expiration. */
  for (i = 0; i < reg->contacts && reg->contact[i].uri.scheme != NULL; i++) {
    DString expires;

    if (reg->contact[i].q < 0) {
      error("InsertCurrentContacts",
            "r=0x%p: contact %d, q = %f was not purged\n",
            r, i, reg->contact[i].q);
      continue;
    }
    
    DStringInit(&expires);
      
    if (reg->contact[i].expires != 0 &&
        (expire_time == 0 || reg->contact[i].expires < expire_time)) {
      expire_time = reg->contact[i].expires;
    }
    gmtime_r(&reg->contact[i].expires, &tm);
    DStringStrftime(&expires, "%a, %d %b %Y %H:%M:%S GMT", &tm);
      
    if (reg->contact[i].nameaddr == NULL) {
      reg->contact[i].nameaddr = SIP_BuildNameAddr(NULL, &reg->contact[i].uri);
    }
      
    HeaderPrintf(&r->headers, 0, "Contact", 
                 "%s; expires=\"%s\"; action=%s; q=%.2f",
                 reg->contact[i].nameaddr,
                 DStringValue(&expires),
                 reg->contact[i].action == L_proxy ? "proxy" : "redirect",
                 reg->contact[i].q);
    
    DStringFree(&expires);
  }
  if (expire_time > 0) {
    HeaderCftime(&r->headers, "Expires", 
                 "%a, %d %b %Y %H:%M:%S GMT", expire_time, 0);
  }
} /* InsertCurrentContacts */


/*
 * REGISTER()
 * Handle a REGISTER request, adding and deleting user entries to/from the
 * database.
 */
void REGISTER(request_t *r)
{
  registration_t reg;

  /* Zero reg, so we can safely call register_free() below. */
  memset(&reg, 0, sizeof(reg));

  /* XXX: I don't like this being here -- it's confusing. */
  ResponseSetStatus(r, 200, NULL); 
  ResponseBuildDefault(r); 

  /* Set up r->destination_user, and check if it's allowed to register */
  if (!SetupAndValidateTarget(r, &r->to.uri)) {
    goto done;
  }

  /* Can 'From' register for 'To'? */
  if (!SetupAndValidateSource(r, &r->from.uri)) {
    goto done;
  }

  /* Make sure the request contains authentication information for
   * from_userid. */
  if (AuthenticateSIP(r, 0, 0, 0) != auth_success) {
    /* The error status appropriate headers (WWW-Auth..) are already set */
    goto done;
  }

  /* Find existing registrations for this use. */
  register_find(r->destination_user->user, &reg);

  /* Perform script uploading. */
  if (!do_script_upload(r, &reg)) {
    debug("REGISTER", "do_script_upload failed\n");
    goto done;
  }

  if (!MergeContactsAndRegistration(r, &reg)) {
    goto done;
  }

  /* See if any subscriptions exist for the user - if yes, send out notify */
  if (num_subscriptions(config.database, r->destination_user->user, 
			NULL, NULL) > 0) {
    send_out_notify_requests(r->expires, r->destination_user->user);
  }
  
  /* Add contacts from 'reg' to response */
  InsertCurrentContacts(r, &reg);

  /* Add body from currently-registered scripts to response */
  do_script_download(r, &reg);

done:
  ResponseSend(r, 0, 1); /* Will also log the request */

  register_free(&reg);

  return;
} /* REGISTER */


/*
 * contact_dup()
 * Duplicate a sip_contact_t
 */
static sip_contact_t contact_dup(const sip_contact_t *contact)
{
  sip_contact_t new_contact;

  memset(&new_contact, 0, sizeof(sip_contact_t));

  new_contact.q        = contact->q;
  new_contact.expires  = contact->expires;
  new_contact.action   = contact->action;
  new_contact.uri      = URI_Copy(contact->uri);
  strassign(&new_contact.nameaddr, contact->nameaddr);
  strassign(&new_contact.comment, contact->comment);
  strassign(&new_contact.methods, contact->methods);

  return new_contact;
} /* contact_dup */


/*
 * contact_free()
 * Free a sip_contact_t
 */
void contact_free(sip_contact_t *contact)
{
  URI_Free(&contact->uri);
  free_valid(contact->nameaddr);
  free_valid(contact->comment);
  free_valid(contact->methods);
  contact->nameaddr = NULL;
  contact->comment = NULL;
  contact->methods = NULL;
} /* contact_free */


/*
 * register_free()
 * Free storage for a registration.
 */
void register_free(registration_t *reg)
{
  int i;

  for (i = 0; i < reg->contacts; i++) {
    contact_free(&reg->contact[i]);
  }
  free_valid(reg->contact);
  reg->contact = 0;
  reg->contacts = 0;

  FreeScriptset(&reg->scriptset);
} /* register_free */


/*
 * uri_in_registration()
 * Return true if the URI 'uri' is present in the contact list of the
 * registration 'reg'.  Compare with SIP URI defaults.
 */
static int uri_in_registration(uri_t uri, registration_t *reg)
{
  int i;

  for (i = 0; i < reg->contacts; i++) {
    if (URI_Equal_With_Defaults(uri, reg->contact[i].uri,
                                sip_uri_defaults)) {
      return 1;
    }
  }
  return 0;
} /* uri_in_registration */


/*
 * contact_uri_str()
 * Given a URI 'uri', return a string representation of that URI, suitable
 * for passing to dbaccess's contact_* tables.
 * The caller must free the resulting string.
 * This calls URI_Canonicalize.
 */
static char *contact_uri_str(const uri_t *uri)
{
  uri_t canon;
  DString canon_str;
    
  canon = URI_Copy(*uri);
  URI_Canonicalize(&canon, sip_uri_defaults);

  DStringInit(&canon_str);
  URI_SPrint(&canon, &canon_str);
  
  URI_Free(&canon);
  
  return DStringValue(&canon_str);
} /* contact_uri_str */


/*
 * registration_append()
 * Append 'uri' to list of contacts in 'reg', with the
 * 'displayname' in the nameaddr form.  If parse && config.HelpResolve,
 * convert host to IP address for SIP URIs. 
 * XXX: Do servers() lookup to get IP addresses; use maddrs rather than
 * re-writing hostnames.  (If we want to keep HelpResolve at all.)
 */
void registration_append(registration_t *reg, uri_t orig_uri,
                         char *displayname, boolean parse,
                         sip_action_t action)
{
  char *nameaddr;
  uri_t uri;

  /* Don't add if name is already on list. */
  if (uri_in_registration(orig_uri, reg)) {
    return;
  }

  uri = URI_Copy(orig_uri);

  if (parse && config.HelpResolve &&
      uri.scheme != NULL && strcasecmp(uri.scheme, "sip") == 0) {
    /* Convert the hostname to an IP address */
    struct in_addr host_addr;

    host_addr = host2ip(uri.host);

    if (host_addr.s_addr != INADDR_ANY) {
      strassign(&uri.host, inet_ntoa(host_addr));
    }
  }

  nameaddr = SIP_BuildNameAddr(displayname, &uri);

  if (!reg->contact) {
    reg->contact = calloc(1, sizeof(sip_contact_t));
  }
  else {
    reg->contact = realloc(reg->contact,
       (reg->contacts+1) * sizeof(sip_contact_t));
    memset(&reg->contact[reg->contacts], 0, sizeof(sip_contact_t));
  }

  reg->contact[reg->contacts].action   = action;
  reg->contact[reg->contacts].q        = 0.0;
  reg->contact[reg->contacts].expires  = 0;
  reg->contact[reg->contacts].uri      = uri;
  reg->contact[reg->contacts].nameaddr = nameaddr;
  reg->contacts++;

  return;
} /* registration_append */


