/*
* route.c - send a SIP request to the address in its "route" header.
*
* This is a simplified version of proxy.c, except that there is only a
* single branch to be tried, so all the "try the next location" logic goes
* away.  The rest of the logic is the same as in that file. (q.v.)
*
* Added bis03 semantics.
*
* Copyright 1998-2000 by Columbia University; all rights reserved */

#include <pthread.h>
#include <time.h>
#include <sys/time.h>
#include <errno.h>     /* ETIME */
#include <stdlib.h>    /* calloc() */
#include <string.h>    /* strerror(), strcmp() */
#include <unistd.h>    /* gettimeofday() */

#include "config.h"
#include "sysdep.h"
#include "strdupn.h"

#include "sipd-policy.h"
#include "proxy-support.h"
#include "rr_support.h"

/* route_init()
 * Called when a transaction being handled by route_policy is started
 * This assumes that we're only called if there's at least one address
 * which it's reasonable to route to.
 * XXX: check what happens if someone tries to route us to something absurd?
 * XXX: add some check to stop people from relaying DoS attacks through us.
 * (Needs spec enhancement, I think)
 */
static int route_init(request_t *r, registration_t *reg, int reg_status)
{
  parser_header_t *h;         /* A "Route" header */
  DString name, addr;
  uri_t uri;
  message_t *req_copy;
  int flag = 1;
  char *maddr;

  /* If route_policy was invoked then h is non-null */
  h = RFC822_Header(r->p.headers, "Route");

  /*
   * Algorithm (or what goes on below):
   *  ---If the top most route header has an maddr parameter and it
   *     does not match us, we are safe. Use the uri created from the
   *     top most route header, and that is the next hop.
   *  ---If the top most route header's maddr matches us, 
   *     then
   *        pop it if there are more route headers. the route
   *        header which is topmost after popping decides the next hop.
   *        if there are no more route headers, nullify the maddr
   *        parameter as it points to us, and route it to the host present
   *        in the "host" param.
   *
   * Why this ?
   *  ---Existing clients dont seem to implement strict bis03 way of
   *     inserting route header by computing Route Set. In this circumstance,
   *     the afore said behavior gives robustness.
   */
  while (h->elements && flag) {
    
    DStringInit(&name);
    DStringInit(&addr);

    if (!h->elements->value ||
      !SIP_NameAddr(h->elements->value, &name, &addr)) {
      DStringFree(&name);
      DStringFree(&addr);
      policy_send_new_response(r, 400, "Malformed Route header");
      return 0;
    }

    uri = URI_Parse(DStringValue(&addr));

    if ((maddr = URI_Parameter(uri, "maddr")) != NULL && 
      !strcmp(maddr, config.ProxyName)) {
  
      /* the top most Route header is pointing to us for whatever
       * reason. pop it, and go to the next header
       */
      parser_element_t* to_del = h->elements;
      
      /* if there is no other route header live with this - but delete the
       * maddr parameter as it points to us
      */
      if (to_del->next == NULL) {
        /* XXX: URI_SetParameter only sets value to null. it still
	 * results in ;maddr appended to output. Todo: Delete it.
	 */
	URI_SetParameter(&uri, "maddr", NULL);
        flag = 0;
        continue;
      }

      h->elements = to_del->next;
      if (!h->elements) 
        h->last_element = NULL;
      to_del->next = NULL;

      if (to_del->value && to_del->value != to_del->element)
        free(to_del->value);
      if (to_del->element) 
        free(to_del->element);

      RFC822_ParameterFree(to_del->parameters);
      
      continue;
    }
    else
      flag = 0;
  }
  
  if (!h->elements) {
    policy_send_new_response(r, 400, "Malformed Route header");
    return 0;
  }
  
  req_copy = GetRequestAsMessage(r);
  policy_proxy_original_request_to(r, &uri, req_copy, NULL);
  MessageDelete(req_copy);
  policy_set_timeout(r, (double)config.timeout);
  
  DStringFree(&name);
  DStringFree(&addr);
  URI_Free(&uri);

  return 0;
} /* route_init */


/* route_handle_request()
 * Called when a follow-up request (ACK or CANCEL) comes in from
 * the client.
 * Dispatch the request as appropriate
 */
static int route_handle_request(request_t *r, message_t *m)
{
  if (strcmp(m->p.method, "CANCEL") == 0) {
    /* 
     * We were canceled.  Cancel all the branches that aren't done, and
     * send back 487.
     */
    CancelUnfinishedBranches(r);

    if (!r->sent_final) {
      policy_send_new_response(r, 487, NULL);
      policy_set_timeout(r, (double)config.timeout);
    }

    m->handled = finished_handling;
  }
  else if (strcmp(m->p.method, "ACK") == 0) {
    if (!r->sent_200) {
      error("route_handle_request", "ACK arrived at unexpected time");
    }
    else {
      ForwardACKToBranches(r, m);
    }
    m->handled = finished_handling;
  }
  else {
    /* I don't think any other methods should be showing
     * up here. */
    error("route_handle_request",
          "Request with surprising method %s on message queue\n",
          m->p.method);
    m->handled = finished_handling;
  }

  return 0;
} /* route_handle_request */


/*
 * route_handle_response()
 * Called when a response comes in from some branch.
 * Dispatch the response as appropriate.
 */
static int route_handle_response(request_t *r, message_t *m, branch_t *b)
{
  ProxyDispatchResponse(r, m);

  if (!r->sent_200 && r->branches_outstanding == 0) {
    SendBestResponse(r);

    policy_set_timeout(r, (double)config.timeout);
  }

  return 0;
} /* route_handle_response */



/* route_timeout()
 * Called when a timeout expires for the route
 * If we're awaiting responses, just refresh the timeout (the clients will
 * time out and send up a pseudo-response); otherwise, we're done.
 */
static int route_timeout(request_t *r)
{
  if (!r->sent_final) {
    policy_set_timeout(r, (double)config.timeout);
  }
  else {
    policy_done(r);
  }

  return 0;
} /* route_timeout */


/* route_cleanup()
 * Called when cleaning up the route policy
 * Nothing to do
 */
static int route_cleanup(request_t *r)
{
  return 0;
} /* route_cleanup */


/* The policy structure describing routeing */
policy_t route_policy =
{
  "route",
  route_init,
  route_handle_request,
  route_handle_response,
  route_timeout,
  route_cleanup
};
