/*
 * script.c: handle registered scripts: add, retrieve, remove
 *
 * Copyright 2000-2001 Columbia University.  All rights reserved.
 *
 * This file contains all the logic for the server to handle scripts (and
 * other content) registered for a user.
 */

/* The script repository contains all the scripts for all the users.
 * It is structured as follows:
 *  - The root of the script repository is specified in config.ScriptRoot.
 *    - Within the repository, subdirectories are named by users, in
 *      the same format as is used for database keys (see register_make_key
 *      in register.c)
 *      - Within each user's directory, subdirectories are named for each
 *        Content-Disposition specified for a script, canonicalized to
 *        all lower case.
 *        - Within the directory, several files appear:
 *          SCRIPT: contains the script.
 *          CONTENT-TYPE: contains the MIME media type of the script.
 *          tmpscXXXXXX: contains temporary copies of the script.
 *
 *        If the CONTENT-TYPE file is missing, "application/octet-stream"
 *          is assumed.  This shouldn't happen if the directory was
 *          created by this code, but this repository structure was
 *          designed so that scripts can be added by hand.
 *
 * It is possible that one transaction (e.g. a REGISTER) modifies
 * the repository at the same time that another one (e.g. INVITE handling)
 * wants to query it.  To avoid this, temporary copies of the script are
 * created, using link() on platforms that support it (for efficiency).
 */

#ifdef __linux__
#define _GNU_SOURCE /* Needed for rwlock declarations from <pthread.h> */
#endif

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>  /* utimes */

#ifndef WIN32
#include <dirent.h>
#else
#include <sys/utime.h> /* struct utimebuf */
#endif

#include <pthread.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>



#include "sysdep.h"

#include "script.h"
#include "error.h"
#include "strdupn.h" /* free_valid() */
#include "config.h"

/* Internal static functions in this file */
static void ReadLockScriptDir(void);
static void WriteLockScriptDir(void);
static void UnlockScriptDir(void);
static int ReadScript(const char* script_dir, const char* disp_str,
                      script_t *script, char **real_file);
static char* CreateTempScript(const char *script_file, const char* script_dir);
static int ReadAndCopyScript(const char* script_dir, const char* disp_str,
                             script_t *script);
static void FreeScript(script_t *script);
static int unlink_if_exists(const char *path);
static int mkdir_if_missing(const char *path);


#ifdef HAVE_PTHREAD_RWLOCK_INIT
static pthread_rwlock_t script_lock = PTHREAD_RWLOCK_INITIALIZER;
#else
static pthread_mutex_t script_lock = PTHREAD_MUTEX_INITIALIZER;
#endif

/*
 * ReadLockScriptDir()
 * Lock the script directory for reading.  While you hold this lock, you are
 * allowed to read from the script directory, to create tempnam()'d 
 * files (provided you are clever about detecting race conditions with
 * tempnam: use either open() with O_CREAT|O_EXCL, or link()), and to delete
 * such of these files that you own.
 * XXX: should be more clever for the case when we haven't got Pthread
 * RW-locks.  Pthread mutex's probably block the system too much.
 */
static void ReadLockScriptDir(void)
{
#ifdef HAVE_PTHREAD_RWLOCK_INIT
  pthread_rwlock_rdlock(&script_lock);
#else
  pthread_mutex_lock(&script_lock);
#endif
} /* ReadLockScriptDir() */


/*
 * WriteLockScriptDir()
 * Lock the script directory for writing.  While you hold this lock, you are
 * allowed to modify the script directory however you like.  (You should,
 * however, leave it in a consistent state when you are done with it,
 * of course.)
 */
static void WriteLockScriptDir(void)
{
#ifdef HAVE_PTHREAD_RWLOCK_INIT
  pthread_rwlock_wrlock(&script_lock);
#else
  pthread_mutex_lock(&script_lock);
#endif
} /* WriteLockScriptDir() */


/*
 * UnlockScriptDir()
 * Unlock the script directory.
 */
static void UnlockScriptDir(void)
{
#ifdef HAVE_PTHREAD_RWLOCK_INIT
  pthread_rwlock_unlock(&script_lock);
#else
  pthread_mutex_unlock(&script_lock);
#endif
} /* UnlockScriptDir() */


/* Definitions of components of the script directory */
#define SCRIPT_FILE "SCRIPT"
#define CONTENT_TYPE_FILE "CONTENT-TYPE"
#define TMP_SCRIPT_PREFIX "tmpsc"

#define DEFAULT_CONTENT_TYPE "application/octet-stream"


/*
 * ReadScript()
 * Read a script directory and populate the script_t pointed to with the
 * resulting information.  Don't create the duplicate scriptfile; leave
 * script->filename NULL.
 * Return 0 on failure, non-zero on success.
 * You should hold the script dir lock (either reading or writing) before
 * calling this.
 */
static int ReadScript(const char* script_dir, const char* disp_str,
                      script_t *script, char **real_file)
{
  char script_file[PATH_MAX];
  char content_type_file[PATH_MAX];
  FILE *content_type_file_handle;
  char content_type[1024]; /* If a content-type is bigger than this,
                              someone's being obnoxious.  Disallow. */
  struct stat script_stat;
  int retval = 0;

  memset(script, 0, sizeof(script_t));

  if (real_file) *real_file = NULL;

  /* Construct pathnames */
  if (strlen(script_dir) + strlen(DIR_SEPARATOR) + strlen(SCRIPT_FILE)
      >= sizeof(script_file) ||
      strlen(script_dir) + strlen(DIR_SEPARATOR) + strlen(CONTENT_TYPE_FILE)
      >= sizeof(content_type_file)) {
    goto fail;
  }
  strcpy(script_file, script_dir);
  strcat(script_file, DIR_SEPARATOR);
  strcat(script_file, SCRIPT_FILE);
  
  if (real_file) strassign(real_file, script_file);

  strcpy(content_type_file, script_dir);
  strcat(content_type_file, DIR_SEPARATOR);
  strcat(content_type_file, CONTENT_TYPE_FILE);


  /* Check to make sure script file exists. */
  if (stat(script_file, &script_stat) != 0)
    goto fail;

  /* Remember the script's last modification time. */
  script->lastmod = script_stat.st_mtime;


  /* Check whether content-type file exists.  If so, read its contents. */
  content_type_file_handle = fopen(content_type_file, "r");
  if (content_type_file_handle) {
    char *nl_pos;
    /* Successfully opened the file. */
    if (fgets(content_type, sizeof(content_type),
              content_type_file_handle) == NULL
        || ferror(content_type_file_handle)) {
      fclose(content_type_file_handle);
      goto fail;
    }
    fclose(content_type_file_handle);
    /* Terminate at trailing newline, if any.  We don't write one, but
     * one might have been added if the repository was edited manually. */
    if ((nl_pos = strchr(content_type, '\n')) != NULL) {
      *nl_pos = '\0';
    }
    if ((nl_pos = strchr(content_type, '\r')) != NULL) {
      *nl_pos = '\0';
    }
    strassign(&script->content_type, content_type);
  }
  else {
    /* If there was no content-type file, use the default content type. */
    strassign(&script->content_type, DEFAULT_CONTENT_TYPE);
  }

  /* Set disposition string in script. */
  strassign(&script->disp_str, disp_str);

  /* Look up disposition string; find corresponding disposition type */
  /* XXX generalize this -- probably should be a table lookup. */
  if (strcasecmp(disp_str, "sip-cgi") == 0) {
    script->disposition = disposition_sip_cgi;
  }
  else if (strcasecmp(disp_str, "cpl") == 0) {
    script->disposition = disposition_cpl;
  }
  else {
    script->disposition = disposition_unknown;
  }
  

  retval = 1;

fail:
  if (retval == 0) {
    /* Free newly-allocated memory */
    free_valid(script->disp_str);
    free_valid(script->content_type);
    free_valid(script->filename);
    if (real_file) {
      free_valid(*real_file);
      *real_file = NULL;
    }
  }
  
  return retval;
} /* ReadScript */

/*
 * CreateTempScript()
 * Create a hard link to 'scriptfile' in 'scriptdir', with a temporary
 * name beginning with TMP_SCRIPT_PREFIX.  Return new filename, or NULL
 * on failure.
 * You should hold the script dir lock (either reading or writing) before
 * calling this.
 * XXX: won't work on Win32, which hasn't got hard links.  Need to copy files
 * instead.
 */
#ifndef WIN32 /* No script for win32 */

static char* CreateTempScript(const char *script_file, const char* script_dir)
{
  char *tmp_file;
  int ret;

  do {
    /* Loop until we link sucessfully; another thread might come in beneath
     * us and use the same tempnam. */
    tmp_file = tempnam(script_dir, TMP_SCRIPT_PREFIX);
    if (tmp_file == NULL) {
      return NULL;
    }

    ret = link(script_file, tmp_file);
    if (ret == -1 && errno != EEXIST) {
      free(tmp_file);
      return NULL;
    }
  } while (ret == -1 && errno == EEXIST);

  return tmp_file;
} /* CreateTempScript */

#else

static char* CreateTempScript(const char *script_file, const char* script_dir)
{
  char *tmp_file;
  int ret;
  int e = 0; /* Win32 error */

  do {
    /* Loop until we link sucessfully; another thread might come in beneath
     * us and use the same tempnam. */
    tmp_file = tempnam(script_dir, TMP_SCRIPT_PREFIX);
    if (tmp_file == NULL) {
      return NULL;
    }

    ret = (CopyFile(script_file, tmp_file, TRUE /* No overwrite */) ? 0 : -1);
    if ( ret == -1 ) {
      e = GetLastError();
    }
    if (ret == -1 && errno != ERROR_ALREADY_EXISTS) {
      free(tmp_file);
      return NULL;
    }
  } while (ret == -1 && errno == ERROR_ALREADY_EXISTS);

  return tmp_file;
} /* CreateTempScript */

#endif

/*
 * ReadAndCopyScript()
 * Read a script directory and populate the script_t pointed to with the
 * resulting information, then create a temporary file for the script.
 * Return 0 on failure, non-zero on success.
 * You should hold the script dir lock (either reading or writing) before
 * calling this.
 */
static int ReadAndCopyScript(const char* script_dir, const char* disp_str,
                             script_t *script)
{
  char *real_scriptfile;

  if (!ReadScript(script_dir, disp_str, script, &real_scriptfile)) {
    return 0;
  }

  script->filename = CreateTempScript(real_scriptfile, script_dir);
  if (script->filename == NULL) {
    FreeScript(script);
    free_valid(real_scriptfile);
    return 0;
  }

  free_valid(real_scriptfile);
  return 1;
} /* ReadAndCopyScript */



/* 
 * FreeScript()
 * Free a script_t.  Delete its temporary file, then free its memory.
 * You should hold a script lock before calling this function.
 */
static void FreeScript(script_t *script)
{
  if (script->filename != NULL) {
    unlink(script->filename); /* Can fail; this is ok. */
  }

  free_valid(script->disp_str);
  free_valid(script->content_type);
  free_valid(script->filename);

  /* Don't free "script" itself, since it's probably a pointer into
   * a larger structure. */
} /* FreeScript */


/*
 * GetScriptset()
 * Given a root directory, and a username, populate the scriptset pointed to
 * by "set" with information about the script set living under that user's
 * directory.  Duplicate individual scripts as needed.
 * Returns non-zero on success, zero on failure.  */

#ifndef WIN32
 
int GetScriptset(const char* base_dir, const char* username, scriptset_t *set)
{
  DIR *scriptdir = NULL;
  struct dirent *subdir;
  struct stat subdir_stats;
  char buf[PATH_MAX];
  char *buf_namepos;
  int buf_nameroom;
  void *tmp_ptr;
  int retval = 0;
  int set_size = 1; /* Assume we normally have only one script. */
  
  ReadLockScriptDir();

  memset(set, 0, sizeof(scriptset_t));

  /* Construct pathnames. */
  if (strlen(base_dir) + strlen(username) + 2 * strlen(DIR_SEPARATOR) + 1
      >= sizeof(buf)) goto fail;
  strcpy(buf, base_dir);
  strcat(buf, DIR_SEPARATOR);
  strcat(buf, username);

  strassign(&set->root_dir, buf);

  strcat(buf, DIR_SEPARATOR);
  buf_namepos = strchr(buf, '\0');
  buf_nameroom = buf + sizeof(buf) - buf_namepos - 1;


  /* Scan directory. */
  scriptdir = opendir(set->root_dir);
  if (scriptdir == NULL) {
    if (errno == ENOENT) {
      /* User doesn't have a directory.  This is fine. */
      retval = 1;
    }
    goto fail;
  }

  /* Allocate memory for scripts */
  set->scripts = (script_t *) malloc(set_size * sizeof(script_t));
  if (set->scripts == NULL) goto fail;

  while ((subdir = readdir(scriptdir)) != NULL) {
    /* Ignore special directories. */
    if (strcmp(subdir->d_name, ".") == 0 ||
        strcmp(subdir->d_name, "..") == 0) {
      continue;
    }

    /* Build pathname of subdir */
    strncpy(buf_namepos, subdir->d_name, buf_nameroom);

    /* If the directory entry isn't itself a directory, skip it. */
    if (stat(buf, &subdir_stats) != 0) goto fail;
    if (!(subdir_stats.st_mode & S_IFDIR))
      continue;
        
    /* Increase allocated memory for scripts, if necessary */
    if (set->count >= set_size) {
      set_size *= 2;
      tmp_ptr = realloc(set->scripts, set_size * sizeof(script_t));
      if (tmp_ptr == NULL) goto fail;
      set->scripts = tmp_ptr;
    }

    /* Read in script info for this directory.  If we sucessfully
     * read it in, increment set->count. */
    if (ReadAndCopyScript(buf, subdir->d_name, &set->scripts[set->count])) {
      set->count++;
    }
  }

  /* Trim set->scripts down to only the size we actually used. */
  if (set->count < set_size) {
    tmp_ptr = realloc(set->scripts, set->count * sizeof(script_t));
    if (tmp_ptr == NULL) goto fail;
    set->scripts = tmp_ptr;
  }
  
  /* Succeeded. */
  retval = 1;

fail:
  if (!retval) {
    int i;
    /* If we're returning failure, clean up allocated memory. */
    for (i = 0; i < set->count; i++) {
      FreeScript(&set->scripts[i]);
    }
    free_valid(set->root_dir);
    free_valid(set->scripts);
  }

  if (scriptdir) closedir(scriptdir); /* May fail; this is okay */
  UnlockScriptDir();
  return retval;
} /* GetScriptset */

#else

int GetScriptset(const char* base_dir, const char* username, scriptset_t *set)
{
  HANDLE scriptdir = NULL;
  WIN32_FIND_DATA subdir;
  struct stat subdir_stats;
  char buf[PATH_MAX];
  char *buf_namepos;
  int buf_nameroom;
  void *tmp_ptr;
  int retval = 0;
  int set_size = 1; /* Assume we normally have only one script. */
  int e = 0;

  ReadLockScriptDir();

  memset(set, 0, sizeof(scriptset_t));

  /* Construct pathnames. */
  if (strlen(base_dir) + strlen(username) + 2 * strlen(DIR_SEPARATOR) + 1
      >= sizeof(buf)) goto fail;
  strcpy(buf, base_dir);
  strcat(buf, DIR_SEPARATOR);
  strcat(buf, username);

  strassign(&set->root_dir, buf);

  strcat(buf, DIR_SEPARATOR);
  buf_namepos = strchr(buf, '\0');
  buf_nameroom = buf + sizeof(buf) - buf_namepos - 1;

  strcat(buf, "*");
  /* Scan directory. */
  scriptdir = FindFirstFile(buf, &subdir);
  if (scriptdir == INVALID_HANDLE_VALUE) {
    if ((e = GetLastError()) == ERROR_FILE_NOT_FOUND ||
         e == ERROR_PATH_NOT_FOUND) {
      /* User doesn't have a directory.  This is fine. */
      retval = 1;
    }
    debug("GetScriptSet", "Failing since no file\n");
    goto fail;
  }

  /* Allocate memory for scripts */
  set->scripts = (script_t *) malloc(set_size * sizeof(script_t));
  if (set->scripts == NULL) goto fail;

  while (e == 0) {
    /* Ignore special directories. */
    if (strcmp(subdir.cFileName, ".") == 0 ||
        strcmp(subdir.cFileName, "..") == 0) {
      if ( !FindNextFile(scriptdir, &subdir) ) {
        e = GetLastError();
        debug("GetScriptSet", "Failing since end of directory\n");
        break;
      }
      continue;
    }

    /* Build pathname of subdir */
    strncpy(buf_namepos, subdir.cFileName, buf_nameroom);

    /* If the directory entry isn't itself a directory, skip it. */
    if (stat(buf, &subdir_stats) != 0) goto fail;
    if (!(subdir_stats.st_mode & S_IFDIR)) {
      if ( !FindNextFile(scriptdir, &subdir) ) {
        e = GetLastError();
        debug("GetScriptSet", "Failing since end of directory\n");
        break;
      }
      continue;
    }
    
    /* Increase allocated memory for scripts, if necessary */
    if (set->count >= set_size) {
      set_size *= 2;
      tmp_ptr = realloc(set->scripts, set_size * sizeof(script_t));
      if (tmp_ptr == NULL) goto fail;
      set->scripts = tmp_ptr;
    }

    /* Read in script info for this directory.  If we sucessfully
     * read it in, increment set->count. */
    if (ReadAndCopyScript(buf, subdir.cFileName, &set->scripts[set->count])) {
      set->count++;
    }

    if ( !FindNextFile(scriptdir, &subdir) ) {
      e = GetLastError();
      debug("GetScriptSet", "Failing since end of directory\n");
      break;
    }
  }

  /* Trim set->scripts down to only the size we actually used. */
  if (set->count < set_size) {
    if ( set->count <= 0 ) {
      free_valid(set->scripts);
      set->scripts = NULL;
    }
    else {
      tmp_ptr = realloc(set->scripts, set->count * sizeof(script_t));
      if (tmp_ptr == NULL) goto fail;
      set->scripts = tmp_ptr;
    }
  }
  
  /* Succeeded. */
  retval = 1;

fail:
  if (!retval) {
    int i;
    /* If we're returning failure, clean up allocated memory. */
    for (i = 0; i < set->count; i++) {
      FreeScript(&set->scripts[i]);
    }
    free_valid(set->root_dir);
    free_valid(set->scripts);
  }

  if (scriptdir != INVALID_HANDLE_VALUE) FindClose(scriptdir); /* May fail; this is okay */
  UnlockScriptDir();
  return retval;
} /* GetScriptset */

#endif

/*
 * FreeScriptset()
 * Free a scriptset, and all its associated scripts, deleting the
 * temporary script files.
 */
void FreeScriptset(scriptset_t *set)
{
  int i;

  ReadLockScriptDir();

  for (i = 0; i < set->count; i++) {
    FreeScript(&set->scripts[i]);
  }
  free_valid(set->root_dir);
  free_valid(set->scripts);

  UnlockScriptDir();
} /* FreeScriptset */


/*
 * GetScriptByDisposition()
 * Given the script_disposition_t disp, return a script_t from
 * the scriptset set corresponding to that disposition, or NULL.
 */
script_t* GetScriptByDisposition(const scriptset_t *set,
                                 script_disposition_t disp)
{
  int i;
  
  if (disp == disposition_invalid) return NULL;
  
  for (i = 0; i < set->count; i++) {
    if (set->scripts[i].disposition == disp) {
      return &set->scripts[i];
    }
  }
  return NULL;
} /* GetScriptByDisposition */


/*
 * GetScriptByDispositionString()
 * Given the script disposition string disp, return a script_t from
 * the scriptset set corresponding to that disposition, or NULL.
 * If found, fill set_pos with the position of the script
 * in the array, or -1.
 */
static script_t* GetScriptByDispositionString(const scriptset_t *set,
                                              const char *disp_str, 
                                              int *set_pos)
{
  int i;
  for (i = 0; i < set->count; i++) {
    if (set->scripts[i].disposition == disposition_invalid) {
      continue;
    }
    if (strcasecmp(set->scripts[i].disp_str, disp_str) == 0) {
      *set_pos = i;
      return &set->scripts[i];
    }
  }
  *set_pos = -1;
  return NULL;
} /* GetScriptByDispositionString */


/* 
 * HasExecutableScript()
 * Return true if the pointed-to scriptset contains an "executable" script:
 * some sort of user policy.
 * XXX: this should be generalized to a table lookup of some sort.
 */
int HasExecutableScript(const scriptset_t *set)
{
  int i;
  for (i = 0; i < set->count; i++) {
    if (set->scripts[i].disposition == disposition_sip_cgi
        || set->scripts[i].disposition == disposition_cpl) {
      return 1;
    }
  }
  return 0;
} /* HasExecutableScript() */


/* 
 * unlink_if_exists()
 * Delete a file, if it doesn't already exist.  Otherwise no-op.
 * Return 0 on success, -1 on failure, just like unlink().
 */
static int unlink_if_exists(const char *path)
{
  int ret;

  ret = unlink(path);
  if (ret == -1 && errno != ENOENT) {
    return -1;
  }

  return 0;
} /* unlink_if_exists */

/*
 * mkdir_if_missing()
 * Create a directory, if it doesn't already exist.  If it *does* exist,
 * set its file permissions appropriately (u+rwx, og-w, rest don't care).
 * Return 0 on success, otherwise -1 and set errno, just like mkdir.
 */
static int mkdir_if_missing(const char *path)
{
  struct stat dir_stats;
  int ret;

  ret = stat(path, &dir_stats);
  if (ret != 0 && errno != ENOENT) {
    /* Couldn't stat dir, but not because it doesn't exist. */
    return -1;
  }
  else if (ret == 0) {
    /* Exists. */
    if (!(dir_stats.st_mode & S_IFDIR)) {
      /* Not a directory */
      errno = ENOTDIR;
      return -1;
    }
    return chmod(path, (dir_stats.st_mode | S_IRWXU) & ~S_IWGRP & ~S_IWOTH);
  }
  else {
    /* Doesn't exist. */
    /* Assume parent directories exist. */
    /* Use 0755 as permissions; let umask take care of clearing out unwanted
       other mode bits. */
#ifndef WIN32
    return mkdir(path, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
#else
    return mkdir(path);
#endif
  }
} /* mkdir_if_missing */


/*
 * set_mtime()
 * Set the mtime of the file 'path' to the specified value.  The
 * file must exist.
 * Return 0 on success, otherwise non-zero.
 */
#ifndef WIN32 
static int set_mtime(const char *path, time_t mtime) {
  struct timeval times[] = {{0, 0}, {0, 0}};
  struct stat file_stat;

  /* The utimes() call sets both atime and mtime.  We don't want to
   * change atime, so stat the file to get its existing value.
   */
  if (stat(path, &file_stat) != 0) {
    return 1;
  }

  times[0].tv_sec = file_stat.st_atime;
  times[1].tv_sec = mtime;
  /* Note that we set tv_usec to zero for both atime and mtime.  This may be
   * wrong on *BSD, which has struct stat->st_atimensec etc., but I'm not
   * going to worry about it. 
   */

  if (utimes(path, times) != 0) {
    return 1;
  }

  return 0;
} /* set_mtime */
  
#else
static int set_mtime(const char *path, time_t mtime) {
  struct utimbuf times;
  struct stat file_stat;

  /* The utimes() call sets both atime and mtime.  We don't want to
   * change atime, so stat the file to get its existing value.
   */
  if (stat(path, &file_stat) != 0) {
    return 1;
  }

  times.actime = file_stat.st_atime;
  times.modtime = mtime;
  /* Note that we set tv_usec to zero for both atime and mtime.  This may be
   * wrong on *BSD, which has struct stat->st_atimensec etc., but I'm not
   * going to worry about it. 
   */

  if (utime(path, &times) != 0) {
    return 1;
  }

  return 0;
} /* set_mtime */
  
#endif

/*
 * write_content()
 * Write 'nbyte' bytes of content from 'buf' to the file 'path'.
 * If mtime is anything other than -1, also set the file's mtime to that
 * value.
 * (Note that this means that we can't set an mtime of time_t -1.  I'm
 * not too worried about this.  -1 was chosen because this is the error
 * return value from mktime()/timegm().)
 * Return -1 on failure, otherwise nbyte.
 * File should not previously exist.
 * Delete the file if the write was only partially sucessful.
 */
static int write_content(const char *path, const char *buf, int nbyte,
                         time_t mtime)
{
  int fd;
  int ret;

  /* Use 0644 as permissions; let umask take care of cleaning up. */
  if ((fd = open(path, O_WRONLY | O_CREAT | O_EXCL,
                 S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) < 0) {
    return -1;
  }
  ret = write(fd, buf, nbyte);
  if (ret < nbyte) {
    /* Either ret == -1, or partial write only. */
    close(fd);
    unlink_if_exists(path);
    return -1;
  }

  if (close(fd) < 0) {
    unlink_if_exists(path);
    return -1;
  }

  if (mtime != -1 && set_mtime(path, mtime) != 0) {
    unlink_if_exists(path);
    return -1;
  }

  return ret;
} /* write_content */


/*
 * AddScriptToSet()
 * Add a script to the specified set.  This causes it both to be written
 * to the actual disk repository, and to be added to the scriptset_t with
 * all the usual rules (such as a temporary script file).
 * If mtime is non-zero, set the script's mtime as well.
 * Returns 0 on failure, non-zero on success.
 */
int AddScriptToSet(scriptset_t *set, const char *disp_str,
                   const char *content_type, const char *content, int length,
                   time_t mtime)
{
  int retval = 0;
  char script_dir[PATH_MAX];
  char script_file[PATH_MAX], content_type_file[PATH_MAX];
  script_t *old_script;
  int pos;

  WriteLockScriptDir();

  /* Construct pathnames */
  if ((strlen(set->root_dir) + strlen (disp_str) + strlen(DIR_SEPARATOR) + 1)
      >= sizeof(script_dir)) goto fail; 
  strcpy(script_dir, set->root_dir);
  strcat(script_dir, DIR_SEPARATOR);
  strcat(script_dir, disp_str);

  if ((strlen(set->root_dir) + strlen (disp_str) + strlen(SCRIPT_FILE)
      + 2 * strlen(DIR_SEPARATOR) + 1) >= sizeof(script_file)) goto fail;
  strcpy(script_file, script_dir);
  strcat(script_file, DIR_SEPARATOR);
  strcat(script_file, SCRIPT_FILE);

  if ((strlen(set->root_dir) + strlen (disp_str) + strlen(CONTENT_TYPE_FILE)
      + 2 * strlen(DIR_SEPARATOR) + 1) >= sizeof(content_type_file)) goto fail;
  strcpy(content_type_file, script_dir);
  strcat(content_type_file, DIR_SEPARATOR);
  strcat(content_type_file, CONTENT_TYPE_FILE);

  /* If an old script exists, free it. */

  old_script = GetScriptByDispositionString(set, disp_str, &pos);

  if (old_script != NULL) {
    FreeScript(old_script);
  }

  /* Try to remove the old files.  If they don't exist, that's okay. */
  /* XXX: can we leave the repository in an unhappy state if we
   * fail later? */
  if (unlink_if_exists(script_file) != 0) {
    goto fail;
  }
  if (unlink_if_exists(content_type_file) != 0) {
    goto fail;
  }

  /* Try to create parent directories. */
  if (mkdir_if_missing(config.ScriptBase) != 0) goto fail;
  if (mkdir_if_missing(set->root_dir) != 0) goto fail;
  if (mkdir_if_missing(script_dir) != 0) goto fail;

  /* Write the content-type file. */
  if (write_content(content_type_file, content_type, strlen(content_type),
                    -1) < 0)
    goto fail;

  /* Write the script. */
  if (write_content(script_file, content, length, mtime) < 0)
    goto fail;

  /* Allocate a new array slot if there wasn't an old script. */ 
  if (old_script == NULL) {
    void *tmp_ptr;
    tmp_ptr = realloc(set->scripts, (set->count + 1) * sizeof(script_t));
    if (tmp_ptr == NULL) goto fail;
    set->count++;
    set->scripts = tmp_ptr;
    pos = set->count - 1;
  }
  
  /* Read the new script into the set. */
  ReadAndCopyScript(script_dir, disp_str, &set->scripts[pos]);

  retval = 1;

fail:
  UnlockScriptDir();
  return retval;
} /* AddScriptToSet */


/*
 * DeleteScriptFromSet()
 * Delete a script from the specified set.  This causes it to be deleted
 * both from the script repository and from the scriptset_t.
 */
int DeleteScriptFromSet(scriptset_t *set, const char *disp_str)
{
  int retval = 0;
  int pos;
  char script_dir[PATH_MAX];
  char script_file[PATH_MAX], content_type_file[PATH_MAX];
  script_t *old_script;

  WriteLockScriptDir();

  /* Construct pathnames */
  if ((strlen(set->root_dir) + strlen (disp_str) + strlen(DIR_SEPARATOR) + 1)
      >= sizeof(script_dir)) goto fail; 
  strcpy(script_dir, set->root_dir);
  strcat(script_dir, DIR_SEPARATOR);
  strcat(script_dir, disp_str);

  if ((strlen(set->root_dir) + strlen (disp_str) + strlen(SCRIPT_FILE)
      + 2 * strlen(DIR_SEPARATOR) + 1) >= sizeof(script_file)) goto fail;
  strcpy(script_file, script_dir);
  strcat(script_file, DIR_SEPARATOR);
  strcat(script_file, SCRIPT_FILE);

  if ((strlen(set->root_dir) + strlen (disp_str) + strlen(CONTENT_TYPE_FILE)
      + 2 * strlen(DIR_SEPARATOR) + 1) >= sizeof(content_type_file)) goto fail;
  strcpy(content_type_file, script_dir);
  strcat(content_type_file, DIR_SEPARATOR);
  strcat(content_type_file, CONTENT_TYPE_FILE);

  /* If an old script existed, free it. */

  old_script = GetScriptByDispositionString(set, disp_str, &pos);

  if (old_script != NULL) {
    FreeScript(old_script);
  }

  /* Try to remove the old files.  If they don't exist, that's okay. */
  if (unlink_if_exists(script_file) != 0) {
    goto fail;
  }
  if (unlink_if_exists(content_type_file) != 0) {
    goto fail;
  }

  /* XXX: We should try to remove the old parent directories.  However,
   * these may still hold temporary scripts, etc...not clear how best
   * to do this. */
  
  /* "Delete" script from set, if it existed. */
  if (pos >= 0) {
    memset(old_script, 0, sizeof(script_t));
    old_script->disposition = disposition_invalid;
  }

  retval = 1;

fail:
  UnlockScriptDir();
  return retval;
} /* DeleteScriptFromSet */


/*
 * GetScriptModificationTime()
 * Get the last-modification time of the script in the set with the given
 * disposition string, if it exists.  If not, return -1.
 * If you already have a script_t*, don't use this function;
 * use the structure's lastmod field instead.
 */
time_t GetScriptModificationTime(scriptset_t *set, const char *disp_str)
{
  script_t *script;
  int pos;

  script = GetScriptByDispositionString(set, disp_str, &pos);

  if (script == NULL) {
    return -1;
  }

  return script->lastmod;
} /* GetScriptModificationTime */


/*
 * ReadScriptToDString()
 * Read a script into a DString.  Return the length of the script, or -1.
 */
int ReadScriptToDString(script_t *script, DString *dstr)
{
  FILE* script_handle;
  char readbuf[BUFSIZ];
  int ret;

  if ( script->filename == NULL ) return -1;
  script_handle = fopen(script->filename, "r");
  if (script_handle == NULL) return -1;

  DStringTrunc(dstr);

  while ((ret = fread(readbuf, 1, sizeof(readbuf), script_handle)) > 0) {
    DStringAppend(dstr, readbuf, ret);
  }
  if (ferror(script_handle)) {
    DStringTrunc(dstr);
    fclose(script_handle);
    return -1;
  }
  fclose(script_handle);
  return DStringLength(dstr);
} /* ReadScriptToDString */


