/* sipd-support.c -- Miscellaneous support functions for sipd. */
/* Copyright 1998-2001 Columbia University.  All rights reserved. */

#include <string.h>
#include <errno.h>

#include "uri.h"
#include "dstring.h"
#include "strdupn.h"

#include "canonicalize.h"
#include "config.h"
#include "querydb.h"
#include "request-sip.h"

#include "sipd-support.h"

/*
 * IsAcceptableURI()
 * Validate that the URI of the given request is one that sipd is
 * willing to handle.  Return false if not, with the response status
 * set appropriately.  The phrasing of the response reason string includes
 * the description in uri_desc.
 */
int IsAcceptableURI(request_t *r, const uri_t *uri, const char *uri_desc)
{
  DString reason;
  int ret = 1;

  DStringInit(&reason);

  /* 
   * sipd handles 'sip' URI for requests.  The domain
   * part is compared against a regular expression for SIP URIs.
   */
  if (uri->scheme == NULL) {
    debug("IsAcceptableURI",
          "r = 0x%p: Missing scheme in %s", r, uri_desc);
    DStringPrintf(&reason, "Ill-formed %s", uri_desc);
    ResponseSetStatus(r, 400, DStringValue(&reason));
    ret = 0;
  }

  else if (strcasecmp(uri->scheme, "sip") != 0 &&
           strcasecmp(uri->scheme, "tel") != 0) {
    DStringPrintf(&reason, "Unsupported scheme %s in %s",
                  uri->scheme, uri_desc);
    ResponseSetStatus(r, 404, DStringValue(&reason));
    ret = 0;
  }

  DStringFree(&reason);
  return ret;
} /* IsAcceptableURI */


/*
 * CanonicalizeAcceptableURI()
 * Check whether the URI 'uri' is acceptable, as defined by IsAcceptableURI.
 * If !remote_ok, check whether it's local as well.
 * If so, fill in canon_res with the result of canonicalization.
 * If not, return false and set the request's response status
 * appropriately.
 */
int CanonicalizeAcceptableURI(request_t *r, const uri_t *uri,
                              canon_result *canon_res,
                              int remote_ok,
                              const char *uri_desc,
                              const char *process_desc)
{
  int ret = 1;
  DString reason;

  if (!IsAcceptableURI(r, uri, uri_desc)) {
    return 0;
  }

  DStringInit(&reason);

  /* Canonicalize the uri to find the proper database entry. */
  canonicalize(uri, canon_res);

  switch(canon_res->status) {
  case canon_unique:
    /* OK. */
    ret = 1;
    break;

  case canon_remote:
    if (remote_ok) {
      ret = 1;
    }
    else {
      /* Won't register (or whatever) a remote address */
      DStringPrintf(&reason, "Non-local %s forbidden in %s",
                    uri_desc, process_desc);
      ResponseSetStatus(r, 403, DStringValue(&reason));
      ret = 0;
    }
    break;

  case canon_multiple:
    DStringPrintf(&reason, "Ambiguous %s in %s", uri_desc, process_desc);
    
    ResponseSetStatus(r, 485, DStringValue(&reason));
    ret = 0;
    break;

  case canon_error:
    /* Canonicalize returned an internal error */
    ResponseSetStatus(r, 500, canon_res->u.error);
    ret = 0;
    break;

  default:
    /* Shouldn't get here; report an error */
    errno = 0;
    error("CanonicalizeAcceptableURI",
          "Request 0x%p: Surprising value %d for canon_res->status\n",
          r, canon_res->status);
    ResponseSetStatus(r, 500, NULL);
    ret = 0;
    break;
  }
  DStringFree(&reason);
  return ret;
} /* CanonicalizeAcceptableURI */


/*
 * CanonicalizeAndLookupURI()
 * Canonicalize the URI 'uri', check if it's acceptable (as above),
 * then find a put_table_entry corresponding to it.  On failure, return
 * NULL and set response statuses appropriately. 
 */
put_table_entry* CanonicalizeAndLookupURI(request_t *r, const uri_t *uri,
                                          int remote_ok,
                                          const char *uri_desc,
                                          const char *process_desc,
                                          put_table_entry *other_user_val)
{
  canon_result canon;
  char *userid = NULL;
  db_err db_ret;
  char *db_ret_str = NULL;
  put_table_entry *ret;
  DString reason;

  DStringInit(&reason);
  memset(&canon, 0, sizeof(canon));

  if (!CanonicalizeAcceptableURI(r, uri, &canon, remote_ok,
                                 uri_desc, process_desc)) {
    ret = NULL;
    goto done;
  }

  userid = get_table_key_from_URI(config.database, canon.u.unique,
                                  &db_ret, &db_ret_str);

  if (userid == NULL) {
    DStringPrintf(&reason, "Can't get database key: %s", db_ret_str);
    ResponseSetStatus(r, 400, DStringValue(&reason));
    ret = NULL;
    goto done;
  }

  if (other_user_val != NULL &&
      strcompare(other_user_val->user, userid, 0) == 0) {
    ret = other_user_val;
  }
  else {
    ret = put_get_by_user(config.database, userid, &db_ret, &db_ret_str);
    
    if (ret == NULL) {
      if (db_ret_invalid(db_ret)) {
        errno = 0;
        error("CanonicalizeAndLookupURI",
              "Request 0x%p: put_get_by_user(%s) returned %d: %s\n",
              r, userid, db_ret, db_ret_str);
      }
      
      request_set_dberr_status(r, db_ret, userid, 500, db_ret_str);
    }
  }

done:
  free_canon_result(&canon);
  free_valid(userid);
  free_valid(db_ret_str);
  DStringFree(&reason);

  return ret;
} /* CanonicalizeAndLookupURI */
                         
