/*
sipd - SIP redirect, registration and proxy server

Copyright 1998-2001 by Columbia University; all rights reserved 

Permission to use, copy, modify, and distribute this software and its
documentation for research and educational purposes and without fee is
hereby granted to non-profit organizations, provided that the above
copyright notice appear in all copies and that both that the copyright
notice and warranty disclaimer appear in supporting documentation, and
that the names of the copyright holders or any of their entities not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  Use of this
software in whole or in parts for direct commercial advantage requires
explicit prior permission. 

The copyright holders disclaim all warranties with regard to this
software, including all implied warranties of merchantability and
fitness.  In no event shall the copyright holders be liable for any
special, indirect or consequential damages or any damages whatsoever
resulting from loss of use, data or profits, whether in an action of
contract, negligence or other tortuous action, arising out of or in
connection with the use or performance of this software.

Written by Henning Schulzrinne, Yan Xu, Jonathan Lennox, Panagiotis
  Sebos, and Xiaotao Wu 
*/

#include <stdlib.h>    /* getopt() */
#include <unistd.h>    /* close() */
#include <stdio.h>     /* fopen() */
#include <string.h>    /* strdup(), strchr() */
#include <pthread.h>   /* pthread_ */
#include <signal.h>    /* signal() */
#include <sys/param.h> /* MAXPATHLEN */

#include "dstring.h"
#include "error.h"
#include "fullpath.h"
#include "raise-rlimits.h"
#include "strdupn.h"   /* strassign */
#include "tcp.h"
#include "ssl_common.h"
#include "ssl_tcp.h"
#include "udp.h"
#include "thread_pool.h"
#include "sysdep.h"

#include "config.h"
#include "sipsnmp_interface.h"

#include "dbapi.h"

#include "sipd.h"
#include "sipd-config.h"


#ifdef USE_PGP_SDK
#include "enginepgp.h"
#include "pgpErrors.h"
#endif

#ifndef USE_MYSQL
#error "sipd needs MySQL database. On unix configure with --with-mysql option. On Win32 add USE_MYSQL macro in libsip, and sipd projects' settings."
#endif

#ifdef USE_SNMP
#define NUM_THREADS 3 /* TCP, UDP, and SNMP */
#else 
#define NUM_THREADS 2 /* TCP and UDP */
#endif

/* Prototypes for external functions without their own header files.
 * (They really ought to have them.) */
extern void check_registration();

/* In RTSP, missing content means zero.  In SIP, it means infinite. */
const int missing_content_length_means_zero = 0; 

/* The name of this server */
const char *server_name = SERVER_NAME;

/*
 * usage()
 * Usage information. Exits program with status code 1.
 */
void usage(char *prog)
{
  fprintf(stderr,
          "SIP Proxy, Redirect and Registration Server %s (built %s %s)\n",
          VERSION, __DATE__, __TIME__);
  fprintf(stderr, "Copyright 1997-2001 Columbia University\n");
  fprintf(stderr, "Usage: %s [options]\n"
          " -s ServerRoot  : Server root. Default is " DEFAULT_SERVER_ROOT "\n"
          " -d ServerRoot  : Same as -s.\n"
          " -l licensefile : License file. Default is sipd.license\n"
          " -f ConfigFile  : Config file. Default is sipd.conf\n"
          " -v             : Verbose mode.\n"
          " -X             : Non daemon mode.\n"
          " -n             : Use IP address instead of domain name.\n",
          prog);
  exit(1);
} /* usage */


/*
 * ParseCommandLine()
 * Parse sipd's command line, and fill in the commandline global.
 * The locations of the license and config files are determined like this:
 * If they are not specified on the command line, they are assumed to
 * have the names DEFAULT_LICENSE_FILE and DEFAULT_CONFIG_FILE relative to
 * ServerRoot.  If they *are* specified on the command line, they
 * are assumed to be relative path names relative to the directory in which
 * sipd was invoked.
 */
static void ParseCommandLine(int argc, char *argv[])
{
  char c;
  extern char *optarg;            /* command line argument value */
  extern int optind;

  /* The effective value of the server root, for locating the license and
   * config file.  We do not just allow commandline.ServerRoot to default
   * to DEFAULT_SERVER_ROOT, because resolution of ServerRoot in the config
   * file is different in the two cases. */
  const char *effective_server_root;

  /* Parse command line options. */
  commandline.OrigDir  = getcwd(NULL, MAXPATHLEN);
  commandline.ServerRoot  = NULL;
  commandline.LicenseFile = NULL;
  commandline.ConfigFile  = NULL;
  commandline.do_daemon   = 1;
  commandline.debug_flag  = 0;
  commandline.use_numeric_address = 0;

  while ((c = getopt(argc, argv, "d:s:l:f:vXn")) != EOF) {
    switch (c) {
    case 's':
    case 'd':
      strassign(&commandline.ServerRoot, optarg);
      break;
    case 'l':
      strassign(&commandline.LicenseFile, optarg);
      break;
    case 'f':
      strassign(&commandline.ConfigFile, optarg);
      break;
    case 'v':
      commandline.debug_flag = 1;
      break;
    case 'X':
      commandline.do_daemon = 0;
      break;
    case 'n':
      commandline.use_numeric_address = 1;
      break;
    case '?':
    default:
      usage(argv[0]);
      break;
    }
  }

  effective_server_root =
    commandline.ServerRoot != NULL ? commandline.ServerRoot
    : DEFAULT_SERVER_ROOT;

  if (commandline.LicenseFile == NULL) {
    strassign(&commandline.LicenseFile,
              fullpath(effective_server_root, DEFAULT_LICENSE_FILE));
  }
  if (commandline.ConfigFile == NULL) {
    strassign(&commandline.ConfigFile,
              fullpath(effective_server_root, DEFAULT_CONFIG_FILE));
  }

  /* debug_flag is a global used by libcine, in error.c. */
  debug_flag = commandline.debug_flag;
} /* ParseCommandLine */


#ifdef USE_SNMP
/*
 * RunSNMPThread()
 * Start the SNMP thread.
 */
static void RunSNMPThread(pthread_t *tid)
{
  SnmpData *p_snmpData;
  int e;

  *tid = NULL;
 
  if(!config.StartSnmp)
     return;

  p_snmpData = (SnmpData*) calloc(1, sizeof(SnmpData));
  p_snmpData->sub_agent = SUB_AGENT; /* right now, keep it as subagent. */
  p_snmpData->agentx_socket_path = config.AgentxSocketPath;
  
  if ((e = pthread_create(tid, NULL, 
                                   sipsnmp_main, (void*)p_snmpData)) != 0) {
    error("main", "Cannot create snmp thread: %s\n",strerror(e));
  }
} /* RunSNMPThread */
#endif /* USE_SNMP */


/* 
 * RunThreads()
 * Start the main socket listening threads (tcp and udp) for the server.
 * Also start SNMP, if it's configured.
 */
static void RunThreads(pthread_t tid[NUM_THREADS], int *started_threads) {
  int e;                          /* thread error condition */
  struct tcp_thread_arg *tcp_arg; /* params to TCP/UDP threads */
  struct udp_thread_arg *udp_arg; 
  struct ssl_tcp_thread_arg *ssl_tcp_arg; 

  tid[0] = 0;
  tid[1] = 0;
  *started_threads = 2;

  
  /* Allocate tcp_arg and create TCP thread for SSL */
  if(config.SSLport) {
     printf("*** run SSLport ****\n");
     ssl_tcp_arg = (struct ssl_tcp_thread_arg *) malloc(sizeof(struct ssl_tcp_thread_arg));
     ssl_tcp_arg->server = NULL; /* Not used, since single server */
     ssl_tcp_arg->port   = config.SSLport;
     if ((e = pthread_create(&tid[0], NULL, 
                                   ReceiveSSLTCP, (void *)ssl_tcp_arg)) != 0) {
        fatal("main", "Cannot create SSL thread: %s\n", strerror(e));
     }
  } else {

     /* Allocate tcp_arg and create TCP thread */
     printf("*** run tcp port ****\n");
     tcp_arg = (struct tcp_thread_arg *) malloc(sizeof(struct tcp_thread_arg));
     tcp_arg->server = NULL; /* Not used, since single server */
     tcp_arg->port   = config.port;
     tcp_arg->primaryIP = NULL;

     if (config.PrimaryIP)
        strassign(&tcp_arg->primaryIP, config.PrimaryIP);

     if ((e = pthread_create(&tid[0], NULL, 
                                   ReceiveTCP, (void *)tcp_arg)) != 0) {
        fatal("main", "Cannot create TCP thread: %s\n", strerror(e));
     }


  /* Allocate udp_arg and create UDP thread */
  udp_arg = (struct udp_thread_arg *) malloc(sizeof(struct udp_thread_arg));
  udp_arg->server = NULL; /* Not used, since single server */
  udp_arg->port   = config.port;
  if ((e = pthread_create(&tid[1], NULL, 
                                   ReceiveUDP, (void *)udp_arg)) != 0) {
    fatal("main", "Cannot create UDP thread: %s\n", strerror(e));
  }

  }
#ifdef USE_SNMP
  RunSNMPThread(&tid[2]);
  if (config.StartSnmp)
	*started_threads = 3;
#endif

  /* No need to free tcp|udp_arg. The created thread frees them */
} /* RunThreads */

/*
 * main()
 * Initialize and start the server.
 */
int main(int argc, char *argv[])
{
  pthread_t tid[NUM_THREADS];
  int act_threads;
  extern void InitLibSip();
  extern void do_daemon(void);
  extern void setupExitThread(void);

  memset(tid, 0, NUM_THREADS * sizeof(pthread_t));

  startupSocket();

  /* Errors that occur before sipd.conf is read should go to stderr. */
  ErrorInit("stderr");

  raise_rlimits();

  ParseCommandLine(argc, argv);

  check_registration(); /* may terminate program if invalid */

  if (commandline.do_daemon) do_daemon();

  ConfigInit(0);


#ifdef USE_PGP_SDK
  InitializePGPsdk();
#endif /* USE_PGP_SDK */

  atexit(ConfigCleanup);
#ifndef WIN32
  signal(SIGHUP,  ConfigInit);
  signal(SIGPIPE, SIG_IGN); 
#endif

  InitLibSip();      /* initialize libsip */

  setupExitThread();

  if (InitThreadPool(config.ThreadPoolSize)) {
    fatal("main", "InitThreadPool failed...\n");
  }
  
  RunThreads(tid, &act_threads);
  
  debug("main", "sipd server %s (built %s %s) listening on port %d\n",
	VERSION, __DATE__, __TIME__, config.port);
  
  /* Wait for thread completion (which shouldn't happen). */
  WaitForMultipleObjects(act_threads, tid, TRUE, INFINITE);
  
  debug("main", "main returns\n");
  exit(0);
} /* main */


