/*
* SIP test server: 
* [-p clientport] [-P serverport] [-c client] [-u] file
* -u uses UDP.
* Sends file in response to first request
* Fills in 
*   $H: own host network address
*   $T: transport protocol
*   $P: server receive port
*
* Does not implement reliability.
*/

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>   /* inet_addr() */
#include <stdlib.h>      /* getopt() */
#include <stdio.h>       /* fprintf() */
#include <string.h>      /* strlen(), strcpy() */
#include <unistd.h>      /* open(), read(), close() */
#include <fcntl.h>       /* open() */

#include "host2ip.h"

static void usage(char *prog)
{
  fprintf(stderr, "SIP test client %s (c) Columbia University\n", VERSION);
  fprintf(stderr, 
  "Usage: %s [-p port] [-P port] [-c server] [-u] file\n",
    prog);
  fprintf(stderr, "\
 -p port     server port (default 5060)\n\
 -P port     client port (default 5060) (UDP only)\n\
 -c client   client address (default localhost)\n\
 -u          use UDP instead of TCP\n\
");
  exit(1);
} /* usage */


/* Create outgoing packet. */
static void create_packet(char *template, char *packet, 
  int port, int sock_type)
{
  char *s, *d;

  for (s = template, d = packet; *s; s++) {
    if (*s == '$') {
      if (*(s+1) == 'H') {
        gethostname(d, 1000);
      }
      else if (*(s+1) == 'P') {
        sprintf(d, "%d", port);
      }
      else if (*(s+1) == 'T') {
        strcpy(d, sock_type == SOCK_DGRAM ? "UDP" : "TCP");
      }
      else {
        sprintf(d, "$%c", *(s+1));
      }
      d += strlen(d);
      s++;
    }
    else {
      *d++ = *s;
    }
  }
  *d = '\0';
} /* create_packet */


int main(int argc, char *argv[])
{
  int bindsock;                /* socket to bind to */
  int isock;                   /* socket to read from (requests) */ 
  int osock;                   /* socket to write to (resposnes) */
  struct sockaddr_in sin, from;
  int fromlen;
  char template[8192];         /* request template read from file */
  char packet[8192];           /* packet to be sent/received */
  int src_port;                /* source port */
  int clientport = 5060;       /* destination (UDP client) port */
  int serverport = 5060;       /* local (server) port */
  int sock_type = SOCK_STREAM; /* TCP socket */
  int c;                       /* character */
  int n;                       /* bytes read */
  char *client = "localhost";  /* name of server host */
  int fd;                      /* file for template */
  struct timeval now;

  while ((c = getopt(argc, argv, "c:p:P:u")) != EOF) {
    switch (c) {
    case 'c':
      client = optarg;
      break;
    case 'p':
      serverport = atoi(optarg);
      break;
    case 'P':
      clientport = atoi(optarg);
      break;
    case 'u':
      sock_type = SOCK_DGRAM;
      break;
    case '?':
      usage(argv[0]);
      break;
    }
  }

  if (argc <= optind) {
    usage(argv[0]);
  }

  /* Read file. */
  if ((fd = open(argv[optind], O_RDONLY)) < 0) {
    perror("open");
  }
  if ((n = read(fd, template, sizeof(template)-1)) <= 0) {
    perror("read");
  }
  template[n] = '\0';
  close(fd);

  /* Create socket for incoming requests. */
  if ((bindsock = socket(PF_INET, sock_type, 0)) < 0) {
    perror("socket");
  }
  from.sin_family      = AF_INET;
  from.sin_port        = htons(serverport);
  from.sin_addr.s_addr = INADDR_ANY;
  if (bind(bindsock, (struct sockaddr *)&from, sizeof(from)) < 0) {
    perror("bind");
  }
  fromlen = sizeof(from);
  if (getsockname(bindsock, (struct sockaddr *)&from, &fromlen) < 0) {
    perror("getsockname");
  }
  src_port = ntohs(from.sin_port);

  /* Socket for incoming requests. */
  if (sock_type == SOCK_STREAM) {
    if (listen(bindsock, 1) < 0) {
      perror("listen");
    }
    /* Wait for an incoming request to respond to. */
    if ((isock = accept(bindsock, (struct sockaddr *)&from, &fromlen)) < 0) {
      perror("accept");
    }
  }
  else {
    /* UDP listens directly on the binding port */
    isock = bindsock;
  }

  /* Socket for outgoing responses */
  if (sock_type == SOCK_DGRAM) {
    if ((osock = socket(PF_INET, sock_type, 0)) < 0) {
      perror("socket");
    }
    sin.sin_family  = AF_INET;
    sin.sin_port    = htons(clientport);
    sin.sin_addr    = host2ip(client);
    if (connect(osock, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
      perror("connect");
    }
  }
  else {
    /* TCP responds on the same socket */
    osock = isock;
  }

  /* read from network and print */
  do {
    if ((c = recvfrom(isock, packet, sizeof(packet)-1, 0, 
         (struct sockaddr *)&from, &fromlen)) < 0) {
      perror("recvfrom");
    }
    packet[c] = '\0';
    gettimeofday(&now, 0);    
    printf("Received t=%ld.%03ld from=%s:%d:\n%s\n------------------------------------------\n",
      (long)now.tv_sec, (long)now.tv_usec/1000,
      inet_ntoa(from.sin_addr), from.sin_port, packet);

    /* Send response */
    if (template[0] != '\0') {
      create_packet(template, packet, src_port, sock_type);

      if (write(osock, packet, strlen(packet)) < 0) {
	perror("write response");
      }

      printf("Written from source=%d to destination=%d:\n%s\n------------------------------------------\n",
	     ntohs(from.sin_port), ntohs(sin.sin_port), packet);
      
      /* don't transmit responses to subsequent requests */
      template[0] = '\0';
    }
  } while (c > 0);

  return 0;
} /* main */
