
//package SipSImp;

import java.util.*;
import org.ietf.sip.*;


/**
 * Represents value of SIP From and To headers and serves as the base
 * class for the representation of Contact header values.
 *
 * <p>A SipAddress can be thought of as representing something that conforms
 * to the following grammar [RFC 2543, section 6.21]:
 * <pre>
 *   SipAddress     = (name-addr | addr-spec) *(";" addr-params)
 *   name-addr      = [display-name] "<" addr-spec ">"
 *   addr-spec      = SIP-URL | URI
 *   display-name   = *token | quoted-string
 * </pre>
 *
 * <p>Note that a SipAddress can encapsulate either a SIP URL or an arbitrary
 * non-SIP URI. These are represented in this class by SipURL and String
 * attributes respectively, exactly one of which must (normally) be non-null.
 *
 * URL referes to SipURLImpl class (and parameters of such address are
 * stored in URL's parameters.
 * URI is kept as a string with no parameters.
 */
public class SipAddressImpl implements SipAddress {

   protected String displayName;
   protected SipURL URL;
   protected String URI;
   private boolean sipUrlFlag;
   protected HashMap parameterMap;

   public SipAddressImpl() {
      displayName = null;
      URL = null;
      URI = null;
      parameterMap = new HashMap();
   }

   public SipAddressImpl(SipAddress sipAddr) {
      if(sipAddr!=null) {
         displayName = sipAddr.getDisplayName();
         URI = sipAddr.getURI();
         SipURL tUrl = sipAddr.getSipURL();
         if(URI==null && tUrl!=null)
            sipUrlFlag = true;
         else
            sipUrlFlag = false;
         if(sipUrlFlag)
            URL = new SipURLImpl(tUrl);
         parameterMap = new HashMap();
         Iterator it = sipAddr.getParameterNames();
         while(it.hasNext()) {
            String k = (String) it.next();
            parameterMap.put(k, sipAddr.getParameter(k));
         }
      }
   }

   public SipAddressImpl(String dName, SipURL url) {
      displayName = dName;
      URL = url;
      sipUrlFlag = true;
      URI = null;
      parameterMap = new HashMap();
   }

   public SipAddressImpl(String dName, SipURL url, HashMap params) {
      displayName = dName;
      URL = url;
      sipUrlFlag = true;
      URI = null;
      parameterMap = new HashMap(params);
   }

   public SipAddressImpl(String dName, String uri) {
      displayName = dName;
      URL = null;
      sipUrlFlag = false;
      URI = uri;
      parameterMap = new HashMap();
   }

   public SipAddressImpl(String dName, String uri, HashMap params) {
      displayName = dName;
      URL = null;
      sipUrlFlag = false;
      URI = uri;
      parameterMap = new HashMap(params);
   }

    /**
     * Returns the display name associated with this SipAddress.
     * This may be null.
     */
    public String getDisplayName() {
      return displayName;
    }

    /**
     * Returns the SipURL associated with this instance, or null if this
     * SipAddress encapsulates a non-SIP URI.
     */
    public SipURL getSipURL() {
      return URL;
    }

    /**
     * Returns the non-SIP URI represented by this instance, or null if
     * this SipAddress encapsulates a SIP URL.
     */
    public String getURI() {
      return URI;
   }

   /**
    * Returns the value of the named parameter.
    */
   public String getParameter(String name) {
      if(parameterMap.containsKey(name)) {
         return (String) parameterMap.get(name);
      }
      else {
         return null;
      }
   }

   /**
    * Returns an iterator over all parameter names (Strings) set in this
    * SipAddress.
    */
   public Iterator getParameterNames() {
       Set n = parameterMap.keySet();
       return n.iterator();
   }


   //XXX: keep this interface immutable?
   public void setDisplayName(String name) {
      displayName = name;
   }

   public void setSipURL(SipURL url) {
      if(url!=null) {
         URL = url;
         sipUrlFlag = true;
      }
   }

   public void setURI(String uri) {
      if(uri!=null) {
         URI = uri;
         sipUrlFlag = false;
      }
   }

   /**
    *
    * ?: check with SipURL which does not allow
    * addition of new parameters
    */
   public void addParameter(String name, String value) {
      parameterMap.put(name, value);
   }

   public void removeParameter(String name) {
      if(parameterMap.containsKey(name)) {
         parameterMap.remove(name);
      }
   }
}
