
//package SipSImp;

import java.util.*;
import org.ietf.sip.*;
import java.net.InetAddress;

import java.security.AccessControlException;

/**
 * Factory methods for SipURLs, Contacts, SipAddresses, and Transactions.
 * @see SipTransaction#createRequest
 * @see SipRequest#createResponse
 */
public class SipFactoryImpl implements SipFactory {

   /**
    * Returns a new SIP transaction object with the specified parameters.
    * This can be used to crate a new SipRequest object which is used to
    * initiate new transactions.
    *
    * <p>The servlet engine generates a new Call-ID and CSeq number for the
    * transaction.
    *
    * NOTE: new local transactions are not supported at this time!
    *
    * @param from      the From header of the returned transaction
    * @param to        the To header of the returned transaction
    */
   public SipTransaction createSipTransaction(SipAddress from, SipAddress to) {

      // creation of new (local) transaction is not supported
      // by sipd server at this time
      throw new UnsupportedOperationException("SipFactory: creation of local transactions not supported by server");

      /* *** code below should remain unused for now ***
      String host = "null";
      try {
         host = InetAddress.getLocalHost().getHostName();
      }catch (Exception e){}
      Random rd = new Random();
      int cID = rd.nextInt();
      int cSeq = rd.nextInt(1000000000);
      String callID = "" + cID + "@";
      if(host!=null)
         callID += host;
      else
         callID += "127.0.0.1";
      if(from!=null && to!=null)
         return new SipTransactionImpl(callID, from, to, ""+cSeq);
      else
         return null;
      *** */
   }


   /**
    * Parses the argument as a SIP URL.
    * @param sipURL    a sip: url as defined in RFC 2543, section 2.
    * @throws ParseException   if the argument is not a valid SIP URL
    */
   public SipURL createSipURL(String sipURL) throws ParseException {
      SipURLImpl ret = null;
      if(sipURL!=null) {
         if(sipURL.indexOf("sip:")>0)
            ret = parseSipUrl(sipURL);
         else
            throw new ParseException("SipFactory: createSipURL: invalid URL string");
      }
      return ret;
   }

   /**
    * Creates a new SIP URL with the specified user and host and the
    * default SIP port number.
    * @param user  the user part if the SIP URL. This must be null for
    *              SipURLs used as request-URIs in REGISTER requests.
    * @param host  the host part of the SIP URL. This may be either a
    *              domain name or a dotted-decimal IP address.
    */
   public SipURL createSipURL(String user, String host) {
      return new SipURLImpl(user, null, host, -1, null);
   }


   /**
    * Creates a new SIP URL with the specified user, host and port number.
    * @param user  the user part if the SIP URL. This must be null for
    *              SipURLs used as request-URIs in REGISTER requests.
    * @param host  the host part of the SIP URL. This may be either a
    *              domain name or a dotted-decimal IP address.
    * @param port  the SIP URL port number
    */
   public SipURL createSipURL(String user, String host, int port) {
      return new SipURLImpl(user, null, host, port, null);
   }


   /**
    * Parses the specified string as a SipAddress, suitable for example,
    * for use as the From or To field of SIP messages.
    * @throws ParseException   if the argument is not a valid SipAddress
    * @see SipAddress
    */
   public SipAddress createSipAddress(String sipAddress) throws ParseException {
      boolean url_uri;
      SipAddressImpl ret;
      sipAddress.trim();
      String dName, addrParams, url;
      if(sipAddress.indexOf('<')>=0 && sipAddress.indexOf('>')>0) {
         url = sipAddress.substring(sipAddress.indexOf('<')+1, sipAddress.indexOf('>'));
         dName = sipAddress.substring(0, sipAddress.indexOf('<'));
         dName.trim();
         if(dName.length()==0)
            dName = null;
         try{
            addrParams = sipAddress.substring(sipAddress.indexOf('>')+1, sipAddress.length());
            addrParams.trim();
            if(addrParams.length()==0)
               addrParams = null;
         }catch(Exception e) {
            addrParams = null;
         }
      }
      else {
         url = sipAddress;
         dName = null;
         addrParams = null;
      }

      if(url==null || url.length()<3)
         throw new ParseException("SipFactory: could not parse URL");
      if(url.indexOf("sip:")>0) {
         //it's a sipURL
         SipURL URL = parseSipUrl(url);
         HashMap params = parseParams(addrParams);
         ret = new SipAddressImpl(dName, URL, params);
      }
      else {
         // url is a URI string
         HashMap params = parseParams(addrParams);
         ret = new SipAddressImpl(dName, url, params);
      }
      return ret;
   }


   /**
    * Creates a new Sipaddress with the specified SipURL and no display-name.
    */
   public SipAddress createSipAddress(SipURL sipURL) {
      return new SipAddressImpl(null, sipURL);
   }


   /**
    * Creates a new SipAddress with the specified SipURL and display-name.
    */
   public SipAddress createSipAddress(SipURL sipURL, String displayName) {
      return new SipAddressImpl(displayName, sipURL);
   }


   /**
    * Parses the argument as a Contact specification.
    * @throws ParseException   if the argument is not a valid Contact
    */
   public Contact createContact(String contact) throws ParseException {
      contact.trim();
      String s = null;
      if(contact.startsWith("Contact:")) {
         s = contact.substring("Contact:".length(),contact.length());
      }
      else if(contact.startsWith("m:")) {
         s = contact.substring("m:".length(),contact.length());
      }
      else
         s = contact;
      s = s.trim();
      String dName=null, addrParams=null, url=null;

      if(s.indexOf('<')>=0 && s.indexOf('>')>0) {
         url = s.substring(s.indexOf('<')+1, s.indexOf('>'));
         dName = s.substring(0, s.indexOf('<')).trim();
         if(dName.length()==0)
            dName = null;
         try{
            addrParams = s.substring(s.indexOf('>')+1, s.length()).trim();
            if(addrParams.length()==0)
               addrParams = null;
         }catch(Exception e) {
            addrParams = null;
         }
      }
      else {
         url = s;
         dName = null;
         addrParams = null;
      }

      SipAddressImpl addr=null;
      if(url==null || url.length()<3)
         throw new ParseException("SipFactory: createContact: exception parsing contact");
      if(url.indexOf("sip:")>0) {
         SipURL URL = parseSipUrl(url);
         addr = new SipAddressImpl(dName, URL);
      }
      else {
         // url is a URI string
         addr = new SipAddressImpl(dName, url);
      }
      HashMap params = parseParams(addrParams);

      float q = -1;
      String action=null, expires=null;

      if(params.containsKey("q")){
         String v = (String) params.get("q");
         try{
            q = Float.parseFloat(v);
         }catch(Exception e) {
            throw new ParseException("SipFactory: createContact: exception parsing contact: q-value");
         }
      }
      if(params.containsKey("action")) {
         action = (String) params.get("action");
      }
      if(params.containsKey("expires")) {
         expires = (String) params.get("expires");
      }

      ContactImpl c = new ContactImpl(q, action, expires, addr);
      return c;
   }


   /**
    * Creates a Contact with the specified SipURL and display-name.
    */
   public Contact createContact(SipURL url, String displayName) {
      SipAddressImpl sipAddr = new SipAddressImpl(displayName, url);
      return new ContactImpl((float) 0, "", "", sipAddr);
   }


   /**
    * Creates a Contact with the specified URI and display-name. The URI
    * need not be a SIP URL.
    */
   public Contact createContact(String uri, String displayName) {
      SipAddressImpl sipAddr = new SipAddressImpl(displayName, uri);
      return new ContactImpl((float) 0, "", "", sipAddr);
   }


   /**
    * Creates a Contact with details copied from the specified SIP address.
    */
   public Contact createContact(SipAddress addr) {
      return new ContactImpl((float) 0, "", "", addr);
   }


   public static SipURLImpl parseSipUrl(String url) throws ParseException {
      if(!url.startsWith("sip:") || url.length()<4)
         throw new ParseException("SipFactory: invalid URL to parse");

      String user=null, host=null, passwd=null;
      String paramStr=null;
      int port = -1;

      // separate parameters
      if(url.indexOf(';')>0) {
         paramStr = url.substring(url.indexOf(';')+1, url.length()).trim();
         if(paramStr.length()<1)
            paramStr = null;
         url = url.substring(0, url.indexOf(';')).trim();
      }
      url = url.substring(url.indexOf("sip:")+4,url.length()).trim();
      if(url.length()<1)
         throw new ParseException("SipFactory: invalid URL to parse");
      if(url.indexOf('@')>0) {
         String left = url.substring(0,url.indexOf('@'));
         String right = url.substring(url.indexOf('@')+1, url.length());
         if(left.length()<1 || right.length()<1)
            throw new ParseException("SipFactory: could not parse URL");
         if(left.indexOf(':')>0) {
            user = left.substring(0, left.indexOf(':'));
            passwd = left.substring(left.indexOf(':')+1, left.length());
         }
         else
            user = left;
         if(right.indexOf(':')>0) {
            try {
               host = right.substring(0, right.indexOf(':'));
               String p = right.substring(right.indexOf(':')+1, right.length());
               port = Integer.parseInt(p);
            }catch(Exception ex) {
               throw new ParseException("SipFactory: could not parse URL");
            }
         }
         else
            host = right;
      }
      else {
         if(url.indexOf(':')>0) {
            try{
               host = url.substring(0, url.indexOf(':')).trim();
               String p = url.substring(url.indexOf(':')+1, url.length());
               port = Integer.parseInt(p);
            }catch(Exception ex) {
               throw new ParseException("SipFactory: could not parse URL");
            }
         }
         else {
            host = url;
         }
      }
      HashMap params = parseParams(paramStr);

      SipURLImpl ret = new SipURLImpl(user, passwd, host, port, params);
      return ret;
   }


   private static HashMap parseParams(String sp) throws ParseException {
      HashMap ret = new HashMap();
      if(sp==null)
         return ret;
      StringTokenizer tk = new StringTokenizer(sp, ";");
      while(tk.hasMoreTokens()) {
         String tmp = tk.nextToken().trim();
         if(tmp.length()>1) {
            if(tmp.indexOf('=')>0) {
               String key = tmp.substring(0, tmp.indexOf('=')).trim();
               String val = tmp.substring(tmp.indexOf('=')+1, tmp.length()).trim();
               ret.put(key, val);
            }
            else
               ret.put(tmp, null);
         }
      }
      return ret;
   }
}
