
//package SipSImplementation;

import java.util.*;
import java.net.InetAddress;
import java.io.*;

// requires Java 1.3
import javax.naming.directory.BasicAttribute;

import org.ietf.sip.*;


/**
 * Base interface for SipRequest and SipResponse.
 *
 * <p>As SIP entities can receive and send requests as well as responses,
 * there is a need for SIP servlets to be able to set and get headers,
 * content, and other information in both requests and response. This
 * interface reflects the commonality between all SIP messages.
 */
public class SipMessageImpl implements SipMessage {

   protected String method;
   protected String protocol;
   protected InetAddress remoteHost;
   protected int serverPort;
   protected String serverProtocol;
   protected SipTransactionImpl transaction;

   // headers stored in Vector of SipHeaderUnit
   protected Vector headers;
   protected String content;
   protected String contentType;
   protected int contentLen;

   // original message from sipd that
   // was used to create this message;
   protected message_t sipdMsg;

   public SipMessageImpl() {
      method = null;
      protocol = null;
      remoteHost = null;
      serverPort = 0;
      serverProtocol = null;
      transaction = null;
      headers = new Vector();
      content = null;
      contentType = null;
      contentLen = 0;
      sipdMsg = null;
   }


   public SipMessageImpl(SipMessageImpl msg) {
      method = new String(msg.getMethod());
      protocol = new String(msg.getProtocol());
      serverPort = msg.getServerPort();
      serverProtocol = new String(msg.getServerProtocol());
      if(msg.getRemoteHost() != null) {
         try{
            remoteHost = InetAddress.getByName(msg.getRemoteHost());
         } catch(Exception ex) {
            remoteHost = null;
         }
      }
      transaction = msg.transaction;
      headers = new Vector(msg.headers);
      content = msg.content!=null ? new String(msg.content):null;
      contentType = msg.getContentType()!=null ? new String(msg.getContentType()):null;
      contentLen = msg.getContentLength();
      sipdMsg = msg.sipdMsg;
   }


   public SipMessageImpl(String m, String p, String rHost, int sPort, String sP,
                          SipTransactionImpl t, List hdrs, String c, String cType, int cLen) {
      method = m;
      protocol = p;
      remoteHost = null;
      serverPort = 0;
      serverProtocol = sP;
      if(rHost != null) {
         try{
            serverPort = sPort;
            remoteHost = InetAddress.getByName(rHost);
         } catch(Exception ex) {
            remoteHost = null;
            serverPort = 0;
         }
      }
      transaction = t;
      headers = new Vector();
      // hdrs stores header lines (<header>: <value>)
      if(hdrs!=null) {
         Iterator it = hdrs.iterator();
         while(it.hasNext()) {
            String elem = (String) it.next();
            int ind = elem.indexOf(":");
            if(ind>1 && elem.length()>ind+2) {
               String hd = elem.substring(0, ind);
               String v = elem.substring(ind+2, elem.length()).trim();
               headers.add(new SipHeaderUnit(hd, v));

               /*** alternative implementation with HashMap ***
               if(headers.contains(hd)) {
                  Vector vals = (Vector) headers.get(hd);
                  vals.add(v);
                  headers.put(hd, vals);
               }
               else {
                  Vector val = new Vector();
                  val.add(v);
                  headers.put(hd, val);
               }**/
            }
         }
      }
      content = c;
      contentType = cType;
      contentLen = cLen;
      sipdMsg = null;
   }


   /**
    * Returns the SIP method of this message. For requests this is the
    * value used in the request line and for responses it is taken from
    * the CSeq header field, which echoes the method used in the
    * corresponding request.
    */
   public String getMethod() {
      return method;
   }

   /**
    * Returns the name and version of the protocol of this message.
    * This is of the form &lt;protocol> "/" &lt;major-version-number> "."
    * &lt;minor-version-number>, e.g. "SIP/2.0".
    */
   public String getProtocol() {
      return protocol;
   }

   /**
    * Returns the IP address of the host sending this message, or the
    * loopback address if this message was created locally.
    *
    * <p>Note that the address returned is that of the previous hop,
    * which will generally not be that of the host originating the
    * request.
    */
   public String getRemoteAddr() {
      if(remoteHost!=null) {
         return remoteHost.getHostAddress();
      }
      else
         return "127.0.0.1";
   }

   /**
    * Returns the fully qualified domain name of the host sending the
    * message, or null if this request was created locally. The returned
    * string may be a dotted-decimal IP address if the server cannot or
    * chooses not to resolve IP addresses.
    */
   public String getRemoteHost() {
      String ret = null;
      if(remoteHost!=null) {
         try {
            ret = remoteHost.getHostName();
         } catch(Exception ex) {
            ret = null;
         }
      }
      else
         ret = null;
      return ret;
   }

   /**
    * Returns the host name used to access this server. May be either a
    * fully qualified domain name or a dotted decimal IP address.
    */
   public String getServerName() {
      String ret = null;
      try {
         ret = InetAddress.getLocalHost().getHostAddress();
      }catch(Exception ex) {}
      return ret;
   }

   /**
    * Returns the port number on which the message was received, or -1 if
    * this message was created locally.
    */
   public int getServerPort(){
      return serverPort;
   }

   /**
    * Returns the name of the transport protocol used to receive this
    * message, e.g. "TCP" or "UDP". Returns null for locally created
    * messages.
    */
   public String getServerProtocol(){
      return serverProtocol;
   }



   // ~~~~ headers ~~~~

   /**
    * Returns the transaction associated with this message.
    */
   public SipTransaction getTransaction(){
      return transaction;
   }


   /**
    * Returns the value of the specified header field.
    */
   public String getHeader(String name){
      SipHeaderUnit comp = new SipHeaderUnit(name, null);
      if(headers.contains(comp)) {
         SipHeaderUnit v = (SipHeaderUnit) headers.elementAt(headers.indexOf(comp));
         return v.value;
      }
      else
         return null;
   }


   /**
    * Returns an iterator over all headers in the message (Strings).
    */
   public Iterator getHeaderNames(){
      Iterator it = headers.iterator();
      if(it.hasNext()) {
         ArrayList ret = new ArrayList();
         while(it.hasNext()) {
            SipHeaderUnit u = (SipHeaderUnit) it.next();
            ret.add(new String(u.header));
         }
         return ret.iterator();
      }
      else
         return null;
   }


   /**
    * Returns an iterator over the values of all headers with the
    * specified name.
    * @param name  the header name
    */
   public Iterator getHeaders(String name){
      Iterator it = headers.iterator();
      if(it.hasNext()) {
         ArrayList ret = new ArrayList();
         SipHeaderUnit comp = new SipHeaderUnit(name, null);
         while(it.hasNext()) {
            SipHeaderUnit u = (SipHeaderUnit) it.next();
            if(u.equals(comp))
               ret.add(u.value);
         }
         return ret.iterator();
      }
      else
         return null;
   }


   /**
    * Sets a header field with the specified name and value. If a header
    * with the same name existed it is overwritten.
    * @param name  the header name
    * @param value the header value
    */
   public void setHeader(String name, String value){
      SipHeaderUnit nh = new SipHeaderUnit(name, value);
      if(headers.contains(nh)) {
         int i = headers.indexOf(nh);
         headers.setElementAt(nh, i);
         if(sipdMsg!=null) {
            try {
               parser_header_t ph = headerStrToObj(name, value);
               sipdMsg.p.headers.set(i, ph);
            }catch(Exception ex) {
               System.out.println(ex.toString());
            }
         }
      }
      else {
         this.addHeader(name, value);
      }
   }


   /**
    * Adds a header field with the specified name and value.
    * @param name  the header name
    * @param value the header value
    */
   public void addHeader(String name, String value){
      SipHeaderUnit nh = new SipHeaderUnit(name, value);

      // sipd message synchronization
      parser_header_t ph = null;
      try {
         ph = headerStrToObj(name, value);
      }catch(Exception ex) {
         System.out.println(ex.toString());
         ph = null;
      }

      if(headers.contains(nh)) {
         // add another header before one in the message
         // that has same header name
         int i = headers.indexOf(nh);
         headers.insertElementAt(nh, i);
         if(sipdMsg!=null && ph!=null) {
            try {
               sipdMsg.p.headers.add(i, ph);
            }catch(Exception ex) { System.out.println(ex.toString()); }
         }
      }
      else {
         // insert new header at appropriate place in the message
         int ni = getHeaderGroupID(name);
         Iterator it = headers.iterator();
         boolean done = false;
         while(it.hasNext()) {
            // search for insertion place in the list of headers
            SipHeaderUnit hd = (SipHeaderUnit) it.next();
            int hi = getHeaderGroupID(hd.header);
            int put = headers.indexOf(hd);
            if(hi<ni)
               continue;
            else if(hi==ni && (hd.header).compareToIgnoreCase(name) >= 0) {
               headers.insertElementAt(nh, put);
               if(sipdMsg!=null && ph!=null)
                  try {
                     sipdMsg.p.headers.add(put, ph);
                  }catch(Exception ex) {}
               done = true;
               break;
            }
            else if(hi>ni) {
               headers.insertElementAt(nh, put);
               if(sipdMsg!=null && ph!=null)
                  try {
                     sipdMsg.p.headers.add(put, ph);
                  }catch(Exception ex) {}
               done = true;
               break;
            }
         }
         if(!done) {
            headers.add(new SipHeaderUnit(name, value));
            if(sipdMsg!=null && ph!=null)
               sipdMsg.p.headers.add(ph);
         }
      }
   }


   /**
    * Returns the Call-ID of this transaction.  Equivalent to
    * <code>getTransaction().getCallID()</code>.
    * @see SipTransaction#getCallID
    */
   public String getCallID(){
      return transaction.getCallID();
   }


   /**
    * Returns the parsed value of the From header field.
    * Equivalent to <code>getTransaction().getFrom()</code>.
    * @see SipTransaction#getFrom
    */
   public SipAddress getFrom(){
      return transaction.getFrom();
   }


   /**
    * Returns the parsed value of the To header field.
    * Equivalent to <code>getTransaction().getTo()</code>.
    * @see SipTransaction#getTo
    */
   public SipAddress getTo(){
      return transaction.getTo();
   }


   /**
    * Returns a list of all Contacts in this message. This is a List of
    * Contact objects corresponding to the parsed value of all Contact
    * headers in this message.
    */
   public List getContacts() {
      SipHeaderUnit h = new SipHeaderUnit(SipConstants.CONTACT, null);
      if(headers.contains(h)) {
         Iterator it = headers.iterator();
         ArrayList list = new ArrayList();
         // contact string parsed into contact object
         // using parsing method from SipFactory
         SipFactory f = new SipFactoryImpl();
         while(it.hasNext()) {
            SipHeaderUnit orig = (SipHeaderUnit) it.next();
            if(orig.equals(h)) {
               String ct = new String(orig.value);
               try {
                  Contact c = f.createContact(ct);
                  list.add(c);
               }catch(Exception ex) {}
            }
         }
         return list;
      }
      else
         return null;
   }


   /**
    * Sets the set of Contact objects for this message.
    * @param contacts  a list of Contact objects.
    */
   public void setContacts(List contacts){
      // new list of contacts replaces (sets) existing
      // contact headers in the message
      if(contacts.size() > 0) {
         SipHeaderUnit h = new SipHeaderUnit(SipConstants.CONTACT, null);
         // remove all current contact headers
         while(headers.contains(h)) {
            int i = headers.indexOf(h);
            headers.remove(h);
            if(sipdMsg!=null) {
               try {
                  sipdMsg.p.headers.remove(i);
               }catch(Exception ex) {}
            }
         }
         Iterator it = contacts.iterator();
         while(it.hasNext()) {
            Object o = it.next();
            if(o instanceof Contact) {
               Contact c = (Contact) o;
               this.addHeader(SipConstants.CONTACT, c.toString());
            }
         }
      }
   }


   // ~~~~ content ~~~~

   /**
    * Returns the length in number of bytes of the content part of this
    * message. This directly reflects the value of the Content-Length header
    * field.
    */
   public int getContentLength(){
      return contentLen;
   }


   /**
    * Returns the value of the Content-Type header field.
    */
   public String getContentType(){
      return contentType;
   }

   /**
    * Sets the value of the Content-Type header field.
    * @param type  the MIME type of the message body
    */
   public void setContentType(String type){
      contentType = type;
      if(sipdMsg!=null)
         sipdMsg.body.content_type = type;
      // update the header
      this.setHeader("Content-Type", contentType);
   }

   /**
    * Returns the content of this message as a byte array.
    */
   public byte[] getContent(){
      return content.getBytes();
   }

   /**
    * Attempts to parse the content of this message. The type of object
    * returned will depend on the Content-Type of the message. A server is
    * required at a minimum to be able to parse SDP [RFC 2327]. A future
    * version of this API will include an interface for accessing SDP and
    * pin down what this method returns for SDP.
    *
    * @return an object representing the body of the message, or null if
    *      the content length is 0.
    * @throws ParseException   if the server does not know how to parse
    *      the content type of the body.
    *
    * Currently, the method will parse SDP. Parsed SDP is returned
    * as a Vector of BasicAttribute types (where attribute ID is SDP type
    * and attribute value is SDP value for the type).
    */
   public Object getParsedContent() throws ParseException{
      String cType = new String(contentType);
      cType.trim();
      Vector ret = null;
      if(cType.equalsIgnoreCase("application/sdp")) {
         String c = new String(content);
         BufferedReader bf = new BufferedReader(new StringReader(c));
         // vector of SDP lines
         ret = new Vector();
         try {
            String l = bf.readLine();
            while(l!=null) {
               int ind = l.indexOf("=");
               if(ind > 0) {
                  // ind should be 1
                  String type = l.substring(0,ind);
                  String val = l.substring(ind+1, l.length());
                  BasicAttribute attr = new BasicAttribute(type, true);
                  attr.add(val);
                  ret.add(attr);
               }
               l = bf.readLine();
            }
         } catch(IOException ex) {}
      }
      else {
         throw new ParseException("SipMessage: Unsupported content type");
      }
      return ret;
   }

   /**
    * Returns an object representing the media description contained within
    * the body of this message, or null if the body is empty or no recognized
    * media format was found.
    *
    * Returns SDP media description types and their values
    * as they are presented in the message body.
    *
    * Each SDP media description line from msg body is put as an
    * entry in a vector. Each entry is of type BasicAttribute
    */
   public Object getMediaDescription() throws ParseException{
      String cType = new String(contentType);
      cType.trim();
      Vector ret = null;
      if(cType.equalsIgnoreCase("application/sdp")) {
         String c = new String(content);
         BufferedReader bf = new BufferedReader(new StringReader(c));
         // vector of SDP lines
         ret = new Vector();
         try {
            String l = bf.readLine();
            while(l!=null) {
               int ind = l.indexOf("=");
               if(ind > 0) {
                  // ind should be 1
                  String type = l.substring(0,ind);
                  // check for SDP types decribing media
                  if(l.equals("m") || l.equals("i") || l.equals("c") || l.equals("b")
                     || l.equals("k") || l.equals("a")) {
                     String val = l.substring(ind+1, l.length());
                     BasicAttribute attr = new BasicAttribute(type, true);
                     attr.add(val);
                     ret.add(attr);
                  }
               }
               l = bf.readLine();
            }
         } catch(IOException ex) {}
      }
      else {
         throw new ParseException("SipMessage: Unsupported content type");
      }
      return ret;
   }

   /**
    * Sets the body of the content to the specified byte array.
    * @param buf   the raw bytes constituting the message body.
    */
   public void setContent(byte[] buf){
      String cnt = new String(buf);
      content = cnt;
      contentLen = content.length();
      if(sipdMsg!=null) {
         sipdMsg.body.content = cnt;
         sipdMsg.body.content_length = cnt.length();
      }
      // update the header
      this.setHeader("Content-Length", ""+contentLen);
   }



   // internal methods (not defined by Sips API
   /**
    * returns object corresponding to sipd message type
    */
   public message_t getSipdMsg() {
      return sipdMsg;
   }

   /**
    * sets sipdMsg of this message object to another
    * java representation of sipd's message_t
    * this method must be used carefully (a potential cause
    * of inconsistencies between the object and its underlying
    * sipd message_t type)
    */
   public void setSipdMsg(message_t msg) {
      sipdMsg = msg;
   }


   /**
    * internal method for determining type of header
    * (general, request, entity, etc..) associate with
    * header name
    * used when new headers are inserted in the message
    */
   protected int getHeaderGroupID(String header) {
      int ret = -1;
      if(header==null)
         return -1;
      if(header.length() <= 1)
         return -1;
      // general header
      if( header.equalsIgnoreCase("Accept") || header.equalsIgnoreCase("Accept-Encoding")
         || header.equalsIgnoreCase("Accept-Language") || header.equalsIgnoreCase("Call-ID")
         || header.equalsIgnoreCase("Contact") || header.equalsIgnoreCase("CSeq")
         || header.equalsIgnoreCase("Date") || header.equalsIgnoreCase("Encryption")
         || header.equalsIgnoreCase("Expires") || header.equalsIgnoreCase("From")
         || header.equalsIgnoreCase("Record-Route") || header.equalsIgnoreCase("Timestamp")
         || header.equalsIgnoreCase("To") || header.equalsIgnoreCase("Via") )
         ret = 1;
      // request header
      else if( header.equalsIgnoreCase("Authorization") || header.equalsIgnoreCase("Contact")
              || header.equalsIgnoreCase("Hide") || header.equalsIgnoreCase("Max-Forwards")
              || header.equalsIgnoreCase("Organization") || header.equalsIgnoreCase("Priority")
              || header.equalsIgnoreCase("Proxy-Authorization") || header.equalsIgnoreCase("Proxy-Require")
              || header.equalsIgnoreCase("Route") || header.equalsIgnoreCase("Require")
              || header.equalsIgnoreCase("Response-Key") || header.equalsIgnoreCase("Subject")
              || header.equalsIgnoreCase("User-Agent") )
         ret = 2;
      // response header
      else if( header.equalsIgnoreCase("Allow") || header.equalsIgnoreCase("Proxy-Authenticate")
              || header.equalsIgnoreCase("Retry-After") || header.equalsIgnoreCase("Server")
              || header.equalsIgnoreCase("Unsupported") || header.equalsIgnoreCase("Warning")
              || header.equalsIgnoreCase("WWW-Authenticate") )
         ret = 3;
      // entity header
      else if( header.equalsIgnoreCase("Content-Encoding") || header.equalsIgnoreCase("Content-Length")
              || header.equalsIgnoreCase("Content-Type") )
         ret = 4;
      else
         // unrecognized headers are supposed assumed to belong to
         // group of entity headers
         ret = 4;

      return ret;
   }




   /**
    * methods for transfering parser header and element objects
    * (from sipd) into a single string value associated with a
    * given string element
    * used also by SipSContainer
    */
   public static String headerObjToStr(parser_header_t hdr) {
      String ret = "";
      if(hdr.elements==null)
         return ret;
      int cnt = hdr.elements.size();
      for(int i=0; i<cnt; i++) {
         String cval = new String();
         parser_element_t pe = (parser_element_t) hdr.elements.get(i);
         if(pe.element!=null && pe.element.length()>0) {
            cval = pe.element;
         }
         else {
            cval = (pe.value!=null ? pe.value : "");
            int pcnt = (pe.parameters!=null ? pe.parameters.size() : -1);
            for(int j=0; j<pcnt; j++) {
               parser_parameter_t pr = (parser_parameter_t) pe.parameters.get(j);
               if(pr.attribute != null) {
                  cval += ";"+pr.attribute;
                  cval += ((pr.value==null || pr.value.length()==0)? "" : "="+pr.value);
               }
            }
         }
         if(ret.equals(""))
            ret = cval;
         else
            ret += ", " + cval;
      }
      return ret;
   }


   public static parser_header_t headerStrToObj(String name, String hdr) {
      if(name==null || hdr==null)
         return null;
      parser_header_t ret = new parser_header_t();
      ret.name = new String(name);
      ret.elements = new LinkedList();
      // processe each element (where elements are comma-separated)
      StringTokenizer tk = new StringTokenizer(hdr, ",");
      while(tk.hasMoreTokens()) {
         String str = tk.nextToken().trim();
         parser_element_t pe = new parser_element_t();
         pe.element = new String(str);
         pe.comment = "";
         pe.parameters = new LinkedList();
         String prs = null;      // string with parameters
         if(str.indexOf("<")>0 && str.indexOf(">")>0) {
            // check for Contact or URL delimenters: "<" ">"
            // treat such URL as a
            pe.value = str.substring(0,str.indexOf(">"));
            pe.value += ">";
            try{
               prs = str.substring(str.indexOf(">")+1, str.length());
            } catch(Exception ex) {
               prs = null;
            }
            if(prs!=null && prs.indexOf(";")<0)
               prs = null;
         }
         else if(str.indexOf(";")>0) {
            pe.value = str.substring(0, str.indexOf(";")).trim();
            prs = str.substring(str.indexOf(";"), str.length());
         }
         else {
            // element with single value
            pe.value = str;
            prs = null;
         }
         if(prs!=null) {
            StringTokenizer tkPrs = new StringTokenizer(prs, ";");
            while(tkPrs.hasMoreTokens()) {
               String param = tkPrs.nextToken();
               parser_parameter_t pr = new parser_parameter_t();
               if(param.length()<1)
                  continue;
               if(param.indexOf("=")>0){
                  pr.attribute = param.substring(0, param.indexOf("=")).trim();
                  try {
                     pr.value =  param.substring(param.indexOf("=")+1, param.length()).trim();
                  }catch(Exception ex) {
                     pr.value = null;
                  }
               }
               else {
                  pr.attribute = param;
                  pr.value = null;
               }
               pe.parameters.add(pr);
            }
         }
         ret.elements.add(pe);
      }
      return ret;
   }
}

