
//package SipSImp;

import java.util.*;
import java.net.*;
import org.ietf.sip.*;

import java.security.AccessControlException;

/**
 * Represents SIP requests. This class provides access to all parts of SIP
 * requests, both for incoming and outgoing messages.
 */
public class SipRequestImpl extends SipMessageImpl implements SipRequest {

   protected String reqURI;
   // SipURL derived from request-URI
   protected SipURLImpl url;

   protected String authType;
   protected boolean authenticated;   // is the request authenticated
   protected String authUser;

   // true if request was received
   // false is created locally (currently, no locally
   // created requests are allowed)
   protected boolean remoteOrigin;

   public SipRequestImpl() {
      super();
      reqURI = null;
      url = null;
      authType = null;
      authenticated = false;
      authUser = null;
   }

   public SipRequestImpl(String uri, String auth, boolean aValid,
                         String aUser,  boolean orig, SipMessageImpl msg) {
      super(msg);
      reqURI = uri;
      if(uri!=null)
         url = reqUriToSipUrl(uri);
      authType = auth;
      authenticated = aValid;
      authUser = aUser;
      remoteOrigin = orig;
   }

   public SipRequestImpl(String uri, String auth, boolean aValid, String aUser, boolean orig,
                        String m, String p, String rHost, int sPort, String sP,
                        SipTransactionImpl t, List hdrs, String c, String cType, int cLen) {
      super(m, p, rHost, sPort, sP, t, hdrs, c, cType, cLen);
      reqURI = uri;
      if(uri!=null)
         url = reqUriToSipUrl(uri);
      authType = auth;
      authenticated = aValid;
      authUser = aUser;
      remoteOrigin = orig;
   }


   /**
    * Returns the SIP URL as it appeared in the request line of this request,
    * or null if this object was created locally.
    */
   public SipURL getRequestURI() {
      return url;
   }

   /**
    * Returns the authentication scheme used for this message. This is the
    * value of the <em>auth-scheme</em> part of the Authorization header,
    * and can take the values "basic", "digest", and "pgp".
    */
   public String getAuthType() {
      return authType;
   }

   /**
    * Returns the identity of the caller as verified by the authentication
    * protocol, or null if the request hasn't been authenticated.
    */
   public String getRemoteUser() {
      if(authenticated)
         return authUser;
      else
         return null;
   }


   /**
    * Add a Record-Route header for this proxy to the request.
    * This is only relevant for requests being proxied.
    * @param recordIt  true if a Record-Route header should be added to
    *                  the message for this proxy, and false if it should
    *                  be removed.
    * @return  true if the operation succeeded, and false otherwise
    */
   public boolean recordRoute(boolean recordIt) {
      // recordRoute based on url in Request-URI
      boolean ret = false;
      if(url==null)
         return false;
      String rrStr = "<" + url.toString() + ">";
      String rrVal = this.getHeader("Record-Route");
      if(rrVal == null) {
         this.addHeader("Record-Route", rrStr);
         ret = true;
      }
      else {
         // append at the front of current value
         if(rrVal.length()>0)
            rrStr += ", ";
         rrStr += rrVal;
         this.setHeader("Record-Route", rrStr);
         ret = true;
      }
      return ret;
   }


   /**
    * Creates a response for this request. The request object will have
    * From, To, Call-ID, CSeq, Via etc. initialized from this request
    * object, and the servlet shouldn't attempt to set these.
    */
   public SipResponse createResponse() {
      message_t sMsg = new message_t();
      sMsg.type = sipdMsg.type;
      sMsg.handled = sipdMsg.handled;
      sMsg.p = new parser_t();
      sMsg.body = new body_t();
      sMsg.body.content_length = 0;
      sMsg.source = sipdMsg.source;
      sMsg.source_port = sipdMsg.source_port;
      sMsg.p.type = sipdMsg.p.type;
      sMsg.p.method = new String(sipdMsg.p.method);
      sMsg.p.URI_string = sipdMsg.p.URI_string!=null ? new String(sipdMsg.p.URI_string):null;
      sMsg.p.URI = sipdMsg.p.URI;
      sMsg.p.status = 0;
      sMsg.p.reason = null;
      sMsg.p.protocol = new String(sipdMsg.p.protocol);
      sMsg.p.version = new String(sipdMsg.p.version);
      sMsg.p.header_body = sipdMsg.p.header_body!=null ? new String(sipdMsg.p.header_body):null;
      sMsg.p.headers = new LinkedList();
      sMsg.p.header_length = 0;

      SipMessageImpl msg = new SipMessageImpl(this);
      // clear headers
      msg.headers = new Vector();
      msg.content = null;
      msg.contentType = null;
      msg.contentLen = 0;
      msg.sipdMsg = sMsg;

      // locally created response
      SipResponseImpl response = new SipResponseImpl(0, "", null, false, msg);
      int hdrs = headers.size();
      for(int i=0; i<hdrs; i++) {
         SipHeaderUnit sh = (SipHeaderUnit) headers.elementAt(i);
         String name = (sh.header!=null ? new String(sh.header) : null);
         if(name==null)
            continue;
         String value = (sh.value!=null ? new String(sh.value) : null);
         int ins = getHeaderGroupID(name);
         // general and response header
         if(ins==1 || ins==3) {
            response.addHeader(name, value);
         }
         // entity header
         else if(ins==4) {
            if(name.equalsIgnoreCase("Content-Length")
               || name.equalsIgnoreCase("Content-Type"))
               continue; // do not add these headers
            else
               response.addHeader(name, value);
         }
      }
      response.addHeader("Content-Length", ""+0);
      return response;
   }


   /**
    * Send this message to the specified callee. This method is used both
    * for initiating new requests and to forward received requests in
    * proxies.
    *
    * @param url   the address to which to send this request. This will
    *              appear in the request URI of the message.
    * @return  a token which can be used for matching responses to this
    *          request. Should only be used for testing reference equality.
    */
   public Object send(SipURL url) {

      Random rd = new Random();
      Integer ret = new Integer(rd.nextInt() % 1000000);

      // wrapper for native function
      SipRequestWrapper rw = new SipRequestWrapper();

      String host = url.getHost();
      int port = url.getPort();
      if(host==null)
         return null;
      // use associate request_t from sip transaction
      // and message_t from sip message
      request_t r = transaction.getSipdTransaction();
      if(r==null)
         return null;
      if(sipdMsg==null)
         return null;

      // use request_t structure to mark destination
      r.sin = host;
      r.sin_port = port>0 ? port : 5060;
      r.via.sin = r.sin;
      r.via.sin_port = r.sin_port;

      String viaHost = null;
      try {
         InetAddress adr = InetAddress.getLocalHost();
         viaHost = adr.getHostName();
      }catch(Exception ex) {}
      if(viaHost==null)
         return null;

      String viaProxy = new String(protocol);
      if(r.type==1)
         viaProxy += "/UDP";
      else if(r.type==2)
         viaProxy += "/TCP";
      viaProxy += " " + viaHost;
      this.addHeader("Via", viaProxy);

      String reqUri = method + " " + url.toString() + " " + protocol;
      r.p.URI_string = reqUri;

      // copy headers to request_t
      r.headers = new LinkedList(this.headers);

      r.response_body.content = content;
      r.response_body.content_length = contentLen;
      r.response_body.content_type = contentType;

      // put SipURL into uri_t format
      uri_t uri = new uri_t();
      uri.user = url.getUser();
      uri.password = url.getPassword();
      uri.host = url.getHost();
      uri.port = url.getPort();
      uri.parameters = new HashMap();
      Iterator it = url.getParameterNames();
      while(it.hasNext()) {
         String p = (String) it.next();
         String v = url.getParameter(p);
         uri.parameters.put(p, v);
      }
      r.p.URI = uri;

      // send the message
      rw.processSend(r, sipdMsg, uri);
      return ret;
   }



   /**
    * internal method:
    * returns SipURL object associated with sip URL
    * found in request-URI
    */
   private SipURLImpl reqUriToSipUrl(String reqUri) {
      String uri = reqUri.trim();
      SipURLImpl ret = null;
      // parsing Req-URI: method SP url SP version
      int c1 = uri.indexOf(' ');
      int c2 = uri.lastIndexOf(' ');
      if(c1<0 || c2<=c1)
         return null;
      String url = uri.substring(c1+1, c2);
      try{
         SipFactoryImpl fct = new SipFactoryImpl();
         ret = fct.parseSipUrl(url);
      }catch(Exception ex) {
         ret = null;
      }
      return ret;
   }
}
