
//package SipSImplementation;

import java.util.*;
import java.net.*;
import org.ietf.sip.*;


/**
 * Represents SIP responses.
 */
public class SipResponseImpl extends SipMessageImpl implements SipResponse {

   protected int status;
   protected String reasonPhrase;
   protected Object requestToken;

   /* true if response was received by servlet engine from
    * the network(from sipd), false if response was created
    * locally (inside servlet engine)
    */
   protected boolean remoteOrigin;

   public SipResponseImpl() {
      super();
      status = -1;
      reasonPhrase = null;
      requestToken = null;
      remoteOrigin = false;
   }

   public SipResponseImpl(int stat, String reas, Object reqTok, boolean orig,
                          String m, String p, String rHost, int sPort, String sP,
                          SipTransactionImpl t, List hdrs, String c, String cType, int cLen) {
      super(m, p, rHost, sPort, sP, t, hdrs, c, cType, cLen);
      status = stat;
      reasonPhrase = reas;
      requestToken = reqTok;
      remoteOrigin = orig;
   }

   public SipResponseImpl(int stat, String reas, Object reqTok, boolean orig, SipMessageImpl msg) {
      super(msg);
      status = stat;
      reasonPhrase = reas;
      /*
      if(sipdMsg!=null) {
         sipdMsg.p.status = stat;
         sipdMsg.p.reason = reas;
      }*/
      requestToken = reqTok;
      remoteOrigin = orig;
   }

   /**
    * Returns the SIP status code of this response. This is a three
    * digit (decimally speaking) integer, usually one of the RFC 2543
    * defined SipConstants.SC_*.
    * @see SipConstants
    */
   public int getStatus() {
      return status;
   }

   /**
    * Returns the reason phrase of this response.
    */
   public String getReasonPhrase() {
      return reasonPhrase;
   }

   /**
    * Returns the "request token" associated with this response.
    * The request token identifies a particular request and is useful only
    * for matching up responses to requests. This is particularly useful
    * in parallel, forking proxies.
    */
   public Object getRequestToken() {
      return requestToken;
   }

   /**
    * Sets the status code and message for this response.
    * @param statusCode    the SIP status code, one of SipConstants.SC_*
    * @param reason        the human readable reason phrase
    * @see SipConstants
    */
   public void setStatus(int code, String reason) {
      status = code;
      reasonPhrase = reason;
      if(sipdMsg!=null) {
         sipdMsg.p.status = code;
         sipdMsg.p.reason = reason;
      }
   }

   /**
    * Sets the status code for this response.
    * @param statusCode    the SIP status code, one of SipConstants.SC_*
    * @param reason        the human readable reason phrase
    * @see SipConstants
    */
   public void setStatus(int code) {
      status = code;
      reasonPhrase = null;
      if(sipdMsg!=null) {
         sipdMsg.p.status = code;
         sipdMsg.p.reason = null;
      }
   }

   /**
    * Causes this response object to get transmitted. The protocol and
    * destination is chosen according to the value of the Via headers of
    * this object.
    *
    * <p>If this server is acting as a proxy for this transaction,
    * the Via headers were in the response when it was received and this
    * method will cause it to strip off the top-most Via address in
    * accordance with RFC 2543 and transmit the request to the next hop.
    * Otherwise, if the response was created by this servlet from a request
    * object, then the Via headers were derived from that request.
    */
   public void send() {
      SipResponseWrapper rw = new SipResponseWrapper();

      // use associate request_t from sip transaction
      // and message_t from sip message
      request_t r = transaction.getSipdTransaction();
      if(r==null)
         return;
      if(sipdMsg==null)
         return;
      if(status<=0)
         return;

      // >> testing
      remoteOrigin = false;

      String host = null;
      int port = 5060;
      if(remoteOrigin) {
         // remove top Via
         int i = headers.indexOf(new SipHeaderUnit("Via", null));
         if(i<0)
            return;
         headers.removeElementAt(i);
         try {
            sipdMsg.p.headers.remove(i);
         }catch(Exception ex) {}

         // get first (top) via header
         String viaStr = this.getHeader("Via");
         if(viaStr==null)
            return;
         int vInd = viaStr.indexOf(";");
         if(vInd <=0)
            vInd = viaStr.length();
         viaStr = viaStr.substring(viaStr.indexOf(" "), vInd).trim();
         vInd = viaStr.indexOf(":");
         if(vInd>0) {
            host = viaStr.substring(0,vInd);
            try{
               port = Integer.parseInt(viaStr.substring(vInd+1, viaStr.length()).trim());
            }catch(Exception ex) {
               return;
            }
         }
         else {
            host = viaStr;
         }
      }
      else {
         // send to where it came from
         host = sipdMsg.source;
         port = sipdMsg.source_port;
      }
      // use request_t structure to mark destination
      r.sin = host;
      r.sin_port = port>0 ? port : 5060;
      r.via.sin = r.sin;
      r.via.sin_port = r.sin_port;

      r.status = status;
      r.reason = reasonPhrase;

      // copy headers to request_t
      r.headers = new LinkedList(this.headers);

      r.response_body.content = content;
      r.response_body.content_length = contentLen;
      r.response_body.content_type = contentType;
      rw.processSend(r, sipdMsg, status, reasonPhrase);
   }
}
