
//package SipSImp;

import java.util.*;
import org.ietf.sip.*;

/**
 * Represents SIP transactions. A transaction is identified by its CSeq
 * number within a call leg. As calls and call legs are not represented
 * explicitly in the API, transactions encapsulate Call-ID, From and To
 * headers as well as the CSeq number.
 *
 * <p>These transaction parameters doesn't change during the lifetime of
 * a transaction, and hence are read-only attributes oif this class.
 *
 * <p>SipTransactions created locally,
 * e.g. by a servlet, are initialized with values for the From and To
 * headers and the request method, whereas the value of the Call-ID, CSeq
 * number and possibly tags associated with the From or To headers cannot
 * be set by applications.
 *
 * @see SipFactory#createSipTransaction
 */
public class SipTransactionImpl implements SipTransaction {

   private String callID;
   private SipAddress from, to;
   private String cSeq;
   private HashMap attributes;

   /* set of listeners (sip servlet instances)
    * currently, only one servlet instance should be used
    * to process transactions requests/responses
    * servlets should not be allowed to call method addListener
    * because user servlets are not allowed to create local transactions
    */
   protected Vector listeners;

   // original sipd request (transaction)
   protected request_t sipdTransaction;

   public SipTransactionImpl() {
      callID = null;
      from = null;
      to = null;
      cSeq = null;
      attributes = new HashMap();
      listeners = new Vector(1);
      sipdTransaction = null;
   }

   public SipTransactionImpl(String cID, SipAddress f, SipAddress t, String cs, HashMap attr) {
      callID = cID;
      from = f;
      to = t;
      cSeq = cs;
      if(attr!=null)
         attributes = new HashMap(attr);
      else
         attributes = null;
      listeners = new Vector(1);
      sipdTransaction = null;
   }

   public SipTransactionImpl(String cID, SipAddress f, SipAddress t, String cs) {
      callID = cID;
      from = f;
      to = t;
      cSeq = cs;
      attributes = new HashMap();
      listeners = new Vector(1);
      sipdTransaction = null;
   }

   /**
    * Returns the Call-ID of this transaction. If this transaction was
    * created locally the Call-ID is generated by the servlet engine.
    * Servlets should never attempt to set the Call-ID using
    * SipMessage.setHeader.
    */
   public String getCallID() {
      return callID;
   }

   /**
    * Returns the value of the From header associated with this transaction.
    */
   public SipAddress getFrom() {
      return from;
   }

   /**
    * Returns the value of the To header associated with this transaction.
    */
   public SipAddress getTo(){
      return to;
   }

   /**
    * Returns the value of the CSeq (command sequence) attribute associated
    * with this transaction. This is either the value used in an incoming
    * SIP message or one generated by the servlet engine.
    */
   public String getCSeq(){
      return cSeq;
   }

   /**
    * Returns a new request object belonging to this transaction.
    *
    * <p>This is used for new transactions created using the SipFactory
    * interface for initiating requests from a servlet, but can also be
    * used by servlets in proxies in order, for example, to cancel
    * outstanding requests.
    *
    * @method  the SIP request method.
    */
   public SipRequest createRequest(String method){
      String msg = "SipTransaction: currently not supported; no locally generated transactions allowed";
      throw new UnsupportedOperationException(msg);
   }

   /**
    * Returns the value of the specified attribute, or null if it is
    * undefined.
    */
   public Object getAttribute(String name){
      return attributes.get(name);
   }

   /**
    * Returns an iterator over all defined attributes.
    */
   public Iterator getAttributeNames(){
      return attributes.keySet().iterator();
   }

   /**
    * Associates the named attribute with the specified value. A set of
    * servlets associated with the same servlet engine sees
    */
   public void setAttribute(String name, Object value){
      attributes.put(name, value);
   }

   /**
    * Removes the specified attribute.
    */
   public void removeAttribute(String name){
      attributes.remove(name);
   }

   /**
    * Register the specified servlet to receive notifications of
    * incoming SIP messages within this transaction. Requests and responses received
    */
   public void addListener(SipServlet servlet){

      throw new UnsupportedOperationException("SipTransaction: user addition of listeners currently not supported");
      // method below is provided by API, but not
      // allowed in the current implementation because
      //
      // listeners.add(servlet);
   }

   /**
    * Deregister the specified servlet for receiving notifications
    * relating to this transaction.
    */
   public void removeListener(SipServlet servlet){
      listeners.remove(servlet);
   }

   // methods included below are not defined by API
   // these methods are used by servlet engine
   public void setSipServletListener(SipServlet servlet) {
      // with current implementation (where users are not allowed to
      // add servlets to transactions) there will be a single listener
      // for a transaction
      listeners.removeAllElements();
      listeners.setSize(1);
      listeners.setElementAt(servlet, 0);
      listeners.trimToSize();
   }

   // currently, only one listener is allowed for a transaction
   public SipServlet getSipServletListener() {
      SipServlet srv = null;
      try{
         srv = (SipServlet) listeners.firstElement();
      }catch(NoSuchElementException ex) {
         srv = null;
      }
      return srv;
   }

   // sets/gets corresponding sipd type
   public request_t getSipdTransaction(){
      return sipdTransaction;
   }

   public void setSipdTransaction(request_t trans) {
      sipdTransaction = trans;
   }
}
