
//package SipSImp;

import java.util.*;
import org.ietf.sip.*;

/**
 * Class SipURL represents addresses of SIP endpoints. SIP URLs are used
 * in the request-URI of SIP requests as well as in To, From, and Contact
 * header fields. See [RFC 2543, section 2] for the definiton of SIP URLs.
 *
 * @see SipAddress
 * @see Contact
 */
public class SipURLImpl implements SipURL {

   String user;
   String password;
   String host;
   int port;
   HashMap parameterMap;

   public SipURLImpl() {
      user = null;
      password = null;
      host = null;
      port = 0;
      parameterMap = new HashMap();
   }

   public SipURLImpl(SipURL sUrl) {
      user = sUrl.getUser();
      password = sUrl.getPassword();
      host = sUrl.getHost();
      port = sUrl.getPort();
      parameterMap = new HashMap();
      Iterator it = sUrl.getParameterNames();
      while(it.hasNext()) {
         String p = (String) it.next();
         parameterMap.put(p, sUrl.getParameter(p));
      }
   }

   public SipURLImpl(String u, String pass, String h, int p, HashMap params) {
      user = u;
      password = pass;
      host = h;
      port = p;
      if(port < 0) port = 0;
      if(params == null)
         parameterMap = new HashMap();
      else
         parameterMap = new HashMap(params);
   }

   /**
    * Returns the user part of this SipURL. This may be null if the URL was
    * obtained from, or is intended to be used in, a REGISTER request.
    */
   public String getUser() {
      return user;
   }

   /**
    * Set the user part of this SipURL.
    */
   public void setUser(String userArg) {
      user = userArg;
   }

   /**
    * Returns the password of this SipURL. This may be null.
    */
   public String getPassword() {
      return password;
   }

   /**
    * Sets the password associated with this SipURL.
    */
   public void setPassword(String passwordArg)  {
      password = passwordArg;
   }

   /**
    * Returns  the host name of this SipURL. This can be either a domain
    * name  or a dotted-decimal IP address.
    */
   public String  getHost() {
      return host;
   }

   /**
    * Sets  the host name of this SipURL.
    */
   public void setHost(String hostArg) {
      host = hostArg;
   }

   /**
    * Returns  the port number of this SipURL. This is 5060 by default,
    * the default SIP port number.
    */
   public int  getPort() {
      return port;
   }

   /**
    * Sets  the port number of this SipURL.
    */
   public void setPort(int portArg) {
      port = portArg;
   }

   /**
    * Returns  the value associated with the named parameter, the
    * @param name  the  name of the parameter
    * @return  the value of the named parameter, or the empty string
    *    for boolean parameters, and null if the parameter is not defined
    */
   public String getParameter(String name) {
      if(parameterMap.containsKey(name)) {
         return (String) parameterMap.get(name);
      }
      else {
         return null;
      }
   }

   /**
    * Sets  the specified boolean parameter.
    */
   public void setParameter(String  name) {
      if(parameterMap.containsKey(name)) {
         parameterMap.put(name, null);
      }
   }

   /**
    * Sets  the named parameter to the specified value.
    * This  implementation will not allow adding of new
    * mappings (replaces existing one).
    */
   public void setParameter(String  name, String value) {
      if(parameterMap.containsKey(name)) {
         parameterMap.put(name, value);
      }
   }

   /**
    * Removes  the specified parameter from this SipURL.
    */
   public void removeParameter(String  name) {
      if(parameterMap.containsKey(name)) {
         parameterMap.remove(name);
      }
   }

   /**
    * Returns  an Iterator over all specified parameters.
    * These are all  Strings.
    */
   public Iterator getParameterNames() {
       Set n = parameterMap.keySet();
       return n.iterator();
   }

   /**
    * Returns  the textual representation corresponding to this SipURL.
    * This  must be a valid SIP URL according to RFC 2543.
    */
   public String toString() {
      String ret = new String("sip:");
      if(user != null && host != null) {
         ret += user;
         if(password!=null)
            ret += ":"+password;
         ret += "@"+host;
         if(port>0)
            ret += ":"+port;
         Iterator it = parameterMap.keySet().iterator();
         while(it.hasNext()) {
            String param = (String) it.next();
            String val = (String) parameterMap.get(param);
            if(val!=null)
               ret += ";"+param+"="+val;
            else
               ret += ";"+param;  // boolean parameter
         }
      }
      else
         ret = null;
      return ret;
   }
}
