
#include <jni.h>
#include <string.h>
#include <stdlib.h>


/*
 The file contains two sets of functions:
   First set allocates objects in JVM and allow to set those
 objects to specific values. When assigning values, objects
 whose values are set are located in parameter lists. Return values are
 codes (negative when errors encountered, positive if ok).
  The second set retrieves values from java objects. In most cases,
 function's return value contains the result.
*/

int allocateJObject(JNIEnv *env, jobject *obj, char *class)
{
   jclass cls;
   cls = (*env)->FindClass(env, class);
   if (cls == NULL)  {
      return -1;
   }
   *obj = (*env)->AllocObject(env, cls);
   if ((*obj) == NULL) {
      return -2;
   }
   return 1;
}


jobject allocateEmptyJObject(JNIEnv *env, char *class)
{
   jclass cls;
   cls = (*env)->FindClass(env, class);
   if (cls == NULL)  {
      return NULL;
   }
   return (*env)->AllocObject(env, cls);
}


jobject allocateNewJObject(JNIEnv *env, char *class)
{
   jclass cls;
   jmethodID cid;

   cls = (*env)->FindClass(env, class);
   if (cls == NULL)
      return NULL;
   cid = (*env)->GetMethodID(env, cls, "<init>", "()V");
   if(cid == NULL)
      return NULL;
   return (*env)->NewObject(env, cls, cid);
}


int setJObjectStringField(JNIEnv *env, jobject *obj, jclass cls, char *field, char *value)
{
   int ret;
   jstring jstr;
   jfieldID fid;

   if(value != NULL) {
      fid = (*env)->GetFieldID(env, cls, field, "Ljava/lang/String;");
      if (fid == 0) {
        return -1;
      }
      jstr = (*env)->NewStringUTF(env, value);
      if (jstr == 0) {
        return -2;
      }
      (*env)->SetObjectField(env, *obj, fid, jstr);
      return 1;
   }
   else
      return 2;
}


int setJObjectIntField(JNIEnv *env, jobject *obj, jclass cls, char *field, int value)
{
   jint ji;
   jfieldID fid;

   ji = (jint) value;
   fid = (*env)->GetFieldID(env, cls, field, "I");
   if (fid == 0) {
      return -1;
   }
   (*env)->SetIntField(env, *obj, fid, ji);
   return 1;
}


int setJObjectDoubleField(JNIEnv *env, jobject *obj, jclass cls, char *field, double value)
{
   jdouble jd;
   jfieldID fid;

   jd = (jdouble) value;
   fid = (*env)->GetFieldID(env, cls, field, "D");
   if (fid == 0) {
      return -1;
   }
   (*env)->SetDoubleField(env, *obj, fid, jd);
   return 1;
}


int setJObjectLongField(JNIEnv *env, jobject *obj, jclass cls, char *field, long value)
{
   jlong jl;
   jfieldID fid;

   jl = (jlong) value;
   fid = (*env)->GetFieldID(env, cls, field, "J");
   if (fid == 0) {
      return -1;
   }
   (*env)->SetLongField(env, *obj, fid, jl);
   return 1;
}


int setJObjectObjectField(JNIEnv *env, jobject *obj, jclass cls, char *field, char *descr, jobject value)
{
   int ret;
   jfieldID fid;

   fid = (*env)->GetFieldID(env, cls, field, descr);
   if (fid == 0) {
      return -1;
   }
   (*env)->SetObjectField(env, *obj, fid, value);
   return 1;
}


int addJHashMapObjectPair(JNIEnv *env, jobject *obj, jclass cls, char *key, char *value)
{
   int ret = 1;
   jstring jkey, jvalue;
   jmethodID mid;

   if(key == NULL)
      return -2;

   mid = (*env)->GetMethodID(env, cls, "put",
         "(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;");
   if(mid == 0)
      return -1;
   jkey = (*env)->NewStringUTF(env, key);
   jvalue = (*env)->NewStringUTF(env, value);
   if(jkey==0 || jvalue==0){
      return -1;
   }

   (*env)->CallObjectMethod(env, *obj, mid, jkey, jvalue);
   return ret;
}


int addJListObjectObject(JNIEnv *env, jobject *obj, jclass cls, jobject jval)
{
   int ret = 1;
   jmethodID mid;
   jboolean res;
   mid = (*env)->GetMethodID(env, cls, "add", "(Ljava/lang/Object;)Z");
   if(mid == NULL)
      return -1;
   res = (*env)->CallBooleanMethod(env, *obj, mid, jval);
   return res;
}



/**
 * Functions for retrieving field values from Java objects
 *
 */

char *getJObjectStringField(JNIEnv *env, jobject obj, jclass cls, char *field)
{
   jfieldID fid;
   jstring jstr;
   const char *str;
   char *outStr;

   if(cls==NULL)
      cls = (*env)->GetObjectClass(env, obj);
   if(cls==NULL)
      return NULL;

   if((fid = (*env)->GetFieldID(env, cls, field, "Ljava/lang/String;"))==NULL)
      return NULL;
   jstr = (*env)->GetObjectField(env, obj, fid);
   if(jstr == NULL)
      return NULL;
   if((str = (*env)->GetStringUTFChars(env, jstr, NULL))==NULL)
      return NULL;
   outStr = strdup(str);
   (*env)->ReleaseStringUTFChars(env, jstr, str);

   return outStr;
}



char *getUTFString(JNIEnv *env, jstring jstr)
{
   const char *str;
   char *outStr;

   if(jstr==NULL)
      return NULL;
   if((str = (*env)->GetStringUTFChars(env, jstr, NULL))==NULL)
      return NULL;
   outStr = strdup(str);
   (*env)->ReleaseStringUTFChars(env, jstr, str);
   return outStr;
}



int getJObjectIntField(JNIEnv *env, jobject obj, jclass cls, char *field)
{
   jfieldID fid;
   int ret;
   jint r;

   if(cls==NULL)
      cls = (*env)->GetObjectClass(env, obj);
   if(cls==NULL)
      return -1;
   fid = (*env)->GetFieldID(env, cls, field, "I");
   r = (*env)->GetIntField(env, obj, fid);
   ret = (int) r;
   return ret;
}


long getJObjectLongField(JNIEnv *env, jobject obj, jclass cls, char *field)
{
   jfieldID fid;
   long ret;
   jlong r;

   if(cls==NULL)
      cls = (*env)->GetObjectClass(env, obj);
   if(cls==NULL)
      return -1;
   fid = (*env)->GetFieldID(env, cls, field, "J");
   r = (*env)->GetLongField(env, obj, fid);
   ret = (long) r;
   return ret;
}


double getJObjectDoubleField(JNIEnv *env, jobject obj, jclass cls, char *field)
{
   jfieldID fid;
   double ret;
   jdouble d;

   if(cls==NULL)
      cls = (*env)->GetObjectClass(env, obj);
   if(cls==NULL)
      return -1;
   fid = (*env)->GetFieldID(env, cls, field, "D");
   d = (*env)->GetDoubleField(env, obj, fid);
   return (double) d;
}


jobject getJObjectObjectField(JNIEnv *env, jobject parentObj, jclass cls, char *field, char *descr)
{
   jfieldID fid;

   if(cls==NULL)
      cls = (*env)->GetObjectClass(env, parentObj);
   if(cls==NULL)
      return NULL;
   if((fid = (*env)->GetFieldID(env, cls, field, descr))==NULL)
      return NULL;
   return (*env)->GetObjectField(env, parentObj, fid);
}



/* list/vector handling functions */
int getListSize(JNIEnv *env, jobject obj, jclass cls)
{
   jint s;
   jmethodID mid;

   if(cls==NULL)
      cls = (*env)->GetObjectClass(env, obj);
   if(cls==NULL)
      return -1;
   mid = (*env)->GetMethodID(env, cls, "size", "()I");
   if(mid==NULL)
      return -1;
   s = (*env)->CallIntMethod(env, obj, mid);
   return (int) s;
}


jobject getListElement(JNIEnv *env, jobject obj, jclass cls, int index)
{
   jint i = (jint) index;
   jmethodID mid;

   if(cls==NULL)
      cls = (*env)->GetObjectClass(env, obj);
   if(cls==NULL)
      return NULL;
   mid = (*env)->GetMethodID(env, cls, "get", "(I)Ljava/lang/Object;");
   if(mid==NULL)
      return NULL;
   return (*env)->CallObjectMethod(env, obj, mid, i);
}


jobject getMapKeyList(JNIEnv *env, jobject obj, jclass cls)
{
   jmethodID mid1, cid;
   jobject set;
   jint i;
   jclass setCls, listCls;

   if(cls==NULL)
      cls = (*env)->GetObjectClass(env, obj);
      //cls = (*env)->FindClass(env, "java/util/HashMap");
   if(cls==NULL)
      return NULL;
   /* HashMap is the expected object */
   mid1 = (*env)->GetMethodID(env, cls, "keySet", "()Ljava/util/Set;");
   if(mid1==NULL)
      return NULL;
   set = (*env)->CallObjectMethod(env, obj, mid1);
   if(set==NULL)
      return NULL;
   if((setCls = (*env)->GetObjectClass(env, set))==NULL)
      return NULL;
   if((listCls = (*env)->FindClass(env, "java/util/ArrayList"))==NULL)
      return NULL;
   cid = (*env)->GetMethodID(env, listCls, "<init>", "(Ljava/util/Collection;)V");
   if(cid==NULL)
      return NULL;
   return (*env)->NewObject(env, listCls, cid, set);
}


jobject getMapValue(JNIEnv *env, jobject obj, jclass cls, jobject pObj)
{
   jmethodID mid;

   if(cls==NULL)
      cls = (*env)->GetObjectClass(env, obj);
      //cls = (*env)->FindClass(env, "java/util/HashMap");
   if(cls==NULL)
      return NULL;
   /* HashMap is the expected object */
   mid = (*env)->GetMethodID(env, cls, "get", "(Ljava/lang/Object;)Ljava/lang/Object;");
   if(mid==NULL)
      return NULL;
   return (*env)->CallObjectMethod(env, obj, mid, pObj);
}