
/**
 * Loads a user servlet contained in Java class file
 * (or looks for .java file and compiles it first into .class file
 * if .class is not found).
 * Assumes that user files are stored according to sipd policy
 * for storing scripts (see script.c on script_t type).
 */

import java.io.*;

class ServletClassLoader extends ClassLoader {

   /* User servlets are stored according to format
    * specified in script.c (script_t type):
    * scripts_root -> user_dir -> disposition_dir -> script_file
    */
   String rootPath;        // base root where user scripts are
   String disposition;
   char separator;
   String userID;

   public ServletClassLoader() {
      rootPath = null;
      disposition = new String("sip_servlet");
      separator = '/';
      userID = null;
   }

   public ServletClassLoader(char sptr, String root, String dispos, String uid) {
      rootPath = root.trim();
      disposition = dispos.trim();
      separator = sptr;
      userID = uid;
   }

   /**
    * finds class and returns its bytes
    *
    * user of this loader will call "loadClass" method
    * from the parent class which will call this method
    * to load class and resolve it
    */
   public Class findClass(String name) {
      Class newC = null;
      name.trim();
      if(name.endsWith(".class") || name.endsWith(".java")) {
         name = name.substring(0, name.lastIndexOf('.'));
      }
      byte[] b = loadClassData(name);
      if(b==null)
         return null;
      else {
         return defineClass(name, b, 0, b.length);
      }
   }


   /**
    * loads class file from user directory of servlets
    */
   private byte[] loadClassData(String name) {
      String path = absServletName(name);
      String clsPath = path + ".class";
      String srcPath = path + ".java";
      File cls = new File(clsPath);
      File src = new File(srcPath);
      if(cls.exists()) {
         // reads class bytes from file in user
         // disposition directory
         return getClassBytes(cls);
      }
      else if(src.exists()){
         // attempt to compile Java source if it exists
         // and its class cant't be found
         String cmd[] = new String[2];
         cmd[0] = SipSConfig.JDKHome + separator + "bin" + separator + "javac";
         cmd[1] = src.getAbsolutePath();

         Runtime r = Runtime.getRuntime();
         try {
            Process p = r.exec(cmd);
            int i = p.waitFor();
         }catch(Exception e) {
            return null;
         }
         if(cls.exists()) {
            return getClassBytes(cls);
         }
         else
            return null;
      }
      else
         return null;
   }

   /**
    * returns absolute path to servlet class file
    */
   private String absServletName(String className) {
      String ret;
      String r = new String(rootPath);
      String u = new String(userID);
      String d = new String(disposition);

      // assumption is that user directory and disposition directory
      // have no dir separators in the middle (are single directories)
      u.replace(separator,' '); u.trim(); u += separator;
      d.replace(separator,' '); d.trim(); d += separator;
      if(!r.endsWith(""+separator))
         r += separator;
      ret = r + u + d + className;
      return ret;
   }

   /**
    * reads bytes from class file to a byte array
    */
   private byte[] getClassBytes(File clsFile) {
      byte b[] = null;
      int len = (int) clsFile.length();
      b = new byte[len];
      try{
         int rLen;
         FileInputStream fIn = new FileInputStream(clsFile);
         rLen = fIn.read(b, 0, len);
         fIn.close();
      } catch(Exception ex) {
         b = null;
      }
      return b;
   }
}
