
/**
 * Servlet engine for sipd
 * - Receives request_t and message_t objects from sipd
 * - Loads and manages instances of user servlets
 * - Instantiates SipS Objects and passes them to instances
 *   of user servlets
 * - Manages instances of current transactions
 */

import java.net.*;
import java.util.*;

// import SipS API and its implementation
//import SipSImplementation.*;
import org.ietf.sip.*;

public class SipSContainer
{
   // synchronized map of current transactions
   private static SipTransactionInstanceContainer instanceMap;

   // servlet config (contains ServletContext, SipFactory, ContactDatabase)
   private static ServletConfigImpl servletConfig;

   /**
    * start-up and shutdown
    */
   public static void startUpContainer() {
      System.out.println(">SIPS: starting up container ");
      instanceMap = new SipTransactionInstanceContainer();
      servletConfig = new ServletConfigImpl(new HashMap(), null, null);
   }

   /**
    * container shutdown
    */
   public static void shutDownContainer() {
      System.out.println(">SIPS: shut down ");
      // calls destroy method of each servlet instance in instanceMap
      destroyActiveTransactions();
      synchronized(instanceMap) { instanceMap = null; }
   }


   /**
    * transaction handling
    *
    * initiates transaction (when called from policy_core.c
    */
   public static int transactionInit(request_t req, registration_t reg) {
      System.out.println(" >sips: init trans: "+ req.callID);
      int ret = 1;
      String callID = req.callID;

      // check if transaction with such callID exists in the map
      // do not allow same transaction
      if(hasTransaction(callID))
         return -1;

      // build SipS transaction from sipd structure
      SipTransactionImpl t = formSipTransaction(req);
      if(t==null)
         return -1;

      // user -> owner of servlet
      String user = req.p.URI.user;
      if(user==null || reg.sipsClass==null )
         return -1;

      ServletClassLoader ld = new ServletClassLoader('/', reg.rootDir, reg.dispString, user );

      Class c = null;
      try {
         c = ld.loadClass(reg.sipsClass);
      } catch (Exception ex) {
         return -2;
      }
      if(c==null) return -2;
      try {
         Object o = c.newInstance();
         if(o instanceof SipServlet) {
            SipServlet sips = (SipServlet) o;

            try {
               sips.init(servletConfig);
            } catch(Throwable th) {
               System.out.println(" servlet's init failed ");
               return -10;
            }

            t.setSipServletListener(sips);
            SipTransactionInstance trans = new SipTransactionInstance();
            trans.callID = callID;
            trans.active = true;
            if(sips instanceof SingleThreadModel)
               trans.singleServletMode = true;
            else
               trans.singleServletMode = false;
            trans.transaction = t;

            add(callID, trans);
         }
         else
            return -3;
      }catch(Exception ex) {
         System.out.println("Exception during initiating servlet; call-ID:" + callID);
         return -4;
      }
      return ret;
   }



   /**
    * processes subsequent transaction requests,
    * called by policy_core
    */
   public static  int transactionRequest(request_t req, message_t msg) {
      System.out.println(" >sips: processing request, transaction callID: "+ req.callID);
      int ret = 1;
      String callID = req.callID;

      if(hasTransaction(callID)) {
         SipTransactionInstance trans = get(callID);
         trans.transaction.setSipdTransaction(req);
         SipServlet sips = trans.transaction.getSipServletListener();
         SipMessageImpl m= formSipMessage(req, msg, trans.transaction);
         if(m==null)
            return -1;
         if(m.getSipdMsg()==null)
            m.setSipdMsg(msg);  // necessary
         SipRequest sipReq = formSipRequest(trans.transaction, m);
         if(sipReq==null)
            return -1;
         if(sips!=null) {
            try {
               // call servlet's method
               sips.gotRequest(sipReq);
            } catch(Throwable th) {
               System.out.println("Exception caught in servlet's gotRequest(); call-ID:" + callID);
               return -10;
            }
         }
      }
      else
         return -1;

      return ret;
   }



   /**
    * transaction handling
    *
    */
   public static  int transactionResponse(request_t req, message_t msg, branch_t branch) {
      System.out.println(" >sips: processing response, transaction callID: "+ req.callID);
      int ret=1;
      String callID = req.callID;

      if(hasTransaction(callID)) {
         // get transaction from map
         SipTransactionInstance trans = get(callID);

         trans.transaction.setSipdTransaction(req);
         SipServlet sips = trans.transaction.getSipServletListener();
         SipMessageImpl m= formSipMessage(req, msg, trans.transaction);
         if(m==null)
            return -1;
         if(m.getSipdMsg()==null)
            m.setSipdMsg(msg);  // necessary
         SipResponse sipResp = formSipResponse(trans.transaction, m);
         if(sipResp==null)
            return -1;
         if(sips!=null) {
            try {
               // call servlet's method
               sips.gotResponse(sipResp);
            } catch(Throwable th) {
               System.out.println("Exception caught in servlet's gotResponse(); call-ID:" + callID);
               return -10;
            }
         }
      }
      else
         return -1;

      return ret;
   }



   public static  int transactionCleanUp(request_t req) {
      System.out.println(" >sips: transaction clean-up, callID: "+ req.callID);
      int ret=1;
      String callID = req.callID;
      if(hasTransaction(callID)) {
         SipTransactionInstance trans = get(callID);
         SipServlet sips = trans.transaction.getSipServletListener();
         if(sips!=null) {
            try {
               sips.destroy();
               trans.transaction.removeListener(sips);
            } catch(Throwable th) {
               System.out.println("Exception caught in servlet's destroy(); call-ID:" + callID);
               return -10;
            }
         }
         instanceMap.remove(callID);
      }
      return ret;
   }



   /**
    * synchronized functions for updating map of instances
    */
   private  synchronized static SipTransactionInstance get(String cID) {
      return instanceMap.get(cID);
   }

   private  synchronized static boolean add(String callID, SipTransactionInstance trans) {
      return instanceMap.add(callID, trans);
   }

   private  synchronized static void remove(String cID) {
      instanceMap.remove(cID);
   }

   private  synchronized static boolean hasTransaction(String callID) {
      return instanceMap.hasInstance(callID);
   }

   private  synchronized static void destroyActiveTransactions() {
      Iterator mapKeys = instanceMap.transactionIDIterator();
      while(mapKeys.hasNext()) {
         String cID = (String) mapKeys.next();
         SipTransactionInstance trans = (SipTransactionInstance) instanceMap.get(cID);

         SipServlet srv = trans.transaction.getSipServletListener();
         if(srv!=null) {
            try {
               // call servlets destroy method
               // if the transaction still is holding on to its instance
               srv.destroy();
            }catch(Throwable t) {}
            trans.transaction.removeListener(srv);
         }
         // servlet instance removed from map
         // garbage collected by jvm (since no more references)
         remove(cID);
      }
   }




   /***
    * set of functions for translation of sipd to sips types
    */
   static SipTransactionImpl formSipTransaction(request_t req) {
      SipTransactionImpl t = null;
      if(req.callID != null && req.to!=null && req.from!=null) {
         SipAddressImpl to, from;
         to = formSipAddress(req.to);
         from = formSipAddress(req.from);
         HashMap hmap = new HashMap();
         hmap.put("time-received", ""+req.time_received);
         hmap.put("time-expires", ""+req.expires);
         t = new SipTransactionImpl(req.callID, from, to, ""+req.cseq.seq, null);
      }
      return t;
   }


   static SipMessageImpl formSipMessage(request_t req, message_t msg, SipTransactionImpl t) {
      SipMessageImpl m = null;
      String protocol = msg.p.protocol +"/"+msg.p.version;
      String sProt = "";
      if(req.type==2) sProt = "TCP";
      else if(req.type==1) sProt = "UDP";

      m = new SipMessageImpl(msg.p.method, protocol, msg.source, msg.source_port,
                             sProt, t, null, msg.body.content, msg.body.content_type,
                             msg.body.content_length);
      // needed
      m.setSipdMsg(msg);
      if(msg.p.headers!=null) {
         LinkedList ls = null;
         synchronized(msg.p.headers) {
            ls = new LinkedList(msg.p.headers);
         }
         for(int i=0; i<ls.size(); i++) {
            parser_header_t pr = (parser_header_t) ls.get(i);
            String name = pr.name;
            String value = SipMessageImpl.headerObjToStr(pr);
            m.addHeader(name,value);
         }
      }
      return m;
   }


   static SipRequestImpl formSipRequest(SipTransactionImpl t, SipMessageImpl m) {
      message_t msg = m.getSipdMsg();
      request_t tr = t.getSipdTransaction();
      if(tr==null || msg==null)
         return null;
      String uri = msg.p.URI_string;
      String aMethod = tr.authorization.method;
      String aUid = tr.authorization.userid;
      boolean aValid = (tr.authorization.valid>0 ? true : false);

      SipRequestImpl r = new SipRequestImpl(uri, aMethod, aValid, aUid, true, m);
      return r;
   }


   static SipResponseImpl formSipResponse(SipTransactionImpl t, SipMessageImpl m) {
      message_t msg = m.getSipdMsg();
      request_t tr = t.getSipdTransaction();
      if(tr==null || msg==null)
         return null;
      int status = msg.p.status;
      String reason = msg.p.reason;
      SipResponseImpl r = new SipResponseImpl(status, reason, null, true, m);
      return r;
   }


   static SipAddressImpl formSipAddress(sip_addr_t addr) {
      SipAddressImpl sipAddr= null;
      SipURLImpl url = null;
      String dName = addr.display_name;
      if(addr.uri.host!=null)
         url = new SipURLImpl(addr.uri.user, addr.uri.password,
                           addr.uri.host, addr.uri.port, addr.uri.parameters);
      if(url!=null)
         sipAddr = new SipAddressImpl(dName, url);

      return sipAddr;
   }

}




/**
 * synchronized servlet map
 *
 */
class SipTransactionInstanceContainer
{
   // constructor creates synchronized map
   // which should be thread-safe
   // maps transactions by callID (assumed unique)
   private Map transactionMap;

   public SipTransactionInstanceContainer() {
      transactionMap = Collections.synchronizedMap(new HashMap());
   }

   public boolean add(String cID, SipTransactionInstance tr) {
      if(!transactionMap.containsKey(cID)) {
         transactionMap.put(cID, tr);
         return true;
      }
      else
         return false;
   }

   public boolean hasInstance(String callID) {
      return transactionMap.containsKey(callID);
   }

   public SipTransactionInstance get(String callID) {
      if(transactionMap.containsKey(callID))
         return (SipTransactionInstance) (transactionMap.get(callID));
      else
         return null;
   }

   public void remove(String callID) {
      transactionMap.remove(callID);
   }

   public Iterator transactionIDIterator() {
      return transactionMap.keySet().iterator();
   }
}



/**
 * instance of servlet as maintained by map
 */
class SipTransactionInstance
{
   public String callID;
   public boolean active;
   public boolean singleServletMode;
   public SipTransactionImpl transaction;
}
