/*
request.c - Create, search, process, destroy requests.

Currently, pending requests are in a linked list; for efficiency, they
should be in a hash table.

Copyright 1998-1999 by Columbia University; all rights reserved
*/

#include <stdio.h>
#include <stdlib.h>     /* calloc() */
#include <assert.h>
#include <pthread.h>
#include <unistd.h>     /* close() */
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include "request-sip.h"
#include "sip.h"
#include "strdupn.h"
#include "sysdep.h"
#include "rfc822.h"
#include "config.h"
#include "authenticate.h"
#include "libsip.h"

#include "log.h"
#include "msgflow.h"

extern double gettimeofdayf(void);

static request_t *R;          /* list of pending requests */
/* Lock for 'R'.  To avoid deadlock, always lock R_lock before locking
 * an individual request's lock, if you need to hold both locks
 * simultaneously. */
static pthread_mutex_t R_lock = PTHREAD_MUTEX_INITIALIZER;

/*
* RequestCreate()
* Create a new request object.
*/
request_t *RequestCreate(int type, int socket, struct sockaddr_in sin)
{
  request_t *r = calloc(1, sizeof(request_t));

  r->version       = NULL;
  r->callID        = NULL;
  r->contact       = NULL;
  r->user_agent    = NULL;
  r->reason        = NULL;
  r->headers.headers = NULL;
  r->headers.last  = NULL;
  r->next          = NULL;

  r->type          = type;
  r->socket        = socket;
  r->source_sin    = sin;
  r->sin           = sin;
  r->time_received = gettimeofdayf();
  r->status        = 0; /* because later code will inspect it */
  pthread_mutex_init(&r->mutex, NULL);
  pthread_cond_init(&r->cond, NULL);

  memset(&r->remote_sin,0,sizeof(r->remote_sin));
  r->call = NULL;
  r->server = NULL;
  memset(&r->outbound_proxy, 0, sizeof(r->outbound_proxy));
  r->preferred_type = 0;

  return r;
} /* RequestCreate */


/*
 * RequestAdd()
 * If it's valid (i.e., findable by RequestSearch), add a request to the
 * head of the list.  Return 1 if it was added, 0 otherwise.
 */
int RequestAdd(request_t *r)
{
  /* We could also check if we have a CSeq, but both 0 and (ulong)-1 are
   * valid according to the spec... */
  if (r->callID == NULL) {
    return 0;
  }

  pthread_mutex_lock(&R_lock);
  r->next = R;
  R = r;
  pthread_mutex_unlock(&R_lock);

  return 1;
} /* RequestAdd */


/*
 * RequestFree()
 * Free the request and the data associated with it.
 * Remove the request from the linked list of requests if it is on it.
 * Note: you should *NOT* have the request locked before calling this.
 * Note: make sure that your program flows allow this to be called from
 * only one location for each request.  The locks do not protect against
 * RequestFree()/RequestFree() race conditions, though we are protected
 * against RequestFree()/RequestSearch() races.
 */
void RequestFree(request_t *r)
{
  request_t **rp;
  struct message_t *m, *m_next;
  int i;

  if (!r) return;
  /* Remove request from the request list. */
  /* Lock R_lock before r->mutex. */
  pthread_mutex_lock(&R_lock);
  pthread_mutex_lock(&r->mutex);

  debug("RequestFree", "Freeing %p\n", r);
  for (rp = &R; *rp != NULL; rp = &(*rp)->next) {
    if (*rp == r) {
      *rp = (*rp)->next;
      break;
    }
  }

  pthread_mutex_unlock(&R_lock);

  /* mark so that it's more obvious if old request get re-used */
  r->status = 0;

  /* Free storage of request. */
  HTTP_Free(&r->p);
  free_valid(r->version);
  /* r->request_body.content_type is a pointer owned by the parse structure,
     and it was freed by HTTP_Free; so don't free it again here. */
  free_valid(r->request_body.content);

  free_valid(r->callID);
  free_valid(r->via.hostname);
  free_valid(r->via.branch.str);

  URI_Free(&r->to.uri);
  free_valid(r->to.display_name);

  URI_Free(&r->from.uri);
  free_valid(r->from.display_name);

  URI_Free(&r->outbound_proxy);

  SIP_FreeContact(r->contact);
  free_valid(r->user_agent);

  free_valid(r->reason);
  HeaderFree(&r->headers);

  free_valid(r->response_body.content_type);
  free_valid(r->response_body.content);

  HTTP_AuthorizationFree(&r->authorization);

  /* Free the message queue */
  for (m = r->mqueue; m != NULL; m = m_next) {
    m_next = m->next;
    MessageDelete(m);
  }

  /* Free URIs from each branch */
  for (i = 0; i < r->branch_count; i++) {
    URI_Free(&r->branches[i].uri);

    /*
     * Free any unprocessed message here. e.g. shut_down.
     * We must free this here since policy_core may post
     * shutdown message to client after the client thread
     * has already stopped. Other fields of branch_t are
     * freed by the client thread.
     */
    for (m = r->branches[i].mlist; m != NULL; m = m_next) {
      m_next = m->next;
      MessageDelete(m);
    }
  }

  /* Free the branch structure */
  free_valid(r->branches);

  pthread_mutex_unlock(&r->mutex); /* Release lock before deleting */
  pthread_mutex_destroy(&r->mutex);
  pthread_cond_destroy(&r->cond);
  free(r);

  return;
} /* RequestFree */


/*
 * RequestPrint()
 * Prints the existing requests on the output stream FILE *fp
 */
void RequestPrint(FILE *fp)
{
  request_t *rp;

  pthread_mutex_lock(&R_lock);
  for (rp = R; rp; rp = rp->next) {
    /* ADd prints here */
  }
  pthread_mutex_unlock(&R_lock);

  return;
} /* RequestPrint */


/*
 * RequestSearch()
 * Determine if the request 'r' with hash 'hash' exists in the list of
 * pending requests.  A request is deemed identical if it has the same
 * Request-URI, To, From, Call-ID, and sequence number.  Return the matching
 * request if it exists, otherwise NULL.
 * N.B.: The returned request has its mutex locked!  You must unlock it
 * when you're done with it!  */
request_t *RequestSearch(request_t *r, unsigned long hash)
{
  request_t *rp, *result = NULL;

  pthread_mutex_lock(&R_lock);
  for (rp = R; rp; rp = rp->next) {
    if (rp != r
        && (hash == 0 || rp->hash == hash)
        && (rp->p.URI_string == NULL || r->p.URI_string == NULL
            || URI_Equal_With_Defaults(rp->p.URI, r->p.URI, sip_uri_defaults))
        && URI_Equal_With_Defaults(rp->from.uri, r->from.uri, sip_uri_defaults)
        && URI_Equal_With_Defaults(rp->to.uri, r->to.uri, sip_uri_defaults)
        && strcmp(rp->callID, r->callID) == 0
        && rp->cseq.seq == r->cseq.seq) {
      pthread_mutex_lock(&rp->mutex);
      result = rp;
      break;
    }
  }
  pthread_mutex_unlock(&R_lock);

  return result;
} /* RequestSearch */

/*
 * RequestFind()
 * Determine if the request corresponding to message 'm' with hash 'hash'
 * exists in the list of
 * pending requests.  A request is deemed identical if it has the same
 * Request-URI, To, From, Call-ID, and sequence number.  Return the matching
 * request if it exists, otherwise NULL.
 * N.B.: The returned request has its mutex locked!  You must unlock it
 * when you're done with it!  */
request_t *RequestFind(parser_t *p, unsigned long hash)
{
  request_t *rp, *result = NULL;
  sip_addr_t to, from;
  char * callid;
  sip_cseq_t cseq;

  to     = SIP_To(p);
  from   = SIP_From(p);
  callid = SIP_CallID(p);
  cseq   = SIP_CSeq(p);

  pthread_mutex_lock(&R_lock);
  for (rp = R; rp; rp = rp->next) {

    if ((hash == 0 || rp->hash == hash)
        && (rp->p.URI_string == NULL || p->URI_string == NULL
            || URI_Equal_With_Defaults(rp->p.URI, p->URI, sip_uri_defaults))
        && URI_Equal_With_Defaults(rp->from.uri, from.uri, sip_uri_defaults)
        && URI_Equal_With_Defaults(rp->to.uri, to.uri, sip_uri_defaults)
        && strcmp(rp->callID, callid) == 0
        && rp->cseq.seq == cseq.seq) {
      result = rp;
      pthread_mutex_lock(&rp->mutex);
      break;
    }
  }
  pthread_mutex_unlock(&R_lock);

  URI_Free(&to.uri);
  URI_Free(&from.uri);
  free_valid(to.display_name);
  free_valid(from.display_name);

  return result;
} /* RequestSearch */


/*
 * SendFailureResponse()
 * For a request which has failed, send back a failure response, if it
 * wasn't an ACK request.  The request's response status should be set.
 * If "built" is true, the request's response headers should also already
 * be set; otherwise the default response headers will be built.
 * NOTE: because malformed requests don't end up on the request list, if
 * this request was an INVITE it can't be ACK'd.
 * HACK: don't retransmit them; let the server retransmit if necessary.
 * Since we're immediately sending a definitive (4xx or 5xx) response,
 * reliability works okay.
 * Technically, this violates RFC 2543.  However, you're not actually
 * obligated to send back responses to malformed requests at all...
 */
static void SendFailureResponse(request_t *r, int built)
{
  char *method = HTTP_Method(&r->p);
  if (strcasecmp(method, "ACK") == 0) {
    /* Don't respond to mal-formed ACKs.  Everything else gets a response. */
    debug("SendFailureResponse", "r = %p: Rejected ACK\n", r);
  }
  else {
    if (!built) {
      ResponseBuildDefault(r);
    }
    /* For ResponseSend, we must have the mutex locked. */
    pthread_mutex_lock(&r->mutex);
    ResponseSend(r, 0, 1);
  }

  free_valid(method);
}


/*
 * ValidateIncomingRequest()
 * A new request has arrived.  Verify that it is correct and processable; if
 * not, send the appropriate error response, and return false.
 */
static int ValidateIncomingRequest(request_t *r)
{
  char *method = NULL, *version = NULL, *protocol = NULL;
  uri_t uri;                   /* request URI */
  parser_header_t *h;
  parser_element_t *e;
  int handle_locally;
  int built = 0;

  if (r->status >= 400) {
    debug("ValidateIncomingRequest", "r = %p: (r->status = %d) >= 400\n",
          r, r->status);
    goto fail;
  }

  method = HTTP_Method(&r->p);
  version = HTTP_Version(&r->p);
  protocol = HTTP_Protocol(&r->p);

  if (method == NULL) {
    debug("ValidateIncomingRequest", "r = %p: method == NULL\n", r);
    ResponseSetStatus(r, 400, NULL);
    goto fail;
  }

  if (protocol == NULL) {
    debug("ValidateIncomingRequest", "r = %p: protocol == NULL\n", r);
    ResponseSetStatus(r, 400, NULL);
    goto fail;
  }

  if (version == NULL) {
    debug("ValidateIncomingRequest", "r = %p: version == NULL\n", r);
    ResponseSetStatus(r, 400, NULL);
    goto fail;
  }

  if (strcmp(protocol, "SIP") != 0) {
    debug("ValidateIncomingRequest", "r = %p: protocol == %s\n",
          r, protocol);
    ResponseSetStatus(r, 505, NULL);
    goto fail;
  }

  if (strcmp(version, "2.0") != 0) {
    debug("ValidateIncomingRequest", "r = %p: version == %s\n",
          r, version);
    ResponseSetStatus(r, 505, NULL);
    goto fail;
  }

  uri = HTTP_URI(&r->p);
  /* sipd supports the 'sip' and 'tel' URI schemes. */
  if (uri.scheme == NULL ||
      (strcmp(uri.scheme, "sip") != 0 && strcmp(uri.scheme, "tel") != 0)) {
    debug("ValidateIncomingRequest",
          "r = %p: uri.scheme = %s is not sip or tel\n",
          r, uri.scheme != NULL ? uri.scheme : "NULL");
    ResponseSetStatus(r, 400, NULL);
    goto fail;
  }

  handle_locally = SIP_HandleRequestLocally(method, uri);
  if (handle_locally) {
    h = RFC822_Header(r->p.headers, "Require");
  }
  else {
    h = RFC822_Header(r->p.headers, "Proxy-Require");
  }

  if (h != NULL) {
    /* TODO: recognize locally-understood extensions. */
    debug("ValidateIncomingRequest", "r = %p: %s present (420)\n",
          r, handle_locally ? "Require" : "Proxy-Require");
    ResponseSetStatus(r, 420, NULL);
    ResponseBuildDefault(r);
    /* Add "Unsupported" headers */
    for (e = h->elements; e != NULL && e->element != NULL; e = e->next) {
      HeaderAppend(&r->headers, "Unsupported", e->element);
    }
    built = 1;
    goto fail;
  }

  /* If some mandatory headers are missing, return 400. */
  if (!SIP_Has_Mandatory_Headers(&r->p)) {
    debug("ValidateIncomingRequest", "r = %p: mandatory headers missing\n",
          r);
    ResponseSetStatus(r, 400, NULL);
    goto fail;
  }

  free_valid(method);
  free_valid(protocol);
  free_valid(version);

  return 1;

fail:
  SendFailureResponse(r, built);

  free_valid(method);
  free_valid(protocol);
  free_valid(version);

  return 0;
} /* ValidateIncomingRequest */


/*
 * HashString()
 * Hash a null-terminated string into an unsigned long, using the Tcl hash
 * function (multiply by 9 and add).  Helper function for HashString().
 * Question: should this be defined in some other file?
 */
static unsigned long HashString(unsigned long start_hash, char *str)
{
  char *c;
  unsigned long hash = start_hash;

  for (c = str; *c != '\0'; c++) {
    hash = ((hash << 3) + hash);
    hash += (unsigned long) *c;
  }

  return hash;
} /* HashString */


/*
 * HashInt()
 * Hash an unsigned long into an unsigned long, exactly as if we sprintf'd the
 * value being added to the hash into a string with "%lu" and then called
 * HashString() on it.
 * Because of this "precisely the same" requirement, we first unpack
 * the unsigned long in decimal into a buffer, then add the buffer in
 * the opposite order (most-significant-digit first).
 */
static unsigned long HashInt(unsigned long start_hash,
                             unsigned long start_value)
{
  unsigned long hash = start_hash;
  unsigned long value = start_value;
  int buf[sizeof(unsigned long)*3];
  int idx;

  for (idx = 0;
       idx < sizeof(unsigned long) * 3 && (value != 0 || idx == 0);
       idx++) {
    buf[idx] = value % 10;
    value /= 10;
  }

  for (idx--; idx >= 0; idx--) {
    hash = ((hash << 3) + hash);
    hash += buf[idx] + '0';
  }

  return hash;
} /* HashInt */


/*
 * HashURI()
 * Hash a uri_t structure into an unsigned long.  The value hashed must
 * be the same if two uri's compare as equal according to
 * URI_Equal_With_Defaults, regardless of their actual form.
 * Currently we accomplish this by calling URI_Canonicalize, URI_SPrint, and
 * hashing the resulting string.
 */
static unsigned long HashURI(unsigned long start_hash, uri_t *uri)
{
  unsigned long hashval = start_hash;

  uri_t canon;
  DString dstr;

  canon = URI_Copy(*uri);
  URI_Canonicalize(&canon, sip_uri_defaults);

  DStringInit(&dstr);
  URI_SPrint(&canon, &dstr, 0);

  hashval = HashString(hashval, DStringValue(&dstr));

  DStringFree(&dstr);
  URI_Free(&canon);

  return hashval;
} /* HashURI */


/*
 * ComputeRequestHash()
 * Compute a request's identifying hash.  Base this on its
 * Request-URI, To, From, Call-ID, and CSeq headers.  Used for loop
 * detection and for fast corrolation of retransmissions and responses.
 */
unsigned long ComputeRequestHash(request_t *r)
{
  unsigned long hashval = 0;

  /* Request-URI */
  hashval = HashURI(hashval, &r->p.URI);
  hashval = HashString(hashval, ":");

  /* To */
  hashval = HashURI(hashval, &r->to.uri);
  hashval = HashString(hashval, ":");

  /* From */
  hashval = HashURI(hashval, &r->from.uri);
  hashval = HashString(hashval, ":");

  /* Call-ID */
  hashval = HashString(hashval, r->callID);
  hashval = HashString(hashval, ":");

  /* CSeq */
  hashval = HashInt(hashval, r->cseq.seq);

  /* Reserve hashval == 0 for special cases */
  if (hashval == 0) {
    hashval = 1;
  }

  return hashval;
} /* ComputeRequestHash */


/*
 * SetRequestVia()
 * Set r->via appropriately, based on the topmost via and the address from
 * which the request was received.  Remember the value for a "received" tag
 * to add to the topmost via if appropriate.
 */
static void SetRequestVia(request_t *r)
{
  r->via          = SIP_Via(&r->p, 1, 1, config.ProxyName);

  if (r->via.hostname == NULL) {
    /* Missing or invalid Via header, fall back to default. */
    debug("SetRequestVia",
      "r = %p: r->via.hostname == NULL\n", r);
    r->via.sin = r->sin;
    r->via.received = r->sin.sin_addr;
  }
  if (r->via.sin.sin_addr.s_addr != r->sin.sin_addr.s_addr) {
    /* Via has wrong host addr. */
    /* Specify actual addr so we an add "received" tag to outgoing requests */
    debug("SetRequestVia",
          "r = %p: topmost via header does not match received addr (%s/%s)\n",
          r, r->via.hostname, inet_ntoa(r->sin.sin_addr));
    r->via.sin.sin_addr = r->sin.sin_addr;
    r->via.received = r->sin.sin_addr;
  }
  r->sin          = r->via.sin;  /* All other attributes of response addr
                                    (notably port) come from Via. */
} /* SetRequestVia */


/*
 * RequestIsNotLooped()
 * Check if this request has been looped -- i.e., if our proxy name appears
 * in its Via headers, and the hash part of that Via header's branch ID
 * corresponds to this request's request-hash.  Return false if there is a
 * loop, i.e. if this request should no longer be processed.
 */
static int RequestIsNotLooped(request_t *r)
{
  sip_via_t via;
  int idx;

  /* Check for Via loops */
  for (idx = 1, via = SIP_Via(&r->p, idx, 0, config.ProxyName);
       via.hostname != NULL;
       idx++, via = SIP_Via(&r->p, idx, 0, config.ProxyName)) {
    if (via.islocal && via.branch.hash == r->hash) {
      /* Via loop! */
      /* XXX: check for hash collisions */
      ResponseSetStatus(r, 482, NULL);
      SendFailureResponse(r, 0);
      free_valid(via.hostname);
      free_valid(via.branch.str);
      return 0;
    }
    free_valid(via.hostname);
    free_valid(via.branch.str);
  }

  return 1;
} /* RequestIsNotLooped */






#define REQ1_REMOVE_FOR_SIPS
#ifndef REQ1_REMOVE_FOR_SIPS

/*
* RequestProcessThread()
* Process a SIP request according to the request type.
* This is the thread which actually does the processing work.
*/
static void *RequestProcessThread(void* ptr)
{
  char *method;                /* method of request */
  request_t *rp;               /* matching earlier request */
  parser_header_t *h;
  request_t *r = (request_t *) ptr;

  debug("RequestProcessThread", "entering\n");

  /* These items are needed for ResponseBuildDefault or ResponseSend -- set
   * them early so they're set for failure responses. */
  r->cseq         = SIP_CSeq(&r->p);
  strassign(&r->callID, SIP_CallID(&r->p));
  SetRequestVia(r);

  debug("RequestProcessThread", "callid %s\n",
        r->callID != NULL ? r->callID : "NULL");

  if (!ValidateIncomingRequest(r)) {
    debug("RequestProcessThread", "ValidateIncomingRequest failed\n");
    return NULL;
  }

  ResponseSetStatus(r, 100, NULL);
  r->version = HTTP_Version(&r->p);

  method = HTTP_Method(&r->p);

  r->contact      = SIP_Contact(&r->p);
  r->expires      = SIP_Expires(&r->p);

  r->to           = SIP_To(&r->p);
  r->from         = SIP_From(&r->p);

  r->hash         = ComputeRequestHash(r);

  if (!RequestIsNotLooped(r)) {
    debug("RequestProcessThread", "request is looped\n");
    return NULL;
  }

  if ((h = RFC822_Header(r->p.headers, "Content-Type")))
    r->request_body.content_type = h->elements->value;

  DRAW_MESSAGE_FLOW({
    DrawMessageFlow("SIPAPI","Remote","This",method);
  });

  /* Does this match an existing request? */
  rp = RequestSearch(r, r->hash);
  debug("RequestProcessThread", "Request %s rp=%p r=%p\n", method, rp, r);

  if (rp) {
    debug("RequestProcessThread", "Duplicate request\n");
    if (!strcmp(method, "ACK")) {
      debug("RequestProcessThread", "ACK\n");
      /* Method is ACK */
      if (rp->sent_200) {
        /* This request (which is probably a proxy) has sent a 200, so
         * this ACK probably needs to be handled end-to-end.  Hand it off to
         * the policy core. */
        debug("RequestProcessThread", "Already sent 200\n");
        MessageEnqueue(&rp->mqueue, message_real_message, &r->p,
                       &r->request_body, r->source_sin,NULL,
                       nobody_handled);
        /* XXX the rp->acked should be set to 1 regardless
         * of whether rp->sent_200 is true or not.
         * Jonathan, is this correct ? */
        rp->acked = 1;
      } else {
        /* just set flag. */
        debug("RequestProcessThread", "not sent 200\n");
        rp->acked = 1;
      }
      pthread_cond_broadcast(&rp->cond);
      /* Don't send a response to an ACK. */
      r->status = 0;
    }
    else if (!strcmp(method, "CANCEL")) {
      debug("RequestProcessThread", "CANCEL\n");
      /* Method is CANCEL. */
      /* Hand it up to the policy core. */
      MessageEnqueue(&rp->mqueue, message_real_message, &r->p,
                     &r->request_body, r->source_sin, NULL, nobody_handled);


      pthread_cond_broadcast(&rp->cond);

      /* Send the CANCEL response here, regardless of what the policy
         does with it.  (XXX Is this right???) */
      ResponseSetStatus(r, 200, NULL);
      ResponseBuildDefault(r);
    }
    else {
      /*
       * Method is BYE, INVITE, OPTIONS, REGISTER.  If existing request,
       * copy response from rp, but have to use new Via.
       */
      debug("RequestProcessThread", "just copying response\n");
      ResponseCopy(r, rp);
    }
    pthread_mutex_unlock(&rp->mutex);
    if (r->status > 0) {
      pthread_mutex_lock(&r->mutex);
      ResponseSend(r, 0, 1); /* Set cleanup status to 1 */
      debug("RequestProcessThread", "status +ve deleted\n");
    } else {
      debug("RequestProcessThread", "Do not call ResponseSend %d\n",
            r->status);
      RequestFree(r);
    }
  }
  else {
    /* New request. */
    if (!strcmp(method, "ACK")) {
      debug("RequestProcessThread", "ACK for unknown request\n");
      /* ACK of an unknown request -- ignore. */
      RequestFree(r);
    }
    else if (!strcmp(method, "CANCEL")) {
      debug("RequestProcessThread", "request is being CANCELed\n");
      /* CANCEL of an unknown request */
      ResponseSetStatus(r, 481, NULL);
      ResponseBuildDefault(r);
      ResponseSend(r, 0, 1); /* the cleanup status is 1 */
    }
    else {
      /*
       * Method is BYE, INVITE, OPTIONS, REGISTER.  Insert at beginning
       * of request list and handle as appropriate.  Once the request
       * goes on the request list it's visible to other threads, so it
       * must be locked.  (It's also necessary to have the mutex locked
       * before calling ResponseSend() -- this covers it for INVITE and
       * REGISTER.)
       */
      pthread_mutex_lock(&r->mutex);
      debug("RequestProcessThread", "Request is added in list\n");
      RequestAdd(r);

      /*
       * OnIncomingRequest() returns -1 if the request is
       * not handled by this function. Otherwise
       * it returns 0. If return value is 0 then
       * no further action need to be taken here.
       */
      if ( OnIncomingRequest(r) < 0 ) {
         /* libsip++ is not present. Process the
          * request here itself */
         if (!strcmp(method, "REGISTER") ) {
           REGISTER((void *)r);
         }
         else {
           /* All other types of requests are handled by Method(). */
           Method((void *)r);
         }
      }
    }
  }
  free_valid(method);
  return NULL;
} /* RequestProcessThread */


/*
 * RequestProcess()
 * Process a SIP request or response; create the appropriate thread to do
 * the actual processing work.
 */
void RequestProcess(request_t *r)
{
  pthread_t th;                 /* id of new thread */
  int thr_err;                  /* thread error indication */
  void* (*thread_func)(void *); /* Function to start */

  if (r->p.type == http_request) {
    thread_func = RequestProcessThread;
  } else if (r->p.type == http_response) {
    thread_func = ResponseProcessThread;
  } else {
    /* If it's neither a request nor a response, just ignore it. */
    debug("RequestProcess", "Unknown message (neither request nor response\n");
    RequestFree(r);
    return;
  }

  debug("RequestProcess", "Creating processor thread\n");
  if ((thr_err =
       pthread_create(&th, NULL, thread_func, (void *)r)) != 0) {
    error("RequestProcess", "couldn't start thread: %d", thr_err);
  }
} /* RequestProcess */

#endif



