
/**************************************************************
* Program:  sips_wrapper.c
*
* Matthias Rusinski
* msr52@columbia.edu
*
* Contains functions for assigning values to sipd objects
* based on their corresponding Java objects.
* Used by sip servlets for sending responses and requests.
*/

#include <stdio.h>
#include <stdlib.h>     /* calloc() */
#include <assert.h>
#include <pthread.h>
#include <unistd.h>     /* close() */
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <jni.h>

#include "headers.h"
#include "JNI_util.h"
#include "policy.h"
#include "message.h"
#include "request-sip.h"
#include "sip.h"

#include "sips_wrapper.h"


/**
 * assigns values to request_t type from corresponding
 * java object
 *
 */
int build_request_t(JNIEnv *env, jobject req, request_t *r)
{
   int ret = 1;
   jclass jRCls, jMCls, jPCls, jBCls, jBrCls, jVectCls, SipHeaderCls;
   jobject obj = NULL, hdObj = NULL;
   jobject p, b, auth, to, from, cont;
   int i;
   double d;
   long l;
   char *temp;
   sip_contact_t *c;

   jRCls = (*env)->GetObjectClass(env, req);

   if((jPCls = (*env)->FindClass(env, "parser_t"))==NULL)
      return -1;
   if((jBCls = (*env)->FindClass(env, "body_t"))==NULL)
      return -1;
   if((jBrCls = (*env)->FindClass(env, "branch_t"))==NULL)
      return -1;
   if((jVectCls = (*env)->FindClass(env, "java/util/Vector"))==NULL)
      return -1;

   /* build request_r from req jobject */
   if(r==NULL) return -2;

   r->time_received = getJObjectDoubleField(env, req, jRCls, "time_received");
   i = getJObjectIntField(env, req, jRCls, "type");
   if(i==1)
      r->type = SOCK_DGRAM;
   else if(i==2)
      r->type = SOCK_STREAM;
   r->socket = getJObjectIntField(env, req, jRCls, "socket");

   r->source_sin.sin_family = AF_INET;
   temp = getJObjectStringField(env, req, jRCls, "source_sin");
   if(temp)
      r->source_sin.sin_addr.s_addr = inet_addr(temp);
   free_valid(temp); temp = NULL;
   i = getJObjectIntField(env, req, jRCls, "source_sin_port");
   r->source_sin.sin_port = htons(i);

   r->sin.sin_family = AF_INET;
   temp = getJObjectStringField(env, req, jRCls, "sin");
   if(temp) {
      r->sin.sin_addr.s_addr = inet_addr(temp);
   }
   free_valid(temp); temp = NULL;
   i = getJObjectIntField(env, req, jRCls, "sin_port");
   r->sin.sin_port = htons(i);

   p = getJObjectObjectField(env, req, jRCls, "p", "Lparser_t;");
   if(p==NULL) return -1;
   if(build_parser_t(env, p, &(r->p)) < 0)
      return -1;

   b = getJObjectObjectField(env, req, jRCls, "request_body", "Lbody_t;");
   if(b==NULL) return -1;
   if(build_body_t(env, b, &(r->request_body)) < 0)
      return -1;

   auth = getJObjectObjectField(env, req, jRCls, "authorization", "Lhttp_authorization_t;");
   if(auth==NULL) return -1;
   if(build_http_authorization_t(env, auth, &(r->authorization)) < 0)
      return -1;

   r->status = getJObjectIntField(env, req, jRCls, "status");
   temp = getJObjectStringField(env, req, jRCls, "reason");
   r->reason = temp ? strdup(temp):NULL;
   free_valid(temp); temp = NULL;

   /* headers retrieved */
   /* these headers are the headers to be sent as the request/response
      these parsed headers should also be found in parset_t of message_t
      that came to 'processSend' with this reqeust_t */
   if((SipHeaderCls = (*env)->FindClass(env, "LSipHeaderUnit;"))==NULL)
      return -1;
   hdObj = getJObjectObjectField(env, req, jRCls, "headers", "Ljava/util/LinkedList;");
   if(hdObj!=NULL) {
      int i;
      char *tmp, *hdName, *hdVal;
      int hSize  = getListSize(env, hdObj, NULL);
      for(i=0; i<hSize; i++) {
         //parser_header_t *cur;
         jobject hdUnit = getListElement(env, hdObj, NULL, i);
         if(hdUnit==NULL)
            return -1;
         tmp = getJObjectStringField(env, hdUnit, SipHeaderCls, "header");
         if(tmp){
            hdName = strdup(tmp);
            free_valid(tmp);
         }
         else
            continue;
         tmp = getJObjectStringField(env, hdUnit, SipHeaderCls, "value");
         hdVal = tmp ? strdup(tmp):NULL;
         free_valid(tmp);

         /* append to headers */
         HeaderAppend(&r->headers, hdName, hdVal);
      }
   }
   else
      return -1;

   b = NULL;
   b = getJObjectObjectField(env, req, jRCls, "response_body", "Lbody_t;");
   if(b==NULL) return -1;
   if(build_body_t(env, b, &(r->response_body)) < 0)
      return -1;

   r->server = NULL;
   temp = getJObjectStringField(env, req, jRCls, "version");
   r->version = temp ? strdup(temp):NULL;
   free_valid(temp); temp = NULL;

   obj = getJObjectObjectField(env, req, jRCls, "cseq", "Lsip_cseq_t;");
   if(obj==NULL)
      return -1;
   r->cseq.seq = getJObjectIntField(env, obj, NULL, "seq");
   temp = getJObjectStringField(env, obj, NULL, "method");
   if(temp!=NULL)
      strncpy(r->cseq.method, temp, 20);
   free_valid(temp); temp = NULL;

   temp = getJObjectStringField(env, req, jRCls, "callID");
   r->callID = temp ? strdup(temp):NULL;
   free_valid(temp);

   to = getJObjectObjectField(env, req, jRCls, "to", "Lsip_addr_t;");
   if(to==NULL)
      return -1;
   if(build_sip_addr_t(env, to, &(r->to)) < 0)
      return -1;
   from = getJObjectObjectField(env, req, jRCls, "from", "Lsip_addr_t;");
   if(from==NULL)
      return -1;
   if(build_sip_addr_t(env, from, &(r->from)) < 0)
      return -1;

   obj = NULL;
   obj = getJObjectObjectField(env, req, jRCls, "via", "Lsip_via_t;");
   if(obj==NULL)
      return -1;

   if(build_sip_via_t(env, obj, &(r->via)) < 0)
      return -1;

   r->expires = (time_t) getJObjectLongField(env, req, jRCls, "expires");

   cont = getJObjectObjectField(env, req, jRCls, "contact", "Lsip_contact_t;");
   c = (sip_contact_t *) malloc(sizeof(sip_contact_t));
   if(c==NULL)
      return -1;
   if(cont) {
      if(build_sip_contact_t(env, cont, c) < 0)
         return -1;
      r->contact = c;
   }

   temp = getJObjectStringField(env, req, jRCls, "user_agent");
   r->user_agent = temp ? strdup(temp):NULL;
   free_valid(temp);

   r->hash = (unsigned long) getJObjectLongField(env, req, jRCls, "hash");
   r->acked = getJObjectIntField(env, req, jRCls, "acked");

   obj=NULL;
   obj = getJObjectObjectField(env, req, jRCls, "authenticate", "Lhttp_authenticate_t;");
   if(obj==NULL)
      return -1;
   if(build_http_authenticate_t(env, obj, &(r->authenticate)) < 0)
      return -1;

   pthread_mutex_init(&r->mutex, NULL);
   pthread_cond_init(&r->cond, NULL);

   r->mqueue = NULL;

   r->sent_final = getJObjectIntField(env, req, jRCls, "sent_final");
   r->sent_200 = getJObjectIntField(env, req, jRCls, "sent_200");
   r->done = getJObjectIntField(env, req, jRCls, "done");

   r->policy = NULL;

   obj=NULL;
   obj = getJObjectObjectField(env, req, jRCls, "timeout", "Ltimespec;");
   if(obj!=NULL) {
      r->timeout.tv_sec = getJObjectLongField(env, obj, NULL, "tv_sec");
      r->timeout.tv_nsec = getJObjectLongField(env, obj, NULL, "tv_nsec");
   }

   r->policy_info = NULL;

   /* branch information */
   r->branches = NULL;
   r->branch_count = getJObjectIntField(env, req, jRCls, "branch_count");
   r->branch_space = getJObjectIntField(env, req, jRCls, "branch_space");
   r->branches_outstanding = getJObjectIntField(env, req, jRCls, "branches_outstanding");

   r->call = NULL;

   obj = NULL;
   obj = getJObjectObjectField(env, req, jRCls, "outbound_proxy", "Luri_t;");
   if(obj!=NULL) {
      if(build_uri_t(env, obj, &(r->outbound_proxy)) < 0)
         return -1;
   }

   i = getJObjectIntField(env, req, jRCls, "preferred_type");
   if(i==1)
      r->preferred_type = SOCK_DGRAM;
   else if(i==2)
      r->preferred_type = SOCK_STREAM;
   else
      r->preferred_type = i;

   r->next = NULL;

   return ret;
}



/**
 * Assigns values to message_t from Java object
 */
int build_message_t(JNIEnv *env, jobject msg, message_t *m)
{
   int ret = 1;
   jclass jMCls;
   jobject obj = NULL;
   jobject p, b;
   int i;
   char *temp;

   jMCls = (*env)->GetObjectClass(env, msg);

   /*
   if((jPCls = (*env)->FindClass(env, "parser_t"))==NULL)
      return -1;
   if((jBCls = (*env)->FindClass(env, "body_t"))==NULL)
      return -1;
   */

   /* build request_r from req jobject */
   if(m==NULL) return -2;

   m->type = getJObjectIntField(env, msg, jMCls, "type");
   m->handled = getJObjectIntField(env, msg, jMCls, "handled");

   p = getJObjectObjectField(env, msg, jMCls, "p", "Lparser_t;");
   if(p==NULL) return -1;
   if(build_parser_t(env, p, &(m->p)) < 0)
      return -1;

   b = getJObjectObjectField(env, msg, jMCls, "body", "Lbody_t;");
   if(b==NULL) return -1;
   if(build_body_t(env, b, &(m->body)) < 0)
      return -1;

   m->source.sin_family = AF_INET;
   temp = getJObjectStringField(env, msg, jMCls, "source");
   if(temp)
      m->source.sin_addr.s_addr = inet_addr(temp);
   free_valid(temp); temp = NULL;
   i = getJObjectIntField(env, msg, jMCls, "source_port");
   m->source.sin_port = htons(i);

   m->branch = NULL;
   m->policy_info = NULL;
   m->next = NULL;

   return ret;
}




int build_parser_t(JNIEnv *env, jobject parser, parser_t *p)
{
   jclass pCls;
   char *temp;
   int i;
   jclass headCls, elemCls, paramCls, listCls;
   jobject hList, eList, pList, uri;

   pCls = (*env)->GetObjectClass(env, parser);

   if((headCls = (*env)->FindClass(env, "parser_header_t"))==NULL)
      return -1;
   if((elemCls = (*env)->FindClass(env, "parser_element_t"))==NULL)
      return -1;
   if((paramCls = (*env)->FindClass(env, "parser_parameter_t"))==NULL)
      return -1;
   if((listCls = (*env)->FindClass(env, "java/util/LinkedList"))==NULL)
      return -1;

   p->type = getJObjectIntField(env, parser, pCls, "type");
   temp = getJObjectStringField(env, parser, pCls, "method");
   strassign(&(p->method), temp);
   free_valid(temp); temp = NULL;
   temp = getJObjectStringField(env, parser, pCls, "URI_string");
   strassign(&(p->URI_string), temp);
   free_valid(temp); temp = NULL;

   uri = getJObjectObjectField(env, parser, pCls, "URI", "Luri_t;");
   if(uri==NULL)
      return -1;
   if(build_uri_t(env, uri, &(p->URI)) < 0)
      return -1;

   p->status = getJObjectIntField(env, parser, pCls, "status");
   temp = getJObjectStringField(env, parser, pCls, "reason");
   p->reason = temp ? strdup(temp):NULL;
   free_valid(temp); temp=NULL;
   temp = getJObjectStringField(env, parser, pCls, "protocol");
   p->protocol = temp ? strdup(temp):NULL;
   free_valid(temp); temp = NULL;
   temp = getJObjectStringField(env, parser, pCls, "version");
   p->version = temp ? strdup(temp):NULL;
   free_valid(temp); temp = NULL;
   temp = getJObjectStringField(env, parser, pCls, "header_body");
   p->header_body = temp ? strdup(temp):NULL;
   free_valid(temp); temp = NULL;

   /* .. headers */
   p->headers = NULL;
   hList = getJObjectObjectField(env, parser, pCls, "headers", "Ljava/util/LinkedList;");
   if(hList!=NULL) {
      int i;
      char *tmp;
      int hSize  = getListSize(env, hList, NULL);
      for(i=hSize-1; i>=0; i--) {
         parser_header_t *cur;
         jobject jh = getListElement(env, hList, NULL, i);
         if(jh==NULL)
            return -1;
         cur = (parser_header_t *) malloc(sizeof(parser_header_t));
         /* insert at beginning */
         cur->next = p->headers;
         p->headers = cur;

         tmp = getJObjectStringField(env, jh, headCls, "name");
         cur->name = tmp ? strdup(tmp):NULL;
         free_valid(tmp);

         cur->elements = NULL;
         eList = getJObjectObjectField(env, jh, headCls, "elements", "Ljava/util/LinkedList;");
         if(eList != NULL) {
            int j;
            char *tmp2;
            int eSize = getListSize(env, eList, listCls);
            for(j=eSize-1; j>=0; j--) {
               parser_element_t *cElem;
               jobject je = getListElement(env, eList, listCls, j);
               if(je==NULL)
                  return -1;
               cElem = (parser_element_t *) malloc(sizeof(parser_element_t));

               if(cur->elements == NULL)
                  cur->last_element = cElem;
               cElem->next = cur->elements;
               cur->elements = cElem;

               tmp2 = getJObjectStringField(env, je, elemCls, "element");
               cElem->element = tmp2 ? strdup(tmp2):NULL;
               free_valid(tmp2);
               tmp2 = getJObjectStringField(env, je, elemCls, "value");
               cElem->value = tmp2 ? strdup(tmp2):NULL;
               free_valid(tmp2);
               tmp2 = getJObjectStringField(env, je, elemCls, "comment");
               cElem->comment = tmp2 ? strdup(tmp2):NULL;
               free_valid(tmp2);

               cElem->parameters = NULL;
               pList = getJObjectObjectField(env, je, elemCls, "parameters", "Ljava/util/LinkedList;");
               if(pList!=NULL) {
                  int k;
                  char *tmp3;
                  int pSize = getListSize(env, pList, listCls);
                  for(k=pSize-1; k>=0; k--) {
                     parser_parameter_t *pr = (parser_parameter_t *) malloc(sizeof(parser_parameter_t));
                     jobject jp = getListElement(env, pList, listCls, k);
                     if(jp==NULL)
                        return -1;
                     pr->next = cElem->parameters;
                     cElem->parameters = pr;

                     tmp3 = getJObjectStringField(env, jp, paramCls, "attribute");
                     pr->attribute = tmp3 ? strdup(tmp3):NULL;
                     free_valid(tmp3);
                     tmp3 = getJObjectStringField(env, jp, paramCls, "value");
                     pr->value = tmp3 ? strdup(tmp3):NULL;
                     free_valid(tmp3);
                     pr->value_type = getJObjectIntField(env, jp, paramCls, "value_type");
                  }
               }
            }
         }
      }
   }
   p->header_length = getJObjectIntField(env, parser, pCls, "header_length");
}



int build_http_authorization_t(JNIEnv *env, jobject auth, http_authorization_t *a)
{
   jclass aCls;
   char *temp;
   int i;

   aCls = (*env)->GetObjectClass(env, auth);
   temp = getJObjectStringField(env, auth, aCls, "http_method");
   a->http_method = temp?strdup(temp):NULL;
   free_valid(temp); temp = NULL;
   temp = getJObjectStringField(env, auth, aCls, "method");
   a->method = temp?strdup(temp):NULL;
   free_valid(temp); temp = NULL;
   temp = getJObjectStringField(env, auth, aCls, "userid");
   a->userid = temp?strdup(temp):NULL;
   free_valid(temp); temp = NULL;
   temp = getJObjectStringField(env, auth, aCls, "userid_mapped");
   a->userid_mapped = temp?strdup(temp):NULL;
   free_valid(temp); temp = NULL;
   temp = getJObjectStringField(env, auth, aCls, "password");
   a->password = temp?strdup(temp):NULL;
   free_valid(temp); temp = NULL;
   temp = getJObjectStringField(env, auth, aCls, "realm");
   a->realm = temp?strdup(temp):NULL;
   free_valid(temp); temp = NULL;
   temp = getJObjectStringField(env, auth, aCls, "nonce");
   a->nonce = temp?strdup(temp):NULL;
   free_valid(temp); temp = NULL;
   temp = getJObjectStringField(env, auth, aCls, "digest_uri");
   a->digest_uri = temp?strdup(temp):NULL;
   free_valid(temp); temp = NULL;
   temp = getJObjectStringField(env, auth, aCls, "response_digest");
   a->response_digest = temp?strdup(temp):NULL;
   free_valid(temp); temp = NULL;
   temp = getJObjectStringField(env, auth, aCls, "entity_digest");
   a->entity_digest = temp?strdup(temp):NULL;
   free_valid(temp); temp = NULL;
   temp = getJObjectStringField(env, auth, aCls, "opaque");
   a->opaque = temp?strdup(temp):NULL;
   free_valid(temp); temp = NULL;
   temp = getJObjectStringField(env, auth, aCls, "algorithm");
   a->algorithm = temp?strdup(temp):NULL;
   free_valid(temp); temp = NULL;
   temp = getJObjectStringField(env, auth, aCls, "version");
   a->version = temp?strdup(temp):NULL;
   free_valid(temp); temp = NULL;
   temp = getJObjectStringField(env, auth, aCls, "signed_by");
   a->signed_by = temp?strdup(temp):NULL;
   free_valid(temp); temp = NULL;
   temp = getJObjectStringField(env, auth, aCls, "signature");
   a->signature = temp?strdup(temp):NULL;
   free_valid(temp); temp = NULL;
   a->valid = getJObjectIntField(env, auth, aCls, "valid");

   return 1;
}



int build_body_t(JNIEnv *env, jobject body, body_t *b)
{
   jclass bCls;
   char *hold;
   int i;

   bCls = (*env)->GetObjectClass(env, body);
   if(bCls==NULL)
      return -1;

   hold = getJObjectStringField(env, body, bCls, "content_type");
   b->content_type = (hold ? strdup(hold) : NULL);
   free_valid(hold); hold = NULL;
   b->content_length = getJObjectIntField(env, body, bCls, "content_length");
   hold = getJObjectStringField(env, body, bCls, "content");
   b->content = (hold ? strdup(hold) : NULL);
   free_valid(hold); hold = NULL;
   return 1;
}



int build_http_authenticate_t(JNIEnv *env, jobject auth, http_authenticate_t *a)
{
   jclass aCls;
   char *temp;
   int i;

   aCls = (*env)->GetObjectClass(env, auth);
   temp = getJObjectStringField(env, auth, aCls, "http_method");
   a->http_method = temp?strdup(temp):NULL;
   free_valid(temp); temp=NULL;
   temp = getJObjectStringField(env, auth, aCls, "method");
   a->method = temp?strdup(temp):NULL;
   free_valid(temp); temp=NULL;
   temp = getJObjectStringField(env, auth, aCls, "realm");
   a->realm = temp?strdup(temp):NULL;
   free_valid(temp); temp=NULL;
   temp = getJObjectStringField(env, auth, aCls, "domain");
   a->domain = temp?strdup(temp):NULL;
   free_valid(temp); temp=NULL;
   temp = getJObjectStringField(env, auth, aCls, "nonce");
   a->nonce = temp?strdup(temp):NULL;
   free_valid(temp); temp=NULL;
   temp = getJObjectStringField(env, auth, aCls, "opaque");
   a->opaque = temp?strdup(temp):NULL;
   free_valid(temp); temp=NULL;
   temp = getJObjectStringField(env, auth, aCls, "stale");
   a->stale = temp?strdup(temp):NULL;
   free_valid(temp); temp=NULL;
   temp = getJObjectStringField(env, auth, aCls, "algorithm");
   a->algorithm = temp?strdup(temp):NULL;
   free_valid(temp); temp=NULL;

   return 1;
}



int build_sip_addr_t(JNIEnv *env, jobject addr, sip_addr_t *ad)
{
   int ret = -1;
   jclass cls;
   char *temp;
   jobject uri;

   cls = (*env)->GetObjectClass(env, addr);

   temp = getJObjectStringField(env, addr, cls, "display_name");
   ad->display_name = temp?strdup(temp):NULL;
   free_valid(temp); temp=NULL;

   uri = getJObjectObjectField(env, addr, cls, "uri", "Luri_t;");
   if(uri==NULL)
      return -1;
   if(build_uri_t(env, uri, &(ad->uri)) < 0)
      return -1;
}



int build_uri_t(JNIEnv *env, jobject uri, uri_t *u)
{
   jclass cls;
   char *temp;
   int i;
   jobject hdrs, prms, hdList, prList;
   uri_nv_t *h, *p;

   cls = (*env)->GetObjectClass(env, uri);
   if(cls==NULL)
      return -1;

   temp = getJObjectStringField(env, uri, cls, "scheme");
   u->scheme = (temp ? strdup(temp) : NULL);
   free_valid(temp); temp=NULL;
   temp = getJObjectStringField(env, uri, cls, "user");
   u->user = (temp ? strdup(temp) : NULL);
   free_valid(temp); temp=NULL;
   temp = getJObjectStringField(env, uri, cls, "password");
   u->password = (temp ? strdup(temp) : NULL);
   free_valid(temp); temp=NULL;
   temp = getJObjectStringField(env, uri, cls, "host");
   u->host = (temp ? strdup(temp) : NULL);
   free_valid(temp); temp=NULL;
   u-> port= getJObjectIntField(env, uri, cls, "port");
   temp = getJObjectStringField(env, uri, cls, "path");
   u->path = (temp ? strdup(temp) : NULL);
   free_valid(temp); temp=NULL;
   temp = getJObjectStringField(env, uri, cls, "param");
   u->param = (temp ? strdup(temp) : NULL);
   free_valid(temp); temp=NULL;
   temp = getJObjectStringField(env, uri, cls, "search");
   u->search = (temp ? strdup(temp) : NULL);
   free_valid(temp); temp=NULL;

   hdrs = getJObjectObjectField(env, uri, cls, "headers", "Ljava/util/HashMap;");
   if(hdrs!=NULL) {
      hdList = getMapKeyList(env, hdrs, NULL);
      if(hdList!=NULL) {
         int hSize  = getListSize(env, hdList, NULL);
         char *key, *val;
         uri_nv_t *h = NULL;
         if(hSize>=0)
            h = (uri_nv_t *) malloc( (hSize+1)*sizeof(uri_nv_t));
         for(i=0; i<hSize; i++) {
            jobject jh = getListElement(env, hdList, NULL, i);
            jobject jv;
            if(jh!=NULL)
               jv = getMapValue(env, hdrs, NULL, jh);
            else
               break;
            key = getUTFString(env, jh);
            val = jv ? getUTFString(env, jv) : NULL;
            h[i].name = key ? strdup(key):NULL;
            h[i].value = val ? strdup(val):NULL;
            free_valid(key);
            free_valid(val);
         }
         if(hSize>=0) {
            h[hSize].name = NULL;
            h[hSize].value = NULL;
         }
         u->headers = h;
      }
   }
   prms = getJObjectObjectField(env, uri, cls, "parameters", "Ljava/util/HashMap;");
   if(prms!=NULL) {
      prList = getMapKeyList(env, prms, NULL);
      if(prList!=NULL) {
         int hSize  = getListSize(env, prList, NULL);
         char *key, *val;
         uri_nv_t *h = NULL;
         if(hSize>=0)
            h = (uri_nv_t *) malloc( (hSize+1)*sizeof(uri_nv_t));
         for(i=0; i<hSize; i++) {
            jobject jh = getListElement(env, prList, NULL, i);
            jobject jv;
            if(jh!=NULL)
               jv = getMapValue(env, prms, NULL, jh);
            else
               break;
            key = getUTFString(env, jh);
            val = jv ? getUTFString(env, jv) : NULL;
            h[i].name = key ? strdup(key):NULL;
            h[i].value = val ? strdup(val):NULL;
            free_valid(key);
            free_valid(val);
         }
         if(hSize>=0) {
            h[hSize].name = NULL;
            h[hSize].value = NULL;
         }
         u->parameters = h;
      }
   }
}



int build_sip_via_t(JNIEnv *env, jobject via, sip_via_t *v)
{
   int ret = -1;
   int i;
   jclass cls;
   char *temp;
   long l;

   cls = (*env)->GetObjectClass(env, via);

   i = getJObjectIntField(env, via, cls, "type");
   if(i==1)
      v->type = SOCK_DGRAM;
   else if(i==2)
      v->type = SOCK_STREAM;

   v->sin.sin_family = AF_INET;
   temp = getJObjectStringField(env, via, cls, "sin");
   if(temp)
      v->sin.sin_addr.s_addr = inet_addr(temp);
   free_valid(temp); temp = NULL;
   i = getJObjectIntField(env, via, cls, "sin_port");
   v->sin.sin_port = htons(i);

   temp = getJObjectStringField(env, via, cls, "received");
   if(temp)
      v->received.s_addr = inet_addr(temp);
   free_valid(temp); temp=NULL;

   temp = getJObjectStringField(env, via, cls, "hostname");
   v->hostname = temp?strdup(temp):NULL;
   free_valid(temp); temp=NULL;
   v->islocal = getJObjectIntField(env, via, cls, "islocal");

   temp = getJObjectStringField(env, via, cls, "b_str");
   v->branch.str = temp?strdup(temp):NULL;
   free_valid(temp); temp=NULL;
   l = getJObjectLongField(env, via, cls, "b_hash");
   v->branch.hash = (unsigned long) l;
   i = getJObjectIntField(env, via, cls, "b_id");
   v->branch.id = i;

   return 1;
}



int build_sip_contact_t(JNIEnv *env, jobject contact, sip_contact_t *c)
{
   int ret = -1;
   jclass cls;
   char *temp;
   jobject uri;
   int i;
   double d;

   cls = (*env)->GetObjectClass(env, contact);

   c->expires = (time_t) getJObjectLongField(env, contact, cls, "expires");
   c->action = getJObjectIntField(env, contact, cls, "action");
   d = getJObjectDoubleField(env, contact, cls, "q");
   c->q = (float) d;

   uri = getJObjectObjectField(env, contact, cls, "uri", "Luri_t;");
   if(uri==NULL)
      return -1;
   if(build_uri_t(env, uri, &(c->uri)) < 0)
      return -1;

   temp = getJObjectStringField(env, contact, cls, "nameaddr");
   c->nameaddr = temp?strdup(temp):NULL;
   free_valid(temp); temp=NULL;
   temp = getJObjectStringField(env, contact, cls, "comment");
   c->comment = temp?strdup(temp):NULL;
   free_valid(temp); temp=NULL;

   return 1;
}