package org.ietf.sip;

import java.util.Iterator;

/**
 * Represents SIP Contact addresses, [RFC 2543, section 6.13].
 * 
 * <p>A Contact extends a SipAddress with three well-defined parameters:
 * the q-value, and action, and an expires indication. This interfaces
 * provides getters and setters for each of these.
 */
public interface Contact extends SipAddress {
    /**
     * Returns a static Contact instance representing the special "wildcard"
     * Contact representing the "*" wildcard which can be used as the
     * value of the Contact header in REGISTER requests, see
     * [RFC 2543, section 6.13].
     * 
     * <p>Applications are guaranteed always to have the same instance
     * returned by invocation sof this method.
     */
    public Contact getWildcard();
    
    /**
     * Returns the qvalue of this contact. This defaults to 1.0.
     */
    public float getQ();
    
    /**
     * Set the qvalue of this Contact. A higher value indicates higher
     * preference.
     * @param q the new "quality value". Should be between 0 and 1.
     */
    public void setQ(float q);
    
    /**
     * Returns the value of the action parameter of this Contact. This
     * indicates a client preference for proxying or redirecting requests.
     * This method may return null.
     */
    public String getAction();
    
    /**
     * Sets the action parameter of this Contact.
     * @param action    the new value of this Contacts action attribute.
     *                  Should be null, "proxy", or "redirect".
     */
    public void setAction(String action);
    
    /**
     * Returns the value of the expires parameter of this Contact. This is
     * treated as an opaque String by class Contact, but must conform to
     * RFC 2543, i.e. can either be an integer specifying delta-seconds,
     * or an absolute RFC 1123 formatted datetime.
     * 
     * <p>Use SipUtils utility methods to obtain corresponding numeric values,
     * either as delta-seconds or as absolute Java time.
     * 
     * @see SipUtils#getJavaTime
     * @see SipUtils#getSipDeltaTime
     */
    public String getExpires();
    
    /**
     * Sets the value of the expires header.
     * @param expires the expiration time of this Contact. Should be an
     *  absolute time in RFC 1123 format or an integer specifying
     *  delta-seconds.
     */
    public void setExpires(String expires);
}
