package org.ietf.sip;

import java.util.Iterator;

/**
 * Represents value of SIP From and To headers and serves as the base
 * class for the representation of Contact header values.
 * 
 * <p>A SipAddress can be thought of as representing something that conforms
 * to the following grammar [RFC 2543, section 6.21]:
 * <pre>
 *   SipAddress     = (name-addr | addr-spec) *(";" addr-params)
 *   name-addr      = [display-name] "<" addr-spec ">"
 *   addr-spec      = SIP-URL | URI
 *   display-name   = *token | quoted-string
 * </pre>
 * 
 * <p>Note that a SipAddress can encapsulate either a SIP URL or an arbitrary
 * non-SIP URI. These are represented in this class by SipURL and String
 * attributes respectively, exactly one of which must (normally) be non-null.
 */
public interface SipAddress {
    /**
     * Returns the display name associated with this SipAddress.
     * This may be null.
     */
    public String getDisplayName();
    
    /**
     * Returns the SipURL associated with this instance, or null if this
     * SipAddress encapsulates a non-SIP URI.
     */
    public SipURL getSipURL();
    
    /**
     * Returns the non-SIP URI represented by this instance, or null if
     * this SipAddress encapsulates a SIP URL.
     */
    public String getURI();
    
    /**
     * Returns the value of the named parameter.
     */
    public String getParameter(String name);
    
    /**
     * Returns an iterator over all parameter names (Strings) set in this
     * SipAddress.
     */
    public Iterator getParameterNames();
    
    /*
    //XXX: keep this interface immutable?
    public void setDisplayName(String name);
    public void setSipURL(SipURL url);
    public void setURL(String url);
    public void addParameter(String name, String value);
    public void removeParameter(String name);
    */
}
