package org.ietf.sip;

import java.security.AccessControlException;

/**
 * Factory methods for SipURLs, Contacts, SipAddresses, and Transactions.
 * @see SipTransaction#createRequest
 * @see SipRequest#createResponse
 */
public interface SipFactory {
    /**
     * Returns a new SIP transaction object with the specified parameters.
     * This can be used to crate a new SipRequest object which is used to
     * initiate new transactions.
     * 
     * <p>The servlet engine generates a new Call-ID and CSeq number for the
     * transaction.
     * 
     * <p>Note that in order to receive responses to requests made within the
     * returned transaction the caller must register itself as a listener with
     * the transaction.
     * 
     * @param from      the From header of the returned transaction
     * @param to        the To header of the returned transaction
     */
    public SipTransaction createSipTransaction(SipAddress from, SipAddress to);
    
    /**
     * Parses the argument as a SIP URL.
     * @param sipURL    a sip: url as defined in RFC 2543, section 2.
     * @throws ParseException   if the argument is not a valid SIP URL
     */
    public SipURL createSipURL(String sipURL) throws ParseException;
    
    /**
     * Creates a new SIP URL with the specified user and host and the
     * default SIP port number.
     * @param user  the user part if the SIP URL. This must be null for
     *              SipURLs used as request-URIs in REGISTER requests.
     * @param host  the host part of the SIP URL. This may be either a
     *              domain name or a dotted-decimal IP address.
     */
    public SipURL createSipURL(String user, String host);
    
    /**
     * Creates a new SIP URL with the specified user, host and port number.
     * @param user  the user part if the SIP URL. This must be null for
     *              SipURLs used as request-URIs in REGISTER requests.
     * @param host  the host part of the SIP URL. This may be either a
     *              domain name or a dotted-decimal IP address.
     * @param port  the SIP URL port number
     */
    public SipURL createSipURL(String user, String host, int port);
    
    /**
     * Parses the specified string as a SipAddress, suitable for example,
     * for use as the From or To field of SIP messages.
     * @throws ParseException   if the argument is not a valid SipAddress
     * @see SipAddress
     */
    public SipAddress createSipAddress(String sipAddress) throws ParseException;
    
    /**
     * Creates a new Sipaddress with the specified SipURL and no display-name.
     */
    public SipAddress createSipAddress(SipURL sipURL);
    
    /**
     * Creates a new SipAddress with the specified SipURL and display-name.
     */
    public SipAddress createSipAddress(SipURL sipURL, String displayName);
    
    
    /**
     * Parses the argument as a Contact specification.
     * @throws ParseException   if the argument is not a valid Contact
     */
    public Contact createContact(String contact) throws ParseException;
    
    /**
     * Creates a Contact with the specified SipURL and display-name.
     */
    public Contact createContact(SipURL url, String displayName);
    
    /**
     * Creates a Contact with the specified URI and display-name. The URI
     * need not be a SIP URL.
     */
    public Contact createContact(String uri, String displayName);
    
    /**
     * Creates a Contact with details copied from the specified SIP address.
     */
    public Contact createContact(SipAddress addr);
}
