package org.ietf.sip;

import java.util.Iterator;
import java.util.List;

/**
 * Base interface for SipRequest and SipResponse.
 * 
 * <p>As SIP entities can receive and send requests as well as responses,
 * there is a need for SIP servlets to be able to set and get headers,
 * content, and other information in both requests and response. This
 * interface reflects the commonality between all SIP messages.
 */
public interface SipMessage extends SipConstants {
    /**
     * Returns the SIP method of this message. For requests this is the
     * value used in the request line and for responses it is taken from
     * the CSeq header field, which echoes the method used in the
     * corresponding request.
     */
    public String getMethod();
    
    /**
     * Returns the name and version of the protocol of this message.
     * This is of the form &lt;protocol> "/" &lt;major-version-number> "."
     * &lt;minor-version-number>, e.g. "SIP/2.0".
     */
    public String getProtocol();
    
    /**
     * Returns the IP address of the host sending this message, or the
     * loopback address if this message was created locally.
     * 
     * <p>Note that the address returned is that of the previous hop,
     * which will generally not be that of the host originating the
     * request.
     */
    public String getRemoteAddr();
    
    /**
     * Returns the fully qualified domain name of the host sending the
     * message, or null if this request was created locally. The returned
     * string may be a dotted-decimal IP address if the server cannot or
     * chooses not to resolve IP addresses.
     */
    public String getRemoteHost();
    
    /**
     * Returns the host name used to access this server. May be either a
     * fully qualified domain name or a dotted decimal IP address.
     */
    public String getServerName();
    
    /**
     * Returns the port number on which the message was received, or -1 if
     * this message was created lcoally.
     */
    public int getServerPort();
    
    /**
     * Returns the name of the transport protocol used to receive this
     * message, e.g. "TCP" or "UDP". Returns null for locally created
     * messages.
     */
    public String getServerProtocol();
    
    
    // ~~~~ headers ~~~~
    
    /**
     * Returns the transaction associated with this message.
     */
    public SipTransaction getTransaction();
    
    /**
     * Returns the value of the specified header field.
     */
    public String getHeader(String name);
    
    /**
     * Returns an iterator over all headers in the message (Strings).
     */
    public Iterator getHeaderNames();
    
    /**
     * Returns an iterator over the values of all headers with the
     * specified name.
     * @param name  the header name
     */
    public Iterator getHeaders(String name);
    
    /**
     * Sets a header field with the specified name and value. If a header
     * with the same name existed it is overwritten.
     * @param name  the header name
     * @param value the header value
     */
    public void setHeader(String name, String value);
    
    /**
     * Adds a header field with the specified name and value.
     * @param name  the header name
     * @param value the header value
     */
    public void addHeader(String name, String value);
    
    /**
     * Returns the Call-ID of this transaction.  Equivalent to
     * <code>getTransaction().getCallID()</code>.
     * @see SipTransaction#getCallID
     */
    public String getCallID();

    /**
     * Returns the parsed value of the From header field.
     * Equivalent to <code>getTransaction().getFrom()</code>.
     * @see SipTransaction#getFrom
     */
    public SipAddress getFrom();
    
    /**
     * Returns the parsed value of the To header field.
     * Equivalent to <code>getTransaction().getTo()</code>.
     * @see SipTransaction#getTo
     */
    public SipAddress getTo();
    
    /**
     * Returns a list of all Contacts in this message. This is a List of
     * Contact objects corresponding to the parsed value of all Contact
     * headers in this message.
     */
    public List getContacts();
    
    /**
     * Sets the set of Contact objects for this message.
     * @param contacts  a list of Contact objects.
     */
    public void setContacts(List contacts);
    
    // ~~~~ content ~~~~
    
    
    /**
     * Returns the length in number of bytes of the content part of this
     * message. This directly reflects the value of the Content-Length header
     * field.
     */
    public int getContentLength();
    
    
    /**
     * Returns the value of the Content-Type header field.
     */
    public String getContentType();
    
    /**
     * Sets the value of the Content-Type header field.
     * @param type  the MIME type of the message body
     */
    public void setContentType(String type);
    
    
    /**
     * Returns the content of this message as a byte array.
     */
    public byte[] getContent();
    
    /**
     * Attempts to parse the content of this message. The type of object
     * returned will depend on the Content-Type of the message. A server is
     * required at a minimum to be able to parse SDP [RFC 2327]. A future
     * version of this API will include an interface for accessing SDP and
     * pin down what this method returns for SDP.
     *
     * <p>If this message has a MIME multipart Content-Type, an object
     * representing the entire multipart body will be returned (XXX: a
     * <code>javax.mail.Multipart</code>?). To obtain just the media
     * description, use getMediaDescription().
     * 
     * @return an object representing the body of the message, or null if
     *      the content length is 0.
     * @throws ParseException   if the server does not know how to parse
     *      the content type of the body.
     */
    public Object getParsedContent() throws ParseException;
    
    /**
     * Returns an object representing the media description contained within
     * the body of this message, or null if the body is empty or no recognized
     * media format was found.
     * 
     * <p>Unlike getParsedContent() this method will descend into a multipart
     * body to find the part containing the medis description.
     */
    public Object getMediaDescription() throws ParseException;
    
    /**
     * Sets the body of the content to the specified byte array.
     * @param buf   the raw bytes constituting the message body.
     */
    public void setContent(byte[] buf);
}
