package org.ietf.sip;

import java.security.AccessControlException;

/**
 * Represents SIP requests. This class provides access to all parts of SIP
 * requests, both for incoming and outgoing messages.
 */
public interface SipRequest extends SipMessage {
    /**
     * Returns the SIP URL as it appeared in the request line of this request,
     * or null if this object was created locally.
     */
    public SipURL getRequestURI();
    
    /**
     * Returns the authentication scheme used for this message. This is the
     * value of the <em>auth-scheme</em> part of the Authorization header,
     * and can take the values "basic", "digest", and "pgp".
     */
    public String getAuthType();
    
    /**
     * Returns the identity of the caller as verified by the authentication
     * protocol, or null if the request hasn't been authenticated.
     */
    public String getRemoteUser();
    
    /**
     * Add a Record-Route header for this proxy to the request.
     * This is only relevant for requests being proxied.
     * @param recordIt  true if a Record-Route header should be added to
     *                  the message for this proxy, and false if it should
     *                  be removed.
     * @return  true if the operation succeeded, and false otherwise
     */
    public boolean recordRoute(boolean recordIt);
    
    /**
     * Creates a response for this request. The request object will have
     * From, To, Call-ID, CSeq, Via etc. initialized from this request
     * object, and the servlet shouldn't attempt to set these.
     */
    public SipResponse createResponse();
    
    /**
     * Send this message to the specified callee. This method is used both
     * for initiating new requests and to forward received requests in
     * proxies.
     * 
     * @param url   the address to which to send this request. This will
     *              appear in the request URI of the message.
     * @return  a token which can be used for matching responses to this
     *          request. Should only be used for testing reference equality.
     */
    public Object send(SipURL url);
}
