package org.ietf.sip;

import java.util.Iterator;

/**
 * A simple abstract adapter class for SIP servlets.
 * 
 * <p>In addition to providing default implementations for all methods,
 * also dispatches incoming request to other methods based on the SIP
 * request method.
 * 
 * The default implementation of callbacks (gotRequest, gotResponse, do*)
 * all return false to indicate to the servlet engine that it should perform
 * the default action for the message.
 */
public abstract class SipServletAdapter implements SipServlet, ServletConfig {
    protected ServletConfig config;
    
    /**
     * Invoked by the servlet engine to initialize this servlet. This may
     * be a good place to acquire resources. Implementations will typically
     * store the config parameter so as to be able to return it in subsequent
     * calls to getServletConfig.
     * 
     * <p>Servlets that override this method should remember to invoke
     * <code>super(config)</code>.
     * 
     * @param config    provides access to parameters of this servlet
     *                  as well as to the servlet context.
     */
    public void init(ServletConfig config) {
        this.config = config;
    }
    
    /**
     * Invoked for incoming SIP requests which are associated with this
     * servlet. The mechanism for associating SIP messages with servlets
     * is implementation dependent.
     * 
     * <p>This method dispatches the request object to one of the do*
     * methods based on the request method.
     * 
     * @return false if the server should apply default treatment to this
     *               request and true otherwise
     */
    public boolean gotRequest(SipRequest req)
        throws ServletException 
    {
        String method = req.getMethod();
        if (method.equals(INVITE)) {
            return doInvite(req);
        } else if (method.equals(ACK)) {
            return doAck(req);
        } else if (method.equals(OPTIONS)) {
            return doOptions(req);
        } else if (method.equals(BYE)) {
            return doBye(req);
        } else if (method.equals(CANCEL)) {
            return doCancel(req);
        } else if (method.equals(REGISTER)) {
            return doRegister(req);
        } else {
            return false;
        }
    }
    
    /**
     * Invoked for INVITE requests. By default just returns false.
     * @return false if the server should apply default treatment to this
     *               request and true otherwise
     */
    public boolean doInvite(SipRequest req) throws ServletException {
        return false;
    }
    
    /**
     * Invoked for ACK requests. By default just returns false.
     * @return false if the server should apply default treatment to this
     *               request and true otherwise
     */
    public boolean doAck(SipRequest req) throws ServletException {
        return false;
    }
    
    /**
     * Invoked for OPTIONS requests. By default just returns false.
     * @return false if the server should apply default treatment to this
     *               request and true otherwise
     */
    public boolean doOptions(SipRequest req) throws ServletException {
        return false;
    }
    
    /**
     * Invoked for BYE requests. By default just returns false.
     * @return false if the server should apply default treatment to this
     *               request and true otherwise
     */
    public boolean doBye(SipRequest req) throws ServletException {
        return false;
    }
    
    /**
     * Invoked for CANCEL requests. By default just returns false.
     * @return false if the server should apply default treatment to this
     *               request and true otherwise
     */
    public boolean doCancel(SipRequest req) throws ServletException {
        return false;
    }
    
    /**
     * Invoked for REGISTER requests. By default just returns false.
     * @return false if the server should apply default treatment to this
     *               request and true otherwise
     */
    public boolean doRegister(SipRequest req) throws ServletException {
        return false;
    }
    
    /**
     * Invoked for SIP responses which are received in response to a
     * request which originated with, or was proxied by, this servlet.
     * 
     * <p>By default this method just returns false.
     * 
     * @return false if the server should apply default treatment to this
     *               request and true otherwise
     */
    public boolean gotResponse(SipResponse res) throws ServletException {
        return false;
    }
    
    /**
     * The empty implementation.
     */
    public void destroy() {}
    
    /**
     * Returns the ServletConfig object passed to this servlet in init().
     */
    public ServletConfig getServletConfig() {
        return config;
    }
    
    /**
     * Returns the value of a named init parameter.
     * This is a convenience method equivalent to:
     * <pre>
     *   getServletConfig().getInitParameter(name);
     * </pre>
     */
    public String getInitParameter(String name) {
        return getServletConfig().getInitParameter(name);
    }
    
    
    /**
     * Returns an iterator over the configuration parameters.
     * This is a convenience method equivalent to:
     * <pre>
     *   getServletConfig().getInitParameterNames();
     * </pre>
     */
    public Iterator getInitParameterNames() {
        return getServletConfig().getInitParameterNames();
    }
    
    
    /**
     * Returns the servlet context. This is a convenience method.
     * It is equivalent to:
     * <pre>
     *   getServletConfig().getServletContext();
     * </pre>
     */
    public SipServletContext getServletContext() {
        return getServletConfig().getServletContext();
    }
    
    /**
     * Equivalent to <code>getServletContext().log(msg, t)</code>.
     * @see SipServletContext#log(String)
     */
    public void log(String msg) {
        getServletContext().log(msg);
    }
    
    /**
     * Equivalent to <code>getServletContext().log(msg, t)</code>.
     * @see SipServletContext#log(String,Throwable)
     */
    public void log(String msg, Throwable t) {
        getServletContext().log(msg, t);
    }
    
    /**
     * Returns null.
     */
    public String getServletInfo() {
        return null;
    }
}
