package org.ietf.sip;

import java.util.Iterator;

/**
 * Represents SIP transactions. A transaction is identified by its CSeq
 * number within a call leg. As calls and call legs are not represented
 * explicitly in the API, transactions encapsulate Call-ID, From and To
 * headers as well as the CSeq number.
 * 
 * <p>These transaction parameters doesn't change during the lifetime of
 * a transaction, and hence are read-only attributes oif this class.
 * 
 * <p>SipTransactions created locally,
 * e.g. by a servlet, are initialized with values for the From and To
 * headers and the request method, whereas the value of the Call-ID, CSeq
 * number and possibly tags associated with the From or To headers cannot
 * be set by applications.
 * 
 * @see SipFactory#createSipTransaction
 */
public interface SipTransaction {
    /**
     * Returns the Call-ID of this transaction. If this transaction was
     * created locally the Call-ID is generated by the servlet engine.
     * Servlets should never attempt to set the Call-ID using
     * SipMessage.setHeader.
     */
    public String getCallID();
    
    /**
     * Returns the value of the From header associated with this transaction.
     */
    public SipAddress getFrom();
    
    /**
     * Returns the value of the To header associated with this transaction.
     */
    public SipAddress getTo();
    
    /**
     * Returns the value of the CSeq (command sequence) attribute associated
     * with this transaction. This is either the value used in an incoming
     * SIP message or one generated by the servlet engine.
     */
    public String getCSeq();
    
    /**
     * Returns a new request object belonging to this transaction.
     * 
     * <p>This is used for new transactions created using the SipFactory
     * interface for initiating requests from a servlet, but can also be
     * used by servlets in proxies in order, for example, to cancel
     * outstanding requests.
     * 
     * @method  the SIP request method.
     */
    public SipRequest createRequest(String method);
    
    /**
     * Returns the value of the specified attribute, or null if it is
     * undefined.
     */
    public Object getAttribute(String name);
    
    /**
     * Returns an iterator over all defined attributes.
     */
    public Iterator getAttributeNames();
    
    /**
     * Associates the named attribute with the specified value. A set of
     * servlets associated with the same servlet engine sees
     */
    public void setAttribute(String name, Object value);
    
    /**
     * Removes the specified attribute.
     */
    public void removeAttribute(String name);
    
    /**
     * Register the specified servlet to receive notifications of
     * incoming SIP messages within this transaction. Requests and responses received
     */
    public void addListener(SipServlet servlet);
    
    /**
     * Deregister the specified servlet for receiving notifications
     * relating to this transaction.
     */
    public void removeListener(SipServlet servlet);
}
