package org.ietf.sip;

import java.util.Iterator;

/**
 * Class SipURL represents addresses of SIP endpoints. SIP URLs are used
 * in the request-URI of SIP requests as well as in To, From, and Contact
 * header fields. See [RFC 2543, section 2] for the definiton of SIP URLs.
 * 
 * @see SipAddress
 * @see Contact
 */
public interface SipURL {
    /**
     * Returns the user part of this SipURL. This may be null if the URL was
     * obtained from, or is intended to be used in, a REGISTER request.
     */
    public String getUser();
    
    /**
     * Set the user part of this SipURL.
     */
    public void setUser(String user);
    
    /**
     * Returns the password of this SipURL. This may be null.
     */
    public String getPassword();
    
    /**
     * Sets the password associated with this SipURL.
     */
    public void setPassword(String password);
    
    /**
     * Returns the host name of this SipURL. This can be either a domain
     * name or a dotted-decimal IP address.
     */
    public String getHost();
    
    /**
     * Sets the host name of this SipURL.
     */
    public void setHost(String host);
    
    /**
     * Returns the port number of this SipURL. This is 5060 by default,
     * the default SIP port number.
     */
    public int getPort();
    
    /**
     * Sets the port number of this SipURL.
     */
    public void setPort(int port);
    
    /**
     * Returns the value associated with the named parameter, the 
     * @param name  the name of the parameter
     * @return the value of the named parameter, or the empty string
     *      for boolean parameters, and null if the parameter is not defined
     */
    public String getParameter(String name);
    
    /**
     * Sets the specified boolean parameter.
     */
    public void setParameter(String name);
    
    /**
     * Sets the named parameter to the specified value.
     */
    public void setParameter(String name, String value);
    
    /**
     * Removes the specified parameter from this SipURL.
     */
    public void removeParameter(String name);
    
    /**
     * Returns an Iterator over all specified parameters.
     * These are all Strings.
     */
    public Iterator getParameterNames();
    
    /**
     * Returns the textual representation corresponding to this SipURL.
     * This must be a valid SIP URL according to RFC 2543.
     */
    public String toString();
}
