/***************************** FILE HEADER *****************************
#               Copyright (c) 1999 by Columbia University              *
#                        All Rights Reserved                           *
#***********************************************************************
#***********************************************************************
#
#  File Name    : main.cpp
#
#  Module       : Main file for A simple SIP User Agent.
#
#  Author(s)    : Kundan Singh (kns10@cs.columbia.edu)
#
#  Created      : 25 Nov 1999
#
#  Description  : 
#                
#  Notes        : 
#
#***********************************************************************
#                Modification History (Reverse Order)
#-----------------------------------------------------------------------
#   Date         Author              Description
#-----------------------------------------------------------------------
#              |             | 
# 28 Jun 2000  | Kundan      | Restructured.
#-----------------------------------------------------------------------
*/

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "iptelapi.h"
#include "sipapi.h"
#include "radio.h"  /* Audio tool API */
#include "command.h"

#ifdef __cplusplus
extern "C" {
#endif

#ifndef WIN32 /* For unix */
#include "unistd.h"
#include <sys/types.h>
#include <sys/socket.h>
#include "netinet/in.h"
#include "arpa/inet.h"
#else
#include "winsock2.h" /* for sockets */
extern int getopt (int argc, char *const *argv, const char *shortopts);
extern char *optarg;
#endif

#ifdef __cplusplus
}
#endif

/**
 * MySIPUserAgent and MySIPCall
 */
class MySIPUserAgent : public SIPEndpoint
{
public:
  MySIPUserAgent(bool IsServer = true,bool IsRegistrar=false);
  MySIPUserAgent(const IptelIPAddress & ip,bool IsRegistrar=false);
  virtual IptelCall *AllocateNewCallObject(const char *callid);
};

class MySIPCall : public SIPCall
{
public:
  MySIPCall(const char *callid=NULL) : SIPCall(callid) { };
  MySIPCall(SIPEndpoint & ep, const char *callid = NULL) : SIPCall(ep,callid) { };
  virtual ~MySIPCall() { cout << "~MySIPCall()" << endl;/* default destructor */ };

  virtual void   OnNewIncomingCall(IptelEndpoint & ep);
  virtual int    OnHangup();
  virtual void   OnCallEstablished();
  virtual int    OnCallRejected();
  virtual void   OnNewMessageBody(const SIPMessageBody & body);
  virtual int    OnRemoteSessionChanged(const SessionDescription & sdp);
};

/* XXX should use a mutex here */
 MySIPCall *call = NULL;
 AudioTool *audio = NULL;
 bool      oBusy = false;
 bool      oLicensed = false; /* to limit features in unlicensed
                                 version */
 bool      oUseDevice = true;  /* whether to use the audio device */

 bool       measure_latency = false;

GenericHeaders globals;

//////////////// Method defn for SIP ///////////////////////

MySIPUserAgent::MySIPUserAgent(bool IsServer,bool IsRegistrar)
: SIPEndpoint(IsServer,IsRegistrar)
{
  /* Do nothing here */
}

MySIPUserAgent::MySIPUserAgent(const IptelIPAddress & ip,bool IsRegistrar)
: SIPEndpoint(ip,IsRegistrar)
{
  /* Do nothing here */
}

IptelCall *MySIPUserAgent::AllocateNewCallObject(const char *callid)
{
  /* Allocate a new SIP call */
  return new MySIPCall(*this, callid);
}

void MySIPCall::OnNewIncomingCall(IptelEndpoint & ep)
{
  IptelAddress dest    = GetSelfAddress();
  IptelAddress src     = GetRemoteAddress();

  //cout << "On incoming call from " << dest << " for " << src << endl;

  if ( oBusy ) {
    Reject(486, "Busy here");
    /* How do we delete the call object here XXX */
    return;
  }

#if 0
  if ( !Authenticate("basic", "./passwd") ) {
    cerr << "Can not be authenticated, rejecting" << endl;
    Reject(); /* Error code are already set */
    return;
  }
  else {
    cout << "Call is authenticated" << endl;
  }
#endif

  /*
   * Set the global variable for call.
   * Accept or reject is called in UI thread 
   */

  oBusy = true;

  call = this;

  cout << "Incoming call from "  << src << " to " << dest << endl;
  const char *subj = GetSubject();
  if ( subj != NULL ) {
    cout << "Subject is " << subj << endl;
  }
  cout << "accept/reject/noresponse ? " << flush;
}

int   MySIPCall::OnHangup()
{
  cout << "Call closed by remote" << endl;
  oBusy = false;
  call = NULL;
  audio->StopTool();
  SIPCall::OnHangup();
  return 0;
}

void   MySIPCall::OnCallEstablished()
{
  cout << "Call Established" << endl;
  SIPCall::OnCallEstablished();

  if ( IsActive() ) {
    
    cout << "Accepted by remote" << endl;
    
    SessionDescription *session;
    MediaInfo media;
    
    /* Get remote IP and port */
    session = GetRemoteSessionDescription();
    
    cout << "Remote Session " << *session << endl;
    
    MediaInfo rem_media;

    session->FindSession(rem_media, MediaInfo::Receive, MediaInfo::Audio);
    
    IptelIPAddress rem_addr = rem_media.GetRxAddress();

    if ( audio != NULL ) {
      struct sockaddr_in addr;
      rem_addr.GetIpAddress(addr);
      addr.sin_port = htons(rem_media.GetRxPort());
      audio->SetRemoteAddress(addr);
      audio->StartTool(measure_latency ? 0 : 1);
    }
    else {
      cerr << "No audio tool" << endl;
    }
  }  
}

int   MySIPCall::OnCallRejected()
{
  cout << "Call is Rejected by remote" << endl;
  cout << GetStatus() << " " << GetStatusString() << endl;

  oBusy = false;

  audio->StopTool();

  int s = GetStatus();

  if ( s == 401 ) {
    const char *auth = GetIncomingResponseHeader("WWW-Authenticate");
    if ( auth != NULL ) {
      cout << "WWW-Authenticate: " << auth << endl;
    }
    cout << "Set your password and try again." << endl;
  }
  else if ( ( s >= 300) && ( s < 400)) {
    const char *contact = GetIncomingResponseHeader("Contact");
    if ( contact != NULL ) {
      cout << "Contact: " << contact << endl;
    }
    cout << "Try the new address given above." << endl;
  }

  call = NULL;
  SIPCall::OnCallRejected();
  return 0;
}

void  MySIPCall::OnNewMessageBody(const SIPMessageBody &body)
{
  cout << "----- New Message Body -----\n"
       << body.GetContentType() 
       << "[" << body.GetContentDisposition() << "]"
       << "[" << body.GetContentHandling() << "]" << "\n"
       << body.GetContent() << "\n"
       << "----------------------------" << endl;
}

int   MySIPCall::OnRemoteSessionChanged(const SessionDescription & sdp)
{
  /*
   * We always accept change in SDP
   */
  cout << "Remote SDP is changed, trying to accept" << endl;
  return 0;
}

/* Main routine */

int main(int argc, char *argv[])
{
  int c;
  char cmd[200];
  unsigned short listening_port = 5060;
  char *command_file = NULL;
  bool interactive = true;
  unsigned short mediaport = 0;

  static char *mycommands = 
    " General format for any command is\n"
    " <commandname>  <list-of-options> <list-of-arguments>\n"
    " Different commands are invite, accept, reject, authenticate\n"
    " redirect, register, unregister, help, exit, quit, log, set,\n"
    " unset, and ?\n"
    " Options are of the form param=value. Every command defines\n"
    " a set of supported options. Additional options are ignored\n"
    " Typical options are\n"
    "  transport with values udp, tcp or multicast\n"
    "  expires with the expiry timeout in seconds\n"
    "  user with values ip or phone\n"
    "  sdp with the sdp identifier\n"
    "  body with the body identifier\n"
    "  cgi or cpl with the corresponding file name\n"
    " \n"
    " For example, \n"
    " you can place an outgoing call as\n"
    " invite transport=tcp sip:kns10@cs sip:ind.cs.columbia.edu\n"
    "\n"
    "Visit http://www.cs.columbia.edu/~kns10/software/sipua for\n"
    "a complete list of commands and user manual\n"
    "\n"
    ;
  
  // Others to be implemented
  // ,verbose,save filename, dump sip filename, dump h323 filename
          
  const char *servername = "SIPUA-Columbia-University";

  cout << "SIPUA, (c) 2000, Columbia University" << endl;
  cout << "Visit http://www.cs.columbia.edu/~kns10/software/sipua for more information." << endl;

  char *audio_tool = NULL;
  const char *usage = "[options]\n"
                      " -v          : verbose mode\n"
                      " -o tracefile: put debug trace in this file.\n"
                      " -a audiotool: specify a different audio tool.\n"
                      " -p port     : listening SIP port, default is 5060.\n"
                      " -f file     : command file to be executed.\n"
                      " -n          : non interactive mode, forbackground.\n"
                      " -d          : use dotted decimal for self address.\n"
                      " -t          : print timeing (latency) information.\n"
                      " -P port     : media self port number.Default is 10000\n"
    ;

  /* Parse command line options. */
  while ((c = getopt(argc, argv, "hvo:a:p:f:nP:dt")) != EOF) {
    switch (c) {
      case 'v':
        SIPLibrary::DebugOn();
        break;
      case 'o':
        SIPLibrary::DebugOn(optarg);
        break;
      case 'a':
        audio_tool = strdup(optarg);
        break;
      case 'p':
        listening_port = atoi(optarg);
        break;
      case 'f' :
        command_file = strdup(optarg);
        break;
      case 'n' :
        interactive = false;
        break;
      case 'P' :
        mediaport = atoi(optarg);
        break;
      case 'd' :
        SIPLibrary::oUseNumericAddress = true;
        break;
      case 't' :
        measure_latency = true;
        break;
      case '?':
        /* no break */
      case 'h':
        cerr << "Usage: " << argv[0] << " "  << usage << endl;
        exit(1);
    } 
  }

  /* Create the listening servers */
  IptelIPAddress listeningAddress;
  listeningAddress.SetPort(listening_port);

  MySIPUserAgent sip(listeningAddress,
                     false);
  IptelAddress localname = SIPLibrary::HostName;
  localname.SetUser("sipua");
  sip.SetLocalName(localname);
 
  int e =
    SIPLibrary::CheckLicense("sipua.license");

  /*
  if ( e > 0 ) {
    cerr << "Your registration file is invalid or has expired.\n"
      "You should obtain a new registration by "
      "sending email to hgs@cs.columbia.edu.\n" << endl;
    exit(0);
  }
  else if ( e < 0 ) {
    cerr << "Unlicensed copy. Contact hgs@cs.columbia.edu for a license." << endl;
    if ( audio_tool != NULL ) {
      cerr << "External audio tool is not supported in unlicensed version."<< endl;
      audio_tool = NULL;
    }
  }
  else if ( e == 0 ) {
    oLicensed = true;
  }
  */
  if ( e == 0 ) {
    oLicensed = true;
  }
  else {
    cerr << "Unlicensed copy. Contact hgs@cs.columbia.edu for a license." << endl;
    if ( audio_tool != NULL ) {
      cerr << "External audio tool is not supported in unlicensed version."<< endl;
      audio_tool = NULL;
    }
  }

  /* Initialize debug options */
  SIPLibrary::SetName(servername);
  SIPLibrary::ErrorOn("stderr");

  cout << "SIP " << sip << endl;
  
  /* Allocate the audio tool */
  audio = new AudioTool(audio_tool); /* use built-in audio */
  audio->SetSelfPort(mediaport);

  cout << "Waiting for connections..." << endl;

  cout << "Type help to get list of commands" << endl;

  bool oExitFlag = false;
  FILE *command_fp = NULL;
  while ( !oExitFlag ) {

    if ( command_file ) {
      if ( command_fp == NULL ) {
        command_fp = fopen(command_file, "r");
        if ( command_fp == NULL ) {
          free(command_file);
          command_file = NULL;
          continue;
        }
      }

      if ( fgets(cmd, sizeof(cmd), command_fp) == NULL ) {
        free(command_file);
        command_file = NULL;
        continue;
      }
      if ( strlen(cmd) >= 1 ) {
        cmd[strlen(cmd)-1] = 0;
      }
    }
    else {
      if ( interactive ) {
        cout << "sipua> " << flush;     /* prompt for command */
        strcpy(cmd,"");
        fscanf(stdin,"%[^\n]",cmd); /* read command */
        getchar();                      /* remove \n from stdin */
      }
      else {
        /* Do nothing in non-interactive mode. wait for ctrl-C */
        while ( 1 ) {
#ifdef WIN32
          Sleep(10000);
#else
          sleep(10);
#endif
        }
      }
    }

    CommandObject command(cmd, globals); /* Create the command object */

    if ( command == NULL )
       continue;

    if ( command == "exit" || command == "quit" ) {
      oExitFlag = true;
    }
    else if ( command == "help" ) {
      cout << "Commands supported are" << endl << mycommands << flush;
    }
    else if ( command ==  "?") {
      if ( command.GetArgsCount() <= 0 ) {
        fprintf(stderr,"%s\n", mycommands);
        continue;
      }
      if ( command.IsParameter(1,"call") ) {
        if ( call != NULL ) {
          cout << *call << endl;
        }
        else {
          cout << "No call present" << endl;
        }
      }
      else if ( command.IsParameter(1,"ep") ) {
        cout << sip << endl;
      }
      else if ( command.IsParameter(1,"calls")) {
        cout << "Numbers of calls: " << SIPLibrary::CallCount << endl;
      }
      else if ( command.IsParameter(1,"eps")) {
        cout << "Number of endpoints: " << SIPLibrary::EndpointCount
             << endl;
      }
      else {
        fprintf(stderr,"%s\n", mycommands);
      }
    }
    /* ---- LOGging mechanism --- */
    else if ( command == "log" ) {
      if ( command.GetArgsCount() <= 0 ) {
        fprintf(stderr,"%s\n", mycommands);
        continue;
      }
      if ( command.IsParameter(1,"msgflow") ) {
        if ( ((const char *) command[2]) == "off" ) {
          SIPLibrary::MessageFlowOn(NULL);
        }
        else {
          if ( command.GetArgsCount() >= 2 ) {
            if ( command.IsParameter(2, "stdout") ) {
              SIPLibrary::MessageFlowOn(stdout);
            }
            else {
              SIPLibrary::MessageFlowOn(stderr);
            }
          }
          else {
            SIPLibrary::MessageFlowOn(stderr);
          }
        }
      }
      else if ( command.IsParameter(1,"debug") ) {
        if ( command[2] == "off" ) {
          SIPLibrary::DebugOff();
        }
        else {
          if ( command.GetArgsCount() >= 2 ) {
            SIPLibrary::DebugOn((const char *) command[2]);
          }
          else {
            SIPLibrary::DebugOn();
          }
        }
      }
      else if ( command.IsParameter(1, "error") ) {
        fprintf(stderr,"errors log can not be modified\n");
      }
      else {
        fprintf(stderr,"%s\n", mycommands);
      }

    }
    /* ---- SIP Invite ---- */
    else if ( command == "invite") {
      if ( command.GetArgsCount() <= 0) {
        fprintf(stderr,"usage: invite [params...] sipurl_of_remote [request_uri]\n");
        continue;
      }

      if ( oBusy ) {
        cerr << "Already busy in a call" << endl;
        continue;
      }

      oBusy = true;

      /* Formate self URL */
      char *logname = getenv("LOGNAME");
      logname = (logname == NULL) ? getenv("USERNAME") : logname ;
      IptelAddress src = SIPLibrary::HostName;
      src.SetScheme("sip");
      src.SetUser(logname == NULL ? "unknown" : logname);

      if (((const char *) command["from"]) != NULL ) {
        /* Set the from address to this */
        src = (const char *) command["from"];
        src.SetScheme("sip");
        cout << "Local address is " << src << endl;
      }

      /* Get the destination URL */
      IptelAddress dest = (const char *) command[1];
      dest.SetScheme("sip");

      MySIPCall *mycall;
      mycall = new MySIPCall(sip);
      call = mycall;

      if ( command.GetArgsCount() >= 2 ) {
        /* Read the request URI */
        IptelAddress requesturi = (const char *) command[2];
        requesturi.SetScheme("sip");
        mycall->SetRequestURI(requesturi);
      }

      SessionDescription *session;
      MediaInfo media;

      /* Allocate self session info */
      session = new SessionDescription;

      media.SetDirection(MediaInfo::Receive);
      media.SetDataType(MediaInfo::Audio);
      IptelIPAddress rxaddr(SIPLibrary::HostInAddr);
      media.SetRxAddress(rxaddr);
      media.SetRxPort(audio->GetSelfPort());
      FormatInfo f(0);
      media.AddFormat(f); /* PCMU */
      session->AddSession(media);
      cout << "Self Session " << *session << endl;
      mycall->SetSelfSessionDescription(session);

      if ( ((const char *) globals["Password"]) != NULL ) {
        char *u;
        if (((const char *) globals["UserId"]) != NULL ) {
          u = strdup( ((const char *) globals["UserId"]));
        }
        else {
          u = src.GetUser();
        }
        
        mycall->SetAuthorization(u, (const char *) globals["Password"]); 
        /* Set basic authenticate */
        if ( u ) {
          free(u);
        }
      }
      if ( ((const char *) globals["Subject"]) != NULL ) {
        mycall->SetSubject(globals["Subject"]);
      }

      cout << "Calling " << dest << " from " << src  << endl;

      /* Set the outbound proxy address */
      if ( ((const char *) globals["proxy"]) != NULL ) {
        mycall->SetProxyAddress((const char *) globals["proxy"]);
      }

      if ( ((const char *) command["transport"]) != NULL ) {
        const char *transport = (const char *) command["transport"];
        if ( !strcmp(transport, "tcp") ) {
          mycall->SetTransportType(SOCK_STREAM);
        }
        else {
          mycall->SetTransportType(SOCK_DGRAM);
        }
      }

      mycall->Initiate(dest,src);

    }
    else if ( command == "reinvite" ) {
      if ( call ) {
        call->Reinvite();
      }
      else {
        cout << "Not in a call" << endl;
        continue;
      }
    }
    else if ( command == "bye") {

      if ( !oBusy  ) {
        cerr << "Not in a call" << endl;
        continue;
      }

      if ( call != NULL ) {
        audio->StopTool();  
        call->Hangup();
        delete call;
        call = NULL;
        oBusy = false;
      }
      else {
        cerr << "ERROR: call pointer is NULL" << endl;
      }
    }
    else if ( command == "set" || command == "unset") {
      if ( command.GetArgsCount() <= 0 ) {
        fprintf(stderr,"mycommands\n");
        continue;
      }
      if ( command == "set" ) {
        if ( command.GetArgsCount() < 2 ) {
          const char *value = globals[(const char *) command[1]];
          if (value) {
            printf("%s is %s\n", (const char *) command[1], value);
          }
          else {
            printf("%s is not set\n", (const char *) command[1]);
          }
        }
        else {
          globals -= (const char *) command[1];
          globals += GenericHeaderValue((const char *) command[1], command.TillEnd(2));
        }
      }
      else {
        globals -= (const char *) command[1];
      }
    }
    else if (command == "get-ogrqh" || command == "get-ogrsh" 
         ||  command == "get-inrqh" || command == "get-inrsh" ) {
      if ( command.GetArgsCount() <= 0 ) {
        fprintf(stderr, "get-??? header-name\n");
        continue;
      }

      if ( call == NULL ) {
        fprintf(stderr, "No active call\n");
      }
      else {
        const char * v;
        if ( command == "get-ogrqh" ) {
          v = call->GetOutgoingRequestHeader((const char *)command[1]);
        }
        else if ( command == "get-ogrsh" ) {
          v = call->GetOutgoingResponseHeader((const char *)command[1]);
        }
        else if ( command == "get-inrqh" ) {
          v = call->GetIncomingRequestHeader((const char *)command[1]);
        }
        else if ( command == "get-inrsh" ) {
          v = call->GetIncomingResponseHeader((const char *)command[1]);
        }
        printf("%s: %s\n", (const char *) command[1], ((v != NULL)? v : "<absent>"));
      }
    }
    else if ( command == "register" || command == "unregister"
           || command == "register-get" ) {
      if ( command.GetArgsCount() <= 0) {
        fprintf(stderr, 
          "usage: register  [transport=tcp|udp] [proxy=wqr] [content=filename] "
          "[content-type=mimetype] [from=local] [contact=abc@xyz] [expires=n] "
          "url [registrar]\n"
          "       unregister [transport=tcp|udp] [proxy=wqr] [from=local] [contact=*|abc@xyz] "
          "url [registrar]\n"
          "       register-get [transport=tcp|udp] [proxy=wqr] [from=local] url [registrar]\n");
        continue;
      }
      IptelAddress sipurl = (const char *) command[1];

      /* Register with SIP server */
      SIPRegistrationInfo reginfo;
      reginfo.localName = sipurl;
      IptelAddress registrar = sipurl;

      /* Override the registrar address */
      if ( command.GetArgsCount() >= 2) {
          registrar = (const char *) command[2];
      }

      registrar.SetUser(NULL); /* No user name is registrar address */
      registrar.SetScheme("sip");
      reginfo.registrar = registrar;

      if ( !(command == "register-get") ) {
        if ( ((const char *) command["contact"]) == NULL ) {
          reginfo.contactAddress = SIPLibrary::HostName;
          reginfo.contactAddress.SetScheme("sip");
          char *logname = getenv("LOGNAME"); /* Try LOGNAME and USERNAME
                                                env variables */
          logname = (logname == NULL) ? getenv("USERNAME") : logname ;
          if ( logname != NULL) {
            reginfo.contactAddress.SetUser(logname);
          }

          IptelIPAddress listening_addr;
          listening_addr = sip.GetServerAddress();
          int listening_port = listening_addr.GetPort();
          if ( listening_port != 0 && listening_port != 5060) {
            reginfo.contactAddress.SetPort(listening_port);
          }
        }
        else {
            reginfo.contactAddress = (const char *) command["contact"];
        }
      }

      if (((const char *) command["from"]) != NULL ) {
        /* Set the from address to this */
        const char *from = (const char *) command["from"];
        if ( !strcmp(from, "local" ) ) {
          reginfo.fromAddress = SIPLibrary::HostName;
          reginfo.fromAddress.SetScheme("sip");
          char *logname = getenv("LOGNAME"); /* Try LOGNAME and USERNAME
                                                env variables */
          logname = (logname == NULL) ? getenv("USERNAME") : logname ;
          if ( logname != NULL) {
            reginfo.fromAddress.SetUser(logname);
          }
        }
        else {
          reginfo.fromAddress = from;
          reginfo.fromAddress.SetScheme("sip");
        }
      }

      /* Set the outbound proxy */
      if ( globals["proxy"] != NULL ) {
        reginfo.outbound_proxy = (const char *) globals["proxy"];
      }
      if ( ((const char *) command["proxy"]) != NULL ) {
        /* Set the proxy to this address */
        reginfo.outbound_proxy = (const char *) command["proxy"];
      }

      /* Set the expires header */
      if ( ((const char *) command["expires"]) != NULL) {
        reginfo.expiry = (long)(atoi((const char *) command["expires"]));
      }
      else if ( command == "register" ) {
        reginfo.expiry = (long) 3600;
      }

      /* Check if there is any body */
      const char *content = NULL;
      const char *content_type = NULL;
      if ( ((const char *) command["content"]) != NULL ) {
        content = (const char *) command["content"];
        content_type = "text/x-perl";
      }
      if ( ((const char *) command["content-type"]) != NULL ) {
        content_type = (const char *) command["content-type"];
      }

      if ( ((const char *) globals["Password"]) != NULL ) {
        char *u;
        if (((const char *) globals["UserId"]) != NULL ) {
          u = strdup( ((const char *) globals["UserId"]));
        }
        else {
          u = reginfo.fromAddress.GetUser();

          if ( !u ) {
            u = reginfo.localName.GetUser();
          }
        }

        if ( u ) {
          sip.SetAuthorization(u, (const char *) globals["Password"]); 
        }

        if ( u ) {
          free(u);
        }
      }

      if ( ((const char *) command["transport"]) != NULL ) {
        const char *transport = (const char *) command["transport"];
        if ( !strcmp(transport, "tcp") ) {
          reginfo.type = SOCK_STREAM;
        }
        else {
          reginfo.type = SOCK_DGRAM;
        }
      }

      cout << "Registering to server as " << reginfo.localName << " at " <<
        reginfo.registrar << " with contact " <<
        reginfo.contactAddress << endl;
      if ( command == "register" ||
           command == "register-get") {
        if ( content == NULL ) {
          sip.Register(reginfo);
        }
        else {
          sip.Register(reginfo, content, content_type);
        }
      }
      else {
        sip.Unregister(reginfo);
      }
    }
    else if ( command == "options") {
      if ( command.GetArgsCount() <= 0 ) {
        fprintf(stderr, "options url [from]\n");
        continue;
      }
      IptelAddress sipurl = (const char *) command[1];
      IptelAddress from;

      if ( command.GetArgsCount() >= 2) {
          from = (const char *) command[2];
      }
      else {
        from = SIPLibrary::HostName;
        from.SetScheme("sip");
        char *logname = getenv("LOGNAME"); /* Try LOGNAME and USERNAME
                                              env variables */
        logname = (logname == NULL) ? getenv("USERNAME") : logname ;
        if ( logname != NULL) {
          from.SetUser(logname);
        }
      }

      sip.GetRemoteSessionDescription(sipurl, from);
    }
    else if ( command == "authenticate" ) {
      if ( !oBusy ) {
        cerr << "No pending incoming call" << endl;
        continue;
      }
      
      MySIPCall *mycall;
      mycall = call;
      if ( mycall == NULL ) {
        cerr << "ERROR: Call pointer is NULL" << endl;
        oBusy = false;
        continue;
      }
      
      const char *method = "basic";
      const char *file = "./passwd";
      if (((const char *) command["method"]) != NULL ) {
        method = (const char *) command["method"];
      }
      if (((const char *) command["file"]) != NULL ) {
        file = (const char *) command["file"];
      }

      if ( !mycall->Authenticate(method, file) ) {
        cerr << "Can not be authenticated, rejecting" << endl;
        mycall->Reject(); /* error code is already set */
        delete mycall;
        call = NULL; /* XXX free ?? */
        oBusy = false;
        audio->StopTool();
      }
      else {
        cout << "Call is authenticated\n you can accept the call (accept/reject)?" << endl;
      }
    }
    else if (    command == "accept" 
              || command == "reject"
              || command == "noresponse" 
              || command == "redirect") {

        if ( !oBusy ) {
          cerr << "No pending incoming call" << endl;
          continue;
        }
        if ( command == "noresponse" ) {
          /* pretend as if not received any incoming call */
          oBusy = false;
          call = NULL;
          continue;
        }

        MySIPCall *mycall;
        mycall = call;
        if ( mycall == NULL ) {
          cerr << "ERROR: Call pointer is NULL" << endl;
          oBusy = false;
          continue;
        }

        if ( ((const char *) command["content"]) != NULL ) {
          SIPMessageBody body;
          FILE *fp;
          char text[1500]; /* note this limitation XXX */
          strcpy(text, "");

          if ((fp = fopen((const char *) command["content"], "rb")) != NULL) {
            int length = fread(text, 1, sizeof(text) - 1, fp);
            text[length] = 0;
            if ( length > 0 ) {
              body.SetContent(text);
              body.SetContentLength(length);
            }
            else {
              body.SetContent((const char *) command["content"]);
            }
          }
          else {
            /* File is empty use whatever is given on the command line.
            */
            body.SetContent((const char *) command["content"]);
          }

          if ( ((const char *) command["content-type"]) != NULL ) {
            body.SetContentType((const char *) command["content-type"]);
          }
          if ( ((const char *) command["content-disposition"]) != NULL ) {
            body.SetContentDisposition((const char *) command["content-disposition"]);
          }
          if ( ((const char *) command["content-handling"]) != NULL ) {
            body.SetContentHandling((const char *) command["content-handling"]);
          }
          mycall->SetMessageBody(body);
        }

        if ( command == "reject" ) {
            int status = 0;
            const char *status_str = (const char *) command["status"];
            const char *reason_str = NULL;

            if ( status_str && *status_str) {
              status = atoi(status_str);
            }
            if ( command.GetArgsCount() >= 1 ) {
              reason_str = command.TillEnd(1);
            }

            cout << "Call rejecting" << endl;
            if ( status < 400 || status >= 700) {
              /* Do not send back incorrect status */
              if ( reason_str && *reason_str ) {
                status = 603;
              }
              else {
                status = 0;
              }
            }
            if ( status == 0 ) {
              mycall->Reject();
            }
            else {
              mycall->Reject(status, (const char *)(reason_str &&
                                                    *reason_str 
                                                    ? reason_str
                                                    : NULL) );
            }

            delete mycall;
            call = NULL; /* XXX free ?? */
            oBusy = false;
            audio->StopTool();
            continue;
        }
        else if ( command ==  "redirect") {          
          if ( command.GetArgsCount() < 1) {
            fprintf(stderr,"usage: redirect uri\n");
          }
          else {
            cout << "Call redirecting to " << ((const char *) command[1]) << endl;
            const char *reason_str = NULL;

            if ( command.GetArgsCount() >= 2 ) {
              reason_str = command.TillEnd(2);
            }

            mycall->Redirect((const char *) command[1], 
                            reason_str && *reason_str ? 
                            reason_str : "Redirected by user");

            call = NULL;
            oBusy = false;
            audio->StopTool();
          }
          continue;
        }

        IptelAddress dest    = mycall->GetSelfAddress();
        IptelAddress src     = mycall->GetRemoteAddress();

        /* --- get remote session --- */
        SessionDescription *session;
        session = mycall->GetRemoteSessionDescription();
        if ( session != NULL ) {
          cout << "Remote Session is " << *session << endl;
          MediaInfo rem_media;  
          if ( session->FindSession(rem_media, 
            MediaInfo::Receive, MediaInfo::Audio) ) {
  
            IptelIPAddress rem_addr;
            rem_addr = rem_media.GetRxAddress();
            struct sockaddr_in addr;
            rem_addr.GetIpAddress(addr);
            addr.sin_port = htons(rem_media.GetRxPort());
            audio->SetRemoteAddress(addr);
          }
        }
        else {
          cout << "No remote session in INVITE" << endl;
        }

        /* --- Allocate self session info --- */
        session = new SessionDescription;
  
        MediaInfo media;
        media.SetDirection(MediaInfo::Receive);
        media.SetDataType(MediaInfo::Audio);
        /* self IP address */
        media.SetRxAddress(IptelIPAddress(SIPLibrary::HostInAddr));
        media.SetRxPort(audio->GetSelfPort());
        FormatInfo f(0);
        media.AddFormat(f);
        session->AddSession(media);
  
        cout << "Self Session is " << *session << endl;

        mycall->SetSelfSessionDescription(session);

        cout << "Accepting the call..." << endl;
        mycall->Accept();

        audio->StartTool(measure_latency ? 0 : 1);
    }
   
    else if ( command ==  "sleep") {
      if ( command.GetArgsCount() < 1) {
        fprintf(stderr,"usage: sleep n\n");
      }
      else {
        cout << "Waiting for " << ((const char *) command[1]) <<" seconds"<< endl;

        int sec = atoi((const char *) command[1]);
        if ( sec <= 0 ) {
          sec = 0;
        }
#ifdef WIN32
        Sleep(sec * 1000);
#else
        sleep(sec);
#endif
        cout << "Wait existing" << endl;

      }

    }
    
    else {
      fprintf(stderr,"Unknown command: %s\n", cmd);
    }
  }

  cout << "Exiting" << endl;
  //exit(0);
  return 0; /* Not required */
}

/************************** END OF FILE ********************************
*/

