/***************************** FILE HEADER *****************************
#               Copyright (c) 1999 by Columbia University              *
#                        All Rights Reserved                           *
#***********************************************************************
#***********************************************************************
#
#  File Name    : rtspapi.c
#
#  Module       : RTSP client API specification.
#
#  Author(s)    : Tarun Kapoor
#
#  Created      : 8 Feb 2000
#
#  Description  : RTSP Library class implementations
#                
#  Notes        : 
#
#***********************************************************************
#                Modification History (Reverse Order)
#-----------------------------------------------------------------------
#   Date         Author              Description
#-----------------------------------------------------------------------
#              |             | 
#              |             | 
#-----------------------------------------------------------------------
*/

#include <ptlib.h>
#include "rtspapi.h"  /* rtspapi calls are declared here */
#include "rtspparser.h"

#define new PNEW

PString NO_TIME_TYPE("none"); /* used for no range */
PString NPT_TYPE("npt"); 
PString SMPT_TYPE("smpte");
PString CLOCK_TYPE("clock");

static char * RTSP_CONNECT_FAILED = "Rtsp Connect Failed";
static char * TRANSPORT = "Transport";
static char * SERVER_PORT = "server_port";

/**************** Globals *****************************/

/*************** Time Point ***************************/
TimePoint::TimePoint(const PString& time, const PString& type)
{
  point = time;
  pointType = type;
}

TimePoint::TimePoint(const TimePoint& time)
:point(time.point)
{
}

TimePoint::~TimePoint()
{
}

/*************** Time Range ***************************/
TimeRange::TimeRange(const PString& start, const PString& end, const PString& type)
{
  startTime = start;
  endTime   = end;
  rangeType = type;
  startPlay = PString("");
}

TimeRange::TimeRange(const TimeRange& range)
: startTime(range.startTime), endTime(range.endTime), rangeType(range.rangeType)
{
}

TimeRange::~TimeRange()
{
}

/*************** Rtsp Connection ***************************/
RtspConnection::RtspConnection(PURL url,RtspConnection::ModeType mode)
{
  uri = url;
  if ( url.GetScheme() == "rtsp" ) {
    sock = new PTCPSocket;
  }
  else {
    sock = new PUDPSocket;
  }

  WORD port;
  /* if not port specified, use the default port of 554 */
  if ( (port = url.GetPort()) == 0 ) {
    port = 554; 
  }

  if ( url.GetScheme() == "rtsp") {
    sock->SetPort(port);
  }
  else {
    cerr << "Error: Probably the userid is not defined." << endl;
  }

  cseq = 1;
  clientPorts[0] = 0;
  clientPorts[1] = 0;

  this->mode = mode;
  destAddr = PString("");
}

RtspConnection::~RtspConnection()
{
  if ( sock != NULL )  {   
    if (!sock->Close())
      cerr << "Socket shutdown unsuccessful" << endl;
    delete sock;
  }
}

void RtspConnection::SetDestinationAddress(PString addr)
{
  destAddr = addr;
}

void RtspConnection::SetClientPort(unsigned short port)
{
  clientPorts[0] = port;
  clientPorts[1] = port + 1;
}

/* This method will be used by a client to determine where to send the RTP */
/* packets for a record mode                                               */
const PString RtspConnection::GetServerAddress()
{
  PIPSocket::Address addr;
  if ( sock != NULL && sock->GetPeerAddress(addr) ) 
  {
    return PString(addr.AsString());
  }
  else
  {
    return PString("");
    /* IP address of peer Not found  or socket not connected */
  }
}

const unsigned short RtspConnection::GetServerPort()
{
  PString serverPort = 
    parser.GetParameter(PString(TRANSPORT),PString(SERVER_PORT));
   
  unsigned short port = (unsigned short)serverPort.AsInteger();
  return port;
}

int RtspConnection::Describe(char *sdp) /* DESCRIBE */
{
  /* first check to make sure socket is not already connected, before trying to */
  /* connect */
  if (!sock->IsOpen())
  {
    /* Connect the socket to remote */
    if (!sock->Connect(uri.GetHostName()))
    {
      this->SetLastError(PString(RTSP_CONNECT_FAILED));
      return -1;
    }
  }  

  RtspMessage msg;

  msg += "DESCRIBE " + uri.AsString() + " RTSP/1.0\r\n";
  msg += "CSeq: ";
  char sequence[10];
  sprintf(sequence,"%d",cseq);
  msg += sequence;
  msg += "\r\n";
  msg += "Accept: application/sdp\r\n";
  msg += "\r\n";

  sock->Write((const unsigned char *)msg,
              msg.GetLength());

  if (sock->GetReadTimeout() > 1000)
    sock->SetReadTimeout(1000); /* Max 1 sec timout */

  PString line;
  parser.SetNewResponse(); /* indicate to the parser that a new response is 
			    * about to come.
                            */

  /* read until no more lines in message */
  while ( (line = this->ReadLine()) != PString(""))
  {
    parser.ParseResponse(line);
  }
  char secondPart[1500];
  sock->Read(secondPart,1500); /* second part of response */
  strncpy(sdp,secondPart,sock->GetLastReadCount());
  
  if (this->SuccessfulRtspMessage())
  {
    cseq++;
    return 0;
  }
  else
  {
    return -1;
  } 
}

int RtspConnection::Setup(char * mode,const char *body) /* SETUP */
{
  /* first check to make sure socket is not already connected, before trying to */
  /* connect */
  if (!sock->IsOpen())
  {
    /* Connect the socket to remote */
    if (!sock->Connect(uri.GetHostName()))
    {
      this->SetLastError(PString(RTSP_CONNECT_FAILED));
      return -1;
    } 
  }

  RtspMessage msg;
  char sequence[10];
  char clientPort1[10];
  char clientPort2[10];

  sprintf(clientPort1,"%d",clientPorts[0]);
  sprintf(clientPort2,"%d",clientPorts[1]);

  msg += "SETUP " + uri.AsString() + " RTSP/1.0\r\n";
  msg += "CSeq: ";

  sprintf(sequence,"%d",cseq);  
  msg += sequence;
  msg += "\r\n";
  msg += "Transport: RTP/AVP;unicast";
  if (destAddr != PString(""))
  {
    msg+= ";destination=" + destAddr;
  }
  msg += ";client_port="; 
  msg += clientPort1; 
  msg += "-";
  msg += clientPort2;  
  msg += ";mode=";
  msg += mode;
  msg += "\r\n";

  if ( body != NULL ) {
    msg += "Content-Type: text/plain";
    msg += "\r\n";
    msg += "Content-Length: ";
    int length = strlen(body);
    char bodyLength[10];
    sprintf(bodyLength,"%d",length);
    msg += bodyLength;
    msg += "\r\n";
  }

  msg += "\r\n";
  if ( body != NULL ) {
    msg += body;
  }

  sock->Write((const unsigned char *)msg,
              msg.GetLength());

  if (sock->GetReadTimeout() > 1000)
    sock->SetReadTimeout(1000); /* Max 1 sec timout */

  PString line;
  parser.SetNewResponse(); /* indicate to the parser that a new response is 
			    * about to come.
                            */

  /* read until no more lines in message */
  while ( (line = this->ReadLine()) != PString("")) 
  {
    parser.ParseResponse(PString(line));  
  }

  if (this->SuccessfulRtspMessage())
  {
    cseq++;
    return 0;
  }
  else
  {
    return -1;
  }
}

int RtspConnection::Teardown() /* TEARDOWN */
{
  /* We can safely assume that when calling teardown a socket connection */
  /* has already been established, not requiring us to create a socket   */
  /* connection just to call the TearDown message                        */

  if (!sock->IsOpen())
  {
    /* Connect the socket to remote */
    if (!sock->Connect(uri.GetHostName()))
    {
      this->SetLastError(PString(RTSP_CONNECT_FAILED));
      return -1;
    } 
  }

  RtspMessage msg;

  msg += "TEARDOWN " + uri.AsString() + " RTSP/1.0\r\n";
  msg += "CSeq: ";
  char sequence[10];
  sprintf(sequence,"%d", cseq);
  msg += sequence;
  msg += "\r\n";
  msg += "Session:" + GetSessionId() + "\r\n";
  msg += "\r\n";

  sock->Write((const unsigned char *)msg,
              msg.GetLength());

  PString line;
  parser.SetNewResponse(); /* indicate to the parser that a new response is
                            * about to come.                               
			    */

  /* read until no more lines in message */
  while ( (line = this->ReadLine()) != PString("")) 
  {
    parser.ParseResponse(line);
  }

  if (this->SuccessfulRtspMessage())
  {
    cseq++;
    return 0;
  }
  else
  {
    return -1;
  } 
}

int RtspConnection::Play(TimeRange & range) /* PLAY */
{
  /* A play message assumes that the socket has already been connected to a */
  /* Rtsp uri, since a setup message has to be sent first, which connects   */
  /* the socket                                                             */

  if (!sock->IsOpen())
  {
    /* Connect the socket to remote */
    if (!sock->Connect(uri.GetHostName()))
    {
      this->SetLastError(PString(RTSP_CONNECT_FAILED));
      return -1;
    } 
  }

  RtspMessage msg;

  bool hasRange;
  if ( range.GetRangeType() == NO_TIME_TYPE )
    hasRange = false;
  else
    hasRange = true; 

  msg += "PLAY " + uri.AsString() + " RTSP/1.0\r\n";
  msg += "CSeq: ";
  char sequence[10];
  sprintf(sequence,"%d",cseq);
  msg += sequence;
  msg += "\r\n";
  msg += "Session:" + GetSessionId() + "\r\n";
  if (hasRange)
  {
    msg += "Range: " + range.GetRangeType() + "="; 
    msg += range.GetStartTime() + "-" + range.GetEndTime();
    if ( range.GetStartPlay() != PString("") )
    {
      msg += ";time=" + range.GetStartPlay();
    }
    msg += "\r\n";
  }
  msg += "\r\n";

  sock->Write((const unsigned char *)msg,
              msg.GetLength());

  PString line;
  parser.SetNewResponse(); /* indicate to the parser that a new response is 
			    * about to come.                                
			    */

  /* read until no more lines in message */
  while ( (line = this->ReadLine()) != PString("")) 
  {
    parser.ParseResponse(line);
  }

  if (this->SuccessfulRtspMessage())
  {
    cseq++;

    PString rangeheader("Range");
    PString rangevalue = parser.GetHeaderValue(rangeheader);
    
    if ( rangevalue != "" ) {
      PStringArray splits = rangevalue.Tokenise("=");
      PString type("");
      PString begin(""), end("");
      if (splits.GetSize() > 0 ) {
        type = splits[0];
        if ( splits.GetSize() > 1 ) {
          PStringArray splits2 = splits[1].Tokenise("-");
          if ( splits2.GetSize() > 0 ) {
            begin = splits2[0];
            if ( splits2.GetSize() > 1 ) {
              end = splits2[1];
            }
          }
        }
        
        range = TimeRange(begin, end, type);
      }
    }

    return 0;
  }
  else
  {
    return -1;
  } 
}

int RtspConnection::Record(TimeRange range) /* RECORD */
{
  /* A record message assumes that the socket has already been connected to a */
  /* Rtsp uri, since an announce and setup message has to be sent first,      */ 
  /* which connects the socket                                                */

  if (!sock->IsOpen())
  {
    /* Connect the socket to remote */
    if (!sock->Connect(uri.GetHostName()))
    {
      this->SetLastError(PString(RTSP_CONNECT_FAILED));
      return -1;
    } 
  }


  RtspMessage msg;

  bool hasRange;
  if ( range.GetRangeType() == NO_TIME_TYPE )
    hasRange = false;
  else
    hasRange = true; 

  msg += "RECORD " + uri.AsString() + " RTSP/1.0\r\n";
  msg += "CSeq: ";
  char sequence[10];
  sprintf(sequence,"%d",cseq);
  msg += sequence;
  msg += "\r\n";
  msg += "Session:" + GetSessionId() + "\r\n";
  if (hasRange)
  {
    msg += "Range: " + range.GetRangeType() + "="; 
    msg += range.GetStartTime() + "-" + range.GetEndTime() + "\r\n";
  }
  msg += "\r\n";

  sock->Write((const unsigned char *)msg,
              msg.GetLength());

  PString line;
  parser.SetNewResponse(); /* indicate to the parser that a new response is 
                            * about to come.                                
			    */

  /* read until no more lines in message */
  while ( (line = this->ReadLine()) != PString("")) 
  {
    parser.ParseResponse(line);
  }

  if (this->SuccessfulRtspMessage())
  {
    cseq++;
    return 0;
  }
  else
  {
    return -1;
  } 
}

int RtspConnection::Record(TimeRange range, long maxsize) /* RECORD */
{
  if (!sock->IsOpen())
  {
    /* Connect the socket to remote */
    if (!sock->Connect(uri.GetHostName()))
    {
      this->SetLastError(PString(RTSP_CONNECT_FAILED));
      return -1;
    } 
  }

  RtspMessage msg;

  bool hasRange;
  if ( range.GetRangeType() == NO_TIME_TYPE )
    hasRange = false;
  else
    hasRange = true; 

  msg += "RECORD " + uri.AsString() + " RTSP/1.0\r\n";
  msg += "CSeq: ";
  char sequence[10];
  sprintf(sequence,"%d",cseq);
  msg += sequence;
  msg += "\r\n";
  msg += "Session:" + GetSessionId() + "\r\n";
  if (hasRange)
  {
    msg += "Range: " + range.GetRangeType() + "="; 
    msg += range.GetStartTime() + "-" + range.GetEndTime() + "\r\n";
  }

  /* This is the only line different from previous method */
  msg += "MaxSize: ";
  msg += PString(PString::Signed, maxsize);
  msg += "\r\n";

  msg += "\r\n";

  sock->Write((const unsigned char *)msg,
              msg.GetLength());

  PString line;
  parser.SetNewResponse(); /* indicate to the parser that a new response is 
                            * about to come.                                
			    */

  /* read until no more lines in message */
  while ( (line = this->ReadLine()) != PString("")) 
  {
    parser.ParseResponse(line);
  }

  if (this->SuccessfulRtspMessage())
  {
    cseq++;
    return 0;
  }
  else
  {
    return -1;
  } 
}


int RtspConnection::Pause(TimePoint time) /* PAUSE */
{
  if (!sock->IsOpen())
  {
    /* Connect the socket to remote */
    if (!sock->Connect(uri.GetHostName()))
    {
      this->SetLastError(PString(RTSP_CONNECT_FAILED));
      return -1;
    } 
  }

  RtspMessage msg;

  bool hasPoint;
  if ( time.GetPointType() == NO_TIME_TYPE )
    hasPoint = false;
  else
    hasPoint = true; 

  msg += "PAUSE " + uri.AsString() + " RTSP/1.0\r\n";
  msg += "CSeq: ";
  char sequence[10];
  sprintf(sequence,"%d",cseq);
  msg += sequence;
  msg += "\r\n";
  msg += "Session:" + GetSessionId() + "\r\n";
  if (hasPoint)
  {
    msg += "Range: " + time.GetPointType() + "=" + time.GetTimePoint() + "\r\n"; 
  }
  msg += "\r\n";

  sock->Write((const unsigned char *)msg,
              msg.GetLength());

  PString line;
  parser.SetNewResponse(); /* indicate to the parser that a new response is 
                            * about to come.                                
			    */
  
  /* read until no more lines in message */
  while ( (line = this->ReadLine()) != PString("")) 
  {
    parser.ParseResponse(line);
  }

  if (this->SuccessfulRtspMessage())
  {
    cseq++;
    return 0;
  }
  else
  {
    return -1;
  } 
}

int RtspConnection::Announce(PString sdp) /* ANNOUNCE */
{
  /* first check to make sure socket is not already connected, before trying to */
  /* connect */
  if (!sock->IsOpen())
  {
    /* Connect the socket to remote */
    if (!sock->Connect(uri.GetHostName()))
    {
      this->SetLastError(PString(RTSP_CONNECT_FAILED));
      return -1;
    } 
  }

  RtspMessage msg;

  msg += "ANNOUNCE " + uri.AsString() + " RTSP/1.0\r\n";
  msg += "CSeq: ";
  char sequence[10];
  sprintf(sequence,"%d",cseq);
  msg += sequence;
  msg += "\r\n";
  PString sessionId = GetSessionId();
  if (sessionId != PString(""))
    msg += "Session:" + sessionId + "\r\n";
  msg += "Content-Type: application/sdp";
  msg += "\r\n"; 
  msg += "Content-Length: ";
  int length = strlen(sdp);
  char sdpLength[10];
  sprintf(sdpLength,"%d",length);
  msg += sdpLength;
  msg += "\r\n";
  msg += "\r\n";
  msg += sdp;
  msg += "\r\n";

  sock->Write((const unsigned char *)msg,
              msg.GetLength());

  if (sock->GetReadTimeout() > 1000)
    sock->SetReadTimeout(1000); /* Max 1 sec timout */

  PString line;
  parser.SetNewResponse(); /* indicate to the parser that a new response is
			    * about to come.                                
			    */
  
  /* read until no more lines in message */
  while ( (line = this->ReadLine()) != PString("")) 
  {
    parser.ParseResponse(line);
  }

  if (this->SuccessfulRtspMessage())
  {
    cseq++;
    return 0;
  }
  else
  {
    return -1;
  } 
}

int RtspConnection::Seek(const TimePoint & time)
{
  return -1;
}

int RtspConnection::Read(char *buf,int size)
{
  return -1;
}

int RtspConnection::Write(char *buf, int size)
{
  return -1;
}

void RtspConnection::SetLastError(PString error)
{
  err = error;
}

PString RtspConnection::GetLastError()
{
  return err;
}


bool RtspConnection::SuccessfulRtspMessage()
{
  int responseCode = parser.GetLastResponseCode();
  if (responseCode >= 300)
  {
    this->SetLastError(PString(responseCode));
    return false;
  }
  return true;
}

PString RtspConnection::ReadLine()
{

  PString line; /* initialize an empty PString */
  char ch;
  int byte;
  byte = sock->ReadChar();
  if (byte != -1)
  {
    ch = (char)byte;
    while (ch != '\n')
    {
      line += ch;
      byte = sock->ReadChar();
      if (byte == -1)
        return PString("");
      ch = (char)byte;
    }
    if (IsResponseLineEmpty(line))
      return PString("");
    else
      return line;
  }
  else
  {
    return PString("");
  }
} 

bool RtspConnection::IsResponseLineEmpty(PString line)
{
  PINDEX i = 0;

  while (i < line.GetLength())
  {
    /* the string must contain any character which is not a space or '\n' or 
     * '\r' to be considered worth storing in the parser.
     */
    if ( (line[i] != ' ') && (line[i] != '\n') && (line[i] != '\r') ) {
      return false;
    }
    i++;
  }

  return true;
  
}

/*************** Rtsp Message ***************************/
RtspMessage::RtspMessage(const PString & str)
:PString(str)
{
}


///////////////////// End of File ////////////////////////////////////////


/************************** END OF FILE ********************************
*/

