/***************************** FILE HEADER *****************************
#               Copyright (c) 2000 by Columbia University              *
#                        All Rights Reserved                           *
#***********************************************************************
#***********************************************************************
#
#  File Name    : rtspapi.h
#
#  Module       : RTSP Library.
#
#  Author(s)    : Tarun Kapoor
#
#  Created      : 8 Feb 2000
#
#  Description  : Contains classes used for the rtsp library api's.
#                
#  Notes        : 
#
#***********************************************************************
#                Modification History (Reverse Order)
#-----------------------------------------------------------------------
#   Date         Author              Description
#-----------------------------------------------------------------------
#              |				| 
#-----------------------------------------------------------------------
*/

#ifndef _RTSP_API_H_
#define _RTSP_API_H_

/******************* INCLUDES ******************************************
 ***********************************************************************
 */
#include <ptlib.h>  /* PString */
#include <ptlib/sockets.h>
#include <ptclib/url.h>
#include <ptclib/http.h> /* PHTTP */
#include "rtspparser.h"
/******************* CLASSES *******************************************
 ***********************************************************************
 */

extern PString NO_TIME_TYPE; 
extern PString NPT_TYPE; 
extern PString SMPT_TYPE;
extern PString CLOCK_TYPE;

class TimePoint : public PObject
{
  PCLASSINFO(TimePoint,PObject)

  public:
    TimePoint(const PString& time, const PString& type);
    TimePoint(const TimePoint& time);
    virtual ~TimePoint();
    const PString& GetTimePoint() { return point; }
    const PString& GetPointType() { return pointType; }

  private:
    PString point;
    PString pointType;
};

class TimeRange : public PObject
{
  PCLASSINFO(TimeRange,PObject)

  public:
    /* constructor from all range types*/
    TimeRange(const PString& start, const PString& end, const PString& type);
    TimeRange(const TimeRange& range);
    virtual ~TimeRange();

    const PString& GetStartTime() { return startTime; }
    const PString& GetEndTime()   { return endTime; }
    const PString& GetRangeType() { return rangeType; }
    const PString& toString();

    const PString& GetStartPlay() { return startPlay; } 
    /* not called in constructor, must be explicitly set if desired in the */
    /* necessary RTSP messages                                             */
    void SetStartPlay(const PString& start) { startPlay = start; } 

  private:
    PString startTime;
    PString endTime;
    PString rangeType;
    PString startPlay;
};

/**
 * RtspConnection: This class represesnts a RtspConnection.
 * One such object is associated with each URI/URL that
 * is used.
 */
class RtspConnection : public PObject
{
  PCLASSINFO(RtspConnection, PObject)

  public :

    enum ModeType {
      RTSP_RDONLY = 0x0001,
      RTSP_WRONLY = 0x0002
    };

    RtspConnection(PURL url, ModeType mode);
    virtual ~RtspConnection();

    int Setup(char * mode,const char *body = NULL);  /* SETUP */
    int Play(TimeRange & range);                /* PLAY */
    int Record(TimeRange range);                /* RECORD */
    int Record(TimeRange range, long maxsize);  /* RECORD for voice
                                                   mail */
    int Pause(TimePoint time);                  /* PAUSE */
    int Describe(char *sdp);                    /* DESCRIBE */
    int Announce(PString sdp);                  /* ANNOUNCE */
    int Teardown();                             /* TEARDOWN */

    int Seek(const TimePoint & time);
    int Read(char *buf,int size);
    int Write(char *buf, int size);

    /* these will be used for playing */
    void SetDestinationAddress(PString addr);
    void SetClientPort(unsigned short port);

    /* these will be used for recording */
    const PString GetServerAddress();
    const unsigned short GetServerPort();

    PString GetLastError();

  private :
    PIPSocket       *sock;    /* socket used for communication */
    PURL            uri;      /* remote URI */
    PString         err;
    RtspParser      parser;   /* used for parsing RTSP response messages */
                              /* from the server.                        */
    void SetLastError(PString error);
    bool SuccessfulRtspMessage();
    
    /* return each line of an RTSP response message.  If the line is an empty
     * line (using IsResponseLineEmpty()), then return empty string.  If there 
     * are no more lines to read from the socket, then also return empty string
     */
    PString ReadLine();
 
    /* check to make sure a response line is not empty (e.g., the DESCRIBE 
     * response returns an empty line before the start of the SDP message 
     */
    bool IsResponseLineEmpty(PString line); 

    /* return the current session id from the parser. */
    const PString GetSessionId()
    { return parser.GetHeaderValue(PString("Session")); }

    int             cseq;           /* cseq number */
    unsigned short  clientPorts[2]; /* client port numbers */
    int             mode;           /* mode type - Read only (play) or write 
                                     * only (record) 
                                     */
    PString         destAddr;       /* a client can redirect the stream to a
                                     * a different destination address    
                                     */
};

/* 
 * RtspMessage: This represesnt a RTSP message.  It is a PString specific only 
 * to a RTSP message. 
 */
class RtspMessage : public PString
{
  PCLASSINFO(RtspMessage, PString)
  public:
    RtspMessage()  { };
    RtspMessage(const PString & str);
    ~RtspMessage() { };
};

#endif /* _RTSP_API_H_ */

/************************** END OF FILE ********************************
*/
