/***************************** FILE HEADER *****************************
#               Copyright (c) 1999 by Columbia University              *
#                        All Rights Reserved                           *
#***********************************************************************
#***********************************************************************
#
#  File Name    : main.cpp
#
#  Module       : Main file for Voice Mail Server.
#
#  Author(s)    : Kundan Singh (kns10@cs.columbia.edu)
#
#  Created      : 15 Nov 1999
#
#  Description  : 
#                
#  Notes        : 
#
#***********************************************************************
#                Modification History (Reverse Order)
#-----------------------------------------------------------------------
#   Date         Author              Description
#-----------------------------------------------------------------------
#              |             | 
#              |             | 
#-----------------------------------------------------------------------
*/

/* Common includes */
#include <iostream.h>
#include <stdio.h>

/* Portable library includes */
#include <ptlib.h>
#include <ptclib/inetmail.h>
#include <ptlib/sockets.h>
#include <ptclib/url.h>

/* SIP API */
#include "iptelapi.h"
#include "sipapi.h"
#include "rtspapi.h"
#include "vmail.h"

#include "dbapi.h"

extern "C" {
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/resource.h>

#ifdef HAVE_ACL
#include "sys/acl.h"
#else
#warning "HAVE_ACL is not defined"
#endif
}

/* extern "C" const char *server_name = "Columbia-Voice-Mail-System";*/

/* 
 * Response mode for URI parameter in recording 
 */
static char * BUSY = "onbusy";
static char * NO_RESPONSE = "noresponse";

/* Local globals */
static bool oLicensed = false;
static bool oEmailNotification = true;
static PString notify_tcl;

static unsigned short configport = 10022;

/*
 * Global trace option.
 */
bool sipum_debug_flag = true;
#define TRACE(x) if (sipum_debug_flag) { cout << x << endl; }
 
/**
 * SIPUserAgent and SIPIncomingCall
 */

class SIPUserAgent : public SIPEndpoint
{
public:
  SIPUserAgent();
  virtual IptelCall *AllocateNewCallObject(const char *callid);
};

class RtspThread; // forward declaration.

class SIPIncomingCall : public SIPCall
{
public:
  SIPIncomingCall(const char *callid=NULL) : SIPCall(callid) 
  { rtsp= NULL; };
  SIPIncomingCall(SIPEndpoint & ep, const char * callid=NULL) 
    : SIPCall(ep, callid) { rtsp= NULL;};
  ~SIPIncomingCall();
  virtual void   OnNewIncomingCall(IptelEndpoint & ep);
  virtual int    OnHangup();
  RtspThread     *rtsp; /* RTSP object associated with this call */
};

/*
 * Thread to handle RTSP requests, this
 * is needed because SIP library does not use PThread
 * while RtspConnection->Sockets need PThread::Current()
 */
class RtspThread : public PThread
{
public:
  RtspThread(SIPIncomingCall & call);
  void Main();
  //void OnHangup();

  /* 
   * IsWaiting() returns TRUE if SIP call is not hungup 
   */
  BOOL IsWaiting();
  void SetWaiting(BOOL wait);

  /* RTSP connections-- mutex */
  PSyncPoint flag;

private:
  /* Associated SIP call */
  SIPIncomingCall & call;

  /* Remote and self SIP addresses */
  IptelAddress remote;
  IptelAddress self;
  IptelAddress requesturi;

  /* RTSP connection objects */
  RtspConnection  *rtsp_play;
  RtspConnection  *rtsp_record;

  /* waiting == TRUE means SIP call is not hung up */
  BOOL            waiting;

  /* Message id of the message */
  int msgid;

  /* Various user params */
  PString         userlogin;
  PString         msgurl;
  PString         httpurl;
  PString         callurl;
  PString         subject;
  PString         priority;
  PString         organization;

  /* Mode in which the vmail has received the call */
  enum { RecordMail, RetrieveMail, DeleteMail } mode;

  /* Mode for recording */
  enum { Busy, NoResponse, Default } recordMode;
};

//////////////// Method defn for SIP ///////////////////////

SIPUserAgent::SIPUserAgent()
  : SIPEndpoint(true,false)
{
}

IptelCall *SIPUserAgent::AllocateNewCallObject(const char * callid)
{
  return new SIPIncomingCall(*this, callid);
}

SIPIncomingCall::~SIPIncomingCall()
{
  if ( rtsp != NULL ) {
    /* Should terminate the thread here XXX */
    // rtsp->oTerminate = true;
    //Don;t delete autoDeleteThread ; delete rtsp;
  }
}

void SIPIncomingCall::OnNewIncomingCall(IptelEndpoint & /* ep ignored */)
{
  rtsp = new RtspThread(*this);
  rtsp->Resume();
}

RtspThread::RtspThread(SIPIncomingCall & c)
  : PThread(10000),call(c),userlogin(""),msgurl(""),httpurl(""),callurl(""),
    subject(""),priority(""),organization("")
{
  char *p;
  rtsp_play   = NULL;
  rtsp_record = NULL;
  mode       = RtspThread::RecordMail;
  recordMode = RtspThread::Default;
  remote     = call.GetRemoteAddress();


  /* XXX self should be To: header of Request URI */
  self       = call.GetSelfAddress();
  requesturi = call.GetRequestURI();
  TRACE("request uri is " << requesturi);

  if (( p = (char *) call.GetHeader("Subject")) != NULL ) {
    subject = p;
  }
  if (( p = (char *) call.GetHeader("Priority")) != NULL ) {
    priority =  p;
  }
  if (( p = (char *) call.GetHeader("Organization")) != NULL ) {
    organization = p;
  }
}

void RtspThread::Main()
{
  int e = 0; /* Indicates error */

  /* XXX lock mutex */
  BOOL mailrecorded = FALSE;

  TRACE("SIP: Incoming Call");

  /* Extract the user name from the SIP url */
  IptelAddress to = requesturi;

  PString command = "";
  PStringArray splits;

  /* Extract the user portion from requestURI */
  char *c = to.GetUser();
  if ( c == NULL ) {
    userlogin = VoiceMailInfo::GetAdministratorLogin();
  }
  else {
    /* Assuming the user portion is of the form
       <user>-*-<command> where command is 
       either retrieve, delete or noresponse, busy.
       or the user portion is simply <user>.
    */
    userlogin = PString(c);

    PString un = userlogin;
    splits = un.Tokenise("-",FALSE);
    if ( splits.GetSize() > 0 ) {
      userlogin = splits[0];
      if ( splits.GetSize() > 1 ) { 
        command = splits[splits.GetSize() - 1];
      }
    }
    else {
      userlogin = "";
      /* Command is "" */
    }
  }

  TRACE("Command is " << command);

  if ( userlogin == "" ) {
    userlogin = VoiceMailInfo::GetAdministratorLogin();  
    /* Send it to default operator */
    /* XXX we should reject the call. Just for testing purpose
       we are forwarding it to the default user */
  }

  VoiceMailInfo vmailinfo(userlogin);

  /* ------ retrieve the voice mail ------ */

  if ( command == "retrieve" ) {

#if 0
    if ( VoiceMailInfo::IsAuthenticationPresent() &&
         !call.Authenticate("basic",
                            VoiceMailInfo::GetAuthenticationFile())) {
      /* Change this to validate the password from the
       * password file. This is because the
       * request uri is kns10-7128182-retrieve@ and not kns10@...
       * XXX
       */
      TRACE("Can not be authenticated, rejecting");
      call.Reject(); /* error code is already set */
      return;
    }
#endif

    /* Get the message id. Msg id if present is second element */
    msgid = atoi((const char *)splits[1]);

    /* Retrive the voice mail for the user */
    mode = RtspThread::RetrieveMail;

    TRACE("SIP: retrieving voice mail");

    SessionDescription *session;
    if ( (session = call.GetRemoteSessionDescription()) == NULL ) {
      TRACE("SIP: Session Description absent, rejecting ...");
      call.Reject(400, "No session description");
      return;
    }


    /* 
     * determine if accept header contains text/html.  if it does AND the SIP
     * URL did not contain a voicemail number, then send the web page of the 
     * cgi script to view all the voicemails, rather than, connecting to the 
     * voicemail server via rtsp and rtp
     */
    bool html = false;
    /* XXX need to test this logic */
    if (msgid == 0) {
      char * media = (char *)(call.GetHeader("Accept"));
      if (accept != NULL) {
	PString mediaTypes = PString(media);
	PStringArray types = mediaTypes.Tokenise(",",FALSE);
	for (unsigned int i = 0; ( (i < types.GetSize()) || (!html) ); i++) {
	  if (types[i] == "text/html") {
            html = true;
	  }
	}
      }
    }
    
    /* Retrive an HTML page only if the
       Accept:text/html is present and 
       no message id number is specified */
    if ((html) && (msgid == 0)) {
      /* create the html content body for the response */
      PString contentBody;
      contentBody += "<html>";
      contentBody += "<body>";
      contentBody += "click ";
      contentBody += "<a href=\"";
      contentBody += vmailinfo.GetHTTPHomeURL();
      contentBody += "\">here</a>";
      contentBody += " to view your inbox";
      contentBody += "</body>";
      contentBody += "</html>";

      call.SetContentInfo("text/html",(const char*)contentBody);
    }  
    else { 
      /* 
       * rtsp connection and rtp used as normal to retrieve the specified
       * voicemail.
       */

      msgurl  = vmailinfo.GetRTSPMessageURL(msgid);
      
      rtsp_play = new RtspConnection(msgurl, 
                                     RtspConnection::RTSP_RDONLY);
      
      struct sockaddr_in rtp_sin;
      MediaInfo media;
      if ( session->FindSession(media, MediaInfo::Receive,
                                MediaInfo::Audio) ) {
        rtp_sin = (sockaddr_in) media.GetRxAddress();
      }

      rtsp_play->SetDestinationAddress(
                PString(inet_ntoa(rtp_sin.sin_addr))); /* XXX not reentrant*/
      rtsp_play->SetClientPort(ntohs(rtp_sin.sin_port));
      
      e = rtsp_play->Setup("play");
      
      if ( e == 0 ) {
        /* Allocate self session info */
        session = new SessionDescription;
        MediaInfo mediainfo;
        mediainfo.SetDirection(MediaInfo::Receive);
        mediainfo.SetDataType(MediaInfo::Audio);
        mediainfo.SetRxAddress("0.0.0.0");
        mediainfo.SetRxPort(0); /* XXX port number is 0 to put media on hold */
        FormatInfo f(0);
        mediainfo.AddFormat(f); /* XXX get from rtspd instead of PCMU */
        session->AddSession(mediainfo);
        
        /* Set the session description and Accept the SIP call */
        call.SetSelfSessionDescription(session);
        call.Accept();
        
        TRACE("Playing " << *session);
        
        TimeRange range(PString(""),PString(""),NO_TIME_TYPE);
        rtsp_play->Play(range);
      }
      else {
        call.Reject(500, "RTSP SETUP for play, failed");
      }
    }
  } /* end of retrieve */

  /* ------- DELETE the voice mail ------- */

  else if ( command == "delete") {
    msgid = atoi((const char *)splits[1]);

    /* self URL is of the form <user>-<msgid>-delete@<server> */
    /* Delete the voice mail for the user */
    mode = RtspThread::DeleteMail;

    TRACE( "SIP: deleting voice mail" );

    SessionDescription *session;
    if ( (session = call.GetRemoteSessionDescription()) == NULL ) {
      TRACE( "SIP: Session Description absent, rejecting ..." );
      call.Reject(400, "No session description");
      return;
    }

    msgurl  = vmailinfo.GetRTSPMessageURL(msgid);

    rtsp_record   = new RtspConnection(msgurl, 
                                       RtspConnection::RTSP_WRONLY);

    struct sockaddr_in rtp_sin;
    MediaInfo media;
    if ( session->FindSession(media, MediaInfo::Receive,
                              MediaInfo::Audio) ) {
      rtp_sin = (sockaddr_in) media.GetRxAddress();
    }
    
    rtsp_record->SetDestinationAddress(
               PString(inet_ntoa(rtp_sin.sin_addr))); /* XXX not reentrant*/
    rtsp_record->SetClientPort(ntohs(rtp_sin.sin_port));

    e = rtsp_record->Setup("record");

    if ( e == 0 ) {
      /* Allocate self session info */
      session = new SessionDescription;
      MediaInfo mediainfo;
      mediainfo.SetDirection(MediaInfo::Receive);
      mediainfo.SetDataType(MediaInfo::Audio);
      mediainfo.SetRxAddress("0.0.0.0");
      mediainfo.SetRxPort(0); /* XXX port number is 0 to put media on hold */
      FormatInfo f(0);
      mediainfo.AddFormat(f); /* PCMU */
      session->AddSession(mediainfo);
      
      /* Set the session description and Accept the SIP call */
      call.SetSelfSessionDescription(session);
      call.Accept();
      
      TRACE( "Deleting mail " << *session );
    }
    else {
      call.Reject(500, "RTSP SETUP for record, failed");
    }

    /* note that message RECORD is not sent */
  } /* end of delete */

  /* ----- Recording the voice mail ----- */

  else {
    /*
     * parse Request-URI for busy, noresponse, or welcome message to play
     * depending on mode parameter in Request-URI.  If there is no mode
     * parameter, then the traditional wait for a specified time out
     * and then record message is processed.  Otherwise, immediately 
     * process message
     */

    char * recordType = (char *)(call.GetParameter("Request-URI","mode"));

    waiting = TRUE;

    bool oExists = false;
    if (recordType != NULL) {  
      /* and it is of type noresponse or onbusy then
       * set the record mode appropriately. */
      
      if ( (strcmp(recordType,BUSY) == 0) ) {
        recordMode = RtspThread::Busy;
        TRACE( "SIP: recording for busy message" );
        oExists = true;
      }
      else if ( (strcmp(recordType,NO_RESPONSE) == 0) ) {
        recordMode = RtspThread::NoResponse;
        TRACE( "SIP: recording for no response message" );
        oExists = true;
      }
    }
    if ( !oExists ) {
      /* Check if uri is of form <user>-onbusy@server */
      if ( command == "onbusy" ) {
        recordMode = RtspThread::Busy;
        TRACE( "SIP: recording for busy message, req-URI" );
        oExists = true;
      }
      /* check if uri is of form <user>-noresponse@server */
      else if ( command == "noresponse" ) {
        recordMode = RtspThread::NoResponse;
        TRACE( "SIP: recording for no response message, req-URI");
        oExists = true;
      }
    }
    
    if ( !oExists ) {
      int VMAIL_WAIT_TIMEOUT = vmailinfo.GetWaitTimeout();
      
      TRACE( "SIP: waiting for " << VMAIL_WAIT_TIMEOUT
           << " seconds..." );

      /* Call PThread->Sleep(millisec); */
      Sleep(VMAIL_WAIT_TIMEOUT * 1000);

      TRACE( "SIP: wait completed" );
      /*
       * If the call was hungup then waiting was set to false.
       */
      /* XXX use mutex to set and test waiting variable */
    }

    if ( waiting ) {
      
      TRACE( "SIP: recording..." );

      msgid = vmailinfo.GetNewMessageID();

      /* Call URL is of the form sip:<user>-<msgid>-retrieve@<server> */
      callurl = vmailinfo.GetSIPMessageURI(msgid);

      /* HTTP URL is of the form http:// */
      httpurl = vmailinfo.GetHTTPMessageURL(msgid);

      /*
       * PString("http://www.cs.columbia.edu/~") +
       * userlogin + PString("/vmail/vmail.cgi");
       */

      /* msg URL is of the form rtsp://<server>/<user>/path.au */
      msgurl  = vmailinfo.GetRTSPMessageURL(msgid);

      PString welcomeMess;
      if (recordMode == RtspThread::Busy) {
	welcomeMess = vmailinfo.GetWelcomeURL(BUSY);
      }
      else if (recordMode == RtspThread::NoResponse) {
	welcomeMess =
          vmailinfo.GetWelcomeURL(NO_RESPONSE);
      }
      else {
        /* Default is typically no response */
	welcomeMess = vmailinfo.GetWelcomeURL();
      }

      SessionDescription *session;
      TRACE( "Remote Session" );
      if ( (session = call.GetRemoteSessionDescription()) == NULL ) {
        TRACE( "Session Description absent, rejecting ..." );
        call.Reject(400, "No session description");
        return;
      }

      struct sockaddr_in rtp_sin;
      MediaInfo media;
      if ( session->FindSession(media, MediaInfo::Receive,
                                MediaInfo::Audio) ) {
        rtp_sin = (sockaddr_in) media.GetRxAddress();
      }
      
      TRACE( "New Rtsp connection play: " << welcomeMess );
      rtsp_play   = new RtspConnection(welcomeMess,
                                       RtspConnection::RTSP_RDONLY);
      TRACE( "New Rtsp connection record: " << msgurl);
      rtsp_record = new RtspConnection(msgurl,
                                     RtspConnection::RTSP_WRONLY);

      rtsp_play->SetDestinationAddress(PString(inet_ntoa(rtp_sin.sin_addr)));
      rtsp_play->SetClientPort(ntohs(rtp_sin.sin_port));
      rtsp_record->SetClientPort(ntohs(rtp_sin.sin_port));

      time_t t = time(NULL);
      PString date = (const char *)ctime(&t);
      date = date.Left(date.GetLength()-1); /* remove \n */

      /* Add Priority, subject and organization in the subject
       * or body of RTSP SETUP for recording. */

      /* XXX chenge this to have \n insted of | */
      PString recordinfo = 
        PString("From:") + remote.GetAddressString() +
        PString("\nTo:")  + self.GetAddressString() +
        PString("\nStatus:N") +  /* New mail */
        PString("\nDate:") + date;

      if ( priority != "" ) {
        recordinfo += PString("\nPriority:") + priority;
      }
      if ( subject != "" ) {
        recordinfo += PString("\nSubject:") + subject;
      }
      if ( organization != "" ) {
        recordinfo += PString("\nOrganization:") + organization;
      }
      recordinfo += "\n";

      TRACE( "VMAIL: record info in RTSP setup\n" << recordinfo );

      e = rtsp_play->Setup("play");
      TRACE( "VMAIL: Setup-play complete:" << e);
      
      e = rtsp_record->Setup("record", recordinfo);
      TRACE( "VMAIL: Setup-record complete:" << e);

      if ( e == 0 ) {
        /* Allocate self session info */
        session = new SessionDescription;
        int selfPortNumber;
        selfPortNumber = rtsp_record->GetServerPort();
        MediaInfo mediainfo;
        mediainfo.SetDirection(MediaInfo::Receive);
        mediainfo.SetDataType(MediaInfo::Audio);
        PString serverAddr = vmailinfo.GetRTSPServerIP(msgurl); 
        //rtsp_record->GetServerAddress(); XXX
        mediainfo.SetRxAddress((const char *) serverAddr);
        mediainfo.SetRxPort(selfPortNumber);
        FormatInfo f(0);
        mediainfo.AddFormat(f); /* get it from RTSP describe insted of
                                   PCMU, XXX */
        session->AddSession(mediainfo);
        
        TRACE( "Recording Session " << *session );

        /* Set the session description and Accept the SIP call */
        call.SetSelfSessionDescription(session);
        call.Accept();

        TimeRange range(PString(""),PString(""),NO_TIME_TYPE);
        
        rtsp_play->Play(range);

        PString duration_str = range.GetEndTime();
        double duration_sec = duration_str.AsReal();
        if ( duration_sec > 0.0 ) {
          long duration_ms = (long) (duration_sec * 1000.0); /* milliseconds */
          PTimeInterval interval(duration_ms);
          TRACE( "SIP: sleeping for " << duration_ms << " millisec");
          
          PThread::Sleep(interval);
          TRACE( "SIP: sleep completed" );
        }

        /* Wait for Play NPT time. XXX 
           or invoke delayed Record here. */
        
        if ( waiting ) {
          long sizelimit = vmailinfo.GetPerMsgSizeLimit();
          
          rtsp_record->Record(TimeRange(PString(""),PString(""),
                                        NO_TIME_TYPE),
                              sizelimit );
          mailrecorded = TRUE;
        }
        else {
          TRACE( "SIP: call is closed while msg was played" );
        }
      }
      else {
        call.Reject(500, "RTSP SETUP for record, failed");
      }

      waiting = FALSE;

    } /* is active */
    else {
      TRACE( "SIP: call is already closed" );
    }
  }

  /* XXX Unlock mutex */

  TRACE( "SIP: waiting for call close mutex" );
  flag.Wait();
  TRACE( "SIP: wait completed" );

  /* XXX lock mutex */
  
  TRACE( "SIP: Call Hung up" );

  if ( mode == RtspThread::RecordMail ) {

    if ( rtsp_play != NULL ) {
      TRACE( "VMAIL: Deleting RTSP play" );
      rtsp_play->Teardown();
      delete rtsp_play;
      rtsp_play = NULL;
    }

    if ( rtsp_record != NULL ) {
      TRACE( "VMAIL: Deleting RTSP record" );
      rtsp_record->Teardown();
      delete rtsp_record;
      rtsp_record = NULL;
    }

    if ( mailrecorded ) {

#if 0 // #ifdef HAVE_ACL // No ACL is used.
      /* 
       * Set ACL permissions to administrator. This is supported 
       * only on solaris.
       */
      if ( userlogin == VoiceMailInfo::GetAdministratorLogin() ) {
        /* Mail is for administrator, no need to set ACL */
      }
      else {
        int count;
        aclent_t *acls;

        /* File name is /home/userlogin/vmail/inbox/msgid.au */
        PString filename = vmailinfo.GetFileName(msgid);

        /*
          PString("/home/") + userlogin +
          PString("/vmail/inbox/") +
          PString(PString::Unsigned, msgid) + 
          PString(".au"); 
        */

        if ((count = acl(filename, GETACLCNT, 0,NULL)) < 0) {
          perror("acl GETALENT");
        }
        else if ((acls = (aclent_t *)malloc((count+1) *
                                            sizeof(aclent_t)))
                 == NULL ) {
          TRACE( "Memory allocation for acl failed" );
        }
        else if ((count = acl(filename, GETACL, count, acls)) < 0) {
          perror("acl GETACL");
        }
        else {
          /* Set the ACL for administrator also */
          acls[count].a_type = USER;
          acls[count].a_id   = VoiceMailInfo::GetUserId(
                               VoiceMailInfo::GetAdministratorLogin());
          acls[count].a_perm = 6; /* rw- */
          if ( acl(filename, SETACL, count+1, acls) != 0) {
            perror("acl SETACL");
          }
        } 
      }
#endif /* HAVE_ACL */

      /* Send a notification to the user */
      PSMTPClient mailclient;

      PString remoteemail = vmailinfo.GetEmailAddress();
      PStringArray splits = remoteemail.Tokenise("@");

      TRACE( "VMAIL: Sending email to " << remoteemail );
    
      /* Connect to server */
      if ( !oEmailNotification || mailclient.Connect(splits[1]) ) {

        TRACE( "VMAIL: SMTP client connected" );
        char *selfuri = NULL;
        selfuri = remote.GetURI();
        IptelAddress senderemail = selfuri;
        if ( selfuri ) {
          free(selfuri);
        }

        senderemail.SetScheme(NULL); /* No SIP scheme */

        TRACE( "VMAIL: Self Email is " << senderemail.GetAddressString());

        if ( !oEmailNotification ||
             mailclient.BeginMessage((const char *)
                                     (senderemail.GetAddressString()),
                                     splits[0]) ||
             mailclient.BeginMessage((const char *)
                                     ("voicemail"),
                                     splits[0]) ) {

          TRACE( "VMAIL: Mail from: " << senderemail.GetAddressString() );
          TRACE( "VMAIL: Rcpt to: " << remoteemail );
    
          /*
           *  -- Email format is as follows --
           *  To: <SIP-URL of callee without scheme >
           *  From: <SIP-URL of caller without scheme >
           *  X-SendProblemsTo: <email of vmail server administrators>
           *  
           *  You have got a new voice mail from
           *  <clickable SIP-URL of caller>
           *  You can access the voice mail from any of the following URLs
           *  (using quicktime, or other RTSP client)
           *  <clickable rtsp message URL>
           *  (using ephone, or other SIP user agent)
           *  <clickable sip url to retrieve the mail>
           *  Web page clickable address for retrieving the mails.
           *
           *  Have a nice day
           *  CINEMA administrator
           */
          PString command(notify_tcl);
          command += PString(" user=") + userlogin;
          selfuri = self.GetURI();
          if ( selfuri ) {
            command += PString(" \'to=")   + selfuri + PString("\'");
            free(selfuri);
          }
          char * remoteuri = remote.GetURI();
          command += PString(" \'from=") + remoteuri + PString("\'");
          if ( remoteuri ) {
            free(remoteuri);
          }

          if ( priority != "" ) {
            command += PString(" \'priority=") + priority + PString("\'");
          }
          command += PString(" \'rtspurl=") + msgurl + PString("\'");
          command += PString(" \'httpurl=") + httpurl + PString("\'");
          command += PString(" \'sipurl=")  + callurl + PString("\'");

          command += PString(" subject=\'") + subject + PString("\'");
          if ( organization != "" ) {
            command += PString(" organization=\'") + organization +
              PString("\'");
          }

          FILE *mailcmd;
          PString message;
          
          TRACE( "VMAIL: command is " << command );

          if ((mailcmd = popen(command, "r")) == NULL ) {
            perror("popen");
            TRACE( "VMAIL: Can not execute " << command );

            /* Command execution failed, make default */
            message += "To: ";
            message += remoteemail;
            message +="\n";
            message += "Subject: voice mail--" + subject + "\n";
            if ( priority != "") {
              message += "Priority: " + priority + "\n";
            }
            if ( organization != "") {
              message += "Organization: " + organization + "\n";
            }
            message += "X-SendProblemsTo: kns10@cs.columbia.edu,hgs@cs.columbia.edu\n";
            message += "You have got a new voice mail from \n";
            message +=  remote.GetAddressString();
            message += ". <a href=\"data:application/sip,";
            message +=  remote.GetAddressString();
            message += "\">(Reply)</a>\n";
            
            message += "You can access the voice mail from any of the following URLs\n";
            message += "\n(using quicktime, or other RTSP client)\n";
            message += msgurl;
            message += ". <a href=\"data:application/rtsp,";
            message += msgurl;
            message += "\">(Click here)</a>\n";
            
            message += "\n(using ephone, or other SIP user agent)\n";
            message += callurl;
            message += "\n";
            message += ". <a href=\"data:application/sip,";
            message +=  callurl;
            message += "\">(Click here)</a>\n";
            
            message += "\n(using netscape, or other HTTP browser)\n";
            message += httpurl;
            message += "\n";
            message += "\nHave a nice day\nCINEMA administrator\n";
          }
          else {
            /* Read from the mailcommand */
            char line[200];
            while ((fgets(line, sizeof(line), mailcmd)) > 0 ) {
              message += line;
            }
            pclose(mailcmd);
          }

          TRACE( "--- email --- \n" << message );
          if ( oEmailNotification &&
               mailclient.Write((const char *) message, message.GetLength() )) {
            mailclient.EndMessage();
            TRACE( "--- email sent --- " );
          }
        }
        else {
          TRACE( "BeginMessage failed" );
        }
      }
    }

  } /* mode is RecordMail */
  else if ( mode == RtspThread::RetrieveMail ) {
    /* Terminate retrieval session */
    if ( rtsp_play != NULL ) {
      rtsp_play->Teardown();
      delete rtsp_play;
      rtsp_play = NULL;
    }
  }
  else if ( mode == RtspThread::DeleteMail ) {
    /* Terminate retrieval session */
    if ( rtsp_record != NULL ) {
      rtsp_record->Teardown();
      delete rtsp_record;
      rtsp_record = NULL;
    }
  }
  }

BOOL RtspThread::IsWaiting()
{
  return waiting;
}

void RtspThread::SetWaiting(BOOL wait)
{
  waiting = wait;
}

int   SIPIncomingCall::OnHangup()
{
  if ( rtsp ) {
    rtsp->SetWaiting(FALSE);
    rtsp->flag.Signal();
    rtsp = NULL;
  }

  SIPCall::OnHangup();
  return 0; /* XXX this causes a memory leak,
               shoule return 0 to free the object */
}

////////////////// Main routine /////////////////////

class TestApp : public PProcess
{
  PCLASSINFO(TestApp, PProcess)
  public:
    TestApp();
    ~TestApp();

    void Main();
};

PCREATE_PROCESS(TestApp);

TestApp::TestApp()
  : PProcess("Columbia University", "SIPUM",
             1, 0, AlphaCode, 1)
{
}


TestApp::~TestApp()
{
}

void TestApp::Main()
{
  struct rlimit rbuf; /* rlimit */
  PArgList & args = GetArguments();
  args.Parse("h-help."
             "v-verbose."
             "X-non-daemon."
             "r-no-registration."
             "d-default-domain:"
             "s-sip-server-address:"
             "l-license-file:"
             "e-no-email-notification."
             "n-notify-tcl-script:"
             "c-config-port:"
             "k-numeric-address."
             "a-passwd-file:"
             "f-local-database-file:"
             "D-mysql-server-host:"
             "U-mysql-server-userid:"
             "P-mysql-server-password:"
             "S-mysql-server-database:"
          , FALSE);

  cout << GetName() 
       << " Version " << GetVersion(TRUE)
       << " by " << GetManufacturer()
       << " on " << GetOSClass() << ' ' << GetOSName()
       << " (" << GetOSVersion() << '-' << GetOSHardware() << ')' << endl;

  if ( args.HasOption('k') ) {
    /* This should be done before any instance of SIPUserAgent is
     * created. */
    SIPLibrary::oUseNumericAddress = true;
  }


  /* Create the listening servers */
  SIPUserAgent sip;

  SIPLibrary::SetName("SIPUM-Columbia-University-Voice-Mail-System");

  if ( args.HasOption('h') ) {
    cout << "Usage : " << GetFile().GetFileName() << " -h\n"
    "      : " << GetFile().GetFileName() << 
      " [options]\n"
      "v        : Verbose mode\n"
      "X        : non daemon mode. (Currently this is default)\n"
      "r        : Do not register the users\n"
      "e        : Do not send email notification\n"
      "u user   : Administrator user login.\n"
      "           Default value is the unix login id of the\n"
      "           effective user id of the application.\n"
      "d domain : A different default domain.\n" 
      "           This overrides the entry in config file.\n"
      "s host   : Host name of SIP server.\n"
      "           This overrides the entry in config file.\n"
      "           Other values are read from config file\n"
      "l file   : Specify the license file if licensing is on.\n"
      "           Default is \"sipum.license\" in current directory.\n"
      "n file   : Specify the notify.tcl script file path.\n"
      "           Default is ./script/notify.tcl.\n"
      "c configport: Specify the TCP port to be used for configuration.\n"
      "           Default is " << configport << ".\n"
      "k        : Use numeric IP address (dotted-decimal) instead of DNS.\n"
      "f local-database: Name of local gdbm database file.\n"
      "D mysql-host : MySQL server host.\n"
      "U mysql-user : MySQL user id. Default is root.\n"
      "P mysql-password: MySQL user password.\n"
      "S mysql-database: MySQL database name. Default is sip.\n"
         << endl;

    exit(0);
  }

  /* Check for validity of license */
  int e = 
  SIPLibrary::CheckLicense(args.GetOptionString('l', "sipum.license"));

  if ( e > 0 ) {
    cerr << "Your registration file is invalid or has expired.\n"
      "You should obtain a new registration by "
      "sending email to hgs@cs.columbia.edu.\n" << endl;
    exit(0);
  }
  else if ( e < 0 ) {
    cerr << "Unlicensed copy. Contact hgs@cs.columbia.edu for a license." << endl;
    exit(0);
  }
  else if ( e == 0 ) {
    oLicensed = true;
  }

  /* Check verbose mode */
  if ( args.HasOption('v') ) {
    SIPLibrary::DebugOn();
    sipum_debug_flag = true;
  }
  else {
    sipum_debug_flag = false;
  }

  /* Update default user login */
  char *un;
  if ( !args.HasOption('u') &&  (un = getenv("LOGNAME"))!= NULL) {
    TRACE( "user name is " << un );
    VoiceMailInfo::SetAdministratorLogin(un);
  }
  else {
    VoiceMailInfo::SetAdministratorLogin(args.GetOptionString('u',"default"));
  }

  /* Get the config port */
  if ( args.HasOption('c') ) {
    PString portstr = args.GetOptionString('c',"");
    configport = atoi(portstr);
  }
  TRACE( "TCP port for configuration is " << configport );


  notify_tcl = args.GetOptionString('n',"./script/notify.tcl");

  /* Update other options */
  if ( args.HasOption('d') ) {
    VoiceMailInfo::SetDefaultDomain(
                   args.GetOptionString('d',"cs.columbia.edu"));
  }
  if ( args.HasOption('s') ) {
    VoiceMailInfo::SetDefaultServer(
                   args.GetOptionString('s',"ind.cs.columbia.edu"));
  }
  if ( args.HasOption('e') ) {
    oEmailNotification = false;
  }

  if ( args.HasOption('a') ) {
    VoiceMailInfo::SetAuthenticationFile(
                  args.GetOptionString('a',""));
  }

  if ( !args.HasOption('D') && !args.HasOption('f')) {
    cerr << "Error: either -f or -D option is mandatory." << endl;
    exit(1);
  }

  if ( args.HasOption('D') ) {
#ifdef USE_MYSQL
    VoiceMailInfo::use_local_db = false;
    /* Connect to the MySQL database */
    PString host = args.GetOptionString('D', "muni");
    PString user = args.GetOptionString('U', "root");
    PString password = args.GetOptionString('P', "tiger");
    PString database = args.GetOptionString('S', "sip");

    if ( password != "NULL") {
      initializeConnection((char *)(const char *)user, 
                         (char *)(const char *)password, 
                         (char *)(const char *)database, 
                         (char *)(const char *)host,
                         0);
    }
    else {
      initializeConnection((char *)(const char *)user, 
                         NULL, 
                         (char *)(const char *)database, 
                         (char *)(const char *)host,
                         0);
    }
#else
    cerr << "Error: not compiled with mysql..." << endl;
    exit(1);
#endif
  }
  else if ( args.HasOption('f')) {
    VoiceMailInfo::use_local_db = true;
    /* Update the config file name */
    PString filename = args.GetOptionString('f',"vmail.conf");
    VoiceMailInfo::SetConfigFile(filename);
  }


  /* Set the open file limit. Works for solaris */
  memset(&rbuf, 0, sizeof(rbuf));
  if ( getrlimit(RLIMIT_NOFILE, &rbuf) == 0 ) {
    rbuf.rlim_cur = rbuf.rlim_max;
    if ( setrlimit(RLIMIT_NOFILE, &rbuf) != 0 ) {
      perror("setrlimit RLIMIT_NOFILE");
    }
  }
  
  /* Set the core file rlimit. Works for solaris */
  memset(&rbuf, 0, sizeof(rbuf));
  if ( getrlimit(RLIMIT_CORE, &rbuf) == 0 ) {
    rbuf.rlim_cur = rbuf.rlim_max;
    if ( rbuf.rlim_cur > 10000000L ) {
      rbuf.rlim_cur = 10000000L;
    }
    if ( setrlimit(RLIMIT_CORE, &rbuf) != 0 ) {
      perror("setrlimit RLIMIT_CORE");
    }
  }

  /* Register the users with appropriate servers */
  if ( !args.HasOption('r') ) {
    TRACE( "Registering..." );

    VoiceMailInfo::RegisterAllUsers(sip);

    TRACE( "VMAIL: Registered" );
  } /* option n */
    
  TRACE( "Done..." );

  int sock;
  int opt;
  if ( (sock = socket(AF_INET, SOCK_STREAM, 0)) < 0 ) {
    perror("socket");
    goto failurelabel;
  }

  /* Set "reuseaddr" (and "reuseport", if it exists) so that we don't
   * have to wait for TIME_WAIT to expire if we crash the server while
   * connections are open. */
  opt = 1;
  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *)&opt, sizeof(opt)) < 0) {
    perror("setsockopt(SO_REUSEADDR)");
    goto failurelabel;
  }

#ifdef SO_REUSEPORT
  opt = 1;
  if (setsockopt(sock, SOL_SOCKET, SO_REUSEPORT, (char *)&opt, sizeof(opt)) < 0) {
    perror("setsockopt(SO_REUSEPORT)");
    goto failurelabel;
  }
#endif

  struct sockaddr_in sin;
  sin.sin_family = AF_INET;
  sin.sin_addr.s_addr = INADDR_ANY;
  sin.sin_port = htons(configport);

  if (( bind(sock, (const sockaddr *)&sin, sizeof(sin)) != 0 )) {
    perror("bind");
    goto failurelabel;
  }
  
  if (listen(sock, 1000) < 0) {
    perror("listen");
    goto failurelabel;
  }

  while ( 1 ) {
    int newsock;
    int len = sizeof(sin);
    newsock = accept(sock, (sockaddr *) &sin, &len);
    if ( newsock < 0 ) {
      perror("accept");
    }
    else {
      /*
       * We should unregister previous registrations,
       * especially those which are removed XXX 
       */
      TRACE( "VMAIL: Restarted" );
      if ( !args.HasOption('r') ) {
        TRACE( "VMAIL: Registering..." );
        VoiceMailInfo::RegisterAllUsers(sip);
        TRACE( "VMAIL: Registered" );
      }
      else {
        TRACE( "VMAIL: Not registering since -r option" );
      }
    }
#ifdef WIN32
    closesocket(newsock);
#else
    close(newsock);
#endif
  }

  exit(0);

 failurelabel:
  while ( 1 ) {
    /* Wait till the process is terminated :-) */
    WaitForTermination();//sleep(10000)
  }

  /* Should never reach here */

  exit (0);
}


void VoiceMailInfo::RegisterAllUsers(SIPEndpoint & sip) 
{
  PStringArray users = VoiceMailInfo::GetUsers();
  
  for ( int i=0; i< users.GetSize(); i++) {
    PString user = users[i];
    
    VoiceMailInfo vmailinfo(user);
    RegistrationInfoList regs = 
      vmailinfo.GetRegistrationInfo();
    for ( int j =0; j<regs.GetSize(); j++) {
      
      RegistrationInfoObject reginfo = regs[j];
      SIPRegistrationInfo iptelreg;
      iptelreg.localName = reginfo.localName;
      iptelreg.registrar = reginfo.registrar;
      
      iptelreg.localName.SetScheme("sip");
      iptelreg.registrar.SetScheme("sip");
      
      /* Set the contact address */
      iptelreg.contactAddress = SIPLibrary::HostName;
      IptelAddress useraddress((const char *)user);
      char *userid = useraddress.GetUser();
      iptelreg.contactAddress.SetUser((const char *) userid);
      if ( userid ) {
        free(userid);
      }
      iptelreg.contactAddress.SetScheme("sip");
      
      TRACE( "VMAIL: user: " << iptelreg.localName << "; registrar: " << iptelreg.registrar << "; contact: " << iptelreg.contactAddress );

      iptelreg.expiry = 3600;

      sip.Register(iptelreg);
    }
  }
}
/************************** END OF FILE ********************************
*/
