/*
* Search in alias file for any matching aliases.
* Lines beginning with white space are treated as continuation lines;
* lines beginning with # are comments. Aliases may recurse.
* Prints aliases to stdout, one-per-line.
*
* Usage:
*   aliases [-a alias_file]
*
* Also looks through YP mail.aliases to get aliases, but silently skips
* it if it doesn't exist in the default domain.
*/

#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "copyright.h"

/* YP */
#include <rpcsvc/ypclnt.h>

typedef struct list_t {
  struct list_t *next, *prev;  /* next and previous entry */
  char *word;                  /* dynamically allocated */
} list_t;


/*
* Add word to end of list.
*/
static list_t *list_add(list_t *list, char *word)
{
  list_t *p, *new;

  new = calloc(sizeof(list_t), 1);
  new->word = word;
  if (!list) {
    list = new;
  }
  else {
    for (p = list; p->next; p = p->next) ;
    p->next = new;
    new->prev = p;
  }
  return list;
} /* list_add */


/*
* Replace list element 'what' by list 'with'. Free word.
*/
static list_t *list_replace(list_t *list, list_t *what, list_t *with)
{
  if (with && list && what) {
    with->prev = what->prev;
    with->next = what->next;
    if (what->prev) what->prev->next = with;
    else list = with;
    if (what->next) what->next->prev = with;
    free(what->word);
    free(what);
  }
  return list;
} /* list_replace */


/*
* Output list to stdout.
*/
static void list_print(list_t *list)
{
  list_t *p;

  for (p = list; p; p = p->next) {
    printf("%s\n", p->word);
  }
} /* list_print */


/*
* Read alias file containing one name per line. Return space-separated
* list, dynamically allocated.
*/
static list_t *alias_file(char *filename)
{
  FILE *f;  
  char *word;   /* start of word */
  char *s;
  list_t *list = 0;
  char line[1024];

  if (!(f = fopen(filename, "r"))) {
    perror(filename);
    return NULL;
  }
  while (fgets(line, sizeof(line), f)) {
    /* Strip leading and trailing white space. */
    for (word = line; *word && isspace((int)*word); word++) ;
    for (s = word; *s && !isspace((int)*s); s++) ;
    *s = '\0';
    if (word[0] != '\0' && word[0] != '#' && word[0] != '\n') {
      list = list_add(list, strdup(word));
    }
  }
  fclose(f);
  return list;
} /* alias_file */


/*
* Try to find name in NIS map. If found, return all matches (comma
* separated).  The value is dynamically allocated and has to be 
* freed by the caller.
*/
static char *yp_find(char *name)
{
  char *domain;
  char *value;
  int valuelen;
  int r;

  if ((r = yp_get_default_domain(&domain))) {
    return 0;
  }
  if ((r = yp_match(domain, "mail.aliases", name, strlen(name) + 1,
    &value, &valuelen))) {
    return 0;
  }
  return value;
} /* yp_find */


/*
* Return matching line, if any, from the aliasfile.
*/
static char *file_find(char *aliasfile, char *name)
{
  FILE *f;
  char line[1024];
  char *e = 0;
  int found = 0;  /* we found a match */

  if (!(f = fopen(aliasfile, "r"))) {
    perror(aliasfile);
    return NULL;
  }
  while (fgets(line, sizeof(line), f)) {
    if (!line[0] || line[0] == '#' || line[0] == '\n') {
      if (found) break;
      else continue;
    }
    /* end of continuation lines */
    else if (found && !isspace((int)line[0])) {
      break;
    }
    else if (!strncasecmp(line, name, strlen(name))) {
      char *s = &line[strlen(name)]; /* character after first word */

      if (isspace((int)*s)) {
        for ( ; *s && *s != ':'; s++) ;
      }
      if (*s == ':') {
        found = 1;
        s++;
      }
      if (found) {
        if (e) {
          e = realloc(e, strlen(e) + strlen(s) + 1);
          e = strcat(e, s);
        }
        else e = strdup(s);
      }
    }
  }
  fclose(f);
  return e;
} /* file_find */


/*
* Return list of expansions.
*/
static list_t *expand(char *aliasfile, char *name)
{
  char *e;       /* dynamically allocated string */
  char *s, *d;
  list_t *list = 0;
  char *bow, *eow;  /* beginning and one char beyond end of word */ 
  char *name_stripped;

  /* Strip name */
  if (!name || !name[0]) return 0;

  /*
   * Remove all but numbers, digits and @, . from name to allow handling of
   * phone numbers such as +49-30-254-99-200.
   */
  name_stripped = malloc(strlen(name) + 1);
  for (s = name, d = name_stripped; *s; s++) {
    if ((!isspace((int)*s) && !ispunct((int)*s)) || strchr("@.-", *s)) *d++ = *s;
  }
  *d = '\0';
  name = name_stripped;

  /* Try YP first. */
  if (!(e = yp_find(name))) {
    if (!(e = file_find(aliasfile, name))) {
      return 0;
    }
  }

  /* Check if alias is of form :include:file list */
  for (s = e; *s && isspace((int)*s); s++) ;
  if (!strncmp(s, ":include:", strlen(":include:"))) {
    for (d = s; *d && !isspace((int)*d); d++) ;
    *d = '\0';  /* remove trailing white space */
    list = alias_file(s + strlen(":include:"));
  }
  /* Convert comma-separated entries into list. */
  else {
    bow = eow = 0;
    for (; *s; s++) {
      if (*s == ',') *s = ' ';
      if (!bow) {
        if (!isspace((int)*s)) bow = s;
      }
      if (bow) {
        if (isspace((int)*s)) eow = s;
        else if (*(s+1) == '\0') eow = s+1;
      }
      if (eow) {
        *eow = '\0';
        list = list_add(list, strdup(bow));
        bow = eow = 0;
      }
    }
  }
  free(e);
  free(name);
  return list;
} /* expand */


/*
* Search in file 'aliasfile' for name 'name'.
*/
list_t *aliases(char *aliasfile, char *name)
{
  list_t *list = 0;
  list_t *p;
  list_t *r;

  list = expand(aliasfile, name);

  p = list;
  while (p) {
    list = list_replace(list, p, (r = expand(aliasfile, p->word)));
    /* If not replaced, move pointer. */
    p = r ? r : p->next;
  }

  return list;
} /* aliases */


static void usage(char *pname)
{
  fprintf(stderr, "Usage: %s [-a aliasfile] name\n", pname);
  exit(1);
} /* usage */


int main(int argc, char *argv[])
{
  extern char *optarg;
  extern int optind;
  char *aliasfile = "/etc/aliases";
  int c;

  while ((c = getopt(argc, argv, "a:")) != EOF) {
    switch (c) {
    case 'a':
      aliasfile = optarg;
      break;
    case '?':
      usage(argv[0]);
      break;
    }
  }
  /* Not enough arguments. */
  if (optind >= argc) {
    usage(argv[0]);
  }

  /* Search for aliases in system alias file. */
  list_print(aliases(aliasfile, argv[optind]));
  return 0;
} /* main */
