/*
* Search in alias file for any matching aliases.
* Lines beginning with white space are treated as continuation lines;
* lines beginning with # are comments. Aliases may recurse.
* Return an appropriate canonicalize_result.
*
* Also looks through YP mail.aliases to get aliases, but silently skips
* it if it doesn't exist in the default domain.
*/

#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include "sysdep.h"
#include "strdupn.h"

#include "canon_core.h"

/* YP */
#include <rpcsvc/ypclnt.h>

typedef struct list_t {
  struct list_t *next, *prev;  /* next and previous entry */
  char *word;                  /* dynamically allocated */
} list_t;


/*
* Add word to end of list.
*/
static list_t *list_add(list_t *list, char *word)
{
  list_t *p, *new;

  new = calloc(sizeof(list_t), 1);
  new->word = word;
  if (!list) {
    list = new;
  }
  else {
    for (p = list; p->next; p = p->next) ;
    p->next = new;
    new->prev = p;
  }
  return list;
} /* list_add */


/*
* Replace list element 'what' by list 'with'. Free word.
*/
static list_t *list_replace(list_t *list, list_t *what, list_t *with)
{
  if (with && list && what) {
    with->prev = what->prev;
    with->next = what->next;
    if (what->prev) what->prev->next = with;
    else list = with;
    if (what->next) what->next->prev = with;
    free(what->word);
    free(what);
  }
  return list;
} /* list_replace */


/*
 * Free an entire list.
 */
static void list_free(list_t *list)
{
  list_t *list_next;
  
  for (; list != NULL; list = list_next) {
    list_next = list->next;
    free(list->word);
    free(list);
  }
}


/*
* Read alias file containing one name per line. Return space-separated
* list, dynamically allocated.
*/
static list_t *alias_file(char *filename, char **errorstr)
{
  FILE *f;  
  char *word;   /* start of word */
  char *s;
  list_t *list = 0;
  char line[1024];

  if (!(f = fopen(filename, "r"))) {
    seterror(errorstr, "Alias file %s: %s", filename, strerror(errno));
    return NULL;
  }
  while (fgets(line, sizeof(line), f)) {
    /* Strip leading and trailing white space. */
    for (word = line; *word && isspace((int)*word); word++) ;
    for (s = word; *s && !isspace((int)*s); s++) ;
    *s = '\0';
    if (word[0] != '\0' && word[0] != '#' && word[0] != '\n') {
      list = list_add(list, strdup(word));
    }
  }
  fclose(f);
  return list;
} /* alias_file */


/*
* Try to find name in NIS map. If found, return all matches (comma
* separated).  The value is dynamically allocated and has to be 
* freed by the caller.
*/
static char *yp_find(char *name)
{
  char *domain;
  char *value;
  int valuelen;
  int r;

  if ((r = yp_get_default_domain(&domain))) {
    return 0;
  }
  if ((r = yp_match(domain, "mail.aliases", name, strlen(name) + 1,
    &value, &valuelen))) {
    return 0;
  }
  return value;
} /* yp_find */


/*
* Return matching line, if any, from the aliasfile.
*/
static char *file_find(char *aliasfile, char *name, char **errorstr)
{
  FILE *f;
  char line[1024];
  char *e = 0;
  int found = 0;  /* we found a match */

  if (!(f = fopen(aliasfile, "r"))) {
    seterror(errorstr, "Alias file %s: %s", aliasfile, strerror(errno));
    return NULL;
  }
  while (fgets(line, sizeof(line), f)) {
    if (!line[0] || line[0] == '#' || line[0] == '\n') {
      if (found) break;
      else continue;
    }
    /* end of continuation lines */
    else if (found && !isspace((int)line[0])) {
      break;
    }
    else if (!strncasecmp(line, name, strlen(name))) {
      char *s = &line[strlen(name)]; /* character after first word */

      if (isspace((int)*s)) {
        for ( ; *s && *s != ':'; s++) ;
      }
      if (*s == ':') {
        found = 1;
        s++;
      }
      if (found) {
        if (e) {
          e = realloc(e, strlen(e) + strlen(s) + 1);
          e = strcat(e, s);
        }
        else e = strdup(s);
      }
    }
  }
  fclose(f);
  return e;
} /* file_find */


/*
* Return list of expansions.
*/
static list_t *expand(char *aliasfile, char *name, char **errorstr)
{
  char *e;       /* dynamically allocated string */
  char *s, *d;
  list_t *list = 0;
  char *bow, *eow;  /* beginning and one char beyond end of word */ 
  char *name_stripped;

  /* Strip name */
  if (!name || !name[0]) return 0;

  /*
   * Remove all but numbers, digits and @, . from name to allow handling of
   * phone numbers such as +49-30-254-99-200.
   */
  name_stripped = malloc(strlen(name) + 1);
  for (s = name, d = name_stripped; *s; s++) {
    if ((!isspace((int)*s) && !ispunct((int)*s)) || strchr("@.-", *s)) *d++ = *s;
  }
  *d = '\0';
  name = name_stripped;

  /* Try YP first. */
  if (!(e = yp_find(name))) {
    if (!(e = file_find(aliasfile, name, errorstr))) {
      free(name);
      return 0;
    }
  }

  /* Check if alias is of form :include:file list */
  for (s = e; *s && isspace((int)*s); s++) ;
  if (!strncmp(s, ":include:", strlen(":include:"))) {
    for (d = s; *d && !isspace((int)*d); d++) ;
    *d = '\0';  /* remove trailing white space */
    list = alias_file(s + strlen(":include:"), errorstr);
  }
  /* Convert comma-separated entries into list. */
  else {
    bow = eow = 0;
    for (; *s; s++) {
      if (*s == ',') *s = ' ';
      if (!bow) {
        if (!isspace((int)*s)) bow = s;
      }
      if (bow) {
        if (isspace((int)*s)) eow = s;
        else if (*(s+1) == '\0') eow = s+1;
      }
      if (eow) {
        *eow = '\0';
        list = list_add(list, strdup(bow));
        bow = eow = 0;
      }
    }
  }
  free(e);
  free(name);
  return list;
} /* expand */


/*
* Search in file 'aliasfile' for name 'name'.
*/
list_t *alias_search(char *aliasfile, char *name, char **errorstr)
{
  list_t *list = 0;
  list_t *p;
  list_t *r;

  list = expand(aliasfile, name, errorstr);

  p = list;
  while (p) {
    list = list_replace(list, p, (r = expand(aliasfile, p->word, errorstr)));
    /* If not replaced, move pointer. */
    p = r ? r : p->next;
  }

  return list;
} /* alias_search */


static canonicalize_status translate_list(list_t *list, uri_t *uri,
                                          char *errorstr,
                                          canonicalize_result *result);
static int merge_uri(char *alias, uri_t *uri);

/*
 * aliases()
 * The interface to canonicalize.c.  Convert the list_t result
 * of alias_search to a canonicalize_result, with URIs appropriately
 * altered.
 */
canonicalize_status aliases(uri_t* uri, canonicalize_result* result)
{
  list_t *alias_list;
  canonicalize_status status;
  char *errorstr = NULL;

  if (!options.do_aliases) {
    return CANONICALIZE_NO_TRANSFORMATION;
  }

  if (strcasecmp(uri->scheme, "sip") != 0) {
    return CANONICALIZE_NO_TRANSFORMATION;
  }

  if (uri->user == NULL || uri->user[0] == '\0') {
    return CANONICALIZE_NO_TRANSFORMATION;
  }

  alias_list = alias_search(options.aliases_file, uri->user, &errorstr);

  status = translate_list(alias_list, uri, errorstr, result);

  list_free(alias_list);
  free_valid(errorstr);

  return status;
}


/* translate_list
 * translate a list_t* into a canonicalize result.
 * There are five possibilities:
 * 1. No aliases found.
 * 2. A single alias found, of the form 'lennox' (local host)
 * 3. A single alias found, of the form 'lennox@orion' (local network)
 * 4. A single alias found, of the form 'lennox@example.com' (remote host)
 * 5. Multiple aliases found, of the forms given in {2,3,4}.
 */
static canonicalize_status translate_list(list_t *list, uri_t *uri,
                                          char *errorstr,
                                          canonicalize_result *result)
{
  int i;
  list_t *list_p; 

  if (list == NULL) {
    if (errorstr != NULL) {
      result->nameaddrs = (char **) calloc(sizeof(char *), 1);
      if (result->nameaddrs != NULL) {
        strassign(&result->nameaddrs[0], errorstr);
      }
      return CANONICALIZE_ERROR;
    }
    else {
      return CANONICALIZE_NO_TRANSFORMATION;
    }
  }

  if (list->next == NULL) {
    /* Only one alias found.  This is therefore a unique transformation,
       cases 2, 3, or 4. */
    /* If merge_uri doesn't change anything, then nothing happened,
       and we want to continue onward to the next attempted
       transformation. */
    if (merge_uri(list->word, uri)) {
      return CANONICALIZE_UNIQUE_TRANSFORMATION;
    }
    else {
      return CANONICALIZE_NO_TRANSFORMATION;
    }
  }

  /* If we get here, there are multiple aliases. */
  /* Count them. */
  result->count = 0;
  for (list_p = list; list_p != NULL; list_p = list_p->next) {
    result->count++;
  }
  result->nameaddrs = (char **) calloc(sizeof(char *), result->count);
  if (result->nameaddrs == NULL) {
    return CANONICALIZE_ERROR;
  }

  for (list_p = list, i = 0; list_p != NULL; list_p = list_p->next, i++) {
    uri_t tmp_uri;

    tmp_uri = URI_Copy(*uri);
    merge_uri(list_p->word, &tmp_uri);

    result->nameaddrs[i] = URI_AsString(tmp_uri);

    URI_Free(&tmp_uri);
  }

  return CANONICALIZE_MULTIPLE_TRANSFORMATION;
}


/*
 * merge_uri()
 * merge the alias 'alias' into the uri 'uri'.
 * There are three possibilities for the alias:
 * 1. The form 'lennox' (local host)
 * 2. The form 'lennox@orion' (local network)
 * 3. The form 'lennox@example.com' (remote host)
 * In the first two cases, we leave the 'host' part of the URI alone;
 * in the latter case, we replace it.
 * Return true if we've changed the URI in some way.
 */
static int merge_uri(char *alias, uri_t *uri)
{
  char *s, *host;
  if ((s = strchr(alias, '@')) != NULL) {
    /* '@' sign present: case 2 or 3 */
    *s = '\0';
    host = s + 1;
  }
  else {
    host = NULL;
  }

  if (host && strchr(host, '.') != NULL) {
    /* case 3 */
    strassign(&uri->user, alias);
    strassign(&uri->host, host);
    return 1;
  }
  else {
    /* case 1 or 2 */
    if (strcmp(uri->user, alias) == 0) {
      /* No change */
      return 0;
    }
    else {
      strassign(&uri->user, alias);
      return 1;
    }
  }
}
