/*
 * canon_core.c
 * Convert URIs into their canonical forms, as understood by an
 * Internet telephony server (in particular sipd.)
 *
 * Most of the code is copied from the canonicalize.c (sipd v1.17).
 * Changes are in the way canonicalize is compiled - now as a shared
 * library, with an exported interface defined in 
 * libsip/canonicalizer_interface.h
 *
 * Copyright 2001 Columbia University.  All rights reserved.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include "sysdep.h"
#include "strdupn.h"
#include "error.h"

#include "uri.h"
#include "canon_core.h"

/* Global configuration options */
struct canonicalize_options options;

/* Order of canonicalization mappings */
static canonicalize_fn *transformations[] =
#ifndef WIN32
    { &database, &aliases, &namemapper, &dialplan };
#else
    { &database, &aliases, &dialplan };
#endif

static const int transformations_count =
    sizeof(transformations) / sizeof(transformations[0]);

/*
 * init_opts()
 * Initialize global 'options' struct.
 */
static void init_opts(void)
{
  memset(&options, 0, sizeof(options));

  strassign(&options.aliases_file, DEFAULT_ALIASES_FILE);
  strassign(&options.dialplan_file, DEFAULT_DIALPLAN_FILE);
} /* init_opts */

/*
 * usage()
 * Print library usage, and exit.
 */
static void usage(char *pname)
{
  fprintf(stderr, "Usage:  %s [-a] [-n] [-d]\n", pname);
  fprintf(stderr, "\t[-A aliasfile] [-D dial plan file] [-T database table] [-u]\n");
  exit(2);
} /* usage */


/*
 * parse_opts()
 * Parse command line options, and fill in the 'options' global struct.
 * Exit on error.
 */
static void parse_opts(int argc, char *argv[])
{
  extern char *optarg;
  int i;
  
  for(i=0; i<argc; i++) {
   if(strcmp(argv[i], "-a") == 0) 
     options.do_aliases = 1;
   if(strcmp(argv[i], "-n") == 0) 
     options.do_namemapper = 1;
   if(strcmp(argv[i],"-d") == 0)
      options.do_dialplan = 1;
   if(strcmp(argv[i],"-u") == 0)
      options.require_userparam = 1;
   if(strcmp(argv[i],"-A") == 0) {
      if(i+1 < argc) {
         strassign(&options.aliases_file, argv[i+1]); 
         ++i;
      }
   }
   if(strcmp(argv[i],"-D") == 0) {
      if(i+1 < argc) {
         strassign(&options.dialplan_file, argv[i+1]); 
         ++i;
      }
   }
   
   if(strcmp(argv[i],"-T") == 0) {
      if(i+1 < argc) {
         strassign(&options.table_uri, argv[i+1]); 
         ++i;
      }
   }
  }
  return;
} /* parse_opts */


/*
 * print_error()
 * Print an error message.  If either the pointer passed, or the string
 * it points to, is null, the most likely reason is a memory allocation
 * failure.
 */
static void print_error(char **error_str)
{
  if (error_str != NULL && *error_str != NULL) {
    printf(":ERROR %s\n", *error_str);
  }
  else {
    printf(":ERROR Memory allocation failure\n");
  }
  fflush(stdout);
} /* print_error */


/*
 * free_canonical_result()
 * Free a canonical_result structure.
 */
static void free_canonical_result(canonicalize_result *result)
{
  int i;
  for (i = 0; i < result->count; i++) {
    free_valid(result->nameaddrs[i]);
  }
  free_valid(result->nameaddrs);
  result->nameaddrs = NULL;
  result->count = 0;
} /* free_canonical_result  */


/*
 * canonicalize_uri()
 * Canonicalize input_uri, and return canonicalized URI,
 * 
 * Returns 0 on success, 1 on error.
 */
int canonicalize_uri(uri_t* input_uri, canon_result* out_result)
{
  int i;
  canonicalize_status status = CANONICALIZE_NO_TRANSFORMATION;
  canonicalize_result result = {0, NULL};
  DString name, addr;

  if (!input_uri || !out_result) {
    error("canonicalize_one_uri","Parameters should be non null...\n");
    return 0;
  }

  if (!input_uri->scheme) {
    /* 
     * if there was an error in uri, copy input to output 
     */
    out_result->status = canon_unique;
    *out_result->u.unique = URI_Copy(*input_uri);
    return 0;
  }
  
  for (i = 0; i < transformations_count; i++) {
    status = (*transformations[i])(input_uri, &result);
    if (status != CANONICALIZE_NO_TRANSFORMATION) {
      break;
    }
  }

  switch(status) {
  case CANONICALIZE_NO_TRANSFORMATION:
  case CANONICALIZE_UNIQUE_TRANSFORMATION:
    
    out_result->u.unique = (uri_t*) malloc(sizeof(uri_t));
    
    if (out_result->u.unique == NULL) {
      out_result->status = canon_error;
      out_result->u.error = NULL;        /* assigned at the end */
      break;
    }
    
    out_result->status = canon_unique;
    *out_result->u.unique = URI_Copy(*input_uri);
    break;
  
  case CANONICALIZE_MULTIPLE_TRANSFORMATION:
    
    out_result->u.multiple.nameaddrs = 
      (struct canon_nameaddrs *) calloc(result.count, 
					sizeof(struct canon_nameaddrs));
    
    if (out_result->u.multiple.nameaddrs == NULL) {
      out_result->status = canon_error;
      out_result->u.error = NULL;
      break;
    }

    DStringInit(&name);
    DStringInit(&addr);

    out_result->status = canon_multiple;
    out_result->u.multiple.nameaddr_count = result.count;
    
    for (i = 0; i < result.count; i++) {
      
      DStringTrunc(&name);
      DStringTrunc(&addr);
      
      SIP_NameAddr(result.nameaddrs[i], &name, &addr);
      
      strassign(&out_result->u.multiple.nameaddrs[i].name, 
		DStringValue(&name));
      
      out_result->u.multiple.nameaddrs[i].addr = 
	URI_Parse(DStringValue(&addr));
    }
    
    DStringFree(&name);
    DStringFree(&addr);
    
    break;
    
  case CANONICALIZE_ERROR:
    
    out_result->status = canon_error;

    if (result.nameaddrs && *result.nameaddrs) {
      seterror(&out_result->u.error,"Canonicalization error: %s\n", *result.nameaddrs);
    }
    else {
      out_result->u.error = NULL;
    }

    break;
  
  default:
    out_result->status = canon_error;
    seterror(&out_result->u.error, "Unknown status code: %d\n",status);
    break;
  }
  
  free_canonical_result(&result);
  
  if (out_result->status == canon_error) {
    if (out_result->u.error == NULL)
      seterror(&out_result->u.error, "Memory allocation failure\n");
    
    return 1;
  }
 
  return 0;
  
} /* canonicalize_uri */

/*
 * InitCanonicalize2.
 *
 * Called by sipd during init, with appropriate command line options.
 * Initialize those modules which have been enabled.
 * Print an error string and exit if any of them fail.
 */
int InitCanonicalize2(int argc, char *argv[], cinema_table *database_table)
{
  char *error_str;
  
  init_opts();
  
  parse_opts(argc, argv);
  
  if (options.do_dialplan) {
    if (init_dialplan(&error_str) != 0) {
      print_error(&error_str);
      free(error_str);
      return -1;
    }
  }
  
  if (options.do_namemapper) {
#ifndef WIN32
    if (init_namemapper(&error_str) != 0) {
      print_error(&error_str);
      free(error_str);
      return -1;
    }
#else
    error("init_modules", "No namemapper implemented for win32\n");
#endif
  }

  /*
   * Reuse the same database connection if possible (i.e, database_table
   * != NULL).
   */
  if (options.table_uri != NULL || database_table != NULL) {
    if (init_database(&error_str, database_table) != 0) {
      print_error(&error_str);
      free(error_str);
      return -1;
    }
  }

  debug("InitCanonicalize2","Started canonicalizer successfully...\n");
  return 0;
    
} /* InitCanonicalize2 */

/*
 * InitCanonicalize.
 *
 * Called by sipd during init, with appropriate command line options.
 * Simply call InitCanonicalize2 with last arg as null.
 *
 */
int InitCanonicalize(int argc, char *argv[])
{

  return InitCanonicalize2(argc, argv, NULL);

} /* InitCanonicalizer */


/*
 * CloseCanonicalize
 *
 * Cleanup resources. Generally, this will never need to be used
 * as rightnow there are no callers (i.e., we run as long as the 
 * executable hosting us runs)
 *
 */
int CloseCanonicalize()
{
  
  return 0;
  
} /* CloseCanonicalize */


/*
 * seterror()
 * Set an error string.  Support function for other utilities.
 */
void seterror(char **error, const char *format, ...)
{
  DString errorstr;
  va_list ap;

  va_start(ap, format);

  DStringInit(&errorstr);
  DStringVPrintf(&errorstr, format, ap);
  
  va_end(ap);

  *error = DStringValue(&errorstr);
} /* seterror */


/*
 * The _init function will be called by the loader, when we are
 * loaded into the process address space. The _fini function will be
 * called by loader when the shared library is unloaded. This works
 * for all unix platforms.
 *
 * For windows, we use DllMain.
 */
#ifndef WIN32

void _init() {

  return;
} /* _init */

void _fini() {
  
  return;
} /* _fini */

#else

BOOL APIENTRY DllMain (int handle, long reason, void* arg) {

  return 1;

} /* DllMain */

#endif

