/*
 * canonicalize.c
 * Convert URIs into their canonical forms, as understood by an
 * Internet telephony server (in particular sipd.)
 *
 * Copyright 2001 Columbia University.  All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>

#include "sysdep.h"
#include "strdupn.h"

#include "canon_core.h"
#include "error.h"

/* copy of some functions from libsip - as libsip wont get linked
 * when canonicalize executable is built. Making libsip as a dependency
 * again wont work.
 */
static void free_result(canon_result *result);
static char *SIP_BuildNameAddr(const char *name, const uri_t *uri);

/*
 * init_stdio()
 * Initialize stdio.  stdout needs to be line-buffered, or Bad Things
 * can happen.
 */
static void init_stdio(void)
{
  setvbuf(stdout, NULL, _IOLBF, 0);
}

/*
 * usage()
 * Print program usage, and exit - moved to canon_core.c as Init
 * Canonicalizer does all these.
 */

/*
 * print_multiple()
 * Print a multiple result set.
 */
static void print_multiple(canon_result *result)
{
  int i;

  char *p;

  printf(":MULTIPLE %d\n", result->u.multiple.nameaddr_count);
  for (i = 0; i < result->u.multiple.nameaddr_count; i++) {
    p = SIP_BuildNameAddr(result->u.multiple.nameaddrs[i].name, 
			  &result->u.multiple.nameaddrs[i].addr);
    if (p) { 
      printf(":+%s\n", p);
      free(p);
    }
  }
  fflush(stdout);
} /* print_multiple */


/*
 * print_error()
 * Print an error message.  If either the pointer passed, or the string
 * it points to, is null, the most likely reason is a memory allocation
 * failure.
 */
static void print_error(char **error_str)
{
  if (error_str != NULL && *error_str != NULL) {
    printf(":ERROR %s\n", *error_str);
  }
  else {
    printf(":ERROR Memory allocation failure\n");
  }
  fflush(stdout);
} /* print_error */


/*
 * canonicalize_one_uri()
 * Read one URI from stdin, canonicalize it, and print the canonicalize URI,
 * or an error message, to stdout.
 * Return zero if stdin has been closed.  In this case, we should exit.
 */
int canonicalize_one_uri(void)
{
  char buf[BUFSIZ];
  char *input;
  char *canonical;
  uri_t uri;
  canon_result out_result;

  input = fgets(buf, BUFSIZ, stdin);
  if (input == NULL) {
    if (ferror(stdin)) {
      printf(":ERROR Error on stdin: %s\n", strerror(errno));
      fflush(stdout);
      return 0;
    }
    else if (feof(stdin)) {
      printf(":ERROR EOF on stdin\n");
      fflush(stdout);
      return 0;
    }
  }

  uri = URI_Parse(input);

  if ( uri.scheme == NULL ) {
    /* 
     * if there was an error in uri.
     * Just print out the input to the output.
     */
    printf("%s\n", input);
    fflush(stdout);
    URI_Free(&uri);
    return 1;
  }

  canonicalize_uri(&uri, &out_result);
  
  switch(out_result.status) {
  case CANONICALIZE_NO_TRANSFORMATION:
  case CANONICALIZE_UNIQUE_TRANSFORMATION:
    canonical = URI_AsString(uri);
    printf("%s\n", canonical);
    fflush(stdout);
    free_valid(canonical);
    break;
  case CANONICALIZE_MULTIPLE_TRANSFORMATION:
    print_multiple(&out_result);
    break;
  case CANONICALIZE_ERROR:
    print_error(&out_result.u.error);
    break;
  default:
    printf(":ERROR Unknown status code %d\n", out_result.status);
    fflush(stdout);
    break;
  }

  URI_Free(&uri);
  free_result(&out_result);
  
  return 1;
}

/*
 * main()
 * Main program.  Parse options, initialize enabled modules, then
 * read and canonicalize URIs.
 */
int main(int argc, char *argv[])
{

  init_stdio();

  if(InitCanonicalize(argc, argv) != 0) 
    return 0;

  /* Successful startup */
  printf(":OK\n");
  fflush(stdout);
  
  while(1) {
    if (!canonicalize_one_uri()) {
      break;
    }
  }
  
  exit(0);
} /* main */


/*
 * from libsip/canonicalize.c/free_canon_result, not included to avoid cyclic
 * dependencies.
 */
void free_result(canon_result *result) {

  int i;

  switch (result->status) {
  case canon_unique:
  case canon_remote:
    URI_Free(result->u.unique);
    free_valid(result->u.unique);
    result->u.unique = NULL;
    break;

  case canon_multiple:
    for (i = 0; i < result->u.multiple.nameaddr_count; i++) {
      URI_Free(&result->u.multiple.nameaddrs[i].addr);
      free_valid(result->u.multiple.nameaddrs[i].name);
    }
    free_valid(result->u.multiple.nameaddrs);
    result->u.multiple.nameaddrs = NULL;
    break;

  case canon_error:
    free_valid(result->u.error);
    break;

  default:
    error("free_canon_result",
          "Unknown result status %d\n", result->status);
    break;
  }
} /* free_result */


/*
 * SIP_BuildNameAddr()
 * Given a name and address, print the nameaddr to a string.
 * It is the caller's responsibility to free the string.
 *
 * This is used to build Contact headers.
 */
static char *SIP_BuildNameAddr(const char *name, const uri_t *uri)
{
  DString nameaddr, addrstr;

  DStringInit(&nameaddr);
  DStringInit(&addrstr);

  if (name && *name) {
    /* Display name not null or empty */
    DStringAppend(&nameaddr, "\"", -1);
    DStringAppendQuoted(&nameaddr, name, "\\\"");
    DStringAppend(&nameaddr, "\" ", -1);
  }

  URI_SPrint(uri, &addrstr);

  DStringAppend(&nameaddr, "<", -1);
  DStringAppendD(&nameaddr, &addrstr);
  DStringAppend(&nameaddr, ">", -1);

  DStringFree(&addrstr);

  return DStringValue(&nameaddr);
} /* SIP_BuildNameAddr */


/*
 * SIP_NameAddr() 
 * Break up name-addr (To, From, Contact, Route) addresses:
 * '[name] <url>' or 'url'.  Return false if the argument isn't
 * of that form.  DStrings must already be initialized.
 * addr may have surrounding whitespace and/or <> remaining; URI_Parse
 * can handle this.
 */
int SIP_NameAddr(char *v, DString *name, DString *addr)
{
  char *s, *uri_start = 0;
  int backslash = 0;

  if (v) {
    /* "name" <...> */
    if (*v == '"') {
      for (s = v+1; *s; s++) {
        if (*s == '\\') {
          backslash = 1;
        }
        else if (*s != '"' || backslash) {
          DStringAppend(name, s, 1);
          backslash = 0;
        }
        /* reached end of quoted string (") */
        else {
          break;
        }
      }
      DStringAppend(addr, s+1, -1);
    }
    /* name <...> */
    else if ((uri_start = strchr(v, '<')) != NULL) {
      /* remove any trailing white space after name */
      for (s = uri_start-1; (s > v) && isspace((int)*s); s--) ;
      DStringAppend(name, v, s - v + 1);
      DStringAppend(addr, uri_start, -1);
    }
    /* URL only */
    else {
      DStringAppend(addr, v, -1);
    }
  }
  return 1;
} /* SIP_NameAddr */
