/* database.c - Implement the transformations associated with
 * a database 'aliases' table.
 * 
 * Copyright 2001 Columbia University.  All rights reserved.
 */

#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>
#include <regex.h>

#include "sysdep.h"
#include "dstring.h"
#include "strdupn.h"

#include "dbaccess.h"

#include "canon_core.h"

static cinema_table *table = NULL;

#ifdef __GNUC__
/* This function uses printf-style formats, so tell GNUC to do printf
 * format checking. */
#define ATTRIBUTE_PRINTF(fmt, args) \
   __attribute__ ((format (printf, fmt, args)))
#else
#define ATTRIBUTE_PRINTF(fmt, args)
#endif

/* A useful function for setting error strings */
static void set_result_error(canonicalize_result *result,
                             const char *format, ...)
     ATTRIBUTE_PRINTF(2, 3);

#undef ATTRIBUTE_PRINTF


/*
 * set_result_error()
 * Set a canon_result to an error string.
 */
static void set_result_error(canonicalize_result *result,
                             const char *format, ...)
{
  DString errorstr;
  va_list ap;

  result->nameaddrs = (char **) calloc(sizeof(char *), 1);
  if (result->nameaddrs == NULL)
    return;

  va_start(ap, format);

  DStringInit(&errorstr);
  DStringVPrintf(&errorstr, format, ap);
  
  va_end(ap);

  result->nameaddrs[0] = DStringValue(&errorstr);
} /* set_result_error */


/*
 * init_database()
 * Initialize the connection to the MySQL database only if we were
 * not supplied with a reference to an existing connection. The
 * reason we have this explicit, is that when we load ourselves
 * as a shared object into sipd's address space, the fastsql
 * singleton wont be visible here, unless we pass a reference like
 * this.
 *
 * On error, return non-zero and fill in errorstr if possible.
 */
int init_database(char **errorstr, cinema_table *database_table)
{
  char *db_error_str;
  db_err db_ret;

  if(database_table) {
    table = database_table;
    return 0;
  }
  
  table = new_cinema_table(options.table_uri, &db_ret, &db_error_str);
  if (table == NULL) {
    seterror(errorstr, "Could not open database table: %s", db_error_str);
    return 1;
  }
  return 0;
} /* init_database */


/*
 * database()
 * The core canonicalization routine for the database module.
 */
canonicalize_status database(uri_t* uri, canonicalize_result* result)
{
  char *db_error_str = NULL;
  db_err db_ret;
  char *key = NULL;
  alias_table_entry *entry = NULL;
  canonicalize_status ret = CANONICALIZE_ERROR;
  uri_t *primary_uri;

  if (options.table_uri == NULL) {
    return CANONICALIZE_NO_TRANSFORMATION;
  }

  if (strcasecmp(uri->scheme, "sip") != 0 &&
      strcasecmp(uri->scheme, "tel") != 0) {
    return CANONICALIZE_NO_TRANSFORMATION;
  }

  if ((key = get_table_key_from_URI(table, uri,
                                    &db_ret, &db_error_str)) == NULL) {
    /* The URI isn't in a form which the database can transform to one of its
       keys. */
    /* XXX TODO: Should this be an error instead? */
    return CANONICALIZE_NO_TRANSFORMATION;
  }

  if ((entry = alias_get_by_alias(table, key,
                                  &db_ret, &db_error_str)) == NULL) {
    if (db_ret == DB_NOTFOUND) {
      ret = CANONICALIZE_NO_TRANSFORMATION;
      goto done;
    }
    else {
      set_result_error(result, "Database lookup: %s", db_error_str);
      ret = CANONICALIZE_ERROR;
      goto done;
    }
  }
  if (entry->primary_user == NULL) {
    /* XXX: Shouldn't happen?  Allow it... */
    ret = CANONICALIZE_NO_TRANSFORMATION;
    goto done;
  }

  if ((primary_uri = 
       get_URI_from_table_key(table, entry->primary_user,
                              &db_ret, &db_error_str)) == NULL) {
    set_result_error(result,
                     "Database lookup: Could not transform primary user %s into URI",
                     entry->primary_user);
    ret = CANONICALIZE_ERROR;
    goto done;
  }

  URI_Free(uri);
  *uri = *primary_uri;
  ret = CANONICALIZE_UNIQUE_TRANSFORMATION;
  
done:
  free_valid(db_error_str);
  free_valid(key);
  if (entry != NULL) {
    free_alias_table_entry(entry);
  }

  return ret;
} /* database */


