/* dialplan.c - Implement the transformations associated with
 * a telephone dialplan.
 * 
 * Copyright 2001 Columbia University.  All rights reserved.
 */

#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>
#include <regex.h>

#include "sysdep.h"
#include "glob2regex.h"
#include "dstring.h"
#include "strdupn.h"
#include "pattern.h"

#include "canon_core.h"

/* Pattern types. */
typedef enum {unknown_pat, glob_pat, regex_pat} plan_pat_type;

/*
 * An entry from the dialplan file.
 * Since we always walk the dialplans in order, a singly-linked list is a
 * good-enough data structure.
 */
typedef struct dialplan_entry {
  pattern_t pattern;
  char *result;
  int lineno;

  struct dialplan_entry *next;
} dialplan_entry;

/* The current dialplan file. */
static dialplan_entry *dialplans;

/*
 * istelephone()
 * Apply a heuristic to guess whether a number is a telephone number.
 * TODO: make this smarter -- recognize more of the tel syntax.
 *
 * Added a patch for
 
 * canonicalize only recognizes SIP URIs with a leading + if you set the -u
 * parameter. I suggest that dialplan.c:istelephone is extended to recognize
 * URIs with a leading + as a telephone number as well so that (i.e. without
 * the -u parameter).
 */
static int istelephone(const char *user)
{
  const char *ptr = user;
  int number = 0;

  if (strlen(user) > 1 && *ptr == '+' &&
      *(ptr+1) >= '0' && *(ptr+1) <= '9' ) {
    number = 1;
    ptr++;
  }

  for (; *ptr; ptr++ ) {
    if (*ptr >= '0' && *ptr <= '9') {
      number = 1;
    } else {
      return 0;
    }
  }
 
  return number;
} /* istelephone */

/*
 * free_dialplan()
 * Free a single dialplan entry.
 */
static void free_dialplan(dialplan_entry *plan)
{
  pattern_free(&plan->pattern);
  free_valid(plan->result);
  free_valid(plan);
} /* free_dialplan */


/*
 * free_dialplans()
 * Free a list of dialplan entries.
 */
static void free_dialplans(dialplan_entry *plans)
{
  dialplan_entry *plan, *plan_next;

  for (plan = plans; plan != NULL; plan = plan_next) {
    plan_next = plan->next;
    free_dialplan(plan);
  }
} /* free_dialplans */


/*
 * parse_dialplan_entry()
 * Parse one dialplan entry line, filling in the plan structure.
 * On error, return false, and fill in 'error' if possible
 */
static int parse_dialplan_entry(dialplan_entry *plan, const char* line,
                                int lineno, char **error)
{
  const char *p, *n;
  char errbuf[1024];
  char *pattern;
  int quote_next = 0;
  int ret;

  /* Find the pattern.  The pattern terminates with the first
   * non-quoted space. */
  for (p = line; quote_next || (*p != ' ' && *p != '\t' && *p != '\n'); p++) {
    switch (*p) {
    case '\\':
      quote_next = 1;
      break;

    case '\n':
      quote_next = 0;
      break;

    default:
      quote_next = 0;
      continue;
    }
  }

  pattern = strdupn(line, (p - line));
  if (pattern == NULL) {
    *error = NULL;
    return 0;
  }
    
  ret = pattern_compile(&plan->pattern, pattern, errbuf, sizeof(errbuf));
  free_valid(pattern);
  if (!ret) {
    seterror(error, "%s at line %d", errbuf, lineno);
    return ret;
  }

  /* Skip over intermediate whitespace */
  while (*p != '\0' && *p != '\n' && isspace((int)*p)) {
    p++;
  }

  /* Find terminating newline */
  if (*p == '\0' || *p == '\n' || (n = strchr(p, '\n')) == NULL) {
    /* Didn't find a second field before '\n'.
    * This means we need to suppress the dialplan translation for
    * this entry. */
    plan->result = strdup("");
  }
  else {
    plan->result = strdupn(p, (n - p));
    if (plan->result == NULL) {
      *error = NULL;
      return 0;
    }
  }

  plan->lineno = lineno;

  return 1;
} /* parse_dialplan_entry */


/*
 * init_dialplan()
 * Initialize the dialplan module, by reading the dialplan file.
 * May be called more than once -- if so, free the pre-existing dialplan
 * information.  (This would typically occur on receipt of a SIGHUP signal.)
 * Return non-zero on error, with a reason string pointed to by "error" if
 * possible.
 */
int init_dialplan(char **error)
{
  dialplan_entry *new_plan, **prev_plans;
  FILE *f;
  char line[1024];
  int line_cnt;
  *error = NULL;

  if (dialplans != NULL) {
    free_dialplans(dialplans);
    dialplans = NULL;
  }
  
  if (!(f = fopen(options.dialplan_file, "r"))) {
    seterror(error, "Dialplan: %s: %s", options.dialplan_file,
             strerror(errno));
    return -1;
  }

  prev_plans = &dialplans;

  line_cnt = 0;
  while (fgets(line, sizeof(line), f)) {
    line_cnt++;
    if (!line[0] || line[0] == '#' || line[0] == '\n') {
      continue;
    }
    new_plan = (dialplan_entry *) malloc(sizeof(dialplan_entry));
    if (new_plan == NULL) {
      *error = NULL;  /* Don't try to allocate any further memory. */
      goto bailout;
    }
    memset(new_plan, 0, sizeof(dialplan_entry));

    if (!parse_dialplan_entry(new_plan, line, line_cnt, error)) {
      free_dialplan(new_plan);
      goto bailout;
    }
    
    *prev_plans = new_plan;
    new_plan->next = NULL;
    prev_plans = &new_plan->next;
  }
  fclose(f);

  return 0;

bailout:
  free_dialplans(dialplans);
  dialplans = NULL;
  return -1;
} /* init_dialplan() */


/*
 * dialplan()
 * The core canonicalization routine for the dialplan module.
 */
canonicalize_status dialplan(uri_t* uri, canonicalize_result* result)
{
  char *tel_field;
  char *param;
  char errbuf[1024];

  dialplan_entry *plan;

  if (!options.do_dialplan) {
    return CANONICALIZE_NO_TRANSFORMATION;
  }

  if (strcasecmp(uri->scheme, "sip") == 0) {
    /* Telephone numbers in SIP URIs are in the 'user' field, if the
     * URI has a user=phone parameter. */
    if (uri->user == NULL || uri->user[0] == '\0') {
      return CANONICALIZE_NO_TRANSFORMATION;
    }

    if ( options.require_userparam ) {
      if ((param = URI_Parameter(*uri, "user")) == NULL ||
          strcasecmp(param, "phone") != 0) {
        return CANONICALIZE_NO_TRANSFORMATION;
      }
    }
    else {
      /* Do not require userparam. So check if the 
       * user field is a telephone number or not.
       */
      if (!istelephone(uri->user)) {
        return CANONICALIZE_NO_TRANSFORMATION;
      }
    }

    tel_field = uri->user;
  }
  else if (strcasecmp(uri->scheme, "tel") == 0) {
    /* Telephone numbers in TEL URIs are in the 'host' field with
     * our URI parser. */
    if (uri->host == NULL || uri->host[0] == '\0') {
      return CANONICALIZE_NO_TRANSFORMATION;
    }

    tel_field = uri->host;
  }
  else {
    /* Not sip or tel; don't transform. */ 
    return CANONICALIZE_NO_TRANSFORMATION;
  }

  for (plan = dialplans; plan != NULL; plan = plan->next) {
    int ret;
    char *new_uri_str;
    ret = pattern_interpolate(tel_field, &plan->pattern, plan->result,
                              &new_uri_str, errbuf, sizeof(errbuf));
    if (ret == 0)
      continue;
    if (ret < 0) {
      if ( ret == -2 ) {
        /* Skip this; no transformation */
        return CANONICALIZE_NO_TRANSFORMATION;
      }

      result->nameaddrs = malloc(sizeof(char *));
      if (result->nameaddrs != NULL) {
        seterror(&result->nameaddrs[0], "%s at %d", errbuf, plan->lineno);
        result->count = 1;
      }
      /* Otherwise this stays NULL, so the error message will be "out of
         memory", which is correct */
      return CANONICALIZE_ERROR;
    }
    /* else ret > 0 */
    URI_Free(uri);
    *uri = URI_Parse(new_uri_str);
    free_valid(new_uri_str);
    return CANONICALIZE_UNIQUE_TRANSFORMATION;
  }

  return CANONICALIZE_NO_TRANSFORMATION;
} /* dialplan */


