/*
 * namemapper.c
 * Copyright 1997, 2001 Columbia University
 *
 * Search in system password files for names that match the user part of the 
 * passed-in URI.
 *
 * This works as follows.  A name (both the queried name and the database
 * name) is treated as a series of words separated by spaces:
 *   Apostrophes are removed; other punctuation characters are equivalent
 *    to spaces (that is, they delimit words). 
 *   The order of words within the query or the DND entry is not
 *    significant.
 *   Upper-case/lower-case distinctions are not significant.
 *
 * A query matches a database name if:
 *   Each query word matches a word in the DND entry, either as an initial
 *    substring or completely (an "exact match").
 *   At least one multiletter query word exactly matches a word in the DND
 *    entry.  
 *   A given word in the DND entry may be used as a match for a query word
 *    only once.
 *
 * As a special case, a one-word query can match a) usernames, or 
 * b) patterns of the form firstlast, firstmlast, flast, or fmlast.  The
 * definitions of "first", "middle", and "last" are determined by a
 * heuristic: the first word of a name is the first name, if the name has
 * more than one word; the last word of the name, optionally preceded by
 * some number of words beginning with lower-case letters, is the last name;
 * and any words in between these two are the middle name.  See the function
 * nameparts().
 *
 * Matching is implemented using an in-memory hash table, which is built
 * by the function init_namemapper().
 *
 * The source database is the 'gecos' field of the system password database.
 * It can contain strings of the following formats:
 *   John Doe (some comment)
 *   John Doe,comment
 *   J. P. Doe
 *
 * If multiple matches are found, nameaddr forms are returned.
 */

#include <pwd.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>

#include "strdupn.h"
#include "dstring.h"

#include "dict.h"
#include "ht.h"

#include "canon_core.h"

/* A list of words.  Should be sorted, downcased, and alnum only. */
typedef struct {
  int count;
  char **words;
} wordlist;

/* The ways a name can match */
/* If multiple matches are found for a given username, the one with the
 * first type from this list is preferred. */
typedef enum {
  match_username,
  match_word,
  match_firstlast,
  match_firstmlast,
  match_firstmiddlelast,
  match_fmlast,
  match_flast,
} match_type_t;

/* The stored value of a namemap lookup. */
typedef struct {
  char *username;
  char *gecos;
  wordlist wordlist;
} namemap_entry_t;

/* An element of a linked list of namemap entries.
 * The return value of namemap_search, or a chain hanging
 * off the hash table.
 * Remembers why the entry 'entry' was stored in this list.
 * The entry shouldn't normally be modified, because it's shared between
 * many different list elements.  (When freeing the table, assume that
 * the 'match_username' match is the canonical one.)
 */
typedef struct namemap_list_t {
  match_type_t type;
  const namemap_entry_t *entry;
  struct namemap_list_t *next;
} namemap_list_t;

/* The namemap hash table */
static dict_h names = NULL;
/* And its error code */
static int names_errno;

/* An entry in the namemap hash table. */
typedef struct {
  char *keyword;
  namemap_list_t *entries;
} namemap_hash_entry_t;

/* Initialize the namemap hash table. */
static int init_names_hash(char **errorstr);




static void free_wordlist(wordlist *words);

/*
 * dupn_and_downcase()
 * Duplicate the string 'string', downcasing it and skipping over all
 * non alphanumeric characters.
 */
static char *dupn_and_downcase(const char *str, int len)
{
  const char *s;
  char *dupstr, *t;

  if (len < 0) {
    len = strlen(str);
  }

  dupstr = malloc(len+1);
  if (dupstr == NULL)
    return NULL;

  for (s = str, t = dupstr; *s != '\0' && s - str < len ; s++) {
    if (isalnum((int)*s)) {
      *t = tolower((int)*s);
      t++;
    }
    /* Else ignore */
  }

  *t = '\0';

  return dupstr;
} /* dupn_and_downcase */


/*
 * add_new_word()
 * Add the word that spans from start to end-1 to wordlist.
 * On memory error, set count to -1.
 */
static void add_new_word(wordlist *words, const char *start,
                         const char *end)
{
  char **new_words;

  if (words->count == -1) {
    /* Previous error */
    return;
  }
  
  words->count++;

  new_words = realloc(words->words, words->count * sizeof(char *));
  if (new_words == NULL) {
    free_wordlist(words);
    return;
  }
  words->words = new_words;

  words->words[words->count-1] = dupn_and_downcase(start, end-start);
  if (words->words[words->count-1] == NULL) {
    free_wordlist(words);
  }

  return;
} /* add_new_word */


/*
 * cmp_strs()
 * Compare two pointers to strings for equality.
 * Case-sensitive comparision since we've already case-canonicalized.
 */
static int cmp_strs(const char **a, const char **b)
{
  return strcompare(*a, *b, 0);
} /* cmp_strs */


/*
 * build_wordlist()
 * Split an input name into a wordlist.  Strip punctuation, downcase, and
 * sort.
 * On error, set 'wordlist' to an error state, i.e. with count == -1.
 */
static wordlist build_wordlist(const char *name)
{
  const char *s, *word_start;
  wordlist ret = {0, NULL};
  int in_word;

  in_word = 0;
  word_start = NULL;

  /* Scan name. */
  for (s = name; *s != '\0'; s++) {
    if (isalnum((int) *s)) {
      if (!in_word) {
        word_start = s;
      }

      in_word = 1;
      continue;
    }
    else if (*s == '\'') {
      /* Ignore.  Doesn't affect in_word. */
      s++;
      /* Make sure we didn't just fall off the end of the string. */
      if (*s == '\0') {
        break;
      }
      else {
        continue;
      }
    }
    else {
      /* Not alnum and not an apostrophe.  A separator. */
      if (in_word) {
        add_new_word(&ret, word_start, s);
      }
      in_word = 0;
    }
  }

  /* If we ended with a word, add it to the word list too. */
  if (in_word) {
    add_new_word(&ret, word_start, s);
  }

  /* Sort the word list */

  if (ret.count != -1) {
    qsort((void *)ret.words, ret.count, sizeof(char *),
          (int(*)(const void*, const void*))cmp_strs);
  }

  return ret;
} /* build_wordlist */


/*
 * dup_wordlist()
 * Duplicate a wordlist.
 */
static wordlist dup_wordlist(const wordlist *oldlist)
{
  wordlist newlist;
  int i;

  if (oldlist->count == -1) {
    newlist.count = -1;
    newlist.words = NULL;
    return newlist;
  }

  newlist.count = oldlist->count;
  newlist.words = calloc(sizeof(char *), newlist.count);
  if (newlist.words == NULL) {
    newlist.count = -1;
    return newlist;
  }

  for (i = 0; i < newlist.count; i++) {
    newlist.words[i] = strdup(oldlist->words[i]);
    if (newlist.words[i] == NULL) {
      free_wordlist(&newlist);
      break;
    }
  }

  return newlist;
} /* dup_wordlist */


/*
 * free_wordlist()
 * Free a wordlist, and reset its values to error.
 */
static void free_wordlist(wordlist *words)
{
  int i;
  if (words->count == -1) {
    return;
  }

  for (i = 0; i < words->count; i++) {
    free_valid(words->words[i]);
  }
  free_valid(words->words);
  words->words = NULL;
  words->count = -1;
  return;
} /* free_wordlist */


/*
 * wordlists_match()
 * Determine if the wordlist 'keywords' is a wordlist-prefix of 'words',
 * as defined in the comment at the top of the file.
 * We know both wordlists are sorted, so we just need to walk through
 * the lists in order.
 * If we got here, we already know one word matched -- we found 'words'
 * for a match_word entry in the hash table.
 */
static int wordlists_match(const wordlist keywords, const wordlist words)
{
  int kpos, wpos;
  int cmp;

  kpos = 0;
  wpos = 0;

  while (kpos < keywords.count) {
    cmp = strncmp(keywords.words[kpos], words.words[wpos],
                  strlen(keywords.words[kpos]));
    if (cmp == 0) {
      /* the words at kpos and wpos matched.  Increment both.
       * If we've exhaused keywords, we matched sucessfully; else if we've
       * exhaused words we've failed.
       */
      kpos++;
      wpos++;
      if (kpos >= keywords.count) {
        return 1;
      }
      else if (wpos >= words.count) {
        return 0;
      }
    }
    else if (cmp < 0) {
      /* The word at kpos is earlier than the word at wpos.  There is therefore
       * no word in words that the word at kpos can match, so fail. */
      return 0;
    }
    else if (cmp > 0) {
      /* The word at kpos is later than the word at wpos.  The word at kpos
       * might match a later word in words, if there are any.  Increment
       * wpos, then fail if we've exhausted words. */ 
      wpos++;
      if (wpos >= words.count) {
        return 0;
      }
    }
  }

  /* If we get get here, we've somehow fallen off keywords without
   * going through the cmp == 0 clause.  Something's wrong. */
  return 0;
} /* wordlists_match */


/*
 * nameparts()
 * Given a name, expressed as a string, of the type 'first middle (...) last',
 * extract out the first, middle, and last name parts.
 * There can be several middle names, and the last name may be of the form
 * 'van de Bruijn', i.e., with lower-case prepositions.
 * 
 * Returns the name parts in fml[], setting entries to "" if missing.  Parts
 * are allocated dynamically and need to be freed by calling routine.
 *
 * Each part is converted to a single all-lower-case word, with punctuation
 * stripped (so that obrien will match O'Brien).
 *
 * If there is only one part, it is considered to be the last name.
 */
static void nameparts(const char *orig_name, DString fml[3])
{
  int n;   /* number of words in name */
  int i;
  char *s, *t;
  char *name;
  char *word[] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
  int last_name = 0;  /* index of last name */

  name = strdup(orig_name); /* make local copy */

  /* Skip leading whitespace */
  for (s = name; *s != '\0' && isspace((int)*s); s++)
    ; /* Do nothing */

  /* split into words at ., _ or white space */
  word[0] = s;
  for (n = 1; *s != '\0'; s++) {
    if (strchr("_. \t", *s)) {
      *s = '\0';
      if ((strlen(s+1) > 0) && (strchr("_. \t", *(s+1)) == 0)) {
        word[n++] = s+1;
      }
    }
  }

  DStringInit(&fml[0]);
  DStringInit(&fml[1]);
  DStringInit(&fml[2]);

  /* 
   * Set last_name to the index of the first word of the last name.  If the
   * last word is capitalized, back up over previous uncapitalized words --
   * we assume they're prepositions which are part of the last name (van,
   * de).
   */
  
  last_name = n - 1; /* n is at least 1 thanks to loop above */

  if (isupper((int)word[last_name][0])) {
    for (i = last_name - 1; i >= 0; i--) {
      if (islower((int)*word[i])) {
        last_name = i;
      }
      else break;
    }
  }

  /* Copy last name(s) to fml[2]. */
  for (i = last_name; i < n; i++) {
    DStringAppend(&fml[2], word[i], -1);
  }

  /* Copy middle name(s) to fml[1]. */
  for (i = 1; i < last_name; i++) {
    DStringAppend(&fml[1], word[i], -1);
  }

  /* Copy first name to fml[0]. */
  if (last_name > 0) {
    DStringAppend(&fml[0], word[0], -1);
  }

  /* Convert all names to lower case, stripping punctuation. */
  for (i = 0; i < 3; i++) {
    if (DStringLength(&fml[i]) == 0)
      continue;
    /* XXX: scribbling on the inside of a DString.  Evil. */
    for (s = t = DStringValue(&fml[i]); *s != '\0'; s++) {
      if (isalnum((int)*s)) {
        *t = tolower((int)*s);
        t++;
      }
    }
    *t = '\0';
    fml[i].length = (t - fml[i].string);
  }

  free(name);
} /* nameparts */


/*
 * prepend_elt()
 * Prepend an entry to a list.  Return false on error.
 */
static int prepend_elt(namemap_list_t **list, namemap_list_t *elt)
{
  elt->next = *list;
  *list = elt;

  return 1;

} /* prepend_elt */


static void free_entry(namemap_entry_t *entry);

/*
 * build_entry()
 * Build a namemap_entry_t.  Return it, or NULL.
 */
static namemap_entry_t* build_entry(const char *username, const char *gecos,
                                    const wordlist *wordlist,
                                    char **errorstr)
{
  namemap_entry_t *entry;

  entry = calloc(1, sizeof(namemap_entry_t));
  if (entry == NULL) {
    return NULL;
  }

  entry->username = strdup(username);
  entry->gecos    = strdup(gecos);
  entry->wordlist = dup_wordlist(wordlist);

  if (entry->username == NULL || entry->gecos == NULL ||
      entry->wordlist.count == -1) {
    free_entry(entry);
    *errorstr = NULL;
    return NULL;
  }
  return entry;
} /* build_entry */


/*
 * free_entry()
 * Free a namemap entry.
 * Note that this takes a non-const argument, so if you're getting this
 * entry out of a namemap_list_t you have to explicitly cast away const.
 */
static void free_entry(namemap_entry_t *entry)
{
  free_valid(entry->username);
  free_valid(entry->gecos);
  free_wordlist(&entry->wordlist);
  free_valid(entry);
} /* free_entry */


/*
 * add_entry()
 * Add an entry to a list.
 * Return false on error.
 */
static int add_entry(namemap_list_t **list, match_type_t type,
                     const namemap_entry_t *entry)
{
  namemap_list_t *elt;

  elt = (namemap_list_t *) calloc(sizeof(namemap_list_t), 1);
  if (elt == NULL) {
    return 0;
  }
  
  elt->type = type;
  elt->entry = entry;

  if (!prepend_elt(list, elt)) {
    free_valid(elt);
    return 0;
  }

  return 1;
} /* add_entry */


/*
 * free_list()
 * free a namemap_list_t list.
 * If check_canon is true, free entries if they're pointed to by
 * a match_username entry.
 */
static void free_list(namemap_list_t **list, int check_canon) 
{
  namemap_list_t *elt, *elt_next;

  for (elt = *list; elt != NULL; elt = elt_next) {
    elt_next = elt->next;
    if (check_canon && elt->type == match_username) {
      /* Cast away const */
      free_entry((namemap_entry_t *)elt->entry);
    }
    free_valid(elt);
  }
  *list = NULL;
} /* free_list */


/*
 * cmp_elts()
 * Compare two (pointers to pointers to) namemap list elements.
 * Comparison in this case is defined as comparison on the duple
 * {entry->username, match_type}.
 * Case-sensitive comparision since we've already case-canonicalized.
 */
static int cmp_elts(const namemap_list_t **a, const namemap_list_t **b)
{
  int ret;

  ret = strcompare((*a)->entry->username, (*b)->entry->username, 0);
  if (ret != 0) return ret;

  if ((*a)->type < (*b)->type) {
    return -1;
  }
  else if ((*a)->type == (*b)->type) {
    return 0;
  }
  else {
    return 1;
  }
} /* cmp_elts */


/*
 * sort_list_by_username_and_matchtype()
 * Sort a list by its username and match type.  Used by
 * make_username_list_unique so it can operate in a single pass.
 * Return false on failure.
 */
static int sort_list_by_username_and_matchtype(namemap_list_t **list)
{
  namemap_list_t **elts;
  namemap_list_t **prev_next;
  namemap_list_t *elt;
  int elts_len;
  int i;

  elts_len = 0;

  for (elt = *list; elt != NULL; elt = elt->next) {
    elts_len++;
  }

  elts = (namemap_list_t **) calloc(elts_len, sizeof(namemap_list_t *));
  if (elts == NULL) {
    return 0;
  }
  
  /* Copy linked list to array */
  for (elt = *list, i = 0;
       elt != NULL && i < elts_len;
       elt = elt->next, i++) {
    elts[i] = elt;
  }

  /* Sort the array */
  qsort((void *)elts, elts_len, sizeof(namemap_list_t *),
        (int(*)(const void*, const void*))cmp_elts);

  /* Convert the array back to a linked list. */
  prev_next = list;
  for (i = 0; i < elts_len; i++) {
    *prev_next = elts[i];
    prev_next = &elts[i]->next;
  }

  *prev_next = NULL;

  free_valid(elts);
  return 1;
} /* sort_list_by_username_and_matchtype */


/*
 * make_username_list_unique()
 * Make each username appear only once in the list.  This is done because
 * an entry for a given user can pop up more than once -- i.e. a match on
 * jonathan.lennox will match "Jonathan Lennox" once for the first name and
 * once for the last.
 * Return false on failure.
 */
static int make_username_list_unique(namemap_list_t **list)
{
  namemap_list_t *elt, *new_next;

  /* Handle the trivial cases trivially.  These should be by far the
   * commonest cases. */
  if (*list == NULL || (*list)->next == NULL) {
    return 1;
  }
  
  if (!sort_list_by_username_and_matchtype(list)) {
    return 0;
  }

  for (elt = *list; elt != NULL; elt = elt->next) {
    if (elt->next != NULL &&
        strcompare(elt->entry->username, elt->next->entry->username, 1) == 0) {
      /* This elt has the same username as the next elt.  Delete the next
       * one. */
      new_next = elt->next->next;
      elt->next->next = NULL;
      free_list(&elt->next, 0);
      elt->next = new_next;
    }
  }

  return 1;
} /* make_username_list_unique */


/* Helper functions for the names hashtable */

/*
 * hash_string()
 * Hash a string.  Use the usual Tcl strategy for string hashes: multiply
 * by 9 and add.
 * XXX this should be located somewhere common in libcine, probably.
 */
static int hash_string(const char *str) {
  int result = 0;

  if (str == NULL) return 0;

  while (*str != '\0') {
    result += (result << 3) + *str;
    str++;
  }

  return result;
} /* hash_string */


/*
 * hash_entry()
 * Hash the key of an entry.
 */
static int hash_entry(const namemap_hash_entry_t *entry)
{
  if (entry == NULL) return 0;
  return hash_string(entry->keyword);
} /* hash_entry */


/*
 * namemap_obj_cmp()
 * libdict passes this function two dict_obj's when it's comparing
 * keys in the hash table
 */
static int namemap_obj_cmp(const namemap_hash_entry_t *entry1,
                           const namemap_hash_entry_t *entry2)
{
  return strcmp(entry1->keyword, entry2->keyword);
} /* namemap_obj_cmp */


/*
 * namemap_key_cmp()
 * libdict passes this function a dict_key and a dict_obj when it's looking
 * up the key in the hash table
 */
static int namemap_key_cmp(const char *key, const namemap_hash_entry_t *entry)
{
  return strcmp(key, entry->keyword);
} /* namemap_key_cmp */


/*
 * init_names_hash()
 * Initialize the 'names' hash table.
 * Return false on failure, filling in errorstr if possible.
 */
static int init_names_hash(char **errorstr)
{
  struct ht_args argsp = {0, 0, 
                          (ht_func) &hash_entry,
                          (ht_func) &hash_string};

  names = ht_create((int (*) (dict_obj, dict_obj)) &namemap_obj_cmp,
                    (int (*) (dict_key, dict_obj)) &namemap_key_cmp,
                    DICT_UNORDERED|DICT_UNIQUE_KEYS|DICT_RETURN_ERROR,
                    &names_errno, &argsp);
  if (names == NULL) {
    seterror(errorstr, "Couldn't create hash table: error %d",
             names_errno);
    return 0;
  }
  return 1;
} /* init_names_hash */



/*
 * insert_elt_in_hash()
 * Insert a list elt in the hash, under the keyword 'keyword'.
 * If there is already an element chain hanging off 'keyword',
 * append this entry to it, otherwise create a new chain.
 * Return false on failure, setting errorstr if possible.
 */
static int insert_elt_in_hash(namemap_list_t *elt, const char *keyword, 
                              char **errorstr)
{
  namemap_hash_entry_t *entry;
  int err;

  entry = (namemap_hash_entry_t *) ht_search(names, (dict_key) keyword);

  if (entry == NULL) {
    entry = calloc(1, sizeof(namemap_hash_entry_t));
    if (entry == NULL) {
      *errorstr = NULL;
      return 0;
    }
    entry->keyword = strdup(keyword);

    if ((err = ht_insert(names, (dict_obj) entry)) != DICT_OK) {
      seterror(errorstr, "insert_elt_in_hash: error %d", err);
      goto fail;
    }
  }

  if (!prepend_elt(&entry->entries, elt)) {
    *errorstr = NULL;
    goto fail;
  }

  return 1;

fail:
  free_valid(entry->keyword);
  free_valid(entry);
  return 0;
} /* insert_elt_in_hash */


/*
 * insert_entry_keyword_in_hash()
 * Insert the entry 'entry' into the hash, with the keyword 'keyword'
 * and match type 'type'.
 * Return false on failure, with errorstr set if possible.
 */
static int insert_entry_keyword_in_hash(const char *orig_keyword,
                                        const namemap_entry_t *entry,
                                        match_type_t type,
                                        char **errorstr)
{
  namemap_list_t *elt;
  char *keyword = NULL;

  elt = (namemap_list_t *) calloc(sizeof(namemap_list_t), 1);
  if (elt == NULL) {
    *errorstr = NULL;
    return 0;
  }

  elt->type  = type;
  elt->entry = entry;  

  keyword = dupn_and_downcase(orig_keyword, -1);

  if (!insert_elt_in_hash(elt, keyword, errorstr)) {
    goto fail;
  }

  free_valid(keyword);
  return 1;
fail:
  free_valid(elt);
  free_valid(keyword);
  return 0;
} /* insert_entry_keyword_in_hash */


/*
 * insert_name_fml_in_hash()
 * Insert the specified entry into the names table, with the
 * keywords 'firstlast', 'flast', 'firstmiddlelast', 'firstmlast', and
 * 'fmlast'. 
 * Return false on failure, filling in errorstr if possible. 
 */
static int insert_name_fml_in_hash(const namemap_entry_t *entry,
                                   const wordlist *wordlist,
                                   char **errorstr)
{
  DString fml[3];
  DString key;
  int i;
  int ret;

  for (i = 0; i < 3; i++) {
    DStringInit(&fml[i]);
  }

  nameparts(entry->gecos, fml);

  DStringInit(&key);

  if (DStringLength(&fml[0]) == 0) {
    /* No first name.  Therefore no middle name either.  Just return. */
    ret = 1;
    goto done;
  }

  /* Build firstlast */
  DStringAppendD(&key, &fml[0]);
  DStringAppendD(&key, &fml[2]);
  if (!insert_entry_keyword_in_hash(DStringValue(&key), entry,
                                    match_firstlast, errorstr)) {
    ret = 0;
    goto done;
  }

  /* Build flast */
  DStringTrunc(&key);
  DStringAppend(&key, DStringValue(&fml[0]), 1); /* First character only */
  DStringAppendD(&key, &fml[2]);
  if (!insert_entry_keyword_in_hash(DStringValue(&key), entry,
                                    match_flast, errorstr)) {
    ret = 0;
    goto done;
  }

  if (DStringLength(&fml[1]) == 0) {
    /* No middle name.  Therefore we're done. */
    ret = 1;
    goto done;
  }

  /* Build firstmiddlelast */
  DStringTrunc(&key);
  DStringAppendD(&key, &fml[0]);
  DStringAppendD(&key, &fml[1]);
  DStringAppendD(&key, &fml[2]);
  if (!insert_entry_keyword_in_hash(DStringValue(&key), entry,
                                    match_firstmiddlelast, errorstr)) {
    ret = 0;
    goto done;
  }

  /* Build firstmlast */
  DStringTrunc(&key);
  DStringAppendD(&key, &fml[0]);
  DStringAppend(&key, DStringValue(&fml[1]), 1);  /* First character only */
  DStringAppendD(&key, &fml[2]);
  if (!insert_entry_keyword_in_hash(DStringValue(&key), entry,
                                    match_firstmlast, errorstr)) {
    ret = 0;
    goto done;
  }

  /* Build fmlast */
  DStringTrunc(&key);
  DStringAppend(&key, DStringValue(&fml[0]), 1); /* First character only */
  DStringAppend(&key, DStringValue(&fml[1]), 1);  /* First character only */
  DStringAppendD(&key, &fml[2]);
  if (!insert_entry_keyword_in_hash(DStringValue(&key), entry,
                                    match_fmlast, errorstr)) {
    ret = 0;
    goto done;
  }

  ret = 1;

done:
  for (i = 0; i < 3; i++) {
    DStringFree(&fml[i]);
  }
  DStringFree(&key);
  return ret;
} /* insert_name_fml_in_hash */


/*
 * insert_name_in_hash()
 * Insert the specified username/gecos into the names table, under all the
 * keys for which it can be looked up.
 * Return false on failure, filling in errorstr if possible.
 */
static int insert_name_in_hash(const char *username, const char *gecos,
                               char **errorstr)
{
  wordlist words;
  int i;
  namemap_entry_t *entry;

  words = build_wordlist(gecos);
  if (words.count == -1) {
    /* This can only happen because of memory allocation failure. */
    *errorstr = NULL;
    return 0;
  }

  entry = build_entry(username, gecos, &words, errorstr);
  if (entry == NULL) {
    goto fail;
  }
   
  /* Insert a user entry based on the username. */
  if (!insert_entry_keyword_in_hash(username, entry,
                                    match_username, errorstr)) {
    /* Must free entry here, since the username entry is the 'canonical'
     * entry in the hash table. */
    free_entry(entry);
    goto fail;
  }

  /* After this point entry must *not* be freed, until the table is. */

  /* Insert a user entry based on the individual words. */
  for (i = 0; i < words.count; i++) {
    if (strlen(words.words[i]) > 1) {
      if (!insert_entry_keyword_in_hash(words.words[i], entry,
                                        match_word, errorstr)) {
        goto fail;
      }
    }
  }

  /* Insert a user entry based on first-middle-last rules. */
  if (!insert_name_fml_in_hash(entry, &words, errorstr)) {
    goto fail;
  }

  free_wordlist(&words);
  return 1;
  
fail:
  free_wordlist(&words);
  return 0;
} /* insert_name_in_hash */


/*
 * fill_names_hash()
 * Fill the 'names' hash table based on the password file.
 * Return false on failure, filling in errorstr if possible.
 */
static int fill_names_hash(char **errorstr)
{
  struct passwd *pw;
  char *gecos, *s;

  setpwent();
  while ((pw = getpwent())) {
    /* Remove any trailing comments and anything after a comma (since
       Doe, John seems to be rarely used in passwd files). */
    gecos = strdup(pw->pw_gecos);
    s = strpbrk(gecos, ",/(");
    if (s) *s = '\0'; 

    if (!insert_name_in_hash(pw->pw_name, gecos, errorstr)) {
      return 0;
    }
    
    free_valid(gecos);
  }
  endpwent();
  return 1;
} /* fill_names_hash */


/*
 * free_names_hash()
 * Free the 'names' hash table.
 */
static void free_names_hash(void)
{
  namemap_hash_entry_t *entry;

  while ((entry = (namemap_hash_entry_t *) ht_minimum(names)) != NULL) {
    ht_delete(names, (dict_obj) entry);
    free_valid(entry->keyword);
    free_list(&entry->entries, 1);
    free_valid(entry);
  }

  ht_destroy(names);

  names = NULL;
} /* free_names_hash */


/*
 * namemapper_search()
 * The key 'name' has been presented to us.  Return all the names it matches.
 * Return false on error.
 */
int namemapper_search(const char *name, namemap_list_t **result)
{
  wordlist keywords;
  char *keyword;
  int i;
  namemap_hash_entry_t *hash_entry;
  namemap_list_t *elt;

  keywords = build_wordlist(name);
  if (keywords.count < 0) {
    return 0;
  }

  if (keywords.count == 0) {
    /* 'name' translated to an empty keywords set. */
    free_wordlist(&keywords);
    return 1;
  }
   
  for (i = 0; i < keywords.count; i++) {
    keyword = keywords.words[i];
    hash_entry = (namemap_hash_entry_t *) ht_search(names, (dict_key) keyword);

    if (hash_entry != NULL) {
      for (elt = hash_entry->entries; elt != NULL; elt = elt->next) {
        /* Non-word matches apply only if a single keyword was given. */
        if (i == 0 && keywords.count == 1) {
          if (elt->type != match_word) {
            add_entry(result, elt->type, elt->entry);
          }
        }
        /* Word matches apply for all keywords, but the wordlists must
         * match */
        if (elt->type == match_word &&
            wordlists_match(keywords, elt->entry->wordlist)) {
          add_entry(result, elt->type, elt->entry);
        }
      }
    }
  }    

  free_wordlist(&keywords);

  if (!make_username_list_unique(result)) {
    return 0;
  }

  return 1;
} /* namemapper_search */


static canonicalize_status translate_list(namemap_list_t *list,
                                          int namemap_status,
                                          uri_t *uri,
                                          canonicalize_result *result);
static int merge_uri(char *name, uri_t *uri);
static char* build_nameaddr(const char *name, const uri_t *uri);


/*
 * init_namemapper()
 * Initialize and fill the namemapper hash table.
 * On error, return non-zero and fill in errorstr if possible.
 */
int init_namemapper(char **errorstr)
{
  if (names != NULL) {
    free_names_hash();
  }

  if (!init_names_hash(errorstr)) {
    return 1;
  }

  if (!fill_names_hash(errorstr)) {
    free_names_hash();
    return 1;
  }

  return 0;
} /* init_namemapper */


/*
 * namemapper()
 * The core canonicalization routine for the namemapper sub-module
 */
canonicalize_status namemapper(uri_t* uri, canonicalize_result* result)
{
  namemap_list_t *namemap_result = NULL;
  int namemap_status;
  canonicalize_status status;

  if (!options.do_namemapper) {
    return CANONICALIZE_NO_TRANSFORMATION;
  }

  if (strcasecmp(uri->scheme, "sip") != 0) {
    return CANONICALIZE_NO_TRANSFORMATION;
  }

  if (uri->user == NULL || uri->user[0] == '\0') {
    return CANONICALIZE_NO_TRANSFORMATION;
  }

  namemap_status = namemapper_search(uri->user, &namemap_result);

  status = translate_list(namemap_result, namemap_status, uri, result);

  free_list(&namemap_result, 0);

  return status;
} /* namemapper */


/*
 * translate_list()
 * Translate the result of namemapper_search() into a canonicalize_result().
 * If any of the matches was a "U" match, accept it as the canonical
 * match.  Otherwise, if we have multiple matches, return a multiple
 * transformation.
 * If the username accepted is the same as the original username, call
 * it no transformation; otherwise it's a unique transformation.
 */
static canonicalize_status translate_list(namemap_list_t *list,
                                          int namemap_status,
                                          uri_t *uri,
                                          canonicalize_result *result)
{
  int count, i;
  namemap_list_t *elt;

  if (list == NULL) {
    return CANONICALIZE_NO_TRANSFORMATION;
  }

  if (list->next == NULL) {
    /* Single match */
    if (merge_uri(list->entry->username, uri)) {
      return CANONICALIZE_UNIQUE_TRANSFORMATION;
    }
    else {
      return CANONICALIZE_NO_TRANSFORMATION;
    }
  }

  count = 0;
  for (elt = list; elt != NULL; elt = elt->next) {
    if (elt->type == match_username) {
      /* Unique "U" match.*/
      if (merge_uri(elt->entry->username, uri)) {
        return CANONICALIZE_UNIQUE_TRANSFORMATION;
      }
      else {
        return CANONICALIZE_NO_TRANSFORMATION;
      }
    }
    count++;
  }

  /* We have multiple matches, none of which are "U". */
  result->count = count;
  result->nameaddrs = calloc(sizeof(char *), count);
  if (result->nameaddrs == NULL) {
    return CANONICALIZE_ERROR;
  }

  for (elt = list, i = 0; elt != NULL; elt = elt->next, i++) {
    uri_t tmp_uri;
    
    tmp_uri = URI_Copy(*uri);
    merge_uri(elt->entry->username, &tmp_uri);

    result->nameaddrs[i] = build_nameaddr(elt->entry->gecos, &tmp_uri);

    URI_Free(&tmp_uri);
  }

  return CANONICALIZE_MULTIPLE_TRANSFORMATION;
}


/*
 * merge_uri()
 * Merge a name into a uri.  If the uri's user field was already
 * identical to the given name, return 0 and do nothing, else return 1.
 */
static int merge_uri(char *name, uri_t *uri)
{
  if (strcmp(uri->user, name) == 0) {
    /* No change */
    return 0;
  }
  else {
    strassign(&uri->user, name);
    return 1;
  }
} /* merge_uri */


/*
 * build_nameaddr()
 * Given a name and address, print the nameaddr to a string.
 * It is the caller's responsibility to free the string.
 *
 * Note: this function is copied from libsip/sip.c; I didn't want to have
 * to link in libsip.
 */
static char* build_nameaddr(const char *name, const uri_t *uri)
{
  DString nameaddr, addrstr;

  DStringInit(&nameaddr);
  DStringInit(&addrstr);

  if (name && *name) {
    /* Display name not null or empty */
    DStringAppend(&nameaddr, "\"", -1);
    DStringAppendQuoted(&nameaddr, name, "\\\"");
    DStringAppend(&nameaddr, "\" ", -1);
  }

  URI_SPrint(uri, &addrstr);

  DStringAppend(&nameaddr, "<", -1);
  DStringAppendD(&nameaddr, &addrstr);
  DStringAppend(&nameaddr, ">", -1);

  DStringFree(&addrstr);

  return DStringValue(&nameaddr);
} /* build_nameaddr */

