/*
 * Location server with whod
 * Usage:
 *   lswhod [-D] [-d directory] [-i interval]
 *
 * Writes files to world-readable per-user sub-directories given by
 * 'location' with file name "user/host", where 'host' is the name of
 * the server lswhod is running on.  If the user is just logged in
 * locally on that host, the file has zero length.  If the user is
 * logged in remotely and/or locally, the file contains the names of the
 * hosts, including 'localhost' for the local (console) login.  This
 * design was chosen since POSIX limits path names to 255 characters,
 * making it difficult to encode the names of remote hosts into the file
 * name.  Per-user subdirectories are chosen to reduce the directory
 * scan time.
 *
 * Updates periodically every 'interval' minutes from who (utmp)
 * database (default:  3).  It only writes one entry per user and server
 * host, with console being favored.
 *
 * The -D flag puts lswhod into daemon mode.
 *
 * Loosely based on earlier work by Frank Oertel and Christian Zahl;
 * modified by Henning Schulzrinne, Columbia University, 1997-07-13
 * 
 * Copyright (c) 1997 by Columbia University; all rights reserved
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>      /* getopt() */
#include <signal.h>
#include <sys/signal.h>
#include <unistd.h>      /* access() */
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#ifdef HAVE_UTMPX_H
#  include <utmpx.h>
#else
#  include <utmp.h>
#endif
#include <limits.h>
#include <sys/utsname.h>
#include "copyright.h"

#ifndef HAVE_GETUTENT
#include "getutent.h"
#endif

#ifndef LSWHOD_DIR
#  define LSWHOD_DIR "/usr/tmp"
#endif
#ifndef LSWHOD_INTERVAL
#  define LSWHOD_INTERVAL 3
#endif

#if defined(HAVE_UTMP_UT_USER) && !defined(HAVE_UTMP_UT_NAME)
#define ut_name ut_user
#endif

typedef struct user_t {
  char *user;    /* user name */
  int console;   /* user is on console during this iteration */
  int present;   /* this user is still here */
  int host_list_modified;  /* host has been added or deleted during iteration */
  struct user_t *next;     /* next user */
  struct host_t {
    char *host;            /* name of remote host */
    int present;           /* iteration host last seen */
    struct host_t *next;   /* next host in list */
  } *host_list;
} user_t;

static user_t *user_list;
static int verbose = 0;
static struct utsname utsname;


/*
* Update internal database. If a user is not in database, add it.
* The key is 'username.localhost|remote' (one for each remote login).
*/
static void lswhod_update(int iteration, char *myhost)
{
#ifdef HAVE_UTMPX_H
  struct utmpx *ut;
#else
  struct utmp  *ut;
#endif
  char fn[1024];   /* file name */
  struct host_t *h, **h_prior, *h_next;
  user_t *u, **u_prior, *u_next;
  FILE *f;

  /* Read all user utmp entries. */
  while (1) {
#ifdef HAVE_UTMPX_H
    if ((ut = getutxent ()) == NULL)
#else
    if ((ut = getutent ()) == NULL)
#endif
      break;
#if HAVE_UTMP_UT_TYPE
    if (ut->ut_type != USER_PROCESS)
      continue;
#endif

    /*
     * Search for user in our internal linked list. If not there, append
     * to end.
     */
    u_prior = &user_list;
    for (u = user_list; u; u = u->next) {
      if (strcmp(u->user, ut->ut_name) == 0) {
        break;
      }
      u_prior = &u->next;
    }
    if (!u) {
      u = calloc(1, sizeof(user_t));
      u->user = strdup(ut->ut_name);
      /* Ensure that a new record yields a file. */
      u->host_list_modified = 1;  
      *u_prior = u;
      if (verbose) printf("%d: new user %s\n", iteration, u->user);
    }
    if (u->present < iteration) u->present = iteration;
    if (strcmp(ut->ut_line, "console") == 0) {
      u->console = iteration;
      if (verbose) printf("%d: console %s\n", iteration, u->user); 
    }
#ifdef HAVE_UTMP_UT_SYSLEN
    if (ut->ut_syslen > 0) {
#else
    if (ut->ut_host[0]) {
#endif
      /* Search for host in hostlist; add if not there. */
      h_prior = &u->host_list;
      for (h = u->host_list; h; h = h->next) {
        if (strcmp(h->host, ut->ut_host) == 0) {
          break;
        }
      }
      if (!h) {
        h = calloc(1, sizeof(struct host_t));
        h->host = strdup(ut->ut_host);
        u->host_list_modified = iteration;
        *h_prior = h;
      }
      h->present = iteration;
      if (verbose) printf("%d: %s from host %s\n", iteration, u->user, h->host);
    }
  }
#ifdef HAVE_UTMPX_H
  endutxent();
#else
  endutent();
#endif

  /*
   * Remove all stale users, i.e., those not updated during this
   * round, from the list and from the file system.
   */
  for (u = user_list; u; u = u_next) {
    u_next = u->next;

    /*
     * Check if any remote hosts in list are no longer in use.
     */
    for (h = u->host_list; h; h = h_next) {
      h_next = h->next;
      if (h->present < iteration) {
        u->host_list_modified = iteration;
        *h_prior = h->next;
        if (h->host) free(h->host);
        free(h);
      }
    }

    /* If user directory does not exist, create it. */
    if (access(u->user, R_OK | X_OK) < 0) {
      mkdir(u->user, 0777);
      if (verbose) printf("Created directory %s\n", u->user);
    }

    /* User was not found during this iteration. */
    sprintf(fn, "%s/%s", u->user, myhost);
    if (u->present < iteration) {
      *u_prior = u->next;
      if (u->user) free(u->user);
      free(u);
      /* Delete file. */
      if (unlink(fn) < 0) {
        perror("unlink update");
      }
    }
    /*
     * If host list has been modified, rewrite file. If also logged in
     * locally, add own host name as first line in file.
     */
    else if (u->host_list_modified == iteration) {
      if ((f = fopen(fn, "w"))) {
        if (u->host_list && u->console) {
          fprintf(f, "%s\n", myhost);
        }
        for (h = u->host_list; h; h = h->next) {
          fprintf(f, "%s\n", h->host);
        }
        fclose(f);
      } else perror(fn);
    }
    u_prior = &u->next;
  }
} /* lswhod_update */


/*
* Remove all entries for this host from file system.
*/
static void lswhod_clean(char *myhost)
{
  DIR *users, *hosts;
  struct dirent *du, *dh;
  char fn[1024];

  if ((users = opendir(".")) == NULL) {
    perror("read user directory");
    return;
  }
  for (du = readdir(users); du != NULL; du = readdir(users)) {
    /* Is this a directory and not . or ..? */
    if (!strcmp(du->d_name, ".") ||
        !strcmp(du->d_name, "..") ||
        !(hosts = opendir(du->d_name))) continue;
    for (dh = readdir(hosts); dh != NULL; dh = readdir(hosts)) {
      if (!strcmp(dh->d_name, myhost)) {
        sprintf(fn, "%s/%s", du->d_name, dh->d_name);
        if (unlink(fn)) {
          perror(fn);
        }
      }
    }
    closedir(hosts);
  }
  closedir(users);
} /* lswhod_clean */


/*
* Signal handler: clean up database when program is interrupted.
*/
static void handler(int sig)
{
  lswhod_clean(utsname.nodename);
  exit(0);
} /* handler */


static void usage(char *progname)
{
  fprintf(stderr, "Usage: %s [-v] [-D] [-d directory] [-i interval]\n",
    progname);
  exit(1);
} /* usage */


/*
* Main program.
*/
int main(int argc, char *argv[])
{
  extern char *optarg;
  extern int optind; 
  char *directory = LSWHOD_DIR;
  double interval = LSWHOD_INTERVAL;  /* minutes */
  int iteration;        /* how many times have we done this? */
  int c;
  int daemon = 0;

  /* Get commandline arguments. */
  while ((c = getopt(argc, argv, "Dd:i:v")) != EOF) {
    switch(c) {
    case 'd':
      directory = optarg;
      break;
    case 'i':
      interval = atof(optarg);
      break;
    case 'D':
      daemon = 1;
      break;
    case 'v':
      verbose = 1;
      break;
    case '?':
      usage(argv[0]);
      break;
    }
  }

  signal(SIGTERM, handler);
  signal(SIGINT,  handler);
  signal(SIGQUIT, handler);
  signal(SIGHUP,  handler);
  uname(&utsname);

  /*
   * Initialize the daemon mode.
   */
  if (daemon) {
    switch (fork()) {
      case -1:
        exit (-1);
      case 0:
        break;
      default:
        exit(0);
    }
    setsid();  /* release controlling terminal */
  }
  if (access(directory, R_OK | X_OK) < 0) {
    if (mkdir(directory, 0777) < 0) {
      perror("Cannot create directory");
      exit(1);
    }
    if (verbose) printf("Created directory %s\n", directory);
  }
  chdir(directory);
  umask(0);
  lswhod_clean(utsname.nodename);
  for (iteration = 1; ; iteration++) {
    if (verbose) printf("Iteration %d\n", iteration);
    lswhod_update(iteration, utsname.nodename);
    sleep((int)(interval * 60));
  }
  return 0;
} /* main */
