/*
* Searches in system password files for names that match the command
* line argument.
*
* Usage:
*
*   namemapper [-v] [-m] name
*
* The -m flag enables matching on middle names.  If either the password
* file entry or the name provided doesn't have a middle name, it is
* ignored.
* 
* Input argument examples:
*   jqd        (as user name)
*   John       (only first name)
*   Doe        (only last name)
*   J.Q.Doe
*   John.Q.Doe
*   John_Doe
*   JohnDoe
*   JDoe
*   JohnQDoe
*
*   All names are case-insensitive.
*
* The password file 'gecos' field can contain strings of the following
* formats:
*   John Doe (some comment)
*   John Doe,comment
*   J. P. Doe
*
* The program prints zero or more strings of the format
*   U|L|F user_name name
*
* The first letter (U, L, F) designates if the match was on user name,
* last name (and first name or middle name), or first name (if only one
* name).
*
* (c) 1997 by Henning Schulzrinne, Columbia University
*/

#include <pwd.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include "copyright.h"

#define VERSION "1.0"

/*
* Extract parts of names of the type 'first middle (...) last', where
* there can be several middle names and the last name may be of the form
* 'van de Bruijn', i.e., with lower-case prepositions.
* 
* Returns name parts in fml[], setting entries to "" if missing.  Parts
* are allocated dynamically and need to be freed by calling routine.
*/
static void nameparts(char *name, char **fml)
{
  int n;   /* number of words in name */
  int len;
  int i;
  char *s;
  char *word[] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
  int last_name = 0;  /* index of last name */

  name = strdup(name); /* make local copy */

  /* split into words at ., _ or white space */
  for (s = name; *s && isspace((int)*s); s++) ;
  word[0] = s;
  for (n = 1; *s; s++) {
    if (strchr("_. \t", *s)) {
      *s = '\0';
      if ((strlen(s+1) > 0) && (strchr("_. \t", *(s+1)) == 0)) {
        word[n++] = s+1;
      }
    }
  }

  /* 
   * Is the last name capitalized and does it have prepositions (van,
   * de)? Copy all parts of last name into fml[2].
   */
  if (n > 0) {
    last_name = n-1;
    len = strlen(word[last_name]);
    if (isupper((int)*word[last_name])) {
      for (i = last_name-1; i >= 0; i--) {
        if (islower((int)*word[i])) {
          last_name = i;
          len += strlen(word[last_name]);
        }
        else break;
      }
    }
  }

  /* Copy last name(s) to fml[2]. */
  fml[2] = malloc(len+1);
  strcpy(fml[2], "");
  for (i = last_name, s = fml[2]; i < n; i++) {
    strcpy(s, word[i]);
    s += strlen(word[i]);
  }

  /* Find and copy middle name(s). */
  for (i = 1, len = 0; i < last_name; i++) {
    len += strlen(word[i]);
  }
  fml[1] = malloc(len+1);
  strcpy(fml[1], "");
  for (i = 1, s = fml[1]; i < last_name; i++) {
    strcpy(s, word[i]);
    s += strlen(word[i]);
  }

  /* First name. */
  fml[0] = strdup((last_name > 0) ? word[0] : "");

  /* Convert all names to lower case. */
  for (i = 0; i < 3; i++) {
    for (s = fml[i]; *s; s++) {
      *s = tolower((int)*s);
    }
  }

  free(name);
} /* nameparts */


static void usage(char *pname)
{
  fprintf(stderr, "%s %s\nUsage: %s [-m] name\n", pname, VERSION, pname);
  exit(1);
} /* usage */


int main(int argc, char *argv[])
{
  struct passwd *pw;
  char *fml_key[3] = {0,0,0}; /* first, middle, last */
  char *fml[3] = {0,0,0};
  char *gecos;     /* current name from user list */
  char *s;
  int first_len;   /* length of first name */
  int middle_len;  /* length of middle name */
  int match_middle = 0;  /* middle name must match, too, if present */
  int verbose = 0;       /* show progress */

  int i;
  extern char *optarg;
  extern int optind;

  /* Parse arguments */
  while ((i = getopt(argc, argv, "mv")) != EOF) {
    switch (i) {
    case 'm':
      match_middle = 1;
      break;
    case 'v':
      verbose = 1;
      break;
    case '?':
      usage(argv[0]);
      break;
    }
  } 
  /* Not enough arguments. */
  if (optind >= argc) {
    usage(argv[0]);
  }

  /* Extract name parts from input; in reverse to get parts right. */
  nameparts(argv[optind], fml_key);
  first_len = strlen(fml_key[0]);
  middle_len = strlen(fml_key[1]);

  /* Get login names. */
  setpwent();
  while ((pw = getpwent())) {

    /* Remove any trailing comments and anything after a comma (since
       Doe, John seems to be rarely used in passwd files). */
    gecos = strdup(pw->pw_gecos);
    s = strpbrk(gecos, ",/(");
    if (s) *s = '\0'; 

    if (verbose) printf("%s\n", gecos);

    /* Extract name parts. */
    nameparts(gecos, fml);

    /* See if equal to user name. */
    if (strcasecmp(fml_key[2], pw->pw_name) == 0) {
      printf("U %s %s\n", pw->pw_name, pw->pw_gecos);
    }
    /* See if name matches: last name must match; if first name, they
       must match to the shorter of the two. We ignore middle names;
       maybe we should have a quality-of-match indication? */
    else if (strcmp(fml_key[2], fml[2]) == 0) {
      int len;

      len = strlen(fml[0]);
      if (len > first_len) len = first_len;
      if (strncmp(fml_key[0], fml[0], len) == 0) {
        if (match_middle) {
          len = strlen(fml[1]);
          if (len > middle_len) len = middle_len;
          if (strncmp(fml_key[1], fml[1], len)) {
            continue; 
          }
        }
        printf("L %s %s\n", pw->pw_name, gecos);
      }
    }
    /* Maybe the name given was really just a first name (joe). */
    else if ((fml_key[0][0] == '\0') && strcmp(fml_key[2], fml[0]) == 0) {
      printf("F %s %s\n", pw->pw_name, gecos);
    }
    /*
	 * If just a single name, maybe it is actually first and last name
     * concatenated.
     */
    else if (fml_key[0][0] == '\0') {
      /* JohnDoe */
      if (strlen(fml_key[2]) == strlen(fml[0]) + strlen(fml[2])) {
        if (!strncasecmp(fml_key[2], fml[0], strlen(fml[0]))
          && !strcasecmp(&fml_key[2][strlen(fml[0])], fml[2])) {
          printf("L %s %s\n", pw->pw_name, gecos);
        }
      }
      /* JohnQDoe */
      else if (strlen(fml_key[2]) == strlen(fml[0]) + strlen(fml[1]) + strlen(fml[2])) {
        if (!strncasecmp(fml_key[2], fml[0], strlen(fml[0]))
          && !strncasecmp(&fml_key[2][strlen(fml[0])], fml[1], strlen(fml[1]))
          && !strcasecmp(&fml_key[2][strlen(fml[0]) + strlen(fml[1])], fml[2])) {
          printf("L %s %s\n", pw->pw_name, gecos);
        }
      }
      /* JQDoe */
      else if (strlen(fml_key[2]) == 1 + strlen(fml[1]) + strlen(fml[2])) {
        if (fml_key[2][0] == fml[0][0]
          && !strncasecmp(&fml_key[2][1], fml[1], strlen(fml[1]))
          && !strcasecmp(&fml_key[2][1 + strlen(fml[1])], fml[2])) {
          printf("L %s %s\n", pw->pw_name, gecos);
        }
      }
      /* JDoe */
      else if (strlen(fml_key[2]) == 1 + strlen(fml[2])) {
        if (fml_key[2][0] == fml[0][0]
          && !strcasecmp(&fml_key[2][1], fml[2])) {
          printf("L %s %s\n", pw->pw_name, gecos);
        }
      }
    }

    /* Free name parts */
    for (i = 0; i < 3; i++) {
      if (fml[i]) {
        free(fml[i]);
        fml[i] = 0;
      }
    }
    free(gecos);
  }
  endpwent();
  for (i = 0; i < 3; i++) {
    if (fml_key[i]) free(fml_key[i]);
  }
  return 0;
} /* main */
