/*
* tracker
*
* Starting from an initial user and host name, the program searches for
* the most recent sightings of the user, making use of finger output
* containing idle time and which host the user logged in from.
*
* Difficulties:
* . Cannot know time zone of remote system, so comparing login times is
*   not possible.
* . Since the remote system does not track the name of the user on the
*   telnet client system, this only works if a user logs in under his
*   own name.
*
* Usage
*   tracker [-v] user@host
*   The -v flag enables tracing output.
*
* Formats understood:
*   On since D on T from H
*   I Idle Time
*
*   Last seen at H on D
*
*   Login       Name               TTY         Idle    When    Where
*   user     Real Name             T           h:m  Day HH:MM  host
*   (...)
*
*   user name Last login D on H from T (H)
*
* Copyright (c) 1997 by Henning Schulzrinne, Columbia University
*/

#ifdef __linux__
#define _GNU_SOURCE /* Needed for strptime declaration from <time.h> */
#endif

#include <stdio.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include "copyright.h"
#define MAX_IDLE 1000000
#define IDLE_UNKNOWN  -1       /* no idle time provided yet */

extern char *finger(int verbose, char *host, char *user);
extern int hostfix(char *domain, char *host);

/* list of hosts */
typedef struct hlist_t {
  int idle_given;         /* number of times this entry had an idle time */
  char *host;             /* current location or where logged in from */
  time_t login_time;
  int idle_time;
  int console;            /* user is on console */
  struct hlist_t *next;
} hlist_t;

/*
* List of hosts where user is currently logged in (idle_time <
* MAX_IDLE and no login_host).
*/
static hlist_t current;

/*
* List of hosts where user might be.
*/
static hlist_t maybe;

int verbose = 0;

/*
* If equal, update. Otherwise, insert at end.
*/
static void hlist_insert_(hlist_t *head, char *host, time_t login_time,
  int idle_time, int console)
{
  hlist_t *h;

  for (h = head; h; h = h->next) {
    /* same host is in list already; update idle time */
    if (h->host && (strcmp(h->host, host) == 0)) {
      if (idle_time == IDLE_UNKNOWN) h->idle_given--;
      else {
        h->idle_given++;
        if (idle_time < h->idle_time || h->idle_time == IDLE_UNKNOWN)
          h->idle_time = idle_time;
      }
      if (login_time > 0 && login_time > h->login_time)
        h->login_time = login_time;
      if (console) h->console = console;
      break;
    }
    else if (h->next == 0) {
      hlist_t *next = h->next;
      /* printf("insert %p %s\n", head, host); */
      h->next = malloc(sizeof(hlist_t));
      h->next->host = strdup(host);
      h->next->login_time = login_time;
      h->next->idle_time  = idle_time;
      h->next->console    = console;
      h->next->next       = next;
      h->next->idle_given = IDLE_UNKNOWN ? 0 : 1;
      break;
    }
  }
} /* hlist_insert_ */


/*
* Insert host into global list.
*/
static void hlist_insert(char *host, char *login_host, time_t login_time,
  int idle_time, int console)
{
  /* User presently there, not just logged in remotely. */
  if ((!login_host || !*login_host) && (idle_time < MAX_IDLE)) {
    hlist_insert_(&current, host, login_time, idle_time, console);
  }

  /* User either was there or is there, but logged in remotely. */
  if (*login_host) {
    hlist_insert_(&maybe, login_host, login_time, idle_time, console);
  }
} /* hlist_insert */


static void hlist_print(hlist_t *hlist)
{
  hlist_t *h;

  for (h = hlist->next; h; h = h->next) {
    printf("%s %c %1.1f %s", h->host, 
      h->console ? 'C' : '-',
      h->idle_given <= 0 ? 0.0 : 
        (h->idle_time == MAX_IDLE ? 1./0. : h->idle_time/60.), 
      ctime(&h->login_time));
  }
} /* hlist_print */


/*
* Extract date (text until "on" or EOL or end-of-string) from
* text. We append the current year if 
*/
static time_t parse_date(char *text)
{
  char *s, *d;
  char time_string[80];
  time_t t = 0;
  struct tm tm;

  d = time_string;
  for (s = text; *s && *s != '\n'; s++) {
    if (strncmp(s, "on", 2) == 0) break;
    *d++ = *s;
  }
  *d = '\0';

     /* Sat May 17 21:39:36 1997 */
  if (strptime(time_string, "%a %b %d %H:%M:%S %Y", &tm) ||
     /* Sat May 17 21:39 */
      strptime(time_string, "%a %b %d %H:%M", &tm) ||
     /* Sep 6, 1996 */
      strptime(time_string, "%b %d, %Y", &tm) ||
     /* Fri 16:10 */
      strptime(time_string, "%a %H:%M", &tm) ||
     /* Jun 6 07:37:35 */
     strptime(time_string, "%b %d %H:%M:%S", &tm)) {
    /* get current year if no year in string */
    if (tm.tm_year == 0 || tm.tm_mon == 0) {
      struct tm *now;

      t = time(0);
      now = localtime(&t);
      if (tm.tm_year == 0) tm.tm_year = now->tm_year;
      if (tm.tm_mon == 0) tm.tm_mon = now->tm_mon;
    }
    t = mktime(&tm);
  }
  else {
    fprintf(stderr, "Illegal date %s\n", time_string);
  }
  return t;
} /* parse_date */


/*
* Parse idle time and convert into seconds.
* format: d day[s] h hour[s] m minute[s] s second[s] Idle Time
*     or  [h:]m or empty
*/
int parse_idle(char *text)
{
  int n;
  int duration = 0;
  int num[2] = {0,0};
  char word[2][80] = {"",""};

  /* See if this is N days, M minutes, etc. */
  n = sscanf(text, "%d %s %d %s", &num[0], word[0], &num[1], word[1]);
  if (n >= 2) {
    int i;

    for (i = 0; i < 2; i++) {
      if (strncmp(word[i], "day", 3) == 0) {
        duration = num[i] * 86400;
      }
      else if (strncmp(word[i], "hour", 4) == 0) {
        duration += num[i] * 3600;
      }
      else if (strncmp(word[i], "minute", 6) == 0) {
        duration += num[i] * 60;
      }
      else if (strncmp(word[i], "second", 6) == 0) {
        duration += num[i];
      }
    }
  }

  if (duration == 0) {
    char *next;

    n = strtol(text, &next, 10);
    if (*next == ':') {  /* 10:06 */
      duration = n * 3600;
      n = strtol(next+1, &next, 10);
      duration += n * 60;
    }
    else if (*next == 'd') { /* 14d */
      duration = n * 86400;
    }
    else if (*next == ' ') { /* 32 */
      duration = n * 60;
    }
  }

  return duration;
} /* parse_idle */


/*
* Login     Name               TTY         Idle    When    Where
* 0         10                 29          41      49      57
* user    Some Name            pts/2       17:  Thu 16:10  wallace
*/
static int parse_tabular1(char *line, char *login_host,
 time_t *login_time, int *idle_time, int *console)
{
  if (!line[0]) return -1;
  else {
    *idle_time = parse_idle(&line[41]);
    *login_time = parse_date(&line[46]);
    scanf(&line[57], "%s", login_host);
    *console = strstr(line, "console") != 0;
  }
  return 3;
} /* parse_tabular1 */


/*
*   User     Real Name         What    Idle  TTY  Host      Console Location
* 0                                   36         47       56   
* hgs      Some Name                  17:11   pc meteor   (193.175.132.42)
*/
static int parse_tabular2(char *line, char *login_host,
  time_t *login_time, int *idle_time, int *console)
{
  if (!line[0]) return -1;
  else {
    *idle_time = parse_idle(&line[36]);
    *login_time = 0;
    sscanf(&line[56], "(%[A-Za-z.0-9])", login_host);
    *console = strstr(line, "console") != 0;
  }
  return 4;
} /* parse_tabular2 */


/*
* Last seen at H on D
* (GNU finger)
*/
static int parse_Lastseen(char *line, char *login_host, 
  time_t *login_time, int *idle_time, int *console)
{
  const char *Lastseenat = "Last seen at ";
  const char *on = " on ";
  char *s;

  if (strncmp(line, Lastseenat, strlen(Lastseenat)) == 0) {
    sscanf(login_host, "%s", line + strlen(Lastseenat));
    if ((s = strstr(line, on))) {
      *login_time = parse_date(s + strlen(on));
    }   
    *idle_time = MAX_IDLE;
    *console = 0;
    return 6;
  }
  else return -1;
} /* parse_Lastseen */


/*
* ??? Last login D on T from H (host)
*/
static int parse_Lastlogin(char *line, char *login_host,
  time_t *login_time, int *idle_time, int *console)
{
  const char *Lastlogin = "Last login ";
  const char *from = "from ";
  char remote[80];
  char *s;

  if ((s = strstr(line, Lastlogin))) {
    strcpy(remote, "");
    *login_time = parse_date(s + strlen(Lastlogin));
    *console = strstr(line, "console") != 0;
    if ((s = strstr(s, from))) {
      sscanf(s + strlen(from), "%s (%[A-Za-z0-9.])", login_host, remote);
      /* the real host is actually in parentheses */
      if (remote[0]) {
        strcpy(login_host, remote);
      }
    }
    *idle_time = MAX_IDLE;
    return 5;
  }
  else return -1;
} /* parse_lastlogin */


/*
* User     Real Name         What    Idle  TTY  Host      Console Location
*/
static int parse_UserReal(char *line, char *login_host,
  time_t *login_time, int *idle_time, int *console)
{
  const char *UserReal = "  User     Real Name         What";
  if (strncmp(line, UserReal, strlen(UserReal)) == 0) return 2;
  return -1;
} /* parse_UserReal */


/*
* Login     Name               TTY         Idle    When    Where
*/
static int parse_LoginName(char *line, char *login_host,
  time_t *login_time, int *idle_time, int *console)
{
  const char *LoginName = "Login     Name               TTY";
  if (strncmp(line, LoginName, strlen(LoginName)) == 0) return 1;
  return -1;
} /* parse_LoginName */


/*
*  On since D on T from H
*  where H is a host name or :0 (ignored); possibly followed by
*   N Idle Time
*  or some other line
*/
static int parse_Onsince(char *line, char *login_host,
  time_t *login_time, int *idle_time, int *console)
{
  const char *Onsince = "On since";
  const char *from = " from ";
  char *s;

  if (strncmp(line, Onsince, strlen(Onsince)) == 0) {
    *login_time = parse_date(&line[strlen(Onsince)]);
    *console = strstr(line, "on console") != 0;
    if ((s = strstr(line, from))) {
      sscanf(s + strlen(from), "%s", login_host);
      /* host name :0 is not very useful */
      if (login_host[0] == ':') strcpy(login_host, "");
    }
    *idle_time = IDLE_UNKNOWN; /* default, may be corrected later */
    return 5;
  }
  else return -1;
} /* parse_Onsince */


/*
* N Idle Time
*/
static int parse_IdleTime(char *line, char *login_host,
  time_t *login_time, int *idle_time, int *console)
{
  const char *IdleTime = "Idle Time";

  if (strstr(line, IdleTime)) {
    *idle_time = parse_idle(line);
    return 6;
  }
  return -1;
} /* parse_IdleTime */


/*
* Parse output of finger command, line by line.
* Parse_* functions return:
*  -1 no match
*   1 Parse_tabular1 header
*   2 Parse_tabular2 header
*   3 Parse_tabular1 
*   4 Parse_tabular2 
*   5 On since, keep state since 'Idle Time' may follow
*   6 one-liners
*  >2 valid host
*/
void parse_finger(char *host, char *text)
{
  char *eol;    /* end of line */
  char *line;   /* current line from finger output */
  time_t login_time;
  int idle_time = 10000000;
  int console = 0;   /* user is on console */
  char login_host[80];
  int state = -1;
  int ((*fnc[])(char *, char *, time_t *, int *, int *)) = {
    parse_Lastlogin,
    parse_Lastseen,
    parse_Onsince,
    parse_UserReal,
    parse_LoginName,
    0
  };

  /* split finger output into lines */
  for (line = text; *line; ) {
    /* mark end of line (CRLF) */
    eol = strchr(line, '\r');
    if (eol) *eol = '\0';

    /* initialize host data */
    if (state != 5) {
      strcpy(login_host, "");
      login_time = 0;
      idle_time  = MAX_IDLE;
      console    = 0;
    }

    if (state == 1 || state == 3) {
      state = parse_tabular1(line, login_host, &login_time, &idle_time,
                             &console);
    }
    else if (state == 2 || state == 4) {
      state = parse_tabular2(line, login_host, &login_time, &idle_time,
                             &console);
    }
    else if (state == 5) {
      state = parse_IdleTime(line, login_host, &login_time, &idle_time,
                             &console);
    }
    /* if not in table, try different line types */
    else {
      int i;

      for (i = 0; fnc[i]; i++) {
        if ((state = (*fnc[i])(line, login_host, &login_time,
                               &idle_time, &console)) >= 0)
          break;
      }
    }

    /* Entry found; fix up login host. */
    if (state > 2) {
      /* Add domain of 'host' argument if no dot in login host. 
       * Unfortunately, if host name already is the domain, we loose, so
       * we try the name and then add the full name.
       */
      hostfix(host, login_host);
      if (verbose) {
        printf("%s: from %s, %s idle %1.5g min., since %s", 
          host, 
          (*login_host) ? login_host : "(local)", 
          console ? "console" : "",
          idle_time == MAX_IDLE ? 1./0. : 
            (idle_time == IDLE_UNKNOWN ? -1 : idle_time/60.), 
          ctime(&login_time));
      }
      hlist_insert(host, login_host, login_time, idle_time, console);
    }
    line = eol+2; /* skip \r\n */
  }
} /* parse_finger */


static void usage(char *pname)
{
  fprintf(stderr, "Usage: %s [-v] user@host\n", pname);
  exit(1);
} /* usage */


int main(int argc, char *argv[])
{
  char *txt = 0;
  hlist_t *h;
  int c;
  extern int optind;
  char *user, *host;

  /* Parse arguments */
  while ((c = getopt(argc, argv, "v")) != EOF) {
    switch (c) {
    case 'v':
      verbose = 1;
      break;
    case '?':
      usage(argv[0]);
      break;
    }
  } 
  /* Not enough arguments. */
  if (optind >= argc) {
    usage(argv[0]);
  }

  /* Split user@host */
  user = argv[optind];
  host = strchr(argv[optind], '@');
  if (!host) usage(argv[0]);
  *host = '\0';
  host++;

  hlist_insert_(&maybe, host, 0, 0, 0);

  for (h = maybe.next; h; h = h->next) {
    txt = finger(1, user, h->host);
    /* Some finger daemon don't understand the /W option. */
    if (!txt) return 1;
    if (strlen(txt) < 10) txt = finger(0, user, h->host); 
    if (!txt) return 1;
    /* might insert more elements */
    parse_finger(h->host, txt);
  }

  /* output where user is (should sort by idle_time...) */
  hlist_print(&current);

  return 0;
} /* main */
