/*
* Internet header parser.
*/

#include <malloc.h>     /* calloc(), free() */
#include <stdio.h>      /* sscanf() */
#include <string.h>     /* strcpy(), strncpy(), strlen() */
#include <stdlib.h>     /* atoi(), atof() */
#include <ctype.h>      /* isspace() */
#include "parser.h"
#include "dstring.h"

/*
***
*** Utility functions. ***
***
*/

/*
* islws()
* Returns 1 if character is LWS (linear white space, tab or space), 0
* otherwise.
*/
static int islws(unsigned char c)
{
  return ((c == '\t') || (c == ' '));
} /* islws */


/*
* iseol()
* Returns 1 if character marks end of line (LF (\n) or CR (\r))
*/
static int iseol(unsigned char c)
{
  return ((c == '\r') || (c == '\n'));
} /* iseol */
 

/*
* strdupn()
* Create a buffer from 'n' characters starting at 's'. Pad with zero and
* remove any embedded CR and LF.
*/
char *strdupn(char *string, int n)
{
  char *buffer = malloc(n+1);
  char *s, *d;
  int i;

  for (s = string, d = buffer, i = 0; *s && i < n; s++, i++) {
    if (!iseol(*s)) *d++ = *s;
  }
  *d = '\0';

  return buffer;
} /* strdupn */


/*
* isunstructured()
* Is this header unstructured (no comma, no parameters)?
*/
static bool isunstructured(char *name)
{
  char *list[] = {
     "date", 
     "from",
     "content-type",
     "referer",
     "server",
     "subject",
     "user-agent", 
     NULL};
  int i;

  for (i = 0; list[i]; i++) {
    if (!strcasecmp(list[i], name)) return true;
  }
  return false;
} /* isunstructured */
  


/*
***
*** Parser ***
***
*/


/*
* RFC822_ParseHeader()
* Parse a single element into top-level value (ignoring comments) and
* parameters, e.g., foo (comment) ;parameter1=x;parameter2="a string".
* If 'all_parameter' is set, the first value is also treated as a parameter.
* Return a pointer to a list of parameters.
*/
parser_parameter_t *RFC822_ParseHeader(char *element,
  char **value, int all_parameter)
{
  unsigned char *s;
  unsigned char *start;
  int level = 0;      /* parentheses in comments */
  enum {H_space, H_comment, H_global, H_semicolon, H_attribute, H_eq, H_value,
    H_quoted_value, H_done} state = H_space, next = H_global;
  DString v, vv;
  parser_parameter_t *p_first = NULL, *p_last = NULL, *p;

  if (all_parameter) next = H_attribute;
  DStringInit(&v);
  for (s = element; state != H_done; s++) {
    switch(state) {
    case H_space:
      if (*s == '(') {
        state = H_comment;
      }
      else if (*s == '\0') state = H_done;
      else if (!isspace(*s)) {
        state = next;
        start = s;
        s--;
      }
      break;

    case H_comment:
      if (*s == '(') {
        level++;
      }
      else if (*s == ')') {
        level--;
        if (level == 0) {
          state = next;
        }
      }
      else if (*s == '\0') state = H_done;
      break;

    /* top-level value */
    case H_global:
      if (*s == ';') {
        state = H_semicolon;
        s--;
      }
      else if (*s == '(') {
        state = H_comment;
        next  = H_global;
        level = 0;
      }
      else if (*s == '\0') state = H_done;
      else if (!isspace(*s)) {
        DStringAppend(&v, s, 1);
      }
      break;

    /* looking for semicolon */
    case H_semicolon:
      if (*s == ';') {
        state = H_space;
        next  = H_attribute;
      }
      else if (*s == '\0') state = H_done;
      break;

    /* name of parameter, followed by equal sign or space */
    case H_attribute:
      if (isspace(*s) || (*s == '=') || (*s == ';') || (*s == '\0')) {
        p = calloc(1, sizeof(parser_parameter_t));
        if (p_last)   p_last->next = p;
        if (!p_first) p_first = p;
        p_last = p;
        p->attribute = strdupn(start, s - start);
        DStringInit(&vv);
        if (*s == ';') state = H_semicolon;
        else {
          state = H_space;
          next  = H_eq;
        }
        s--;
      }
      break;

    /* looking for equal */
    case H_eq:
      if (*s == '=') {
        state = H_space;
        next  = H_value;
      }
      else if (*s == '\0') state = H_done;
      break;

    /* unquoted value */
    case H_value:
      if ((s == start) && (*s == '"')) {
        state = H_quoted_value;
      }
      else if ((*s == ';') || (*s == '\0')) {
        p->value = DStringValue(&vv);
        state = H_semicolon;
        s--;
      }
      else if (!isspace(*s)) {
        DStringAppend(&vv, s, 1);
      }
      break;

    /* value enclosed in quotation marks */ 
    case H_quoted_value:
      if ((*s == '"')) {
        p->value = DStringValue(&vv);
        state = H_semicolon;
      }
      else if (*s == '\0') {
        p->value = DStringValue(&vv);
        state = H_done;
      }
      else if (!isspace(*s)) {
        DStringAppend(&vv, s, 1);
      }
      break;

    case H_done:
      break;
    }
  } 

  /* Return first (global) value. */
  if (value) *value = DStringValue(&v);
  else DStringFree(&v);

  return p_first;
} /* RFC822_ParseHeader */


/*
 * RFC822_Header
 * Find specific header in parsed message.
 */
parser_header_t *RFC822_Header(parser_header_t *h, char *name)
{
  for (; h; h = h->next) {
    if (!strcasecmp(h->name, name)) return h;
  }
  return NULL;
} /* HTTP_Header */


static void element(parser_element_t *e, int unstructured, char *start,
  char *eov)
{
  if (eov) e->element = strdupn(start, eov - start);
  if (unstructured) e->value = e->element;
  else e->parameters = RFC822_ParseHeader(e->element, &e->value, 0);
} /* element */


/*
* RFC822_Parse:
* Extracts header from block and passes it to RFC822_??? for processing.
* Stop at \n\n (LF LF) or \r\n\r\n (CRLF CRLF). Create list of headers.
* Handle case of comma-divided list as well as several headers with same
* name. Deals with continuation lines. Return list of headers.
*/
parser_header_t *RFC822_Parse(char *header, int *length)
{
  int blank = 0;         /* blank lines */
  unsigned char *s;      /* string pointer */
  unsigned char *start;  /* start of block */
  unsigned char *eov; /* end of value */
  parser_header_t  *h_list = NULL, *h, *h_last = NULL;
  parser_element_t *e, *e_last = NULL;
  enum {P_lws, P_continuation, P_name, P_colon, P_unstructured, P_value,
    P_quoted, P_done} state = P_lws, next = P_name;
  bool unstructured = 0;
  
  for (s = header; *s && state != P_done; s++) {
    switch(state) {
    case P_lws:
      if (!islws(*s)) {
        state = next;
        start = s;
        s--;
      }
      break;

    /* Continuation line. */
    case P_continuation:
      if (iseol(*s)) { 
        if (*s == '\n') blank++;
        /* We have reached the end of the header. */
        if (blank == 2) {
          state = P_done;
          element(e, unstructured, start, eov);
        }
      }
      else if (islws(*s)) {
        state = next;  /* this was a continuation line */
        blank = 0;
      }
      /* New header, not a continuation line. */
      else { 
        /* Finish up value. */
        element(e, unstructured, start, eov);
        state = P_lws;
        next  = P_name;
        s--;
        blank = 0;
      }
      break;

    case P_name:
      eov = 0;  /* no value yet */
      if (isspace(*s) || *s == ':') {
        char *name;

        /* Check if name exists already. */
        name = strdupn(start, s - start);
        h = RFC822_Header(h_list, name);
        if (!h) {
          h = calloc(1, sizeof(parser_header_t));
          h->name = name;
          if (h_last) h_last->next = h; 
          if (!h_list) h_list = h;
          h_last = h;
          e_last = NULL;
        }
        /* If name exists, append as element. */
        else {
          free(name);
        }
        unstructured = isunstructured(h->name);
        e = calloc(1, sizeof(parser_element_t));
        if (e_last) e_last->next = e;
        if (!h->elements) h->elements = e;
        e_last = e;

        s--;
        state = P_colon;
      }
      break;

    case P_colon:
      if (*s == ':') {
        state = P_lws;
        next  = unstructured ? P_unstructured : P_value;
      }
      else if (iseol(*s)) {
        s--;
        next  = P_colon;
        state = P_continuation;
      }
      break;

    case P_unstructured:
      if (iseol(*s)) {
        eov = s;
        s--;
        next  = P_unstructured;
        state = P_continuation;
      }
      break;

    case P_value:
      if (iseol(*s)) {
        eov = s;
        s--;
        next  = P_value;
        state = P_continuation;
      }
      else if (*s == '"') {
        state = P_quoted;
      }
      /* Create a new header replica (same name). */
      else if ((*s == ',') || (*(s+1) == '\0')) {
        /* Finish old element. */
        element(e, unstructured, start, eov);

        /* Allocate and link new element. */
        e = calloc(1, sizeof(parser_element_t));
        if (e_last) e_last->next = e;
        if (!h->elements) h->elements = e;
        e_last = e;
      }
      break;

    case P_quoted:
      if (iseol(*s)) {
        s--;
        state = P_continuation;
      }
      else if (*s == '"') {
        state = P_value;
      }
      break;

    case P_done:
      /* We should never get here. */
      break;
    }
  } 

  if (length) *length = (char *)s - header;

  return h_list;
} /* RFC822_Parse */


/*
* RFC822_ParameterFree()
* Free a parameter list.
*/
void RFC822_ParameterFree(parser_parameter_t *plist)
{
  parser_parameter_t *p, *p_next;

  for (p = plist; p; p = p_next) {
    p_next = p->next;
    if (p->attribute) free(p->attribute);
    if (p->value)     free(p->value);
    free(p);
  }
} /* RFC822_ParameterFree */


/*
* RFC822_Free()
* Free a list of headers.
*/
void RFC822_Free(parser_header_t *hlist)
{
  parser_header_t    *h, *h_next;
  parser_element_t   *e, *e_next;

  for (h = hlist; h; h = h_next) {
    h_next = h->next;

    if (h->name) {
      free(h->name);
      h->name = 0;
    }

    for (e = h->elements; e; e = e_next) {
      e_next = e->next;
      if (e->element) free(e->element);
      if (e->value && e->value != e->element) free(e->value);
      RFC822_ParameterFree(e->parameters);
      free(e);
    }
    free(h);
  }
} /* RFC822_Free */


/*
* RFC822_Attribute1()
* Fetch an attribute value from a parameter for a given parameter list. 
* If 'name' is 0, retrieve the value.  If attribute has no value, return
* "".  If the attribute doesn't exist, return 0.
*/
char *RFC822_Attribute1(parser_parameter_t *plist, char *attribute)
{
  parser_parameter_t *p;

  for (p = plist; p; p = p->next) {
    if (!strcasecmp(p->attribute, attribute))
      return p->value ? (char *)p->value : "";
  }
  return NULL;
} /* RFC822_Attribute1 */


/*
 * RFC822_Attribute()
 * Fetch an attribute from a parameter list for a given top-level element.
 * If 'name' is 0, retrieve the value. If attribute has no value, return "".
 */
char *RFC822_Attribute(parser_header_t *h, int index, char *attribute)
{
  parser_element_t   *e;
  int n = 0;

  for (e = h->elements; e && n < index; e = e->next, n++) {
  } 

  if (e) {
    if (attribute) return RFC822_Attribute1(e->parameters, attribute);
    else return e->value;
  }

  return NULL;
} /* RFC822_Attribute */
