#include "parser.h"
#include "parser_http.h"

#include <stdlib.h>  /* atoi() */
#include <string.h>  /* strchr(), strncpy() */
#include <ctype.h>   /* isspace(), isalpha() */


/*
* HTTP_Parse: parse HTTP-like request into basic objects: first line
* components, header, body. Invoke RFC822_Parse to parse header lines.
*/
parser_t HTTP_Parse(char *msg)
{
  parser_t p;
  unsigned char *s;
  bool isResponse = true;
  enum {P_space, P_protocol, P_version, P_status,
        P_reason, P_URI, P_header, P_body, P_done} 
       state = P_space, next = P_protocol;
  unsigned char *start;              /* start of atom */
  int header_length = 0;             /* length of header */

  p.method      = NULL;              /* init the status line */
  p.URI         = NULL;
  p.reason      = NULL;
  p.protocol    = NULL;
  p.version     = NULL;
  p.status      = -1;
  p.headers     = NULL;              /* assume no headers */
  p.body        = NULL;

  for (s = msg; *s && state != P_done; s++) {        
    switch(state) {
    /* Space between atoms. */
    case P_space:
      if (!isspace(*s)) {
        state = next;
        start = s;
        s--;
      }
      break;

    /* protocol (RTSP/1.0) or method (GET) */
    case P_protocol:
      if (*s =='/') {
        p.protocol = strdupn(start, s - start);
        state = P_version;
        start = s+1;
      }
      else if (isspace(*s)) {
        isResponse = false;
        p.method = strdupn(start, s - start);
        state = P_space;
        next  = P_URI;
      }
      else if (!isalpha(*s)) goto syntax;
      break;
      
    /* Version */
    case P_version:
      if (isspace(*s)) {
        p.version = strdupn(start, s - start);
        state = P_space;
        next  = isResponse ? P_status : P_header;
        s--;
      }
      else if (!(*s == '.') && !isdigit(*s)) goto syntax;
      break;
      
    /* status (number) */
    case P_status:
      if (isspace(*s)) {
        p.status  = atoi(start);
        state     = P_space;
        next      = P_reason;
      }
      else if (!isdigit(*s)) goto syntax;
      break;
      
    /* Reason phrase */
    case P_reason:
      if ((*s =='\n') || (*s == '\r')) {
        p.reason = strdupn(start, s - start);
        state = P_space; 
        next  = P_header;
      }
      break;
      
    /* request URI */
    case P_URI:
      if (isspace(*s)) {
        p.URI = strdupn(start, s - start);
        state = P_space; 
        next  = P_protocol;
      }
      break;

    /* remainder of header, until blank line */
    case P_header:
      p.headers = RFC822_Parse(s, &header_length);
      s += header_length - 1;
      state = P_body;
      break;

    /* message body */
    case P_body:
      p.body = strdup(s);  /* rest of string */
      state = P_done;
      break;

    case P_done:
      break;
    }
  }  
  return p;

syntax:
  return p;
} /* HTTP_Parse */


/*
 * HTTP_Free()
 * Free resources allocated.
 */
void HTTP_Free(parser_t *p)
{
  if (p->method)      free(p->method);
  if (p->URI)         free(p->URI);
  if (p->reason)      free(p->reason);
  if (p->protocol)    free(p->protocol);
  if (p->version)     free(p->version);
  if (p->body)        free(p->body);
  RFC822_Free(p->headers);
} /* HTTP_Free */


/* 
 * HTTP_Method
 * Return request method.
 */
char *HTTP_Method(parser_t *p)
{
  return p ? p->method : NULL;
} /* HTTP_Method */


/*
 * HTTP_URI()
 * Return URI.
 * Should be un-escaped here.
 */
char *HTTP_URI(parser_t *p)
{
  return p ? p->URI : NULL;
} /* HTTP_URI */


/*
 * HTTP_Status:
 * Return status code.
 */
int HTTP_Status(parser_t *p)
{
  return p ? p->status : NULL;
} /* HTTP_Status */


/*
 * HTTP_Reason
 * Return reason phrase.
 */
char *HTTP_Reason(parser_t *p)
{
  return p ? p->reason : NULL;
} /* HTTP_Reason */


/*
 * HTTP_Protocol
 * Return protocol name.
 */
char *HTTP_Protocol(parser_t *p)
{
  return p ? p->protocol : NULL;
} /* HTTP_Protocol */


/* 
 * HTTP_Version()
 * Return version number (major.minor form), NULL if none.
 */
char *HTTP_Version (parser_t *p)
{
  return p ? p->version : NULL;
} /* HTTP_Version */


/* 
 * HTTP_ContentLength()
 * Content-Length: 12345
 * Return length of message body; -1 if no content length.
 */
int HTTP_ContentLength(parser_t *p)
{
  parser_header_t *h = RFC822_Header(p->headers, "Content-Length");  
  int len = -1;

  if (h) len = atoi(h->elements->value);
  return len;
} /* HTTP_ContentLength */


/*
 * HTTP_ContentType()
 * Content-Type: base/sub
 * Return message body MIME type and subtype.
 */
http_content_type_t HTTP_ContentType(parser_t *p)
{
  parser_header_t *h = RFC822_Header(p->headers, "Content-Type");
  http_content_type_t c = {"", ""};
  char *slash, *v;
  
  if (h) {
    v = h->elements->value;
    if ((slash = strchr(v, '/'))) {
      if (slash - v < MAX_MIME_TYPE) {
        strncpy(c.type, v, slash - v);
        c.type[slash - v] = '\0';
      }
      if ((strlen(v) - (slash - v)) < MAX_MIME_TYPE) {
        strcpy(c.subtype, slash+1);
      }
    }
  }
  return c;
} /* HTTP_ContentType */


/*
 * HTTP_Server()
 * Server: Apache
 * Return server description, NULL.
 */
http_server_t HTTP_Server(parser_t *p)
{
  parser_header_t *h = RFC822_Header(p->headers, "Server");
  http_server_t s = {""};

  if (h) s.type = h->elements->value;
  return s;
} /* HTTP_Server */


/*
 * HTTP_UserAgent:
 * User-Agent: Netsoft
 * Return client description.
 */
http_agent_t HTTP_UserAgent(parser_t *p)
{
  parser_header_t *h = RFC822_Header(p->headers, "User-Agent");
  http_agent_t a = {""};

  if (h) a.type = h->elements->value;
  return a;
} /* HTTP_UserAgent */
