#include "parser_rtsp.h"

#include <stdlib.h>   /* atoi(), atof(), strtol() */
#include <string.h>   /* strchr() */
#include <stdio.h>    /* sscanf() */
#include <sys/time.h> /* gettimeofday() */

extern struct in_addr host2ip(char *host);

int RTSP_CSeq(parser_t *P)
{
  parser_header_t *h = RFC822_Header(P->headers, "CSeq");  

  return h ? atoi(RFC822_Attribute(h, 0, 0)) : -1;
} /* RTSP_CSeq */


/*
 * RTSP_Range()
 * Range: clock=; time=
 * Currently
 * Return range description.
 */
rtsp_range_t RTSP_Range(parser_t *P)
{
  parser_header_t *h = RFC822_Header(P->headers, "Range");  
  rtsp_range_t r;
  parser_parameter_t *p;
  char *dash;
  char *v;

  r.effective = -1.;
  r.absolute  = 0;
  r.begin     = -1.;
  r.end       = -1.;
  if (h) {
    v = RFC822_Attribute(h, 0, "time");
    if (v) r.effective = strtoutc(v);
    else {
      struct timeval tv;

      gettimeofday(&tv, 0);
      r.effective = tv.tv_sec + tv.tv_usec / 1e6;
    }

    if ((v = RFC822_Attribute(h, 0, 0))) {
      p = RFC822_ParseHeader(v, 0, 1);
      if ((v = RFC822_Attribute1(p, "clock"))) {
        r.absolute = 1;
        r.begin = strtoutc(v);
        if ((dash = strchr(v, '-'))) r.end = strtoutc(dash+1);
      }
      else if ((v = RFC822_Attribute1(p, "npt"))) {
        r.absolute = 0;
        r.begin = strtonpt(v);
        if ((dash = strchr(v, '-'))) r.end = strtonpt(dash+1);
      }
      RFC822_ParameterFree(p);
    }
  }
  return r;
} /* RTSP_Range */


/*
 * RTSP_Scale(): 
 * Scale: 5.1
 * Return play rate description.
 */
float RTSP_Scale(parser_t *p)
{
  parser_header_t *h = RFC822_Header(p->headers, "Scale");
  float f = 1.0;

  if (h) { 
    f = atof(h->elements->value);
  }
  return f;
} /* RTSP_Scale */


/*
 * RTSP_Speed():
 * Speed: 5.1
 * Return delivery rate.
 */
float RTSP_Speed(parser_t *p)
{
  parser_header_t *h = RFC822_Header(p->headers, "Speed");
  float f = 1.0;

  if (h) { 
    f = atof(h->elements->value);
  }
  return f;
} /* RTSP_Scale */


/*
 * RTSP_Session:
 * Session: some string
 * Return session identifier.
 */
rtsp_session_t RTSP_Session(parser_t *p)
{
  parser_header_t *h = RFC822_Header(p->headers, "Session");
  rtsp_session_t s = {0, 0};
  char *v;

  if (h) {
    s.id = h->elements->value; 
    v = RFC822_Attribute(h, 0, "timeout");
    if (v) s.timeout = atoi(v);
  }
  return s;
} /* RTSP_Session */


/*
* RTSP_Timestamp():
* Timestamp: 38383.3 1.03
*/
rtsp_timestamp_t RTSP_Timestamp(parser_t *P)
{
  parser_header_t *h = RFC822_Header(P->headers, "Timestamp");
  rtsp_timestamp_t t = {0,0};
  char *v;

  if (h && (v = RFC822_Attribute(h, 0, 0))) {
    sscanf(v, "%lf %lf", &t.timestamp, &t.delay);
  }

  return t;
} /* RTSP_Timestamp */


/*
 * RTSP_Transport(): 
 * Transport: RTP/AVP;multicast;compressed;ttl=127;mode="PLAY"
 * Returns a transport description.
 */
rtsp_transport_t RTSP_Transport(parser_t *p)
{
  parser_header_t *h = RFC822_Header(p->headers, "Transport");
  char *v;
  rtsp_transport_t r;

  r.protocol           = P_RTP;
  r.profile            = Pr_AVP;
  r.port[0]            = 3458;
  r.port[1]            = 3459;
  r.client_port[0]     = r.port[0];
  r.client_port[1]     = r.port[1];
  r.server_port[0]     = r.port[0];
  r.server_port[1]     = r.port[1];
  r.ttl                = 0;
  r.layers             = 1;
  r.ssrc               = 0;
  r.multicast          = B_FALSE;
  r.interleaved        = 0;
  r.compressed         = 0;
  r.append             = 0;
  r.destination.s_addr = INADDR_ANY;
  r.transport          = T_UDP;

  if (h) {
    v = RFC822_Attribute(h, 0, "port");
    if (v) sscanf(v, "%u-%u", &r.port[0], &r.port[1]);

    v = RFC822_Attribute(h, 0, "client_port");
    if (v) sscanf(v, "%u-%u", &r.client_port[0], &r.client_port[1]);

    v = RFC822_Attribute(h, 0, "server_port");
    if (v) sscanf(v, "%u-%u", &r.server_port[0], &r.server_port[1]);

    v = RFC822_Attribute(h, 0, "ssrc");
    if (v) r.ssrc = strtol(v,NULL,16);

    v = RFC822_Attribute(h, 0, "ttl");
    if (v) r.ttl = atoi(v);

    v = RFC822_Attribute(h, 0, "interleaved");
    if (v) r.interleaved = atoi(v);

    v = RFC822_Attribute(h, 0, "compressed");
    if (v) r.compressed = B_TRUE;

    v = RFC822_Attribute(h, 0, "destination"); 
    if (v) r.destination = host2ip(v);

    v = RFC822_Attribute(h, 0, "multicast");
    if (v) r.multicast = B_TRUE;

    v = RFC822_Attribute(h, 0, "udp/rtp");
    r.transport = v ? T_UDP :T_TCP;
  }
 
  return r;
} /* RTSP_Transport */


/*
* strtonpt()
* 124.45 or 12:05:04.3
* Parse an NPT timestamp.
*/
double strtonpt(char *npt)
{
  int hh = 0, mm = 0, ss = 0;
  int n;              /* number of elements */
  char *decimal;
  double t;           /* converted time */

  if (strchr(npt, ':')) {
    n = sscanf(npt, "%d:%d:%d", &hh, &mm, &ss);
    t = hh*3600 + mm*60 + ss;      
  }
  else {
    n = sscanf(npt, "%d", &ss);
    t = ss;
  }
  if ((decimal = strchr(npt, '.'))) {
    t += atof(decimal);
  }
  return t;
} /* strtonpt */


/*
* strtoutc()
* 19961108T143720.25Z
*/
double strtoutc(char *utc)
{
  double t;
  char *decimal;   /* position of decimal point */
  struct tm tm;

  memset(&tm, 0, sizeof(tm));
  sscanf(utc, "%4d%2d%2dT%2d%2d%2d", &tm.tm_year, &tm.tm_mon,
    &tm.tm_mday, &tm.tm_hour, &tm.tm_min, &tm.tm_sec);
  tm.tm_year -= 1900;
  t = mktime(&tm);

  if ((decimal = strchr(utc, '.'))) {
    t += atof(decimal);
  }

  return t;
} /* strtoutc */
