#include <stdio.h>
#include <thread.h>
#include <sys/errno.h>
#include "queue.h"
#include "rtsp_parser.h"


int
QueGet(QUEUE *pQueue, void **ppMsg)
{
	QMSG	*pQMsg;

	/* acquire semaphore and decrement semaphore count */
	SemaWait(&pQueue->qNotEmpty);

	/* now remove msg from queue head under mutex protection */
	mutex_lock(&pQueue->qMutex);

	pQMsg = RemoveMsgFromQueueHead(pQueue);

 	mutex_unlock(&pQueue->qMutex);

	/* rest of operations are done outside queue CS */

	if (pQueue->qMax != QUE_NO_LIMIT) /* bounded queue */
       	SemaRelease(&pQueue->qNotFull, 1);

	*ppMsg = pQMsg->pMsg;

	free (pQMsg);

	return (0);
 }

int
QuePut(QUEUE *pQueue, void *pMsg)
{
	if (pQueue->qMax != QUE_NO_LIMIT) /* bounded queue */
		SemaWait(&pQueue->qNotFull);

	/* now insert msg at end of queue under mutex protection */
    mutex_lock(&pQueue->qMutex);

	InsertMsgAtQueueTail(pQueue, pMsg);

	/* exit queue CS */
    mutex_unlock(&pQueue->qMutex);

	/* notify consumers */
	SemaRelease(&pQueue->qNotEmpty, 1);

	return(0);
}


int
QueCreate(QUEUE *pQueue, int qLimit)
{

	mutex_init(&pQueue->qMutex, USYNC_THREAD, NULL);

	SemaInit(&pQueue->qNotEmpty, 0); /* none initally available */

	if (qLimit != QUE_NO_LIMIT)
		SemaInit(&pQueue->qNotFull, qLimit); /* upper limit on qmsgs */

	pQueue->qMax = qLimit;
	pQueue->qMsgHead = NULL;
	pQueue->qMsgTail = NULL;

	return (0);
}

int
InsertMsgAtQueueTail(QUEUE *pQueue, void *pMsg)
{
	QMSG	*pQMsg;

	int msglen;

	pQMsg = (QMSG *) calloc(1,sizeof(QMSG));
        if (pQMsg == NULL)
                printf("InsertMsgAtQueueTail(): PANIC>> calloc return NULL\n");
 
        msglen=strlen((char *)pMsg)+1 ;
        pQMsg->pMsg = (char *)calloc(msglen,sizeof(char));
        bzero(pQMsg->pMsg, msglen);
        bcopy((char *)pMsg, pQMsg->pMsg, msglen);

	//pQMsg->pMsg = pMsg; 

	if (pQueue->qMsgHead == NULL)
	{
		pQueue->qMsgHead = pQMsg;
		pQueue->qMsgTail= pQMsg;
	}
	else
	{
		pQueue->qMsgTail->next = pQMsg;
		pQueue->qMsgTail = pQMsg;
	}
	return(0);
}

QMSG *
RemoveMsgFromQueueHead(QUEUE *pQueue)
{
	QMSG	*pQMsg;

	pQMsg = pQueue->qMsgHead;

	if (pQueue->qMsgHead == pQueue->qMsgTail)
		pQueue->qMsgTail = NULL;

	pQueue->qMsgHead = pQueue->qMsgHead->next;

	return(pQMsg);
}


QMSG *
QueModify(QUEUE *pQueue, long Time)
{
http_parser_t        *p;
http_header_t        *h;
rtsp_range_t         range;

QMSG    	     *pQMsg, *tmp;
char		     *request;
 
printf("QueModify(): --inside Time = %ld \n", Time);
// modify the queue under mutex protection 
mutex_lock(&pQueue->qMutex);


if ( pQueue->qMsgHead == NULL ){ //queue is empty
     mutex_unlock(&pQueue->qMutex);
     return(0);
}

pQMsg = pQueue->qMsgHead;
tmp=pQMsg;

while (pQMsg != NULL ) {

   p = HTTP_Parse( pQMsg->pMsg );
   request = HTTP_Request( p );
   if ( strcmp(request,"PLAY") != 0)
        printf("ModifyQueueHead(): Error in the play queue -get\n");

        h = HTTP_Header(p,"range");
 	if ( h ) {
	// printf("QueModify(): The current msg is %s\n", pQMsg->pMsg);
 	range=RTSP_Range(h,p);
        if ( range.begin.isValid && range.end.isValid){ // range is valid
	if ( range.begin.sec > Time ){
	     // discard this msg
	     if ( pQMsg == pQueue->qMsgHead){ // in the first msg of the queue
		  pQueue->qMsgHead = pQueue->qMsgHead->next;
                  tmp=pQueue->qMsgHead;
		  pQMsg=tmp;
	      }
	      else{
	          if ( pQMsg == pQueue->qMsgTail ){ // it is the last msg in the queue
		       pQueue->qMsgTail = tmp;
		       pQMsg = pQMsg->next;	
		  }
		  else {
	             tmp->next=pQMsg->next;
		     pQMsg = pQMsg->next;
		  }
	      }
	      printf("QueModify(): The play-request has been DISCARDED\n");
        } // the msg was discarded
	else{ // it might needed to update the time
	    tmp=pQMsg;
	    pQMsg=pQMsg->next;
	    if ( range.begin.sec <= Time && range.end.sec > Time){
		range.end.sec=Time;
            } 
	} // it might have updated the time range of the request
	} // it was a msg with valid range
     	else // the range is not valid --give error msg
       	    printf("ModifyQueueHead(): Error in the Range.begin or Range.end variable \n");
        }
	else printf("ModifyQueue(): Error in the http_header\n");
        
}// end of while
// release the lock mutex protection --modification is complete
mutex_unlock(&pQueue->qMutex);

return(0);
}


int
QueDestroy(QUEUE *pQueue)
{

;
}

int
QueEmpty(QUEUE *pQueue)
{
	pQueue->qMsgHead = NULL;
        pQueue->qMsgTail = NULL;

        return (0);

}

int 
SemaInit(SEMA *pSema, int semCount)
{
	sema_init(&pSema->sema, semCount, USYNC_THREAD, NULL);
	return(0);
}

int 
SemaWait(SEMA *pSema)
{
	sema_wait(&pSema->sema);
	return (0);
}

int 
SemaRelease(SEMA *pSema, int nRelease)
{
	int	i; 
	
	for (i = 0; i < nRelease; ++i)
		sema_post(&pSema->sema);
		
	return(0);
	
}

int
SemaDestroy(SEMA *pSema)
{
	/* left as an exercise */
	return(-1);
}
