

/* rtp_api.h : RTP API types, structures, and functions that a user
	       of RTP might require

Copyright 1997 Lucent Technologies; all rights reserved
*/

#ifndef _RTP_API_H
#define _RTP_API_H

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

typedef long int context;
typedef long int person;
typedef long int int32;
typedef short int int16;
typedef signed char int8;

typedef unsigned long int u_int32;
typedef unsigned short int u_int16;
typedef unsigned char u_int8;

typedef struct {
	int sock;
} socktype;


typedef struct {
  unsigned int version:2;  /* protocol version */
  unsigned int p:1;        /* padding flag */
  unsigned int x:1;        /* header extension flag */
  unsigned int cc:4;       /* CSRC count */
  unsigned int m:1;        /* marker bit */
  unsigned int pt:7;       /* payload type */
  u_int16 seq;             /* sequence number */
  u_int32 ts;              /* timestamp */
  u_int32 ssrc;            /* synchronization source */
  u_int32 csrc[1];	   /* CSRC list */
} rtp_hdr_t;



typedef struct {           /* RTP Header Extension */
  u_int16 etype;             /* extension type */
  u_int16 elen;              /* extension length */
  char epayload[1];            /* extension payload */
} rtp_hdr_ext;



typedef struct {
  rtp_hdr_t *RTP_header;
  rtp_hdr_ext *RTP_extension;
  char *payload;         /* payload portion of RTP packet */
  long payload_len;	   /* The length of the payload */
} rtp_packet;



typedef struct {                 /* REPORT BLOCK */
    u_int32 ssrc;                /* ssrc identifier */
    unsigned int frac_lost:8;    /* Fraction lost */
    unsigned int cum_lost:24;    /* Cumulative pkts lost */
    u_int32 highest_seqno;       /* highest seqno received */
    u_int32 jitter;              /* interarrival jitter */
    u_int32 lsr;                 /* last SR timestamp */
    u_int32 dlsr;                /* delay since last SR */
  } rtcp_report_block;



typedef struct {          /* SDES item */
  u_int32 ssrc;		  /* SSRC of the member described by the SDES item */
  int chunkno;            /* The index of the SDES chunk */
  unsigned int type:8;    /* SDES item type */
  unsigned int len:8;     /* SDES item length */
  char *description;  /* SDES item description (255 char limit) */
  int offset;         /* Describes the position of the type field
	                 octet within int32: for internal use of
			 iterator */
} rtcp_sdes_item;



typedef struct {         /* BYE component */
  u_int32 ssrccsrc;      /* The ssrc or csrc that is leaving */
} rtcp_bye_block;





typedef struct {                 /* RTCP Packet */
  unsigned int version:2;        /* protocol version */
  unsigned int p:1;              /* padding flag */
  unsigned int rc:5;             /* report block count */
  unsigned int pt:8;             /* payload type */
  u_int16 len;           /* length */

  union {

    struct {                     /* SR specific fields */
      u_int32 ssrc;              /* SSRC of sender */
      u_int32 ntp_stamp_msw;     /* NTP timestamp most sig. word */
      u_int32 ntp_stamp_lsw;     /* NTP timestamp least sig. word */
      u_int32 rtp_stamp;         /* RTP timestamp */
      u_int32 pkt_count;         /* Sender's packet count */
      u_int32 oct_count;         /* Sender's octet count */
      rtcp_report_block rblock[1]; /* report blocks */
    } sr;

    struct {                      /* RR specific fields */
      u_int32 ssrc;               /* SSRC of sender */
      rtcp_report_block rblock[1]; /* report blocks */
    } rr;
    
    struct {                       /* BYE specific fields */
      rtcp_bye_block byeblock[1];  /* bye SSRCs, CSRCs */
	/* We can't place the reason for leaving info in
	   this part of the packet.  That info must be
	   accessed separately */

    } bye;


    struct {                       /* APP specific fields */
      u_int32 ssrccsrc; /* the ssrc or csrc */
      u_int32 name;     /* an ascii name */
      char appdata[1];  /* application dependent data */
    } app;

  } rtcp_hdr_variant;
} rtcp_overlay;



typedef struct {
  rtcp_overlay *overlay;
  union {
    struct {
      char* extension;
    } sr_or_rr;

    struct {
      char* remainder;
    } sdes;

    struct {
      u_int8 reason_length;
      char* reason;
    } bye;
  } non_overlay;
} rtcp_packet; 

struct link;

typedef struct { /* Iterator used to walk through lists of members */
  struct link *cur_link;
  int go_forward; /* Should iterate up or down list */
} member_iterator;


typedef enum {
  NO_FIELD,
  RTP_MI_CNAME,
  RTP_MI_NAME,
  RTP_MI_EMAIL,
  RTP_MI_PHONE,
  RTP_MI_LOC,
  RTP_MI_TOOL,
  RTP_MI_NOTE,
  RTP_MI_IMG,
  RTP_MI_DOOR,
  RTP_MI_SOURCE,
  RTP_MI_PRIV,
  RTP_MI_SSRC,
  
  RTP_MI_SR_NTPTIMESTAMP,
  RTP_MI_SR_RTPTIMESTAMP,
  RTP_MI_SR_PACKETCOUNT,
  RTP_MI_SR_BYTECOUNT,

  RTP_MI_STAT_HIGHESTSEQNUM,
  RTP_MI_STAT_STARTSEQNUM,
  RTP_MI_STAT_NUMBERLOST,

  RTP_MI_RR_NUMBERSENDERS
} memberinfo;




typedef enum {
  RTP_PF_RFC1890,            /* Standard AV */
} rtpprofile;



typedef enum {
  RTP_OK,                       /* No error */
  RTP_CANT_ALLOC_MEM,		/* Memory allocation failed */
  RTP_TOO_MANY_CONTEXTS,	/* Max # of contexts exceeded */
  RTP_UNKNOWN_CONTEXT,          /* Context never created, or destroyed */
  RTP_NOSOCKET,                 /* Socket not open for r/w */
  RTP_CANT_GET_SOCKET,		/* Can't obtain a new socket */
  RTP_CANT_BIND_SOCKET,		/* Can't bind the socket */
  RTP_CANT_SET_SOCKOPT,		/* Unable to set certain socket options */
  RTP_SOCKET_WRITE_FAILURE,	/* Unable to write to a socket */
  RTP_SOCKET_READ_FAILURE,	/* Unable to read from a socket */
  RTP_SOCKET_MISMATCH,		/* Socket does not belong to session */
  RTP_FIXED_WHEN_OPEN,		/* change not allowed during open session */
  RTP_BAD_ADDR,                 /* The address specified is bad */
  RTP_BAD_PORT,                 /* The port specified is bad */
  RTP_CANT_SEND,		/* Sending packets not allowed because
				   destination address was never set */
  RTP_CANT_CLOSE_SESSION,       /* Unable to close the current session */
  RTP_BAD_MI,                   /* No such membership info */
  RTP_BAD_IP_VERSION,           /* Invalid IP version specified */
  RTP_BAD_PROFILE,              /* Invalid rtpprofile specified */
  RTP_SOCKET_EMPTY,             /* no data to receive at socket */
  RTP_PACKET_LOOPBACK,		/* packet sent from local address with
				   SSRC identical to current SSRC or 
				   identical to the previous SSRC
				   (before a collision).  i.e. it
				   is believed the packet looped back */
			
  RTP_INSUFFICIENT_BUFFER,      /* allocated buffer too small */
  RTP_CSRC_LIST_NOT_SET,	/* CSRC list not yet set */
  RTP_CSRC_APPEARS_TWICE,       /* 2 CSRCs on list have same SSRC */
  RTP_BAD_PACKET_FORMAT,        /* Packet has incorrect format */
  RTP_NO_SUCH_PERSON,           /* Person not session member */
  RTP_PERSON_DATA_PROTECTED,	/* Not allowed to alter this field of data
				   for this person */
  RTP_NO_SUCH_SENDER,           /* No such sender in current session */
  RTP_NO_SUCH_SDES,             /* No such SDES field */
  RTP_CANT_USE_ENCRYPTION,	/* Encryption /decryption cannot be used
				   because the encryption / decryption
				   functions were not set. */
  RTP_DECRYPTION_FAILURE,       /* Packet was unsuccessfully decrypted */
  RTP_END_OF_LIST,              /* Called by iterator when no more members
				   exist on the list */
  RTP_CANT_CALL_FUNCTION,	/* Function call not allowed at this
				   time (to prevent recursion that would
				   cause side effects) */
  RTP_SSRC_IN_USE		/* Attempt to create a CSRC with an
				   SSRC value already used in the CSRC
				   list */

} rtperror;


#define RTP_FLAG_NEW_MEMBER        1   /* New Member(s) joined session */
#define RTP_FLAG_NEW_SENDER        2   /* Member (re)started sending */
#define RTP_FLAG_EXPIRED_MEMBER    4   /* Member(s) expired from session */
#define RTP_FLAG_EXPIRED_SENDER    8   /* Member(s) expired as senders */
#define RTP_FLAG_MEMBER_LEAVES     16  /* Member(s) sent BYE packet */
#define RTP_FLAG_OBSERVE_COLLISION 32  /* Collision detected */
#define RTP_FLAG_A_CSRC_COLLIDES   64  /* Member(s) have same SSRC as a
	                                  local member's SSRC or as a 
					  member in local CSRC list */
#define RTP_FLAG_UNIQUE_ID_REMAP   128 /* A member involved in a collision
					  returned to the session with
					  a new SSRC and was initially
					  thought to be a new member,
					  and was assigned another
					  unique ID.  It has just
					  been realized that this
					  is in fact not a new member,
					  and so the API will revert
					  to using the original unique
					  ID for the member. */
#define RTP_FLAG_MEMBER_INFO_CHANGES 256 /* One of a member's SDES fields
					    has changed. */


rtperror RTPCreate(context *the_context);


rtperror  RTPDestroy(context cid);

rtperror RTPOpenConnection(context cid);


rtperror RTPCloseConnection(context cid);


rtperror RTPSend(context cid, int32 tsinc, int8 marker,
   int16 pti, int8 *payload, int len);


rtperror RTPSessionSetAddr(context cid, int32 value);
rtperror RTPSessionGetAddr(context cid, int32 *value);

rtperror RTPSessionSetLocalAddr(context cid, int32 value);
rtperror RTPSessionGetLocalAddr(context cid, int32 *value);

rtperror RTPSessionSetPort(context cid, int16 value);
rtperror RTPSessionGetPort(context cid, int16 *value);

rtperror RTPSessionSetEncryption(context cid, int value);
rtperror RTPSessionGetEncryption(context cid, int *value);
rtperror RTPSessionSetEncryptionFuncs(context cid, 
				 void (*init)(context, void*),
				 void (*encrypt)(context, char*, int, void*),
				 void (*decrypt)(context, char*, int, void*));

rtperror RTPSessionSetTTL(context cid, int8 value);
rtperror RTPSessionGetTTL(context cid, int8 *value);

rtperror RTPSessionSetRTPStampRate(context cid, int32 payload_type,
	int32 usec);
rtperror RTPSessionGetRTPStampRate(context cid, int32 payload_type,
	int32 *usec);

rtperror RTPSessionSetIPVersion(context cid, int32 value);
rtperror RTPSessionGetIPVersion(context cid, int32 *value);

rtperror RTPSessionSetKey(context cid, void* value);
rtperror RTPSessionGetKey(context cid, void** value);

rtperror RTPSessionSetReconsideration(context cid, int32 value);
rtperror RTPSessionGetReconsideration(context cid, int32 *value);

rtperror RTPSessionSetExtension(context cid, rtp_hdr_ext *the_ext);
rtperror RTPSessionGetExtension(context cid, rtp_hdr_ext **the_ext);

rtperror RTPSessionSetBandwidth(context cid, float session_bwidth,
				float rtcp_fraction);
rtperror RTPSessionGetBandwidth(context cid, float *session_bwidth,
				float *rtcp_fraction);

rtperror RTPSessionSetUserInfo(context cid, void* info);
rtperror RTPSessionGetUserInfo(context cid, void** info);

rtperror RTPSessionSetProfile(context cid, rtpprofile pf);
rtperror RTPSessionGetProfile(context cid, rtpprofile *pf);

rtperror RTPSessionSetCSRCList(context cid, int32 *value, int32 size);
rtperror RTPSessionGetCSRCList(context cid, int32 *value, int32 *size);

rtperror RTPSessionGetUniqueIDForCSRC(context cid, int32 ssrc, person *p);

rtperror RTPSessionGetMemberList(context cid, member_iterator *iter);
rtperror RTPSessionGetMemberListLen(context cid, long *the_len);


int RTPValidRTPPacket(char *rtppacket, long len, int32 trykey,
		      char *decodebuffer);
int RTPValidRTCPPacket(char *rtcppacket, long len, int32 trykey,
		       char *decodebuffer);

rtperror RTPSessionGetRTPSocket(context cid, socktype *value);

rtperror RTPSessionGetRTCPSocket(context cid, socktype *value);

rtperror RTPSetNewMemberCallBack(context cid, void (*f)(context, person, unsigned int));

rtperror RTPSetNewSenderCallBack(context cid, void (*f)(context, person, unsigned int));

rtperror RTPSetExpiredMemberCallBack(context cid, void (*f)(context, person, unsigned int));

rtperror RTPSetExpiredSenderCallBack(context cid, void (*f)(context, person, unsigned int));

rtperror RTPSetByeMemberCallBack(context cid, void (*f)(context, person, unsigned int));

rtperror RTPSetChangedMemberInfoCallBack(context cid,
	void (*f)(context, person, memberinfo, char*, char*, unsigned int));

rtperror RTPSetCollidedMemberCallBack(context cid, 
	void (*f)(context, person, person, unsigned int));

rtperror RTPSetRevertingIDCallBack(context cid, 
	void (*f)(context, person, person, void*, unsigned int));

rtperror RTPCurrentMember(context cid, member_iterator *iter,
				person *the_member);
rtperror RTPNextMember(context cid, member_iterator *iter,
				person *the_member);

rtperror RTPReceive(context cid, int socket,
                    char *rtp_pkt_stream, int *len);


rtp_packet RTPGetRTPPacket(char *rtppacket, long pktlen);
int32 RTPPacketGetCSRC(rtp_packet *rtppkt, int csrc_no);

int RTPSplitCompoundRTCP(char *rtcppacket, char *indpkts[], long len);
rtcp_packet RTPGetRTCPPacket(char *rtcppacket);
rtcp_sdes_item InitSDESItemIter(rtcp_packet *rtcpsdes);
rtcp_sdes_item GetNextItem(rtcp_sdes_item *prev_item);


rtcp_report_block RTPGetReportBlock(rtcp_packet *rtcprrsr, int blockno);

rtcp_bye_block RTPGetByeBlock(rtcp_packet *rtcpbye, int blockno);

rtperror RTPMemberInfoSetSDES(context cid, person p, int SDES_field,	
	char *the_info);
rtperror RTPMemberInfoGetSDES(context cid, person p, memberinfo SDES_field,
	char *the_info);


rtperror RTPMemberInfoSetNTP(context cid, person p, 
	struct timeval NTP_time);
	
rtperror RTPMemberInfoGetNTP(context cid, person p,
	 struct timeval *NTP_time);


rtperror RTPMemberInfoSetRTP(context cid, person p, 
		int32 rtp_stamp);
	
rtperror RTPMemberInfoGetRTP(context cid, person p,
	 int32 *RTP_time);


rtperror RTPMemberInfoSetPktCount(context cid, person p, 
		int32 count);
	
rtperror RTPMemberInfoGetPktCount(context cid, person p,
	 int32 *count);


rtperror RTPMemberInfoSetOctCount(context cid, person p, 
		int32 count);
	
rtperror RTPMemberInfoGetOctCount(context cid, person p,
	 int32 *count);


rtperror RTPMemberInfoSetSSRC(context cid, person p, 
		int32 ssrc);
	
rtperror RTPMemberInfoGetSSRC(context cid, person p,
	 int32 *ssrc);


	
rtperror RTPMemberInfoGetSSRCStatListLen(context cid, person p,
	 int *len);


rtperror RTPMemberInfoSetSSRCStatList(context cid, person p, 
		int32 *ssrc_list, int len);
	
rtperror RTPMemberInfoGetSSRCStatList(context cid, person p,
	 int32 *ssrc_list, int *len);


rtperror RTPMemberInfoSetUserInfo(context cid, person p, 
		void* info);
	
rtperror RTPMemberInfoGetUserInfo(context cid, person p,
		void** info);


rtperror RTPSenderInfoGetFracLost(context cid, person p, int32 ssrc,
		int8 *fraclost);


rtperror RTPSenderInfoGetCumLost(context cid, person p, int32 ssrc,
		int32 *cumlost);


rtperror RTPSenderInfoGetHighestSeqNo(context cid, person p, int32 ssrc,
		int32 *seqno);


rtperror RTPSenderInfoGetJitter(context cid, person p, int32 ssrc,
		int32 *jitter);


rtperror RTPSenderInfoGetLSRStamp(context cid, person p, int32 ssrc,
		int32 *lsr);


rtperror RTPSenderInfoGetDLSRStamp(context cid, person p, int32 ssrc,
		int32 *dlsr);


rtperror RTPFindMember(context cid, memberinfo mi, char *sdes_field,
		 person *p);

rtperror RTPRemapPerson(context cid, person *p);

rtperror RTPMostRecentRTPTime(context cid, person p,
			      struct timeval *the_time);
rtperror RTPMostRecentRTCPTime(context cid, person p,
			      struct timeval *the_time);
rtperror RTPMostRecentRTPPerson(context cid, person *p);
rtperror RTPMostRecentRTCPPerson(context cid, person *p);

void RTPSchedule(context cid, int32 opaque, struct timeval *tp);

rtperror RTPExecute(context cid, int32 opaque);



char *RTPStrError();
#endif

