/*
* Columbia University RTSP Server
*
* Author: Maria Papadopouli and Henning Schulzrinne
* Department of Computer Science
* Columbia University  
* Last updated: October 16th, 1997
*/

/*
* We use the following threads:
* - TCP accept, creates TCP connection thread
* - one per TCP connection to receive requests
* - receive UDP requests and enter PLAY, PAUSE onto queue
* - one per stream to get P/P requests from the queue and serve packets
*   created on SETUP; it waits on a condition variable for new requests
*
* Requests are created by the network interface and either queued or
* destroyed when processed.
*/

#include <stdlib.h>
#include <unistd.h>    /* close() */
#include <stdio.h>     /* fopen() */
#include <string.h>    /* strdup() */
#include <pthread.h>   /* pthread_ */

#include "stream.h"

struct {
  int port;      /* RTSP server port */
} config = {
  5540,          /* port */
};

#define MAX_THREADS 100
#define MAX_REQUEST 8192  /* large enough for UDP packet */

extern int errno;
extern int debug_flag;   /* debug level */

FILE  *log;
pthread_t tid[MAX_THREADS];

static request_t *RequestCreate(int type, int socket)
{
  request_t *r = calloc(1, sizeof(request_t));

  r->type   = type;
  r->socket = socket;

  return r;
} /* RequestCreate */


/*
* Setup()
* Create a new stream.
*/
static int Setup(request_t *r)
{
  Stream *s = StreamCreate();

  /* Mark request as belonging to the new stream. */
  r->stream  = s;
  r->session = s->session;

  /* Create socket for outgoing RTP data. */

  return 200;
} /* SetupRequest */


/*
* Pause()
* Insert pause request into request queue.
*/
static int Pause(request_t *r)
{
  Stream *s = r->stream;

  return 200;
} /* PauseRequest */


/*
* Teardown()
* Remove stream; implicit PAUSE.
*/
static int Teardown(request_t *r)
{
  Stream *s = StreamGet(r->session);

  Pause(r);
  StreamFree(r->session);
  return 200;
} /* Teardown */


/*
* Play()
* Add a play request to the queue, sorted by effective date. 
* If the PLAY request has no range and queue is empty, re-start playing
* at the current time.
* Return status.
*/
static int Play(request_t *r)
{
  Stream *s = r->stream;

  /* Lock stream state. */
  pthread_mutex_lock(&s->lock);  

  /* If paused and nothing else pending... */
  if (s->state == S_PAUSE && !s->first) {
    s->state = S_PLAY;
    /* XXX wakeup the player thread */
  }
  /* Empty queue: signal to stream thread. */
  else if (s->first) {
    s->state = S_PLAY;
    s->first = r;
    /* XXX wakeup the player thread */
  }
  /* Non-empty queue: insert according to effective date. */
  else {
    s->state = S_PLAY;
    s->last = r;
    if (!s->first) s->first = r;
  }

  /* Debug: show request queue */
  if (debug_flag) {
    request_t *rr;

    printf("Requests for stream %d\n", s->session);
    for (rr = s->first; rr; rr = rr->next) {
      printf("  %s: %f %f-%f\n", 
        rr->URI, rr->range.effective, rr->range.begin, rr->range.end);
    }
  }

  /* Unlock stream. */
  pthread_mutex_unlock(&s->lock);  

  return 200;
} /* Play */


/*
* ProcessRequest()
* Process RTSP messages according to the request type.
*/
static int ProcessRequest(request_t *r)
{
  int status;              /* error status of request */
  char *method;            /* method of request */
  rtsp_session_t session;

  if (!(method = HTTP_Method(&r->p))) {
    printf("MsgsHandler(): Error in parsing method\n");
    return 0;
  }

  /* Get request information. */
  r->URI     = HTTP_URI(&r->p);
  r->seq     = RTSP_CSeq(&r->p);
  r->version = HTTP_Version(&r->p);

  if (!strcmp(method, "SETUP")) {
    r->transport = RTSP_Transport(&r->p);
    status = Setup(r);
  }
  else {
    session = RTSP_Session(&r->p);
    if (!session.id) {
      Response(r, 454, "No session identifier");
      return -1;
    }
    r->session = atoi(session.id);
    r->stream  = StreamGet(r->session);
    if (!r->stream) {
      Response(r, 454, 0);
      return -1;
    }
    r->range   = RTSP_Range(&r->p);

    if      (!strcmp(method, "PLAY"))     status = Play(r);
    else if (!strcmp(method, "PAUSE"))    status = Pause(r);
    else if (!strcmp(method, "TEARDOWN")) status = Teardown(r);
  }
  Response(r, status, 0);
 
  HTTP_Free(&r->p);
  return 0;
} /* ProcessRequest */


/*
* Process incoming TCP requests and send out responses.
*/
static void *ReceiveTCPRequests(void *arg)
{
  int s = (int)arg;        /* socket for data */
  char msg[2048];          /* XXX: should be extended dynamically */
  int pt = 0;              /* index within message */
  unsigned char c;         /* we can ignore Unicode here */
  enum {Header, CR, CRNL, CRNLCR, NL, ContentLength, Content} state = Header;
  int left = sizeof(msg);  /* bytes left to read */
  request_t *r;            /* request parameters */
  int n;                   /* bytes read */
  int i;

  r = RequestCreate(SOCK_STREAM, s);
  for (;;) {
    /* Wait for request. */
    if ((n = read(s, &msg[pt], left)) < 0) {
      perror("ReceiveTCPRequest:read");
      return NULL;
    }
    if (n > 0) {
      /* Is there an empty line yet? */
      if (state < Content) {
        for (i = pt; i < pt + n; i++) {
          c = msg[i];
          switch (state) {
          case Header:
            if (c == '\r') state = CR;
            else if (c == '\n') state = NL;
            break;
          case CR:
            if (c == '\n') state = CRNL;
            else state = Header;
            break;
          case CRNL:
            if (c == '\r') state = CRNLCR;
            else state = Header;
            break;
          case CRNLCR:
            if (c == '\n') state = ContentLength;
            else state = Header;
            break; 
          case NL:
            if (c == '\n') state = ContentLength;
            else state = Header;
            break; 
          case Content:
          case ContentLength:
            break;
          }
          if (state == ContentLength) break;
        }
        /* Parse to get Content-Length. */ 
        if (state == ContentLength) {
          r->p = HTTP_Parse(msg);
          left = HTTP_ContentLength(&r->p);
          state = Content;
        }
      }
      pt += n;
      left -= n;
    }
    /* We now have a complete message, with header and content. */
    if (n == 0 || left == 0) {
      /* Process message. */
      ProcessRequest(r);

      /* Reset state for next message. */
      state = Header;
      pt    = 0;
      left  = sizeof(msg);
      r = RequestCreate(SOCK_STREAM, s);
    }
  }
  return NULL;
} /* ReceiveTCPRequests */


/*
* TCP RTSP receiver. Creates new thread for each accepted connection.
*/
static void *ReceiveTCP(void *arg)
{
  struct sockaddr_in sin;  /* address for binding and connections */
  int sinlen;              /* length of address */
  int s;                   /* socket for incoming TCP connection requests */
  int t;                   /* new socket */
  int thr_err;             /* thread error condition */
  pthread_t th;            /* id of new thread */

  /* Create TCP listening socket. */
  if ((s = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
    perror("ReceiveTCP:socket");
    return NULL;
  }

  /* Bind to server address. */
  sin.sin_family      = AF_INET;
  sin.sin_port        = htons(config.port);
  sin.sin_addr.s_addr = INADDR_ANY;
  if (bind(s, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
    perror("ReceiveTCP:bind");
    return NULL;
  }
 
  if (listen(s, 5) < 0) {
    perror("ReceiveTCP:listen");
    return NULL;
  }

  /* Block waiting for connection. */
  for (;;) {
    sinlen = sizeof(sin);
    if ((t = accept(s, (struct sockaddr *)&sin, &sinlen)) < 0) {
      perror("ReceiveTCP:accept");
      return NULL;
    }

    /* Create a new thread for each connection. */
    if ((thr_err = pthread_create(&th, NULL, ReceiveTCPRequests, NULL)) != 0) {
      fprintf(stderr, "ReceiveTCP:thread: %s\n", strerror(thr_err));
    }
  }
  return NULL;
} /* ReceiveTCP */


/*
* UDP RTSP receiver: get message from socket and pass to ProcessRTSPRequest().
*/
static void *ReceiveUDP(void *arg)
{
  int sock;                /* socket for incoming requests */
  struct sockaddr_in sin;  /* address of incoming request */
  int n, sinlen;
  request_t *r;            /* incoming request */
  char msg[MAX_REQUEST];   /* incoming message */

  /* Create UDP socket. */
  if ((sock = socket(AF_INET, SOCK_DGRAM, 0) ) < 0) {
    perror("ReceiveUDP:socket");
    return NULL;
  }
 
  memset(&sin, 0, sizeof(sin));
  sin.sin_family      = AF_INET;
  sin.sin_addr.s_addr = htonl(INADDR_ANY);
  sin.sin_port        = htons(config.port);
 
  if (bind(sock, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
    perror("ReceiveUDP(): bind");
    return NULL;
  }
 
  for (;;) {
    r = RequestCreate(SOCK_DGRAM, sock);
    sinlen = sizeof(r->sin);
	if ((n = recvfrom(sock, msg, MAX_REQUEST, 0, (struct sockaddr *)&r->sin,
                      &sinlen)) < 0) {
      perror("ReceiveUDP(): recvfrom");
      break;
    } else {
      debug("ReceiveUDP", "received %s", msg);
      r->p = HTTP_Parse(msg);
      ProcessRequest(r);
    }
    free(msg);
  }
  close(sock);
  
  return NULL;
} /* ReceiveUDP */
 


/*
* Usage information. Exits program with status code 1.
*/
void usage(char *prog)
{
  fprintf(stderr, "Usage: %s [-p port] [-l logfile] [-d]\n", prog);
  exit(1);
} /* usage */


/*
* Main function for server.
*/
int main(int argc, char *argv[])
{
  int i;
  int thr_err;
  extern char *optarg;
  extern int optind;
  int c;
  char *logfile = "rtsp_log";
  extern void ResponseInit(void);

  /* Parse command line options. */
  while ((c = getopt(argc, argv, "dl:p")) != EOF) {
    switch (c) {
      case 'd':
        debug_flag = 1;
        break;
        break;
      case 'l':
        logfile = strdup(optarg);
        break;
      case 'p':
        config.port = atoi(optarg);
        break;
      case '?':
        usage(argv[0]);
        break;
    } 
  }

  debug("", "RTSP server listening on port %d", config.port);

  /* Create log file. */
  if ((log = fopen(logfile, "a")) == NULL) {
    perror("cannot open log file");
    return 1;
  }

  ResponseInit();

  if ((thr_err = pthread_create(&tid[0], NULL, ReceiveUDP, NULL)) != 0) {
    fprintf(stderr, "Cannot create UDP thread: %s\n", strerror(thr_err));
  }
  if ((thr_err = pthread_create(&tid[1], NULL, ReceiveTCP,  NULL)) != 0) {
    fprintf(stderr, "Cannot create TCP thread: %s\n", strerror(thr_err));
  }

  /* Wait for thread completion. */
  for (i = 0; i < MAX_THREADS; i++) {
    if (tid[i]) pthread_join(tid[i], NULL);
  }

  fclose(log);
  return 0;
} /* main */
