/*
* Columbia University RTSP Server
*
* Author: Maria Papadopouli and Henning Schulzrinne
* Department of Computer Science
* Columbia University  
* Last updated: October 16th, 1997
*/

/*
* We have the following threads:
* - TCP accept, creates TCP connection thread
* - one per TCP connection to receive requests
* - receive UDP requests and enter PLAY, PAUSE onto queue
* - one per stream to get P/P requests from the queue and serve packets
*/

#include <stdlib.h>

#include "server.h"
#include "request.h"
#include "parser.h"
#include "parser_http.h"
#include "parser_rtsp.h"

extern void     *rtcp_sender(void *pThrArg);

struct {
  int port;      /* RTSP server port */
} config = {
  6000,          /* port */
};


extern int errno;
extern int debug;   /* debug level */
extern mutex_t  rtp_createMutex;

FILE  *log;
 
static int nThread = 0;
mutex_t    nThreadMutex;
thread_t        threadId[THREADS_NUM];

u_long          inow; // Internal time format (ms)
static struct   timeval   start = { 0, 0 };
struct timeval  now;

/*
* TCP RTSP receiver. Creates new thread for each accepted connection.
*/
static void *ReceiveTCP(void *arg)
{
  struct sockaddr_in sin;  /* address for binding and connections */
  int sinlen;              /* length of address */
  int s;                   /* socket for incoming TCP connection requests */
  int thr_err;             /* thread error condition */
  thread_t th;             /* id of new thread */

  /* Create TCP listening socket. */
  if ((s = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
    perror("ReceiveTCP:socket");
    return -1;
  }

  /* Bind to server address. */
  sin.sin_family      = AF_INET;
  sin.sin_port        = htons(conf.port);
  sin.sin_addr.s_addr = INADDR_ANY;
  if (bind(s, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
    perror("ReceiveTCP:bind");
    return -1;
  }
 
  if (listen(s, 5) < 0) {
    perror("ReceiveTCP:listen");
    return -1;
  }

  /* Block waiting for connection. */
  for (;;) {
    sinlen = sizeof(sin);
    if ((t = accept(s, (struct sockaddr *)&sin, &sinlen)) < 0) {
      perror("ReceiveTCP:accept");
      return -1;
    }

    /* Create a new thread for each connection. */
    if ((thr_err = thr_create(NULL, 0, ReceiveTCPRequests, (void *)t, 0, &th)) != 0) {
      fprintf(stderr, "ReceiveTCP:thread: %s\n", strerror(thr_err));
    }
  }
  return NULL;
} /* ReceiveTCP */


/*
* Process incoming TCP requests and send out responses.
*/
static void *ReceiveTCPRequests(void *arg)
{
  int s = (int)arg;        /* socket for data */
  char msg[2048];          /* XXX: should be extended dynamically */
  int pt = 0;              /* index within message */
  unsigned char c;         /* we can ignore Unicode here */
  enum {Header, CR, CRNL, CRNLCR, NL, ContentLength, Content, EOM} state = Header;
  int content_length = 0;  /* length of message content */
  int left = sizeof(msg);  /* bytes left to read */
  http_header_t *h;        /* parsed header */
  request_t r;             /* request parameters */
  int i;

  r.type   = SOCK_STREAM;
  r.socket = s;
  for (;;) {
    /* Wait for request. */
    if ((n = read(s, &msg[pt], left)) < 0) {
      perror("ReceiveTCPRequest:read");
      return -1;
    }
    if (n > 0) {
      /* Is there an empty line yet? */
      if (state < Content) {
        for (i = pt; i < pt + n; i++) {
          c = msg[i];
          switch (state) {
          case Header:
            if (c == '\r') state = CR;
            else if (c == '\n') state = NL;
            break;
          case CR:
            if (c == '\n') state = CRNL;
            else state = Header;
            break;
          case CRNL:
            if (c == '\r') state = CRNLCR;
            else state = Header;
            break;
          case CRNLCR:
            if (c == '\n') state = ContentLength;
            else state = Header;
            break; 
          case NL:
            if (c == '\n') state = ContentLength;
            else state = Header;
            break; 
          }
          if (state == ContentLength) break;
        }
        /* Parse to get Content-Length. */ 
        if (state == ContentLength) {
          r.p = HTTP_Parse(msg);
          left = HTTP_ContentLength(&r.p, "Content-Length");
          state = Content;
        }
      }
      pt += n;
      left -= n;
    }
    /* We now have a complete message, with header and content. */
    if (n == 0 && left == 0) {
      /* Process message. */
      ProcessRTSPRequest(r);

      /* Reset state for next message. */
      state = Header;
      pt    = 0;
      left  = sizeof(msg);
    }
  }
  return NULL;
} /* ReceiveTCPRequests */


/*
* UDP RTSP receiver: get message from socket and pass to ProcessRTSPRequest().
*/
static void *ReceiveUDP(void *arg)
{
  int s;                   /* socket for incoming UDP requests */
  struct sockaddr_in sin;  /* address of incoming request */
  int n, sinlen;
  char *msg;               /* message buffer (local to function) */
  request_t r;             /* incoming request */

  /* Create UDP socket. */
  if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0) ) < 0) {
    perror("ReceiveUDP:socket");
    return -1;
  }
 
  memset(&sin, 0, sizeof(sin));
  sin.sin_family      = AF_INET;
  sin.sin_addr.s_addr = htonl(INADDR_ANY);
  sin.sin_port        = htons(conf.port);
 
  if (bind(s, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
    perror("ReceiveUDP(): bind");
    return 1;
  }
 
  r.socket = s;
  r.type   = SOCK_DGRAM;
  for (;;) {
    if ((msg = (char *)calloc(MAXMSG, sizeof(char))) == NULL) {
      error("ReceiveUDP", "calloc() failed");
      return 1;
    }

    sinlen = sizeof(r.sin);
	if ((n = recvfrom(sockfd, msg, MAXMSG, 0, (struct sockaddr *)&r.sin,
                      &sinlen)) < 0) {
      perror("ReceiveUDP(): recvfrom");
      break;
    } else {
      debug("ReceiveUDP", "received %s", msg);
      r.p = HTTP_Parse(msg);
      ProcessRTSPRequest(r);
    }
    free(msg);
  }
  close(sockfd);
  
  return NULL;
} /* ReceiveUDP */
 

/*
* ProcessRTSPRequest
* Process RTSP messages according to the request type.
*/
static int ProcessRTSPRequest(request_t r)
{
  int status;
  char *method;
  rtsp_session_t session;
  int id;
  int seq_num;
      
  if (!(method = HTTP_Method(r.p))) {
    printf("MsgsHandler(): Error in parsing method\n");
    return 0;
  }

  seq = RTSP_CSeq(p);
  if (!strcmp(method, "SETUP")) {
    r.transport = RTSP_Transport(&r.p);
    status = SetupRequest(r);
  }
  else {
    session = RTSP_Session(&r.p);
    if (session) {
      Response(r, 454, "No session identifier");
      return -1;
    }
    r.session = atoi(session.id);

    if (!strcmp(request, "PAUSE")) {
      debug("MsgHandler", "PAUSE for session %d", r.session);
      status = PauseRequest(r);
    }
    else if (!strcmp(request, "TEARDOWN")) {
      debug("MsgHandler", "TEARDOWN for session %d", r.session);
      status = TeardownRequest(r);
    }
    else if (!strcmp(request, "PLAY")) {
      debug("MsgHandler", "PLAY for session %d", r.session);
      status = PlayRequest(r);
    }
    else {
	  debug("MsgHandler", "Unknown request %s for session %d", method,
	    r.session);
    }
  }

  if (status < 0) {
  }
 
  HTTP_Free(&r.p);
  return 0;
} /* ProcessRTSPRequest */


/*
* current request -- is the request that is currently served.
*/
void *StreamControlPause(void *argid)
{
  int    s,id;
  http_parser_t  *p;
  http_header_t   *h;
  rtsp_range_t    range;
  char    *pause_req;

  id=atoi((char *)argid);

  if ( (s=GetStream( id )) == INVALID_STREAM ){
    printf("StreamControlPause(%d): Error >>NOT valid Stream Id !\n",id );
    return(0);
  }
 

  for ( ; ; ) {


    QueGet(&Stream[s].PauseReqQue, (void **) &pause_req);

    printf("StreamControlPause(%d): Process the Pause Req (NPT=%ld)\n",Stream[s].id, Stream[s].NPT); 
    p=HTTP_Parse(pause_req);
    if ( p == NULL ){
      printf("StreamControlPause(): Error >>HTTP_Parse() return is NULL\n");
      printf("The msg was %s\n",pause_req);
      return(0);
    }
 


    h=HTTP_Header(p,"range");

    if ( h ) {
      range = RTSP_Range(h,p);

      /* If a pause time is specified, it is range.begin. */
      if (range.begin.isValid) {

  printf("StreamControlPause(%d): Pause time for the Stream is \t%d:%d\n",id ,range.begin.sec,range.begin.microsec);

  mutex_lock(&Stream[s].nptMutex);
  Stream[s].PauseTime = range.begin.sec;
  Stream[s].PauseSet = 1;
  mutex_unlock(&Stream[s].nptMutex);

      }
      else // invalid pause header
  printf("StreamControlPause(%d): Error>> NOT valid Pause-Range in Request!\n",Stream[s].id);
    } // valid header-range
    else {  /* no header */
      mutex_lock(&Stream[s].nptMutex);
      Stream[s].PauseTime = Stream[s].NPT;
      Stream[s].PauseSet = 1;
      mutex_unlock(&Stream[s].nptMutex);
    }

    free_http_parser_t(p);
  }

  return(NULL);
} /* StreamControlPause() */


/*
* 
*/
void * StreamControlPlay(void *argid)
{
  int  s,id; 
  char  *play_req;
  http_parser_t        *p;
  http_header_t *h;
  rtsp_range_t range;

  char            *request;
  char    *file, *uri;

 
  id = atoi((char *)argid);


  if ( (s=GetStream( id )) == INVALID_STREAM ){
    printf("StreamControlPlay(%d): Error >>NOT valid stream id !\n", id);
    return(0);
  }
  
  printf("I AM IN STHE StreamControlPlay()\n");
  // Process the Requests that are pending in the Request Queue
  for ( ; ; ) {

    // first check if TEARDOWN REQUEST IS RECEIVED and if it is, STOP THE DELIVERY
    mutex_lock( &Stream[s].stateMutex);
  if ( Stream[s].state == TEARDOWN_STATE){
       printf("StreamControlPlay(%d): TEARDOWN REQ RECEIVED>> stop the data delivery for stream \n",Stream[s].id);
       mutex_unlock(&Stream[s].stateMutex);
       break;
  }
  mutex_unlock( &Stream[s].stateMutex);

  QueGet( &Stream[s].PlayReqQue, (void **) &play_req );
  printf("StreamControlPlay(): get the play_req from the queue %s\n", play_req);
  p = HTTP_Parse( play_req );
  if ( p == NULL ){
      printf("StreamControlPlay(): Error >>HTTP_Parse() return is NULL\n");
      printf("The msg was %s\n",play_req);
      return(0);
    }

  request = HTTP_Request( p );
  if ( request == NULL){
      printf("StreamControlPlay(): Error >>HTTP_Request() unexpected return of NULL\n");
      return(0);
  }
  
    // It is Pause Req
    if (strcmp(request,"PAUSE") == 0) 
      // You basically discard it, since the "actual" process 
      // of the Pause Req is done by the PauseReqIssued() and
      // the StreamControlPause() threads
      // the ControlPauseReq() is used more as a "flag" that
      // helps in modifying or discarding the Play Req that
      // have been received before that Pause Req 
      ControlPauseReq(s);

    else{
      // It is Play Req
      printf("StreamControlPlay( %d ): Play Req. in the queue!\n", id);
      h = HTTP_Header(p,"range");
      uri = HTTP_URI(p);
      file = AbsPath(uri);

      // determine the Range of the Play Req
      // check if the Play Req specifies the Range
      if ( h ) {
   range=RTSP_Range(h,p);

    // check the Range of the Play Req
   if ( !range.begin.isValid) {
     mutex_lock(&Stream[s].nptMutex);
     range.begin.sec=Stream[s].NPT;
     mutex_unlock(&Stream[s].nptMutex);
     range.begin.microsec=0;
    }
          if ( !range.end.isValid) {
      range.end.sec=END_OF_PLAYBACK;
      range.end.microsec=0;
           }
      }
      else{
     mutex_lock(&Stream[s].nptMutex);
           range.begin.sec=Stream[s].NPT;
           mutex_unlock(&Stream[s].nptMutex);

           range.begin.microsec=0;
           range.end.sec=END_OF_PLAYBACK;
           range.end.microsec=0;
       }
       // Now the Range of the Play Req is set
       // Process the Play Req : it is ready for retrieval 

       ProcPlayReq(s, file, range.begin.sec, range.end.sec);

    }// it was a play request to be processed
    free_http_parser_t(p);
  }

  // Close the connection of stream s
  CloseConnection(s);
 
 // Release the resources that were allocated for stream s
  ReleaseResources(s);

  return(NULL);
} /* StreamControlPlay() */


/*
* ProcPlayReq() modifies or discards the PlayReq that have been
* received before a Pause Req that now is pending.
*/
void ProcPlayReq(int s, char *file, long begin, long end)
{
  struct timeval tvorig,tvend;
  mutex_t mp;

  printf("Playback for %d from %ld(sec) to %ld(sec)\n",Stream[s].id, begin, end);
  mutex_lock(&Stream[s].stateMutex);

  // test if after this play request a Pause Req. has been received
  if ( Stream[s].PauseFlag >0 ) { 
    mutex_unlock(&Stream[s].stateMutex);
 
    mutex_lock(&Stream[s].nptMutex);
    if (begin > Stream[s].PauseTime ) {
      mutex_unlock(&Stream[s].nptMutex);
      printf("StreamControlPlay(%d): !!!! Discard the req with Range (%ld,%ld)\n", Stream[s].id, begin, end);
    }// msg discarded
    else { 
      // check if a modification in the range of the Play Req. is needed
      if ( end > Stream[s].PauseTime ){
  // mutex_unlock(&Stream[s].nptMutex);
  printf("StreamControlPlay(%d): !!!! Modify the req with range (%ld,%ld) to (%ld,%ld)\n",Stream[s].id, begin, end, begin, Stream[s].PauseTime);
  end = Stream[s].PauseTime;
      }
 
      mutex_unlock(&Stream[s].nptMutex);
      gettimeofday(&tvorig, (struct timezone *)NULL);
      RetrieveAndSend(s,file, begin, end);
      gettimeofday(&tvend, (struct timezone *)NULL);
      printf("Retrieval/Sent(%d) completed/paused!!! : the duration was  %ld(sec) %ld(usec) \n",Stream[s].id, tvend.tv_sec-tvorig.tv_sec,tvend.tv_usec-tvorig.tv_usec)
  ;
    } // end of modification
  }// end of the case that a Pause Req. has been received after that Play Req.

  // if no Pause Req. has been received after this Play Req. 
  else{ 
    mutex_unlock(&Stream[s].stateMutex);
    printf("StreamControlPlay(%d): The PlayReq. will be retrieved *without* any modification\n", Stream[s].id);

    gettimeofday(&tvorig, (struct timezone *)NULL);
    RetrieveAndSend(s,file, begin, end);
    gettimeofday(&tvend, (struct timezone *)NULL);

    printf("Retrieval/Sent(%d) completed/paused!!! : of Duration %ld(sec) %ld(usec)\n",Stream[s].id, tvend.tv_sec-tvorig.tv_sec,tvend.tv_usec-tvorig.tv_usec);
  }// end of the case that no Paue Req. has been received after that Play Req.
}


// decrease the Pause Flag that is used as a "marker" for the Play Req.
// that have been received before that Pause Req. and need to be 
// processed accordingly --by modifying their Range (end.sec) or
// discard them.
// the Pause Flag need to be integers, because more than one Pause Req.
// might have been accepted and pending in the Queue.
// However, the modification in the Range is done in respect with the
// PauseTime --there is *only one* PauseTime, and that is the most recent
// Pause Req. that is accepted.
void ControlPauseReq(int s)
{
  mutex_lock(&Stream[s].stateMutex);
  Stream[s].PauseFlag--;
  mutex_unlock(&Stream[s].stateMutex);
}


/*
* Handle PLAY requests for session.
*/
void PlayRequest(request_t r)
{
  int s, reqlen;
  http_parser_t *p;
  char *request, *play_req;

  if (INVALID_STREAM == (s = GetStream(session))) {
    printf("PlayReqIssued(%d): Error >>NOT a valid stream !\n",id);
    return(-1);
  }

  p = HTTP_Parse(req);
  if (p==NULL) {
    printf("PlayReqIssued(%d): Error >>HTTP_Parse() return is NULL\n",id);
    printf("the msg was %s\n",req);
    return(0);
  }

  request = HTTP_Request(p);
  if (request == NULL){
    printf("PlayReqIssued(%d): Error >>HTTP_Request() unexpected return NULL\n",id);
    return(0);
  }

  if (strcmp(request,"PLAY") != 0)
    printf("PlayReqIssued(): Error>> This Req is NOT a Play Req!\n");

  // reqlen=strlen(req);
  // play_req=(char *)calloc(reqlen+1,sizeof(char));
  // if (play_req == NULL){
  // printf("PlayReqIssued(): Error>> Problem in calloc()\n");
  // return(0);
  // }
  // strcpy(play_req,req);

  mutex_lock(&Stream[s].stateMutex);
  if ( Stream[s].state == INVALID_STATE ){
    printf("PlayReqIssued() : Error>> INVALID STATE\n");
    mutex_unlock(&Stream[s].stateMutex);
    return(0);
  }

  mutex_unlock(&Stream[s].stateMutex);

  mutex_lock(&Stream[s].nptMutex);
  if ( Stream[s].PauseSet == 1 ) {
    Stream[s].PauseSet = 0;
    printf("PlayReqIssued(%d): Stream was in PAUSE-STATE, but NOW is back to PLAY-STATE !!!!!\n",id);
  }
  mutex_unlock(&Stream[s].nptMutex);


  QuePut(&Stream[s].PlayReqQue, req);
  printf("PlayReqIssued(%d): Queue the Play Req ...\n",id);

  free_http_parser_t( p );
}


PauseReqIssued(u_long id, char *req)
{
  int      s, reqlen;
  char     *request, *pause_req;
  http_parser_t        *p;
  mutex_t      mp;
 
  if ( INVALID_STREAM == (s=GetStream(id)) ){
    printf("PauseReqIssued(%d): Error >>NOT a valid Stream Id !\n",id);
    return(-1);
  }

  mutex_lock(&Stream[s].stateMutex);
  if ( Stream[s].state == INVALID_STATE ){
    printf("StreamControlPause(%d) : Error >>INVALID STATE\n",id);
    mutex_unlock(&Stream[s].stateMutex);
    return(0);
  }
  mutex_unlock(&Stream[s].stateMutex);

  p = HTTP_Parse( req );
  if ( p == NULL){
    printf("PauseReqIssued(%d): Error >>HTTP_Parse(): uexpected return NULL\n",id);
    return(0);
  }

  request = HTTP_Request( p );
  if ( request == NULL){
    printf("PauseReqIssued(%d): Error >>HTTP_Request(): uexpected return NULL\n",id);
    return(0);
  }

  if (strcmp(request,"PAUSE") != 0)
    printf("PauseReqIssued(%d): Error >>This Req is NOT a Pause Req!\n",Stream[s].id);

  reqlen=strlen(req);
  // pause_req=(char *)calloc(reqlen+1,sizeof(char));
  //if (pause_req == NULL){
  //        printf("PauseReqIssued(): Error>> Problem in calloc()\n");
  //        return(0);
  // }

  //strcpy(pause_req, req);

  mutex_lock(&Stream[s].stateMutex);
  Stream[s].PauseFlag++;
  mutex_unlock(&Stream[s].stateMutex);

  QuePut(&Stream[s].PauseReqQue, req);
  QuePut(&Stream[s].PlayReqQue,  req);
  printf("PauseReqIssued(%d): Queue Pause Req in the Pause-Queue...(NPT=%ld)\n",Stream[s].id, Stream[s].NPT);
  printf("PauseReqIssued(%d): Queue Pause Req in the Play-Queue...\n",Stream[s].id);

  free_http_parser_t( p );
}


/*
* SetupReqIssued
* Handle SETUP requests.
* Initialize the resources for the new stream.
*/
void SetupReqIssued(int sockfd, struct sockaddr cli_addr, int port,
  int cli_len, int seq_num)
{
  u_long  id;
  char  *argid;
  int  s;
  int  init_err, thr_err1, thr_err2, conn_err;

  // initialize a thread to serve it

  printf("SetupReqIssued(): Setup Req Received! :\n");
  printf("The port number is %d\n",port);

  // assign a new id;
  id = ScheduleStream();

  argid=(char *)malloc(sizeof(char)*100);
  if (argid == NULL){
    printf("SetupReqIssued(): Error>> Problem in malloc()\n");
    RemoveStream(id);
    return(0);
  }
  sprintf(argid,"%d",id);
  
  printf("SetupReqIssued: suggested id for the stream = %d (argid = %s) \n",id,argid);
  if ( INVALID_STREAM == (s=GetStream(id)) ){
    printf("SetupReqIssued(): Error >> NOT valid Stream id!\n");
    printf("SetupReqIssued(): UNSUCCESFULL SETUP for REQ = %d !\n", id);
    RemoveStream(id);
    return(0);
  }

  mutex_init(&Stream[s].mp ,USYNC_THREAD,0);
  mutex_init(&Stream[s].retrievMutex ,USYNC_THREAD,0);
  mutex_init(&Stream[s].nptMutex ,USYNC_THREAD,0);
  mutex_init(&Stream[s].rtp_sendMutex ,USYNC_THREAD,0);

  init_err = cond_init(&Stream[s].cvp,USYNC_THREAD,0);
  // the third argument is ignored
 
  if ( init_err == EINVAL || init_err == EFAULT) {
    printf("SetupReqIssued(): Error >>cond_init() error occured ...");
    printf("SetupReqIssued(): UNSUCCESFULL SETUP for REQ = %d !\n", id);
    RemoveStream(id);
    return(0);
  }

  init_err = cond_init(&Stream[s].cvp_retr,USYNC_THREAD,0);
  if ( init_err == EINVAL || init_err == EFAULT) {
    printf("SetupReqIssued(): Error>> cond_init() error occured ...");
    printf("SetupReqIssued(): UNSUCCESFULL SETUP for REQ = %d !\n", id);
    RemoveStream(id);
    return(0);
  }

  mutex_lock(&Stream[s].stateMutex);
  Stream[s].state =SETUP_STATE;
  Stream[s].PauseFlag=0;
  mutex_unlock(&Stream[s].stateMutex);
 
  mutex_lock(&Stream[s].nptMutex);
  Stream[s].NPT = 0;
  Stream[s].PauseSet = 0;
  mutex_unlock(&Stream[s].nptMutex);

  thr_err1 = thr_create(0, 0, StreamControlPlay, (void *)argid, THR_BOUND, &threadId[nThread]);

  mutex_lock(&nThreadMutex);
  nThread++;
  mutex_unlock(&nThreadMutex);

  thr_err2 = thr_create(0, 0, StreamControlPause,  (void *)argid, THR_BOUND, &threadId[nThread]);

  mutex_lock(&nThreadMutex);
  nThread++;
  mutex_unlock(&nThreadMutex);

  if (thr_err1 != 0 || thr_err2 != 0){
    printf("SetupReqIssued(): Error>> Unable to create threads for the new stream\n");
    RemoveStream(id);
    return(0);
  }

  /* Create the play and pause queues. */
  QueCreate(&Stream[s].PlayReqQue, QUE_NO_LIMIT);     // play req. queue
  QueCreate(&Stream[s].PauseReqQue, QUE_NO_LIMIT);    // pause req. queue
   
  Stream[s].sockfd = sockfd;
  Stream[s].sinlen = cli_len;

  Stream[s].port   = port;
  Stream[s].cli_addr = (struct sockaddr *)malloc(sizeof(struct sockaddr));
 
  if (Stream[s].cli_addr == NULL){
    printf("SetupReqIssued(): Error >>in malloc !\n");
    RemoveStream(id);
    return(0);
  }
 
  bzero(Stream[s].cli_addr, cli_len);
  bcopy(cli_addr, Stream[s].cli_addr, cli_len);

  conn_err = OpenConnection(s, (struct sockaddr_in *)cli_addr, port, cli_len);
  if (RTP_OK == conn_err){
     printf("SetupReqIssued(): succesful RTP Open Connection for the Req %d\n",id);
     printf("SetupReqIssued(): succesful SETUP for the REQ = %d\n",id);
     AckReq(s,"200", seq_num);
  } else {
    printf("SetupReqIssued(%d): NOT succesful  RTP Open Connection for the Req %d\n",id);
    AckReq(s, "400", seq_num);
    return(0);
  }
} /* SetupReqIssued */


// teardown_request_issued() : free the resources for the stream is stopped.
TeardownReqIssued(u_long id)
{
  int s;

  if (INVALID_STREAM == (s=GetStream(id)) ){
    printf("TeardownReqIssued(%d): Error >>NOT valid Stream Id !\n", id);
    return(-1);
  }

  mutex_lock(&Stream[s].stateMutex);
  Stream[s].state=TEARDOWN_STATE;
  mutex_unlock(&Stream[s].stateMutex);
}


ReleaseResources(int s)
{
  // change the state to INVALID_STATE --the stream is not ACTIVE ANY MORE
  mutex_lock(&Stream[s].stateMutex);
  Stream[s].state=INVALID_STATE;
  Stream[s].id =INVALID_STREAM;
  mutex_unlock(&Stream[s].stateMutex);

  // CLOSE/FREE the resources of that stream 
  // destroy the mutex's
  mutex_destroy(&Stream[s].nptMutex);
  mutex_destroy(&Stream[s].retrievMutex);
  mutex_destroy(&Stream[s].mp);
  mutex_destroy(&Stream[s].rtp_sendMutex);
  mutex_destroy(&Stream[s].rtp_scheduleMutex);
  mutex_destroy(&Stream[s].rtcp_tMutex);

  // Notice that i don't destroy the Stream[s].stateMutex
  // destroy the data queue 
}


/*
* Acknowledge request 'r' with status 'status'.
*/
int AckReq(int s, char *status_num, int seq_num)
{
  char *Packet;
  int send_err;
 
  Packet=(char *)calloc(512,sizeof(char));
 
  // You could include the reason phrase after the status_num as %s
  sprintf(Packet,"%s %s Hello!: ack from %s %d \nSession: %d", "RTSP/1.0", status_num, SERVER_HOST_ADDR, seq_num, Stream[s].id);
 
  if ( (send_err=sendto(Stream[s].sockfd, Packet, 512, 0, Stream[s].cli_addr, Stream[s].sinlen)) <1 ) 
  printf("dg_echo: sendto error\n");
}


/*
* Usage information. Exits program with status code 1.
*/
void usage(char *prog)
{
  fprintf(stderr, "Usage: %s [-p port] [-l logfile] [-d]\n", prog);
  exit(1);
} /* usage */


/*
* Main function for server.
*/
int main(int argc, char *argv[])
{
  int i;
  int thr_err;
  extern char *optarg;
  extern int optind;
  int c;
  char *logfile = "rtsp_log";
  thread_t t_udp, t_tcp;
  extern void ResponseInit(void);

  /* Parse command line options. */
  while ((c = getopt(argc, argv, "dl:p")) != EOF) {
    switch (c) {
      case 'd':
        debug = 1;
        break;
        break;
      case 'l':
        logfile = strdup(optarg);
        break;
      case 'p':
        config.port = atoi(optarg);
        break;
      case '?':
        usage(argv[0]);
        break;
    } 
  }

  /* Create log file. */
  if ((log = fopen(logfile, "a")) == NULL) {
    perror("cannot open log file");
    return 1;
  }

  ResponseInit();

  /* Initialize the message queue. */
  mutex_init(&nThreadMutex, USYNC_THREAD, 0);
  mutex_init(&rtp_createMutex, USYNC_THREAD, 0);
  ScheduleInit();

  mutex_lock(&nThreadMutex);
  nThread++;
  mutex_unlock(&nThreadMutex);

  if ((thr_err = thr_create(NULL, 0, ReceiveUDP,  NULL, 0, &t_udp)) != 0) {
    fprintf(stderr, "Cannot create UDP thread: %s\n", strerror(thr_err));
  }
  if ((thr_err = thr_create(NULL, 0, ReceiveTCP,  NULL, 0, &t_tcp)) != 0) {
    fprintf(stderr, "Cannot create TCP thread: %s\n", strerror(thr_err));
  }

  /* Wait for thread completion. */
  thr_join(t_udp, 0);
  thr_join(t_tcp, 0);

  /* Release the resources. */
  mutex_destroy(&nThreadMutex);
  mutex_destroy(&rtp_createMutex);

  fclose(log);
  return 0;
} /* main */
