/*
* stream.c
*
* For efficiency (and since we get to pick the session numbers), we
* maintain an expanding table of pointers to streams rather than a
* linear linked list. We dynamically allocate each stream entry since
* the base address may change when a new stream is created.
*/

#include "stream.h"
#include <stdio.h>
#include <malloc.h>     /* calloc() */
#include <assert.h>
#include <pthread.h>
 
#define STREAMS_INITIAL 20

static Stream **streams; /* base of array */
static int entries;      /* currently available entries */
static int used;         /* in use */

/*
* StreamCreate()
* Create a new stream. Return pointer to stream object.
*/
Stream *StreamCreate(void)
{
  int n;
  Stream *s;

  /* expand table */
  if (entries == 0) {
    streams = calloc(STREAMS_INITIAL, sizeof(Stream *));
    n = 0;
  }
  else if (used >= entries) {
    entries *= 1.5;
    streams = realloc(streams, entries * sizeof(Stream *));
    n = used+1;
  }  
  else {
    for (n = 0; n < entries; n++) {
      if (streams[n] == NULL) break;
    } 
    assert(n < entries);
  }
  used++;

  s = streams[n] = calloc(1, sizeof(Stream));

  s->session   = n;
  s->npt       = -1.;
  s->state     = S_INVALID;
  s->socket[0] = -1;
  s->socket[1] = -1;
  pthread_mutex_init(&s->lock, NULL);

  return s;
} /* StreamCreate */


/*
* StreamGet()
* Map a session identifier to a stream.
*/
Stream *StreamGet(int session)
{
  if ((session < 0) || (session >= entries)) return NULL;
  assert(session == streams[session]->session);
  return streams[session];
} /* StreamGet */


/*
* StreamFree()
* Release a stream and its resources.
*/
int StreamFree(int n)
{
  if ((n < 0) || (n >= entries)) return -1;
  pthread_mutex_destroy(&streams[n]->lock);
  free(streams[n]);
  return 0;
} /* StreamFree */
